# Configure the compiler options
# ------------------------------

include_guard(GLOBAL)

# Define an interface library that is solely responsible for passing on compiler
# flags. All libraries/executables are linked to this via our wrappers, so users
# should not have to worry about this at all.
add_library(compiler_flags INTERFACE)

if (NOT WIN32)
  # Turn on all the warnings, disable a few we are bad about
  target_compile_options(compiler_flags BEFORE
    INTERFACE
      -Wextra
      -Wall
      -Wno-unused-parameter
      -Wno-unused-function
      -Wno-ignored-attributes
      -Wno-deprecated-declarations
  )

  include(CheckCXXCompilerFlag)
  check_cxx_compiler_flag(-fbracket-depth=2000 HAS_BRACKET_DEPTH)
  if (HAS_BRACKET_DEPTH)
    target_compile_options(compiler_flags
      INTERFACE
        -fbracket-depth=2000
      )
  endif()

  # Check if this is an Apple Silicon host (aka ARM64). If so, we
  # are not using Intel SSE3 and need to add a define for code selection.
  if (APPLE AND CMAKE_HOST_SYSTEM_PROCESSOR STREQUAL "arm64")
    target_compile_definitions(compiler_flags
      INTERFACE
        _APPLE_ARM64
    )
  else()
    target_compile_options(compiler_flags
      INTERFACE
        -msse3
    )
  endif()

  option(COMPILER_PROFILE "Enable compiler profiling" OFF)
  if (COMPILER_PROFILE)
    target_compile_options(compiler_flags
      INTERFACE
        -ftime-trace
    )
  endif()

  # Set build type-specific flags
  set(CMAKE_CXX_FLAGS_RELEASE "-O3 -pedantic")
  set(CMAKE_CXX_FLAGS_DEBUG "-Og -g -fno-omit-frame-pointer")
else()
  target_compile_options(compiler_flags
    INTERFACE
      /bigobj
      /wd4100
      /EHsc
      /permissive-
  )

  target_compile_definitions(compiler_flags
    INTERFACE
      WIN32
      NOMINMAX
      _USE_MATH_DEFINES
      _CRT_SECURE_NOWARNINGS
      H5_BUILT_AS_DYNAMIC_LIB
  )
endif()

# If we are using ninja or make, we still want colors
if (NOT WIN32)
  if (CMAKE_GENERATOR STREQUAL "Ninja" OR
      CMAKE_GENERATOR STREQUAL "Unix Makefiles")
    target_compile_options(compiler_flags
      INTERFACE
        -fdiagnostics-color=always
    )
  endif()
endif()

cmake_policy(SET CMP0067 NEW)
set(CMAKE_CXX_STANDARD 17)
include(CheckCXXSourceCompiles)
check_cxx_source_compiles("
#include <filesystem>
#include <iostream>
int main() {
    std::filesystem::path p = std::filesystem::current_path();
    std::cout << p.string() << std::endl;
    return 0;
}" HAS_STL_FILESYSTEM)
option(USE_BOOST_FILESYSTEM "Use boost filesystem instead of STL <filesystem>." OFF)
if(NOT HAS_STL_FILESYSTEM)
  message(STATUS "STL filesystem not detected, using Boost filesystem.")
  set(USE_BOOST_FILESYSTEM ON)
endif()
if (USE_BOOST_FILESYSTEM)
  target_compile_definitions(compiler_flags INTERFACE USE_BOOST_FILESYSTEM)
endif()

option(ENABLE_IPO "Enable Interprocedural Optimization, aka Link Time Optimization (LTO)" OFF)
if(ENABLE_IPO)
  include(CheckIPOSupported)
  check_ipo_supported(RESULT result OUTPUT output)
  if (result)
    set(CMAKE_INTERPROCEDURAL_OPTIMIZATION TRUE)
  else()
    message(SEND_ERROR "IPO is not supported: ${output}")
  endif()
endif()

install(
  TARGETS "compiler_flags"
  EXPORT "${PROJECT_NAME}_Targets"
  ARCHIVE DESTINATION "${CMAKE_INSTALL_LIBDIR}"
  LIBRARY DESTINATION "${CMAKE_INSTALL_LIBDIR}"
  RUNTIME DESTINATION "${CMAKE_INSTALL_BINDIR}"
  COMPONENT "compiler_flags"
  )
