include_guard(GLOBAL)

include(GNUInstallDirs)
include(GenerateExportHeader)

# Install the include directory, containing all of the header files for the project.
install(
  DIRECTORY "${PROJECT_SOURCE_DIR}/include/"
  DESTINATION "include")

# Install the targets. Note: At this point in the CMake execution, no targets have actually been
# defined yet. The anima_add_library() macro below takes care of setting up the install info for
# each of the libraries, and by the time CMake has descended into all of the lib directories the 
# targets will all be populated.
install(
  EXPORT      "${PROJECT_NAME}_Targets"
  FILE        "${PROJECT_NAME}Targets.cmake"
  NAMESPACE   "${PROJECT_NAME}::"
  DESTINATION "${CMAKE_INSTALL_DATAROOTDIR}/${PROJECT_NAME}/cmake")

add_library("sdk_headers" INTERFACE)
add_library("${PROJECT_NAME}::sdk_headers" ALIAS "sdk_headers")
target_include_directories(
  "sdk_headers"
  INTERFACE
    $<BUILD_INTERFACE:${PROJECT_SOURCE_DIR}/include>
    $<BUILD_INTERFACE:${PROJECT_BINARY_DIR}/include>
    $<INSTALL_INTERFACE:${CMAKE_INSTALL_PREFIX}/include>)
install(
  TARGETS "sdk_headers"
  EXPORT "${PROJECT_NAME}_Targets"
  ARCHIVE DESTINATION "${CMAKE_INSTALL_LIBDIR}"
  LIBRARY DESTINATION "${CMAKE_INSTALL_LIBDIR}"
  RUNTIME DESTINATION "${CMAKE_INSTALL_BINDIR}"
  COMPONENT "sdk_headers")

function(d5_sdk_add_library LIB_NAME)
  # Set up argument parsing
  set(options SHARED STATIC MODULE OBJECT INTERFACE PUBLIC_VISIBILITY_ON_UNIX)
  set(oneValueArgs EXPORT_HEADER_DIR EXPORT_MACRO_NAME)
  set(multiValueArgs)
  cmake_parse_arguments(args "${options}" "${oneValueArgs}" "${multiValueArgs}" ${ARGN})

  # The library name is required
  if(NOT DEFINED LIB_NAME)
    message(FATAL_ERROR "d5_sdk_add_library(): LIB_NAME not defined")
  endif()


  # Handle all of the various library types, shared libraries are the important
  # ones, so keep track of those.
  set(SHARED_LIB FALSE)
  if(args_SHARED)
    add_library(${LIB_NAME} SHARED)
    set(SHARED_LIB TRUE)
  elseif (args_STATIC)
    add_library(${LIB_NAME} STATIC)
  elseif (args_MODULE)
    add_library(${LIB_NAME} MODULE)
  elseif (args_OBJECT)
    add_library(${LIB_NAME} OBJECT)
  elseif (args_INTERFACE)
    add_library(${LIB_NAME} INTERFACE)
  else()
    add_library(${LIB_NAME})
    if(BUILD_SHARED_LIBS)
      set(SHARED_LIB TRUE)
    endif()
  endif()

  # Make sure things work with shared libraries
  if(args_STATIC OR args_OBJECT)
    set_property(TARGET ${LIB_NAME} PROPERTY POSITION_INDEPENDENT_CODE ON)
  endif()

  # Apply all of our warning/optimization flags
  if(NOT args_INTERFACE)
    target_link_libraries(${LIB_NAME} PRIVATE compiler_flags)
  endif()

  # All libraries get a namespaced alias target
  add_library(${PROJECT_NAME}::${LIB_NAME} ALIAS ${LIB_NAME})

  # Handle shared-library specific boilerplate
  if(SHARED_LIB)
    set(LIB_ROOT ${CMAKE_BINARY_DIR}/${CMAKE_INSTALL_LIBDIR})
    set_target_properties(${LIB_NAME}
      PROPERTIES
        OUTPUT_NAME ${PROJECT_NAME}_${LIB_NAME}
        VERSION ${PROJECT_VERSION}
        SOVERSION ${PROJECT_VERSION_MAJOR}.${PROJECT_VERSION_MINOR}
        ARCHIVE_OUTPUT_DIRECTORY ${LIB_ROOT}
        LIBRARY_OUTPUT_DIRECTORY ${LIB_ROOT}
        RUNTIME_OUTPUT_DIRECTORY ${CMAKE_BINARY_DIR}/${CMAKE_INSTALL_BINDIR})
    target_link_libraries(${LIB_NAME} PRIVATE ${CMAKE_DL_LIBS})

    # By default, we have use hidden visibility (only public if the user
    # adds this flag). Set the flags and make an export header.
    if(NOT args_PUBLIC_VISIBILITY_ON_UNIX)
      set_target_properties(
        ${LIB_NAME}
        PROPERTIES
          CXX_VISIBILITY_PRESET hidden
          VISIBILITY_INLINES_HIDDEN 1)

      # Make the parameters for the export header
      if(NOT args_EXPORT_HEADER_DIR)
        set(args_EXPORT_HEADER_DIR "${LIB_NAME}")
      endif()
      if(NOT args_EXPORT_MACRO_NAME)
        string(TOUPPER "D5_${LIB_NAME}_EXPORT" args_EXPORT_MACRO_NAME)
      endif()

      # Generate the export header and add it to the include path
      generate_export_header(
        ${LIB_NAME}
        EXPORT_FILE_NAME "${PROJECT_BINARY_DIR}/include/dirsig5/${args_EXPORT_HEADER_DIR}/Export.h"
        EXPORT_MACRO_NAME "${args_EXPORT_MACRO_NAME}")
      target_include_directories(${LIB_NAME} PUBLIC $<BUILD_INTERFACE:${PROJECT_BINARY_DIR}/include>)
      install(
        FILES "${PROJECT_BINARY_DIR}/include/dirsig5/${args_EXPORT_HEADER_DIR}/Export.h"
        DESTINATION "include/dirsig5/${args_EXPORT_HEADER_DIR}")
    endif()

    # Set the RPATH to see the other libraries
    if(APPLE)
      set_target_properties(${LIB_NAME} PROPERTIES INSTALL_RPATH "@loader_path/../${CMAKE_INSTALL_LIBDIR}")
    else()
      set_target_properties(${LIB_NAME} PROPERTIES INSTALL_RPATH "$ORIGIN/../${CMAKE_INSTALL_LIBDIR}")
    endif()

  endif()
  install(
    TARGETS "${LIB_NAME}"
    EXPORT "${PROJECT_NAME}_Targets"
    ARCHIVE DESTINATION "${CMAKE_INSTALL_LIBDIR}"
    LIBRARY DESTINATION "${CMAKE_INSTALL_LIBDIR}"
    RUNTIME DESTINATION "${CMAKE_INSTALL_BINDIR}"
    COMPONENT "${LIB_NAME}")
endfunction()
