/** Interface file for ConvergenceCriteria
 *
 *  \file ConvergenceCriteria.h
 */

#pragma once

namespace d5::apis {

/// Constants for the source of the convergence parameters
enum ConvergenceSource {
  DEFAULT,       ///< The default (constructor)
  USER_DEFINED,  ///< User defined (e.g., via the command-line)
  PLUGIN_DEFINED ///< Plugin defined
};

/** Defines the conditions under which we consider a solution for a
 *  given problem to be complete.
 *  \ingroup apis
 */
struct ConvergenceCriteria {
  /// Constructor, use default values
  ConvergenceCriteria() = default;

  /** Constructor
   *
   *  \param minSamples  See member description
   *  \param maxSamples  See member description
   *  \param thresh      See member description
   *  \param maxNodes    See member description
   *  \param src         See member description
   */
  ConvergenceCriteria(unsigned minSamples, unsigned maxSamples, double thresh,
                      int maxNodes, ConvergenceSource src = USER_DEFINED)
      : minimumSamples(minSamples), maximumSamples(maxSamples),
        threshold(thresh), maximumNodes(maxNodes), source(src) {}

  /// Minimum samples required for convergence
  unsigned minimumSamples = 20;

  /// Maximum files that can be used to reach the threshold
  unsigned maximumSamples = 100;

  /** Difference between the current mean and prior mean under which we stop
   * as measured for the reference wavelength
   */
  double threshold = 1e-6;

  /// Maximum number of nodes within a path
  int maximumNodes = 4;

  /// The source of the parameters
  ConvergenceSource source = DEFAULT;

  /// Optional decimal fraction of samples that should be tracked as outliers.
  /// The number will be split between the top and bottom samples.
  double outlierCullFraction = 0;
};

} // namespace d5::apis
