/** Interface file for the geographic API
 *
 *  \file GeoInterface.h
 */

#pragma once

#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/IntID.h"
#include "dirsig5/geolib/Location.h"

namespace d5::apis {

/** Provides an API that can be used for geographic queries
 *
 *  \ingroup apis
 */
class GeoInterface {
public:
  /// Destructor
  virtual ~GeoInterface() = default;

  /** Request the position on the WGS84 datum (ellipsoid) that a
   *  ray in a local coordinate system would hit after passing
   *  through the atmosphere associated with the scene.
   *
   *  \param[in] sceneIndex    Index into user scenes, defines the coordinate
   * system for rayOrigin and rayDirection
   *  \param[in] rayOrigin     Origin of the ray in the scene's coordinate
   *                           system
   *  \param[in] rayDirection  Direction of the ray in the scene's coordinate
   *                           system
   *  \param[out] hitLocation  Output parameter for the hit location
   *
   *  \return Whether a valid hit location could be computed
   */
  virtual bool getDatumIntersection(size_t sceneIndex, const Point &rayOrigin,
                                    const Vector &rayDirection,
                                    geolib::Location &hitLocation) = 0;

  /** Request the bounding box of a scene in its own coordinate system.
   *
   *  \param[in] sceneIndex  Index into user scenes
   *  \param[out] boxMin  The minimum point of the scene's bounds
   *  \param[out] boxMax  The maximum point of the scene's bounds
   *
   *  \return Whether the bounds could be computed
   */
  virtual bool getSceneBounds(size_t sceneIndex, Pointf &boxMin,
                              Pointf &boxMax) = 0;

  /** Request the origin of a scene in global coordinates.
   *
   *  \param[in] userSceneID  User-space ID of the scene
   *
   *  \return The origin of the scene
   */
  virtual geolib::Location getSceneOrigin(size_t sceneIndex) = 0;

  /** Request the local ENU coordinates of the given ECEF coordinate.
   *
   * \param[in] sceneIndex   User-space index into scenes
   * \param[in] ecef         ECEF coordinate (meters)
   * \param[in] asVector     treat as a vector
   */
  virtual Point convertECEFtoENU(size_t sceneIndex, Point ecef,
                                 bool asVector = false) = 0;

  /** Request the ECEF coordinates of the given local ENU coordinate.
   *
   * \param[in] sceneIndex   User-space index into scenes
   * \param[in] enu          ENU coordinate (meters)
   * \param[in] asVector     treat as a vector
   */
  virtual Point convertENUtoECEF(size_t sceneIndex, Point enu,
                                 bool asVector = false) = 0;
};

} // namespace d5::apis
