/** Interface file for the material APIs
 *
 *  \file MaterialInfo.h
 */

#pragma once

#include <string>
#include <utility>
#include <vector>

#include "dirsig5/base/IntID.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Pcg.h"
#include "dirsig5/base/SpectralVector.h"

namespace d5::apis {

/// \addtogroup apis
/// \{

/// Constants for the types of materials that DIRSIG supports
enum MaterialType {
  LAMBERTIAN,  ///< Lambertian (diffuse)
  DELTA,       ///< Delta (specular)
  WARD,        ///< Ward (diffuse +specular +anisotropic specular width)
  SQT,         ///< Spherical Quad Tree (data-driven)
  SMDL         ///< Spectral MDL
};

/// Information about a material
struct MaterialInfo {
  /** Constructor
   *
   *  \param t     The type of the material
   *  \param wave  List of wavelengths in microns
   */
  MaterialInfo(MaterialType t, std::vector<float> wave = {})
      : type(t), wavelengths(std::move(wave)) {}

  /// Destructor
  virtual ~MaterialInfo() = default;

  /// Type of the material
  MaterialType type;

  /// Wavelengths in microns
  std::vector<float> wavelengths;
};

/// Information about a Lambertian material
struct MaterialInfoLambertian : MaterialInfo {
  /// Constructor
  MaterialInfoLambertian() : MaterialInfo(LAMBERTIAN) {}

  /** Constructor
   *
   *  \param wave  Wavelengths in microns
   *  \param refl  Reflectance for each wavelength (can be empty for none)
   *  \param tran  Transmission for each wavelength (can be empty for none)
   */
  MaterialInfoLambertian(
      const std::vector<float>& wave,
      std::vector<float> refl,
      std::vector<float> tran = {})
      : MaterialInfo(LAMBERTIAN, wave),
        R(std::move(refl)),
        T(std::move(tran)) {}

  /// Reflectance. If non-empty, must be the same size as wavelengths.
  std::vector<float> R;

  /// Transmittance. If non-empty, must be the same size as wavelengths.
  std::vector<float> T;
};

/** Information for a delta (specular) material.
 *
 *  The Delta function is assumed to be aligned with the specular
 *  reflectance vector. The provided spectral vector is the total reflectance.
 */
struct MaterialInfoDelta : MaterialInfo {
  /// Constructor
  MaterialInfoDelta() : MaterialInfo(DELTA) {}

  /** Constructor
   *
   *  \param wave  Wavelength samples in microns
   *  \param refl  Total reflectance for each wavelength
   */
  MaterialInfoDelta(const std::vector<float>& wave, std::vector<float> refl)
      : MaterialInfo(DELTA, wave), R(std::move(refl)) {}

  /// Total reflectance, must be the same size as wavelengths.
  std::vector<float> R;
};

/** The Ward inputs are the magnitudes of the diffuse and specular lobe
 *  componenets as well as the standard deviation (sigma) values
 *  used to shape the lobe -- each input is defined spectrally.
 */
struct MaterialInfoWard : MaterialInfo {
  MaterialInfoWard() : MaterialInfo(WARD) {}
  MaterialInfoWard(
      const std::vector<float>& wave,
      std::vector<float> spec,
      std::vector<float> diff,
      std::vector<float> sigu,
      std::vector<float> sigv)
      : MaterialInfo(WARD, wave),
        specular(std::move(spec)),
        diffuse(std::move(diff)),
        sigma_u(std::move(sigu)),
        sigma_v(std::move(sigv)) {}

  std::vector<float> specular;
  std::vector<float> diffuse;
  std::vector<float> sigma_u;
  std::vector<float> sigma_v;
};

/** All information about the material is obtained from the
 *  SQT file(s) (no wavelengths are needed).
 *
 *  The provided string RSqtFilename is the filename of
 *  the SQT file to use as reflectance. The provided string TSqtFilename is
 *  the filename of the SQT file to use as transmittance. Only 1 of
 *  the filenames may be empty, in order to specify a purely reflective
 *  (RSqtFilename non-empty, TSqtFilename empty) or purely transmissive
 *  (RSqtFilename empty, TSqtFilename non-empty) Lambertian material. Both
 *  may be non-empty to specify a partially reflective, partially
 *  transmissive SQT material.
 *
 *  If all relevant data is encoded in the SQT file(s), then _no_
 *  wavelengths or spectral vectors need to be provided. However, the
 *  total spectral reflectance and/or transmittance of a wavelength
 *  independent SQT may be overridden by providing the total
 *  spectral reflectance R and/or transmittance T.
 */
struct MaterialInfoSQT : MaterialInfo {
  MaterialInfoSQT() : MaterialInfo(SQT) {}
  MaterialInfoSQT(std::string refl, std::string tran = {})
      : MaterialInfo(SQT),
        RSqtFilename(std::move(refl)),
        TSqtFilename(std::move(tran)) {}

  /// Reflectance SQT filename.
  std::string RSqtFilename = {};

  /// Transmittance SQT filename.
  std::string TSqtFilename = {};

  /// Reflectance override.
  std::vector<float> R = {};

  /// Transmittance override.
  std::vector<float> T = {};
};

/** Information for a Spectral MDL (SMDL) material.
 */
struct MaterialInfoSmdl : MaterialInfo {
  /// Constructor
  MaterialInfoSmdl() : MaterialInfo(SMDL)
  {
  }

  /** Constructor
   *
   *  \param wvl  The wavelength samples in microns
   */
  MaterialInfoSmdl(const std::vector<float>& wvl)
      : MaterialInfo(SMDL, wvl)
  {
  }

  /// The wavelengths available in the SMDL module
  std::vector<float> wavelengths;

  /// A callback to get the total reflectance
  typedef std::function<void(const d5::Vectorf&,
      const d5::IntIDStack*,
      d5::SpectralVector&)> TotalReflectanceCallback;

  /// A callback to compute the scattering value
  typedef std::function<void(const d5::Vectorf&,
      const d5::Vectorf&,
      const d5::IntIDStack*,
      d5::SpectralVector&,
      float)> ScatteringValueCallback;

  /// A callback to sample (reflect/scatter an incident vector)
  typedef std::function<void(d5::RandEngine&,
      const d5::Vectorf&,
      const d5::IntIDStack*,
      d5::Vectorf&,
      d5::SpectralVector&)> SampleCallback;

  TotalReflectanceCallback totalReflectanceCallback;
  ScatteringValueCallback scatteringValueCallback;
  SampleCallback sampleCallback;
};

/// \}
} // namespace d5::apis
