/** Interface file for the NestedMotion API
 *
 *  \file NestedMotion.h
 */

#pragma once

#include "dirsig5/apis/RayTracingInterface.h"
#include "dirsig5/base/Geometry.h"

namespace d5::apis {

/** Wraps the a chain of nested motion model callbacks for a dynamic instance
 *  \ingroup apis
 */
struct NestedMotion {
  /** Adds a callback to the chain
   *
   *  \param callback  The dynamic instance callback to add
   */
  void addCallback(const DynamicCallback &callback) {
    callbacks.push_back(callback);
  }

  /// Gets a callback that chains all added callbacks together
  [[nodiscard]] DynamicCallback getAggregateCallback() const {
    return [this](float stateTime) { return transform(stateTime); };
  }

  /** Gets the aggregate dual-quaternion transformation
   *
   *  \param sceneID    ID of the scene
   *  \param baseID     ID of the base geometry
   *  \param stateTime  Relative time in seconds from the current temporal
   *                    state
   */
  [[nodiscard]] DualQuaternion transform(float stateTime) const {
    DualQuaternion chain;
    for (auto call = callbacks.rbegin(); call != callbacks.rend(); ++call)
      chain = (*call)(stateTime)*chain;
    return chain;
  }

  /// The nested callback chain
  std::vector<DynamicCallback> callbacks;
};
} // namespace d5::apis
