/** Interface file for DIRSIG's object API
 *
 *  \file ObjectInterface.h
 */

#pragma once

#include <functional>
#include <unordered_set>
#include <utility>

#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/DualQuaternion.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/IntID.h"
#include "dirsig5/base/Truth.h"

namespace d5::apis {

/// \addtogroup apis
/// \{

using SceneCallback = std::function<void()>;

using ObjectCallback = std::function<void(IntID)>;

using DynamicCallback = std::function<DualQuaternion(float)>;

using AlphaCallback = std::function<float(float, float, int)>;

/** Parameters are reference date/time, capture start, time in capture. */
using EnabledStateCallback =
    std::function<float(const DateTime &, float, float)>;

using SceneTransformCallback = std::function<void(Point &)>;

struct Node;

using NodeCallback = std::function<void(size_t, const Node *)>;

/** Provides a high-level API for defining an object in a scene.
 *
 *  An object within a scene is rarely a unique location in memory describing
 *  the geometry as it is within the scene. Instead, an object is usually
 *  a geometry definition that has been instanced into the scene through
 *  an affine transform. The object interface is an abstracted wrapper around
 *  an instance that allows for querying a specific object at runtime,
 *  regardless of how it is implemented internally.
 *
 *  The object interface is not a plugin interface in the usual sense since
 *  it is implemented internally by concrete objects, but it is used by other
 *  plugin interfaces.
 */
class ObjectInterface {
public:
  /** Constructor
   *
   *  \param isStatic_  Whether this is a object is a static instance
   *  \param sceneID_   ID of the scene containing the object (internal)
   *  \param baseID_    ID of the base geometry for the object
   *  \param instID_    ID of the instance this object is representing
   */
  ObjectInterface(bool isStatic_, IntID sceneID_, IntID baseID_, IntID instID_)
      : isStatic(isStatic_), sceneID(sceneID_), baseID(baseID_),
        instID(instID_) {}

  /// Destructor
  virtual ~ObjectInterface() = default;

  /** Is guaranteed to be static or not?
   *
   *  \note
   *  This should be modified _only_ by derived class implementations. This
   *  used to be implemented as a virtual function, but it is more performant
   *  to make it available as a flag to avoid the table dispatch.
   */
  bool isStatic = true;

  /** Get the internal scene, base geometry and instance IDs.
   *  Note that external requests use the user scene index use plugin resources.
   *  The user scene index can be specifically requested below
   */
  template <typename Int1, typename Int2, typename Int3>
  void getInternalIdentifiers(Int1 &scene, Int2 &base, Int3 &inst) const {
    scene = sceneID;
    base = baseID;
    inst = instID;
  }

  /// Set the actual user scene index for this object. This should only be used
  /// internally.
  void setUserSceneIndex(size_t userSceneIndex_) const {
    DIRSIG_THROW_IF(this->userSceneIndex != size_t(-1),
                    "Attempting to reset the user scene index is not allowed!");
    this->userSceneIndex = userSceneIndex_;
  }

  /** Add a scene transformation callback. This is used internally
   *  to transform points from the scene to the global coordinate space.
   */
  void setSceneTransformation(SceneTransformCallback transform,
                              SceneTransformCallback inverseTransform) const {
    sceneTransform = std::move(transform);
    sceneInverseTransform = std::move(inverseTransform);
  }

  /** Apply transform to a point.
   *
   *  Transforms a given point based on the current time in seconds
   *  (relative to the temporal state). The input point should be given
   *  in the object local coordinate system (i.e. [0,0,0] will return the
   *  current location of the object). The returned position will be
   *  in the ENU coordinate frame of the primary user scene.
   */
  Point transformPoint(Point point, float time) const {
    point = doTransformPoint(point, time);
    if (sceneTransform)
      sceneTransform(point);
    return point;
  }

  /** Apply transform to a vector.
   *
   *  Transforms a given directional vector based on the current time in
   *  seconds (relative to the temporal state). Note that this can be used
   *  to find the pointing direction if the original pointing convention
   *  is known (e.g. the pointing vector of a car originally facing +Y
   *  can be queried using [0,1,0])
   */
  virtual Vector transformVector(const Vector &vector, float time) const = 0;

  /// Apply inverse transform to a point.
  Point inverseTransformPoint(Point point, float time) const {
    point = doInverseTransformPoint(point, time);
    if (sceneInverseTransform)
      sceneInverseTransform(point);
    return point;
  }

  /// Apply inverse transform to a vector.
  virtual Vector inverseTransformVector(const Vector &vector,
                                        float time) const = 0;

  /** Disable the object. After disabling, all rays will pass through
   *  the object, unchanged. Disabling an object that is already
   *  disabled will have no effect.
   */
  virtual void disableObject() const = 0;

  /** Enable the object. The object starts out enabled. This function
   *  may be used to re-enable an object that has been disabled. Enabling
   *  an object that is already enabled will have no effect.
   */
  virtual void enableObject() const = 0;

  /** Check if an object is currently enabled at the current time in
   *  seconds (relative to the temporal state, default is zero).
   */
  virtual bool isObjectEnabled(float t = 0.f) const = 0;

  /** Get tags associated with this object. Tags for this object are
   *  added to provided (unsorted) set. The set is not cleared first!
   */
  virtual void getTagList(std::unordered_set<std::string> &tags) const = 0;

  /** Get the center of the object at a given time.
   *
   *  Transforms the center of the object's bounding box for the given
   *  time. This is often preferable to transforming the local object origin
   *  in case it is offset from the expected location (e.g. the origin
   *  of a tree is often at its base, not midway up the trunk). Time is
   *  relative to the current temporal state. We also return the radius
   *  for the object at the given time.
   */
  Point getCenter(float time, float &radius) const {
    Point center = doGetCenter(time, radius);
    if (sceneTransform)
      sceneTransform(center);
    return center;
  }

  /// Request that abundance of this instance is tracked by the wrapper.
  void requestAbundance(InstanceAbundanceCallback callback) const {
    abundanceCallback = std::move(callback);
    trackAbundance = true;
  }

  /// Check if the abundance of this instance is being tracked.
  bool hasAbundanceCallback() const { return trackAbundance; }

  /// Get the callback to the function to updated abundance info.
  InstanceAbundanceCallback *getAbundanceCallback() const {
    DIRSIG_THROW_IF(not trackAbundance,
                    "Attempted to grab a nonexistent abundance callback!");
    return &abundanceCallback;
  }

  /** Add a state controller callback to the instance.
   *
   *  \note
   *  It is ok to add multiple controllers, the result
   *  is the multiplication of all states at a given time.
   */
  void addStateController(EnabledStateCallback callback) const {
    stateControllers.push_back(std::move(callback));
  }

  /** Set the elapsed time for the current state (i.e. the beginning
   *  of the current time window). This is relative to the reference
   *  date/time so we can adjust the local capture time to be relative
   *  to the reference.
   */
  void setStateElapsedTime(const DateTime &refDateTime, float elapsed) const {
    referenceDateTime = refDateTime;
    elapsedTime = elapsed;
  }

  /** Check if there is a variable state.
   */
  bool hasState(void) const { return (stateControllers.size() > 0); }

  /** Get the state based on state controllers (if any -- default is one).
   *  Note that the state can be any value in [0,1] where any non-zero
   *  value is considered to be "on". Values between 0 and 1 may be used
   *  to describe a partial state, e.g., scaling source intensity.
   *
   *  The provided time is relative to the current temporal state.
   *  Any state value outside the acceptable range will be truncated.
   */
  float getState(float stateTime) const {
    float state = 1;
    for (const auto &stateController : stateControllers)
      state *= stateController(referenceDateTime, elapsedTime, stateTime);
    state = std::max(0.0f, state);
    state = std::min(1.0f, state);
    return state;
  }

  /** Add a part to this object. This is used internally to manipulate objects
   *  that have been broken into parts.
   */
  virtual void addPart(const ObjectInterface *part) const = 0;

  /// Set the keypoints associated with this object.
  void setKeypoints(const Keypoints &points) const { keypoints = points; }

  /// Get the const reference to the keypoints.
  const Keypoints &getKeypoints() const { return keypoints; }

  /// Set the history length needed for thermal scarring.
  void setHistoryLength(float historyLen) const { historyLength = historyLen; }

  /// Get the history length needed for thermal scarring.
  float getHistoryLength() const { return historyLength; }

  size_t getUserSceneIndex() const {
    DIRSIG_THROW_IF(this->userSceneIndex == size_t(-1),
                    "Internal error! User scene index was never set.");
    return this->userSceneIndex;
  }

  void setNodeCallback(NodeCallback callback) const {
    nodeCallback = std::move(callback);
  }

  const NodeCallback &getNodeCallback() const { return nodeCallback; }

protected:
  /// Transforms a local point into world space.
  virtual Point doTransformPoint(const Point &point, float time) const = 0;

  /// Transforms a world point into local space.
  virtual Point doInverseTransformPoint(const Point &point,
                                        float time) const = 0;

  /// Gets the center and radius of the object for a given time.
  virtual Point doGetCenter(float time, float &radius) const = 0;

  /// The ID for the scene this object is in.
  IntID sceneID = INVALID_ID;

  /// Corresponding scene index for the object
  mutable size_t userSceneIndex = size_t(-1);

  /// The ID for the base geometry of this object.
  IntID baseID = INVALID_ID;

  /// The ID for the instance of this base geometry.
  IntID instID = INVALID_ID;

  /// The flag that indicates if abundance is being tracked.
  mutable bool trackAbundance = false;

  /// The (optional) abundance callback function.
  mutable InstanceAbundanceCallback abundanceCallback = {};

  /// The (optional) state controller callback functions.
  mutable std::vector<EnabledStateCallback> stateControllers = {};

  /// The reference date/time.
  mutable DateTime referenceDateTime = {};

  /// The time elapsed since the reference date/time (start of time window).
  mutable float elapsedTime = 0;

  /// The callback to get the object-to-scene transform.
  mutable SceneTransformCallback sceneTransform = {};

  /// The callback to get the scene-to-object transform.
  mutable SceneTransformCallback sceneInverseTransform = {};

  /// The keypoints (labeled 3D locations).
  mutable Keypoints keypoints = {};

  /// The history length needed for thermal scarring.
  mutable float historyLength = 0;

  /// The user node callback
  mutable NodeCallback nodeCallback = {};
};

/// \}
} // namespace d5::apis
