/** Interface file for the top-level DIRSIG plugin API
 *
 *  \file Plugin.h
 */

#pragma once

#include <iomanip>
#include <utility>
#include <vector>

#include "dirsig5/apis/GeoInterface.h"
#include "dirsig5/apis/GetEphemerisInterface.h"
#include "dirsig5/apis/GetWeatherInterface.h"
#include "dirsig5/apis/IndexInterface.h"
#include "dirsig5/apis/MaterialInfo.h"
#include "dirsig5/apis/ObjectInterface.h"
#include "dirsig5/apis/PluginInterface.h"
#include "dirsig5/apis/SceneIntersectionInterface.h"
#include "dirsig5/apis/SceneTagInterface.h"
#include "dirsig5/base/Json.h"
#include "dirsig5/base/State.h"
#include "dirsig5/base/Version.h"

/// \cond HIDDEN_SYMBOLS
namespace spdlog {
class logger;
}
/// \endcond

namespace d5::apis {

/// \addtogroup apis
/// \{

/// \cond HIDDEN_SYMBOLS
// Forward declarations
class Plugin;
class RunInfo;
/// \endcond

/// Simple struct containing common setup information (command line options).
struct ArgumentsIn {
  int argc = 0;
  char** argv = nullptr;
};

/// Holds a map of tag name to object instances
struct TaggedInstanceRequestsIn {
  /// Tag -> list of object instances
  std::unordered_map<std::string, std::vector<const ObjectInterface*>>
    taggedInstances;
};

/** Common objects for tagged object instance queries.
 *
 *  Plugin APIs using this mechanism should derive their setup output
 *  and thier init input frome these structs. Requests are made
 *  during the setup phase and returned during init. Note that
 *  no error is raised if a tag is not found (there just won't be
 *  an entry in the returned map).
 */
struct TaggedInstanceRequestsOut {
  std::vector<std::string> instanceTags;
};

/// optional referenced information about facet vertexes corresponding to a
/// primId
struct HitVertexInfo {
  bool valid = false;
  std::array<Pointf, 3> vs;    ///< optional vertexes
  std::array<Vectorf, 3> vns;  ///< optional vertex normals
};

/// Information about a ray intersection
struct IntersectInfo {
  const Pointf& rayOrigin;      ///< origin of the ray
  const Vectorf& rayDirection;  ///< direction of the ray
  const float& time;            ///< ray origin time[s] capture relative
  const float& tnear;           ///< minimum hit distance [m]
  const float& tfar;            ///< maximum hit distance [m]
  const bool& highPriority;     ///< high priority (direct lighting)?
  float& hitDistance;           ///< distance to the hit if applicable
  Vectorf& normal;              ///< normal vector at the hit
  int& matIndex; /**< optional index that triggers usage of a material plugin
                    (index provided) */
  float& u;      ///< barycentric coordinate u of hit
  float& v;      ///< barycentric coordinate v of hit
  bool& hitBackside;    ///< flag for a backside hit
  size_t& primId;       ///< primitive index (if available)
  double& temperature;  ///< temperature in Kelvin (ignored if <0 or unchanged)
  float* cd =
    nullptr;  /// optional concentration distance if a non-null ptr is set
  HitVertexInfo* vertexInfo = nullptr;  /// optional vertex info
};

using IntersectCallback = std::function<bool(IntersectInfo&)>;

/// This struct provides info about the simulation.
/// 1) It provides access to other plugins that are available.
/// 2) It provides access to geo location tools to go from
///    a scene's local coordinate system to ECEF (and others)
///
/// The available functions are described in the corresponding
/// common library headers.
struct Resources {
  /// Logger to use for log output
  spdlog::logger* logger = nullptr;

  /// Run info object that can be populated with diagnostic information
  RunInfo* runInfo = nullptr;

  /// Plugin API that can be used to communicate with other plugins
  std::shared_ptr<PluginInterface> pluginInterface;

  /// Geographic API that can be used for coordinate transformations
  std::shared_ptr<GeoInterface> geoInterface;

  /// Tag API to query tagged objects
  std::shared_ptr<SceneTagInterface> sceneTagInterface;

  /// Ephemeris API that can be used to get ephemeris information
  std::shared_ptr<GetEphemerisInterface> ephemerisInterface;

  /// Intersection API that can be used intersect a specific scene
  std::shared_ptr<SceneIntersectionInterface> sceneIntersectionInterface;

  /// Weather API that can be used to get weather information
  std::shared_ptr<GetWeatherInterface> weatherInterface;

  /// Simple API for accessing the current scheduler indexes
  std::shared_ptr<IndexInterface> indexInterface;
};

/** Definition of input/output variables for the setup interface. The
 *  output object is derived from the tagged request struct to allow
 *  for requesting access to instances at runtime.
 */
namespace Setup {

/// Input structure for a plugin setup call
struct In : ArgumentsIn {
  /** Instance definition. This void pointer provides access to
   *  the input definition for the particular instance of the
   *  plugin being setup (e.g. access to the input XML file).
   *  It is provided in as generic a form as possible to facilitate
   *  many possible input data formats.
   */
  char* definition = nullptr;

  /// Filename corresponding to the input file (if applicable)
  std::string filename;

  /// Information about resources that are available to the plugin.
  Resources resources;
};

/// Output structure for a plugin setup call
struct Out : public TaggedInstanceRequestsOut {
  /// Generic JSON field that a plugin can populate to publish
  /// useful information to the run log.
  Json pluginData;
};
}  // namespace Setup

/// Definition of input/output variables for the state interface
namespace UpdateState {

struct In {
  In(
    const DateTime& refTime,
    double elapsed,
    double dur,
    const SpectralState::WavelengthSet& wave,
    size_t refWaveIndex,
    size_t senIndex,
    size_t senSpecIndex)
      : referenceTime(refTime),
        elapsedTime(elapsed),
        duration(dur),
        wavelengths(wave),
        referenceWavelengthIndex(refWaveIndex),
        sensorIndex(senIndex),
        sensorSpectralStateIndex(senSpecIndex) {}

  /// Convenience constructor to initialize everything with
  /// a system state object.
  In(const State* state)
      : referenceTime(*state->referenceDateTime),
        elapsedTime(state->elapsedTime),
        duration(state->duration),
        wavelengths(state->spectralStates->at(state->spectralStateIndex)
                      .wavelengthSamples),
        referenceWavelengthIndex(
          state->spectralStates->at(state->spectralStateIndex)
            .referenceWavelengthIndex),
        sensorIndex(state->sensorIndex),
        sensorSpectralStateIndex(state->sensorSpectralStateIndex),
        internalState(state) {}

  /// Reference date/time. All time windows and query times
  /// are given relative to this point.
  ///
  const DateTime& referenceTime;

  /// Time window in [s] for the current temporal state. Queries
  /// should be made with the window, but are not guaranteed to
  /// be so under some circumstances.
  ///
  ///@{
  double elapsedTime = 0, duration = 0;
  ///@}

  /// Wavelength samples for the current spectral state. Passed
  /// as a (const) reference to the internal spectral state vector.
  /// The reference is not guaranteed to be valid after the state
  /// update has returned.
  ///
  const SpectralState::WavelengthSet& wavelengths;

  /// Index of the reference wavelength, i.e. the wavelength that
  /// should be used to drive any behavior that is not descrbed
  /// spectrally.
  ///
  size_t referenceWavelengthIndex = 0;

  /// Index of the sensor that requested this state
  size_t sensorIndex = 0;

  /// Index of the (sensor local) spectral state
  size_t sensorSpectralStateIndex = 0;

  /// Original state if used to set
  const State* internalState = 0;
};

}  // namespace UpdateState

/// This is the top level plugin object from which the APIs are derived.
/// It provides basic functionality to support versioning and interfacing
/// with the plugin manager.
///
class Plugin {
public:
  /// Non-binary status for plugins to return. Values shown explicitly for
  /// clarity.
  enum Status {
    FAIL = 0,        // unrecoverable error (expect abort)
    SUCCESS = 1,     // results are valid
    USE_DEFAULT = 2  // use the default result
  };

  using Version = d5::Version;

  Plugin() = default;

  virtual ~Plugin() = default;

  /// Virtual setup function. Does nothing by default. The setup
  /// is defined generically so plugins can be loaded in without
  /// knowing their identities.
  ///
  Status setup(const Setup::In& in, Setup::Out& out) {
    if (in.definition != nullptr) {
      try {
        inputJson_ = Json::parse(in.definition);
        inputWasJson_ = true;
      } catch (...) {
        jsonParseException_ = std::current_exception();
      }
    }

    resources_ = in.resources;
    if (resources_.logger == nullptr) {
      throw std::runtime_error(
        "No logger was provided to this plugin, this is an internal error");
    }

    auto status = doSetup(in, out);
    setupData_ = out.pluginData;
    return status;
  }

  /// Returns the input JSON (blank if input was not JSON)
  [[nodiscard]] const Json& getInputAsJson() const { return inputJson_; }

  /// Parses the input JSON into a struct of the given type. Throws on
  /// any error.
  template <typename T>
  void parseInputJson(T& value) {
    if (!inputWasJson_) {
      if (jsonParseException_) {
        std::rethrow_exception(jsonParseException_);
      }
      throw std::runtime_error("Input was not valid JSON");
    }
    inputJson_.get_to(value);
  }

  /// Returns any data published by the plugin upon setup()
  [[nodiscard]] const Json& getSetupData() const { return setupData_; }

  /// Get the SDK version used to build the plugin
  [[nodiscard]] const Version& getSdkVersion() const { return sdkVersion_; }

  /// Get the version of the plugin (according to the plugin implementer)
  [[nodiscard]] virtual const Version& getVersion() const {
    static Version version(1, 0);
    return version;
  }

  /// Gets the name of the plugin
  [[nodiscard]] virtual const std::string& getName() const {
    static std::string name = "unknown";
    return name;
  }

  /// Gets the names of the implemented plugin type
  [[nodiscard]] virtual std::vector<std::string> getPluginTypeNames() const {
    return {};
  }

  /// Update state function. This is used internally to trigger
  /// a (virtual) updateState call (so we can track used states)
  Status updatePluginState(const State* state) {
    if (state == currentState) return SUCCESS;
    return updateState(UpdateState::In(currentState = state));
  }

  /// Virtual update state function. Does nothing by default.
  /// This function is used to notify the plugin that the current
  /// state (tempora or spectral) has changed. No output struct is passed
  /// back since this is intended to be a one-way notification.
  ///
  virtual Status updateState(const UpdateState::In& /*in*/) { return SUCCESS; }

  /// Teardown.
  ///
  /// Notification that the simulation is done using this
  /// plugin, even if it may not choose to destroy the plugin at
  /// this time. Memory can be cleaned up with the expectation that
  /// no more calls are expected (this is guaranteed)
  ///
  /// Implementation of this interface is optional and does not
  /// provide or request any information.
  ///
  virtual Status teardown() { return SUCCESS; }

protected:
  template <typename... Args>
  void LogDebug(Args&&... args) const {
    LogDebugHelper(logger(), std::forward<Args>(args)...);
  }

  template <typename... Args>
  void LogInfo(Args&&... args) const {
    LogInfoHelper(logger(), std::forward<Args>(args)...);
  }

  template <typename... Args>
  void LogWarn(Args&&... args) const {
    LogWarnHelper(logger(), std::forward<Args>(args)...);
  }

  template <typename... Args>
  void LogError(Args&&... args) const {
    LogErrorHelper(logger(), std::forward<Args>(args)...);
  }

  void ForwardResources(Resources& r) const { r = resources_; }

private:
  /// Virtual helper function for setup(). Does nothing by default.
  virtual Status doSetup(const Setup::In& /*in*/, Setup::Out& /*out*/) {
    return SUCCESS;
  }

  // Log helper functions
  spdlog::logger& logger() const { return *resources_.logger; }

  template <typename T, typename... Args>
  void LogDebugHelper(T&& logger, Args&&... args) const {
    logger.debug(std::forward<Args>(args)...);
  }

  template <typename T, typename... Args>
  void LogInfoHelper(T&& logger, Args&&... args) const {
    logger.info(std::forward<Args>(args)...);
  }

  template <typename T, typename... Args>
  void LogWarnHelper(T&& logger, Args&&... args) const {
    logger.warn(std::forward<Args>(args)...);
  }

  template <typename T, typename... Args>
  void LogErrorHelper(T&& logger, Args&&... args) const {
    logger.error(std::forward<Args>(args)...);
  }

private:
  bool inputWasJson_ = false;
  std::exception_ptr jsonParseException_ = {};
  Json inputJson_;
  Json setupData_;
  const State* currentState = nullptr;
  Resources resources_;

  /// The SDK version.
  const Version sdkVersion_ = Version(kSdkVersionMajor, kSdkVersionMinor);
};

template <typename... PluginT>
class PluginImplementation : public PluginT... {
public:
  static_assert(sizeof...(PluginT) > 0);

  /// \override
  ///
  /// This collects the plugin type names from the static 
  /// `getPluginTypeName()` methods on the plugin interface types this
  /// plugin is inheriting from.
  ///
  [[nodiscard]] std::vector<std::string> getPluginTypeNames() const final {
    std::vector<std::string> typeNames{};
    (PluginT::getPluginTypeName(typeNames), ...);
    return typeNames;
  }
};

/// \}
}  // namespace d5::apis
