/** Interface file for the plugin interface, which enables communication
 *  between plugins and the core and themselves.
 *
 *  \file PluginInterface.h
 */

#pragma once

#include <memory>
#include <string>
#include <utility>
#include <vector>

#include "dirsig5/sdk_info.h"
#include "dirsig5/base/Version.h"

namespace d5::apis {

/// \addtogroup apis
/// \{

/** Command line options available for the plugin. All options
 *  should be of the form "--option" or "--option=val". The
 *  values below are used to construct the DIRSIG usage message.
 */
struct Option {
  Option(std::string opt, std::string descr)
      : option(std::move(opt)), description(std::move(descr)) {}

  std::string option;
  std::string description;
};

/// A list of Option's
using Options = std::vector<Option>;

/// A request for a plugin
struct PluginRequest {
  // Destructor
  virtual ~PluginRequest() = default;

  /// Name of the plugin
  std::string name;

  /// Inputs for the plugin (format is plugin-specific)
  std::string inputs;

protected:
  PluginRequest(std::string name_, std::string inputs_)
      : name(std::move(name_)), inputs(std::move(inputs_)) {}
};

/// A request for a specific type of plugin
template <class PluginType>
struct PluginRequestT : public PluginRequest {
  /** Constructor
   *
   *  \param name_    Name of the plugin
   *  \param inputs_  Inputs for the plugin
   */
  PluginRequestT(const std::string& name_, const std::string& inputs_)
      : PluginRequest(name_, inputs_) {}

  /// Pointer to the plugin (will be populated by PluginInterface)
  PluginType* plugin = nullptr;
};

/// API that allows the caller to get access to a plugin
class PluginInterface {
public:
  virtual ~PluginInterface() = default;

  /** Request a plugin. Note that the templated form must be used
   *  and the member plugin is null if the plugin could not be
   *  found and false is returned.
   */
  virtual bool requestPlugin(PluginRequest* request) = 0;
};

/** This class provides a templated framework for plugins to implement
 *  their own factory for constructing objects that they define.
 */
template <class PluginType>
class PluginFactory {
public:
  /// Destuctor
  virtual ~PluginFactory() = default;

  /// Get an instance of the plugin. Should return null on failure.
  virtual std::shared_ptr<PluginType> getInstance() = 0;

  /// Get the SDK version used to build the plugin
  [[nodiscard]] const Version& getSdkVersion() const { return sdkVersion_; }

private:
  /// The SDK version.
  const Version sdkVersion_ = Version(kSdkVersionMajor, kSdkVersionMinor);
};

/// \cond HIDDEN_SYMBOLS

namespace plugins {
class AtmosphereInterface;
class BackgroundSkyInterface;
class EphemerisInterface;
class CloudModelInterface;
class DynamicInstanceInterface;
class EarthModelInterface;
class GeometryInterface;
class LaserPulseInterface;
class MaterialInterface;
class MediumBoundaryInterface;
class MediumIOPInterface;
class StateControllerInterface;
class SensorInterface;
class WeatherInterface;
}  // namespace plugins

#define PLUGIN_FACTORY_ADD_METHOD(Name) \
  add(const std::string&, PluginFactory<plugins::Name>*, const Options& = {})

/// \endcond

/// This class is used internally to store factories that generate
/// instances for each implemented interface. The add function
/// allows plugins to register their factories.
///
class PluginFactoryRegistry {
public:
  virtual ~PluginFactoryRegistry() = default;

  /// Add a factory to the registry. Note that a single plugin
  /// can register any number of factories with one or more
  /// interface type.
  ///
  /// The provided name is used to link a request for an object
  /// with this factory. Note that factory names must be unique
  /// and if multiple factories with the same name are provided
  /// only the first such factory is used.
  ///
  /// The (optional) options vector used used only to list
  /// command-line parameters that can be passed to the plugin.
  ///
  ///\{
  virtual void PLUGIN_FACTORY_ADD_METHOD(AtmosphereInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(BackgroundSkyInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(EphemerisInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(CloudModelInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(DynamicInstanceInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(EarthModelInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(GeometryInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(LaserPulseInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(MaterialInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(MediumBoundaryInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(MediumIOPInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(StateControllerInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(SensorInterface) {}
  virtual void PLUGIN_FACTORY_ADD_METHOD(WeatherInterface) {}
  ///\}
};

#undef PLUGIN_FACTORY_ADD_METHOD

/// \}
}  // namespace d5::apis
