/** Interface file for the primitive object API
 *
 *  \file PrimitiveInfo.h
 */

#pragma once

namespace d5::apis {

/** A struct to represent a geometric primitive, defined by a Type and Params.
 *  TODO: this could just be refactored to be a std::variant
 *  \ingroup apis
 */
struct PrimitiveInfo {
  /// Destructor
  ~PrimitiveInfo() = default;

  /// Constants for the type of supported primitives
  enum Type {
    UNKNOWN,
    SPHERE,
    BOX,
    DISK,
    CYLINDER,
    CURVED_FRUSTUM,
    CATENARY_CURVE,
    SINUSOID
  };

  /// The type of primitive
  Type type = UNKNOWN;

  /// The parameters are stored in a union
  union Params {
    Params() = default;

    /// Sphere parameters
    struct Sphere {
      float radius;    ///< Radius of the sphere
      float origin[3]; ///< Center of the sphere
    } sphere;

    /// Axis-aligned box
    struct Box {
      float lower[3]; ///< Lower bound of the box
      float upper[3]; ///< Upper bound of the box
    } box;

    /// Flat disk
    struct Disk {
      float radius;    ///< Radius of the disk
      float origin[3]; ///< Center of the disk
      float normal[3]; ///< Normal vector of the disk
    } disk;

    /// Cylinder
    struct Cylinder {
      float radius;    ///< Radius of the cylinder
      float pointA[3]; ///< Center point of one end
      float pointB[3]; ///< Center point of the other end
      bool capA;       ///< Whether the first end is closed
      bool capB;       ///< Whether the other end is closed
    } cylinder;

    /** A curved frustum (generalization of a cone where the ends can be
     *  rounded rectangles)
     */
    struct CurvedFrustum {
      float topRadius;       ///< Radius of the curves on the top
      float bottomRadius;    ///< Radius of the curves on the bottom
      float bottomCenter[3]; ///< Center of the bottom
      float bottomXWidth;    ///< Side length of the bottom in X
      float bottomYWidth;    ///< Side length of the bottom in Y
      float height;          ///< Height
      float rampLength;      ///< One side of the frustum can be a ramp
      bool topCapFlag;       /// Whether the top is closed
      bool bottomCapFlag;    /// Whether the bottom is closed
      bool flipFlag;         /// Whether the shape is flipped vertically
    } curvedfrustum;

    /// A catenary curve
    struct CatenaryCurve {
      float pointA[3]; ///< One endpoint
      float pointB[3]; ///< The other endpoint
      float sag;       ///< How much the curve sags
      float thickness; ///< Thickness of the curve
    } catenarycurve;

    /// Vertically-extruded sinusoid
    struct Sinusoid {
      float lower[3];       ///< Lower bound of the box
      float upper[3];       ///< Upper bound of the box
      float magnitude;      ///< Amplitude of the sinusoid
      float period;         ///< Period of the sinusoid
      float orientation[3]; ///< Direction of the sinusoid
      float phaseShift;     ///< Phase shift [radians]
    } sinusoid;
  };

  /// Union of all potential primitive parameters
  Params params;
};

} // namespace d5::apis
