/** Interface file for a raster input plugin. This plugin type is significantly
 *  simpler and does not need all of the machinery of the other plugins.
 *
 *  \file RasterInputInterface.h
 */
#pragma once

#include <cmath>
#include <cstddef>
#include <cstdint>
#include <functional>
#include <limits>
#include <optional>
#include <stdexcept>
#include <string>
#include <vector>

#include "dirsig5/base/DType.h"

namespace d5::apis {

/** \defgroup apis_rasterinput  Raster Input Plugin
 *  Holds the API definition for a raster input plugin
 *  \ingroup plugin_apis
 */

/// \addtogroup apis_rasterinput
/// \{

/** XMacro for the list of raster data types
 *
 *  \param X Function macro of the form X(name, value, C++ type, description)
 */
#define D5_RASTER_DATA_TYPES(X)                                                \
  X(Uint8, 1, std::uint8_t, 8 - bit unsigned integer)                          \
  X(Int16, 2, std::int16_t, 16 - bit signed integer)                           \
  X(Int32, 3, std::int32_t, 32 - bit signed integer)                           \
  X(Float32, 4, float, 32 - bit floating point)                                \
  X(Float64, 5, double, 64 - bit floating point)                               \
  X(Uint16, 12, std::uint16_t, 16 - bit unsigned integer)                      \
  X(Uint32, 13, std::uint32_t, 32 - bit unsigned integer)                      \
  X(Int64, 14, std::int64_t, 64 - bit signed integer)                          \
  X(Uint64, 16, std::uint64_t, 64 - bit unsigned integer)

using RasterDataType = DType;

namespace detail {

/// Metafunction to test whether a type has a `value` member
template <typename T, typename = int>
struct has_value_member : std::false_type {};

/// \cond HIDDEN_SYMBOLS

template <typename T>
struct has_value_member<T, decltype((void)T::value, 0)> : std::true_type {};

/// \endcond

/// Convenience alias for has_value_member<T>::value
template <typename T>
inline constexpr bool has_value_member_v = has_value_member<T>::value;

} // namespace detail

/** Enum dispatch function for RasterDataType
 *
 *  \tparam Callable  The callable class template. This template should have a
 *                    static dispatch method and optionally a value
 *                    metafunction that specifies whether the template parameter
 *                    should be accepted.
 *
 *  \param type  The type parameter
 *  \param args  Any arguments to forward to the function
 *
 *  \return The return value of the function
 */
template <template <RasterDataType> typename Callable, typename... Args>
inline decltype(auto) dispatch(RasterDataType type, Args &&...args) {
  auto err = [](RasterDataType type) {
    return std::runtime_error(
        "d5::apis::dispatch(RasterDataType): passed unsupported type: " +
        std::string(to_string(type)) + " (" +
        std::to_string(static_cast<int>(type)) + ")");
  };

#define X_DISPATCH_CASE(T, val, cpp_t, desc)                                   \
  case RasterDataType::T: {                                                    \
    using CallableT = Callable<RasterDataType::T>;                             \
    if constexpr (detail::has_value_member_v<CallableT>) {                     \
      if constexpr (CallableT::value)                                          \
        return CallableT::dispatch(std::forward<Args>(args)...);               \
      else                                                                     \
        throw err(type);                                                       \
    } else                                                                     \
      return CallableT::dispatch(std::forward<Args>(args)...);                 \
  }

  switch (type) {
    D5_RASTER_DATA_TYPES(X_DISPATCH_CASE)
  default:
  case RasterDataType::Unknown:
    throw err(type);
  }

#undef X_DISPATCH_CASE
}

/// The various ways that raster data may be interleaved in memory
enum class RasterInterleave : uint8_t {
  BSQ, ///< Band-sequential (band, row, column indexing)
  BIL, ///< Band interleave-by-line (row, band, column indexing)
  BIP  ///< Band interleave-by-pixel (row, column, band indexing)
};

/** An abstraction for the output buffer so that the raster reading abstraction
 *  is not dependent on this data type.
 */
struct RasterDestination {
  /// Destructor
  virtual ~RasterDestination() = default;

  /// The type of interleaving that the destination buffer is using
  [[nodiscard]] virtual RasterInterleave interleave() const = 0;

  /** Allocates the destination buffer
   *
   *  \param bands     The number of bands in the image
   *  \param rows      The number of rows in the image
   *  \param columns   The number of columns in the image
   *  \param dataType  The data type of the image
   *
   *  \throw std::exception subclass on any failure
   */
  virtual void allocate(size_t bands, size_t rows, size_t columns,
                        RasterDataType dataType) = 0;

  /** Accessor for a read-write pointer to a row in the destination buffer
   *
   *  \param band  The band index of the row (useful for BSQ, BIL)
   *  \param row   The index of the row (useful for all interleaves)
   *
   *  \return Pointer to the beginning of the specified row
   */
  virtual char *row(size_t band, size_t row) = 0;
};

/// Subclass for band-sequential destinations
struct BsqRasterDestination : RasterDestination {
  [[nodiscard]] RasterInterleave interleave() const override {
    return RasterInterleave::BSQ;
  }
};

/// Subclass for band-interleave-by-line destinations
struct BilRasterDestination : RasterDestination {
  [[nodiscard]] RasterInterleave interleave() const override {
    return RasterInterleave::BIL;
  }
};

/// Subclass for band-interleave-by-pixel destinations
struct BipRasterDestination : RasterDestination {
  [[nodiscard]] RasterInterleave interleave() const override {
    return RasterInterleave::BIP;
  }
};

/// Hold the concept of an ROI of an image to load
struct RasterInputRoi {
  size_t startX; ///< Start X coordinate (inclusive)
  size_t startY; ///< Start Y coordinate (inclusive)
  size_t sizeX;  ///< Number of pixels to load in X
  size_t sizeY;  ///< Number of pixels to load in Y

  /** Factory function
   *
   *  \param minX  Start X coordinate (inclusive)
   *  \param minY  Start Y coordinate (inclusive)
   *  \param nX    Number of pixels to load in X
   *  \param nY    Number of pixels to load in Y
   */
  static RasterInputRoi fromMinSize(size_t minX, size_t minY, size_t nX,
                                    size_t nY) {
    return {minX, minY, nX, nY};
  }

  /** Factory function
   *
   *  \param minX  Start X coordinate (inclusive)
   *  \param minY  Start Y coordinate (inclusive)
   *  \param maxX  End X coordinate (inclusive)
   *  \param maxY  End Y coordinate (inclusive)
   */
  static RasterInputRoi fromMinMaxInclusive(size_t minX, size_t minY,
                                            size_t maxX, size_t maxY) {
    return {minX, minY, maxX - minX + 1, maxY - minY + 1};
  }

  /** Factory function
   *
   *  \param minX  Start X coordinate (inclusive)
   *  \param minY  Start Y coordinate (inclusive)
   *  \param maxX  End X coordinate (exclusive)
   *  \param maxY  End Y coordinate (exclusive)
   */
  static RasterInputRoi fromMinMaxExclusive(size_t minX, size_t minY,
                                            size_t maxX, size_t maxY) {
    return {minX, minY, maxX - minX, maxY - minY};
  }
};

/// Metadata about a raster
struct RasterMetadata {
  /// Number of rows in the raster
  size_t rows = 0;

  /// Number of columns in the raster
  size_t cols = 0;

  /// The data type of the raster
  RasterDataType dataType;

  /// Per-band metadata structure
  struct BandMetadata {
    /// Name of the band
    std::string name;

    /// Center of the band's bandpass [micron]
    std::optional<double> centerWavelengthMicron;

    /// Width of the band (FWHM) [micron]
    std::optional<double> widthMicron;
  };

  /// Metadata about each band
  std::vector<std::optional<BandMetadata>> bands;
};

/// Definition for the plugin initialization function
namespace RasterInputInit {

/// Input for initialization (empty)
struct In {};

/// Output for initialization
struct Out {
  /// The extensions that are supported via this plugin
  std::vector<std::string> extensions;

  /// If a file has no supported extension, whether to try anyways with this
  /// plugin. This will only occur if no other plugin has claimed the extension
  /// of the file.
  bool tryWithUnrecognizedExtension = false;

  /// Whether the plugin supports a cheap metadata scan
  bool supportsCheapMetadataScan = false;

  /// Whether the plugin supports a progress callback
  bool supportsProgress = false;

  /// Reason for a failure (if one occurs)
  std::string error;
};
} // namespace RasterInputInit

/// Definition for the lightweight metadata reading API
namespace RasterMetadataRead {

/// Input for a metadata query request
struct In {
  /// Path to the raster to scan
  std::string path;
};

/// Output structure for a raster metadata query
struct Out {
  /// The metadata to be populated
  RasterMetadata metadata;

  /// Error string in the case of a failure
  std::string error;
};

} // namespace RasterMetadataRead

/// Definition for the raster reading function
namespace RasterInputRead {

/// Raster read query definition
struct In {
  /// Path to the raster to read
  std::string path;

  /// Whether a spatial ROI is present. If a
  /// std::numeric_limits<size_t>::max() is seen in the size parameters,
  /// that should be interpreted as "to the end" of that dimension.
  std::optional<RasterInputRoi> roi;

  /// Whether to load the image as grayscale (for RGB images)
  bool loadAsGrayscale = false;

  /// Whether to load default bands (if available and bandsToLoad is empty)
  bool loadDefaultBands = false;

  /// Band indices to load (empty indicates load all bands)
  std::vector<size_t> bandsToLoad;

  /// What step size to take when loading the image
  size_t pixelStride = 1;

  /// Callback function for progress updates (called with 0-1 progress fraction)
  std::function<void(double)> progressCallback;
};

/// Output structure for a read request
struct Out {
  /// Metadata for the image
  RasterMetadata metadata;

  /// In the case of a failed, load, the error message
  std::string error;
};

} // namespace RasterInputRead

/** A plugin that can provide raster reading functionality to DIRSIG5.
 *
 *  A plugin advertises its features to DIRSIG and DIRSIG will consider all
 *  available plugins and route raster read requests to the individual plugins.
 *  Extensions are used as the primary mechanism, but extensionless files can
 *  also be used via plugins that advertise the capability to handle them.
 */
class RasterInputInterface {
public:
  /// Destructor
  virtual ~RasterInputInterface() = default;

  /// Returns the name of the plugin
  [[nodiscard]] virtual std::string name() const = 0;

  /** Initialize the plugin
   *
   *  \param[in] in    Input structure, currently empty
   *  \param[out] out  Output structure, to be filled with information about the
   *                   plugin
   *
   *  \return Whether the plugin was successfully initialized. A failure will
   *          result in the plugin being skipped and a warning being printed. If
   *          false is returned, the failure reason should be populated to
   *          communicate an error condition to the user.
   */
  virtual bool init(const RasterInputInit::In &in,
                    RasterInputInit::Out &out) = 0;

  /** Reads an image
   *
   *  \param[in] in    Query definition, contains read parameters and the
   *                   destination object
   *  \param[in,out] destination  The destination object to which to write the
   *                              raster data
   *  \param[out] out  Output structure, should populate error on a failed read
   *
   *  \return Whether the plugin could read the image. The error should be
   *          populated on a failed read, although it may be displayed to the
   *          user as an error or a warning, depending on the other plugins that
   *          are present.
   */
  virtual bool read(const RasterInputRead::In &in,
                    RasterDestination &destination,
                    RasterInputRead::Out &out) = 0;

  /// \overload
  bool read(const RasterInputRead::In &in, RasterDestination &&destination,
            RasterInputRead::Out &out) {
    return read(in, destination, out);
  }

  /// Reads the metadata of an image
  virtual bool metadata(const RasterMetadataRead::In &,
                        RasterMetadataRead::Out &) {
    return false;
  }
};

class RasterSizeParameters {
public:
  RasterSizeParameters(const RasterInputRead::In &in, size_t imHeight,
                       size_t imWidth)
      : imageHeight_(imHeight), imageWidth_(imWidth) {
    if (in.roi) {
      const auto &r = *in.roi;
      if (r.startX >= imWidth) {
        throw std::runtime_error(
            "invalid ROI provided, starting column was set to " +
            std::to_string(r.startX) + ", but the image only has " +
            std::to_string(imWidth) + " columns");
      } else if (r.startY >= imHeight) {
        throw std::runtime_error(
            "invalid ROI provided, starting row was set to " +
            std::to_string(r.startY) + ", but the image only has " +
            std::to_string(imHeight) + " rows");
      }
      startX_ = r.startX;
      startY_ = r.startY;
      sizeX_ = r.sizeX;
      sizeY_ = r.sizeY;
      if (sizeX_ == std::numeric_limits<size_t>::max()) {
        sizeX_ = imWidth - startX_;
      }
      if (sizeY_ == std::numeric_limits<size_t>::max()) {
        sizeY_ = imHeight - startY_;
      }

      if (r.startX + r.sizeX > imWidth) {
        throw std::runtime_error("invalid ROI provided, ending column was " +
                                 std::to_string(r.startX + r.sizeX) +
                                 ", which is past the image width of " +
                                 std::to_string(imWidth));
      } else if (r.startY + r.sizeY > imHeight) {
        throw std::runtime_error("invalid ROI provided, ending row was " +
                                 std::to_string(r.startY + r.sizeY) +
                                 ", which is past the image height of " +
                                 std::to_string(imHeight));
      }
    } else {
      sizeX_ = imWidth;
      sizeY_ = imHeight;
    }

    stridedSizeX_ = sizeX_ == 0 ? 0 : (sizeX_ - 1) / in.pixelStride + 1;
    stridedSizeY_ = sizeY_ == 0 ? 0 : (sizeY_ - 1) / in.pixelStride + 1;
    progressRowInterval_ = std::max(
        size_t{1}, static_cast<size_t>(std::round(stridedSizeY_ / 100.)));
  }

  [[nodiscard]] size_t startColumn() const { return startX_; }
  [[nodiscard]] size_t startRow() const { return startY_; }
  [[nodiscard]] size_t stridedNumColumns() const { return stridedSizeX_; }
  [[nodiscard]] size_t stridedNumRows() const { return stridedSizeY_; }
  [[nodiscard]] size_t unstridedNumColumns() const { return sizeX_; }
  [[nodiscard]] size_t unstridedNumRows() const { return sizeY_; }
  [[nodiscard]] size_t imageNumRows() const { return imageHeight_; }
  [[nodiscard]] size_t imageNumColumns() const { return imageWidth_; }
  [[nodiscard]] size_t progressRowInterval() const {
    return progressRowInterval_;
  }

private:
  size_t startX_ = 0;
  size_t startY_ = 0;
  size_t sizeX_ = 0;
  size_t sizeY_ = 0;
  size_t stridedSizeX_ = 0;
  size_t stridedSizeY_ = 0;
  size_t imageHeight_ = 0;
  size_t imageWidth_ = 0;
  size_t progressRowInterval_ = 0;
};

inline void SetSizeMetadata(const RasterSizeParameters &params,
                            RasterMetadata &metadata) {
  metadata.cols = params.stridedNumColumns();
  metadata.rows = params.stridedNumRows();
}

/// \}
} // namespace d5::apis
