/** Interface file for the Ray API
 *
 *  \file Ray.h
 */

#pragma once

#include <cmath>
#include <functional>
#include <iostream>
#include <limits>
#include <optional>
#include <random>
#include <stack>
#include <vector>

#include "dirsig5/base/Constants.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/IntID.h"
#include "dirsig5/base/Pcg.h"

#define RTC_MAX_INSTANCE_LEVEL_COUNT 1

namespace d5::apis {

/// \addtogroup apis
/// \{

class ObjectInterface;

enum Coordinates {
  DoubleEarth, ///< Double-precision, earth (ECEF) coordinates
  DoubleScene, ///< Double-precision, scene-local (ENU) coordinates
  SingleScene  ///< Single-precision, scene-local (ENU) coordinates
};

constexpr double RayMax = std::numeric_limits<double>::infinity();

// borrowed from embree
#ifdef _WIN32
#define RTCORE_ALIGN_PROXY(...) __declspec(align(__VA_ARGS__))
#else
#define RTCORE_ALIGN_PROXY(...) __attribute__((aligned(__VA_ARGS__)))
#endif

struct RTCORE_ALIGN_PROXY(16) EmbreeRay {
  struct RTCORE_ALIGN_PROXY(16) RayStruct {
    float org[3];
    float tnear;
    float dir[3];
    float time;
    float tfar;
    unsigned mask;
    unsigned id;
    unsigned flags;
  } ray;

  struct RTCORE_ALIGN_PROXY(16) HitStruct {
    float Ng[3];
    float u;
    float v;
    unsigned primID;
    unsigned geomID;
    unsigned instID[RTC_MAX_INSTANCE_LEVEL_COUNT];
  } hit;
};

/// This is a simple struct for gathering meta data about a hit from
/// the function that sets the scene material ID for a ray. It is
/// optional and is usually only used if truth is being collected. The
/// "collectMetaData" flag in the ray must be set to be used.
struct HitMetaData {
  HitMetaData() {
    matIDStack.reserve(10);
    mixedMats.reserve(10);
  }

  void clear() {
    matIDStack.clear();
    mixedMats.clear();
    hasNull = false;
    u = 0;
    v = 0;
  }

  void clearMixture() {
    mixedMats.clear();
    hasNull = false;
  }

  std::vector<IntID> matIDStack;
  float u = 0;
  float v = 0;
  IntID sceneID = INVALID_ID;
  std::vector<std::pair<float, IntID>> mixedMats;
  bool hasNull = false;
};

/// Padded representation of a RTCRay that is aligned for Embree use. This ray
/// exists within the coordinate system of a certain scene (part of the
/// definition below) or is global.
///
/// The order and type of the variables stored here cannot be changed.
struct Ray {
  Ray(EmbreeRay *embreeRay)
      : coreRayHit(embreeRay), org(&embreeRay->ray.org[0], 3, 1),
        dir(&embreeRay->ray.dir[0], 3, 1), Ng(&embreeRay->hit.Ng[0], 3, 1),
        tnear(embreeRay->ray.tnear), tfar(embreeRay->ray.tfar),
        u(embreeRay->hit.u), v(embreeRay->hit.v), primID(embreeRay->hit.primID),
        instID(embreeRay->hit.instID[0]), geomID(embreeRay->hit.geomID) {
    embreeRay->ray.mask = static_cast<unsigned>(-1);
    embreeRay->ray.time = 0;
  }

  void *coreRayHit;
  Eigen::Map<Eigen::Vector3f> org;
  Eigen::Map<Eigen::Vector3f> dir;
  Eigen::Map<Eigen::Vector3f> Ng;
  float &tnear;
  float &tfar;
  float &u;
  float &v;
  unsigned &primID;
  unsigned &instID;
  unsigned &geomID;

  /// The associated scene ID (if in local coords)
  IntID sceneID = INVALID_ID;

  /// The hit material ID
  IntID matID = INVALID_ID;

  /// The hit object
  const ObjectInterface *hitObject{};

  /// The hit normal (not necessarily the geometric normal, Ng)
  Vectorf normal = {0, 0, 0};

  /// Global, double-precision origin (relative to a scene)
  Point gOrigin = {0, 0, 0};

  /// Global, double-precision direction (relative to a scene)
  Vector gDirection = {0, 0, -1};

  /// Elapsed time since the origin
  double elapsedTime = 0;

  /// Global distance [m]. This value tracks the total distance traveled
  /// by the ray starting from the global origin (i.e. it includes any
  /// travel due to localization, possibly via a non-straight path).
  double propagationDistance = 0;

  /// Flag indicating this ray should be skipped
  bool skip = true;

  /// Flag indicating this ray is for a boolean shadow check
  bool boolShadow = false;

  /// Flag indicating this ray is high priority (is involved in a
  /// direct lighting calculation)
  bool highPriority = true;

  /// Record of the parent scene ID when switching rays over to a proxy
  /// scene (subsequent rays need to return to the original scene)
  ///
  /// The default value (-1) indicates that the scene ID is un-proxied
  IntID proxyParent = INVALID_ID;

  /// The current medium, to quickly eliminate interface self-intersection
  /// checks.
  IntID curMedium = INVALID_ID;

  /// Coordinate system of the ray. Can be either earth (ECEF) or scene
  /// (ENU) and double or single precision in either.
  Coordinates coordinates = SingleScene;

  /// Double-precision origin coordinates in ECEF
  Point ecefOrigin = {};

  /// Double-precision direction coordinates in ECEF
  Vector ecefDirection = {};

  /// Random engine
  RandEngine randEngine = {};

  /// The current index of refraction
  double currentIOR = 1;

  /// The hit meta data
  HitMetaData hitMetaData;

  /// Flag indicating if the hit meta data should be collected
  bool collectMetaData = false;

  /// Optional temperature [Kelvin] that will override the temperature model
  std::optional<double> temperatureK;

  /// Optional concentration distance [conc*m] along the ray
  std::optional<float> concDist;

  struct VertexInfo {
    std::array<Pointf, 3> vs;   // vertexes
    std::array<Vectorf, 3> vns; // vertex normals
  };

  /// Optional information about vertices
  std::optional<VertexInfo> vertexInfo;

  /// Information about primary reflectance versus transmittance counts
  int primaryRCount = -1, primaryTCount = -1;

  /// Copy just local data from another ray
  void copy(const Ray &ray) {
    org = ray.org;
    dir = ray.dir;
    Ng = ray.Ng;
    tnear = ray.tnear;
    tfar = ray.tfar;
    u = ray.u;
    v = ray.v;
    geomID = ray.geomID;
    primID = ray.primID;
    instID = ray.instID;
    sceneID = ray.sceneID;
    matID = ray.matID;
    hitObject = ray.hitObject;
    normal = ray.normal;
    gOrigin = ray.gOrigin;
    gDirection = ray.gDirection;
    elapsedTime = ray.elapsedTime;
    propagationDistance = ray.propagationDistance;
    skip = ray.skip;
    boolShadow = ray.boolShadow;
    highPriority = ray.highPriority;
    proxyParent = ray.proxyParent;
    curMedium = ray.curMedium;
    collectMetaData = ray.collectMetaData;
    hitMetaData = ray.hitMetaData;
    currentIOR = ray.currentIOR;
    coordinates = ray.coordinates;
    randEngine = ray.randEngine;
    temperatureK = ray.temperatureK;
    concDist = ray.concDist;
    vertexInfo = ray.vertexInfo;
    ecefOrigin = ray.ecefOrigin;
    ecefDirection = ray.ecefDirection;
    primaryRCount = ray.primaryRCount;
    primaryTCount = ray.primaryTCount;
  }

  /// Get the local hit point (quick helper function)
  Pointf getLocalHitPoint() { return org + tfar * dir; }

  void initGenericRay() {
    tnear = 1e-3f;
    tfar = RayMax;
    u = 0;
    v = 0;
    geomID = unsigned(-1);
    primID = unsigned(-1);
    instID = unsigned(-1);
    matID = INVALID_ID;
    hitObject = nullptr;
    elapsedTime = 0;
    propagationDistance = 0;
    skip = false;
    boolShadow = false;
    proxyParent = INVALID_ID;
    curMedium = INVALID_ID;
    collectMetaData = false;
    hitMetaData.clear();
    temperatureK.reset();
    concDist.reset();
    vertexInfo.reset();
    primaryRCount = -1;
    primaryTCount = -1;
  }

  /// Initialize a ray based on a scene local coordinate system. Does
  /// not set the origin and direction.
  void initLocalSceneRay(IntID scene) {
    initGenericRay();
    sceneID = scene;
    coordinates = SingleScene;
  }

  /// Initialize a ray based on a scene global coordinate system. Does
  /// not set the origin and direction.
  void initGlobalSceneRay(IntID scene) {
    initGenericRay();
    sceneID = scene;
    coordinates = DoubleScene;
  }

  /// Coordinate agnostic way to propagate a ray a certain distance
  void propagate(float dist) {
    if (coordinates == SingleScene) {
      org += dist * dir;
    } else if (coordinates == DoubleEarth) {
      ecefOrigin += dist * ecefDirection;
    } else {
      gOrigin += dist * gDirection;
    }
  }

  /// Check if we hit the backside according to the ray normal
  bool hitBacksideCheck() {
    if (coordinates == SingleScene) {
      return (dir.dot(normal) > 0.f);
    } else if (coordinates == DoubleEarth) {
      return (ecefDirection.dot(normal.cast<double>()) > 0.f);
    } else {
      return (gDirection.dot(normal.cast<double>()) > 0.f);
    }
  }

  friend std::ostream &operator<<(std::ostream &out, const Ray &ray) {
    Vector org = Vectorf(ray.org).cast<double>();
    Vector dir = Vectorf(ray.dir).cast<double>();
    if (ray.coordinates == DoubleEarth) {
      org = ray.ecefOrigin;
      dir = ray.ecefDirection;
    } else if (ray.coordinates == DoubleScene) {
      org = ray.gOrigin;
      dir = ray.gDirection;
    }
    out << '[' << org[0] << ", " << org[1] << ", " << org[2] << "] ";
    out << '[' << dir[0] << ", " << dir[1] << ", " << dir[2] << "] ";
    out << ray.tnear << " -> " << ray.tfar << ' ';
    out << '('                 //
        << ray.sceneID << ", " //
        << ray.geomID << ", "  //
        << ray.primID << ", "  //
        << ray.instID << ", "  //
        << ray.matID << ") "   //
        << int(ray.skip);
    return out;
  }
};

using Rays = std::vector<Ray *>;

/// \}
} // namespace d5::apis
