#pragma once

#include <cstdint>

#include "dirsig5/apis/ObjectInterface.h"
#include "dirsig5/apis/PrimitiveInfo.h"
#include "dirsig5/apis/Ray.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/Version.h"

namespace d5::apis {

/// \addtogroup apis
/// \{

using FloatBuffer = std::vector<float>;
using IndexBuffer = std::vector<std::uint32_t>;

enum CorePluginType { UNDEFINED = 0, EMBREE };

struct CorePluginInfo {
  CorePluginInfo(CorePluginType t, Version v) : type(t), version(v) {}
  CorePluginType type;
  Version version;
};

/// A scene is intended to be a semi-autonomous group of objects and materials
/// contained within a bounded volume. Scenes should not be defined (loaded)
/// until needed and support level-of-detail (LoD) usage by allowing
/// overlapping scene definitions corresponding to different GSD ranges
/// but the same spatial bounding volume. Distance calculations outside of
/// the scene are performed with double precision, global coordinates, while
/// calculations within a scene are coverted to single precision floats in
/// the local coordinate space.
///
/// This header file defines the publicly available interface to scene
/// definitions within the core.
class RayTracingInterface {
public:
  virtual ~RayTracingInterface() = default;

  /// Get information about this plugin.
  virtual CorePluginInfo getPluginInfo() = 0;

  /// Create a scene (anchored origin, local box version).
  ///
  /// Creates a scene placeholder volume. The first provided scene will be
  /// used by default for GSD ranges not subsequently added.
  ///
  /// Latitude range: [-90:90]
  /// Longitude range: [-180:180]
  ///
  /// Angles are in decimal degrees, distances in meters.
  ///
  /// \param[in] boxMin     Bound box minimum coordinates.
  /// \param[in] boxMax     Bound box maximum coordinates.
  /// \param[in] latLonAlt  Scene origin latitude, longitude, altitude.
  /// \param[in] callback   Scene notification callback.
  ///
  /// \return globally unique ID corresponding to both the proxy volume and
  ///  the default scene that can fill it. Any additional LoD scenes will
  ///  have their own unique ID as well.
  ///
  virtual IntID createScene(Pointf boxMin, Pointf boxMax, Pointf latLonAlt,
                            SceneCallback callback) = 0;

  /// Create base geometry.
  ///
  /// Creates an abstracted representation of an object in the scene. The
  /// geometry does not exist anywhere in the scene until it is actually
  /// instanced but we give it an initial bounding box in its local
  /// coordinate system when it is created (the bounding box is then
  /// modified by each instance transform). Created geometry does not
  /// need to have a type yet (e.g. it could be a facetized or primitive
  /// geometry) -- all that matters is that it will fit within its
  /// bounding box when added. The base geometry ID passed back must
  /// be provided when creating an instance.
  ///
  /// \param[in] sceneID  Unique ID to the scene.
  /// \param[in] boxMin   Bound box minimum coordinates.
  /// \param[in] boxMax   Bound box maximum coordinates.
  /// \param[in] callback Object notification callback.
  ///
  /// \return scene unique ID corresponding to the base geometry.
  ///
  virtual IntID createBaseGeometry(IntID sceneID, Pointf boxMin, Pointf boxMax,
                                   ObjectCallback callback) = 0;

  /// Add base facetization.
  ///
  /// Add the minimal amount of information needed to represent facetized
  /// geometry. The core engine takes control of the memory in the vector.
  ///
  /// Facet indexes are aligned to 16 bytes (V1,V2,V3,M), where the fourth
  /// value is the material index associated with the facet (this can be
  /// overriden by a map). This is the "default" alignment and corresponds
  /// to what is expected in Embree if it is built with striding support
  /// set off (which is faster). The stride (16) is set via a parameter
  /// and cannot be changed -- we just need to agree that's
  /// what it is.
  ///
  /// The vertexes are aligned in the same way (4 element stride) and
  /// therefore have an extra float per vertex (currently unused). The
  /// format of each vertex is [X,Y,Z,Q] (Q may become a per-vertex
  /// temperature in the future.
  ///
  /// \param[in] sceneID           Unique ID to the scene.
  /// \param[in] baseID            Unique ID to the base geometry proxy.
  /// \param[in] verts             Buffer of vertexes (X,Y,Z).
  /// \param[in] vertIndexes       Buffer of vertex indexes.
  /// \param[in] vertNormals       Buffer of vertex normals. _Optional_.
  /// \param[in] vertNormalIndexes Buffer of vertex normal indexes. _Optional_.
  ///
  /// \return success flag (returns false if already defined)
  ///
  virtual bool
  addBaseFacetization(IntID sceneID, IntID baseID, const FloatBuffer &verts,
                      const IndexBuffer &vertIndexes,
                      const FloatBuffer &vertNormals = {},
                      const IndexBuffer &vertNormalIndexes = {}) = 0;

  /// Add base Bezier curves.
  ///
  /// \param[in] sceneID       Unique ID to the scene.
  /// \param[in] baseID        Unique ID to the base geometry proxy.
  /// \param[in] verts         Buffer of vertexes (X,Y,Z,R).
  /// \param[in] vertNormals   Buffer of vertex normals. _Optional_.
  /// \param[in] firstIndexes  Buffer of first-vertex indexes. _Optional_.
  ///
  virtual bool addBaseBezierCurves(IntID sceneID, IntID baseID,
                                   const FloatBuffer &verts,
                                   const FloatBuffer &vertNormals = {},
                                   const IndexBuffer &firstIndexes = {}) {
    return false;
  }

  /// Add base primitive.
  ///
  /// \param[in] sceneID   Unique ID to the scene.
  /// \param[in] baseID    Unique ID to the base geometry proxy.
  /// \param[in] info      Info to construct the primitive.
  ///
  virtual bool addBasePrimitive(IntID sceneID, IntID baseID,
                                const PrimitiveInfo &info) = 0;

  /// Add a static instance.
  ///
  /// Add a static copy of a base geometry to the scene by providing its
  /// unique ID (received when added) and an affine matrix describing its
  /// placement. The matrix is a single precision 3x4 flattened array in
  /// row major order. The core does not control the memory corresponding
  /// to the affine matrix.
  ///
  /// \param[in] sceneID           Unique ID to the scene.
  /// \param[in] baseID            Unique ID to the base geometry proxy.
  /// \param[in] matrix            Matrixf 3x4 matrix in row major
  /// order. \param[in] alphaCallback     Alpha map callback. _Optional_.
  ///
  /// \return scene unique instance ID
  ///
  virtual IntID addStaticInstance(IntID sceneID, IntID baseID, Matrixf matrix,
                                  AlphaCallback alphaCallback = nullptr) = 0;

  /// Add a dynamic instance.
  ///
  /// Add a dynamic instance of a base geometry to the scene by providing its
  /// unique ID (received when added) and a callback function that is used
  /// to update the geometry when the temporal state changes.
  ///
  /// \param[in] sceneID           Unique ID to the scene.
  /// \param[in] baseID            Unique ID to the base geometry proxy.
  /// \param[in] dynamicCallback   Dynamic evaluation callback.
  /// \param[in] alphaCallback     Alpha map callback. _Optional_.
  ///
  /// \return scene unique instance ID
  ///
  virtual IntID addDynamicInstance(IntID sceneID, IntID baseID,
                                   DynamicCallback dynamicCallback,
                                   AlphaCallback alphaCallback = nullptr) = 0;

  /// Build a scene.
  ///
  /// This effectively finalizes the scene and uses the callback to load
  /// up the base objects and instances. Intersections and other
  /// queries will not work until the scene is built (note that we
  /// use this manual construction process for flexibility and to
  /// avoid mutexing the process during multi-threaded queuries).
  ///
  /// \param[in] sceneID   Unique ID to the scene.
  ///
  virtual void buildScene(IntID sceneID) = 0;

  /// Set the reference date time.
  ///
  /// Takes a DateTime representing the calendar date/time to be
  /// used as a reference point. This is not used internally
  /// but is passed to callbacks when the internal temporal state
  /// changes (along with the elapsed time).
  ///
  /// Setting the reference time does _not_ update the internal temporal state.
  ///
  /// An unset reference time will generate a fatal error.
  ///
  virtual void setReferenceDateTime(const DateTime &refDateTime) = 0;

  /// Update the temporal state.
  ///
  /// The temporal state is defined by the elapsed time (measured from
  /// the reference date time) and an optional "duration" that sets the
  /// limit of any linear motion blurring that is performed
  /// (i.e. interpolation between two geometry states).
  ///
  /// When the temporal state is updated any dynamic callbacks
  /// (e.g. geometry deformation or instance motion) are called with the
  /// elapsed time (and reference date time).
  ///
  /// A non-zero duration will trigger a second call, offsetting the
  /// elapsed time by the duration. Rays with timing outside of the
  /// [elapsed, elapsed+duration] window will be clipped to either side.
  ///
  /// It is not possible to set either the elapsed time or the duration
  /// to a negative value (an error will result).
  ///
  /// An unset temporal state will generate a fatal error.
  ///
  /// \param[in] elapsedTime  Time [s] elapsed from the reference.
  /// \param[in] duration     Duration [s] used for linear blur.
  ///
  virtual void updateTemporalState(double elapsedTime, double duration) = 0;

  /// Create a wrapper object interface around an instance and return it.
  ///
  /// \param[in] sceneID   Unique ID to the scene.
  /// \param[in] instID    Unique ID to the instance.
  /// \param[in] tags      Tags to associate with the instance.
  ///
  /// \note
  /// The pointer should be owned by the scene implementation, so
  /// the caller must not delete it!
  ///
  /// \note
  /// The object interface may retain the tags as a pointer, so
  /// the caller must not destroy the tags while using the interface!
  ///
  virtual ObjectInterface *
  getObjectInstance(IntID sceneID, IntID instID,
                    const std::vector<std::string> &tags) = 0;

  /// Intersect ray with a given scene.
  virtual void intersect(IntID sceneID, apis::Ray &ray) = 0;

  /// Intersect multiple rays with a given scene.
  virtual void intersect(IntID sceneID, apis::Rays &rays,
                         bool /*coherent*/ = false) {
    for (apis::Ray *ray : rays)
      intersect(sceneID, *ray);
  }
};

/// \}
} // namespace d5::apis

using GetSceneInterfaceCallback = d5::apis::RayTracingInterface *(*)();
