/** Interface file for RunInfo
 *
 *  \file RunInfo.h
 */

#pragma once

#include <memory>
#include <stdexcept>
#include <string>
#include <utility>
#include <vector>

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Json.h"
#include "dirsig5/sdk_info.h"

namespace d5::apis {

/// \addtogroup apis
/// \{

/// DIRSIG run metadata that would be needed to reproduce it
class RunInfo {
 public:
  /// Metadata about a scene
  struct SceneInfo {
    SceneInfo() = default;

    /** Constructor
     *
     *  \param fname  The filename of the scene
     *  \param md5    MD5 hash of the scene
     *  \param lla    Lat/lon/altitude of the scene anchor point
     *  \param bmin   Minimum point of the scene's bounding box in scene LCS
     *  \param bmax   Maximum point of the scene's bounding box in scene LCS
     */
    SceneInfo(
        std::string fname,
        std::string md5,
        Point lla,
        Vectorf bmin,
        Vectorf bmax)
        : filename(std::move(fname)),
          md5Hash(std::move(md5)),
          latLonAlt(std::move(lla)),
          bboxMin(std::move(bmin)),
          bboxMax(std::move(bmax)) {}

    std::string filename;
    std::string md5Hash;
    Point latLonAlt;
    Vectorf bboxMin;
    Vectorf bboxMax;
  };

  /// Metadata about a plugin
  struct PluginInfo {
    PluginInfo() = default;

    /** Constructor
     *
     *  \param name_    Name of the plugin
     *  \param inputs_  Input JSON data given to plugin setup
     *  \param data     Any additional data the plugin wished to publish
     */
    PluginInfo(std::string name_, std::vector<std::string> typeNames_,
      std::string version_, std::string sdkVersion_,
      Json inputs_, Json data)
        : name(std::move(name_)),
          typeNames(std::move(typeNames_)),
          version(std::move(version_)),
          sdkVersion(std::move(sdkVersion_)),
          inputs(std::move(inputs_)),
          pluginData(std::move(data)) {}

    std::string name;         ///< Name of the plugin
    std::vector<std::string> typeNames;         ///< Type of the plugin
    std::string version;      ///< Version of the plugin
    std::string sdkVersion;   ///< SDK version used by the plugin
    Json inputs;              ///< Inputs given to the plugin
    Json pluginData;          ///< Setup output from the plugin
  };

  /// Gets the singleton RunLog instance
  static RunInfo& getInstance() {
    static RunInfo instance;
    return instance;
  }

  /// Const accessor for command line arguments to the run
  [[nodiscard]] const std::vector<std::string>& commandLineArguments() const {
    return clArgs_;
  }

  /// Non-const accessor for command line arguments to the run
  std::vector<std::string>& commandLineArguments() { return clArgs_; }

  /// Adds a command line argument to the list
  void addCommandLineArgument(const char* arg) { clArgs_.emplace_back(arg); }

  /// Const accessor for the list of plugins loaded by the simulation
  [[nodiscard]] const std::vector<PluginInfo>& plugins() const {
    return plugins_;
  }

  /// Non-const accessor for the list of plugins loaded by the simulation
  std::vector<PluginInfo>& plugins() { return plugins_; }

  /// Gets the index of the plugin in the plugin list
  bool indexForPlugin(const Plugin& plugin, size_t& index) const {
    auto it = indexForPlugin_.find(&plugin);
    if (it != indexForPlugin_.end()) {
      index = it->second;
      return true;
    }
    return false;
  }

  /// Adds a plugin to the list
  void addPlugin(const Plugin& plugin) {
    std::string versionStr;
    auto VersionToString = [](const d5::Version v) {
      return std::to_string(v.majorID) + "." + std::to_string(v.minorID);
    };

    plugins_.emplace_back(
        plugin.getName(), plugin.getPluginTypeNames(),
        VersionToString(plugin.getVersion()),
        VersionToString(plugin.getSdkVersion()),
        plugin.getInputAsJson(), plugin.getSetupData());
    indexForPlugin_[&plugin] = plugins_.size() - 1;
  }

  /// Gets the plugin info for a plugin, throws if not there
  PluginInfo& pluginInfo(const Plugin& plugin) {
    size_t idx;
    if (indexForPlugin(plugin, idx)) return plugins_[idx];
    throw std::out_of_range(
        "The requested plugin has not been added to the RunInfo yet");
  }

  /// Const accessor for the list of scenes loaded by the run
  [[nodiscard]] const std::vector<SceneInfo>& scenes() const { return scenes_; }

  /// Non-const accessor for the list of scenes loaded by the run
  std::vector<SceneInfo>& scenes() { return scenes_; }

  /** Adds a scene to the list
   *
   *  \param fname  The filename of the scene
   *  \param md5    MD5 hash of the scene
   *  \param lla    Lat/lon/altitude of the scene anchor point
   *  \param bmin   Minimum point of the scene's bounding box in scene LCS
   *  \param bmax   Maximum point of the scene's bounding box in scene LCS
   */
  void addScene(
      const std::string& fname,
      const std::string& md5,
      const Point& lla,
      const Vectorf& bmin,
      const Vectorf& bmax) {
    scenes_.emplace_back(fname, md5, lla, bmin, bmax);
  }

 private:
  RunInfo() = default;

 private:
  std::vector<std::string> clArgs_;
  std::vector<PluginInfo> plugins_;
  std::vector<SceneInfo> scenes_;
  std::unordered_map<const Plugin*, size_t> indexForPlugin_;
};

/// Convenience function for getting access to the singleton RunInfo instance
inline RunInfo& GetRunInfo() { return RunInfo::getInstance(); }

// JSON conversions

D5_TO_JSON(RunInfo::PluginInfo, j, pluginInfo) {
  j = Json{{"name", pluginInfo.name},
    {"sdk_version", pluginInfo.sdkVersion},
    {"type_names", pluginInfo.typeNames},
    {"version", pluginInfo.version},
    {"inputs", pluginInfo.inputs}};
  if (!pluginInfo.pluginData.empty()) {
    j["plugin_data"] = pluginInfo.pluginData;
  }
}

D5_FROM_JSON(RunInfo::PluginInfo, j, pluginInfo) {
  j.at("name").get_to(pluginInfo.name);
  j.at("sdk_version").get_to(pluginInfo.sdkVersion);
  j.at("type_names").get_to(pluginInfo.typeNames);
  j.at("version").get_to(pluginInfo.version);
  pluginInfo.inputs = j.at("inputs");
  auto it = j.find("plugin_data");
  if (it != j.end()) {
    pluginInfo.pluginData = it.value();
  }
}

D5_TO_JSON(RunInfo::SceneInfo, j, sceneInfo) {
  j = Json{
      {"filename", sceneInfo.filename},
      {"md5", sceneInfo.md5Hash},
      {"geolocation", sceneInfo.latLonAlt},
      {"bounding_box",
       {{"minimum", sceneInfo.bboxMin}, {"maximum", sceneInfo.bboxMax}}}};
}

D5_FROM_JSON(RunInfo::SceneInfo, j, sceneInfo) {
  j.at("filename").get_to(sceneInfo.filename);
  j.at("md5").get_to(sceneInfo.md5Hash);
  j.at("geolocation").get_to(sceneInfo.latLonAlt);
  j.at("bounding_box").at("minimum").get_to(sceneInfo.bboxMin);
  j.at("bounding_box").at("maximum").get_to(sceneInfo.bboxMax);
}

D5_TO_JSON(RunInfo, j, runInfo) {
  static const std::string kVersion = "5.1.0";
  j = Json{
      {"version", kVersion},
      {"run_info", {{"arguments", runInfo.commandLineArguments()}}},
      {"plugin_list", runInfo.plugins()},
      {"scene_list", runInfo.scenes()}};
}

D5_FROM_JSON(RunInfo, j, runInfo) {
  j.at("run_info").at("arguments").get_to(runInfo.commandLineArguments());
  j.at("plugin_list").get_to(runInfo.plugins());
  j.at("scene_list").get_to(runInfo.scenes());
}

/// \}
}
