/** Interface file for RunLog
 *
 *  \file RunLog.h
 */

#pragma once

#include <memory>
#include <string>
#include <vector>

#include "dirsig5/apis/PlatformInfo.h"
#include "dirsig5/apis/RunInfo.h"
#include "dirsig5/apis/TimeWindow.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Json.h"

namespace d5::apis {

/// \addtogroup apis
/// \{

/// Contains useful output information and log information about a run
class RunLog {
public:
  /// Statistics about the run
  struct RunStats {
    DateTime referenceDateTime;
    DateTime startTime;
    std::string runInfoFilename;
    double setupTimeS = 0;
    double computeTimeS = 0;
  };

  /// Information about exoatmospheric sources
  struct ExoSourceInfo {
    std::string name;
    float zenithDeg;
    float azimuthDeg;
  };

  /// Information about captures
  struct CaptureInfo {
    size_t pluginIndex = 0;
    size_t taskIndex = 0;
    size_t captureIndex = 0;
    std::string description;
    TimeWindow relativeTimeWindow;
    PlatformInfo platformInfo;
    std::vector<ExoSourceInfo> exoSourceInfo;
    Json pluginData;
    double computeTimeS = 0;
  };

  /// Gets the singleton RunLog instance
  static RunLog &getInstance() {
    static RunLog instance;
    return instance;
  }

  /// Gets access to the run statistics
  [[nodiscard]] const RunStats &runStats() const { return runStats_; }
  RunStats &mutableRunStats() { return runStats_; }

  /// Gets access to the captures list
  [[nodiscard]] const std::vector<CaptureInfo> &captures() const {
    return captures_;
  }

  /// Gets mutable access to the captures list
  std::vector<CaptureInfo> &mutableCaptures() { return captures_; }

  /// Gets mutable access to a single capture information
  CaptureInfo &mutableCapture(size_t index) {
    if (captures_.size() <= index)
      captures_.resize(index + 1);
    return captures_[index];
  }

  /// Whether we include the run info in the log
  [[nodiscard]] bool includeRunInfo() const { return includeRunInfo_; }

  /// Sets whether to include the run info in the log
  void setIncludeRunInfo(bool inc) { includeRunInfo_ = inc; }

private:
  RunLog() : runStats_() { runStats_.startTime = DateTime::now(); }

private:
  RunStats runStats_;
  std::vector<CaptureInfo> captures_;
  bool includeRunInfo_ = false;
};

/// Convenience function to get the singleton run log instance
inline RunLog &GetRunLog() { return RunLog::getInstance(); }

D5_TO_JSON(RunLog::RunStats, j, r) {
  j = Json{{"start_time", r.startTime.toIso8601()},
           {"setup_time_s", r.setupTimeS},
           {"compute_time_s", r.computeTimeS}};
  if (r.referenceDateTime.isValid())
    j["reference_date_time"] = r.referenceDateTime.toIso8601();
  if (!r.runInfoFilename.empty())
    j["run_info_filename"] = r.runInfoFilename;
}

D5_FROM_JSON(RunLog::RunStats, j, r) {
  std::string startTime;
  j.at("start_time").get_to(startTime);
  r.startTime = DateTime::fromIso8601(startTime);
  j.at("setup_time_s").get_to(r.setupTimeS);
  j.at("compute_time_s").get_to(r.computeTimeS);

  auto it = j.find("run_info_filename");
  if (it != j.end())
    it.value().get_to(r.runInfoFilename);
}

D5_TO_JSON(RunLog::ExoSourceInfo, j, i) {
  j = Json{{"name", i.name},
           {"zenith_deg", i.zenithDeg},
           {"azimuth_deg", i.azimuthDeg}};
}

D5_FROM_JSON(RunLog::ExoSourceInfo, j, i) {
  j.at("name").get_to(i.name);
  j.at("zenith_deg").get_to(i.zenithDeg);
  j.at("azimuth_deg").get_to(i.azimuthDeg);
}

D5_TO_JSON(RunLog::CaptureInfo, j, c) {
  j = Json{{"plugin_index", c.pluginIndex},
           {"task_index", c.taskIndex},
           {"capture_index", c.captureIndex},
           {"description", c.description},
           {"relative_time_window", c.relativeTimeWindow},
           {"platform_info", c.platformInfo},
           {"exo_source_info", c.exoSourceInfo},
           {"plugin_data", c.pluginData},
           {"compute_time_s", c.computeTimeS}};
}

D5_FROM_JSON(RunLog::CaptureInfo, j, c) {
  j.at("plugin").get_to(c.pluginIndex);
  j.at("description").get_to(c.description);
  j.at("relative_time_window").get_to(c.relativeTimeWindow);
  j.at("platform_info").get_to(c.platformInfo);
  j.at("exo_source_info").get_to(c.exoSourceInfo);
  c.pluginData = j.at("plugin_data");
  j.at("compute_time_s").get_to(c.computeTimeS);
}

D5_TO_JSON(RunLog, j, runLog) {
  j = Json{{"run_info", runLog.runStats()},
           {"capture_list", runLog.captures()}};
  if (runLog.includeRunInfo())
    j["run_info"].merge_patch(GetRunInfo());
}

D5_FROM_JSON(RunLog, j, runLog) {
  j.at("run_info").get_to(runLog.mutableRunStats());
  j.at("capture_list").get_to(runLog.mutableCaptures());
  runLog.setIncludeRunInfo(j.at("run_info").find("run_info_filename") ==
                           j.at("run_info").end());
}

/// \}
} // namespace d5::apis
