/** Interface file for the scene tag API
 *
 *  \file SceneTagInterface.h
 */

#pragma once

#include <string>
#include <vector>

#include "dirsig5/base/Geometry.h"

namespace d5::apis {

/** The scene tag interface is responsible for broadcasting any tags associated
 *  with user scenes. User scenes are referenced by their (0 based) index, where
 *  the first, or primary, scene is usually the first one listed in the jsim.
 *  The interface also allows a subset of scenes to be enabled (by a list
 *  of tags) while all other scenes are disabled (this is useful for dynamically
 *  choosing a subset of overlapping scenes).
 *
 *  \ingroup apis
 */
class SceneTagInterface {
public:
  virtual ~SceneTagInterface() = default;

  /** Request information about what tags are attached to each user
   *  scene. They are provided in index order with each index entry
   *  corresponding to a vector of tags that were associated with
   *  that scene. Tag lists for a given scene may be empty.
   */
  virtual void
  getSceneTags(std::vector<std::vector<std::string>> &sceneTags) = 0;

  /** Set a tag filter. Only user scenes that have one or more of the given
   *  tags will be enabled in the simulation. All other user scenes will
   *  be turned off.
   */
  virtual void setSceneTagFilter(const std::vector<std::string> &tags) = 0;

  /** Add an additional position offset to any scene that is setup relative to
   *  another. In other words, the provided point is added to any pre-existing
   *  scene offset and cannot be added to a "primary" scene. The scene being
   *  offset is referenced by it's tag. Note that additional scene offsets
   *  added in this way are not cumulative (i.e. each call of this function
   *  for a particular scene sets a new, additional to the original position).
   *  If more than one scene have the same, provided tag then all those
   *  scenes will have the additional offset.
   */
  virtual void setAdditionalSceneOffset(const std::string &tag,
                                        const Point &offset) = 0;
};
} // namespace d5::apis
