/** Interface file for the scheduler API
 *
 *  \file SchedulerModule.h
 */

#pragma once

#include <functional>
#include <memory>
#include <vector>

#include "dirsig5/apis/plugins/SensorInterface.h"
#include "dirsig5/base/IntID.h"

namespace d5::apis {

/// \addtogroup apis
/// \{

/// A simple struct describing a capture.
struct Capture {
  /// The index of the sensor responsible for this capture
  std::size_t sensorIndex = 0;

  /// The (optional) starting problem for this capture
  std::size_t problemStartIndex = 0;

  /// A flag that indicates if this capture starts a task
  bool startsTask = false;

  /// A flag that indicates if this capture completes a task
  bool completesTask = false;

  /// The input parameters for the capture for the plugin
  plugins::SensorNextCapture::In cpin;

  /// The output parameters for the capture from the plugin
  plugins::SensorNextCapture::Out cpout;
};

/// Simple struct encompassing run ranges.
struct Ranges {
  std::size_t eventIndex = 0;
  std::size_t problemStart = 0;
  std::size_t problemCount = 0;
};

/// Info shared between module calls
struct SchedulerInfo {
  IntID radBuffID = INVALID_ID;
  IntID geoBuffID = INVALID_ID;
  IntID truthBuffID = INVALID_ID;

  // buffers (geo is [lat,lon,alt] order)
  double *buffer, *geoBuffer, *truthBuffer;
};

/**
 * Signature for the various callbacks
 */
using InitSignature = std::function<bool(std::size_t)>;
using QueueSignature = std::function<void()>;
using ProblemSignature = std::function<void(Ranges, SchedulerInfo &)>;
using ProcessSignature =
    std::function<void(Ranges, std::size_t, std::size_t, SchedulerInfo &)>;
using CaptureStartSignature = std::function<void(std::size_t)>;
using CaptureEndSignature = std::function<void(std::size_t)>;

/**
 * Modules are specific compile-time options for how a particular component
 * of the code is implemented. They are not intended for run-time linking
 * (i.e. they are not plugins).
 *
 * The scheduler module is responsible for defining how captures are processed
 * by the core code. It is separated from the internal Scheduler object to
 * enable alternate implemntations that are tied to site-specific
 * capabilities. For example, the module can be used to setup an MPI
 * implementation to process a capture queues across an HPC environment.
 */
class SchedulerModule {
public:
  SchedulerModule(const std::vector<plugins::SensorInterface *> &_sensors,
                  const std::vector<unsigned> &_wavelengthCounts,
                  std::vector<Capture> &_events)
      : sensors(_sensors), wavelengthCounts(_wavelengthCounts),
        events(_events) {}

  virtual ~SchedulerModule() = default;

  /// Setup the scheduling module and use the callback to setup the
  /// internal scheduler at each node.
  virtual bool init(InitSignature initCallback) = 0;

  /// Generate a queue for each node
  virtual void generateQueue(QueueSignature queueCallback) = 0;

  /// Generate problems across nodes and process them
  virtual void runProblems(ProblemSignature problemCallback,
                           ProcessSignature processCallback,
                           CaptureStartSignature captureStartCallback,
                           CaptureEndSignature captureEndCallback) = 0;

  /** direct, const access to simulation storage */
  const std::vector<plugins::SensorInterface *> &sensors;
  const std::vector<unsigned> &wavelengthCounts;
  std::vector<Capture> &events;
};

class SchedulerModuleFactory {
public:
  SchedulerModuleFactory() = default;
  virtual ~SchedulerModuleFactory() = default;

  virtual std::unique_ptr<SchedulerModule>
  Create(const std::vector<plugins::SensorInterface *> &_sensors,
         const std::vector<unsigned> &_wavelengthCounts,
         std::vector<Capture> &_events) = 0;
};

template <typename T>
class SimpleSchedulerModuleFactory : public SchedulerModuleFactory {
  static_assert(std::is_base_of_v<SchedulerModule, T>,
                "T must be a subclass of SchedulerModule");

public:
  SimpleSchedulerModuleFactory() = default;
  ~SimpleSchedulerModuleFactory() override = default;

  std::unique_ptr<SchedulerModule>
  Create(const std::vector<plugins::SensorInterface *> &sensors,
         const std::vector<unsigned> &wavelengthCounts,
         std::vector<Capture> &events) override {
    return std::make_unique<T>(sensors, wavelengthCounts, events);
  }
};

/// \}
} // namespace d5::apis
