/** Interface file for DIRSIG's concept of a surface
 *
 *  \file SurfaceModel.h
 */

#pragma once

#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/IntID.h"
#include "dirsig5/base/Pcg.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/State.h"

namespace d5::apis {

/// \addtogroup apis
/// \{

/// Utility struct for subclasses to use to store broadband stats
struct BroadbandStats {
  double solarAverage = 0;   ///< Average in the solar bandpass
  double thermalAverage = 0; ///< Average in the thermal bandpass
};

/// Abstract struct for describing the behavior of a surface.
struct SurfaceModel {
  enum class Mode { BRDF, BTDF };

  SurfaceModel(bool isDelta = false) noexcept : isDelta(isDelta) {}

  virtual ~SurfaceModel() = default;

  void setScatterMode(Mode m) { mode = m; }

  /// Set a backside medium (called internally)
  void setBacksideMedium(IntID m) { mediumID = m; }

  /// Check for a backside medium
  bool hasBacksideMedium(IntID &m) const {
    return (m = mediumID) != INVALID_ID;
  }

  /// Update the state and return the total scattering
  virtual void updateState(const State *state) = 0;

  /// Get the total (integrated) scattering based on a local incident
  /// direction at the reference wavelength.
  ///
  virtual float getTotal(const Vectorf &localIncident,
                         const IntIDStack *mediumStack) = 0;

  /// Get the spectral total (integrated) scattering based on a local
  /// incident direction.
  ///
  virtual void getSpectralTotal(const Vectorf &localIncident,
                                const IntIDStack *mediumStack,
                                SpectralVector &total) = 0;

  /// Get the average value of the function hemispherically and over the
  /// solar bandpass.
  ///
  [[nodiscard]] virtual double
  getHemisphericalBroadbandSolarAverage() const = 0;

  /// Get the average value of the function hemispherically and over the
  /// thermal bandpass.
  ///
  [[nodiscard]] virtual double
  getHemisphericalBroadbandThermalAverage() const = 0;

  /// Get the differential value of the function for a pair
  /// of incident and exitent directions for the current
  /// spectral state.
  ///
  /// The optional acceptance angle (given by its cosine) should
  /// only be used for testing a very tight lobe -- it is intended
  /// for "delta" like properties being queried for a non-point source.
  /// If the acceptance angle is not explicitly set, it will default to 1.
  ///
  virtual void getScatteringValue(const Vectorf &localIncident,
                                  const Vectorf &localExitant,
                                  const IntIDStack *mediumStack,
                                  SpectralVector &scattering,
                                  float acceptanceMu) = 0;

  /// Get a sampled vector from the function where the probabilty
  /// of the sample is proportional to the value of the function.
  /// The returned spectral weights should be used to scale
  /// the result as part of a MC integral.
  ///
  virtual void sample(RandEngine &randEngine, const Vectorf &localExitant,
                      const IntIDStack *mediumStack, Vectorf &localIncident,
                      SpectralVector &weights) = 0;

  const bool isDelta = false;

  /// The scattering mode this model should be used for.
  Mode mode = Mode::BRDF;

  /// Backside medium id (used internally)
  IntID mediumID = INVALID_ID;
};

/// \}
} // namespace d5::apis
