#pragma once

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/apis/plugins/SensorInterface.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/IntID.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/State.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_atm Atmosphere Plugin
 *  Holds the API definition for an atmopsphere plugin
 *  \ingroup plugin_apis
 */

/// \addtogroup plugin_apis_atm
/// \{

/// Built-in atmosphere profiles available to be used
enum ATMPROFILE {
  MLS,   ///< mid-latitude summer
  MLW,   ///< mid-latitude winter
  SAS,   ///< sub-arctic summer
  SAW,   ///< sub-arctic winter
  TROP,  ///< tropical
  USSTD, ///< US "standard" atmosphere
  NONE   ///< default value
};

/// Definition of input/output variables for the atmospheric init.
namespace AtmosphereInit {

struct In {
  /** Each spectral state is defined by a set of wavelengths
   *  (order and range can be arbitrary, but usually defines
   *  a fixed spectral window) and a reference wavelength
   *  used for geometric calculations.
   *
   *  Wavelengths are in [microns]
   */
  const std::vector<SpectralState> *spectralStates = nullptr;

  /** Scene origins in latitude, longitude and altitude. The index
   *  into the vector corresponds to (non-negative) scene indexes
   *  passed vi the query. Latitude and longitude in degrees,
   *  altitude in meters (relative to a wgs84 ellipsoid).
   */
  std::vector<Point> sceneOrigins;
};

struct Out;

} // namespace AtmosphereInit

/** Shared structure for the provided query information about
 *  the path and spectral state.
 */
struct QueryInfo {
  /// The time of the query [s] relative to the current temporal state.
  float t = 0;

  /** Scene index. This value provides information about which scene the ray
   *  coordinate system is based in. If less than zero, the coordinates are in
   *  global (ECEF) coordinates.
   */
  IntID sceneIndex = 0;

  /** The view origin of the path in local/global coordinates. The coordinates
   *  are double-precision.
   */
  Point viewPoint = {0, 0, 0};

  /** The view direction of the path in local/global coordinates (normalized).
   *  The coordinates are double-precision.
   */
  Vector direction = {0, 0, 1};

  /** Segment path length in [m]. This may incoporate bending due to atmospheric
   *  refraction. The path length is always non-negative and a value equal to
   *  the std::numeric_limits infinity value indicates an unconstrained path
   *  (e.g. a free path into space)
   */
  double segmentLength = 0;

  /** Segment index along the path. An index of zero is closest to the primary
   *  ray origin.
   */
  size_t segmentIndex = 0;

  /** Spectral state index being requested. References a set of wavelengths
   *  provided during the setup. Indexing starts with zero for the first
   *  element.
   */
  size_t spectralStateIndex = 0;
};

/** Definition of input/output variables for the atmospheric path radiance
 *  calculations.
 */
namespace AtmospherePathRadiance {

struct In : QueryInfo {};

struct Out {
  /** In-scattered radiance [W/cm^2 sr um] from the described path arriving at
   *  the view point. The spectral vector has been pre-sized to the appropriate
   *  length for the current spectral state. The initial state of the spectral
   *  array is guaranteed to be all zeros.
   */
  SpectralVector radiance;
};

} // namespace AtmospherePathRadiance

/** Definition of input/output variables for the atmospheric path transmission
 *  calculations.
 */
namespace AtmospherePathTransmission {

struct In : QueryInfo {};

struct Out {
  /** Transmission along the given path. Transmission is expected to always be
   *  non-negative and the spectral vector has been pre-sized to the appropriate
   *  length for the current spectral state. The initial state of the spectral
   *  array is guaranteed to be all ones (if there is no attenuation to report
   *  for the path then the function can simply return)
   */
  SpectralVector transmission;
};

} // namespace AtmospherePathTransmission

/** Definition of input/output variables for the atmospheric profile
 *  calculations.
 */
namespace AtmosphereProfiles {

/** The inputs to the atmospheric profiles takes the full radiometric query.
 *  However, the provided profiles can independently provide altitudes that may
 *  not correspond to the exact directions and distances in the query.
 */
struct In : QueryInfo {};

struct Out {
  /** Turbulence Profile.
   *
   *  This describes the so-called structure coefficient of the refractive
   *  index, or Cn2 (Cn squared) value of the atmsophere, as a  function of
   *  height [m]. This is a measure of the optical turbulence and can be
   *  important for LASER propagation calculation.  The units of Cn2 are
   *  [m^(-2/3)]. A example Cn2 profile is provided as default and made be left
   *  unchanged if this information is unknown or irrelevant. In the case of an
   *  incomplete profile, a best guess will be made for the remainder of the
   *  atmosphere.
   */
  std::vector<double> turbulenceCn2Profile;

  /** Temperature Profile.
   *
   *  This describes the temperature as a function of altitude in degrees
   *  [Celsius] and [m], respectively. A typical temperature profile is provided
   *  by default and may be left untouched if this information is unavailable.
   *  Among other things, the temperature is used along with pressure (and a
   *  water vapor concentration estimate) to predict the refractive index
   *  profile of the atmosphere. In the case of an incomplete profile, a best
   *  guess will be made for the remainder of the atmosphere.
   */
  std::vector<double> temperatureCProfile;

  /** Pressure Profile.
   *
   *  This describes the pressure as a function of altitude in [mbars] and [m],
   *  respectively. A typical pressure profile is provided by default and may be
   *  left untouched if this information is unavailable. Among other things, the
   *  pressure is used along with temperature (and a water vapor concentration
   *  estimate) to predict the refractive index profile of the atmosphere. In
   *  the case of an incomplete profile, a best guess will be made for the
   *  remainder of the atmosphere.
   */
  std::vector<double> pressureMBProfile;

  /** Dew point Profile.
   *
   *  This describes the dew point as a function of altitude in degrees
   *  [Celsius] and [m], respectively. A typical dew point profile is provided
   *  by default and may be left untouched if this information is unavailable.
   *  Among other things, the dew point is used along with temperature and
   *  pressure to predict the refractive index profile of the atmosphere.  In
   *  the case of an incomplete profile, a best guess will be made for the
   *  remainder of the atmosphere.
   */
  std::vector<double> dewPointCProfile;

  /** Altitudes corresponding to the profiles [m].
   *
   *  The altitude is expected to be non-negative. All provided profiles should
   *  be relative to this data set.
   */
  std::vector<double> altitudeM;

  /** As an alternative to providing data for the profiles, the atmosphere
   *  plugin can choose one of the enumerated built-in profiles
   *  (MLS,MLW,SAS,SAW,TRPO,USSTD). The default value is NONE, which either uses
   *  the provided data, or no profile if no data is provided.
   */
  ATMPROFILE builtInProfiles = NONE;
};

} // namespace AtmosphereProfiles

/** Definition of input/output variables for the
 *  getExoatmosphericSourceIrradiance interface
 */
namespace ExoatmosphericSourceIrradiance {

struct In : QueryInfo {
  /** Index of the source being queried. Corresponds to the exoatmospheric
   * source list provided during init.
   */
  size_t sourceIndex = 0;
};

struct Out {
  /** Spectral irradiance in [W/cm^2 um] for the given source over the distance
   *  to the query point, not considering any transmission or projection
   *  effects. The number of entries is equal to the number of wavelength
   *  samples in the spectral state.
   */
  SpectralVector sourceIrradiance;
};

} // namespace ExoatmosphericSourceIrradiance

/** The atmosphere model in DIRSIG5 encompasses the radiometric impact of
 *  media above the Earth up to the Kalman Line, or 100km above sea
 *  level. Beyond that point, we assume that we are traveling in "space"
 *  where in-scattered path radiometry is negligible.
 *
 *  The primary role of the atmosphere plugin is to provide in-scattered
 *  radiance along a path towards the origin (point) of a request, or
 *  the extinction long a similar path, expressed as the total
 *  transmission. The geometric format of requests follows a: point,
 *  vector, path length format, where the length can often be infinite
 *  (usually the result of interacting with the open sky). Note that
 *  any bending of the path is already taken into account internally
 *  and subsequent path differences are already in the length, if
 *  appropriate.
 *
 *  While not an inherent property of the atmosphere, exoatmospheric
 *  sources drive the energy available for scattering and other
 *  physical processes. An exoatmospheric source (eg. sun, moon)
 *  is handled slightly differently than other sources since they
 *  are assumed to contribute a fairly constant irradiance to the
 *  scene (before and after transmission through the atmosphere).
 *  This means that they are modeled as uni-directional sources
 *  despite the source itself usually emitting in all directions
 *  (i.e. the angular variation is negligible over the solid angle
 *  connecting the source and scene). This means that the source
 *  is fully defined by a pointing vector, the solid angle (assumed
 *  rotationally symmetric), and a spectral irradiance curve.
 *
 *  The secondary role of the atmosphere plugin is to provide
 *  information about the structure of the atmosphere, as needed
 *  to characterize its properties for other models. This includes
 *  information about the turbulence (in the form of a Cn2 parameter),
 *  the temperature (celsius), the pressure (millibars), the dew
 *  point (celsius), the scattering coefficient (1/km), and a 4
 *  component model of the scattering phase function using the
 *  Henyey-Greenstein model. Note that the scattering components
 *  are defined spectrally for each of the state reference wavelengths
 *  and are only used for geometric calculations (the atmospheric
 *  transmission and radiance are assumed to be spectrally varying
 *  within states however). Other structural components of the
 *  atmosphere include the weather history
 *
 *  Atmospheric properties are global in nature, so it is beneficial
 *  to use a matching coordinate system. ECEF is a convenient
 *  global Cartesian coordinate system, and is used internally for
 *  in-atmosphere ray traces. An underlying Earth ellipsoid of
 *  WGS84 (EPSG:4326) is assumed throughout the code and can be
 *  used to convert ECEF coordinates to other geo systems.
 *
 *  Spectral calculations are done in sets of nearly arbitrary
 *  wavelength samplings, corresponding to "spectral states" which
 *  are provided during init. Note that some of the init parameters
 *  are provided at wavelengths not included in the spectral states
 *  in order to allow for inelastic processes and to fit model expectations.
 *
 *  Any errors produced internally should be thrown after describing the
 *  issue as thoroughly as possible in an error message.
 *
 *  All virtual functions should be implemented, but the static
 *  ::getVersion is used to track the version of this interface.
 *  Plugin functions are implemented using the same interface format,
 *  which provides two custom structs, one for the inputs and one
 *  for the outputs.
 */
class AtmosphereInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("atmosphere");
  }

  /** Generate the in-scattered radiance reaching the origin of the query
   *  along a path coming from the provided direction over a distance.
   *  Can be calculated in scene local or global coordinate spaces
   *  depending on inputs.
   *
   *  The default behavior is to nothing to the path.
   */
  using GetPathRadiance = std::function<Status(
      const AtmospherePathRadiance::In &, AtmospherePathRadiance::Out &)>;

  /** Generate the path transmission along a path coming from the
   *  provided direction over a given distance. Can be calculated in
   *  scene local or global coordinate spaces depending on inputs.
   *
   *  The default behavior is to have a unit transmission.
   */
  using GetPathTransmission =
      std::function<Status(const AtmospherePathTransmission::In &,
                           AtmospherePathTransmission::Out &)>;

  /** Generate the atmospheric properties as a function of altitude
   *  along a (possibly slant) path. Property profiles that are
   *  unavailable should be left empty.
   *
   *  There is currently no specific default behavior for this function.
   */
  using GetProfiles = std::function<Status(const AtmosphereProfiles::In &,
                                           AtmosphereProfiles::Out &)>;

  /** This function should provide the spectral irradiance
   *  onto the scene without consideration of transmission
   *  through an atmosphere (handled elsewhere). The spectral
   *  state is provided as an index into the states (see init).
   *  A source index is provided that corresponds to the number
   *  of exoatmospheric sources returned during init.
   *
   *  \note This function will only be requested once per scene
   *  per state and its assumed that the source-Earth distance does
   *  not change signifcantly within the simulation.
   *
   *  \note Variations in the source itself should
   *  be built into this value (e.g. the phase of the moon)
   *
   *  The default behavior is to set the irradiance to zero.
   */
  using GetExoatmosphericSourceIrradiance =
      std::function<Status(const ExoatmosphericSourceIrradiance::In &,
                           ExoatmosphericSourceIrradiance::Out &)>;

  /** Initialization
   *
   *  Provides runtime info to the plugin.
   */
  virtual Status init(const AtmosphereInit::In &in,
                      AtmosphereInit::Out &out) = 0;
};

namespace AtmosphereInit {

struct Out {
  /** List of exoatmospheric sources to consider during radiometry
   *  calculations. The string provided is the name of the source
   *  used for logging simulation info and source information
   *  will be requested by the corresponding index into this list.
   */
  std::vector<std::string> exoatmosphericSources;

  /// Callback functions for directly querying the atmosphere.
  ///\{
  AtmosphereInterface::GetPathRadiance getPathRadiance;
  AtmosphereInterface::GetPathTransmission getPathTransmission;
  AtmosphereInterface::GetProfiles getProfiles;
  AtmosphereInterface::GetExoatmosphericSourceIrradiance
      getExoatmosphericSourceIrradiance;
  ///\}
};
} // namespace AtmosphereInit

/// \}
} // namespace d5::apis::plugins
