#pragma once

#include <functional>

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/SpectralVector.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_bgs  Background Sky Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for a background sky plugin
 */

/// \addtogroup plugin_apis_bgs
/// \{

/// Definition of input/output variables for the background sky init.
namespace BackgroundSkyInit {

struct In {
  /** Each spectral state is defined by a set of wavelengths (order and range
   *  can be arbitrary, but usually defines a fixed spectral window) and a
   *  reference wavelength used for geometric calculations.
   *
   *  Wavelengths are in [microns]
   */
  const std::vector<SpectralState> *spectralStates = nullptr;

  /// Scene origins in latitude, longitude and altitude. The index
  /// into the vector corresponds to (non-negative) scene indexes
  /// passed via the query. Latitude and longitude in degrees,
  /// altitude in meters (relative to a wgs84 ellipsoid).
  std::vector<Point> sceneOrigins;
};

struct Out;

} // namespace BackgroundSkyInit

/** Definition of input/output variables for the background sky path radiance
 *  calculations.
 */
namespace BackgroundSkyPathRadiance {

struct In {
  /// The time of the query [s] relative to the current temporal state.
  float t = 0;

  /** The scene index indicates which scene the ray coordinate system is
   *  based in. If less than zero, the coordinates are in global (ECEF)
   *  coordinates.
   */
  IntID sceneIndex = 0;

  /** The view origin of the path in local/global coordinates. The coordinates
   *  are double-precision.
   */
  Point viewPoint = Point(0, 0, 0);

  /** The view direction of the path in local/global coordinates (normalized).
   *  The coordinates are double-precision.
   */
  Vector direction = Vector(0, 0, 1);
};

struct Out {
  /** In-scattered radiance [W/cm^2 sr um] from the described path arriving at
   *  the view point. The spectral vector has been pre-sized to the appropriate
   *  length for the current spectral state. The initial state of the spectral
   *  array is guaranteed to be all zeros.
   */
  SpectralVector radiance;
};

} // namespace BackgroundSkyPathRadiance

/** DIRSIG's atmospheric plugins can specify path radiances and transmissions,
 *  however, there are a number of illumination sources they can struggle to
 *  represent. Light pollution and starlight are two prime examples of such
 *  sources. These effects may be easier to express through a simple additive
 *  background illumination model, which is where the background sky plugin
 *  comes in.
 *
 *  The primary fole of the background sky plugin is to provide the in scatted
 *  radiance along a path towards the origin (point) of a request. By the nature
 *  of this plugin, these paths must be sky-reaching paths. No transmission
 *  effects are accounted for in this plugin, as those will be handled by the
 *  atmosphere plugin.
 */
class BackgroundSkyInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("backgroundsky");
  }

  /** Generate the in-scattered radiance reaching the origin of the query
   *  along a path coming from the provided direction.
   *
   *  The default behavior is to nothing to the path.
   */
  using GetPathRadiance = std::function<Status(
      const BackgroundSkyPathRadiance::In &, BackgroundSkyPathRadiance::Out &)>;

  /** Initialization
   *
   *  Provides runtime info to the plugin.
   */
  virtual Status init(const BackgroundSkyInit::In &in,
                      BackgroundSkyInit::Out &out) = 0;
};

namespace BackgroundSkyInit {

struct Out {
  /// Callback function for directly querying the background sky.
  BackgroundSkyInterface::GetPathRadiance getPathRadiance;
};

} // namespace BackgroundSkyInit

/// \}
} // namespace d5::apis::plugins
