/** Interface file for CloudModelInterface
 *
 *  \file CloudModelInterface.h
 */

#pragma once

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/State.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_cloudmodel Cloud Model Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for an Cloud model plugin
 */

/// \addtogroup plugin_apis_cloudmodel
/// \{

struct CloudModelOut {
  CloudModelOut() = default;
  virtual ~CloudModelOut() = default;
};

struct OpaqueCloudOut : public CloudModelOut {
  Point ecefHit = {0, 0, 0};  ///< ECEF hit point [out]
  Vectorf normal = {0, 0, 1}; ///< Normal vector at the hit [out]
  float scale = 1.f; ///< Optional scaling of the final reflectance [out]
};

/** Transmit ray output
 */
struct TransmitRayOut : public CloudModelOut {
  std::vector<Point> ecefOrigin;     ///< new ECEF origins [out]
  std::vector<Vector> ecefDirection; ///< new ECEF directions [out]
};

struct CloudIntersectInfo {
  CloudIntersectInfo(RandEngine &_randEngine) : randEngine(_randEngine) {}

  Point ecef1;            ///< ECEF startpoint of the path [in]
  Point ecef2;            ///< ECEF endpoint of the path [in]
  Point lla1;             ///< Lat/Lon/Alt startpoint of the path [in]
  Point lla2;             ///< Lat/Lon/Alt endpoint of the path [in]
  RandEngine &randEngine; ///< random engine for current ray
  float time = 0;         ///< Time [in]
  int primaryRCount = -1; /// primary path problem R count [in]
  int primaryTCount = -1; /// primary path problem T count [in]
  bool freePath = true;   /// track if we never encountered a >0 probability

  ///< get the output struct for the specific type
  template <class Type> Type &getOut(void) {
    if (out == nullptr)
      out = new Type;
    return *dynamic_cast<Type *>(out);
  }

  ~CloudIntersectInfo(void) { delete out; }

private:
  CloudModelOut *out = nullptr; ///< output should be cast for given use
};

/// Callback for the Cloud model intersection.
using CloudIntersectCallback = std::function<bool(CloudIntersectInfo &)>;

/// Indicates the way the returned material index should be used
enum CloudMaterialType {
  OpaqueCloud,   // use a nominal cloud reflectance
  TransmittedRay // change original ray to a new origin/direction on exit (todo)
};

struct OpaqueCloudOptions {};

struct TransmitRayOptions {};

union CloudMaterialOptions {
  OpaqueCloudOptions opaqueCloudOptions;
  TransmitRayOptions transmitRayOptions;
};

/// Definition of input/output variables for the Init interface
namespace CloudModelInit {

struct In {};

struct Out {
  /// Intersection callback.
  CloudIntersectCallback intersectCallback;

  /// Type of materials to use (enumerated type)
  CloudMaterialType cloudMaterialType;

  /// Each material system has its own set of options stored
  /// in a union of the available options structs.
  CloudMaterialOptions cloudMaterialOptions;

  /// The minimum and maximum altitude (WGS84) must be supplied
  double minAltitude = 1.0;
  double maxAltitude = -1.0;
};

} // namespace CloudModelInit

class CloudModelInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("cloudmodel");
  }

  /// Initialization
  virtual Status init(const CloudModelInit::In &in,
                      CloudModelInit::Out &out) = 0;
};

/// \}
} // namespace d5::apis::plugins
