/** Interface file for EarthModelInterface
 *
 *  \file EarthModelInterface.h
 */

#pragma once

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/State.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_earthmodel Earth Model Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for an Earth model plugin
 */

/// \addtogroup plugin_apis_earthmodel
/// \{

struct EarthModelOut {
  EarthModelOut() = default;
  virtual ~EarthModelOut() = default;
};

struct IndexedRGBOut : public EarthModelOut {
  Point ecefHit = {0, 0, 0};  ///< ECEF hit point [out]
  Vectorf normal = {0, 0, 1}; ///< Normal vector at the hit [out]
  size_t rgbIndex = 0;        ///< index to use for the hit [out]
};

/** Revise ray output
 *  Note that the rays should be ordered such that the first hit is used
 */
struct ReviseRayOut : public EarthModelOut {
  std::vector<Point> ecefOrigin;     ///< new ECEF origins [out]
  std::vector<Vector> ecefDirection; ///< new ECEF directions [out]
  size_t sceneIndex = 0;             ///< new (user) scene index
};

struct EarthIntersectInfo {
  Point ecef1;    ///< ECEF endpoint of the path [in]
  Point ecef2;    ///< ECEF endpoint of the path [in]
  Point lla1;     ///< Lat/Lon/Alt endpoint of the path [in]
  Point lla2;     ///< Lat/Lon/Alt endpoint of the path [in]
  float time = 0; ///< Time [in]

  ///< get the output struct for the specific type
  template <class Type> Type &getOut(void) {
    if (out == nullptr)
      out = new Type;
    return *dynamic_cast<Type *>(out);
  }

  ~EarthIntersectInfo(void) { delete out; }

private:
  EarthModelOut *out = nullptr; ///< output should be cast for given use
};

/// Callback for the Earth model intersection.
using EarthIntersectCallback = std::function<bool(EarthIntersectInfo &)>;

/// Indicates the way the returned material index should be used
enum EarthMaterialType {
  IndexedRGB, // convert an index to RGB reflectance
  ReviseRay   // change original ray to a new origin/direction and re-hit
};

struct IndexedRGBOptions {
  /// Scale to apply to 8-bit RGB values to get total reflectance.
  float rgbScale;
};

struct ReviseRayOptions {};

union EarthMaterialOptions {
  IndexedRGBOptions indexedRGBOptions;
  ReviseRayOptions reviseRayOptions;
};

/// Definition of input/output variables for the Init interface
namespace EarthModelInit {

struct In {};

struct Out {
  /// Intersection callback.
  EarthIntersectCallback intersectCallback;

  /// Type of materials to use (enumerated type)
  EarthMaterialType earthMaterialType;

  /// Each material system has its own set of options stored
  /// in a union of the available options structs.
  EarthMaterialOptions earthMaterialOptions;
};

} // namespace EarthModelInit

class EarthModelInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("earthmodel");
  }

  /// Initialization
  virtual Status init(const EarthModelInit::In &in,
                      EarthModelInit::Out &out) = 0;
};

/// \}
} // namespace d5::apis::plugins
