/** Interface file for the ephemeris plugin
 *
 *  \file EphemerisInterface.h
 */

#pragma once

#include <string>
#include <vector>

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_ephem Ephemeris Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for an ephemeris plugin
 */

/// \addtogroup plugin_apis_ephem
/// \{

/// Definition of input/output variables for the ephemeris model init.
namespace EphemerisInit {

struct In {
  /** Scene origins in latitude, longitude and altitude. The index into the
   *  vector corresponds to (non-negative) scene indexes passed via the query.
   *  Latitude and longitude in degrees, altitude in meters (relative to a
   *  WGS84 ellipsoid).
   */
  std::vector<Point> sceneOrigins;

  /// The list of source names needed by the simulation.
  std::vector<std::string> sourceNames;
};

struct Out {};

} // namespace EphemerisInit

/// Definition of input/output variables for the getSourceGeometry interface
namespace EphemerisSourceGeometry {

struct In {
  /// The date/time of the query
  DateTime dateTime;

  /** The index of the source being queried. This index corresponds to the
   *  source list provided during init.
   */
  std::size_t sourceIndex = 0;

  /// The flag to have the plugin suppress and update notifications.
  bool suppressNotifications = false;
};

struct Out {
  /// The ECEF location of the source [m]
  Point sourceEcefLocation = Point(0, 0, 0);

  /// The ECEF velocity of the source [m/s]
  Vector sourceEcefVelocity = Vector(0, 0, 0);

  /** The source half angle in radians. This gives the apparent size of the
   *  source assuming that it is rotationally symmetric (i.e. *  the visible
   *  cross-section is always a circle).
   */
  double sourceHalfAngle = 0.0;

  /// The source phase angle in radians.
  double sourcePhaseAngle = 0.0;

  /** The source fraction. This is used for the Moon to indicate the current
   *  phase fraction.
   */
  double sourceFraction = 1.0;

  /** Flag to indicate that the vector represents an apparent location (as seen
   *  from the given query point, but not necessarily the vector to the
   *  celestial location). This should primarily be used if refraction has
   *  already been taken into account.
   *
   *  If false, the apparent location will be computed internally if
   *  appropriate.
   */
  bool isApparentLocation = false;
};
} // namespace EphemerisSourceGeometry

/** Any errors produced internally should be thrown after describing
 *  the issue as thoroughly as possible in an error message.
 *
 *  All virtual functions should be implemented, but the static getVersion
 *  is used to track the version of this interface. Plugin functions are
 *  implemented using the same interface format, which provides two custom
 *  structs, one for the inputs and one for the outputs.
 */
class EphemerisInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("ephemeris");
  }

  /** Initialization
   *
   *  Provides runtime info to the plugin.
   */
  virtual Status init(const EphemerisInit::In &in, EphemerisInit::Out &out) = 0;

  /** Get source geometry.
   *
   *  This function should provide the ECEF location in kilometers of the
   *  requested source.  This function may be called each time the temporal
   *  state of the simulation changes.
   *
   *  The solid angle of the source is requested so that the
   *  source (assumed to have a circular cross-section) can
   *  be sampled appropriately. Note that some source solid angles
   *  can change significantly depending on time of year (e.g.
   *  the apparent moon radius changes by roughly 12% between
   *  perigee and apogee distances).
   */
  virtual Status getSourceGeometry(const EphemerisSourceGeometry::In &in,
                                   EphemerisSourceGeometry::Out &out) = 0;
};

/// \}
} // namespace d5::apis::plugins
