/** Interface file for a geometry plugin
 *
 *  \file GeometryInterface.h
 */

#pragma once

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/State.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_geom Geometry Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for a geometry plugin
 */

/// \addtogroup plugin_apis_geom
/// \{

/// Definition of input/output variables for the Init interface
namespace GeometryInit {

struct In {};

/** If the geometry is defined with a static facetized object, the object can
 *  be returned here (for internally optimized intersections). 
 *
 *  If intersections will be handled by the plugin, these output variables
 *  should be remain as the defaults.
 *
 *  In both cases, the intersect callback should be supplied here so the plugin
 *  can set the appropriate material index or do any necessary modification of
 *  the hit. If the plugin is doing its own intersection the output variables
 *  will not be set and need to be filled in by the callback. If the
 *  intersection is done internally then they will already be set.
 *
 *  The vector of materials are provided so that an internal material can be
 *  constructed and associated with each hit (the material inputs are defined
 *  in PI::Plugin).
 */
struct Out {
  /// Buffer of vertexes (interlaced XYZ triplets)
  std::vector<float>* vertexes = nullptr;

  /// Buffer of vertex index triangles (interlaced index triplets)
  std::vector<std::uint32_t>* facets = nullptr;

  /// Buffer of vertex normals (may point to an empty vector)
  std::vector<float>* vnData = nullptr;

  /// Buffer of vertex normal indexes (may point to an empty vector)
  std::vector<std::uint32_t>* vnIndexes = nullptr;

  /// Flag that indicates geometry is updated during state changes
  bool stateUpdatesModifyGeometry = false;

  /// Bounding box corners (only used if facets are not supplied)
  Pointf bboxMin, bboxMax;

  /// Intersection callback
  IntersectCallback intersectCallback;

  /// Materials corresponding to the intersection index
  std::vector<MaterialInfo*> materials;

  /** List of instances to be added. The first value is the string tag
   *  associated with a scene (defaults to the first user-defined scene) that
   *  supplies the coordinate system. The second is the first three rows of a
   *  4x4 tranformation matrix.
   *
   *  If the plugin is responsible for the base geometry intersection then the
   *  bounding sphere (encompassing the bounding box) is used to define the
   *  region through which rays will trigger an intersection check.
   *
   *  If no instances are provided then an identity transform is assumed in the
   *  first user scene.
   */
  std::vector<std::pair<std::string, Matrixf>> instances;
};

}

/** Interface for geometry definitions. This allows the implementer
 *  to provide geometry through the plugin interface rather than at
 *  at startup (e.g. in a scene hdf). Multiple instances of geometry can
 *  be created and are represented in the simulation as proxy bounding
 *  boxes that trigger the plugin's intersection routing.
 *
 *  Geometry intersections can either be done entirely by the plugin
 *  (particularly useful if the geometry isn't facetized) or a set of
 *  facets that are passed to the internal ray tracer.
 */
class GeometryInterface : virtual public Plugin {
 public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("geometry");
  }

  /** Initialization
   *
   *  Requests details about the geometry being modeled
   *  including whether intersections should be handled internally
   *  (as static geometry) or in the plugin. For the latter case,
   *  an intersection routine can be provided as a callback function
   *
   *  Callbacks are provided as std::function objects
   *  based on the prototypes given for PI::IntersectCallback
   */
  virtual Status init(const GeometryInit::In& in, GeometryInit::Out& out) = 0;
};

/// \}
}
