#pragma once

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/DualQuaternion.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/State.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_laserpulse Laser Pulse Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for a laser pulse plugin
 */

/// \addtogroup plugin_apis_laserpulse
/// \{

struct LaserPulseInfo {
  /** Initial emission time of the source relative to the current
   *  temporal state [s]. This is the leading edge of the temporal
   *  pulse (it will continue to emit for the temporal width, below).
   *  Note that the timing of the pulse should be based on the
   *  significant portion of the pulse (e.g. a the trailing edges
   *  of a gaussian profile should be clipped)
   */
  double emitTime = 0;

  /** The nominal temporal width [s] of the entire pulse -- this should
   *  represent the simulation relevant portion of the pulse and is used
   *  to bound the pulse spatially in the scene.
   */
  double pulseSupport = 0;

  /** Cartesian location of the pulse. If the scene index is non-negative,
   *  the position will be interpreted local to the given scene.
   *  Otherwise it is assumed to be ECEF.
   */
  Point origin = Point(0, 0, 0);

  /** Pointing direction of the laser source at emission. As with the
   *  origin, the direction is in the active coordinate system based
   *  on the scene index.
   */
  Vector pointing = Vector(0, 0, -1);

  /** Velocity vector of the platform at time of release. This is
   *  used to maintain the source reference frame relative to the
   *  Earth. Units are [m/s] and in the same coordinates as the *
   *  origin and pointing data.
   */
  Vector velocity = Vector(0, 0, 0);

  /** The dual quaternion capturing the location and orientation of
   *  the source at emission.
   */
  DualQuaternion toWorld;

  /** Index of the scene that the source origin exists in. If negative,
   *  the origin and direction are interpreted as ECEF points/vectors.
   *  Scene indexes start with zero and correspond to the order they
   *  are added to the simulation.
   */
  int sceneIndex = 0;

  /// Spectral energy of the pulse (sampled for the current spectral state).
  SpectralVector pulseEnergy;

  /** Angular extent of the pulse in radians. This is the angle over
   *  which the pulse is assumed to be significant (i.e. insignificant
   *  tail ends of a gaussian distribution would not be included). It
   *  is assumed that the "shape" of the pulse within the pulse angle
   *  integrates to the total spectral energy if the entire temporal
   *  distribution is incorporated, but it will be normalized even if
   *  it is not. The shape of the pulse is also assumed to be radially
   *  symmetric.
   */
  double pulseMaxAngle = 0;

  /** Optional ratio of the minimum pulse angle to the maximum (i.e. always <1)
   */
  double pulseAngleRatio = 0.;
};

/** Interface for laser pulses, i.e. fixed time emission of energy from
 *  what is effectively a point source into a narrow solid angle. Even
 *  though pulsed sources typically generate many individual pulses
 *  over time, the object generated by this plugin is a single pulse that
 *  exists independently within a scene. Like other plugins, implementation
 *  of this plugin will generate a factory -- however, rather than
 *  generating objects to be used internally, the factory is used
 *  by other plugins via the "requestPlugin" interface.
 *
 *  Pulses are represented by a location/direction (source of the pulse), its
 *  temporal, spectral, and spatial/angular distribution functions.
 *  The plugin itself is lightweight and should be fully
 *  ready after a setup call and a state update. Note that the pulse
 *  is only defined in the spectral/temporal state it was created in.
 *
 *  Laser pulses are assumed to be rotationally symmetric around the
 *  pointing direction.
 */
class LaserPulseInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("laserpulse");
  }

  /** Get information about the pulse as defined in the
   *  corresponding struct. The info will not be requested until
   *  the setup function has been called and a state update has
   *  been sent. Note that a pulse is only valid for a single
   *  spectral state (the current one) and is defined relative
   *  to the current temporal state.
   */
  virtual void getInfo(LaserPulseInfo &info) const = 0;

  /** Get the fraction of the pulse that exists between two times.
   *  Note that this does not include any angular distribution
   *  across the wavefront of the pulse -- it should be the fraction
   *  of the spatially/angularly integrated spectral energy that
   *  exists between two times assuming a constant medium and relative
   *  to the start of the pulse (emitTime). Both times will be less
   *  than the pulseSupport. The fraction between 0 and pulseSupport
   *  should always be 1.0. Times are in seconds.
   */
  [[nodiscard]] virtual double getTimeFraction(double t1, double t2) const = 0;

  /** Get the angular weight relative to the original direction.
   *  When integrated over the entire solid angle of the the source
   *  (out to pulseMaxAngle) the result is expected to be unit valued
   *  (i.e. it is the angular distribution of the pulse energy). Note
   *  that it will be internally normalized even if is not.
   */
  [[nodiscard]] virtual double getAngularWeight(const Vector &vector) const = 0;
};

/// \}
} // namespace d5::apis::plugins
