/** Interface file for a material plugin
 *
 *  \file MaterialInterface.h
 */

#pragma once

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/apis/SurfaceModel.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/State.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_mat Material Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for a material plugin
 */

/// \addtogroup plugin_apis_mat
/// \{

/// Definition of input/output variables for the init interface
namespace MaterialInit {

struct In {};

struct Out {
  /** Reflectance object for this material. The pointer provided
   *  is required to be non-zero. This pointer will not be freed
   *  and it is expected that this plugin will handle any memory
   *  associated with it during teardown.
   */
  SurfaceModel *reflectanceModel = nullptr;

  /** Transmittance object for this material. If the returned
   *  pointer is null then no transmittance properties will be
   *  added. This pointer will not be freed and it is expected
   *  that this plugin will handle any memory associated with it
   *  during teardown.
   */
  SurfaceModel *transmittanceModel = nullptr;
};

} // namespace MaterialInit

/** This class defines an interface to the Material plugin. Plugins
 *  that implement this interface are responsible for providing a BRDF
 *  and, optionally, a BTDF for the material being defined. Both the BRDF
 *  an BTDF models inherit from the SurfaceModel class and the
 *  provided implementations must implement all virtual functions.
 *
 *  The important functions in the plugin interface itself follow a simple
 *  format using two structs (In/Out) as parameters. These structs have the
 *  same names between functions, but are defined within function
 *  specific namespaces (for instance, the init In/Out structs are defined
 *  within the MaterialInterfaceInit namespace).
 *
 *  The plugin functions are mostly concerned with coordinating interactions
 *  with the calling simulation. As such, there are three primary functions
 *  that must be implemented: "setup", "init", and "updateState". An
 *  optional "teardown" notification is provided as well, in case any
 *  final calculations/output needs to be done before the plugin is unloaded.
 *
 *  The "setup" function is called once at the start of the simulation
 *  and its primary purpose is to provide runtime inputs to the plugin
 *  (in the form of an input stream so no file i/o needs to be performed
 *  in the plugin itself).
 *
 *  The "init" function is called once the simulation is ready to run and
 *  the plugin returns pointers for the reflectane/transmittance objects.
 *
 *  Note that the Material plugin does not get a state change notification,
 *  this is expected to be handled by the provided SurfaceModel objects.
 */
class MaterialInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("material");
  }

  /** Initialization
   *
   *  Requests the individual models for the reflectance and
   *  (optionally) the transmittance.
   */
  virtual Status init(const MaterialInit::In &in, MaterialInit::Out &out) = 0;
};

/// \}
} // namespace d5::apis::plugins
