#pragma once

#include <functional>

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/State.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_mediop Medium IOP Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for a medium Inherent Optical Properties (IOP)
 *  plugin
 */

/// \addtogroup plugin_apis_mediop
/// \{

/** Callback function prototype for querying the absorption coefficient
 *  [1/m] at the currrent state's reference wavelength.
 *
 *  The time passed to this and the other callbacks refers to the elapsed
 *  time in [s] since the start of the current temporal state.
 */
using AbsCoeffCallback =
    std::function<float(const Pointf &location, float time)>;

/** Callback function prototype for querying the scattering coefficient
 *  [1/m] at the currrent state's reference wavelength.
 *
 *  The time passed to this and the other callbacks refers to the elapsed
 *  time in [s] since the start of the current temporal state.
 */
using SctCoeffCallback =
    std::function<float(const Pointf &location, float time)>;

/** Callback function prototype for querying the spectral absorption
 *  coefficients [1/m] in the current state. The result should be assigned
 *  to the passed spectral vector, which is not pre-sized.
 *
 *  If the absorption coefficient is zero or negligible then the return
 *  value should be false. In that case the output spectral vector
 *  does not need to be set/sized. Otherwise, return true with a valid
 *  spectral vector passed.
 */
using SpectralAbsCoeffCallback = std::function<bool(
    const Pointf &location, float time, SpectralVector &abs)>;

/** Callback function prototype for querying the spectral scattering
 *  coefficients [1/m] in the current state. The result should be assigned
 *  to the passed spectral vector, which is not pre-sized.
 *
 *  If the scattering coefficient is zero or negligible then the return
 *  value should be false. In that case the output spectral vector
 *  does not need to be set/sized. Otherwise, return true with a valid
 *  spectral vector passed.
 */
using SpectralSctCoeffCallback = std::function<bool(
    const Pointf &location, float time, SpectralVector &sct)>;

/** Callback function prototype for querying the spectral (real) index of
 *  refraction for a particular location/time. A flag indicates if this
 *  position is just under the surface (e.g. querying a wave for a
 *  refraction angle calculation) versus any point in the medium.
 *  The spectral vector provided has not been pre-sized.
 */
using IORCallback =
    std::function<void(const Pointf &location, float time, bool atInterface,
                       SpectralVector &ior_n, SpectralVector &ior_k)>;

/** Callback function prototype for querying the temperture at a
 *  particular location/time. Returned value is in Kelvin.
 */
using TempKCallback = std::function<float(const Pointf &location, float time)>;

/** Callback function prototype for querying the scattering phase function
 *  probability density function for a particular location/time. The given
 *  angle from the forward direction (expressed as the cosine
 *  of that angle) should be used under an assumption of rotational
 *  symmetry. The PDF is also assumed to be constant across wavelengths
 *  within a spectral state (multiple states can be used if this assumption
 *  is false).
 */
using PdfCallback =
    std::function<float(const Pointf &location, float time, float mu)>;

/** Callback function prototype for sampling the scattering phase function
 *  probability density function for a particular location/time. The
 *  returned angle from the forward direction (expressed as the cosine
 *  of that angle) will be used under an assumption of rotational
 *  symmetry. The PDF is also assumed to be constant across wavelengths
 *  within a spectral state (multiple states can be used if this assumption
 *  is false).
 */
using SamplePdfCallback =
    std::function<float(const Pointf &location, float time)>;

/// Definition of input/output variables for the init interface
namespace MediumIOPInit {

struct In {
  /** Inherent optical properties are tied to a medium bound
   *  by an axis-aligned box in the local scene coordinate
   *  system. When queries are made using the returned callbacks
   *  (below) a position will be given in the query that can
   *  be localized to the medium using this box.
   */
  /// \{
  Pointf mediumMinCorner = {0, 0, 0};
  Pointf mediumMaxCorner = {0, 0, 0};
  /// \}
};

struct Out {
  Out() {
    tempKCallback = [](const Pointf &, float) { return 0.f; };
  }

  /** Callbacks for individual IOPs (see documentation for
   *  function prototypes).
   */
  /// \{
  AbsCoeffCallback absCallback;
  SctCoeffCallback sctCallback;
  SpectralAbsCoeffCallback spectralAbsCallback;
  SpectralSctCoeffCallback spectralSctCallback;
  IORCallback iorCallback;
  PdfCallback pdfCallback;
  TempKCallback tempKCallback;
  SamplePdfCallback samplePdfCallback;
  /// \}

  /** Flag indicating if the IOPs are constant throughout the
   *  volume described by this model. This greatly reduces
   *  runtime as a large number of values can be cached.
   */
  bool hasConstantIOPs = false;

  /** Flag indicating that volumetric calculations should be skipped
   *  and the medium should be treated as a black/gray body. If there
   *  is an active boundary interface then the local refractive index
   *  callback will be used to drive the emissivity and the temperature
   *  callback will drive the blackbody emission. All other callbacks
   *  will be ignored and can be undefined.
   */
  bool emissionOnly = false;
};
} // namespace MediumIOPInit

/** This class defines an interface to the Medium IOP plugin. Plugins
 *  that implement this interface are responsible for providing inherent
 *  optical properties (spectral scattering and absorption) at any
 *  location within the associated medium and at any time. A medium is
 *  defined as any volume within the simulation that represents a change
 *  in radiative transfer methods or properties, such as a body of water
 *  or a cloud.
 *
 *  The important functions in the plugin interface itself follow a simple
 *  format using two structs (In/Out) as parameters. These structs have the
 *  same names between functions, but are defined within function
 *  specific namespaces (for instance, the init In/Out structs are defined
 *  within the MediumIOPInit namespace).
 *
 *  The plugin functions are mostly concerned with coordinating interactions
 *  with the calling simulation. As such, there are three primary functions
 *  that should be implemented: "setup", "init", and "updateState". An
 *  optional "teardown" notification is provided as well, in case any
 *  final calculations/output needs to be done before the plugin is unloaded.
 *
 *  The "setup" function is called once at the start of the simulation
 *  and its primary purpose is to provide runtime inputs to the plugin
 *  (in the form of an input stream so no file i/o needs to be performed
 *  in the plugin itself).
 *
 *  The "init" function is called once the simulation is ready to run.
 *  Information about the associated medium is passed in, and information
 *  about functions that actually do the work are passed back in the form
 *  of function wrappers (i.e. bound to std::function objects). Function
 *  prototypes for these are listed in the next section.
 *
 *  The "updateState" function is used to notify the plugin whenever
 *  the current time window (relative to a reference date time that is
 *  also passed in) or spectral state (wavelength samples in spectral
 *  data and the reference wavelength to use for geometric decisions)
 *  changes. This is an opportunity to prepare the plugin to return
 *  requested values as efficiently as possible based on knowledge
 *  of the state.
 */
class MediumIOPInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("mediumiop");
  }

  /** Initialization
   *
   *  Requests the individual callbacks that will be used for
   *  querying the IOP models for scattering/absorption coefficients
   *  and the scattering phase function probability density
   *  function. Callbacks are provided as std::function objects
   *  based on prototypes given for ScatterCallback,
   *  AbsorptionCallback, and ScatterPdfCallback.
   */
  virtual Status init(const MediumIOPInit::In &in, MediumIOPInit::Out &out) = 0;

  /** Optional flag to indicate that there is no scattering in
   *  this medium (i.e. aborption only). This will be used to
   *  significantly simplify internal calculations.
   */
  [[nodiscard]] virtual bool hasScattering() const { return true; }
};

/// \}
} // namespace d5::apis::plugins
