/** Interface file for a sensor plugin
 *
 *  \file SensorInterface.h
 */

#pragma once

#include <array>
#include <functional>
#include <iostream>
#include <random>
#include <unordered_map>
#include <vector>

#include "dirsig5/apis/ConvergenceCriteria.h"
#include "dirsig5/apis/ObjectInterface.h"
#include "dirsig5/apis/PlatformInfo.h"
#include "dirsig5/apis/Plugin.h"
#include "dirsig5/apis/SourceInfo.h"
#include "dirsig5/apis/TimeWindow.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/DualQuaternion.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Json.h"
#include "dirsig5/base/Pcg.h"
#include "dirsig5/base/State.h"
#include "dirsig5/base/Truth.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_sensor Sensor Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for a sensor plugin
 */

/// \addtogroup plugin_apis_sensor
/// \{

/// The callback function to generate a ray for a given problem.
using SamplerCallback = std::function<bool(
    RandEngine &randEngine, std::size_t taskIndex, std::size_t captureIndex,
    std::size_t problemIndex, std::size_t stateIndex, void *data, Point &origin,
    Vector &direction, float &time, double &weight)>;

/** Callback function to be used if any sources are within the problem
 *  solid angle (only checked if a solid angle and this callback
 *  are defined for a problem).
 */
using SourceCallback = std::function<void(
    std::size_t taskIndex, std::size_t captureIndex, std::size_t problemIndex,
    float visibility, const SourceInfo &)>;

/// Definition of input/output variables for the init interface
namespace SensorInit {

struct In : TaggedInstanceRequestsIn {
  /** While the plugin can override the convergence criteria
   *  for any particular problem, there is a simulation-wide
   *  default convergence set. This input lets the plugin
   *  know what those values are. When the problem info requests
   *  are made later on these default will be passed for
   *  the convergence criteria (they can be overriden at that point).
   */
  ConvergenceCriteria defaultConvergence;
};

struct Out {
  /** A list (vector) of the spectral states that will be
   *  used by this sensor (see SpectralState definition)
   */
  std::vector<SpectralState> spectralStates;

  /** The reference (absolute) date and time that all relative
   *  times will be reference in this plugin.  This should be at
   *  or prior to the very first task that will be defined by this
   *  plugin.
   */
  DateTime referenceTime;

  /** Request that the ephemeris plugin supress any update output
   *  at runtime (this request may not be respected by the
   *  other plugin)
   */
  bool requestEphemerisUpdateSupression = false;
};

} // namespace SensorInit

/// Definition of input/output variables for the getNextTask interface
namespace SensorNextTask {

struct In : TaggedInstanceRequestsIn {
  /** The elapsed time since the reference date/time
   *  provided during init. The function is guaranteed to be
   *  called with an elapsedTime of zero for the first call.
   */
  double elapsedTime = 0;

  /// The index of the current task (indexes start at 0).
  size_t taskIndex = 0;
};

struct Out {
  /** Window definition. Not that the TimeWindow object provided
   *  is initialized so that the task validity flag is false
   *  (i.e. indicating that plugin has no more tasks).  The helper
   *  setWindow function can be used to provide the window (and
   *  set the flag to true) or this can be done manually.
   */
  TimeWindow task;
};

} // namespace SensorNextTask

/// Definition of input/output variables for the collectionStarted interface
namespace SensorCollectionStarted {
struct In {};
struct Out {};
} // namespace SensorCollectionStarted

/// Definition of input/output variables for the collectionCompleted interface
namespace SensorCollectionCompleted {
struct In {};
struct Out {};
} // namespace SensorCollectionCompleted

/// Definition of input/output variables for the taskStarted interface
namespace SensorTaskStarted {

struct In {
  /// The index of the current task (indexes start at 0).
  size_t taskIndex = 0;
};

struct Out {};
} // namespace SensorTaskStarted

/// Definition of input/output variables for the taskCompleted interface
namespace SensorTaskCompleted {

struct In {
  /// The index of the current task (indexes start at 0).
  size_t taskIndex = 0;
};

struct Out {};
} // namespace SensorTaskCompleted

/// Definition of input/output variables for the getNextCapture interface
namespace SensorNextCapture {

struct In {
  /** The elapsed time since the start of the task.
   *  The time will be within the current task window (see taskIndex)
   */
  double elapsedTime = 0;

  /// The index of the current task (indexes start at 0).
  size_t taskIndex = 0;

  /// The index of the current capture (indexes start at 0).
  size_t captureIndex = 0;
};

struct Out {
  /** The "fine scale" time window for the task.
   *  Note that the TimeWindow object provided is initialized so
   *  that the task validity flag is false * (i.e. indicating that
   *  plugin has no more captures within the current task). The
   *  helper setWindow function can be used to provide the window
   *  (and set the flag to true) or this can be done manually.
   */
  TimeWindow capture;

  /** The number of problems in the capture. Problems are solid
   *  angles (possibly infinitely small) for which an average
   *  radiance will be calculated, such as a pixel radiance.
   */
  size_t nProblems = 0;

  /** This spectral states indexes (defined during init) that will
   *  be used for this capture. The values in the vector are indexes
   *  into SensorInit::Out::spectralStates (starting with zero). The
   *  sampler will be called separately for each spectral state.
   */
  std::vector<size_t> spectralStates;

  /** The number of problem groups or "chunks" that the capture
   *  will be broken up into.  Chunks will be equal length problem
   *  sets. Memory in the buffer is allocated to the size of each
   *  chunk (the number of problems times the number of spectral
   *  samples across all spectral states). Each chunk will be
   *  filled and sent to the sensor for processing before starting
   *  a new chunk on the same computation node.
   *
   *  Note that the number of problems (nProblems, above) must be
   *  divisible by the given chunk count. Prime or otherwise
   *  inconvenient problem counts can be padded to a functional
   *  size. Problems beyond the extent of the capture can then
   *  simply be skipped
   */
  size_t chunkCount = 1;

  /** Platform info. Gives a brief summary of what a collect will
   *  look like during the given capture window from a platform
   *  perspective (position, pointing, fov, etc..)
   */
  PlatformInfo info;

  /// The requests for truth to be collected.
  std::vector<Truth> truthRequests;

  /// An (optional) text description of the capture.
  std::string description;

  /** Additional logging data that the plugin would like to publish about this
   * capture.
   */
  Json pluginData;

  /** Request an (optional) gated return. This takes the provided
   *  time window and breaks it up into "gateCount" time chunks.
   *  If a "gateOffset" is also provided then any portion of the
   *  time window before the offset will be ignored and the remainder
   *  will be broken into "gateCount" time chunks. The gate bins
   *  are always uniform in size. A value of zero (default)
   *  indicates that no gate is in use.
   */
  size_t gateCount = 0;

  /** If defined with a non-zero "gateCount" this sets an offset from
   *  the beginning of the time window to start gating from
   *  (the end is still the end of the window). The offset
   *  is in seconds and must be non-negative. The default value
   *  is zero (no offset)
   */
  double gateOffset = 0;
};
} // namespace SensorNextCapture

/// Definition of input/output variables for the taskStarted interface
namespace SensorCaptureStarted {

struct In {
  /// The index of the current task (indexes start at 0).
  size_t taskIndex = 0;

  /// The index of the current capture (indexes start at 0).
  size_t captureIndex = 0;
};

struct Out {};
} // namespace SensorCaptureStarted

/// Definition of input variables for the captureCompleted interface
namespace SensorCaptureCompleted {

struct In {
  /// The index of the current task (indexes start at 0).
  size_t taskIndex = 0;

  /// The index of the current capture (indexes start at 0).
  size_t captureIndex = 0;
};

struct Out {
  /** Additional logging data that the plugin would like to publish
   *  about this capture.
   */
  Json pluginData;
};
} // namespace SensorCaptureCompleted

/// Definition of input/output variables for the getProblemInfo interface
namespace SensorProblemInfo {

struct In {
  /** The elapsed time since the reference date time
   *  provided during init. The time will be within the current
   *  task window (see taskIndex) and at the start of the current
   *  capture (see captureIndex)
   */
  double captureTime = 0;

  /// The index of the current task (indexes start at 0).
  size_t taskIndex = 0;

  /// The index of the current capture (indexes start at 0).
  size_t captureIndex = 0;

  /** The index of the specific problem to be solved.
   *  Note that the index has no meaning internally and is only
   *  used to reference an abstracted radiance solution.
   */
  size_t problemIndex = 0;

  /** The spectral state index indicates which set of wavelength
   *  samples (and reference wavelength) will be used for the
   *  path calculations. A problem info request will be made
   *  for each problem and each spectral state requested. If
   *  this problem does not require a particular spectral state,
   *  the skip flag can be set to true to bypass any calculations.
   */
  size_t spectralStateIndex = 0;
};

struct Out {
  /** The skip flag indicates if this particular problem and spectral
   *  state combination is not needed then the skip flag can be set
   *  to true to bypass any calculation. It is false by default.
   */
  bool skipFlag = false;

  /** Sampler callback. Provides a ray and its weight with each
   *  call. The solution will average the results from each
   *  sample ray (corresponding to a path). The callback
   *  is separate from the plugin interface to allow for
   *  modular definition of samplers within the sensor.
   */
  SamplerCallback callback;

  /** Convergence criteria. This struct provides the information
   *  used to decide when to stop sampling the problem and finalize
   *  the answer. This should only be used to modify the convergence
   *  of a single problem (simulation-wide convergence criteria
   *  will already be assigned).
   */
  ConvergenceCriteria convergenceCriteria = {};

  /** User data pointer to which contextual information can be passed to the
   *  sampler callback. A generic (void) pointer is provided which needs to be
   *  cast to the appropriate data structure.
   *
   *  Note that the user is responsible for allocating and de-allocating the
   *  corresponding memory.
   */
  void *data = nullptr;

  /** The scene index indicates which scene the ray coordinate system is
   *  based in. If less than zero, the coordinates are in global (ECEF)
   *  coordinates.
   */
  IntID sceneIndex = 0;

  /** Solid angle (rotationally symmetric) that bounds the problem.
   *  A zero/negative solid angle indicates that no solid angle
   *  calculations should be done. The sensor plugin is in
   *  charge of determinining if any contribution is real. Note
   *  that the bounds should consider temporal integration.
   */
  double solidAngle = -1;

  /** Center pointing direction of the problem. This is effectively
   *  the center of the rotationally symmetric solid angle and
   *  usually corresponds to the center of the detector (if
   *  the detector is integrating then this may be an effective
   *  center corresponding to the overall bounds).
   */
  Vector solidPointing = {0, 0, 0};

  /** Center pointing origin of the problem. This is effectively
   *  the center of the rotationally symmetric solid angle and
   *  usually corresponds to the center of the detector (if
   *  the detector is integrating then this may be an effective
   *  center corresponding to the overall bounds).
   */
  Vector solidOrigin = {0, 0, 0};

  /** The optional handler for sources found within the bounding
   *  solid angle of the problem. If the handler and the solid
   *  angle definitions exist then this function will be called
   *  so the plugin can access info about the source (location,
   *  pointing, irradiance, etc..) as defined in the common
   *  class header (CLSourceInfo.h).
   */
  SourceCallback sourceHandler;
};
} // namespace SensorProblemInfo

/// Definition of input/output variables for the Radiance Chunk interface
namespace SensorProcessRadianceChunk {

struct In {
  /** The buffer contains all radiance data for this chunk, stored spectrally,
   *  problem by problem.  The radiance units are W/(m^2 sr um).
   */
  const double *buffer = nullptr;

  /** The geo location buffer contains geodetic latitude/longitude/altitude info
   *  for each problem in the buffer (3*problemCount values, in decimal degrees
   *  and meters)
   */
  const double *geoBuffer = nullptr;

  /// The number of problems in the buffer.
  size_t problemCount = 0;

  /** The number of spectral samples in the buffer for each problem.
   *  Each problem is full written out before moving onto the next problem).
   */
  size_t spectralSampleCount = 0;

  /// The index of the current task (indexes start at 0).
  size_t taskIndex = 0;

  /// The index of the current capture (indexes start at 0).
  size_t captureIndex = 0;

  /** The overall problem index for the first set of spectral
   *  values in the supplied buffer.
   */
  size_t firstProblemIndex = 0;

  /** The (optional) gate count. If non-zero, the first set of
   *  pixels corresponds to the passive radiance as usual, but
   *  the remainder of the buffer represents the energy (not power)
   *  within the given gate bins. By default, this value is zero
   *  when no gate info was provided. Data returned for each gate
   *  is sequential,  starting with the first bin after the offset
   *  and providing the entire spectral buffer before moving on
   *  to the next bin.
   */
  size_t gateCount = 0;
};

struct Out {};
} // namespace SensorProcessRadianceChunk

/// Definition of input/output variables for the Truth Chunk interface
namespace SensorProcessTruthChunk {

struct In {
  /** The buffer contains all radiance data for this chunk,
   *  stored spectrally, problem by problem
   */
  const double *buffer = nullptr;

  /// The number of problems in the buffer.
  size_t problemCount;

  /** The number of truth items written per-problem in the buffer
   *  (analagous to the SensorProcessRadianceChunk::spectralSamplesCount)
   */
  size_t truthCount = 0;

  /// The index of the current task (indexes start at 0).
  size_t taskIndex = 0;

  /// The index of the current capture (indexes start at 0).
  size_t captureIndex = 0;

  /** The overall problem index for the first set of spectral
   *  values in the supplied buffer.
   */
  size_t firstProblemIndex = 0;
};

struct Out {};
} // namespace SensorProcessTruthChunk

class SensorInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("sensor");
  }

  /** The callback function to initialize the sensor plugin.
   *
   *  Requests the spectral states being used by this model and
   *  the reference date/time for any tasks/events that will be
   *  defined. This function is called once and is guaranteed to
   *  be called after setup and before any other functions.
   *
   *  This is likely the point at which the sensor will initialize
   *  itself in preparation for the first task request. Any requests
   *  that were made during setup are filled and passed back
   *  through this interface.
   */
  virtual Status init(const SensorInit::In &in, SensorInit::Out &out) = 0;

  /** Get the next task request.
   *
   *  Tasks in DIRSIG are defined as non-overlapping blocks in time
   *  during which capture events can occur for a particular object
   *  (it is possible for multiple plugin object to have overlapping
   *  tasks, but not within a single object).
   *
   *  The time block returned by this function must be at or after
   *  the current time provided. All times are provided in seconds
   *  from the reference date/time. Note that the initial request is
   *  made using the reference time (i.e. an elapsed time of zero).
   *
   *  All tasks must be provided in sequential order.
   */
  virtual Status getNextTask(const SensorNextTask::In &in,
                             SensorNextTask::Out &out) {
    return SUCCESS;
  }

  /** Simulation start notification.
   *
   *  Indication that the collection has started. This is called
   *  before the first task is started.
   */
  virtual Status collectionStarted(const SensorCollectionStarted::In &,
                                   SensorCollectionStarted::Out &) {
    return SUCCESS;
  }

  /** Simulation completed notification.
   *
   *  Indication that the collection has completed. This is called
   *  after the last task is completed.
   */
  virtual Status collectionCompleted(const SensorCollectionCompleted::In &,
                                     SensorCollectionCompleted::Out &) {
    return SUCCESS;
  }

  /** Task start notification.
   *
   *  Indication that the current task has started (the provided
   *  index corresponds to the one provided for getNextTask). This
   *  function is optional and will not be called again until
   *  the corresponding taskStarted function is called.
   */
  virtual Status taskStarted(const SensorTaskStarted::In &,
                             SensorTaskStarted::Out &) {
    return SUCCESS;
  }

  /** Task complete notification.
   *
   *  Indication that the current task is complete (the provided
   *  index corresponds to the one provided for getNextTask). This
   *  function is optional.
   */
  virtual Status taskCompleted(const SensorTaskCompleted::In &,
                               SensorTaskCompleted::Out &) {
    return SUCCESS;
  }

  /** Get the next capture request.
   *
   *  A capture defines a time window within a task during which
   *  radiometry problems are solved. Unlike task windows, capture
   *  windows can overlap, but only one capture will be processed
   *  at a time on a single node. Captures describe radiometry problems in
   *  an abstract way (field of view, number of individual problems
   *  such as pixels, etc..). This information is used internally
   *  to schedule calculations across resources in an efficient way.
   */
  virtual Status getNextCapture(const SensorNextCapture::In &in,
                                SensorNextCapture::Out &out) {
    return SUCCESS;
  }

  /** Capture start notification.
   *
   *  Indication that the provided capture in the given task has
   *  started. This function is optional and will not
   *  be called again on a single node until the corresponding
   *  captureCompleted function has been called.
   */
  virtual Status captureStarted(const SensorCaptureStarted::In &,
                                SensorCaptureStarted::Out &) {
    return SUCCESS;
  }

  /** Capture complete notification.
   *
   *  Indication that the provided capture in the given task is
   *  complete. This function is optional.
   */
  virtual Status captureCompleted(const SensorCaptureCompleted::In &,
                                  SensorCaptureCompleted::Out &) {
    return SUCCESS;
  }

  /** Get information about a problem request.
   *
   *  This function requests information about a single problem that
   *  is indicated by an index into the number of problems provided
   *  from the capture information. The applicable task, state and capture
   *  indexes are provided as well and the function should return
   *  convergence criteria for the solution and a pointer to a
   *  buffer where the solution should be placed. Note that a problem
   *  may be skipped (e.g. the requested spectral state is not
   *  applicable for that pixel).
   *
   *  Rather than sample a problem directly
   *  through the plugin interface, a callback function pointer
   *  is provided. This allows for directing particular problems
   *  to pre-built modules without having to test each sample
   *  request.
   *
   *  There is no corresponding "problemComplete" function as problems
   *  can be written to the supplied buffers in any order internally.
   */
  virtual Status getProblemInfo(const SensorProblemInfo::In &in,
                                SensorProblemInfo::Out &out) {
    return SUCCESS;
  }

  /** Process a radiance capture chunk notification.
   *
   *  Chunks are portions of a capture that are computed and processed
   *  before the next chunk is started (e.g. a sub-set of pixels in
   *  a focal plane, like a line or a rectanagular region).
   *  This function is called whenever a chunk has been filled and
   *  is ready for processing. It will be called in a separate thread
   *  and a double buffer system is used to start processing the
   *  next chunk internally. Note that chunks include space for
   *  all spectral samples across all spectral states defined by a
   *  capture and the buffer should be processed accordingly.
   *  Spectral data is given in the state order and all spectral
   *  information for a problem is filled (or skipped) before another
   *  pixel begins.
   *
   *  Chunks may be available for processing in any order and plugins
   *  should use the provided information to appropriately process
   *  and output the correct portion of the capture.
   */
  virtual Status processRadianceChunk(const SensorProcessRadianceChunk::In &in,
                                      SensorProcessRadianceChunk::Out &out) {
    return SUCCESS;
  }

  /** Process a truth capture chunk notification.
   *
   *  Similar to capture chunk processing but the returned buffer
   *  contains the requested truth data. This function is optional
   *  but should be implemented if truth is requested.
   */
  virtual Status processTruthChunk(const SensorProcessTruthChunk::In &,
                                   SensorProcessTruthChunk::Out &) {
    return SUCCESS;
  }

  std::vector<SensorInterface *> children;
};

/// \}
} // namespace d5::apis::plugins
