#pragma once

#include <functional>
#include <string>
#include <vector>

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/State.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_statecontroller State Controller Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for a state controller plugin
 */

/// \addtogroup plugin_apis_statecontroller
/// \{

/** Callback for the enabled state. The enabled state can be a
 *  floating point value between zero and one, though not all
 *  state checks will know what to do with a non-binary state.
 *  In those cases, any non-zero value will be considered to be "on".
 *  Similarly any enabled state > 1.f will be truncated to one.
 *
 *  Time given is for the current temporal state.
 */
using EnabledStateCallback =
    std::function<float(const DateTime &referenceDateTime, float captureStart,
                        float t // time during capture
                        )>;

/// Definition of input/output variables for the Init interface
namespace StateControllerInit {

struct In {};

struct Out {
  struct Callback {
    std::vector<std::string> tagSet;
    EnabledStateCallback callback;
  };

  /** Vector of callbacks corresponding to the paired
   *  tag set (i.e. each callback is attached to the object(s)
   *  corresponding to all tags in the tag set).
   */
  std::vector<Callback> callbacks;
};
} // namespace StateControllerInit

class StateControllerInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("statecontroller");
  }

  /// Initialization
  ///
  /// Requests callbacks for obtaining the current object states
  /// at a time within the current temporal state.
  ///
  /// Callbacks are provided as std::function objects
  /// based on the prototypes given for IntersectCallback
  ///
  virtual Status init(const StateControllerInit::In &in,
                      StateControllerInit::Out &out) = 0;
};

/// \}
} // namespace d5::apis::plugins
