/** Interface file for a weather plugin
 *
 *  \file WeatherInterface.h
 */

#pragma once

#include <vector>

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Weather.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_weather Weather Plugin
 *  \ingroup plugin_apis
 *  Holds the API definition for a weather plugin
 */

/// \addtogroup plugin_apis_weather
/// \{

/// Definition of input/output variables for the init interface.
namespace WeatherInit {

struct In {
  /** Scene origins in latitude, longitude and altitude. The index
   *  into the vector corresponds to (non-negative) scene indexes
   *  passed vi the query. Latitude and longitude in degrees,
   *  altitude in meters (relative to a WGS84 ellipsoid).
   */
  std::vector<Point> sceneOrigins;
};

struct Out {};
} // namespace WeatherInit

/// Definition of input/output variables for the request interface.
namespace WeatherRequest {

struct In {
  /// The lat/lon/alt for the request
  Point lla = Point(0, 0, 0);

  /// The time for the request
  double time = 0;
};

struct Out {
  /// The requested weather data
  Weather weather;
};
} // namespace WeatherRequest

/** The weather model in DIRSIG5 provides the meteorological data
 *  for the simulation at locations and times.
 *
 *  Any errors produced internally should be thrown after describing the
 *  issue as thoroughly as possible in an error message.
 *
 *  All virtual functions should be implemented, but the static
 *  \c getVersion() is used to track the version of this interface.
 *  Plugin functions are implemented using the same interface format,
 *  which provides two custom structs, one for the inputs and one
 *  for the outputs.
 */
class WeatherInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("weather");
  }

  /// Provides runtime info to the plugin.
  virtual Status init(const WeatherInit::In &in, WeatherInit::Out &out) = 0;

  /// Provides weather data for the requested location and date/time.
  virtual Status request(const WeatherRequest::In &in,
                         WeatherRequest::Out &out) = 0;
};

/// \}
} // namespace d5::apis::plugins
