#pragma once

#include <cstdint>
#include <string>
#include <vector>

#include "dirsig5/atmosphere/Export.h"
#include "dirsig5/base/Json.h"

namespace d5::atm {

/** \defgroup atm Atmosphere
 *  Holds the atmosphere model frontend library
 */

/// \addtogroup atm
/// \{

/** Constants for the various types run request for the backend atmosphere
 *  model to compute.
 */
enum class ModelRequestType : std::uint32_t {
  Unknown = 0x00,       ///< Unknown request type
  ProfileData = 0x01,   ///< Request for an atmospheric profile
  ExoSourceData = 0x02, ///< Request for a exo-atm solar data
  PathData = 0x03,      ///< Request for a path within the atmosphere
  SkyData = 0x04,       ///< Request for a path to the sky (space)
  TotalData = 0x05      ///< Request for a reflected/emitted path
};

/** Constants for features supported by a backend atmosphere model.
 */
enum class BackendFeatureType : std::uint32_t {
  Unknown = 0x00,             ///< Undefined feature
  GenerateProfileData = 0x01, ///< Can generate vertical profile data
  SpectralReflectance = 0x02, ///< Can accept a spectral reflectance data
  SpectralResponse = 0x04     ///< Can accept a spectral response data
};

/// Constants to define the options for a given request.
enum class ModelRequestOptions : std::uint32_t {
  None = 0x00,             ///< No options provided
  SkipRequest = 0x01,      ///< Setup the request but don't run it
  PreserveRequest = 0x02,  ///< Preserve the requests temporary files
  SunSourceRequest = 0x04, ///< Use the sun as the exo-atmospheric source
  MoonSourceRequest = 0x08 ///< Use the sun as the exo-atmospheric source
};

/// Bitwise and operator for testing flags
inline std::uint32_t operator&(ModelRequestOptions a, ModelRequestOptions b) {
  return static_cast<std::uint32_t>(a) & static_cast<std::uint32_t>(b);
}

/// Bitwise or operator for combining
inline ModelRequestOptions operator|(ModelRequestOptions a,
                                     ModelRequestOptions b) {
  return static_cast<ModelRequestOptions>(static_cast<std::uint32_t>(a) |
                                          static_cast<std::uint32_t>(b));
}

/// Bitwise or operator for combining
inline ModelRequestOptions &operator|=(ModelRequestOptions &a,
                                       ModelRequestOptions b) {
  a = a | b;
  return a;
}

/// A struct for a spectral wavelength/value sample.
struct SpectralSample {
  float wavelength = 0.f;
  float value = 0.f;
  SpectralSample(void) : wavelength(0.f), value(0.f) {}
  SpectralSample(float w, float v) : wavelength(w), value(v) {}
};

/// A type for a spectral curve
typedef std::vector<SpectralSample> SpectralCurveData;

/// A type for a set of relative spectral response (RSR) curves
typedef std::vector<SpectralCurveData> SpectralResponseData;

/** A structure to describe a request to a backend atmosphere model.
 *  Depending on the request type, a subset of the variables will be populated.
 */
struct ModelRequest {
  ModelRequestType type = ModelRequestType::Unknown;
  ModelRequestOptions options = ModelRequestOptions::None;
  size_t index = 0;
  std::string description;
  int dayOfYear;
  std::vector<double> wavelengthSamples;
  double groundAltM = 0.0;
  double pathAlt1M = 0.0;
  double pathAlt2M = 0.0;
  double pathLengthM = 0.0;
  double pathZenithDeg = 0.0;
  double pathAzimuthDeg = 0.0;
  double sourceZenithDeg = 0.0;
  double sourceAzimuthDeg = 0.0;
  double sourcePhaseAngleDeg = 0.0;
  const SpectralCurveData *reflData = nullptr;
  const SpectralResponseData *respData = nullptr;
  Json backendInput;
};

/// JSON Parsing for ModelRequest
D5_ATM_EXPORT void from_json(const Json &j, ModelRequest &r);

/// JSON Serialization for ModelRequest
D5_ATM_EXPORT void to_json(Json &j, const ModelRequest &r);

/**
 * A structure to describe a response to a backend atmosphere model.
 * Depending on the original request type, a subset of the variables
 * will be populated.
 */
struct ModelResponse {
  /** The first buffer (primarily used for irradiance or radiances) */
  std::vector<float> buffer1;

  /** The second buffer (primarily used for transmission or extinctions) */
  std::vector<float> buffer2;

  /** The variable used to record the actual range of the path */
  float range = 0;

  /** The table to store the atm profile data */
  std::vector<float> profileTable;
};

NLOHMANN_DEFINE_TYPE_NON_INTRUSIVE(ModelResponse, buffer1, buffer2, range,
                                   profileTable)

/**
 * The interface for a backend atmosphere model.
 */
class D5_ATM_EXPORT ModelBackend {
public:
  /// Virtual destructor
  virtual ~ModelBackend() = default;

  /// Get the name of the backend model
  [[nodiscard]] virtual const std::string &name() const = 0;

  /// Set the flag to save files generated by the backend
  void enableSaveFiles(bool enable) { this->saveFiles_ = enable; }

  /**
   * Check if the backend generates atmosphere profile data.
   * The default is false.
   */
  [[nodiscard]] virtual bool generatesProfileData() const { return false; }

  /**
   * A (optional) test function that confirms the backend is correctly
   * configured to run.  Any messages should be written to the input
   * stream.  The default return value is true.
   */
  [[nodiscard]] virtual bool test(std::ostream &stream) const { return true; }

  /**
   * Provide a spectral reflectance curve.
   * The pointer to the data is assumed to stay valid for the lifespan of
   * the object. This pointer can be included in a request and the
   * apppropriate mapping to the data will be handled by the backend.
   * The function returns true if the data is correctly stored for
   * requests or false if the feature is unsupported by the backend.
   */
  virtual bool addSpectralReflectance(const std::string &name,
                                      const SpectralCurveData *reflData) {
    return false;
  }

  /**
   * Provide a set of relative spectral response (RSR) curves.
   * The pointer to the data is assumed to stay valid for the lifespan of
   * the object. This pointer can be included in a request and the
   * apppropriate mapping to the data will be handled by the backend.
   * The function returns true if the data is correctly stored for
   * requests or false if the feature is unsupported by the backend.
   */
  virtual bool addSpectralResponse(const SpectralResponseData *rsrData) {
    return false;
  }

  /**
   * Print a summary of the backend to the supplied stream.
   */
  virtual void printSummary(std::ostream &stream) const = 0;

  /**
   * Perform the supplied request.
   */
  [[nodiscard]] virtual ModelResponse
  performRequest(const ModelRequest &_request) const = 0;

protected:
  /** The flag to save the input/output file */
  mutable bool saveFiles_ = false;
};

struct CreateModelBackendRequest {
  std::string name;
  Json backendData;
};

D5_ATM_EXPORT void from_json(const Json &j, CreateModelBackendRequest &r);
D5_ATM_EXPORT void to_json(Json &j, const CreateModelBackendRequest &r);

D5_ATM_EXPORT std::unique_ptr<ModelBackend>
createBackend(const CreateModelBackendRequest &req);

/// A callback function for resolving paths (variable substitution and such)
using PathResolutionCallback = std::function<std::string(const std::string &)>;

/** Sets the callback function to resolve paths
 *
 *  \param f  The callback function
 */
D5_ATM_EXPORT void SetPathResolutionCallback(PathResolutionCallback f);

/** Resolves the given path
 *
 *  \param p  The path to resolve
 *
 *  \return The resolved path
 *  \throw std::exception subclass on any error
 */
std::string ResolvePath(const std::string &p);

/// \}
} // namespace d5::atm
