#pragma once

#include <iostream>

#include <functional>
#include <memory>
#include <optional>
#include <stdexcept>
#include <string>

#include "ModtranCommon/HazeModel.h"
#include "ModtranCommon/MultipleScattering.h"
#include "dirsig5/atmosphere/Export.h"
#include "dirsig5/base/Inheritance.h"
#include "dirsig5/base/Json.h"

namespace d5::atm::modtran {

/** \defgroup atm_modtran MODTRAN backend libraries
 *  \ingroup atm
 *  Holds public interface for the MODTRAN backends
 */

/// \addtogroup atm_modtran
/// \{

/// Definition of a MODTRAN profile
struct Profile {
  /// Path to the MODTRAN executable
  std::string exeFilePath;

  /// Path to the MODTRAN data directory
  std::string dataDirPath;
};

/// Callback function used to resolve a MODTRAN profile name
using ModtranProfileCallback = std::function<Profile(const std::string &)>;

/** Sets the callback function to resolve MODTRAN profile names
 *
 *  \param f  The callback function
 */
D5_ATM_EXPORT void SetModtranProfileCallback(ModtranProfileCallback f);

/** Gets the MODTRAN profile for the given name
 *
 *  \c SetModtranProfileCallback must be called before this function
 *
 *  \param name  The name of the profile
 *
 *  \return The MODTRAN profile
 *  \throw std::exception subclass on any error
 */
D5_ATM_EXPORT Profile GetModtranProfile(const std::string &name);

#define D5_ATM_MODTRAN_MODEL_TYPES(X)                                          \
  X(Unknown, -999)                                                             \
  X(MidLatitudeSummer, 2)                                                      \
  X(UserDefinedSingleAltitude, 0)                                              \
  X(Tropical, 1)                                                               \
  X(MidLatitudeWinter, 3)                                                      \
  X(SubArcticSummer, 4)                                                        \
  X(SubArcticWinter, 5)                                                        \
  X(UsStandard1976, 6)                                                         \
  X(UserDefined, 7)                                                            \
  X(UserDefinedPressureDependent, 8)

#define X_ENUM_DECL(enumVal, val) enumVal = val, /* NOLINT */

enum class ModelType : int { D5_ATM_MODTRAN_MODEL_TYPES(X_ENUM_DECL) };

#define X_JSON_DECL(enumVal, val) {ModelType::enumVal, #enumVal},

D5_JSON_SERIALIZE_ENUM(ModelType, {D5_ATM_MODTRAN_MODEL_TYPES(X_JSON_DECL)});

#undef X_JSON_DECL
#undef X_ENUM_DECL

struct Tape5Parameters {
  ModelType atmosphericModel = ModelType::MidLatitudeSummer;
  std::unique_ptr<HazeModel> hazeModel;
  std::unique_ptr<MultipleScattering> multipleScattering;
  float co2MixingRatioPpm = 0.f;
  std::string bandModelName = "";
};

D5_FROM_JSON(Tape5Parameters, j, v) {
  if (auto it = j.find("atmospheric_model"); it != j.end()) {
    v.atmosphericModel = j.at("atmospheric_model").get<ModelType>();
    if (v.atmosphericModel == ModelType::Unknown) {
      throw std::runtime_error("unknown atmospheric model type!");
    }
  }
  if (auto it = j.find("boundary_aerosol_model"); it != j.end()) {
    v.hazeModel = HazeModel::fromJSON(it.value());
  }
  if (auto it = j.find("multiple_scattering"); it != j.end()) {
    v.multipleScattering = MultipleScattering::fromJSON(it.value());
  }
  d5::Optional(j, "co2_mixing_ratio_ppm", v.co2MixingRatioPpm);
  d5::Optional(j, "band_model_name", v.bandModelName);
}

/// \}
} // namespace d5::atm::modtran
