/** Interface file for MODTRAN haze models
 *
 *  \file HazeModel.h
 */

#pragma once

#include <memory>
#include <string>

#include "dirsig5/atmosphere/Export.h"
#include "dirsig5/base/Inheritance.h"
#include "dirsig5/base/Json.h"

namespace d5::atm::modtran {

/// \addtogroup atm_modtran
/// \{

/** XMacro for the MODTRAN haze model types
 *
 *  \param X  Function macro of the form X(enumVal, value) to apply to each
 *            type
 */
#define D5_ATM_MODTRAN_HAZE_MODEL_TYPES(X)                                     \
  X(Unknown, -999)                                                             \
  X(RuralVis23Km, 1)                                                           \
  X(NoAerosolMaybeClouds, -1)                                                  \
  X(NoAerosolOrClouds, 0)                                                      \
  X(RuralVis5Km, 2)                                                            \
  X(NavyMaritime, 3)                                                           \
  X(MaritimeVis23Km, 4)                                                        \
  X(UrbanVis5Km, 5)                                                            \
  X(TroposphericVis50Km, 6)                                                    \
  X(UserDefinedAerosol, 7)                                                     \
  X(AdvectiveFoxVis0_2Km, 8)                                                   \
  X(RadiativeFogVis0_5Km, 9)                                                   \
  X(Desert, 10)

/// \cond HIDDEN_SYMBOLS
#define X_ENUM_DECL(enumVal, val) enumVal = val, /* NOLINT */
#define X_JSON_DECL(enumVal, val) {HazeModel::Type::enumVal, #enumVal},
/// \endcond

/// Top-level class for a MODTRAN haze model
class HazeModel : public EnumeratedSuperclass<HazeModel> {
public:
  /// Constants for the supported types of haze models
  enum class Type : int { D5_ATM_MODTRAN_HAZE_MODEL_TYPES(X_ENUM_DECL) };

  /// Destructor
  virtual ~HazeModel() = default;

  /// Get the type of the haze model
  [[nodiscard]] Type type() const { return type_; };

  // Get the override visibility
  [[nodiscard]] float visibility() const { return visibilityKm_; }
  [[nodiscard]] float &visibility() { return visibilityKm_; }

  /** Dispatch function
   *
   *  \param callback  Callback function that takes the type-specific subclass
   *                   as the first argument (templated)
   *  \param args      Additional arguments to be forwarded to the callback
   *
   *  \return Return value of the callback
   */
  template <typename Callable, typename... Args>
  decltype(auto) dispatch(Callable &&callback, Args &&...args) const;

  /** Makes a HazeModel from JSON
   *
   *  \param j  The input JSON
   *
   *  \return Pointer to the HazeModel type (might be a subclass)
   */
  static std::unique_ptr<HazeModel> fromJSON(const d5::Json &j);

protected:
  /// Constructor
  HazeModel(Type type) : type_(type) {}

private:
  /// Type of the model
  Type type_;

  /// The override visibility (km)
  float visibilityKm_ = 0.f;
};

D5_FROM_JSON(HazeModel, j, v) {
  if (auto it = j.find("visibility_km"); it != j.end()) {
    it->get_to(v.visibility());
  }
}

D5_JSON_SERIALIZE_ENUM(HazeModel::Type,
                       {D5_ATM_MODTRAN_HAZE_MODEL_TYPES(X_JSON_DECL)});

#undef X_ENUM_DECL
#undef X_JSON_DECL

/// Subclass for the user-defined aerosol haze model
struct UserDefinedAerosolHazeModel
    : public HazeModel::Subclass<HazeModel::Type::UserDefinedAerosol> {
public:
  UserDefinedAerosolHazeModel() {
    throw std::runtime_error("user-defined aerosol haze models are not "
                             "supported in tape5 defaults mode yet");
  }
};

D5_FROM_JSON(UserDefinedAerosolHazeModel, j, v) {}

/// Subclass for the Navy Maritime extinction haze model
struct NavyMaritimeHazeModel
    : public HazeModel::Subclass<HazeModel::Type::NavyMaritime> {
public:
  NavyMaritimeHazeModel() = default;
  NavyMaritimeHazeModel(float wsMS, float ws24MS)
      : currentWindSpeedMS(wsMS), averageWindSpeedMS(ws24MS) {}
  ~NavyMaritimeHazeModel() override = default;

  float currentWindSpeedMS = 0.f;
  float averageWindSpeedMS = 0.f;
  int airMassInfluence = 3;
};

D5_FROM_JSON(NavyMaritimeHazeModel, j, v) {
  j.at("current_wind_speed_ms").get_to(v.currentWindSpeedMS);
  j.at("average_wind_speed_ms").get_to(v.averageWindSpeedMS);
  j.at("air_mass_influence").get_to(v.airMassInfluence);
  if (auto it = j.find("visibility_km"); it != j.end()) {
    it->get_to(v.visibility());
  }
}

/// Subclass for the Desert haze model
struct DesertHazeModel
    : public HazeModel::Subclass<HazeModel::Type::NavyMaritime> {
public:
  DesertHazeModel() = default;
  DesertHazeModel(float wsMS) : currentWindSpeedMS(wsMS) {}
  ~DesertHazeModel() override = default;

  float visibilityKm = 0.f;
  float currentWindSpeedMS = 0.f;
};

D5_FROM_JSON(DesertHazeModel, j, v) {
  j.at("current_wind_speed_ms").get_to(v.currentWindSpeedMS);
  if (auto it = j.find("visibility_km"); it != j.end()) {
    it->get_to(v.visibility());
  }
}

/** Dispatch function on the haze model type
 *
 *  \tparam Callable  Templated functor struct, the template argument will be
 *                    the HazeModel subclass for the given type. There must
 *                    be a static dispatch() function invokable with Args.
 *  \tparam Args      Types of arguments to be passed to dispatch()
 *  \param type       The enum type we are dispatching on
 *  \param args       Arguments to be passed on to Callable::dispatch
 */
template <template <class> class Callable, typename... Args>
decltype(auto) dispatch(HazeModel::Type type, Args &&...args) {
  switch (type) {
  case HazeModel::Type::NoAerosolMaybeClouds:
  case HazeModel::Type::NoAerosolOrClouds:
  case HazeModel::Type::RuralVis23Km:
  case HazeModel::Type::RuralVis5Km:
  case HazeModel::Type::MaritimeVis23Km:
  case HazeModel::Type::UrbanVis5Km:
  case HazeModel::Type::TroposphericVis50Km:
  case HazeModel::Type::AdvectiveFoxVis0_2Km:
  case HazeModel::Type::RadiativeFogVis0_5Km:
    return Callable<HazeModel>::dispatch(std::forward<Args>(args)...);
  case HazeModel::Type::UserDefinedAerosol:
    return Callable<UserDefinedAerosolHazeModel>::dispatch(
        std::forward<Args>(args)...);
  case HazeModel::Type::NavyMaritime:
    return Callable<NavyMaritimeHazeModel>::dispatch(
        std::forward<Args>(args)...);
  case HazeModel::Type::Desert:
    return Callable<DesertHazeModel>::dispatch(std::forward<Args>(args)...);
  case HazeModel::Type::Unknown:
  default:
    throw std::runtime_error("unknown boundary aerosol type");
  }
}

// #################################################################
// #              Inline Implementation for HazeModel              #
// #################################################################

namespace detail {
template <class T> struct HazeModelDispatchHelper {
  template <typename Callable, typename... Args>
  static decltype(auto) dispatch(Callable &&callback, const HazeModel &hz,
                                 Args &&...args) {
    if constexpr (std::is_same_v<T, HazeModel>) {
      return callback(hz, std::forward<Args>(args)...);
    } else {
      return callback(hz.as<T>(), std::forward<Args>(args)...);
    }
  }
};

template <class T> struct HazeModelFromJsonHelper {
  template <typename Callable>
  static std::unique_ptr<HazeModel> dispatch(const d5::Json &j,
                                             Callable &&makeSimple) {
    if constexpr (std::is_same_v<T, HazeModel>) {
      return makeSimple();
    } else {
      auto v = std::make_unique<T>();
      from_json(j.at("parameters"), *v);
      return v;
    }
  }
};
} // namespace detail

template <typename Callable, typename... Args>
decltype(auto) HazeModel::dispatch(Callable &&callback, Args &&...args) const {
  return d5::atm::modtran::dispatch<detail::HazeModelDispatchHelper>(
      type(), std::forward<Callable>(callback), *this,
      std::forward<Args>(args)...);
}

inline std::unique_ptr<HazeModel> HazeModel::fromJSON(const d5::Json &j) {
  auto type = j.at("type").get<HazeModel::Type>();
  auto MakeSimple = [=]() {
    auto model = std::unique_ptr<HazeModel>(new HazeModel(type));
    from_json(j.at("parameters"), *model);
    return model;
  };
  return d5::atm::modtran::dispatch<detail::HazeModelFromJsonHelper>(
      type, j, MakeSimple);
}

/// \}
} // namespace d5::atm::modtran
