/** Interface file for MODTRAN multiple scattering models
 *
 *  \file MultiplScattering
 */

#pragma once

#include <memory>
#include <string>

#include "dirsig5/atmosphere/Export.h"
#include "dirsig5/base/Inheritance.h"
#include "dirsig5/base/Json.h"

namespace d5::atm::modtran {

/// \addtogroup atm_modtran
/// \{

/** XMacro for the MODTRAN haze model types
 *
 *  \param X  Function macro of the form X(enumVal, value) to apply to each
 *            type
 */
#define D5_ATM_MODTRAN_MULTI_SCAT_TYPES(X)                                     \
  X(Unknown, -999)                                                             \
  X(Isaac, 1)                                                                  \
  X(None, 0)                                                                   \
  X(Disort, 2)

/// \cond HIDDEN_SYMBOLS
#define X_ENUM_DECL(enumVal, val) enumVal = val, /* NOLINT */
#define X_JSON_DECL(enumVal, val) {MultipleScattering::Type::enumVal, #enumVal},
/// \endcond

/// Top-level class for a multiple scattering model
class MultipleScattering : public EnumeratedSuperclass<MultipleScattering> {
public:
  /// Constants for the supported types of haze models
  enum class Type : int { D5_ATM_MODTRAN_MULTI_SCAT_TYPES(X_ENUM_DECL) };

  /// Destructor
  virtual ~MultipleScattering() = default;

  /// The type of the model
  [[nodiscard]] Type type() const { return type_; };

  /** Dispatch function
   *
   *  \param callback  Callback function that takes the type-specific subclass
   *                   as the first argument (templated)
   *  \param args      Additional arguments to be forwarded to the callback
   *
   *  \return Return value of the callback
   */
  template <typename Callable, typename... Args>
  decltype(auto) dispatch(Callable &&callback, Args &&...args) const;

  /** Makes a MultipleScattering from JSON
   *
   *  \param j  The input JSON
   *
   *  \return Pointer to the MultipleScattering object (might be a subclass)
   */
  static std::unique_ptr<MultipleScattering> fromJSON(const d5::Json &j);

protected:
  /// Constructor
  MultipleScattering(Type type) : type_(type) {}

private:
  /// Type of the model
  Type type_;
};

D5_JSON_SERIALIZE_ENUM(MultipleScattering::Type,
                       {D5_ATM_MODTRAN_MULTI_SCAT_TYPES(X_JSON_DECL)});

#undef X_ENUM_DECL
#undef X_JSON_DECL

/// Subclass for the DISORT multiple scattering model
class DisortMultipleScattering
    : public MultipleScattering::Subclass<MultipleScattering::Type::Disort> {
public:
  DisortMultipleScattering() = default;

  DisortMultipleScattering(int streams) { setStreams(streams); }

  [[nodiscard]] int streams() const { return streams_; }

  void setStreams(int streams) {
    if (streams < 2 || streams > 128 || (streams & (streams - 1)) != 0) {
      throw std::runtime_error(
          "streams must be a power of 2 between 2 and 128");
    }
    streams_ = streams;
  }

private:
  int streams_ = 2;
};

D5_FROM_JSON(DisortMultipleScattering, j, v) {
  v.setStreams(j.at("streams").get<int>());
}

/** Dispatch function on the multi-scattering model type
 *
 *  \tparam Callable  Templated functor struct, the template argument will be
 *                    the MultipleScattering subclass for the given type.
 *                    There must be a static dispatch() function invokable
 *                    with Args.
 *  \tparam Args      Types of arguments to be passed to dispatch()
 *  \param type       The enum type we are dispatching on
 *  \param args       Arguments to be passed on to Callable::dispatch
 */
template <template <class> class Callable, typename... Args>
decltype(auto) dispatch(MultipleScattering::Type type, Args &&...args) {
  switch (type) {
  case MultipleScattering::Type::None:
  case MultipleScattering::Type::Isaac:
    return Callable<MultipleScattering>::dispatch(std::forward<Args>(args)...);
  case MultipleScattering::Type::Disort:
    return Callable<DisortMultipleScattering>::dispatch(
        std::forward<Args>(args)...);
  case MultipleScattering::Type::Unknown:
  default:
    throw std::runtime_error("unknown multiple scattering type");
  }
}

// ##################################################################
// #          Inline Implementation for MultipleScattering          #
// ##################################################################

namespace detail {
template <class T> struct MultipleScatteringDispatchHelper {
  template <typename Callable, typename... Args>
  static decltype(auto) dispatch(Callable &&callback,
                                 const MultipleScattering &ms, Args &&...args) {
    if constexpr (std::is_same_v<T, MultipleScattering>) {
      return callback(ms, std::forward<Args>(args)...);
    } else {
      return callback(ms.as<T>(), std::forward<Args>(args)...);
    }
  }
};

template <class T> struct MultipleScatteringFromJsonHelper {
  template <typename Callable>
  static std::unique_ptr<MultipleScattering> dispatch(const d5::Json &j,
                                                      Callable &&makeSimple) {
    if constexpr (std::is_same_v<T, MultipleScattering>) {
      return makeSimple();
    } else {
      auto v = std::make_unique<T>();
      from_json(j.at("parameters"), *v);
      return v;
    }
  }
};
} // namespace detail

template <typename Callable, typename... Args>
decltype(auto) MultipleScattering::dispatch(Callable &&callback,
                                            Args &&...args) const {
  return d5::atm::modtran::dispatch<detail::MultipleScatteringDispatchHelper>(
      type(), std::forward<Callable>(callback), *this,
      std::forward<Args>(args)...);
}

inline std::unique_ptr<MultipleScattering>
MultipleScattering::fromJSON(const d5::Json &j) {
  auto type = j.at("type").get<MultipleScattering::Type>();
  auto MakeSimple = [=]() {
    return std::unique_ptr<MultipleScattering>(new MultipleScattering(type));
  };
  return d5::atm::modtran::dispatch<detail::MultipleScatteringFromJsonHelper>(
      type, j, MakeSimple);
}

/// \}
} // namespace d5::atm::modtran
