/** Interface file for a bandpass
 *
 *  \file Bandpass.h
 */

#pragma once

#include <cassert>
#include <cstddef>

#include "dirsig5/base/Constants.h"
#include "dirsig5/base/Json.h"
#include "dirsig5/base/State.h"

namespace d5 {

/// \addtogroup common
/// \{

/** Hold a uniformly-sampled spectral bandpass
 *
 *  It is described by a minimum, maximum, sampling delta and the units.
 */
struct Bandpass {
  /// Default constructor, makes an invalid bandpass
  Bandpass() = default;

  /** Constructor, makes a bandpass in microns
   *
   *  \param minVal    Minimum value of the bandpass [micron]
   *  \param maxVal    Maximum value of the bandpass [micron]
   *  \param deltaVal  Spacing between samples in the bandpass [micron]
   *  \param count     The number of samples in the bandpass
   *
   *  TODO: deltaVal and count obviously allow the user to create garbage
   *        objects, we should address this. Also having the members public
   *        can violate invariants of this object.
   */
  Bandpass(double minVal, double maxVal, double deltaVal, size_t count)
      : units(d5::MicronUnits), minimum(minVal), maximum(maxVal),
        delta(deltaVal), pointCount(count) {}

  /** A routine to generate the list of wavelenghts defined by
   * the bandpass.
   */
  [[nodiscard]] std::vector<double> generateWavelengths() const {
    std::vector<double> wavelengths(pointCount);
    for (size_t ii = 0; ii < pointCount; ii++) {
      wavelengths[ii] = minimum + (ii * delta);
    }
    return wavelengths;
  }

  /** A routine to convert a spectral location from one unit to another.
   *
   *  \param value     The spectral location to convert
   *  \param oldUnits  Old units of \c value
   *  \param newUnits  New units for \c value
   */
  [[nodiscard]] double
  convertSpectralUnits(const double &value, const d5::SpectralUnits &oldUnits,
                       const d5::SpectralUnits &newUnits) const;

  /** Gets the spectral "location" at a given index in the bandpass
   *
   *  The user has the option to specify the spectral units to convert the
   *  location to.
   *
   *  \param index     Index in the bandpass
   *  \param reqUnits  If not UnknownUnits, the unit to which to convert the
   *                   spectral location.
   */
  [[nodiscard]] double
  getSpectralLocation(std::size_t index,
                      d5::SpectralUnits reqUnits = d5::UnknownUnits) const {
    assert(index < this->pointCount);
    double value = this->minimum + static_cast<double>(index) * this->delta;

    if (reqUnits == d5::UnknownUnits) {
      return value;
    }
    return convertSpectralUnits(value, this->units, reqUnits);
  }

  /// The spectral units defining the bandpass
  d5::SpectralUnits units = d5::UnknownUnits;

  /// The minimum for the spectral bandpass
  double minimum = -1;

  /// The maximum for the spectral bandpass
  double maximum = -1;

  /// The sampling delta for the spectral bandpass
  double delta = -1;

  /// The number of samples in the spectral bandpass
  std::size_t pointCount = 0;
};

/**
 * Converts a value between two spectral units.  If one of the two spectral
 * units in Unknown, the data value is simple returned.  This simulates
 * stripping the units information when converting to an Unknown and specify new
 * units when converting from an Unknown.
 *
 * \param    value      The scalar value.
 * \param    oldUnits   The old units associated with scalar.
 * \param    newUnits   The new units associated with scalar.
 *
 * \return   The scalar value with the new units.
 */
inline double
Bandpass::convertSpectralUnits(const double &value,
                               const d5::SpectralUnits &oldUnits,
                               const d5::SpectralUnits &newUnits) const {
  // If the units are the same or either one is Unknown,
  // simply return the value
  if ((oldUnits == newUnits) || (oldUnits == d5::UnknownUnits) ||
      (newUnits == d5::UnknownUnits)) {
    return value;
  }

  // Create a variable for the answer
  double ans = 0.0;
  if (oldUnits == d5::NanometerUnits) {
    if (newUnits == d5::MicronUnits) {
      ans = value / 1000.;
    } else if (newUnits == d5::WavenumberUnits) {
      ans = 10000000.0 / value;
    } else if (newUnits == d5::GHzUnits) {
      ans = d5::speedOfLight / value;
    } else if (newUnits == d5::AngstromUnits) {
      ans = 10. * value;
    }
  } else if (oldUnits == d5::MicronUnits) {
    if (newUnits == d5::NanometerUnits) {
      ans = value * 1000.;
    } else if (newUnits == d5::WavenumberUnits) {
      ans = 10000.0 / value;
    } else if (newUnits == d5::GHzUnits) {
      ans = d5::speedOfLight / value / 1000.0;
    } else if (newUnits == d5::AngstromUnits) {
      ans = 10000.0 * value;
    }
  } else if (oldUnits == d5::WavenumberUnits) {
    if (newUnits == d5::NanometerUnits) {
      ans = 10000000.0 / value;
    } else if (newUnits == d5::MicronUnits) {
      ans = 10000.0 / value;
    } else if (newUnits == d5::GHzUnits) {
      ans = d5::speedOfLight * 1e-7 * value;
    } else if (newUnits == d5::AngstromUnits) {
      ans = 100000000.0 / value;
    }
  } else if (oldUnits == d5::GHzUnits) {
    if (newUnits == d5::NanometerUnits) {
      ans = d5::speedOfLight / value;
    } else if (newUnits == d5::MicronUnits) {
      ans = d5::speedOfLight / value / 1000.0;
    } else if (newUnits == d5::WavenumberUnits) {
      ans = 1e7 * value / d5::speedOfLight;
    } else if (newUnits == d5::AngstromUnits) {
      ans = 10. * d5::speedOfLight / value;
    }
  } else if (oldUnits == d5::AngstromUnits) {
    if (newUnits == d5::NanometerUnits) {
      ans = 0.1 * value;
    } else if (newUnits == d5::MicronUnits) {
      ans = value / 10000.0;
    } else if (newUnits == d5::WavenumberUnits) {
      ans = 1.0e8 / value;
    } else if (newUnits == d5::GHzUnits) {
      ans = 10. * d5::speedOfLight / value;
    }
  } else {
    std::string errorMsg;
    errorMsg = "unknown spectral units conversion";
    throw std::runtime_error(errorMsg);
  }
  return ans;
}

D5_TO_JSON(Bandpass, j, bandpass) {
  j = d5::Json{{"units", bandpass.units},
               {"minimum", bandpass.minimum},
               {"maximum", bandpass.maximum},
               {"delta", bandpass.delta},
               {"point_count", bandpass.pointCount}};
}

/// \}

} // namespace d5
