/** Interface file for BinaryStream
 *
 *  \file BinaryStream.h
 */

#pragma once

#include <cstddef>
#include <streambuf>
#include <string>
#include <vector>

namespace d5 {

/// \addtogroup common
/// \{

/** Wrapper around a binary stream object to track read/write stats and a
 *  strongly-typed I/O interface.
 */
template <typename Stream> struct BinaryStream {
  /** Constructor
   *
   *  \param s  The stream to wrap, caller maintains ownership
   */
  BinaryStream(Stream *s) : stream(s) {
    unsigned long ul{1};
    unsigned char uc{};
    std::memcpy(&uc, &ul, 1);
    isLittleEndian = (uc == 1);
  }

  /** Reads an array of values from the stream
   *
   *  \tparam Value          The data type of the values to read
   *  \param[in,out] values  Buffer into which to read
   *  \param[in] count       The number of values to read
   */
  template <typename Value> void read(Value *values, std::size_t count) {
    stream->read(reinterpret_cast<char *>(values), sizeof(Value) * count);
    if constexpr (sizeof(Value) > 1) {
      if (!isLittleEndian) {
        for (std::size_t i = 0; i < count; i++) {
          std::reverse(reinterpret_cast<unsigned char *>(values + i),
                       reinterpret_cast<unsigned char *>(values + i + 1));
        }
      }
    }
    bytesRead += sizeof(Value) * count;
  }

  /** Writes an array of values to the stream
   *
   *  \tparam Value      The data type of the values to write
   *  \param[in] values  Buffer from which to write
   *  \param[in] count   The number of values to write
   */
  template <typename Value> void write(const Value *values, std::size_t count) {
    if constexpr (sizeof(Value) > 1) {
      if (!isLittleEndian) {
        Value value{};
        for (std::size_t i = 0; i < count; i++) {
          value = values[i];
          std::reverse(reinterpret_cast<unsigned char *>(&value),
                       reinterpret_cast<unsigned char *>(&value + 1));
          stream->write(reinterpret_cast<const char *>(&value), sizeof(Value));
          bytesWritten += sizeof(Value);
        }
        return;
      }
    }
    stream->write(reinterpret_cast<const char *>(values),
                  sizeof(Value) * count);
    bytesWritten += sizeof(Value) * count;
  }

  /** Read a single value from the stream
   *
   *  \tparam Value  The data type of the value to read
   *
   *  \return The read value
   */
  template <typename Value> Value read() {
    Value value{};
    read(&value, 1);
    return value;
  }

  /** Write a single value to the stream
   *
   *  \tparam Value  The data type of the value to write
   *  \param value   The value to write
   */
  template <typename Value> void write(const Value &value) { write(&value, 1); }

  /// The wrapped stream
  Stream *stream = nullptr;

  /// The total number of bytes read from the stream
  std::size_t bytesRead = 0;

  /// The total number of bytes written to the stream
  std::size_t bytesWritten = 0;

  bool isLittleEndian{true};
};

/// Deduction guide for BinaryStream
template <typename Stream> BinaryStream(Stream *) -> BinaryStream<Stream>;

/// An input stream buffer wrapping a binary data block.
struct RawIStreambuf : std::streambuf {
  RawIStreambuf(void *buf, std::size_t bufLen) {
    setg(static_cast<char *>(buf), static_cast<char *>(buf),
         static_cast<char *>(buf) + bufLen);
  }
  template <typename Char, typename... Args>
  explicit RawIStreambuf(std::vector<Char, Args...> &buf)
      : RawIStreambuf(buf.data(), sizeof(Char) * buf.size()) {}
  template <typename Char, typename... Args>
  explicit RawIStreambuf(std::basic_string<Char, Args...> &buf)
      : RawIStreambuf(buf.data(), sizeof(Char) * buf.size()) {}
};

/// An output stream buffer wrapping a binary data block.
struct RawOStreambuf : std::streambuf {
  RawOStreambuf(void *buf, std::size_t bufLen) {
    setp(static_cast<char *>(buf), static_cast<char *>(buf) + bufLen);
  }
  template <typename Char, typename... Args>
  explicit RawOStreambuf(std::vector<Char, Args...> &buf)
      : RawOStreambuf(buf.data(), sizeof(Char) * buf.size()) {}
  template <typename Char, typename... Args>
  explicit RawOStreambuf(std::basic_string<Char, Args...> &buf)
      : RawOStreambuf(buf.data(), sizeof(Char) * buf.size()) {}
};

/// \}

} // namespace d5
