/** Interface file for blackbody calculations
 *
 *  \file Blackbody.h
 */

#pragma once

#include "dirsig5/base/SpectralVector.h"

namespace d5 {

/// \addtogroup common
/// \{

/** Computes the Planckian blackbody spectral irradiance for the supplied
 *  temperature and wavelength.
 *
 *  \param tempK             The temperature of the blackbody [K]
 *  \param wavelengthMicron  The wavelength [um]
 *
 *  \return The spectral irradiance at \c wavelengthMicron [W/cm^2/um]
 */
constexpr double BlackbodyIrradianceWcm2um(double tempK,
                                           double wavelengthMicron) {
  if (tempK <= 0)
    return 0;

  constexpr double C1m = 3.741832E-20; // [Watt cm^-2 m^1]
  constexpr double C2m = 1.438786E-02; // [m^-1 K]

  double lm = wavelengthMicron * 1.0e-06;

  double lm5 = lm * lm * lm * lm * lm; // manually compute lm^5

  // Planck's equation
  double E = C1m / (lm5 * (std::exp(C2m / (lm * tempK)) - 1.0));

  // convert the value from W/(cm^2 m) to W/(cm^2 um)
  E *= 1.0e-06;
  return E;
}

/** Generates a Planck blackbody spectral irradiance curve for the supplied
 *  temperature and wavelengths.
 *
 *  \param tempK             The temperature of the blackbody [K]
 *  \param wavelengthMicron  The list of wavelengths [um]
 *
 *  \return The spectral vector of irradiances at each wavelength in
 *          \c wavelengthMicron [W/cm^2/um]
 */
inline SpectralVector
BlackbodyIrradianceWcm2um(double tempK,
                          const std::vector<double> &wavelengthMicron) {
  // loop through each spectral point in the vector
  SpectralVector result(wavelengthMicron.size());
  for (size_t ii = 0; ii < wavelengthMicron.size(); ii++) {
    result[ii] = BlackbodyIrradianceWcm2um(tempK, wavelengthMicron[ii]);
  }

  return result;
}

/// \}

} // namespace d5
