#pragma once

#include <string>
#include <vector>

#include "dirsig5/base/Export.h"

namespace d5 {

/** Manages the DIRSIG cache directory
 *  \ingroup utility
 */
class D5_BASE_EXPORT CacheDirectory {
public:
  /** Constructor.
   *
   *  \param dirPath  Directory path to use. If empty, a default directory
   *                  path is generated.
   *  \param noPurge  Disable purging of the cache directory
   *
   *  \throw std::runtime_error  If `dir` is not empty and does not name a valid
   *                             directory path.
   *
   *  \note
   *  If `dir` is empty, the implementation first looks for an
   *  environment variable `DIRSIG_CACHE_DIR`, such that the user can
   *  specify the location of the cache. If this variable does not
   *  exist or is invalid, then the implemention tries to use a
   *  default location in the user documents folder. If this fails (it
   *  shouldn't), then the implementation defaults to the current
   *  working directory.
   */
  CacheDirectory(const std::string &dirPath = {}, bool noPurge = false);

  /// Read-only access to the cache directory path.
  [[nodiscard]] const std::string &directoryPath() const {
    return directoryPath_;
  }

  /// Contains file with given filename?
  [[nodiscard]] bool contains(const std::string &filename) const;

  /// Read data.
  [[nodiscard]] std::string read(const std::string &filename) const;

  /// Write data.
  void write(const std::string &filename, const std::string &data) const;

  /// Remove files from the cache directory.
  void purge(const std::vector<std::string> &patterns = {});

private:
  /// Directory path.
  std::string directoryPath_;

  /// Purge disabled? (happens if directory path defaults to pwd)
  bool purgeDisabled_ = false;
};

} // namespace d5
