#pragma once

#include <algorithm>
#include <fstream>
#include <string>
#include <vector>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/Json.h"

namespace d5 {

/// A structure to store information about a clock, which can be
/// independent, dependent or externally driven.
struct Clock {
  enum Type {
    UnknownClock,     ///< The clock has an unknown/invalid state
    IndependentClock, ///< The clock has an independent rate
    DependentClock,   ///< The clock has a rate dependent on the master
    ExternalClock     ///< The clock uses triggers from an external file
  };

  /// Make this an independent clock using the supplied rate and offset.
  void setAsIndependent(double rate, double offset = 0) noexcept {
    type = IndependentClock;
    independentRate = rate;
    independentOffset = offset;
  }

  /// Make this a dependent clock using the supplied divider and offset.
  void setAsDependent(double divider, int offset = 0) noexcept {
    type = DependentClock;
    masterClockDivider = divider;
    masterClockOffset = offset;
  }

  /// A mutator to make this an external clock
  void setAsExternal(const std::string &fname) {
    type = ExternalClock;
    filename = fname;
  }

  /// Sets the master clock rate
  ///
  /// If this clock is a dependent clock, then we compute the clock
  /// rate using the clock divider.  Otherwise, we just stick with the
  /// independent clock rate.
  void setMasterClockRate(double rate) {
    switch (type) {
    case DependentClock:
      DIRSIG_THROW_IF(masterClockDivider == 0,
                      "Invalid master clock divider for dependent clock!");
      clockRate = rate / masterClockDivider;
      clockOffset = masterClockOffset / rate;
      break;
    case IndependentClock:
      clockRate = independentRate;
      clockOffset = independentOffset;
      break;
    default:
      break;
    }
  }

  /// Generate a clock triggers for a given time window.
  [[nodiscard]] std::vector<double> generateTriggers(double start,
                                                     double stop) const;

  /// The clock type (e.g. independent, dependent, external, etc. )
  Type type = UnknownClock;

  /// The current clock rate [1/s]
  double clockRate = 0;

  /// The current clock offset [1]
  double clockOffset = 0;

  /// The independent clock rate [1/s]
  double independentRate = -1;

  /// The independent clock offset [s]
  double independentOffset = 0;

  /// The master clock divider (only used for dependent/slave clocks)
  double masterClockDivider = -1;

  /// The master clock offset (only used for dependent/slave clocks)
  int masterClockOffset = 0;

  /// The name of the external file containing trigger times
  std::string filename;

  /// The relative trigger times from the external file
  std::vector<double> triggerTimes;
};

inline std::vector<double> Clock::generateTriggers(double start,
                                                   double stop) const {
  DIRSIG_THROW_IF(not(start <= stop), "Invalid window!");

  double duration = stop - start;
  std::vector<double> triggers;
  if (type == ExternalClock) {
    using namespace std::string_literals;
    auto ifs = d5::OpenIfstream(filename);

    // read all the relative triggers from the file
    std::vector<double> fileTimes;
    while (not ifs.eof()) {
      double triggerTime = 0;
      ifs >> triggerTime;
      if (ifs.fail())
        break;
      fileTimes.push_back(triggerTime);
    }
    ifs.close();

    // now generate (and possibly loop) the triggers over the requested window
    double timeOffset = 0.0;
    size_t timeIndex = 0;
    while (1) {
      // compute the current trigger time
      auto triggerTime = timeOffset + fileTimes[timeIndex];

      if (triggerTime < start) {
        // if the time is before the start, skip it
        ;
      } else if (triggerTime > stop) {
        // if the time is after the stop, bail out
        break;
      } else {
        // otherwise store the time
        triggers.push_back(triggerTime);
      }

      // if we reached the end of the sequence, restart and update the offset
      timeIndex += 1;
      if (timeIndex >= fileTimes.size()) {
        timeOffset += fileTimes.back();
        if (fileTimes.at(0) != 0.0) {
          // starting at the front is OK if the time isn't 0s
          timeIndex = 0;
        } else if (fileTimes.size() > 1) {
          // if the first time was 0s, we don't want to repeat a trigger
          timeIndex = 1;
        } else {
          break;
        }
      }
    }

    // sort all the triggers
    std::sort(triggers.begin(), triggers.end());
  } else {
    DIRSIG_THROW_IF(clockRate <= 0, "Invalid (negative or zero) clock rate!");

    // Generate all the triggers
    int triggerCount = int(1.0 + duration * clockRate);
    if (triggerCount == 0)
      triggerCount = 1;
    for (int tIndex = 0; tIndex < triggerCount; tIndex++) {
      double triggerTime = start + clockOffset + tIndex / clockRate;
      if (triggerTime >= start and triggerTime <= stop)
        triggers.push_back(triggerTime);
    }
  }
  return triggers;
}

/// Conversion from a Clock to JSON
inline void to_json(Json &j, const Clock &c) {
  switch (c.type) {
  case Clock::IndependentClock:
    j = Json{{"type", "independent"},
             {"rate", c.independentRate},
             {"offset", c.independentOffset}};
    break;
  case Clock::DependentClock:
    j = Json{{"type", "dependent"},
             {"master_clock_divider", c.masterClockDivider},
             {"master_clock_offset", c.masterClockOffset}};
    break;
  case Clock::ExternalClock:
    j = Json{{"type", "external"}, {"filename", c.filename}};
    break;
  default:
    j = nullptr;
    break;
  }
}

} // namespace d5
