/** Interface file for color filter array (CFA) patterns.
 *
 *  \file ColorFilterArrays.h
 */

#pragma once

#include <cstddef>
#include <string>

#include "dirsig5/base/Error.h"

namespace d5 {

/// \addtogroup common
/// \{

/**
 * This function defines the Bayer pattern filter layout on a 2D
 * detector array.  The caller supplies the X,Y index pair into the
 * 2D array and the function returns a channel index (0 for red, 1
 * for green and 2 for blue).  The pattern is captured in a 2 x 2
 * repeating pattern, with rows containing either alternating red/green
 * or alternating blue/green filters.  There are 4 possibilities for
 * how the pattern is started, which can be decribed by a 4 character
 * string indicating the 2 x 2 order:
 *
 * BGGR:        GBRG:        GRBG:        RGGB:
 * +---+---+    +---+---+    +---+---+    +---+---+
 * | B | G |    | G | B |    | G | R |    | R | G |
 * +---+---+    +---+---+    +---+---+    +---+---+
 * | G | R |    | R | G |    | B | G |    | G | B |
 * +---+---+    +---+---+    +---+---+    +---+---+
 *
 * We handle the different patterns by offsetting the supplied
 * index values and using a single 2 x 2 row/col logic.
 *
 * \return Returns 0 for red, 1 for green and 2 for blue.
 *
 */
[[nodiscard]] inline size_t bayerPattern(const std::string &pattern,
                                         size_t xIndex, size_t yIndex) {
  long row = yIndex;
  long col = xIndex;
  if (pattern == "BGGR") {
    row += 1;
    col += 1;
  } else if (pattern == "GBRG") {
    row += 1;
  } else if (pattern == "GRBG") {
    col += 1;
  } else if (pattern == "RGGB") {
    ;
  } else {
    DIRSIG_THROW("Invalid Bayer pattern!");
  }

  static const size_t filterPattern[2][2] = {{0, 1}, {1, 2}};

  return filterPattern[row % 2][col % 2];
}

/**
 * This function defines the TrueSense pattern filter layout on a 2D
 * detector array.  This pattern features "white" (unfiltered, pan, etc.),
 * Red, Green and Blue filters that can be modeled as a 4 x 4 repeating
 * pattern.
 *
 * +---+---+---+---+
 * | W | B | W | G |
 * +---+---+---+---+
 * | B | W | G | W |
 * +---+---+---+---+
 * | W | G | W | R |
 * +---+---+---+---+
 * | G | W | R | W |
 * +---+---+---+---+
 *
 * \returns Returns 0 for white, 1 for red, 2 for green and 3 for blue.
 */
[[nodiscard]] inline size_t truesensePattern(size_t xIndex, size_t yIndex) {
  static const size_t filterPattern[4][4] = {
      {0, 3, 0, 2}, {3, 0, 2, 0}, {0, 2, 0, 1}, {2, 0, 1, 0}};

  return filterPattern[yIndex % 4][xIndex % 4];
}

/// \}

} // namespace d5
