#pragma once

#include <functional>
#include <list>
#include <map>
#include <optional>
#include <string>
#include <string_view>
#include <tuple>
#include <type_traits>
#include <utility>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Export.h"
#include "dirsig5/base/StringUtils.h"

namespace d5 {

namespace detail {

template <typename... T> struct TypeList {
  template <size_t N>
  using type = typename std::tuple_element<N, std::tuple<T...>>::type;
};
template <typename T, typename U> struct TypeListCat {
  using type = TypeList<T, U>;
};
template <typename T, typename... U> struct TypeListCat<T, TypeList<U...>> {
  using type = TypeList<T, U...>;
};
template <typename... T, typename U> struct TypeListCat<TypeList<T...>, U> {
  using type = TypeList<T..., U>;
};
template <typename... T, typename... U>
struct TypeListCat<TypeList<T...>, TypeList<U...>> {
  using type = TypeList<T..., U...>;
};
template <size_t N, typename T> struct IndexedType {
  static constexpr size_t index = N;
  using type = T;
};
template <size_t N, typename... T> struct IndexedTypeList;
template <size_t N, typename T, typename... U>
struct IndexedTypeList<N, T, U...> {
  using type =
      typename TypeListCat<IndexedType<N, T>,
                           typename IndexedTypeList<N + 1, U...>::type>::type;
};
template <size_t N, typename T> struct IndexedTypeList<N, T> {
  using type = TypeList<IndexedType<N, T>>;
};
template <> struct IndexedTypeList<0> {
  using type = TypeList<>;
};
template <typename... T>
using IndexedTypeList_t = typename IndexedTypeList<0, T...>::type;

template <typename T>
struct CallableIndexedTypeList
    : CallableIndexedTypeList<decltype(&T::operator())> {};
template <typename T, typename... Args>
struct CallableIndexedTypeList<T(Args...)> {
  using type = IndexedTypeList_t<Args...>;
};
template <typename T, typename... Args>
struct CallableIndexedTypeList<T (*)(Args...)> {
  using type = IndexedTypeList_t<Args...>;
};
template <typename T, typename Class, typename... Args>
struct CallableIndexedTypeList<T (Class:: *)(Args...)> {
  using type = IndexedTypeList_t<Args...>;
};
template <typename T, typename Class, typename... Args>
struct CallableIndexedTypeList<T (Class:: *)(Args...) const> {
  using type = IndexedTypeList_t<Args...>;
};
template <typename T>
struct CallableIndexedTypeList<T &> : CallableIndexedTypeList<T> {};
template <typename T>
struct CallableIndexedTypeList<T &&> : CallableIndexedTypeList<T> {};
template <typename T>
using CallableIndexedTypeList_t = typename CallableIndexedTypeList<T>::type;

template <typename T> struct RemoveOptional {
  using type = T;
};
template <typename T> struct RemoveOptional<std::optional<T>> {
  using type = T;
};
template <typename T> using RemoveOptional_t = typename RemoveOptional<T>::type;

} // namespace detail

class D5_BASE_EXPORT CommandLine {
public:
  CommandLine(bool isTopLevel = true);

  void ignoreUnknown() { mIgnoreUnknown = true; }

private:
  template <typename Arg>
  [[nodiscard]] static constexpr const char *ArgType() noexcept {
    if constexpr (std::is_unsigned_v<Arg>)
      return "uint";
    else if constexpr (std::is_integral_v<Arg>)
      return "int";
    else if constexpr (std::is_floating_point_v<Arg>)
      return "float";
    else
      return "string";
  }

  template <typename Arg> [[nodiscard]] static Arg ArgCast(const char *argv) {
    static_assert(std::is_arithmetic_v<Arg> ||         //
                  std::is_same_v<Arg, char *> ||       //
                  std::is_same_v<Arg, const char *> || //
                  std::is_same_v<Arg, std::string> ||  //
                  std::is_same_v<Arg, std::string_view>);
    if constexpr (std::is_arithmetic_v<Arg>)
      return ParseNumber<Arg>(std::string_view(argv));
    else
      return argv;
  }

public:
  struct Argument {
    bool isRequired{false};
    int argc{0};
    std::string names[2]{};
    std::string typeString;
    std::string helpString;
    std::function<void(char **)> onArgument;
    bool isPresent{false};
    bool isNumericOnly{false};

    [[nodiscard]] std::string niceName() const {
      if (names[1].empty())
        return std::string(names[0]);
      else
        return std::string(names[0]) + "/" + std::string(names[1]);
    }

  private:
    template <typename Handler>
    Argument &initialize(bool required, const char *name, Handler &&handler) {
      isRequired = required;
      names[0] = name;
      auto pos = names[0].find('/');
      if (pos != std::string_view::npos) {
        names[1] = names[0].substr(pos + 1);
        names[0] = names[0].substr(0, pos);
      }
      return initialize(std::forward<Handler>(handler),
                        detail::CallableIndexedTypeList_t<Handler>{});
    }
    template <typename Handler, typename... Args>
    Argument &initialize(Handler &&handler, detail::TypeList<Args...>) {
      argc = sizeof...(Args);
      if constexpr (sizeof...(Args) > 0) {
        constexpr const char *ArgTypes[] = {
            ArgType<std::decay_t<typename Args::type>>()...};
        typeString = ArgTypes[0];
        for (auto itr = &ArgTypes[0] + 1; itr < &ArgTypes[0] + argc; itr++) {
          typeString += ' ';
          typeString += *itr;
        }
        isNumericOnly =
            (std::is_arithmetic_v<std::decay_t<typename Args::type>> && ...);
      }
      onArgument = [&, handler](char **argv) {
        DIRSIG_ERROR_TRACE("handling argument " + niceName(), {
          (void)argv;
          handler(
              ArgCast<std::decay_t<typename Args::type>>(argv[Args::index])...);
        });
      };
      return *this;
    }
    friend class CommandLine;
  };

  struct Command {
    struct ArgumentGroup {
      std::string name;
      std::string helpString;
      std::vector<std::list<Argument>::iterator> arguments;
    };
    std::string name;
    std::vector<ArgumentGroup> argumentGroups;
    std::list<Argument> arguments;
    std::list<Argument> requiredPositionalArguments;
    std::list<Argument>::iterator requiredPositionalArgumentItr;
    std::string defaultPositionalName;
    std::string defaultPositionalHelpString;
    std::function<void(std::string_view)> defaultPositional;
    std::function<void()> onReset;
    std::function<void()> onExecute;
    std::string description;
    /// For backward compatibility with DIRSIG5 component mode.
    bool dontRequirePositionalArguments{false};

    void reset();
    void execute();
    void printHelpAndExit(CommandLine &owner);
  };

public:
  void command(std::string_view name) {
    for (auto itr = mCommands.begin(); itr != mCommands.end(); ++itr)
      if (itr->name == name || (itr->name.empty() && name.empty())) {
        mCommands.splice(mCommands.end(), mCommands, itr);
        return;
      }
    mCommands.emplace_back();
    mCommands.back().name = std::string(name);
    mCommands.back().argumentGroups.emplace_back();
    optionalArgument("-h/--help", [&, commandPtr = &mCommands.back()] {
      commandPtr->printHelpAndExit(*this);
    }) = "Display this help and exit.";
  }

  void describe(std::string_view description) {
    mCommands.back().description = description;
  }

  void argumentGroup(const char *name) {
    mCommands.back().argumentGroups.emplace_back().name = name;
  }

  template <bool IsRequired, typename Handler>
  std::string &argument(const char *name, Handler &&handler) {
    auto &arg = mCommands.back().arguments.emplace_back().initialize(
        IsRequired, name, std::forward<Handler>(handler));
    mCommands.back().argumentGroups.back().arguments.emplace_back(
        --mCommands.back().arguments.end());
    return arg.helpString;
  }

  template <bool IsRequired>
  std::string &argument(const char *name, bool *flag) {
    return argument<IsRequired>(name, [=] { *flag = true; });
  }

  template <bool IsRequired, typename T, typename... U>
  std::string &argument(const char *name, T *arg, U *...args) {
    return argument<IsRequired>(
        name, [=](detail::RemoveOptional_t<T> arg1,
                  detail::RemoveOptional_t<U>... args1) {
          std::tie(*arg, *args...) = std::make_tuple(arg1, args1...);
        });
  }

  template <typename... T>
  std::string &optionalArgument(const char *name, T &&...args) {
    return argument<false>(name, std::forward<T>(args)...);
  }

  template <typename... T>
  std::string &requiredArgument(const char *name, T &&...args) {
    return argument<true>(name, std::forward<T>(args)...);
  }

  template <typename Handler>
  std::string &requiredPositionalArgument(const char *name, Handler &&handler) {
    return mCommands.back()
        .requiredPositionalArguments.emplace_back()
        .initialize(true, name, std::forward<Handler>(handler))
        .helpString;
  }

  template <typename T>
  std::string &requiredPositionalArgument(const char *name, T *arg) {
    return requiredPositionalArgument(name, [=](T arg1) { *arg = arg1; });
  }

  std::string &defaultPositionalArguments(const char *name,
                                          std::vector<std::string> *args) {
    return defaultPositionalArguments(name, [=](std::string_view arg) {
      args->emplace_back(std::string(arg));
    });
  }

  template <typename Handler>
  std::string &defaultPositionalArguments(const char *name, Handler &&handler) {
    mCommands.back().defaultPositionalName = name;
    mCommands.back().defaultPositional = std::forward<Handler>(handler);
    return mCommands.back().defaultPositionalHelpString;
  }

  void dontRequirePositionalArguments() {
    mCommands.back().dontRequirePositionalArguments = true;
  }

  template <typename Handler> void onReset(Handler &&handler) {
    mCommands.back().onReset = std::forward<Handler>(handler);
  }

  template <typename Handler> void onExecute(Handler &&handler) {
    mCommands.back().onExecute = std::forward<Handler>(handler);
  }

  int run(int argc, char **argv);

private:
  const char *mProgramName{nullptr};

  bool mIsTopLevel{true};

  bool mIgnoreUnknown{false};

  std::list<Command> mCommands;

public:
  [[nodiscard]] const char *programName() const noexcept {
    return mProgramName;
  }

  [[nodiscard]] const std::list<Command> &commands() const noexcept {
    return mCommands;
  }
};

} // namespace d5
