/** Interface file for constant definitions
 *
 *  \file Constants.h
 */

#pragma once

#include <cmath>
#include <limits>
#include <string>

namespace d5 {

/** \defgroup common_constants Useful constants for DIRSIG code
 *  Physical, mathematical and geographic constants
 *  \ingroup common
 */

/// Namespace for constants, inlined for syntactic sugar
inline namespace constants {

/// \addtogroup common_constants
/// \{

/// A constant for PI
constexpr double PI = 3.141592653589793;

/// A constant for the inverse of PI
constexpr double IPI = 0.3183098861837907;

/// A constant for 2PI
constexpr double TwoPI = 2 * PI;

/// A constant for PI divided by two (PI/2)
constexpr double PI_2 = PI / 2;

/// A constant for PI divided by four (PI/4)
constexpr double PI_4 = PI / 4;

/// A constant for the square root of two
constexpr double SQRT_2 = 1.4142135623730951;

/// A constant for the inverse square root of two
constexpr double INV_SQRT_2 = 0.7071067811865475;

/// A constant for the square root of PI
constexpr double SQRT_PI = 1.7724538509055159;

/// A constant for PI squared
constexpr double PI_SQR = 9.869604401089358;

/// A constant for the speed of light in a vacuum [m/s]
constexpr double speedOfLight = 2.99792458e+08;

/// Standard acceleration due to gravity [m/s^2]
constexpr double g = 9.80665;

/// The Planck constant (NIST) [Js]
constexpr double planckConstant = 6.62606957e-34;

/// The Boltzmann constant [J/K]
constexpr double boltzmannConstant = 1.380658e-23;

/// The first radiation constant (NIST,2hc^2) [W/m2]
constexpr double rc1L = 3.741771790e-16;

/// The second radiation constant (NIST,hc/Kb) [mK]
constexpr double rc2 = 1.43877736e-2;

/// The mean earth radius (IUGG) [m]
constexpr double meanEarthRadius = 6371009; // [m]

// WGS84 Earth radius (equatorial) [m]
constexpr double WGS84EquatorialRadius = 6378137;

// WGS84 Flattening
constexpr double WGS84Flattening = 1.0 / (298.257223563);

// WGS84 Eccentricity
constexpr double EARTH_ASPECT_RATIO = 1.0 - WGS84Flattening;

// WGS83 Earth radius (polar) [m]
constexpr double WGS84PolarRadius = WGS84EquatorialRadius * EARTH_ASPECT_RATIO;

// lowest altitude/elevation (on ground) on earth (Dead Sea, approx.) [m]
constexpr double MinEarthAltitude = -431.0;

// highest altitude/elevation (on ground) on earth (Mt Everest) [m]
constexpr double MaxEarthAltitude = 8848.0;

/// The mean earth to sun distance [m]
constexpr double meanEarthSunDistance = 149600000.0e+03;

/// The mean earth to moon distance [m]
constexpr double meanEarthMoonDistance = 384400.0e+03;

/// The mean sun half angle on the earth [rad]
constexpr double meanSolarHalfAngle = 0.004655;

/// The mean moon half angle on the earth [rad]
constexpr double meanLunarHalfAngle = 0.004610;

/// The nominal refractive index of air at STP
constexpr double nAir = 1.0002926;

/// The nominal refractive index of void
constexpr double nVoid = 1.0;

/// The number of nanoseconds in a second
constexpr long nanoSecondsPerSecond = 1000000000;

/// The number of seconds in a day
constexpr long secondsPerDay = 86400;

/// The number of seconds in an hour
constexpr long secondsPerHour = 3600;

/// The number of seconds in a minute
constexpr long secondsPerMinute = 60;

/// The minimum hit distance (used for self intersect checks)
constexpr float MIN_HIT_DISTANCE = 1e-4f;

// The minimum offset from the surface (to try to avoid self-intersection)s
constexpr float SURFACE_EPS = 1e-3f;

// The minimum time epsilon [s]
constexpr double TimeEps = 1e-12;

// distance that we ignore scene-to-scene intersections
constexpr float minInterSceneDist = 0.1;

/// The separator for a search path (list of directories)
#ifdef WIN32
const std::string pathSep = ";";
#else
const std::string pathSep = ":";
#endif

/// \}
} // namespace constants

/// \addtogroup common
/// \{

/** Convert an angle in degrees to radians
 *
 *  \param deg  The angle in degrees
 *
 *  \return The angle in radians
 */
constexpr double toRadians(double deg) { return deg * (PI / 180.); }

/** Convert an angle in radians to degrees
 *
 *  \param rad The angle in radians
 *
 *  \return The angle in degrees
 */
constexpr double toDegrees(double rad) { return rad * (180. / PI); }

/** Dew point calculation (Magnus formula)
 *
 *  All based on Arden Buck (1981) for ranges above and below 0 deg C
 *
 *  \param Tc  The air temperature in C
 *  \param RH  The relative humidity [percent]
 *
 *  \return The dew point in degrees C
 */
inline double Tdp(double Tc, double RH) {
  constexpr double Tdp_b_pos = 17.368;
  constexpr double Tdp_c_pos = 238.88;
  constexpr double Tdp_b_neg = 17.966;
  constexpr double Tdp_c_neg = 247.15;
  double gamma = (Tc >= 0)
                     ? std::log(RH / 100.) + Tdp_b_pos * Tc / (Tdp_c_pos + Tc)
                     : std::log(RH / 100.) + Tdp_b_neg * Tc / (Tdp_c_neg + Tc);
  return (Tc >= 0) ? Tdp_c_pos * gamma / (Tdp_b_pos - gamma)
                   : Tdp_c_neg * gamma / (Tdp_b_neg - gamma);
}

/** Calculate the spectral refractive index of water based on the IAPWS
 *
 *  \param wave     The wavelength in microns
 *  \param temp     The temperature in Kelvin
 *  \param density  The density of the water in [kg/m^3]
 *
 *  \return The refractive index of the water
 */
inline double refractiveIndexIAPWS(double wave,             // [um]
                                   double temp = 273.15,    // [K]
                                   double density = 1000) { // [kg/m^3]
  constexpr double a0 = 0.244257733e0;
  constexpr double a1 = 9.75634476e-3;
  constexpr double a2 = -3.73234996e-3;
  constexpr double a3 = 2.68678472e-4;
  constexpr double a4 = 1.58920570e-3;
  constexpr double a5 = 2.45934259e-3;
  constexpr double a6 = 0.900704920e0;
  constexpr double a7 = -1.66626219e-2;
  constexpr double Luv2 = 0.2292029 * 0.2292020;
  constexpr double Lir2 = 5.432937 * 5.432937;
  constexpr double refWave = 0.589;     // [um]
  constexpr double refTemp = 273.15;    // [K]
  constexpr double refDensity = 1000.0; // [kg/m^3]
  double waveRatio = wave / refWave, waveRatio2 = waveRatio * waveRatio;
  double tempRatio = temp / refTemp;
  double densityRatio = density / refDensity;
  double aTerms = a0 + a1 * densityRatio + a2 * tempRatio +
                  a3 * waveRatio2 * tempRatio + a4 / waveRatio2 +
                  a5 / (waveRatio2 - Luv2) + a6 / (waveRatio2 - Lir2) +
                  a7 * densityRatio * densityRatio;
  double a = densityRatio * aTerms;
  return std::sqrt((1 + 2 * a) / (1 - a));
}
/// \}

} // namespace d5
