#pragma once

#include <complex>
#include <cstddef>
#include <cstdint>
#include <string_view>

namespace d5 {

/// Primitive data type enums.
///
/// These necessarily correspond to the integer codes used by ENVI.
///
enum class DType : uint8_t {
  Unknown = 0,
  Uint8 = 1,
  Int16 = 2,
  Int32 = 3,
  Float32 = 4,
  Float64 = 5,
  Complex64 = 6,
  Complex128 = 9,
  Uint16 = 12,
  Uint32 = 13,
  Int64 = 14,
  Uint64 = 16
};

template <DType T> struct dtype_enum_traits {};
template <typename T> struct dtype_type_traits {};

#define DTYPE_TRAITS(T, Size, Type)                                            \
  template <> struct dtype_enum_traits<DType::T> {                             \
    using type = Type;                                                         \
    static constexpr uint8_t size = Size;                                      \
  };                                                                           \
  template <> struct dtype_type_traits<Type> {                                 \
    static constexpr DType value = DType::T;                                   \
  };
DTYPE_TRAITS(Uint8, 1, uint8_t)
DTYPE_TRAITS(Int16, 2, int16_t)
DTYPE_TRAITS(Int32, 4, int32_t)
DTYPE_TRAITS(Float32, 4, float)
DTYPE_TRAITS(Float64, 8, double)
DTYPE_TRAITS(Complex64, 8, std::complex<float>)
DTYPE_TRAITS(Complex128, 16, std::complex<double>)
DTYPE_TRAITS(Uint16, 2, uint16_t)
DTYPE_TRAITS(Uint32, 4, uint32_t)
DTYPE_TRAITS(Int64, 8, int64_t)
DTYPE_TRAITS(Uint64, 8, uint64_t)
#undef DTYPE_TRAITS

/// Returns whether the data type is complex
constexpr bool IsComplex(DType dtype) noexcept {
  return dtype == DType::Complex64 || dtype == DType::Complex128;
}

/// Convenience alias for using type traits for getting the size of the data
template <DType T> constexpr uint8_t ByteSize() noexcept {
  return dtype_enum_traits<T>::size;
}

/// Returns the size of the given data type in bytes
constexpr uint8_t ByteSize(DType dtype) noexcept {
#define DTYPE_CASE(T)                                                          \
  case DType::T:                                                               \
    return ByteSize<DType::T>()
  switch (dtype) {
    DTYPE_CASE(Uint8);
    DTYPE_CASE(Int16);
    DTYPE_CASE(Int32);
    DTYPE_CASE(Float32);
    DTYPE_CASE(Float64);
    DTYPE_CASE(Complex64);
    DTYPE_CASE(Complex128);
    DTYPE_CASE(Uint16);
    DTYPE_CASE(Uint32);
    DTYPE_CASE(Int64);
    DTYPE_CASE(Uint64);
  default:
    break;
  }
#undef DTYPE_CASE
  return 0;
}

template <DType T> using dtype_t = typename dtype_enum_traits<T>::type;

template <typename T> constexpr auto dtype_v = dtype_type_traits<T>::value;

constexpr std::string_view to_string(DType dtype) noexcept {
#define DTYPE_CASE(T)                                                          \
  case DType::T:                                                               \
    return #T
  switch (dtype) {
    DTYPE_CASE(Uint8);
    DTYPE_CASE(Int16);
    DTYPE_CASE(Int32);
    DTYPE_CASE(Float32);
    DTYPE_CASE(Float64);
    DTYPE_CASE(Complex64);
    DTYPE_CASE(Complex128);
    DTYPE_CASE(Uint16);
    DTYPE_CASE(Uint32);
    DTYPE_CASE(Int64);
    DTYPE_CASE(Uint64);
  default:
    break;
  }
#undef DTYPE_CASE
  return {};
}

} // namespace d5
