/** Interface file of DateTime
 *
 *  \file DateTime.h
 */

#pragma once

#include <array>
#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstring>
#include <ctime>
#include <numeric>

#include "dirsig5/base/Constants.h"
#include "dirsig5/base/StringUtils.h"

namespace d5 {

/// \addtogroup common
/// \{

/** Stores the date and time down to nanosecond precision.
 *
 *  The class has been constructed to be very useful including the support of
 *  mathematical operations such as comparisons (A less than B, A equal to B,
 *  etc.) and arithmetic operations including addition, subtraction,
 *  multiplication and division.
 *
 *  This single class supports both absolute date/times and relative
 *  date/times. If the date/time is absolute, then this date/time is
 *  referenced to the start of the Epoch. If not, then it is relative to some
 *  arbitrary time. This latter case only has use for performing some basic
 *  math operations. For example, the user wants to know the absolute
 *  date/time that is N seconds later than a reference date/time.
 *  The current date is an absolute date/time and the offset into the future
 *  is a relative date/time. It might seem confusing to mix both of these
 *  capabilities in one object, but it simplifies the ability to perform
 *  common operations. To ensure that these two aspects of this class are kept
 *  separate, the accessors and mutators have been carefully defined.
 *
 *  This class also supports Daylight Savings Time which is internally treated
 *  simply as an additional offset to the UTC reference.
 *
 *  This class also has some support for ISO 8601 Date/Time representations.
 *  For example, the time string for January 3, 2001 at 2:24 PM and 5.67
 *  seconds in the EST timezone (UTC -5) is: 2001-01-03T14:24:5.67-05:00
 *  Although the ISO 8601 format allows other representations of this date
 *  and time, DIRSIG only supports the above fully extended format.
 *
 *  This class works by using only a handful of variables.  The date and
 *  time is stored as absolute seconds since the Epoch, but split into a
 *  pair of 64-bit integer values -- one for the integer seconds since the
 *  Epoch and one for the fractional seconds.
 */
class DateTime {
public:
  /// Flags for the various date/time attributes
  enum DateTypeFlag {
    Valid = 0x01,          ///< The date/time has been set
    Relative = 0x02,       ///< The time is relative, not absolute
    DaylightSavings = 0x04 ///< The time is in daylight savings
  };

public:
  /// Default constructor, sets absolute to the Epoch (UTC).
  constexpr DateTime() noexcept = default;

  /// Copy constructor
  constexpr DateTime(const DateTime &) noexcept = default;

  /// Absolute seconds from the Epoch constructor
  constexpr DateTime(double seconds) noexcept { setAbsoluteSeconds(seconds); }

  /// Clear the current date and time, resets to the Epoch
  constexpr void clear() noexcept {
    attributes = 0;
    intSeconds = 0;
    fracSeconds = 0;
    utcHourOffset = 0;
  }

  /// Checks if the instance is storing a "valid" date/time
  [[nodiscard]] constexpr bool isValid() const noexcept {
    return attributes & Valid;
  }

  /// Checks if the instance is storing an "absolute" time
  [[nodiscard]] constexpr bool isAbsolute() const noexcept {
    return (attributes & Relative) == 0;
  }

  /// Checks if the instance is storing a "relative" time.
  [[nodiscard]] constexpr bool isRelative() const noexcept {
    return (attributes & Relative) != 0;
  }

  /// Set the absolute date and time.
  void set(int year, int month, int day, int hour, int minute, double seconds,
           double utcOffset, bool /*asDaylightSavings*/ = false) {
    attributes = Valid;
    intSeconds = dateToSeconds(day, month, year);
    intSeconds += hour * 3600 + minute * 60 + int(seconds);
    fracSeconds = (seconds - int(seconds)) * nanoSecondsPerSecond;
    utcHourOffset = utcOffset;
    intSeconds -= int64_t(utcOffset * secondsPerHour);
  }

  /// Set the absolute date and time (with fractional hours)
  void set(int year, int month, int day, double fracHours, double utcOffset,
           bool asDaylightSavings = false) {
    auto hours = static_cast<int>(fracHours);
    fracHours -= hours;
    fracHours *= 60.0;
    auto minutes = static_cast<int>(fracHours);
    fracHours -= minutes;
    fracHours *= 60.0;
    auto seconds = static_cast<int>(fracHours);
    set(year, month, day, hours, minutes, seconds, utcOffset,
        asDaylightSavings);
  }

  /// Set date/time using absolute seconds.
  constexpr void setAbsoluteSeconds(double seconds) noexcept {
    attributes = Valid;
    intSeconds = int64_t(seconds);
    fracSeconds = int64_t((seconds - intSeconds) * nanoSecondsPerSecond);
  }

  /// Set date/time using relative seconds.
  constexpr void setRelativeSeconds(double seconds) noexcept {
    setAbsoluteSeconds(seconds);
    attributes |= Relative;
  }

  /** Set the time with an ISO 8601 string.  This routine can be used
   *  if the user supplies an ASCII string where the format follows one of
   *  the excepted ISO 8601 Date/Time representations.  An example time
   *  string for January 3, 2001 at 2:24 PM and 5.6078 seconds in the EST
   *  timezone (UTC -5) is: 2001-01-03T-5:14:24:5,6078.
   */
  void setIso8601(const std::string &iso8601) try {
    TokenList dateTimeTokens(iso8601, "T");
    TokenList dateTokens(dateTimeTokens.at(0), "-");
    TokenList localOffsetTokens(dateTimeTokens.at(1), "+-Z");
    TokenList localTimeTokens(localOffsetTokens.at(0), ":");
    int year = ParseNumber<int>(dateTokens.at(0));
    int month = ParseNumber<int>(dateTokens.at(1));
    int day = ParseNumber<int>(dateTokens.at(2));
    int hour = ParseNumber<int>(localTimeTokens.at(0));
    int minute = ParseNumber<int>(localTimeTokens.at(1));
    auto second = ParseNumber<double>(localTimeTokens.at(2));
    double offset = 0;
    if (localOffsetTokens.size() == 2) {
      TokenList utcOffsetTokens(localOffsetTokens.at(1), ":");
      offset += ParseNumber<int>(utcOffsetTokens.at(0));
      offset += ParseNumber<int>(utcOffsetTokens.at(1)) / 60.0;
      if (*(localOffsetTokens.at(1).begin() - 1) == '-')
        offset = -offset;
    }
    set(year, month, day, hour, minute, second, offset);
  } catch (const std::exception &except) {
    DIRSIG_THROW("Can't parse an ISO8601 date/time from '" + iso8601 + "'")
        .withDebugMessage(except.what());
  }

  /// Get the fractional part of the seconds.
  [[nodiscard]] double getFractionalSeconds() const {
    return static_cast<double>(fracSeconds) /
           static_cast<double>(nanoSecondsPerSecond);
  }

  /// Get the total integer number of seconds (without the fractional part)
  [[nodiscard]] int64_t getTotalIntegerSeconds() const { return intSeconds; }

  /// Get the total number of seconds including the fractional part.
  [[nodiscard]] double getTotalSeconds() const {
    return static_cast<double>(intSeconds) + getFractionalSeconds();
  }

  /** Get the calender date (day, month, year).
   *
   *  \return The remaining fractional seconds in the day.
   */
  double getCalenderDate(int &day, int &month, int &year,
                         int &dayOfYear) const {
    assert(isAbsolute());
    return secondsToDate(intSeconds + int64_t(utcHourOffset * 3600), day, month,
                         year, dayOfYear) +
           getFractionalSeconds();
  }

  /// Get the day of year between 1 and 365 (or 366 on leap years).
  [[nodiscard]] int getDayOfYear() const {
    assert(isAbsolute());
    int day, month, year, dayOfYear;
    getCalenderDate(day, month, year, dayOfYear);
    return dayOfYear;
  }

  /** Convert the date/time to the Julian Date JD), which is the number
   *  of fractional days since 12:00 January 1, 4713 BC (Monday).  The
   *  calculation is adapted from the one outlined on Wolfram Alpha.
   *  See https://scienceworld.wolfram.com/astronomy/JulianDate.html.
   */
  [[nodiscard]] double toJulianDate() const {
    assert(isAbsolute());
    int day, month, year, dayOfYear;
    double seconds = getCalenderDate(day, month, year, dayOfYear);
    seconds -= getOffset() * d5::secondsPerHour;
    int hour = (int)(seconds / d5::secondsPerHour);
    seconds -= hour * d5::secondsPerHour;
    int minute = (int)(seconds / d5::secondsPerMinute);
    seconds -= minute * d5::secondsPerMinute;
    double utcTime = hour + minute/60.0 + seconds;
    double jd = 367*year 
        - int(7*(year + int((month+9)/12.0))/4)
        - int(3*(int((year+(month-9)/7)/100)+1)/4)
        + int(275*month/9.0) + day + 1721028.5 + utcTime/24.0;
    return jd;
  }

  /** Convert the date/time to a Modified Julian Date (MJD), which is
   *  the number of fractional days since midnight November 17, 1858.
   *  This date corresponds to 2400000.5 days after day 0 of the Julian
   *  calendar.
   */
  [[nodiscard]] double toModifiedJulianDate() const {
    assert(isAbsolute());
    return toJulianDate() - 2400000.5;
  }

  /** Convert the date/time to seconds since the J2000 epoch.
   *  The J2000 epoch is Jan 1, 2000 11:58:55.815683 UTC.
   */
  [[nodiscard]] double toJ2kSeconds() const {
    assert(isAbsolute());
    auto j2000Epoch = fromIso8601("2000-01-01T11:58:55.815683+00:00");
    return this->getTotalSeconds() - j2000Epoch.getTotalSeconds();
  }

  /** Return the date/time in the UTC zone.
   *  Note that the UTC offset is captured in the "intSeconds" and the
   *  "utcHourOffset" tracks what it was in case the user wants to convert
   *  back to a representation with it preserved. Hence, this function
   *  simply needs to zero out the stored offset.
   */
  [[nodiscard]] DateTime toUtc() const noexcept {
    assert(isAbsolute());
    DateTime ans;
    ans.attributes = attributes;
    ans.intSeconds = intSeconds;
    ans.fracSeconds = fracSeconds;
    ans.utcHourOffset = 0.f;
    return ans;
  }

  /// Get the GMT/UTC offset.
  [[nodiscard]] double getOffset() const {
    return utcHourOffset;
  }

  /// Get an ISO 8601 string from the date/time.
  [[nodiscard]] std::string toIso8601() const {
    char str[64] = {};
    if (isAbsolute()) {
      int day, month, year, dayOfYear;
      int second = getCalenderDate(day, month, year, dayOfYear);
      int hour = second / 3600;
      second -= hour * 3600;
      int minute = second / 60;
      second -= minute * 60;
      int utcHour = utcHourOffset;
      int utcMinute = std::abs(utcHourOffset - utcHour) * 60;
      std::snprintf(str, 64, "%04d-%02d-%02dT%02d:%02d:%07.4f%+03d:%02d", year,
                    month, day, hour, minute, second + getFractionalSeconds(),
                    utcHour, utcMinute);
    } else {
      int second = intSeconds;
      int hour = second / secondsPerHour;
      second -= hour * 3600;
      int minute = second / 60;
      second -= minute * 60;
      std::snprintf(str, 64, "%02d:%02d:%07.4f", hour, minute,
                    second + getFractionalSeconds());
    }
    return &str[0];
  }

  /** Make another date/time absolute, assuming this date/time is absolute.
   *
   *  This is a useful idiom for ensuring a date/time is absolute with respect
   *  to a reference date/time.
   */
  [[nodiscard]] DateTime makeAbsolute(const DateTime &other) const {
    assert(isAbsolute());
    return other.isAbsolute() ? other : operator+(other);
  }

  /** Measure relative seconds of another date/time with respect to this
   *  date/time, assuming this date/time is absolute.
   *
   *  This is a useful idiom for converting absolute date/times to relative
   *  date/times with respect to a reference. If `other` is absolute, subtracts
   *  this from it and returns the total seconds. If `other` is relative,
   *  returns the total seconds unchanged. For example,
   *  ```
   *  double elapsed = refTime.measureRelativeSeconds(keyframeTime);
   *  ```
   */
  [[nodiscard]] double measureRelativeSeconds(const DateTime &other) const {
    assert(isAbsolute());
    return other.isAbsolute() ? (other - *this).getTotalSeconds()
                              : other.getTotalSeconds();
  }

  /// Constructs an instance from absolute seconds.
  static DateTime fromAbsoluteSeconds(double seconds) { return seconds; }

  /// Constructs an instance from relative seconds.
  static DateTime fromRelativeSeconds(double seconds) {
    DateTime dateTime;
    dateTime.setRelativeSeconds(seconds);
    return dateTime;
  }

  /// Constructs an instance from an ISO 8601 string.
  static DateTime fromIso8601(const std::string &iso8601) {
    DateTime dateTime;
    dateTime.setIso8601(iso8601);
    return dateTime;
  }

  /// Returns a date/time representing the current time
  static DateTime now() { return fromAbsoluteSeconds(std::time(nullptr)); }

  /** Tests if year is a leap year.
   *
   *  Leap years are every 4 years, however
   *  since a year is actually 365.24 days (and not 365.25 days) a year
   *  that is divisible by 100 but not 400 is NOT a leap year (e.g 1700,
   *  1800, 1900, 2100, 2200, 2300, 2500, 2600, etc. are NOT leap years).
   */
  static bool isLeapYear(int year) {
    if (year % 4 == 0)
      return !(year % 100 == 0 && year % 400 != 0);
    else
      return false;
  }

  /// Compute number of days in a year.
  static int daysInYear(int year) { return isLeapYear(year) ? 366 : 365; }

  /// Convert a date into absolute seconds.
  static int64_t dateToSeconds(int day, int month, int year) {
    assert(day >= 1 && day <= 31);
    assert(month >= 1 && month <= 12);
    assert(year >= 1970);
    int64_t daysSinceEpoch = 0;
    for (int yi = epochYear; yi < year; yi++)
      daysSinceEpoch += daysInYear(yi);
    daysSinceEpoch += day - 1;
    daysSinceEpoch +=
        isLeapYear(year) ? leapYearDays[month] : nonLeapYearDays[month];
    return daysSinceEpoch * secondsPerDay;
  }

  /** Convert absolute seconds into a date.
   *
   *  \return The remaining fractional seconds in the day.
   */
  static int64_t secondsToDate(int64_t seconds, int &day, int &month, int &year,
                               int &dayOfYear) {
    int64_t daysSinceEpoch = seconds / secondsPerDay;

    // figure out the year
    int dayCounter = 0;
    for (year = epochYear; true; year++) {
      int tmp = dayCounter + daysInYear(year);
      if (tmp > daysSinceEpoch)
        break;
      dayCounter = tmp;
    }
    const auto &monthStartDays =
        isLeapYear(year) ? leapYearDays : nonLeapYearDays;

    // figure out the month
    month = 1;
    int yearDays = 0;
    while (month < 12) {
      if (monthStartDays[month + 1] > daysSinceEpoch - dayCounter)
        break;
      yearDays = monthStartDays[month + 1];
      month++;
    }
    dayCounter += yearDays;
    day = daysSinceEpoch - dayCounter + 1;
    dayOfYear = monthStartDays[month] + day;
    return seconds - daysSinceEpoch * secondsPerDay;
  }

  DateTime &operator=(const DateTime &) noexcept = default;

  DateTime &operator=(double rhs) noexcept {
    setAbsoluteSeconds(rhs);
    utcHourOffset = 0; // No other knowledge, assume zero
    return *this;
  }

  DateTime &operator+=(const DateTime &rhs) noexcept {
    intSeconds += rhs.intSeconds;
    fracSeconds += rhs.fracSeconds;
    wrapFraction();
    return *this;
  }

  DateTime &operator-=(const DateTime &rhs) noexcept {
    intSeconds -= rhs.intSeconds;
    fracSeconds -= rhs.fracSeconds;
    wrapFraction();
    return *this;
  }

  DateTime &operator+=(double rhs) {
    DIRSIG_THROW_IF(!std::isfinite(rhs),
                    "Tried to add non-finite number to DateTime!");
    return operator+=(DateTime(rhs));
  }

  DateTime &operator-=(double rhs) {
    DIRSIG_THROW_IF(!std::isfinite(rhs),
                    "Tried to add non-finite number to DateTime!");
    return operator-=(DateTime(rhs));
  }

  template <typename Any> DateTime operator+(const Any &rhs) const {
    return DateTime(*this) += rhs;
  }

  template <typename Any> DateTime operator-(const Any &rhs) const {
    return DateTime(*this) -= rhs;
  }

  double operator*(double rhs) const noexcept {
    return getTotalSeconds() * rhs;
  }

  double operator/(double rhs) const noexcept {
    return getTotalSeconds() / rhs;
  }

  double operator/(const DateTime &rhs) const noexcept {
    return getTotalSeconds() / rhs.getTotalSeconds();
  }

  bool operator==(const DateTime &rhs) const noexcept {
    return attributes == rhs.attributes && //
           intSeconds == rhs.intSeconds && //
           fracSeconds == rhs.fracSeconds;
  }

  bool operator!=(const DateTime &rhs) const noexcept {
    return !operator==(rhs);
  }

  bool operator<(const DateTime &rhs) const noexcept {
    return intSeconds < rhs.intSeconds ||
           (intSeconds == rhs.intSeconds && fracSeconds < rhs.fracSeconds);
  }

  bool operator>(const DateTime &rhs) const noexcept {
    return !operator<=(rhs);
  }

  bool operator<=(const DateTime &rhs) const noexcept {
    return intSeconds < rhs.intSeconds ||
           (intSeconds == rhs.intSeconds && fracSeconds <= rhs.fracSeconds);
  }

  bool operator>=(const DateTime &rhs) const noexcept {
    return !operator<(rhs);
  }

public:
  /// The reference epoch year
  static constexpr int epochYear = 1970;

  /// The cumulative number of days for each month (non-leap year)
  static constexpr std::array<int, 13> nonLeapYearDays = {
      -1, 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};

  /// The cumulative number of days for each month (leap year)
  static constexpr std::array<int, 13> leapYearDays = {
      -1, 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335};

  /// The various attributes of the instance (valid, relative, etc.)
  unsigned char attributes = 0;

  /** The integer number of seconds since the Epoch.
   *
   *  If the object is storing an absolute date/time then this variable
   *  stores the integer seconds the since the Epoch. Otherwise the
   *  number of seconds is relative to an arbitrary reference (only a limited
   *  number of operations are supported for this case).
   */
  int64_t intSeconds = 0;

  /** Fractional seconds.
   *
   *  Since the `intSeconds` variable stores the integer
   *  number of seconds and we need to support fractional seconds, we
   *  introduce the fracSeconds variable to store the fractional second
   *  portion of the time. This could be done several ways but for
   *  simplicity we choose to use a signed 64-bit integer to store the
   *  number of nanoseconds.
   */
  int64_t fracSeconds = 0;

  /** This variable stores the relative offset of a local date/time from
   *  Greenwich Mean Time (GMT) or Coordinated Universal Time (UTC). This
   *  variable is critical in that it allows this class to correctly
   *  compute the broken down date (year, month, day, etc.) from the
   *  integer time. This value is in hours.
   */
  float utcHourOffset = 0;

  void wrapFraction() noexcept {
    int64_t quo = fracSeconds / nanoSecondsPerSecond;
    int64_t rem = fracSeconds - nanoSecondsPerSecond * quo;
    intSeconds += quo;
    fracSeconds = rem;
    if (fracSeconds < 0) {
      fracSeconds += nanoSecondsPerSecond;
      intSeconds--;
    }
  }
};

/// \}

} // namespace d5
