/** Interface file for a dual quaternion
 *
 *  \file DualQuaternion.h
 */

#pragma once

#include <iostream>

#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Json.h"
#include "dirsig5/base/Pcg.h"

namespace d5 {

/// \addtogroup common
/// \{

struct WellDefinedDualQuaternion {
public:
  WellDefinedDualQuaternion() = default;

  WellDefinedDualQuaternion(const Quaternion &r, const Quaternion &d) {
    set(r, d);
  }

  WellDefinedDualQuaternion(const Quaternion &r, const Vector &t) { set(r, t); }

  WellDefinedDualQuaternion(const Eigen::Matrix4d &m) {
    set(Quaternion(Eigen::Matrix3d(m.block(0, 0, 3, 3))),
        Vector(m.col(3).head(3)));
  }

  Quaternion real = Quaternion(1, 0, 0, 0);

  Quaternion dual = Quaternion(0, 0, 0, 0);

  void set(const Quaternion &r, const Quaternion &d) {
    real = r;
    dual = d;
    normalize();
  }

  void set(const Quaternion &r, const Vector &t) {
    real = r;
    dual = Quaternion(0, t(0), t(1), t(2)) * r;
    dual.coeffs() *= 0.5;
    normalize();
  }

  /// Returns the conjugate with respect to quaternion ijk.
  [[nodiscard]] WellDefinedDualQuaternion conjugate() const {
    WellDefinedDualQuaternion result = *this;
    result.real = result.real.conjugate();
    result.dual = result.dual.conjugate();
    return result;
  }

  /// Returns the conjugate with respect to dual unit.
  [[nodiscard]] WellDefinedDualQuaternion dualConjugate() const {
    WellDefinedDualQuaternion result = *this;
    result.dual.coeffs() *= -1;
    return result;
  }

  void normalize() {
    double fac = real.norm();
    if (fac > 0.0) {
      real.coeffs() /= fac;
      dual.coeffs() /= fac;
    }
  }

  [[nodiscard]] WellDefinedDualQuaternion normalized() const {
    WellDefinedDualQuaternion result = *this;
    result.normalize();
    return result;
  }

  [[nodiscard]] WellDefinedDualQuaternion inverse() const {
    WellDefinedDualQuaternion result = conjugate();
    double a = 1.0 / real.dot(real);
    double b = a * a * -2.0 * real.dot(dual);
    result.dual.coeffs() *= a;
    result.dual.coeffs() += b * result.real.coeffs();
    result.real.coeffs() *= a;
    return result;
  }

  WellDefinedDualQuaternion operator+() const { return *this; }

  WellDefinedDualQuaternion operator-() const {
    WellDefinedDualQuaternion result = *this;
    result.real.coeffs() *= -1.0;
    result.dual.coeffs() *= -1.0;
    return result;
  }

  WellDefinedDualQuaternion
  operator+(const WellDefinedDualQuaternion &rhs) const {
    WellDefinedDualQuaternion result;
    result.real.coeffs() = real.coeffs() + rhs.real.coeffs();
    result.dual.coeffs() = dual.coeffs() + rhs.dual.coeffs();
    return result;
  }

  WellDefinedDualQuaternion
  operator-(const WellDefinedDualQuaternion &rhs) const {
    WellDefinedDualQuaternion result;
    result.real.coeffs() = real.coeffs() - rhs.real.coeffs();
    result.dual.coeffs() = dual.coeffs() - rhs.dual.coeffs();
    return result;
  }

  WellDefinedDualQuaternion
  operator*(const WellDefinedDualQuaternion &rhs) const {
    WellDefinedDualQuaternion result;
    result.real = real * rhs.real;
    result.dual = dual * rhs.real;
    result.dual.coeffs() += (real * rhs.dual).coeffs();
    return result;
  }

  WellDefinedDualQuaternion
  operator/(const WellDefinedDualQuaternion &rhs) const {
    return operator*(rhs.inverse());
  }

  Point pointForward(const Point &p) const {
    WellDefinedDualQuaternion d;
    d.real = Quaternion(1, 0, 0, 0);
    d.dual = Quaternion(0, p(0), p(1), p(2));
    return ((*this) * d * conjugate().dualConjugate()).dual.vec();
  }

  Point pointInverse(const Point &p) const { return inverse().pointForward(p); }

  Vector vectorForward(const Vector &v) const {
    return (real * Quaternion(0, v(0), v(1), v(2)) * real.conjugate()).vec();
  }

  Vector vectorInverse(const Vector &v) const {
    return inverse().vectorForward(v);
  }

  friend std::ostream &operator<<(std::ostream &os,
                                  const WellDefinedDualQuaternion &dq) {
    return os << "[("                 //
              << dq.real.w() << ','   //
              << dq.real.x() << ','   //
              << dq.real.y() << ','   //
              << dq.real.z() << "),(" //
              << dq.dual.w() << ','   //
              << dq.dual.x() << ','   //
              << dq.dual.y() << ','   //
              << dq.dual.z() << ")]";
  }
};

/// Decomposed linear interpolation between dual quaternions.
///
/// This is accomplished as a simple linear interpolaion of the location and
/// spherical linear interpolation (SLERP) on the rotation.
///
inline WellDefinedDualQuaternion interpolate(const WellDefinedDualQuaternion &a,
                                             const WellDefinedDualQuaternion &b,
                                             double t) {
  Vector aLoc = a.pointForward(Point(0, 0, 0));
  Vector bLoc = b.pointForward(Point(0, 0, 0));
  return WellDefinedDualQuaternion(a.real.slerp(t, b.real),
                                   (1 - t) * aLoc + t * bLoc);
}
/// Performs a screw linear interpolation (ScLERP) between a pair of
/// DualQuaternion instances.
inline WellDefinedDualQuaternion ScLERP(const WellDefinedDualQuaternion &a,
                                        const WellDefinedDualQuaternion &b,
                                        double t) {
  if (a.real.dot(b.real) < 0)
    return ScLERP(a, -b, t);

  // ScLERP = qa(qa^-1 qb)^t
  WellDefinedDualQuaternion diff = a.conjugate() * b;
  Vector vr(diff.real.x(), diff.real.y(), diff.real.z());
  Vector vd(diff.dual.x(), diff.dual.y(), diff.dual.z());
  double vrLen = std::sqrt(vr.dot(vr));
  double phi = std::atan2(vrLen, diff.real.w());
  if (std::fabs(vrLen) < 1e-5 || !std::isfinite(phi))
    return WellDefinedDualQuaternion(
        a.real.slerp(t, b.real),
        Quaternion((1 - t) * a.dual.w() + t * b.dual.w(),
                   (1 - t) * a.dual.x() + t * b.dual.x(),
                   (1 - t) * a.dual.y() + t * b.dual.y(),
                   (1 - t) * a.dual.z() + t * b.dual.z()));

  double pitch = -diff.dual.w() / vrLen;
  Vector vrHat = vr / vrLen;
  double sinPhit = std::sin(phi * t);
  double cosPhit = std::cos(phi * t);
  Vector v1 = sinPhit * vrHat;
  Vector v2 = sinPhit * (vd - vrHat * pitch * diff.real.w()) / vrLen +
              cosPhit * pitch * t * vrHat;
  WellDefinedDualQuaternion result;
  result.real = Quaternion(cosPhit, v1.x(), v1.y(), v1.z());
  result.dual = Quaternion(sinPhit * -pitch * t, v2.x(), v2.y(), v2.z());
  result = a * result;
  return result;
}

struct DualQuaternion : WellDefinedDualQuaternion {
public:
  DualQuaternion() = default;

  using WellDefinedDualQuaternion::WellDefinedDualQuaternion;

  enum ElementOrder { XYZW, WXYZ };

  /// Array constructor. Constructs the dual quaternion from the supplied
  /// array of values (8 in all). The real quaternion (x,y,z,w) is provided
  /// first, followed by the dual.
  ///
  /// An optional scale term can be provided as well.
  ///
  template <typename Type>
  DualQuaternion(const Type *array, size_t arraySize = 8,
                 ElementOrder order = XYZW) {
    static_assert(std::is_arithmetic_v<Type>);
    if (order == XYZW) {
      real = Quaternion(array[3], array[0], array[1], array[2]);
      dual = Quaternion(array[7], array[4], array[5], array[6]);
    } else {
      real = Quaternion(array[0], array[1], array[2], array[3]);
      dual = Quaternion(array[4], array[5], array[6], array[7]);
    }
    normalize();
    if (arraySize == 11)
      addScaling(Vector(array[8], array[9], array[10]));
  }

  explicit DualQuaternion(const Matrix &affine) {
    Eigen::Matrix3d rotation;
    Eigen::Matrix3d scaling;
    affine.computeRotationScaling(&rotation, &scaling);
    set(Quaternion(rotation), affine.translation());
    addScaling(scaling.diagonal());
  }

  void disableTransform() { disabledFlag = true; }

  bool isDisabled() const { return disabledFlag; }

  /// Returns the conjugate with respect to quaternion ijk.
  [[nodiscard]] DualQuaternion conjugate() const {
    DualQuaternion result = *this;
    result.real = result.real.conjugate();
    result.dual = result.dual.conjugate();
    return result;
  }

  void negate() {
    real.coeffs() *= -1;
    dual.coeffs() *= -1;
  }

  void addScaling(const Vector &s) {
    if (s != Vector(1, 1, 1) && s.allFinite() && s.maxCoeff() > 1e-9) {
      scale = s;
      scaleFlag = true;
    }
  }

  [[nodiscard]] Point pointForward(Point p) const {
    if (scaleFlag)
      p.array() *= scale.array();
    return WellDefinedDualQuaternion::pointForward(p);
  }

  [[nodiscard]] Point pointInverse(Point p) const {
    p = WellDefinedDualQuaternion::pointInverse(p);
    if (scaleFlag)
      p.array() /= scale.array();
    return p;
  }

  [[nodiscard]] Vector vectorForward(Vector v, bool allowScale = true) const {
    if (scaleFlag and allowScale)
      v.array() *= scale.array();
    return WellDefinedDualQuaternion::vectorForward(v);
  }

  [[nodiscard]] Vector vectorInverse(Vector v, bool allowScale = true) const {
    v = WellDefinedDualQuaternion::vectorInverse(v);
    if (scaleFlag and allowScale)
      v.array() /= scale.array();
    return v;
  }

  friend DualQuaternion operator*(const DualQuaternion &lhs,
                                  const DualQuaternion &rhs) {
    WellDefinedDualQuaternion wellDefLhs =
        static_cast<const WellDefinedDualQuaternion &>(lhs);
    WellDefinedDualQuaternion wellDefRhs =
        static_cast<const WellDefinedDualQuaternion &>(rhs);
    if (lhs.scaleFlag) {
      // Warning, this only makes proper sense if scaling is uniform!
      wellDefRhs.dual.x() *= lhs.scale[0];
      wellDefRhs.dual.y() *= lhs.scale[1];
      wellDefRhs.dual.z() *= lhs.scale[2];
      wellDefRhs.dual.w() *= (lhs.scale[0] + lhs.scale[1] + lhs.scale[2]) / 3;
    }
    WellDefinedDualQuaternion wellDefResult = wellDefLhs * wellDefRhs;
    DualQuaternion result;
    result.real = wellDefResult.real;
    result.dual = wellDefResult.dual;
    if (lhs.scaleFlag && !rhs.scaleFlag)
      result.addScaling(lhs.scale);
    if (rhs.scaleFlag && !lhs.scaleFlag)
      result.addScaling(rhs.scale);
    if (lhs.scaleFlag && rhs.scaleFlag)
      result.addScaling(lhs.scale.array() * rhs.scale.array());
    return result;
  }

  operator Matrix() const noexcept {
    // Construct by transforming the standard basis.
    Matrix matrix;
    matrix.setIdentity();
    matrix.affine().col(0) = vectorForward(Vector::UnitX());
    matrix.affine().col(1) = vectorForward(Vector::UnitY());
    matrix.affine().col(2) = vectorForward(Vector::UnitZ());
    matrix.affine().col(3) = pointForward(Vector(0, 0, 0));
    return matrix;
  }

  operator Matrixf() const noexcept { return Matrix(*this).cast<float>(); }

public:
  Vector scale = Vector(1, 1, 1);
  bool scaleFlag = false;
  bool disabledFlag = false;
};

/// Decomposed linear interpolation between dual quaternions.
///
/// This is accomplished as a simple linear interpolaion of the location and
/// spherical linear interpolation (SLERP) on the rotation.
///
inline DualQuaternion interpolate(const DualQuaternion &a,
                                  const DualQuaternion &b, double t) {
  Vector aLoc = a.pointForward(Point(0, 0, 0));
  Vector bLoc = b.pointForward(Point(0, 0, 0));
  return DualQuaternion(a.real.slerp(t, b.real), (1 - t) * aLoc + t * bLoc);
}

inline DualQuaternion ScLERP(const DualQuaternion &a, const DualQuaternion &b,
                             double t) {
  WellDefinedDualQuaternion wellDefResult =
      ScLERP(static_cast<const WellDefinedDualQuaternion &>(a),
             static_cast<const WellDefinedDualQuaternion &>(b), t);
  DualQuaternion result;
  result.real = wellDefResult.real;
  result.dual = wellDefResult.dual;
  if (a.scaleFlag || b.scaleFlag) {
    Vector scaleA = a.scaleFlag ? a.scale : Vector(1, 1, 1);
    Vector scaleB = b.scaleFlag ? b.scale : Vector(1, 1, 1);
    result.addScaling((1 - t) * scaleA + t * scaleB);
  }
  return result;
}

D5_TO_JSON(DualQuaternion, j, dq) {
  j = Json{{"real", dq.real}, {"dual", dq.dual}};
}

D5_FROM_JSON(DualQuaternion, j, dq) {
  j.at("real").get_to(dq.real);
  j.at("dual").get_to(dq.dual);
}

inline BBox GrowBBoxWithDynamicTransform(
    BBox bbox, const std::function<DualQuaternion(double t)> &func, double t0,
    double t1, int maxDepth = 8) {
  BBox worldBBox;
  bool anyEnabled = false;
  bool anyDisabled = false;
  RandEngine rng = {};
  auto expand = [&](double time0, double time1) {
    BBox prevWorldBBox = worldBBox;
    for (int step = 0; step <= 8; step++) {
      double s = (step + std::generate_canonical<float, 32>(rng)) * 0.125;
      DualQuaternion transform = func((1 - s) * time0 + s * time1);
      if (transform.isDisabled()) {
        anyDisabled = true;
        continue;
      }
      anyEnabled = true;
      for (int c = 0; c < 8; c++) {
        auto corner = transform.pointForward(
            Vector((c & 1) == 0 ? bbox.min[0] : bbox.max[0], //
                   (c & 2) == 0 ? bbox.min[1] : bbox.max[1], //
                   (c & 4) == 0 ? bbox.min[2] : bbox.max[2]));
        if (std::isfinite(corner[0]) && std::isfinite(corner[1]) &&
            std::isfinite(corner[2]))
          worldBBox |= corner;
      }
    }
    return worldBBox != prevWorldBBox; // Updated?
  };
  struct Interval {
    double time0;
    double time1;
    int depth;
  };
  std::vector<Interval> intervals;
  intervals.reserve(512);
  intervals.push_back({t0, t1, 0});
  while (!intervals.empty()) {
    Interval interval = intervals.back();
    intervals.pop_back();
    double time0 = interval.time0;
    double time1 = interval.time1;
    int depth = interval.depth;
    if (depth > maxDepth)
      continue;
    if (expand(time0, time1) || depth < 3) {
      intervals.push_back({time0, 0.5 * time0 + 0.5 * time1, depth + 1});
      intervals.push_back({0.5 * time0 + 0.5 * time1, time1, depth + 1});
    }
  }
  // give ourselves a little padding
  worldBBox.expand(0.05);
  if (anyDisabled && !anyEnabled) {
    worldBBox.min = {0, 0, 0};
    worldBBox.max = {0, 0, 0};
  }
  return worldBBox;
}

/// \}

} // namespace d5
