#pragma once

#include <cstdlib>
#include <cstring>
#include <stdexcept>
#include <string>
#include <system_error>
#include <vector>

#if USE_BOOST_FILESYSTEM
#include <boost/system/error_code.hpp>
#endif // #if USE_BOOST_FILESYSTEM

#if __GNUC__ || __clang__
#if __has_include(<cxxabi.h>)
#define DIRSIG_HAS_CXXABI 1
#endif // #if __has_include(<cxxabi.h>)
#endif // #if __GNUC__ || __clang__

#if __GNUC__ || __clang__
#define DIRSIG_ERROR_EXPORT __attribute__((visibility("default")))
#else
#define DIRSIG_ERROR_EXPORT
#endif // #if __GNUC__ || __clang__

#if DIRSIG_HAS_CXXABI
#include <cxxabi.h>
#endif // #if DIRSIG_HAS_CXXABI

namespace d5 {
// NOLINT

/// \addtogroup common
/// \{

namespace detail {

/** Removes the path component from a file path (from __FILE__)
 *
 *  \param filename  The input file path
 *
 *  \return The pointer inside of filename of just the filename part of the path
 */
inline const char *remove_path(const char *filename) {
  if (auto *base = std::strrchr(filename, '/'); base)
    return base + 1;
  if (auto *base = std::strrchr(filename, '\\'); base)
    return base + 1;
  return filename;
}

inline std::string demangle(const char *name) {
  if (!name)
    return {};
#if DIRSIG_HAS_CXXABI
  int status = 0;
  char *cstr = ::abi::__cxa_demangle(name, 0, 0, &status);
  if (!cstr)
    return name;
  std::string str = cstr;
  std::free(cstr);
  return str;
#else
  return name;
#endif // #if DIRSIG_HAS_CXXABI
}

inline std::string currentExceptionName() {
#if DIRSIG_HAS_CXXABI
  return demangle(::abi::__cxa_current_exception_type()->name());
#else
  return "unknown exception";
#endif // #if DIRSIG_HAS_CXXABI
}

} // namespace detail

class DIRSIG_ERROR_EXPORT Error : public std::runtime_error {
public:
  template <typename... Args>
  Error(const char *funcName, const char *fileName, long lineNo,
        std::string message)
      : std::runtime_error(std::move(message)), funcName(funcName),
        fileName(detail::remove_path(fileName)), lineNo(lineNo) {}

  Error &withDebugMessage(std::string message) {
    if (!message.empty())
      debugMessages.emplace_back(std::move(message));
    return *this;
  }
  Error &withDebugMessage(std::error_code ec) {
    if (ec)
      debugMessages.emplace_back("error_code: " + ec.message());
    return *this;
  }
#if USE_BOOST_FILESYSTEM
  Error &withDebugMessage(boost::system::error_code ec) {
    if (ec)
      debugMessages.emplace_back("error_code: " + ec.message());
    return *this;
  }
#endif

  Error &withTraceMessage(std::string message) {
    if (!message.empty())
      traceMessages.emplace_back(std::move(message));
    return *this;
  }
  Error &withTrace(std::string message) {
    return withTraceMessage(std::move(message));
  }

public:
  const char *funcName = nullptr;
  const char *fileName = nullptr;
  long lineNo = 0;
  std::vector<std::string> debugMessages;
  std::vector<std::string> traceMessages;
};

#define DIRSIG_MAKE_ERROR(message)                                             \
  ::d5::Error(__func__, __FILE__, __LINE__, message)

#define DIRSIG_TRY try

#define DIRSIG_CATCH(name, ...)                                                \
  catch (::d5::Error & name) {                                                 \
    __VA_ARGS__                                                                \
  }                                                                            \
  catch (const std::exception &other) {                                        \
    auto name =                                                                \
        DIRSIG_MAKE_ERROR(other.what())                                        \
            .withDebugMessage("Converted from STL exception. (" +              \
                              ::d5::detail::currentExceptionName() + ")");     \
    __VA_ARGS__                                                                \
  }                                                                            \
  catch (...) {                                                                \
    auto name =                                                                \
        DIRSIG_MAKE_ERROR("Unknown error!")                                    \
            .withDebugMessage("Converted from non-STL exception. (" +          \
                              ::d5::detail::currentExceptionName() + ")");     \
    __VA_ARGS__                                                                \
  }

#define DIRSIG_THROW(message) throw DIRSIG_MAKE_ERROR(message)

#define DIRSIG_THROW_FMT(...) DIRSIG_THROW(fmt::format(__VA_ARGS__))

#define DIRSIG_THROW_IF(cond, message)                                         \
  do {                                                                         \
    if ((cond))                                                                \
      DIRSIG_THROW(message).withDebugMessage(                                  \
          "Error condition satisfied: " #cond);                                \
  } while (false)

#define DIRSIG_THROW_IF_FMT(cond, ...)                                         \
  DIRSIG_THROW_IF(cond, fmt::format(__VA_ARGS__))

#define DIRSIG_ERROR_TRACE(where, ...)                                         \
  DIRSIG_TRY{__VA_ARGS__} DIRSIG_CATCH(                                        \
      error, { throw error.withTraceMessage(where); })

/// \}

} // namespace d5
