#pragma once

#include <fstream>
#include <iostream>
#include <memory>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Filesystem.h"

namespace d5 {

/// Verify that the given path exists on the filesystem.
inline void VerifyPathExists(const stdfs::path &path) {
  stdfs_error_code ec;
  if (!stdfs::exists(path, ec))
    DIRSIG_THROW("Path \"" + path.string() + "\" does not exist!")
        .withDebugMessage(ec);
}

/// Verify that the given path is a file.
inline void VerifyPathIsFile(const stdfs::path &path) {
  stdfs_error_code ec;
  if (!stdfs::is_regular_file(path, ec))
    DIRSIG_THROW("Path \"" + path.string() + "\" is not a file!")
        .withDebugMessage(ec);
}

/// Verify that the given path is a directory.
inline void VerifyPathIsDirectory(const stdfs::path &path) {
  stdfs_error_code ec;
  if (!stdfs::is_directory(path, ec))
    DIRSIG_THROW("Path \"" + path.string() + "\" is not a directory!")
        .withDebugMessage(ec);
}

namespace detail {

inline void ThrowIfNotOpen(const std::string &path, std::ifstream &stream) {
  DIRSIG_THROW_IF(!stream.is_open(),
                  "Can't open file \"" + path +
                      "\" for reading: " + std::strerror(errno));
}

inline void ThrowIfNotOpen(const std::string &path, std::ofstream &stream) {
  DIRSIG_THROW_IF(!stream.is_open(),
                  "Can't open file \"" + path +
                      "\" for writing: " + std::strerror(errno));
}

} // namespace detail

/// Open an ifstream, throw on failure.
template <typename... Args>
inline std::ifstream OpenIfstream(const std::string &path, Args &&...args) {
  std::ifstream stream(path, std::forward<Args>(args)...);
  detail::ThrowIfNotOpen(path, stream);
  return stream;
}

/// Open an ofstream, throw on failure.
template <typename... Args>
inline std::ofstream OpenOfstream(const std::string &path, Args &&...args) {
  std::ofstream stream(path, std::forward<Args>(args)...);
  detail::ThrowIfNotOpen(path, stream);
  return stream;
}

namespace detail {

struct IstreamDeleter {
  void operator()(std::istream *stream) const {
    if (stream != &std::cin)
      delete stream;
  }
};

struct OstreamDeleter {
  void operator()(std::ostream *stream) const {
    if (stream != &std::cout && stream != &std::cerr)
      delete stream;
  }
};

} // namespace detail

/// A unique pointer to an istream, doesn't delete `std::cin`.
using IstreamPtr = std::unique_ptr<std::istream, detail::IstreamDeleter>;

/// A unique pointer to an istream, doesn't delete `std::cout` or `std::cerr`.
using OstreamPtr = std::unique_ptr<std::ostream, detail::OstreamDeleter>;

/// Open a filestream, or `std::cin` if `path == "-"`.
inline IstreamPtr OpenIfstreamOrCin(const std::string &path) {
  if (path.empty() or path == "-")
    return IstreamPtr(&std::cin);
  else {
    IstreamPtr stream(new std::ifstream(path));
    detail::ThrowIfNotOpen(path, static_cast<std::ifstream &>(*stream));
    return stream;
  }
}

/// Open a filestream, or `std::cout` if `path == "-"`.
inline OstreamPtr OpenOfstreamOrCout(const std::string &path) {
  if (path.empty() or path == "-")
    return OstreamPtr(&std::cout);
  else {
    OstreamPtr stream(new std::ofstream(path));
    detail::ThrowIfNotOpen(path, static_cast<std::ofstream &>(*stream));
    return stream;
  }
}

inline std::string ReadIfstreamIntoString(const std::string &path) {
  auto ifs = OpenIfstream(path, std::ios::in | std::ios::binary);
  return std::string((std::istreambuf_iterator<char>(ifs)),
                     std::istreambuf_iterator<char>());
}

inline std::string ShowFilename(const std::string &path) {
  if (!stdfs::exists(path))
    return path;
  auto relPath = stdfs::relative(path).string();
  return relPath.size() < path.size() ? relPath : path;
}

} // namespace d5
