#pragma once

#include <algorithm>
#include <complex>
#include <vector>

#include "dirsig5/base/Constants.h"

namespace d5 {

/**
 * Supplies tools for Fourier transforms. Note that the FFTs themselves
 * have been moved to a third party library, Simple-FFT
 */
namespace FourierTools {

template <typename Complex>
inline void primeFactorTransform(int next, int prev, int curr,
                                 const Complex *in, Complex *out) {
  in -= prev * (curr + 1) + 1;
  out -= prev * (next + 1) + 1;
  double sinTheta = std::sin(2 * PI / (next * curr));
  double cosTheta = std::cos(2 * PI / (next * curr));
  Complex delta(cosTheta, -sinTheta);
  Complex omega(1.0);
  for (int j = 1; j <= curr; ++j) {
    for (int k = 1; k <= next; ++k) {
      for (int i = 1; i <= prev; ++i) {
        const Complex *inTmp = in + i + (k + 1) * prev * curr;
        Complex inSum = inTmp[0];
        for (int l = 2; l <= curr; ++l)
          inSum = inSum * omega + inTmp[(1 - l) * prev];
        out[i + (k + j * next) * prev] = inSum;
      }
      omega *= delta;
    }
  }
}

/// Get the spatial frequencies corresponding to a 1D forward FFT
/// for a given array size.
template <typename Float>
inline std::vector<Float> forwardFrequencies(int n, double dx) {
  if (n <= 0)
    return {};
  std::vector<Float> freq(n);
  if (n % 2 == 0) { // even
    for (int ni = 0; ni <= (n / 2) - 1; ni++)
      freq[ni] = ni / (dx * n);
    for (int ni = (n / 2); ni < n; ni++)
      freq[ni] = -int(n - ni) / (dx * n);
  } else {
    for (int ni = 0; ni <= (n - 1) / 2; ni++)
      freq[ni] = ni / (dx * n);
    for (int ni = 1 + (n - 1) / 2; ni < n; ni++)
      freq[ni] = -int(n - ni) / (dx * n);
  }
  return freq;
}

} // namespace FourierTools

} // namespace d5
