/** DIRSIG extensions to the C++ standard functional library
 *
 *  \file Functional.h
 */

#pragma once

#include <functional>
#include <type_traits>
#include <utility>

#include "dirsig5/base/Macros.h"

namespace d5 {

/** \addtogroup common
 *  \{
 */

/** Binds a member function to a function object
 *
 *  \tparam T    Type of the object that has the member function
 *  \tparam F    Type of the member function pointer
 *  \param f     Member function pointer
 *  \param self  Object on which to call the member function
 *
 *  \return Bound member function object
 */
template <typename T, typename F> decltype(auto) BindMemFn(F &&f, T &&self) {
  static_assert(std::is_member_function_pointer_v<std::decay_t<F>>,
                "F must be a member function pointer");
  static_assert(std::is_pointer_v<std::decay_t<T>>,
                "self must be a pointer to the object");

  return [=](auto &&...args) {
    return (self->*f)(std::forward<decltype(args)>(args)...);
  };
}

/** Binds a function to a function object and forwards all arguments
 *
 *  \tparam F Type of the function
 *  \param f  Function to bind
 *
 *  \return Bound function object
 */
template <typename F> decltype(auto) BindFn(F &&f) {
  return
      [&](auto &&...args) { return f(std::forward<decltype(args)>(args)...); };
}

/// \}
} // namespace d5
