/** Common header for geometry utilities
 *
 *  \file Geometry.h
 */

#pragma once

#include <array>
#include <random>

#include <Eigen/Dense>

#include "dirsig5/base/Constants.h"

namespace d5 {

/// \addtogroup common
/// \{

/// \name Common typedefs
/// \{
template <typename Value, int N = Eigen::Dynamic>
using VectorN = Eigen::Matrix<Value, N, 1>;
using Point = Eigen::Vector3d;
using Pointf = Eigen::Vector3f;
using Point2 = Eigen::Vector2d;
using Point2f = Eigen::Vector2f;
using Point2i = Eigen::Vector2<int64_t>;
using Vector = Eigen::Vector3d;
using Vectorf = Eigen::Vector3f;
using Vector2 = Eigen::Vector2d;
using Vector2f = Eigen::Vector2f;
using Matrix = Eigen::Affine3d;
using Matrixf = Eigen::Affine3f;
using Matrix3f = Eigen::Matrix3f;
using Matrix3 = Eigen::Matrix3d;
using Quaternion = Eigen::Quaterniond;
using Quaternionf = Eigen::Quaternionf;
/// \}

/// Get the zenith (declination) angle.
inline double getZenith(const Vector &vec) noexcept {
  return std::clamp(std::acos(vec[2]), 0.0, PI);
}

/// Get the geographic azimuth (CW east of north) angle.
inline double getGeoAzimuth(const Vector &vec) noexcept {
  if (vec[0] == 0 && vec[1] == 0)
    return 0;
  double phi = PI_2 - std::atan2(vec[1], vec[0]);
  while (phi < 0)
    phi += 2 * PI;
  while (phi >= 2 * PI)
    phi -= 2 * PI;
  return phi;
}

/// Make vector from theta (zenith) and phi (azimuth) angles.
inline Vector makeVector(double theta, double phi) noexcept {
  return Vector(std::sin(theta) * std::cos(phi),
                std::sin(theta) * std::sin(phi), std::cos(theta));
}

/// Make vector from theta (zenith) and phi (geographic azimuth) angles.
inline Vector makeGeoVector(double theta, double phi) noexcept {
  return makeVector(theta, PI_2 - phi);
}

/// Make a matrix to convert from ENU to ECEF
inline Matrix3 makeLatLonMatrix(double lat, double lon) noexcept {
  lat = toRadians(lat);
  lon = toRadians(lon);
  Matrix3 matrix;
  matrix(0, 0) = -sin(lon);
  matrix(0, 1) = -sin(lat) * cos(lon);
  matrix(0, 2) = cos(lat) * cos(lon);
  matrix(1, 0) = cos(lon);
  matrix(1, 1) = -sin(lat) * sin(lon);
  matrix(1, 2) = cos(lat) * sin(lon);
  matrix(2, 0) = 0;
  matrix(2, 1) = cos(lat);
  matrix(2, 2) = sin(lat);
  return matrix;
}

/// Make rotation about the X, then Y, then Z axes.
inline Matrix makeRotation(double thetaX, double thetaY,
                           double thetaZ) noexcept {
  Matrix m;
  m = Eigen::AngleAxisd(thetaZ, Vector::UnitZ()) *
      Eigen::AngleAxisd(thetaY, Vector::UnitY()) *
      Eigen::AngleAxisd(thetaX, Vector::UnitX());
  return m;
}

/// Make rotation about the X, then Y, then Z axes.
inline Matrixf makeRotationf(float thetaX, float thetaY,
                             float thetaZ) noexcept {
  Matrixf m;
  m = Eigen::AngleAxisf(thetaZ, Vectorf::UnitZ()) *
      Eigen::AngleAxisf(thetaY, Vectorf::UnitY()) *
      Eigen::AngleAxisf(thetaX, Vectorf::UnitX());
  return m;
}

inline Matrix makeRotation(Point angles, const char *order) noexcept {
  assert(order);
  Matrix m = Matrix::Identity();
  for (int k = 0; k < 3; k++)
    switch (order[k]) {
    case 'X':
    case 'x':
      m = Eigen::AngleAxisd(angles[0], Vector::UnitX()) * m;
      break;
    case 'Y':
    case 'y':
      m = Eigen::AngleAxisd(angles[1], Vector::UnitY()) * m;
      break;
    case 'Z':
    case 'z':
      m = Eigen::AngleAxisd(angles[2], Vector::UnitZ()) * m;
      break;
    default:
      break;
    }
  return m;
}

/// Make rotation about a given axis.
inline Matrix makeRotation(const Vector &axis, double angle) noexcept {
  Matrix m;
  m = Eigen::AngleAxisd(angle, axis);
  return m;
}

/// Make rotation about a given axis.
inline Matrixf makeRotationf(const Vectorf &axis, float angle) noexcept {
  Matrixf m;
  m = Eigen::AngleAxisf(angle, axis);
  return m;
}

/// Make a rotation transforming "from" to "to" vector.
inline Matrix makeRotation(const Vector &from, const Vector &to) noexcept {
  Vector cross = from.cross(to);
  double mu = from.dot(to);
  if (+mu > 1.0 - 1e-8)
    return Matrix::Identity();
  if (-mu > 1.0 - 1e-8) {
    Matrix m = Matrix::Identity();
    m(0, 0) *= -1;
    m(1, 1) *= -1;
    m(2, 2) *= -1;
    return m;
  }
  return makeRotation(cross.normalized(), std::acos(mu));
}

/// Make a rotation transforming "from" to "to" vector.
inline Matrixf makeRotationf(const Vectorf &from, const Vectorf &to) noexcept {
  Vectorf cross = from.cross(to);
  float mu = from.dot(to);
  if (+mu > 1.0f - 1e-8f)
    return Matrixf::Identity();
  if (-mu > 1.0f - 1e-8f) {
    Matrixf m = Matrixf::Identity();
    m(0, 0) *= -1;
    m(1, 1) *= -1;
    m(2, 2) *= -1;
    return m;
  }
  return makeRotationf(cross.normalized(), std::acos(mu));
}

/// Refract a source vector across a boundary.
///
/// Returns false if there is no transmitted vector. Note that ior is
/// the ratio of the incident to transmitted refractive indexes.
///
inline bool refract(const Vectorf &normal, float ior, const Vectorf &source,
                    Vectorf &refracted) {
  Vectorf local = makeRotationf(normal, Vectorf(0, 0, 1)) * source;
  if (local[2] == 0)
    return false;
  if (std::abs(local[2]) > 0.99999999)
    refracted = -source;
  else {
    // refract
    float thetai = std::acos(std::abs(local[2]));
    float thetat = std::asin(std::sin(thetai) * ior);
    if (!std::isfinite(thetat))
      return false;
    float cosThetat = std::cos(thetat);
    float sinThetat = std::sin(thetat);
    float scale = sinThetat / std::hypot(local[0], local[1]);
    refracted = makeRotationf(Vectorf(0, 0, 1), normal) *
                Vectorf(-local[0] * scale, -local[1] * scale, -cosThetat);
  }
  return refracted.dot(normal) <= 0;
}

inline Vectorf sampleCosineHemisphere(float u1, float u2) {
  // sample is generated for the cosine weighted hemisphere
  float cosTheta = std::sqrt(u1);
  float sinTheta = std::sqrt(1 - u1);
  return Vectorf(sinTheta * std::cos(float(2 * PI) * u2),
                 sinTheta * std::sin(float(2 * PI) * u2), cosTheta);
}

template <typename Gen> inline Vectorf sampleCosineHemisphere(Gen &gen) {
  return sampleCosineHemisphere(std::generate_canonical<float, 32>(gen),
                                std::generate_canonical<float, 32>(gen));
}

/// Stores a geodetic (WGS84) coordinate.
struct GeodeticCoord {
  GeodeticCoord() noexcept = default;
  GeodeticCoord(double lat, double lon, double h) noexcept
      : latitude(lat), longitude(lon), height(h) {}
  operator Point() const { return Point(latitude, longitude, height); }

  double latitude = 0;  ///< Latitude [decimal degrees]
  double longitude = 0; ///< Longitude [decimal degrees]
  double height = 0;    ///< Height above ellipsoid [m]
};

/// Stores an Earth-Centered, Earth-Fixed coordinate.
struct EcefCoord {
  EcefCoord() noexcept = default;
  EcefCoord(double a, double b, double c) noexcept : x(a), y(b), z(c) {}
  EcefCoord(const Point &p) noexcept : EcefCoord(p[0], p[1], p[2]) {}
  operator Point() const { return Point(x, y, z); }

  double x = 0; ///< ECEF X coordinate [m]
  double y = 0; ///< ECEF Y coordinate [m]
  double z = 0; ///< ECEF Z coordinate [m]
};

/// Stores a Universal Transverse Mercator (UTM) coordinate.
struct UtmCoord {
  UtmCoord() noexcept = default;
  UtmCoord(long z, char hemi, double east, double north, double h) noexcept
      : zone(z), hemisphere(hemi), easting(east), northing(north), height(h) {}
  operator Point() const { return Point(easting, northing, height); }

  long zone = 1;             ///< UTM zone
  char hemisphere = 'N';     ///< Hemisphere ['N' or 'S']
  double easting = 100000.0; ///< Easting [m]
  double northing = 0;       ///< Northing [m]
  double height = 0;         ///< Height above ellipsoid [m]
};

/// The default refractive index model (no atmosphere profile) is a
/// polynomial fit to the MLS profile at 1.0um. Above 50km is assumed
/// to have n=1
inline double defaultAtmRefractiveIndex(double alt /* [m] */) {
  constexpr double a0 = 1.0002666875;
  constexpr double a1 = -2.481e-5;
  constexpr double a2 = 8.646e-7;
  constexpr double a3 = -1.318e-8;
  constexpr double a4 = 7.357e-11;
  double alt2 = alt * alt;
  double alt3 = alt2 * alt;
  double alt4 = alt2 * alt2;
  return a0 + a1 * alt + a2 * alt2 + a3 * alt3 + a4 * alt4;
}

/// Time taken to travel a path based on `defaultAtmRefractiveIndex()`.
///
/// \param pa Point A.
/// \param pb Point B.
/// \param dz The Z component of the normalized direction from pa to pb.
///
/// \return Time in seconds, not accounting for curvature.
///
inline double defaultAtmPathTime(const Point &pa, const Point &pb, double dz) {
  dz = std::abs(dz);

  constexpr double k1 = -1.00027;
  constexpr double k2 = 0.000012405;
  constexpr double k3 = -2.882e-7;
  constexpr double k4 = 3.295e-9;
  constexpr double k5 = -1.4714e-11;
  constexpr double c = speedOfLight * 1e-3;

  // convert heights to [km]
  double za = std::max(pa[2] * 1e-3, 0.0);
  double zb = std::max(pb[2] * 1e-3, 0.0);
  if (za == zb) // Horizontal?
    return defaultAtmRefractiveIndex(pa[2]) * (pb - pa).norm() / speedOfLight;
  if (za >= 50 && zb >= 50)
    return std::abs(zb - za) / dz / c; // Unit IOR

  // Limit at 50km
  double t = 0.0;
  if (za >= 50)
    t = (za - 50) / dz / c, za = 50;
  if (zb >= 50)
    t = (zb - 50) / dz / c, zb = 50;

  // Integrate.
  double a = za, a2 = a * a, a3 = a2 * a, a4 = a2 * a2;
  double b = zb, b2 = b * b, b3 = b2 * b, b4 = b2 * b2;
  return t + std::abs((a * (k1 + k2 * a + k3 * a2 + k4 * a3 + k5 * a4) -
                       b * (k1 + k2 * b + k3 * b2 + k4 * b3 + k5 * b4))) /
                 dz / c;
}

/// Woo et al "Watertight" ray-triangle intersection test.
inline bool testRayTriangle(const Pointf &org, const Vectorf &dir, float tmin,
                            float &tmax, const Pointf &point0,
                            const Pointf &point1, const Pointf &point2,
                            float &u, float &v) {
  Vectorf::Index maxIndex;
  dir.cwiseAbs().maxCoeff(&maxIndex);
  int order[3] = {};
  order[2] = maxIndex;
  order[0] = (order[2] + 1) % 3;
  order[1] = (order[2] + 2) % 3;
  float shear[3] = {};
  shear[2] = 1 / dir[order[2]];
  shear[0] = -shear[2] * dir[order[0]];
  shear[1] = -shear[2] * dir[order[1]];

  Pointf p[3];
  p[0] = point0 - org;
  p[1] = point1 - org;
  p[2] = point2 - org;
  p[0] = Pointf(p[0][order[0]], p[0][order[1]], p[0][order[2]]);
  p[1] = Pointf(p[1][order[0]], p[1][order[1]], p[1][order[2]]);
  p[2] = Pointf(p[2][order[0]], p[2][order[1]], p[2][order[2]]);

  // Shear in XY.
  Point2f h[3];
  for (int k = 0; k < 3; k++) {
    h[k][0] = p[k][0] + shear[0] * p[k][2];
    h[k][1] = p[k][1] + shear[1] * p[k][2];
  }

  // Barycentric coordinates.
  float b[3] = {h[1][1] * h[2][0] - h[1][0] * h[2][1],
                h[2][1] * h[0][0] - h[2][0] * h[0][1],
                h[0][1] * h[1][0] - h[0][0] * h[1][1]};
  if (b[0] == 0.0f || b[1] == 0.0f || b[2] == 0.0f) {
    b[0] = double(h[1][1]) * h[2][0] - double(h[1][0]) * h[2][1];
    b[1] = double(h[2][1]) * h[0][0] - double(h[2][0]) * h[0][1];
    b[2] = double(h[0][1]) * h[1][0] - double(h[0][0]) * h[1][1];
  }
  if ((b[0] < 0.0f || b[1] < 0.0f || b[2] < 0.0f) &&
      (b[0] > 0.0f || b[1] > 0.0f || b[2] > 0.0f))
    return false;

  // Determinant.
  float q = b[0] + b[1] + b[2];
  if (q == 0.0f)
    return false;

  // Intersect parameter.
  float t = shear[2] * (b[0] * p[0][2] + b[1] * p[1][2] + b[2] * p[2][2]);
  if (((q > 0.0f) && !(t > tmin * q && t < tmax * q)) ||
      ((q < 0.0f) && !(t < tmin * q && t > tmax * q)))
    return false;

  tmax = t / q;
  u = b[1] / q;
  v = b[2] / q;
  return true;
}

namespace detail {

/// An axis-aligned bounding box.
template <typename Float, size_t Dim> struct BBox {
  using Coord = Eigen::Matrix<Float, Dim, 1>;

  /// Default constructor.
  ///
  /// Set the minimum to positive infinity and the maximum to
  /// negative infinity. This prepares the box for growing, such
  /// that the first `operator|=` call puts the box in a valid
  /// state containg only the right-hand side.
  ///
  BBox() noexcept {
    for (size_t i = 0; i < Dim; i++) {
      min[i] = +std::numeric_limits<Float>::infinity();
      max[i] = -std::numeric_limits<Float>::infinity();
    }
  }

  /// Construct from point.
  BBox(const Coord &pt) noexcept : min(pt), max(pt) {}

  /// Construct from minimum and maximum points.
  BBox(const Coord &minPt, const Coord &maxPt) noexcept
      : min(minPt), max(maxPt) {}

  /// Is valid? (Is minimum less than or equal to maximum?)
  bool isValid() const noexcept {
    for (size_t i = 0; i < Dim; i++)
      if (!(min[i] <= max[i]))
        return false;
    return true;
  }

  /// Test if this box contains a point.
  bool contains(const Coord &point) const noexcept {
    for (size_t i = 0; i < Dim; i++)
      if (!(min[i] <= point[i] && point[i] <= max[i]))
        return false;
    return true;
  }

  /// Test if this box contains another box.
  bool contains(const BBox &other) const noexcept {
    for (size_t i = 0; i < Dim; i++)
      if (!(min[i] <= other.min[i] && other.max[i] <= max[i]))
        return false;
    return true;
  }

  /// Test if this box overlaps another box.
  bool overlaps(const BBox &other) const noexcept {
    for (size_t i = 0; i < Dim; i++)
      if (!(min[i] <= other.max[i] && other.min[i] <= max[i]))
        return false;
    return true;
  }

  /// Grow operator.
  BBox &operator|=(const Coord &point) noexcept {
    min = min.cwiseMin(point);
    max = max.cwiseMax(point);
    return *this;
  }

  /// Grow operator.
  BBox &operator|=(const BBox &other) noexcept {
    min = min.cwiseMin(other.min);
    max = max.cwiseMax(other.max);
    return *this;
  }

  /// Index access operator.
  auto &operator[](size_t i) noexcept { return i == 0 ? min : max; }

  /// Index access operator.
  auto &operator[](size_t i) const noexcept { return i == 0 ? min : max; }

  Coord extent() const noexcept { return max - min; }

  int minAxis() const noexcept {
    int axis = 0;
    extent().minCoeff(&axis);
    return axis;
  }

  int maxAxis() const noexcept {
    int axis = 0;
    extent().maxCoeff(&axis);
    return axis;
  }

  /// Test for ray intersection.
  bool intersect(const Coord &org, const Coord &dir, Float *t0,
                 Float *t1) const {
    Coord invDir;
    for (size_t i = 0; i < Dim; i++)
      invDir[i] = Float(1) / dir[i];
    return intersect(org, dir, invDir, t0, t1);
  }

  /// Test for ray intersection.
  bool intersect(const Coord &org, const Coord &, const Coord &invDir,
                 Float *t0, Float *t1) const {
    int isNeg[Dim];
    for (size_t i = 0; i < Dim; i++)
      isNeg[i] = std::signbit(invDir[i]);
    auto pickMin = [&](size_t i) { return (*this)[isNeg[i]][i]; };
    auto pickMax = [&](size_t i) { return (*this)[1 - isNeg[i]][i]; };
    Float tmin = (pickMin(0) - org[0]) * invDir[0];
    Float tmax = (pickMax(0) - org[0]) * invDir[0];
    for (size_t i = 1; i < Dim; i++) {
      Float tmink = (pickMin(i) - org[i]) * invDir[i];
      Float tmaxk = (pickMax(i) - org[i]) * invDir[i];
      if (!(tmin < tmaxk && tmax > tmink))
        return false;
      tmin = std::max(tmin, tmink);
      tmax = std::min(tmax, tmaxk);
    }
    if (t0)
      *t0 = tmin;
    if (t1)
      *t1 = tmax;
    return true;
  }

  bool operator==(const BBox &other) const noexcept {
    return min[0] == other.min[0] && min[1] == other.min[1] &&
           min[2] == other.min[2] && max[0] == other.max[0] &&
           max[1] == other.max[1] && max[2] == other.max[2];
  }

  bool operator!=(const BBox &other) const noexcept {
    return !operator==(other);
  }

  void expand(float eps) noexcept {
    Coord center = (max + min) * Float(0.5);
    Coord extent = (max - min) * Float(0.5);
    min = center - extent * (1 + eps);
    max = center + extent * (1 + eps);
  }

  /// The minimum point.
  Coord min;

  /// The maximum point.
  Coord max;
};

} // namespace detail

using BBox = detail::BBox<double, 3>;
using BBox2 = detail::BBox<double, 2>;
using BBoxf = detail::BBox<float, 3>;
using BBox2f = detail::BBox<float, 2>;

// Same as exp(-x) but it is assumed that -x is actually negative! (i.e., x
// itself is positive).
template <typename Float> [[nodiscard]] inline Float expDecay(Float x) {
  Float x1 = x;
  Float y = 1 + x;
  y += (x *= x1 * Float(1.0 / 2.0));
  y += (x *= x1 * Float(1.0 / 3.0));
  y += (x *= x1 * Float(1.0 / 4.0));
  y += (x *= x1 * Float(1.0 / 5.0));
  return Float(1.0) / y;
}

/// \}

} // namespace d5
