/** Interface file for GridSampler
 *
 *  \file GridSampler.h
 */

#pragma once

#include <algorithm>
#include <random>
#include <stdexcept>
#include <vector>

namespace d5 {

/** A class to generate importance samples from a grid of values. The samples
 *  are returned as an XY pair with the values ranging from [-0.5,+0.5].
 */
class GridSampler {
public:
  struct Sample {
    double x = 0;
    double y = 0;
  };

  template <typename Callable, typename URNG>
  GridSampler(std::size_t xCount, std::size_t yCount,
              std::size_t avgPixelSampleCount, URNG &&rng,
              Callable &&callback) {
    setup(xCount, yCount, avgPixelSampleCount, std::forward<URNG>(rng),
          std::forward<Callable>(callback));
  }

  template <typename Callable, typename URNG>
  void setup(std::size_t xCount, std::size_t yCount,
             std::size_t avgPixelSampleCount, URNG &&rng, Callable &&callback);

  /** Get a sample.
   *
   *  The next sample from the randomized vector of samples is returned.
   *  If we have run out of samples in the vector, we shuffle the samples and
   *  reset the index back to the start. Since this function will most likely
   *  get used in a multi-threaded execution, we must ensure that the shuffle
   *  and index update are guarded.
   */
  template <typename URNG> Sample operator()(URNG &&rng) const {
    // make sure we have samples
    if (samples_.empty())
      return {0, 0};
    std::uniform_int_distribution<std::size_t> urng(0, samples_.size() - 1);
    return samples_[urng(rng)];
  }

private:
  /// The CDF of the grid, row-major ordering
  std::vector<Sample> samples_;
};

// ###########################################################
// #                 Template Implementation                 #
// ###########################################################

template <typename Callable, typename URNG>
void GridSampler::setup(std::size_t xCount, std::size_t yCount,
                        std::size_t avgPixelSampleCount, URNG &&rng,
                        Callable &&callback) {
  std::uniform_real_distribution<double> randGen;

  std::size_t avgSampleCount = xCount * yCount * avgPixelSampleCount;

  // sum the counts across the image
  double total = 0;
  for (std::size_t y = 0; y < yCount; y++) {
    for (std::size_t x = 0; x < xCount; x++) {
      total += callback(x, y);
    }
  }

  // start generating samples for each pixel
  samples_.clear();
  double norm = static_cast<double>(avgSampleCount) / total;
  double xNorm = 1. / static_cast<double>(xCount);
  double yNorm = 1. / static_cast<double>(yCount);
  for (std::size_t yy = 0; yy < yCount; yy++) {
    for (std::size_t xx = 0; xx < xCount; xx++) {
      // figure out how many samples for this pixel
      auto sampleCount = static_cast<int>(callback(xx, yy) * norm);

      // generate the samples for this pixel
      for (int sIndex = 0; sIndex < sampleCount; sIndex++) {
        Sample sample;
        sample.x = (xx + randGen(rng)) * xNorm - 0.5;
        sample.y = (yy + randGen(rng)) * yNorm - 0.5;
        samples_.push_back(sample);
      }
    }
  }
}

} // namespace d5
