#pragma once

#include <cstdint>
#include <cstring>
#include <limits>

namespace d5 {

/// A half-precision float.
struct Half {
public:
  Half() noexcept = default;

  Half(float f) noexcept;

public:
  /// Cast to float.
  operator float() const noexcept;

  /// Cast to double.
  operator double() const noexcept { return double(float(*this)); }

  /// Cast to bit pattern.
  explicit operator std::uint16_t &() noexcept { return b_; }

  /// Cast to bit pattern.
  explicit operator const std::uint16_t &() const noexcept { return b_; }

public:
  /// Absolute value.
  Half abs() const noexcept {
    Half h = *this;
    h.b_ = h.b_ & ~0x8000;
    return h;
  }

  /// Is negative? (possibly `-0.0`)
  bool signbit() const noexcept { return (b_ & 0x8000); }

  /// Is infinite?
  bool isinf() const noexcept { return (b_ & 0x7FFF) == 0x7C00; }

  /// Is not-a-number?
  bool isnan() const noexcept {
    return (b_ & 0x7FFF) != 0x7C00 && (b_ & 0x7C00) == 0x7C00;
  }

  /// Is neither infinite nor not-a-number?
  bool isfinite() const noexcept { return (b_ & 0x7C00) != 0x7C00; }

  /// Is neither infinite nor not-a-number nor subnormal?
  bool isnormal() const noexcept {
    return (b_ & 0x7C00) != 0x7C00 && (b_ & 0x7C00) != 0;
  }

  Half &operator+=(float f) { return *this = float(*this) + f; }
  Half &operator-=(float f) { return *this = float(*this) - f; }
  Half &operator*=(float f) { return *this = float(*this) * f; }
  Half &operator/=(float f) { return *this = float(*this) / f; }

private:
  /// Bit pattern.
  std::uint16_t b_ = 0;

  static_assert(std::numeric_limits<float>::is_iec559,
                "Half requires IEC-559/IEEE-754 single precision floats");
};

inline Half::Half(float f) noexcept {
  std::uint32_t u;
  std::memcpy(&u, &f, sizeof(f));
  if (f == 0) {
    b_ = u >> 16;
  } else {
    std::int32_t e = (u >> 23) & 0x01FF;
    std::int32_t s = (u >> 31) & 0x0001;
    std::int32_t m = u & 0x007FFFFF;
    if (e > 0x0070 && e < 0x008E) {
      e = (e - 0x0070) * 0x0400;
      b_ = e + ((m + ((m >> 13) & 1) + 0x0FFF) >> 13);
    } else if (e > 0x0170 && e < 0x018E) {
      e = (e - 0x0170) * 0x0400 + 0x8000;
      b_ = e + ((m + ((m >> 13) & 1) + 0x0FFF) >> 13);
    } else {
      s <<= 15;
      e &= 255;
      e -= 112;
      if (e < 1) {
        if (e < -10) {
          // Zero.
          b_ = s;
        } else {
          // Subnormal.
          m |= 0x00800000;
          b_ = s | ((m + ((1 << (13 - e)) + ((m >> (14 - e)) & 1) - 1)) >>
                    (14 - e));
        }
      } else if (e == 0x8F) {
        if (m == 0) {
          // Inf.
          b_ = s | 0x7C00;
        } else {
          // NaN.
          m >>= 13;
          b_ = s | 0x7C00 | m | (m == 0);
        }
      } else {
        m += ((m >> 13) & 1) + 0x0FFF;
        if (m & 0x00800000) {
          m = 0;
          e++;
        }
        // Overflow?
        if (e > 30) {
          // Inf.
          b_ = s | 0x7C00;
        } else {
          // Normal.
          b_ = s | (e << 10) | (m >> 13);
        }
      }
    }
  }
}

inline Half::operator float() const noexcept {
  float f;
  std::uint32_t u;
  std::int32_t e = (b_ >> 10) & 0x001F;
  std::int32_t s = (b_ >> 15) & 0x0001;
  std::int32_t m = b_ & 0x03FF;
  if (e == 0) {
    if (m == 0) {
      // Zero.
      u = s << 31;
    } else {
      // Subnormal.
      e += 112;
      while (!(m & 0x0400))
        e--, m <<= 1;
      e++;
      m &= ~0x0400;
      u = (s << 31) | (e << 23) | (m << 13);
    }
  } else if (e == 31) {
    if (m == 0) {
      // Inf.
      u = (s << 31) | 0x7F800000;
    } else {
      // NaN.
      u = (s << 31) | 0x7F800000 | (m << 13);
    }
  } else {
    // Normal.
    e += 112;
    u = (s << 31) | (e << 23) | (m << 13);
  }
  std::memcpy(&f, &u, sizeof(f));
  return f;
}

} // namespace d5
