/** Interface file for some inheritance patterns
 *
 *  \file Inheritance.h
 */

#pragma once

#include <stdexcept>
#include <string>
#include <type_traits>

#include "TypeTraits.h"

namespace d5 {

/** \defgroup common_inheritance Enumerated Inheritance Pattern
 *  Lightweight RTTI replacement for classes where subclasses can be enumerated
 *  \ingroup common
 */

/// \addtogroup common_inheritance
/// \{

namespace detail {

/// Helper struct to determine whether a type has a `subclass_type` member
template <typename T> class has_subclass_type_member {
  using one = char;
  struct two {
    char x[2];
  };

  template <typename C> static one test(decltype(&C::subclass_type));
  template <typename C> static two test(...);

public:
  static constexpr bool value = sizeof(test<T>(0)) == sizeof(char);
};

/// Constexpr test for whether a type has a `subclass_type` method
template <typename T>
inline constexpr bool has_subclass_type_member_v =
    has_subclass_type_member<T>::value;
} // namespace detail

/** A CRTP class that can be used to designate a class to be a superclass in
 *  an "enumerated inheritance" framework. This is essentially a lightweight
 *  RTTI dynamic_cast replacement. The requirements are:
 *
 *  1. An enum exists with a unique element per subclass
 *  2. The base type has a .type() method that returns the type of the subclass.
 *  3. The base type must have a constructor that takes a single parameter of
 *     the enum type (representing the type of the subclass). It is recommended
 *     that this constructor be protected.
 *  4. The base type inherits from this type
 *  5. The base type must have a virtual destructor
 *
 *  Subclasses should then inherit from Derived::Subclass<SubclassType>, where
 *  SubclassType is the enum value that they represent.
 *
 *  Example:
 *  \code
 *  class Animal : public EnumeratedSuperclass<Animal> {
 *   public:
 *    enum class Type { Dog, Cat };
 *
 *    virtual ~Animal() = default;
 *
 *    std::string void speak() = 0;
 *
 *   protected:
 *     Animal(Type type) : type_(type) {}
 *   private:
 *    Type type_;
 *  };
 *
 *  class Dog : public Animal::Subclass<Animal::Type::Dog> {
 *    ~Dog() override = default;
 *    std::string speak() override { return "woof"; }
 *  };
 *
 *  class Cat : public Animal::Subclass<Animal::Type::Cat> {
 *    ~Cat() override = default;
 *    std::string speak() override { return "meow"; }
 *  };
 *  \endcode
 */
template <class Derived> struct EnumeratedSuperclass {
  /// Subclasses should inherit from this type instead of directly from Derived
  template <auto SubclassType> class Subclass : public Derived {
  public:
    static_assert(std::is_same_v<decltype(SubclassType),
                                 decltype(std::declval<Derived>().type())>,
                  "Template parameter must be the same as the return type of "
                  "Derived::type()");

    /// Virtual Destructor
    ~Subclass() override = default;

    /// Advertise the type (used for as<T>() downcasting)
    static constexpr auto subclass_type = SubclassType;

  protected:
    /// Constructor, passes up the type
    Subclass() : Derived(SubclassType) {}
  };

  /// Downcast helper, throws if invalid downcast is performed
  template <typename T> [[nodiscard]] const T &as() const {
    if (!is<T>())
      throw std::runtime_error("invalid downcast");
    return static_cast<const T &>(*this);
  }

  /// Downcast helper, throws if invalid downcast is performed
  template <typename T> [[nodiscard]] T &as() {
    return const_cast<T &>(
        static_cast<const EnumeratedSuperclass<Derived> *>(this)->as<T>());
  }

  /// Tests whether the object is of the given type
  template <typename T> [[nodiscard]] bool is() const {
    static_assert(std::is_base_of_v<Derived, T>);
    static_assert(detail::has_subclass_type_member_v<T>,
                  "T was not derived from EnumeratedSubclass");
    return static_cast<const Derived *>(this)->type() == T::subclass_type;
  }
};

/// \}

} // namespace d5
