/** Interface file for working with JSON
 *
 *  \file Json.h
 */

#pragma once

#include <string>

#include <Eigen/Dense>

#if defined(JSON_DIAGNOSTICS)
#undef JSON_DIAGNOSTICS
#endif
#define JSON_DIAGNOSTICS 1

#if defined(__APPLE__)
#if defined(JSON_HAS_FILESYSTEM)
#undef JSON_HAS_FILESYSTEM
#endif
#define JSON_HAS_FILESYSTEM 0
#define JSON_HAS_CPP_14 1
#endif

#include <nlohmann/json.hpp>

#include "dirsig5/base/Constants.h"
#include "dirsig5/base/Macros.h"

namespace d5 {

/** \defgroup common_json JSON
 *  Some minor additions to nlohmann::json
 *  \ingroup common
 */

/// \addtogroup common_json
/// \{

/// We use nlohmann's library for JSON
using Json = nlohmann::json;

/** Parse a required field out of JSON.
 *
 *  \param j    The JSON containing the field
 *  \param key  The key of the field
 *  \param v    The value to parse into
 */
template <typename Value>
inline void Required(const Json &j, const char *key, Value &v) {
  j.at(key).get_to(v);
}

/// \overload
template <typename Value>
inline void Required(const Json &j, const std::string &key, Value &v) {
  return Required(j, key.c_str(), v);
}

/** Parse an optional field out of JSON.
 *
 *  \param j    The JSON containing the field
 *  \param key  The key of the field
 *  \param v    The value to parse into
 */
template <typename Value>
inline bool Optional(const Json &j, const char *key, Value &v) {
  if (auto it = j.find(key); it != j.end()) {
    it.value().get_to(v);
    return true;
  }
  return false;
}

/// \overload
template <typename Value>
inline bool Optional(const Json &j, const std::string &key, Value &v) {
  return Optional(j, key.c_str(), v);
}

/** Generates the documented function header for a to_json overload
 *
 *  \param T  The type being encoded to JSON
 *  \param j  The name to use for the JSON parameter
 *  \param v  The name to use for the value parameter
 */
#define D5_TO_JSON(T, j, v)                                                    \
  /** Encodes a \ref T to JSON                                                 \
     \param[out] j  The JSON to which to encode                                \
     \param[in] v   The value to encode                                        \
   */                                                                          \
  inline void to_json(::d5::Json &j, const T &v) /* NOLINT */

/** Generates the documented function header for a from_json overload
 *
 *  \param T  The type being decoded from JSON
 *  \param j  The name to use for the JSON parameter
 *  \param v  The name to use for the value parameter
 */
#define D5_FROM_JSON(T, j, v)                                                  \
  /** Decodes a \ref T from JSON                                               \
     \param[in] j   The JSON from which to decode                              \
     \param[out] v  The value to decode                                        \
   */                                                                          \
  inline void from_json(const ::d5::Json &j, T &v) /* NOLINT */

#define D5_JSON_SERIALIZE_ENUM(...)                                            \
  D5_EXPAND(NLOHMANN_JSON_SERIALIZE_ENUM(__VA_ARGS__))                         \
  static_assert(true, "Remember the ; after D5_JSON_SERIALIZE_ENUM")

/// \}

} // namespace d5

/// \cond HIDDEN_SYMBOLS

namespace nlohmann {

/// JSON conversion for generic matrices
template <typename Value, int Rows, int Cols, int... Args>
struct adl_serializer<Eigen::Matrix<Value, Rows, Cols, Args...>> {
  static void to_json(json &j,
                      const Eigen::Matrix<Value, Rows, Cols, Args...> &m) {
    for (int r = 0; r < m.rows(); r++) {
      auto row = json::array();
      for (int c = 0; c < m.cols(); c++)
        row.push_back(m(r, c));
      j.push_back(row);
    }
  }

  static void from_json(const json &j,
                        Eigen::Matrix<Value, Rows, Cols, Args...> &m) {
    // Make sure it is a filled array, where all rows are the same
    if (!j.is_array() || j.empty()) {
      throw std::runtime_error("    [json.exception.parse_error.0] parse "
                               "error: Eigen::Matrix JSON must be 2D array");
    }
    int foundCols = -1;
    for (size_t i = 0; i < j.size(); i++) {
      if (!j[i].is_array()) {
        throw std::runtime_error("    [json.exception.parse_error.0] parse "
                                 "error: Eigen::Matrix JSON must be 2D array");
      }
      if (foundCols == -1) {
        foundCols = static_cast<int>(j[i].size());
      }
      if (foundCols != int(j[i].size())) {
        throw std::runtime_error(
            "    [json.exception.parse_error.0] parse error: Each row of "
            "Eigen::Matrix JSON must have the same size");
      }
    }
    // Resize the matrix if we can
    if ((Rows == Eigen::Dynamic && Cols == Eigen::Dynamic) ||
        (Rows == Eigen::Dynamic && Cols == foundCols) ||
        (Rows == j.size() && Cols == Eigen::Dynamic))
      m.resize(j.size(), foundCols);
    if (int(j.size()) != m.rows() || foundCols != int(m.cols())) {
      throw std::runtime_error("    [json.exception.parse_error.0] parse "
                               "error: Cannot resize matrix "
                               "to fit dimensions found in JSON");
    }
    // Read in the matrix
    for (size_t r = 0; r < j.size(); r++) {
      for (size_t c = 0; c < j[r].size(); c++) {
        j[r][c].get_to(m(r, c));
      }
    }
  }
};

/// JSON conversions specifically for column vectors
template <typename Value, int Rows, int... Args>
struct adl_serializer<Eigen::Matrix<Value, Rows, 1, Args...>> {
  static void to_json(json &j,
                      const Eigen::Matrix<Value, Rows, 1, Args...> &m) {
    j = json::array();
    for (int r = 0; r < m.rows(); r++) {
      j.push_back(m(r, 0));
    }
  }

  static void from_json(const json &j,
                        Eigen::Matrix<Value, Rows, 1, Args...> &m) {
    if (!j.is_array() || j.empty()) {
      throw std::runtime_error("    [json.exception.parse_error.0] parse "
                               "error: Eigen::Matrix JSON must be 2D array");
    }
    // Resize if we can
    if (Rows == Eigen::Dynamic) {
      m.resize(j.size(), 1);
    }
    if (int(j.size()) != m.rows()) {
      throw std::runtime_error("    [json.exception.parse_error.0] parse "
                               "error: Cannot resize matrix "
                               "to fit dimensions found in JSON");
    }
    // Parse the JSON
    for (int r = 0; r < m.rows(); r++) {
      if (j[r].is_number()) {
        j[r].get_to(m(r, 0));
      } else if (j[r].is_array() && j[r].size() == 1u) {
        j[r][0].get_to(m(r, 0));
      } else {
        throw std::runtime_error(
            "    [json.exception.parse_error.0] parse error: Eigen::Vector "
            "JSON must be either a 1D array or a 2D array where all inner "
            "lists are size 1");
      }
    }
  }
};

/// JSON conversion for quaternions
template <typename Value, int Options>
struct adl_serializer<Eigen::Quaternion<Value, Options>> {
  static void to_json(json &j, const Eigen::Quaternion<Value, Options> &q) {
    j = json::array({q.w(), q.x(), q.y(), q.z()});
  }
  static void from_json(const json &j, Eigen::Quaternion<Value, Options> &q) {
    j.at(0).get_to(q.w());
    j.at(1).get_to(q.x());
    j.at(2).get_to(q.y());
    j.at(3).get_to(q.z());
  }
};

/// \endcond
} // namespace nlohmann
