#pragma once

#include <algorithm>
#include <cmath>
#include <functional>
#include <type_traits>
#include <utility>

namespace d5 {

/// \addtogroup common
/// \{

/// Linear interpolate between floating point values.
template <typename Float>
constexpr Float lerp(Float t, Float x0, Float x1) noexcept {
  static_assert(std::is_floating_point_v<Float>);
  return (1 - t) * x0 + t * x1;
}

/** Bilinear interpolate between floating point values.
 *
 *  \note
 *  This function is implemented such that, for `s` and `t`
 *  each equal to either 0 or 1,
 *  ```
 *  bilerp(s, t, x[0][0], x[0][1], x[1][0], x[1][1]) == x[s][t]
 *  ```
 *
 *  \tparam Float The floating point type being used
 *
 */
template <typename Float>
constexpr Float bilerp(Float s, Float t, Float x00, Float x01, Float x10,
                       Float x11) noexcept {
  static_assert(std::is_floating_point_v<Float>);
  return lerp(s, lerp(t, x00, x01), lerp(t, x10, x11));
}

/// Wrap floating point value onto interval [minValue, maxValue].
template <typename Float>
inline Float wrapFloat(Float value, Float minValue, Float maxValue) {
  static_assert(std::is_floating_point_v<Float>);
  if (minValue == maxValue)
    return minValue;
  maxValue -= minValue;
  value -= minValue;
  value = std::remainder(value, maxValue);
  if (value < 0)
    value += maxValue;
  return value + minValue;
}

/// Find adjacent iterators for linear interpolation.
template <typename Iterator, typename Value, typename Func>
inline auto findLerpIterators(Iterator from, Iterator to, const Value &value,
                              Func &&func) {
  Iterator itr0 = std::lower_bound(from, to, value, std::forward<Func>(func));
  Iterator itr1 = itr0;
  if (from != to) {
    if (itr0 != from)
      --itr0;
    if (itr1 == to)
      --itr1;
  }
  return std::make_pair(itr0, itr1);
}

/// Find indexes of adjacent iterators for linear interpolation.
template <typename Iterator, typename Value, typename Func>
inline auto findLerpIndexes(Iterator from, Iterator to, const Value &value,
                            Func &&func) {
  auto itrs = findLerpIterators(from, to, value, std::forward<Func>(func));
  return std::make_pair(std::distance(from, itrs.first), //
                        std::distance(from, itrs.second));
}

/// Overload with comparison function defaulting to `std::less<>()`.
template <typename Iterator, typename Value>
inline auto findLerpIterators(Iterator from, Iterator to, const Value &value) {
  return findLerpIterators(from, to, value, std::less<>());
}

/// Overload with comparison function defaulting to `std::less<>()`.
template <typename Iterator, typename Value>
inline auto findLerpIndexes(Iterator from, Iterator to, const Value &value) {
  return findLerpIndexes(from, to, value, std::less<>());
}

template <typename Float>
inline Float findLerpFactor(Float value, Float minValue, Float maxValue) {
  static_assert(std::is_floating_point_v<Float>);
  if (minValue == maxValue)
    return 0;
  return (value - minValue) / (maxValue - minValue);
}

template <typename Float>
inline Float findLerpFactorClamped(Float value, Float minValue,
                                   Float maxValue) {
  static_assert(std::is_floating_point_v<Float>);
  Float factor = findLerpFactor(value, minValue, maxValue);
  factor = std::fmin(factor, Float(1));
  factor = std::fmax(factor, Float(0));
  return factor;
}

/// \}

} // namespace d5
