#pragma once

#include <map>
#include <memory>
#include <vector>

#include "dirsig5/base/Export.h"
#include "dirsig5/base/Library.h"

namespace d5 {

/// A basic library manager.
///
/// This constitutes the base functionality of the plugin
/// library manager. It serves as a collection of dynamically
/// loaded libraries which all have the same, properly resolved
/// entry-point function name.
///
class D5_BASE_EXPORT LibraryManager {
public:
  LibraryManager(std::string funcName) : funcName(std::move(funcName)) {}

  /// Load libraries from a directory.
  ///
  /// This scans the specified directory for files with dynamic-library
  /// extensions (.so, .dylib, .dll) and loads in all libraries which
  /// successfully resolve the entry point function.
  ///
  /// \returns The number of libraries loaded.
  ///
  long loadLibraryDirectory(const std::string &dirname);

  /// Try to load a library.
  ///
  /// \returns True if success, false if failure. Note that failure
  /// may mean that 1) the file does not exists or is not a dynamic
  /// library or 2) the entry point function could not be resolved.
  ///
  bool loadLibrary(const std::string &filename);

  /// Get the entry point function for the given library.
  ///
  /// \param[in] name
  /// The library name. This may be the fully-specified filename (as
  /// per the string passed to `loadLibrary()`) or just the basename,
  /// excluding the `lib` prefix on UNIX platforms.
  ///
  [[nodiscard]] void *getLibraryFunc(const std::string &name) const;

  /// Is the given library loaded?
  [[nodiscard]] bool hasLibrary(const std::string &name) const {
    return getLibraryFunc(name) != nullptr;
  }

public:
  /// The entry point function name.
  const std::string funcName;

  /// The map of filenames to library/entry-point pairs.
  std::map<std::string, std::pair<Library, void *>> libraries;

public:
  /// Find all candidate library filenames in the given directory.
  ///
  /// This returns all filenames in the given directory ending with
  /// .so, .dll, or .dylib. This is not recursive, i.e., does not search
  /// subdirectories.
  ///
  [[nodiscard]] static std::vector<std::string>
  FindFilenames(const std::string &dirname);
};

} // namespace d5
