#pragma once

#include <array>
#include <memory>
#include <string>

// These are intentionally in this order
// clang-format off
#include <spdlog/spdlog.h> // NOLINT
#include <spdlog/pattern_formatter.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/sinks/stdout_color_sinks.h>
#include <spdlog/fmt/ostr.h>
// clang-format on

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Export.h"

namespace d5 {

/** Helper class to hold options for the logger */
class D5_BASE_EXPORT LoggerOptions {
public:
  /** Accessors and mutators for minimum logging level to print */
  [[nodiscard]] spdlog::level::level_enum minLevel() const { return minLevel_; }
  LoggerOptions &withMinLevel(spdlog::level::level_enum l) {
    minLevel_ = l;
    return *this;
  }

  /** Accessors and mutators for output filename (optional) */
  [[nodiscard]] const std::string &filename() const { return filename_; }
  LoggerOptions &withFilename(const std::string &f) {
    filename_ = f;
    return *this;
  }

private:
  spdlog::level::level_enum minLevel_ = spdlog::level::info;
  std::string filename_;
};

/// Singleton class to hold the default logger for DIRSIG
class D5_BASE_EXPORT Logger {
public:
  /// Singleton instance accessor for the logger
  static Logger &getInstance();

  /** Sets up the default logger. You only need to call this if you want
   *  non-default behavior, as the first call to one of the DIRSIG_ macros will
   *  initialize the singleton if you have not.
   *
   *  \param options  The options to use when setting up the logger
   */
  void reinit(const LoggerOptions &options);

  /// Accessor for the logger
  spdlog::logger *logger() { return logger_.get(); }

private:
  // Constructor, sets the logger up with the default options
  Logger() { reinit(LoggerOptions()); }

  std::unique_ptr<spdlog::logger> logger_;
};

/** Convenience method for calling
 * d5::Logger::getInstance().reinit() */
inline void setupLogger(const LoggerOptions &options = LoggerOptions()) {
  Logger::getInstance().reinit(options);
}

} // namespace d5

/** MSVC workaround for __VA_ARGS__ */
#define DIRSIG_LOG_EXPAND(x) x

/** Macro to log a trace-level message. Trace messages are used for developers
 *  to debug or track fine-level code issues.
 */
#define DIRSIG_TRACE(...)                                                      \
  DIRSIG_LOG_EXPAND(                                                           \
      SPDLOG_LOGGER_TRACE(::d5::Logger::getInstance().logger(), __VA_ARGS__))

/** Macro to log a debug-level message. Debug messages are for use by developers
 *  and/or sysadmins to diagnose larger issues with the program.
 */
#define DIRSIG_DEBUG(...)                                                      \
  DIRSIG_LOG_EXPAND(                                                           \
      SPDLOG_LOGGER_DEBUG(::d5::Logger::getInstance().logger(), __VA_ARGS__))

/** Macro to log an info-level message. Info messages are used to print info
 *  about program state that may be helpful to diagnose issues if something
 *  goes wrong, but is non-critical under normal circumstances.
 */
#define DIRSIG_INFO(...)                                                       \
  DIRSIG_LOG_EXPAND(                                                           \
      SPDLOG_LOGGER_INFO(::d5::Logger::getInstance().logger(), __VA_ARGS__))

/** Macro to log a warning-level message. Warning messages should be used to
 *  signify odd conditions that are recoverable, but may lead to unexpected
 *  behavior.
 */
#define DIRSIG_WARN(...)                                                       \
  DIRSIG_LOG_EXPAND(                                                           \
      SPDLOG_LOGGER_WARN(::d5::Logger::getInstance().logger(), __VA_ARGS__))

/** Macro to log an error-level message. Error messages should be used when
 *  conditions arise that are fatal to the current procedure, but not
 *  necessarily the whole application.
 */
#define DIRSIG_ERROR(...)                                                      \
  DIRSIG_LOG_EXPAND(                                                           \
      SPDLOG_LOGGER_ERROR(::d5::Logger::getInstance().logger(), __VA_ARGS__))

/** Macro to log a fatal-level message. Fatal messages should be used when
 *  conditions arise that are fatal to the entire application.
 */
#define DIRSIG_CRITICAL(...)                                                   \
  DIRSIG_LOG_EXPAND(SPDLOG_LOGGER_CRITICAL(                                    \
      ::d5::Logger::getInstance().logger(), __VA_ARGS__))

namespace d5 {

inline void dumpError(const d5::Error &error) {
  std::string message = error.what();
  if (message.empty())
    message = "Unknown error.";
  if (('a' <= message[0] && message[0] <= 'z') ||
      ('A' <= message[0] && message[0] <= 'Z'))
    // Upper case for consistency
    message[0] = ::toupper(static_cast<unsigned char>(message[0]));
  DIRSIG_ERROR("Error!\n"
               "  {}\n\n"
               "  Thrown by '{}' ({}:{})\n",
               message, d5::detail::demangle(error.funcName), error.fileName,
               error.lineNo);
  if (!error.traceMessages.empty())
    for (auto &trace : error.traceMessages)
      DIRSIG_ERROR("  ... while {}", trace);
  if (!error.debugMessages.empty()) {
    DIRSIG_ERROR("");
    DIRSIG_ERROR("  Debug Log:");
    for (auto &debug : error.debugMessages)
      DIRSIG_ERROR("    - {}", debug);
  }
}

} // namespace d5
