#pragma once

#include <algorithm>
#include <memory>
#include <vector>

namespace d5 {

/** A memory arena.
 *
 * This is a helper structure to allocate memory more efficiently.
 * It allocates memory in larger blocks (65KB by default), then
 * distributes the memory as requested by client code. Sub-allocations
 * may not be freed specifically. All memory must be cleared or reset
 * all at once.
 */
class MemoryArena {
public:
  typedef char Byte;

  typedef std::allocator<char> ByteAllocator;

  MemoryArena(std::size_t blockSize = 0) : blockSize_(blockSize) {
    // Round up to 256 byte interval.
    blockSize_ += 255u;
    blockSize_ &= ~255u;
    if (blockSize_ == 0) {
      blockSize_ = 65536;
    }

    // Allocate initial block.
    block_.size = blockSize_;
    block_.begin = byteAlloc_.allocate(block_.size);
    block_.offset = 0;

    // Reserve blocks.
    freeBlocks_.reserve(4);
    fullBlocks_.reserve(4);
  }

  MemoryArena(const MemoryArena &) = delete;

  ~MemoryArena() {
    byteAlloc_.deallocate(block_.begin, block_.size);
    for (Block &freeBlock : freeBlocks_)
      byteAlloc_.deallocate(freeBlock.begin, freeBlock.size);
    for (Block &fullBlock : fullBlocks_)
      byteAlloc_.deallocate(fullBlock.begin, fullBlock.size);
  }

public:
  void *allocate(std::size_t size) {
    // Round up to 16 byte interval.
    size += 15u;
    size &= ~15u;
    if (size == 0) {
      return nullptr;
    }

    if (block_.size < block_.offset + size) {
      fullBlocks_.emplace_back(block_);
      if (freeBlocks_.size() == 0 || freeBlocks_.back().size < size) {
        // Allocate block.
        block_.size = std::max(blockSize_, size);
        block_.begin = byteAlloc_.allocate(block_.size);
        block_.offset = 0;
      } else {
        // Use free block.
        block_ = freeBlocks_.back();
        freeBlocks_.pop_back();
      }
    }

    Byte *pos = block_.begin + block_.offset;
    block_.offset += size;
    return static_cast<void *>(pos);
  }

  /// Clear.
  void clear() {
    block_.offset = 0;
    freeBlocks_.reserve(freeBlocks_.size() + fullBlocks_.size());
    for (Block &fullBlock : fullBlocks_) {
      freeBlocks_.push_back(fullBlock);
      freeBlocks_.back().offset = 0;
    }
    fullBlocks_.clear();
  }

  /// Clear _and_ free everything.
  void reset() {
    block_.offset = 0;
    for (Block &freeBlock : freeBlocks_)
      byteAlloc_.deallocate(freeBlock.begin, freeBlock.size);
    for (Block &fullBlock : fullBlocks_)
      byteAlloc_.deallocate(fullBlock.begin, fullBlock.size);
    freeBlocks_.clear();
    fullBlocks_.clear();
  }

private:
  struct Block {
    Byte *begin;
    std::size_t offset;
    std::size_t size;
  };

  /// The block size.
  std::size_t blockSize_;

  /// The current block.
  Block block_;

  /// The list of free blocks.
  std::vector<Block> freeBlocks_;

  /// The list of full blocks.
  std::vector<Block> fullBlocks_;

  /// The byte allocator.
  ByteAllocator byteAlloc_;
};

} // namespace d5
