#pragma once

#include <cstddef>
#include <cstring>
#include <memory>
#include <vector>

#include "dirsig5/base/Error.h"

namespace d5 {

/** A memory pool.
 *
 * This is a helper structure to allocate memory more efficiently.
 * It allocates memory for fixed-size elements 1 at a time from fixed-sized
 * pools. Unlike a memory arena, however, this structure also allows elements
 * to be freed and returned to the pool.
 */
class MemoryPool {
private:
  typedef char Byte;

  struct Pool {
    Pool *next;      ///< Pointer to next pool.
    Byte *begin;     ///< Pointer to elements.
    Byte *firstFree; ///< Pointer to first free element.
  };

  typedef std::allocator<Byte> ByteAllocator;

  typedef std::allocator<Pool> PoolAllocator;

public:
  MemoryPool(std::size_t elemSize, std::size_t elemsPerPool)
      : elemSize_(elemSize), elemsPerPool_(elemsPerPool) {
    if (elemSize_ < sizeof(void *))
      elemSize_ = sizeof(void *);
    if (elemsPerPool_ < 1)
      elemsPerPool_ = 1;
  }

  MemoryPool(const MemoryPool &) = delete;

  ~MemoryPool() { reset(); }

public:
  /// Allocate element.
  void *allocate() {
    Pool *pool = head_;
    while (pool and not pool->firstFree)
      pool = pool->next;
    if (not pool)
      pool = AllocPool();
    Byte *elem = pool->firstFree;
    std::memcpy(&pool->firstFree, pool->firstFree, sizeof(void *));
    return static_cast<void *>(elem);
  }

  /// Deallocate element.
  void deallocate(void *ptr) {
    if (ptr == nullptr)
      return;

    Byte *elem = static_cast<Byte *>(ptr);
    Pool *pool = head_;
    while (pool) {
      // Element in range of pool?
      std::ptrdiff_t pool_size = elemSize_ * elemsPerPool_;
      std::ptrdiff_t elem_diff = elem - pool->begin;
      if (elem_diff >= 0 && elem_diff < pool_size) {
        // Element in range of pool, but pointer is not
        // aligned to element boundary?
        DIRSIG_THROW_IF(elem_diff % std::ptrdiff_t(elemSize_) != 0,
                        "Pointer misaligned!");
        break;
      } else
        // Search next pool.
        pool = pool->next;
    }

    // Element not in range of any pool?
    DIRSIG_THROW_IF(not pool, "Pointer not in range of any pool!");

    // Prepend.
    std::memcpy(elem, &pool->firstFree, sizeof(void *));
    pool->firstFree = elem;
  }

  /// Clear.
  void clear() {
    for (Pool *pool = head_; pool; pool = pool->next)
      ClearPool(pool);
  }

  /// Clear and free everything.
  void reset() {
    for (Pool *pool = head_; pool;) {
      // Deallocate.
      Pool *next = pool->next;
      byteAlloc_.deallocate(pool->begin, elemSize_ * elemsPerPool_);
      poolAlloc_.deallocate(pool, 1);
      pool = next; // Increment.
    }
    // Nullify.
    head_ = nullptr;
  }

private:
  std::size_t elemSize_ = 0;     ///< The element size in bytes.
  std::size_t elemsPerPool_ = 0; ///< The elements per pool.
  Pool *head_ = nullptr;         ///< The pool list.
  ByteAllocator byteAlloc_;      ///< The byte allocator.
  PoolAllocator poolAlloc_;      ///< The pool allocator.

private:
  Pool *AllocPool() {
    Pool *pool = poolAlloc_.allocate(1);
    pool->next = nullptr;
    pool->begin = byteAlloc_.allocate(elemSize_ * elemsPerPool_);
    ClearPool(pool);

    // First pool?
    if (head_ == nullptr) {
      head_ = pool; // Set head.
    } else {
      // Append.
      Pool *tail = head_;
      while (tail->next)
        tail = tail->next;
      tail->next = pool;
    }

    return pool;
  }

  void ClearPool(Pool *pool) {
    // Link all elements sequentially.
    pool->firstFree = pool->begin;
    for (std::size_t index = 0; index + 1 < elemsPerPool_; index++) {
      Byte *elem0 = pool->begin + (index + 0) * elemSize_;
      Byte *elem1 = pool->begin + (index + 1) * elemSize_;
      std::memcpy(elem0, &elem1, sizeof(void *));
    }

    Byte *elem0 = pool->begin + (elemsPerPool_ - 1) * elemSize_;
    Byte *elem1 = nullptr;
    std::memcpy(elem0, &elem1, sizeof(void *));
  }
};

} // namespace d5
