#pragma once

#include <array>
#include <complex>

#include "dirsig5/base/Constants.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Pcg.h"

namespace d5 {

/// \addtogroup common
/// \{

/**
 * This is a lightly modified version of the original Boren-Huffman (1983) Mie
 * scattering code to generate a scattering phase function and
 * scattering efficiency factor for single embedded sphere, with heavy use
 * of the version translated to C by P. J. Flatau (Feb 1998).
 */
class Mie {
public:
  /**
   * Calling function for the Mie code. Generates 2*nAngles-1 samples
   * of the scattering phase function (angle from forward) distributed
   * evenly and including 0, 90 and 180. The scattering efficiency
   * factor and average cosine are also output for reference.
   */
  static void
  bhmie(double iorMedium, /* real refractive index of the medium */
        const std::complex<double> &iorSphere, /* complex ior of sphere */
        double radius,                         /* [um] */
        double wavelength,                     /* [um] */
        size_t nAngles,           /* number of output angles in [0,90] */
        std::vector<double> &spf, /* scattering efficiency [0:PI] */
        double &qScat,            /* efficiency factor for scattering */
        double &gScat) {          /* average cosine */
    // relative refractive index
    std::complex<double> iorRelative = iorSphere / iorMedium;

    // size parameter
    double x = TwoPI * radius * iorMedium / wavelength;

    // resize the scattering phase function to hold the output
    spf.resize(nAngles * 2 - 1);
    std::fill(spf.begin(), spf.end(), 0.);

    // vars
    std::complex<double> cxan, cxan1, cxbn, cxbn1, cxxi, cxxi0, cxy, cxxi1;
    double apsi, apsi0, apsi1, chi, chi0, chi1, dang, fn, p, rn, t, theta,
        xstop, ymod;
    double dn, dx, psi, psi0, psi1;
    unsigned n, nmx, nn, nstop;

    static thread_local std::vector<std::complex<double>> cxd;
    std::vector<double> amu(nAngles);
    std::vector<double> pi(nAngles);
    std::vector<double> pi0(nAngles);
    std::vector<double> pi1(nAngles);
    std::vector<double> tau(nAngles);
    std::vector<std::complex<double>> cxs1(2 * nAngles - 1),
        cxs2(2 * nAngles - 1);

    dx = x;
    cxy = x * iorRelative;

    /* Series expansion terminated after NSTOP terms */
    xstop = x + 4. * pow(x, 0.3333) + 2.0;
    nstop = xstop;
    ymod = std::abs(cxy);
    nmx = std::max(xstop, ymod) + 15;

    cxd.resize(nmx);

    // generate cosines of angles
    dang = 0.5f * M_PI / (double)(nAngles - 1);
    for (size_t jj = 1; jj <= nAngles; jj++) {
      theta = (double)(jj - 1) * dang;
      amu[jj - 1] = cosf(theta);
    }

    /* Logarithmic derivative D(J) calculated by downward recurrence
       beginning with initial value (0.,0.) at J=NMX -1 */

    cxd[nmx - 1] = std::complex<double>(0., 0.);
    nn = nmx - 1;

    for (n = 1; n <= nn; n++) {
      rn = nmx - n + 1;
      cxd[nmx - n - 1] = (rn / cxy) - (1. / (cxd[nmx - n] + rn / cxy));
    }

    for (size_t jj = 1; jj <= nAngles; jj++) {
      pi0[jj - 1] = 0.;
      pi1[jj - 1] = 1.;
    }

    nn = 2 * nAngles - 1;
    for (size_t jj = 1; jj <= nn; jj++) {
      cxs1[jj - 1] = std::complex<double>(0., 0.);
      cxs2[jj - 1] = std::complex<double>(0., 0.);
    }

    /* Riccati-Bessel functions with real argument X
       calculated by upward recurrence */

    psi0 = cos(dx);
    psi1 = sin(dx);
    chi0 = -sin(x);
    chi1 = cos(x);
    apsi0 = psi0;
    apsi1 = psi1;
    cxxi0 = std::complex<double>(apsi0, -chi0);
    cxxi1 = std::complex<double>(apsi1, -chi1);
    qScat = 0.;
    gScat = 0.;

    n = 1;
    size_t jj;
    while (n <= nstop) {
      dn = n;
      rn = n;
      fn = (2. * rn + 1.) / (rn * (rn + 1.));
      psi = (2. * dn - 1.) * psi1 / dx - psi0;
      apsi = psi;
      chi = (2. * rn - 1.) * chi1 / x - chi0;
      cxxi = std::complex<double>(apsi, -chi);

      /* Store previous values of AN and BN for use
          in computation of g=<cos(theta)> */

      if (n > 1) {
        cxan1 = cxan;
        cxbn1 = cxbn;
      }

      /* Compute AN and BN:*/
      cxan = (cxd[n - 1] / iorRelative + rn / x) * apsi - apsi1;
      cxan /= ((cxd[n - 1] / iorRelative + rn / x) * cxxi - cxxi1);
      cxbn = (iorRelative * cxd[n - 1] + rn / x) * apsi - apsi1;
      cxbn /= ((iorRelative * cxd[n - 1] + rn / x) * cxxi - cxxi1);

      /* Augment sums for *qsca and g=<cos(theta)> */
      qScat += (2. * rn + 1.) * (std::norm(cxan) + std::norm(cxbn));
      gScat += ((2. * rn + 1.) / (rn * (rn + 1.))) * (std::norm(cxan));

      if (n > 1) {
        gScat +=
            ((rn - 1.) * (rn + 1.) / rn) * (std::norm(cxan1) + std::norm(cxbn));
      }

      for (size_t j = 1; j <= nAngles; j++) {
        jj = 2 * nAngles - j;
        pi[j - 1] = pi1[j - 1];
        tau[j - 1] = rn * amu[j - 1] * pi[j - 1] - (rn + 1.) * pi0[j - 1];
        p = std::pow(-1.0, n - 1);
        cxs1[j - 1] += fn * (cxan * pi[j - 1] + cxbn * tau[j - 1]);

        t = pow(-1.0, n);
        cxs2[j - 1] += fn * (cxan * tau[j - 1] + cxbn * pi[j - 1]);

        if (j != jj) {
          cxs1[jj - 1] += fn * (cxan * pi[j - 1] * p + cxbn * tau[j - 1] * t);
          cxs2[jj - 1] += fn * (cxan * tau[j - 1] * t + cxbn * pi[j - 1] * p);
        }
      }

      psi0 = psi1;
      psi1 = psi;
      apsi1 = psi1;
      chi0 = chi1;
      chi1 = chi;
      cxxi1 = std::complex<double>(apsi1, -chi1);

      n = n + 1;
      rn = n;

      /*  For each angle J, compute pi_n+1
          from PI = pi_n , PI0 = pi_n-1 */

      for (size_t j = 1; j <= nAngles; j++) {
        pi1[j - 1] = ((2. * rn - 1.) / (rn - 1.)) * amu[j - 1] * pi[j - 1];
        pi1[j - 1] -= rn * pi0[j - 1] / (rn - 1.);
        pi0[j - 1] = pi[j - 1];
      }
    }

    gScat = 2. * gScat / qScat;
    qScat *= (2. / (x * x));

    //      std::cout << "qExt: " << (4./(x*x))*cxs1[0].real() << std::endl;
    //      std::cout << "qBack: " << (4./(x*x))*std::abs(cxs1[2*nAngles-2])*
    //            std::abs(cxs1[2*nAngles-2]) << std::endl;

    for (size_t ii = 0; ii < cxs1.size(); ii++) {
      spf[ii] = 0.5 * (std::norm(cxs1[ii]) + std::norm(cxs2[ii]));
    }

    // integrate by interpolation and normalize
    double total = 0.0;
    double dth = 0.001;
    for (double th = 0.5 * dth; th < M_PI; th += dth) {
      size_t thi0 = (th / M_PI) * spf.size();
      size_t thi1 = thi0 + 1;
      double th0 = thi0 * (M_PI / spf.size());
      double th1 = thi1 * (M_PI / spf.size());
      double f = (th - th0) / (th1 - th0);
      double val = (1. - f) * spf[thi0] + f * spf[thi1];
      total += 2. * M_PI * val * sin(th) * dth;
    }

    std::for_each(spf.begin(), spf.end(), [&](double &a) { a /= total; });
  }

  /**
   * Calling function for the Mie code with the addition of a coating layer.
   * Inputs match bhmie with additional inputs about the coating. Code
   * is based off of the code accompanying "Improved algorithm of light
   * scattering by a coated sphere" (2006).
   */
  static void
  miecoat(double iorMedium, /* real refractive index of the medium */
          const std::complex<double> &iorSphere,  /* complex ior of sphere */
          const std::complex<double> &iorCoating, /* complex ior of coating */
          double radiusSphere,      /* radius of the core sphere [um] */
          double radiusCoating,     /* radius including the coating [um] */
          double wavelength,        /* [um] */
          size_t nAngles,           /* number of output angles in [0,90] */
          std::vector<double> &spf, /* scattering efficiency [0:PI] */
          double &qScat,            /* efficiency factor for scattering */
          double &gScat) {          /* average cosine */

    // relative refractive indexes
    std::complex<double> RFREL1 = iorSphere / iorMedium;
    std::complex<double> RFREL2 = iorCoating / iorMedium;

    spf.clear();
    qScat = 0.;
    gScat = 0.;

    // size parameters
    double alpha1 = TwoPI * radiusSphere * iorMedium / wavelength;
    double alpha2 = TwoPI * radiusCoating * iorMedium / wavelength;

    // real and imaginary refractive indexes
    double rm1 = RFREL1.real();
    double im1 = RFREL1.imag();
    double rm2 = RFREL2.real();
    double im2 = RFREL2.imag();

    // arrays that are maintained throughout
    const int j = 100000;
    std::vector<double> r_L11(j), i_L11(j), r_L12(j), i_L12(j), r_L22(j),
        i_L22(j), La1n(j), r_a(j), i_a(j), r_b(j), i_b(j);

    // scattering intensity
    auto intensity = [&](int n_max, double ta,
                         std::array<double, 2> &intensity) {
      int n = 1;
      double pi0 = 0;
      double dpi0 = 0;
      double pi1 = 1;
      double dpi1 = 0;
      double ti1 = cos(ta);
      double nn = double(n);
      double p_nn = (2. * nn + 1.) / nn / (nn + 1.);
      double rs1 = p_nn * (r_a[n] * pi1 + r_b[n] * ti1);
      double rs2 = p_nn * (r_a[n] * ti1 + r_b[n] * pi1);
      double is1 = p_nn * (i_a[n] * pi1 + i_b[n] * ti1);
      double is2 = p_nn * (i_a[n] * ti1 + i_b[n] * pi1);
      for (n = 2; n <= n_max; n++) {
        nn = double(n);
        double pi2 = ((2 * nn - 1) * cos(ta) * pi1 - nn * pi0) / (nn - 1.);
        double dpi2 = (2 * nn - 1) * pi1 + dpi0;
        double ti2 = cos(ta) * pi2 - pow(sin(ta), 2) * dpi2;
        pi0 = pi1;
        pi1 = pi2;
        dpi0 = dpi1;
        dpi1 = dpi2;
        p_nn = (2. * nn + 1.) / nn / (nn + 1.);
        rs1 += p_nn * (r_a[n] * pi2 + r_b[n] * ti2);
        rs2 += p_nn * (r_a[n] * ti2 + r_b[n] * pi2);
        is1 += p_nn * (i_a[n] * pi2 + i_b[n] * ti2);
        is2 += p_nn * (i_a[n] * ti2 + i_b[n] * pi2);
      }
      intensity[0] = rs1 * rs1 + is1 * is1;
      intensity[1] = rs2 * rs2 + is2 * is2;
    };

    // complex calculation
    auto CX = [](double p, double q, double s, double t,
                 std::array<double, 4> &x) {
      double modl = s * s + t * t;
      x[0] = (p * s + q * t) / modl;
      x[1] = (q * s - p * t) / modl;
      x[2] = p * s - q * t;
      x[3] = p * t + q * s;
    };

    // recursion NFunc
    auto recursNFunc = [&](int n, double rma, double ima, double r_Func,
                           double i_Func, std::array<double, 2> &y) {
      double rtmp, itmp, k1 = 1.0, k0 = 0.0;
      std::array<double, 4> com_X;
      CX(double(n), k0, rma, ima, com_X);
      rtmp = com_X[0];
      itmp = com_X[1];
      CX(k1, k0, rtmp - r_Func, itmp - i_Func, com_X);
      y[0] = com_X[0] - rtmp;
      y[1] = com_X[1] - itmp;
    };

    // recursion Xnz
    auto recursXnz = [&](int n, double rma, double ima, double r_Func,
                         double i_Func, std::array<double, 2> &y) {
      std::array<double, 4> com_X;
      CX(n, 0.0, rma, ima, com_X);
      y[0] = com_X[0] + r_Func;
      y[1] = com_X[1] + i_Func;
    };

    // FuncU
    auto FuncU = [&](double r1, double i1, double rF1, double iF1, double r2,
                     double i2, double rF2, double iF2,
                     std::array<double, 2> &y) {
      std::array<double, 4> com_X;
      CX(r1, i1, rF1, iF1, com_X);
      y[0] = com_X[2];
      y[1] = com_X[3];
      CX(r2, i2, rF2, iF2, com_X);
      y[0] -= com_X[2];
      y[1] -= com_X[3];
    };

    // Lentz
    auto FGL = [&](double rma, double ima, int L, double ParaFGL[]) {
      double rfl3, ifl3;
      std::array<double, 4> com_X;
      double k0 = 0.0;
      double k1 = 2.0 * L + 1.0;
      double k2 = -2.0 * L - 3.0;
      CX(k1, k0, rma, ima, com_X);
      double rwl1 = com_X[0];
      double iwl1 = com_X[1];
      CX(k2, k0, rma, ima, com_X);
      double rwl2 = com_X[0];
      double iwl2 = com_X[1];
      double rfl2 = rwl1 + rma / k2;
      double ifl2 = iwl1 + ima / k2;
      double ro2 = rwl2 + rma / k1;
      double io2 = iwl2 + ima / k1;
      double rp2 = rwl2;
      double ip2 = iwl2;
      double rwlk = rwl2;
      double iwlk = iwl2;

      for (int k = 3;; k++) {
        double parak = -(2.0 * (L + k) - 1) / (2.0 * (L + k) - 3);
        rwlk = parak * rwlk;
        iwlk = parak * iwlk;
        k1 = 1.0;
        CX(k1, k0, ro2, io2, com_X);
        double ro3 = rwlk + com_X[0];
        double io3 = iwlk + com_X[1];
        CX(k1, k0, rp2, ip2, com_X);
        double rp3 = rwlk + com_X[0];
        double ip3 = iwlk + com_X[1];
        CX(ro3, io3, rp3, ip3, com_X);
        double rp4 = com_X[0];
        double ip4 = com_X[1];
        CX(rfl2, ifl2, rp4, ip4, com_X);
        rfl3 = com_X[2];
        ifl3 = com_X[3];
        if (fabs(1 - sqrt(rp4 * rp4 + ip4 * ip4)) < 1e-20)
          break;
        else {
          ro2 = ro3;
          io2 = io3;
          rp2 = rp3;
          ip2 = ip3;
          rfl2 = rfl3;
          ifl2 = ifl3;
        }
      }
      k1 = -L;
      k0 = 0.0;
      CX(k1, k0, rma, ima, com_X);
      ParaFGL[1] = rfl3 + com_X[0];
      ParaFGL[2] = ifl3 + com_X[1];
    };

    // Main equations (numbered)
    if (alpha1 < alpha2) {
      double tmp;
      tmp = alpha1;
      alpha1 = alpha2;
      alpha2 = tmp;
      tmp = rm1;
      rm1 = rm2;
      rm2 = tmp;
      tmp = im1;
      im1 = im2;
      im2 = tmp;
    }

    int kim;
    if (im1 < 0 || im2 < 0) {
      im1 = -fabs(im1);
      im2 = -fabs(im2);
      kim = -1;
    } else {
      im1 = fabs(im1);
      im2 = fabs(im2);
      kim = 1;
    }

    int n = 0;
    int n_max = int(alpha1 + 7.5 * pow(alpha1, 0.34) + 2.5);
    assert(n_max < j);

    // double kext = 0.0;
    // double ksca = 0.0;
    std::array<double, 4> com_x;
    std::array<double, 2> com_y;
    double ParaFGL[3];
    double alpha1s = pow(alpha1, 2);
    double rm2a2 = rm2 * alpha2;
    double rm1a2 = rm1 * alpha2;
    double rm1a1 = rm1 * alpha1;
    double im2a2 = im2 * alpha2;
    double im1a2 = im1 * alpha2;
    double im1a1 = im1 * alpha1;
    double k0 = 0.0;
    double k1 = 1.0;
    double r1_tmp, i1_tmp, r2_tmp, i2_tmp, r3_tmp, i3_tmp;

    //	[1]
    FGL(rm1a1, im1a1, n_max, ParaFGL);
    r_L11[n_max] = r2_tmp = ParaFGL[1];
    i_L11[n_max] = i2_tmp = ParaFGL[2];
    for (n = n_max; n >= 1; n--) {
      CX(double(n), k0, rm1a1, im1a1, com_x);
      r1_tmp = com_x[0], i1_tmp = com_x[1];
      CX(k1, k0, r1_tmp + r2_tmp, i1_tmp + i2_tmp, com_x);
      r_L11[n - 1] = r2_tmp = r1_tmp - com_x[0],
                i_L11[n - 1] = i2_tmp = i1_tmp - com_x[1];
    }

    //	[2]
    FGL(rm1a2, im1a2, n_max, ParaFGL);
    r_L12[n_max] = r2_tmp = ParaFGL[1];
    i_L12[n_max] = i2_tmp = ParaFGL[2];
    for (n = n_max; n >= 1; n--) {
      CX(double(n), k0, rm1a2, im1a2, com_x);
      r1_tmp = com_x[0];
      i1_tmp = com_x[1];
      CX(k1, k0, r1_tmp + r2_tmp, i1_tmp + i2_tmp, com_x);
      r_L12[n - 1] = r2_tmp = r1_tmp - com_x[0];
      i_L12[n - 1] = i2_tmp = i1_tmp - com_x[1];
    }

    //	[3]
    FGL(rm2a2, im2a2, n_max, ParaFGL);
    r_L22[n_max] = r2_tmp = ParaFGL[1];
    i_L22[n_max] = i2_tmp = ParaFGL[2];
    for (n = n_max; n >= 1; n--) {
      CX(double(n), k0, rm2a2, im2a2, com_x);
      r1_tmp = com_x[0];
      i1_tmp = com_x[1];
      CX(k1, k0, r1_tmp + r2_tmp, i1_tmp + i2_tmp, com_x);
      r_L22[n - 1] = r2_tmp = r1_tmp - com_x[0];
      i_L22[n - 1] = i2_tmp = i1_tmp - com_x[1];
    }

    //	[4]
    FGL(alpha1, k0, n_max, ParaFGL);
    La1n[n_max] = ParaFGL[1];
    for (n = n_max; n >= 1; n--) {
      La1n[n - 1] = double(n) / alpha1 - 1.0 / (double(n) / alpha1 + La1n[n]);
    }

    double r_Ma1n, r_N12n, r_N11n, r_Na1n, r_Qn, i_Ma1n, i_N12n, i_N11n, i_Na1n,
        i_Qn;

    //	M1(a1):
    r_Ma1n = pow(sin(alpha1) - alpha1 * cos(alpha1), 2) / (alpha1s + 1.0);
    if (kim < 0) {
      i_Ma1n = ((alpha1s - 1.0) * sin(2.0 * alpha1) * 0.5 +
                alpha1 * cos(2.0 * alpha1)) /
               (alpha1s + 1.0);
    } else {
      i_Ma1n = -((alpha1s - 1.0) * sin(2.0 * alpha1) * 0.5 +
                 alpha1 * cos(2.0 * alpha1)) /
               (alpha1s + 1.0);
    }

    //	N0(a1),N0(m1a1),N0(m1a2):
    r_N12n = r_N11n = r_Na1n = 0.0;
    if (kim < 0) {
      i_N12n = i_N11n = i_Na1n = -1.0;
    } else {
      i_N12n = i_N11n = i_Na1n = 1.0;
    }

    //	Q0(m1a2m1a1):
    if (kim <= 0) {
      CX(cos(2.0 * rm1a2) - exp(2.0 * im1a2), sin(2.0 * rm1a2),
         cos(2.0 * rm1a1) - exp(2.0 * im1a1), sin(2.0 * rm1a1), com_x);
      r_Qn = com_x[0] * exp(2.0 * (im1a1 - im1a2));
      i_Qn = com_x[1] * exp(2.0 * (im1a1 - im1a2));
    } else {
      CX(exp(-2.0 * im1a2) - cos(2.0 * rm1a2), sin(2.0 * rm1a2),
         exp(-2.0 * im1a1) - cos(2.0 * rm1a1), sin(2.0 * rm1a1), com_x);
      r_Qn = com_x[0] * exp(2.0 * (im1a2 - im1a1));
      i_Qn = com_x[1] * exp(2.0 * (im1a2 - im1a1));
    }

    for (n = 1; n <= n_max; n++) {
      double nalpha = double(n) / alpha1;

      //	[5]
      if (n > 1) {
        CX(r_Ma1n, i_Ma1n, r_Na1n - nalpha, i_Na1n, com_x);
        r_Ma1n = com_x[0] * (La1n[n - 1] - nalpha),
        i_Ma1n = com_x[1] * (La1n[n - 1] - nalpha);
      }
      //	[6]
      CX(k1, k0, nalpha - r_Na1n, -i_Na1n, com_x);
      r_Na1n = com_x[0] - nalpha, i_Na1n = com_x[1];

      //	[7]
      recursNFunc(n, rm1a1, im1a1, r_N11n, i_N11n, com_y);
      r_N11n = com_y[0], i_N11n = com_y[1];

      //	[8]
      recursNFunc(n, rm1a2, im1a2, r_N12n, i_N12n, com_y);
      r_N12n = com_y[0], i_N12n = com_y[1];

      //	[9]
      recursXnz(n, rm1a1, im1a1, r_L11[n], i_L11[n], com_y);
      r1_tmp = com_y[0], i1_tmp = com_y[1];
      recursXnz(n, rm1a2, im1a2, r_N12n, i_N12n, com_y);
      r2_tmp = com_y[0], i2_tmp = com_y[1];
      CX(r1_tmp, i1_tmp, r2_tmp, i2_tmp, com_x);
      r3_tmp = com_x[2], i3_tmp = com_x[3];

      recursXnz(n, rm1a2, im1a2, r_L12[n], i_L12[n], com_y);
      r1_tmp = com_y[0], i1_tmp = com_y[1];
      recursXnz(n, rm1a1, im1a1, r_N11n, i_N11n, com_y);
      r2_tmp = com_y[0], i2_tmp = com_y[1];
      CX(r1_tmp, i1_tmp, r2_tmp, i2_tmp, com_x);
      r2_tmp = com_x[2], i2_tmp = com_x[3];

      CX(r3_tmp, i3_tmp, r2_tmp, i2_tmp, com_x);
      r1_tmp = com_x[0], i1_tmp = com_x[1];
      CX(r_Qn, i_Qn, r1_tmp, i1_tmp, com_x);
      r_Qn = com_x[2], i_Qn = com_x[3];

      //	[10]
      double ru1a, ru2a, ru3a, ru4a, ru5a, ru6a, iu1a, iu2a, iu3a, iu4a, iu5a,
          iu6a;

      FuncU(rm2, im2, r_N12n, i_N12n, rm1, im1, r_L22[n], i_L22[n], com_y);
      ru1a = com_y[0], iu1a = com_y[1];

      FuncU(1.0, 0.0, r_L11[n], i_L11[n], rm1, im1, La1n[n], 0.0, com_y);
      ru2a = com_y[0], iu2a = com_y[1];

      FuncU(rm2, im2, r_L12[n], i_L12[n], rm1, im1, r_L22[n], i_L22[n], com_y);
      ru3a = com_y[0], iu3a = com_y[1];

      FuncU(1.0, 0.0, r_N11n, i_N11n, rm1, im1, La1n[n], 0.0, com_y);
      ru4a = com_y[0], iu4a = com_y[1];

      FuncU(1.0, 0.0, r_L11[n], i_L11[n], rm1, im1, r_Na1n, i_Na1n, com_y);
      ru5a = com_y[0], iu5a = com_y[1];

      FuncU(1.0, 0.0, r_N11n, i_N11n, rm1, im1, r_Na1n, i_Na1n, com_y);
      ru6a = com_y[0], iu6a = com_y[1];

      //	[11]
      double ru1b, ru2b, ru3b, ru4b, ru5b, ru6b, iu1b, iu2b, iu3b, iu4b, iu5b,
          iu6b;
      FuncU(rm1, im1, r_N12n, i_N12n, rm2, im2, r_L22[n], i_L22[n], com_y);
      ru1b = com_y[0], iu1b = com_y[1];

      FuncU(rm1, im1, r_L11[n], i_L11[n], 1.0, 0.0, La1n[n], 0.0, com_y);
      ru2b = com_y[0], iu2b = com_y[1];

      FuncU(rm1, im1, r_L12[n], i_L12[n], rm2, im2, r_L22[n], i_L22[n], com_y);
      ru3b = com_y[0], iu3b = com_y[1];

      FuncU(rm1, im1, r_N11n, i_N11n, 1.0, 0.0, La1n[n], 0.0, com_y);
      ru4b = com_y[0], iu4b = com_y[1];

      FuncU(rm1, im1, r_L11[n], i_L11[n], 1.0, 0.0, r_Na1n, i_Na1n, com_y);
      ru5b = com_y[0], iu5b = com_y[1];

      FuncU(rm1, im1, r_N11n, i_N11n, 1.0, 0.0, r_Na1n, i_Na1n, com_y);
      ru6b = com_y[0], iu6b = com_y[1];

      //	[12]
      CX(r_Qn, i_Qn, ru3a, iu3a, com_x);
      r1_tmp = com_x[2], i1_tmp = com_x[3];
      FuncU(ru1a, iu1a, ru2a, iu2a, r1_tmp, i1_tmp, ru4a, iu4a, com_y);
      r2_tmp = com_y[0], i2_tmp = com_y[1];
      FuncU(ru1a, iu1a, ru5a, iu5a, r1_tmp, i1_tmp, ru6a, iu6a, com_y);
      r3_tmp = com_y[0], i3_tmp = com_y[1];
      CX(r2_tmp, i2_tmp, r3_tmp, i3_tmp, com_x);
      r1_tmp = com_x[0], i1_tmp = com_x[1];
      CX(r_Ma1n, i_Ma1n, r1_tmp, i1_tmp, com_x);
      r_a[n] = com_x[2], i_a[n] = com_x[3];

      //	[13]
      CX(r_Qn, i_Qn, ru3b, iu3b, com_x);
      r1_tmp = com_x[2], i1_tmp = com_x[3];
      FuncU(ru1b, iu1b, ru2b, iu2b, r1_tmp, i1_tmp, ru4b, iu4b, com_y);
      r2_tmp = com_y[0], i2_tmp = com_y[1];
      FuncU(ru1b, iu1b, ru5b, iu5b, r1_tmp, i1_tmp, ru6b, iu6b, com_y);
      r3_tmp = com_y[0], i3_tmp = com_y[1];
      CX(r2_tmp, i2_tmp, r3_tmp, i3_tmp, com_x);
      r1_tmp = com_x[0], i1_tmp = com_x[1];
      CX(r_Ma1n, i_Ma1n, r1_tmp, i1_tmp, com_x);
      r_b[n] = com_x[2], i_b[n] = com_x[3];

      //	[14]
      // double para=2.0*double(n)+1.0;
      // kext+=para*(r_a[n]+r_b[n]);
      // ksca+=para*(pow(r_a[n],2)+pow(i_a[n],2)+pow(r_b[n],2)+pow(i_b[n],2));
    }

    // finalize

    size_t fullAngles = 2 * nAngles - 1;
    double dang = 0.5 * M_PI / (double)(nAngles - 1);
    spf.resize(fullAngles);
    std::array<double, 2> intens;
    for (size_t ii = 0; ii < fullAngles; ii++) {
      double th = ii * dang;
      intensity(n_max, th, intens);
      spf[ii] = intens[0] + intens[1];
    }

    // integrate by interpolation and normalize
    double total = 0.0;
    double dth = 0.001;
    for (double th = 0.5 * dth; th < M_PI; th += dth) {
      size_t thi0 = (th / M_PI) * spf.size();
      size_t thi1 = thi0 + 1;
      double th0 = thi0 * (M_PI / spf.size());
      double th1 = thi1 * (M_PI / spf.size());
      double f = (th - th0) / (th1 - th0);
      double val = (1. - f) * spf[thi0] + f * spf[thi1];
      total += 2. * M_PI * val * sin(th) * dth;
    }

    std::for_each(spf.begin(), spf.end(), [&](double &a) { a /= total; });
  }
};

/**
 * This class represents an arbitrary size distribution of spheres of
 * a given range of radii (in microns). A depth parameter (which may
 * be ignored) allows the distribution to vary in one dimension. The
 * min/max radius is over the entire definition and has no dependencies.
 *
 * The units of the size distribution are [m^-3 um^-1] (i.e.
 * per cubic meter volumetricaly, per micron radially)
 */
class SphereSizeDistribution {
public:
  SphereSizeDistribution() {}

  virtual ~SphereSizeDistribution() {}

  virtual double minRadius() = 0;

  virtual double maxRadius() = 0;

  virtual double n(double radius, double depth) = 0;

  /**
   * Evaluate the number density over the range radii between
   * Rmin and Rmax (in units of [m^-3]
   */
  double N(double depth) {
    static constexpr double xs[15] = {
        5.1471842555317695833025213166722573749141453666569564e-02,
        1.5386991360858354696379467274325592041855197124433846e-01,
        2.5463692616788984643980512981780510788278930330251842e-01,
        3.5270472553087811347103720708937386065363100802142562e-01,
        4.4703376953808917678060990032285400016240759386142440e-01,
        5.3662414814201989926416979331107279416417800693029710e-01,
        6.2052618298924286114047755643118929920736469282952813e-01,
        6.9785049479331579693229238802664006838235380065395465e-01,
        7.6777743210482619491797734097450313169488361723290845e-01,
        8.2956576238276839744289811973250191643906869617034167e-01,
        8.8256053579205268154311646253022559005668914714648423e-01,
        9.2620004742927432587932427708047400408647453682532906e-01,
        9.6002186496830751221687102558179766293035921740392339e-01,
        9.8366812327974720997003258160566280194031785470971136e-01,
        9.9689348407464954027163005091869528334088203811775079e-01};
    static constexpr double ws[15] = {
        1.0285265289355884034128563670541504386837555706492822e-01,
        1.0176238974840550459642895216855404463270628948712684e-01,
        9.9593420586795267062780282103569476529869263666704277e-02,
        9.6368737174644259639468626351809865096406461430160245e-02,
        9.2122522237786128717632707087618767196913234418234107e-02,
        8.6899787201082979802387530715125702576753328743545344e-02,
        8.0755895229420215354694938460529730875892803708439299e-02,
        7.3755974737705206268243850022190734153770526037049438e-02,
        6.5974229882180495128128515115962361237442953656660378e-02,
        5.7493156217619066481721689402056128797120670721763134e-02,
        4.8402672830594052902938140422807517815271809197372736e-02,
        3.8799192569627049596801936446347692033200976766395352e-02,
        2.8784707883323369349719179611292043639588894546287496e-02,
        1.8466468311090959142302131912047269096206533968181403e-02,
        7.9681924961666056154658834746736224504806965871517212e-03};
    double result = 0;
    double minR = minRadius();
    double maxR = maxRadius();
    for (int k = 0; k < 15; k++) {
      double n0 = n((1 - xs[k]) * minR + xs[k] * maxR, depth);
      double n1 = n((1 - xs[k]) * maxR + xs[k] * minR, depth);
      result += ws[k] * (n0 + n1) * 0.5;
    }
    result *= maxR - minR;
    return result;
  }
};

/**
 * This is the bubble size distribution model of Hall and Novarini[1995]
 * as given in Ma[2015]. It is intended for the range beteen 10um and 150um
 * and varies with depth. The model is for entrained bubbles under
 * a given windspeed measured at 10m above sea level.
 *
 * Depth is measured in positive meters down from mean sea level.
 */
class HallNovariniBubbleModel final : public SphereSizeDistribution {
public:
  HallNovariniBubbleModel(double U10 /* wind speed @10m */) {
    this->constPartA = 1.6e4 * pow(U10 / 13., 3.);
    if (U10 <= 7.5)
      this->constPartB = -(1. / 0.4);
    else
      this->constPartB = -(1. / (0.4 + 0.115 * (U10 - 7.5)));
  }

  ~HallNovariniBubbleModel() override {}

  virtual double minRadius() override { return 10.; }

  virtual double maxRadius() override { return 150.; }

  virtual double n(double radius, double depth) override {
    if (radius < this->minRadius())
      return 0.;
    if (radius > this->maxRadius())
      return 0.;

    double rref = 54.4 + 1.984e-6 * depth;
    double Gexp =
        (radius <= rref) ? 4. : 4.37 + (depth / 2.55) * (depth / 2.55);
    double G = pow(rref / radius, Gexp);

    return this->constPartA * G * exp(depth * this->constPartB);
  }

private:
  double constPartA, constPartB;
};

/**
 * This is the bubble size distribution model Leifer et al [2007] for
 * large, shallow diffuse plumes.
 *
 * Depth is measured in positive meters down from mean sea level.
 */
class LeiferNSDiBubbleModel final : public SphereSizeDistribution {
public:
  LeiferNSDiBubbleModel() {}

  ~LeiferNSDiBubbleModel() override {}

  virtual double minRadius() override { return 200.; }

  virtual double maxRadius() override { return 3000.; }

  virtual double n(double radius, double depth) override {
    if (radius < this->minRadius())
      return 0.;
    if (radius > this->maxRadius())
      return 0.;

    if (radius < 800) {
      // const double S1 = -1.;
      return 1. / (radius / 50.);
    } else {
      // note that isn't quite right and under-represents larger radii
      const double S2 = -3.5;
      return pow(radius / 362.289, S2);
    }
  }
};

/**
 * This is a simple bubble distribution for a given range of radii (um)
 * and a fixed density.
 */
class BubbleRangeModel final : public SphereSizeDistribution {
public:
  BubbleRangeModel(double _minRad, double _maxRad, double _density)
      : minRad(_minRad), maxRad(_maxRad), density(_density) {}

  ~BubbleRangeModel() override {}

  virtual double minRadius() override { return minRad; }

  virtual double maxRadius() override { return maxRad; }

  virtual double n(double radius, double depth) override {
    if (radius < this->minRadius())
      return 0.;
    if (radius > this->maxRadius())
      return 0.;

    return density;
  }

private:
  double minRad, maxRad, density;
};

/**
 * This is a rain drop size distribution model based on the work of
 * Marhsall and Palmer [1948] assuming stratiform rain.
 */
class MarshallPalmerModel final : public SphereSizeDistribution {
public:
  MarshallPalmerModel(double _rainRate /* rain rate mm/hr */) {
    this->constPart = 4.1 * pow(_rainRate, -0.21); // [mm^-1]
  }

  ~MarshallPalmerModel() override {}

  virtual double minRadius() override { return 600.; }

  virtual double maxRadius() override { return 1400; }

  virtual double n(double radius, double depth) override {
    if (radius < this->minRadius())
      return 0.;
    if (radius > this->maxRadius())
      return 0.;
    double dMM = 2 * radius * 1e-3;
    const double N0 = 8000.; // [m^-3 mm^-1]
    double nMM = N0 * exp(-this->constPart * dMM);
    return nMM * 1e3;
  }

private:
  double constPart;
};

/**
 * This is a water droplet size distribution from MODTRAN driven by
 * the equation n(r) = Kr^a * exp(-b*r), where b and a are fit
 * parameters, r is the droplet radius and K is a normalization constant.
 */
class ModifiedGammaModel final : public SphereSizeDistribution {
public:
  ModifiedGammaModel(double _a = 6.0, double _b = 1.5) : a(_a), b(_b) {
    double tot = 0;
    double step = 0.01;
    this->minRad = 0.;
    this->maxRad = 20;
    this->K = 1.0;
    bool tail = false;
    for (double r = minRad + 0.5 * step; r < maxRad; r += step) {
      double nr = this->n(r, 0.);
      if (nr < 1e-3) {
        if (tail) {
          maxRad = r;
          break;
        } else {
          minRad = r;
        }
      } else
        tail = true;
      tot += step * nr;
    }
    this->K = 1. / tot;
  }

  ~ModifiedGammaModel() override {}

  virtual double minRadius() override { return this->minRad; }

  virtual double maxRadius() override { return this->maxRad; }

  virtual double n(double radius, double depth) override {
    if (radius < this->minRadius())
      return 0.;
    if (radius > this->maxRadius())
      return 0.;
    return K * pow(radius, a) * exp(-b * radius);
  }

private:
  double minRad = 0.5;
  double maxRad = 12.0;
  double a, b, K;
};

/**
 * Single generator function that produces full inherent optical properties
 * (IOPs) for a given distribution of spheres. For now we assume there is
 * no absorption.
 *
 * The scattering phase function is given at 0.1 [deg] samples from the
 * forward direction [0:180].
 */
static void calculateMieIOPs(
    double wl,        /* wavelength in [microns] */
    double iorMedium, /* real refractive index of the medium */
    const std::complex<double> &iorSphere,      /* complex ior of sphere */
    SphereSizeDistribution *sphereDistribution, /* distribution model */
    double depth, /* sphere distribution parameter */
    std::vector<double> &spf, double &scatCoeff,
    double coatThickness = 0.0, /* optional coating thickness [microns] */
    const std::complex<double> &iorCoat = {0, 0}, /* coating ior (optional) */
    double nSteps = 100) {
  if (coatThickness > 0)
    assert(abs(iorCoat) > 0);
  size_t nAngles = 901;

  // adjust the delta based on the actual range
  double delta =
      (sphereDistribution->maxRadius() - sphereDistribution->minRadius()) /
      nSteps;
  double qS, gS;
  std::vector<double> tmp;

  double totB = 0.;
  for (double r = sphereDistribution->minRadius() + 0.5 * delta;
       r < sphereDistribution->maxRadius(); r += delta) {
    double nrd = sphereDistribution->n(r, depth);
    scatCoeff = 1e-12 * M_PI * r * r * nrd;
    totB += scatCoeff * delta;
  }

  size_t index = 0;

  // do an approximate discrete integral
  for (double r = sphereDistribution->minRadius() + 0.5 * delta;
       r < sphereDistribution->maxRadius(); r += delta) {
    double nrd = sphereDistribution->n(r, depth);
    scatCoeff = 1e-12 * M_PI * r * r * nrd;
    if (index == 0) {
      if (coatThickness <= 0) {
        Mie::bhmie(iorMedium, iorSphere, r, wl, nAngles, spf, qS, gS);
      } else {
        Mie::miecoat(iorMedium, iorSphere, iorCoat, r, r + coatThickness, wl,
                     nAngles, spf, qS, gS);
      }
      std::transform(spf.begin(), spf.end(), spf.begin(),
                     [&](double &a) { return a * (scatCoeff / totB) * delta; });
    } else {
      if (coatThickness <= 0) {
        Mie::bhmie(iorMedium, iorSphere, r, wl, nAngles, tmp, qS, gS);
      } else {
        Mie::miecoat(iorMedium, iorSphere, iorCoat, r, r + coatThickness, wl,
                     nAngles, tmp, qS, gS);
      }
      std::transform(spf.begin(), spf.end(), tmp.begin(), spf.begin(),
                     [&](double &a, double &b) {
                       return a + b * (scatCoeff / totB) * delta;
                     });
    }

    index++;
  }

  scatCoeff = totB;
}

/// \}

} // namespace d5
