/** Interface file for dealing with mixin architectures
 *
 *  This file is based off of libciabatta, but customized a bit for our use
 *  case. https://github.com/atomgalaxy/libciabatta
 *
 *  \file Mixin.h
 */

#pragma once

#include <utility>

#include "dirsig5/base/TypeTraits.h"

namespace d5 {

/** \defgroup common_mixin Mixin Architecture
 *  \ingroup common
 *  Holds template framework for the mixin architecture pattern.
 *
 *  Mixin architecture is the idea of building a complex object from a number
 *  of smaller building blocks. Thi architecture is useful when there are a
 *  number of types that do not fit into an inheritance tree with each other,
 *  but share common functionality that the developer does not wish to repeat
 *  in each type. The core mechanism is implemented by defining a number of
 *  mixin types, which are all CRTP classes and then having the combined object
 *  (generally referred to as \c Concrete), inherit from mixin, where each
 *  desired building block being listed in the template parameters.
 *
 *  The general pattern for making a mixin is:
 *  \code
 *  template <typename Base>
 *  struct MyMixin : Base {
 *    // interface to add...
 *  };
 *  \endcode
 *  and then making a mixin class is as simple as:
 *  \code
 *  struct MyClass : mixin<Mixin1, Mixin2, ...> {
 *    // interfaces from all mixins are inherited
 *  };
 *  \endcode
 */

/// \addtogroup common_mixin
/// \{

/** The base class for mixins to allow communications between mixins
 *
 *  \tparam Mixin      The mixin for which this class is serving as the base
 *  \tparam MixinType  mixin<Mixins...>, the mixin type with all used mixins
 */
template <template <class...> class Mixin, class MixinType> struct MixinBase {
  /// Type alias for the derived class
  using self_t = MixinType;

  /** Get a reference to the concrete object, can be used from a mixin to
   *  communicate with other mixins
   */
  /// \{
  decltype(auto) self() & { return static_cast<self_t &>(*this); }
  decltype(auto) self() && { return static_cast<self_t &&>(*this); }
  decltype(auto) self() const & { return static_cast<const self_t &>(*this); }
  decltype(auto) self() const && { return static_cast<const self_t &&>(*this); }
  /// \}
};

/** Main mixin template, mixin classes should inherit from this template.
 *
 *  When inheriting from this template, the class will inherit from each mixin.
 *  For example:
 *
 *  \code
 *  template <class Base>
 *  struct A : Base {
 *    int a;
 *  };
 *
 *  template <class Base>
 *  struct B : Base {
 *    int b;
 *  };
 *
 *  class C : public mixin<A, B> {};
 *  \endcode
 *
 *  In this case, the inheritance tree would look like:
 *  \dot
 *  digraph inheritance_tree {
 *    node [shape=oval, fontname=Courier, fontsize=10];
 *    C
 *    m [label="mixin<A, B>"]
 *    A [label="A<MixinBase<A, mixin<A, B>>>"]
 *    B [label="B<MixinBase<B, mixin<A, B>>>"]
 *    mbA [label="MixinBase<A, mixin<A, B>>"]
 *    mbB [label="MixinBase<B, mixin<A, B>>"]
 *    C -> m
 *    m -> A
 *    m -> B
 *    A -> mbA
 *    B -> mbB
 *  }
 *  \enddot
 *  And \c C will wind up with the \c a and \c b public data members.
 */
template <template <class...> class... Mixins>
struct mixin : Mixins<MixinBase<Mixins, mixin<Mixins...>>>... {};

/** Metafunction to get the mixin type from a concrete class
 *
 *  For example:
 *  \code
 *  struct C : mixin<A, B> {};
 *  static_assert(std::is_same_v<mixin_type_t<C>, mixin<A, B>>);
 *  \endcode
 */
template <class T> struct mixin_type {
private:
  template <template <class...> class... Mixins>
  static decltype(static_cast<const mixin<Mixins...> &>(std::declval<T>()),
                  mixin<Mixins...>{})
  test(const mixin<Mixins...> &);

  static void test(...);

public:
  using type = decltype(test(std::declval<T>()));
};

/// Concenience alias for mixin_type
template <class T> using mixin_type_t = typename mixin_type<T>::type;

/// Metafunction to test whether \c T is a mixin or inherits from a mixin
template <class T>
using is_mixin = std::negation<std::is_same<mixin_type_t<T>, void>>;

/// Convenience alias for is_mixin
template <class T> inline constexpr bool is_mixin_v = is_mixin<T>::value;

/** Dispatch metafunction for mixin classes
 *
 *  For a metafunction of the form
 *  \code
 *  template <template <class...> class... Mixins>
 *  struct metafunc {
 *    static constexpr bool value = ...; // or type
 *  };
 *  \endcode
 *  this dispatcher can allow it to be called from just the concrete type. For
 *  example:
 *  \code
 *  template <class T>
 *  struct nicer_metafunc {
 *    static constexpr bool value =
 *        mixin_metafunction_t<T, metafunc>::value; // or type
 *  };
 *  \endcode
 *
 *  For an example, see has_mixin.
 */
template <class T, template <template <class...> class...> class Metafunction,
          typename... Args>
struct mixin_metafunction {
private:
  /** Expression SFINAE statement, if \c Concrete can be upcast to a
   *  \c mixin<Concrete, Mixins...> (this is how mixin_type worked), then the
   *  return type is \c Metafunction<Concrete, Mixins...> (decltype of an object
   *  of that type gives the type back).
   */
  template <template <class...> class... Mixins>
  static decltype(static_cast<const mixin<Mixins...> &>(std::declval<T>()),
                  Metafunction<Mixins...>{})
  test(const mixin<Mixins...> &);

  /** Base case for if \c Concrete cannot be upcast (static_assert prevents this
   *  from ever happening, but here for completeness)
   */
  static void test(...);

public:
  static_assert(is_mixin_v<T>, "T must be a mixin class");

  /** This will be the type for the metafunction with the mixins as template
   *  arguments
   */
  using type = decltype(test(std::declval<T>()));
};

/// Convenience alias mixin_metafunction::type
template <typename T,
          template <template <class...> class...> class Metafunction>
using mixin_metafunction_t = typename mixin_metafunction<T, Metafunction>::type;

/** Metafunction for testing whether the given concrete mixin class has the
 *  given mixin.
 *
 *  \tparam Concrete  The concrete mixin class
 *  \tparam Mixin     The mixin to test for
 */
template <typename T, template <class...> class Mixin> struct has_mixin {
  /** Implementation metafunction
   *
   *  The first (unnamed) parameter is \c Concrete, which we already have from
   *  the parent struct.
   *
   *  \tparam Mixins  The mixins in \c Concrete (came from mixin_metafunction)
   */
  template <template <class...> class... Mixins> struct impl {
    /// Simply test whether Mixin is found in Mixins
    static constexpr bool value = is_any_tmpl_v<Mixin, Mixins...>;
  };

  /// Index of the mixin in the type
  static constexpr bool value = mixin_metafunction_t<T, impl>::value;
};

/// Convenience alias for has_mixin::value
template <typename T, template <class...> class Mixin>
inline constexpr bool has_mixin_v = has_mixin<T, Mixin>::value;

template <class T, template <class...> class Mixin> struct mixin_base {
private:
  static_assert(has_mixin_v<T, Mixin>, "Derived does not have Mixin");

public:
  using type = Mixin<MixinBase<Mixin, mixin_type_t<T>>>;
};

/// Convenience alias for mixin_base
template <typename T, template <class...> class Mixin>
using mixin_base_t = typename mixin_base<T, Mixin>::type;

/** Casts a concrete mixin class to a specific mixin base class.
 *
 *  The specifc return type can be found via the mixin_base metafunction
 *
 *  \tparam T      The concrete mixin class
 *  \tparam Mixin  The mixin to which to cast
 *
 *  \param m  The mixin object to cast
 *
 *  \return The casted version of \c m
 */
template <template <class...> class Mixin, typename T>
constexpr decltype(auto) mixin_cast(const T &m) {
  return static_cast<const mixin_base_t<T, Mixin> &>(m);
}

/// \overload
template <template <class...> class Mixin, typename T>
constexpr decltype(auto) mixin_cast(T &m) {
  return static_cast<mixin_base_t<T, Mixin> &>(m);
}

/** Allows mixin classes with multiple template parameters.
 *
 *  Basic usage:
 *  ```
 *  template <typename Arg1, typename Arg2, typename Base>
 *  struct my_mixin : Base {
 *    Arg1 a1;
 *    Arg2 a2;
 *  };
 *
 *  struct myclass : mixin<myclass,
 *                        template_mixin<my_mixin, int, float>::mixin> {};
 *  ```
 */
template <template <class...> class Mixin, typename... Args>
struct template_mixin {
  template <typename Base> using mixin = Mixin<Args..., Base>;
};

/// Trivial example mixin that adds nothing to Base's interface
template <typename Base> struct NoMixin : Base {};

/// \}
} // namespace d5
