/** Interface file for a named parameter store
 *
 *  \file NamedParameterStore.h
 */

#pragma once

#include <any>
#include <map>
#include <stdexcept>

#include "TypeTraits.h"

namespace d5 {

struct DefaultNamedParameterStoreTag {};

template <typename NameT, NameT Name,
          typename Tag = DefaultNamedParameterStoreTag>
struct NamedParameterStoreTraits {
  static_assert(
      always_false<Tag>,
      "No NamedParameterStoreTraits for this name type/value combination");
};

template <typename NameT, typename Tag = DefaultNamedParameterStoreTag,
          typename Compare = std::less<NameT>>
class NamedParameterStore {
private:
  template <NameT Name>
  using ExpectedType =
      typename NamedParameterStoreTraits<NameT, Name, Tag>::type;

  template <NameT Name, typename T> constexpr void CheckExpectedType() {
    static_assert(std::is_convertible_v<std::decay_t<T>, ExpectedType<Name>>,
                  "Incorrect type passed for NamedParameterStore::set()");
  }

public:
  using enum_t = NameT;
  using map_t = std::map<NameT, std::any, Compare>;
  using iterator = typename map_t::iterator;

  template <NameT Name> using type_for = ExpectedType<Name>;

  NamedParameterStore() = default;
  NamedParameterStore(const NamedParameterStore &) = default;
  NamedParameterStore(NamedParameterStore &&) = default;
  NamedParameterStore &operator=(const NamedParameterStore &) = default;
  NamedParameterStore &operator=(NamedParameterStore &&) = default;

  auto begin() { return map_.begin(); }
  auto begin() const { return map_.begin(); }
  auto cbegin() const { return map_.cbegin(); }
  auto end() { return map_.end(); }
  auto end() const { return map_.end(); }
  auto cend() const { return map_.cent(); }

  [[nodiscard]] bool empty() const { return map_.empty(); }
  [[nodiscard]] auto size() const { return map_.size(); }
  void clear() { map_.clear(); }

  // equivalent of operator[] = val
  template <NameT Name> void set(ExpectedType<Name> val) {
    map_[Name] = std::move(val);
  }

  // in-place construction, will not overwrite
  template <NameT Name, typename... Args>
  std::pair<bool, iterator> emplace(Args &&...args) {
    return map_.emplace(Name, std::forward<Args>(args)...);
  }

  // Const access
  template <NameT Name> const auto &at() const {
    return AnyCast<Name>(any_at<Name>());
  }

  // Mutable access
  template <NameT Name> auto &at() { return AnyCast<Name>(any_at<Name>()); }

  template <NameT Name> auto erase() { return map_.erase(Name); }

  template <NameT Name> [[nodiscard]] auto count() const {
    return map_.count(Name);
  }

  template <NameT Name> [[nodiscard]] bool contains() const {
    return count<Name>() > 0u;
  }

  // Const downcasting
  template <NameT Name> const auto &AnyCast(const std::any &anyVal) const try {
    return std::any_cast<const ExpectedType<Name> &>(anyVal);
  } catch (const std::exception &e) {
    throw std::runtime_error(
        "the incorrect type was for this parameter was stored in the map");
  }

  // Non-const downcasting
  template <NameT Name> auto &AnyCast(std::any &anyVal) const try {
    return std::any_cast<ExpectedType<Name> &>(anyVal);
  } catch (const std::exception &e) {
    throw std::runtime_error(
        "the incorrect type was for this parameter was stored in the map");
  }

private:
  template <NameT Name> [[nodiscard]] const std::any &any_at() const try {
    return map_.at(Name);
  } catch (const std::exception &e) {
    throw std::out_of_range("NamedParameterStore::at(): no parameter with that "
                            "name was in the map");
  }

  template <NameT Name> std::any &any_at() try {
    return map_.at(Name);
  } catch (const std::exception &e) {
    throw std::out_of_range("NamedParameterStore::at(): no parameter with that "
                            "name was in the map");
  }

private:
  map_t map_;
};

} // namespace d5
