/** Interface file for the PCG, which is our random number facility
 *
 *  \file Pcg.h
 */

#pragma once

#include <cassert>
#include <cstdint>
#include <limits>
#include <random>

namespace d5 {

/// \addtogroup common
/// \{

/// A 32-bit PCG XSH-RR random number generator.
class Pcg32 {
public:
  using State = std::uint64_t;

  static constexpr State multiplier = 6364136223846793005ULL;

  static constexpr State default_increment = 1442695040888963407ULL;

  using Result = std::uint32_t;

  using result_type = std::uint32_t;

  static constexpr Result min() noexcept { return 0; }

  static constexpr Result max() noexcept {
    return std::numeric_limits<Result>::max();
  }

public:
  Pcg32() noexcept { state = state * multiplier + increment; }

  Pcg32(State seed) noexcept : state(seed) {
    state = state + increment;
    state = state * multiplier + increment;
  }

  Pcg32(State seed, State seq) noexcept : state(seed) {
    stream(seq);
    state = state + increment;
    state = state * multiplier + increment;
  }

public:
  void seed(State s) noexcept { *this = Pcg32(s); }

  [[nodiscard]] State stream() const noexcept { return increment >> 1; }

  State stream(State seq) noexcept {
    State res = stream();
    increment = (seq << 1) | 1;
    return res;
  }

  void discard(State n) noexcept {
    state = lcg_seek(state, multiplier, increment, n);
  }

  Result operator()() noexcept {
    State prev_state = state;
    state = state * multiplier + increment;
    return output(prev_state);
  }

  Result operator()(Result b) noexcept {
    Result rmin = (-b) % b;
    while (true) {
      Result r = operator()();
      if (r >= rmin)
        return r % b;
    }
  }

  bool operator==(const Pcg32 &other) const noexcept {
    return state == other.state && increment == other.increment;
  }

public:
  State state = 0;

  State increment = default_increment;

  static Result output(State state) noexcept {
    auto rotr = [](Result v, unsigned r) {
      return (v >> r) | (v << ((-r) & (sizeof(v) * 8 - 1)));
    };
    return rotr(Result(((state >> 18) ^ state) >> 27), (state >> 59) & 31);
  }

  static State lcg_seek(State x, State a, State b, State n) noexcept {
    // Fast power by squaring.
    State an = 1;
    State bn = 0;
    while (n > 0) {
      if (n & 1) {
        an *= a;
        bn *= a;
        bn += b;
      }
      b *= a + 1;
      a *= a;
      n >>= 1;
    }
    x *= an;
    x += bn;
    return x;
  }

  static State lcg_distance(State x, State a, State b, State xn) noexcept {
    assert(((a & 3) == 1) || !((b & 1) == 1));
    State p = 1;
    State r = 0;
    while (x != xn) {
      if ((x & p) != (xn & p)) {
        x *= a;
        x += b;
        r |= p;
      }
      p <<= 1;
      b *= a + 1;
      a *= a;
    }
    return r;
  }
};

inline double canonical(Pcg32 &random) noexcept {
  return static_cast<double>(random()) * 2.3283064365386963e-10;
}

inline double sampleUniform(Pcg32 &random, double u0 = 0,
                            double u1 = 1) noexcept {
  double u = canonical(random);
  return (1 - u) * u0 + u * u1;
}

inline double sampleNormalDistribution(Pcg32 &random, double mu = 0,
                                       double sigma = 1) noexcept {
  auto erfInverse = [](double y) {
    double w = -std::log((1 - y) * (1 + y));
    double x = 0;
    if (w < 5) {
      w = w - 2.5f;
      x = std::fma(w, +2.81022636e-08, +3.43273939e-7);
      x = std::fma(w, x, -3.52338770e-6);
      x = std::fma(w, x, -4.39150654e-6);
      x = std::fma(w, x, +2.18580870e-4);
      x = std::fma(w, x, -1.25372503e-3);
      x = std::fma(w, x, -4.17768164e-3);
      x = std::fma(w, x, +2.46640727e-1);
      x = std::fma(w, x, +1.50140941);
    } else {
      w = std::sqrt(w) - 3;
      x = std::fma(x, -2.00214257e-4, +1.00950558e-4);
      x = std::fma(w, x, +1.34934322e-3);
      x = std::fma(w, x, -3.67342844e-3);
      x = std::fma(w, x, +5.73950773e-3);
      x = std::fma(w, x, -7.62246130e-3);
      x = std::fma(w, x, +9.43887047e-3);
      x = std::fma(w, x, +1.00167406);
      x = std::fma(w, x, +2.83297682);
    }
    x *= y;
    return x;
  };
  return mu +
         1.4142135623730951 * sigma * erfInverse(2 * canonical(random) - 1);
}

using RandEngine = Pcg32;

/// \}

} // namespace d5
