/** Interface file for a pixel array.
 *
 *  \file PixelArray.h
 */

#pragma once

#include <iostream>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Pcg.h"

namespace d5 {

/**
 * Defines a 2D array of pixels and a mechanism to generate a random sample
 * within the area of a specified pixel.  The size defines the area of the
 * pixel and the stride defines the distance from center to center of
 * adjacent pixels. Hence, size equals stride means perfectly packed pixels
 * (no gaps).  The units for the sizes, strides and offsets are note defined
 * here, but are assumed to be the same across these parameters.
 */
struct PixelArray {
  /** The size of a pixel in the X dimension */
  float xPixelSize;

  /** The size of a pixel in the X dimension */
  float yPixelSize;

  /** The stride of the pixel pixel centers the X dimension */
  float xPixelStride;

  /** The stride of the pixel pixel centers the Y dimension */
  float yPixelStride;

  /** The number of pixels in the X dimension */
  size_t xPixelCount = 0;

  /** The number of pixels in the Y dimension */
  size_t yPixelCount = 0;

  /** The offset of the center of the array in the X dimension */
  float xArrayOffset = 0.f;

  /** The offset of the center of the array in the Y dimension */
  float yArrayOffset = 0.f;

  /**
   * Generate a random sample within the area of the specified pixel.
   */
  Point operator()(d5::RandEngine &randEngine, size_t xIndex, size_t yIndex) {
    // create a random number generator for spatial sampling
    static thread_local std::uniform_real_distribution<float> randGen;

    DIRSIG_THROW_IF(xPixelCount == 0,
                    "PixelArray::operator(): invalid X array dimensions!");
    DIRSIG_THROW_IF(yPixelCount == 0,
                    "PixelArray::operator(): invalid Y array dimensions!");
    DIRSIG_THROW_IF(xIndex >= xPixelCount,
                    "PixelArray::operator(): invalid X pixel index!");
    DIRSIG_THROW_IF(yIndex >= yPixelCount,
                    "PixelArray::operator(): invalid Y pixel index!");

    auto xPos = xIndex - (xPixelCount - 1.0f) / 2.0f;
    xPos *= xPixelStride;
    xPos += xPixelSize * (randGen(randEngine) - 0.5f);
    xPos += xArrayOffset;

    auto yPos = yIndex - (yPixelCount - 1.0f) / 2.0f;
    yPos *= yPixelStride;
    yPos += yPixelSize * (randGen(randEngine) - 0.5f);
    yPos += yArrayOffset;

    return Point(xPos, yPos, 0);
  }

  /** A helper function for the X length of the array */
  double getXLength() const {
    return (this->xPixelCount * this->xPixelStride) + this->xPixelSize;
  }

  /** A helper function for the Y length of the array */
  double getYLength() const {
    return (this->yPixelCount * this->yPixelStride) + this->yPixelSize;
  }
};

} // namespace d5
