#pragma once

#include "dirsig5/base/Export.h"
#include "dirsig5/base/LibraryManager.h"
#include "dirsig5/base/StandardPaths.h"

namespace d5 {

template <typename PluginType>
class PluginLibraryManager : public LibraryManager {
public:
  using EntryPointFunc = PluginType *(*)();

  PluginLibraryManager() : LibraryManager("entrypoint") {}

  PluginLibraryManager(std::string funcName)
      : LibraryManager(std::move(funcName)) {}

  void loadBuiltin(const std::string &name, EntryPointFunc func) {
    builtins[name] = func;
  }

  void loadDefaultLocations(const std::string &subdirName) {
    loadLibraryDirectory(StandardPaths::UserPlugins(subdirName));
    loadLibraryDirectory(StandardPaths::DirsigPlugins(subdirName));
  }

  [[nodiscard]] bool hasPlugin(const std::string &name) const {
    return builtins.find(name) != builtins.end() or hasLibrary(name);
  }

  [[nodiscard]] std::unique_ptr<PluginType>
  getInstance(const std::string &name) const {
    EntryPointFunc func = nullptr;
    if (auto itr = builtins.find(name); itr != builtins.end())
      func = itr->second;
    else
      func = reinterpret_cast<EntryPointFunc>(getLibraryFunc(name));
    if (!func) {
      using namespace std::string_literals;
      throw std::runtime_error("Can't get instance of '"s + name +
                               "': plugin not found"s);
    }
    std::unique_ptr<PluginType> plugin;
    try {
      plugin.reset(func());
    } catch (const std::exception &error) {
      using namespace std::string_literals;
      throw std::runtime_error(
          "Can't get instance of '"s + name +
          "': plugin entry point function threw an exception: "s +
          error.what());
    }
    return plugin;
  }

public:
  std::map<std::string, EntryPointFunc> builtins;
};

} // namespace d5
