/** Interface file for a set of spherical harmonics
 *
 *  \file SHQueryObjects.h
 */

#pragma once

#include <fstream>
#include <stdexcept>
#include <vector>

#include "dirsig5/base/Geometry.h"
#include "sh/spherical_harmonics.h"

namespace d5 {

/** A struct that wraps a set of spherical harmonics coefficients for a
 *  a particular order. It provides a manual sum that can be much faster
 *  than the third party spherical-harmonics library since it eliminates
 *  coefficients that are lower than a particular threshold.
 *
 *  The (optional) threshold argument sets a minimum coefficient magnitude
 *  required to be part of the sum.
 */
struct SHQueryObject {
  SHQueryObject() = default;

  ~SHQueryObject() = default;

  void updateState(const State *st) {
    if (this->order != 0) {
      size_t coeffCount = sh::GetCoefficientCount(order);
      this->stateCoeffs.resize(coeffCount);
      for (size_t ii = 0; ii < coeffCount; ii++) {
        this->stateCoeffs[ii].resize(
            st->spectralStates->at(st->spectralStateIndex)
                .wavelengthSamples.size());
        interpolate(
            this->origWavelengths,
            &this->origCoeffs[this->origWavelengths.size() * ii],
            st->spectralStates->at(st->spectralStateIndex).wavelengthSamples,
            this->stateCoeffs[ii]);
      }
    }
#if 0    
    this->dumpCoeffs();
#endif
  }

  double getDepth(void) const { return this->depth; }

  // Set coeffs with a binary file stream (double depth,
  // int n_wavelengths, double wavelengths, int order, double coefficients,
  // with all spectral points given for each coefficient before the next)
  SHQueryObject(std::ifstream &stream, double coeffThreshold = 1e-4) {
    stream.read((char *)&this->depth, sizeof(double));
    stream.read((char *)&this->nWavelengths, sizeof(int));
    this->origWavelengths.resize(this->nWavelengths);
    stream.read((char *)this->origWavelengths.data(),
                sizeof(double) * this->origWavelengths.size());
    stream.read((char *)&this->order, sizeof(int));
    this->origCoeffs.resize(nWavelengths *
                            sh::GetCoefficientCount(this->order));
    stream.read((char *)this->origCoeffs.data(),
                sizeof(double) * this->origCoeffs.size());
    DIRSIG_THROW_IF(not stream,
                    "Unable to read spherical harmonic coefficients from "
                    "input stream!");
  }

  // Query the spherical harmonics for a given vector
  SpectralVector query(const d5::Vectorf dir) {
    DIRSIG_THROW_IF(
        order == 0 || stateCoeffs.size() == 0,
        "Attempting to use spherical harmonics at depth without values!");

    static thread_local SpectralVector vals;
    vals.resize(stateCoeffs[0].size());
    vals.setData(0.0);

    // it is (a lot) faster to work with angles for non-trivial orders
    double ph, th, coeff, harmonic;
    sh::ToSphericalCoords(dir.cast<double>().normalized(), &ph, &th);
    int l, m;
    size_t jj = 0;
    for (l = 0; l <= order; l++) {
      for (m = -l; m <= l; m++) {
        harmonic = sh::EvalSH(l, m, ph, th);
        for (size_t iw = 0; iw < stateCoeffs[0].size(); iw++) {
          coeff = this->stateCoeffs[jj][iw];
          if (fabs(coeff) >= this->coeffThreshold) {
            vals[iw] += coeff * harmonic;
          }
        }
        jj++;
      }
    }
    // don't allow negative values
    vals.clampNegative();

    return vals;
  }

  void dumpCoeffs(void) {
    for (size_t iw = 0; iw < stateCoeffs[0].size(); iw++) {
      std::ofstream out("coeffs" + std::to_string(iw) + ".dat");
      for (int t = 0; t < 90; t++) {
        double theta = t * M_PI / 180;
        for (int p = -180; p < 180; p++) {
          double phi = p * M_PI / 180.0;
          d5::Vectorf dir(cos(phi) * sin(theta), sin(phi) * sin(theta),
                          cos(theta));
          double val = this->query(dir)[iw];
          out << val << ' ';
        }
        out << std::endl;
      }
      out.close();
    }
  }

private:
  double depth;
  int nWavelengths;
  std::vector<double> origWavelengths;
  int order = 0;
  std::vector<double> origCoeffs;
  double coeffThreshold = 1e-4;

  std::vector<SpectralVector> stateCoeffs;
};

} // namespace d5
