/** Interface file for scope guards, a generic RAII facility
 *
 *  \file ScopeGuard.h
 */

#pragma once

#include <exception>
#include <iostream>
#include <type_traits>
#include <utility>

namespace d5 {

/// \addtogroup common
/// \{

namespace detail {

/// Helper non-templated base class for ScopeGuard
struct ScopeGuardBase {
  [[nodiscard]] bool dismissed() const { return dismissed_; }

  void dismiss() { dismissed_ = true; }

  template <typename Callable> static void RunFunction(Callable &fn) noexcept {
    try {
      fn();
    } catch (const std::exception &e) {
      std::cerr << "Caught an exception in a scope guard destructor. Calling "
                   "std::terminate().\n";
      std::cerr << "what(): " << e.what() << "\n";
      std::terminate();
    } catch (...) {
      std::cerr << "Caught an exception in a scope guard destructor. Calling "
                   "std::terminate().\n";
      std::terminate();
    }
  }

private:
  bool dismissed_ = false;
};

} // namespace detail

/// A scope guard to call a function upon destruction, unless it is dismissed.
template <typename Callable> struct ScopeGuard : detail::ScopeGuardBase {
  explicit ScopeGuard(const Callable &fn) : fn_(fn) {}

  explicit ScopeGuard(Callable &&fn) : fn_(std::move(fn)) {}

  ~ScopeGuard() {
    if (!dismissed())
      RunFunction(fn_);
  }

private:
  Callable fn_;
};

/** Make a scope guard.
 *
 *  \param fn  The function to run upon scope exit
 */
template <class Callable> inline auto make_guard(Callable &&fn) {
  return ScopeGuard<std::decay_t<Callable>>(std::forward<Callable>(fn));
}

/// \}

} // namespace d5
