#pragma once

#include <algorithm>
#include <cmath>
#include <iostream>
#include <numeric>
#include <utility>
#include <vector>

#include <Eigen/Dense>

#include "Error.h"

#define MAX_WAVELENGTHS 500000

namespace d5 {

/// \addtogroup common
/// \{

class SpectralVector {
public:
  SpectralVector() = default;

  SpectralVector(const SpectralVector &) = default;

  SpectralVector(SpectralVector &&) = default;

  SpectralVector(Eigen::ArrayXd points) : eigenData(std::move(points)) {}

  SpectralVector(std::size_t pointCount, double value = 0) {
    DIRSIG_THROW_IF(pointCount > MAX_WAVELENGTHS, "Point count too large!");
    eigenData.resize(static_cast<int>(pointCount));
    if (pointCount > 0)
      setData(value);
  }

  /// Set the content of the vector.
  void setData(double value) {
    DIRSIG_THROW_IF(std::isnan(value), "Attempting to fill with NaNs!");
    eigenData.fill(value);
  }

  /// Clear the contents of the vector (set to zero)
  void clear() { eigenData.fill(0); }

  /// Set spectral data at specific index
  void setData(std::size_t spectralIndex, double value) {
    DIRSIG_THROW_IF(spectralIndex >= size(), "Out of bounds index!");
    DIRSIG_THROW_IF(std::isnan(value), "Attempting to set NaN by index!");
    eigenData(static_cast<int>(spectralIndex)) = value;
  }

  /// Copy data from another vector
  void copyData(const SpectralVector &copy) { eigenData = copy.eigenData; }

  /// Get spectral data at specific index
  [[nodiscard]] double getData(size_t spectralIndex) const {
    DIRSIG_THROW_IF(spectralIndex >= size(), "Out of bounds index!");
    return eigenData(static_cast<int>(spectralIndex));
  }

  /// Get spectral data at specific index
  double &getData(std::size_t spectralIndex) {
    DIRSIG_THROW_IF(spectralIndex >= size(), "Out of bounds index!");
    return eigenData(static_cast<int>(spectralIndex));
  }

  /// Get the total number of spectral points in the vector
  [[nodiscard]] std::size_t getPointCount() const noexcept {
    return eigenData.rows();
  }

  void setPointCount(size_t pointCount) {
    DIRSIG_THROW_IF(pointCount > MAX_WAVELENGTHS, "Point count too large!");
    if (pointCount != size())
      eigenData.resize(static_cast<int>(pointCount));
  }

  /// Alternate form of getPointCount()
  [[nodiscard]] std::size_t size() const noexcept { return eigenData.rows(); }

  /// Alternate form of setPointCount()
  void resize(std::size_t pointCount) { setPointCount(pointCount); }

  double *begin() noexcept { return data(); }

  double *end() noexcept { return data() + size(); }

  [[nodiscard]] const double *begin() const noexcept { return data(); }

  [[nodiscard]] const double *end() const noexcept { return data() + size(); }

  /// Get raw data.
  double *data() noexcept { return eigenData.data(); }

  /// Get raw data.
  [[nodiscard]] const double *data() const noexcept { return eigenData.data(); }

  [[nodiscard]] Eigen::ArrayXd &array() noexcept { return eigenData; }

  [[nodiscard]] const Eigen::ArrayXd &array() const noexcept {
    return eigenData;
  }

  /// Gets data at an absolute index
  double operator[](size_t spectralIndex) const {
    return getData(spectralIndex);
  }

  /// Gets data at an absolute index
  double &operator[](size_t spectralIndex) { return getData(spectralIndex); }

  SpectralVector &operator=(const SpectralVector &) = default;

  SpectralVector &operator=(SpectralVector &&) = default;

  SpectralVector &operator+=(const SpectralVector &rhs) {
    DIRSIG_THROW_IF(size() != rhs.size(), "Mismatched vector sizes!");
    eigenData += rhs.eigenData;
    return *this;
  }

  SpectralVector &operator-=(const SpectralVector &rhs) {
    DIRSIG_THROW_IF(size() != rhs.size(), "Mismatched vector sizes!");
    eigenData -= rhs.eigenData;
    return *this;
  }

  SpectralVector &operator*=(const SpectralVector &rhs) {
    DIRSIG_THROW_IF(size() != rhs.size(), "Mismatched vector sizes!");
    eigenData *= rhs.eigenData;
    return *this;
  }

  SpectralVector &operator/=(const SpectralVector &rhs) {
    DIRSIG_THROW_IF(size() != rhs.size(), "Mismatched vector sizes!");
    eigenData /= rhs.eigenData;
    return *this;
  }

  SpectralVector &operator+=(double rhs) {
    DIRSIG_THROW_IF(std::isnan(rhs), "Attempting to add NaN!");
    eigenData += rhs;
    return *this;
  }

  SpectralVector &operator-=(double rhs) {
    DIRSIG_THROW_IF(std::isnan(rhs), "Attempting to subtract NaN!");
    eigenData -= rhs;
    return *this;
  }

  SpectralVector &operator*=(double rhs) {
    DIRSIG_THROW_IF(std::isnan(rhs), "Attempting to multiply NaN!");
    eigenData *= rhs;
    return *this;
  }

  SpectralVector &operator/=(double rhs) {
    DIRSIG_THROW_IF(std::isnan(rhs), "Attempting to divide NaN!");
    eigenData /= rhs;
    return *this;
  }

  template <typename Vec> SpectralVector &operator+=(const Vec &rhs) {
    DIRSIG_THROW_IF(size() != size_t(rhs.rows()), "Mismatched vector sizes!");
    eigenData += rhs;
    return *this;
  }

  template <typename Vec> SpectralVector &operator-=(const Vec &rhs) {
    DIRSIG_THROW_IF(size() != size_t(rhs.rows()), "Mismatched vector sizes!");
    eigenData -= rhs;
    return *this;
  }

  template <typename Vec> SpectralVector &operator*=(const Vec &rhs) {
    DIRSIG_THROW_IF(size() != size_t(rhs.rows()), "Mismatched vector sizes!");
    eigenData *= rhs;
    return *this;
  }

  template <typename Vec> SpectralVector &operator/=(const Vec &rhs) {
    DIRSIG_THROW_IF(size() != size_t(rhs.rows()), "Mismatched vector sizes!");
    eigenData /= rhs;
    return *this;
  }

  friend auto operator+(const SpectralVector &lhs, const SpectralVector &rhs) {
    DIRSIG_THROW_IF(lhs.size() != rhs.size(), "Mismatched vector sizes!");
    return lhs.eigenData + rhs.eigenData;
  }

  friend auto operator-(const SpectralVector &lhs, const SpectralVector &rhs) {
    DIRSIG_THROW_IF(lhs.size() != rhs.size(), "Mismatched vector sizes!");
    return lhs.eigenData - rhs.eigenData;
  }

  friend auto operator*(const SpectralVector &lhs, const SpectralVector &rhs) {
    DIRSIG_THROW_IF(lhs.size() != rhs.size(), "Mismatched vector sizes!");
    return lhs.eigenData * rhs.eigenData;
  }

  friend auto operator/(const SpectralVector &lhs, const SpectralVector &rhs) {
    DIRSIG_THROW_IF(lhs.size() != rhs.size(), "Mismatched vector sizes!");
    return lhs.eigenData / rhs.eigenData;
  }

  friend auto operator*(double lhs, const SpectralVector &rhs) {
    return lhs * rhs.eigenData;
  }

  friend auto operator*(const SpectralVector &lhs, double rhs) {
    return lhs.eigenData * rhs;
  }

  friend SpectralVector exp(const SpectralVector &sv) {
    return Eigen::ArrayXd(sv.eigenData.exp());
  }

  [[nodiscard]] SpectralVector pow(double value) const {
    return Eigen::ArrayXd(eigenData.pow(value));
  }

  /// Get the minimum value in the spectral vector
  [[nodiscard]] double min() const noexcept { return eigenData.minCoeff(); }

  /// Gets the maximum value in the spectral vector
  [[nodiscard]] double max() const noexcept { return eigenData.maxCoeff(); }

  /// Gets the average value of the entire spectral vector
  [[nodiscard]] double average() const noexcept { return eigenData.mean(); }

  /// Check if any of the spectral data exceeds the supplied value.
  [[nodiscard]] bool maxExceeds(double maxVal) const noexcept {
    return eigenData.maxCoeff() > maxVal;
  }

  /// Clamp negative values (to zero)
  void clampNegative() {
    eigenData = eigenData.max(Eigen::ArrayXd::Zero(eigenData.rows()));
  }

  /// Print the vector
  friend std::ostream &operator<<(std::ostream &output,
                                  const SpectralVector &sv) {
    if (sv.getPointCount() > 0) {
      output << '\n';
      for (size_t point = 0; point < sv.getPointCount(); point++)
        output << sv.getData(point) << '\n';
    }
    return output;
  }

  // the vector is represented by a column array (Eigen)
  Eigen::ArrayXd eigenData;
};

namespace detail {

/// Helper struct to determine whether a type has a `void resize(size_t)` method
template <typename T> struct HasResizeMethod {
  template <typename U, void (U:: *)(size_t)> struct SFINAE {};

  template <typename U> static char Test(SFINAE<U, &U::resize> *);

  template <typename U> static int Test(...);

  static const bool Has = sizeof(Test<T>(0)) == sizeof(char);
};

/// Constexpr test for whether a type has a `void resize(size_t)` method
template <typename T>
inline constexpr bool has_resize_v =
    std::integral_constant<bool, HasResizeMethod<std::decay_t<T>>::Has>();

} // namespace detail

/** This is a simple helper interpolator function to go from
 *  one spectral vector with given wavelengths to another
 *  spectral vector via linear interpolation. Extrapolation
 *  is done by simply repeating the edge value;
 *
 *  The size of the original data is set by the input wavelengths
 *  (which must implement the size() accessor). The corresponding
 *  input data is not checked and can be a pointer directly into
 *  memory. Similarly, the output wavelengths sets the output data
 *  size. The must implement size() and resize(..), respectively.
 */
template <typename WaveA, typename DataA, typename WaveB, typename DataB>
inline void interpolate(const WaveA &oldWave, const DataA &oldData,
                        const WaveB &newWave, DataB &&newData) {
  // Check if the output argument has a resize method. If so, resize to the
  // correct size, otherwise throw if it is the wrong size
  if constexpr (detail::has_resize_v<DataB>)
    newData.resize(newWave.size());
  else
    DIRSIG_THROW_IF(
        newData.size() != newWave.size(),
        "Output spectral vector has no resize() and is a different size "
        "than output wavelengths");
  DIRSIG_THROW_IF(oldWave.size() == 0,
                  "Attempting to interpolate a null vector!");
  DIRSIG_THROW_IF(newWave.size() == 0,
                  "Attempting to interpolate a null vector!");
  if (oldWave.size() == 1) {
    for (size_t i = 0; i < newData.size(); i++)
      newData[i] = oldData[0];
    return;
  }
  std::vector<size_t> ind(oldWave.size());
  std::iota(ind.begin(), ind.end(), 0);
  std::sort(ind.begin(), ind.end(),
            [&](size_t i1, size_t i2) { return oldWave[i1] < oldWave[i2]; });
  ind.erase(std::unique(ind.begin(), ind.end(),
                        [&](size_t i1, size_t i2) {
                          return oldWave[i1] == oldWave[i2];
                        }),
            ind.end());

  for (size_t i = 0; i < newWave.size(); i++) {
    double wave = newWave[i];
    double waveA = oldWave[ind.front()];
    double waveB = oldWave[ind.back()];
    DIRSIG_THROW_IF(!std::isfinite(wave),
                    "Attempting to interpolate to a non-finite wavelength!");
    if (wave <= waveA) { // Lower edge?
      newData[i] = oldData[ind.front()];
      DIRSIG_THROW_IF(!std::isfinite(newData[i]),
                      "Attempting to interpolate to a non-finite value!");
    } else if (wave >= waveB) {
      newData[i] = oldData[ind.back()];
      DIRSIG_THROW_IF(!std::isfinite(newData[i]),
                      "Attempting to interpolate to a non-finite value!");
    } else {
      // take a reasonable guess at where were are in the sorted vector
      auto j = static_cast<std::size_t>((wave - waveA) / (waveB - waveA) *
                                        static_cast<double>(ind.size()));
      // back up if we have to find the actual lower bounds
      while (wave < oldWave[ind[j]])
        j--;
      // now find the upper bounds and linearly interpolate
      for (; j < ind.size(); j++) {
        double upper = oldWave[ind[j]];
        DIRSIG_THROW_IF(
            !std::isfinite(upper),
            "Attempting to interpolate to a non-finite wavelength!");
        if (wave < upper) {
          double lower = oldWave[ind[j - 1]];
          double fract = (wave - lower) / (upper - lower);
          newData[i] =
              (1 - fract) * oldData[ind[j - 1]] + fract * oldData[ind[j]];
          DIRSIG_THROW_IF(!std::isfinite(lower),
                          "Attempting to interpolate to a "
                          "non-finite wavelength!");
          DIRSIG_THROW_IF(!std::isfinite(newData[i]),
                          "Attempting to interpolate to a non-finite value!");
          break;
        }
      }
    }
  }
}

/// \}

} // namespace d5
