/** Interface file for states
 *
 *  \file State.h
 */

#pragma once

#include <algorithm>
#include <cmath>
#include <numeric>
#include <type_traits>
#include <vector>

#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Json.h"

namespace d5 {

/// \addtogroup common
/// \{

/// An enumerated type to track spectral units
enum SpectralUnits {
  UnknownUnits = 0x00,    ///< unknown
  MicronUnits = 0x01,     ///< microns
  NanometerUnits = 0x02,  ///< nanometers
  WavenumberUnits = 0x04, ///< wavenumbers [cm^-1]
  GHzUnits = 0x08,        ///< gigahertz
  AngstromUnits = 0x10    ///< angstrom
};

/// JSON serialization for SpectralUnits
D5_JSON_SERIALIZE_ENUM(SpectralUnits, {{UnknownUnits, "unknown"},
                                       {MicronUnits, "microns"},
                                       {NanometerUnits, "nanometers"},
                                       {WavenumberUnits, "wavenumbers"},
                                       {GHzUnits, "GHz"},
                                       {AngstromUnits, "angstroms"}});

/** This structure defines a single spectral state, including a set
 *  of wavelength samples and a reference wavelength.
 *
 *  Each unique spectral state is computed independently within
 *  DIRSIG 5, meaning that a single computational path is used
 *  for each radiometry calculation.
 *
 *  In other words, spectral states should be chosen to minimize
 *  spatio-spectral variation (e.g. the spread of light
 *  through a glass prism). A spectral state should encompass
 *  a range of wavelengths for which photons are expected to
 *  behave similarly as they interact with the scene or as they
 *  are being sampled at the detector.
 *
 *  Each spectral state is defined by providing the wavelength
 *  samples at which radiometry calculations are done (arbitrarily
 *  defined in Microns). The spatial behavior of that state is determined
 *  by the "reference" wavelength -- an index provided into
 *  the wavelength samples which governs the definition and use
 *  of properties within the simulation when geometrically processing
 *  that state. Note that spectral calculations are still done over
 *  the full wavelength samples, the reference wavelength only governs
 *  importance sampled choices (e.g. scattering direction at a path node)
 *
 *  It is possible to use either extreme of this definition -- a
 *  wide spectral response could be modeled as a single spectral
 *  state (losing wavelength-dependent variation) or by assigning
 *  a separate state to each spectral sample (at the cost of
 *  additional runtime). In general, it is recommended to use the
 *  state to represent "chunks" of spectral samples thata are
 *  expected to behave similarly.
 */
struct SpectralState {
  using WavelengthSet = std::vector<double>;
  /// Wavelengths [microns]
  std::vector<double> wavelengthSamples;

  /// Index of reference wavelength (for wavelength-dependent calculations)
  size_t referenceWavelengthIndex = 0;

  double referenceWavelength() const {
    return wavelengthSamples[referenceWavelengthIndex];
  }

  bool operator==(const SpectralState &rhs) const {
    return referenceWavelengthIndex == rhs.referenceWavelengthIndex &&
           wavelengthSamples == rhs.wavelengthSamples;
  }

  bool operator!=(const SpectralState &rhs) const { return !operator==(rhs); }
};

struct StateDateTime {
  /// Reference date time.
  const DateTime *referenceDateTime = nullptr;

  /// Elapsed time from reference [s].
  double elapsedTime = 0;
};

/** Struct definition of a spectral and temporal state of the system. This
 *  defines the wavelengths being computed and the current temporal information.
 *  Note that the temporal state defines a reference date/time and the
 *  current "fine" time scale window relative to that reference (during
 *  a fine scale window no geometry/instance altering changes are made).
 *  An optional gate offset is provided (default is zero) that indicates
 *  if one or more gate collection are in use that open the gate after
 *  the elapsed time (i.e. the gate offset will be between zero and duration).
 *
 *  This struct also has a flag that indicates if any of the wavelengths
 *  within the state samples should trigger passive temperature prediction.
 *  This is assumed to occur at wavelengths greater than 3 microns.
 */
struct State : StateDateTime {
  /// Fine time-scale duration of the state [s].
  double duration = 0;

  /// Current spectral state index.
  size_t spectralStateIndex = 0;

  /// Const access to the spectral states.
  const std::vector<SpectralState> *spectralStates = nullptr;

  const SpectralState &currentSpectralState() const {
    return spectralStates->at(spectralStateIndex);
  }

  /// Minimum offset from the elapsed time to the opening of a gate.
  double gateOffset = 0;

  /// The index of the sensor this spectral state belongs to.
  size_t sensorIndex = 0;

  /// The original index of the spectral state.
  size_t sensorSpectralStateIndex = 0;

  /// Long enough wavelengths to trigger passive temperature prediction (>3um)?
  bool requiresPassiveTemperaturePrediction = false;
};

/**
 * Computes a quantized time in seconds.
 *
 *  \param t           The unquantized time in seconds
 *  \param timeDeltaS  The time interval in seconds to which to quantize
 */
static unsigned int QuantizeTime(double t, int timeDeltaS) {
  auto qt = static_cast<unsigned int>(std::round(t / timeDeltaS));
  qt *= timeDeltaS;
  return qt;
}

/// Constants for pre-defined spectral ranges.
enum class PredSpectralRange {
  UV,   ///< Ultraviolet (0.20 - 0.40 microns @ 0.001 microns)
  Vis,  ///< Visible (0.35 - 0.80 microns @ 0.001 microns)
  Nir,  ///< Near Infrared (0.70 - 1.40 microns @ 0.001 microns)
  Swir, ///< Short-wave Infrared (1.0 - 2.5 microns @ 0.001 microns)
  Mwir, ///< Mid-wave Infrared (3.0 - 5.0 microns @ 0.005 microns)
  Lwir  ///< Long-wave Infrared (7.5 - 14.0 microns @ 0.005 microns)
};

/// Simple struct to hold a uniformly sampled range of wavelengths.
struct UniformlySampledSpectralRange {
  double startMicron = 0;    ///< The start of the wavelength range in microns
  double endMicron = -1;     ///< The end of the wavelength range in microns
  double deltaMicron = 1e-6; ///< The delta of the wavelength range in microns

  /// Th minimum delta allowed in a range
  static constexpr double minDelta = 1e-6;

  /** Constructor
   *
   *  \param start  The start of the wavelength range in microns
   *  \param end    The end of the wavelength range in microns
   *  \param delta  The wavelength delta in microns
   */
  UniformlySampledSpectralRange(double start, double end, double delta)
      : startMicron(start), endMicron(end),
        deltaMicron(std::max(delta, minDelta)) {}

  /** Constructor
   *
   *  \param pred  One of the predefined spectral ranges
   */
  UniformlySampledSpectralRange(PredSpectralRange pred) {
    switch (pred) {
    case PredSpectralRange::UV:
      startMicron = 0.20, endMicron = 0.40, deltaMicron = 0.001;
      break;
    case PredSpectralRange::Vis:
      startMicron = 0.35, endMicron = 0.80, deltaMicron = 0.001;
      break;
    case PredSpectralRange::Nir:
      startMicron = 0.70, endMicron = 1.40, deltaMicron = 0.001;
      break;
    case PredSpectralRange::Swir:
      startMicron = 1.00, endMicron = 2.55, deltaMicron = 0.001;
      break;
    case PredSpectralRange::Mwir:
      startMicron = 2.80, endMicron = 5.20, deltaMicron = 0.005;
      break;
    case PredSpectralRange::Lwir:
      startMicron = 7.50, endMicron = 14.00, deltaMicron = 0.005;
      break;
    }
  }

  /** Constructor
   *
   *  \param pred   One of the predefined spectral ranges
   *  \param delta  The spectral delta override [micron]
   */
  UniformlySampledSpectralRange(PredSpectralRange pred, double delta)
      : UniformlySampledSpectralRange(pred) {
    deltaMicron = delta;
  }
};

/** Holds a number of uniformly-sampled spectral ranges to form a more complex
 *  range.
 */
class UniformlySampledSpectralRanges {
public:
  /// Constructor, empty range
  UniformlySampledSpectralRanges() = default;

  /// Constructor, takes a single range to start with
  UniformlySampledSpectralRanges(const UniformlySampledSpectralRange &r)
      : ranges_({r}) {}

  /// Emplaces a new UniformlySampledSpectralRange into the ranges.
  template <typename... Args> void emplace(Args &&...args) {
    insert(UniformlySampledSpectralRange(std::forward<Args>(args)...));
  }

  /** Inserts a new UniformlySampledSpectralRange into the ranges. This range
   *  may not appear as given, as overlap and sampling rate conflicts are
   *  resolved in this function.
   */
  void insert(const UniformlySampledSpectralRange &r);

  /// Clears the ranges
  void clear() { ranges_.clear(); }

  /// Gets the list of wavelengths that sample all of the ranges as a whole
  template <typename Float = double>
  [[nodiscard]] std::vector<Float> wavelengths() const;

  /** Gets the list of ranges. They will always be sorted in ascending order of
   *  starting wavelengths and will never overlap.
   */
  [[nodiscard]] const std::vector<UniformlySampledSpectralRange> &
  ranges() const {
    return ranges_;
  }

private:
  /** Sorts the list of ranges. Elements are sorted in ascending order of the
   *  start wavelength. If two ranges have the same start, the one with the
   *  larger range comes first, so that on an insert, you can assume you are
   *  only ever contained in a range that comes immediately before you in the
   *  list.
   */
  void sortRanges() {
    std::sort(ranges_.begin(), ranges_.end(), [](const auto &a, const auto &b) {
      if (a.startMicron < b.startMicron)
        return true;
      if (b.startMicron < a.startMicron)
        return false;
      return a.endMicron > b.endMicron;
    });
  }

private:
  std::vector<UniformlySampledSpectralRange> ranges_;
};

/// \cond HIDDEN_SYMBOLS

inline void
UniformlySampledSpectralRanges::insert(const UniformlySampledSpectralRange &r) {
  // Precondition: ranges_ is a list of ranges sorted in ascending order by
  // starting wavelengths. The elements are non-overlapping.

  // Ignore any malformed ranges
  if (r.endMicron < r.startMicron)
    return;

  // Add the new range to the back of the list. This breaks the precondition.
  ranges_.push_back(r);

  // Sort by ascrending start wavelength. If the starting wavelength matches a
  // previously existing one, the bigger range will be first, so that range i
  // may be contained in i - 1, but not the other way around (unless they are
  // identical).
  sortRanges();

  // Now we need to restore the precondition of non-overlapping ranges.
  // Go through and if any two ranges overlap, split at one of the endpoints,
  // preferring the one with the higher resolution.
  for (size_t i = 1; i < ranges_.size(); i++) {
    auto &cur = ranges_[i];
    auto &prev = ranges_[i - 1];

    // If there is no overlap, move on to the next range
    if (cur.startMicron >= prev.endMicron)
      continue;

    // Figure out whether we are entirely contained in the previous range and
    // if we are higher resolution than the previous range. By precondition,
    // we know our start point is >= their start point.
    bool containedInPrev = cur.endMicron <= prev.endMicron;
    bool higherResThanPrev = cur.deltaMicron < prev.deltaMicron;

    // In the case where we are contained, but have a coarser resolution,
    // we are redundant and can be dropped.
    if (containedInPrev && !higherResThanPrev) {
      ranges_.erase(ranges_.begin() + i);
      i--;
      continue;
    }

    // Figure out if we share an endpoint
    bool sameEnd = cur.endMicron == prev.endMicron;

    // If we are the same range (we have a higher resolution), drop the previous
    // range, as they are redundant.
    if (sameEnd && cur.startMicron == prev.startMicron) {
      ranges_.erase(ranges_.begin() + i - 1);
      i--;
      continue;
    }

    // If we get here and are contained, then we know we are higher-res and our
    // start points are different. Handle the case where our endpoints are the
    // same, in which case, we want to trim the previous range to end at our
    // start and then use our higher-res in our range.
    if (containedInPrev && sameEnd) {
      prev.endMicron = cur.startMicron;
      continue;
    }

    // If we get here and our contained, then we know our start and end points
    // are both different and we are higher-res. In this case, we need to do a
    // 3-way split, using the lower-res on the two ends and leaving the
    // higher-res in the middle. By precondition, we know that the less than
    // side will not overlap with any other ranges, so we only need to recurse
    // on the greater than side.
    if (containedInPrev) {
      auto prevEnd = prev.endMicron;
      prev.endMicron = cur.startMicron;
      emplace(cur.endMicron, prevEnd, prev.deltaMicron);
      return;
    }

    // If the current range has a higher spectral resolution than the previous
    // range, then we "win" the merge and trim the previous range. Otherwise,
    // the previous range wins and we'll trim our range
    if (cur.deltaMicron < prev.deltaMicron) {
      prev.endMicron = cur.startMicron;
    } else {
      cur.startMicron = prev.endMicron;
    }
  }
}

template <typename Float>
std::vector<Float> UniformlySampledSpectralRanges::wavelengths() const {
  static_assert(std::is_same_v<Float, float> || std::is_same_v<Float, double>,
                "Float must be float or double");
  std::vector<Float> out;

  // Go through all the ranges and generate samples (keep track of the minimum
  // delta for de-duping later.
  double minDelta = 1e10;
  for (const auto &r : ranges_) {
    minDelta = std::min(minDelta, r.deltaMicron);
    double w;
    for (w = r.startMicron; w <= r.endMicron; w += r.deltaMicron)
      out.push_back(Float(w));
    if (w - r.endMicron < 0.01 * r.deltaMicron)
      out.push_back(Float(r.endMicron));
  }

  // De-duplicate the wavelenghts
  out.erase(
      std::unique(out.begin(), out.end(),
                  [&](auto a, auto b) { return fabs(a - b) < 0.1 * minDelta; }),
      out.end());

  return out;
}

/// \endcond

/// \}

} // namespace d5
