/** Interface file for string utilities
 *
 *  \file StringUtils.h
 */

#pragma once

#include <algorithm>
#include <cctype>
#include <climits>
#include <iterator>
#include <limits>
#include <sstream>
#include <string>
#include <string_view>
#include <type_traits>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/string/Numeric.h"

namespace d5 {

/** \defgroup common_string String Utilities
 *  Additions to the C++ Standard's string class
 *  \ingroup common
 */

/// \addtogroup common_string
/// \{

/// Type check for whether T is either a string or a string_view
template <typename T>
using is_string_or_view = std::disjunction<std::is_same<T, std::string>,
                                           std::is_same<T, std::string_view>>;

/// Convenience alias for is_string_or_view<T>::value
template <typename T>
constexpr bool is_string_or_view_v = is_string_or_view<T>::value;

inline void transformToLower(std::string &s) {
  std::transform(s.begin(), s.end(), s.begin(),
                 [](unsigned char ch) { return std::tolower(ch); });
}

inline void transformToUpper(std::string &s) {
  std::transform(s.begin(), s.end(), s.begin(),
                 [](unsigned char ch) { return std::toupper(ch); });
}

template <typename String> inline std::string toLower(String &&s) {
  static_assert(is_string_or_view_v<std::decay_t<String>>);
  std::string t(s.begin(), s.end());
  transformToLower(t);
  return t;
}

template <typename String> inline std::string toUpper(String &&s) {
  static_assert(is_string_or_view_v<std::decay_t<String>>);
  std::string t(s.begin(), s.end());
  transformToUpper(t);
  return t;
}

template <typename String> inline std::string_view trimmedView(String &&s) {
  static_assert(is_string_or_view_v<std::decay_t<String>>);
  auto pos0 = s.data();
  auto pos1 = s.data() + s.size() - 1;
  auto end = s.data() + s.size();
  while (true) {
    if (pos0 == end)
      return {};
    if (not std::isspace(*pos0))
      break;
    ++pos0;
  }
  while (pos1 > pos0 and std::isspace(*pos1))
    --pos1;
  return {pos0, size_t(pos1 - pos0 + 1)};
}

template <typename String> inline std::string trimmed(String &&s) {
  static_assert(is_string_or_view_v<std::decay_t<String>>);
  return std::string{trimmedView(s)};
}

template <typename Iterator, typename Sentinel, typename ToPrintable>
inline std::string Join(Iterator beg, Sentinel end, const char *const delimiter,
                        ToPrintable toPrintable) {
  if (beg == end)
    return std::string();

  std::ostringstream os;
  do {
    os << toPrintable(*beg);
    ++beg;
    if (beg != end) {
      os << delimiter;
    } else
      break;
  } while (true);

  return os.str();
}

template <typename Iterator, typename Sentinel>
inline std::string Join(Iterator &&beg, Sentinel &&end,
                        const char *const delimiter) {
  return Join(std::forward<Iterator>(beg), std::forward<Sentinel>(end),
              delimiter, [](auto &&x) { return std::forward<decltype(x)>(x); });
}

template <typename Range>
inline std::string Join(const Range &elements, const char *const delimiter) {
  return Join(std::begin(elements), std::end(elements), delimiter);
}

template <typename Range, typename ToPrintable>
inline std::string Join(const Range &elements, const char *const delimiter,
                        ToPrintable &&toPrintable) {
  return Join(std::begin(elements), std::end(elements), delimiter,
              std::forward<ToPrintable>(toPrintable));
}

template <typename T> inline size_t FindFirstOfRequired(T &&s, char c) {
  static_assert(is_string_or_view_v<std::decay_t<T>>,
                "T must be string or string_view");
  auto p = s.find_first_of(c);
  if (p != std::decay_t<T>::npos)
    return p;
  throw std::runtime_error(std::string("expected '") + c + "' in \"" +
                           std::string(s) + "\"");
}

inline bool endsWith(std::string_view haystack, std::string_view needle) {
  if (haystack.size() >= needle.size())
    return haystack.compare(haystack.size() - needle.size(), needle.size(),
                            needle) == 0;
  else
    return false;
}

inline std::pair<std::string_view, std::string_view>
splitExtension(std::string_view s) {
  std::string_view s0 = s;
  size_t pos = s.find_last_of('/');
  if (pos == std::string_view::npos)
    pos = s.find_last_of('\\'); // Windows?
  if (pos != std::string_view::npos)
    s = s.substr(pos + 1);
  std::string_view extension = s.substr(s.find_first_of('.'));
  std::string_view base(s0.data(), extension.data() - s0.data());
  return {base, extension};
}

struct TokenList {
  struct IteratorSentinel {};

  struct Iterator {
    using value_type = std::string_view;
    using reference = std::string_view;
    using difference_type = std::ptrdiff_t;
    using iterator_category = std::input_iterator_tag;

    constexpr Iterator() = default;
    constexpr Iterator(std::string_view input, std::string_view delim)
        : input(input), delim(delim) {
      pos0 = input.begin();
      pos1 = input.begin();
      operator++();
    }

    constexpr std::string_view operator*() noexcept {
      return std::string_view(&*pos0, pos1 - pos0);
    }
    constexpr Iterator &operator++() noexcept {
      while (pos1 < input.end() && delim.find(*pos1) != delim.npos)
        ++pos1;
      pos0 = pos1;
      while (pos1 < input.end() && delim.find(*pos1) == delim.npos)
        ++pos1;
      return *this;
    }
    constexpr Iterator operator++(int) noexcept {
      Iterator copy = *this;
      operator++();
      return copy;
    }
    constexpr bool operator==(IteratorSentinel) const noexcept {
      return pos0 == input.end();
    }
    constexpr bool operator!=(IteratorSentinel) const noexcept {
      return pos0 != input.end();
    }

    std::string_view input;
    std::string_view::iterator pos0 = {};
    std::string_view::iterator pos1 = {};
    std::string_view delim;
  };
  constexpr TokenList() noexcept = default;
  constexpr TokenList(std::string_view input, std::string_view delim) noexcept
      : input(input), delim(delim) {}
  [[nodiscard]] constexpr Iterator begin() const noexcept {
    return Iterator(input, delim);
  }
  [[nodiscard]] constexpr IteratorSentinel end() const noexcept { return {}; }
  [[nodiscard]] constexpr bool empty() const noexcept {
    return begin() == end();
  }
  [[nodiscard]] constexpr size_t size() const noexcept {
    size_t n = 0;
    for (auto itr = begin(); itr != end(); ++itr)
      ++n;
    return n;
  }
  [[nodiscard]] std::string_view operator[](size_t pos) const {
    auto itr = begin();
    while (pos-- != 0)
      ++itr;
    return *itr;
  }
  [[nodiscard]] std::string_view at(size_t pos) const {
    DIRSIG_THROW_IF(pos >= size(), "Index out of range!");
    return operator[](pos);
  }

public:
  std::string_view input;
  std::string_view delim;
};

/// \}

} // namespace d5
