/** Interface file for customization points through tag_invoke.
 *
 *  This file is largely taken from libunifex, which is the basis for the
 *  proposal to add tag_invoke to the standard library.
 *
 *  \file TagInvoke.h
 */

#pragma once

#include <type_traits>
#include <utility>

#include "dirsig5/base/Macros.h"

namespace d5 {

/// Detail namespace for implementing tag_invoke
namespace _tag_invoke {

/// Base case of tag_invoke, this is never called directly
void tag_invoke();

/** This is just a proxy struct that allows tag_invoke to be passed around as
 *  a value to functions like the standard algorithms
 */
struct _fn {
  /// Call operator, simply a passthrough to tag_invoke
  template <typename CPO, typename... Args>
  constexpr auto operator()(CPO &&cpo, Args &&...args) const
      noexcept(noexcept(tag_invoke(D5_FWD(cpo), D5_FWD(args)...)))
          -> decltype(tag_invoke(D5_FWD(cpo), D5_FWD(args)...)) {
    return tag_invoke(D5_FWD(cpo), D5_FWD(args)...);
  }
};

/** Dummy type to use for a return type in the positive case of an expression
 *  SFINAE test
 */
using yes_type = char;

/// Negative case return type for an expression SFINAE test
using no_type = char (&)[2];

/** Expression SFINAE test function for tag_invoke. This overload will be seen
 *  if the tag_invoke() function call is well-formed.
 */
template <typename CPO, typename... Args>
auto try_tag_invoke(int) //
    noexcept(noexcept(tag_invoke(std::declval<CPO &&>(),
                                 std::declval<Args &&>()...)))
        -> decltype(static_cast<void>(tag_invoke(std::declval<CPO &&>(),
                                                 std::declval<Args &&>()...)),
                    yes_type{});

/** This overload has the lowest match priority, so will be selected if
 *  the tag_invoke() is not well-formed (the expression SFINAE check fails)
 */
template <typename CPO, typename... Args>
static no_type try_tag_invoke(...) noexcept(false);
} // namespace _tag_invoke

/// Detail namespace for tag_invoke
namespace _tag_invoke_cpo {

/** A global tag_invoke passthrough object. This allows tag_invoke to be passed
 *  as a callback to other functions.
 */
inline constexpr _tag_invoke::_fn tag_invoke{};
} // namespace _tag_invoke_cpo

// Pull tag_invoke into this namespace. If we just declare it here, we would
// get errors about tag_invoke being declared as a non-function when people
// tried to implement tag_invoke() for their use cases.
using namespace _tag_invoke_cpo;

/// Type trait to get the return type of a tag_invoke() call
template <typename CPO, typename... Args>
using tag_invoke_result_t =
    decltype(tag_invoke(std::declval<CPO &&>(), std::declval<Args &&>()...));

/** Tests if the given CPO/Args combination would result in a valid
 *  tag_invoke call
 */
template <typename CPO, typename... Args>
inline constexpr bool is_tag_invocable_v =
    (sizeof(_tag_invoke::try_tag_invoke<CPO, Args...>(0)) ==
     sizeof(_tag_invoke::yes_type));

/** Tests if the given CPO/Args combination would result in a valid
 *  tag_invoke call that is noexcept
 */
template <typename CPO, typename... Args>
inline constexpr bool is_nothrow_tag_invocable_v =
    noexcept(_tag_invoke::try_tag_invoke<CPO, Args...>(0));
} // namespace d5

/** Defines a customization point object and type
 *
 *  \param T  The name of the customization point object, the type will be T
 *            with a _t appended.
 */
#define D5_DEFINE_TAG_INVOKE_CPO(T)                                            \
  inline constexpr struct T##_t final {                                        \
    constexpr T##_t() = default;                                               \
                                                                               \
    template <typename... Args>                                                \
    constexpr auto operator()(Args &&...args) const                            \
        noexcept(d5::is_nothrow_tag_invocable_v<T##_t, Args...>)               \
            -> d5::tag_invoke_result_t<T##_t, Args...> {                       \
      return d5::tag_invoke(*this, std::forward<Args>(args)...);               \
    }                                                                          \
  } T /*NOLINT*/;                                                              \
  static_assert(true, "Remember the ; after D5_DEFINE_TAG_INVOKE_CPO")
