/** Interface file for temperature utilities
 *
 *  \file Temperature.h
 */

#pragma once

#include "Json.h"

namespace d5 {

/// \addtogroup common
/// \{

/// Converts the given temperature from Celsius to Kelvin
constexpr double CelsiusToKelvin(double c) { return c + 273.15; }

/// Converts the given temperature from Kelvin to Celsius
constexpr double KelvinToCelsius(double k) { return k - 273.15; }

/// Converts the given temperature from Fahrenheit to Celsius
constexpr double FahrenheitToCelsius(double f) { return (5. / 9.) * (f - 32); }

/// Converts the given temperature from Celsius to Fahrenheit
constexpr double CelsiusToFahrenheit(double c) { return (9. / 5.) * c + 32; }

/// Converts the given temperature from Fahrenheit to Kelvin
constexpr double FahrenheitToKelvin(double f) {
  return CelsiusToKelvin(FahrenheitToCelsius(f));
}

/// Converts the given temperature from Kelvin to Fahrenheit
constexpr double KelvinToFahrenheit(double k) {
  return CelsiusToFahrenheit(KelvinToCelsius(k));
}

/// Represents a temperature and its units
class Temperature {
public:
  /// Supported temperature units
  enum Units { Kelvin, Celsius, Fahrenheit };

  /** Converts a temperature from one unit to another
   *
   *  \param t     The value of the temperature to convert
   *  \param from  The units of \c t
   *  \param to    The units to which to convert
   *
   *  \return The converted temperature in units of \c from
   */
  static double Convert(double t, Units from, Units to) {
    switch (from) {
    case Kelvin:
      switch (to) {
      case Celsius:
        return KelvinToCelsius(t);
      case Fahrenheit:
        return KelvinToFahrenheit(t);
      case Kelvin:
        break;
      }
      break;
    case Celsius:
      switch (to) {
      case Kelvin:
        return CelsiusToKelvin(t);
      case Fahrenheit:
        return CelsiusToFahrenheit(t);
      case Celsius:
        break;
      }
      break;
    case Fahrenheit:
      switch (to) {
      case Kelvin:
        return FahrenheitToKelvin(t);
      case Celsius:
        return FahrenheitToCelsius(t);
      case Fahrenheit:
        break;
      }
      break;
    }
    return t;
  }

  /// Default constructor. Absolute zero.
  Temperature() = default;

  /// Copy constructor
  Temperature(const Temperature &) = default;

  /// Move constructor
  Temperature(Temperature &&) noexcept = default;

  /// Copy assignment operator
  Temperature &operator=(const Temperature &) = default;

  /// Move assignment operator
  Temperature &operator=(Temperature &&) noexcept = default;

  /// Constructor
  Temperature(double t, Units units) : tempK_(Convert(t, units, Kelvin)) {}

  /// Factory function for creating a temperature from Kelvin
  static Temperature fromKelvin(double k) { return Temperature(k); }

  /// Factory function for creating a temperature from Celsius
  static Temperature fromCelsius(double c) {
    return Temperature(CelsiusToKelvin(c));
  }

  /// Factory function for creating a temperature from Fahrenheit
  static Temperature fromFahrenheit(double f) {
    return Temperature(f, Fahrenheit);
  }

  void setKelvin(double t) { *this = fromKelvin(t); }

  void setCelsius(double t) { *this = fromCelsius(t); }

  void setFahrenheit(double t) { *this = fromFahrenheit(t); }

  /// Templated accessor for the temperature value
  template <Units U> [[nodiscard]] double get() const {
    if constexpr (U == Kelvin)
      return tempK_;
    else if constexpr (U == Celsius)
      return KelvinToCelsius(tempK_);
    else if constexpr (U == Fahrenheit)
      return KelvinToFahrenheit(tempK_);
    else
      return 0;
  }

  /// Convenience accessor for Kelvin
  [[nodiscard]] double kelvin() const { return get<Kelvin>(); }

  /// Convenience accessor for Celsius
  [[nodiscard]] double celsius() const { return get<Celsius>(); }

  /// Convenience accessor for Celsius
  [[nodiscard]] double fahrenheit() const { return get<Fahrenheit>(); }

  bool operator==(const Temperature &other) const {
    return tempK_ == other.tempK_;
  }

  bool operator!=(const Temperature &other) const {
    return tempK_ != other.tempK_;
  }

  bool operator<(const Temperature &other) const {
    return tempK_ < other.tempK_;
  }

  bool operator<=(const Temperature &other) const {
    return tempK_ <= other.tempK_;
  }

  bool operator>(const Temperature &other) const {
    return tempK_ > other.tempK_;
  }

  bool operator>=(const Temperature &other) const {
    return tempK_ >= other.tempK_;
  }

private:
  /// Construct directly from a Kelvin temperature
  explicit Temperature(double k) : tempK_(k) {}

private:
  /// Temperature stored in Kelvin
  double tempK_ = 0;
};

D5_TO_JSON(Temperature, j, t) { j = Json{{"k", t.get<Temperature::Kelvin>()}}; }

D5_FROM_JSON(Temperature, j, t) { t = Temperature::fromKelvin(j.at("k")); }

/// \}

} // namespace d5
