#pragma once

#include <chrono>
#include <string>

#include "dirsig5/base/Export.h"

namespace d5 {

/** Simple timer with a start, stop and ability to get the duration
 *  \ingroup utility
 */
class D5_BASE_EXPORT Timer {
public:
  /// Constants for the units of time
  enum TimeUnit {
    kHours = 0,
    kMinutes,
    kSeconds,
    kMilliseconds,
    kTimeUnitCount
  };

  Timer() = default;

  [[nodiscard]] bool isRunning() const { return running_; }

  /// Start the timer, return whether it was successful
  bool start() {
    if (running_)
      return false;
    beginTime_ = std::chrono::steady_clock::now();
    running_ = true;
    return true;
  }

  /// Stop the timer, return whether it was successful
  bool stop() {
    if (not running_)
      return false;
    endTime_ = std::chrono::steady_clock::now();
    running_ = false;
    return true;
  }

  /// Get the elapsed time duration as a floating point number
  [[nodiscard]] double duration(TimeUnit timeUnit) const;

  /// Prints the elapsed time duration as a string
  ///
  /// \param format A string which contains the format specifiers and
  /// placement
  ///               for the time units.  Valid specifiers are:
  ///               \li \c %h - hours
  ///               \li \c %m - minutes
  ///               \li \c %s - seconds
  ///               \li \c %v - milliseconds
  ///               The specifier may include a precision decimal (such as
  ///               %2.2s)
  ///
  /// \return The time duration based on the format string
  ///
  std::string format(const char *format) const;

private:
  /// Whether the timer is currently running
  bool running_ = false;

  /// The start point in time
  std::chrono::time_point<std::chrono::steady_clock> beginTime_;

  /// The stop point in time
  std::chrono::time_point<std::chrono::steady_clock> endTime_;
};
} // namespace d5
