#pragma once

#include <array>
#include <functional>
#include <string>
#include <unordered_map>
#include <vector>

#include "dirsig5/base/IntID.h"
#include "dirsig5/base/Json.h"

namespace d5 {

/** \defgroup common_truth Truth Requests
 *  Holds common definitions around model truth requests
 *  \ingroup common
 */

/// \addtogroup common_truth
/// \{

/** XMacro for the list of truth values
 *
 *  \param X  Function macro of the form X(name, description, format_str)
 */
#define D5_TRUTH_TYPES(X)                                                      \
  X(AVG_HIT_X, "Average X of hit point in local coords [m]",                   \
    "Scene ENU X Coordinate [m]")                                              \
  X(AVG_HIT_Y, "Average Y of hit point in local coords [m]",                   \
    "Scene ENU Y Coordinate [m]")                                              \
  X(AVG_HIT_Z, "Average Z of hit point in local coords [m]",                   \
    "Scene ENU Z Coordinate [m]")                                              \
  X(AVG_DISTANCE, "Average distance to first surface[m]", "Distance [m]")      \
  X(AVG_NRM_X, "Average X of the normal", "Surface Normal X Component")        \
  X(AVG_NRM_Y, "Average Y of the normal", "Surface Normal Y Component")        \
  X(AVG_NRM_Z, "Average Z of the normal", "Surface Normal Z Component")        \
  X(AVG_COSINE, "Average cosine of view angle relative to surface",            \
    "View Angle Cosine")                                                       \
  X(SHAPE_FACTOR, "Fraction of hemisphere that is unshadowed[0 - 1]",          \
    "Sky Fraction")                                                            \
  X(SAMPLE_COUNT, "Number of samples used", "Sample Count")                    \
  X(BOUNCE_COUNT, "Average number of intersection nodes in path",              \
    "Bounce Count")                                                            \
  X(SUN_SHADOW, "Fraction of pixel exposed to sun w / transmission",           \
    "Sun Fraction")                                                            \
  X(SUN_EXPOSURE, "Sun shadow with cosine of facing surface", "Sun Exposure")  \
  X(MOON_SHADOW, "Fraction of pixel exposed to moon w / transmission",         \
    "Moon Fraction")                                                           \
  X(MOON_EXPOSURE, "Moon shadow with cosine of facing surface",                \
    "Moon Exposure")                                                           \
  X(TOP_MAT_ID, "Dominant material index in a pixel(scene based)",             \
    "Dominant Material Index")                                                 \
  X(TOP_OPT_ID, "Dominant optical prop index within a pixel",                  \
    "Dominant Optical Property Index")                                         \
  X(TOP_SCENE_ID, "Top scene id in a pixel([0 + ] scene[-] internal)",         \
    "Scene Index")                                                             \
  X(TOP_INST_ID, "Top instance id within a pixel(scene based)",                \
    "Instance Index")                                                          \
  X(TOP_GEOM_ID, "Top(base) geometry id in a pixel(scene based)",              \
    "Geometry Index")                                                          \
  X(TOP_PRIM_ID, "Top primitive(facet) id in a pixel(scene based)",            \
    "Primitive Index")                                                         \
  X(AVG_GEO_LAT, "Average geodetic latitude[deg + N]",                         \
    "Geodetic Latitude [degrees +N]")                                          \
  X(AVG_GEO_LON, "Average geodetic longitude[deg + E]",                        \
    "Geodetic Longitude [degrees +E]")                                         \
  X(AVG_GEO_ALT, "Average geodetic altitude above sea level[m]",               \
    "WGS84 Altitude [m]")                                                      \
  X(SOURCE_COUNT, "Number of sources used for a problem", "Source Count")      \
  X(BARY_U, "Barycentric coordinates(v0 weight)", "Barycentric U Coordinate")  \
  X(BARY_V, "Barycentric coordinates(v1 weight)", "Barycentric V Coordinate")  \
  X(BARY_W, "Barycentric coordinates(v2 weight)", "Barycentric W Coordinate")  \
  X(AVG_U, "Average UV coordinate(U)", "Texture U Coordinate")                 \
  X(AVG_V, "Average UV coordinate(V)", "Texture V Coordinate")                 \
  X(AVG_TEMP, "Average temperature[K]", "Temperature [K]")                     \
  X(TOP_TEMP_MODEL, "Top temperature model index",                             \
    "Dominant Temperature Model Index")                                        \
  X(MAT_ABUNDANCE, "Abundance of materials in pixels",                         \
    "Mat ID '{}' Abundance [fraction]")                                        \
  X(INST_ABUNDANCE, "Abundance of tagged instances in pixels",                 \
    "Tag '{}' Abundance [fraction]")                                           \
  X(PATH_RADIANCE, "Radiance along primary path segment toward sensor",        \
    "Path Radiance")                                                           \
  X(PATH_TRANSMISSION, "Transmission along primary path segment",              \
    "Path Transmission")                                                       \
  X(PATH_ANGLE_ZENITH, "Zenith angle for initial, average path segment",       \
    "Path Zenith [degrees declination]")                                       \
  X(PATH_ANGLE_AZIMUTH, "Azimuth angle for initial, average path segment",     \
    "Path Azimuth [degrees East of North]")                                    \
  X(PATH_RIGHT_ACCENSION,                                                      \
    "Right-Accension angle for initial, average path segment",                 \
    "Path Right Accension [degrees]")                                          \
  X(PATH_DECLINATION, "Declination angle for initial, average path segment",   \
    "Path Declination [degrees]")                                              \
  X(PATH_LIGHT_TIME, "Light time for the initial path segment",                \
    "Path Light Time [seconds]")                                               \
  X(PROP_START_DIR_X,                                                          \
    "Initial X component of propagation (at sensor) in ECEF coords",           \
    "Path Direction ECEF X Component (at sensor)")                             \
  X(PROP_START_DIR_Y,                                                          \
    "Initial Y component of propagation (at sensor) in ECEF coords",           \
    "Path Direction ECEF Y Component (at sensor)")                             \
  X(PROP_START_DIR_Z,                                                          \
    "Initial Z component of propagation (at sensor) in ECEF coords",           \
    "Path Direction ECEF Z Component (at sensor)")                             \
  X(PROP_END_DIR_X,                                                            \
    "Final X component of propagation (at scene) in ECEF coords",              \
    "Path Direction ECEF X Component (at scene)")                              \
  X(PROP_END_DIR_Y,                                                            \
    "Final Y component of propagation (at scene) in ECEF coords",              \
    "Path Direction ECEF Y Component (at scene)")                              \
  X(PROP_END_DIR_Z,                                                            \
    "Final Z component of propagation (at scene) in ECEF coords",              \
    "Path Direction ECEF Z Component (at scene)")                              \
  X(SOLAR_IRRADIANCE, "Average solar irradiance onto the surface",             \
    "Solar Irradiance (w/ cos)")                                               \
  X(LUNAR_IRRADIANCE, "Average lunar irradiance onto the surface",             \
    "Lunar Irradiance (w/ cos)")                                               \
  X(MED_HIT_X, "Average x - location hit point in a medium[m]",                \
    "Scene ENU X Coordinate (medium) [m]")                                     \
  X(MED_HIT_Y, "Average y - location hit point in a medium[m]",                \
    "Scene ENU Y Coordinate (medium) [m]")                                     \
  X(MED_HIT_Z, "Average z - location hit point in a medium[m]",                \
    "Scene ENU Z Coordinate (medium) [m]")                                     \
  X(VOL_PATH_LENGTH, "Average path length in a volume[m]",                     \
    "Volume Path Length [m]")                                                  \
  X(VOL_CONC_LENGTH, "Average concentration length in a volume[conc*m]",       \
    "Volume Concentration Length [conc*m]")                                    \
  X(AVG_ECEF_X, "Average X of hit point in ECEF coords [m]",                   \
    "ECEF X Coordinate [m]")                                                   \
  X(AVG_ECEF_Y, "Average Y of hit point in ECEF coords [m]",                   \
    "ECEF Y Coordinate [m]")                                                   \
  X(AVG_ECEF_Z, "Average Z of hit point in ECEF coords [m]",                   \
    "ECEF Z Coordinate [m]")                                                   \
  X(CONVERGENCE_STATUS, "Flag indicating how the convergence finalized",       \
    "Convergence Status")                                                      \
  X(CLOUD_LAYER_FRACTION, "Fraction of the pixel that is a cloud layer",       \
    "Cloud layer fraction")

/// \cond HIDDEN_SYMBOLS
#define X_ENUM_DECL(name, desc, fmtStr) name, /**< desc */
#define X_JSON_ENTRY(name, desc, fmtStr) {name, desc},
#define X_FMT_CASE(name, desc, fmtStr)                                         \
  case TruthType::name:                                                        \
    return fmtStr;
/// \endcond

/** Constants for all truth values available within DIRSIG.
 *
 *  Truth can be written in arbitrary order to an output buffer.
 *  Spectral values will be returned for the reference wavelength of the
 *  current spectral state.
 */
enum TruthType {
  UNKNOWN = 0, ///< Undefined truth type (should not be used)
  D5_TRUTH_TYPES(X_ENUM_DECL)
};

#undef X_ENUM_DECL

D5_JSON_SERIALIZE_ENUM(TruthType,
                       {{UNKNOWN, "Unknown"}, D5_TRUTH_TYPES(X_JSON_ENTRY)});

#undef X_JSON_ENTRY

/** Returns a format string for an truth image band (to be evaluated with
 *  fmt::format and the parameters for the truth request
 *
 *  \param type  The type of the truth request
 *
 *  \return Format string
 */
inline std::string_view TruthBandNameFmtStr(TruthType type) {
  using namespace d5;
  switch (type) {
    D5_TRUTH_TYPES(X_FMT_CASE)
  case TruthType::UNKNOWN:
    break;
  }
  return "Unknown";
}

#undef X_FMT_CASE

/// Structure to represent a request for a type of truth
struct Truth {
  /// Default constructor
  Truth() = default;

  /// Constructor with a type
  Truth(TruthType t) : type(t) {}

  /// Assignment from a truth type
  Truth &operator=(const TruthType &t) {
    type = t;
    return *this;
  }

  /// Implicit conversion to a truth type
  operator TruthType() const { return type; }

  /// Truth type
  TruthType type = UNKNOWN;

  /// Name/value pair for parameters
  struct Parameter {
    std::string name;
    std::string value;

    NLOHMANN_DEFINE_TYPE_INTRUSIVE(Parameter, name, value);
  };

  /// Optional truth parameters
  std::vector<Parameter> parameters;
};

D5_TO_JSON(Truth, j, t) {
  j = Json{{"type", t.type}};
  if (!t.parameters.empty())
    j["parameters"] = t.parameters;
}

D5_FROM_JSON(Truth, j, t) {
  j.at("type").get_to(t.type);
  Optional(j, "parameters", t.parameters);
}

/** Callback for the instance abundance truth. Will be called during
 *  the truth finalization for a problem if the stored count is
 *  greater than zero. The abundance is expressed as the fractional
 *  number of primary rays that intersected the instance.
 */
using InstanceAbundanceCallback =
    std::function<void(size_t, size_t, size_t, float)>;

/** This struct is used to create a request for instance truth (probably
 *  through an instance wrapper request). The count is used to track
 *  the number of primary intersections within a problem and callback that
 *  is triggered if the instance is found in a problem (it can also be
 *  stored as a normal truth output).
 */
struct InstanceAbundance {
  IntID sceneID = INVALID_ID;
  IntID baseID = INVALID_ID;
  unsigned instID = 0;
  InstanceAbundanceCallback *callback = nullptr;
  size_t bufferIndex = 0;
  size_t count = 0;
};

using Keypoints =
    std::unordered_map<std::string, std::vector<std::array<float, 3>>>;

/// \}

} // namespace d5
