/** Common extensions to type_traits
 *
 *  \file TypeTraits.h
 */

#pragma once

#include <type_traits>

// we use the top-level namespace for non-verbose re-use

namespace d5 {

/// \addtogroup common_type_traits
/// \{

/// Always false, useful for static asserts
template <class... T> inline constexpr bool always_false = false;

/** Tests if two templates are equal to each other
 *
 *  \tparam A  The first template
 *  \tparam B  The second template
 */
template <template <class...> class A, template <class...> class B>
struct is_same_tmpl : std::false_type {};

/// True case when the templates are equal
template <template <class...> class A>
struct is_same_tmpl<A, A> : std::true_type {};

/// Convenience alias for is_same_tmpl
template <template <class...> class A, template <class...> class B>
inline constexpr bool is_same_tmpl_v = is_same_tmpl<A, B>::value;

/** Tests if the given class template is found in any of the following template
 *  parameters.
 *
 *  \tparam T     The template to search for
 *  \tparam Rest  The templates in which to search
 */
template <template <class...> class T, template <class...> class... Rest>
struct is_any_tmpl : std::false_type {};

/// Base case of only one left
template <template <class...> class T, template <class...> class First>
struct is_any_tmpl<T, First> : is_same_tmpl<T, First> {};

/// General search case for is_any_tmpl
template <template <class...> class T, template <class...> class First,
          template <class...> class... Rest>
struct is_any_tmpl<T, First, Rest...>
    : std::integral_constant<bool, is_same_tmpl<T, First>::value ||
                                       is_any_tmpl<T, Rest...>::value> {};

/// Convenience shorthand for is_any_tmpl
template <template <class...> class... Ts>
inline constexpr bool is_any_tmpl_v = is_any_tmpl<Ts...>::value;

/// Backport of C++20 std::remove_cvref
template <class T> struct remove_cvref {
  using type = std::remove_cv_t<std::remove_reference_t<T>>;
};

/// Convenience alias for remove_cvref::type
template <class T> using remove_cvref_t = typename remove_cvref<T>::type;

/// Tests whether a struct has a static constexpr bool value
template <typename T, typename = int>
struct has_value_member : std::false_type {};

/// True case for has_value_member
template <typename T>
struct has_value_member<T, decltype((void)T::value, 0)> : std::true_type {};

/// Convenience alias for has_value_member<T>::value
template <typename T>
inline constexpr bool has_value_member_v = has_value_member<T>::value;

/// Gets the underlying value of an enum
template <typename T> inline constexpr auto underlying_value(T v) {
  static_assert(std::is_enum_v<T>, "T must be an enum");
  return static_cast<std::underlying_type_t<T>>(v);
}

/// \}
} // namespace d5
