/** Utility functions for std::variant
 *
 *  \file variant.h
 */

#pragma once

#include <type_traits>
#include <variant>

namespace d5 {

/// \addtogroup common
/// \{

namespace detail {

/// Templated struct to inherit () operator from multiple classes
template <typename... Ts> struct overload : Ts... {
  using Ts::operator()...;
};

/// Deduction guide
template <typename... Ts> overload(Ts...) -> overload<Ts...>;
} // namespace detail

/** Helper function to make using std::visit easier with lambda functions
 *
 *  \param variant   The std::variant to be used
 *  \param visitors  A variable number of visitors to use with std::visit
 *
 *   \return The return type of the visitor being used
 *
 * Example usage:
 * \code
 * using var_t = std::variant<int, double, std::string>;
 *
 * std::vector<var_t> testers = {"Hello", 1, 5.0};
 *
 * for (const auto& test : testers) {
 *   overloaded_visit(
 *     test, [](int i) { std::cout << i + i << std::endl; },
 *     [](double d) { std::cout << d * d << std::endl; },
 *     [](const std::string& s) { std::cout << s + " world" << std::endl; });
 * }
 *
 * // Result:
 * // Hello World
 * // 2
 * // 25
 * \endcode
 */
template <typename Variant, typename... Visitors>
decltype(auto) overloaded_visit(Variant &&variant, Visitors &&...visitors) {
  return std::visit(detail::overload{std::forward<Visitors>(visitors)...},
                    std::forward<Variant>(variant));
}

/// \cond HIDDEN_SYMBOLS
template <class T, class VariantT> struct is_variant_member;
/// \endcond

/// Tests if type is contained in a given variant
template <class T, class... VariantMembers>
struct is_variant_member<T, std::variant<VariantMembers...>>
    : public std::disjunction<std::is_same<T, VariantMembers>...> {};

/// Alias for is_variant_member<T, ...>::value
template <class T, class... VariantMembers>
constexpr bool is_variant_member_v =
    is_variant_member<T, VariantMembers...>::value;

/// \}

} // namespace d5
