/** Interface file for the Weather class
 *
 *  \file Weather.h
 */

#pragma once

#include <optional>

#include "dirsig5/base/Json.h"
#include "dirsig5/base/Temperature.h"

namespace d5 {

/// \addtogroup common
/// \{

/// Holds a weather data record for a single time point and spatial point.
struct Weather {
public:
  /// The air temperature.
  Temperature airTemperature = Temperature::fromCelsius(20);

  /// Air pressure in millibars.
  double airPressureMbar = 1013.2;

  /// Relative humidity between 0 and 1.
  double relativeHumidity = 0.4;

  /// Wind speed in meters per second.
  double windSpeedMS = 5;

  /// Get the dew point (approximate if not set first).
  [[nodiscard]] Temperature dewPoint() const {
    if (measuredDewPoint_)
      return *measuredDewPoint_;
    return Temperature::fromCelsius(airTemperature.get<Temperature::Celsius>() -
                                    20 * (1 - relativeHumidity));
  }

  /// Set the dew point.
  void setDewPoint(const std::optional<Temperature> &dewPoint) {
    measuredDewPoint_ = dewPoint;
  }

  /// Set the dew point.
  Weather &withDewPoint(const Temperature &dewPoint) {
    setDewPoint(dewPoint);
    return *this;
  }

  /// Clear the dew point (calls to \c dewPoint() will approximate from
  /// relative humidity.)
  void clearDewPoint() { measuredDewPoint_.reset(); }

  /// Get whether a measured dew point has been set.
  [[nodiscard]] bool hasMeasuredDewPoint() const {
    return measuredDewPoint_.has_value();
  }

  /// The direct insolation in watts per meter squared.
  double directInsolationWM2 = 80;

  /// The diffuse insolation in watts per meter squared.
  double diffuseInsolationWM2 = 10;

  /// The (optional) wind direction (radians E of N),
  /// the default direction is coming from the East
  static constexpr double defaultWindDirectionRad = 0.5 * M_PI;
  double windDirectionRad = defaultWindDirectionRad;

private:
  std::optional<Temperature> measuredDewPoint_;
};

D5_TO_JSON(Weather, j, w) {
  j = Json{{"airTemp", w.airTemperature},
           {"airPressureMbar", w.airPressureMbar},
           {"relativeHumidity", w.relativeHumidity},
           {"windSpeedMS", w.windSpeedMS},
           {"directInsolationWM2", w.directInsolationWM2},
           {"diffuseInsolationWM2", w.diffuseInsolationWM2}};
  if (w.hasMeasuredDewPoint())
    j["dewPoint"] = w.dewPoint();
  // only write the wind direction if it's not the default
  if (w.windDirectionRad != Weather::defaultWindDirectionRad) {
    j["windDirection"] = w.windDirectionRad;
  }
}

D5_FROM_JSON(Weather, j, w) {
  j.at("airTemp").get_to(w.airTemperature);
  j.at("airPressureMbar").get_to(w.airPressureMbar);
  j.at("relativeHumidity").get_to(w.relativeHumidity);
  j.at("windSpeedMS").get_to(w.windSpeedMS);
  j.at("directInsolationWM2").get_to(w.directInsolationWM2);
  j.at("diffuseInsolationWM2").get_to(w.diffuseInsolationWM2);
  auto it = j.find("dewPoint");
  if (it != j.end()) {
    w.setDewPoint(it.value().get<Temperature>());
  }
  it = j.find("windDirection");
  if (it != j.end()) {
    it.value().get_to(w.windDirectionRad);
  }
}

/// \}

} // namespace d5
