/** Integration functions
 *
 *  \file Integration.h
 */

#pragma once

#include <utility>
#include <vector>

namespace d5 {

inline namespace math {

/// \addtogroup common_math
/// \{

/** Integration function for ordered paired data. This assumes the function
 *  being integrated has an integral from -infinity to infinity of 1.
 *
 *  \param sampledData  Ordered pair data. Should be sorted, but this is not
 *                      checked.
 *  \param x1           Lower bound of integration
 *  \param x2           Upper bound of integration
 *
 *  \return The numerical integral
 */
template <typename T>
T IntegrateSampledKernelFunction(
    const std::vector<std::pair<double, T>> &sampledData, double x1,
    double x2) {
  // check for valid domain
  if (x2 < x1)
    return 0.0;

  // grab limits
  auto lowerLimit = sampledData.front().first;
  auto upperLimit = sampledData.back().first;

  // check early exits
  if (x2 <= lowerLimit)
    return 0.0;
  if (x1 >= upperLimit)
    return 0.0;
  if (x1 < lowerLimit && x2 > upperLimit)
    return 1.0;

  // indexes surrounding the endpoints
  std::size_t index1a = 0, index1b = 0, index2a = 0, index2b = 0;

  // brute force the search
  std::size_t ii = 1;
  bool x1Set = false;
  for (; ii < sampledData.size(); ii++) {
    if (!x1Set) {
      if (x1 < sampledData[ii].first) {
        index1a = ii - 1;
        index1b = ii;
        x1Set = true;
      }
    }

    if (x2 < sampledData[ii].first) {
      index2a = ii - 1;
      index2b = ii;
      break;
    }

    // check for integral beyond end of function
    if (ii == (sampledData.size() - 1)) {
      index2a = ii - 1;
      index2b = ii;
    }
  }

  if (x1 < lowerLimit)
    x1 = lowerLimit;
  if (x2 > upperLimit)
    x2 = upperLimit;

  // integrate full sections between the endpoints
  double integral = 0;
  for (size_t ii = index1b; ii < index2a; ii++) {
    // assume linearity between sample points
    double width = sampledData[ii + 1].first - sampledData[ii].first;
    double height = 0.5 * (sampledData[ii + 1].second + sampledData[ii].second);
    integral += width * height;
  }

  // add in the endcaps
  double fraction1 = (x1 - sampledData[index1a].first) /
                     (sampledData[index1b].first - sampledData[index1a].first);
  double fraction2 = (x2 - sampledData[index2a].first) /
                     (sampledData[index2b].first - sampledData[index2a].first);

  // look for overlap
  if (index1a == index2a) {
    double fraction = 0.5 * (fraction2 + fraction1);
    double width = x2 - x1;
    double height = (1.0 - fraction) * sampledData[index1a].second +
                    fraction * sampledData[index1b].second;
    integral += width * height;
  } else {
    double width1 = sampledData[index1b].first -
                    (fraction1 * sampledData[index1b].first +
                     (1 - fraction1) * sampledData[index1a].first);
    double height1 = 0.5 * ((1 + fraction1) * sampledData[index1b].second +
                            (1 - fraction1) * sampledData[index1a].second);
    integral += width1 * height1;

    double width2 = (fraction2 * sampledData[index2b].first +
                     (1 - fraction2) * sampledData[index2a].first) -
                    sampledData[index2a].first;
    double height2 = 0.5 * (fraction2 * sampledData[index2b].second +
                            (2 - fraction2) * sampledData[index2a].second);
    integral += width2 * height2;
  }

  return integral;
}

/// \}
} // namespace math

} // namespace d5
