/** Interpolation routines
 *
 *  \file Interpolation.h
 */

#pragma once

#include <vector>

namespace d5 {

inline namespace math {

/// \addtogroup common_math
/// \{

/**
 * How extrapolation case should be handled.
 * \ingroup math
 */
enum class ExtrapolationPolicy {
  Linear, ///< Keep doing linear with two end pts.
  Flat,   ///< Just fill using closest end point.
  Fill    ///< Fill using supplied fill-value.
};

template <typename T>
T LinearInterpolation(
    const std::vector<std::pair<double, T>> &sampledData, double x,
    ExtrapolationPolicy extrapolate = ExtrapolationPolicy::Fill,
    const T &fillValue = 0) {
  // grab limits
  double lowerLimit = sampledData.front().first;
  double upperLimit = sampledData.back().first;
  double epsilon = 1e-9 * (upperLimit - lowerLimit);
  lowerLimit -= epsilon;
  upperLimit += epsilon;

  // check early exits
  if (x < lowerLimit) {
    if (extrapolate == ExtrapolationPolicy::Fill) {
      return fillValue;
    } else if (extrapolate == ExtrapolationPolicy::Flat) {
      return sampledData.front().second;
    } else {
      // "extrapolate" a single point
      if (sampledData.size() == 1) {
        return sampledData[0].second;
      }

      T m = (sampledData[1].second - sampledData[0].second) /
            (sampledData[1].first - sampledData[0].first);
      T b = sampledData[0].second - m * sampledData[0].first;

      return m * x + b;
    }
  }

  if (x > upperLimit) {
    if (extrapolate == ExtrapolationPolicy::Fill) {
      return fillValue;
    } else if (extrapolate == ExtrapolationPolicy::Flat) {
      return sampledData.back().second;
    } else {
      // "extrapolate" a single point
      if (sampledData.size() == 1) {
        return sampledData[0].second;
      }

      size_t v = sampledData.size() - 1;
      size_t u = v - 1;
      T m = (sampledData[v].second - sampledData[u].second) /
            (sampledData[v].first - sampledData[u].first);
      T b = sampledData[v].second - m * sampledData[v].first;

      return m * x + b;
    }
  }

  // brute force the search
  size_t ii = 1;
  for (; ii < sampledData.size(); ii++) {
    if (x <= sampledData[ii].first) {
      break;
    }
  }

  if (ii >= sampledData.size())
    return sampledData.back().second;

  size_t upperIndex = ii;
  size_t lowerIndex = ii - 1;

  double fraction =
      (x - sampledData[lowerIndex].first) /
      (sampledData[upperIndex].first - sampledData[lowerIndex].first);
  if (fraction < 0)
    fraction = 0;

  return sampledData[lowerIndex].second +
         fraction *
             (sampledData[upperIndex].second - sampledData[lowerIndex].second);
}

/// \}
} // namespace math

} // namespace d5
