/** Interface file for kernel functions
 *
 *  \file KernelFunctions
 */

#pragma once

#include <algorithm>
#include <cmath>
#include <stdexcept>
#include <utility>
#include <vector>

#include "Integration.h"
#include "Interpolation.h"
#include "dirsig5/base/Constants.h"

namespace d5 {

inline namespace math {

/** \defgroup common_math_kernels Kernel functions
 *  \ingroup common_math
 */

/// \addtogroup common_math_kernels
/// \{

/** Namespace for kernel functions.
 *
 *  Kernel functions are one-dimensional functions with limited (practical)
 *  support (kernels). This namespace supports evaluation of the function at a
 *  given independent value and integration. The functions are always normalized
 *  so that the total integral is one.
 *
 *  These functions are expected to be used as std::function callbacks with
 *  bound parameters.
 */
namespace kernels {

/** The Epanechnikov kernel is a parabolic shape that resembles an arc of a
 *  ellipse between its limits, though the ends straighten out a bit.
 */
class Epanechnikov {
public:
  /// Evaluates the kernel at the given value
  [[nodiscard]] constexpr double evaluate(double x) const {
    // convert to the standard form
    double u = (x - center_) / half_;

    if (u > 1 || u < -1)
      return 0;

    constexpr double k = 0.75;

    return (1. / half_) * k * (1.0 - u * u);
  }

  /// Integrates the kernel between x1 and x2
  [[nodiscard]] constexpr double integrate(double x1, double x2) const {
    if (x2 <= x1)
      return 0.0;

    // convert to the standard form
    double u1 = std::max(-1.0, (x1 - center_) / half_);
    double u2 = std::min(+1.0, (x2 - center_) / half_);

    // make sure we didn't make this easy..
    if (u1 == u2)
      return 0.0;
    if (u2 < -1)
      return 0.0;
    if (u1 > +1)
      return 0.0;

    constexpr double k1 = -0.75;
    constexpr double k2 = +1.0 / 3.0;

    return k1 * (k2 * (u2 * u2 * u2 - u1 * u1 * u1) - (u2 - u1));
  }

  /// Gets the lower bound of the kernel
  [[nodiscard]] constexpr double getLower() const { return center_ - half_; }

  /// Gets the upper bound of the kernel
  [[nodiscard]] constexpr double getUpper() const { return center_ + half_; }

  /// Get the center of the kernel
  [[nodiscard]] constexpr double getCenter() const { return center_; }

  /// Get the FWHM of the kernel
  [[nodiscard]] constexpr double getFwhm() const { return 2.0 * half_; }

  /** Factory function
   *
   *  \param center  Center of the kernel
   *  \param half    Half-width of the kernel
   */
  static constexpr Epanechnikov fromCenterHalfWidth(double center,
                                                    double half) {
    return Epanechnikov(center, half);
  }

private:
  /// Constructor
  constexpr Epanechnikov(double c, double h) : center_(c), half_(h) {}

  /// Center of the function
  double center_ = 0;

  /// Half-width of the kernel (x of the 0)
  double half_ = 1;
};

/** The Gaussian kernel implements a normal distribution and is parameterized
 *  by the mean and standard deviation.
 */
class Gaussian {
public:
  /// Evaluate the kernel at x
  [[nodiscard]] double evaluate(double x) const {
    auto u = (x - mean_) / (SQRT_2 * sigma_);

    if (u > 7.109 || u < -7.109)
      return 0.0;

    constexpr double k = 1.0 / (SQRT_2 * SQRT_PI);

    return (k / sigma_) * exp(-u * u);
  }

  /// Integrate the kernel between x1 and x2
  [[nodiscard]] double integrate(double x1, double x2) const {
    if (x2 <= x1)
      return 0.0;

    // convert to the standard form
    auto u1 = std::max(-7.109, (x1 - mean_) / (d5::SQRT_2 * sigma_));
    auto u2 = std::min(+7.109, (x2 - mean_) / (d5::SQRT_2 * sigma_));

    // make sure we didn't make this easy..
    if (u1 == u2)
      return 0.0;

    // compute the lower cdf
    double result = 0.0;
    if (u1 < 0.0) {
      result += 0.5 * std::erf(-u1);
    } else {
      result -= 0.5 * std::erf(+u1);
    }

    // compute the upper cdf
    if (u1 < 0.0) {
      result -= 0.5 * std::erf(-u2);
    } else {
      result += 0.5 * std::erf(+u2);
    }

    return result;
  }

  /// Get the lower bound of the kernel
  [[nodiscard]] constexpr double getLower() const {
    return mean_ - 5.04 * sigma_;
  }

  /// Get the upper bound of the kernel
  [[nodiscard]] constexpr double getUpper() const {
    return mean_ + 5.04 * sigma_;
  }

  /// Get the center of the kernel
  [[nodiscard]] constexpr double getCenter() const { return mean_; }

  /// Get the FWHM of the kernel
  [[nodiscard]] constexpr double getFwhm() const { return sigma_ / 0.42466; }

  /// Factory function
  constexpr static Gaussian fromMeanStdDev(double mean, double sigma) {
    return Gaussian(mean, sigma);
  }
  constexpr static Gaussian fromFwhm(double mean, double fwhm) {
    return Gaussian(mean, FwhmToStdDev(fwhm));
  }

  /// Computes the standard deviation from the full-width half-max
  constexpr static double FwhmToStdDev(double fwhm) {
    // FWHM = 2sqrt(2 ln(2)) * sigma
    // sigma = FWHM / (2 * sqrt(2 * ln(2)))
    // sigma = 0.42466090014400953 * FWHM
    constexpr double k = 0.42466090014400953;
    return k * fwhm;
  }

private:
  /// Constructor
  constexpr Gaussian(double mean, double sigma) : mean_(mean), sigma_(sigma) {}

  /// Mean of the function
  double mean_;

  /// Standard deviation of the function
  double sigma_;
};

/** A Lorentizian kernel function is defined by the scale parameter (HWHM).
 *  We will arbitrarily set the limits of the Lorentzian to 1e6 times the scale
 *  parameter.
 */
class Lorentzian {
public:
  /// Evaluate the kernel at x
  [[nodiscard]] constexpr double evaluate(double x) const {
    // convert to the standard form
    double u = (x - center_);

    if (u > 1e6 || u < -1e6)
      return 0.0;

    return (hwhm_ / PI) * 1.0 / (u * u + hwhm_ * hwhm_);
  }

  /// Integrate the kernel between x1 and x2
  [[nodiscard]] double integrate(double x1, double x2) const {
    if (x2 <= x1)
      return 0.0;

    // convert to the standard form
    double u1 = std::max(-1e6, (x1 - center_) / hwhm_);
    double u2 = std::min(+1e6, (x2 - center_) / hwhm_);

    return IPI * (atan(u2) - atan(u1));
  }

  /// Get the lower bound of the kernel
  [[nodiscard]] constexpr double getLower() const {
    return center_ - hwhm_ * 1e6;
  }

  /// Get the upper bound of the kernel
  [[nodiscard]] constexpr double getUpper() const {
    return center_ + hwhm_ * 1e6;
  }

  /// Get the center of the kernel
  [[nodiscard]] constexpr double getCenter() const { return center_; }

  /// Get the FWHM of the kernel
  [[nodiscard]] constexpr double getFwhm() const { return 2.0 * hwhm_; }

  /// Factory function
  constexpr static Lorentzian fromMeanHWHM(double mean, double hwhm) {
    return Lorentzian(mean, hwhm);
  }

private:
  /// Constructor
  constexpr Lorentzian(double mean, double hwhm) : center_(mean), hwhm_(hwhm) {}

  /// Center of the kernel
  double center_;

  /// Scale factor (HWHM)
  double hwhm_;
};

/** The rectangular kernel has a constant value betweem ots limits.
 *
 *  Since we are dealing with a quantized system, we will not use the standard
 *  convention of setting the value equal to half the height at the two limits.
 */
class Rect {
public:
  /// Evaluate the function at x
  [[nodiscard]] constexpr double evaluate(double x) const {
    // convert to the standard form
    double u = (x - center_) / (2. * halfWidth_);

    if (u > 0.5 || u < -0.5)
      return 0;

    return 0.5 / halfWidth_;
  }

  /// Integrate the function between x1 and x2
  [[nodiscard]] constexpr double integrate(double x1, double x2) const {
    if (x2 <= x1)
      return 0.0;

    // convert to the standard form
    double u1 = std::min(1.0, std::max(-1.0, (x1 - center_) / halfWidth_));
    double u2 = std::max(-1.0, std::min(+1.0, (x2 - center_) / halfWidth_));

    return 0.5 * (u2 - u1);
  }

  /// Get the lower bound of the kernel
  [[nodiscard]] constexpr double getLower() const {
    return center_ - halfWidth_;
  }

  /// Get the upper bound of the kernel
  [[nodiscard]] constexpr double getUpper() const {
    return center_ + halfWidth_;
  }

  /// Get the center of the kernel
  [[nodiscard]] constexpr double getCenter() const { return center_; }

  /// Get the FWHM of the kernel
  [[nodiscard]] constexpr double getFwhm() const { return 2.0 * halfWidth_; }

  /// Factory function
  constexpr static Rect fromCenterHalfWidth(double c, double halfWidth) {
    return Rect(c, halfWidth);
  }

private:
  /// Constructor
  constexpr Rect(double c, double h) : center_(c), halfWidth_(h) {}

  /// Center of the function
  double center_;

  /// Half-width of the function
  double halfWidth_;
};

/** The Sinc kernel is defined as sin(pi*x)/(pi*x).
 *
 *  The integral is solved via a Pade approximant or u <= 4 and the remainder of
 *  the integral (about 0.025) is "stuffed in" beyond that (this is also
 *  mirrored in the negative direction).
 */
class Sinc {
public:
  // constants for the pade approximant of the sine integral
  static constexpr double k1 = -4.54393409816329991e-2;
  static constexpr double k2 = +1.15457225751016682e-3;
  static constexpr double k3 = -1.41018536821330254e-5;
  static constexpr double k4 = +9.43280809438713025e-8;
  static constexpr double k5 = -3.53201978997168357e-10;
  static constexpr double k6 = +7.08240282274875911e-13;
  static constexpr double k7 = -6.05338212010422477e-16;
  static constexpr double k8 = +1.01162145739225565e-2;
  static constexpr double k9 = +4.99175116169755106e-5;
  static constexpr double k10 = +1.55654986308745614e-7;
  static constexpr double k11 = +3.28067571055789734e-10;
  static constexpr double k12 = +4.5049097575386581e-13;
  static constexpr double k13 = +3.21107051193712168e-16;
  static constexpr double f1 = 7.44437068161936700618e2;
  static constexpr double f2 = 1.96396372895146869801e5;
  static constexpr double f3 = 2.37750310125431834034e7;
  static constexpr double f4 = 1.43073403821274636888e9;
  static constexpr double f5 = 4.33736238870432522765e10;
  static constexpr double f6 = 6.40533830574022022911e11;
  static constexpr double f7 = 4.20968180571076940208e12;
  static constexpr double f8 = 1.00795182980368574617e13;
  static constexpr double f9 = 4.94816688199951963482e12;
  static constexpr double f10 = -4.94701168645415959931e11;
  static constexpr double f11 = 7.46437068161927678031e2;
  static constexpr double f12 = 1.97865247031583951450e5;
  static constexpr double f13 = 2.41535670165126845144e7;
  static constexpr double f14 = 1.47478952192985464958e9;
  static constexpr double f15 = 4.58595115847765779830e10;
  static constexpr double f16 = 7.08501308149515401563e11;
  static constexpr double f17 = 5.06084464593475076774e12;
  static constexpr double f18 = 1.43468549171581016479e13;
  static constexpr double f19 = 1.11535493509914254097e13;
  static constexpr double g1 = 8.1359520115168615e2;
  static constexpr double g2 = 2.35239181626478200e5;
  static constexpr double g3 = 3.12557570795778731e7;
  static constexpr double g4 = 2.06297595146763354e9;
  static constexpr double g5 = 6.83052205423625007e10;
  static constexpr double g6 = 1.09049528450362786e12;
  static constexpr double g7 = 7.57664583257834349e12;
  static constexpr double g8 = 1.81004487464664575e13;
  static constexpr double g9 = 6.43291613143049485e12;
  static constexpr double g10 = -1.36517137670871689e12;
  static constexpr double g11 = 8.19595201151451564e2;
  static constexpr double g12 = 2.40036752835578777e5;
  static constexpr double g13 = 3.26026661647090822e7;
  static constexpr double g14 = 2.23355543278099360e9;
  static constexpr double g15 = 7.87465017341829930e10;
  static constexpr double g16 = 1.39866710696414565e12;
  static constexpr double g17 = 1.17164723371736605e13;
  static constexpr double g18 = 4.01839087307656620e13;
  static constexpr double g19 = 3.99653257887490811e13;

  /// Integral of sin(x)/x from 0 to x
  [[nodiscard]] constexpr static double StandardIntegral0to(double x) {
    bool flip = false;
    if (x < 0) {
      x *= -1.0;
      flip = true;
    }

    double x_2 = x * x;
    double x_4 = x_2 * x_2;
    double x_6 = x_4 * x_2;
    double x_8 = x_4 * x_4;
    double x_10 = x_8 * x_2;
    double x_12 = x_8 * x_4;
    double x_14 = x_8 * x_6;

    double val = 0;

    if (x < 4.0) {
      val = x *
            (1. + k1 * x_2 + k2 * x_4 + k3 * x_6 + k4 * x_8 + k5 * x_10 +
             k6 * x_12 + k7 * x_14) /
            (1. + k8 * x_2 + k9 * x_4 + k10 * x_6 + k11 * x_8 + k12 * x_10 +
             k13 * x_12);
    } else {
      double x_16 = x_8 * x_8;
      double x_18 = x_8 * x_10;
      double x_20 = x_8 * x_12;

      double f =
          (1. / x) *
          (1. + f1 / x_2 + f2 / x_4 + f3 / x_6 + f4 / x_8 + f5 / x_10 +
           f6 / x_12 + f7 / x_14 + f8 / x_16 + f9 / x_18 + f10 / x_20) /
          (1. + f11 / x_2 + f12 / x_4 + f13 / x_6 + f14 / x_8 + f15 / x_10 +
           f16 / x_12 + f17 / x_14 + f18 / x_16 + f19 / x_18);

      auto g = (1. / x_2) *
               (1. + g1 / x_2 + g2 / x_4 + g3 / x_6 + g4 / x_8 + g5 / x_10 +
                g6 / x_12 + g7 / x_14 + g8 / x_16 + g9 / x_18 + g10 / x_20) /
               (1. + g11 / x_2 + g12 / x_4 + g13 / x_6 + g14 / x_8 +
                g15 / x_10 + g16 / x_12 + g17 / x_14 + g18 / x_16 + g19 / x_18);

      val = 0.5 * PI - f * cos(x) - g * sin(x);
    }

    if (flip)
      val *= -1.0;
    return val;
  }

  /// Evaluate the function at x
  [[nodiscard]] constexpr double evaluate(double x) const {
    // convert to the standard form
    double u = PI * (x - center_) / halfWidth_;

    if (u == 0)
      return 1.0 / halfWidth_;

    return sin(u) / (halfWidth_ * u);
  }

  /// Integrate the function between x1 and x2
  [[nodiscard]] constexpr double integrate(double x1, double x2) const {
    if (x2 <= x1)
      return 0.0;

    // convert to the standard form
    double u1 = PI * (x1 - center_) / halfWidth_;
    double u2 = PI * (x2 - center_) / halfWidth_;

    return (StandardIntegral0to(u2) - StandardIntegral0to(u1)) / PI;
  }

  /// Get the lower bound of the kernel
  [[nodiscard]] constexpr double getLower() const {
    return center_ - 73.55 * halfWidth_;
  }

  /// Get the upper bound of the kernel
  [[nodiscard]] constexpr double getUpper() const {
    return center_ + 73.55 * halfWidth_;
  }

  /// Get the center of the kernel
  [[nodiscard]] constexpr double getCenter() const { return center_; }

  /// Get the FWHM of the kernel
  [[nodiscard]] constexpr double getFwhm() const { return 2.0 * halfWidth_; }

  /// Factory function
  constexpr static Sinc fromCenterHalfWidth(double c, double h) {
    return Sinc(c, h);
  }

private:
  /// Constructor
  constexpr Sinc(double c, double h) : center_(c), halfWidth_(h) {}

  /// Center of the function
  double center_;

  /// Half-width of the function (x of first zero)
  double halfWidth_;
};

/// A SincSquared is simply the square of a Sinc
class SincSquared {
public:
  /// Evaluate the function at x
  [[nodiscard]] constexpr double evaluate(double x) const {
    auto y = Sinc::fromCenterHalfWidth(center_, halfWidth_).evaluate(x);
    return halfWidth_ * y * y;
  }

  /// Integrate the function between x1 and x2
  [[nodiscard]] double integrate(double x1, double x2) const {
    double u1 = (x1 - center_) / halfWidth_;
    double u2 = (x2 - center_) / halfWidth_;

    double sina = sin(PI * u1);
    double sinb = sin(PI * u2);

    return (-PI * Sinc::StandardIntegral0to(2.0 * u1 * PI) + sina * sina / u1 +
            PI * Sinc::StandardIntegral0to(2.0 * u2 * PI) - sinb * sinb / u2) /
           (PI * PI);
  }

  /// Get the lower bound of the kernel
  [[nodiscard]] constexpr double getLower() const {
    return center_ - 99.295 * halfWidth_;
  }

  /// Get the upper bound of the kernel
  [[nodiscard]] constexpr double getUpper() const {
    return center_ + 99.295 * halfWidth_;
  }

  /// Get the center of the kernel
  [[nodiscard]] constexpr double getCenter() const { return center_; }

  /// Get the FWHM of the kernel
  [[nodiscard]] constexpr double getFwhm() const { return 2.0 * halfWidth_; }

  /// Factory function
  constexpr static SincSquared fromCenterHalfWidth(double c, double h) {
    return SincSquared(c, h);
  }

private:
  /// Constructor
  constexpr SincSquared(double c, double h) : center_(c), halfWidth_(h) {}

  /// Center of the function
  double center_;

  /// Half-width of the function (x of first zero)
  double halfWidth_;
};

/** A SuperGaussian is a Gaussian function raised to a power
 *
 *  It is parameterized by the center and standard deviation of the underlying
 *  Gaussian and the power.
 */
class SuperGaussian {
public:
  /// Evaluate the function at x
  [[nodiscard]] double evaluate(double x) const {
    return SuperGaussian::eval(x, center_, stdDev_, order_);
  }

  /// Integrate the function from x1 to x2
  [[nodiscard]] double integrate(double x1, double x2) const {
    if (x2 <= x1)
      return 0.0;

    double integral = 0.0;

    // convert to the standard form
    double u1 = std::max(-1.5, (x1 - center_) / stdDev_);
    double u2 = std::min(+1.5, (x2 - center_) / stdDev_);

    if (u1 < -1.49999999 && u2 > 0.) {
      u1 = 0.0;
      integral += 0.5;
    }

    if (u2 > +1.499999999 && u1 < 0.) {
      u2 = 0.0;
      integral += 0.5;
    }

    // brute force the remainder
    const double step = (u2 - u1) / 100.;
    double u = 0;
    for (u = u1 + 0.5 * step; u < u2; u += step) {
      integral += eval(u, 0.0, 1.0, order_) * step;
    }

    return integral;
  }

  /// Get the lower bound of the kernel
  [[nodiscard]] constexpr double getLower() const {
    return center_ - 1.18 * stdDev_;
  }

  /// Get the upper bound of the kernel
  [[nodiscard]] constexpr double getUpper() const {
    return center_ + 1.18 * stdDev_;
  }

  /// Get the center of the kernel
  [[nodiscard]] constexpr double getCenter() const { return center_; }

  /// Get the FWHM of the kernel
  [[nodiscard]] constexpr double getFwhm() const { return 2.0 * stdDev_; }

  /// Factory function
  constexpr static SuperGaussian
  fromCenterStdDevOrder(double center, double stdDev, int order) {
    if (order <= 2 || order % 2 != 0) {
      throw std::runtime_error(
          "The power on a super Gaussian must be an even integer > 2");
    }
    return SuperGaussian(center, stdDev, order);
  }

private:
  /// Generic evaluation (since integral needs to use it)
  static double eval(double x, double center = 0.0, double sigma = 1.0,
                     int order = 10) {
    // convert to the standard form
    double u = fabs((x - center) / (sigma));

    if (u > 1.5)
      return 0.0;

    double u2 = u * u;
    double un = u2;

    for (int oi = 1; oi < order; oi++) {
      un *= u2;
    }

    double value = exp(-2.0 * un);

    double n = order;

    // fold in normalization
    double normalization =
        1. / (sigma * pow(2.0, ((n - 1) / n)) * std::tgamma((n + 1) / n));
    value *= normalization;

    return value;
  }

  /// Constructor
  constexpr SuperGaussian(double c, double s, int o)
      : center_(c), stdDev_(s), order_(o) {}

  /// Center of the function
  double center_;

  /// Standard deviation of the underlying Gaussian
  double stdDev_;

  /// Exponent for the Gaussian
  int order_;
};

/** The Tri function is a simple function that has a maximum at the center
 *  and linearly falls off towards the bounds in either direction
 */
class Tri {
public:
  /// Evaluate the function at x
  [[nodiscard]] double evaluate(double x) const {
    // convert to the standard form
    auto u = std::abs((x - center_) / (2. * halfWidth_));

    if (u > 0.5)
      return 0.0;

    auto h = (1.0 / halfWidth_);
    return h - 2 * h * u;
  }

  [[nodiscard]] constexpr double integrate(double x1, double x2) const {
    auto u1 = std::max(-1.0, (x1 - center_) / halfWidth_);
    auto u2 = std::min(+1.0, (x2 - center_) / halfWidth_);
    if (u1 >= +1.0)
      return 0.0;
    if (u2 <= -1.0)
      return 0.0;

    auto iu1 =
        (u1 >= 0.0) ? 0.5 + 0.5 * (2.0 - u1) * u1 : 0.5 + 0.5 * (2.0 + u1) * u1;
    auto iu2 =
        (u2 >= 0.0) ? 0.5 + 0.5 * (2.0 - u2) * u2 : 0.5 + 0.5 * (2.0 + u2) * u2;

    return iu2 - iu1;
  }

  /// Factory function
  constexpr static Tri fromCenterHalfWidth(double c, double h) {
    return Tri(c, h);
  }

  /// Get the lower bound of the kernel
  [[nodiscard]] constexpr double getLower() const {
    return center_ - halfWidth_;
  }

  /// Get the upper bound of the kernel
  [[nodiscard]] constexpr double getUpper() const {
    return center_ + halfWidth_;
  }

  /// Get the center of the kernel
  [[nodiscard]] constexpr double getCenter() const { return center_; }

  /// Get the FWHM of the kernel
  [[nodiscard]] constexpr double getFwhm() const { return halfWidth_; }

private:
  /// Constructor
  constexpr Tri(double c, double h) : center_(c), halfWidth_(h) {}

  /// Center of the function
  double center_;

  /// Half-width of the function (location of zero)
  double halfWidth_;
};

/// Custom kernel function via sampled data
class SampledData {
public:
  /**
   * Set samples. Returns the integral before normalizing.
   */
  double setSamples(std::vector<std::pair<double, double>> samples) {
    // store the samples
    if (samples.empty()) {
      throw std::runtime_error("Zero samples provided!\n");
    }
    sampledData_ = std::move(samples);

    // sort the samples and get the lower and upper bounds
    std::sort(sampledData_.begin(), sampledData_.end());

    // look for the max value
    double maxValue = sampledData_.front().second;
    for (size_t ii = 1; ii < sampledData_.size(); ii++) {
      if (maxValue < sampledData_[ii].second) {
        maxValue = sampledData_[ii].second;
      }
    }
    double halfMaxValue = maxValue / 2.0;

    // look for the lower half-max point
    double lowerWavelength = sampledData_.front().first;
    for (auto iter = begin(sampledData_); iter != end(sampledData_); ++iter) {
      if (iter->second > halfMaxValue) {
        lowerWavelength = iter->first;
        break;
      }
    }

    // look for the upper half-max point
    double upperWavelength = sampledData_.back().first;
    for (auto iter = rbegin(sampledData_); iter != rend(sampledData_); ++iter) {
      if (iter->second > halfMaxValue) {
        upperWavelength = iter->first;
        break;
      }
    }

    // compute the center and FWHM
    fwhm_ = upperWavelength - lowerWavelength;
    center_ = lowerWavelength + fwhm_ / 2.0;

    // set the lower and upper limits from the data
    lowerLimit_ = sampledData_.front().first;
    upperLimit_ = sampledData_.back().first;

    // get the initial integral
    double integral = integrate(lowerLimit_, upperLimit_);
    if (integral <= 0.0) {
      throw std::runtime_error("Integral is not positive!\n");
    }

    // normalize the function and get the maximum value
    maxValue_ = -1e99;
    for (auto &xy : sampledData_) {
      xy.second /= integral;
      if (maxValue_ < xy.second) {
        maxValue_ = xy.second;
      }
    }

    return integral;
  }

  /// Get the raw samples
  [[nodiscard]] const std::vector<std::pair<double, double>> &
  getSamples() const {
    return this->sampledData_;
  }

  /// Evaluate the function at x
  [[nodiscard]] double evaluate(double x) const {
    return LinearInterpolation(sampledData_, x);
  }

  /// Integrate the function from x1 to x2
  [[nodiscard]] double integrate(double x1, double x2) const {
    return IntegrateSampledKernelFunction(sampledData_, x1, x2);
  }

  /// Get the lower limit of the function
  [[nodiscard]] double getLower() const { return lowerLimit_; }

  /// Get the upper limit of the function
  [[nodiscard]] double getUpper() const { return upperLimit_; }

  /// Get the center of the kernel
  [[nodiscard]] constexpr double getCenter() const { return center_; }

  /// Get the FWHM of the kernel
  [[nodiscard]] constexpr double getFwhm() const { return fwhm_; }

private:
  /// The sampled data (normalized)
  std::vector<std::pair<double, double>> sampledData_;

  /// The lower limit of the function
  double lowerLimit_;

  /// The upper limit of the function
  double upperLimit_;

  /// The center of the function
  double center_;

  /// The full-wdith at half max (FWHM) of the function
  double fwhm_;

  /// The maximum value of the function
  double maxValue_;
};

} // namespace kernels
/// \}
} // namespace math

} // namespace d5
