// -*- C++ -*-
//===------------------------------ span ---------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===---------------------------------------------------------------------===//

#pragma once

#include <array>    // for array
#include <cstddef>  // for ptrdiff_t
#include <cstddef>  // for byte
#include <iterator> // for iterators
#include <limits>
#include <type_traits> // for remove_cv, etc

namespace d5 {

/** \defgroup common_span C++20 span<T> Backport
 *  A view to a strongly-typed sequential block in memory
 *  \ingroup common
 */

/// \addtogroup common_span
/// \{

#ifdef __cpp_inline_variables
inline constexpr size_t dynamic_extent = std::numeric_limits<size_t>::max();
#else
constexpr size_t dynamic_extent = std::numeric_limits<size_t>::max();
#endif

template <typename _Tp, size_t _Extent = dynamic_extent> class span;

template <class _Tp> struct __is_span_impl : public std::false_type {};

template <class _Tp, size_t _Extent>
struct __is_span_impl<span<_Tp, _Extent>> : public std::true_type {};

template <class _Tp>
struct __is_span : public __is_span_impl<typename std::remove_cv<_Tp>::type> {};

template <class _Tp> struct __is_std_array_impl : public std::false_type {};

template <class _Tp, size_t _Sz>
struct __is_std_array_impl<std::array<_Tp, _Sz>> : public std::true_type {};

template <class _Tp>
struct __is_std_array
    : public __is_std_array_impl<typename std::remove_cv<_Tp>::type> {};

template <class _Tp, class _ElementType, class = void>
struct __is_span_compatible_container : public std::false_type {};

#ifdef __cpp_lib_void_t
template <class... Args> using VoidT = std::void_t<Args...>;
#else
template <class...> using VoidT = void;
#endif

template <class _Tp, class _ElementType>
struct __is_span_compatible_container<
    _Tp, _ElementType,
    VoidT<
        // is not a specialization of span
        typename std::enable_if<!__is_span<_Tp>::value, std::nullptr_t>::type,
        // is not a specialization of array
        typename std::enable_if<!__is_std_array<_Tp>::value,
                                std::nullptr_t>::type,
        // is_array_v<Container> is false,
        typename std::enable_if<!std::is_array<_Tp>::value,
                                std::nullptr_t>::type,
// data(cont) and size(cont) are well formed
#if __cplusplus >= 201703L
        decltype(std::data(std::declval<_Tp>())),
        decltype(std::size(std::declval<_Tp>())),
#else
        decltype(std::declval<_Tp>().data()),
        decltype(std::declval<_Tp>().size()),
#endif
// remove_pointer_t<decltype(data(cont))>(*)[] is convertible to
// ElementType(*)[]
#if __cplusplus >= 201703L
        typename std::enable_if<
            std::is_convertible_v<std::remove_pointer_t<decltype(std::data(
                                      std::declval<_Tp &>()))> (*)[],
                                  _ElementType (*)[]>,
            std::nullptr_t>::type>
#else
        typename std::enable_if<
            std::is_convertible<
                typename std::remove_pointer<
                    decltype(std::declval<_Tp &>().data())>::type (*)[],
                _ElementType (*)[]>::value,
            std::nullptr_t>::type>
#endif
    > : public std::true_type {
};

template <typename _Tp, size_t _Extent> class span {
public:
  //  constants and types
  using element_type = _Tp;
  using value_type = typename std::remove_cv<_Tp>::type;
  using size_type = size_t;
  using difference_type = ptrdiff_t;
  using pointer = _Tp *;
  using const_pointer = const _Tp *;
  using reference = _Tp &;
  using const_reference = const _Tp &;
  using iterator = pointer;
  using const_iterator = const_pointer;
  using reverse_iterator = std::reverse_iterator<iterator>;
  using const_reverse_iterator = std::reverse_iterator<const_iterator>;

  static constexpr size_type extent = _Extent;

  // [span.cons], span constructors, copy, assignment, and destructor
  constexpr span() noexcept {
    static_assert(
        _Extent == 0,
        "Can't default construct a statically sized span with size > 0");
  }

  constexpr span(const span &) noexcept = default;

  constexpr span &operator=(const span &) noexcept = default;

  constexpr span(pointer __ptr, size_type /*__count*/) : __data{__ptr} {}
  constexpr span(pointer __f, pointer /*__l*/) : __data{__f} {}

  constexpr span(element_type (&__arr)[_Extent]) noexcept : __data{__arr} {}
  constexpr span(std::array<value_type, _Extent> &__arr) noexcept
      : __data{__arr.data()} {}
  constexpr span(const std::array<value_type, _Extent> &__arr) noexcept
      : __data{__arr.data()} {}

  template <class _OtherElementType>
  constexpr span(
      const span<_OtherElementType, _Extent> &__other,
      typename std::enable_if<std::is_convertible<_OtherElementType (*)[],
                                                  element_type (*)[]>::value,
                              std::nullptr_t>::type = nullptr)
      : __data{__other.data()} {}

  template <class _OtherElementType>
  constexpr span(
      const span<_OtherElementType, dynamic_extent> &__other,
      typename std::enable_if<std::is_convertible<_OtherElementType (*)[],
                                                  element_type (*)[]>::value,
                              std::nullptr_t>::type = nullptr) noexcept
      : __data{__other.data()} {}

  //  ~span() noexcept = default;

  template <size_t _Count>
  constexpr span<element_type, _Count> first() const noexcept {
    static_assert(_Count <= _Extent, "Count out of range in span::first()");
    return {data(), _Count};
  }

  template <size_t _Count>
  constexpr span<element_type, _Count> last() const noexcept {
    static_assert(_Count <= _Extent, "Count out of range in span::last()");
    return {data() + size() - _Count, _Count};
  }

  constexpr span<element_type, dynamic_extent>
  first(size_type __count) const noexcept {
    return {data(), __count};
  }

  constexpr span<element_type, dynamic_extent>
  last(size_type __count) const noexcept {
    return {data() + size() - __count, __count};
  }

  template <size_t _Offset, size_t _Count = dynamic_extent>
  constexpr auto subspan() const noexcept
      -> span<element_type,
              _Count != dynamic_extent ? _Count : _Extent - _Offset> {
    static_assert(_Offset <= _Extent, "Offset out of range in span::subspan()");
    return {data() + _Offset,
            _Count == dynamic_extent ? size() - _Offset : _Count};
  }

  constexpr span<element_type, dynamic_extent>
  subspan(size_type __offset,
          size_type __count = dynamic_extent) const noexcept {
    if (__count == dynamic_extent)
      return {data() + __offset, size() - __offset};
    return {data() + __offset, __count};
  }

  [[nodiscard]] constexpr size_type size() const noexcept { return _Extent; }
  [[nodiscard]] constexpr size_type size_bytes() const noexcept {
    return _Extent * sizeof(element_type);
  }
  [[nodiscard]] constexpr bool empty() const noexcept { return _Extent == 0; }

  constexpr reference operator[](size_type __idx) const noexcept {
    return __data[__idx];
  }

  constexpr reference front() const noexcept { return __data[0]; }

  constexpr reference back() const noexcept { return __data[size() - 1]; }

  constexpr pointer data() const noexcept { return __data; }

  // [span.iter], span iterator support
  constexpr iterator begin() const noexcept { return iterator(data()); }
  constexpr iterator end() const noexcept { return iterator(data() + size()); }
  constexpr const_iterator cbegin() const noexcept {
    return const_iterator(data());
  }
  constexpr const_iterator cend() const noexcept {
    return const_iterator(data() + size());
  }
  constexpr reverse_iterator rbegin() const noexcept {
    return reverse_iterator(end());
  }
  constexpr reverse_iterator rend() const noexcept {
    return reverse_iterator(begin());
  }
  constexpr const_reverse_iterator crbegin() const noexcept {
    return const_reverse_iterator(cend());
  }
  constexpr const_reverse_iterator crend() const noexcept {
    return const_reverse_iterator(cbegin());
  }

  constexpr void swap(span &__other) noexcept {
    pointer __p = __data;
    __data = __other.__data;
    __other.__data = __p;
  }

private:
  pointer __data = nullptr;
};

template <typename _Tp> class span<_Tp, dynamic_extent> {
private:
public:
  //  constants and types
  using element_type = _Tp;
  using value_type = typename std::remove_cv<_Tp>::type;
  using size_type = size_t;
  using difference_type = ptrdiff_t;
  using pointer = _Tp *;
  using const_pointer = const _Tp *;
  using reference = _Tp &;
  using const_reference = const _Tp &;
  using iterator = pointer;
  using const_iterator = const_pointer;
  using reverse_iterator = std::reverse_iterator<iterator>;
  using const_reverse_iterator = std::reverse_iterator<const_iterator>;

  static constexpr size_type extent = dynamic_extent;

  // [span.cons], span constructors, copy, assignment, and destructor
  constexpr span() noexcept = default;

  constexpr span(const span &) noexcept = default;

  constexpr span &operator=(const span &) noexcept = default;

  constexpr span(pointer __ptr, size_type __count)
      : __data{__ptr}, __size{__count} {}
  constexpr span(pointer __f, pointer __l)
      : __data{__f}, __size{static_cast<size_t>(distance(__f, __l))} {}

  template <size_t _Sz>
  constexpr span(element_type (&__arr)[_Sz]) noexcept
      : __data{__arr}, __size{_Sz} {}

  template <size_t _Sz>
  constexpr span(std::array<value_type, _Sz> &__arr) noexcept
      : __data{__arr.data()}, __size{_Sz} {}

  template <size_t _Sz>
  constexpr span(const std::array<value_type, _Sz> &__arr) noexcept
      : __data{__arr.data()}, __size{_Sz} {}

  template <class _Container>
  constexpr span(_Container &__c,
                 typename std::enable_if<
                     __is_span_compatible_container<_Container, _Tp>::value,
                     std::nullptr_t>::type = nullptr)
#if __cplusplus >= 201703L
      : __data{std::data(__c)}, __size{std::size(__c)} {}
#else
      : __data{__c.data()}, __size{__c.size()} {
  }
#endif

        template <class _Container>
        constexpr span(
            const _Container &__c,
            typename std::enable_if<
                __is_span_compatible_container<const _Container, _Tp>::value,
                std::nullptr_t>::type = nullptr)
#if __cplusplus >= 201703L
      : __data{std::data(__c)}, __size{std::size(__c)} {}
#else
      : __data{__c.data()}, __size{__c.size()} {
  }
#endif

        template <class _OtherElementType, size_t _OtherExtent>
        constexpr span(const span<_OtherElementType, _OtherExtent> &__other,
                       typename std::enable_if<
                           std::is_convertible<_OtherElementType (*)[],
                                               element_type (*)[]>::value,
                           std::nullptr_t>::type = nullptr) noexcept
      : __data{__other.data()}, __size{__other.size()} {
  }

  //    ~span() noexcept = default;

  template <size_t _Count>
  constexpr span<element_type, _Count> first() const noexcept {
    return {data(), _Count};
  }

  template <size_t _Count>
  constexpr span<element_type, _Count> last() const noexcept {
    return {data() + size() - _Count, _Count};
  }

  [[nodiscard]] constexpr span<element_type, dynamic_extent>
  first(size_type __count) const noexcept {
    return {data(), __count};
  }

  [[nodiscard]] constexpr span<element_type, dynamic_extent>
  last(size_type __count) const noexcept {
    return {data() + size() - __count, __count};
  }

  template <size_t _Offset, size_t _Count = dynamic_extent>
  [[nodiscard]] constexpr span<_Tp, dynamic_extent> subspan() const noexcept {
    return {data() + _Offset,
            _Count == dynamic_extent ? size() - _Offset : _Count};
  }

  [[nodiscard]] constexpr span<element_type, dynamic_extent>
  subspan(size_type __offset,
          size_type __count = dynamic_extent) const noexcept {
    if (__count == dynamic_extent)
      return {data() + __offset, size() - __offset};
    return {data() + __offset, __count};
  }

  [[nodiscard]] constexpr size_type size() const noexcept { return __size; }
  [[nodiscard]] constexpr size_type size_bytes() const noexcept {
    return __size * sizeof(element_type);
  }
  [[nodiscard]] constexpr bool empty() const noexcept { return __size == 0; }

  constexpr reference operator[](size_type __idx) const noexcept {
    return __data[__idx];
  }

  [[nodiscard]] constexpr reference front() const noexcept { return __data[0]; }

  [[nodiscard]] constexpr reference back() const noexcept {
    return __data[size() - 1];
  }

  [[nodiscard]] constexpr pointer data() const noexcept { return __data; }

  // [span.iter], span iterator support
  [[nodiscard]] constexpr iterator begin() const noexcept {
    return iterator(data());
  }
  [[nodiscard]] constexpr iterator end() const noexcept {
    return iterator(data() + size());
  }
  [[nodiscard]] constexpr const_iterator cbegin() const noexcept {
    return const_iterator(data());
  }
  [[nodiscard]] constexpr const_iterator cend() const noexcept {
    return const_iterator(data() + size());
  }
  [[nodiscard]] constexpr reverse_iterator rbegin() const noexcept {
    return reverse_iterator(end());
  }
  [[nodiscard]] constexpr reverse_iterator rend() const noexcept {
    return reverse_iterator(begin());
  }
  [[nodiscard]] constexpr const_reverse_iterator crbegin() const noexcept {
    return const_reverse_iterator(cend());
  }
  [[nodiscard]] constexpr const_reverse_iterator crend() const noexcept {
    return const_reverse_iterator(cbegin());
  }

  constexpr void swap(span &__other) noexcept {
    pointer __p = __data;
    __data = __other.__data;
    __other.__data = __p;

    size_type __sz = __size;
    __size = __other.__size;
    __other.__size = __sz;
  }

private:
  pointer __data = nullptr;
  size_type __size = 0;
};

template <class _Tp, size_t _Extent>
constexpr void swap(span<_Tp, _Extent> &__lhs,
                    span<_Tp, _Extent> &__rhs) noexcept {
  __lhs.swap(__rhs);
}

#ifdef __cpp_deduction_guides

//  Deduction guides
template <class _Tp, size_t _Sz> span(_Tp (&)[_Sz]) -> span<_Tp, _Sz>;

template <class _Tp, size_t _Sz> span(std::array<_Tp, _Sz> &) -> span<_Tp, _Sz>;

template <class _Tp, size_t _Sz>
span(const std::array<_Tp, _Sz> &) -> span<const _Tp, _Sz>;

template <class _Container>
span(_Container &) -> span<typename _Container::value_type>;

template <class _Container>
span(const _Container &) -> span<const typename _Container::value_type>;

#endif

/// \}
} // namespace d5
