/** Interface file for numeric parsing
 *
 *  \file Numeric.h
 */

#pragma once

#include <climits>
#include <cstdlib>
#include <limits>
#include <stdexcept>
#include <string>
#include <string_view>
#include <type_traits>

namespace d5 {

/// \addtogroup common_string
/// \{

/// Constants for the outcomes of a numeric parse operation
enum class NumericParseStatus {
  Success = 0,
  Invalid,
  Underflow,
  Overflow,
  UnspecifiedError
};

namespace detail {

template <typename T>
inline NumericParseStatus ParseFloatingPoint(const char *str, char **str_end,
                                             T &val) {
  static_assert(std::is_floating_point_v<T>, "T must be a float type");
  using Type = std::remove_cv_t<T>;

  errno = 0;
  if constexpr (std::is_same_v<Type, float>) {
    val = strtof(str, str_end);
  } else if constexpr (std::is_same_v<Type, double>) {
    val = strtod(str, str_end);
  } else {
    val = strtold(str, str_end);
  }

  if (errno == 0 && str != *str_end)
    return NumericParseStatus::Success;
  if (str == *str_end)
    return NumericParseStatus::Invalid;
  if (errno == ERANGE)
    return NumericParseStatus::Overflow;
  return NumericParseStatus::UnspecifiedError;
}

template <typename T>
inline NumericParseStatus ParseUnsignedInt(const char *str, char **str_end,
                                           T &val) {
  static_assert(std::is_unsigned_v<T>, "T must be an unsigned type");
  using Type = std::remove_cv_t<T>;

  unsigned long long v;
  v = strtoull(str, str_end, 10);

  if (errno == 0 && str != *str_end) {
    if constexpr (std::is_same_v<Type, unsigned long long>) {
      val = v;
      return NumericParseStatus::Success;
    } else {
      if (v <= std::numeric_limits<Type>::max()) {
        val = static_cast<Type>(v);
        return NumericParseStatus::Success;
      } else {
        return NumericParseStatus::Overflow;
      }
    }
  }

  if (str == *str_end)
    return NumericParseStatus::Invalid;
  if (errno == ERANGE)
    return NumericParseStatus::Overflow;
  return NumericParseStatus::UnspecifiedError;
}

template <typename T>
inline NumericParseStatus ParseSignedInt(const char *str, char **str_end,
                                         T &val) {
  static_assert(std::is_signed_v<T>, "T must be a signed type");
  using Type = std::remove_cv_t<T>;

  long long v;
  v = strtoll(str, str_end, 10);

  if (errno == 0 && str != *str_end) {
    if constexpr (std::is_same_v<Type, long long>) {
      val = v;
      return NumericParseStatus::Success;
    } else {
      if (v <= std::numeric_limits<Type>::max() &&
          v >= std::numeric_limits<Type>::lowest()) {
        val = static_cast<Type>(v);
        return NumericParseStatus::Success;
      } else if (v > std::numeric_limits<Type>::max()) {
        return NumericParseStatus::Overflow;
      } else {
        return NumericParseStatus::Underflow;
      }
    }
  }

  if (str == *str_end)
    return NumericParseStatus::Invalid;
  if (errno == ERANGE) {
    return (v == LLONG_MAX) ? NumericParseStatus::Overflow
                            : NumericParseStatus::Underflow;
  }
  return NumericParseStatus::UnspecifiedError;
}
} // namespace detail

template <typename T>
inline NumericParseStatus ParseNumber(const char *str, char **str_end, T &val) {
  // Make sure we are a number and get the base type
  static_assert(std::is_arithmetic_v<T>, "T must be a numeric type");
  using Type = std::remove_cv_t<T>;

  errno = 0;

  // Floating point parse
  if constexpr (std::is_floating_point_v<Type>) {
    return detail::ParseFloatingPoint(str, str_end, val);
  } else if constexpr (std::is_unsigned_v<Type>) {
    return detail::ParseUnsignedInt(str, str_end, val);
  } else {
    return detail::ParseSignedInt(str, str_end, val);
  }
}

template <typename T> class NumericParseError : public std::runtime_error {
public:
  static_assert(std::is_arithmetic_v<T>, "T must be a numeric type");

  NumericParseError(NumericParseStatus status, std::string_view s)
      : std::runtime_error(message(status, s)) {}

private:
  static std::string typeName() {
    if constexpr (std::is_floating_point_v<T>) {
      return "floating point";
    } else if constexpr (std::is_unsigned_v<T>) {
      return "unsigned integral";
    } else {
      return "integral";
    }
  }

  static std::string message(NumericParseStatus status, std::string_view s) {
    switch (status) {
    case NumericParseStatus::Success:
      return "internal logic error: NumericParseError made with a successful "
             "status";
    case NumericParseStatus::Invalid:
      return "unable to parse " + typeName() + " value from \"" +
             std::string(s) + "\"";
    case NumericParseStatus::Underflow:
      return typeName() + " underflow when parsing \"" + std::string(s) + "\"";
    case NumericParseStatus::Overflow:
      return typeName() + " overflow when parsing \"" + std::string(s) + "\"";
    case NumericParseStatus::UnspecifiedError:
      return "unspecified error when parsing " + typeName() + " value from \"" +
             std::string(s) + "\"" + typeName();
    }
    return "unknown error while parsing \"" + std::string(s) + "\"";
  }
};

template <typename T> inline T ParseNumber(std::string_view s) {
  char *end = nullptr;
  T val;
  auto res = ParseNumber(s.data(), &end, val);
  if (res == NumericParseStatus::Success)
    return val;
  else
    throw NumericParseError<T>(res, s);
}

template <typename T> inline void ParseNumber(std::string_view s, T &val) {
  char *end = nullptr;
  auto res = ParseNumber(s.data(), &end, val);
  if (res != NumericParseStatus::Success)
    throw NumericParseError<T>(res, s);
}

/// \}

} // namespace d5
