#pragma once

#include <vector>

#include "detail/Hashable.h"
#include "dirsig5/brdf/Export.h"

/** \defgroup brdf BRDF Library
 *
 *  Provides a library of BRDF models for evaluation and optimization
 */

/// Main namespace for the BRDF library
namespace d5::brdf {

/// \addtogroup brdf
/// \{

/// Represents a spectral BRDF sample for a single set of angles
struct D5_BRDF_EXPORT BrdfSample {
  std::vector<double> brdf;     ///< BRDF values for the sample
  double viewZenithRad;         ///< Viewer zenith angle [radians]
  double viewAzimuthRad;        ///< Viewer azimuth angle [radians]
  double illuminantZenithRad;   ///< Illuminant zenith angle [radians]
  double illuminantAzimuthRad;  ///< Illuminant azimuth angle [radians]
};

/// Top-level abstraction for a BRDF model
class D5_BRDF_EXPORT BrdfModel : public detail::Hashable {
public:
  /** Constructor
   *
   *  \param wavelengthMicrons  The wavelength samples [microns]. Does not need
   *                            to be sorted, but will be sorted internally.
   */
  explicit BrdfModel(std::vector<double> wavelengthMicrons);

  /// Virtual destructor
  virtual ~BrdfModel() = default;

  /// Results of an optimize request
  enum OptimizeStatus {
    /// Optimization failure
    Failed = 0,

    /// Optimization was successful and converged
    Success,

    /// Optimization did not fail but did not converge
    DidNotConverge,

    /// Optimization is not supported for this type
    NotSupported
  };

  /// The wavelength samples in microns
  [[nodiscard]] const std::vector<double>& wavelengthMicrons() const {
    return wavelengthMicrons_;
  }

  /** Samples the BRDF at the given angular inputs.
   *
   *  \param illumZenithRad     Illuminant zenith angle to sample [radians]
   *  \param illumAzimuthRad    Illuminant azimuth angle to sample [radians]
   *  \param viewZenithRad      View zenith angle to sample [radians]
   *  \param viewAzimuthRad     View azimuth angle to samples [radians]
   *
   *  \return The BRDF value
   */
  [[nodiscard]] virtual std::vector<double> sample(
    double illumZenithRad,
    double illumAzimuthRad,
    double viewZenithRad,
    double viewAzimuthRad) const = 0;

  /**
   * Attempt to simplify the BRDF spectrally for efficient sampling, this is
   * done in three parts:
   *
   * 1) Shrink the list of wavelengths (in microns) to those that BRDF
   * parameters are actually known for. In most cases, the BRDF is only
   * described at a discrete set of wavelengths and interpolated for all others.
   * This function reduces \c wavelengthMicrons() to that list of known
   * wavelengths. Note that the total (integrated) reflectance may be known for
   * many more wavelengths than the parameters.
   *
   * 2) Re-intialize or otherwise modify the property so that spectral vectors
   * driving and produced by this model are sampled at the wavelengths given in
   * 1). Note that interpolating the previous state to the new one for each call
   * to the BRDF functions is not efficient.
   *
   * 3) Return the total, integrated reflectance for a nadir source (i.e.
   * positioned along [0,0,1] in surface local coordinates) as it would be for
   * the original spectral sampling.
   *
   * \param[out] nadirDHR  The DHR spectrum. The wavelengths for this spectrum
   *                       are \c wavelengthMicrons() prior to this call, so be
   *                       sure to have a copy of that still around, as it
   *                       changes as a result of this call.
   */
  void simplifySpectralBrdf(std::vector<double>& nadirDHR);

  /** Samples the BRDF at the given angular inputs. Samples are computed for
   *  every combination of the input angles.
   *
   *  \param illumZenithRad     Illuminant zenith angles to sample [radians]
   *  \param illumAzimuthRad    Illuminant azimuth angles to sample [radians]
   *  \param viewZenithRad      View zenith angles to sample [radians]
   *  \param viewAzimuthRad     View azimuth angles to samples [radians]
   *
   *  \return A list of the sampled values
   */
  [[nodiscard]] virtual std::vector<BrdfSample> sampleAngleCombinations(
    const std::vector<double>& illumZenithRad,
    const std::vector<double>& illumAzimuthRad,
    const std::vector<double>& viewZenithRad,
    const std::vector<double>& viewAzimuthRad) const;

  /** Optimize the BRDF to fit the given samples. The object will be changed
   *  if and only if the returned status is either Success or DidNotConverge.
   *
   *  \param wavelengthsMicrons  Wavelengths for the samples [microns]
   *  \param samples             The BRDF samples to which to fit
   *
   *  \return The result of the optimization
   */
  virtual OptimizeStatus optimize(
    const std::vector<double>& wavelengthsMicrons,
    const std::vector<BrdfSample>& samples) {
    (void)wavelengthsMicrons;
    (void)samples;
    return NotSupported;
  }

  void getHashInputs(const HashCallback& callback) const final;

private:
  /** Initialization method. This should be called directly from the subclass
   *  constructor and should be used to perform any non-inline initialization.
   *
   *  The method will also be used to re-initialize after a call to
   *  \c simplifySpectralBrdf().
   */
  virtual void init() {}

  /** Virtual helper function for simplifySpectralBrdf. This method should be
   *  overridden to perform the simplification and return the new set of
   *  wavelengths to be used, which will then be set in this class. If no
   *  simplification can be performed, just return an empty vector. The nadirDHR
   *  parameter should always be set, regardless of whether simplification could
   *  be performed.
   */
  [[nodiscard]] virtual std::vector<double> doSimplifySpectralBrdf(
    std::vector<double>& nadirDHR) = 0;

  /** Virtual helper function for getHashValues(). This method should be
   *  overridden to add type-specific input values to the hash inputs.
   */
  virtual void doGetHashInputs(const HashCallback& callback) const = 0;

private:
  /// The wavelengths to sample at [microns]
  std::vector<double> wavelengthMicrons_;
};

/// \}

}  // namespace d5::brdf
