#pragma once

#include <cmath>
#include <complex>
#include <vector>

#include "MuellerMatrix.h"
#include "dirsig5/brdf/Export.h"

namespace d5::brdf {

/** Contains theoretical Fresnel reflectance calculation in static methods
 *  \ingroup brdf
 */
class D5_BRDF_EXPORT FresnelReflectance {
public:
  /** Angles associated with BRDF calculation. Stashing them here keeps all the
   *  variables organized and saves us from computing them multiple times.
   */
  struct D5_BRDF_EXPORT BrdfAngles {
    BrdfAngles(double _thetaI, double _phiI, double _thetaR, double _phiR);

    double thetaI;
    double phiI;
    double thetaR;
    double phiR;
    double deltaPhi;
    double cosThetaI;
    double sinThetaI;
    double cosThetaR;
    double sinThetaR;
    double tanThetaI;
    double tanThetaR;
    double beta;
    double cosBeta;
    double sinBeta;
    double theta;
    double cosTheta;
    double tanTheta;
    double cosEtaI;
    double sinEtaI;
    double cosEtaR;
    double sinEtaR;
    double signEta;
    double sinEtaRsinEtaI;
    double cosEtaRcosEtaI;
    double sinEtaRcosEtaI;
    double cosEtaRsinEtaI;
  };

  /**
   * Microfacet Mueller Matrix Generator.
   *
   * The following function contains the microfacet Mueller matrix used to
   * polarize the Torrance-Sparrow microfacet BRDF as described by Priest and
   * Germer.  The parameters of the function (\f$\eta_i\f$, \f$\eta_r\f$, and
   * \f$\beta\f$) are defined by Priest and Germer as follows:  The parameter
   * \f$\eta_i\f$ and \f$\eta_r\f$ are rotation angles to and from the surface
   * and microfacet reference planes.  The parameter \f$\beta\f$ is the angle
   * of incidence upon the microfacet.
   *
   * This function is in part an implementation
   * of the various terms in the Mueller Matrix (\f$M_{j,k}\f$) defined in
   * Equation 15 from Priest and Germer.  The integer arguement specifies
   * which absolute index to use from the stored bandpass list.
   */
  [[nodiscard]] static std::vector<MuellerMatrix> computeReflectance(
    const std::vector<std::complex<double>>& nk, const BrdfAngles& angles);

  [[nodiscard]] static std::vector<MuellerMatrix> computeTransmission(
    const std::vector<std::complex<double>>& nk, const BrdfAngles& angles);

  static void compute(
    const std::vector<std::complex<double>>& nk,
    const BrdfAngles& angles,
    std::vector<MuellerMatrix>& r,
    std::vector<MuellerMatrix>& t);

private:
  /**
   * Compute the Etas.
   *
   * The calculation of \f$\eta_i\f$ and \f$\eta_r\f$ is slightly tricky in
   * that we need to protect against the in-plane scattering case (ie. the
   * surface normal and the microfacet normal lie in the same plane).  The
   * in-plane scattering case results in a \f$0 / 0\f$ singularity which in
   * reality evaluates to 1 using L'Hopital's rule.  In other words, \f$\eta_i
   * = \eta_r = 0\f$ for the in-plane scattering case.
   *
   * Here we simply calculate the cosine of \f$\eta_i\f$ and \f$\eta_r\f$
   * becuase this the natural form of the equation defining these two angles
   * and all subsequent calculations only require the cosines of \f$eta_i\f$
   * and \f$eta_r\f$ and this saves calls to the std::acos() and std::cos()
   * functions.
   *
   * In addition, we have to determine the sign of \f$\eta_i\f$ and
   * \f$\eta_r\f$ for use when calculating the sine of \f$\eta_i\f$ and
   * \f$\eta_r\f$.  This is accomplished by examining the value of
   * \f$\Delta\phi\f$.  If \f$\Delta\phi > 180\f$ then the sign of
   * \f$\eta_i\f$ and \f$\eta_r\f$ is negative.
   *
   * The formulas for computing \f$\cos(\eta_i)\f$ and \f$\cos(\eta_r)\f$ are
   * taken from Equation 19 in Preist and Germer:
   *
   * \f[
   *     \cos( \eta_i ) =
   *         \frac
   *             {\frac
   *                 {\cos( \theta_i ) + \cos( \theta_r )}
   *                 {2\cos( \beta )}
   *             - \cos( \theta_i) \cos( \beta )
   *             }
   *             {\sin( \theta_i) \sin( \beta ) }
   * \f]
   * \f[
   *     \cos( \eta_r ) =
   *         \frac
   *             {\frac
   *                 {\cos( \theta_i ) + \cos( \theta_r )}
   *                 {2\cos( \beta )}
   *             - \cos( \theta_r) \cos( \beta )
   *             }
   *             {\sin( \theta_r) \sin( \beta ) }
   * \f]
   *
   * \param[in,out] angles  The angles structure to process
   */
  static void computeEtas(BrdfAngles& angles);

  /**
   * Compute the cosine of theta.
   *
   * Computes the cosine of \f$\theta\f$ which is the zenith angle of the
   * microfacet.  This value is computed using Equation 12 in Priest and
   * Germer:
   *
   * \f[
   *     \cos( \theta ) = \frac
   *         {\cos( \theta_i ) + \cos( \theta_r )}
   *         {1 \cos( \beta )}
   * \f]
   *
   * \param cosThetaI  Cosine of the illuminant zenith angle
   * \param cosThetaR  Cosine of the view zenith angle
   * \param cosBeta    Cosine of incident angle onto the microfacet
   *
   * \return  The cosine of the microfacet zenith angle
   */
  static double computeCosTheta(
    double cosThetaI, double cosThetaR, double cosBeta) {
    double cosTheta = (cosThetaI + cosThetaR) / (2.0 * cosBeta);
    if (cosTheta > 1.0) cosTheta = 1.0;
    if (cosTheta < -1.0) cosTheta = -1.0;
    return cosTheta;
  }

  /**
   * Compute beta.
   *
   * Computes the angle \f$\beta\f$ which is the angle of incidence upon the
   * specular microfacet.  This function computes \f$\beta\f$ by solving from
   * \f$\beta\f$ in Equation 12 in Priest and Germer:
   *
   *
   * \f[
   *     \cos( 2 \beta ) =
   *         \cos( \theta_i ) \cos( \theta_r) +
   *         \sin( \theta_i ) \sin( \theta_r ) \cos( \phi_i - \phi_r )
   * \f]
   *
   * \f[
   *     \beta = \frac
   *         { \cos^{-1} \left[
   *             \cos( \theta_i ) \cos( \theta_r) +
   *             \sin( \theta_i ) \sin( \theta_r ) \cos( \phi_i - \phi_r )
   *             \right ] }
   *         {2}
   * \f]
   *
   * \param cosThetaI Cosine of illuminant zenith angle
   * \param cosThetaR Cosine of view zenith angle
   * \param sinThetaI Sine of the illuminant zenith angle
   * \param sinThetaR Sine of the view zenith angle
   * \param deltaPhi  Difference in view and illuminant azimuth angles
   *
   * \return
   */
  static double computeBeta(
    double cosThetaI,
    double cosThetaR,
    double sinThetaI,
    double sinThetaR,
    double deltaPhi) {
    return 0.5 *
           acos(cosThetaI * cosThetaR + sinThetaI * sinThetaR * cos(deltaPhi));
  }

  /**
   * Parallel Reflection Coefficients.
   *
   * The Fresnel amplitude coefficient for reflection of light parallel to the
   * plane of incidence.  Some times this is refered to as \f$\rho_{\pi}\f$ or
   * \f$\rho_p\f$.
   *
   */
  [[nodiscard]] static std::complex<double> fresnelReflectionParallel(
    const std::complex<double>& nI,
    const std::complex<double>& nT,
    double cosThetaI,
    const std::complex<double>& sqrt1mbb);

  /**
   * Perpendicular Reflection Coefficients.
   *
   * Calculates the Fresnel amplitude coefficient for reflection of light
   * perpendicular to the plane of incidence.  Some times this is refered to
   * as \f$\rho_{\sigma}\f$ or \f$\rho_s\f$.
   *
   */
  [[nodiscard]] static std::complex<double> fresnelReflectionPerpendicular(
    const std::complex<double>& nI,
    const std::complex<double>& nT,
    double cosThetaI,
    const std::complex<double>& sqrt1mbb);
};

}  // namespace d5::brdf
