#pragma once

#include <algorithm>

#include "dirsig5/brdf/BrdfModel.h"

namespace d5::brdf {

class D5_BRDF_EXPORT Lambertian final : public BrdfModel {
public:
  struct Parameters {
    struct SpectralParameters {
      /// Wavelength in micrometers.
      double wavelengthMicrons = 0;

      /// Reflectance coefficient in [0, 1].
      double reflectance = 0.8;

      /// Transmittance coefficient in [0, 1].
      double transmittance = 0.0;
    };

    std::vector<SpectralParameters> spectral;
  };

  Lambertian(std::vector<double> wavelengthMicrons, Parameters params)
      : BrdfModel(std::move(wavelengthMicrons)), params_(std::move(params)) {
    std::sort(
      params_.spectral.begin(), params_.spectral.end(),
      [](const auto& lhs, const auto& rhs) {
        return lhs.wavelengthMicrons < rhs.wavelengthMicrons;
      });
    init();
  }

  [[nodiscard]] const Parameters& parameters() const { return params_; }

  [[nodiscard]] std::vector<double> sample(
    double illumZenithRad,
    double illumAzimuthRad,
    double viewZenithRad,
    double viewAzimuthRad) const override;

private:
  void init() override;

  [[nodiscard]] std::vector<double> doSimplifySpectralBrdf(
    std::vector<double>& nadirDHR) override;

  void doGetHashInputs(const HashCallback& callback) const override;

private:
  Parameters params_;

  /// Interpolated reflectance.
  std::vector<double> reflectanceOverPi_;

  /// Interpolated transmittance.
  std::vector<double> transmittanceOverPi_;
};

}  // namespace d5::brdf
