#pragma once

#include <cmath>

#include "FresnelReflectance.h"
#include "detail/Hashable.h"
#include "dirsig5/brdf/Export.h"

namespace d5::brdf {

/** \defgroup brdf_microfacetmodels Microfacet Models
 *  Common functions for microfacet-based BRDF models
 *  \ingroup brdf
 */

/// \addtogroup brdf_microfacetmodels
/// \{

/** Microfacet orientation probability functor interface.
 *
 *  The microfacet probability distribution function, \f$p(\theta_N)\f$, may be
 *  thought of as providing the "spread" of the specular microfacet reflections
 *  according to the surface roughness statistics.
 */
class D5_BRDF_EXPORT OrientationProbability : public detail::Hashable {
public:
  /// Virtual destructor
  virtual ~OrientationProbability() = default;

  /** Compute the orientation probability.
   *
   *  \param angles  Microfacet orientation angle relative to macrosurface
   *                 normal.
   *
   * \return The microfacet orientation probability for requested angle
   */
  [[nodiscard]] virtual double operator()(
    const FresnelReflectance::BrdfAngles& angles) const = 0;
};

/** Shadowing functor interface.
 *
 *  When the observer is not coincident with the incident irradiance, there are
 *  shadowing effects which modify the signature.  That is, some microfacet
 *  elements may not be illuminated due to others blocking the illumination
 *  source.  It is also possible that some illuminated microfacet elements are
 *  not visible by the observer due to line-of-sight blocking by microfacets.
 *  The net result is an attenuation of the specular reflectance due to what is
 *  collectively referred to as shadowing effects.
 */
class D5_BRDF_EXPORT ShadowingFunction : public detail::Hashable {
public:
  /// Virtual destructor
  virtual ~ShadowingFunction() = default;

  /** Compute the shadowing
   *
   * \param angles  Microfacet orientation angle relative to macrosurface
   *                normal.
   *
   * \return Value of the shadowing function (in the range [0,1]).
   */
  [[nodiscard]] virtual double operator()(
    const FresnelReflectance::BrdfAngles& angles) const = 0;
};

/// Volumetric reflectance functor interface
class D5_BRDF_EXPORT VolumetricTerm : public detail::Hashable {
public:
  /// Virtual destructor
  virtual ~VolumetricTerm() = default;

  /** Compute the volumetric reflectance
   *
   * \param angles  Source and view zenith angles
   *
   * \return  The volumetric reflectance component
   */
  [[nodiscard]] virtual double operator()(
    const FresnelReflectance::BrdfAngles& angles) const = 0;
};

/** Gaussian distribution for microfacet orientation.
 *
 *  The Gaussian microfacet probability distribution function
 *  \f$p_G(\theta_N)\f$ is that used by the CDPriestGermer model:
 *  \f[
 *    p_G(\theta_N) = \frac{B}{2 \pi \sigma^2 \cos^3(\theta_N)}
 *                    \mathrm{exp}\left[\frac{-\tan^2(\theta_N)}
 *                                           {2 \sigma^2} \right]
 *  \f]
 */
class D5_BRDF_EXPORT OPGaussian : public OrientationProbability {
public:
  /** Constructor
   *
   * \param bias    Bias parameter
   * \param stddev  Standard deviation
   */
  OPGaussian(double bias, double stddev) : B_(bias), var_(stddev * stddev) {}

  /// Virtual Destructor
  ~OPGaussian() override = default;

  [[nodiscard]] double operator()(
    const FresnelReflectance::BrdfAngles& angles) const override;

  void getHashInputs(const HashCallback& callback) const override;

private:
  /// Bias parameter (always 1 in Priest-Germer)
  double B_;

  /// Variance
  double var_;
};

/** Cauchy distribution for microfacet orientation
 *
 *  The modified Cauchy microfacet probability distribution function
 *  \f$p_C(\theta_N)\f$ is adapted from that used by the more recent versions of
 *  the NEF. It is given by:
 *
 *  \f[
 *    p_C(\theta_N) = \frac{B}{\cos^3(\theta_N) \left[\sigma^2 +
 *                    \tan^2(\theta_N)\right]}  \.
 *  \f]
 */
class D5_BRDF_EXPORT OPCauchy : public OrientationProbability {
public:
  /** Constructor
   *
   * \param bias    Bias parameter
   * \param stddev  Standard deviation
   */
  OPCauchy(double bias, double stddev) : B_(bias), var_(stddev * stddev) {}

  /// Virtual destructor
  ~OPCauchy() override = default;

  [[nodiscard]] double operator()(
    const FresnelReflectance::BrdfAngles& angles) const override;

  void getHashInputs(const HashCallback& callback) const override;

private:
  /// Bias parameter (always 1 in Priest-Germer)
  double B_;

  /// Variance
  double var_;
};

/** Maxwell-Beard shadowing function.  It is given by:
 *
 *  \f[
 *    S_{MB} = \frac{1+\frac{\theta_N}{\Omega} \
 *             e^{-2\beta/\tau}}{1+\frac{\theta_N}{\Omega}} \ ,
 *  \f]
 *
 *  where \f$\Omega\f$ and \f$\tau\f$ are the fit parameters.
 */
class D5_BRDF_EXPORT SFMaxwellBeard : public ShadowingFunction {
public:
  /** Constructor
   *
   *  \param omega  Omega fit parameter
   *  \param tau    Tau fit parameter
   */
  SFMaxwellBeard(double omega, double tau)
      : invOmega_(1.0 / omega), invTau_(-2.0 / tau) {}

  /// Virtual Destructor
  ~SFMaxwellBeard() override = default;

  [[nodiscard]] double operator()(
    const FresnelReflectance::BrdfAngles& angles) const override;

  void getHashInputs(const HashCallback& callback) const override;

private:
  /// Fit parameters
  double invOmega_, invTau_;
};

/** Sancer shadowing function.  It is given by:
 *
 *  \f[
 *    S_{Sancer} = \frac{1+\frac{\theta_N}{\Omega} \
 *             e^{-2\beta/\tau}}{1+\frac{\theta_N}{\Omega}} \ ,
 *  \f]
 *
 *  where \f$\sigma\f$ is the fit parameters.
 */
class D5_BRDF_EXPORT SFSancer : public ShadowingFunction {
public:
  /** Constructor
   *
   *  \param sigma  Sigma fit parameter
   */
  explicit SFSancer(double sigma) : sigma_(sigma) {}

  /// Virtual destructor
  ~SFSancer() override = default;

  [[nodiscard]] double operator()(
    const FresnelReflectance::BrdfAngles& angles) const override;

  void getHashInputs(const HashCallback& callback) const override;

private:
  /// Fit parameters
  double sigma_;
};

/** Maxwell-Beard functor for volumetric reflectance.
 *
 *  Only one representation of the volumetric component is currently
 *  being considered, that from the NEF Maxwell-Beard BRDF model. This
 *  term is (currently) completely randomly polarized and is given by
 *
 *  \f[
 *    V_{MB} = \rho_D + \frac{2 \rho_V}{\cos(\theta_i) +
 *             \cos(\theta_r)} \ ,
 *  \f]
 *
 *  where \f$\rho_D\f$ and \f$\rho_V\f$ are two free fit parameters.
 */
class D5_BRDF_EXPORT VTMaxwellBeard : public VolumetricTerm {
public:
  /** Constructor
   *
   *  \param rho_D  Rho_D fit parameter
   *  \param rho_V  Rho_V fit parameter
   */
  VTMaxwellBeard(double rho_D, double rho_V) : rho_D_(rho_D), rho_V_(rho_V) {}

  /// Virtual destructor
  ~VTMaxwellBeard() override = default;

  [[nodiscard]] double operator()(
    const FresnelReflectance::BrdfAngles& angles) const override;

  void getHashInputs(const HashCallback& callback) const override;

public:
  /// Fit parameters
  double rho_D_, rho_V_;
};

/// \}

}  // namespace d5::brdf
