#pragma once

#include "BrdfModel.h"
#include "FresnelReflectance.h"
#include "dirsig5/brdf/Export.h"

namespace d5::brdf {

/** Implementation of the Priest-Germer BRDF model
 *  \ingroup brdf
 */
class D5_BRDF_EXPORT PriestGermer : public BrdfModel {
public:
  struct Parameters {
    /// Diffuse renormalization?
    bool diffuseRenormalization = false;

    /// Spectrally-varying parameters.
    struct SpectralParameters {
      /// Wavelength in micrometers.
      double wavelengthMicrons = 0;

      /// Roughness \f$ sigma \f$.
      double sigma = 1;

      /// Complex index-of-refraction (IOR).
      std::complex<double> ior = {1, 0};
    };

    /// Spectrally-varying parameters.
    std::vector<SpectralParameters> spectral;
  };

  /// Constructor.
  PriestGermer(std::vector<double> wavelengthMicrons, Parameters params)
      : BrdfModel(std::move(wavelengthMicrons)), params_(std::move(params)) {
    // Sort spectral samples by wavelength.
    std::sort(
      params_.spectral.begin(), params_.spectral.end(),
      [](const auto& lhs, const auto& rhs) {
        return lhs.wavelengthMicrons < rhs.wavelengthMicrons;
      });
    // Delegate.
    init();
  }

  /// The parameters with which the model was constructed
  [[nodiscard]] const Parameters& parameters() const { return params_; }

  [[nodiscard]] std::vector<double> sample(
    double illumZenithRad,
    double illumAzimuthRad,
    double viewZenithRad,
    double viewAzimuthRad) const override;

private:
  void init() override;

  [[nodiscard]] std::vector<double> doSimplifySpectralBrdf(
    std::vector<double>& nadirDHR) override;

  void doGetHashInputs(const HashCallback& callback) const override;

private:
  /// Parameters.
  Parameters params_;

  /// Interpolated roughness \f$ \sigma \f$.
  std::vector<double> sigma_;

  /// Interpolated roughness, pre-computed term \f$ 1 / (2\sigma^2) \f$.
  std::vector<double> halfInvSigma2_;

  /// Interpolated complex index-of-refraction (IOR).
  std::vector<std::complex<double>> ior_;
};

}  // namespace d5::brdf
