#pragma once

#include <string>

#include "BrdfModel.h"
#include "dirsig5/brdf/Export.h"

namespace d5::brdf {

/// \addtogroup brdf
/// \{

/** An implementation of the Ross-Li BRDF, originally introduced in:
 *
 *  Wanner, W., X. Li, and A. H. Strahler (1995), On the derivation of kernels
 *  for kernel-driven models of bidirectional reflectance, J. Geophys. Res.,
 *  100(D10), 21077-21089.
 */
class D5_BRDF_EXPORT RossLi : public BrdfModel {
public:
  /// The types of Ross (volume-scattering) kernels
  enum class RossKernel {
    /// The thick (large LAI) kernel (equation 7)
    Thick,

    /// The thin (small LAI) kernel (equation 13)
    Thin
  };

  /// The types of Li (geometric optics) kernels
  enum class LiKernel {
    /// The sparse canopy kernel (equation 32)
    Sparse,

    /// The dense canopy kernel (equation 47)
    Dense
  };

  /// Input configuration parameters for a Ross-Li BRDF
  struct Parameters {
    /// The type of Ross kernel to use
    RossKernel rossKernel = RossKernel::Thick;

    /// The type of Li kernel to use
    LiKernel liKernel = LiKernel::Sparse;

    /** Whether to use the "reciprocal form" of the sparse Li kernel. This adds
     *  in an additional \f$\sec\theta'_i\f$ term to the last term in equation
     *  32. This is the more common form of the kernel used in MODIS products.
     */
    bool reciprocalLi = true;

    /// The length to width ratio of tree crowns
    double br = 1;

    /// The tree height to crown width ratio of trees
    double hb = 2;

    /// Parameters to the BRDF that vary spectrally
    struct SpectralParameters {
      /// The wavelength for the sample in microns
      double wavelengthMicrons;

      /// The diffuse portion of the BRDF (isotropic scattering)
      double isotropic;

      /// The weight on geometric-optics scattering (Li kernel)
      double geometric;

      /// The weight on the volume-scattering (Ross kernel)
      double volume;
    };

    /// Weights for each wavelength
    std::vector<SpectralParameters> spectral;
  };

  /** Constructor
   *
   *  \param wavelengthMicrons  The spectral sampling for the BRDF parameters
   *  \param params             The parameters for the BRDF
   */
  RossLi(std::vector<double> wavelengthMicrons, Parameters params)
      : BrdfModel(std::move(wavelengthMicrons)), params_(std::move(params)) {
    init();
  }

  /// Virtual destructor
  ~RossLi() override = default;

  /// The parameters with which to BRDF was constructed
  [[nodiscard]] const Parameters& parameters() const { return params_; }

  [[nodiscard]] std::vector<double> sample(
    double illumZenithRad,
    double illumAzimuthRad,
    double viewZenithRad,
    double viewAzimuthRad) const override;

private:
  void init() override;

  [[nodiscard]] std::vector<double> doSimplifySpectralBrdf(
    std::vector<double>& nadirDHR) override;

  void doGetHashInputs(const HashCallback& callback) const override;

private:
  Parameters params_;

  // Cache variables for holding interpolated paramter values
  std::vector<double> isoParams_, volParams_, geoParams_;
};

/// Converts a RossKernel to a string
D5_BRDF_EXPORT inline std::string to_string(RossLi::RossKernel k) {
  switch (k) {
    case RossLi::RossKernel::Thick:
      return "Thick";
    case RossLi::RossKernel::Thin:
      return "Thin";
    default:
      return "Unknown";
  }
}

/// Converts a LiKernel to a string
D5_BRDF_EXPORT inline std::string to_string(RossLi::LiKernel k) {
  switch (k) {
    case RossLi::LiKernel::Sparse:
      return "Sparse";
    case RossLi::LiKernel::Dense:
      return "Dense";
    default:
      return "Unknown";
  }
}

/// \}

}  // namespace d5::brdf
