#pragma once

#include <algorithm>

#include "BrdfModel.h"
#include "dirsig5/brdf/Export.h"

namespace d5::brdf {

/** A simple unpolarized version of the ShellBackground BRDF model.
 *  \ingroup brdf
 */
class D5_BRDF_EXPORT ShellBackground : public BrdfModel {
public:
  struct Parameters {
    /// Unpolarized fit coefficient \f$ a \f$.
    double a = 0;

    /// Unpolarized fit coefficient \f$ b \f$.
    double b = 0;

    /// Ground sample distance (GSD) in inches.
    double gsdInches = 6;

    /// Fit sample.
    struct FitSample {
      /// Integrated reflectance \f$ \rho \f$.
      double rho = 1;

      /// Roujean coefficient \f$ k_0 \f$.
      double k0 = 0;

      /// Roujean coefficient \f$ k_1 \f$.
      double k1 = 0;

      /// Roujean coefficient \f$ k_2 \f$.
      double k2 = 0;
    };

    /// Emissivity sample.
    struct EmissivitySample {
      /// Wavelength in micrometers.
      double wavelengthMicrons = 0;

      /// Emissivity.
      double emissivity = 0;
    };

    /// Fit samples.
    std::vector<FitSample> fit;

    /// Emissivity samples.
    std::vector<EmissivitySample> emissivity;
  };

  /// Constructor.
  ShellBackground(std::vector<double> wavelengthMicrons, Parameters params)
      : BrdfModel(std::move(wavelengthMicrons)), params_(std::move(params)) {
    // Ensure 2 or more fit samples.
    if (params_.fit.empty()) params_.fit.emplace_back();
    if (params_.fit.size() == 1) {
      params_.fit.push_back(params_.fit.front());
      params_.fit.back().rho += 0.1;
    }
    // Ensure 1 or more emissivity samples.
    if (params_.emissivity.empty()) {
      params_.emissivity.emplace_back();
    }
    // Sort fit samples by reflectance.
    std::sort(
      params_.fit.begin(), params_.fit.end(),
      [](const auto& lhs, const auto& rhs) { return lhs.rho < rhs.rho; });
    // Sort emissivity samples by wavelength.
    std::sort(
      params_.emissivity.begin(), params_.emissivity.end(),
      [](const auto& lhs, const auto& rhs) {
        return lhs.wavelengthMicrons < rhs.wavelengthMicrons;
      });
    // Delegate.
    init();
  }

  /// The parameters with which the model was constructed
  [[nodiscard]] const Parameters& parameters() const { return params_; }

  [[nodiscard]] std::vector<double> sample(
    double illumZenithRad,
    double illumAzimuthRad,
    double viewZenithRad,
    double viewAzimuthRad) const override;

private:
  void init() override;

  [[nodiscard]] std::vector<double> doSimplifySpectralBrdf(
    std::vector<double>& nadirDHR) override;

  void doGetHashInputs(const HashCallback& callback) const override;

private:
  /// Parameters.
  Parameters params_;
  /// Interpolated emissivity curve.
  std::vector<double> emissivity_;
  double f00Sigma_ = 0;
};

}  // namespace d5::brdf
