#pragma once

#include <complex>
#include <memory>
#include <vector>

#include "BrdfModel.h"
#include "MicroFacetModels.h"
#include "dirsig5/brdf/Export.h"

namespace d5::brdf {

/** Target pBRDF model described by James Shell.
 *
 *  <I>Reference</I>: Shell, J. R, "Polarimetric Remote Sensing in the Visible
 *                    to Near Infrared", Phd Thesis.
 *
 *  This BRDF model is most useful for materials which are homogeneous
 *  at spatial scales of interest to remote sensing (typically man-made
 *  materials).
 *
 *  \ingroup brdf
 */
class D5_BRDF_EXPORT ShellTarget : public BrdfModel {
public:
  /// Constants for the various spectral interpolation schemes
  enum class Interpolation {
    /// Linear interpolation of modeled BRDF between reference wavelengths
    IgnoreEmissivityCurve = 0,

    /** Linear interpolation of 1st surface scatter term of BRDF between
     *  reference wavelengths. Apply wavelength dependent gain to "bulk"
     *  scatter BRDF term such that the resulting hemispherical integration of
     *  full BRDF matches the lab measured DHR spectrum.
     */
    DiffuseOnly = 1,

    /** Linear interpolation of both 1st surface and "bulk" scatter terms of
     *  BRDF between reference wavelengths with an additional wavelength
     *  dependent gain applied to model such that the resulting hemispherical
     *  integration of modeled BRDF matches the lab measured DHR spectrum.
     */
    DiffuseAndSpecular = 2
  };

  /// Parameters to this BRDF model
  struct Parameters {
    /// The spectral interpolation scheme to use
    Interpolation interpolation = Interpolation::DiffuseOnly;

    /// Wavelengths in microns for the high-resolution DHR
    std::vector<double> dhrWavelengths;

    /// High-resolution DHR spectrum values
    std::vector<double> dhr;

    /// Parameters which vary spectrally
    struct SpectralParameters {
      /// The wavelength for this sample in microns
      double wavelengthMicrons;

      /// Complex index of refraction
      std::complex<double> ior;

      /// DHR at this wavelength
      double dhr;

      /// Gain factor used to modulate transmission
      double tgain = 1;

      /// Orientation probability density function for microfacets
      std::shared_ptr<OrientationProbability> orientPdf;

      /// Microfacet shadowing function (must be SFMaxwellBeard)
      std::shared_ptr<ShadowingFunction> shadowFunc;

      /// Volumetric scattering term function (must be VTMaxwellBeard)
      std::shared_ptr<VolumetricTerm> volumeTerm;
    };

    /// List of spectrally-varying parameters
    std::vector<SpectralParameters> spectral;

    /// Convenience typedef for an iterator in \c spectral
    using spectral_it = typename std::vector<SpectralParameters>::iterator;
  };

  /** Constructor
   *
   *  \param wavelengthMicrons  Spectral sampling desired for this model
   *  \param params             Parameters for the model
   */
  ShellTarget(std::vector<double> wavelengthMicrons, Parameters params);

  /// Virtual destructor
  ~ShellTarget() override = default;

  /// The parameters with which the model was constructed
  [[nodiscard]] const Parameters& parameters() const { return params_; }

  [[nodiscard]] std::vector<double> sample(
    double illumZenithRad,
    double illumAzimuthRad,
    double viewZenithRad,
    double viewAzimuthRad) const override;

private:
  void init() override;

  [[nodiscard]] std::vector<double> doSimplifySpectralBrdf(
    std::vector<double>& nadirDHR) override;

  void doGetHashInputs(const HashCallback& callback) const override;

private:
  Parameters params_;

  // Local cache parameters to prevent recomputation
  std::vector<std::complex<double>> nk_;
  std::vector<std::pair<Parameters::spectral_it, Parameters::spectral_it>>
    fitIters_;
};

}  // namespace d5::brdf
