#pragma once

#include <array>
#include <cmath>
#include <cstdint>
#include <limits>
#include <string>
#include <unordered_map>
#include <utility>
#include <vector>

#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/span.h"
#include "dirsig5/d4legacy/Export.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_geometry
/// \{

/// Holds the data for a facetized object
class D5_D4LEGACY_EXPORT FacetObject {
public:
  // ##############################
  // # Maps (for Python bindings) #
  // ##############################

  /// The vertexes.
  [[nodiscard]] auto mapv() {
    return Eigen::Map<Eigen::MatrixXf>(v_.data(), v_.size() / 3, 3);
  }

  [[nodiscard]] auto mapv(uint32_t i) {
    return Eigen::Map<Eigen::Vector3f>(vertex(i));
  }

  [[nodiscard]] auto mapv(uint32_t i) const {
    return Eigen::Map<const Eigen::Vector3f>(vertex(i));
  }

  /// The vertex normals.
  [[nodiscard]] auto mapvn() {
    return Eigen::Map<Eigen::MatrixXf>(vn_.data(), vn_.size() / 3, 3);
  }

  [[nodiscard]] auto mapvn(uint32_t i) {
    return Eigen::Map<Eigen::Vector3f>(vertexNormal(i));
  }

  [[nodiscard]] auto mapvn(uint32_t i) const {
    return Eigen::Map<const Eigen::Vector3f>(vertexNormal(i));
  }

  /// The vertex texture coordinates.
  [[nodiscard]] auto mapvt() {
    return Eigen::Map<Eigen::MatrixXf>(vt_.data(), vt_.size() / 2, 2);
  }

  [[nodiscard]] auto mapvt(uint32_t i) {
    return Eigen::Map<Eigen::Vector2f>(textureCoord(i));
  }

  [[nodiscard]] auto mapvt(uint32_t i) const {
    return Eigen::Map<const Eigen::Vector2f>(textureCoord(i));
  }

  /// The facet vertex indexes.
  [[nodiscard]] auto mapf() {
    return Eigen::Map<Eigen::MatrixX<uint32_t>>(f_.data(), f_.size() / 3, 3);
  }

  /// The facet vertex normal indexes.
  [[nodiscard]] auto mapfn() {
    return Eigen::Map<Eigen::MatrixX<uint32_t>>(
      vnFacets_.data(), vnFacets_.size() / 3, 3);
  }

  /// The facet vertex texture coordinate indexes.
  [[nodiscard]] auto mapft() {
    return Eigen::Map<Eigen::MatrixX<uint32_t>>(
      vtFacets_.data(), vtFacets_.size() / 3, 3);
  }

  /// The per-facet thicknesses.
  [[nodiscard]] auto mapthickness() {
    return d5::span<float>{facetThickness_.data(), facetThickness_.size()};
  }

  /// The per-facet temperatures.
  [[nodiscard]] auto maptemperature() {
    return d5::span<float>{facetTempK_.data(), facetTempK_.size()};
  }

  // #############################
  // #     Original Filename     #
  // #############################

  bool hasOriginalFilename() const { return !originalFilename_.empty(); }

  void setOriginalFilename(const std::string& s) { originalFilename_ = s; }

  const std::string& originalFilename() const { return originalFilename_; }

  // ############################
  // #         Part API         #
  // ############################

  /// Represents a part of the facetized object
  struct Part {
    /// Name of the part
    std::string name;

    /// Number of the facets in the part
    size_t numFacets = 0;
  };

  /// List of parts in the object
  [[nodiscard]] const std::vector<Part>& parts() const { return parts_; }

  /// Adds a part to the object and returns the local index of it
  uint32_t addPart(const std::string& name);

  /// Gets a part by name (nullptr if it doesn't exist)
  [[nodiscard]] const Part* getPartByName(const std::string& name) const;

  /// Gets a part by name (-1 if it doesn't exist)
  [[nodiscard]] int getPartIndexByName(const std::string& name) const;

  // Removes any empty parts from the model
  void removeEmptyParts() {
    if (parts_.empty()) return;
    for (int64_t i = static_cast<int64_t>(parts_.size()) - 1; i >= 0; --i) {
      if (parts_[i].numFacets == 0) {
        parts_.erase(parts_.begin() + i);
        for (auto& p : partForFacet_) {
          if (p > i) p--;
        }
      }
    }
  }

  /// Removes a part from the object
  void removePart(const std::string& name);

  // ############################
  // #        Vertex API        #
  // ############################

  /// The vertex buffer for the object. Vertices are stored in XYZ order.
  [[nodiscard]] const std::vector<float>& vertexData() const { return v_; }

  /// The number of vertices
  [[nodiscard]] uint32_t numVertices() const {
    return v_.size() / kVertexStride;
  }

  /// Accessor for a single vertex
  [[nodiscard]] const float* vertex(uint32_t i) const {
    return &v_[kVertexStride * i];
  }

  /// \overload
  [[nodiscard]] float* vertex(uint32_t i) { return &v_[kVertexStride * i]; }

  /// Adds a vertex to the object and returns it index
  uint32_t addVertex(float x, float y, float z) {
    v_.push_back(x);
    v_.push_back(y);
    v_.push_back(z);
    return numVertices() - 1;
  }

  /// Reserves space for the given number of vertices
  void reserveVertices(uint32_t num) { v_.reserve(kVertexStride * num); }

  // ###########################
  // #        Facet API        #
  // ###########################

  /** The facet buffer. Every 3 elements represent a triangle and are indices
   *  that can be passed to \c vertex().
   */
  [[nodiscard]] const std::vector<uint32_t>& facetData() const { return f_; }

  /// Accessor for a single facet
  [[nodiscard]] const uint32_t* facet(uint32_t i) const {
    return &f_[kFacetStride * i];
  }

  /// The number of facets
  [[nodiscard]] uint32_t numFacets() const { return f_.size() / kFacetStride; }

  /// \overload
  [[nodiscard]] uint32_t* facet(uint32_t i) { return &f_[kFacetStride * i]; }

  /// Reserves memory for the given number of facets
  void reserveFacets(size_t n) {
    f_.reserve(kFacetStride * n);
    partForFacet_.reserve(n);
  }

  /// Adds a facet to the given part
  uint32_t addFacetToPart(
    uint32_t partIdx, uint32_t v0, uint32_t v1, uint32_t v2, uint32_t matId) {
    uint32_t i = numFacets();
    f_.push_back(v0);
    f_.push_back(v1);
    f_.push_back(v2);
    parts_[partIdx].numFacets++;
    partForFacet_.push_back(partIdx);

    // Every facet must have a material
    matForFacet_.push_back(matId);

    // If there are texture or normal facets, they must be the same size
    if (!vtFacets_.empty()) vtFacets_.resize(f_.size(), 0);
    if (!vnFacets_.empty()) vnFacets_.resize(f_.size(), kNoVertexNormal);

    return i;
  }

  /// Returns the part index for the given facet index
  [[nodiscard]] uint32_t partForFacet(uint32_t i) const {
    return partForFacet_[i];
  }

  // TODO Use Eigen
  /// Calculates the facet normal for the given facet index
  [[nodiscard]] std::array<float, 3> facetNormal(uint32_t i) const {
    const float* v0 = vertex(facet(i)[0]);
    const float* v1 = vertex(facet(i)[1]);
    const float* v2 = vertex(facet(i)[2]);
    std::array<float, 3> e1 = {v1[0] - v0[0], v1[1] - v0[1], v1[2] - v0[2]};
    std::array<float, 3> e2 = {v2[0] - v0[0], v2[1] - v0[1], v2[2] - v0[2]};
    std::array<float, 3> n = {
      e1[1] * e2[2] - e1[2] * e2[1], e1[2] * e2[0] - e1[0] * e2[2],
      e1[0] * e2[1] - e1[1] * e2[0]};
    float l = std::sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);
    if (l > 0) {
      n[0] /= l;
      n[1] /= l;
      n[2] /= l;
    }
    return n;
  }

  // #################################
  // #      Facet Thickness API      #
  // #################################

  /** The facet thickness in meters. Element indices correspond to the facet
   *  that would be returned by \c facet() with that index. Can be empty.
   */
  [[nodiscard]] auto& facetThicknessesM() { return facetThickness_; }

  [[nodiscard]] auto& facetThicknessesM() const { return facetThickness_; }

  [[nodiscard]] bool hasFacetThicknesses() const {
    return !facetThickness_.empty();
  }

  /// Sets the thickness in meters for the given facet
  void setFacetThicknessM(uint32_t facetIdx, float thickness) {
    if (facetIdx >= facetThickness_.size()) {
      facetThickness_.resize(numFacets(), 0);
      if (facetIdx >= facetThickness_.size()) return;
    }
    facetThickness_[facetIdx] = thickness;
  }

  void removeFacetThicknesses() { facetThickness_ = {}; }

  // ###################################
  // #      Facet Temperature API      #
  // ###################################

  /** The facet temperatures in Kelvin. Element indices correspond to the
   * facet facet that would be returned by \c facet() with that index. Can be
   * empty.
   */
  [[nodiscard]] auto& facetTemperaturesK() { return facetTempK_; }

  [[nodiscard]] auto& facetTemperaturesK() const { return facetTempK_; }

  [[nodiscard]] bool hasFacetTemperatures() const {
    return !facetTempK_.empty();
  }

  /// Sets the temperature in Kelvin for the given facet (-1 for calculated)
  void setFacetTemperatureK(uint32_t facetIdx, float k) {
    if (facetIdx >= facetTempK_.size()) {
      facetTempK_.resize(numFacets(), -1);
      if (facetIdx >= facetTempK_.size()) return;
    }
    facetTempK_[facetIdx] = k;
  }

  void removeFacetTemperatures() { facetTempK_ = {}; }

  // ##################################
  // #       Facet Material API       #
  // ##################################

  /** The facet material indices. Element indices correspond to the facet
   *  facet that would be returned by \c facet() with that index.
   */
  [[nodiscard]] const std::vector<uint32_t>& facetMaterialIndices() const {
    return matForFacet_;
  }

  /// Sets the facet material index for the given facet
  void setFacetMaterialIndex(uint32_t facetIdx, uint32_t matId) {
    if (facetIdx >= matForFacet_.size()) {
      matForFacet_.resize(numFacets(), 0);
      if (facetIdx >= matForFacet_.size()) return;
    }
    matForFacet_[facetIdx] = matId;
  }

  // ##############################
  // #        Material API        #
  // ##############################

  /// Adds a material label to the object and returns the local index of it
  uint32_t addMaterialLabel(const std::string& label);

  /// Returns the number of materials in the object
  [[nodiscard]] uint32_t numMaterials() const { return materialLabels_.size(); }

  /// Returns the name of the material with the given index
  [[nodiscard]] const std::string& materialLabel(uint32_t i) const {
    return materialLabels_[i];
  }

  /// Sets the name material with the given index
  void setMaterialLabel(uint32_t i, std::string name) {
    materialLabels_[i] = std::move(name);
  }

  // ###################################
  // #      Facet Emissivity API       #
  // ###################################

  /// The thermal emissivity of the facets.
  [[nodiscard]] const std::vector<float>& facetThermalEmissivities() const {
    return facetThermalEmissivity_;
  }

  /// Whether the object has facet temperatures
  [[nodiscard]] bool hasFacetThermalEmissivities() const {
    return !facetThermalEmissivity_.empty();
  }

  /// Sets the thermal emissivity [0 to 1] for the given facet
  void setFacetThermalEmissivity(uint32_t facetIdx, float tE) {
    if (facetIdx >= facetThermalEmissivity_.size()) {
      facetThermalEmissivity_.resize(numFacets(), -1);
      if (facetIdx >= facetThermalEmissivity_.size()) return;
    }
    facetThermalEmissivity_[facetIdx] = tE;
  }

  void removeFacetThermalEmissivities() { facetThermalEmissivity_ = {}; }

  // ######################################
  // #       Texture Coordinate API       #
  // ######################################

  /** The data buffer for UV texture coordinates. Coordinates are stored as UV
   *  coordinate pairs.
   */
  [[nodiscard]] const std::vector<float>& textureCoords() const { return vt_; }

  /// The number of texture coordinates
  [[nodiscard]] uint32_t numTextureCoords() const {
    return vt_.size() / kVtStride;
  }

  /// Accessor for a single set of texture coordinates
  [[nodiscard]] const float* textureCoord(uint32_t i) const {
    return &vt_[kVtStride * i];
  }

  /// \overload
  [[nodiscard]] float* textureCoord(uint32_t i) { return &vt_[kVtStride * i]; }

  /// Adds a texture coordinate to the object and returns its index
  uint32_t addTextureCoord(float u, float v) {
    vt_.push_back(u);
    vt_.push_back(v);
    return numTextureCoords() - 1;
  }

  /// Reserves space for texture coordinates
  void reserveTextureCoords(uint32_t num) { vt_.reserve(kVtStride * num); }

  void removeTextureCoords() {
    vt_ = {};
    vtFacets_ = {};
  }

  bool hasTextureCoords() const { return !vt_.empty(); }

  // #################################
  // #       Texture Facet API       #
  // #################################

  /** Facet buffer for texture coordinates. Facets are stored as index triples
   *  into \c textureCoords(). Facets correspond to facets in \c facetData().
   */
  [[nodiscard]] const std::vector<uint32_t>& textureFacets() const {
    return vtFacets_;
  }

  /// The number of texture facets in the object
  [[nodiscard]] size_t numTextureFacets() const {
    return vtFacets_.size() / kFacetStride;
  }

  /// Accessor for a single texture facet
  [[nodiscard]] const uint32_t* textureFacet(uint32_t i) const {
    return &vtFacets_[kFacetStride * i];
  }

  /// \overload
  [[nodiscard]] uint32_t* textureFacet(uint32_t i) {
    return &vtFacets_[kFacetStride * i];
  }

  /// Sets the value of the given texture facet
  void setTextureFacet(uint32_t i, uint32_t v0, uint32_t v1, uint32_t v2) {
    if (i < numFacets()) {
      vtFacets_.resize(f_.size(), 0);
      auto* f = textureFacet(i);
      f[0] = v0;
      f[1] = v1;
      f[2] = v2;
    }
  }

  // #################################
  // #       Vertex Normal API       #
  // #################################

  /// Data buffer for the vertex normals. Normals are stored as XYZ triples.
  [[nodiscard]] const std::vector<float>& vertexNormals() const { return vn_; }

  /// The number of vertex normals
  [[nodiscard]] uint32_t numVertexNormals() const {
    return vn_.size() / kVnStride;
  }

  /// Accessor for a single vertex normal
  [[nodiscard]] const float* vertexNormal(uint32_t i) const {
    return &vn_[kVnStride * i];
  }

  /// \overload
  float* vertexNormal(uint32_t i) { return &vn_[kVnStride * i]; }

  /// Adds a vertex normal to the object
  uint32_t addVertexNormal(float x, float y, float z) {
    vn_.push_back(x);
    vn_.push_back(y);
    vn_.push_back(z);
    return numVertexNormals() - 1;
  }

  /// Reserves space for the vertex normals
  void reserveVertexNormals(uint32_t num) { vn_.reserve(kVnStride * num); }

  void removeVertexNormals() {
    vn_ = {};
    vnFacets_ = {};
  }

  bool hasVertexNormals() const { return !vn_.empty(); }

  // #######################################
  // #       Vertex Normal Facet API       #
  // #######################################

  static constexpr auto kNoVertexNormal = std::numeric_limits<uint32_t>::max();

  /** Facet buffer for vertex normals. Facets are stored as index triples
   *  into \c vertexNormals(). Facets correspond to facets in \c facetData().
   */
  [[nodiscard]] const std::vector<uint32_t>& vertexNormalFacets() const {
    return vnFacets_;
  }

  /// The number of vertex normal facets in the object
  [[nodiscard]] size_t numVertexNormalFacets() const {
    return vnFacets_.size() / kFacetStride;
  }

  /// Accessor for a single normal facet
  [[nodiscard]] const uint32_t* vertexNormalFacet(uint32_t i) const {
    return &vnFacets_[kFacetStride * i];
  }

  /// \overload
  [[nodiscard]] uint32_t* vertexNormalFacet(uint32_t i) {
    return &vnFacets_[kFacetStride * i];
  }

  /// Sets the value of the given normal facet
  void setVertexNormalFacet(uint32_t i, uint32_t v0, uint32_t v1, uint32_t v2) {
    if (i < numFacets()) {
      vnFacets_.resize(f_.size(), kNoVertexNormal);
      auto* f = vertexNormalFacet(i);
      f[0] = v0;
      f[1] = v1;
      f[2] = v2;
    }
  }

  // ################################
  // #       Point Source API       #
  // ################################

  /// Represents a point source in the object
  struct PointSource {
    /// Index of the vertex for the point source
    size_t vertexIndex;

    /// What material the point source is
    uint32_t materialIndex;

    /// The direction the point source is pointing
    std::array<float, 3> direction;

    /// What part the point source is in
    uint32_t partIndex;
  };

  /// The list of point sources in the object
  [[nodiscard]] const std::vector<PointSource>& pointSources() const {
    return pointSources_;
  }

  /// \overload
  [[nodiscard]] PointSource& addPointSource() {
    return pointSources_.emplace_back();
  }

  // ###########################################
  // #        Point Source Material API        #
  // ###########################################

  /// Adds a material label to the object and returns the local index of it
  uint32_t addPointSourceMaterialLabel(const std::string& label);

  /// Returns the number of materials in the object
  [[nodiscard]] uint32_t numPointSourceMaterials() const {
    return pointSourceMaterialLabels_.size();
  }

  /// Returns the name of the material with the given index
  [[nodiscard]] const std::string& pointSourceMaterialLabel(uint32_t i) const {
    return pointSourceMaterialLabels_[i];
  }

  // ##############################
  // #        Keypoint API        #
  // ##############################

  struct KeypointValue {
    size_t partIndex = 0;
    std::vector<Point> points;
  };

  /** Map of keypoint names to a list of 3D points in the object's local
   *  coordinate system.
   */
  using Keypoints = std::unordered_map<std::string, KeypointValue>;

  [[nodiscard]] const Keypoints& keypoints() const { return keypoints_; }

  void addKeypointToPart(
    size_t partIndex, std::string name, std::vector<Point> points) {
    keypoints_[std::move(name)] = {partIndex, std::move(points)};
  }

  void removeAllUnusedAttributes();

  // ###############################
  // #       Transform API         #
  // ###############################

  void transform(Eigen::Matrix4f);

  void transform(Eigen::Matrix3f);

  void translate(Eigen::Vector3f);

  void scale(Eigen::Vector3f);

  void scale(float);

  void rotate(float theta, Eigen::Vector3f axis) {
    transform(Eigen::Matrix3f(Eigen::AngleAxisf(theta, axis)));
  }

  void rotateX(float theta) { rotate(theta, Eigen::Vector3f::UnitX()); }

  void rotateY(float theta) { rotate(theta, Eigen::Vector3f::UnitY()); }

  void rotateZ(float theta) { rotate(theta, Eigen::Vector3f::UnitZ()); }

  void swapAxes(uint32_t ax0, uint32_t ax1);

  BBoxf computeExtent() const;

  Vectorf computeExtentCenter() const;

  Vectorf computeCentroid() const;

  // ##############################
  // #       File write API       #
  // ##############################

  void writeObj(std::ostream&) const;

  void writeGdb(std::ostream&) const;

  std::string summary() const;

  void validateIndexes() const;

private:
  /// The original filename this came from, if applicable.
  std::string originalFilename_;

  /// Vertex buffer (stored [x0,y0,z0,x1,y1,z1,...])
  std::vector<float> v_;

  /// The stride of v_
  static constexpr size_t kVertexStride = 3;

  /// Facet buffer (stored [v00,v01,v02,v10,v11,v12,...])
  std::vector<uint32_t> f_;

  /// The stride of f_
  static constexpr size_t kFacetStride = 3;

  /** Facet material index (stored [mat0,mat1,...]). Must have the same number
   *  of entries as the number of facets
   */
  std::vector<uint32_t> matForFacet_;

  /** Facet part index (stored [part0,part1,...]). Must have the same number
   *  of entries as the number of facets
   */
  std::vector<uint32_t> partForFacet_;

  /** Facet thickness [m] (stored [t0,t1,...]). If non-empty, must have one
   *  entry for every facet.
   */
  std::vector<float> facetThickness_;

  /** Facet temperature [K] (stored [t0,t1,...]). If non-empty, must have one
   *  entry for every facet.
   */
  std::vector<float> facetTempK_;

  std::vector<float> facetThermalEmissivity_;

  /// Texture vertex buffer (stored [u0,v0,u1,v1,...])
  std::vector<float> vt_;

  /// The stride of vt_
  static constexpr size_t kVtStride = 2;

  /** Texture vertex facet buffer (stored [vt00,vt01,vt02,vt10,vt11,vt12,...])
   *  If non-empty, must have the same number of facets as f_.
   */
  std::vector<uint32_t> vtFacets_;

  /// Vertex normal buffer (stored [vnx0,vny0,vnz0,vnx1,vny1,vnz1,...])
  std::vector<float> vn_;

  /// The stride of vt_
  static constexpr size_t kVnStride = 3;

  /** Vertex normal facet buffer (stored [vn00,vn01,vn02,vn10,vn11,vn12,...])
   *  If non-empty, must have the same number of facets as f_.
   */
  std::vector<uint32_t> vnFacets_;

  /// List of parts
  std::vector<Part> parts_;

  /// The set of material labels used on facets
  std::vector<std::string> materialLabels_;

  /// Mapping of material label to index in materialLabels_
  std::unordered_map<std::string, uint32_t> matLabelToLocalId_;

  /// List of point sources in the object
  std::vector<PointSource> pointSources_;

  /// The set of material labels used on point sources
  std::vector<std::string> pointSourceMaterialLabels_;

  /// Mapping of material label to index in pointSourceMaterialLabels_
  std::unordered_map<std::string, uint32_t> pointSourceMatLabelToLocalId_;

  Keypoints keypoints_;
};

/** Splits parts off of this facet object into their own objects
 *
 *  \param[in,out] fo            Facet object to process
 *  \param[in] partNamesToSplit  Names of the parts to split
 *
 *  \return List of the split parts (in the order the names are given)
 *  \throw std::exception subclass on any error
 */
D5_D4LEGACY_EXPORT std::vector<FacetObject> SplitOffParts(
  FacetObject& fo, const std::vector<std::string>& partNamesToSplit);

/// \}
}  // namespace d5::d4legacy
