/** Interface file for facet object I/O
 *
 *  \file FacetObjectIO.h
 */

#pragma once

#include <string>

#include "FacetObject.h"
#include "Geometry.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/misc/Units.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_geometry
/// \{

/// Request object for reading facet objects
class FacetObjectReadRequest {
public:
  /// The filename to load
  [[nodiscard]] const std::string& filename() const { return filename_; }

  /// Sets the filename
  FacetObjectReadRequest& withFilename(std::string f) {
    filename_ = std::move(f);
    fromMusesPlugin_ = 0;
    return *this;
  }

  /// Determines if request is from MuSES Plugin
  [[nodiscard]] int MuSES_Plugin() const { return fromMusesPlugin_; }

  /// sends back weather start in seconds from 12/31/1899 @ noon
  [[nodiscard]] double MuSES_WeatherStart() const { return weatherStartTime_; }

  /// sends back weather duration in seconds
  [[nodiscard]] int MuSES_WeatherDuration() const { return weatherDuration_; }

  /// sends back muses temperature calc step size
  [[nodiscard]] int MuSES_StepSize() const { return stepSize_; }

  /// sends back weather filename
  [[nodiscard]] const std::string& MuSES_WeatherFilename() const {
    return weatherFilename_;
  }

  /// sends back weather filename
  [[nodiscard]] const std::string& MuSES_Filename() const {
    return newFilename_;
  }

  /// sends back weather filename
  [[nodiscard]] Point MuSES_Origin() const { return origin_; }

  /// sends back timeZone
  [[nodiscard]] float MuSES_timeZone() const { return timeZone_; }

  /// sends back solutionStartTime
  [[nodiscard]] double MuSES_solutionStartTime() const {
    return solutionStartTime_;
  }

  /// sends back solutionEndTime
  [[nodiscard]] double MuSES_solutionEndTime() const {
    return solutionEndTime_;
  }

  /// Whether to swap the Y and Z dimensions
  [[nodiscard]] bool swapYZ() const { return swapYZ_; }

  /// Sets whether to swap the Y and Z dimensions
  FacetObjectReadRequest& withSwapYZ(bool s) {
    swapYZ_ = s;
    return *this;
  }

  /// The units of the file
  [[nodiscard]] DistanceUnits units() const { return units_; }

  /// Sets the units of the file
  FacetObjectReadRequest& withUnits(DistanceUnits units) {
    units_ = units;
    return *this;
  }

  /// The time step index for MuSES TDF reader, or -1 for most recent
  [[nodiscard]] int timeStepIndex() const { return timeStepIndex_; }

  /// Sets the time step index for MuSES TDF reader
  FacetObjectReadRequest& withTimeStepIndex(int index) {
    timeStepIndex_ = index;
    fromMusesPlugin_ = 0;
    return *this;
  }

  [[nodiscard]] bool useTdfMaterialNames() const {
    return useTdfMaterialNames_;
  }

  FacetObjectReadRequest& withUseTdfMaterialNames(bool flag) {
    useTdfMaterialNames_ = flag;
    return *this;
  }

  /// Saves off an updated version of the tdf file
  FacetObjectReadRequest& withMusesPlugin(
    std::string origFilename,
    std::string newFilename,
    std::string weatherFilename,
    double weatherStartTime,
    int weatherDuration,
    int stepSize,
    Point origin,
    float timeZone,
    double solutionStartTime,
    double solutionEndTime,
    int timeStepIndex) {
    filename_ = std::move(origFilename);
    newFilename_ = std::move(newFilename);
    weatherFilename_ = std::move(weatherFilename);
    weatherStartTime_ = std::move(weatherStartTime);
    weatherDuration_ = std::move(weatherDuration);
    fromMusesPlugin_ = 1;
    timeStepIndex_ = timeStepIndex;
    stepSize_ = stepSize;
    origin_ = origin;
    timeZone_ = timeZone;
    solutionStartTime_ = solutionStartTime;
    solutionEndTime_ = solutionEndTime;
    return *this;
  }

  /// Map of keypoints to a list of 3D points in the object's local
  /// coordinates
  using Keypoints = std::vector<FacetObjectWithProperties::Keypoint>;

  /// Returns the keypoint list
  [[nodiscard]] const Keypoints& keypoints() const { return keypoints_; }

  /// Sets the keypoint list
  FacetObjectReadRequest& withKeypoints(const Keypoints& keypoints) {
    keypoints_ = keypoints;
    return *this;
  }

  FacetObjectReadRequest& withTemperature(double temp) {
    temperature_ = temp;
    return *this;
  }

  [[nodiscard]] double temperature() const { return temperature_; }

  /// Equality operator
  bool operator==(const FacetObjectReadRequest& other) const {
    return swapYZ_ == other.swapYZ_ && units_ == other.units_ &&
           filename_ == other.filename_ &&
           timeStepIndex_ == other.timeStepIndex_ &&
           temperature_ == other.temperature_ && keypoints_ == other.keypoints_;
  }

  bool operator!=(const FacetObjectReadRequest& other) const {
    return !operator==(other);
  }

private:
  std::string filename_;
  bool swapYZ_ = false;
  DistanceUnits units_ = DistanceUnits::Meters;
  Keypoints keypoints_;
  double temperature_ = -1;

  /// Time step index, specifically for MuSES TDF reader
  int timeStepIndex_ = -1;
  bool useTdfMaterialNames_ = false;

  std::string weatherFilename_;
  std::string newFilename_;
  double weatherStartTime_;
  double solutionStartTime_;
  double solutionEndTime_;
  int weatherDuration_;
  int stepSize_;
  float timeZone_;
  Point origin_;
  int fromMusesPlugin_ = 0;
};

/** Reads a facet object from a file
 *
 *  \param request  The read request to perform
 *
 *  \return The read facet object
 *  \throw A std::exception subclass on any error
 */
D5_D4LEGACY_EXPORT FacetObject
ReadFacetObjectFromFile(const FacetObjectReadRequest& request);

/// \}
}  // namespace d5::d4legacy
