#pragma once

#include <array>
#include <map>
#include <memory>
#include <optional>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <unordered_map>
#include <utility>
#include <vector>

#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Inheritance.h"
#include "dirsig5/d4legacy/geometry/HeightModel.h"
#include "dirsig5/d4legacy/misc/FileLocator.h"
#include "dirsig5/d4legacy/misc/Units.h"
#include "dirsig5/d4legacy/motion/FlexMotion.h"
#include "dirsig5/d4legacy/motion/Spin.h"
#include "dirsig5/d4legacy/tag/TagFileElement.h"
#include "dirsig5/d4legacy/xml/xml.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_geometry
/// \{

/// \cond HIDDEN_SYMBOLS
class GeometryList;
/// \endcond

/// Base class for all types of geometry supported by DIRSIG4
class D5_D4LEGACY_EXPORT Geometry : public EnumeratedSuperclass<Geometry> {
public:
  /// Constants for each type of geometry
  enum class Type {
    Unknown = 0,
    GdbFacetObject,
    ObjFacetObject,
    TdfFacetObject,
    HdfFacetObject,
    OdbInclude,
    GlistInclude,
    Box,
    CurvedFrustum,
    Cylinder,
    Sphere,
    Sinusoid,
    SecchiDisk,
    Disk,
    RegularGrid,
    SurfacePlane,
    CatenaryCurve,
    HeightField,
    PointSource,
    BezierCurveSet
  };

  virtual ~Geometry() = default;

  [[nodiscard]] Type type() const noexcept { return type_; }

  /// The line number this came from
  std::optional<size_t> lineNumber;

  /// Calls the given callable with the appropriate downcasted suclass
  template <typename Callable, typename... Args>
  decltype(auto) dispatch(Callable&& callback, Args&&... args) const;

  static std::shared_ptr<Geometry> FromXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});

  virtual void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});

  virtual void ReadTag(const TagFileSection& section);

protected:
  Geometry(Type type) : type_(type) {}

private:
  Type type_ = Type::Unknown;
};

/// Convert a Geometry::Type to a string
D5_D4LEGACY_EXPORT std::string to_string(Geometry::Type type);

/// Base class to allow the attachment of part properties to a facet object
class D5_D4LEGACY_EXPORT FacetObjectWithProperties {
public:
  struct Part {
    bool enabled = true;
    std::string name;
    std::string materialId;
    std::optional<Point> origin;
    std::variant<std::nullptr_t, std::string, SpinMotion, FlexMotion> motion =
      nullptr;
    bool operator==(const Part& other) const {
      if (
        enabled != other.enabled || name != other.name ||
        materialId != other.materialId || origin != other.origin ||
        motion.index() != other.motion.index()) {
        return false;
      }
      return motion.index() == 0 ||
             (motion.index() == 1 &&
              std::get<1>(motion) == std::get<1>(other.motion));
    }
    bool operator!=(const Part& other) const { return !operator==(other); }
  };

  struct Keypoint {
    std::string name;
    std::string part;
    std::vector<Point> points;
    bool operator==(const Keypoint& other) const {
      return name == other.name && part == other.part && points == other.points;
    }
  };

  std::string filename;

  std::vector<Part> partProperties;

  std::vector<Keypoint> keypoints;

  DistanceUnits units = DistanceUnits::Meters;

  bool swapYZ = false;

  double defaultTemperature = -1;

  std::map<std::string, std::string> materialRemap;

protected:
  void FacetObjectReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});
};

// Trivial subclasses of Geometry

class D5_D4LEGACY_EXPORT GdbFacetObject final
    : public Geometry::Subclass<Geometry::Type::GdbFacetObject>,
      public FacetObjectWithProperties {
public:
  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override {
    FacetObjectReadXML(node, facade, fileLocator);
  }
};

class D5_D4LEGACY_EXPORT ObjFacetObject final
    : public Geometry::Subclass<Geometry::Type::ObjFacetObject>,
      public FacetObjectWithProperties {
public:
  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override {
    FacetObjectReadXML(node, facade, fileLocator);
  }
};

class D5_D4LEGACY_EXPORT TdfFacetObject final
    : public Geometry::Subclass<Geometry::Type::TdfFacetObject>,
      public FacetObjectWithProperties {
public:
  int timeStepIndex = -1;
  bool useTdfMaterialNames = false;

public:
  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

class D5_D4LEGACY_EXPORT HdfFacetObject final
    : public Geometry::Subclass<Geometry::Type::HdfFacetObject>,
      public FacetObjectWithProperties {
public:
  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override {
    FacetObjectReadXML(node, facade, fileLocator);
  }
};

template <typename T>
using is_facet_object = std::disjunction<
  std::is_same<std::decay_t<T>, GdbFacetObject>,
  std::is_same<std::decay_t<T>, ObjFacetObject>,
  std::is_same<std::decay_t<T>, TdfFacetObject>,
  std::is_same<std::decay_t<T>, HdfFacetObject>>;

template <typename T>
constexpr bool is_facet_object_v = is_facet_object<T>::value;

constexpr bool IsFacetObject(Geometry::Type type) noexcept {
  return type == Geometry::Type::GdbFacetObject ||
         type == Geometry::Type::ObjFacetObject ||
         type == Geometry::Type::TdfFacetObject ||
         type == Geometry::Type::HdfFacetObject;
}

class D5_D4LEGACY_EXPORT OdbInclude final
    : public Geometry::Subclass<Geometry::Type::OdbInclude> {
public:
  /// The filename to include, should end in `.odb`.
  std::string filename;

public:
  void ReadXML(
    const xml::Node& node,
    const geolib::Facade*,
    const FileLocator& fileLocator = {}) override {
    filename = xml::readFilename(
      node, "filename", fileLocator, FileLocator::GeometryList);
  }
};

class D5_D4LEGACY_EXPORT GlistInclude final
    : public Geometry::Subclass<Geometry::Type::GlistInclude> {
public:
  /// The filename to include, should end in `.glist`.
  std::string filename;

public:
  void ReadXML(
    const xml::Node& node,
    const geolib::Facade*,
    const FileLocator& fileLocator = {}) override {
    filename = xml::readFilename(
      node, "filename", fileLocator, FileLocator::GeometryList);
  }
};

template <typename T>
using is_glist_include = std::disjunction<
  std::is_same<std::decay_t<T>, GlistInclude>,
  std::is_same<std::decay_t<T>, OdbInclude>>;

template <typename T>
constexpr bool is_glist_include_v = is_glist_include<T>::value;

class D5_D4LEGACY_EXPORT RegularGrid final
    : public Geometry::Subclass<Geometry::Type::RegularGrid> {
public:
  /// The filename
  std::string filename;

  /// The insertion point
  Point insertPoint = {0, 0, 0};

  /// The voxel size in each axis
  Point voxelSize = {0, 0, 0};

  /// The type of the grid
  std::string mediumType;

public:
  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Geometry subclass to represent a height field.
class D5_D4LEGACY_EXPORT HeightField final
    : public Geometry::Subclass<Geometry::Type::HeightField> {
public:
  // Whether to write to an OBJ
  bool writeObjs = false;

  // Material ID for the object
  std::string materialId;

  bool boxMode = false;

  int cacheSize = 20;

  std::unique_ptr<HeightModel> heightModel;

public:
  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Geometry subclass to represent a point source.
class D5_D4LEGACY_EXPORT PointSource final
    : public Geometry::Subclass<Geometry::Type::PointSource> {
public:
  /// The material ID.
  std::string matId;

  /// The pointing vector.
  Vector pointingVec = {0, 0, 1};

public:
  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Geometry subclass to represent a bezier curve set.
class D5_D4LEGACY_EXPORT BezierCurveSet final
    : public Geometry::Subclass<Geometry::Type::BezierCurveSet> {
public:
  std::string matId;

  /// The vertex (X,Y,Z,R) data.
  std::vector<float> vertexData;

  /// The vertex normal (X,Y,Z) data. This is optional.
  std::vector<float> normalData;

  /// The first indexes of each 4-vertex curve. This is optional.
  std::vector<uint32_t> firstIndexes;

public:
  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// \}
}  // namespace d5::d4legacy

#ifndef D4LEGACY_INCLUDED_FROM_PRIMITIVES_H
#ifndef D4LEGACY_INCLUDED_FROM_GEOMETRY_DISPATCH_H
#define D4LEGACY_INCLUDED_FROM_GEOMETRY_H
#include "detail/GeometryDispatch.h"
#endif
#endif
