/** Interface file for height models
 *
 *  \file HeightModel.h
 */

#pragma once

#include <iosfwd>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <variant>

#include "dirsig5/base/Inheritance.h"
#include "dirsig5/d4legacy/Export.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_geometry
/// \{

class D5_D4LEGACY_EXPORT HeightModel
    : public EnumeratedSuperclass<HeightModel> {
public:
  enum class Type { Unknown = 0, Wave, Image };

  virtual ~HeightModel() = default;

  [[nodiscard]] Type type() const { return type_; }

  /// Calls the given callable with the appropriate downcasted suclass
  template <typename Callable, typename... Args>
  decltype(auto) dispatch(Callable&& callback, Args&&... args) const;

protected:
  HeightModel(Type type) : type_(type) {}

private:
  Type type_;
};

/// Output operator
D5_D4LEGACY_EXPORT std::ostream& operator<<(
  std::ostream& os, const HeightModel& hm);

/// Convert a HeightModel::Type to a string
D5_D4LEGACY_EXPORT std::string to_string(HeightModel::Type type);

/// Data model for wave-based height models
class D5_D4LEGACY_EXPORT WaveHeightModel
    : public HeightModel::Subclass<HeightModel::Type::Wave> {
public:
  ~WaveHeightModel() override = default;

  /// Cosine^2s spreading function
  struct Cos2S {
    /// Exponent (divided by 2) of the cosine
    double s;
  };

  /// Constant wind model
  struct ConstWind {
    /// Wind speed in m/s @10m
    double speed10;

    /// Direction of the wind (deg CW from North)
    double direction;
  };

  /// Joint North Sea Wave Project (JONSWAP) spectrum
  struct JONSWAP {
    /// Peak enhancement factor
    double gamma = 3.3;

    /// Fetch length [meters]
    double fetch = 1e6;
  };

  /// Pierson-Moskowitz spectrum
  struct PM {
    double alpha = 8.1e-3;
    double g = 9.80665;
  };

  /// The wave spread model (only cos^2s is supported)
  Cos2S cos2SWaveSpread;

  /// The wind model (only constant is supported)
  ConstWind constWind;

  /// The wave spectrum model
  std::variant<JONSWAP, PM> waveSpectrum;

  size_t nElements = 1024;
};

/// Data model for image height models
class D5_D4LEGACY_EXPORT ImageHeightModel
    : public HeightModel::Subclass<HeightModel::Type::Image> {
public:
  ~ImageHeightModel() override = default;

  double xWidth = 0;
  double yWidth = 0;
  std::string filename;
  double dcHeight = 0;
};

// ##############################################
// #   Template Implementation of HeightField   #
// ##############################################

/// \cond HIDDEN_SYMBOLS

template <typename Callable, typename... Args>
decltype(auto) HeightModel::dispatch(
  Callable&& callback, Args&&... args) const {
  switch (type_) {
    case Type::Wave:
      return callback(as<WaveHeightModel>(), std::forward<Args>(args)...);
    case Type::Image:
      return callback(as<ImageHeightModel>(), std::forward<Args>(args)...);
    case Type::Unknown:
    default:
      throw std::runtime_error(
        "HeightModel::dispatch() encountered unknown type");
  }
}

/// \endcond

/// \}
}  // namespace d5::d4legacy
