#pragma once

#include <array>
#include <map>
#include <ostream>
#include <stdexcept>
#include <string>
#include <type_traits>

#include "Rotation.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/motion/FlexMotion.h"
#include "dirsig5/d4legacy/motion/GenericMotion.h"
#include "dirsig5/d4legacy/xml/xml.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_geometry
/// \{

/// Base class for all instance types
class D5_D4LEGACY_EXPORT Instance : public EnumeratedSuperclass<Instance> {
public:
  /// Constants for the supported types of instances
  enum class Type {
    Tsr,            ///< translation, scale, rotation
    Affine,         ///< affine transform
    DeltaMotion,    ///< Delta (keyframe) motion (.mov file)
    GenericMotion,  ///< Generic motion (.ppd file)
    FlexMotion,     ///< Flexible motion
    Decal           ///< decal instance
  };

protected:
  Instance(Type type) : type_(type) {}

public:
  virtual ~Instance() = default;

  /// The type of the instance
  [[nodiscard]] Type type() const { return type_; }

  /// Tests whether the given type is a static instance type
  static bool IsStatic(Type type) {
    return type == Type::Tsr || type == Type::Affine;
  }

  /// \overload
  static bool IsStatic(const Instance& inst) { return IsStatic(inst.type()); }

  /// The name of the instance (can be empty)
  std::string name;

  /// The (optional) material remapping table
  std::map<std::string, std::string> origToNewMaterialId;

  /// The (optional) set of instance tags
  std::vector<std::string> tags;

  /// The time window string for the instance (can be empty)
  std::string timeWindow;

  /// Whether the instance has a time window
  [[nodiscard]] bool hasTimeWindow() const { return !timeWindow.empty(); }

  /// The (optional) filename of the control table
  std::string controlTableFilename;

  /// The material ID associated with the source
  std::string sourceMaterialId;

  /// Whether this instance is a source
  [[nodiscard]] bool isSource() const { return !sourceMaterialId.empty(); }

  /// An override index to force which geometry population member is used
  std::optional<size_t> geometryPopulationIndex;

  /// The line number the instance came from
  std::optional<size_t> lineNumber;

  [[nodiscard]] static std::shared_ptr<Instance> FromXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});

  void ReadXML(
    const xml::Node& node,
    const xml::Node& dataNode,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});

protected:
  virtual void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) = 0;

private:
  Type type_;
};

inline std::ostream& operator<<(std::ostream& os, const Instance& inst) {
  if (!inst.name.empty()) os << "Name: \"" << inst.name << "\"\n";
  if (!inst.tags.empty()) {
    os << "Tags: \"" << inst.tags[0] << "\"";
    for (size_t i = 1; i < inst.tags.size(); i++) {
      os << ", \"" << inst.tags[i] << "\"";
    }
    os << "\n";
  }
  if (inst.hasTimeWindow()) {
    os << "Time window: \"" << inst.timeWindow << "\"\n";
  }
  if (!inst.controlTableFilename.empty()) {
    os << "Control table: \"" << inst.controlTableFilename << "\"\n";
  }
  if (inst.isSource()) {
    os << "Source, material ID = \"" << inst.sourceMaterialId << "\"\n";
  }
  return os;
}

// #################################
// #      Instance Subclasses      #
// #################################

template <Instance::Type T>
class StaticInstance : public Instance::Subclass<T> {
public:
  /// Tag of the instance this instance is anchored to
  std::string anchor;

  /// Whether the instance should be rotated to the anchor's normal
  bool anchorRotation = false;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& = {}) final {
    if (auto attr = node.attribute("anchor"); attr) {
      anchor = attr.as_string();
      anchorRotation = node.attribute("anchorrotation").as_bool();
    }
    ReallyFinishReadXML(node, facade);
  }

protected:
  virtual void ReallyFinishReadXML(
    const xml::Node& node, const geolib::Facade* facade) = 0;
};

/// Subclass for a translation/scale/rotation instance
class D5_D4LEGACY_EXPORT TsrInstance
    : public StaticInstance<Instance::Type::Tsr> {
public:
  double xTranslate = 0;
  double yTranslate = 0;
  double zTranslate = 0;
  double xScale = 1;
  double yScale = 1;
  double zScale = 1;
  double xRotate = 0;
  double yRotate = 0;
  double zRotate = 0;
  RotationOrder rotationOrder = RotationOrder::Zyx;

private:
  void ReallyFinishReadXML(
    const xml::Node& node, const geolib::Facade* facade) override;
};

/// Subclass for an affine transform instance
class D5_D4LEGACY_EXPORT AffineInstance
    : public StaticInstance<Instance::Type::Affine> {
public:
  /// The affine transform matrix in row-major order.
  std::array<double, 16> affine;

private:
  void ReallyFinishReadXML(
    const xml::Node& node, const geolib::Facade* facade) override;
};

/// Subclass for a delta motion instance
class D5_D4LEGACY_EXPORT DeltaMotionInstance
    : public Instance::Subclass<Instance::Type::DeltaMotion> {
public:
  /// The filename of the motion data, should end in `.mov`.
  std::string filename;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Subclass for a generic motion instance
class D5_D4LEGACY_EXPORT GenericMotionInstance
    : public Instance::Subclass<Instance::Type::GenericMotion> {
public:
  /// The filename of the motion data, should end in `.ppd`.
  std::string filename;

  /// Option to start hidden, which is always inline.
  bool startHidden = false;

  /// Option to end hidden, which is always inline.
  bool endHidden = false;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Suclass for a flex motion instance
class D5_D4LEGACY_EXPORT FlexMotionInstance
    : public Instance::Subclass<Instance::Type::FlexMotion> {
public:
  FlexMotion flexMotion;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Subclass for decal instances
class D5_D4LEGACY_EXPORT DecalInstance
    : public Instance::Subclass<Instance::Type::Decal> {
public:
  std::string materialId;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Simple test for whether the given instance is anchored
inline bool IsAnchored(const Instance& instance) {
  switch (instance.type()) {
    case Instance::Type::Tsr:
      return !instance.as<TsrInstance>().anchor.empty();
    case Instance::Type::Affine:
      return !instance.as<AffineInstance>().anchor.empty();
    default:
      return false;
  }
}

/// \}
}  // namespace d5::d4legacy
