#pragma once

#include <array>
#include <optional>
#include <random>
#include <string>
#include <variant>
#include <vector>

#include "dirsig5/base/Constants.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Pcg.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/geometry/Instance.h"
#include "dirsig5/d4legacy/xml/xml.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_geometry
/// \{

struct D5_D4LEGACY_EXPORT Population {
  struct MaterialVariant {
    std::string id;

    struct DistributionEntry {
      double weight = 1;
      std::string materialId;
    };

    std::vector<DistributionEntry> distribution;

    std::optional<long> seed;
  };

  std::vector<MaterialVariant> materialVariants;

  void ReadXML(const xml::Node& node);
};

struct D5_D4LEGACY_EXPORT RandomFill {
  std::string saveFilename;
  bool loadInstances = false;

  static std::vector<std::shared_ptr<Instance>> ReadSavefile(
    const std::string& filename);

  static void WriteSavefile(
    const std::string& filename,
    const std::vector<std::array<float, 12>>& instances,
    const std::vector<size_t>& geometryIndexes = {});

  struct D5_D4LEGACY_EXPORT Box {
    Point lowerExtent;
    Point upperExtent;
    size_t count;

    /** Generates random locations inside the box
     *
     *  \param rng       The random number generator to be used
     *  \param callback  Callback function to be called with each generated
     *                   sample.
     */
    template <typename RNG, typename Callable>
    void Generate(RNG&& rng, Callable&& callback) const {
      for (size_t i = 0; i < count; i++) {
        auto inst = std::make_shared<TsrInstance>();
        inst->xTranslate = sampleUniform(rng, lowerExtent[0], upperExtent[0]);
        inst->yTranslate = sampleUniform(rng, lowerExtent[1], upperExtent[1]);
        inst->zTranslate = sampleUniform(rng, lowerExtent[2], upperExtent[2]);
        callback(std::move(inst));
      }
    }

    void ReadXML(const xml::Node& node);
  };

  struct D5_D4LEGACY_EXPORT Anchor {
    std::string filename;
    double fillFraction;
    double reverseFraction;
    std::string anchor;
    Point posDev;
    Point rotDevRad;
    double minScale;
    double maxScale;
    std::optional<long> seed{};

    struct AnchorEntry {
      double x;
      double y;
      double rx;
      double ry;
    };

    static std::vector<AnchorEntry> ReadAnchorListFile(
      const std::string& filename);

    /** Generates random locations based on the anchor list
     *
     *  \param anchors   The list of anchors at which to potentially place
     *                   samples.
     *  \param callback  Function to be called with each generated sample.
     */
    template <typename Callable>
    void Generate(
      RandEngine& rng,
      const std::vector<AnchorEntry>& anchors,
      Callable&& callback) const {
      // Process each anchor in the list independently
      for (const auto& a : anchors) {
        // Draw a number to determine if the anchor is occupied
        if (sampleUniform(rng) > this->fillFraction) continue;

        // This is a vector in the orientation of the anchor
        Vector d(a.rx, a.ry, 0);
        d.normalize();

        // Compute the translation (Gaussian offset from anchor point)
        auto inst = std::make_shared<TsrInstance>();
        auto m = makeRotation(Vector(0, 1, 0), Vector(a.rx, a.ry, 0));
        Vector dPos = m * Vector(posDev[0], posDev[1], posDev[2]);
        inst->xTranslate = sampleNormalDistribution(rng, a.x, dPos(0));
        inst->yTranslate = sampleNormalDistribution(rng, a.y, dPos(1));
        inst->zTranslate = sampleNormalDistribution(rng, 0, dPos(2));
        inst->xScale = inst->yScale = inst->zScale =
          sampleUniform(rng, this->minScale, this->maxScale);

        // Compute the rotation
        double rot = getGeoAzimuth(d);
        if (sampleUniform(rng) < this->reverseFraction) {
          rot += (rot < M_PI) ? M_PI : -M_PI;
        }
        inst->xRotate = sampleNormalDistribution(rng, 0, this->rotDevRad[0]);
        inst->yRotate = sampleNormalDistribution(rng, 0, this->rotDevRad[1]);
        inst->zRotate = sampleNormalDistribution(rng, rot, this->rotDevRad[2]);
        inst->anchor = anchor;

        callback(std::move(inst));
      }
    }

    void ReadXML(const xml::Node& node);
  };

  struct D5_D4LEGACY_EXPORT DensityMap {
    /// Tag to anchor fill instances to.
    std::string anchor;

    /// Seed.
    std::optional<long> seed{};

    /// Target number of fill instances.
    size_t count;

    /// Minimum distance between fill instances.
    double minDist;

    /// Maximum distance between fill instances.
    double maxDist;

    /// Minimum scale.
    double minScale;

    /// Maximum scale.
    double maxScale;

    /// Material to associate with this density map.
    std::string materialId;

    std::vector<std::string> moreMaterialIds{};

    /// Random azimuth orientation?
    bool randomOrientation = false;

    void ReadXML(const xml::Node& node);
  };

  struct D5_D4LEGACY_EXPORT PolyFill {
    struct D5_D4LEGACY_EXPORT Polygon {
      /// Center point.
      std::array<double, 2> centerPoint = {0, 0};

      /// Points in CCW order.
      std::vector<std::array<double, 2>> points;

      /// Bound minimum.
      std::array<double, 2> boundMin = {0, 0};

      /// Bound maximum.
      std::array<double, 2> boundMax = {0, 0};

      /// Contains point?
      [[nodiscard]] bool contains(const std::array<double, 2>& point) const;

      /// Calculate area.
      [[nodiscard]] double calculateArea() const;
    };

    /// Read polygon file.
    static std::vector<Polygon> ReadPolyFile(const std::string& filename);

    /// Polygon filename.
    std::string polyFilename;

    /// Random seed.
    std::optional<long> seed{};

    /// Minimum number of fill instances.
    size_t minCount = 0;

    /// Maximum number of fill instances.
    size_t maxCount = 0;

    /// Minimum scale.
    double minScale = 1;

    /// Maximum scale.
    double maxScale = 1;

    /// Material to anchor fill instances to.
    std::string materialId;

    /// Random azimuth orientation?
    bool randomOrientation = false;

    void ReadXML(const xml::Node& node);
  };

  std::variant<Box, Anchor, DensityMap, PolyFill> fill;

  void ReadXML(const xml::Node& node);
};

/// \}
}  // namespace d5::d4legacy
