#pragma once

#include <array>
#include <limits>
#include <ostream>
#include <string>
#include <vector>

#ifndef D4LEGACY_INCLUDED_FROM_GEOMETRY_H
#define D4LEGACY_INCLUDED_FROM_PRIMITIVES_H
#include "dirsig5/d4legacy/geometry/Geometry.h"
#endif

#include "dirsig5/base/Mixin.h"

namespace d5::d4legacy {

/** \defgroup d4legacy_geometry_primitives Primitives
 *  Holds legacy data models for primtive objects
 *  \ingroup d4legacy_geometry
 */

/// \addtogroup d4legacy_geometry_primitives
/// \{

namespace PrimitiveMixin {

/// Mixin for a fixed > 1 number of material IDs
template <class N, class Base>
struct MaterialIds : Base {
  static_assert(N::value >= 2, "N must be >= 2");
  std::array<std::string, N::value> materialIds;
};

/// Mixin for one material ID
template <class Base>
struct MaterialIds<std::integral_constant<std::size_t, 1>, Base> : Base {
  std::string materialId;
};

/// Alias for one material ID
template <typename Base>
using SingleMaterialId =
  MaterialIds<std::integral_constant<std::size_t, 1>, Base>;

/// Max value of a size_t
inline constexpr auto AnyN = std::numeric_limits<size_t>::max();

/// Mixin for an arbitrary number of material IDs
template <class Base>
struct MaterialIds<std::integral_constant<std::size_t, AnyN>, Base> : Base {
  std::vector<std::string> materialIds;
};

/// Alias for an arbitrary number of material IDs
template <typename Base>
using MultipleMaterialIds =
  MaterialIds<std::integral_constant<std::size_t, AnyN>, Base>;

/// Mixin for a Kelvin temperature
template <class Base>
struct TempK : Base {
  double tempK = -1;
};

/// Mixin for a concentration in parts per million [ppm]
template <class Base>
struct ConcentrationPpm : Base {
  double concentrationPpm;
};

/// Mixin for a N concentrations in parts per million and weightings [ppm]
template <class Base>
struct ConcentrationsPpm : Base {
  std::vector<double> concentrationsPpm;
  std::vector<double> weightings;
};

}  // namespace PrimitiveMixin

/// Base class for all primitives, allows passing policy template parameters
template <Geometry::Type T, template <class> class... Mixins>
class D5_D4LEGACY_EXPORT Primitive : public Geometry::Subclass<T>,
                                     public mixin<Mixins...> {
  using mixin_t = mixin<Mixins...>;

public:
  template <template <class> class Mixin>
  static constexpr bool HasMixin() {
    return has_mixin_v<mixin_t, Mixin>;
  }

  static_assert(
    !HasMixin<PrimitiveMixin::ConcentrationsPpm>() ||
      HasMixin<PrimitiveMixin::MultipleMaterialIds>(),
    "If a primitive supports concentrations, it must support multiple "
    "materials");

  static_assert(
    !HasMixin<PrimitiveMixin::ConcentrationPpm>() ||
      HasMixin<PrimitiveMixin::SingleMaterialId>(),
    "If a primitive supports a concetration, it must support only one "
    "material ID");

  // It also needs at least one material, but I cna't figure out how to check
  // that, since it is a template mixin

  void ReadXML(
    const xml::Node& node, const geolib::Facade*, const FileLocator&) final {
    this->lineNumber = xml::lineno(node);
    if constexpr (HasMixin<PrimitiveMixin::SingleMaterialId>())
      this->materialId = xml::requiredChild(node, "matid").text().as_string();
    else {
      // We need specialized logic for SecchiDisk...
      if constexpr (T == Geometry::Type::SecchiDisk)
        this->materialIds = {
          xml::requiredChild(node, "matid_a").text().as_string(),
          xml::requiredChild(node, "matid_b").text().as_string()};
      else
        this->materialIds.push_back(
          xml::requiredChild(node, "matid").text().as_string());
    }

    if constexpr (HasMixin<PrimitiveMixin::ConcentrationPpm>()) {
      double defaultConc = node.attribute("volume_boundary").as_bool() ? 1 : 0;
      this->concentrationPpm =
        xml::readNumber(node, "concentration", defaultConc);
    }
    if constexpr (HasMixin<PrimitiveMixin::ConcentrationsPpm>()) {
      double defaultConc = node.attribute("volume_boundary").as_bool() ? 1 : 0;
      this->concentrationsPpm.push_back(
        xml::readNumber(node, "concentration", defaultConc));
      this->weightings.push_back(1.0);
    }
    if constexpr (HasMixin<PrimitiveMixin::TempK>()) {
      this->tempK = xml::readNumber(node, "temperature", -1);
    }
    FinishReadXML(node);
  }

  void ReadTag(const TagFileSection& section) final {
    this->lineNumber = section.lineNumber();
    if constexpr (HasMixin<PrimitiveMixin::SingleMaterialId>())
      section.requiredVariable("MATERIAL_ID", 0, this->materialId);
    else
      section.requiredVariable("MATERIAL_IDS", 0, this->materialIds);
    if constexpr (HasMixin<PrimitiveMixin::ConcentrationsPpm>()) {
      section.optionalVariable("CONCENTRATIONS", 0, this->concentrationsPpm);
      section.optionalVariable("WEIGHTINGS", 0, this->weightings);
    }
    if constexpr (HasMixin<PrimitiveMixin::TempK>()) {
      section.optionalVariable("TEMPERATURE", 0, this->tempK);
    }
    FinishReadTag(section);
  }

protected:
  virtual void FinishReadXML(const xml::Node& node) = 0;

  virtual void FinishReadTag(const TagFileSection& section) = 0;
};

// ##########################################
// #   is_primitive Type Trait Definition   #
// ##########################################

namespace detail {

/// Metafunction to test whether T is a primitive
template <typename, typename>
struct is_primitive : std::false_type {};

/// \cond HIDDEN_SYMBOLS

// Positive case, only consider types that are subclasses of Geometry::Suclass
template <typename T>
struct is_primitive<
  T,
  std::enable_if_t<
    std::is_base_of_v<Geometry, T> &&
    d5::detail::has_subclass_type_member_v<T>>> {
  // Positive case, T must be able to be upcasted to Primitive, using the
  // suclass_type embedded via Geometry::Suclass and an arbitrary set of
  // mixins
  template <template <class> class... Mixins>
  static decltype(static_cast<const Primitive<T::subclass_type, Mixins...>&>(std::declval<T>()), std::true_type{})
  test(const Primitive<T::subclass_type, Mixins...>&);

  static std::false_type test(...);

public:
  /// Whether T is a primitive or not
  static constexpr bool value =
    decltype(is_primitive::test(std::declval<T>()))::value;
};

/// \endcond
}  // namespace detail

/// Tests whether T is a Primitive subclass
template <typename T>
struct is_primitive {
  static constexpr bool value =
    detail::is_primitive<remove_cvref_t<T>, void>::value;
};

/// Convenience alias for is_primitive
template <typename T>
inline constexpr bool is_primitive_v = is_primitive<T>::value;

// ##########################################
// #   Primitive Object Class Definitions   #
// ##########################################

using BoxPrimT = Primitive<
  Geometry::Type::Box,
  PrimitiveMixin::MultipleMaterialIds,
  PrimitiveMixin::ConcentrationsPpm,
  PrimitiveMixin::TempK>;

/// Represents a box primitive
class D5_D4LEGACY_EXPORT Box : public Primitive<
                                 Geometry::Type::Box,
                                 PrimitiveMixin::MultipleMaterialIds,
                                 PrimitiveMixin::ConcentrationsPpm,
                                 PrimitiveMixin::TempK> {
public:
  /// Lower extent in each axis.
  Point lowerExtent = {0, 0, 0};

  /// Upper extent in each axis.
  Point upperExtent = {0, 0, 0};

private:
  void FinishReadXML(const xml::Node& node) override;

  void FinishReadTag(const TagFileSection& section) override;
};

/// Represents a Catenary curve primitive
class D5_D4LEGACY_EXPORT CatenaryCurve : public Primitive<
                                           Geometry::Type::CatenaryCurve,
                                           PrimitiveMixin::SingleMaterialId> {
public:
  /// Thickness [m]
  double thickness = 0.01;

  /// Sag parameter
  double sag = 20;

  /// First endpoint of the curve
  Point pointA = {-15, 0, 30};

  /// Second endpoint of the curve
  Point pointB = {15, 0, 30};

private:
  void FinishReadXML(const xml::Node& node) override;

  void FinishReadTag(const TagFileSection& section) override;
};

/// Represents a curved frustum primitive
class D5_D4LEGACY_EXPORT CurvedFrustum : public Primitive<
                                           Geometry::Type::CurvedFrustum,
                                           PrimitiveMixin::SingleMaterialId> {
public:
  /// Height of frustum.
  double height = 1;

  /// Center location.
  Point center = {0, 0, 0};

  /// Bottom width in X.
  double bottomXWidth = 10;

  /// Bottom width in Y.
  double bottomYWidth = 10;

  /// Bottom radius.
  double bottomRadius = 5;

  /// Top radius.
  double topRadius = 0;

  /// Z-rotation.
  double zRotation = 0;

  /// Ramp length.
  double rampLength = 0;

private:
  void FinishReadXML(const xml::Node& node) override;

  void FinishReadTag(const TagFileSection& section) override;
};

/// Represents a cylinder primitive
class D5_D4LEGACY_EXPORT Cylinder : public Primitive<
                                      Geometry::Type::Cylinder,
                                      PrimitiveMixin::SingleMaterialId,
                                      PrimitiveMixin::ConcentrationPpm,
                                      PrimitiveMixin::TempK> {
public:
  /// Radius.
  double radius = 1;

  /// Point A.
  Point pointA = {0, 0, -0.5};

  /// Point B.
  Point pointB = {0, 0, 0.5};

  /// Show cap at point A?
  bool capA = true;

  /// Show cap at point B?
  bool capB = true;

private:
  void FinishReadXML(const xml::Node& node) override;

  void FinishReadTag(const TagFileSection& section) override;
};

/// Represents a sphere primitive
class D5_D4LEGACY_EXPORT Sphere : public Primitive<
                                    Geometry::Type::Sphere,
                                    PrimitiveMixin::MultipleMaterialIds,
                                    PrimitiveMixin::ConcentrationsPpm,
                                    PrimitiveMixin::TempK> {
public:
  /// Radius.
  double radius = 1;

  /// Center.
  Point center = {0, 0, 0};

private:
  void FinishReadXML(const xml::Node& node) override;

  void FinishReadTag(const TagFileSection& section) override;
};

/// Represents a sinusoid primtive
class D5_D4LEGACY_EXPORT Sinusoid : public Primitive<
                                      Geometry::Type::Sinusoid,
                                      PrimitiveMixin::SingleMaterialId> {
public:
  /// Lower extent in each axis.
  Point lowerExtent = {0, 0, 0};

  /// Upper extent in each axis.
  Point upperExtent = {0, 0, 0};

  /// Magnitude.
  double magnitude = 1;

  /// Period.
  double period = 1;

  /// Orientation in XY.
  Point orientation = {1, 0, 0};

  /// Phase shift.
  double phaseShift = 0;

private:
  void FinishReadXML(const xml::Node& node) override;

  void FinishReadTag(const TagFileSection& section) override;
};

/// Represents a Secchi disk primitive
class D5_D4LEGACY_EXPORT SecchiDisk
    : public Primitive<
        Geometry::Type::SecchiDisk,
        template_mixin<
          PrimitiveMixin::MaterialIds,
          std::integral_constant<std::size_t, 2>>::mixin,
        PrimitiveMixin::TempK> {
public:
  /// Radius.
  double radius = 1;

  /// Center.
  Point center = {0, 0, 0};

  /// Normal direction.
  Point normal = {0, 0, 1};

private:
  void FinishReadXML(const xml::Node& node) override;

  void FinishReadTag(const TagFileSection& section) override;
};

/// Represents a disk primitive
class D5_D4LEGACY_EXPORT Disk : public Primitive<
                                  Geometry::Type::Disk,
                                  PrimitiveMixin::SingleMaterialId,
                                  PrimitiveMixin::TempK> {
public:
  /// Radius.
  double radius = 1;

  /// Center.
  Point center = {0, 0, 0};

  /// Normal direction.
  Point normal = {0, 0, 1};

private:
  void FinishReadXML(const xml::Node& node) override;

  void FinishReadTag(const TagFileSection& section) override;
};

/// Represents a surface plane primitive
class D5_D4LEGACY_EXPORT SurfacePlane : public Primitive<
                                          Geometry::Type::SurfacePlane,
                                          PrimitiveMixin::SingleMaterialId,
                                          PrimitiveMixin::TempK> {
public:
  /// Anchor point.
  Point anchor = {0, 0, 0};

  /// Has checkers?
  bool checkers = false;

  /// If has checkers, checker material ID.
  std::string checkersMaterialId;

  /// If has checkers, checker width.
  double checkerWidth = 1;

  /// Overrides for a specific checker
  struct CheckerOverride {
    /// X-index for the checker
    int xIndex;

    /// Y-index for the checker
    int yIndex;

    /// The material ID to override with
    std::string materialId;
  };

  /// Checker material overrides
  std::vector<CheckerOverride> checkerOverrides;

private:
  void FinishReadXML(const xml::Node& node) override;

  void FinishReadTag(const TagFileSection& section) override;
};

/// \}
}  // namespace d5::d4legacy

#ifndef D4LEGACY_INCLUDED_FROM_GEOMETRY_H
#ifndef D4LEGACY_INCLUDED_FROM_GEOMETRY_DISPATCH_H
#include "detail/GeometryDispatch.h"
#endif
#endif
