#pragma once

#include <cstdint>
#include <stdexcept>
#include <string>

#include <Eigen/Dense>

namespace d5::d4legacy {

/// \addtogroup d4legacy_geometry
/// \{

/// Supported Euler angle rotation orders
enum class RotationOrder : uint8_t { Xyz, Xzy, Yxz, Yzx, Zxy, Zyx };

/// Parses a RotationOrder from a string
inline RotationOrder ParseRotationOrder(const std::string& s) {
  if (s == "xyz") return RotationOrder::Xyz;
  if (s == "xzy") return RotationOrder::Xzy;
  if (s == "yxz") return RotationOrder::Yxz;
  if (s == "yzx") return RotationOrder::Yzx;
  if (s == "zxy") return RotationOrder::Zxy;
  if (s == "zyx") return RotationOrder::Zyx;
  throw std::invalid_argument("unrecognized rotation order: " + s);
}

/// Converts a rotation order to a string
inline std::string to_string(RotationOrder o) {
  switch (o) {
    case d5::d4legacy::RotationOrder::Xyz:
      return "xyz";
    case d5::d4legacy::RotationOrder::Xzy:
      return "xzy";
    case d5::d4legacy::RotationOrder::Yxz:
      return "yxz";
    case d5::d4legacy::RotationOrder::Yzx:
      return "yzx";
    case d5::d4legacy::RotationOrder::Zxy:
      return "zxy";
    case d5::d4legacy::RotationOrder::Zyx:
      return "zyx";
    default:
      throw std::invalid_argument("invalid rotation order");
  }
}

/// Makes a rotation matrix
inline Eigen::Matrix3d MakeRotationMatrix(
  double rx, double ry, double rz, RotationOrder order) {
  auto Rx = Eigen::AngleAxisd(rx, Eigen::Vector3d::UnitX()),
       Ry = Eigen::AngleAxisd(ry, Eigen::Vector3d::UnitY()),
       Rz = Eigen::AngleAxisd(rz, Eigen::Vector3d::UnitZ());
  switch (order) {
    case d5::d4legacy::RotationOrder::Xyz:
      return Rz * Ry * Rx * Eigen::Matrix3d::Identity();
    case d5::d4legacy::RotationOrder::Xzy:
      return Ry * Rz * Rx * Eigen::Matrix3d::Identity();
    case d5::d4legacy::RotationOrder::Yxz:
      return Rz * Rx * Ry * Eigen::Matrix3d::Identity();
    case d5::d4legacy::RotationOrder::Yzx:
      return Rx * Rz * Ry * Eigen::Matrix3d::Identity();
    case d5::d4legacy::RotationOrder::Zxy:
      return Ry * Rx * Rz * Eigen::Matrix3d::Identity();
    case d5::d4legacy::RotationOrder::Zyx:
      return Rx * Ry * Rz * Eigen::Matrix3d::Identity();
    default:
      throw std::invalid_argument("invalid rotation order");
  }
}

/// \}
}  // namespace d5::d4legacy
