/** Interface file for bulk properties
 *
 *  \file BulkProperties.h
 */

#pragma once

#include <memory>
#include <optional>
#include <stdexcept>
#include <string>
#include <utility>
#include <vector>

#include "dirsig5/base/Inheritance.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/mat/SimpleSpectralData.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_mat
/// \{

// ###############################################
// #     AbsorptionProperty Class Definition     #
// ###############################################

/// Data model for absorption properties
class D5_D4LEGACY_EXPORT AbsorptionProperty
    : public EnumeratedSuperclass<AbsorptionProperty> {
public:
  /// Types of supported absorption properties
  enum class Type {
    None = 0,
    ClassicAbsorption,
    SimpleAbsorption,
    VoxelizedAbsorption
  };

  /// Virtual destructor
  virtual ~AbsorptionProperty() = default;

  /// Returns the type of property
  [[nodiscard]] Type type() const { return type_; }

  /// Prints the property to a stream
  virtual void PrintToStream(std::ostream&) const {}

protected:
  explicit AbsorptionProperty(Type type) : type_(type) {}

private:
  Type type_;
};

// ###############################################
// #     ScatteringProperty Class Definition     #
// ###############################################

/// Data model for scattering properties
class D5_D4LEGACY_EXPORT ScatteringProperty
    : public EnumeratedSuperclass<ScatteringProperty> {
public:
  /// Types of supported scattering properties
  enum class Type { None = 0, SimpleScattering };

  /// Virtual destructor
  virtual ~ScatteringProperty() = default;

  /// Returns the type of property
  [[nodiscard]] Type type() const { return type_; }

  /// Prints the property to a stream
  virtual void PrintToStream(std::ostream&) const {}

protected:
  explicit ScatteringProperty(Type type) : type_(type) {}

private:
  Type type_;
};

// ##################################################
// #     PhaseFunctionProperty Class Definition     #
// ##################################################

/// Data model for phase function properties
class D5_D4LEGACY_EXPORT PhaseFunctionProperty
    : public EnumeratedSuperclass<PhaseFunctionProperty> {
public:
  /// Types of supported phase function properties
  enum class Type { None = 0, HenyeyGreenstein, Tabulated };

  /// Virtual destructor
  virtual ~PhaseFunctionProperty() = default;

  /// Returns the type of property
  [[nodiscard]] Type type() const { return type_; }

  /// Prints the property to a stream
  virtual void PrintToStream(std::ostream&) const {}

protected:
  explicit PhaseFunctionProperty(Type type) : type_(type) {}

private:
  Type type_;
};

// ###########################################
// #     BulkProperties Class Definition     #
// ###########################################

struct BulkProperties {
  std::unique_ptr<AbsorptionProperty> absorption;
  std::unique_ptr<ScatteringProperty> scattering;
  std::unique_ptr<PhaseFunctionProperty> phaseFunction;
};

// #########################################
// #     AbsorptionProperty Subclasses     #
// #########################################

/// Classic absorption data model
class D5_D4LEGACY_EXPORT ClassicAbsorption
    : public AbsorptionProperty::Subclass<
        AbsorptionProperty::Type::ClassicAbsorption> {
public:
  ~ClassicAbsorption() override = default;

  /// The absorption filename (.abs or .gas)
  std::string filename;
};

/// Simple absorption data model
class D5_D4LEGACY_EXPORT SimpleAbsorption
    : public AbsorptionProperty::Subclass<
        AbsorptionProperty::Type::SimpleAbsorption> {
public:
  ~SimpleAbsorption() override = default;

  /// Filename of the simple spectral data
  std::string filename;

  /// Format of filename
  SimpleSpectralData::Format format = SimpleSpectralData::Format::Unknown;

  /// Whether to use concentration scaling
  bool useConcentrationScaling = false;

  /// The broadband solar absorption
  std::optional<double> solarAbsorption;
};

/// Voxelized absorption data model
class D5_D4LEGACY_EXPORT VoxelizedAbsorption
    : public AbsorptionProperty::Subclass<
        AbsorptionProperty::Type::VoxelizedAbsorption> {
public:
  ~VoxelizedAbsorption() override = default;

  /// Parameter filename
  std::string filename;
};

// #########################################
// #     ScatteringProperty Subclasses     #
// #########################################

/// Simple absorption data model
class D5_D4LEGACY_EXPORT SimpleScattering
    : public ScatteringProperty::Subclass<
        ScatteringProperty::Type::SimpleScattering> {
public:
  ~SimpleScattering() override = default;

  /// Filename of the simple spectral data
  std::string filename;

  /// Format of filename
  SimpleSpectralData::Format format = SimpleSpectralData::Format::Unknown;

  /// Whether to use concentration scaling
  bool useConcentrationScaling = false;
};

// ############################################
// #     PhaseFunctionProperty Subclasses     #
// ############################################

/// Henyey-Greenstein phase function data model
class HenyeyGreensteinPhaseFunction
    : public PhaseFunctionProperty::Subclass<
        PhaseFunctionProperty::Type::HenyeyGreenstein> {
public:
  ~HenyeyGreensteinPhaseFunction() override = default;

  /// Filename of the simple spectral data
  std::string filename;

  /// Format of filename
  SimpleSpectralData::Format format = SimpleSpectralData::Format::Unknown;
};

/// Tabulated phase function data model
class TabulatedPhaseFunction : public PhaseFunctionProperty::Subclass<
                                 PhaseFunctionProperty::Type::Tabulated> {
public:
  ~TabulatedPhaseFunction() override = default;

  /// Filename of the parameter data
  std::string filename;
};

/// \}
}  // namespace d5::d4legacy
