/** Interface file for EmissivityProperty
 *
 * \file EmissivityProperty.h
 */

#pragma once

#include <iostream>
#include <optional>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <vector>

#include "dirsig5/base/Inheritance.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/scene/Maps.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_mat
/// \{

/// Base class for DIRSIG4 emissivity properties
class D5_D4LEGACY_EXPORT EmissivityProperty
    : public EnumeratedSuperclass<EmissivityProperty> {
public:
  /// Various types of supported emissivity properties
  enum class Type { None = 0, ClassicEmissivity = 1 };

  /// Virtual destructor
  virtual ~EmissivityProperty() = default;

  /// Returns the type of the property
  [[nodiscard]] Type type() const { return type_; }

  /// Prints the material to a stream
  virtual void PrintToStream(std::ostream&) const {}

protected:
  explicit EmissivityProperty(Type type) : type_(type) {}

private:
  Type type_ = Type::None;
};

/// ClassicEmissivity material data model
class D5_D4LEGACY_EXPORT ClassicEmissivity final
    : public EmissivityProperty::Subclass<
        EmissivityProperty::Type::ClassicEmissivity> {
public:
  /// Destructor
  ~ClassicEmissivity() override = default;

  /// Filename for the emissivity data
  std::string filename;

  /// The optional texture map specification for the material
  std::optional<d5::d4legacy::TextureMapEntry> textureMap;

  void PrintToStream(std::ostream& os) const override {
    os << "ClassicEmissivity: " << filename << "\n";
  }
};

/// Emissivity property type to string.
inline std::string to_string(EmissivityProperty::Type type) {
  switch (type) {
    case EmissivityProperty::Type::None:
      return "None";
    case EmissivityProperty::Type::ClassicEmissivity:
      return "ClassicEmissivity";
    default:
      break;
  }
  return "Unknown";
}

/// \}
}  // namespace d5::d4legacy
