/** Interface file for ExtinctionProperty
 *
 * \file ExtinctionProperty.h
 */

#pragma once

#include <iostream>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <variant>
#include <vector>

#include "dirsig5/base/Inheritance.h"
#include "dirsig5/base/span.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/mat/SimpleSpectralData.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_mat
/// \{

/// Base class for DIRSIG4 extinction properties
class D5_D4LEGACY_EXPORT ExtinctionProperty
    : public EnumeratedSuperclass<ExtinctionProperty> {
public:
  /// Various types of supported extinction properties
  enum class Type {
    None = 0,
    WardBtdf,
    ClassicExtinction,
    SimpleTransmission,
    PMFresnelBtdf,
    SphericalDataBtdf,
    DeltaTransmission
  };

  /// Virtual destructor
  virtual ~ExtinctionProperty() = default;

  /// Returns the type of the property
  [[nodiscard]] Type type() const { return type_; }

  /// Prints the material to a stream
  virtual void PrintToStream(std::ostream&) const {}

protected:
  /** Constructor
   *
   *  \param type  The type of property
   */
  explicit ExtinctionProperty(Type type) : type_(type) {}

private:
  Type type_ = Type::None;
};

/// WardBTDF material data model
class D5_D4LEGACY_EXPORT WardBtdf final
    : public ExtinctionProperty::Subclass<ExtinctionProperty::Type::WardBtdf> {
public:
  /// Destructor
  ~WardBtdf() override = default;

  // The diffuse weight of the BRDF
  float diffuseWeight = 0;

  // The specular weight of the BRDF
  float specularWeight = 0;

  // The width of the specular BRDF in the parallel direction
  float specularSigmaParallel = 0;

  // The width of the specular BRDF in the orthogonal direction
  float specularSigmaOrthogonal = 0;

  void PrintToStream(std::ostream& os) const override {
    os << "WardBTDF:\n  Diffuse = " << diffuseWeight
       << "\n  Specular = " << specularWeight
       << "\n  Specular sigma (parallel) = " << specularSigmaParallel
       << "\n  Specular sigma (orthogonal) = " << specularSigmaOrthogonal
       << "\n";
  }
};

/// ClassicExtinction material data model
class D5_D4LEGACY_EXPORT ClassicExtinction final
    : public ExtinctionProperty::Subclass<
        ExtinctionProperty::Type::ClassicExtinction> {
public:
  struct ExtinctionData {
    std::vector<float> wavelengthsMicron;
    std::vector<float> extinctionsInvKm;
  };

  /// Destructor
  ~ClassicExtinction() override = default;

  /// Filename for the .ext file holding the extinction data
  std::string filename;

  void PrintToStream(std::ostream& os) const override {
    os << "ClassicExtinction: " << filename << "\n";
  }
};

D5_D4LEGACY_EXPORT ClassicExtinction::ExtinctionData ReadExtinctionFile(
  const char* filename, span<const float> wavelengthsScene = {});

/// SimpleTransmission material data model
class D5_D4LEGACY_EXPORT SimpleTransmission final
    : public ExtinctionProperty::Subclass<
        ExtinctionProperty::Type::SimpleTransmission> {
public:
  /// Destructor
  ~SimpleTransmission() override = default;

  /// Filename of the simple spectral data
  std::string filename;

  /// Format of the data
  SimpleSpectralData::Format format;

  std::variant<int, std::string> sliIndex = -1;

  void PrintToStream(std::ostream& os) const override {
    os << "SimpleTransmission:\n  Filename: " << filename << "\n";
  }
};

class D5_D4LEGACY_EXPORT DeltaTransmission final
    : public ExtinctionProperty::Subclass<
        ExtinctionProperty::Type::DeltaTransmission> {
public:
  /// Filename of the simple spectral data
  std::string filename;

  /// Format of the data
  SimpleSpectralData::Format format;

  std::variant<int, std::string> sliIndex = -1;

  void PrintToStream(std::ostream& os) const override {
    os << "DeltaTransmission:\n  Filename: " << filename << "\n";
  }
};

/// SimpleTransmission material data model
class D5_D4LEGACY_EXPORT PMFresnelBtdf final
    : public ExtinctionProperty::Subclass<
        ExtinctionProperty::Type::PMFresnelBtdf> {
public:
  /// Destructor
  ~PMFresnelBtdf() override = default;

  void PrintToStream(std::ostream& os) const override {
    os << "PMFresnelBtdf\n";
  }
};

/// Arbitrary SQT BRDF data model
class D5_D4LEGACY_EXPORT SphericalDataBtdf final
    : public ExtinctionProperty::Subclass<
        ExtinctionProperty::Type::SphericalDataBtdf> {
public:
  /// Virtual destructor
  ~SphericalDataBtdf() override = default;

  /// The SQT filename
  std::string sqtFilename;

  /// The DHR filename
  std::string dhrFilename;

  bool normalizeDhr = false;

  /// \override
  void PrintToStream(std::ostream& os) const override {
    os << "SphericalDataBrdf:"
       << "\n  SQT filename: " << sqtFilename << "\n";
  }
};

/// Extinction property type to string.
inline std::string to_string(ExtinctionProperty::Type type) {
  switch (type) {
    case ExtinctionProperty::Type::None:
      return "None";
    case ExtinctionProperty::Type::ClassicExtinction:
      return "ClassicExtinction";
    case ExtinctionProperty::Type::SimpleTransmission:
      return "SimpleTransmission";
    case ExtinctionProperty::Type::PMFresnelBtdf:
      return "PMFresnelBtdf";
    case ExtinctionProperty::Type::SphericalDataBtdf:
      return "SphericalDataBtdf";
    case ExtinctionProperty::Type::DeltaTransmission:
      return "DeltaTransmission";
    default:
      break;
  }
  return "Unknown";
}

/// \}
}  // namespace d5::d4legacy
