/** Interface file for MatFile
 *
 * \file MatFile.h
 */

#pragma once

#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/mat/MatFileEntry.h"
#include "dirsig5/d4legacy/misc/FileLocator.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_mat
/// \{

/// A material file.
class D5_D4LEGACY_EXPORT MatFile {
public:
  /// Default constructible.
  MatFile() = default;

  /// Non-copyable.
  MatFile(const MatFile&) = delete;

  /// Default movable.
  MatFile(MatFile&&) = default;

  /// Non-copyable.
  MatFile& operator=(const MatFile&) = delete;

  /// Default movable.
  MatFile& operator=(MatFile&&) = default;

  /// Read-write access to the list of entries.
  std::vector<MatFileEntry>& entries() { return entries_; }

  /// Read-only access to the list of entries
  [[nodiscard]] const std::vector<MatFileEntry>& entries() const {
    return entries_;
  }

  /// Read-write access to the list of includes.
  std::vector<std::string>& includes() { return includes_; }

  /// Read-only access to the list of includes.
  [[nodiscard]] const std::vector<std::string>& includes() const {
    return includes_;
  }

  /// Error mode.
  enum class ErrorMode : uint32_t {
    Default = 0,
    PermissiveBit = 0x0001,
    Permissive = PermissiveBit
  };

  /// Bit-or operator.
  friend constexpr ErrorMode operator|(ErrorMode lhs, ErrorMode rhs) {
    return ErrorMode(uint32_t(lhs) | uint32_t(rhs));
  }

  /// Bit-and operator.
  friend constexpr ErrorMode operator&(ErrorMode lhs, ErrorMode rhs) {
    return ErrorMode(uint32_t(lhs) & uint32_t(rhs));
  }

  /** Load a .mat file from a file.
   *
   *  \param[in] filename     A filename.
   *  \param[in] specificId   A specific ID to load, or empty to load all
   *                          entries.
   *  \param[in] errorMode    Error mode.
   *  - If `errorMode` is `ErrorMode::Default`, then a runtime error
   *    is thrown with a descriptive message when an error is encountered.
   *
   *  - If `errorMode` is `ErrorMode::Permissive`, then the implementation
   *    dumps any messages which would be thrown under `ErrorMode::Default`
   *    to `std::cerr` as warnings, and any material entry which fails
   *    to load is skipped.
   *
   *  \note This function returns null only if `errorMode` is
   *        `ErrorMode::Permissive` and the implementation fails to open
   *        `filename`.
   */
  static std::unique_ptr<MatFile> LoadFromFile(
    const std::string& filename,
    const std::string& specificId = {},
    ErrorMode errorMode = ErrorMode::Default);

  /// Output operator.
  friend std::ostream& operator<<(std::ostream& os, const MatFile& matFile) {
    for (const auto& entry : matFile.entries()) {
      os << entry;
    }
    return os;
  }

private:
  std::vector<MatFileEntry> entries_;
  std::vector<std::string> includes_;
};

/// \}
}  // namespace d5::d4legacy
