/** Interface file for MatFileEntry
 *
 * \file MatFileEntry.h
 */

#pragma once

#include <memory>
#include <optional>
#include <sstream>

#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/mat/BulkProperties.h"
#include "dirsig5/d4legacy/mat/EmissivityProperty.h"
#include "dirsig5/d4legacy/mat/ExtinctionProperty.h"
#include "dirsig5/d4legacy/mat/ReflectanceProperty.h"
#include "dirsig5/d4legacy/mat/TemperatureSolver.h"
#include "dirsig5/d4legacy/scene/Maps.h"
#include "dirsig5/d4legacy/sources/SourceInfo.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_mat
/// \{

/// A data model for the surface properties section in the .mat file.
class D5_D4LEGACY_EXPORT SurfaceProperties {
public:
  /// Default constructible.
  SurfaceProperties() = default;

  /// Non-copyable.
  SurfaceProperties(const SurfaceProperties&) = delete;

  /// Default movable.
  SurfaceProperties(SurfaceProperties&&) = default;

  /// Non-copyable.
  SurfaceProperties& operator=(const SurfaceProperties&) = delete;

  /// Default movable.
  SurfaceProperties& operator=(SurfaceProperties&&) = default;

  /// Has reflectance property?
  [[nodiscard]] bool hasReflectanceProperty() const {
    return reflProp_ != nullptr;
  }

  /// Has reflectance property of type?
  [[nodiscard]] bool hasReflectanceProperty(ReflectanceProperty::Type t) const {
    return hasReflectanceProperty() && reflProp_->type() == t;
  }

  /// Get read-only reference to reflectance property.
  [[nodiscard]] const ReflectanceProperty& reflectanceProperty() const {
    return *reflProp_;
  }

  /// Set reflectance property.
  void setReflectanceProperty(std::unique_ptr<ReflectanceProperty> prop) {
    reflProp_ = std::move(prop);
  }

  /// Has emissivity property?
  [[nodiscard]] bool hasEmissivityProperty() const {
    return emsProp_ != nullptr;
  }

  /// Has emissivity property of type?
  [[nodiscard]] bool hasEmissivityProperty(EmissivityProperty::Type t) const {
    return hasEmissivityProperty() && emsProp_->type() == t;
  }

  /// Get read-only reference to emissivity property.
  [[nodiscard]] const EmissivityProperty& emissivityProperty() const {
    return *emsProp_;
  }

  /// Get read-write reference to emissivity property.
  [[nodiscard]] EmissivityProperty& emissivityProperty() { return *emsProp_; }

  /// Set emissivity property.
  void setEmissivityProperty(std::unique_ptr<EmissivityProperty> prop) {
    emsProp_ = std::move(prop);
  }

  /// Has extinction property?
  [[nodiscard]] bool hasExtinctionProperty() const {
    return extProp_ != nullptr;
  }

  /// Has extinction property of type?
  [[nodiscard]] bool hasExtinctionProperty(ExtinctionProperty::Type t) const {
    return hasExtinctionProperty() && extProp_->type() == t;
  }

  /// Get read-only reference to extinction property.
  [[nodiscard]] const ExtinctionProperty& extinctionProperty() const {
    return *extProp_;
  }

  /// Set extinction property.
  void setExtinctionProperty(std::unique_ptr<ExtinctionProperty> prop) {
    extProp_ = std::move(prop);
  }

  /// Stream output operator.
  friend std::ostream& operator<<(
    std::ostream& os, const SurfaceProperties& props) {
    if (props.hasReflectanceProperty())
      props.reflectanceProperty().PrintToStream(os);
    if (props.hasEmissivityProperty())
      props.emissivityProperty().PrintToStream(os);
    if (props.hasExtinctionProperty())
      props.extinctionProperty().PrintToStream(os);
    return os;
  }

public:
  double reflPropWeight = 1;
  double tranPropWeight = 1;

private:
  std::unique_ptr<ReflectanceProperty> reflProp_;
  std::unique_ptr<EmissivityProperty> emsProp_;
  std::unique_ptr<ExtinctionProperty> extProp_;
};

/// Types of rad solvers.
enum class RadSolverType {
  Unknown = 0,
  Null,
  Classic,
  Simple,
  Mirror,
  Generic,
  Fresnel
};

/// Definition of a normal map entry
struct NormalMapEntry {
  /// Name of the map
  std::string name;

  /// Whether the map is enabled
  bool enabled = true;

  /// The (optional) scale to apply
  float scale = 1.0;

  /// The projector for the map
  Projector projector;

  /// The filename for the map
  std::string imageFilename;
};

/// An entry in a DIRSIG4 material (.mat) file.
class D5_D4LEGACY_EXPORT MatFileEntry {
public:
  /// Default constructible.
  MatFileEntry() {
    surfaceProps_.emplace_back(std::make_unique<SurfaceProperties>());
  }

  /// Non-copyable.
  MatFileEntry(const MatFileEntry&) = delete;

  /// Default movable.
  MatFileEntry(MatFileEntry&&) = default;

  /// Non-copyable.
  MatFileEntry& operator=(const MatFileEntry&) = delete;

  /// Default movable.
  MatFileEntry& operator=(MatFileEntry&&) = default;

  /// Gets the ID of the material.
  [[nodiscard]] const std::string& id() const { return id_; }

  /// Sets the ID of the material.
  void setId(const std::string& id) { id_ = id; }

  /// Gets the name of the material.
  [[nodiscard]] const std::string& name() const { return name_; }

  /// Sets the name of the material.
  void setName(const std::string& name) { name_ = name; }

  /// The author of the material
  std::string author;

  /// The organization responsible for the material
  std::string organization;

  /// The description of the material
  std::string description;

  /// Gets the rad solver type.
  [[nodiscard]] RadSolverType radSolverType() const { return radSolverType_; }

  /// Sets the rad solver type.
  void setRadSolverType(RadSolverType type) { radSolverType_ = type; }

  /// Gets the thickness of the material.
  [[nodiscard]] float thicknessCm() const {
    if (thicknessCm_ != -1) return thicknessCm_;
    if (tempSolver_->type() == TemperatureSolver::Type::Therm) {
      return static_cast<float>(
        tempSolver_->as<ThermTemperatureSolver>().material.thickness);
    }
    throw std::runtime_error(
      "requested a thickness, but material entry has none");
  }

  /// Sets the thickness of the material.
  void setThicknessCm(float cm) { thicknessCm_ = cm; }

  [[nodiscard]] std::size_t numSurfaceProperties() const {
    return surfaceProps_.size();
  }

  /// Gets read-only access to the surface properties
  [[nodiscard]] const SurfaceProperties& surfaceProperties(size_t i = 0) const {
    return *surfaceProps_[i];
  }

  /// Gets read-write access to the surface properties.
  SurfaceProperties& surfaceProperties(size_t i = 0) {
    while (surfaceProps_.size() <= i)
      surfaceProps_.emplace_back(new SurfaceProperties());
    return *surfaceProps_[i];
  }

  /// The bulk properties (if any)
  std::optional<BulkProperties> bulkProperties;

  /// Gets non-owning access to the temperature solver. May be null.
  TemperatureSolver* tempSolver() { return tempSolver_.get(); }

  /// Gets non-owning access to the temperature solver. May be null.
  [[nodiscard]] const TemperatureSolver* tempSolver() const {
    return tempSolver_.get();
  }

  /// Sets the temperature solver.
  void setTempSolver(std::unique_ptr<TemperatureSolver> ts) {
    tempSolver_ = std::move(ts);
  }

  /// Whether the entry has a bump map
  [[nodiscard]] bool hasBumpMap() const { return bumpMap_ != nullptr; }

  /// The bump map (call hasBumpMap() first)
  [[nodiscard]] const BumpMapEntry& bumpMap() const { return *bumpMap_; }

  /// Sets the bump map
  void setBumpMap(std::unique_ptr<BumpMapEntry> map) {
    bumpMap_ = std::move(map);
  }

  /// Whether the entry has a density map
  [[nodiscard]] bool hasDensityMap() const { return densityMap_ != nullptr; }

  /// The density map (call hasDensityMap() first)
  [[nodiscard]] const DensityMapEntry& densityMap() const {
    return *densityMap_;
  }

  /// Sets the density map
  void setDensityMap(std::unique_ptr<DensityMapEntry> map) {
    densityMap_ = std::move(map);
  }

  /// Whether the entry has a material map
  [[nodiscard]] bool hasNormalMap() const { return normalMap_ != nullptr; }

  /// The material map (call hasMaterialMap() first)
  [[nodiscard]] const NormalMapEntry& normalMap() const { return *normalMap_; }

  /// Sets the material map
  void setNormalMap(std::unique_ptr<NormalMapEntry> map) {
    normalMap_ = std::move(map);
  }

  /// Whether the entry has a material map
  [[nodiscard]] bool hasMaterialMap() const { return matMap_ != nullptr; }

  /// The material map (call hasMaterialMap() first)
  [[nodiscard]] const MaterialMapEntry& materialMap() const { return *matMap_; }

  /// Sets the material map
  void setMaterialMap(std::unique_ptr<MaterialMapEntry> map) {
    matMap_ = std::move(map);
  }

  /// Whether the entry has a material map
  [[nodiscard]] bool hasMixtureMap() const { return mixMap_ != nullptr; }

  /// The material map (call hasMaterialMap() first)
  [[nodiscard]] const MixtureMapEntry& mixtureMap() const { return *mixMap_; }

  /// Sets the material map
  void setMixtureMap(std::unique_ptr<MixtureMapEntry> map) {
    mixMap_ = std::move(map);
  }

  /// Whether this is a dummy entry (no surface/map/bulk properties)
  [[nodiscard]] bool isDummy() const {
    return (surfaceProps_.empty() or
            (surfaceProps_.size() == 1 and
             !surfaceProps_[0]->hasReflectanceProperty() and
             !surfaceProps_[0]->hasEmissivityProperty() and
             !surfaceProps_[0]->hasExtinctionProperty())) and
           !bulkProperties and !hasMaterialMap() and !hasMixtureMap() and
           (radSolverType_ == RadSolverType::Unknown or
            radSolverType_ == RadSolverType::Null) and
           !source and !bundleFilename;
  }

  /// The bundle filename
  std::optional<std::string> bundleFilename;

  /// The source info (if there is any)
  std::optional<SourceInfo> source;

  /// Stream out.
  friend std::ostream& operator<<(std::ostream& os, const MatFileEntry& entry) {
    os << "Material:\n";
    os << "  ID = " << entry.id() << "\n";
    os << "  NAME = " << entry.name() << "\n";
    std::stringstream ss;
    std::string line;
    ss << entry.surfaceProperties();
    while (std::getline(ss, line)) os << "  " << line << "\n";
    if (auto ts = entry.tempSolver()) {
      ss.clear();
      ts->PrintToStream(ss);
      while (std::getline(ss, line)) os << "  " << line << "\n";
    }
    return os << std::endl;
  }

private:
  std::string id_;
  std::string name_;
  RadSolverType radSolverType_ = RadSolverType::Unknown;
  std::unique_ptr<TemperatureSolver> tempSolver_;
  float thicknessCm_ = -1;
  std::vector<std::unique_ptr<SurfaceProperties>> surfaceProps_;
  std::unique_ptr<BumpMapEntry> bumpMap_;
  std::unique_ptr<DensityMapEntry> densityMap_;
  std::unique_ptr<NormalMapEntry> normalMap_;
  std::unique_ptr<MaterialMapEntry> matMap_;
  std::unique_ptr<MixtureMapEntry> mixMap_;
};

/// \}
}  // namespace d5::d4legacy
