/** Interface file for ReflectanceProperty
 *
 * \file ReflectanceProperty.h
 */

#pragma once

#include <array>
#include <iostream>
#include <memory>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <variant>
#include <vector>

#include "dirsig5/base/Inheritance.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/mat/EmissivityData.h"
#include "dirsig5/d4legacy/mat/SimpleSpectralData.h"
#include "dirsig5/d4legacy/scene/Maps.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_mat
/// \{

/// Base class for DIRSIG4 reflectance properties
class D5_D4LEGACY_EXPORT ReflectanceProperty
    : public EnumeratedSuperclass<ReflectanceProperty> {
public:
  /// Types of supported reflectance properties
  enum class Type {
    None = 0,
    WardBrdf,
    SimpleReflectance,
    PriestGermer,
    ShellTarget,
    ShellBackground,
    ReflectanceMap,
    RgbImage,
    SphericalDataBrdf,
    PMFresnelBrdf,
    RossLi,
    DeltaReflectance,
    CurveMap
  };

  /// Virtual destructor
  virtual ~ReflectanceProperty() = default;

  /// Returns the type of parameter
  [[nodiscard]] Type type() const { return type_; }

  /// Prints the material to a stream
  virtual void PrintToStream(std::ostream&) const {}

protected:
  explicit ReflectanceProperty(Type type) : type_(type) {}

private:
  Type type_ = Type::None;
};

/// WardBRDF material data model
class D5_D4LEGACY_EXPORT WardBrdf final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::WardBrdf> {
public:
  /// Destructor
  ~WardBrdf() override = default;

  // The diffuse weight of the BRDF
  float diffuseWeight = 0;

  // The specular weight of the BRDF
  float specularWeight = 0;

  // The width of the specular BRDF in the parallel direction
  float specularSigmaParallel = 0;

  // The width of the specular BRDF in the orthogonal direction
  float specularSigmaOrthogonal = 0;

  void PrintToStream(std::ostream& os) const override {
    os << "WardBRDF:\n  Diffuse = " << diffuseWeight
       << "\n  Specular = " << specularWeight
       << "\n  Specular sigma (parallel) = " << specularSigmaParallel
       << "\n  Specular sigma (orthogonal) = " << specularSigmaOrthogonal
       << "\n";
  }
};

/// SimpleReflectance material data model
class D5_D4LEGACY_EXPORT SimpleReflectance final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::SimpleReflectance> {
public:
  /// Destructor
  ~SimpleReflectance() override = default;

  /// Filename of the simple spectral data
  std::string filename;

  /// Format of filename
  SimpleSpectralData::Format format = SimpleSpectralData::Format::Unknown;

  /// The SLI index, if necessary
  std::variant<int, std::string> sliIndex = -1;
};

class D5_D4LEGACY_EXPORT DeltaReflectance final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::DeltaReflectance> {
public:
  /// Filename of the simple spectral data
  std::string filename;

  /// Format of filename
  SimpleSpectralData::Format format = SimpleSpectralData::Format::Unknown;

  /// The SLI index, if necessary
  std::variant<int, std::string> sliIndex = -1;

  void PrintToStream(std::ostream& os) const override {
    os << "DeltaReflectance";
  }
};

/// Priest-Germer BRDF data model
class D5_D4LEGACY_EXPORT PriestGermer final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::PriestGermer> {
public:
  /// Virtual destructor
  ~PriestGermer() override = default;

  struct SpectralParameters {
    double wavelengthMicron;
    double surfNormSigmaRad;
    double ciorReal;
    double ciorImag;
  };

  /// The parameter file
  std::string paramFilename;

  /// Whether the BRDF has polarization enabled
  bool polarizationEnabled = false;

  /// Whether the BRDF has diffuse contribution enabled
  bool diffuseContributionEnabled = false;

  /// Whether the BRDF has diffuse renormalization enabled
  bool diffuseRenormalizationEnabled = false;

  /// Whether the BRDF has the shadowing function enabled
  bool shadowingFunctionEnabled = false;

  void PrintToStream(std::ostream& os) const override {
    os << "PriestGermer:\n  Parameters: " << paramFilename
       << "\n  Polarization enabled: " << polarizationEnabled
       << "\n  Diffuse contribution enabled: " << diffuseContributionEnabled
       << "\n  Diffuse renormalization enabled: "
       << diffuseRenormalizationEnabled
       << "\n  Shadowing function enabled: " << shadowingFunctionEnabled
       << "\n";
  }
};

D5_D4LEGACY_EXPORT std::vector<PriestGermer::SpectralParameters>
ReadPriestGermerParametersFile(const char* filename);

/// Shell Target BRDF data model
class D5_D4LEGACY_EXPORT ShellTarget final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::ShellTarget> {
public:
  /// Virtual destructor
  ~ShellTarget() override = default;

  /// Orient function.
  enum class Orient { Cauchy, Gaussian };

  /// Struct to represent each spectral sample
  struct FitParams {
    double wavelengthMicrons = 0;
    double n = 1;
    double k = 0;
    double dhr = 0;
    double tgain = 1;
    Orient orient = Orient::Cauchy;
    double orientMu = 0, orientSigma = 1;
    double shadowFuncTau = 0, shadowFuncOmega = 0;
    double volumeFuncRhoD = 0, volumeFuncRhoV = 0;
  };

  /// The filename of the fit data
  std::string fitFilename;

  /// The filename of the emisssivity data (may be empty)
  std::string emsFilename;

  /// The interpolation method
  int interpolation = 1;

  void PrintToStream(std::ostream& os) const override {
    os << "ShellTarget:\n  Interpolation: " << interpolation
       << "\n  Fit filename: " << fitFilename
       << "\n  Emissivity filename: " << emsFilename << "\n";
  }
};

D5_D4LEGACY_EXPORT std::vector<ShellTarget::FitParams> ReadShellTargetFitFile(
  const char* filename);

/// Shell Background BRDF data model
class D5_D4LEGACY_EXPORT ShellBackground final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::ShellBackground> {
public:
  /// Virtual destructor
  ~ShellBackground() override = default;

  /// Shared fit parameters
  ///
  /// \note
  /// For use by d5brdf, the only necessary shared fit parameters are
  /// `a`, `b` and `gsdInches`. The `gsdInches` parameter was not needed
  /// in DIRSIG4 because it could be deduced on the fly. This is not the
  /// case in DIRSIG5. All other parameters concern polarization, which
  /// is not modeled by d5brdf.
  ///
  struct SharedFitParams {
    float rho0 = 0;
    float p1 = 0;
    float p2 = 0;
    float p3 = 0;
    float p4 = 0;
    float pf1 = 0;
    float pf2 = 0;
    float pf3 = 0;
    float pf4 = 0;
    float a = 0;
    float b = 0;
    float c = 0;
    float d = 0;
    float e = 0;
    float f = 0;
    float gsdInches = 6;
  };

  /// Spectrally-varying fit parameters
  struct SpectralFitParams {
    float wavelengthMicrons = 0;
    float rho = 0;
    float k0 = 0;
    float k1 = 0;
    float k2 = 0;
  };

  struct FitParameters {
    SharedFitParams shared;
    std::vector<SpectralFitParams> spectral;
  };

  /// Filename of the fit parameters
  std::string fitFilename;

  /// Filename of the emissivity file (may be empty)
  std::string emsFilename;

  void PrintToStream(std::ostream& os) const override {
    os << "ShellBackground:"
       << "\n  Fit filename: " << fitFilename
       << "\n  Emissivity filename: " << emsFilename << "\n";
  }
};

D5_D4LEGACY_EXPORT ShellBackground::FitParameters ReadShellBackgroundFitFile(
  const char* filename);

/// Reflectance map data model
class D5_D4LEGACY_EXPORT ReflectanceMap final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::ReflectanceMap> {
public:
  /// Virtual destructor
  ~ReflectanceMap() override = default;

  /// The data model for the map
  ReflectanceMapEntry map;
};

/// RGB image data model
class D5_D4LEGACY_EXPORT RgbImage final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::RgbImage> {
public:
  /// Virtual destructor
  ~RgbImage() override = default;

  /// The image filename
  std::string filename;

  /// The scale for the digital counts in the image
  double scale = 1;

  /// The map projector
  Projector projector;
};

class D5_D4LEGACY_EXPORT CurveMap final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::CurveMap> {
public:
  /// The ENVI SLI (Spectral LIbrary) filename.
  std::string sliFilename;

  /// The (optional) offset for indexes into the SLI file
  int sliIndexOffset = 0;

  /// The image filename.
  std::string imageFilename;

  /// The projector information.
  Projector projector;
};

/// Arbitrary SQT BRDF data model
class D5_D4LEGACY_EXPORT SphericalDataBrdf final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::SphericalDataBrdf> {
public:
  /// Virtual destructor
  ~SphericalDataBrdf() override = default;

  /// The SQT filename
  std::string sqtFilename;

  /// The DHR filename
  std::string dhrFilename;

  bool normalizeDhr = false;

  void PrintToStream(std::ostream& os) const override {
    os << "SphericalDataBrdf:"
       << "\n  SQT filename: " << sqtFilename << "\n";
  }
};

/// Fresnel BRDF
class D5_D4LEGACY_EXPORT PMFresnelBrdf final
    : public ReflectanceProperty::Subclass<
        ReflectanceProperty::Type::PMFresnelBrdf> {
public:
  /// Virtual destructor
  ~PMFresnelBrdf() override = default;

  void PrintToStream(std::ostream& os) const override { os << "PMFresnelBrdf"; }
};

/// Ross-Li BRDF
class D5_D4LEGACY_EXPORT RossLi final
    : public ReflectanceProperty::Subclass<ReflectanceProperty::Type::RossLi> {
public:
  /// Virtual destructor
  ~RossLi() override = default;

  enum RossKernel { Thick, Thin };

  /// The Ross kernel to use
  RossKernel rossKernel = Thick;

  enum LiKernel { Dense, Sparse };

  /// The Li kernel to use
  LiKernel liKernel = Sparse;

  /// Base-to-radius ratio
  double br = 1;

  /// Height-to-base ratio
  double hb = 2;

  /** Whether to assume objects drivig BRDF (normally trees) are vetrically
   *  aligned
   */
  bool verticalFlag = false;

  /// Whether to use the reciprocal Li form (sparse only)
  bool reciprocalLi = true;

  /// Emissivity filename
  std::string emissivityFilename;

  /// Texture filename
  std::string textureFilename;

  /// Bandpass for the texture
  std::array<double, 2> textureBandpass = {0.5, 0.6};

  /// Texture stats [mean, std. dev]
  std::array<double, 2> textureStats = {0, -1};

  /// Wavelength-specific fit parameters
  struct FitParameters {
    double wavelengthMicron;
    double fIso;
    double fVol;
    double fGeo;
  };

  /// Fit parameters
  std::vector<FitParameters> fitParameters;

  /// Filename for a fit file (if parameters are external)
  std::string fitFilename;

  /// Range bound parameters
  struct RangeParameters {
    int sourceMin;
    int sourceMax;
    int viewMin;
    int viewMax;
    std::string filename;
  };

  std::vector<RangeParameters> rangeParameters;

  void PrintToStream(std::ostream& os) const override { os << "RossLi"; }
};

D5_D4LEGACY_EXPORT std::vector<RossLi::FitParameters> ReadRossLiFitFile(
  const char* filename);

/// Reflectance property type to string.
inline std::string to_string(ReflectanceProperty::Type type) {
  switch (type) {
    case ReflectanceProperty::Type::None:
      return "None";
    case ReflectanceProperty::Type::WardBrdf:
      return "WardBrdf";
    case ReflectanceProperty::Type::SimpleReflectance:
      return "SimpleReflectance";
    case ReflectanceProperty::Type::PriestGermer:
      return "PriestGermer";
    case ReflectanceProperty::Type::ShellTarget:
      return "ShellTarget";
    case ReflectanceProperty::Type::ShellBackground:
      return "ShellBackground";
    case ReflectanceProperty::Type::ReflectanceMap:
      return "ReflectanceMap";
    case ReflectanceProperty::Type::RgbImage:
      return "RgbImage";
    case ReflectanceProperty::Type::SphericalDataBrdf:
      return "SphericalDataBrdf";
    case ReflectanceProperty::Type::PMFresnelBrdf:
      return "PMFresnelBrdf";
    case ReflectanceProperty::Type::RossLi:
      return "RossLi";
    case ReflectanceProperty::Type::DeltaReflectance:
      return "DeltaReflectance";
    default:
      break;
  }
  return "Unknown";
}

/// \}
}  // namespace d5::d4legacy
