/** Interface file for TemperatureSolver
 *
 * \file TemperatureSolver.h
 */

#pragma once

#include <array>
#include <iomanip>
#include <iostream>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <vector>

#include "dirsig5/base/Inheritance.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/mat/Therm.h"
#include "dirsig5/d4legacy/scene/Maps.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_mat
/// \{

/// Base class for DIRSIG4 temperature solvers
class D5_D4LEGACY_EXPORT TemperatureSolver
    : public EnumeratedSuperclass<TemperatureSolver> {
public:
  /// Types of temperature solvers
  enum class Type : uint32_t {
    None = 0,
    Therm,
    Balfour,
    DataDriven,
    Map,
    MappedBalfour,
    MappedTherm
  };

  /// Virtual destructor
  virtual ~TemperatureSolver() = default;

  /// The type of this solver
  [[nodiscard]] Type type() const { return type_; }

  /// Prints the material to a stream
  virtual void PrintToStream(std::ostream&) const {}

protected:
  explicit TemperatureSolver(Type type) : type_(type) {}

private:
  Type type_;
};

/// Subclass for THERM temperature solvers
class D5_D4LEGACY_EXPORT ThermTemperatureSolver final
    : public TemperatureSolver::Subclass<TemperatureSolver::Type::Therm> {
public:
  ~ThermTemperatureSolver() override = default;

  void PrintToStream(std::ostream& os) const override {
    os << "THERM:\n"
       << material << "\n"
       << "\nEnable solar shadows: " << std::boolalpha << enableSkyShadows
       << "\nEnable Timestep Optimization: " << enableTimestepOptimization
       << "\nMinimum Optimized Timesteps: " << minimumOptimizedTimesteps
       << "\n";
  }

  ThermMaterialProps material;
  bool enableSkyShadows = false;
  bool enableTimestepOptimization = false;
  int minimumOptimizedTimesteps = 0;
};

/// Subclass for Balfour temperature solvers
class D5_D4LEGACY_EXPORT BalfourTemperatureSolver final
    : public TemperatureSolver::Subclass<TemperatureSolver::Type::Balfour> {
public:
  ~BalfourTemperatureSolver() override = default;

  void PrintToStream(std::ostream& os) const override {
    os << "Balfour:\n  Coefficients:";
    for (auto a : coefficients) os << " " << a;
  }

  /// Balfour coefficients
  std::array<double, 6> coefficients;
};

/// Subclass for the data-driven temperature solvers
class D5_D4LEGACY_EXPORT DataDrivenTemperatureSolver final
    : public TemperatureSolver::Subclass<TemperatureSolver::Type::DataDriven> {
public:
  ~DataDrivenTemperatureSolver() override = default;

  void PrintToStream(std::ostream& os) const override {
    os << "Data-Driven: ";
    if (filename.empty())
      os << tempK;
    else
      os << filename;
  }

  /// Temperature in Kelvin (-1 = not given)
  float tempK = -1;

  /// Filename containing temperature time series
  std::string filename;
};

/// Subclass for temperature maps
class D5_D4LEGACY_EXPORT TemperatureMapSolver final
    : public TemperatureSolver::Subclass<TemperatureSolver::Type::Map> {
public:
  ~TemperatureMapSolver() override = default;

  /// The temperature map specification
  TemperatureMapEntry map;
};

/// Subclass for temperature maps
class D5_D4LEGACY_EXPORT MappedBalfourSolver final
    : public TemperatureSolver::Subclass<
        TemperatureSolver::Type::MappedBalfour> {
public:
  ~MappedBalfourSolver() override = default;

  /// The temperature map specification
  BalfourMapEntry map;
};

/// Subclass for THERM parameter maps
class D5_D4LEGACY_EXPORT MappedThermSolver final
    : public TemperatureSolver::Subclass<TemperatureSolver::Type::MappedTherm> {
public:
  ~MappedThermSolver() override = default;

  /// The THERM map specification
  ThermMapEntry map;
};

/// Temperature solver type to string.
inline std::string to_string(TemperatureSolver::Type type) {
  switch (type) {
    case TemperatureSolver::Type::None:
      return "None";
    case TemperatureSolver::Type::Therm:
      return "Therm";
    case TemperatureSolver::Type::Balfour:
      return "Balfour";
    case TemperatureSolver::Type::DataDriven:
      return "DataDriven";
    case TemperatureSolver::Type::Map:
      return "Map";
    case TemperatureSolver::Type::MappedBalfour:
      return "MappedBalfour";
    case TemperatureSolver::Type::MappedTherm:
      return "MappedTherm";
    default:
      break;
  }
  return "Unknown";
}

/// \}
}  // namespace d5::d4legacy
