#pragma once

#include <cstdint>
#include <functional>
#include <string>
#include <unordered_map>
#include <vector>

#include "dirsig5/d4legacy/Export.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_misc
/// \{

class D5_D4LEGACY_EXPORT FileLocator {
public:
  enum FileType : uint32_t {
    Any = 0,
    GeometryList,
    Geometry,
    Emissivity,
    Extinction,
    Absorption,
    Map,
    Source,
    BundledMaterial
  };

  void addPath(FileType type, std::vector<std::string> directories) {
    if (auto it = pathLookups_.find(type); it != pathLookups_.end()) {
      for (auto& d : directories) it->second.push_back(std::move(d));
    } else {
      pathLookups_.emplace(type, std::move(directories));
    }
  }

  FileLocator& withPath(FileType type, std::vector<std::string> directories) {
    addPath(type, std::move(directories));
    return *this;
  }

  void addFilter(const std::function<std::string(const std::string&)>& filter) {
    filters_.push_back(filter);
  }

  // Forward along a basic unordered_map interface
  [[nodiscard]] auto begin() { return pathLookups_.begin(); }
  [[nodiscard]] auto begin() const { return pathLookups_.begin(); }
  [[nodiscard]] auto end() { return pathLookups_.end(); }
  [[nodiscard]] auto end() const { return pathLookups_.end(); }
  [[nodiscard]] size_t size() const { return pathLookups_.size(); }
  [[nodiscard]] auto find(FileType type) { return pathLookups_.find(type); }
  [[nodiscard]] auto find(FileType type) const {
    return pathLookups_.find(type);
  }
  void erase(FileType type) { pathLookups_.erase(type); }

  [[nodiscard]] std::string locateFile(const std::string& filename) const;
  [[nodiscard]] std::string locateFile(
    const std::string& filename, FileType type) const;

  [[nodiscard]] std::string requiredFile(const std::string& filename) const;
  [[nodiscard]] std::string requiredFile(
    const std::string& filename, FileType type) const;

  [[nodiscard]] bool searchCwd() const { return searchCwd_; }
  void setSearchCwd(bool search) { searchCwd_ = search; }

  static FileLocator MakeRelativeToFile(const std::string& filename);

private:
  bool searchCwd_ = true;
  std::unordered_map<FileType, std::vector<std::string>> pathLookups_;
  std::vector<std::function<std::string(const std::string&)>> filters_;
};

/// \}
}  // namespace d5::d4legacy
