#pragma once

#include <memory>
#include <optional>
#include <string>
#include <type_traits>

#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Inheritance.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/geometry/Rotation.h"
#include "dirsig5/d4legacy/motion/GenericMotion.h"
#include "dirsig5/d4legacy/motion/Spin.h"
#include "dirsig5/d4legacy/xml/xml.h"
#include "dirsig5/flex_motion/Motion.h"
#include "dirsig5/geolib/Location.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_motion
/// \{

class D5_D4LEGACY_EXPORT JitterModel
    : public EnumeratedSuperclass<JitterModel> {
public:
  enum class Type { Correlated, Uncorrelated };

  virtual ~JitterModel() = default;

  /// Type of the location engine
  [[nodiscard]] Type type() const { return type_; }

  /// Deep copy operation
  [[nodiscard]] virtual std::unique_ptr<JitterModel> clone() const = 0;

  [[nodiscard]] static std::unique_ptr<JitterModel> FromXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});

  virtual void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) = 0;

protected:
  JitterModel(Type t) : type_(t) {}

private:
  Type type_;
};

class WithJitterModels {
public:
  virtual ~WithJitterModels() = default;

  [[nodiscard]] bool hasJitter(int axis) const { return jm_[axis] != nullptr; }

  [[nodiscard]] JitterModel::Type jitterType(int axis) const {
    return jm_[axis]->type();
  }

  void setJitter(int axis, std::unique_ptr<JitterModel> jm) {
    jm_[axis] = std::move(jm);
  }

  template <typename T = JitterModel>
  [[nodiscard]] const T& jitter(int axis) const {
    if constexpr (std::is_same_v<T, JitterModel>)
      return *jm_[axis];
    else
      return jm_[axis]->as<T>();
  }

  template <typename T = JitterModel>
  [[nodiscard]] T& jitter(int axis) {
    if constexpr (std::is_same_v<T, JitterModel>)
      return *jm_[axis];
    else
      return jm_[axis]->as<T>();
  }

protected:
  void cloneJitterModels(const WithJitterModels& other) {
    if (this != &other)
      for (int ax = 0; ax < 3; ax++)
        jm_[ax] = other.jm_[ax] ? other.jm_[ax]->clone() : nullptr;
  }

  std::unique_ptr<JitterModel> jm_[3] = {};
};

/// Base class for location engines in a flex motion model
class D5_D4LEGACY_EXPORT LocationEngine
    : public EnumeratedSuperclass<LocationEngine>,
      public WithJitterModels {
public:
  /// Types of location engines
  enum class Type { Fixed, Waypoints, Straight, Circling, Sgp4 };

  /// Virtual destructor
  virtual ~LocationEngine() = default;

  /// Type of the location engine
  [[nodiscard]] Type type() const { return type_; }

  /// Deep copy operation
  [[nodiscard]] virtual std::unique_ptr<LocationEngine> clone() const = 0;

  [[nodiscard]] static std::unique_ptr<LocationEngine> FromXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});

  virtual void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) = 0;

protected:
  LocationEngine(Type t) : type_(t) {}

private:
  Type type_;
};

/// Base class for rotation engines in a flex motion model
class D5_D4LEGACY_EXPORT RotationEngine
    : public EnumeratedSuperclass<RotationEngine>,
      public WithJitterModels {
public:
  /// Types of rotation engines
  enum class Type { EulerAngles, Spin, Quaternions, Velocity, LookAt };

  /// Reference frames for rotation engines to use
  enum class Frame { Enu, Ecef };

  /// Virtual destructor
  virtual ~RotationEngine() = default;

  /// Type of the rotation engine
  [[nodiscard]] Type type() const { return type_; }

  /// Deep copy operation
  [[nodiscard]] virtual std::unique_ptr<RotationEngine> clone() const = 0;

  [[nodiscard]] static std::unique_ptr<RotationEngine> FromXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});

  virtual void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) = 0;

protected:
  RotationEngine(Type t) : type_(t) {}

private:
  Type type_;
};

D5_D4LEGACY_EXPORT std::string to_string(RotationEngine::Frame f);

/** Represents a flex motion model, composed of a LocationEngine and an
 *  RotationEngine
 */
class D5_D4LEGACY_EXPORT FlexMotion {
public:
  FlexMotion() = default;

  FlexMotion(const FlexMotion& other)
      : le_(other.le_ ? other.le_->clone() : nullptr),
        re_(other.re_ ? other.re_->clone() : nullptr) {}

  FlexMotion(FlexMotion&&) = default;

  FlexMotion& operator=(const FlexMotion& other) {
    if (this != &other) {
      le_ = other.le_ ? other.le_->clone() : nullptr;
      re_ = other.re_ ? other.re_->clone() : nullptr;
    }
    return *this;
  }

  FlexMotion& operator=(FlexMotion&&) = default;

  /// Whether we have a location engine
  [[nodiscard]] bool hasLocationEngine() const { return le_ != nullptr; }

  /// Whether we have an rotation engine
  [[nodiscard]] bool hasRotationEngine() const { return re_ != nullptr; }

  /// The type of the location engine
  [[nodiscard]] LocationEngine::Type locationEngineType() const {
    return le_->type();
  }

  /// The type of the rotation engine
  [[nodiscard]] RotationEngine::Type rotationEngineType() const {
    return re_->type();
  }

  /// Accessor for the location engine
  template <typename T = LocationEngine>
  const T& locationEngine() const {
    if constexpr (std::is_same_v<T, LocationEngine>)
      return *le_;
    else
      return le_->as<T>();
  }

  /// \overload
  template <typename T = LocationEngine>
  T& locationEngine() {
    if constexpr (std::is_same_v<T, LocationEngine>)
      return *le_;
    else
      return le_->as<T>();
  }

  /// Accessor for the rotation engine
  template <typename T = RotationEngine>
  const T& rotationEngine() const {
    if constexpr (std::is_same_v<T, RotationEngine>)
      return *re_;
    else
      return re_->as<T>();
  }

  /// \overload
  template <typename T = RotationEngine>
  T& rotationEngine() {
    if constexpr (std::is_same_v<T, RotationEngine>)
      return *re_;
    else
      return re_->as<T>();
  }

  /// Sets the location engine
  void setLocationEngine(std::unique_ptr<LocationEngine> le) {
    le_ = std::move(le);
  }

  /// Sets the rotation engine
  void setRotationEngine(std::unique_ptr<RotationEngine> re) {
    re_ = std::move(re);
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});

private:
  std::unique_ptr<LocationEngine> le_;
  std::unique_ptr<RotationEngine> re_;
};

// #################################
// #   LocationEngine Subclasses   #
// #################################

/// Holds subclasses of LocationEngine
namespace LocationEngines {

/// Subclass for a fixed location engine
class D5_D4LEGACY_EXPORT Fixed final
    : public LocationEngine::Subclass<LocationEngine::Type::Fixed> {
public:
  geolib::Location location = {nullptr};

  [[nodiscard]] std::unique_ptr<LocationEngine> clone() const override {
    auto c = std::make_unique<Fixed>();
    c->cloneJitterModels(*this);
    c->location = location;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Subclass for a waypoints location engine
class D5_D4LEGACY_EXPORT Waypoints final
    : public LocationEngine::Subclass<LocationEngine::Type::Waypoints> {
public:
  struct LocationEntry {
    DateTime dateTime;
    geolib::Location location = {nullptr};
  };

  std::vector<LocationEntry> waypoints;
  bool loop = false;

  std::string frameType;
  std::string timeType;
  std::string delimiter;
  std::optional<std::string> externalFilename;
  std::optional<std::string> stkFilename;

  [[nodiscard]] std::vector<LocationEntry> LoadWaypointsFromString(
    const std::string& data, const geolib::Facade* facade) const;
  [[nodiscard]] std::vector<LocationEntry> LoadWaypointsFromExternalFile(
    const char* filename, const geolib::Facade* facade) const;

  [[nodiscard]] std::unique_ptr<LocationEngine> clone() const override {
    auto c = std::make_unique<Waypoints>();
    c->cloneJitterModels(*this);
    c->waypoints = waypoints;
    c->frameType = frameType;
    c->timeType = timeType;
    c->delimiter = delimiter;
    c->externalFilename = externalFilename;
    c->stkFilename = stkFilename;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Subclass for the straight path location engine
class D5_D4LEGACY_EXPORT Straight final
    : public LocationEngine::Subclass<LocationEngine::Type::Straight> {
public:
  double headingDegEofN = 0;
  double velocityMS = 0;
  geolib::Location startLocation = {nullptr};
  DateTime startTime;

  [[nodiscard]] std::unique_ptr<LocationEngine> clone() const override {
    auto c = std::make_unique<Straight>();
    c->cloneJitterModels(*this);
    c->headingDegEofN = headingDegEofN;
    c->velocityMS = velocityMS;
    c->startLocation = startLocation;
    c->startTime = startTime;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Subclass for the circling location engine
class D5_D4LEGACY_EXPORT Circling final
    : public LocationEngine::Subclass<LocationEngine::Type::Circling> {
public:
  geolib::Location centerLocation = {nullptr};
  double radiusM = 0;
  double velocityMS = 0;
  DateTime startTime;

  [[nodiscard]] std::unique_ptr<LocationEngine> clone() const override {
    auto c = std::make_unique<Circling>();
    c->cloneJitterModels(*this);
    c->centerLocation = centerLocation;
    c->radiusM = radiusM;
    c->velocityMS = velocityMS;
    c->startTime = startTime;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Subclass for the circling location engine
class D5_D4LEGACY_EXPORT Sgp4 final
    : public LocationEngine::Subclass<LocationEngine::Type::Sgp4> {
public:
  std::string tleLine1;
  std::string tleLine2;

  [[nodiscard]] std::unique_ptr<LocationEngine> clone() const override {
    auto c = std::make_unique<Sgp4>();
    c->cloneJitterModels(*this);
    c->tleLine1 = tleLine1;
    c->tleLine2 = tleLine2;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

}  // namespace LocationEngines

// ####################################
// #   RotationEngine Subclasses   #
// ####################################

/// Holds subclasses of RotationEngine
namespace RotationEngines {

/// A constraint (up) vector for an rotation engine
struct ConstraintVector {
  /// Coordinate system for the vector
  enum Option {
    UnknownConstraint = 0x00,  ///< Unknown
    UseVelocity = 0x01,        ///< Use the current velocity vector
    UseConstant = 0x02,        ///< Use a constant vector
    VectorFrameEnu = 0x04,     ///< Constant vector frame is Scene ENU
    VectorFrameEcef = 0x08,    ///< Constant vector frame is ECEF
    UseConstantEnu = UseConstant | VectorFrameEnu,
    UseConstantEcef = UseConstant | VectorFrameEcef
  };

  /// Vector pointing in the "up" direction
  Vector up = Vector(0, 0, 1);

  /// Coordinate system for the up vector
  uint32_t option = UseConstantEnu;

  void ReadXML(const xml::Node& node);
};

/// Subclass for the Euler angles rotation engine
class D5_D4LEGACY_EXPORT EulerAngles final
    : public RotationEngine::Subclass<RotationEngine::Type::EulerAngles> {
public:
  struct EulerAngleEntry {
    DateTime dateTime;
    std::array<double, 3> angles;
  };

  std::vector<EulerAngleEntry> eulerAngles;
  RotationOrder rotationOrder = RotationOrder::Zyx;
  Frame frame = Frame::Enu;
  std::string timeType;
  std::string delimiter;
  std::optional<std::string> externalFilename;
  bool loop = false;

  [[nodiscard]] std::vector<EulerAngleEntry> LoadAnglesFromString(
    const std::string& data) const;
  [[nodiscard]] std::vector<EulerAngleEntry> LoadAnglesFromFile(
    const char* filename) const;

  [[nodiscard]] std::unique_ptr<RotationEngine> clone() const override {
    auto c = std::make_unique<EulerAngles>();
    c->cloneJitterModels(*this);
    c->eulerAngles = eulerAngles;
    c->rotationOrder = rotationOrder;
    c->frame = frame;
    c->timeType = timeType;
    c->delimiter = delimiter;
    c->externalFilename = externalFilename;
    c->loop = loop;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Subclass for the spin rotation engine
class D5_D4LEGACY_EXPORT Spin final
    : public RotationEngine::Subclass<RotationEngine::Type::Spin> {
public:
  struct RateEntry {
    DateTime dateTime;
    double rateRadS;
  };

  Vector axis = Vector(0, 0, 1);
  std::vector<RateEntry> rates;

  Frame frame = Frame::Enu;
  std::string timeType;
  std::string delimiter;
  std::optional<std::string> externalFilename;

  [[nodiscard]] std::vector<RateEntry> LoadRatesFromString(
    const std::string& data) const;
  [[nodiscard]] std::vector<RateEntry> LoadRatesFromFile(
    const char* filename) const;

  [[nodiscard]] std::unique_ptr<RotationEngine> clone() const override {
    auto c = std::make_unique<Spin>();
    c->cloneJitterModels(*this);
    c->axis = axis;
    c->rates = rates;
    c->frame = frame;
    c->timeType = timeType;
    c->delimiter = delimiter;
    c->externalFilename = externalFilename;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Subclass for the quaternion rotation engine
class D5_D4LEGACY_EXPORT Quaternions final
    : public RotationEngine::Subclass<RotationEngine::Type::Quaternions> {
public:
  struct QuaternionEntry {
    DateTime dateTime;
    Quaternion quaternion;
  };

  std::vector<QuaternionEntry> quaternions;
  Frame frame = Frame::Enu;
  std::string timeType;
  std::string delimiter;
  std::optional<std::string> externalFilename;
  std::optional<std::string> stkFilename;
  bool loop = false;

  [[nodiscard]] std::vector<QuaternionEntry> LoadQuaternionsFromString(
    const std::string& data) const;
  [[nodiscard]] std::vector<QuaternionEntry> LoadQuaternionsFromFile(
    const char* filename) const;

  [[nodiscard]] std::unique_ptr<RotationEngine> clone() const override {
    auto c = std::make_unique<Quaternions>();
    c->cloneJitterModels(*this);
    c->quaternions = quaternions;
    c->frame = frame;
    c->timeType = timeType;
    c->delimiter = delimiter;
    c->externalFilename = externalFilename;
    c->stkFilename = stkFilename;
    c->loop = loop;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Subclass for the velocity rotation engine
class D5_D4LEGACY_EXPORT Velocity final
    : public RotationEngine::Subclass<RotationEngine::Type::Velocity> {
public:
  ConstraintVector constraint;

  [[nodiscard]] std::unique_ptr<RotationEngine> clone() const override {
    auto c = std::make_unique<Velocity>();
    c->cloneJitterModels(*this);
    c->constraint = constraint;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

/// Subclass for the look at rotation engine
class D5_D4LEGACY_EXPORT LookAt final
    : public RotationEngine::Subclass<RotationEngine::Type::LookAt> {
public:
  std::unique_ptr<LocationEngine> locEngine;
  ConstraintVector constraint;

  [[nodiscard]] std::unique_ptr<RotationEngine> clone() const override {
    auto c = std::make_unique<LookAt>();
    c->cloneJitterModels(*this);
    c->locEngine = locEngine->clone();
    c->constraint = constraint;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

}  // namespace RotationEngines

namespace JitterModels {

class D5_D4LEGACY_EXPORT Correlated final
    : public JitterModel::Subclass<JitterModel::Type::Correlated> {
public:
  struct Component {
    double frequency = 0;
    double magnitude = 0;
    double phase = 0;
  };

  std::vector<Component> components;

  [[nodiscard]] std::unique_ptr<JitterModel> clone() const override {
    auto c = std::make_unique<Correlated>();
    c->components = components;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

class D5_D4LEGACY_EXPORT Uncorrelated final
    : public JitterModel::Subclass<JitterModel::Type::Uncorrelated> {
public:
  uint32_t seed = 0;
  double mean = 0;
  double stdev = 0;

  [[nodiscard]] std::unique_ptr<JitterModel> clone() const override {
    auto c = std::make_unique<Uncorrelated>();
    c->seed = seed;
    c->mean = mean;
    c->stdev = stdev;
    return c;
  }

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {}) override;
};

}  // namespace JitterModels

/// Convert SpinMotion to FlexMotion.
D5_D4LEGACY_EXPORT FlexMotion SpinMotionToFlexMotion(const SpinMotion&);

/// Convert GenericMotion to FlexMotion.
D5_D4LEGACY_EXPORT FlexMotion GenericMotionToFlexMotion(const GenericMotion&);

/// Load FlexMotion into runtime instance of flex_motion::Motion
D5_D4LEGACY_EXPORT std::shared_ptr<flex_motion::Motion> LoadRuntimeFlexMotion(
  const FlexMotion&);

/// \}
}  // namespace d5::d4legacy
