#pragma once

#include <optional>
#include <variant>

#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/geometry/Rotation.h"
#include "dirsig5/d4legacy/platform/Clock.h"
#include "dirsig5/d4legacy/xml/xml.h"

namespace d5::d4legacy::platform {

struct D5_D4LEGACY_EXPORT Mount {
public:
  struct CommandedOrTabulated {
    /// The clock.
    Clock clock{};

    /// The rotation order.
    RotationOrder rotationOrder{RotationOrder::Xyz};

    struct Keyframe {
      /// The date-time. Should be present if fraction is not.
      std::optional<DateTime> dateTime{};

      /// The interpolation fraction. Should be present if dateTime is not.
      std::optional<double> fraction{};

      /// The rotation angle in X in radians.
      double rotationX{};

      /// The rotation angle in Y in radians.
      double rotationY{};

      /// The rotation angle in Z in radians.
      double rotationZ{};

      /// The translation in X.
      double translationX{};

      /// The translation in Y.
      double translationY{};

      /// The translation in Z.
      double translationZ{};

      void ReadXML(const xml::Node& node, double angularUnits = 1) {
        *this = {};
        if (node.child("datetime"))
          dateTime.emplace(xml::readDatetime(node, "datetime"));
        if (auto child = node.child("fraction"))
          fraction.emplace(child.text().as_double());
        rotationX = angularUnits * xml::readNumber(node, "xrotation");
        rotationY = angularUnits * xml::readNumber(node, "yrotation");
        rotationZ = angularUnits * xml::readNumber(node, "zrotation");
        translationX = xml::readNumber(node, "xtranslation");
        translationY = xml::readNumber(node, "ytranslation");
        translationZ = xml::readNumber(node, "ztranslation");
      }
    };

    /// The keyframes.
    std::vector<Keyframe> keyframes{};

    /// The external filename, if applicable.
    std::optional<std::string> filename{};

    void ReadXML(const xml::Node& node) {
      *this = {};
      clock.ReadXML(xml::requiredChild(node, "clock"));
      rotationOrder = xml::readRotationOrderAttribute(node, "rotationorder");
      auto angularUnits =
        xml::readAngularUnitsAttributeAsMultiplier(node, "angularunits");
      for (auto child : node.children("entry"))
        keyframes.emplace_back().ReadXML(child, angularUnits);
      if (auto child = node.child("filename"))
        filename.emplace(child.text().as_string());
    }
  };

  struct LineOrWhisk {
    /// The clock.
    Clock clock{};

    /// The start angle in radians.
    double startAngle{};

    /// The stop angle in radians.
    double stopAngle{};

    /// Read XML. The given node is understood to be `<data>`.
    void ReadXML(const xml::Node& node) {
      clock.ReadXML(xml::requiredChild(node, "clock"));
      auto angularUnits =
        xml::readAngularUnitsAttributeAsMultiplier(node, "angularunits");
      auto scanRange = xml::requiredChild(node, "scanrange");
      startAngle =  //
        angularUnits * xml::readRequiredNumber(scanRange, "start");
      stopAngle =  //
        angularUnits * xml::readRequiredNumber(scanRange, "stop");
    }
  };

  struct Lemniscate {
    /// The clock.
    Clock clock{};

    /// The minimum across-track angle in radians.
    double minAcrossTrackAngle{};

    /// The maximum across-track angle in radians.
    double maxAcrossTrackAngle{};

    /// The minimum along-track angle in radians.
    double minAlongTrackAngle{};

    /// The maximum along-trackpecificData angle in radians.
    double maxAlongTrackAngle{};

    /// Read XML. The given node is understood to be `<data>`.
    void ReadXML(const xml::Node& node) {
      clock.ReadXML(xml::requiredChild(node, "clock"));
      auto angularUnits =
        xml::readAngularUnitsAttributeAsMultiplier(node, "angularunits");
      auto across = xml::requiredChild(node, "acrosstrack");
      minAcrossTrackAngle =
        angularUnits * xml::readRequiredNumber(across, "minimumangle");
      maxAcrossTrackAngle =
        angularUnits * xml::readRequiredNumber(across, "maximumangle");
      auto along = xml::requiredChild(node, "alongtrack");
      minAlongTrackAngle =
        angularUnits * xml::readRequiredNumber(along, "minimumangle");
      maxAlongTrackAngle =
        angularUnits * xml::readRequiredNumber(along, "maximumangle");
    }
  };

  struct Python {
    /// The clock.
    std::optional<Clock> clock{};

    /// The module name.
    std::string moduleName{};

    /// Read XML. The given node is understood to be `<data>`.
    void ReadXML(const xml::Node& node) {
      *this = {};
      if (auto clockNode = node.child("clock"))
        clock.emplace().ReadXML(clockNode);
      moduleName = xml::requiredChild(node, "module").text().as_string();
    }
  };

  struct Static {
    /// The rotation order.
    RotationOrder rotationOrder{RotationOrder::Xyz};

    /// The rotation angle in X in radians.
    double rotationX{};

    /// The rotation angle in Y in radians.
    double rotationY{};

    /// The rotation angle in Z in radians.
    double rotationZ{};

    /// Read XML. The given node is understood to be `<data>`.
    void ReadXML(const xml::Node& node) {
      *this = {};
      rotationOrder = xml::readRotationOrderAttribute(node, "rotationorder");
      auto angularUnits =
        xml::readAngularUnitsAttributeAsMultiplier(node, "angularunits");
      rotationX = angularUnits * xml::readNumber(node, "xrotation");
      rotationY = angularUnits * xml::readNumber(node, "yrotation");
      rotationZ = angularUnits * xml::readNumber(node, "zrotation");
    }

    /// Convert to rotation matrix.
    [[nodiscard]] Eigen::Matrix3d toMatrix() const {
      return MakeRotationMatrix(rotationX, rotationY, rotationZ, rotationOrder);
    }
  };

  struct Tracking {
    /// The target name.
    std::string targetName{};

    /// The target offset.
    Vector targetOffset{};

    /// Read XML. The given node is understood to be `<data>`.
    void ReadXML(const xml::Node& node) {
      targetName = xml::requiredChild(node, "targetname").text().as_string();
      targetOffset = xml::readCoordinate(node, "targetoffset");
    }
  };

  std::string name{};

  std::string type{};

  std::variant<
    std::monostate,
    CommandedOrTabulated,
    LineOrWhisk,
    Lemniscate,
    Python,
    Static,
    Tracking>
    typeVariant;

public:
  /// Read XML. The given node is understood to be `<mount>`.
  void ReadXML(const xml::Node& node);
};

}  // namespace d5::d4legacy::platform
