#pragma once

#include <map>

#include "dirsig5/d4legacy/xml/xml.h"

namespace d5::d4legacy::platform {

struct D5_D4LEGACY_EXPORT SpectralResponse final {
public:
  struct Bandpass final {
    /// The minimum wavelength in microns.
    double minimum{0.4};
    /// The maximum wavelength in microns.
    double maximum{0.7};
    /// The sampling delta in microns.
    double delta{0.01};
  };

  struct Channel final {
    /// The name.
    std::string name{};
    /// The shape, e.g., triangle, rectangle, gaussian.
    std::string shape{};
    /// The center in microns.
    double center{0.5};
    /// The width in microns.
    double width{0.1};
    /// The gain.
    double gain{1.0};
    /// The bias.
    double bias{0.0};
    /// Normalize the response curve? (Before gain and bias).
    bool normalize{true};
    /// Tabulated data if applicable.
    std::vector<std::pair<double, double>> tabulated{};
  };

  struct ChannelPattern final {
    /// The name, i.e., bayer or truesense.
    std::string name;
    /// The filter-to-channel name mapping.
    std::map<std::string, std::string> filterToChannel;
  };

  /// The bandpass.
  Bandpass bandpass{};
  /// The channels.
  std::vector<Channel> channels{};
  /// The channel pattern if applicable.
  std::optional<ChannelPattern> channelPattern{};

public:
  /// Read XML. The given node is understood to be `<spectralresponse>`.
  void ReadXML(const xml::Node& node) {
    auto bandpassNode = xml::requiredChild(node, "bandpass");
    auto channelsNode = xml::requiredChild(node, "channellist");
    auto units =
      xml::readWavelengthUnitsAttribute(bandpassNode, "spectralunits");
    bandpass = {
      toMicrons(xml::readNumber(bandpassNode, "minimum", 0.4), units),
      toMicrons(xml::readNumber(bandpassNode, "maximum", 0.7), units),
      toMicrons(xml::readNumber(bandpassNode, "delta", 0.01), units)};
    channels.clear();
    for (auto channelNode : channelsNode.children("channel")) {
      auto& channel = channels.emplace_back();
      channel.name = channelNode.attribute("name").as_string();
      channel.shape = channelNode.attribute("shape").as_string();
      channel.center = xml::readNumber(channelNode, "center", 0.5);
      channel.width = xml::readNumber(channelNode, "width", 0.1);
      channel.gain = channelNode.attribute("gain").as_double(1.0);
      channel.bias = channelNode.attribute("bias").as_double(0.0);
      channel.normalize = channelNode.attribute("normalize").as_bool();
      for (auto entry : channelNode.child("entry"))
        channel.tabulated.emplace_back(
          xml::readNumber(entry, "spectralpoint"),  //
          xml::readNumber(entry, "value"));
    }
    channelPattern.reset();
    if (auto channelPatternNode = node.child("channelpattern")) {
      channelPattern = ChannelPattern();
      channelPattern->name = channelPatternNode.attribute("name").as_string();
      for (auto mappingNode : channelPatternNode.children("mapping")) {
        std::string key{mappingNode.attribute("filtername").as_string()};
        std::string val{mappingNode.attribute("channelname").as_string()};
        channelPattern->filterToChannel[key] = std::move(val);
      }
    }
  }
};

}  // namespace d5::d4legacy::platform
