#pragma once

#include <array>
#include <cstdint>
#include <ostream>
#include <stdexcept>
#include <string>
#include <unordered_map>
#include <vector>

#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Inheritance.h"
#include "dirsig5/base/Json.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/mat/Therm.h"
#include "dirsig5/d4legacy/misc/FileLocator.h"
#include "dirsig5/d4legacy/tag/TagFileElement.h"
#include "dirsig5/d4legacy/xml/xml.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_scene
/// \{

/// The various types of maps in DIRSIG4
enum class MapType : uint32_t {
  Unknown = 0,
  Bump,
  Decal,
  Density,
  Material,
  Mixture,
  Noise,
  Texture,
  Reflectance,
  Temperature,
  Balfour,
  Therm
};

/// Converts the enum to a string
D5_D4LEGACY_EXPORT std::string to_string(MapType t);

/// Holds information about a map projection
struct D5_D4LEGACY_EXPORT Projector {
  /// Projection types supported in DIRSIG4
  enum class Type { UV = 0, Drape };

  /// The type of the projector
  Type type = Type::UV;

  /// The origin of the drape projection
  Point drapeOrigin;

  /// The GSD of the drape projection
  double drapeGsd = 0;

  /// Whether to flip in the X dimension
  bool flipX = false;

  /// Whether to flip in the Y dimension
  bool flipY = false;

  /// The various types of map extensions
  enum class Extend { ClampToEdge, Repeat, Mirror };

  /// How to extend the map in X
  Extend extendX = Extend::Mirror;

  /// How to extend the map in Y
  Extend extendY = Extend::Mirror;

  /// Min and max normalized U coordinates
  float minU = 0, maxU = 1;

  /// Min and max normalized V coordinates
  float minV = 0, maxV = 1;

  /// Whether the image uses a Cartesian origin
  bool useCartesianOrigin = false;

  enum class Sampling { Nearest, Linear };

  Sampling sampling = Sampling::Nearest;

  /// Read from XML.
  void ReadXML(const xml::Node& node, const geolib::Facade* facade);

  /// Read from tag file section.
  void ReadTag(const TagFileSection& section);
};

/** Class to represent an image source. Normally this is a filenames, but
 *  might be tiles, which we will encode into a fake filename. Since 99% of
 *  the time, the source is a filename, we won't make an inheritance structure
 *  here.
 */
class D5_D4LEGACY_EXPORT ImageSource {
public:
  /// Constants for the types of sources we support
  enum class Type { Filename = 0, WangTiles = 1 };

  /// Constructor
  ImageSource() = default;

  /// Constructor
  ImageSource(Type type, std::string filename)
      : type_(type), filename_(std::move(filename)) {}

  /// Copy constructor
  ImageSource(const ImageSource&) = default;

  /// Move constructor
  ImageSource(ImageSource&&) noexcept = default;

  /// Copy assignment
  ImageSource& operator=(const ImageSource&) = default;

  /// Move assignment
  ImageSource& operator=(ImageSource&&) noexcept = default;

  /// The type of the source
  [[nodiscard]] Type type() const { return type_; }

  /// Set the type of the source
  void setType(Type type) { type_ = type; }

  /// The filename (or encoding) of the source
  [[nodiscard]] const std::string& filename() const { return filename_; }

  /// Set the filename (or encoding) of the source
  void setFilename(std::string filename) { filename_ = std::move(filename); }

  /// Equality operator
  bool operator==(const ImageSource& other) const {
    return type_ == other.type_ && filename_ != other.filename_;
  }

  void ReadXML(const xml::Node& node, const FileLocator& fileLocator = {});

  void ReadTag(
    const TagFileSection& section, const FileLocator& fileLocator = {});

private:
  Type type_ = Type::Filename;
  std::string filename_;
};

/// Holds the information about a Wang tile image source
struct D5_D4LEGACY_EXPORT WangTileSource {
  /// Number of tiles in X and Y
  int xCount, yCount;

  /// Holds information about each tile
  struct Tile {
    std::array<int, 4> nesw;
    std::string filename;
  };

  /// The tiles that make up the image source
  std::vector<Tile> tiles;

  void ReadXML(const xml::Node& node, const FileLocator& fileLocator = {});
};

NLOHMANN_DEFINE_TYPE_NON_INTRUSIVE(WangTileSource::Tile, nesw, filename)
NLOHMANN_DEFINE_TYPE_NON_INTRUSIVE(WangTileSource, xCount, yCount, tiles)

/// Represents an entry in the map list in the scene file
class D5_D4LEGACY_EXPORT MapEntry : public EnumeratedSuperclass<MapEntry> {
protected:
  MapEntry(MapType type) : type_(type) {}

public:
  virtual ~MapEntry() = default;

  /// The type of map this entry encodes
  [[nodiscard]] MapType type() const { return type_; }

  /// The name of the map
  std::string name;

  /// The map projector
  Projector projector;

  /// Whether the map is enabled in the scene
  bool enabled = true;

  /// The material IDs this map is associated with
  std::vector<std::string> materialIds;

  /// Calls the given callable with the appropriate downcasted suclass
  template <typename Callable>
  decltype(auto) dispatch(Callable&& callback);

  /// Calls the given callable with the appropriate downcasted suclass
  template <typename Callable>
  decltype(auto) dispatch(Callable&& callback) const;

  [[nodiscard]] static std::unique_ptr<MapEntry> FromXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});

  void ReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator = {});

  void ReadTag(
    const TagFileSection& section, const FileLocator& fileLocator = {});

protected:
  virtual void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) = 0;

  virtual void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) = 0;

private:
  MapType type_ = MapType::Unknown;
};

class D5_D4LEGACY_EXPORT BumpMapEntry final
    : public MapEntry::Subclass<MapType::Bump> {
public:
  /// The source of the image for the map
  ImageSource source;

  /// The scale factor on the bump map
  double scale = 1;

  /// Equality operator
  bool operator==(const BumpMapEntry& other) const {
    return source == other.source && scale == other.scale;
  }

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) override;

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override;
};

class D5_D4LEGACY_EXPORT DecalMapEntry final
    : public MapEntry::Subclass<MapType::Decal> {
public:
  /// The glist file describing the decal map
  std::string glist;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) override;

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override;
};

class D5_D4LEGACY_EXPORT DensityMapEntry final
    : public MapEntry::Subclass<MapType::Density> {
public:
  /// The source of the image for the map
  ImageSource source;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade*,
    const FileLocator& fileLocator) override {
    source.ReadXML(xml::requiredChild(node, "image"), fileLocator);
  }

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override {
    source.ReadTag(section, fileLocator);
  }
};

/// Subclass of MapEntry for material maps
class D5_D4LEGACY_EXPORT MaterialMapEntry final
    : public MapEntry::Subclass<MapType::Material> {
public:
  using Lut = std::unordered_map<uint8_t, std::string>;

  /// The source of the image for the map
  ImageSource source;

  /// The lookup table for the materials in this map
  Lut materialLut;

  /// Whether this is a mixing map
  bool isMixing = false;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) override;

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override;
};

/// Subclass of MapEntry for material mixture maps
class D5_D4LEGACY_EXPORT MixtureMapEntry final
    : public MapEntry::Subclass<MapType::Mixture> {
public:
  /// The source of the image for the map
  ImageSource source;

  /// The material IDs for each band of the mixture map
  std::vector<std::string> bandMaterialIds;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) override;

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override;
};

/// Subclass of MapEntry for noise maps
class D5_D4LEGACY_EXPORT NoiseMapEntry final
    : public MapEntry::Subclass<MapType::Noise> {
public:
  /// The scale on the noise map
  double scale = 1;

  /// The order of the noise map
  int order = 1;

  /// The magnitude of the noise map
  double magnitude = 0.4;

  /// The beta parameter of the noise map
  double beta = 2.0;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) override;

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override;
};

/// Subclass of MapEntry for texture maps
class D5_D4LEGACY_EXPORT TextureMapEntry final
    : public MapEntry::Subclass<MapType::Texture> {
public:
  /// Holds information about each band of the texture map
  struct Band {
    /// The minimum wavelength of the bandpass [micron]
    double bandpassMin = 0;

    /// The maximum wavelength of the bandpass [micron]
    double bandpassMax = 0;

    /// The source of the image for the band
    ImageSource source;

    /// A list of digital counts in the map to ignore
    std::vector<uint8_t> ignoreDCs;

    /// The optional mean of the digital counts
    double dcMean = 0;

    /// The optional standard deviation of the digital counts
    double dcStdDev = -1;
  };

  /// The band metadata of the texture map
  std::vector<Band> bands;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) override;

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override;
};

/// Subclass of MapEntry for reflectance maps
class D5_D4LEGACY_EXPORT ReflectanceMapEntry final
    : public MapEntry::Subclass<MapType::Reflectance> {
public:
  /// The source of the image for the map
  ImageSource source;

  /// The gain of the map
  double gain = 1;

  /// The bias of the map
  double bias = 0;

  /// Whether to clip the values in the map to [0,1]
  bool clipValues = true;

  /// Band centers [microns] (overrides the image headers if given)
  std::vector<double> bandCenters;

  /// Whether this map should be converted to a mixture map
  bool convertToMixture = false;

  /// The SQT material to multiply, if applicable.
  std::string sqtFilename;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) override;

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override;
};

class D5_D4LEGACY_EXPORT TemperatureMapEntry final
    : public MapEntry::Subclass<MapType::Temperature> {
public:
  /// The source of the image for the map
  ImageSource source;

  /// The gain of the map
  double gain = 1;

  /// The bias of the map
  double bias = 0;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) override;

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override;

  void AccountForUnits(std::string_view units);
};

class D5_D4LEGACY_EXPORT BalfourMapEntry final
    : public MapEntry::Subclass<MapType::Balfour> {
public:
  enum class Type { Classic = 0, Iterative };

  /// The source of the image for the map
  ImageSource source;

  /// The gain of the map
  Type mapType = Type::Classic;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) override;

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override;
};

class D5_D4LEGACY_EXPORT ThermMapEntry final
    : public MapEntry::Subclass<MapType::Therm> {
public:
  /// The source of the image for the map
  ImageSource source;

  ThermMaterialProps thermDefaults;

private:
  void FinishReadXML(
    const xml::Node& node,
    const geolib::Facade* facade,
    const FileLocator& fileLocator) override;

  void FinishReadTag(
    const TagFileSection& section, const FileLocator& fileLocator) override;
};

// ########################################
// #   MapEntry template implementation   #
// ########################################

/// \cond HIDDEN_SYMBOLS

#define D4LEGACY_MAP_ENTRY_CONST const
#define D4LEGACY_MAP_ENTRY_DISPATCH                           \
  template <typename Callable>                                \
  decltype(auto) MapEntry::dispatch(Callable&& callback)      \
    D4LEGACY_MAP_ENTRY_CONST {                                \
    switch (type()) {                                         \
      case d5::d4legacy::MapType::Bump:                       \
        return callback(as<BumpMapEntry>());                  \
      case d5::d4legacy::MapType::Decal:                      \
        return callback(as<DecalMapEntry>());                 \
      case d5::d4legacy::MapType::Density:                    \
        return callback(as<DensityMapEntry>());               \
      case d5::d4legacy::MapType::Material:                   \
        return callback(as<MaterialMapEntry>());              \
      case d5::d4legacy::MapType::Mixture:                    \
        return callback(as<MixtureMapEntry>());               \
      case d5::d4legacy::MapType::Noise:                      \
        return callback(as<NoiseMapEntry>());                 \
      case d5::d4legacy::MapType::Texture:                    \
        return callback(as<TextureMapEntry>());               \
      case d5::d4legacy::MapType::Reflectance:                \
        return callback(as<ReflectanceMapEntry>());           \
      case d5::d4legacy::MapType::Temperature:                \
        return callback(as<TemperatureMapEntry>());           \
      case d5::d4legacy::MapType::Balfour:                    \
        return callback(as<BalfourMapEntry>());               \
      case d5::d4legacy::MapType::Therm:                      \
        return callback(as<ThermMapEntry>());                 \
      case d5::d4legacy::MapType::Unknown:                    \
        break;                                                \
    }                                                         \
    throw std::runtime_error("encountered unknown map type"); \
  }

D4LEGACY_MAP_ENTRY_DISPATCH
#undef D4LEGACY_MAP_ENTRY_CONST
#define D4LEGACY_MAP_ENTRY_CONST
D4LEGACY_MAP_ENTRY_DISPATCH
#undef D4LEGACY_MAP_ENTRY_DISPATCH

/// \endcond

/// \}
}  // namespace d5::d4legacy
