#pragma once

#include <array>
#include <cstdint>
#include <cstdlib>
#include <iterator>
#include <memory>
#include <ostream>
#include <sstream>
#include <stdexcept>
#include <string>
#include <string_view>
#include <type_traits>
#include <vector>

#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Inheritance.h"
#include "dirsig5/base/StringUtils.h"
#include "dirsig5/d4legacy/Export.h"

namespace d5::d4legacy {

/// \addtogroup d4legacy_tag
/// \{

/** A class for storing a generic "element" in a tag file.
 */
class D5_D4LEGACY_EXPORT TagFileElement
    : public EnumeratedSuperclass<TagFileElement> {
public:
  /// The types of elements
  enum Type {
    UnknownElement,
    SectionBegin,
    SectionEnd,
    Variable,
    IncludeFile,
    UnformattedData
  };

  /// Virtual destructor
  virtual ~TagFileElement() = default;

  /// The type of the element
  [[nodiscard]] Type type() const { return elementType_; }

  /// The name of the tag element
  [[nodiscard]] const std::string& tagName() const { return tagName_; }

  /// Sets the name of the tag
  void setTagName(std::string tagName) { tagName_ = std::move(tagName); }

  /// The line number in the file of the element
  [[nodiscard]] int64_t lineNumber() const { return lineNumber_; }

  /// Sets the line number of the element
  void setLineNumber(int64_t lineNumber) { lineNumber_ = lineNumber; }

  /// Virtual output function to a stream
  virtual void display(std::ostream& os) const { os << tagName() << std::endl; }

protected:
  TagFileElement(Type type) : elementType_(type) {}

private:
  /** A variable defining what type of element this is */
  Type elementType_ = UnknownElement;

  /** The name of this element */
  std::string tagName_;

  /** The line this element started on in the file */
  int64_t lineNumber_ = 0;
};

/** Represents a data item in a DIRSIG3 tag file.
 */
class D5_D4LEGACY_EXPORT TagFileData
    : public TagFileElement::Subclass<TagFileElement::UnformattedData> {
public:
  /// Constructor
  TagFileData() = default;

  /// Destructor
  ~TagFileData() override = default;

  /// Sets the internal data
  void setData(std::string rawData) { rawData_ = std::move(rawData); }

  /// Gets the internal data
  [[nodiscard]] const std::string& data() const { return rawData_; }

  void display(std::ostream& os) const override { os << data() << std::endl; }

private:
  std::string rawData_;
};

/** A class for storing a "section" in a tag file.
 */
class D5_D4LEGACY_EXPORT TagFileSection
    : public TagFileElement::Subclass<TagFileElement::SectionBegin> {
private:
  using vec_type = std::vector<std::unique_ptr<TagFileElement>>;

public:
  /// A custom iterator that dereferences the unique_ptr in out vector
  template <bool Const>
  class dereferencing_iterator {
  public:
    /// Determines the base iterator type based on the const-ness
    using it_t =
      std::conditional_t<Const, vec_type::const_iterator, vec_type::iterator>;

    // Standard iterator typedef boilerplate
    using iterator_category = std::random_access_iterator_tag;
    using value_type =
      std::conditional_t<Const, const TagFileElement&, TagFileElement&>;
    using difference_type = typename it_t::difference_type;
    using reference = value_type;
    using pointer =
      std::conditional_t<Const, const TagFileElement*, TagFileElement*>;

    // Make an iterator from the underlying iterator type implicitly
    dereferencing_iterator(const it_t& i) : it_(i) {}

    // Forward iterator requirements
    reference operator*() const { return **it_; }

    pointer operator->() const { return it_->get(); }

    dereferencing_iterator& operator++() {
      ++it_;
      return *this;
    }

    dereferencing_iterator operator++(int) { return iterator(it_++); }

    // Bidirectional iterator requirements
    dereferencing_iterator& operator--() {
      --it_;
      return *this;
    }

    dereferencing_iterator operator--(int) { return iterator(it_--); }

    // Random access iterator requirements
    reference operator[](const difference_type& n) const { return *it_[n]; }

    dereferencing_iterator& operator+=(const difference_type& n) {
      it_ += n;
      return *this;
    }

    dereferencing_iterator operator+(const difference_type& n) const {
      return iterator(it_ + n);
    }

    dereferencing_iterator& operator-=(const difference_type& n) {
      it_ -= n;
      return *this;
    }

    dereferencing_iterator operator-(const difference_type& n) const {
      return iterator(it_ - n);
    }

    difference_type operator-(const dereferencing_iterator& other) const {
      return it_ - other.it_;
    }

    // Comparison checks
    bool operator==(const dereferencing_iterator& other) const {
      return it_ == other.it_;
    }
    bool operator!=(const dereferencing_iterator& other) const {
      return it_ != other.it_;
    }
    bool operator<(const dereferencing_iterator& other) const {
      return it_ < other.it_;
    }
    bool operator>(const dereferencing_iterator& other) const {
      return it_ > other.it_;
    }
    bool operator<=(const dereferencing_iterator& other) const {
      return it_ <= other.it_;
    }
    bool operator>=(const dereferencing_iterator& other) const {
      return it_ >= other.it_;
    }

  private:
    it_t it_;
  };

  // Make nice aliases for our custom iterator
  using iterator = dereferencing_iterator<false>;
  using const_iterator = dereferencing_iterator<true>;

public:
  /// Default constructor
  TagFileSection() = default;

  /// Move constructor
  TagFileSection(TagFileSection&&) = default;

  /// Move assignment
  TagFileSection& operator=(TagFileSection&&) = default;

  /// Virtual destructor
  ~TagFileSection() override = default;

  /// The name of this section
  [[nodiscard]] const std::string& name() const { return tagName(); }

  /// Sets the name of this section
  void setName(std::string name) { setTagName(std::move(name)); }

  /// The depth of this section in the document
  [[nodiscard]] int32_t sectionDepth() const { return sectionDepth_; }

  /// Sets the depth of this section in the document
  void setSectionDepth(int32_t sectionDepth) { sectionDepth_ = sectionDepth; }

  /// Get the non-const iterator for the first sub-element in the section.
  [[nodiscard]] iterator begin() { return elementVector_.begin(); }

  /// Get the const iterator for the first sub-element in the section.
  [[nodiscard]] const_iterator begin() const { return elementVector_.begin(); }

  /// Get the non-const iterator for the last sub-element in the section.
  [[nodiscard]] iterator end() { return elementVector_.end(); }

  /// Get the non-const iterator for the last sub-element in the section.
  [[nodiscard]] const_iterator end() const { return elementVector_.end(); }

  /// Add an element to this section.
  void push_back(std::unique_ptr<TagFileElement> element) {
    elementVector_.push_back(std::move(element));
  }

  /// The number of elements in this section.
  [[nodiscard]] size_t size() const { return elementVector_.size(); }

  /// The number of elements in this section with a given name.
  [[nodiscard]] size_t elementCount(std::string_view name) const;

  /// Get a non-const element from this section using an index
  [[nodiscard]] TagFileElement& operator[](size_t index) {
    return *elementVector_[index];
  }

  /// Get an const element from this section using an index
  [[nodiscard]] const TagFileElement& operator[](size_t index) const {
    return *elementVector_[index].get();
  }

  /** Get an element with a given name from this section.
   *
   *  \param name          The name of the element to get
   *  \param instance      If more than one element exist with the same given
   *                       name, then the instance can be used to specify
   *                       which to return
   */
  [[nodiscard]] const TagFileElement* element(
    std::string_view name, int instance = 0) const;

  /// \overload
  [[nodiscard]] TagFileElement* element(
    std::string_view name, int instance = 0) {
    return const_cast<TagFileElement*>(
      static_cast<const TagFileSection*>(this)->element(name, instance));
  }

  /** A version of getElement that throws a std::domain_error if the element
   *  does not exist.
   */
  [[nodiscard]] const TagFileElement& requiredElement(
    std::string_view name, int instance = 0) const;

  /// \overload
  [[nodiscard]] TagFileElement& requiredElement(
    std::string_view name, int instance = 0) {
    return const_cast<TagFileElement&>(
      static_cast<const TagFileSection*>(this)->requiredElement(
        name, instance));
  }

  /** Get a section with a given name from this section. If the section does
   *  not extist, a null pointer is returned.
   *
   *  \param name      The name of the section to get
   *  \param instance  If more than one section with the given name exists,
   *                   which instance to return
   */
  [[nodiscard]] const TagFileSection* subsection(
    std::string_view name, int instance = 0) const;

  /// \overload
  [[nodiscard]] TagFileSection* subsection(
    std::string_view name, int instance = 0) {
    return const_cast<TagFileSection*>(
      static_cast<const TagFileSection*>(this)->subsection(name, instance));
  }

  /** Get a section with a given name from this section. Throws if the section
   *  does not exist
   *
   *  \param name      The name of the section to get
   *  \param instance  If more than one section with the given name exists,
   *                   which instance to return
   */
  [[nodiscard]] const TagFileSection& requiredSubsection(
    std::string_view name, int instance = 0) const;

  /// \overload
  [[nodiscard]] TagFileSection& requiredSubsection(
    std::string_view name, int instance = 0) {
    return const_cast<TagFileSection&>(
      static_cast<const TagFileSection*>(this)->requiredSubsection(
        name, instance));
  }

  /** Simple helper struct for optional variables that allow the assignment of
   *  a default value. Use an expression of the form:
   *
   *  std::string val;
   *  section.optionalVariable("name", 0, val).Or("not_there");
   */
  template <typename T>
  class OptVarContinuation {
  public:
    OptVarContinuation(bool _found, T& ref) : found(_found), ref_(ref) {}

    void Or(const T& defaultVal) {
      if (!found) ref_ = defaultVal;
    }

    bool found;

  private:
    T& ref_;
  };

  /** Gets the value of a variable that may or may not be present
   *
   *  \param[in] name      The name of the variable
   *  \param[in] instance  The instance of the variable if > 1 are present
   *  \param[out] val      The output variable
   *
   *  \return A continuation struct that can be used to test if the variable
   * was there and set a default value.
   */
  template <typename T>
  OptVarContinuation<T> optionalVariable(
    std::string_view name, int instance, T& val) const;

  /// \overload
  template <typename T>
  [[nodiscard]] T optionalVariable(
    std::string_view name, int instance = 0, const T& defaultValue = T()) const;

  /** Gets the value of a variable and throws if it is not there
   *
   *  \param[in] name      The name of the variable
   *  \param[in] instance  The instance of the variable if > 1 are present
   *  \param[out] v        The output variable
   */
  template <typename T>
  void requiredVariable(std::string_view name, int instance, T& v) const;

  /// \overload
  template <typename T>
  [[nodiscard]] T requiredVariable(
    std::string_view name, int instance = 0) const {
    T v;
    requiredVariable(name, instance, v);
    return v;
  }

  /// Display (write) the contents of this section to the output stream.
  void display(std::ostream& os) const override;

private:
  /** The depth of this section in the hierarchy */
  int32_t sectionDepth_ = 0;

  /** The vector of sub-elements in this section */
  vec_type elementVector_;
};

template <typename T>
struct from_tag_string {};

/** A class for storing a "variable" in a tag file
 */
class D5_D4LEGACY_EXPORT TagFileVariable
    : public TagFileElement::Subclass<TagFileElement::Variable> {
public:
  ~TagFileVariable() override = default;

  /// The value of the variable as a string
  [[nodiscard]] const std::string& value() const { return value_; }

  /// Sets the value of the variable as a string
  void setValue(std::string s) { value_ = std::move(s); }

  /// Templated accessor for the value
  template <typename T>
  void get_to(T& v) const {
    from_tag_string<std::decay_t<T>>()(value_, v);
  }

  /// Templated accessor for the value
  template <typename T>
  [[nodiscard]] T get() const {
    T v;
    get_to(v);
    return v;
  }

  void display(std::ostream& os) const override {
    os << tagName() << " = " << value_ << std::endl;
  }

private:
  /** The value associated with the variable */
  std::string value_;
};

// Template implementation
/// \cond HIDDEN_SYMBOLS

template <typename T>
auto TagFileSection::optionalVariable(
  std::string_view name, int instance, T& v) const -> OptVarContinuation<T> {
  const auto* e = element(name, instance);
  if (e && e->type() == TagFileElement::Variable) {
    e->as<TagFileVariable>().get_to(v);
    return OptVarContinuation<T>(true, v);
  }

  return OptVarContinuation<T>(false, v);
}

template <typename T>
T TagFileSection::optionalVariable(
  std::string_view name, int instance, const T& defaultValue) const {
  T v;
  optionalVariable(name, instance, v).Or(defaultValue);
  return v;
}

template <typename T>
inline void TagFileSection::requiredVariable(
  std::string_view name, int instance, T& v) const {
  const auto* e = element(name, instance);
  if (!e || e->type() != TagFileElement::Variable) {
    std::ostringstream errorMsg;
    errorMsg << "TagFileSection::requiredVariable:\n"
             << "    Failed to find required variable!\n"
             << "    Variable = " << name << "\n"
             << "    Section starts at line number = " << lineNumber() << "\n";
    throw std::domain_error(errorMsg.str());
  }
  e->as<TagFileVariable>().get_to(v);
}

template <>
inline void TagFileVariable::get_to(bool& v) const {
  v = value_ == "TRUE";
}

template <>
inline void TagFileVariable::get_to(std::string& v) const {
  v = trimmed(value_);
}

template <>
inline void TagFileVariable::get_to(double& v) const {
  v = ParseNumber<double>(value_);
}

template <>
inline void TagFileVariable::get_to(float& v) const {
  v = ParseNumber<float>(value_);
}

template <>
inline void TagFileVariable::get_to(int32_t& v) const {
  v = ParseNumber<int32_t>(value_);
}

template <>
inline void TagFileVariable::get_to(int64_t& v) const {
  v = ParseNumber<int64_t>(value_);
}

template <>
inline void TagFileVariable::get_to(uint32_t& v) const {
  v = ParseNumber<uint32_t>(value_);
}

template <>
inline void TagFileVariable::get_to(uint64_t& v) const {
  v = ParseNumber<uint64_t>(value_);
}

namespace tag {

template <typename T>
inline void ParseVector(const std::string& str, std::vector<T>& v) {
  if constexpr (std::is_same_v<T, std::string>) {
    auto isDelimiter = [](const std::string& delims, char c) {
      return delims.find(c) != std::string::npos;
    };

    // walk through the source string and make tokens
    std::vector<std::string> tokenVector;
    auto stringEnd = str.end();
    for (auto it = str.begin(); it != stringEnd; ++it) {
      std::string token;
      std::insert_iterator<std::string> ii(token, token.begin());
      while (it != stringEnd && isDelimiter(", \t", *it)) it++;
      while (it != stringEnd && !isDelimiter(", \t", *it)) *ii++ = *it++;
      v.push_back(token);
      if (it == stringEnd) break;
    }
  } else {
    std::istringstream iss(str);
    T a;
    while (iss >> a) {
      v.push_back(std::move(a));
      while (iss.good()) {
        if (char ch = iss.peek(); ch == ',' || ch == ' ' || ch == '\t')
          iss.ignore();
        else
          break;
      }
    }
  }
}

template <typename T>
inline std::vector<T> ParseVector(const std::string& str) {
  std::vector<T> v;
  ParseVector(str, v);
  return v;
}

template <typename T, size_t N>
inline void ParseArray(const std::string& str, std::array<T, N>& a) {
  std::istringstream iss(str);
  if constexpr (std::is_same_v<T, std::string>) {
    auto tokens = ParseVector<std::string>(str);
    if (tokens.size() != N)
      throw std::runtime_error(
        "incorrect number of strings (" + std::to_string(tokens.size()) +
        " vs. " + std::to_string(N));
    for (size_t i = 0; i < N; i++) a[i] = std::move(tokens[i]);
  } else {
    for (size_t i = 0; i < N; i++) {
      if (!(iss >> a[i]))
        throw std::runtime_error(
          "unable to parse element " + std::to_string(i) + " in array");
      while (iss.good()) {
        if (char ch = iss.peek(); ch == ',' || ch == ' ' || ch == '\t')
          iss.ignore();
        else
          break;
      }
    }
  }
}

template <typename T, size_t N>
inline std::array<T, N> ParseArray(const std::string& str) {
  std::array<T, N> a;
  ParseArray(str, a);
  return a;
}

inline Point ParsePoint(const std::string& str) {
  auto a = ParseArray<double, 3>(str);
  return Point(a[0], a[1], a[2]);
}

}  // namespace tag

template <typename T, size_t N>
struct from_tag_string<std::array<T, N>> {
  void operator()(const std::string& s, std::array<T, N>& a) const {
    d5::d4legacy::tag::ParseArray(s, a);
  }
};

template <typename T>
struct from_tag_string<std::vector<T>> {
  void operator()(const std::string& s, std::vector<T>& v) const {
    d5::d4legacy::tag::ParseVector(s, v);
  }
};

template <>
struct from_tag_string<Point> {
  void operator()(const std::string& s, Point& v) const {
    std::array<double, 3> a;
    d5::d4legacy::tag::ParseArray(s, a);
    std::copy(a.begin(), a.end(), v.data());
  }
};

/// \endcond

/// \}
}  // namespace d5::d4legacy
