#pragma once

#include <stdexcept>
#include <string>
#include <string_view>

#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/StringUtils.h"
#include "dirsig5/d4legacy/Export.h"
#include "dirsig5/d4legacy/geometry/Rotation.h"
#include "dirsig5/d4legacy/misc/FileLocator.h"
#include "dirsig5/d4legacy/misc/Units.h"
#include "dirsig5/d4legacy/xml/pugixml/pugixml.hpp"
#include "dirsig5/geolib/Facade.h"
#include "dirsig5/geolib/Location.h"

namespace d5::d4legacy::xml {

using Document = pugi::xml_document;
using Node = pugi::xml_node;
using Attr = pugi::xml_attribute;

/// Load an XML document from a file.
D5_D4LEGACY_EXPORT Document loadFile(const char* filename);

/// Load an XML document from a string.
D5_D4LEGACY_EXPORT Document loadString(const char* file_content);

/// \overload
inline Document loadFile(const std::string& filename) {
  return loadFile(filename.c_str());
}

/// \overload
inline Document loadString(const std::string& file_content) {
  return loadString(file_content.c_str());
}

D5_D4LEGACY_EXPORT Node requiredChild(const Node& node, const char* name);

D5_D4LEGACY_EXPORT Attr requiredAttribute(const Node& node, const char* name);

D5_D4LEGACY_EXPORT long lineno(const Node& node);

template <typename Range>
inline size_t size(const Range& range) {
  return std::distance(range.begin(), range.end());
}

inline double readNumber(
  const Node& node, const char* name, double or_else = 0.0) {
  auto subnode = node.child(name);
  if (!subnode) return or_else;
  return subnode.text().as_double(or_else);
}

inline double readRequiredNumber(const Node& node, const char* name) {
  return requiredChild(node, name).text().as_double();
}

D5_D4LEGACY_EXPORT DateTime
readDatetime(const Node& node, const char* name = "datetime");

/// Read XYZ.
///
/// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/// <node><x>1</x><y>2</y><z>3</z></node>
/// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
inline Point readCoordinate(const Node& node, Point or_else = {0, 0, 0}) {
  return {
    readNumber(node, "x", or_else[0]), readNumber(node, "y", or_else[1]),
    readNumber(node, "z", or_else[2])};
}

/// Read XYZ wrapped in another node, or use a default value.
///
/// This is helpful to read XML following this pattern:
/// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/// <node>
///   <name>
///     <wrap><x>1</x><y>2</y><z>3</z></wrap>
///   </name>
/// </node>
/// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/// where `<wrap>` can have any name, or not be present at all. If
/// present, it is typically `<point>`, `<vector>`, or `<cartesiantriple>`
/// in D4-era XML files.
///
inline Point readCoordinate(
  const Node& node, const char* name, Point or_else = {0, 0, 0}) {
  auto subnode = node.child(name);
  if (!subnode) return or_else;
  if (size(subnode.children()) <= 1) {
    auto triplet = subnode.first_child();  // <point>, <vector>, ...
    if (!triplet) return or_else;
    return readCoordinate(triplet, or_else);
  } else
    return readCoordinate(subnode, or_else);
}

D5_D4LEGACY_EXPORT std::string readFilename(
  const Node& node,
  const char* name,
  const FileLocator& fileLocator,
  const FileLocator::FileType type = FileLocator::Any);

D5_D4LEGACY_EXPORT DistanceUnits
readDistanceUnitsAttribute(const Node& node, const char* name = "units");

D5_D4LEGACY_EXPORT WavelengthUnits
readWavelengthUnitsAttribute(const Node& node, const char* name = "units");

D5_D4LEGACY_EXPORT double readAngularUnitsAttributeAsMultiplier(
  const Node& node, const char* name = "angularunits");

D5_D4LEGACY_EXPORT RotationOrder readRotationOrderAttribute(
  const Node& node, const char* name = "rotationorder");

D5_D4LEGACY_EXPORT geolib::Location readLocation(
  const Node& node, const geolib::Facade* facade);

inline std::vector<std::string> readTags(const Node& node) {
  std::vector<std::string> tags;
  if (auto attr = node.attribute("tags"); attr)
    for (auto tag : TokenList(attr.as_string(), " ,"))
      tags.emplace_back(std::string(tag));
  return tags;
}

}  // namespace d5::d4legacy::xml
