/** Interface file for dealing with byte orders and byte order swapping
 *
 *  \file Endian.h
 */

#pragma once

#include <cstdint>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <utility>

namespace d5::envi {

/// Constants for the different type of endianness
enum class Endian : std::uint8_t {
  Unknown, ///< Unknown
  Little,  ///< Little, least significant byte first
  Big      ///< Big, most significant byte first
};

inline std::string to_string(Endian e) {
  switch (e) {
  case Endian::Unknown:
    return "unknown";
  case Endian::Little:
    return "little endian";
  case Endian::Big:
    return "big endian";
  }

  throw std::domain_error("invalid endian value");
}

/// Function to test for the endian-ness of the native system.
inline Endian nativeByteOrder() {
  union probe {
    std::uint16_t num;
    unsigned char bytes[sizeof(uint16_t)];
  };

  // initialze the first member of p with an unsigned 1
  probe p = {1U};

  // on a big endian architecture, p.bytes[0] equals 0,
  // on a little endian architecture, p.bytes[0] equals 1
  if (p.bytes[0] == 0U) {
    return Endian::Big;
  } else if (p.bytes[0] == 1U) {
    return Endian::Little;
  } else {
    return Endian::Unknown;
  }
}

/** Swaps the endian of the given value
 *
 *  \tparam T   The type of the value, must be either 1, 2, 4 or 8 byte
 *              arithmetic type
 *  \param val  The value to endian swap
 */
template <typename T> void swapEndian(T &val) {
  static_assert(std::is_arithmetic_v<T>, "Only arithmetic types are supported");
  static_assert(sizeof(T) == 1 || sizeof(T) == 2 || sizeof(T) == 4 ||
                    sizeof(T) == 8,
                "T must either be 1, 2, 4, or 8 bytes");
  if constexpr (sizeof(T) == 1)
    return;

  auto data = reinterpret_cast<char *>(&val);
  if constexpr (sizeof(T) == 2) {
    auto tmp = data[1];
    data[1] = data[0];
    data[0] = tmp;
  } else if constexpr (sizeof(T) == 4) {
    char tmp = data[3];
    data[3] = data[0];
    data[0] = tmp;

    tmp = data[2];
    data[2] = data[1];
    data[1] = tmp;
  } else {
    auto tmp = data[7];
    data[7] = data[0];
    data[0] = tmp;

    tmp = data[6];
    data[6] = data[1];
    data[1] = tmp;

    tmp = data[5];
    data[5] = data[2];
    data[2] = tmp;

    tmp = data[4];
    data[4] = data[3];
    data[3] = tmp;
  }
}
} // namespace d5::envi
