/** Interface file for the ENVI image header
 *
 *  \file Header.h
 */

#pragma once

#include <array>
#include <optional>
#include <string>
#include <unordered_map>
#include <vector>

#include "Endian.h"
#include "Interleave.h"
#include "WavelengthUnits.h"
#include "dirsig5/base/DType.h"
#include "dirsig5/envi/Export.h"

namespace d5::envi {

struct Header {
  std::string acquisitionTime = "";
  std::vector<std::string> bandNames;
  std::size_t bands = 0;
  std::vector<double> badBandMultipliers;
  Endian byteOrder = Endian::Unknown;
  double cloudCover = 0;

  enum class ComplexFunction { Real, Imaginary, Power, Magnitude, Phase };
  ComplexFunction complexFunction = ComplexFunction::Power;

  std::string coordSystemString;

  std::vector<double> dataGainValues; // [W / (m^2 * micron * sr)]

  // double dataIgnoreValue;  // ambiguous whether scalar or vector

  std::string dataUnits = "";

  std::vector<double> dataOffsetValues;

  std::vector<double> dataReflectanceGainValues;

  std::vector<double> dataReflectanceOffsetValues;

  DType dataType = DType::Uint8;

  std::vector<std::size_t> defaultBands;

  std::string defaultStretch;

  std::size_t demBand = 0;

  std::string demFile;

  // File type must be an image and is omitted

  std::string description = "";

  std::vector<double> fwhm;

  std::vector<double> geoPoints;

  std::size_t headerOffset = 0;

  Interleave interleave = Interleave::BIL;

  std::size_t lines = 0;

  std::string mapInfo;

  double pixelSize = 0;

  // Skipping legacy support (projection info, read procedures)

  std::optional<double> reflectanceScaleFactor;

  std::string rpcInfo;

  std::size_t samples = 0;

  std::string securityTag;

  std::string sensorType;

  std::optional<double> solarIrradiance; // W/(m^2 * micron)

  std::optional<double> sunAzimuthDegCwFromN;

  std::optional<double> sunElevationDeg;

  std::vector<double> wavelength;

  WavelengthUnits wavelengthUnits = WavelengthUnits::Micrometers;

  std::optional<std::size_t> xStart, yStart;

  std::vector<std::string> spectraNames; // for SLI

  std::unordered_map<std::string, std::string> nonstandardFields;

  // Skipping z-plot (ambiguous documentation)
};

D5_ENVI_EXPORT const std::string &to_string(Header::ComplexFunction c);
} // namespace d5::envi
