/** Interface file for ENVI image I/O
 *
 *  \file ImageIO.h
 */

#pragma once

#include <array>
#include <fstream>
#include <optional>
#include <string>
#include <vector>

#include "dirsig5/base/DType.h"
#include "dirsig5/envi/Header.h"
#include "dirsig5/envi/HeaderIO.h"
#include "dirsig5/envi/ImageReadRequest.h"
#include "dirsig5/envi/Interleave.h"
#include "dirsig5/envi/readers/BilReader.h"
#include "dirsig5/envi/readers/BipReader.h"
#include "dirsig5/envi/readers/BsqReader.h"

namespace d5::envi {

struct ImageDestination {
  virtual void prepare(const Header &header) {}
  virtual void allocate(size_t bands, size_t rows, size_t columns,
                        DType dataType) = 0;
  [[nodiscard]] virtual Interleave interleave() const = 0;

  virtual char *row(size_t band, size_t row) = 0;
};

struct BsqImageDestination : ImageDestination {
  [[nodiscard]] Interleave interleave() const override {
    return Interleave::BSQ;
  }
};

struct BilImageDestination : ImageDestination {
  [[nodiscard]] Interleave interleave() const override {
    return Interleave::BIL;
  }
};

struct BipImageDestination : ImageDestination {
  [[nodiscard]] Interleave interleave() const override {
    return Interleave::BIP;
  }
};

template <typename Destination>
Header LoadImage(const ImageReadRequest &request, Destination &&dst) {
  static_assert(!std::is_const_v<Destination>, "dst must be non-const");
  static_assert(std::is_base_of_v<ImageDestination, std::decay_t<Destination>>,
                "dst must be a subclass of ImageDestination");

  // Load the header
  Header hdr;
  std::string hdrFilename;
  {
    std::ifstream ifs;
    if (request.hasHeaderFilename()) {
      hdrFilename = request.headerFilename();
      ifs.open(hdrFilename);
      if (!ifs) {
        throw std::runtime_error("unable to open ENVI header \"" +
                                 request.headerFilename() +
                                 "\":" + strerror(errno));
      }
    } else {
      std::array<std::string, 2> paths{request.imageFilename() + ".hdr",
                                       request.imageFilename()};
      if (paths[1].size() < 3)
        throw std::runtime_error("input path too short");
      paths[1][paths[1].size() - 3] = 'h';
      paths[1][paths[1].size() - 2] = 'd';
      paths[1][paths[1].size() - 1] = 'r';

      std::size_t pathIdx = 0;
      ifs.open(paths[0]);
      if (!ifs) {
        pathIdx = 1;
        auto firstErr = errno;
        ifs.open(paths[1]);
        if (!ifs) {
          throw std::runtime_error("unable to open ENVI header, tried:\n  \"" +
                                   paths[0] + "\": " + strerror(firstErr) +
                                   "\n  \"" + paths[1] +
                                   "\": " + strerror(errno));
        }
      }
      hdrFilename = std::move(paths[pathIdx]);
    }

    try {
      hdr = ReadHeader(ifs);
    } catch (const std::exception &e) {
      throw std::runtime_error("while loading header from \"" + hdrFilename +
                               "\":\n" + e.what());
    }
  }

  dst.prepare(hdr);

  // Open the data file
  std::ifstream ifs(request.imageFilename(), std::ios_base::binary);
  if (!ifs) {
    throw std::runtime_error(std::string("unable to open ENVI image: ") +
                             strerror(errno));
  }

  // Determine the read sizes and allocate the output
  auto params = DetermineReadParameters(hdr, request);
  Allocate(dst, params, hdr);

  // Read in based on the interleave
  switch (hdr.interleave) {
  case Interleave::BSQ:
    BsqRead(ifs, hdr, request, params, dst);
    break;
  case Interleave::BIL:
    BilRead(ifs, hdr, request, params, dst);
    break;
  case Interleave::BIP:
    BipRead(ifs, hdr, request, params, dst);
    break;
  default:
    throw std::runtime_error("invalid interleave in the header");
  }

  request.postProgress(1);

  if (hdr.byteOrder != nativeByteOrder()) {
    SwapEndian(dst, params, hdr.dataType);
  }

  return hdr;
}

D5_ENVI_EXPORT void SwapEndian(ImageDestination &dst,
                               const ReadParameters &params, DType dtype);
} // namespace d5::envi
