/** Interface file for ENVI image I/O
 *
 *  \file ImageIO.h
 */

#pragma once

#include <array>
#include <functional>
#include <optional>
#include <stdexcept>
#include <string>
#include <vector>

#include "Interleave.h"

namespace d5::envi {

struct ImageReadRequest {
public:
  [[nodiscard]] const std::string &imageFilename() const {
    if (!imageFilename_.empty())
      return imageFilename_;
    throw std::runtime_error("no image filename was given with this request");
  }

  ImageReadRequest &withImageFilename(std::string filename) {
    imageFilename_ = std::move(filename);
    return *this;
  }

  [[nodiscard]] bool hasHeaderFilename() const {
    return !headerFilename_.empty();
  }

  [[nodiscard]] const std::string &headerFilename() const {
    return headerFilename_;
  }

  ImageReadRequest &withHeaderFilename(std::string filename) {
    headerFilename_ = std::move(filename);
    return *this;
  }

  [[nodiscard]] bool hasRoi() const { return roi_.has_value(); }
  [[nodiscard]] const std::array<std::size_t, 4> &roi() const {
    if (!hasRoi())
      throw std::runtime_error("image read request has no ROI");
    return *roi_;
  }

  ImageReadRequest &withRoiFromMinSize(size_t minX, size_t minY, size_t sizeX,
                                       size_t sizeY) {
    roi_.reset();
    auto &r = roi_.emplace();
    r[0] = minX;
    r[1] = minY;
    r[2] = sizeX;
    r[3] = sizeY;
    return *this;
  }

  ImageReadRequest &withRoiFromMinMaxInclusive(size_t minX, size_t minY,
                                               size_t maxX, size_t maxY) {
    roi_.reset();
    auto &r = roi_.emplace();
    r[0] = minX;
    r[1] = minY;
    r[2] = maxX - minX + 1;
    r[3] = maxY - minY + 1;
    return *this;
  }

  ImageReadRequest &withRoiFromMinMaxExclusive(size_t minX, size_t minY,
                                               size_t maxX, size_t maxY) {
    roi_.reset();
    auto &r = roi_.emplace();
    r[0] = minX;
    r[1] = minY;
    r[2] = maxX - minX;
    r[3] = maxY - minY;
    return *this;
  }

  [[nodiscard]] bool hasBandsToLoad() const { return !bandsToLoad_.empty(); }
  [[nodiscard]] const std::vector<std::size_t> &bandsToLoad() const {
    return bandsToLoad_;
  }
  ImageReadRequest &withBandsToLoad(std::vector<std::size_t> bandsToLoad) {
    bandsToLoad_ = std::move(bandsToLoad);
    return *this;
  }

  [[nodiscard]] bool loadDefaultBands() const { return loadDefaultBands_; }
  ImageReadRequest &withLoadDefaultBands(bool loadDefaultBands) {
    loadDefaultBands_ = loadDefaultBands;
    return *this;
  }

  [[nodiscard]] size_t pixelStride() const { return pixelStride_; }
  ImageReadRequest &withPixelStride(size_t stride) {
    pixelStride_ = stride;
    return *this;
  }

  ImageReadRequest &withProgressCallback(std::function<void(double)> f) {
    progressCallback_ = std::move(f);
    return *this;
  }

  void postProgress(double p) const {
    if (progressCallback_)
      progressCallback_(p);
  }

private:
  std::string imageFilename_;
  std::string headerFilename_;
  std::optional<std::array<std::size_t, 4>> roi_;
  std::vector<std::size_t> bandsToLoad_;
  bool loadDefaultBands_ = false;
  size_t pixelStride_ = 1;
  std::function<void(double)> progressCallback_;
};
} // namespace d5::envi
