#pragma once

#include <variant>

#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/envi/Export.h"

namespace d5::envi {

class D5_ENVI_EXPORT SpectralLibrary {
public:
  SpectralLibrary() noexcept = default;

  void clear() {
    description.clear();
    wavelength.clear();
    spectra.clear();
    spectraNames.clear();
    reflectanceScale = 1;
  }

  /// \name IO functions
  ///
  /// \note
  /// Wherever a filename is given, it is understood to be the name of
  /// the _image_ file on disk, typically ending with either `.sli` or `.img`.
  /// The header file is formed by adding `.hdr`, though the read functions
  /// will also detect header filenames formed by replacing the extension
  /// with `.hdr`.
  ///
  /// \note
  /// All functions throw `std::runtime_error` on failure.
  ///
  /// \{

  /// Read in a specific curve from file, either by index or spectrum name.
  void read(const std::string &filename,
            const std::variant<int, std::string> &indexOrName);

  /// Read in from file.
  void read(const std::string &filename);

  /// Write to file.
  void write(const std::string &filename) const;

  /// \}

  /// Reinterpolate the spectra to new wavelengths.
  template <typename Container>
  void reinterpolate(const Container &newWavelength) {
    for (auto &spectrum : spectra) {
      SpectralVector newSpectrum;
      interpolate(wavelength, spectrum, newWavelength, newSpectrum);
      spectrum = newSpectrum;
    }
    wavelength = newWavelength;
  }

  struct Statistics {
    Eigen::VectorXd mean; ///< The mean.
    Eigen::VectorXd var;  ///< The variance (diagonal of the covariance matrix).
    Eigen::MatrixXd cov;  ///< The covariance matrix.
  };

  /// Compute statistics.
  [[nodiscard]] Statistics computeStatistics() const;

  /// Look-up curve by index (unsafely!).
  [[nodiscard]] auto &operator[](size_t index) noexcept {
    return spectra[index];
  }

  /// Look-up curve by index (unsafely!), const variant.
  [[nodiscard]] auto &operator[](size_t index) const noexcept {
    return spectra[index];
  }

public:
  /// A string description of the spectra in the library.
  std::string description;

  /// The wavelengths of the spectra in micrometers.
  std::vector<double> wavelength;

  /// The spectra. Every curve should be the same size as `wavelength`.
  std::vector<SpectralVector> spectra;

  /// The spectra names. If non-empty, should be the same size as `spectra`.
  std::vector<std::string> spectraNames;

  /// The reflectance scale factor.
  double reflectanceScale = 1;
};

} // namespace d5::envi
