/** ENVI statistics file.
 *
 *  \file StatisticsFile.h
 */
#pragma once

#include <vector>

#include "dirsig5/envi/Export.h"

namespace d5::envi {

/**
 * A class to store the contents of an ENVI statistics (.sta) file.
 */
class D5_ENVI_EXPORT StatisticsFile {

public:
  /**
   * The header structure contains information about the dimensions of the
   * data and what region-of-interest (ROI) was used to product it.
   */
  struct Header {
    int32_t id;
    int32_t sampleCount;
    int32_t lineCount;
    int32_t bandCount;
    int32_t dataType;
    int32_t roiIndex;
    int32_t startSample;
    int32_t endSample;
    int32_t startLine;
    int32_t endLine;
  };

  /**
   * This structure contains the wavelengths and the offsets to
   * other data in the file.
   */
  struct Offsets {
    std::vector<int32_t> histogramOffsets;
    int32_t covarianceOffset;
    int32_t roiNameLength;
    std::string roiName;
    std::vector<float> wavelengths;
  };

  /**
   * This structure tracks which bands have stats.
   */
  struct BandFlags {
    char *hasStats;
  };

  /**
   * This structure contains the min, max, meand and stddev for each band.
   */
  struct BandStats {
    std::vector<double> bandMin;
    std::vector<double> bandMax;
    std::vector<double> bandMean;
    std::vector<double> bandStddev;
  };

  /**
   * This struct contains how many bands are in the covariance data
   * and then the list of which indexes.
   */
  struct CovarianceHeader {
    int32_t bandCount;
    std::vector<int32_t> bandIndexes;
  };

  /**
   * This struct contains the spectral covariance matrix, eigen matrix and
   * eigen value vector.
   */
  struct CovarianceData {
    std::vector<double> covarianceMatrix;
    std::vector<double> eigenMatrix;
    std::vector<double> eigenValues;
  };

  /**
   * Read a binary STA file into a data structure.
   */
  bool read(const std::string &filename);

  Header header;
  Offsets offsets;
  BandFlags bandFlags;
  BandStats bandStats;
  CovarianceHeader covarianceHeader;
  CovarianceData covarianceData;
  bool swapBytes;

private:
  /**
   * A function to perform endian related byte swapping on any data type.
   */
  template <typename T> T byteSwap(T val) {
    T retVal;
    char *pVal = (char *)&val;
    char *pRetVal = (char *)&retVal;
    int size = sizeof(T);
    for (int i = 0; i < size; i++) {
      pRetVal[size - 1 - i] = pVal[i];
    }
    return retVal;
  }
};

} // namespace d5::envi
