/** Interface file for reading BIL images
 *
 *  \file BilReader.h
 */

#pragma once

#include <cstring>
#include <istream>
#include <memory>
#include <stdexcept>

#include "dirsig5/envi/Header.h"
#include "dirsig5/envi/ImageReadRequest.h"
#include "dirsig5/envi/readers/Common.h"

namespace d5::envi {

template <typename Destination>
void BilRead(std::istream &ifs, const Header &hdr, const ImageReadRequest &req,
             const ReadParameters &params, Destination &&dst) {
  // Pre-compute some sizing info
  size_t imgBandRowSize = hdr.samples * params.byteSize;
  size_t imgFullRowSize = imgBandRowSize * hdr.bands;
  size_t imgWindowedBandRowSize = params.nonStridedNumCols * params.byteSize;
  size_t bipPixSize = params.numBands * params.byteSize;

  // A helper to go to the start of a row to read
  auto GoToBandRow = [&](size_t b, size_t r) {
    ifs.seekg(hdr.headerOffset + r * imgFullRowSize + b * imgBandRowSize +
                  params.startCol * params.byteSize,
              std::ios_base::beg);
  };

  // Local buffer (if needed)
  bool useLocalBuffer =
      dst.interleave() == Interleave::BIP || req.pixelStride() != 1;
  std::unique_ptr<char[]> localBuffer;
  if (useLocalBuffer) {
    localBuffer = std::make_unique<char[]>(imgWindowedBandRowSize);
  }

  // Go through each band/row, reading line-by-line
  for (size_t r = 0; r < params.stridedNumRows; r++) {
    size_t row = params.startRow + r * req.pixelStride();
    for (size_t b = 0; b < params.numBands; b++) {
      char *dstBuff = dst.row(b, r);
      char *srcBuff = useLocalBuffer ? localBuffer.get() : dstBuff;
      if (!dstBuff)
        continue;

      // Read into either the destination or the local buffer
      size_t band = params.useBandRemapping ? params.bandsToLoad[b] : b;
      GoToBandRow(band, row);
      ifs.read(srcBuff, imgWindowedBandRowSize);
      if (static_cast<size_t>(ifs.gcount()) != imgWindowedBandRowSize) {
        throw std::runtime_error("unexpected EOF when reading band " +
                                 std::to_string(band) + " line " +
                                 std::to_string(row));
      }

      // If we used a local buffer, move into the user buffer
      if (dst.interleave() == Interleave::BIP) {
        // Copy each pixel into its appropriate location
        for (size_t dstCol = 0; dstCol < params.stridedNumCols; dstCol++) {
          size_t srcCol = dstCol * req.pixelStride();
          memcpy(dstBuff + dstCol * bipPixSize + b * params.byteSize,
                 srcBuff + srcCol * params.byteSize, params.byteSize);
        }
      } else if (req.pixelStride() != 1) {
        for (size_t dstCol = 0; dstCol < params.stridedNumCols; dstCol++) {
          size_t srcCol = dstCol * req.pixelStride();
          memcpy(dstBuff + dstCol * params.byteSize,
                 srcBuff + srcCol * params.byteSize, params.byteSize);
        }
      }
    }

    // Post a progress update occasionally
    if ((r + 1) % params.progressRowInterval == 0) {
      req.postProgress((r + 1.) / params.stridedNumRows);
    }
  }
}
} // namespace d5::envi
