/** Interface file for reading BIP images
 *
 *  \file BipReader.h
 */

#pragma once

#include <cstring>
#include <istream>
#include <memory>
#include <stdexcept>

#include "dirsig5/envi/Header.h"
#include "dirsig5/envi/ImageReadRequest.h"
#include "dirsig5/envi/readers/Common.h"

namespace d5::envi {

template <typename Destination>
void BipRead(std::istream &ifs, const Header &hdr, const ImageReadRequest &req,
             const ReadParameters &params, Destination &&dst) {
  // Pre-compute some sizing info
  size_t imgFullPixelSize = hdr.bands * params.byteSize;
  size_t imgFullRowSize = hdr.samples * imgFullPixelSize;
  size_t imgWindowedRowSize = params.nonStridedNumCols * imgFullPixelSize;
  size_t dstBipPixSize = params.numBands * params.byteSize;

  // A helper to go to the start of a row to read
  auto GoToRow = [&](size_t r) {
    ifs.seekg(hdr.headerOffset + r * imgFullRowSize +
                  params.startCol * imgFullPixelSize,
              std::ios_base::beg);
  };

  // Full image load case into a BIP buffer (optimized simple load case)
  if (dst.interleave() == Interleave::BIP && req.pixelStride() == 1 &&
      !params.useBandRemapping) {
    for (size_t r = 0; r < params.stridedNumRows; r++) {
      char *dstBuff = dst.row(0, r);
      if (!dstBuff)
        continue; // For SLI curve selection
      size_t row = params.startRow + r * req.pixelStride();
      GoToRow(row);
      ifs.read(dstBuff, imgWindowedRowSize);
      if (static_cast<size_t>(ifs.gcount()) != imgWindowedRowSize) {
        throw std::runtime_error(
            "unexpected EOF when reading line " + std::to_string(row) +
            ", expected " + std::to_string(imgWindowedRowSize) +
            " bytes, but got only " + std::to_string(ifs.gcount()));
      }
      if ((r + 1) % params.progressRowInterval == 0) {
        req.postProgress((r + 1.) / params.stridedNumRows);
      }
    }
    return;
  }

  auto srcBuff = std::make_unique<char[]>(imgWindowedRowSize);

  // Go through each row, reading the image in. For now, we're just going to
  // read the full windowed row and filter down.
  for (size_t r = 0; r < params.stridedNumRows; r++) {
    size_t row = params.startRow + r * req.pixelStride();
    GoToRow(row);

    ifs.read(srcBuff.get(), imgWindowedRowSize);
    if (static_cast<size_t>(ifs.gcount()) != imgWindowedRowSize) {
      throw std::runtime_error("unexpected EOF when reading line " +
                               std::to_string(row));
    }

    switch (dst.interleave()) {
    // For BSQ and BIL, we need to go value by value and transpose
    case Interleave::BSQ:
    case Interleave::BIL:
      for (size_t b = 0; b < params.numBands; b++) {
        auto *dstBuff = dst.row(b, r);
        if (!dstBuff)
          continue; // For SLI curve selection
        auto band = params.useBandRemapping ? params.bandsToLoad[b] : b;
        for (size_t dstCol = 0; dstCol < params.stridedNumCols; dstCol++) {
          auto srcCol = dstCol * req.pixelStride();
          memcpy(dstBuff + dstCol * params.byteSize,
                 srcBuff.get() + srcCol * imgFullPixelSize +
                     band * params.byteSize,
                 params.byteSize);
        }
      }
      break;

    // For BIP to get here, we either need a stride or a spectral subset
    case Interleave::BIP: {
      auto *dstBuff = dst.row(0, r);
      if (!dstBuff)
        break; // For SLI curve selection
      if (params.useBandRemapping) {
        for (size_t dstCol = 0; dstCol < params.stridedNumCols; dstCol++) {
          auto srcCol = dstCol * req.pixelStride();
          for (size_t dstBand = 0; dstBand < params.numBands; dstBand++) {
            memcpy(dstBuff + dstCol * dstBipPixSize + dstBand * params.byteSize,
                   srcBuff.get() + srcCol * imgFullPixelSize +
                       params.bandsToLoad[dstBand] * params.byteSize,
                   params.byteSize);
          }
        }
      } else {
        for (size_t dstCol = 0; dstCol < params.stridedNumCols; dstCol++) {
          auto srcCol = dstCol * req.pixelStride();
          memcpy(dstBuff + dstCol * imgFullPixelSize,
                 srcBuff.get() + srcCol * imgFullPixelSize, imgFullPixelSize);
        }
      }
      break;
    }
    }

    if ((r + 1) % params.progressRowInterval == 0) {
      req.postProgress((r + 1.) / params.stridedNumRows);
    }
  }
}
} // namespace d5::envi
