/** Common utilities for image reading
 *
 *  \file Common.h
 */

#pragma once

#include <cstddef>
#include <cstdint>
#include <limits>

#include "dirsig5/envi/Header.h"
#include "dirsig5/envi/ImageReadRequest.h"

namespace d5::envi {

struct ReadParameters {
  /// Start column for the read
  std::size_t startCol;

  /// Start row for the read
  std::size_t startRow;

  /// Number of columns to read (strided)
  std::size_t stridedNumCols;

  /// Number of rows to read (strided)
  std::size_t stridedNumRows;

  /// Number of rows to read (non-strided)
  std::size_t nonStridedNumCols;

  /// Number of bytes per image value
  std::uint8_t byteSize;

  /// Whether to use the band remapping in the request
  bool useBandRemapping;

  /// Number of bands to read
  std::size_t numBands;

  /// List of bands to load (optional)
  std::vector<size_t> bandsToLoad;

  /// Interval at which to post progress updates (rows)
  std::size_t progressRowInterval;
};

/** Determine some derived parameters from the image header and the read request
 *
 *  \param hdr  The image header
 *  \param req  The user's read request
 *
 *  \return A structure of the derived parameters
 */
inline ReadParameters DetermineReadParameters(const Header &hdr,
                                              const ImageReadRequest &req) {
  std::size_t startCol = req.hasRoi() ? req.roi()[0] : 0;
  std::size_t startRow = req.hasRoi() ? req.roi()[1] : 0;
  std::size_t numCols = req.hasRoi() ? req.roi()[2] : hdr.samples;
  std::size_t numRows = req.hasRoi() ? req.roi()[3] : hdr.lines;
  if (numCols == std::numeric_limits<std::size_t>::max()) {
    numCols = hdr.samples - startCol;
  }
  if (numRows == std::numeric_limits<std::size_t>::max()) {
    numRows = hdr.lines - startRow;
  }

  // Check for an invalid ROI
  if (startCol + numCols > hdr.samples) {
    throw std::runtime_error(
        "invalid ROI, requested column index " + std::to_string(startCol) +
        " - " + std::to_string(startCol + numCols - 1) +
        ", but image only has " + std::to_string(hdr.samples) + " columns");
  }
  if (startRow + numRows > hdr.lines) {
    throw std::runtime_error(
        "invalid ROI, requested row index " + std::to_string(startRow) + " - " +
        std::to_string(startRow + numRows - 1) + ", but image only has " +
        std::to_string(hdr.lines) + " rows");
  }

  bool useBandRemapping = req.hasBandsToLoad();
  auto bandsToLoad = req.bandsToLoad();

  // If there were no bands requested (req.bandsToLoad) but the header has
  // the "default_bands" set and the request says it's OK to use those bands
  // (req.loadDefaultBands), then push those default bands into bandsToLoad.
  if (bandsToLoad.empty() && req.loadDefaultBands() &&
      !hdr.defaultBands.empty()) {
    bandsToLoad.insert(bandsToLoad.begin(), hdr.defaultBands.begin(),
                       hdr.defaultBands.end());
    useBandRemapping = true;
  }

  // In the case of the user passing all bands in order ({0, 1, 2, ...}),
  // ignore the band remapping
  if (useBandRemapping && bandsToLoad.size() == hdr.bands) {
    useBandRemapping = false;
    for (size_t i = 0; i < hdr.bands; i++) {
      if (bandsToLoad[i] != i) {
        useBandRemapping = true;
        break;
      }
    }
  }
  if (useBandRemapping) {
    for (auto b : bandsToLoad) {
      if (b >= hdr.bands) {
        throw std::runtime_error("invalid bands, requested band index " +
                                 std::to_string(b) + " but image only has " +
                                 std::to_string(hdr.bands) + " bands");
      }
    }
  }

  size_t numBands = (bandsToLoad.size() > 0) ? bandsToLoad.size() : hdr.bands;
  size_t stridedNumCols =
      (numCols == 0) ? 0 : (numCols - 1) / req.pixelStride() + 1;
  size_t stridedNumRows =
      (numRows == 0) ? 0 : (numRows - 1) / req.pixelStride() + 1;
  return ReadParameters{
      startCol, startRow, stridedNumCols, stridedNumRows, numCols,
      ByteSize(hdr.dataType), useBandRemapping, numBands, bandsToLoad,
      /* progressRowInterval = */
      std::max(std::size_t{1},
               static_cast<std::size_t>(std::round(stridedNumRows / 100.)))};
}

/** Allocate the destination image to the appropriate size
 *
 *  \param[in,out] dst  The image destination
 *  \param[in] params   Read parameters from \c DetermineReadParameters
 *  \param[in] hdr      The image header
 */
template <typename Destination>
void Allocate(Destination &&dst, const ReadParameters &params,
              const Header &hdr) {
  dst.allocate(params.numBands, params.stridedNumRows, params.stridedNumCols,
               hdr.dataType);
}
} // namespace d5::envi
