#pragma once

#include <memory>
#include <string>
#include <vector>

#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/State.h"
#include "dirsig5/flex_motion/Export.h"
#include "dirsig5/geolib/Facade.h"
#include "dirsig5/geolib/Location.h"

namespace d5::flex_motion {

struct LocationState {
  Point location = {0, 0, 0};

  Vector velocity = {0, 0, 0};
};

/// This is the base class for all the location engines.
/// The location engines primary job is to compute the location
/// and velocity (both in Scene ENU) for a given time.
class D5_FM_EXPORT LocationEngine {
public:
  LocationEngine() = default;

  virtual ~LocationEngine() = default;

  [[nodiscard]] virtual bool isStatic() const { return false; }

  void updateState(const StateDateTime &state, const geolib::Facade *facade);

  [[nodiscard]] LocationState get(const DateTime &dateTime) const;

public:
  DateTime startDateTime = DateTime::fromRelativeSeconds(0);

  /// The coordinate frame (default is Scene ENU).
  geolib::Location::Type frame = geolib::Location::EnuType;

protected:
  struct Cache {
    DateTime referenceDateTime = {};
  } cache = {};

  virtual void doUpdateState(const StateDateTime &, const geolib::Facade *) {}

  [[nodiscard]] virtual LocationState doGet(const DateTime &absoluteTime,
                                            double relativeTime) const = 0;
};

inline void LocationEngine::updateState(const StateDateTime &state,
                                        const geolib::Facade *facade) {
  cache.referenceDateTime = *state.referenceDateTime;
  doUpdateState(state, facade);
}

inline LocationState LocationEngine::get(const DateTime &dateTime) const {
  DateTime absoluteTime = cache.referenceDateTime.makeAbsolute(dateTime);
  double relativeTime =
      (absoluteTime - cache.referenceDateTime.makeAbsolute(startDateTime))
          .getTotalSeconds();
  return doGet(absoluteTime, relativeTime);
}

} // namespace d5::flex_motion
