#pragma once

#include <memory>

#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/DualQuaternion.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/State.h"
#include "dirsig5/flex_motion/Export.h"
#include "dirsig5/flex_motion/JitterModel.h"
#include "dirsig5/geolib/Facade.h"

namespace d5::flex_motion {

/// The parent class for algorithmic and data driven motion. Derived classes
/// are used to model platform motion (vehicle motion) and sub-platform
/// motion (vehicle-relative motion, or mounts).
class D5_FM_EXPORT Motion {
public:
  Motion() = default;

  virtual ~Motion() = default;

  void updateState(const StateDateTime &, const geolib::Facade *);

  /// Get the motion state for a given relative time.
  DualQuaternion getRelative(double relativeTime) const;

  /// Get the motion state for a given date/time.
  DualQuaternion get(const DateTime &dateTime) const;

  struct Cache {
    DateTime referenceDateTime = {};
    double elapsedTime = 0;
  } cache = {};

  std::shared_ptr<JitterModel> locationJitter[3] = {};

  std::shared_ptr<JitterModel> rotationJitter[3] = {};

  DateTime jitterStartTime = DateTime::fromRelativeSeconds(0);

  /// An enum to store the coordinate frame (ENU vs. ECEF) for the motion
  enum CoordinateFrame {
    UnknownFrame,
    EnuFrame,
    EcefFrame
  } frame = UnknownFrame;

protected:
  virtual void doUpdateState(const StateDateTime &, const geolib::Facade *) {}

  virtual DualQuaternion doGet(const DateTime &) const = 0;
};

inline void Motion::updateState(const StateDateTime &state,
                                const geolib::Facade *facade) {
  cache.referenceDateTime = *state.referenceDateTime;
  cache.elapsedTime = state.elapsedTime;
  doUpdateState(state, facade);
}

} // namespace d5::flex_motion
