#pragma once

#include "dirsig5/flex_motion/LocationEngine.h"

namespace d5::flex_motion {

/// This is the base class for all the orientation engines.
/// The orientation engines primary job is to compute the orientation
/// for a given time.  This current orientation is described by an
/// affine transform stored in a matrix object.  Note that the
/// current design allows orientation calculations to be driven by
/// the current.
class D5_FM_EXPORT RotationEngine {
public:
  /// This enumerated type is used to define constraint vector options
  /// when we need a second vector to fully define the orientation.
  enum ConstraintOption {
    UnknownConstraint = 0x00, ///< Unknown
    UseVelocity = 0x01,       ///< Use the current velocity vector
    UseConstant = 0x02,       ///< Use a constant vector
    VectorFrameEnu = 0x04,    ///< Constant vector frame is Scene ENU
    VectorFrameEcef = 0x08    ///< Constant vector frame is ECEF
  };

  RotationEngine() = default;

  virtual ~RotationEngine() = default;

  [[nodiscard]] virtual bool isStatic() const { return false; }

  void updateState(const StateDateTime &state, const geolib::Facade *facade);

  [[nodiscard]] Quaternion get(const DateTime &dateTime) const;

public:
  /// The (optional) location engine.
  std::shared_ptr<LocationEngine> locationEngine = {};

  /// The (optional) reference date/time.
  DateTime startDateTime = DateTime::fromRelativeSeconds(0);

  /// The coordinate frame (default is Scene ENU).
  geolib::Location::Type frame = geolib::Location::EnuType;

  /// The constraint option (default/fallback to constant ENU Z+)
  unsigned int constraintFlags = UseConstant | VectorFrameEnu;

  /// The constraint vector.
  Vector constraintVector = {0, 0, 1};

protected:
  struct Cache {
    DateTime referenceDateTime = {};

    /// The transform from ECEF to Scene ENU.
    Matrix ecefToEnuTransform = Matrix::Identity();

    /// The transform from the body coordinate system to Scene ENU.
    Quaternion bodyToEnuTransform = {1, 0, 0, 0};

  } cache = {};

  virtual void doUpdateState(const StateDateTime &, const geolib::Facade *) {}

  [[nodiscard]] virtual Quaternion doGet(const DateTime &absoluteTime,
                                         double relativeTime) const = 0;
};

inline void RotationEngine::updateState(const StateDateTime &state,
                                        const geolib::Facade *facade) {
  DIRSIG_THROW_IF(
      constraintFlags == UnknownConstraint,
      "Flex::RotationEngine::updateState: Unknown constraint option!");
  DIRSIG_THROW_IF(facade == nullptr,
                  "Flex::RotationEngine::updateState: geolib facade is null!");
  cache.referenceDateTime = *state.referenceDateTime;
  cache.ecefToEnuTransform = geolib::Location::computeEcefToEnuMatrix(facade);
  cache.bodyToEnuTransform = Quaternion{1, 0, 0, 0};
  doUpdateState(state, facade);
}

inline Quaternion RotationEngine::get(const DateTime &dateTime) const {
  DateTime absoluteTime = cache.referenceDateTime.makeAbsolute(dateTime);
  double relativeTime =
      (absoluteTime - cache.referenceDateTime.makeAbsolute(startDateTime))
          .getTotalSeconds();
  return doGet(absoluteTime, relativeTime);
}

} // namespace d5::flex_motion
