#pragma once

#include <GeographicLib/Geocentric.hpp>
#include <GeographicLib/Geodesic.hpp>
#include <GeographicLib/LocalCartesian.hpp>
#include <GeographicLib/UTMUPS.hpp>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Geometry.h"

namespace d5::geolib {

static GeographicLib::Geocentric earth = GeographicLib::Geocentric::WGS84();

/** A facade (wrapper) around GeographicLib (developed by Charles Karney).
 *  \ingroup geolib
 */
class Facade {
public:
  /// The constructor takes the geodetic coordinate of the Scene/Local ENU
  /// coordinate system so that all geographic coordinates can be
  /// converted into and out of that coordinate system.
  Facade(double lat, double lon, double h) {
    DIRSIG_THROW_IF(!(lat >= -90 and lat <= +90), "Latitude must be +/- 90!");
    DIRSIG_THROW_IF(!(lon >= -180 and lon <= +180),
                    "Longitude must be +/- 180!");
    projection = new GeographicLib::LocalCartesian(lat, lon, h, earth);

    // create the ECEF to ENU rotation transform
    Eigen::Matrix3d matrix;
    double lam = lon * (M_PI / 180.0);
    double phi = lat * (M_PI / 180.0);
    double sinLam = std::sin(lam);
    double cosLam = std::cos(lam);
    double sinPhi = std::sin(phi);
    double cosPhi = std::cos(phi);
    matrix(0, 0) = -sinLam;
    matrix(1, 0) = cosLam;
    matrix(2, 0) = 0.0;
    matrix(0, 1) = -sinPhi * cosLam;
    matrix(1, 1) = -sinPhi * sinLam;
    matrix(2, 1) = cosPhi;
    matrix(0, 2) = cosPhi * cosLam;
    matrix(1, 2) = cosPhi * sinLam;
    matrix(2, 2) = sinPhi;
    ecefToEnu = new Quaternion(matrix.transpose());
  }

  Facade(const Facade &) = delete;

  Facade(const GeographicLib::LocalCartesian *proj) {
    DIRSIG_THROW_IF(proj == nullptr, "Projection must be non-null!");
    projection = new GeographicLib::LocalCartesian(proj->LatitudeOrigin(),
                                                   proj->LongitudeOrigin(),
                                                   proj->HeightOrigin(), earth);
  }

  ~Facade() {
    delete projection;
    delete ecefToEnu;
  }

  // get the ECEF to ENU rotation
  [[nodiscard]] Quaternion getEcefToEnuRotation() const { return *ecefToEnu; }

  // convert a geodetic coordinate to a Scene ENU point
  [[nodiscard]] Point from(const GeodeticCoord &coord) const {
    Point point = {0, 0, 0};
    projection->Forward(coord.latitude, coord.longitude, coord.height, //
                        point[0], point[1], point[2]);
    return point;
  }

  // convert a ECEF coordinate to a Scene ENU point
  [[nodiscard]] Point from(const EcefCoord &coord) const {
    GeodeticCoord geo = {};
    earth.Reverse(coord.x, coord.y, coord.z, //
                  geo.latitude, geo.longitude, geo.height);
    return from(geo);
  }

  // convert a UTM coordinate to a Scene ENU point
  [[nodiscard]] Point from(const UtmCoord &coord) const {
    GeodeticCoord geo = {};
    geo.height = coord.height;
    GeographicLib::UTMUPS::Reverse(int(coord.zone), coord.hemisphere == 'N',
                                   coord.easting, coord.northing, geo.latitude,
                                   geo.longitude);
    return from(geo);
  }

  // convert an ECEF quaternion to a Scene ENU quaternion
  [[nodiscard]] Quaternion fromEcef(const Quaternion &rot) const {
    return (*ecefToEnu) * rot;
  }

  // convert a Scene ENU point to a UTM coord
  [[nodiscard]] UtmCoord toUtm(const Point &point) const {
    GeodeticCoord geo = toGeodetic(point);
    UtmCoord utm = {};
    utm.height = geo.height;
    bool northp = false;
    int zone = 0;
    GeographicLib::UTMUPS::Forward(geo.latitude, geo.longitude, zone, northp,
                                   utm.easting, utm.northing);
    if (northp)
      utm.hemisphere = 'N';
    else
      utm.hemisphere = 'S';
    utm.zone = zone;
    return utm;
  }

  // convert a Scene ENU point to ECEF coord
  [[nodiscard]] EcefCoord toEcef(const Point &point) const {
    GeodeticCoord geo = toGeodetic(point);
    EcefCoord ecef = {};
    earth.Forward(geo.latitude, geo.longitude, geo.height, ecef.x, ecef.y,
                  ecef.z);
    return ecef;
  }

  // convert a Scene ENU point to a geodetic coord
  [[nodiscard]] GeodeticCoord toGeodetic(const Point &point) const {
    GeodeticCoord geo = {};
    projection->Reverse(point[0], point[1], point[2], //
                        geo.latitude, geo.longitude, geo.height);
    return geo;
  }

  // convert a Scene ENU quaternion to an ECEF quaternion
  [[nodiscard]] Quaternion toEcef(const Quaternion &rot) const {
    return (*ecefToEnu).inverse() * rot;
  }

  // convert a Scene ENU matrix to an ECEF matrix
  [[nodiscard]] Matrix toEcef(const Matrix &mat) const {
    return (*ecefToEnu).inverse() * mat;
  }

  GeographicLib::LocalCartesian *projection = nullptr;

  Quaternion *ecefToEnu = nullptr;
};
} // namespace d5::geolib
