/** Interface file for Location
 *
 *  \file Location.h
 */

#pragma once

#include "dirsig5/geolib/Facade.h"

namespace d5::geolib {

/** A flexible object to store a location, including both Scene ENU
 *  locations and a variety of geographic location types including UTM,
 *  geodetic, and ECEF.  This class leverages the GeoLibFacade, which
 *  provides a streamlined facade to GeographicLib.
 *  \ingroup geolib
 */
class Location {
public:
  enum Type {
    UnknownType,  ///< No coordinate supplied
    UtmType,      ///< Coordinate supplied as UTM
    GeodeticType, ///< Coordinate supplied as lat,long,alt
    EcefType,     ///< Coordinate supplied as ECEF
    EnuType       ///< Coordinate supplied as local ENU
  };

  Location(const Facade *f) : facade(f) {}

  /// Templated constructor (on coordinate type)
  template <typename CoordType>
  Location(const CoordType &coord, const Facade *f) : facade(f) {
    setLocation(coord);
  }

  operator const Point &() const { return localCartesian; }

  operator Point &() { return localCartesian; }

  /// Clear the current location.
  void clear() { localCartesian = {0, 0, 0}; }

  /// Set location with a UTM coordinate
  void setLocation(const UtmCoord &utmCoord) {
    localCartesian = facade->from(utmCoord);
  }

  /// Set location with a geodetic coordinate
  void setLocation(const GeodeticCoord &geodeticCoord) {
    localCartesian = facade->from(geodeticCoord);
  }

  /// Set location with an ECEF coordinate.
  void setLocation(const EcefCoord &ecefCoord) {
    localCartesian = facade->from(ecefCoord);
  }

  /// Set the location with a local ENU coordinate
  void setLocation(const Point &point) { localCartesian = point; }

  /// Get the location in UTM coordinates.
  [[nodiscard]] UtmCoord getUtmLocation() const {
    return facade->toUtm(localCartesian);
  }

  /// Get location in geodetic coordinates.
  [[nodiscard]] GeodeticCoord getGeodeticLocation() const {
    return facade->toGeodetic(localCartesian);
  }

  /// Get location in ECEF coordinates.
  [[nodiscard]] EcefCoord getEcefLocation() const {
    return facade->toEcef(localCartesian);
  }

  /// Get the local cartesian location.
  [[nodiscard]] const Point &getLocalCartesian() const {
    return localCartesian;
  }

  /// Matrix to convert from UTM to ENU (user reference frame)
  [[nodiscard]] static Matrix computeUtmToEnuMatrix(const Location &frame) {
    auto enuFrame = static_cast<Point>(frame);

    // compute some points in UTM space to make basis vectors
    Point originUtm =
        Location(enuFrame + Point(0, 0, 0), frame.facade).getUtmLocation();
    Point xUtm =
        Location(enuFrame + Point(100, 0, 0), frame.facade).getUtmLocation();
    Point yUtm =
        Location(enuFrame + Point(0, 100, 0), frame.facade).getUtmLocation();
    Point zUtm =
        Location(enuFrame + Point(0, 0, 100), frame.facade).getUtmLocation();

    // compute the basis vectors
    Vector xUtmVector(xUtm - originUtm);
    Vector yUtmVector(yUtm - originUtm);
    Vector zUtmVector(zUtm - originUtm);

    // normalize the basis vectors
    xUtmVector.normalize();
    yUtmVector.normalize();
    zUtmVector.normalize();

    // make a matrix from the basis vectors
    Matrix utmToEnu;
    for (int ii = 0; ii < 3; ii++) {
      utmToEnu(0, ii) = xUtmVector(ii);
      utmToEnu(1, ii) = yUtmVector(ii);
      utmToEnu(2, ii) = zUtmVector(ii);
    }

    return utmToEnu;
  }

  /// Matrix to convert from ECEF to ENU (user reference frame)
  static Matrix computeEcefToEnuMatrix(const Location &frame) {
    return Location::constructMatrix(frame.getGeodeticLocation(), false);
  }

  /// Matrix to convert from ENU to ECEF (user reference frame)
  static Matrix computeEnuToEcefMatrix(const Location &frame) {
    return Location::constructMatrix(frame.getGeodeticLocation(), true);
  }

private:
  /// Common code to construct ENU/ECEF matrices
  static Matrix constructMatrix(const GeodeticCoord &origin, bool EnuToEcef) {
    double lam = origin.longitude * (M_PI / 180.0);
    double phi = origin.latitude * (M_PI / 180.0);
    double sinLam = std::sin(lam), cosLam = std::cos(lam);
    double sinPhi = std::sin(phi), cosPhi = std::cos(phi);
    Matrix result = Matrix::Identity();
    result(0, 0) = -sinLam;
    result(1, 0) = cosLam;
    result(2, 0) = 0.0;
    result(0, 1) = -sinPhi * cosLam;
    result(1, 1) = -sinPhi * sinLam;
    result(2, 1) = cosPhi;
    result(0, 2) = cosPhi * cosLam;
    result(1, 2) = cosPhi * sinLam;
    result(2, 2) = sinPhi;
    if (not EnuToEcef) { // Need to transpose the rotation elements?
      double d10 = result(1, 0);
      result(1, 0) = result(0, 1);
      result(0, 1) = d10;
      double d20 = result(2, 0);
      result(2, 0) = result(0, 2);
      result(0, 2) = d20;
      double d12 = result(1, 2);
      result(1, 2) = result(2, 1);
      result(2, 1) = d12;
    }
    return result;
  }

  /// The location in local cartesian (native DIRSIG) coords
  Point localCartesian = {0, 0, 0};

  /// Optional explicit geo lib facade
  const Facade *facade = nullptr;
};
} // namespace d5::geolib
