/** Interface file for the OutputFile class
 *
 *  \file OutputFile.h
 */

#pragma once

#include <memory>
#include <string>

#include "OutputFileDriver.h"
#include "dirsig5/io/Export.h"

namespace d5::io {

/** Generic interface to an output file, modeled after ofstream.
 *
 *  In reality, no I/O happens in this class, this class simply passed through
 *  operations to a driver, which implements the specifics of the storage.
 *  However, the user should treat this class similar to an ofstream used in
 *  binary mode (with write()).
 */
class D5_IO_EXPORT OutputFile {
public:
  /// The filename of the current file (if any)
  [[nodiscard]] const std::string &filename() const { return filename_; }

  /// Returns whether there is a currently open file
  [[nodiscard]] bool is_open() const { return callbacks_.is_open(); }

  /** Opens a file with the given filanem in the given mode
   *
   *  \param filename  The filename to open
   *  \param mode      The mode in which to open (only accepts output modes)
   */
  void open(const std::string &filename,
            std::ios_base::openmode mode = std::ios_base::out |
                                           std::ios_base::binary) {
    filename_ = filename;
    callbacks_.open(filename, mode);
  }

  /// Closes the currently open file
  void close() {
    callbacks_.close();
    filename_.clear();
  }

  /** Writes binary data to the current position in the file
   *
   *  \tparam T     The type of the data being written
   *  \param data   The data to write
   *  \param count  The number of elements in \c data
   *
   *  \return Whether the write was successful
   */
  template <typename T> bool write(const T *data, std::size_t count) {
    return callbacks_.write2(reinterpret_cast<const char *>(data),
                             count * sizeof(T));
  }

  /** Writes binary data to an arbitrary position in the file
   *
   *  \tparam T      The type of the data being written
   *  \param data    The data to write
   *  \param count   The number of elements in \c data
   *  \param offset  The position in the file to which to write
   *
   *  \return Whether the write was successful
   */
  template <typename T>
  bool write(const T *data, std::size_t count, std::size_t offset) {
    return callbacks_.write3(reinterpret_cast<const char *>(data),
                             count * sizeof(T), offset);
  }

  /// Returns the current position in the file
  [[nodiscard]] std::size_t tellp() const { return callbacks_.tellp(); }

  /** Sets the current position in the file
   *
   *  \param pos  The new position in the file
   *
   *  \return Whether the operation was successful
   */
  bool seekp(std::size_t pos) { return callbacks_.seekp1(pos); }

  /** Sets the the current position in the file
   *
   *  \param off  Relative position (positive or negative) to set the output
   *              position indicator to.
   *  \param dir  Defines base position to apply the relative offset to.
   */
  bool seekp(std::int64_t off, std::ios_base::seekdir dir) {
    return callbacks_.seekp2(off, dir);
  }

  /// Flushes any buffered data to the file
  void flush() { callbacks_.flush(); }

  /// Returns a description of the last failed operation
  [[nodiscard]] std::string lastError() const { return callbacks_.lastError(); }

  /** Self-contained write operation for a whole file
   *
   *  \param filename  Filename to which to write
   *  \param buffer    Contents of the file to write
   *  \param size      Size of \c buffer in bytes
   *
   *  \return Whether the write was successful
   */
  bool writeFile(const std::string &filename, const char *buffer,
                 std::size_t size) {
    return callbacks_.writeFile(filename, buffer, size);
  }

  /** Sets the preferred driver for OutputFile
   *
   *  The default if this function is not called is to use the local filesystem
   *  driver ("local_output_file")
   *
   *  \param name  The name of the driver
   */
  static void SetPreferredDriver(const std::string &name);

  /// Factory function to use the local filesystem driver
  static OutputFile MakeLocal();

  /** Factory function with a specific driver
   *
   *  \param name  The name of the driver
   */
  static OutputFile MakeWithDriver(const std::string &name);

  /// Factory function that uses the preferred driver
  static OutputFile MakeWithPreferredDriver();

private:
  /// Constructor, takes a driver
  OutputFile(std::unique_ptr<OutputFileDriver> plugin);

  /// Filename of the currently open file
  std::string filename_;

  /// Callbacks retrieved from the driver
  OutputFileDriver::Callbacks callbacks_;

  /// The driver
  std::unique_ptr<OutputFileDriver> plugin_;

  /// Holds the preferred driver name
  static std::string s_preferredDriver;
};
} // namespace d5::io
