/** Interface file for the OutputFileDriver interface
 *
 *  \file OutputFileDriver.h
 */

#pragma once

#include <cstdint>
#include <functional>
#include <ios>
#include <string>

namespace d5::io {

/** Driver interface for an output file. Intended to be subclassed in plugin
 *  libraries. The driver works by populating a lot of callback functions, which
 *  are largely inspired by the std::ofstream interface.
 */
class OutputFileDriver {
public:
  /// Container type for all of the callbacks needed to implement the I/O driver
  struct Callbacks {
    std::function<void(const std::string &filename,
                       std::ios_base::openmode mode)>
        open;

    std::function<bool()> is_open;

    std::function<void()> close;

    std::function<bool(const char *data, std::size_t size)> write2;

    std::function<bool(const char *data, std::size_t size, std::size_t offset)>
        write3;

    std::function<std::size_t()> tellp;

    std::function<bool(std::size_t pos)> seekp1;

    std::function<bool(std::int64_t off, std::ios_base::seekdir dir)> seekp2;

    std::function<bool()> flush;

    std::function<std::string()> lastError;

    std::function<bool(const std::string &filename, const char *buffer,
                       std::size_t size)>
        writeFile;

    operator bool() const {
      return open && is_open && close && write2 && write3 && tellp && seekp1 &&
             seekp2 && flush && lastError && writeFile;
    }
  };

  /// Virtual destructor
  virtual ~OutputFileDriver() = default;

  /// Get the callbacks for this driver
  virtual Callbacks getCallbacks() = 0;
};
} // namespace d5::io
