/** Interface file for a DIRSIG LIDAR bin file
 *
 *  \file BinOutputFile.h
 */

#pragma once

#include <cstdint>
#include <fstream>
#include <limits>
#include <vector>

#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/DualQuaternion.h"
#include "dirsig5/io/Export.h"
#include "dirsig5/io/OutputFile.h"

namespace d5::io {

/** An output file handler that produces a LIDAR waveform file.
 *  \ingroup io
 */
class D5_IO_EXPORT BinOutputFile {
public:
  /**
   * \struct FileHeader
   *
   * File header for LIDAR waveform BIN file.
   */
  struct D5_IO_EXPORT FileHeader {
    /** The "magic string" used to verify the file type ("DIRSIGPROTO") */
    char fileIdentifier[12];

    /** The version of file format specification. */
    char fileFormatRev;

    /** The byte order, 0 indicates big endian, 1 for little endian */
    char byteOrder;

    /** The date/time that file was created. Format: YYYYMMDDhhmm.ss */
    char creationDateTime[16];

    /** The DIRSIG version used to create file.  */
    char dirsigVersion[33];

    /** The arbitrary user-supplied text describing simulation. */
    char simulationDesc[257];

    /** The latitude of scene origin in degrees using +East [-180,180) */
    double sceneOriginLatitude = 0;

    /** The longitude of scene origin in degrees using +North [-90,90] */
    double sceneOriginLongitude = 0;

    /** The height of scene origin above WGS84 reference ellipsoid [m] */
    double sceneOriginHeight = 0;

    /** The description of mount type for laser */
    char laserMountType[17];

    /** The description of mount type for detector */
    char detectorMountType[17];

    /** The number of detector elements in the X dimension */
    std::uint32_t xDetectorCount = std::numeric_limits<std::uint32_t>::max();

    /** The number of detector elements in the Y dimension */
    std::uint32_t yDetectorCount = std::numeric_limits<std::uint32_t>::max();

    /** The distance between detector centers [microns] in X dimension */
    double xDetectorPitch = -1;

    /** The distance between detector centers [microns] in Y dimension */
    double yDetectorPitch = -1;

    /** The distance between array center and optical axis [microns] */
    double xArrayOffset = 0;

    /** The distance between array center and optical axis [microns] */
    double yArrayOffset = 0;

    /** The first radial lens distortion coefficient k1 */
    double lensK1 = 0;

    /** The second radial lens distortion coefficient k2 */
    double lensK2 = 0;

    /** The number of tasks in file */
    std::uint32_t taskCount = 0;

    /** The ID for the focal plane array */
    std::uint16_t fpaId = 0;

    /**
     * Constructor
     */
    FileHeader();

    /**
     * Setup the header with supplied data.
     */
    void setup(const d5::GeodeticCoord &_sceneOrigin,
               std::uint32_t _xPixelCount, std::uint32_t _yPixelCount,
               double _xPixelPitch, double _yPixelPitch, double _xArrayOffset,
               double _yArrayOffset, std::uint32_t taskCount);

    /**
     * Print the header as ASCII/Text to a stream.
     */
    void print(std::ostream &_stream) const;
  };

  /**
   * \struct TaskHeader
   *
   * Task header for LIDAR waveform BIN file.
   */
  struct D5_IO_EXPORT TaskHeader {
    /** The task index (starts at 0) */
    std::uint32_t taskIndex = 0;

    /** The ASCII/Text description of the task */
    char description[65];

    /** The date/time for task start. Format: YYYYMMDDhhmm.ss */
    char startDateTime[16];

    /** The date/time for task stop. Format: YYYYMMDDhhmm.ss */
    char stopDateTime[16];

    /** The instrument focal length in [mm] */
    double focalLength = 0;

    /** The nominal laser pulse repetition frequency [Hz] */
    double pulseRepetitionFreq = 0;

    /** The nominal laser pulse duration [sec] */
    double pulseDuration = 0;

    /** The nominal laser pulse energy [J] */
    double pulseEnergy = 0;

    /** The laser spectral center[microns] */
    double laserSpectralCenter = 0;

    /** The laser spectral width [microns] */
    double laserSpectralWidth = 0;

    /** The number of pulses in this task */
    std::uint32_t pulseCount = 0;

    /**
     * Constructor
     */
    TaskHeader();

    /**
     * Setup the header with supplied receiver data.
     */
    void setupRx(std::uint32_t _taskIndex, std::uint32_t _pulseCount,
                 double _focalLength, const d5::DateTime &_taskStart,
                 const d5::DateTime &_taskEnd);

    /**
     * Setup the header with the supplied transmitter data.
     */
    void setupTx(double _pulseRate, double _pulseEnergy, double _pulseDuration,
                 double _laserCenter, double _laserWidth);

    /**
     * Print the header as ASCII/Text to a stream.
     */
    void print(std::ostream &stream) const;
  };

  /**
   * \struct PulseHeader
   *
   * The header for a pulse in the LIDAR waveform BIN file.
   */
  struct D5_IO_EXPORT PulseHeader {
    /** The pulse time relative to task start [sec] */
    double pulseTime = -1;

    /** The time gating start time relative to task start [sec] */
    double timeGateStart = -1;

    /** The time gating stop time relative to task start [sec] */
    double timeGateStop = -1;

    /** The number of temporal samples (equally spaced) over time gating */
    std::uint32_t timeGateBinCount = 0;

    /** The number of samples (equally spaced) within a time bin */
    std::uint32_t samplesPerTimeBin = 1;

    /** The platform location (x,y,z) in scene local coordinates [m] */
    double platformLocation[3];

    /** The platform Euler angles (x,y,z) applied in "yzx" order [rad] */
    double platformRotation[3];

    /** The Tx to mount transform (4x4 affine matrix, row major) */
    double transmitterToMount[16];

    /** The Tx pointing Euler angles (x,y,z) applied in "yzx" order [rad]*/
    double transmitterMountPointing[3];

    /** The Tx mount to platform transform (4x4 matrix, row major) */
    double transmitterMountToPlatform[16];

    /** The Rx to mount transform (4x4 affine matrix, row major) */
    double receiverToMount[16];

    /** The Rx pointing Euler angles (x,y,z) applied in "yzx" order [rad] */
    double receiverMountPointing[3];

    /** The Rx mount to platform transform (4x4 matrix, row major) */
    double receiverMountToPlatform[16];

    /**
     * Data type used in pulse results with only type 5 (double precision)
     * being currently supported. Corresponds to the ENVI type numbers.
     */
    int pulseDataType = 5;

    /**
     * The compression mode.
     * 0 indicates no compression,
     * 1 indicates zlib compression of pulse data
     */
    char dataCompression = 0;

    /** The pulse index (starts at 0) within a task */
    std::uint32_t pulseIndex = 0;

    /** The number of bytes of (possibly compressed) pulse data */
    std::uint64_t pulseDataBytes = 0;

    /** The system polarization, transmit (optics, scan mirrors, etc...) */
    double systemPolTransmit[16];

    /** The system polarization, receive (optics, scan mirrors, etc...) */
    double systemPolReceive[16];

    /**
     * Constructor.
     */
    PulseHeader();

    /**
     * Setup the header with supplied data.
     */
    void setup(std::uint32_t _pulseIndex, double _pulseTime,
               double _gateStartTime, double _gateStopTime,
               std::uint32_t _gateTimeCount,
               const d5::DualQuaternion &_platformState,
               const d5::DualQuaternion &_platformAttach,
               const d5::DualQuaternion &_mountState,
               const d5::DualQuaternion &_mountAttach,
               unsigned long _pulseDataBytes);

    /**
     * Print the header as ASCII/Text to a stream.
     */
    void print(std::ostream &stream) const;
  };

  /**
   * The constructor.
   */
  BinOutputFile() = default;

  /**
   * The destructor.
   */
  ~BinOutputFile() {
    // close the file
    this->close();
  }

  /**
   * Open the output file.
   */
  bool open();

  /**
   * Set the filename.
   */
  void setFilename(const std::string &_filename) { this->filename = _filename; }

  /**
   * Get the filename.
   */
  [[nodiscard]] const std::string &getFilename() const {
    return this->filename;
  }

  /**
   * Get the file header.
   */
  FileHeader &getFileHeader() { return this->fileHeader; }

  /**
   * Get the task header.
   */
  TaskHeader &getTaskHeader() { return this->taskHeader; }

  /**
   * Get the pulse header.
   */
  PulseHeader &getPulseHeader() { return this->pulseHeader; }

  /**
   * Resize the pulse data buffer.
   */
  void resizePulseBuffer(size_t timeBinCount);

  /**
   * Get the pulse data buffer.
   */
  std::vector<double> &getPulseBuffer() {
    return this->uncompressedPulseBuffer;
  }

  /**
   * Write a FileHeader to the output file.
   */
  bool writeFileHeader();

  /**
   * Write a TaskHeader to the output file.
   */
  bool writeTaskHeader();

  /**
   * Write a PulseHeader to the output file.
   */
  bool writePulseHeader();

  /**
   * Write the uncompressed pulse data to the output file.
   */
  bool writePulseBuffer();

  /**
   * Write the compressed pulse data to the output file.
   */
  bool writePulseBuffer(const void *buffer, size_t byteCount);

  /**
   * Close the output file.
   */
  bool close();

private:
  /** The filename */
  std::string filename;

  /**
   * The uncompressed pulse cube.  It is buffered until the cube is
   * complete and it can be compressed for output.
   */
  std::vector<double> uncompressedPulseBuffer;

  /** The buffer for any generated error messages.  */
  std::string errorMsg;

  /** The current file header */
  FileHeader fileHeader;

  /** The current task header */
  TaskHeader taskHeader;

  /** The current pulse header */
  PulseHeader pulseHeader;

  /** The position of the pulse count in the current task header */
  size_t pulseCountPosition = 0;

  /** The underlying output file object (we only support local I/O) */
  d5::io::OutputFile file = d5::io::OutputFile::MakeLocal();
};
} // namespace d5::io
