/** Interface file for EnviOutputFile
 *
 *  \file EnviOutputFile.h
 */

#pragma once

#include <cassert>

#include "dirsig5/io/Export.h"
#include "dirsig5/io/OutputFile.h"
#include "dirsig5/io/image/ImageOutputFile.h"

namespace d5::io {

/**
 * \class EnviOutputFile
 * \ingroup PluginToolkitLib
 *
 * An output file handler that produces an ENVI image and header file pair.
 */
class EnviOutputFile : public ImageOutputFile {
public:
  /**
   * The constructor.
   */
  EnviOutputFile() = default;

  /**
   * The destructor.
   * Free the internal scratch buffer if one was allocated.
   */
  ~EnviOutputFile() override {
    // close the file
    this->close();

    // if we have a scratch buffer, free it up

    delete[] this->scratchBuffer;
  }

  /**
   * Open the output file.
   */
  bool open() override;

  /**
   * Get the current position in the output file.
   */
  [[nodiscard]] size_t getPosition() const override {
    assert(this->dataSize != 0);
    return this->imgFile.tellp() / this->dataSize;
  }

  /**
   * Write the buffer to the output file at the specified offset.
   */
  bool writeBuffer(size_t _pixelOffset) override;

  /**
   * This function is called to get the last error.
   */
  std::string lastError() const override { return this->errorString; }

  /**
   * Close the output file.
   */
  bool close() override;

private:
  /**
   * Write the header file.
   */
  bool writeHeaderFile();

  /**
   * Set the name of the output file.
   */
  void doSetFilename() override {
    this->hdrFilename = this->getFilename() + ".hdr";
  }

  /** The filename for the header data file */
  std::string hdrFilename;

  /** The binary file */
  d5::io::OutputFile imgFile = d5::io::OutputFile::MakeWithPreferredDriver();

  /** The current output data type */
  DataType dataType = UnknownType;

  /** The current output data type size (per pixel, per band) */
  size_t dataSize = 0;

  /** The size of the scratch output buffer if we are not writing doubles */
  size_t scratchBufferSize = 0;

  /** The scratch output buffer if we are not writing doubles */
  char *scratchBuffer = nullptr;

  /** Holds the previous offset written (for incremental header writes) */
  std::size_t offsetCount_ = 0;

  /** Holds the error string */
  std::string errorString = "";
};
} // namespace d5::io
