/**
 * \file ImageOutputFile.h
 *
 * Copyright (c) 2011 Rochester Institute of Technology.
 * All rights reserved.
 */

#pragma once

#include <array>
#include <cstdint>
#include <iostream>
#include <memory>
#include <string>
#include <utility>
#include <vector>

#include "dirsig5/base/Macros.h"
#include "dirsig5/base/NamedParameterStore.h"
#include "dirsig5/io/Export.h"

namespace d5::io {

#define D5_IO_IMAGE_OUTPUT_FILE_TAGS(X)                                        \
  X(description, std::string)                                                  \
  X(samples, std::uint64_t)                                                    \
  X(lines, std::uint64_t)                                                      \
  X(bands, std::uint64_t)                                                      \
  X(header_offset, std::uint64_t)                                              \
  X(x_start, std::uint64_t)                                                    \
  X(y_start, std::uint64_t)                                                    \
  X(pixel_size, D5_SINGLE_ARG(std::array<double, 2>))                          \
  X(file_type, std::string)                                                    \
  X(data_type, d5::io::ImageOutputFile::DataType)                              \
  X(data_units, std::string)                                                   \
  X(interleave, d5::io::ImageOutputFile::InterleaveMethod)                     \
  X(sensor_type, std::string)                                                  \
  X(byte_order, int)                                                           \
  X(map_info, std::vector<std::string>)                                        \
  X(z_plot_range, D5_SINGLE_ARG(std::array<double, 2>))                        \
  X(z_plot_titles, std::vector<std::string>)                                   \
  X(z_plot_average, D5_SINGLE_ARG(std::array<double, 2>))                      \
  X(default_bands, std::vector<std::uint64_t>)                                 \
  X(default_stretch, std::string)                                              \
  X(band_names, std::vector<std::string>)                                      \
  X(wavelength_units, std::string)                                             \
  X(wavelength, std::vector<double>)                                           \
  X(fwhm, std::vector<double>)                                                 \
  X(classes, std::uint64_t)                                                    \
  X(class_lookup, std::vector<int>)                                            \
  X(class_names, std::vector<std::string>)                                     \
  X(spectra_names, std::vector<std::string>)                                   \
  X(coordinate_system_string, std::string)                                     \
  X(geo_points, std::string)                                                   \
  X(rpc_info, std::string)                                                     \
  X(reflectance_scale_factor, double)                                          \
  X(projection_info, std::string)                                              \
  X(dem_band, std::uint64_t)                                                   \
  X(dem_file, std::string)                                                     \
  X(bbl, std::vector<double>)                                                  \
  X(data_gain_values, std::vector<double>)                                     \
  X(data_ignore_value, std::vector<double>)                                    \
  X(data_offset_values, std::vector<double>)                                   \
  X(complex_function, std::string)                                             \
  X(acquisition_time, std::string)                                             \
  X(integration_time, double)

/** An interface for describing a generic output file created by DIRSIG
 *  (primarily the sensor plugins).
 *
 *  \ingroup io
 */
class D5_IO_EXPORT ImageOutputFile {
public:
#define X_ENUM_DECL(name, type) name,
  /**
   * The list of tags that can be used to define parameters describing
   * the output file.
   */
  enum Tags { D5_IO_IMAGE_OUTPUT_FILE_TAGS(X_ENUM_DECL) };
#undef X_ENUM_DECL

  /**
   * An enumerated type for the image data type
   */
  enum DataType {
    UnknownType = 0,
    ByteType = 1,
    SignedIntegerType = 2,
    SignedLongType = 3,
    SinglePrecisionType = 4,
    DoublePrecisionType = 5,
    SingleComplexType = 6,
    invalid1,
    invalid2,
    DoubleComplexType = 9,
    invalid3,
    invalid4,
    UnsignedIntegerType = 12,
    UnsignedLongType = 13,
    SignedLongLongType = 14,
    invalid5,
    UnsignedLongLongType = 16
  };

  /**
   * An enumerated type for the image band interleave method.
   */
  enum class InterleaveMethod {
    Unknown,
    BSQ, /**< Band-sequential interleave */
    BIL, /**< Band-interleaved-by-line   */
    BIP  /**< Band-interleaved-by-pixel  */
  };

  /// Typedef for our metadata type
  using Metadata = d5::NamedParameterStore<Tags>;

  /**
   * The default constructor.
   */
  ImageOutputFile() = default;

  /**
   * The destructor.
   */
  virtual ~ImageOutputFile() = default;

  /**
   * Set a value by the enumerated tag.
   *
   * \tparam tag   The tag of the value being set
   * \tparam T     The type of the value being set
   * \param value  The value being set
   */
  template <Tags tag, class T> void setValue(T &&value) {
    metadata_.set<tag>(std::forward<T>(value));
  }

  /**
   * Get a value by the enumerated tag.  This function returns true if the
   * tagName is found, false otherwise.
   */
  template <Tags tag> bool getValue(Metadata::type_for<tag> &tagValue) const {
    if (metadata_.contains<tag>()) {
      tagValue = metadata_.at<tag>();
      return true;
    }
    return false;
  }

  template <Tags tag> auto getValue() { return metadata_.at<tag>(); }

  template <typename Visitor> void visitMetadata(Visitor &&visitor) const;

  /**
   * Get the output filename
   */
  [[nodiscard]] const std::string &getFilename() const {
    return this->filename;
  }

  /**
   * Set the output filename.
   */
  void setFilename(std::string filename) {
    this->filename = std::move(filename);
    doSetFilename();
  }

  /**
   * Open the output file.
   */
  virtual bool open() = 0;

  /**
   * Get the current position in the output file.
   */
  [[nodiscard]] virtual size_t getPosition() const = 0;

  /**
   * Close the output file.
   */
  virtual bool close() = 0;

  /**
   * Create a buffer to store a fixed number of output pixels.  The
   * returned buffer is sized to hold all the bands for each pixel.
   */
  double *setupBuffer(std::size_t pixelCount);

  /**
   * An accessor to retrieve a previously setup buffer.  If the buffer
   * has not been setup, then the returned pointer is 0.
   */
  double *getBuffer() { return this->pixelBuffer.get(); }

  /**
   * This function is called to write the output to a specific offset in
   * the file.
   */
  virtual bool writeBuffer(size_t _fileOffset) = 0;

  /**
   * This function is called to get the last error.
   */
  virtual std::string lastError() const { return std::string(); }

  /**
   * A factory function that produces different derived class.
   */
  static ImageOutputFile *create(const std::string &name);

protected:
  /**
   * Get the size of the current data type in bytes.
   */
  [[nodiscard]] size_t getDataSize() const;

  [[nodiscard]] DataType getType() const {
    return ImageOutputFile::UnknownType;
  }

  [[nodiscard]] InterleaveMethod getInterleave() const {
    return ImageOutputFile::InterleaveMethod::Unknown;
  }

  /** File metadata */
  d5::NamedParameterStore<Tags> metadata_;

  /** The number of elements in the internal buffer */
  size_t pixelBufferSize = 0;

  /** The internal buffer */
  std::unique_ptr<double[]> pixelBuffer = nullptr;

private:
  /**
   * Virtual helper for setFilename(). Does nothing by default.
   */
  virtual void doSetFilename() {}

  /** The filename of the output */
  std::string filename;
};
} // namespace d5::io

#define X_TYPE_TRAIT(n, t)                                                     \
  template <>                                                                  \
  struct d5::NamedParameterStoreTraits<d5::io::ImageOutputFile::Tags,          \
                                       d5::io::ImageOutputFile::Tags::n> {     \
    using type = t;                                                            \
  };

D5_IO_IMAGE_OUTPUT_FILE_TAGS(X_TYPE_TRAIT)

#undef X_TYPE_TRAIT

namespace d5::io {

inline std::string to_string(ImageOutputFile::Tags tag) {
#define X_TO_STRING_CASE(n, t)                                                 \
  case ImageOutputFile::Tags::n:                                               \
    return #n;
  switch (tag) { D5_IO_IMAGE_OUTPUT_FILE_TAGS(X_TO_STRING_CASE); }
  return "unknown (" +
         std::to_string(
             static_cast<std::underlying_type_t<ImageOutputFile::Tags>>(tag)) +
         ")";
#undef X_TO_STRING_CASE
}

template <typename Visitor>
void ImageOutputFile::visitMetadata(Visitor &&visitor) const {
#define X_VISIT_CASE(n, t)                                                     \
  case Tags::n:                                                                \
    visitor(nameVal.first, metadata_.AnyCast<n>(nameVal.second));              \
    break;

  for (const auto &nameVal : metadata_) {
    switch (nameVal.first) { D5_IO_IMAGE_OUTPUT_FILE_TAGS(X_VISIT_CASE) }
  }

#undef X_VISIT_CASE
}

inline size_t ImageOutputFile::getDataSize() const {
  // get the data type enum
  int dataType = metadata_.at<data_type>();

  std::size_t dataSize = 0;
  switch (dataType) {
  case ByteType:
    dataSize = 1;
    break;
  case SignedIntegerType:
  case UnsignedIntegerType:
    dataSize = 2;
    break;
  case SignedLongType:
  case SinglePrecisionType:
  case UnsignedLongType:
    dataSize = 4;
    break;
  case DoublePrecisionType:
  case SingleComplexType:
  case SignedLongLongType:
  case UnsignedLongLongType:
    dataSize = 8;
    break;
  case DoubleComplexType:
    dataSize = 16;
    break;
  }

  return dataSize;
}
} // namespace d5::io
