#pragma once

#include "dirsig5/base/Geometry.h"

namespace d5::raster {

enum class SampleMode { Nearest, Linear, Cubic };
enum class ExtendMode { Clamp, Repeat, Mirror };
enum class ProjectorMode { Drape, UV };

[[nodiscard]] inline float doExtend(ExtendMode mode, float u) noexcept {
  if (mode == ExtendMode::Clamp)
    return std::fmin(std::fmax(u, 0.0f), 1.0f);
  if (mode == ExtendMode::Repeat)
    return u - std::floor(u);
  if (mode == ExtendMode::Mirror) {
    int q = 0;
    u = std::remquo(u, 1.0f, &q);
    if (u < 0)
      u++, q++;
    if (q & 1)
      u = 1 - u;
    return u == 1 ? 0 : u;
  }
  return u;
}

[[nodiscard]] inline int64_t doExtend(ExtendMode mode, int64_t i,
                                      int64_t n) noexcept {
  if (mode == ExtendMode::Clamp)
    return i <= 0 ? 0 : i < n ? i : n - 1;
  if (mode == ExtendMode::Repeat)
    return i >= 0 ? i % n : (n + i % n) % n;
  if (mode == ExtendMode::Mirror) {
    int64_t r = i % n;
    int64_t q = i / n;
    if (r < 0)
      r += n, q++;
    if (q & 1)
      r = n - r - 1;
    return r;
  }
  return i;
}

struct Coords {
  Point2i xy = {0, 0}; ///< The pixel (x, y) indexes.
  Point2f st = {0, 0}; ///< The pixel (s, t) coordinates.
  Point2f uv = {0, 0}; ///< The surface (u, v) coordinates.
};

struct Projector {
  ProjectorMode mode = ProjectorMode::UV;
  ExtendMode extendModeX = ExtendMode::Repeat;
  ExtendMode extendModeY = ExtendMode::Repeat;
  SampleMode sampleMode = SampleMode::Nearest;

  /// \name UV parameters
  /// \{
  Point2f uvMin = {0, 0}; ///< The min UV coordinates.
  Point2f uvMax = {1, 1}; ///< The max UV coordinates.
  /// \}

  /// \name Drape parameters
  /// \{
  Point2f drapeOrigin = {0, 0};    ///< The origin in meters.
  Point2f drapePixelSize = {1, 1}; ///< The pixel size in meters.
  /// \}

  [[nodiscard]] Point2i extend(Point2i index, Point2i count) const noexcept {
    return {doExtend(extendModeX, index[0], count[0]),
            doExtend(extendModeY, index[1], count[1])};
  }

  /// Construct map coordinates from image XY.
  [[nodiscard]] Coords projectXY(Point2f xy, Point2i count) const noexcept {
    int64_t xI = int64_t(std::floor(xy[0]));
    int64_t yI = int64_t(std::floor(xy[1]));
    return {extend({xI, yI}, count), {xy[0] - xI, xy[1] - yI}, {}};
  }

  /// Construct map coordinates from surface UV.
  [[nodiscard]] Coords projectUV(Point2f uv, Point2i count) const noexcept {
    uv[0] = doExtend(extendModeX, uv[0]);
    uv[1] = doExtend(extendModeY, uv[1]);
    uv[0] = (1 - uv[0]) * uvMin[0] + uv[0] * uvMax[0];
    uv[1] = (1 - uv[1]) * uvMin[1] + uv[1] * uvMax[1];
    Coords coords = projectXY(Point2f{uv[0] * float(count[0]), //
                                      uv[1] * float(count[1])},
                              count);
    coords.uv = uv;
    return coords;
  }

  /// Construct map coordinates from hit point and surface UV.
  [[nodiscard]] Coords projectHitAndUV(const Pointf &hitPoint, Point2f uv,
                                       Point2i count) const noexcept {
    if (mode == ProjectorMode::UV)
      return projectUV(uv, count);
    else {
      Point2f xy = {(hitPoint[0] - drapeOrigin[0]) / drapePixelSize[0],
                    (hitPoint[1] - drapeOrigin[1]) / drapePixelSize[1]};
      Coords coords = projectXY(xy, count);
      coords.uv = uv;
      return coords;
    }
  }
};

} // namespace d5::raster
