#pragma once

#include <cassert>
#include <memory>
#include <optional>
#include <stdexcept>
#include <string>
#include <utility>
#include <vector>

#include "dirsig5/base/DType.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/Logger.h"
#include "dirsig5/base/span.h"
#include "dirsig5/raster/Export.h"

namespace d5::raster {

/** A weakly-type container for an image data array (rows, columns, channels).
 *  Data is stored in band interleave by pixel (BIP) order. These objects have
 *  shared_ptr semantics (copies are shallow).
 */
class D5_RASTER_EXPORT Raster {
public:
  using DataType = DType;

  /// Default constructor. Makes an empty image.
  Raster() = default;

  /// Copy constructor. Shallow copy of the image.
  Raster(const Raster &) = default;

  /// Move constructor.
  Raster(Raster &&) noexcept = default;

  /// Copy assigment. Shallow copy of the image
  Raster &operator=(const Raster &) = default;

  /// Move assignement.
  Raster &operator=(Raster &&) noexcept = default;

  /** Constructor. Allocates the image.
   *
   *  \param rows      The number of rows in the image
   *  \param cols      The number of columns in the image
   *  \param channels  The number of channels in the image
   *  \param type      The data type of the image
   *  \param step      The line width (bytes) of the images. If not given, will
   *                   be auto-computed to enforce 32-bit alignment of lines.
   */
  Raster(size_t rows, size_t cols, size_t channels, DType type,
         size_t step = 0) {
    create(rows, cols, channels, type, step);
  }

  /// Destructor
  ~Raster() = default;

  /** Allocates the image array, dereferencing the previous image
   *
   *  \param rows      The number of rows in the image
   *  \param cols      The number of columns in the image
   *  \param channels  The number of channels in the image
   *  \param type      The data type of the image
   *  \param step      The line width (bytes) of the images. If not given, will
   *                   be auto-computed to enforce 32-bit alignment of lines.
   */
  void create(size_t rows, size_t cols, size_t channels, DType type,
              size_t step = 0);

  /// Whether the image is empty
  [[nodiscard]] bool empty() const noexcept {
    return rows_ * cols_ * channels_ == 0;
  }

  /// The type of the image.
  [[nodiscard]] DType type() const noexcept { return type_; }

  /// The number of rows in the image.
  [[nodiscard]] size_t rows() const noexcept { return rows_; }

  /// The stride in bytes between rows.
  [[nodiscard]] size_t rowStride() const noexcept { return rowStride_; }

  /// The number of columns in the image.
  [[nodiscard]] size_t cols() const noexcept { return cols_; }

  /// The number of channels in the image.
  [[nodiscard]] size_t channels() const noexcept { return channels_; }

  /// The raw data pointer.
  [[nodiscard]] uint8_t *rawData() noexcept { return buffer_.get(); }

  /// The raw data pointer, const variant.
  [[nodiscard]] const uint8_t *rawData() const noexcept {
    return buffer_.get();
  }

  /// The raw data pointer as a span.
  [[nodiscard]] span<uint8_t> rawDataSpan() noexcept {
    return {rawData(), rows_ * rowStride_};
  }

  /// The raw data pointer as a span, const variant.
  [[nodiscard]] span<const uint8_t> rawDataSpan() const noexcept {
    return {rawData(), rows_ * rowStride_};
  }

  [[nodiscard]] uint8_t *rawRow(size_t r) noexcept {
    return rawData() + std::min<size_t>(r, rows_ - 1) * rowStride_;
  }

  [[nodiscard]] const uint8_t *rawRow(size_t r) const noexcept {
    return rawData() + std::min<size_t>(r, rows_ - 1) * rowStride_;
  }

  /** Accessor for a row of the image.
   *
   *  \param r  The index of the row in the image
   *
   *  \return Pointer to the beginning of the row. Data is stored in band
   *          interleave by pixel order (BIP).
   */
  template <typename Ty> [[nodiscard]] auto *row(size_t r) const noexcept {
    return reinterpret_cast<const Ty *>(rawRow(r));
  }

  /// \overload
  template <typename Ty> [[nodiscard]] auto *row(size_t r) noexcept {
    return const_cast<Ty *>(static_cast<const Raster *>(this)->row<Ty>(r));
  }

  /** Accessor for a pixel in the image.
   *
   *  \param r  The index of the row of the pixel
   *  \param c  The index of the column of the pixel
   *  \return Pointer to the first channel in the pixel
   */
  template <typename Ty>
  [[nodiscard]] auto pixel(size_t r, size_t c) const noexcept {
    return span<const Ty>(this->row<Ty>(r) +
                              std::min<size_t>(c, cols_ - 1) * channels_,
                          channels_);
  }

  /// \overload
  template <typename Ty> [[nodiscard]] auto pixel(size_t r, size_t c) noexcept {
    return span<Ty>(this->row<Ty>(r) +
                        std::min<size_t>(c, cols_ - 1) * channels_,
                    channels_);
  }

  /** Value accessor.
   *
   *  \param r        The index of the row of the pixel
   *  \param c        The index of the column of the pixel
   *  \param channel  The index of the channel
   *
   *  \return The value of the pixel in the requested channel
   */
  template <typename Ty>
  [[nodiscard]] auto &at(size_t r, size_t c, size_t ch = 0) const {
    DIRSIG_THROW_IF_FMT(r >= rows_ || c >= cols_ || ch >= channels_,
                        "Out of bounds index! ({}, {}, {}) for image "
                        "dimensions ({}, {}, {})", //
                        r, c, ch, rows_, cols_, channels_);
    return pixel<Ty>(r, c)[ch];
  }

  /// \overload
  template <typename Ty>
  [[nodiscard]] auto &at(size_t r, size_t c, size_t ch = 0) {
    return const_cast<Ty &>(
        static_cast<const Raster *>(this)->at<Ty>(r, c, ch));
  }

  /** Generic function dispatcher. Calls the given callable with a
   *  strongly-typed RasterT<Ty> version of this image.
   *
   *  \param callable  Function that takes a strongly-typed RasterT<Ty> as
   *                   the first parameter.
   *  \param args      Any additional arguments to be forwarded to the callable.
   *
   *  \return The return value of the callable.
   */
  template <typename Callable, typename... Args>
  decltype(auto) dispatch(Callable &&callable, Args &&...args);

  /// \overload
  template <typename Callable, typename... Args>
  decltype(auto) dispatch(Callable &&callable, Args &&...args) const;

  /// Flips the image
  void flip(bool flipX, bool flipY);

  /// Metadata about each channel of the image
  struct ChannelMetadata {
    /// Name of the channel
    std::string name;

    /// Center of the channel's bandpass [micron]
    std::optional<double> centerWavelengthMicron;

    /// Width of the channel's bandpass (FWHM) [micron]
    std::optional<double> widthMicron;
  };

  /// Read-write access for channel metadata
  [[nodiscard]] ChannelMetadata &channelMetadata(size_t i) noexcept {
    return channelMetadata_[i];
  }

  /// Read-only access for channel metadata
  [[nodiscard]] const ChannelMetadata &
  channelMetadata(size_t i) const noexcept {
    return channelMetadata_[i];
  }

private:
  std::shared_ptr<uint8_t> buffer_;
  DType type_ = DType::Unknown;
  size_t rows_ = 0, rowStride_ = 0;
  size_t cols_ = 0;
  size_t channels_ = 0;
  std::vector<ChannelMetadata> channelMetadata_;
};

/** Strongly-typed subclass of Raster for syntatic sugar and usage in
 *  dispatch().
 */
template <typename Ty> class D5_RASTER_EXPORT RasterT : public Raster {
public:
  /// Advertise the type
  using value_type = Ty;

  /// Default constructor. Makes an empty image.
  RasterT() = default;

  /// Copy constructor. Shallow copy of the image.
  RasterT(const RasterT &) = default;

  /// Move constructor
  RasterT(RasterT &&) noexcept = default;

  /// Copy assignment. Shallow copy of the image.
  RasterT &operator=(const RasterT &) = default;

  /// Move assignment
  RasterT &operator=(RasterT &&) noexcept = default;

  /** Constructor. Makes a strongly-typed view on the given image.
   *
   *  \param da  The Raster we are referencing
   *  \throw std::domain_error if the types do not match
   */
  RasterT(const Raster &da) : Raster(da) {
    DIRSIG_THROW_IF_FMT(dtype_v<Ty> != type(),
                        "Wrong data type! {}, expected {}", to_string(type()),
                        to_string(dtype_v<Ty>));
  }

  /** Constructor. Allocates the image array.
   *
   *  \param rows      The number of rows in the image
   *  \param cols      The number of columns in the image
   *  \param channels  The number of channels in the image
   *  \param step      The line width (bytes) of the images. If not given, will
   *                   be auto-computed to enforce 32-bit alignment of lines.
   */
  RasterT(size_t rows, size_t cols, size_t channels, size_t step = 0) {
    create(rows, cols, channels, step);
  }

  /** Allocates the image array, dereferencing the previous image
   *
   *  \param rows      The number of rows in the image
   *  \param cols      The number of columns in the image
   *  \param channels  The number of channels in the image
   *  \param step      The line width (bytes) of the images. If not given, will
   *                   be auto-computed to enforce 32-bit alignment of lines.
   */
  void create(size_t rows, size_t cols, size_t channels, size_t step = 0) {
    Raster::create(rows, cols, channels, dtype_v<Ty>, step);
  }

  /// Row accessor.
  [[nodiscard]] auto *row(size_t r) noexcept { return Raster::row<Ty>(r); }

  /// Row accessor, const variant.
  [[nodiscard]] auto *row(size_t r) const noexcept {
    return Raster::row<Ty>(r);
  }

  /// Pixel accessor.
  [[nodiscard]] auto pixel(size_t r, size_t c) noexcept {
    return Raster::pixel<Ty>(r, c);
  }

  /// Pixel accessor, const variant.
  [[nodiscard]] auto pixel(size_t r, size_t c) const noexcept {
    return Raster::pixel<Ty>(r, c);
  }

  /// Value accessor.
  [[nodiscard]] auto &operator()(size_t r, size_t c, size_t ch = 0) {
    return Raster::at<Ty>(r, c, ch);
  }

  /// Value accessor, const variant.
  [[nodiscard]] auto &operator()(size_t r, size_t c, size_t ch = 0) const {
    return Raster::at<Ty>(r, c, ch);
  }
};

// ####################################################
// # Type trait implementation for DType
// ####################################################

/// Returns the size of the underlying data type for the given array
template <typename Ty>
constexpr uint8_t ByteSize(const RasterT<Ty> &rast) noexcept {
  return ByteSize(dtype_v<typename RasterT<Ty>::value_type>);
}

// ######################################
// # Inline Implementation for Raster
// ######################################

template <typename Callable, typename... Args>
inline decltype(auto) Raster::dispatch(Callable &&callable, Args &&...args) {
#define DTYPE_CASE(Ty)                                                         \
  case DType::Ty: {                                                            \
    RasterT<dtype_t<DType::Ty>> a(*this);                                      \
    return callable(a, std::forward<Args>(args)...);                           \
  }
  switch (type_) {
    DTYPE_CASE(Uint8)
    DTYPE_CASE(Int16)
    DTYPE_CASE(Int32)
    DTYPE_CASE(Float32)
    DTYPE_CASE(Float64)
    /*DTYPE_CASE(Complex64)*/
    /*DTYPE_CASE(Complex128)*/
    DTYPE_CASE(Uint16)
    DTYPE_CASE(Uint32)
    DTYPE_CASE(Int64)
    DTYPE_CASE(Uint64)
  default:
    DIRSIG_THROW("Invalid data type");
  }
#undef DTYPE_CASE
}

template <typename Callable, typename... Args>
inline decltype(auto) Raster::dispatch(Callable &&callable,
                                       Args &&...args) const {
#define DTYPE_CASE(Ty)                                                         \
  case DType::Ty: {                                                            \
    RasterT<dtype_t<DType::Ty>> a(*this);                                      \
    const auto &aRef = a;                                                      \
    return callable(aRef, std::forward<Args>(args)...);                        \
  }
  switch (type_) {
    DTYPE_CASE(Uint8)
    DTYPE_CASE(Int16)
    DTYPE_CASE(Int32)
    DTYPE_CASE(Float32)
    DTYPE_CASE(Float64)
    /*DTYPE_CASE(Complex64)*/
    /*DTYPE_CASE(Complex128)*/
    DTYPE_CASE(Uint16)
    DTYPE_CASE(Uint32)
    DTYPE_CASE(Int64)
    DTYPE_CASE(Uint64)
  default:
    DIRSIG_THROW("Invalid data type");
  }
#undef DTYPE_CASE
}

} // namespace d5::raster
