#pragma once

#include "dirsig5/raster/Projector.h"
#include "dirsig5/raster/Raster.h"

namespace d5::raster {

class RasterMap : public Raster {
public:
  /// Default constructor. Makes an empty image.
  RasterMap() = default;

  /// Copy constructor. Shallow copy of the image.
  RasterMap(const RasterMap &) = default;

  /// Move constructor
  RasterMap(RasterMap &&) noexcept = default;

  /// Copy assignment. Shallow copy of the image.
  RasterMap &operator=(const RasterMap &) = default;

  /// Move assignment
  RasterMap &operator=(RasterMap &&) noexcept = default;

  RasterMap(const Raster &rast) : Raster(rast) {}

  template <typename Value, int Channels = Eigen::Dynamic>
  using Pixel = Eigen::Matrix<Value, Channels, 1>;

  template <typename Value, int Channels = Eigen::Dynamic>
  using PixelView = Eigen::Map<Pixel<Value, Channels>>;

  template <typename Value, int Channels = Eigen::Dynamic>
  using PixelConstView = Eigen::Map<const Pixel<Value, Channels>>;

  [[nodiscard]] Point2i countXY() const noexcept {
    return {int64_t(cols()), int64_t(rows())};
  }

  [[nodiscard]] Point2i extendXY(const Point2i &xy) const noexcept {
    return projector.extend(xy, countXY());
  }

  [[nodiscard]] Coords projectXY(const Point2f &xy) const noexcept {
    return projector.projectXY(xy, countXY());
  }

  [[nodiscard]] Coords projectUV(const Point2f &uv) const noexcept {
    return projector.projectUV(uv, countXY());
  }

  [[nodiscard]] Coords projectHitAndUV(const Pointf &hitPoint,
                                       const Point2f &uv) const noexcept {
    return projector.projectHitAndUV(hitPoint, uv, countXY());
  }

  template <typename Value, int Channels = Eigen::Dynamic>
  [[nodiscard]] auto pixelView(const Point2i &xy) noexcept {
    if constexpr (Channels != Eigen::Dynamic)
      return PixelView<Value, Channels>(pixel<Value>(xy[1], xy[0]).data());
    else
      return PixelView<Value, Channels>(pixel<Value>(xy[1], xy[0]).data(),
                                        channels(), 1);
  }

  template <typename Value, int Channels = Eigen::Dynamic>
  [[nodiscard]] auto pixelView(const Point2i &xy) const noexcept {
    if constexpr (Channels != Eigen::Dynamic)
      return PixelConstView<Value, Channels>(pixel<Value>(xy[1], xy[0]).data());
    else
      return PixelConstView<Value, Channels>(pixel<Value>(xy[1], xy[0]).data(),
                                             channels(), 1);
  }

  template <typename Value, int Channels = Eigen::Dynamic,
            typename Float = std::common_type_t<Value, float>>
  [[nodiscard]] Pixel<Float, Channels> sample(const Coords &coords) const {
    if (projector.sampleMode == SampleMode::Nearest)
      return Pixel<Value, Channels>(pixelView<Value, Channels>(coords.xy))
          .template cast<Float>();
    else {
      auto s = Float(coords.st[0]);
      auto t = Float(coords.st[1]);
      auto p00 = pixelView<Value, Channels>(coords.xy);
      auto p01 =
          pixelView<Value, Channels>(extendXY(coords.xy + Point2i(0, 1)));
      auto p10 =
          pixelView<Value, Channels>(extendXY(coords.xy + Point2i(1, 0)));
      auto p11 =
          pixelView<Value, Channels>(extendXY(coords.xy + Point2i(1, 1)));
      auto lerp = [](Float t, const auto &a, const auto &b) {
        return (1 - t) * a + t * b;
      };
      if constexpr (Channels != Eigen::Dynamic) {
        assert(channels() == size_t(Channels));
        return lerp(s,      //
                    lerp(t, //
                         Pixel<Value, Channels>(p00).template cast<Float>(),
                         Pixel<Value, Channels>(p01).template cast<Float>()),
                    lerp(t, //
                         Pixel<Value, Channels>(p10).template cast<Float>(),
                         Pixel<Value, Channels>(p11).template cast<Float>()));
      } else {
        Pixel<Float, Channels> pixel;
        pixel.resize(channels());
        for (size_t i = 0; i < channels(); i++)
          pixel[i] = lerp(s, //
                          lerp(t, Float(p00[i]), Float(p01[i])),
                          lerp(t, Float(p10[i]), Float(p11[i])));
        return pixel;
      }
    }
  }

  Projector projector;
};

} // namespace d5::raster
