#pragma once

#include <iostream>
#include <memory>

#include "dirsig5/sqt/Export.h"
#include "dirsig5/sqt/SpectralCurve.h"

/** Request object for configuring a call to \c raw2sqt */
class Raw2SqtRequest {
public:
  Raw2SqtRequest& withNadirMagnitudeIn(
    std::shared_ptr<d5::sqt::SpectralCurve> nadirMag) {
    nadirMagnitudeIn = nadirMag;
    return *this;
  }

  /** Sets the input stream for the .raw data */
  Raw2SqtRequest& withRawIn(std::shared_ptr<std::istream> in) {
    rawIn = std::move(in);
    return *this;
  }

  /** Sets the output stream for the SQT data */
  Raw2SqtRequest& withSqtOut(std::shared_ptr<std::ostream> out) {
    sqtOut = std::move(out);
    return *this;
  }

  /** Sets the maximum depth for the SQT data */
  Raw2SqtRequest& withMaxDepth(unsigned depth) {
    maxDepth = depth;
    return *this;
  }

  /** Sets whether interpolation should be used */
  Raw2SqtRequest& withCompressThreshold(double value) {
    compressThreshold = value;
    return *this;
  }

  /** Sets whether interpolation should be used */
  Raw2SqtRequest& withInterpolation(bool flag) {
    interpolation = flag;
    return *this;
  }

  /** Sets whether verbose output should be generated */
  Raw2SqtRequest& withVerboseOutput(bool flag) {
    verboseOutput = flag;
    return *this;
  }

public:
  std::shared_ptr<d5::sqt::SpectralCurve> nadirMagnitudeIn;
  std::shared_ptr<std::istream> rawIn;
  std::shared_ptr<std::ostream> sqtOut;
  unsigned maxDepth = 7;
  double compressThreshold = 0.05;
  bool interpolation = true;
  bool verboseOutput = false;
};

/** Structure for wrapping the result of a raw2sqt request */
class Raw2SqtResult {
public:
  [[nodiscard]] bool success() const { return success_; }

  operator bool() const { return success_; }

  [[nodiscard]] const std::string& error() const { return error_; }

  static Raw2SqtResult Success() { return Raw2SqtResult(); }

  static Raw2SqtResult Failure(std::string err) {
    Raw2SqtResult r;
    r.success_ = false;
    r.error_ = std::move(err);
    return r;
  }

private:
  bool success_ = true;
  std::string error_;
};

/**
 * Convert a raw stream into an SQT stream
 */
[[nodiscard]] D5_SQT_EXPORT Raw2SqtResult
raw2sqt(const Raw2SqtRequest& request);
