#pragma once

#include "dirsig5/sqt/ScatteringFunctionSlice.h"

namespace d5::sqt {

class D5_SQT_EXPORT ScatteringFunction {
public:
  ScatteringFunction() = default;

  ScatteringFunction(const ScatteringFunction&) = delete;

  explicit ScatteringFunction(std::istream& in) { read(in); }

  explicit ScatteringFunction(std::string buf)
      : ScatteringFunction(buf.data(), buf.size()) {}

  explicit ScatteringFunction(void* buf, size_t bufLen);

  ~ScatteringFunction() { clear(); }

  /// Has data in the upper hemisphere?
  [[nodiscard]] bool hasBrdf() const noexcept {
    return flags & UPPER_HEMISPHERE_BIT;
  }

  /// Has data in the lower hemisphere?
  [[nodiscard]] bool hasBtdf() const noexcept {
    return flags & LOWER_HEMISPHERE_BIT;
  }

  /// Is cosine weighted?
  [[nodiscard]] bool isCosineWeighted() const noexcept {
    return flags & COSINE_WEIGHT_BIT;
  }

  /// Clear.
  void clear();

  /// Read from file.
  void read(const std::string& filename);

  /// Read from arbitrary stream.
  void read(std::istream&);

  /// Write to file.
  void write(const std::string& filename) const;

  /// Write to arbitrary stream.
  void write(std::ostream&) const;

  /// Write to string.
  std::string writeToString() const;

  /// Update state and recompute cached data.
  void updateState(const State* state) {
    for (Slice& slice : slices) slice.slice->updateState(state);
  }

  [[nodiscard]] float computeBroadbandSolarAverage() const;

  [[nodiscard]] float computeBroadbandThermalAverage() const;

  void forceNormalize();

  void overrideNadirMagnitude(const SpectralCurve& magnitude);

  void optimizeMagnitude(float tolerance = 0.005f);

  /// Get the integrated magnitude at the given wavelength.
  ///
  /// \param[in] omegaO        The output (or viewing) direction.
  /// \param[in] wave          The wavelength in microns.
  /// \param[in] includeUpper  Include upper hemisphere (BRDF) magnitude?
  /// \param[in] includeLower  Include lower hemisphere (BTDF) magnitude?
  ///
  [[nodiscard]] float getMagnitude(
    Vectorf omegaO,
    float wave,
    bool includeUpper = true,
    bool includeLower = true) const;

  /// Get the integrated magnitude at the state wavelengths.
  ///
  /// \param[in] omegaO        The output (or viewing) direction.
  /// \param[in] includeUpper  Include upper hemisphere (BRDF) magnitude?
  /// \param[in] includeLower  Include lower hemisphere (BTDF) magnitude?
  ///
  /// \note
  /// The `updateState()` method must be called beforehand in order
  /// for the implementation to know the state wavelengths!
  ///
  [[nodiscard]] SpectralVector getMagnitudeAtStateWavelengths(
    const Vectorf& omegaO,
    bool includeUpper = true,
    bool includeLower = true) const {
    SpectralVector magnitude;
    getMagnitudeAtStateWavelengths(
      omegaO, includeUpper, includeLower, magnitude);
    return magnitude;
  }

  /// \overload
  ///
  /// The non-allocating implementation.
  void getMagnitudeAtStateWavelengths(
    Vectorf omegaO,
    bool includeUpper,
    bool includeLower,
    SpectralVector& magnitude) const;

  /// Get scattering value.
  ///
  /// \param[in] omegaO  The output (or viewing) direction.
  /// \param[in] omegaI  The incident direction.
  /// \param[in] wave    The wavelength in microns.
  ///
  [[nodiscard]] float getValue(
    Vectorf omegaO, Vectorf omegaI, float wave) const;

  /// Get scattering value at state wavelengths.
  ///
  /// \param[in]  omegaO  The output (or viewing) direction.
  /// \param[in]  omegaI  The incident direction.
  ///
  /// \note
  /// The `updateState()` method must be called beforehand in order
  /// for the implementation to know the state wavelengths!
  ///
  [[nodiscard]] SpectralVector getValueAtStateWavelengths(
    const Vectorf& omegaO, const Vectorf& omegaI) const {
    SpectralVector value;
    getValueAtStateWavelengths(omegaO, omegaI, value);
    return value;
  }

  /// \overload
  ///
  /// The non-allocating implementation.
  void getValueAtStateWavelengths(
    Vectorf omegaO, Vectorf omegaI, SpectralVector& value) const;

  /// Get the probably density function (PDF) value.
  ///
  /// \param[in] omegaO      The output (or viewing) direction.
  /// \param[in] omegaI      The incident direction.
  /// \param[in] allowUpper  Allow upper hemisphere (BRDF) sampling?
  /// \param[in] allowLower  Allow lower hemisphere (BTDF) sampling?
  ///
  [[nodiscard]] float getPdf(
    Vectorf omegaO,
    Vectorf omegaI,
    bool allowUpper = true,
    bool allowLower = true) const;

  /// Sample direction from the probability density function (PDF).
  ///
  /// \param[in] omegaO      The output (or viewing) direction.
  /// \param[in] allowUpper  Allow upper hemisphere (BRDF) sampling?
  /// \param[in] allowLower  Allow lower hemisphere (BTDF) sampling?
  /// \param[out] weights    _Optional._ Weights at state wavelengths.
  ///
  [[nodiscard]] Vectorf samplePdf(
    RandEngine& prng,
    Vectorf omegaO,
    bool allowUpper = true,
    bool allowLower = true,
    SpectralVector* weights = nullptr) const;

public:
  Flags flags = 0;

  struct Slice {
    /// The output azimuth in [0,pi].
    float phiO = 0;

    /// The incident scattering function slice.
    ScatteringFunctionSlice* slice = nullptr;
  };

  std::vector<Slice> slices;

private:
  struct OmegaOFrame {
    OmegaOFrame() = default;
    OmegaOFrame(Vectorf omegaO) {
      omegaO.normalize();  // Just to be safe
      isFlipped = omegaO[2] < 0;
      cosThetaO = std::abs(omegaO[2]);
      sinThetaO = std::hypot(omegaO[0], omegaO[1]);
      thetaO = std::atan2(sinThetaO, cosThetaO);
      if (sinThetaO > 1e-5f) {
        cosPhiO = omegaO[0] / sinThetaO;
        sinPhiO = omegaO[1] / sinThetaO;
        phiO = std::atan2(sinPhiO, cosPhiO);
      }
    }

    bool isFlipped = false;
    float thetaO = 0;
    float cosThetaO = 1;
    float sinThetaO = 0;
    float phiO = 0;
    float cosPhiO = 1;
    float sinPhiO = 0;

    [[nodiscard]] Vectorf localOmegaO() const noexcept {
      return {sinThetaO, 0, cosThetaO};
    }

    [[nodiscard]] Vectorf localOmegaI(Vectorf omegaI) const noexcept {
      float omegaI0 = omegaI[0];
      float omegaI1 = omegaI[1];
      omegaI[0] = +cosPhiO * omegaI0 + sinPhiO * omegaI1;
      omegaI[1] = -sinPhiO * omegaI0 + cosPhiO * omegaI1;
      if (isFlipped) omegaI[2] *= -1;
      return omegaI.normalized();
    }

    [[nodiscard]] Vectorf worldOmegaI(Vectorf omegaI) const noexcept {
      float omegaI0 = omegaI[0];
      float omegaI1 = omegaI[1];
      omegaI[0] = cosPhiO * omegaI0 - sinPhiO * omegaI1;
      omegaI[1] = sinPhiO * omegaI0 + cosPhiO * omegaI1;
      if (isFlipped) omegaI[2] *= -1;
      return omegaI.normalized();
    }
  };

  struct FindSlicesResult {
    float fac = 0;
    const Slice* slice0 = nullptr;
    const Slice* slice1 = nullptr;
  };

  [[nodiscard]] FindSlicesResult findSlices(float phiO) const;
};

}  // namespace d5::sqt
