#pragma once

#include "dirsig5/sqt/SpectralSphereQuadTree.h"

namespace d5::sqt {

class D5_SQT_EXPORT ScatteringFunctionSlice {
public:
  ScatteringFunctionSlice() = default;

  ScatteringFunctionSlice(const ScatteringFunctionSlice&) = delete;

  ScatteringFunctionSlice(std::istream& in) { read(in); }

  ~ScatteringFunctionSlice() { clear(); }

  /// Clear.
  void clear();

  /// Read from arbitrary stream.
  void read(std::istream&);

  /// Write to arbitrary stream.
  void write(std::ostream&) const;

  /// Update state and recompute cached data.
  void updateState(const State* state) {
    for (Sphere& sphere : spheres) sphere.quadTree->updateState(state);
  }

  /// Get the integrated magnitude at the given wavelength.
  ///
  /// \param[in] thetaO        The viewing angle.
  /// \param[in] wave          The wavelength in microns.
  /// \param[in] includeUpper  Include upper hemisphere (BRDF) magnitude?
  /// \param[in] includeLower  Include lower hemisphere (BTDF) magnitude?
  ///
  [[nodiscard]] float getMagnitude(
    float thetaO,
    float wave,
    bool includeUpper = true,
    bool includeLower = true) const;

  /// Get the integrated magnitude at the state wavelengths.
  ///
  /// \param[in] thetaO        The viewing angle.
  /// \param[in] includeUpper  Include upper hemisphere (BRDF) magnitude?
  /// \param[in] includeLower  Include lower hemisphere (BTDF) magnitude?
  ///
  /// \note
  /// The `updateState()` method must be called beforehand in order
  /// for the implementation to know the state wavelengths!
  ///
  [[nodiscard]] SpectralVector getMagnitudeAtStateWavelengths(
    float thetaO,  //
    bool includeUpper = true,
    bool includeLower = true) const {
    SpectralVector magnitude;
    getMagnitudeAtStateWavelengths(
      thetaO, includeUpper, includeLower, magnitude);
    return magnitude;
  }

  /// \overload
  ///
  /// The non-allocating implementation.
  void getMagnitudeAtStateWavelengths(
    float thetaO,
    bool includeUpper,
    bool includeLower,
    SpectralVector& magnitude) const;

  /// Get scattering value.
  ///
  /// \param[in] thetaO  The output (or viewing) angle.
  /// \param[in] omegaI  The incident direction.
  /// \param[in] wave    The wavelength in microns.
  ///
  [[nodiscard]] float getValue(float thetaO, Vectorf omegaI, float wave) const;

  /// Get scattering value at state wavelengths.
  ///
  /// \param[in] thetaO  The output (or viewing) angle.
  /// \param[in] omegaI  The incident direction.
  ///
  /// \note
  /// The `updateState()` method must be called beforehand in order
  /// for the implementation to know the state wavelengths!
  ///
  [[nodiscard]] SpectralVector getValueAtStateWavelengths(
    float thetaO, const Vectorf& omegaI) const {
    SpectralVector value;
    getValueAtStateWavelengths(thetaO, omegaI, value);
    return value;
  }

  /// \overload
  ///
  /// The non-allocating implementation.
  void getValueAtStateWavelengths(
    float thetaO, Vectorf omegaI, SpectralVector& value) const;

  /// Get the probably density function (PDF) value.
  ///
  /// \param[in] thetaO      The output (or viewing) angle.
  /// \param[in] omegaI      The incident direction.
  /// \param[in] allowUpper  Allow upper hemisphere (BRDF) sampling?
  /// \param[in] allowLower  Allow lower hemisphere (BTDF) sampling?
  ///
  [[nodiscard]] float getPdf(
    float thetaO,
    Vectorf omegaI,
    bool allowUpper = true,
    bool allowLower = true) const;

  /// Sample direction from the probability density function (PDF).
  ///
  /// \param[in] thetaO      The output (or viewing) angle.
  /// \param[in] allowUpper  Allow upper hemisphere (BRDF) sampling?
  /// \param[in] allowLower  Allow lower hemisphere (BTDF) sampling?
  ///
  [[nodiscard]] Vectorf samplePdf(
    RandEngine& prng,
    float thetaO,
    bool allowUpper = true,
    bool allowLower = true) const;

public:
  struct Sphere {
    /// The output zenith in [0,pi/2].
    float thetaO = 0;

    /// The spectral sphere quad tree of incident scattering values.
    SpectralSphereQuadTree* quadTree = nullptr;
  };

  std::vector<Sphere> spheres;

private:
  struct FindSpheresResult {
    float fac = 0;
    const Sphere* sphere0 = nullptr;
    const Sphere* sphere1 = nullptr;
  };

  [[nodiscard]] FindSpheresResult findSpheres(float thetaO) const;
};

}  // namespace d5::sqt
