#pragma once

#include <ios>
#include <vector>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/span.h"
#include "dirsig5/sqt/Export.h"

namespace d5::sqt {

class D5_SQT_EXPORT SpectralCurve {
public:
  SpectralCurve() = default;

  template <typename Float1, typename Float2>
  SpectralCurve(span<const Float1> waves, span<const Float2> values) {
    static_assert(std::is_floating_point_v<Float1>);
    static_assert(std::is_floating_point_v<Float2>);
    DIRSIG_THROW_IF(waves.size() != values.size(), "Inconsistent sizes!");
    samples.resize(waves.size());
    for (size_t index = 0; index < samples.size(); index++) {
      samples[index].wave = waves[index];
      samples[index].value = values[index];
    }
  }

  explicit SpectralCurve(std::istream& in) { read(in); }

  void read(std::istream&);

  void write(std::ostream&) const;

  void clear() noexcept { samples.clear(); }

  [[nodiscard]] std::vector<float> unzipWaves() const {
    std::vector<float> waves(samples.size());
    for (const Sample& sample : samples) waves.push_back(sample.wave);
    return waves;
  }

  [[nodiscard]] std::vector<float> unzipValues() const {
    std::vector<float> values(samples.size());
    for (const Sample& sample : samples) values.push_back(sample.value);
    return values;
  }

  [[nodiscard]] float getValue(float wave) const;

  template <typename Float>
  [[nodiscard]] SpectralVector getValue(span<const Float> waves) const {
    static_assert(std::is_floating_point_v<Float>);
    SpectralVector values;
    values.resize(waves.size());
    for (size_t index = 0; index < waves.size(); index++)
      values[index] = getValue(waves[index]);
    return values;
  }

  template <typename Float>
  [[nodiscard]] SpectralVector getValue(const std::vector<Float>& waves) const {
    return getValue(span<const Float>{waves.data(), waves.size()});
  }

  [[nodiscard]] SpectralVector getValue(const SpectralVector& waves) const {
    return getValue(span<const double>{waves.data(), waves.size()});
  }

  [[nodiscard]] float getIntegral(float wave0, float wave1) const;

  template <typename Float>
  void resampleTo(span<const Float> newWaves);

  template <typename Float, typename Allocator>
  void resampleTo(const std::vector<Float, Allocator>& newWaves) {
    resampleTo(span<const Float>{newWaves.data(), newWaves.size()});
  }

  template <typename Float>
  void resampleToUnion(span<const Float> otherWaves);

  template <typename Float, typename Allocator>
  void resampleToUnion(const std::vector<Float, Allocator>& otherWaves) {
    resampleToUnion(span<const Float>{otherWaves.data(), otherWaves.size()});
  }

  void optimizeLinear(float tolerance = 0.005f);

  SpectralCurve& operator*=(const SpectralCurve& other) {
    for (Sample& sample : samples) sample.value *= other.getValue(sample.wave);
    return *this;
  }

  SpectralCurve& operator/=(const SpectralCurve& other) {
    for (Sample& sample : samples) sample.value /= other.getValue(sample.wave);
    return *this;
  }

  void clampToZeroOne() {
    for (Sample& sample : samples) {
      sample.value = std::fmax(sample.value, 0.0f);
      sample.value = std::fmin(sample.value, 1.0f);
    }
  }

  [[nodiscard]] float getMinWave() const noexcept {
    return samples.front().wave;
  }
  [[nodiscard]] float getMaxWave() const noexcept {
    return samples.back().wave;
  }

public:
  struct Sample {
    float wave = 0;   ///< Wavelength in microns.
    float value = 0;  ///< Value.
  };

  std::vector<Sample> samples;
};

}  // namespace d5::sqt
