#pragma once

#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/State.h"
#include "dirsig5/base/span.h"
#include "dirsig5/sqt/SpectralCurve.h"
#include "dirsig5/sqt/SphereQuadTree.h"

namespace d5::sqt {

class D5_SQT_EXPORT SpectralSphereQuadTree {
public:
  SpectralSphereQuadTree() = default;

  SpectralSphereQuadTree(const SpectralSphereQuadTree&) = delete;

  SpectralSphereQuadTree(std::istream& in) { read(in); }

  ~SpectralSphereQuadTree() { clear(); }

  struct RawData {
    /// \name Configuration
    /// \{

    /// The flags.
    Flags flags = 0;

    /// The maximum depth of the quad trees.
    unsigned maxDepth = 7;

    /// Clamp magnitude such that it never exceeds 1?
    bool clampMagnitude = true;

    /// Compression threshold. Set to zero to disable.
    float compressThreshold = 0.03f;

    /// Wave optimization threshold. Set to zero to disable.
    float waveOptimizeThreshold = 0.05f;

    /// Interpolate input data?
    bool interpolation = true;

    /// Verbose output?
    bool verboseOutput = false;

    /// \}

    using Waves = std::vector<float>;

    using SampleVectors =
      Eigen::Matrix<float, Eigen::Dynamic, 3, Eigen::RowMajor>;

    using SampleSpectra =
      Eigen::Matrix<float, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>;

    /// \name Samples
    /// \{

    /// The vector of wavelengths.
    ///
    /// This is a vector of `waveCount` wavelengths, which must be
    /// sorted in ascending order.
    Waves waves;

    /// The matrix of sample vectors.
    ///
    /// This is a `(sampleCount, 3)` matrix where the i'th row is the
    /// normalized direction vector for the i'th sample.
    SampleVectors sampleVectors;

    /// The matrix of sample spectra.
    ///
    /// This is a `(sampleCount, waveCount)` matrix where the i'th row contains
    /// the spectral vector for the i'th sample.
    SampleSpectra sampleSpectra;

    /// \}

    void resize(size_t sampleCount, size_t waveCount) {
      waves.resize(waveCount);
      sampleVectors.resize(Eigen::Index(sampleCount), 3);
      sampleSpectra.resize(Eigen::Index(sampleCount), Eigen::Index(waveCount));
    }
  };

  void buildFromRawData(const RawData& rawData);

  void clear();

  void read(std::istream&);

  void write(std::ostream&) const;

  void updateState(const State* state);

  /// Get the integrated magnitude at the given wavelength.
  ///
  /// \param[in] wave          The wavelength in microns.
  /// \param[in] includeUpper  Include the upper hemisphere integral?
  /// \param[in] includeLower  Include the lower hemisphere integral?
  ///
  [[nodiscard]] float getMagnitude(
    float wave,                //
    bool includeUpper = true,  //
    bool includeLower = true) const;

  /// Get the integrated magnitude at the state wavelengths.
  ///
  /// \param[in]  includeUpper  Include the upper hemisphere integral?
  /// \param[in]  includeLower  Include the lower hemisphere integral?
  /// \param[out] magnitude     The returned magnitude.
  ///
  [[nodiscard]] SpectralVector getMagnitudeAtStateWavelengths(
    bool includeUpper = true,  //
    bool includeLower = true) const {
    SpectralVector magnitude;
    getMagnitudeAtStateWavelengths(includeUpper, includeLower, magnitude);
    return magnitude;
  }

  /// \overload
  ///
  /// The non-allocating implementation.
  void getMagnitudeAtStateWavelengths(
    bool includeUpper, bool includeLower, SpectralVector& magnitude) const {
    magnitude.resize(cache.spectralState->wavelengthSamples.size());
    magnitude.setData(0);
    if (includeUpper) magnitude += cache.upperMagnitude;
    if (includeLower) magnitude += cache.lowerMagnitude;
  }

  /// Get the value at the given wavelength.
  [[nodiscard]] float getValue(const Vectorf& vec, float wave) const {
    return weight.getValue(wave) * interpolateSphere(vec, wave);
  }

  /// Get the value at the state wavelengths.
  ///
  /// \note
  /// The `updateState()` method must be called beforehand in order
  /// for the implementation to know the state wavelengths!
  ///
  [[nodiscard]] SpectralVector getValueAtStateWavelengths(
    const Vectorf& vec) const {
    SpectralVector value;
    getValueAtStateWavelengths(vec, value);
    return value;
  }

  /// \overload
  ///
  /// The non-allocating implementation.
  void getValueAtStateWavelengths(
    const Vectorf& vec, SpectralVector& value) const {
    interpolateSphere(
      vec,
      {cache.spectralState->wavelengthSamples.data(),
       cache.spectralState->wavelengthSamples.size()},
      value);
    value *= cache.weight;
  }

  /// Get the probably density function (PDF) value.
  ///
  /// \param[in] vec         The vector.
  /// \param[in] allowUpper  Allow upper hemisphere (BRDF) sampling?
  /// \param[in] allowLower  Allow lower hemisphere (BTDF) sampling?
  ///
  [[nodiscard]] float getPdf(
    Vectorf vec, bool allowUpper = true, bool allowLower = true) const;

  /// Sample direction from the probability density function (PDF).
  ///
  /// \param[in] allowUpper  Allow upper hemisphere (BRDF) sampling?
  /// \param[in] allowLower  Allow lower hemisphere (BTDF) sampling?
  ///
  [[nodiscard]] Vectorf samplePdf(
    RandEngine& prng, bool allowUpper = true, bool allowLower = true) const;

public:
  struct Sphere {
    float wave = 0;
    SphereQuadTree* quadTree = nullptr;
  };

  std::vector<Sphere> spheres;

  SpectralCurve weight;

private:
  struct Cache {
    size_t spectralStateIndex = size_t(-1);
    const SpectralState* spectralState = nullptr;
    SpectralVector weight;
    SpectralVector upperMagnitude;
    SpectralVector lowerMagnitude;
    float pdfFac = 0;
    const SphereQuadTree* pdfQuadTree0 = nullptr;
    const SphereQuadTree* pdfQuadTree1 = nullptr;
  } cache;

  [[nodiscard]] float interpolateSphere(Vectorf vec, float wave) const;

  void interpolateSphere(
    Vectorf vec, span<const double> waves, SpectralVector& values) const;

  struct FindSpheresResult {
    float fac = 0;
    const Sphere* sphere0 = nullptr;
    const Sphere* sphere1 = nullptr;
  };

  [[nodiscard]] FindSpheresResult findSpheres(float wave) const;

  friend class ScatteringFunction;
};

}  // namespace d5::sqt
