#pragma once

#include <array>
#include <bitset>
#include <cmath>
#include <iomanip>
#include <iostream>
#include <random>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Pcg.h"
#include "dirsig5/sqt/Export.h"

namespace d5::sqt {

/// This class implements a sphere pixelation based on the HEALPix
/// or Hierarchical Equal Area isoLatitude Pixelization structure described
/// by Gorski99 and further derived by Roukema08. It intentionally does
/// _not_ use the opensource, but GPL-licensed code available at
/// http://sourceforge.net/projects/healpix/ due to license restrictions.
/// All code presented here is derived directly from the two listed papers
/// and equation references (Roukema08) are given as much as possible.
/// The HEI in the name refers to the three requirements proposed by
/// Gorski and satisfied by the model (Hierarchical, Equal-area, and
/// Isolatitude).
///
/// HEALPix divides the sphere into 12 base diamond shapes, 4 equatorial
/// and 8 polar. Within each base shape a simple binary fractional mapping
/// of the shape is used to perform pixel indexing in and out of the sphere.
///
/// This class provides basic utility functions on the sphere and is meant
/// to be wrapped with additional code to be useful. The "order" of
/// pixelation describes how many times each sub-element is split in two
/// along each of two dimensions. Therefore the number of cells is equal
/// 12 * 2^(2*order), so 12, 48, 192, 768, 3072 are the sizes of the
/// first five orders.
///
class D5_SQT_EXPORT SphereHEI {
public:
  /// Region/fraction/fraction data structure for finding neighbors.
  struct RFF {
    unsigned R = 0;
    float f1 = 0;
    float f2 = 0;
  };

  /// The constructor fixes the order of the partition (the number of
  /// times the 12 diamond shapes are sub-divided into 4 child cells).
  /// The order is limited to 8 (roughly 3 million cells).
  SphereHEI(unsigned ord) : order(ord) {
    DIRSIG_THROW_IF(order > 8, "Requested partition order exceeds max (8)!");
    cellCount = 12 << (2 * order);
  }

  /// Get the number of equal-area cells stored on the sphere
  size_t getCellCount() const { return cellCount; }

  /// Get the current split (sub region width)
  float getSplit() const { return 1.0f / (1 << order); }

  /// Convert a normalized vector to an index. The vector is assumed to
  /// be pre-normalized. Results from an un-normalized vector are
  /// undefined.
  size_t vectorToIndex(const Vectorf& vec) const {
    unsigned region = 0;
    float f1 = 0;
    float f2 = 0;
    int f = 0;
    vectorToFractions(vec, region, f1, f2);
    encodeFractions(f1, f2, f);
    return region * (1 << 2 * order) + f;
  }

  /// Convert an index into the pixelation to a 3D, normalized
  /// vector on the sphere at the cell center. We follow HEALPix
  /// conventions for consistency.
  void indexToCenterVector(size_t index, Vectorf& vec) const;

  Vectorf indexToCenterVector(size_t index) const {
    Vectorf vec;
    indexToCenterVector(index, vec);
    return vec;
  }

  /// Convert an index into the pixelation to a list of vertexes in the
  /// cell (gnuplot polygon formatting). Output is written to the given stream
  void indexToCellList(size_t index, std::ostream& out) const;

  /// Convert an index into a set of 4 vertexes forming the shape of the
  /// corresponding cell. Results are added to the given vector.
  void indexToCell(size_t index, std::vector<Vectorf>& vecs) const;

  /// Convert an index into a uniform sample within that cell. We
  /// use a simple barycentric coordinate sampling scheme based on
  /// uniformly sampling the 2D fractions and using the angular
  /// representation of the 4 corners when possible (vector mixing
  /// is expensive). The resulting sample vector is normalized.
  void indexToSampleVector(
    size_t index, Vectorf& vec, RandEngine& randEngine) const;

  /// Get the index of the neighboring region in the direction (-1,0,+1)
  /// relative to the current region. If no neighbor exists in that
  /// particular direction then a -1 is returned (only for diagonals).
  int getRegionNeighbor(unsigned ri, int d1, int d2) const;

  /// Get the closest neighbors to a point in a cell (including the current
  /// cell). The list starts with the current cell then each neighbor
  /// in counter-clockwise order from (+1,0). If a neighbor does not
  /// exist then a -1 is returned for the region index.
  /// The neighbors are returned as region and fractions, not indexes.
  void getCorners(const Vectorf& v, std::array<RFF, 9>& ns) const;

  /// Write an OBJ for the entire partition -- shared vertices are not
  /// consolidated. The resulting OBJ file is written to the provided
  /// stream. No material is assigned, the file is limited to vertex
  /// and facet entries.
  void writeOBJ(std::ostream& out) const {
    size_t vIndex = 1;
    for (size_t ii = 0; ii < cellCount; ii++) {
      unsigned region = ii / (1 << 2 * order);
      int f = ii - (region * (1 << 2 * order));
      writeOBJFacet(region, f, vIndex, out, 1.f);
    }
  }

  /// Write a text file for the entire partition that is suitable for
  /// plotting in gnuplot or similar to the output stream provided. The
  /// optional step count determines of frequently the reulting polygons
  /// are sampled on the sphere.
  void writeTXT(std::ostream& out, float stepCount = 20) const {
    for (size_t ii = 0; ii < cellCount; ii++) {
      unsigned region = ii / (1 << 2 * order);
      int f = ii - (region * (1 << 2 * order));
      writeTXTFacet(region, f, out, stepCount);
    }
  }

  /// Write an OBJ facet based on index. A running vertex index is
  /// passed back and forth so we can track the appropriate facet.
  void writeOBJFacet(
    unsigned region,
    size_t index,
    size_t& vIndex,
    std::ostream& out,
    float mag) const;

  /// Write a TXT facet based on index. The provided step count
  /// determines how often the polygon is sampled on the sphere.
  void writeTXTFacet(
    unsigned region, size_t index, std::ostream& out, float stepCount) const;

  /// Write an SVG cell. The provided value must be between 0 and 1.
  void writeSVGFacet(
    unsigned region,
    size_t index,
    unsigned width,
    std::ostream& out,
    float val,
    bool bw) const;

  /// Write an abstract image of the grid that can be used for a textured uv
  /// space on a sphere. Output is a float in [0,1]. The image
  /// is sized to the given 2D array and is mapped into equally spaced angles
  /// in the range of [-pi/2,pi/2] and [-pi,pi] in latitude and longitude.
  void writeGridTexture(std::vector<std::vector<float>>& img) const;

  /// Convert a fraction within a region to a vector
  void fractionsToAngles(
    unsigned region, float f1, float f2, float& theta, float& phi) const;

  /// Convert a set of fractions within a region to a vector
  void fractionsToVector(
    unsigned region, float f1, float f2, Vectorf& vec) const {
    float theta = 0;
    float phi = 0;
    fractionsToAngles(region, f1, f2, theta, phi);
    anglesToVector(theta, phi, vec);
  }

  /// Convert a normalized vector to a region and fractions
  void vectorToFractions(
    const Vectorf& vec, unsigned& region, float& f1, float& f2) const;

  /// Convert a binary fraction to the corresponding fractions
  void decodeFractions(int f, float& f1, float& f2, float& curSplit) const {
    // fractions encoded in binary fraction, Equation 15
    curSplit = 0.5;
    f1 = 0;
    f2 = 0;
    unsigned bits = 2 * order;
    unsigned mask = 2 << (bits - 2);
    if (order > 0) {
      do {
        if (f & mask) f2 += curSplit;
        if (f & (mask >> 1)) f1 += curSplit;
        mask = mask >> 2;
        curSplit *= 0.5f;

      } while (mask != 0);
    }
    f1 += curSplit;
    f2 += curSplit;
  }

  /// Convert fractions to binary fraction
  void encodeFractions(float f1, float f2, int& f) const {
    // fractions encoded in binary fraction, Equation 15
    f = 0;
    unsigned bits = 2 * order;
    unsigned mask = 2 << (bits - 2);
    if (order > 0) {
      float curSplit = 0.5f;
      do {
        if (f2 > curSplit) f2 -= curSplit, f |= mask;
        if (f1 > curSplit) f1 -= curSplit, f |= mask >> 1;
        mask = mask >> 2;
        curSplit *= 0.5f;
      } while (mask != 0);
    }
  }

  size_t cellCount = 0;

  unsigned order = 0;

  /// Simple spherical to cartesian conversion helper.
  static void anglesToVector(float theta, float phi, Vectorf& vec) {
    float sinTheta = std::sin(theta);
    float cosTheta = std::cos(theta);
    vec[0] = sinTheta * std::cos(phi);
    vec[1] = sinTheta * std::sin(phi);
    vec[2] = cosTheta;
  }

  /// Simple spherical to cartesian conversion helper.
  static void vectorToAngles(const Vectorf& vec, float& theta, float& phi) {
    theta = std::atan2(std::hypot(vec[0], vec[1]), vec[2]);
    phi = std::atan2(vec[1], vec[0]);
  }
};

}  // namespace d5::sqt
