#pragma once

#include "dirsig5/sqt/Export.h"
#include "dirsig5/sqt/SphereHEI.h"

namespace d5::sqt {

/// A node in the quad tree. Each node contains four children as well as the
/// cumulative, contained value. Negative values are not allowed and are used
/// to mark uninitialized values. A node with any null children is a leaf node.
struct D5_SQT_EXPORT SphereNode {
  SphereNode() = default;

  ~SphereNode();

  /// Helper to delete and nullify children
  void deleteChildren() {
    if (children) {
      delete[] children;
      children = nullptr;
    }
  }

  /// Recursively expand the quad-tree to the given local depth.
  void expand(unsigned depth);

  /// Accumulate child nodes by summing their values. If one or more of the
  /// values are undefined then the average of the defined values is used
  /// instead, multiplied by four. If all four children are undefined then the
  /// average of each defined sibling will be used.
  void accumulate(unsigned curDepth = 0);

  /// Average any multi-sample cells
  void averageSamples();

  /// Recursively scale all the values by a constant value.
  void scale(float k);

  /// Recursively sample the tree based on a random value (within total)
  /// and return the index of the final cell.
  void sample(float xi, unsigned& order, size_t& index) const;

  /// Recursively compress each cell, replacing similar children with
  /// their parent node. The given threshold determines the minimum
  /// amount of the pdf in each leaf.
  void compress(float threshold, size_t curDepth = 0, size_t index = 0);

  /// Recursively count the leaves in the tree.
  void countLeaves(size_t& counter) const;

  /// Recursively walk and index until we can return a value.
  float walkIndex(
    size_t index,
    unsigned& curDepth,
    uint32_t maxDepth,
    const std::vector<SphereHEI>& partitions,
    uint32_t stopDepth = -1) const;

  /// Read a binary file back in sequential order. The call should be
  /// primed with the first depth. Since multiple regions are contained
  /// in the binary file, the last data depth and value is passed
  /// for subsequent calls.
  void read(
    size_t cellCount,
    unsigned& dataDepth,
    float& dataValue,
    unsigned curDepth,
    std::istream& in,
    uint32_t& readCount);

  /// Recursively write out the leaf data in this quad-tree. We track
  /// and write the current depth so we can step through during the
  /// inverse process and store heterogeneous trees. Note that each
  /// leaf is stored in 32-bits, where those 4 bytes are split between
  /// the depth and an integral representation of the fraction (all
  /// quad-trees are non-negative and normalized during finalization
  /// so we know that each value will be less than one).
  void write(unsigned curDepth, std::ostream& out) const;

  /// Recursively write out the leaf data as OBJ facets with the data
  /// value represented with vertex color. We track the current and
  /// have the maximum depth so we can adjust the pre-integrated value
  /// appropriately.
  void writeOBJ(
    unsigned region,
    unsigned curDepth,
    uint32_t maxDepth,
    size_t index,
    const std::vector<SphereHEI>& partitions,
    size_t& vIndex,
    std::ostream& out,
    bool useMag) const;

  /// Recursively write out the leaf data as an SVG with the data
  /// value represented by color. We track the current depth and
  /// have the maximum depth so we can adjust the pre-integrated value
  /// appropriately.
  void writeSVG(
    unsigned region,
    unsigned curDepth,
    uint32_t maxDepth,
    float width,
    size_t index,
    float solidAngle,
    const std::vector<SphereHEI>& partitions,
    std::ostream& out,
    bool bw) const;

  /// Recursively write out the leaf data as a text file.
  void writeTXT(
    unsigned region,
    unsigned curDepth,
    uint32_t maxDepth,
    size_t index,
    const std::vector<SphereHEI>& partitions,
    std::ostream& out) const;

  /// Children, null means that this is a leaf.
  SphereNode* children = nullptr;

  /// Value stored in-node, non-leaf nodes sum child values.
  float value = -1;

  /// Number of samples added to this node.
  uint32_t count = 0;
};

}  // namespace d5::sqt
