#pragma once

#include <array>
#include <cstddef>
#include <iostream>
#include <random>
#include <string>
#include <vector>

#include "dirsig5/base/Error.h"
#include "dirsig5/sqt/Export.h"
#include "dirsig5/sqt/SphereHEI.h"
#include "dirsig5/sqt/SphereNode.h"

namespace d5::sqt {

typedef std::uint32_t Flags;

/// Is non-zero in the upper hemisphere?
constexpr Flags UPPER_HEMISPHERE_BIT = 0x01;

/// Is non-zero in the lower hemisphere?
constexpr Flags LOWER_HEMISPHERE_BIT = 0x02;

/// Is non-zero over the full sphere?
constexpr Flags FULL_SPHERE = LOWER_HEMISPHERE_BIT | UPPER_HEMISPHERE_BIT;

/// Is cosine-weighted?
constexpr Flags COSINE_WEIGHT_BIT = 0x04;

/// This class implements a quad-tree storage mechanism built around the
/// Hierarchical Equal Area isoLatitude Pixelization structure modeled
/// by sqt::SphereHEI.
///
/// The quad tree depth describes how many times each of the initial 12
/// regions are split into 4 child cells. A depth of zero is not allowed
/// and the maximum is defined by sqt::SphereHEI.
///
/// \section Cosine-weighting
/// This class is commonly used to represent cosine-weighted functions,
/// such that initialization is performed as follows.
/// ~~~{cpp}
/// SphereQuadTree tree(initalDepth);
/// for (auto sample : samples) {
///     float cosTheta = std::abs(sample.vec[2]);
///     tree.addVectorData(sample.vec, sample.val * cosTheta);
/// }
/// tree.finalize();
/// tree.compress(eps); // Some epsilon, e.g., 1e-6
/// ~~~
/// This works, but is rather memory-inefficient in practice, as the
/// cosine envelope tends to make the tree less compressible.
///
/// To cope with this, this class accepts a cosine-weighting flag
/// which permits evaluation and sampling of a non-cosine-weighted function
/// as if it is cosine weighted, thus permitting cosine-weighted usage while
/// preserving compressibility. Pass `true` as the second argument to the
/// constructor to turn on cosine-weighting.
/// ~~~{cpp}
/// SphereQuadTree tree(initialDepth, true); // Cosine-weighting on
/// for (auto sample : samples) {
/// //  float cosTheta = std::abs(sample.vec[2]);
/// //  tree.addVectorData(sample.vec, sample.val * cosTheta);
///     tree.addVectorData(sample.vec, sample.val); // No weighting here
/// }
/// tree.finalize();
/// tree.compress(eps); // Some epsilon, e.g., 1e-6
/// ~~~
///
class D5_SQT_EXPORT SphereQuadTree {
public:
  SphereQuadTree(std::uint32_t initialDepth, Flags descriptorFlags)
      : maxDepth(initialDepth), flags(descriptorFlags) {
    DIRSIG_THROW_IF(maxDepth == 0, "Invalid maximum depth!");
    DIRSIG_THROW_IF(flags == 0, "Invalid flags!");
    partitions.reserve(maxDepth);
    for (std::uint32_t depth = 0; depth <= maxDepth; depth++)
      partitions.emplace_back(depth);
    for (SphereNode& region : regions) region.expand(maxDepth);
    solidAngle = getSolidAngleAtDepth(maxDepth);
  }

  explicit SphereQuadTree(std::istream& in) { read(in); }

  /// Add vector sample. Only valid before finalization.
  void addVectorData(Vectorf vec, float value);

  /// Finalize the quad-tree. Averages, accumulates, and normalizes.
  void finalize();

  /// Compress the quad tree. If all child nodes have a value within
  /// the given threshold from their mean, they will be replaced
  /// by their parent.
  void compress(float threshold);

  /// Common function to read input stream
  void read(std::istream& in);

  /// Write the entire quad-tree out (binary) to the given stream. The
  /// magnitude and solid angle are written first, followed by the quad-tree
  /// for each region
  void write(std::ostream& out) const;

  /// Write an OBJ visualization. Compressed cells are written out
  /// as single (poly) facets
  void writeOBJ(std::ostream& out, bool useMag) const {
    size_t vertexIndex = 1;
    size_t regionIndex = 0;
    for (const SphereNode& region : regions)
      region.writeOBJ(
        regionIndex++, 0, maxDepth, 0, partitions, vertexIndex, out, useMag);
  }

  /// Write an SVG visualization. Compressed cells are written out
  /// as single polygons in a 2D polar plot.
  void writeSVG(std::ostream& out, bool bw) const;

  /// Write contents to a text file as `<index> <value>`.
  void writeTXT(std::ostream& out) const {
    size_t regionIndex = 0;
    for (const SphereNode& region : regions)
      region.writeTXT(regionIndex++, 0, maxDepth, 0, partitions, out);
  }

public:
  /// Get the active leaf cell count.
  [[nodiscard]] size_t getCellCount() const {
    size_t counter = 0;
    for (const SphereNode& region : regions) region.countLeaves(counter);
    return counter;
  }

  /// Get solid angle at given partition depth.
  [[nodiscard]] float getSolidAngleAtDepth(unsigned depth) const {
    DIRSIG_THROW_IF(depth >= partitions.size(), "Depth out of range!");
    return float(4 * M_PI / double(partitions[depth].getCellCount()));
  }

  /// Has data in the upper hemisphere?
  [[nodiscard]] bool hasUpperHemisphere() const noexcept {
    return flags & UPPER_HEMISPHERE_BIT;
  }

  /// Has data in the lower hemisphere?
  [[nodiscard]] bool hasLowerHemisphere() const noexcept {
    return flags & LOWER_HEMISPHERE_BIT;
  }

  /// Is cosine weighted?
  [[nodiscard]] bool isCosineWeighted() const noexcept {
    return flags & COSINE_WEIGHT_BIT;
  }

  /// Get the integrated magnitude.
  [[nodiscard]] float getMagnitude() const noexcept { return magnitude; }

  /// Get the integrated magntiude of the upper hemisphere.
  [[nodiscard]] float getUpperMagnitude() const noexcept {
    return magnitude * upperFraction;
  }

  /// Get the integrated magntiude of the lower hemisphere.
  [[nodiscard]] float getLowerMagnitude() const noexcept {
    return magnitude * lowerFraction;
  }

  /// Set the integrated magnitude.
  void setMagnitude(float value) noexcept { magnitude = value; }

  /// Scale the magnitude.
  void scaleMagnitude(float scale) noexcept { magnitude *= scale; }

  /// Clamp the magnitude to make sure it is less than or equal to 1.
  void clampMagnitude() noexcept { magnitude = std::min(1.0f, magnitude); }

  /// Get the value at the given vector.
  [[nodiscard]] float getValue(const Vectorf& vec) const {
    return magnitude * getPdf(vec);
  }

  /// Get the PDF at the given vector.
  ///
  /// \note
  /// This is the probability density of vectors generated by `samplePdf()`.
  [[nodiscard]] float getPdf(
    Vectorf vec, bool allowUpper = true, bool allowLower = true) const;

  /// Sample the quad-tree to generate a vector.
  [[nodiscard]] Vectorf samplePdf(
    RandEngine& randEngine,
    bool allowUpper = true,
    bool allowLower = true) const;

private:
  /// Convert an index into the lowest depth node.
  SphereNode* indexToNode(size_t index);

  float indexToValue(size_t index);

  /// Flag indicating whether the quad-tree is finalized (and normalized)
  bool finalized = false;

  /// Maximum depth of any of the quad-trees.
  uint32_t maxDepth = 0;

  /// Flags.
  uint32_t flags = 0;

  /// Helper HEALPix partitions
  std::vector<SphereHEI> partitions;

  /// Top level parition regions (12)
  std::array<SphereNode, 12> regions;

  /// Solid angle of a full depth cell.
  float solidAngle = 0;

  /// Total value of the quad-tree (sum of 12 regions before normalization).
  float magnitude = 0;

  /// Relative sum of the upper hemisphere (includes half of equator sum).
  float upperFraction = 0;

  /// Relative sum of the lower hemisphere (includes half of equator sum).
  float lowerFraction = 0;
};

}  // namespace d5::sqt
