#pragma once

#include <stdexcept>
#include <vector>

#include "dirsig5/base/Constants.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Pcg.h"
#include "dirsig5/base/Temperature.h"
#include "dirsig5/base/Weather.h"
#include "dirsig5/base/span.h"
#include "dirsig5/thermal/ShadowHistory.h"

namespace d5::thermal {

/** Evaluate the classic Balfour model
 *
 *  \param params        Coefficient vector for the Balfour model
 *  \param weather       Weather data for the requested time
 *  \param cosSunNormal  Cosine of the angle between the sun and the surface
 *                       normal (pass 0 if in shadow).
 *
 *  \return The predicted temperature
 */
template <typename T>
inline Temperature ClassicBalfour(
  span<const T, 5> params, const Weather& weather, float cosSunNormal) {
  auto airTempK = weather.airTemperature.get<Temperature::Kelvin>();
  auto tempC = params[0] +
               params[1] * weather.directInsolationWM2 * cosSunNormal +
               params[2] * weather.diffuseInsolationWM2 + params[3] * airTempK +
               params[4] * airTempK * weather.windSpeedMS;
  return Temperature::fromCelsius(tempC);
}

template <typename T, typename ShadowHistory, typename GetWeather>
Temperature IterativeBalfour(
  span<const T, 6> params,
  std::uint16_t seed,
  std::size_t numSamples,
  const Vectorf& surfaceNormal,
  ShadowHistory&& getShadowHistory,
  GetWeather&& getWeather) {
  assert_shadow_history_callback(getShadowHistory);
  static_assert(
    std::is_invocable_r_v<Weather, GetWeather, float>,
    "getWeather must be a weather callback");

  static thread_local std::vector<float> timesS;
  timesS.resize(numSamples);

  // Compute the sample times and get the shadow history for those times
  RandEngine rng(seed);
  double randParam = -std::log(static_cast<double>(params[5]));
  for (std::size_t i = 0; i < numSamples; i++) {
    auto u = std::generate_canonical<double, 32>(rng);  // u in [0, 1)
    u = (u + i) / numSamples;  // u in [i/numSamples, (i + 1)/numSamples)
    timesS[i] = 3600 * std::log1p(-u) / randParam;  // t = log(1-u) / -log(a5)
  }
  auto shadowHistory = getShadowHistory(timesS);

  double t = 0;
  for (std::size_t i = 0; i < numSamples; i++) {
    auto wth = getWeather(timesS[i]);
    const auto& history = shadowHistory[i];
    float sunCos =
      (1 - history.shadowFraction) * surfaceNormal.dot(history.solarVector);
    if (sunCos < 0) sunCos = 0;
    t += ClassicBalfour(params.template subspan<0, 5>(), wth, sunCos).kelvin();
  }
  return Temperature::fromKelvin(t / numSamples);
}

}  // namespace d5::thermal
