#pragma once

#include <array>
#include <stdexcept>
#include <vector>

#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Weather.h"
#include "dirsig5/thermal/Export.h"

namespace d5::thermal {

/** Compute time steps for THERM.
 *
 *  The general idea is that we do some large, coarse timesteps for the
 *  equilibrium temperatures further back and switch to shorter, fine
 *  timesteps for the equilibrium temperatures close to our prediction time.
 *
 *  Only a subset of these stepping times may actually be needed for surfaces
 *  with less thermal interia.
 *
 *  \return List of timesteps given in seconds expected by THERM.
 */
D5_THERMAL_EXPORT const std::vector<float>& ThermTimeSteps();

/** Convert a history length to the appropriate start time step index.
 *
 *  \param[in] historyLength
 *  The history length in seconds. This should be a positive number (worth
 *  noting because the time steps from `ThermTimeSteps()` are negative
 *  numbers).
 */
D5_THERMAL_EXPORT size_t HistoryLengthToTimeStepIndex(double historyLength);

/**
 * Invariant data for RIT-THERM model.
 *
 * This class holds all the data that can be shared between THERM models.  We
 * attempt to precompute as much as possible (e.g. effective sky emissivity) in
 * order to speed up the computationally expensive process of thermal modeling.
 *
 * This class has a superset of the functionality in ThermPreferences.  It is
 * also designed to integrate more cleanly with THERM itself.  In fact,
 * RIT-THERM will be using this data structure directly.
 */
class D5_THERMAL_EXPORT RitThermInvariants {
public:
  /** Describes the invariants related to a given THERM timestep.
   *  \ingroup core_components
   */
  struct ThermTimeStepInvariants {
    /// Time in seconds relative to the temporal state
    double relativeTimeS = 0;

    /// Time as THERM understands it [decimal hours]
    double timeHr = 0;

    /// The weather data
    Weather weather;

    /// Sun vector
    Vector sunVector;

    /// Cosine of sun vector zenith
    float cosSunZenith = 0;

    /// Sine of sun vector zenith
    float sinSunZenith = 0;

    /// Sun vector azimuth [geo,degrees]
    float sunAzimuthDegrees = 0;

    /// Effective sky emissivity
    float effSkyEmissivity = 0;
  };

public:
  using value_type = ThermTimeStepInvariants;
  using size_type = std::vector<value_type>::size_type;
  using reference = value_type&;
  using const_reference = const value_type&;
  using iterator = std::vector<value_type>::iterator;
  using const_iterator = std::vector<value_type>::const_iterator;

  /// Default constructor
  RitThermInvariants() = default;

  /// Constructor with data
  RitThermInvariants(std::vector<value_type> data) { setData(std::move(data)); }

  /// Sets the data from a list of time step invariants
  void setData(std::vector<value_type> data);

  /** Retrieve invariant data for a given timestep.
   *
   *  \param i  Time step index
   *
   *  \return Invariant data associated with timestep
   */
  [[nodiscard]] const_reference at(std::size_t i) const { return data_.at(i); }

  /** Retrieve invariant data for a given timestep.
   *
   *  \param i  Time step index
   *
   *  \return Invariant data associated with timestep
   */
  [[nodiscard]] const_reference operator[](std::size_t i) const {
    return data_[i];
  }

  /// Get the invariant data for the first timestep
  [[nodiscard]] const_reference front() const { return data_.front(); }

  /// Get the invariant data for the last timestep
  [[nodiscard]] const_reference back() const { return data_.back(); }

  /// Gets the beginning iterator for the timesteps
  [[nodiscard]] const_iterator begin() const { return data_.begin(); }

  /// Gets the ending iterator for the timesteps
  [[nodiscard]] const_iterator end() const { return data_.end(); }

  /// Returns whether there are no timesteps
  [[nodiscard]] bool empty() const { return data_.empty(); }

  /// Gets the number of timesteps
  [[nodiscard]] size_type size() const { return data_.size(); }

  /** Compute time elapsed since previous timestep.
   *
   *  \param i  Timestep index, must be in [1, size() - 1]
   *
   *  \return   time elapsed since previous timestep [hours]
   */
  [[nodiscard]] double timeStepDelta(size_type i) const {
    if (i == 0 || i >= size()) {
      throw std::domain_error("RitThermInvariants::timeStepDelta");
    }

    // Time since previous entry
    auto dt = data_[i].timeHr - data_[i - 1].timeHr;

    // Handle midnight boundary
    dt = (dt < 0) ? dt + 24 : dt;

    return dt;
  }

  /** Get air temperature at sunrise
   *
   *  This is linearly interpolated from the weather data during
   *  RitThermInvariants::init() using the sunrise time.
   */
  [[nodiscard]] Temperature sunriseAirTemp() const { return sunriseAirTemp_; }

  /// Get decimal time of highest temperature [decimal hours]
  [[nodiscard]] double peakTempTimeHr() const { return peakTempTime_; }

  /// Get highest air temperature
  [[nodiscard]] Temperature peakAirTemp() const { return peakAirTemp_; }

  static void ComputeEffectiveSkyEmissivity(value_type& stepData);

  template <typename GetWeather, typename GetSolarVector>
  static RitThermInvariants Create(
    const DateTime& referenceDateTime,
    GetWeather&& getWeather,
    GetSolarVector&& getSolarVector);

private:
  /// Perform initialization tasks
  void init();

  /// Contains invariants information for each timestep.
  std::vector<ThermTimeStepInvariants> data_;

  /** Temperature at sunrise [C].  This is linearly interpolated from the
   *  weather data.
   */
  Temperature sunriseAirTemp_;

  /// Time for highest temperature in weather history [decimal hours].
  double peakTempTime_ = 0;

  /// Highest temperature in weather history [C].
  Temperature peakAirTemp_;
};

// #######################################################
// #               Template Implementation               #
// #######################################################

template <typename GetWeather, typename GetSolarVector>
RitThermInvariants RitThermInvariants::Create(
  const DateTime& referenceDateTime,
  GetWeather&& getWeather,
  GetSolarVector&& getSolarVector) {
  const auto& timesteps = d5::thermal::ThermTimeSteps();

  // Setup the invariants for each timestep
  std::vector<d5::thermal::RitThermInvariants::value_type> tInvariants(
    timesteps.size());
  for (std::size_t i = 0; i < timesteps.size(); i++) {
    auto& inv = tInvariants[i];

    inv.relativeTimeS = timesteps[i];
    {
      int d, m, y, doy;
      double fracSec =
        (referenceDateTime + inv.relativeTimeS).getCalenderDate(d, m, y, doy);
      inv.timeHr = fracSec / 3600.;
    }
    inv.weather = getWeather(timesteps[i]);
    inv.sunVector = getSolarVector(timesteps[i]);
    inv.cosSunZenith = inv.sunVector(2);
    inv.sinSunZenith = sqrt(1 - inv.cosSunZenith * inv.cosSunZenith);
    inv.sunAzimuthDegrees =
      static_cast<float>(toDegrees(getGeoAzimuth(inv.sunVector)));
    d5::thermal::RitThermInvariants::ComputeEffectiveSkyEmissivity(inv);
  }

  return RitThermInvariants(tInvariants);
}
}  // namespace d5::thermal
