/** Implementation file for atmospheric model backends
 *
 *  \file Atmosphere.cpp
 */
#include <spdlog/fmt/fmt.h>

#include "dirsig5/atmosphere/Atmosphere.h"

#include "backends/Dummy/DummyBackend.h"
#include "backends/ModtranJson/ModtranJsonBackend.h"
#include "backends/ModtranTape/ModtranTapeBackend.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/Logger.h"
#include "dirsig5/base/Paths.h"

void d5::atm::from_json(const nlohmann::json &j, ModelRequest &r) {
  r.type = static_cast<ModelRequestType>(j.at("type").get<uint32_t>());
  r.options = static_cast<ModelRequestOptions>(j.at("options").get<uint32_t>());
  j.at("index").get_to(r.index);
  j.at("dayOfYear").get_to(r.dayOfYear);
  j.at("wavelengthSamples").get_to(r.wavelengthSamples);
  j.at("groundAlt").get_to(r.groundAltM);
  j.at("pathAlt1").get_to(r.pathAlt1M);
  j.at("pathAlt2").get_to(r.pathAlt2M);
  j.at("pathLength").get_to(r.pathLengthM);
  j.at("pathZenith").get_to(r.pathZenithDeg);
  j.at("pathAzimuth").get_to(r.pathAzimuthDeg);
  j.at("sourceZenith").get_to(r.sourceZenithDeg);
  j.at("sourceAzimuth").get_to(r.sourceAzimuthDeg);
  j.at("sourcePhaseAngle").get_to(r.sourcePhaseAngleDeg);
  if (auto it = j.find("backendInput"); it != j.end()) {
    r.backendInput = it.value();
  }
}

void d5::atm::to_json(Json &j, const ModelRequest &r) {
  j = Json{{"type", static_cast<uint32_t>(r.type)},
           {"options", static_cast<uint32_t>(r.options)},
           {"index", r.index},
           {"dayOfYear", r.dayOfYear},
           {"wavelengthSamples", r.wavelengthSamples},
           {"groundAlt", r.groundAltM},
           {"pathAlt1", r.pathAlt1M},
           {"pathAlt2", r.pathAlt2M},
           {"pathLength", r.pathLengthM},
           {"pathZenith", r.pathZenithDeg},
           {"pathAzimuth", r.pathAzimuthDeg},
           {"sourceZenith", r.sourceZenithDeg},
           {"sourceAzimuth", r.sourceAzimuthDeg},
           {"sourcePhaseAngle", r.sourcePhaseAngleDeg},
           {"backendInput", r.backendInput}};
}

void d5::atm::from_json(const Json &j, CreateModelBackendRequest &r) {
  j.at("name").get_to(r.name);
  r.backendData = j;
  r.backendData.erase("name");
}
void d5::atm::to_json(Json &j, const CreateModelBackendRequest &r) {
  j = Json{{"name", r.name}};
  j.merge_patch(r.backendData);
}

std::unique_ptr<d5::atm::ModelBackend>
d5::atm::createBackend(const CreateModelBackendRequest &req) {
  std::unique_ptr<ModelBackend> modelBackend;
  if ((req.name == "dummy") || (req.name == "uniform") ||
      (req.name == "simple")) {
    modelBackend = std::make_unique<DummyBackend>();
  } else if (req.name == "modtran_tape") {
    modelBackend =
        std::make_unique<modtran::tape::ModtranTapeBackend>(req.backendData);
  } else if (req.name == "modtran_json") {
    modelBackend =
        std::make_unique<modtran::json::ModtranJsonBackend>(req.backendData);
  } else {
    DIRSIG_THROW_FMT("unknown/unsupported backend model: name = '{}'",
                     req.name);
  }

  return modelBackend;
}

static d5::atm::PathResolutionCallback s_pr = [](const std::string &s) {
  return s;
};

void d5::atm::SetPathResolutionCallback(PathResolutionCallback f) {
  s_pr = std::move(f);
}

std::string d5::atm::ResolvePath(const std::string &p) {
  DIRSIG_ERROR_TRACE("attempting to resolve path '" + p + "'",
                     { return s_pr(d5::ReplaceEnvVars(p)); });
}
