/** Implementation file for the dummy backend
 *
 *  \file DummyBackend.cpp
 */

#include "DummyBackend.h"

#include <limits>

#include "../Common.h"

namespace {

void computeBlackbodyIrradiance(double tempK, double minLambda,
                                double deltaLambda, size_t lambdaCount,
                                std::vector<double> &wavelengths,
                                std::vector<double> &irradiance) {
  wavelengths.resize(lambdaCount);
  for (size_t ll = 0; ll < lambdaCount; ll++) {
    wavelengths[ll] = minLambda + (ll * deltaLambda);
  }

  irradiance = d5::atm::BlackbodyIrradiance(tempK, wavelengths);
}
} // namespace

namespace d5::atm {

ModelResponse DummyBackend::performRequest(const ModelRequest &request) const {
  static const double infiniteLength = std::numeric_limits<double>::max();

  size_t lambdaCount = request.wavelengthSamples.size();
  double minLambda = request.wavelengthSamples.front();
  double maxLambda = request.wavelengthSamples.back();
  double deltaLambda = (maxLambda - minLambda) / (lambdaCount - 1.);
  assert(deltaLambda != 0.0);

  // setup the response
  ModelResponse response;
  response.buffer1.resize(request.wavelengthSamples.size());
  response.buffer2.resize(request.wavelengthSamples.size());

  switch (request.type) {
  case ModelRequestType::ProfileData:
    break;
  case ModelRequestType::ExoSourceData:
    if (request.options & ModelRequestOptions::SunSourceRequest) {
      this->computeExoSolar(minLambda, deltaLambda, lambdaCount, response);
    } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
      this->computeExoLunar(minLambda, deltaLambda, lambdaCount, response);
    }
    break;
  case ModelRequestType::PathData:
    if (request.pathLengthM == infiniteLength) {
      this->computeSkyPath(minLambda, deltaLambda, lambdaCount, response);
    } else {
      this->computeOtherPath(minLambda, deltaLambda, lambdaCount, response);
      response.range = static_cast<float>(request.pathLengthM);
    }
    break;
  default:
    break;
  }

  return response;
}

// This model simply uses a scaled 5800K blackbody distribution.
void DummyBackend::computeExoSolar(double minLambda, double deltaLambda,
                                   std::size_t lambdaCount,
                                   ModelResponse &response) const {
  std::vector<double> wavelengths, irradiance;
  computeBlackbodyIrradiance(5800, minLambda, deltaLambda, lambdaCount,
                             wavelengths, irradiance);
  for (size_t ii = 0; ii < lambdaCount; ii++) {
    response.buffer1[ii] = static_cast<float>(irradiance[ii] / 5.0e+04);
    response.buffer2[ii] = 1.0f;
  }
}

// This model simply uses a scaled 5800K blackbody distribution and
// a spectrally flat lunar reflectance.
void DummyBackend::computeExoLunar(double minLambda, double deltaLambda,
                                   std::size_t lambdaCount,
                                   ModelResponse &response) const {
  std::vector<double> wavelengths, irradiance;
  computeBlackbodyIrradiance(5800, minLambda, deltaLambda, lambdaCount,
                             wavelengths, irradiance);
  for (size_t ii = 0; ii < lambdaCount; ii++) {
    response.buffer1[ii] =
        static_cast<float>(irradiance[ii] * 0.18 / PI * 6.436e-05 / 5.0e+04);
    response.buffer2[ii] = 1.0f;
  }
}

/**
 * This model simply uses an expoential function that mimicks a clear sky
 * radiance. The extinction along these paths is 0.
 */
void DummyBackend::computeSkyPath(double minLambda, double deltaLambda,
                                  std::size_t lambdaCount,
                                  ModelResponse &response) const {
  for (size_t ii = 0; ii < lambdaCount; ii++) {
    double wl = minLambda + (ii * deltaLambda);
    response.buffer1[ii] = 0.00035 / pow(wl, 2.6);
    response.buffer2[ii] = 0.0f;
  }
}

/**
 * For all non-sky paths, the radiance and extinction are 0.
 * respectively.
 */
void DummyBackend::computeOtherPath(double /*minLambda*/,
                                    double /*deltaLambda*/,
                                    std::size_t lambdaCount,
                                    ModelResponse &response) const {
  for (size_t ii = 0; ii < lambdaCount; ii++) {
    response.buffer1[ii] = 0.0f;
    response.buffer2[ii] = 0.0f;
  }
}
} // namespace d5::atm
