#include "ModtranJsonBackend.h"

#include <atomic>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <limits>
#include <mutex>
#include <sstream>

#include <spdlog/fmt/fmt.h>

#include <boost/process.hpp>
namespace bp = boost::process;

#include <filesystem>
namespace fs = std::filesystem;

#include "../Common.h"
#include "OutputHelper.h"
#include "dirsig5/atmosphere/backends/ModtranJson/ModtranJson.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/StandardPaths.h"

static std::atomic<int> s_dirSeqNum = 0;
static const std::string kFormat1ProfileHeader = "1 ATMOSPHERIC PROFILES";
static const std::string kFormat2ProfileHeader = " ATMOSPHERIC PROFILES";

namespace {

template <typename Arg, typename... Rest>
fs::path PathCatHelper(fs::path &base, Arg &&arg, Rest &&...rest) {
  base /= arg;
  return PathCat(base, std::forward<Rest>(rest)...);
}

template <typename Arg> fs::path PathCatHelper(fs::path &base, Arg &&arg) {
  base /= arg;
  return base;
}

template <typename... Args>
fs::path PathCat(const fs::path &base, Args &&...args) {
  auto p = base;
  return PathCatHelper(p, std::forward<Args>(args)...);
}

std::ifstream Ifstream(const std::string &f) {
  std::ifstream ifs(f);
  DIRSIG_THROW_IF(!ifs, std::string("unable to open file: ") + strerror(errno));
  return ifs;
}

std::ofstream Ofstream(const std::string &f) {
  std::ofstream ofs(f);
  DIRSIG_THROW_IF(!ofs, std::string("unable to open file: ") + strerror(errno));
  return ofs;
}

auto StreamSize(std::istream &is) {
  auto curPos = is.tellg();
  is.seekg(0, std::ios::beg);
  auto fsize = is.tellg();
  is.seekg(0, std::ios::end);
  fsize = is.tellg() - fsize;
  is.seekg(curPos, std::ios::beg);
  return fsize;
}

// constants for a dew point calculation (Magnus formula)
// all based on Arden Buck (1981) for ranges above and below 0 deg C
static const double Tdp_b_pos = 17.368; // [unitless]
static const double Tdp_c_pos = 238.88; // [unitless]
static const double Tdp_b_neg = 17.966; // [unitless]
static const double Tdp_c_neg = 247.15; // [unitless]

// functions for calculating dew point [deg C] from RH [%]
double Tdp_gamma(double Tc, double RH) {
  return (Tc >= 0) ? log(RH / 100.) + Tdp_b_pos * Tc / (Tdp_c_pos + Tc)
                   : log(RH / 100.) + Tdp_b_neg * Tc / (Tdp_c_neg + Tc);
}

double Tdp(double Tc, double RH) {
  const double gamma = Tdp_gamma(Tc, RH);
  return (Tc >= 0) ? Tdp_c_pos * gamma / (Tdp_b_pos - gamma)
                   : Tdp_c_neg * gamma / (Tdp_b_neg - gamma);
}
} // namespace

namespace d5::atm::modtran::json {

void from_json(const nlohmann::json &j, Config &c) {
  if (auto it = j.find("modtran_profile"); it != j.end()) {
    c.profile = it.value();
  } else {
    Profile p;
    try {
      p.exeFilePath = ResolvePath(j.at("modtran_exe").get<std::string>());
      p.dataDirPath = ResolvePath(j.at("modtran_data").get<std::string>());
      c.profile = p;
    } catch (const std::exception &e) {
      DIRSIG_THROW("Error resolving MODTRAN executable and data path!\n" +
                   std::string(e.what()));
    }
    if (p.exeFilePath.empty() || p.dataDirPath.empty()) {
      DIRSIG_THROW("If 'modtran_profile' is not given, 'modtran_exe' and "
                   "'modtran_data' must be non-empty strings");
    }
  }

  if (auto it = j.find("modtran_options"); it != j.end()) {
    it.value().get_to(c.modtranOptions);
  }

  if (auto it = j.find("json_contents"); it != j.end()) {
    JsonContents json;
    it.value().get_to(json.contents);
    c.json = std::move(json);
  } else if (it = j.find("json_filename"); it != j.end()) {
    JsonFilename json;
    it.value().get_to(json.filename);
    json.filename = ResolvePath(json.filename);
    c.json = std::move(json);
  } else {
    c.json = nullptr;
  }

  // check for the (optional) flag to extract profiles
  if (auto it = j.find("extract_profile"); it != j.end()) {
    it.value().get_to(c.extractProfile);
  }
}

void from_json(const nlohmann::json &j, RequestBackendInput &v) {
  if (auto it = j.find("json_contents"); it != j.end()) {
    JsonContents json;
    it.value().get_to(json.contents);
    v.json = std::move(json);
  } else if (it = j.find("json_filename"); it != j.end()) {
    JsonFilename json;
    it.value().get_to(json.filename);
    json.filename = ResolvePath(json.filename);
    v.json = std::move(json);
  } else {
    v.json = nullptr;
  }
}

static const double infiniteLength = std::numeric_limits<double>::max();
static const float infiniteExt = std::numeric_limits<float>::max();

ModtranJsonBackend::ModtranJsonBackend(const nlohmann::json &input) {
  // create an empty folder for the MODTRAN runs
  std::string dirName = "modtran_runs" + std::to_string(s_dirSeqNum++);
  DIRSIG_ERROR_TRACE("removing \"" + dirName + "\"",
                     { fs::remove_all(dirName); })
  DIRSIG_ERROR_TRACE("creating \"" + dirName + "\"", {
    fs::create_directory(dirName);
    workdir_ = fs::current_path();
    workdir_ /= dirName;
  })

  // setup the list of extra MODTRAN files to purge after a run
  extraFilenames_ = {"tmp.mc", "tmp.psc", "tmp.tp8", "tmp.7sr"};

  Config c;
  DIRSIG_ERROR_TRACE("parsing the " + name() + " input JSON",
                     { input.get_to(c); })

  // check if a profile name was provided
  if (std::holds_alternative<std::string>(c.profile)) {
    // extract the profile name
    profile_ = GetModtranProfile(std::get<std::string>(c.profile));
  } else {
    profile_ = std::move(std::get<Profile>(c.profile));
  }
  exeArgs_ = std::move(c.modtranOptions);

  // check if the MODTRAN program exists
  DIRSIG_THROW_IF(!fs::exists(profile_.exeFilePath),
                  "Invalid path for MODTRAN executable: '" +
                      profile_.exeFilePath + "'");
  modtranExe_ = profile_.exeFilePath;

  // check if the MODTRAN DATA directory is valid
  DIRSIG_THROW_IF(!fs::is_directory(profile_.dataDirPath),
                  "Invalid path for MODTRAN data folder: '" +
                      profile_.dataDirPath + "'");

  // make sure we have a training slash
  if (auto c = profile_.dataDirPath.back(); c != '/' && c != '\\') {
    profile_.dataDirPath.push_back('/');
  }

  // Initialize the default json if one is given
  if (!std::holds_alternative<std::nullptr_t>(c.json)) {
    defaultTemplateJson_ = std::make_unique<Json>();
    loadJson(c.json, *defaultTemplateJson_);
    if (std::holds_alternative<JsonFilename>(c.json)) {
      defaultJsonFilename_ = std::get<JsonFilename>(c.json).filename;
    }
  }

  // check for the (optional) flag to extract profiles
  extractProfile_ = c.extractProfile;
}

ModtranJsonBackend::~ModtranJsonBackend() {
  // if we are not saving files, then remove the entire "modtran_runs"
  // folder
  try {
    if (!saveFiles_)
      fs::remove_all(workdir_);
  } catch (...) {
  }
}

void ModtranJsonBackend::printSummary(std::ostream &stream) const {
  stream << "ModtranJson Backend Configuration:\n";
  if (!defaultJsonFilename_.empty()) {
    stream << "    MODTRAN JSON = '" << defaultJsonFilename_ << "'\n";
  }
  stream << "    MODTRAN executable = '" << profile_.exeFilePath << "'\n"
         << "    MODTRAN DATA = '" << profile_.dataDirPath << "'\n";
  if (extractProfile_) {
    stream << "    Extracting atmospheric profile\n";
  }
  stream << std::endl;
}

/**
 *
 */
bool ModtranJsonBackend::test(std::ostream &stream) const {
  stream << "Verifying MODTRAN configuration:" << std::endl;
  stream << "    Checking if MODTRAN program path exists ... " << std::flush;
  if (!fs::exists(profile_.exeFilePath)) {
    stream << std::endl
           << "MODTRAN program file does not exist!" << std::endl
           << "    Path = '" << profile_.exeFilePath << "'" << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  stream << "    Checking if MODTRAN data path exists ... " << std::flush;
  if (!fs::exists(profile_.dataDirPath)) {
    stream << std::endl
           << "MODTRAN data directory does not exist!" << std::endl
           << "    Path = '" << profile_.dataDirPath << "'" << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  stream << "    Checking if MODTRAN data is a directory ... " << std::flush;
  if (!fs::is_directory(profile_.dataDirPath)) {
    stream << std::endl
           << "MODTRAN data path does not refer to a directory!" << std::endl
           << "    Path = '" << profile_.dataDirPath << "'" << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  stream << "    Reading a MODTRAN example JSON file ... " << std::flush;
  JsonFilename jsonPath;
  jsonPath.filename =
      d5::StandardPaths::DirsigHome() + "/lib/data/tape5/mls.json";
  d5::Json testJson;
  try {
    loadJson(jsonPath, testJson);
  } catch (...) {
    stream << std::endl
           << "Exception occured while reading the test input file!"
           << std::endl
           << "    File path = '" << jsonPath.filename << "'" << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  stream << "    Running a simple MODTRAN example ... " << std::flush;
  ModelRequest testRequest;
  testRequest.index = 0;
  testRequest.type = d5::atm::ModelRequestType::ExoSourceData;
  testRequest.options = d5::atm::ModelRequestOptions::SunSourceRequest;
  testRequest.dayOfYear = 182;
  testRequest.groundAltM = 0.0;
  testRequest.pathAlt1M = 0.0;
  testRequest.sourceZenithDeg = 10.0;
  testRequest.sourceAzimuthDeg = 180.0;
  static const float minWvl = 0.350;
  static const float maxWvl = 15.000;
  static const float deltaWvl = 0.050;
  size_t wvlCount = size_t((maxWvl - minWvl) / deltaWvl) + 1;
  for (size_t ww = 0; ww < wvlCount; ww++) {
    testRequest.wavelengthSamples.push_back((ww * deltaWvl) + minWvl);
  }
  setupJson(testRequest, testJson);
  std::string runPath = "test_run";
  try {
    fs::create_directory(runPath);
  } catch (const std::exception &e) {
    std::cerr << "An exception occured while creating run directory!"
              << std::endl
              << "    Directory = '" << runPath << "'" << std::endl
              << "    Exception = '" << e.what() << "'" << std::endl;
    return false;
  } catch (...) {
    std::cerr << std::endl
              << "An exception occured while trying to run MODTRAN!"
              << std::endl;
    return false;
  }

  // setup a simple exo-atm source run
  setupExoSource(runPath, testRequest, testJson);

  try {
    runModtran(runPath, testRequest);
  } catch (const std::exception &e) {
    std::cerr << std::endl
              << "An exception occured while trying to run MODTRAN!"
              << std::endl
              << "   Executable = '" << modtranExe_ << "'" << std::endl
              << "   Exception  = '" << e.what() << "'" << std::endl;
    return false;
  } catch (...) {
    std::cerr << std::endl
              << "An exception occured while trying to run MODTRAN!"
              << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  stream << "    Checking for MODTRAN output files ... " << std::flush;
  ModelResponse testResponse;
  try {
    readExoSourceResult(runPath, testRequest, testResponse);
  } catch (const std::exception &e) {
    std::cerr << "Exception while occured while reading MODTRAN output!"
              << std::endl
              << "   '" << e.what() << "'" << std::endl;
    return false;
  } catch (...) {
    std::cerr << "Exception while occured while reading MODTRAN output!"
              << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  // if we got this far, things look good
  fs::remove_all(runPath);
  return true;
}

/**
 * This function adds a spectral reflectance curve to a background
 * reflectance (aka "spectral albedo") database created in the run
 * directory.  The format of the MODTRAN spectral albedo file is
 * described in the MODTRAN documentation.
 */
bool ModtranJsonBackend::addSpectralReflectance(
    const std::string &name, const SpectralCurveData *reflData) {
  // if this data is already in the map, return
  if (reflDataMap_.find(reflData) != reflDataMap_.end()) {
    return true;
  }

  // scan the data to make sure its valid
  for (auto pt : *reflData) {
    if ((pt.value < 0) || (pt.value > 1.0)) {
      return false;
    }
  }

  // open the local spectral albedo file
  std::ofstream specAlbFile(workdir_.string() + "/spec_alb.txt",
                            std::ofstream::out | std::ofstream::app);
  if (specAlbFile.fail()) {
    return false;
  }

  // figure out if we have any curves we have already
  size_t curveIndex = reflDataMap_.size() + 1;

  // write the curve header (auto generated names for now)
  specAlbFile << curveIndex << " " << name << std::endl;

  // write the curve data, one wavelength/value pair per line
  for (auto pt : *reflData) {
    specAlbFile << pt.wavelength << " " << pt.value << std::endl;
  }

  if (specAlbFile.fail()) {
    return false;
  }

  specAlbFile.close();

  // add this curve to our map
  reflDataMap_[reflData] = curveIndex;

  return true;
}

/**
 * This function adds a relative spectral response function file
 * in the run directory.  The format of the MODTRAN instrument filter
 * function file is described in the MODTRAN documentation.
 */
bool ModtranJsonBackend::addSpectralResponse(
    const SpectralResponseData *respData) {
  // if this data is already in the map, return
  if (respDataMap_.find(respData) != respDataMap_.end()) {
    return true;
  }

  // figure out if we have any functions we have already
  size_t respIndex = respDataMap_.size() + 1;

#ifdef _ENABLE_RSP
  // open the local filter function file
  std::ofstream fltFile(workdir_.string() + "/rsr" + std::to_string(respIndex) +
                            ".txt",
                        std::ofstream::out);
  if (fltFile.fail()) {
    return false;
  }

  // write the header
  fltFile << "Microns: User-supplied instrument filter function" << std::endl;

  // write the curve data, one wavelength/value pair per line
  for (size_t ci = 0; ci < respData->size(); ci++) {
    fltFile << "Channel #" << ci + 1 << std::endl;
    auto &curve = respData->at(ci);
    for (auto pt : curve) {
      fltFile << pt.wavelength << " " << pt.value << std::endl;
    }
  }

  if (fltFile.fail()) {
    return false;
  }

  fltFile.close();
#endif

  // add this curve to our map
  respDataMap_[respData] = respIndex;

  return true;
}

void ModtranJsonBackend::loadJson(const JsonFilename &filename,
                                  d5::Json &json) const {
  DIRSIG_ERROR_TRACE("loading JSON file from \"" + filename.filename + "\"", {
    // open the external file
    auto fileStream = Ifstream(filename.filename);

    // read the contents
    json = d5::Json::parse(fileStream);
    json = json["MODTRAN"][0]["MODTRANINPUT"];
  })
}

void ModtranJsonBackend::loadJson(const JsonContents &contents,
                                  d5::Json &json) const {
  DIRSIG_ERROR_TRACE("reading user-provided JSON contents", {
    // make a stream to read from the content string
    std::istringstream contentsStream(contents.contents);

    // read the contents
    json = d5::Json::parse(contentsStream);
    json = json["MODTRAN"][0]["MODTRANINPUT"];
  })
}

void ModtranJsonBackend::setupJson(const ModelRequest &request,
                                   d5::Json &json) const {
  // set the name
  json["NAME"] = "tmp";

  // make sure we run in MODTRAN mode
  json["RTOPTIONS"]["MODTRN"] = "RT_MODTRAN";

  // explicitly set the path to the DATA directory path
  json["FILEOPTIONS"]["DATDIR"] = profile_.dataDirPath;

  // disable binary output
  json["FILEOPTIONS"]["BINARY"] = false;

  // disable profile generation
  json["FILEOPTIONS"]["NOPRNT"] = 1;

  // let MODTRAN select the band model file
  json["SPECTRAL"]["LBMNAM"] = "f";
  json["SPECTRAL"]["BMNAME"] = "";

  // disable channel scanning output
  json["SPECTRAL"]["FILTNM"] = "";

  // set the day (impacts Earth-Sun distance) and use the default earth radius
  json["GEOMETRY"]["IDAY"] = request.dayOfYear;
  json["GEOMETRY"]["RAD_E"] = 0.0;

  // setup for this spectral state
  auto delta =
      (request.wavelengthSamples.back() - request.wavelengthSamples.front()) /
      (request.wavelengthSamples.size() - 1.);
  json["SPECTRAL"]["V1"] = request.wavelengthSamples.front();
  json["SPECTRAL"]["V2"] = request.wavelengthSamples.back();
  json["SPECTRAL"]["DV"] = delta;
  json["SPECTRAL"]["FWHM"] = delta;
  json["SPECTRAL"]["FLAGS"] = "mraa   ";

  if (request.reflData != nullptr) {
    setupSpectralReflectance(request.reflData, json);
  }
#ifdef _ENABLE_RSP
  if (request.respData != nullptr) {
    setupSpectralResponse(request.respData, json);
  }
#endif
}

ModelResponse
ModtranJsonBackend::performRequest(const ModelRequest &request) const {
  // parse the input JSON
  RequestBackendInput in;
  DIRSIG_ERROR_TRACE("parsing MODTRAN backend input JSON",
                     { request.backendInput.get_to(in); })

  // make a copy of the template
  d5::Json json;
  if (!loadJson(in.json, json)) {
    if (defaultTemplateJson_) {
      json = *defaultTemplateJson_;
    } else {
      DIRSIG_THROW("No template JSON was specified in the request "
                   "input and no default was provided");
    }
  }

  // do the baseline setup for the request
  setupJson(request, json);

  auto runPath = PathCat(workdir_, std::to_string(request.index));
  try {
    fs::create_directory(runPath);
  } catch (const std::exception &e) {
    DIRSIG_THROW("unable to create directory \"" + runPath.string() +
                 "\": " + e.what());
  }

  ModelResponse response;

  try {
    switch (request.type) {
    case ModelRequestType::ProfileData:
      setupProfile(runPath, request, json);
      runModtran(runPath, request);
      readProfileResult(runPath, request, response);
      break;
    case ModelRequestType::ExoSourceData:
      setupExoSource(runPath, request, json);
      runModtran(runPath, request);
      readExoSourceResult(runPath, request, response);
      break;
    case ModelRequestType::PathData:
      if (request.pathLengthM == infiniteLength) {
        setupSkyPath(runPath, request, json);
        runModtran(runPath, request);
        readSkyPathResult(runPath, request, response);
      } else {
        setupGeneralPath(runPath, request, json);
        runModtran(runPath, request);
        readGeneralPathResult(runPath, request, response);
      }
      break;
    case ModelRequestType::SkyData:
      setupTotalSky(runPath, request, json);
      runModtran(runPath, request);
      readTotalSkyResult(runPath, request, response);
      break;
    case ModelRequestType::TotalData:
      setupTotalPath(runPath, request, json);
      runModtran(runPath, request);
      readTotalPathResult(runPath, request, response);
      break;
    default:
      break;
    }
  } catch (...) {
    // if something bad happened, make sure we save the output and then rethrow
    saveFiles_ = true;
    throw;
  }

  // cleanup the files used for this request
  cleanupRequest(runPath);

  return response;
}

/**
 * To setup a spectral reflectance, we find the curve in the database
 * we previously constructed via our map that resolves pointers to
 * indexes. Then we let the MODTRAN tape5 helper set it up.  MODTRAN
 * allows you to request a curve using either the name of index and we
 * are using the index for simplicity.  We use the short relative path
 * to the spectral albedo database file (previously generated) to avoid
 * overflowing the MODTRAN variable for the filename, which has a limited
 * length.
 */
void ModtranJsonBackend::setupSpectralReflectance(
    const SpectralCurveData *reflData, d5::Json &json) const {
  // find the index for this data in our map
  std::unordered_map<const SpectralCurveData *, size_t>::const_iterator
      findIter = reflDataMap_.find(reflData);
  DIRSIG_THROW_IF(
      findIter == reflDataMap_.end(),
      "Supplied spectral reflectance data was not previously added!");
#ifdef _FIXED
  json.useLambertianReflectance("../spec_alb.txt",
                                std::to_string(findIter->second));
#endif
}

/**
 * To setup a spectral response, we find the response in the database
 * we previously constructed via our map that resolves pointers to
 * indexes. Then we let the MODTRAN tape5 helper set it up.  MODTRAN
 * allows you to request a curve using either the name of index and we
 * are using the index for simplicity.  We use the short relative path
 * to the spectral filter function file (previously generated) to avoid
 * the MODTRAN variable for the filename has a limited length.
 */
void ModtranJsonBackend::setupSpectralResponse(
    const SpectralResponseData *respData, d5::Json &json) const {
  // find the index for this data in our map
  std::unordered_map<const SpectralResponseData *, size_t>::const_iterator
      findIter = respDataMap_.find(respData);
  DIRSIG_THROW_IF(findIter == respDataMap_.end(),
                  "Supplied spectral response data was not previously added!");
#ifdef _FIXED
  json.useFilterFunction("../rsr1.txt");
#endif
}

void ModtranJsonBackend::setupProfile(const fs::path &runPath,
                                      const ModelRequest &request,
                                      d5::Json &json) const {

  // use "directly transmitted solar/lunar irradiance" mode
  json["RTOPTIONS"]["IEMSCT"] = "RT_SOLAR_IRRADIANCE";

  // enable profile generation
  json["FILEOPTIONS"]["NOPRNT"] = 0;

  // set the ground altitude (convert to km)
  json["SURFACE"]["GNDALT"] = static_cast<float>(request.groundAltM / 1000.0);

  // use a "slant path to space or ground"
  json["GEOMETRY"]["ITYPE"] = 3;
  json["GEOMETRY"]["H1ALT"] = static_cast<float>(request.pathAlt1M / 1000.0);
  json["GEOMETRY"]["H2ALT"] = 0.0f;
  json["GEOMETRY"]["OBSZEN"] = 0.0f;
  json["GEOMETRY"]["ANGLEM"] = 0.0f;
  json["GEOMETRY"]["BETA"] = 0.0f;
  json["GEOMETRY"]["LENN"] = 0;
  json["GEOMETRY"]["RAD_E"] = 0.0f;
  json["GEOMETRY"]["BCKZEN"] = 0.0f;

  // save the file into the run folder
  writeJson(PathCat(runPath, "tmp.json").string(), json);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream << "Atmospheric profile run" << std::endl;
  writeReadme(runPath, readmeStream.str());
}

void ModtranJsonBackend::setupExoSource(const fs::path &runPath,
                                        const ModelRequest &request,
                                        d5::Json &json) const {
  // use "directly transmitted solar/lunar irradiance" mode
  std::string sourceName = "";
  if (request.options & ModelRequestOptions::SunSourceRequest) {
    json["RTOPTIONS"]["IEMSCT"] = "RT_SOLAR_IRRADIANCE";
    sourceName = "Sun";
  } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
    json["RTOPTIONS"]["IEMSCT"] = "RT_LUNAR_IRRADIANCE";
    sourceName = "Moon";
  } else {
    DIRSIG_THROW("ModtranJson: Invalid exo-atm source in request!");
  }

  // set the ground altitude (convert to km)
  json["SURFACE"]["GNDALT"] = static_cast<float>(request.groundAltM / 1000.0);

  // use a "slant path to space or ground"
  json["GEOMETRY"]["ITYPE"] = 3;
  json["GEOMETRY"]["H1ALT"] = static_cast<float>(request.pathAlt1M / 1000.0);
  json["GEOMETRY"]["H2ALT"] = 0.0;
  json["GEOMETRY"]["OBSZEN"] = static_cast<float>(request.sourceZenithDeg);
  json["GEOMETRY"]["ANGLEM"] = 0.0;

  // check if we need to "fake" the source angle
  if (request.sourceZenithDeg > 90.0) {
    json["GEOMETRY"]["OBSZEN"] = 90.0;
  }

  // save the file into the run folder
  writeJson(PathCat(runPath, "tmp.json").string(), json);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream << "Exo-atmospheric source run" << std::endl
               << "Day of year = " << request.dayOfYear << std::endl
               << "Ground altitude = " << request.groundAltM << " [meters]"
               << std::endl
               << "Path altitude = " << request.pathAlt1M << " [meters]"
               << std::endl
               << "Source is the " << sourceName << std::endl
               << "Source zenith = " << request.sourceZenithDeg << " [degrees]"
               << std::endl
               << "Source azimuth = " << request.sourceAzimuthDeg
               << " [degrees]" << std::endl;
  writeReadme(runPath, readmeStream.str());
}

void ModtranJsonBackend::setupSkyPath(const fs::path &runPath,
                                      const ModelRequest &request,
                                      d5::Json &json) const {
  // use a "slant path to space or ground"
  json["GEOMETRY"]["ITYPE"] = 3;

  // use "scattered solar/lunar radiance" mode
  std::string sourceName = "";
  if (request.options & ModelRequestOptions::SunSourceRequest) {
    json["RTOPTIONS"]["IEMSCT"] = "RT_SOLAR_AND_THERMAL";
    sourceName = "Sun";
  } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
    json["RTOPTIONS"]["IEMSCT"] = "RT_LUNAR_AND_THERMAL";
    sourceName = "Moon";
  } else {
    DIRSIG_THROW("ModtranJson: Invalid exo-atm source in request!");
  }

  // set the ground altitude (convert to km)
  json["SURFACE"]["GNDALT"] = static_cast<float>(request.groundAltM / 1000.0);

  // setup the basic geometry (convert to km)
  json["GEOMETRY"]["H1ALT"] = static_cast<float>(request.pathAlt1M / 1000.0);
  json["GEOMETRY"]["H2ALT"] = 0.0;
  json["GEOMETRY"]["OBSZEN"] = static_cast<float>(request.pathZenithDeg);
  json["GEOMETRY"]["HRANGE"] = 0.0;
  json["GEOMETRY"]["BETA"] = 0.0;
  json["GEOMETRY"]["LENN"] = 0;
  json["GEOMETRY"]["RAD_E"] = 0.0;
  json["GEOMETRY"]["BCKZEN"] = 0.0;

  // compute the relative azimuth between the path and source (-180:180)
  auto pathSourceAzimuth = request.pathAzimuthDeg - request.sourceAzimuthDeg;
  if (pathSourceAzimuth < -180.0)
    pathSourceAzimuth += 360.0;
  else if (pathSourceAzimuth > +180.0)
    pathSourceAzimuth -= 360.0;

  json["AEROSOLS"]["HGPF"] = 0.0;
  json["GEOMETRY"]["IPARM"] = 2;
  json["GEOMETRY"]["PARM1"] = static_cast<float>(pathSourceAzimuth);
  json["GEOMETRY"]["PARM2"] = static_cast<float>(request.sourceZenithDeg);
  json["GEOMETRY"]["PARM3"] = 0.0;
  json["GEOMETRY"]["PARM4"] = 0.0;
  json["GEOMETRY"]["GMTIME"] = 0.0;
  json["GEOMETRY"]["ANGLEM"] = 0.0;
  json["GEOMETRY"]["TRUEAZ"] = static_cast<float>(request.pathAzimuthDeg);

  // save the file into the run folder
  writeJson(PathCat(runPath, "tmp.json").string(), json);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream << "Sky path run" << std::endl
               << "Day of year = " << request.dayOfYear << std::endl
               << "Ground altitude = " << request.groundAltM << " [meters]"
               << std::endl
               << "Path altitude = " << request.pathAlt1M << " [meters]"
               << std::endl
               << "Path zenith = " << request.pathZenithDeg << " [degrees]"
               << std::endl
               << "Path azimuth = " << request.pathAzimuthDeg << " [degrees]"
               << std::endl
               << "Source is the " << sourceName << std::endl
               << "Source zenith = " << request.sourceZenithDeg << " [degrees]"
               << std::endl
               << "Source azimuth = " << request.sourceAzimuthDeg
               << " [degrees]" << std::endl;
  writeReadme(runPath, readmeStream.str());
}

void ModtranJsonBackend::setupGeneralPath(const fs::path &runPath,
                                          const ModelRequest &request,
                                          d5::Json &json) const {
  // use a "slant path to between two altitudes"
  json["GEOMETRY"]["ITYPE"] = 2;

  // use "scattered solar/lunar radiance" mode
  std::string sourceName = "";
  if (request.options & ModelRequestOptions::SunSourceRequest) {
    json["RTOPTIONS"]["IEMSCT"] = "RT_SOLAR_AND_THERMAL";
    sourceName = "Sun";
  } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
    json["RTOPTIONS"]["IEMSCT"] = "RT_LUNAR_AND_THERMAL";
    sourceName = "Moon";
  } else {
    DIRSIG_THROW("ModtranJson: Invalid exo-atm source in request!");
  }

  // convert the altitudes and ranges to kilometers and quantize to meters
  auto gndaltKm = floor(request.groundAltM * 1.0e+03) / 1.0e+06;
  auto h1Km = floor(request.pathAlt1M * 1.0e+03) / 1.0e+06;
  auto h2Km = floor(request.pathAlt2M * 1.0e+03) / 1.0e+06;
  auto rangeKm = floor(request.pathLengthM * 1.0e+03) / 1.0e+06;

  // set the ground altitude
  json["SURFACE"]["GNDALT"] = static_cast<float>(gndaltKm);

  // setup the path geometry (TOA in MODTRAN is 100 km)
  std::string pathType = "General path run ";
  std::string pathMethod;
  if (h1Km < 100.0) {
    // inside the atmosphere, use H1, BCKZEN and RANGE
    if (request.pathZenithDeg == 0) {
      json["GEOMETRY"]["H1ALT"] = static_cast<float>(h1Km);
      json["GEOMETRY"]["H2ALT"] =
          h1Km + std::cos(d5::toRadians(request.pathZenithDeg)) * rangeKm;
      json["GEOMETRY"]["HRANGE"] = 0.0f;
      json["GEOMETRY"]["OBSZEN"] = 0.0f;
      json["GEOMETRY"]["BCKZEN"] =
          static_cast<float>(180.0 - request.pathZenithDeg);
      pathType += "(inside atmosphere, straight up)";
      pathMethod = fmt::format(
          "ITYPE = 2, using H1 = {:.3f}, H2 = {:.3f} and BCKZEN = {:.3f}", h1Km,
          h2Km, 180.0 - request.pathZenithDeg);
    } else if (request.pathZenithDeg <= 90.0) {
      json["GEOMETRY"]["H1ALT"] = static_cast<float>(h1Km);
      json["GEOMETRY"]["H2ALT"] = 0.0f;
      json["GEOMETRY"]["HRANGE"] = static_cast<float>(rangeKm - 0.001);
      json["GEOMETRY"]["OBSZEN"] = static_cast<float>(request.pathZenithDeg);
      json["GEOMETRY"]["BCKZEN"] = 0.0f;
      pathType += "(inside atmosphere, uplooking)";
      pathMethod = fmt::format(
          "ITYPE = 2, using H1 = {:.3f}, RANGE = {:.3f} and OBSZEN = {:.3f}",
          h1Km, rangeKm, request.pathZenithDeg);
    } else {
      json["GEOMETRY"]["H1ALT"] = static_cast<float>(h1Km);
      json["GEOMETRY"]["H2ALT"] = 0.0f;
      json["GEOMETRY"]["HRANGE"] = static_cast<float>(rangeKm - 0.001);
      json["GEOMETRY"]["OBSZEN"] = 0.0f;
      json["GEOMETRY"]["BCKZEN"] =
          static_cast<float>(180.0 - request.pathZenithDeg);
      pathType += "(inside atmosphere, downlooking)";
      pathMethod = fmt::format(
          "ITYPE = 2, using H1 = {:.3f}, RANGE = {:.3f} and BCKZEN = {:.3f}",
          h1Km, rangeKm, 180.0 - request.pathZenithDeg);
    }
  } else {
    // above the atmosphere, use H1, H2 and BCKZEN (convert to km)
    json["GEOMETRY"]["H1ALT"] = static_cast<float>(request.pathAlt1M / 1000.0);
    json["GEOMETRY"]["H2ALT"] = static_cast<float>(request.pathAlt2M / 1000.0);
    json["GEOMETRY"]["BCKZEN"] =
        static_cast<float>(180 - request.pathZenithDeg);
    json["GEOMETRY"]["HRANGE"] = 0.0;
    pathType += "(above atmosphere, downlooking)";
    pathMethod = fmt::format(
        "ITYPE = 2, using H1 = {:.3f}, H2 = {:.3f} and BCKZEN = {:.3f}", h1Km,
        h2Km, 180.0 - request.pathZenithDeg);

    // make sure H2 is never below the ground
    if (request.pathAlt2M < request.groundAltM) {
      json["GEOMETRY"]["H2ALT"] = request.groundAltM / 1000.0;
    }
  }

  // if nadir, BCKZEN = 0 and we need to use OBSZEN
  if (request.pathZenithDeg == 180.0) {
    json["GEOMETRY"]["OBSZEN"] = request.pathZenithDeg;
    pathMethod += " (swapped BCKZEN -> OBSZEN)";
  }

  // fill in the rest of the geometry
  json["GEOMETRY"]["BETA"] = 0.0;
  json["GEOMETRY"]["LENN"] = 0;
  json["GEOMETRY"]["RAD_E"] = 0.0;

  // compute the relative azimuth between the path and source (-180:180)
  auto pathSourceAzimuth = request.pathAzimuthDeg - request.sourceAzimuthDeg;
  if (pathSourceAzimuth < -180.0) {
    pathSourceAzimuth += 360.0;
  } else if (pathSourceAzimuth > +180.0) {
    pathSourceAzimuth -= 360.0;
  }

  json["AEROSOLS"]["HGPF"] = 0.0;
  json["GEOMETRY"]["IPARM"] = 2;
  json["GEOMETRY"]["PARM1"] = static_cast<float>(pathSourceAzimuth);
  json["GEOMETRY"]["PARM2"] = static_cast<float>(request.sourceZenithDeg);
  json["GEOMETRY"]["PARM3"] = 0.0;
  json["GEOMETRY"]["PARM4"] = 0.0;
  json["GEOMETRY"]["GMTIME"] = 0.0;
  json["GEOMETRY"]["ANGLEM"] = 0.0;
  json["GEOMETRY"]["TRUEAZ"] = static_cast<float>(request.pathAzimuthDeg);

  // save the file into the run folder
  writeJson(PathCat(runPath, "tmp.json").string(), json);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream << pathType << std::endl
               << pathMethod << std::endl
               << "Day of year = " << request.dayOfYear << std::endl
               << "Ground altitude = " << request.groundAltM << " [meters]"
               << std::endl
               << "Path altitude = " << request.pathAlt1M << " [meters]"
               << std::endl
               << "Path range = " << request.pathLengthM << " [meters]"
               << std::endl
               << "Path zenith = " << request.pathZenithDeg << " [degrees]"
               << std::endl
               << "Path azimuth = " << request.pathAzimuthDeg << " [degrees]"
               << std::endl
               << "Source is the " << sourceName << std::endl
               << "Source zenith = " << request.sourceZenithDeg << " [degrees]"
               << std::endl
               << "Source azimuth = " << request.sourceAzimuthDeg
               << " [degrees]" << std::endl;
  writeReadme(runPath, readmeStream.str());
}

/**
 * To get the total sky irradiance from MODTRAN, the common trick is to
 * look down at a Lambertian 100% reflector and get the reflected radiance
 * from it.
 */
void ModtranJsonBackend::setupTotalSky(const fs::path &runPath,
                                       const ModelRequest &request,
                                       d5::Json &json) const {
  // use a "slant path to space or ground"
  json["GEOMETRY"]["ITYPE"] = 3;

  // use "scattered solar/lunar radiance" mode
  std::string sourceName = "";
  if (request.options & ModelRequestOptions::SunSourceRequest) {
    json["RTOPTIONS"]["IEMSCT"] = "RT_SOLAR_AND_THERMAL";
    sourceName = "Sun";
  } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
    json["RTOPTIONS"]["IEMSCT"] = "RT_LUNAR_AND_THERMAL";
    sourceName = "Moon";
  } else {
    DIRSIG_THROW("ModtranJson: Invalid exo-atm source in request!");
  }

  // set the ground altitude (convert to km) and surface reflectance to 1
  json["SURFACE"]["GNDALT"] = static_cast<float>(request.groundAltM / 1000.0);
  json["SURFACE"]["SURREF"] = 1.00;

  // setup the basic geometry (convert to km)
  json["GEOMETRY"]["H1ALT"] = 0.001;
  json["GEOMETRY"]["H2ALT"] = 0.000;
  json["GEOMETRY"]["OBSZEN"] = 180.0;
  json["GEOMETRY"]["HRANGE"] = 0.0;
  json["GEOMETRY"]["BETA"] = 0.0;
  json["GEOMETRY"]["LENN"] = 0;
  json["GEOMETRY"]["RAD_E"] = 0.0;
  json["GEOMETRY"]["BCKZEN"] = 0.0;

  // compute the relative azimuth between the path and source (-180:180)
  auto pathSourceAzimuth = request.pathAzimuthDeg - request.sourceAzimuthDeg;
  if (pathSourceAzimuth < -180.0)
    pathSourceAzimuth += 360.0;
  else if (pathSourceAzimuth > +180.0)
    pathSourceAzimuth -= 360.0;

  json["AEROSOLS"]["HGPF"] = 0.0;
  json["GEOMETRY"]["IPARM"] = 2;
  json["GEOMETRY"]["PARM1"] = static_cast<float>(pathSourceAzimuth);
  json["GEOMETRY"]["PARM2"] = static_cast<float>(request.sourceZenithDeg);
  json["GEOMETRY"]["PARM3"] = 0.0;
  json["GEOMETRY"]["PARM4"] = 0.0;
  json["GEOMETRY"]["GMTIME"] = 0.0;
  json["GEOMETRY"]["ANGLEM"] = 0.0;
  json["GEOMETRY"]["TRUEAZ"] = static_cast<float>(request.pathAzimuthDeg);

  // save the file into the run folder
  writeJson(PathCat(runPath, "tmp.json").string(), json);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream << "Total sky run" << std::endl
               << "Day of year = " << request.dayOfYear << std::endl
               << "Ground altitude = " << request.groundAltM << " [meters]"
               << std::endl
               << "Source is the " << sourceName << std::endl
               << "Source zenith = " << request.sourceZenithDeg << " [degrees]"
               << std::endl
               << "Source azimuth = " << request.sourceAzimuthDeg
               << " [degrees]" << std::endl;
  writeReadme(runPath, readmeStream.str());
}

/**
 * This is the most traditional MODTRAN setup.
 */
void ModtranJsonBackend::setupTotalPath(const fs::path &runPath,
                                        const ModelRequest &request,
                                        d5::Json &json) const {
  // use a "slant path to space or ground"
  json["GEOMETRY"]["ITYPE"] = 3;

  // use "scattered solar/lunar radiance" mode
  std::string sourceName = "";
  if (request.options & ModelRequestOptions::SunSourceRequest) {
    json["RTOPTIONS"]["IEMSCT"] = "RT_SOLAR_AND_THERMAL";
    sourceName = "Sun";
  } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
    json["RTOPTIONS"]["IEMSCT"] = "RT_LUNAR_AND_THERMAL";
    sourceName = "Moon";
  } else {
    DIRSIG_THROW("ModtranJson: Invalid exo-atm source in request!");
  }

  // set the ground altitude (convert to km)
  json["SURFACE"]["GNDALT"] = static_cast<float>(request.groundAltM / 1000.0);

  // setup the basic geometry (convert to km)
  json["GEOMETRY"]["H1ALT"] = static_cast<float>(request.pathAlt1M / 1000.0);
  json["GEOMETRY"]["H2ALT"] = 0.0;
  json["GEOMETRY"]["OBSZEN"] = 0.0;
  json["GEOMETRY"]["HRANGE"] =
      static_cast<float>((request.pathLengthM - 1) / 1000.0);
  json["GEOMETRY"]["BETA"] = 0.0;
  json["GEOMETRY"]["LENN"] = 0;
  json["GEOMETRY"]["RAD_E"] = 0.0;
  json["GEOMETRY"]["BCKZEN"] = static_cast<float>(180 - request.pathZenithDeg);

  // if nadir, BCKZEN = 0 and we need to use OBSZEN
  if (request.pathZenithDeg == 180.0) {
    json["GEOMETRY"]["OBSZEN"] = request.pathZenithDeg;
  } else {
    json["GEOMETRY"]["OBSZEN"] = 0.0;
  }

  // compute the relative azimuth between the path and source (-180:180)
  auto pathSourceAzimuth = request.pathAzimuthDeg - request.sourceAzimuthDeg;
  if (pathSourceAzimuth < -180.0)
    pathSourceAzimuth += 360.0;
  else if (pathSourceAzimuth > +180.0)
    pathSourceAzimuth -= 360.0;

  json["AEROSOLS"]["HGPF"] = 0.0;
  json["GEOMETRY"]["IPARM"] = 2;
  json["GEOMETRY"]["PARM1"] = static_cast<float>(pathSourceAzimuth);
  json["GEOMETRY"]["PARM2"] = static_cast<float>(request.sourceZenithDeg);
  json["GEOMETRY"]["PARM3"] = 0.0;
  json["GEOMETRY"]["PARM4"] = 0.0;
  json["GEOMETRY"]["GMTIME"] = 0.0;
  json["GEOMETRY"]["ANGLEM"] = 0.0;
  json["GEOMETRY"]["TRUEAZ"] = static_cast<float>(request.pathAzimuthDeg);

  // save the file into the run folder
  writeJson(PathCat(runPath, "tmp.json").string(), json);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream << "Total radiance run" << std::endl
               << "Day of year = " << request.dayOfYear << std::endl
               << "Ground altitude = " << request.groundAltM << " [meters]"
               << std::endl
               << "Path altitude = " << request.pathAlt1M << " [meters]"
               << std::endl
               << "Path range = " << request.pathLengthM << " [meters]"
               << std::endl
               << "Path zenith = " << request.pathZenithDeg << " [degrees]"
               << std::endl
               << "Path azimuth = " << request.pathAzimuthDeg << " [degrees]"
               << std::endl
               << "Source is the " << sourceName << std::endl
               << "Source zenith = " << request.sourceZenithDeg << " [degrees]"
               << std::endl
               << "Source azimuth = " << request.sourceAzimuthDeg
               << " [degrees]" << std::endl;
  writeReadme(runPath, readmeStream.str());
}

void ModtranJsonBackend::runModtran(const fs::path &runPath,
                                    const ModelRequest &request) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest)
    return;

#ifdef _DEBUG
  std::cout << "Running '" << runPath.string() << "'" << std::endl;
  std::cout << profile_.exeFilePath << " ";
  for (const auto &e : exeArgs_)
    std::cout << e << " ";
  std::cout << std::endl;
#endif

  // extend the arguments
  std::vector<std::string> args;
  args.push_back("tmp.json");
  args.push_back(profile_.dataDirPath);

  // setup and run the MODTRAN process
  auto outfile = PathCat(runPath, "modtran_stdout.txt"),
       errfile = PathCat(runPath, "modtran_stderr.txt");
  std::error_code ec;
  bp::child proc(modtranExe_.string(), bp::args(args), bp::std_out > outfile,
                 bp::std_err > errfile, bp::start_dir = runPath.string(), ec);

  // If we have an error here the setup of the exe failed
  DIRSIG_THROW_IF(ec, "unable to start '" + modtranExe_.string() + "': " +
                          std::to_string(ec.value()) + " - " + ec.message());

  proc.wait(ec);
}

void ModtranJsonBackend::writeReadme(const fs::path &runPath,
                                     const std::string &contents) const {
  DIRSIG_ERROR_TRACE("writing README to \"" + runPath.string() + "\"", {
    auto p = PathCat(runPath, "README.txt");
    auto ofs = Ofstream(p.string());
    ofs << contents;
  })
}

void ModtranJsonBackend::cleanupRequest(const fs::path &runPath) const try {
  // if we are saving files, just remove the extra files
  if (saveFiles_) {
    for (const auto &f : extraFilenames_)
      fs::remove(PathCat(runPath, f));
  } else {
    // otherwise, purge the whole request directory
    fs::remove_all(runPath);
  }
} catch (...) {
}

void ModtranJsonBackend::readProfileResult(const fs::path &runPath,
                                           const ModelRequest &request,
                                           ModelResponse &response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest)
    return;

  // make room for the profile data (100 altitudes x 4 values)
  response.profileTable.resize(400);

  // open the MODTRAN output "scanned" file
  auto tp6Path = PathCat(runPath, "tmp.tp6").string();
  std::ifstream tp6File;
  DIRSIG_ERROR_TRACE("opening MODTRAN output file \"" + tp6Path + "\"",
                     { tp6File = Ifstream(tp6Path); })

  // check if the file is empty
  DIRSIG_THROW_IF(StreamSize(tp6File) == 0,
                  "MODTRAN output file \"" + tp6Path + "\" is empty");

  // read line-by-line to the first profile table and note the format
  std::string format = "Unknown";
  std::string tapeLine;
  while (std::getline(tp6File, tapeLine)) {
    // check if this is the start of a profile table
    if (!tapeLine.compare(0, kFormat1ProfileHeader.size(),
                          kFormat1ProfileHeader)) {
      format = "Format1";
      break;
    } else if (!tapeLine.compare(0, kFormat2ProfileHeader.size(),
                                 kFormat2ProfileHeader)) {
      format = "Format2";
      break;
    }
  }

  // check if we didn't find the tables

  DIRSIG_ERROR_TRACE("parsing MODTRAN output file \"" + tp6Path + "\"", {
    DIRSIG_THROW_IF(format == "Unknown", "failed to find atmospheric profile");

    // grab what we need
    if (format == "Format1") {
      this->readProfileFormat1(tp6File, response);
    } else if (format == "Format2") {
      this->readProfileFormat2(tp6File, response);
    }
  })
}

/**
 * This function reads the 2nd atmospheric profile from a "tape6"
 * file generated by MODTRAN4.
 */
void ModtranJsonBackend::readProfileFormat1(std::istream &tapeStream,
                                            ModelResponse &response) const {
  // find the next table (has temp, pressure, humidity)
  std::string line;
  while (std::getline(tapeStream, line)) {
    // check if this is the start of a profile table
    if (!line.compare(0, kFormat1ProfileHeader.size(), kFormat1ProfileHeader)) {
      break;
    }
  }

  // read the blank line, the two header lines and the next blank line
  std::getline(tapeStream, line);
  std::getline(tapeStream, line);
  std::getline(tapeStream, line);
  std::getline(tapeStream, line);

  // read the whole table
  int altIndex = 0;
  std::vector<double> lineTokens;
  while (std::getline(tapeStream, line) && !line.empty()) {
    // make sure we don't have too many altitudes
    DIRSIG_THROW_IF(altIndex >= 100,
                    "too many altitudes in the atmospheric profile table");

    // split the line into the column tokens and check the count
    lineTokens.clear();
    double val;
    std::istringstream ss(line);
    while (ss >> val)
      lineTokens.push_back(val);
    DIRSIG_THROW_IF(lineTokens.size() != 14,
                    "row " + std::to_string(altIndex + 1) +
                        ": unexpected number of columns in the "
                        "atmospheric profile table (" +
                        std::to_string(lineTokens.size()) + ", expected 14)");

    // parse the columns and convert
    double altKm = lineTokens[1];
    double tempC = lineTokens[3] - 273.15;
    double pressureMb = lineTokens[2];
    double rh = lineTokens[10];
    response.profileTable[(altIndex * 4) + 0] = altKm;
    response.profileTable[(altIndex * 4) + 1] = tempC;
    response.profileTable[(altIndex * 4) + 2] = pressureMb;
    response.profileTable[(altIndex * 4) + 3] = Tdp(tempC, rh);

    altIndex++;
  }
}

/**
 * This function reads the 2nd atmospheric profile from a "tape6"
 * file generated by MODTRAN5 or MODTRAN6.
 */
void ModtranJsonBackend::readProfileFormat2(std::istream &tapeStream,
                                            ModelResponse &response) const {
  // find the next table
  std::string line;
  while (std::getline(tapeStream, line)) {
    // check if this is the start of a profile table
    if (!line.compare(0, kFormat2ProfileHeader.size(), kFormat2ProfileHeader)) {
      break;
    }
  }

  // read the blank line and the two header lines
  std::getline(tapeStream, line);
  std::getline(tapeStream, line);
  std::getline(tapeStream, line);

  // read the whole table
  int altIndex = 0;
  std::vector<double> lineTokens;
  while (std::getline(tapeStream, line) && !line.empty()) {
    // make sure we don't have too many altitudes
    DIRSIG_THROW_IF(altIndex >= 100,
                    "too many altitudes in the atmospheric profile table");

    // split the line into the column tokens and check the count
    lineTokens.clear();
    double val;
    std::istringstream ss(line);
    while (ss >> val)
      lineTokens.push_back(val);
    DIRSIG_THROW_IF(lineTokens.size() != 14,
                    "row " + std::to_string(altIndex + 1) +
                        ": unexpected number of columns in the "
                        "atmospheric profile table (" +
                        std::to_string(lineTokens.size()) + ", expected 14)");

    // parse the columns and convert
    double altKm = lineTokens[1];
    double tempC = lineTokens[3] - 273.15;
    double pressureMb = lineTokens[2];
    double rh = lineTokens[10];
    response.profileTable[(altIndex * 4) + 0] = altKm;
    response.profileTable[(altIndex * 4) + 1] = tempC;
    response.profileTable[(altIndex * 4) + 2] = pressureMb;
    response.profileTable[(altIndex * 4) + 3] = Tdp(tempC, rh);

    altIndex++;
  }
}

void ModtranJsonBackend::readExoSourceResult(const fs::path &runPath,
                                             const ModelRequest &request,
                                             ModelResponse &response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest)
    return;

  // open the MODTRAN tape7 output file
  auto tp7Path = PathCat(runPath, "tmp.7sc").string();
  if (request.respData != nullptr) {
    tp7Path = PathCat(runPath, "tmp.tp7").string();
  }
  std::ifstream tp7File;
  try {
    tp7File = Ifstream(tp7Path);
  } catch (const std::exception &e) {
    DIRSIG_THROW("ModtranJsonBackend::readExoSourceResult:\n" +
                 std::string("Unable to open MODTRAN output file '") + tp7Path +
                 "'\n" + e.what());
  }

  // check if the file is empty
  DIRSIG_THROW_IF(StreamSize(tp7File) == 0,
                  "MODTRAN output file \"" + tp7Path + "\" is empty");

#ifdef _DEBUG
  std::cout << "ReadExoSourceResult:" << std::endl
            << "Filename = " << tp7Path << std::endl;
#endif

  // read the tape7 header
  float h1, h2, angle, range, v1, v2, dv;
  DIRSIG_ERROR_TRACE(
      "reading header of MODTRAN output file \"" + tp7Path + "\"",
      { LegacyOutput::readHeader(tp7File, h1, h2, angle, range, v1, v2, dv); })
  size_t vCount = static_cast<size_t>((v2 - v1) / dv);

  if (request.respData == nullptr) {
    // read the result directly into the output
    response.buffer1.resize(request.wavelengthSamples.size());
    response.buffer2.resize(request.wavelengthSamples.size());
    LegacyOutput::readIrradianceData(tp7File, response.buffer1,
                                     response.buffer2);
  } else {
    // read the raw data into a temporary buffer
    std::vector<float> totalRad(vCount);
    std::vector<float> totalTau(vCount);
    LegacyOutput::readIrradianceData(tp7File, totalRad, totalTau);

    // integrate the raw data with the channels
    std::vector<ChannelResponse> channels;
    setupChannelResponses(v1, v2, dv, vCount, request.respData, channels);
    response.buffer1.resize(request.respData->size());
    response.buffer2.resize(request.respData->size());
    for (size_t ci = 0; ci < channels.size(); ci++) {
      response.buffer1[ci] = channels[ci].integrate(totalRad, true);
      response.buffer2[ci] = channels[ci].average(totalTau);
    }
  }

  DIRSIG_THROW_IF(!tp7File,
                  "read failed on MODTRAN output file \"" + tp7Path + "\"");

  // check if we need to "correct" the transmission
  if (request.sourceZenithDeg >= 90.0) {
    for (size_t ii = 0; ii < request.wavelengthSamples.size(); ii++) {
      response.buffer2[ii] = 0.0;
    }
  }
}

void ModtranJsonBackend::readSkyPathResult(const fs::path &runPath,
                                           const ModelRequest &request,
                                           ModelResponse &response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest)
    return;

  // open the MODTRAN tape7 output file
  auto tp7Path = PathCat(runPath, "tmp.7sc").string();
  if (request.respData != nullptr) {
    tp7Path = PathCat(runPath, "tmp.tp7").string();
  }
  std::ifstream tp7File;
  try {
    tp7File = Ifstream(tp7Path);
  } catch (const std::exception &e) {
    DIRSIG_THROW("ModtranJsonBackend::readSkyPathResult:\n" +
                 std::string("Unable to open MODTRAN output file '") + tp7Path +
                 "'\n" + e.what());
  }

  // check if the file is empty
  DIRSIG_THROW_IF(StreamSize(tp7File) == 0,
                  "MODTRAN output file '" + tp7Path + "' is empty");

#ifdef _DEBUG
  std::cout << "ReadSkyPathResult:" << std::endl
            << "Filename = " << tp7Path << std::endl;
#endif

  // read the tape7 header
  float h1, h2, angle, range, v1, v2, dv;
  DIRSIG_ERROR_TRACE(
      "reading header of MODTRAN output file \"" + tp7Path + "\"",
      { LegacyOutput::readHeader(tp7File, h1, h2, angle, range, v1, v2, dv); })
  size_t vCount = static_cast<size_t>((v2 - v1) / dv);

  // store the path length (convert to meters)
  response.range = range * 1000.0;

  if (request.respData == nullptr) {
    // read the result directly into the output
    response.buffer1.resize(request.wavelengthSamples.size());
    response.buffer2.resize(request.wavelengthSamples.size());
    LegacyOutput::readTotalRadianceData(tp7File, response.buffer1,
                                        response.buffer2);
  } else {
    // read the raw data into a temporary buffer
    std::vector<float> totalRad(vCount);
    std::vector<float> totalTau(vCount);
    LegacyOutput::readTotalRadianceData(tp7File, totalRad, totalTau);

    // integrate the raw data with the channels
    std::vector<ChannelResponse> channels;
    setupChannelResponses(v1, v2, dv, vCount, request.respData, channels);
    response.buffer1.resize(request.respData->size());
    response.buffer2.resize(request.respData->size());
    for (size_t ci = 0; ci < channels.size(); ci++) {
      response.buffer1[ci] = channels[ci].integrate(totalRad, true);
      response.buffer2[ci] = channels[ci].average(totalTau);
    }
  }

  DIRSIG_THROW_IF(!tp7File,
                  "read failed on MODTRAN output file \"" + tp7Path + "\"");
}

void ModtranJsonBackend::readGeneralPathResult(const fs::path &runPath,
                                               const ModelRequest &request,
                                               ModelResponse &response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest)
    return;

  // open the MODTRAN tape7 output file
  auto tp7Path = PathCat(runPath, "tmp.7sc").string();
  if (request.respData != nullptr) {
    tp7Path = PathCat(runPath, "tmp.tp7").string();
  }
  std::ifstream tp7File;
  try {
    tp7File = Ifstream(tp7Path);
  } catch (const std::exception &e) {
    DIRSIG_THROW("ModtranJsonBackend::readGeneralPathResult:\n" +
                 std::string("Unable to open MODTRAN output file '") + tp7Path +
                 "'\n" + e.what());
  }

  // check if the file is empty
  DIRSIG_THROW_IF(StreamSize(tp7File) == 0,
                  "MODTRAN output file \"" + tp7Path + "\" is empty");

#ifdef _DEBUG
  std::cout << "ReadGeneralPathResult:" << std::endl
            << "Filename = " << tp7Path << std::endl;
#endif

  // read the tape7 header
  float h1, h2, angle, range, v1, v2, dv;
  DIRSIG_ERROR_TRACE(
      "reading header of MODTRAN output file \"" + tp7Path + "\"",
      { LegacyOutput::readHeader(tp7File, h1, h2, angle, range, v1, v2, dv); })
  size_t vCount = static_cast<size_t>((v2 - v1) / dv);

  // if we started above the atmosphere, don't let MODTRAN reduce the range
  auto origRangeKm = request.pathLengthM / 1000.0;
  if ((request.pathAlt1M > 100.0e+03) && (range < origRangeKm)) {
    range = origRangeKm;
  }

  // store the path length (convert to meters)
  response.range = range * 1000.0;

  if (request.respData == nullptr) {
    // read the result directly into the output
    response.buffer1.resize(request.wavelengthSamples.size());
    response.buffer2.resize(request.wavelengthSamples.size());
    LegacyOutput::readPathRadianceData(tp7File, response.buffer1,
                                       response.buffer2);
  } else {
    // read the raw data into a temporary buffer
    std::vector<float> totalRad(vCount);
    std::vector<float> totalTau(vCount);
    LegacyOutput::readPathRadianceData(tp7File, totalRad, totalTau);

    // integrate the raw data with the channels
    std::vector<ChannelResponse> channels;
    setupChannelResponses(v1, v2, dv, vCount, request.respData, channels);
    response.buffer1.resize(request.respData->size());
    response.buffer2.resize(request.respData->size());
    for (size_t ci = 0; ci < channels.size(); ci++) {
      response.buffer1[ci] = channels[ci].integrate(totalRad, true);
      response.buffer2[ci] = channels[ci].average(totalTau);
    }
  }

  DIRSIG_THROW_IF(!tp7File,
                  "read failed on MODTRAN output file \"" + tp7Path + "\"");

  // normalize and convert the data
  for (size_t ii = 0; ii < response.buffer1.size(); ii++) {
    response.buffer1[ii] = response.buffer1[ii] / response.range;
    if (response.buffer2[ii] > 0.0) {
      response.buffer2[ii] = -log(response.buffer2[ii]) / response.range;
    } else {
      response.buffer2[ii] = infiniteExt;
    }
  }
}

void ModtranJsonBackend::readTotalSkyResult(const fs::path &runPath,
                                            const ModelRequest &request,
                                            ModelResponse &response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest)
    return;

  // open the MODTRAN tape7 output file
  auto tp7Path = PathCat(runPath, "tmp.7sc").string();
  if (request.respData != nullptr) {
    tp7Path = PathCat(runPath, "tmp.tp7").string();
  }
  std::ifstream tp7File;
  try {
    tp7File = Ifstream(tp7Path);
  } catch (const std::exception &e) {
    DIRSIG_THROW("ModtranJsonBackend::readTotalSkyResult:\n" +
                 std::string("Unable to open MODTRAN output file '") + tp7Path +
                 "'\n" + e.what());
  }

  // check if the file is empty
  DIRSIG_THROW_IF(StreamSize(tp7File) == 0,
                  "MODTRAN output file \"" + tp7Path + "\" is empty");

#ifdef _DEBUG
  std::cout << "ReadTotalSkyResult:" << std::endl
            << "Filename = " << tp7Path << std::endl;
#endif

  // read the tape7 header
  float h1, h2, angle, range, v1, v2, dv;
  DIRSIG_ERROR_TRACE(
      "reading header of MODTRAN output file \"" + tp7Path + "\"",
      { LegacyOutput::readHeader(tp7File, h1, h2, angle, range, v1, v2, dv); })
  size_t vCount = static_cast<size_t>((v2 - v1) / dv);

  // store the path length (convert to meters)
  response.range = range * 1000.0;

  if (request.respData == nullptr) {
    // read the result directly into the output
    response.buffer1.resize(request.wavelengthSamples.size());
    response.buffer2.resize(request.wavelengthSamples.size());
    LegacyOutput::readTotalRadianceData(tp7File, response.buffer1,
                                        response.buffer2);
  } else {
    // read the raw data into a temporary buffer
    std::vector<float> totalRad(vCount);
    std::vector<float> totalTau(vCount);
    LegacyOutput::readTotalRadianceData(tp7File, totalRad, totalTau);

    // integrate the raw data with the channels
    std::vector<ChannelResponse> channels;
    setupChannelResponses(v1, v2, dv, vCount, request.respData, channels);
    response.buffer1.resize(request.respData->size());
    response.buffer2.resize(request.respData->size());
    for (size_t ci = 0; ci < channels.size(); ci++) {
      response.buffer1[ci] = channels[ci].integrate(totalRad, true);
      response.buffer2[ci] = channels[ci].average(totalTau);
    }
  }

  DIRSIG_THROW_IF(!tp7File,
                  "read failed on MODTRAN output file \"" + tp7Path + "\"");
}

/**
 * This is the reader for a "total" (L-shaped path) run in MODTRAN.
 */
void ModtranJsonBackend::readTotalPathResult(const fs::path &runPath,
                                             const ModelRequest &request,
                                             ModelResponse &response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest)
    return;

  // open the MODTRAN tape7 output file
  auto tp7Path = PathCat(runPath, "tmp.7sc").string();
  if (request.respData != nullptr) {
    tp7Path = PathCat(runPath, "tmp.tp7").string();
  }
  std::ifstream tp7File;
  try {
    tp7File = Ifstream(tp7Path);
  } catch (const std::exception &e) {
    DIRSIG_THROW("ModtranJsonBackend::readTotalPathResult:\n" +
                 std::string("Unable to open MODTRAN output file '") + tp7Path +
                 "'\n" + e.what());
  }

  // check if the file is empty
  DIRSIG_THROW_IF(StreamSize(tp7File) == 0,
                  "MODTRAN output file \"" + tp7Path + "\" is empty");

#ifdef _DEBUG
  std::cout << "ReadTotalPathResult:" << std::endl
            << "Filename = " << tp7Path << std::endl;
#endif

  // read the tape7 header
  float h1, h2, angle, range, v1, v2, dv;
  DIRSIG_ERROR_TRACE(
      "reading header of MODTRAN output file \"" + tp7Path + "\"",
      { LegacyOutput::readHeader(tp7File, h1, h2, angle, range, v1, v2, dv); })
  size_t vCount = static_cast<size_t>((v2 - v1) / dv);

  if (request.respData == nullptr) {
    // read the result directly into the output
    response.buffer1.resize(request.wavelengthSamples.size());
    response.buffer2.resize(request.wavelengthSamples.size());
    LegacyOutput::readSensorAndSurfaceData(tp7File, response.buffer1,
                                           response.buffer2);
  } else {
    // read the raw data into a temporary buffer
    std::vector<float> sensorRad(vCount);
    std::vector<float> surfaceRad(vCount);
    LegacyOutput::readSensorAndSurfaceData(tp7File, sensorRad, surfaceRad);

    // integrate the raw data with the channels
    std::vector<ChannelResponse> channels;
    setupChannelResponses(v1, v2, dv, vCount, request.respData, channels);
    response.buffer1.resize(request.respData->size());
    response.buffer2.resize(request.respData->size());
    for (size_t ci = 0; ci < channels.size(); ci++) {
      response.buffer1[ci] = channels[ci].integrate(sensorRad, true);
      response.buffer2[ci] = channels[ci].integrate(surfaceRad, true);
    }
  }

  DIRSIG_THROW_IF(!tp7File,
                  "read failed on MODTRAN output file \"" + tp7Path + "\"");
}

void ModtranJsonBackend::setupChannelResponses(
    float v1, float v2, float dv, size_t vCount,
    const SpectralResponseData *respData,
    std::vector<ChannelResponse> &channels) const {
  channels.resize(respData->size());

  for (size_t ci = 0; ci < respData->size(); ci++) {
    auto &rsp = respData->at(ci);
    float firstFreq = 10000.0 / rsp.back().wavelength;
    firstFreq = std::min(std::max(firstFreq, v1), v2);
    float lastFreq = 10000.0 / rsp.front().wavelength;
    lastFreq = std::min(std::max(lastFreq, v1), v2);
    size_t freqCount = static_cast<size_t>((lastFreq - firstFreq) / dv);

    //
    // The interpolator needs the old and new data to be on the same
    // spectral units.  But we are interpolating to wavenumbers, which
    // are also going to be in decreasing order as wavelengths. So we
    // compute the target wavelengths in increasing order, perform the
    // interpolation and then reverse them.
    //
    std::vector<float> oldWavelengths(rsp.size());
    std::vector<float> oldData(rsp.size());
    for (size_t ii = 0; ii < rsp.size(); ii++) {
      oldWavelengths[ii] = rsp[ii].wavelength;
      oldData[ii] = rsp[ii].value;
    }
    std::vector<float> newWavelengths(freqCount);
    std::vector<float> newWidths(freqCount);
    for (size_t ii = 0; ii < freqCount; ii++) {
      float centerFreq = lastFreq - (ii * dv);
      newWavelengths[ii] = 10000.0 / centerFreq;
      float lowerWave = 10000.0 / (centerFreq + dv / 2);
      float upperWave = 10000.0 / (centerFreq - dv / 2);
      newWidths[ii] = upperWave - lowerWave;
    }

    std::vector<float> newData(freqCount);
    d5::interpolate(oldWavelengths, oldData, newWavelengths, newData);

    // create the channel response (reverse the weights)
    ChannelResponse &channel = channels[ci];
    channel.startIndex = static_cast<long>((firstFreq - v1) / dv);
    channel.stopIndex = static_cast<long>((lastFreq - v1) / dv);
    channel.pointCount = channel.stopIndex - channel.startIndex;
    channel.weights.resize(newData.size());
    std::reverse_copy(std::begin(newData), std::end(newData),
                      std::begin(channel.weights));
    channel.widths.resize(newWidths.size());
    std::reverse_copy(std::begin(newWidths), std::end(newWidths),
                      std::begin(channel.widths));

#ifdef _DEBUG
    std::cout << "firstFreq = " << firstFreq << std::endl;
    std::cout << "lastFreq  = " << lastFreq << std::endl;
    std::cout << "freqCount = " << freqCount << std::endl;
    std::cout << "Channel #" << ci << std::endl
              << "    Start index = " << channel.startIndex << std::endl
              << "    Start freq = " << (v1 + channel.startIndex * dv)
              << std::endl
              << "    Stop index = " << channel.stopIndex << std::endl
              << "    Stop freq = " << (v1 + channel.stopIndex * dv)
              << std::endl
              << "    Freq count = " << freqCount << std::endl;
    for (size_t ii = 0; ii < freqCount; ii++) {
      std::cout << 10000.0 / newWavelengths[ii] << " " << newWavelengths[ii]
                << " " << newData[ii] << std::endl;
    }
#endif
  }
}

bool ModtranJsonBackend::writeJson(const std::string &filename,
                                   d5::Json &json) {
  d5::Json runJson;
  runJson["MODTRANINPUT"] = json;

  d5::Json runArray;
  runArray.push_back(runJson);

  d5::Json outputJson;
  outputJson["MODTRAN"] = runArray;

  auto out = Ofstream(filename);
  out << std::setw(4) << outputJson << std::endl;

  return true;
}

} // namespace d5::atm::modtran::json
