#pragma once

#include <iostream>
#include <unordered_map>
#include <vector>

#include <filesystem>

#include "dirsig5/atmosphere/Atmosphere.h"
#include "dirsig5/atmosphere/backends/ModtranCommon.h"
#include "dirsig5/atmosphere/backends/ModtranJson/ModtranJson.h"

namespace d5::atm::modtran::json {

/// Implementation of an atmospheric model backend using MODTRAN with tape files
class ModtranJsonBackend : public ModelBackend {
public:
  /** Constructor
   *
   *  \param input  JSON encoding of a MODTRAN configuration
   */
  explicit ModtranJsonBackend(const nlohmann::json& input);

  /// Destructor
  ~ModtranJsonBackend() override;

  /// \override
  [[nodiscard]] const std::string& name() const override {
    static const std::string kName = "ModtranJson";
    return kName;
  }

  /**
   * The test function for this backend confirms that the user provided
   * a viable MODTRAN configuration and runs a test case to confirm the
   * correct operation.
   * \override
   */
  [[nodiscard]] bool test(std::ostream& stream) const override;

  /**
   * This backend supports the ability to supply a spectral reflectance.
   * \override
   */
  bool addSpectralReflectance(const std::string& name,
      const SpectralCurveData* reflData) override;

  /**
   * This backend supports the ability to supply a relative spectral response.
   * \override
   */
  bool addSpectralResponse(const SpectralResponseData* respData) override;

  /// \override
  void printSummary(std::ostream& stream) const override;

  /// \override
  [[nodiscard]] ModelResponse performRequest(
      const ModelRequest& request) const override;

  /**
   * The MODTRAN backend has an internal flag that controls
   * atmospheric profile generation.
   */
  [[nodiscard]] bool generatesProfileData() const override {
    return extractProfile_;
  }

private:
  /// Load an external "template" JSON file
  void loadJson(const JsonFilename& filename, d5::Json& json) const;

  /// Load a "template" JSON from the string contents
  void loadJson(const JsonContents& contents, d5::Json& json) const;

  /// Load a "tape5" template
  bool loadJson(const JsonVariant& in, d5::Json& json) const {
    return std::visit(
        [&](auto&& arg) {
          if constexpr (!std::is_same_v<
                            std::decay_t<decltype(arg)>, std::nullptr_t>) {
            this->loadJson(arg, json);
            return true;
          }
          return false;
        },
        in);
  }

  /// Do the baseline setup of the JSON for the request
  void setupJson(const ModelRequest& request, d5::Json& json) const;

  /// Add the lambertian reflectance setup
  void setupSpectralReflectance(
      const SpectralCurveData* reflData, d5::Json& json) const;

  /// Add the spectral response setup
  void setupSpectralResponse(
      const SpectralResponseData* respData, d5::Json& json) const;

  /**
   * Setup an atmospheric profile run.
   */
  void setupProfile(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      d5::Json& json) const;

  /**
   * Setup an exo-atmospheric source irradiance run.
   */
  void setupExoSource(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      d5::Json& json) const;

  /**
   * Setup an radiance and transmission along a path to the sky run.
   */
  void setupSkyPath(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      d5::Json& json) const;

  /**
   * Setup an radiance and transmission along non-sky paths run.
   */
  void setupGeneralPath(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      d5::Json& json) const;

  /**
   * Setup a total sky irradiance run.
   */
  void setupTotalSky(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      d5::Json& json) const;

  /**
   * Setup a total path radiance run.
   */
  void setupTotalPath(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      d5::Json& json) const;

  /**
   * Run MODTRAN
   */
  void runModtran(
      const std::filesystem::path& runPath,
      const ModelRequest& _request) const;

  /**
   * Write the contents of a README file
   */
  void writeReadme(
      const std::filesystem::path& runPath,
      const std::string& contents) const;

  /**
   * Cleanup/remove the directory created for the model request.
   */
  void cleanupRequest(const std::filesystem::path& runPath) const;

  /**
   * Read the result of a profile run.
   */
  void readProfileResult(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      ModelResponse& response) const;

  void readProfileFormat1(
      std::istream& tapeStream, ModelResponse& response) const;

  void readProfileFormat2(
      std::istream& tapeStream, ModelResponse& response) const;

  /**
   * Read the result of an exo-atmospheric source run.
   */
  void readExoSourceResult(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      ModelResponse& response) const;

  /**
   * Read the result of a sky path run.
   */
  void readSkyPathResult(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      ModelResponse& response) const;

  /**
   * Read the result of a general path run.
   */
  void readGeneralPathResult(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      ModelResponse& response) const;

  /**
   * Read the result of a total sky run.
   */
  void readTotalSkyResult(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      ModelResponse& response) const;

  /**
   * Read the result of a total radiance run.
   */
  void readTotalPathResult(
      const std::filesystem::path& runPath,
      const ModelRequest& request,
      ModelResponse& response) const;

private:
  /**
   * A type to store a channel.
   */
  struct ChannelResponse {
    bool normalize = false;
    long startIndex = 0l;
    long stopIndex = 0l;
    long pointCount = 0l;
    std::vector<float> weights = {};
    std::vector<float> widths = {};

    /**
     * Compute the spectrally weighted integral of the input data.  The
     * integrral can be normalized by the area of the weighting function,
     * which would produce spectral radiance (e.g. radiance per spectral unit.
     */
    float integrate(const std::vector<float>& data, bool normalize = false) {
      float result = 0.f;
      float normWeight = 0.f;
      for (size_t ii = 0; ii < this->weights.size(); ii++) {
#ifdef _DEBUG
        std::cout << ii << " " << data[this->startIndex + ii] << " "
                  << this->weights[ii] << std::endl;
#endif
        result +=
            data[this->startIndex + ii] * this->weights[ii] / this->widths[ii];
        normWeight += this->weights[ii];
      }
      if (normalize) {
        result /= normWeight;
      }
      return result;
    }

    /**
     * Compute the spectrally weighted average of the input data.
     */
    float average(const std::vector<float>& data) {
      float result = 0.f;
      float normWeight = 0.f;
      for (size_t ii = 0; ii < this->weights.size(); ii++) {
        result += data[this->startIndex + ii] * this->weights[ii];
        normWeight += this->weights[ii];
      }
      result /= normWeight;
      return result;
    }
  };

  /**
   * This function creates a set of channel weighting functions that can be
   * applied to MODTRAN output that is in wavenumbers (native resolution).
   */
  void setupChannelResponses(
      float v1,
      float v2,
      float dv,
      size_t vCount,
      const SpectralResponseData* respData,
      std::vector<ChannelResponse>& channels) const;

  /**
   * Write the JSON input file.
   */
  static bool writeJson(const std::string& filename, d5::Json& json);

  /** The input template JSON file MODTRAN */
  std::string defaultJsonFilename_;

  /** The default template MODTRAN input file */
  std::unique_ptr<d5::Json> defaultTemplateJson_;

  /** The current (working) folder for the MODTRAN runs */
  std::filesystem::path workdir_;

  /** The MODTRAN profile to use */
  Profile profile_;

  /** Path to the MODTRAN executable */
  std::filesystem::path modtranExe_;

  /** The arguments for the MODTRAN executable */
  std::vector<std::string> exeArgs_;

  /** The flag to enable/disable atmosphere profile extraction */
  bool extractProfile_ = false;

  /** The extra files produced by MODTRAN to remove */
  std::vector<std::string> extraFilenames_;

  /** The map to resolve reflectance curves */
  std::unordered_map<const SpectralCurveData*, size_t> reflDataMap_;

  /** The map to resolve response functions */
  std::unordered_map<const SpectralResponseData*, size_t> respDataMap_;
};
}
