#include "OutputHelper.h"

#include <cassert>
#include <cmath>
#include <fstream>
#include <iostream>
#include <limits>
#include <sstream>

#ifdef WIN32
#include <stdio.h>
#endif

void LegacyOutput::readHeader(std::istream &tapeStream, float &h1, float &h2,
                              float &angle, float &range, float &v1, float &v2,
                              float &dv) {
  // read the 11 line header
  std::string line;
  for (size_t lineIndex = 1; lineIndex <= 11; lineIndex++) {
    if (!std::getline(tapeStream, line)) {
      throw std::runtime_error("unexpected end of file while reading header");
    }

    // if this is the 6th line, this is card 3 and we want to extract stuff
    if (lineIndex == 6) {
      std::istringstream cardStream(line);
      cardStream >> h1 >> h2 >> angle >> range;
      continue;
    }

    // if this is the 9th line, this is card 4 and we want to extract stuff
    if (lineIndex == 9) {
      std::istringstream cardStream(line);
      cardStream >> v1 >> v2 >> dv;
      continue;
    }
  }
}

void LegacyOutput::readIrradianceData(std::istream &tapeStream,
                                      std::vector<float> &totalIrrad,
                                      std::vector<float> &totalTau) {
  // read the rest of the file, which should be nothing but data
  double lambdaValue, irradValue, tauValue, skipValue;
  std::string line;
  for (size_t ii = 0; ii < totalIrrad.size(); ii++) {
    std::getline(tapeStream, line);
    std::istringstream lineStream(line);
    lineStream >> lambdaValue >> tauValue >> skipValue >> irradValue;

#ifdef _DEBUG
    std::cout << ii << ", " << lambdaValue << ", " << irradValue << ", "
              << tauValue << std::endl;
#endif

    // store the values in the response
    totalIrrad[ii] = irradValue;
    totalTau[ii] = tauValue;
  }
}

/**
 * Summary of columns in a "tape7" file in radiance mode
 * 0  -> Frequency/Wavelength
 * 1  -> Total transmission
 * 2  -> Path thermal
 * 3  -> Thermal scattered (empty with DISORT)
 * 4  -> Surface emission
 * 5  -> Total solar scattered (single + multiple)
 * 6  -> Single solar scattered
 * 7  -> Total ground reflected (includes direct and diffuse)
 * 8  -> Direct reflected (direct only)
 * 9  -> Total radiance
 * For just the path radiance data, we want the "path thermal" and "solar
 * scattered" (combines single and multiple):
 */
void LegacyOutput::readPathRadianceData(std::istream &tapeStream,
                                        std::vector<float> &pathRad,
                                        std::vector<float> &pathTau) {
  // read the rest of the file, which should be nothing but data
  double lambdaValue, trans, thermalPath, thermalScat, surfEmis;
  double solarScat, singleScat, grndRefl, drctRefl, totalRad;
  std::string line;
  for (size_t ii = 0; ii < pathRad.size(); ii++) {
    std::getline(tapeStream, line);
    std::istringstream lineStream(line);
    auto pos = lineStream.tellg();
    lineStream >> lambdaValue >> trans >> thermalPath >> thermalScat;

    // correct if THRML_SCT was a missing (MODTRAN4/MODTRAN5 w/ DISORT)
    if (lineStream.tellg() - pos > 50) {
      surfEmis = thermalScat;
      thermalScat = 0.0;
    } else {
      lineStream >> surfEmis;
    }
    lineStream >> solarScat >> singleScat >> grndRefl >> drctRefl >> totalRad;

#ifdef _DEBUG
    std::cout << ii << std::fixed << ", " << lambdaValue << ", " << std::fixed
              << trans << ", " << std::scientific << thermalPath << ", "
              << std::scientific << thermalScat << ", " << std::scientific
              << solarScat << ", " << std::scientific << singleScat << ", "
              << std::scientific << totalRad << ", " << std::endl;
#endif
    pathRad[ii] = thermalPath + solarScat;
    pathTau[ii] = trans;

    // address SOL_SCAT vs. MULT_SCAT + SING_SCAT changes
    if (solarScat < singleScat) {
      pathRad[ii] += singleScat;
    }
  }
}

void LegacyOutput::readTotalRadianceData(std::istream &tapeStream,
                                         std::vector<float> &totalRad,
                                         std::vector<float> &totalTau) {
  // read the rest of the file, which should be nothing but data
  double lambdaValue, radValue, tauValue, skipValue;
  std::string line;
  for (size_t ii = 0; ii < totalRad.size(); ii++) {
    std::getline(tapeStream, line);
    std::istringstream lineStream(line);
    auto pos = lineStream.tellg();
    lineStream >> lambdaValue >> tauValue >> skipValue >> skipValue;

    // check if there was a missing column (MODTRAN4/MODTRAN5 w/ DISORT)
    if (lineStream.tellg() - pos < 50) {
      lineStream >> skipValue;
    }
    lineStream >> skipValue >> skipValue >> skipValue >> skipValue;
    lineStream >> radValue;

#ifdef _DEBUG
    std::cout << std::fixed << ii << ", " << lambdaValue << ", "
              << std::scientific << radValue << ", " << tauValue << std::endl;
#endif

    totalRad[ii] = radValue;
    totalTau[ii] = tauValue;
  }
}

void LegacyOutput::readSensorAndSurfaceData(std::istream &tapeStream,
                                            std::vector<float> &sensorRad,
                                            std::vector<float> &surfaceRad) {
  // read the rest of the file, which should be nothing but data
  float lambdaValue, totalTrans, pathThermal, scatThermal, surfEmis, singScat,
      multScat, grndRefl, drctRefl, totalRad;
  std::string line;
  for (size_t ii = 0; ii < sensorRad.size(); ii++) {
    std::getline(tapeStream, line);
    std::istringstream lineStream(line);
    auto pos = lineStream.tellg();
    lineStream >> lambdaValue >> totalTrans >> pathThermal >> scatThermal;

    // if there was a missing column (MODTRAN4/MODTRAN5 w/ DISORT)
    if (lineStream.tellg() - pos < 50) {
      lineStream >> surfEmis;
    } else {
      // "THRML_SCT" was missing, we read "SURF_EMIS" instead
      surfEmis = scatThermal;
      scatThermal = 0.0;
    }
    lineStream >> singScat >> multScat >> grndRefl >> drctRefl >> totalRad;

    // build up the surface leaving radiance
    double surfRad = grndRefl + surfEmis;

#ifdef _DEBUG
    std::cout << std::fixed << ii << ", " << lambdaValue << ", "
              << std::scientific << totalRad << ", " << surfRad << std::endl;
#endif

    sensorRad[ii] = totalRad;
    surfaceRad[ii] = surfRad;
  }
}
