#include "ModtranTapeBackend.h"

#include <atomic>
#include <fstream>
#include <iostream>
#include <limits>
#include <mutex>
#include <sstream>

#include <spdlog/fmt/fmt.h>

#include <boost/process.hpp>
namespace bp = boost::process;

#include <filesystem>
namespace fs = std::filesystem;

#include "../Common.h"
#include "dirsig5/atmosphere/backends/ModtranTape/ModtranTape.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/StandardPaths.h"

static std::atomic<int> s_dirSeqNum = 0;
static const std::string kFormat1ProfileHeader = "1 ATMOSPHERIC PROFILES";
static const std::string kFormat2ProfileHeader = " ATMOSPHERIC PROFILES";

namespace {

template <typename Arg, typename... Rest>
fs::path PathCatHelper(fs::path& base, Arg&& arg, Rest&&... rest) {
  base /= arg;
  return PathCat(base, std::forward<Rest>(rest)...);
}

template <typename Arg>
fs::path PathCatHelper(fs::path& base, Arg&& arg) {
  base /= arg;
  return base;
}

template <typename... Args>
fs::path PathCat(const fs::path& base, Args&&... args) {
  auto p = base;
  return PathCatHelper(p, std::forward<Args>(args)...);
}

std::ifstream Ifstream(const std::string& f) {
  std::ifstream ifs(f);
  if (!ifs) {
    std::string errMsg;
    errMsg  = "Unable to open input file!\n";
    errMsg += "Filename = '" + f + "'";
    throw std::runtime_error(errMsg);
  }
  return ifs;
}

std::ofstream Ofstream(const std::string& f) {
  std::ofstream ofs(f);
  if (!ofs) {
    std::string errMsg;
    errMsg  = "Unable to open output file!\n";
    errMsg += "Filename = '" + f + "'";
    throw std::runtime_error(errMsg);
  }
  return ofs;
}

auto StreamSize(std::istream& is) {
  auto curPos = is.tellg();
  is.seekg(0, std::ios::beg);
  auto fsize = is.tellg();
  is.seekg(0, std::ios::end);
  fsize = is.tellg() - fsize;
  is.seekg(curPos, std::ios::beg);
  return fsize;
}

// constants for a dew point calculation (Magnus formula)
// all based on Arden Buck (1981) for ranges above and below 0 deg C
static const double Tdp_b_pos = 17.368;  // [unitless]
static const double Tdp_c_pos = 238.88;  // [unitless]
static const double Tdp_b_neg = 17.966;  // [unitless]
static const double Tdp_c_neg = 247.15;  // [unitless]

// functions for calculating dew point [deg C] from RH [%]
double Tdp_gamma(double Tc, double RH) {
  return (Tc >= 0) ? log(RH / 100.) + Tdp_b_pos * Tc / (Tdp_c_pos + Tc)
                   : log(RH / 100.) + Tdp_b_neg * Tc / (Tdp_c_neg + Tc);
}

double Tdp(double Tc, double RH) {
  const double gamma = Tdp_gamma(Tc, RH);
  return (Tc >= 0) ? Tdp_c_pos * gamma / (Tdp_b_pos - gamma)
                   : Tdp_c_neg * gamma / (Tdp_b_neg - gamma);
}
}

namespace d5::atm::modtran::tape {

void from_json(const nlohmann::json& j, Config& c) {
  if (auto it = j.find("modtran_profile"); it != j.end()) {
    c.profile = it.value();
  } else {
    Profile p;
    try {
      p.exeFilePath = ResolvePath(j.at("modtran_exe").get<std::string>());
      p.dataDirPath = ResolvePath(j.at("modtran_data").get<std::string>());
      c.profile = p;
    } catch (const std::exception& e) {
      DIRSIG_THROW("Error resolving MODTRAN executable and data path!\n"
        + std::string(e.what()));
    }
    if (p.exeFilePath.empty() || p.dataDirPath.empty()) {
      DIRSIG_THROW("If 'modtran_profile' is not given, 'modtran_exe' and "
                   "'modtran_data' must be non-empty strings");
    }
  }

  if (auto it = j.find("modtran_options"); it != j.end()) {
    it.value().get_to(c.modtranOptions);
  }
  else {
    c.modtranOptions.push_back("-modroot");
  }

  if (auto it = j.find("tape5_contents"); it != j.end()) {
    Tape5Contents tp5;
    it.value().get_to(tp5.contents);
    c.tape5 = std::move(tp5);
  } else if (it = j.find("tape5_filename"); it != j.end()) {
    Tape5Filename tp5;
    it.value().get_to(tp5.filename);
    tp5.filename = ResolvePath(tp5.filename);
    c.tape5 = std::move(tp5);
  } else if (it = j.find("tape5_parameters"); it != j.end()) {
    c.tape5 = it.value().get<Tape5Parameters>();
  } else {
    c.tape5 = nullptr;
  }

  // check for the (optional) flag to extract profiles
  if (auto it = j.find("extract_profile"); it != j.end()) {
    it.value().get_to(c.extractProfile);
  }
}

void from_json(const nlohmann::json& j, RequestBackendInput& v) {
  if (auto it = j.find("tape5_contents"); it != j.end()) {
    Tape5Contents tp5;
    it.value().get_to(tp5.contents);
    v.tape5 = std::move(tp5);
  } else if (it = j.find("tape5_filename"); it != j.end()) {
    Tape5Filename tp5;
    it.value().get_to(tp5.filename);
    tp5.filename = ResolvePath(tp5.filename);
    v.tape5 = std::move(tp5);
  } else if (auto it = j.find("tape5_parameters"); it != j.end()) {
    Tape5Parameters tp5;
    it.value().get_to(tp5);
    v.tape5 = std::move(tp5);
  } else {
    v.tape5 = nullptr;
  }
}

static const double infiniteLength = std::numeric_limits<double>::max();
static const float infiniteExt = std::numeric_limits<float>::max();

ModtranTapeBackend::ModtranTapeBackend(const nlohmann::json& input) {
  // create an empty folder for the MODTRAN runs
  std::string dirName = "modtran_runs" + std::to_string(s_dirSeqNum++);
  DIRSIG_ERROR_TRACE("removing \"" + dirName + "\"", {
    fs::remove_all(dirName);
  })
  DIRSIG_ERROR_TRACE("creating \"" + dirName + "\"", {
    fs::create_directory(dirName);
    workdir_ = fs::current_path();
    workdir_ /= dirName;
  })

  // setup the list of extra MODTRAN files to purge after a run
  extraFilenames_ = {"tmp.mc", "tmp.psc", "tmp.tp8", "tmp.7sr"};

  Config c;
  DIRSIG_ERROR_TRACE(
      "parsing the " + name() + " input JSON", { input.get_to(c); })

  // check if a profile name was provided
  if (std::holds_alternative<std::string>(c.profile)) {
    // extract the profile name
    profile_ = GetModtranProfile(std::get<std::string>(c.profile));
  } else {
    profile_ = std::move(std::get<Profile>(c.profile));
  }
  exeArgs_ = std::move(c.modtranOptions);

  // check if the MODTRAN program exists
  DIRSIG_THROW_IF( !fs::exists(profile_.exeFilePath),
      "Invalid path for MODTRAN executable: '" + profile_.exeFilePath + "'");
  modtranExe_ = profile_.exeFilePath;

  // check if the MODTRAN DATA directory is valid
  DIRSIG_THROW_IF( !fs::is_directory(profile_.dataDirPath),
      "Invalid path for MODTRAN data folder: '" + profile_.dataDirPath + "'");

  // make sure we have a training slash
  if (auto c = profile_.dataDirPath.back(); c != '/' && c != '\\') {
    profile_.dataDirPath.push_back('/');
  }

  // Initialize the default tape5 if one is given
  if (!std::holds_alternative<std::nullptr_t>(c.tape5)) {
    defaultTemplateTape5_ = std::make_unique<Tape5>();
    loadTape5(c.tape5, *defaultTemplateTape5_);
    if (std::holds_alternative<Tape5Filename>(c.tape5)) {
      defaultTape5Filename_ = std::get<Tape5Filename>(c.tape5).filename;
    }
  }

  // check for the (optional) flag to extract profiles
  extractProfile_ = c.extractProfile;
}

ModtranTapeBackend::~ModtranTapeBackend() {
  // if we are not saving files, then remove the entire "modtran_runs"
  // folder
  try {
    if (!saveFiles_) fs::remove_all(workdir_);
  } catch (...) {
  }
}

void ModtranTapeBackend::printSummary(std::ostream& stream) const {
  stream << "ModtranTape Backend Configuration:\n";
  if (!defaultTape5Filename_.empty()) {
    stream << "    MODTRAN tape5 = '" << defaultTape5Filename_ << "'\n";
  }
  stream << "    MODTRAN executable = '" << profile_.exeFilePath << "'\n"
         << "    MODTRAN DATA = '" << profile_.dataDirPath << "'\n";
  if (extractProfile_) stream << "    Extracting atmospheric profile\n";
  stream << std::endl;
}

/**
 *
 */
bool ModtranTapeBackend::test(std::ostream& stream) const {
  stream << "Verifying MODTRAN configuration:" << std::endl;
  stream << "    Checking if MODTRAN program path exists ... " << std::flush;
  if (!fs::exists(profile_.exeFilePath)) {
    stream << std::endl
           << "MODTRAN program file does not exist!" << std::endl
           << "    Path = '" << profile_.exeFilePath << "'" << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  stream << "    Checking if MODTRAN data path exists ... " << std::flush;
  if (!fs::exists(profile_.dataDirPath)) {
    stream << std::endl
           << "MODTRAN data directory does not exist!" << std::endl
           << "    Path = '" << profile_.dataDirPath << "'" << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  stream << "    Checking if MODTRAN data is a directory ... " << std::flush;
  if (!fs::is_directory(profile_.dataDirPath)) {
    stream << std::endl
           << "MODTRAN data path does not refer to a directory!" << std::endl
           << "    Path = '" << profile_.dataDirPath << "'" << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  stream << "    Reading a MODTRAN example tape5 file ... " << std::flush;
  Tape5Filename tape5Path;
  Tape5 testTape5;
  tape5Path.filename =
      d5::StandardPaths::DirsigHome() + "/lib/data/tape5/mls.tp5";
  try {
    loadTape5(tape5Path, testTape5);
  } catch (...) {
    stream << std::endl
           << "Exception occured while reading the test input file!" << std::endl
           << "    File path = '" << tape5Path.filename << "'" << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  stream << "    Running a simple MODTRAN example ... " << std::flush;
  ModelRequest testRequest;
  testRequest.index = 0;
  testRequest.type = d5::atm::ModelRequestType::ExoSourceData;
  testRequest.options = d5::atm::ModelRequestOptions::SunSourceRequest;
  testRequest.dayOfYear = 182;
  testRequest.groundAltM = 0.0;
  testRequest.pathAlt1M = 0.0;
  testRequest.sourceZenithDeg = 10.0;
  testRequest.sourceAzimuthDeg = 180.0;
  static const float minWvl =   0.350;
  static const float maxWvl =  15.000;
  static const float deltaWvl = 0.050;
  size_t wvlCount = size_t((maxWvl - minWvl) / deltaWvl)+1;
  for( size_t ww = 0; ww < wvlCount; ww++ ) {
    testRequest.wavelengthSamples.push_back((ww * deltaWvl) + minWvl);
  }
  setupTape5(testRequest, testTape5);
  std::string runPath = "test_run";
  try {
    fs::create_directory(runPath);
  } catch (const std::exception& e) {
    std::cerr
        << "Exception occured while creating run directory!" << std::endl
        << "    Directory = '" << runPath << "'" << std::endl
        << "    Exception = '" << e.what() << "'" << std::endl;
    return false;
  }
  setupExoSource(runPath, testRequest, testTape5);
  try {
    runModtran(runPath, testRequest);
  }
  catch (const std::exception& e) {
    std::cerr
        << "Exception occured while trying to run MODTRAN!" << std::endl
        << "   '" << e.what() << "'" << std::endl;
    return false;
  }
  catch( ... ) {
    std::cerr
        << "Exception occured while trying to run MODTRAN!" << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  stream << "    Checking for MODTRAN output files ... " << std::flush;
  ModelResponse testResponse;
  try {
    readExoSourceResult(runPath, testRequest, testResponse);
  }
  catch (const std::exception& e) {
    std::cerr << std::endl
        << "Exception while occured while reading MODTRAN output!" << std::endl
        << e.what() << std::endl;
    return false;
  }
  catch( ... ) {
    std::cerr << std::endl
        << "Exception while occured while reading MODTRAN output!" << std::endl;
    return false;
  }
  stream << "passed!" << std::endl;

  // if we got this far, things look good
  fs::remove_all(runPath);
  return true;
}

/**
 * This function adds a spectral reflectance curve to a background
 * reflectance (aka "spectral albedo") database created in the run
 * directory.  The format of the MODTRAN spectral albedo file is
 * described in the MODTRAN documentation.
 */
bool ModtranTapeBackend::addSpectralReflectance(
    const std::string& name,
    const SpectralCurveData* reflData) {
  // if this data is already in the map, return
  if (reflDataMap_.find(reflData) != reflDataMap_.end()) {
    return true;
  }

  // scan the data to make sure its valid
  for (auto pt : *reflData) {
    if ((pt.value < 0) || (pt.value > 1.0)) {
      return false;
    }
  }

  // open the local spectral albedo file
  std::ofstream specAlbFile(
      workdir_.string() + "/spec_alb.txt",
      std::ofstream::out | std::ofstream::app);
  if (specAlbFile.fail()) {
    return false;
  }

  // figure out if we have any curves we have already
  size_t curveIndex = reflDataMap_.size() + 1;

  // write the curve header (auto generated names for now)
  specAlbFile << curveIndex << " " << name << std::endl;

  // write the curve data, one wavelength/value pair per line
  for (auto pt : *reflData) {
    specAlbFile << pt.wavelength << " " << pt.value << std::endl;
  }

  if (specAlbFile.fail()) {
    return false;
  }

  specAlbFile.close();

  // add this curve to our map
  reflDataMap_[reflData] = curveIndex;

  return true;
}

/**
 * This function adds a relative spectral response function file
 * in the run directory.  The format of the MODTRAN instrument filter
 * function file is described in the MODTRAN documentation.
 */
bool ModtranTapeBackend::addSpectralResponse(
    const SpectralResponseData* respData) {
  // if this data is already in the map, return
  if (respDataMap_.find(respData) != respDataMap_.end()) {
    return true;
  }

  // figure out if we have any functions we have already
  size_t respIndex = respDataMap_.size() + 1;

  // open the local filter function file
  std::ofstream fltFile(
      workdir_.string() + "/rsr" + std::to_string(respIndex) + ".txt",
      std::ofstream::out);
  if (fltFile.fail()) {
    return false;
  }

  // write the header
  fltFile << "Microns: User-supplied instrument filter function" << std::endl;

  // write the curve data, one wavelength/value pair per line
  for (size_t ci = 0; ci < respData->size(); ci++) {
    fltFile << "Channel #" << ci + 1 << std::endl;
    auto& curve = respData->at(ci);
    for (auto pt : curve) {
      fltFile << pt.wavelength << " " << pt.value << std::endl;
    }
  }

  // make sure all the writes succeeded
  if (fltFile.fail()) {
    return false;
  }

  // close the file
  fltFile.close();

  // add this curve to our map
  respDataMap_[respData] = respIndex;

  return true;
}

void ModtranTapeBackend::loadTape5(
    const Tape5Filename& filename, Tape5& tape5) const {
  DIRSIG_ERROR_TRACE("loading tape5 file from \"" + filename.filename + "\"", {
    // open the external file
    auto tape5File = Ifstream(filename.filename);

    // the parser is not thread safe, so we add a mutex guard here
    static std::mutex readMutex;
    std::lock_guard readGuard(readMutex);

    // read the file
    tape5.read(tape5File);
  })
}

void ModtranTapeBackend::loadTape5(
    const Tape5Contents& contents, Tape5& tape5) const {
  DIRSIG_ERROR_TRACE("reading user-provided tape5 contents", {
    // make a stream to read from the content string
    std::istringstream contentsStream(contents.contents);

    // read the contents
    tape5.read(contentsStream);
  })
}

void ModtranTapeBackend::loadTape5(
  const Tape5Parameters& params, Tape5& tape5)
    const {
  DIRSIG_ERROR_TRACE( "loading user-provided tape5 parameters", {
    tape5.fromParameters(params);
  })
}

void ModtranTapeBackend::setupTape5(const ModelRequest& request,
    Tape5& tape5) const
{
  // make sure we run in MODTRAN mode
  tape5.card1.modtrn = 'm';

  // disable binary output
  tape5.card1.cbinout = ' ';

  // disable channel scanning output
  tape5.card1a.lfltnm = 'f';

  // explicitly set the path to the DATA directory path
  tape5.card1a.cdtdir = 't';
  strncpy(tape5.card1a4.datdir, profile_.dataDirPath.c_str(), 80);
  tape5.card1a4.datdir[profile_.dataDirPath.size()] = '\0';

  // set the day (impacts Earth-Sun distance) and use the default earth radius
  tape5.card3alt.iday = request.dayOfYear;
  tape5.card3alt.ro = 0.0;
  tape5.card3a1.iday = request.dayOfYear;

  // setup the tape5 for this spectral state
  auto delta =
      (request.wavelengthSamples.back() - request.wavelengthSamples.front()) /
      (request.wavelengthSamples.size() - 1.);
  tape5.card4.v1 = request.wavelengthSamples.front();
  tape5.card4.v2 = request.wavelengthSamples.back();
  tape5.card4.dv = delta;
  tape5.card4.fwhm = delta;
  tape5.card4.flags[0] = 'm';
  tape5.card4.flags[1] = 'r';
  tape5.card4.flags[2] = 'a';
  tape5.card4.flags[3] = 'a';
  tape5.card5.irpt = 0;

  if (request.reflData != nullptr) {
    setupSpectralReflectance(request.reflData, tape5);
  }
#ifdef _ENABLE_RSP
  if (request.respData != nullptr) {
    setupSpectralResponse(request.respData, tape5);
  }
#endif
}

ModelResponse ModtranTapeBackend::performRequest(const ModelRequest& request) const {
  // Parse the input JSON
  RequestBackendInput in;
  DIRSIG_ERROR_TRACE("parsing MODTRAN tape backend input JSON", {
    request.backendInput.get_to(in);
  })

  // make a copy of the template tape5
  Tape5 tape5;
  if (!loadTape5(in.tape5, tape5)) {
    if (defaultTemplateTape5_) {
      tape5 = *defaultTemplateTape5_;
    } else {
      DIRSIG_THROW("No template tape5 was specified in the request "
                   "input and no default was provided");
    }
  }

  // do the baseline tape5 setup for the request
  setupTape5(request, tape5);

  auto runPath = PathCat(workdir_, std::to_string(request.index));
  try {
    fs::create_directory(runPath);
  } catch (const std::exception& e) {
    DIRSIG_THROW(
        "unable to create directory \"" + runPath.string() + "\": " + e.what());
  }

  ModelResponse response;

  try {
    switch (request.type) {
      case ModelRequestType::ProfileData:
        setupProfile(runPath, request, tape5);
        runModtran(runPath, request);
        readProfileResult(runPath, request, response);
        break;
      case ModelRequestType::ExoSourceData:
        setupExoSource(runPath, request, tape5);
        runModtran(runPath, request);
        readExoSourceResult(runPath, request, response);
        break;
      case ModelRequestType::PathData:
        if (request.pathLengthM == infiniteLength) {
          setupSkyPath(runPath, request, tape5);
          runModtran(runPath, request);
          readSkyPathResult(runPath, request, response);
        } else {
          setupGeneralPath(runPath, request, tape5);
          runModtran(runPath, request);
          readGeneralPathResult(runPath, request, response);
        }
        break;
      case ModelRequestType::SkyData:
        setupTotalSky(runPath, request, tape5);
        runModtran(runPath, request);
        readTotalSkyResult(runPath, request, response);
        break;
      case ModelRequestType::TotalData:
        setupTotalPath(runPath, request, tape5);
        runModtran(runPath, request);
        readTotalPathResult(runPath, request, response);
        break;
      default:
        break;
    }
  } catch (...) {
    // if something bad happened, make sure we save the output and then rethrow
    saveFiles_ = true;
    throw;
  }

  // cleanup the files used for this request
  cleanupRequest(runPath);

  return response;
}

/**
 * To setup a spectral reflectance, we find the curve in the database
 * we previously constructed via our map that resolves pointers to
 * indexes. Then we let the MODTRAN tape5 helper set it up.  MODTRAN
 * allows you to request a curve using either the name of index and we
 * are using the index for simplicity.  We use the short relative path
 * to the spectral albedo database file (previously generated) to avoid
 * overflowing the MODTRAN variable for the filename, which has a limited
 * length.
 */
void ModtranTapeBackend::setupSpectralReflectance(
    const SpectralCurveData* reflData, Tape5& tape5) const {
  // find the index for this data in our map
  std::unordered_map<const SpectralCurveData*, size_t>::const_iterator
      findIter = reflDataMap_.find(reflData);
  DIRSIG_THROW_IF(
      findIter == reflDataMap_.end(),
      "Supplied spectral reflectance data was not previously added!");

  tape5.useLambertianReflectance(
      "../spec_alb.txt", std::to_string(findIter->second));
}

/**
 * To setup a spectral response, we find the response in the database
 * we previously constructed via our map that resolves pointers to
 * indexes. Then we let the MODTRAN tape5 helper set it up.  MODTRAN
 * allows you to request a curve using either the name of index and we
 * are using the index for simplicity.  We use the short relative path
 * to the spectral filter function file (previously generated) to avoid
 * the MODTRAN variable for the filename has a limited length.
 */
void ModtranTapeBackend::setupSpectralResponse(
    const SpectralResponseData* respData, Tape5& tape5) const {
  // find the index for this data in our map
  std::unordered_map<const SpectralResponseData*, size_t>::const_iterator
      findIter = respDataMap_.find(respData);
  DIRSIG_THROW_IF(
      findIter == respDataMap_.end(),
      "Supplied spectral response data was not previously added!");
  tape5.useFilterFunction("../rsr1.txt");
}

void ModtranTapeBackend::setupProfile(
    const fs::path& runPath, const ModelRequest& request, Tape5& tape5) const {
  // use a "slant path to space or ground"
  tape5.card1.itype = 3;

  // use "directly transmitted solar/lunar irradiance" mode
  tape5.card1.iemsct = 3;

  // we need to make sure the profiles are output
  tape5.card1.noprt = 0;

  // setup the basic geometry (convert to km)
  tape5.card2.gndalt = static_cast<float>(request.groundAltM / 1000.0);
  tape5.card3alt.h1 = static_cast<float>(request.pathAlt1M / 1000.0);
  tape5.card3alt.h2 = 0.0f;
  tape5.card3alt.isourc = 0;
  tape5.card3alt.angle = 0.0f;
  tape5.card3alt.anglem = 0.0f;

  // save the tape5 file into the run folder
  tape5.write(PathCat(runPath, "tmp.tp5").string(), Tape5::QuietIO);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream << "Atmospheric profile run" << std::endl;
  writeReadme(runPath, readmeStream.str());
}

void ModtranTapeBackend::setupExoSource(
    const fs::path& runPath, const ModelRequest& request, Tape5& tape5) const {
  // use a "slant path to space or ground"
  tape5.card1.itype = 3;

  // use "directly transmitted solar/lunar irradiance" mode
  tape5.card1.iemsct = 3;

  // we don't need the profiles to be generated
  tape5.card1.noprt = 1;

  // figure out what source we are using
  std::string sourceName = "";
  if (request.options & ModelRequestOptions::SunSourceRequest) {
    sourceName = "Sun";
    tape5.card3alt.isourc = 0;
  } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
    sourceName = "Moon";
    tape5.card3alt.isourc = 1;
  } else {
    DIRSIG_THROW("ModtranTape: Invalid exo-atm source in request!");
  }

  // setup the basic geometry (convert to km)
  tape5.card2.gndalt = static_cast<float>(request.groundAltM / 1000.0);
  tape5.card3alt.h1 = static_cast<float>(request.pathAlt1M / 1000.0);
  tape5.card3alt.h2 = 0.0f;
  tape5.card3alt.angle = static_cast<float>(request.sourceZenithDeg);
  tape5.card3alt.anglem = 0.0f;

  // check if we need to "fake" the source angle
  if (tape5.card3alt.angle > 90.0) {
    tape5.card3alt.angle = 90.0f;
  }

  // save the tape5 file into the run folder
  tape5.write(PathCat(runPath, "tmp.tp5").string(), Tape5::QuietIO);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream
      << "Exo-atmospheric source run" << std::endl
      << "Day of year = " << request.dayOfYear << std::endl
      << "Ground altitude = " << request.groundAltM << " [meters]" << std::endl
      << "Path altitude = " << request.pathAlt1M << " [meters]" << std::endl
      << "Source is the " << sourceName << std::endl
      << "Source zenith = " << request.sourceZenithDeg << " [degrees]"
          << std::endl
      << "Source azimuth = " << request.sourceAzimuthDeg << " [degrees]"
          << std::endl;
  writeReadme(runPath, readmeStream.str());
}

void ModtranTapeBackend::setupSkyPath(
    const fs::path& runPath, const ModelRequest& request, Tape5& tape5) const {
  // use a "slant path to space or ground"
  tape5.card1.itype = 3;

  // use "scattered solar/lunar radiance" mode
  tape5.card1.iemsct = 2;

  // we don't need the profiles to be generated
  tape5.card1.noprt = 1;

  // setup the basic geometry (convert to km)
  std::string pathMethod = "ITYPE = 3, using H1 and OBSZEN";
  tape5.card2.gndalt = static_cast<float>(request.groundAltM / 1000.0);
  tape5.card3.h1 = static_cast<float>(request.pathAlt1M / 1000.0);
  tape5.card3.h2 = 0.0;
  tape5.card3.obszen = static_cast<float>(request.pathZenithDeg);
  tape5.card3.range = 0.0;
  tape5.card3.beta = 0.0;
  tape5.card3.lenn = 0.0;
  tape5.card3.ro = 0.0;
  tape5.card3.bckzen = 0.0;

  // uplooking case, so if we are doing multiple scattering do it at H1
  if (tape5.card1.imult != 0) {
    tape5.card1.imult = 1;
  }

  // compute the relative azimuth between the path and source (-180:180)
  auto pathSourceAzimuth = request.pathAzimuthDeg - request.sourceAzimuthDeg;
  if (pathSourceAzimuth < -180.0) {
    pathSourceAzimuth += 360.0;
  } else if (pathSourceAzimuth > +180.0) {
    pathSourceAzimuth -= 360.0;
  }

  // figure out what source we are using
  std::string sourceName = "";
  if (request.options & ModelRequestOptions::SunSourceRequest) {
    sourceName = "Sun";
    tape5.card3a1.isourc = 0;
  } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
    sourceName = "Moon";
    tape5.card3a1.isourc = 1;
  } else {
    DIRSIG_THROW("ModtranTape: Invalid exo-atm source in request!");
  }

  tape5.card3a1.iparm = 2;
  tape5.card3a2.parm1 = static_cast<float>(pathSourceAzimuth);
  tape5.card3a2.parm2 = static_cast<float>(request.sourceZenithDeg);
  tape5.card3a2.parm3 = 0;
  tape5.card3a2.parm4 = 0;
  tape5.card3a2.time = 0.0;
  tape5.card3a2.anglem = 0.0;
  tape5.card3a2.g = 0.0;
  tape5.card3a2.psipo = static_cast<float>(request.pathAzimuthDeg);

  // save the tape5 file into the run folder
  tape5.write(PathCat(runPath, "tmp.tp5").string(), Tape5::QuietIO);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream
      << "Sky path run" << std::endl
      << pathMethod << std::endl
      << "Day of year = " << request.dayOfYear << std::endl
      << "Ground altitude = " << request.groundAltM << " [meters]" << std::endl
      << "Path altitude = " << request.pathAlt1M << " [meters]" << std::endl
      << "Path zenith = " << request.pathZenithDeg << " [degrees]" << std::endl
      << "Path azimuth = " << request.pathAzimuthDeg << " [degrees]"
          << std::endl
      << "Source is the " << sourceName << std::endl
      << "Source zenith = " << request.sourceZenithDeg << " [degrees]"
          << std::endl
      << "Source azimuth = " << request.sourceAzimuthDeg << " [degrees]"
          << std::endl;
  writeReadme(runPath, readmeStream.str());
}

void ModtranTapeBackend::setupGeneralPath(
    const fs::path& runPath, const ModelRequest& request, Tape5& tape5) const {
  // use a "slant path to between two altitudes"
  tape5.card1.itype = 2;

  // use "scattered solar/lunar radiance" mode
  tape5.card1.iemsct = 2;

  // we don't need the profiles to be generated
  tape5.card1.noprt = 1;

  // if we are doing multiple scattering, figure out where to bias it
  if (tape5.card1.imult != 0) {
    if (request.pathAlt1M > request.pathAlt2M) {
      // downlooking case, do multiple scattering at H2
      tape5.card1.imult = -1;
    }
    else {
      // uplooking case, do multiple scattering at H1
      tape5.card1.imult = +1;
    }
  }

  // convert the altitudes and ranges to kilometers and quantize to meters
  float gndaltKm = (float)floor(request.groundAltM * 1.0e+03) / 1.0e+06;
  float h1Km = (float)floor(request.pathAlt1M * 1.0e+03) / 1.0e+06;
  float h2Km = (float)floor(request.pathAlt2M * 1.0e+03) / 1.0e+06;
  float rangeKm = (float)floor(request.pathLengthM * 1.0e+03) / 1.0e+06;

  // set the ground altitude
  tape5.card2.gndalt = gndaltKm;

  // setup the path geometry (TOA in MODTRAN is 100 km)
  std::string pathType = "General path run ";
  std::string pathMethod;
  if (h1Km < 100.0) {
    if (request.pathZenithDeg == 0) {
      tape5.card3.h1 = h1Km;
      if (rangeKm != 0.0) {
        tape5.card3.h2 = h1Km + rangeKm;
      }
      else {
        tape5.card3.h2 = 0.0;
      }
      tape5.card3.range = 0.0f;
      tape5.card3.obszen = 0.0f;
      tape5.card3.bckzen = static_cast<float>(180.0-request.pathZenithDeg);
      pathType += "(inside atmosphere, straight up)";
      pathMethod = fmt::format(
          "ITYPE = 2, using H1 = {:.3f}, H2 = {:.3f} and BCKZEN = {:.3f}",
          tape5.card3.h1, tape5.card3.h2, tape5.card3.bckzen );
    }
    else if (request.pathZenithDeg == 180) {
      tape5.card3.h1 = h1Km;
      if (rangeKm != 0.0) {
        tape5.card3.h2 = h1Km - rangeKm;
      }
      else {
        tape5.card3.h2 = gndaltKm;
      }
      tape5.card3.range = 0.0f;
      tape5.card3.obszen = static_cast<float>(request.pathZenithDeg);
      tape5.card3.bckzen = 0.0f;
      pathType += "(inside atmosphere, straight down)";
      pathMethod = fmt::format(
          "ITYPE = 2, using H1 = {:.3f}, H2 = {:.3f} and OBSZEN = {:.3f}",
          tape5.card3.h1, tape5.card3.h2, tape5.card3.obszen );
    }
    else if (request.pathZenithDeg <= 90.0) {
      tape5.card3.h1 = h1Km;
      tape5.card3.h2 = 0.0f;
      tape5.card3.range = static_cast<float>(rangeKm);
      tape5.card3.obszen = static_cast<float>(request.pathZenithDeg);
      tape5.card3.bckzen = 0.0f;
      pathType += "(inside atmosphere, uplooking)";
      pathMethod = fmt::format(
          "ITYPE = 2, using H1 = {:.3f}, RANGE = {:.3f} and OBSZEN = {:.3f}",
          tape5.card3.h1, tape5.card3.range, tape5.card3.obszen );
    }
    else {
      tape5.card3.h1 = h1Km;
      tape5.card3.h2 = 0.0f;
      tape5.card3.range = static_cast<float>(rangeKm);
      tape5.card3.obszen = 0.0f;
      tape5.card3.bckzen = static_cast<float>(180.0-request.pathZenithDeg);
      pathType += "(inside atmosphere, downlooking)";
      pathMethod = fmt::format(
          "ITYPE = 2, using H1 = {:.3f}, RANGE = {:.3f} and BCKZEN = {:.3f}",
          tape5.card3.h1, tape5.card3.range, tape5.card3.bckzen );
    }
  } else {
    // above the atmosphere, use H1, H2 and BCKZEN
    tape5.card3.h1 = h1Km;
    tape5.card3.h2 = h2Km;
    tape5.card3.obszen = 0.0f;
    tape5.card3.range = 0.0f;
    tape5.card3.bckzen = static_cast<float>(180.0-request.pathZenithDeg);
    pathType += "(above atmosphere, downlooking)";
    pathMethod = fmt::format(
        "ITYPE = 2, using H1 = {:.3f}, H2 = {:.3f} and BCKZEN = {:.3f}",
        tape5.card3.h1, tape5.card3.h2, tape5.card3.bckzen );

    // if nadir, BCKZEN = 0 (ambiguous) and we need to use OBSZEN
    if (request.pathZenithDeg == 180.0) {
      tape5.card3.obszen = static_cast<float>(request.pathZenithDeg);
      pathMethod += " (swapped BCKZEN -> OBSZEN)";
    }

    // make sure H2 is never below the ground
    if (tape5.card3.h2 < tape5.card2.gndalt) {
      tape5.card3.h2 = tape5.card2.gndalt;
    }
  }

  // fill in the rest of the geometry
  tape5.card3.beta = 0.0f;
  tape5.card3.lenn = 0.0f;
  tape5.card3.ro = 0.0f;

  // compute the relative azimuth between the path and source (-180:180)
  auto pathSourceAzimuth = request.pathAzimuthDeg - request.sourceAzimuthDeg;
  if (pathSourceAzimuth < -180.0)
    pathSourceAzimuth += 360.0;
  else if (pathSourceAzimuth > +180.0)
    pathSourceAzimuth -= 360.0;

  // figure out what source we are using
  std::string sourceName = "";
  if (request.options & ModelRequestOptions::SunSourceRequest) {
    sourceName = "Sun";
    tape5.card3a1.isourc = 0;
  } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
    sourceName = "Moon";
    tape5.card3a1.isourc = 1;
  } else {
    DIRSIG_THROW("ModtranTape: Invalid exo-atm source in request!");
  }

  tape5.card3a1.iparm = 2;
  tape5.card3a2.parm1 = static_cast<float>(pathSourceAzimuth);
  tape5.card3a2.parm2 = static_cast<float>(request.sourceZenithDeg);
  tape5.card3a2.parm3 = 0;
  tape5.card3a2.parm4 = 0;
  tape5.card3a2.time = 0.0;
  tape5.card3a2.anglem = 0.0;
  tape5.card3a2.g = 0.0;
  tape5.card3a2.psipo = static_cast<float>(request.pathAzimuthDeg);

  // save the tape5 file into the run folder
  tape5.write(PathCat(runPath, "tmp.tp5").string(), Tape5::QuietIO);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream
      << pathType << std::endl
      << pathMethod << std::endl
      << "Day of year = " << request.dayOfYear << std::endl
      << "Ground altitude = " << request.groundAltM << " [meters]" << std::endl
      << "Path altitude = " << request.pathAlt1M << " [meters]" << std::endl
      << "Path range = " << request.pathLengthM << " [meters]" << std::endl
      << "Path zenith = " << request.pathZenithDeg << " [degrees]" << std::endl
      << "Path azimuth = " << request.pathAzimuthDeg << " [degrees]"
          << std::endl
      << "Source is the " << sourceName << std::endl
      << "Source zenith = " << request.sourceZenithDeg << " [degrees]"
          << std::endl
      << "Source azimuth = " << request.sourceAzimuthDeg << " [degrees]"
          << std::endl;
  writeReadme(runPath, readmeStream.str());
}

/**
 * To get the total sky irradiance from MODTRAN, the common trick is to
 * look down at a Lambertian 100% reflector and get the reflected radiance
 * from it.
 */
void ModtranTapeBackend::setupTotalSky(
    const fs::path& runPath, const ModelRequest& request, Tape5& tape5) const {
  // use a "slant path to space or ground"
  tape5.card1.itype = 3;

  // use "scattered solar/lunar radiance" mode
  tape5.card1.iemsct = 2;

  // we don't need the profiles to be generated
  tape5.card1.noprt = 1;

  // save the old Card 1 SURREF and force it to "1.00"
  char savedSurRef[8];
  strncpy(savedSurRef, tape5.card1.surref, 8);
  strncpy(tape5.card1.surref, "  1.000\0", 8);

  // setup the basic geometry (convert to km)
  tape5.card2.gndalt = static_cast<float>(request.groundAltM / 1000.0);
  tape5.card3.h1 = 0.001;
  tape5.card3.h2 = 0.000;
  tape5.card3.obszen = 180.0;
  tape5.card3.range = 0.0;
  tape5.card3.beta = 0.0;
  tape5.card3.lenn = 0.0;
  tape5.card3.ro = 0.0;
  tape5.card3.bckzen = 0.0;

  // compute the relative azimuth between the path and source (-180:180)
  auto pathSourceAzimuth = request.pathAzimuthDeg - request.sourceAzimuthDeg;
  if (pathSourceAzimuth < -180.0)
    pathSourceAzimuth += 360.0;
  else if (pathSourceAzimuth > +180.0)
    pathSourceAzimuth -= 360.0;

  // figure out what source we are using
  std::string sourceName = "";
  if (request.options & ModelRequestOptions::SunSourceRequest) {
    sourceName = "Sun";
    tape5.card3a1.isourc = 0;
  } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
    sourceName = "Moon";
    tape5.card3a1.isourc = 1;
  } else {
    DIRSIG_THROW("ModtranTape: Invalid exo-atm source in request!");
  }


  // make sure we bias the multiple scattering to H1
  if (tape5.card1.imult != 0) {
    tape5.card1.imult = +1;
  }

  tape5.card3a1.iparm = 2;
  tape5.card3a2.parm1 = static_cast<float>(pathSourceAzimuth);
  tape5.card3a2.parm2 = static_cast<float>(request.sourceZenithDeg);
  tape5.card3a2.parm3 = 0;
  tape5.card3a2.parm4 = 0;
  tape5.card3a2.time = 0.0;
  tape5.card3a2.anglem = 0.0;
  tape5.card3a2.g = 0.0;
  tape5.card3a2.psipo = static_cast<float>(request.pathAzimuthDeg);

  // save the tape5 file into the run folder
  tape5.write(PathCat(runPath, "tmp.tp5").string(), Tape5::QuietIO);

  // restore the Card1 SURREF
  strncpy(tape5.card1.surref, savedSurRef, 8);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream << "Total sky run" << std::endl
               << "Day of year = " << request.dayOfYear << std::endl
               << "Ground altitude = " << request.groundAltM << " [meters]"
               << std::endl
               << "Source is the " << sourceName << std::endl
               << "Source zenith = " << request.sourceZenithDeg << " [degrees]"
               << std::endl
               << "Source azimuth = " << request.sourceAzimuthDeg
               << " [degrees]" << std::endl;
  writeReadme(runPath, readmeStream.str());
}

/**
 * This is the most traditional MODTRAN setup.
 */
void ModtranTapeBackend::setupTotalPath(
    const fs::path& runPath, const ModelRequest& request, Tape5& tape5) const {
  // use a "slant path to space or ground"
  tape5.card1.itype = 3;

  // use "scattered solar/lunar radiance" mode
  tape5.card1.iemsct = 2;

  // we don't need the profiles to be generated
  tape5.card1.noprt = 1;

  // if we are doing multiple scattering, figure out where to bias it
  if(tape5.card1.imult != 0) {
    if(request.pathAlt1M > request.pathAlt2M) {
      // downlooking case, do multiple scattering at H2
      tape5.card1.imult = -1;
    }
    else {
      // uplooking case, do multiple scattering at H1
      tape5.card1.imult = +1;
    }
  }

  // convert the altitudes and ranges to kilometers and quantize to meters
  float gndaltKm = (float)floor(request.groundAltM * 1.0e+03) / 1.0e+06;
  float h1Km = (float)floor(request.pathAlt1M * 1.0e+03) / 1.0e+06;
  float rangeKm = (float)floor(request.pathLengthM * 1.0e+03) / 1.0e+06;

  std::string pathType = "Total radiance run";

  // setup the path geometry
  tape5.card2.gndalt = gndaltKm;
  tape5.card3.h1 = h1Km;
  tape5.card3.h2 = 0.0;
  tape5.card3.obszen = 0.0;
  tape5.card3.range = rangeKm;
  tape5.card3.beta = 0.0;
  tape5.card3.lenn = 0.0;
  tape5.card3.ro = 0.0;

  // if nadir, BCKZEN = 0 (ambiguous) and we need to use OBSZEN
  std::string pathMethod = "";
  if (tape5.card3.bckzen == 0.0) {
    tape5.card3.obszen = static_cast<float>(request.pathZenithDeg);
    pathMethod = fmt::format(
        "ITYPE = 2, using H1 = {:.3f}, H2 = {:.3f} and OBSZEN = {:.3f}",
        tape5.card3.h1, tape5.card3.h2, tape5.card3.obszen );
  }
  else {
    tape5.card3.bckzen = static_cast<float>(180.0-request.pathZenithDeg);
    pathMethod = fmt::format(
        "ITYPE = 2, using H1 = {:.3f}, H2 = {:.3f} and BCKZEN = {:.3f}",
        tape5.card3.h1, tape5.card3.h2, tape5.card3.bckzen );
  }

  // compute the relative azimuth between the path and source (-180:180)
  auto pathSourceAzimuth = request.pathAzimuthDeg - request.sourceAzimuthDeg;
  if (pathSourceAzimuth < -180.0) {
    pathSourceAzimuth += 360.0;
  } else if (pathSourceAzimuth > +180.0) {
    pathSourceAzimuth -= 360.0;
  }

  // figure out what source we are using
  std::string sourceName = "";
  if (request.options & ModelRequestOptions::SunSourceRequest) {
    sourceName = "Sun";
    tape5.card3a1.isourc = 0;
  } else if (request.options & ModelRequestOptions::MoonSourceRequest) {
    sourceName = "Moon";
    tape5.card3a1.isourc = 1;
  } else {
    DIRSIG_THROW("ModtranTape: Invalid exo-atm source in request!");
  }

  tape5.card3a1.iparm = 2;
  tape5.card3a2.parm1 = static_cast<float>(pathSourceAzimuth);
  tape5.card3a2.parm2 = static_cast<float>(request.sourceZenithDeg);
  tape5.card3a2.parm3 = 0;
  tape5.card3a2.parm4 = 0;
  tape5.card3a2.time = 0.0;
  tape5.card3a2.anglem = 0.0;
  tape5.card3a2.g = 0.0;
  tape5.card3a2.psipo = static_cast<float>(request.pathAzimuthDeg);

  // save the tape5 file into the run folder
  tape5.write(PathCat(runPath, "tmp.tp5").string(), Tape5::QuietIO);

  // create the contents for the README
  std::ostringstream readmeStream;
  if (!request.description.empty()) {
    readmeStream << request.description << std::endl;
  }
  readmeStream
      << pathType << std::endl
      << pathMethod << std::endl
      << "Day of year = " << request.dayOfYear << std::endl
      << "Ground altitude = " << request.groundAltM << " [meters]" << std::endl
      << "Path altitude = " << request.pathAlt1M << " [meters]" << std::endl
      << "Path range = " << request.pathLengthM << " [meters]" << std::endl
      << "Path zenith = " << request.pathZenithDeg << " [degrees]" << std::endl
      << "Path azimuth = " << request.pathAzimuthDeg << " [degrees]"
          << std::endl
      << "Source is the " << sourceName << std::endl
      << "Source zenith = " << request.sourceZenithDeg << " [degrees]"
          << std::endl
      << "Source azimuth = " << request.sourceAzimuthDeg << " [degrees]"
          << std::endl;
  writeReadme(runPath, readmeStream.str());
}

void ModtranTapeBackend::runModtran(
    const fs::path& runPath, const ModelRequest& request) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest) return;

  // create the "modroot" files (unique names for MODTRAN4 and MODTRAN5)
  auto modRootFile = PathCat(runPath, "modroot.in");
  DIRSIG_ERROR_TRACE("writing file \"" + modRootFile.string() + "\"", {
    auto ofs = Ofstream(modRootFile.string());
    ofs << "tmp";
  })
  modRootFile = PathCat(runPath, "mod5root.in");
  DIRSIG_ERROR_TRACE("writing file \"" + modRootFile.string() + "\"", {
    auto ofs = Ofstream(modRootFile.string());
    ofs << "tmp";
  })

#ifdef _DEBUG
  std::cout << "Running '" << runPath.string() << "'" << std::endl;
  std::cout << profile_.exeFilePath << " ";
  for (const auto& e : exeArgs_)
    std::cout << e << " ";
  std::cout << std::endl;
#endif

  // setup and run the MODTRAN process
  auto outfile = PathCat(runPath, "modtran_stdout.txt"),
       errfile = PathCat(runPath, "modtran_stderr.txt");
  std::error_code ec;
  bp::child proc(
      modtranExe_.string(), bp::args(exeArgs_), bp::std_out > outfile,
      bp::std_err > errfile, bp::start_dir = runPath.string(), ec);

  // If we have an error here the setup of the exe failed
  DIRSIG_THROW_IF( ec,
      "Unable to start MODTRAN: '" + modtranExe_.string() + "' -> " + ec.message());

  // Wait for the process to complete
  proc.wait(ec);

  // Check if things went OK
  DIRSIG_THROW_IF( ec,
      "unable to run MODTRAN: " + std::to_string(ec.value()) + " - " + ec.message());
}

void ModtranTapeBackend::writeReadme(
    const fs::path& runPath, const std::string& contents) const {
  DIRSIG_ERROR_TRACE("writing README to \"" + runPath.string() + "\"", {
    auto p = PathCat(runPath, "README.txt");
    auto ofs = Ofstream(p.string());
    ofs << contents;
  })
}

void ModtranTapeBackend::cleanupRequest(const fs::path& runPath) const try {
  // if we are saving files, just remove the extra files
  if (saveFiles_) {
    for (const auto& f : extraFilenames_) fs::remove(PathCat(runPath, f));
  } else {
    // otherwise, purge the whole request directory
    fs::remove_all(runPath);
  }
} catch (...) {
}

void ModtranTapeBackend::readProfileResult(
    const fs::path& runPath,
    const ModelRequest& request,
    ModelResponse& response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest) return;

  // make room for the profile data (100 altitudes x 4 values)
  response.profileTable.resize(400);

  // open the MODTRAN output "scanned" file
  auto tp6Path = PathCat(runPath, "tmp.tp6").string();
  std::ifstream tp6File;
  DIRSIG_ERROR_TRACE("opening MODTRAN output file \"" + tp6Path + "\"", {
    tp6File = Ifstream(tp6Path);
  })

  // check if the file is empty
  DIRSIG_THROW_IF(
      StreamSize(tp6File) == 0,
      "MODTRAN output file \"" + tp6Path + "\" is empty");

  // read line-by-line to the first profile table and note the format
  std::string format = "Unknown";
  std::string tapeLine;
  while (std::getline(tp6File, tapeLine)) {
    // check if this is the start of a profile table
    if (!tapeLine.compare(
            0, kFormat1ProfileHeader.size(), kFormat1ProfileHeader)) {
      format = "Format1";
      break;
    } else if (!tapeLine.compare(
                   0, kFormat2ProfileHeader.size(), kFormat2ProfileHeader)) {
      format = "Format2";
      break;
    }
  }

  // check if we didn't find the tables

  DIRSIG_ERROR_TRACE("parsing MODTRAN output file \"" + tp6Path + "\"", {
    DIRSIG_THROW_IF(format == "Unknown", "failed to find atmospheric profile");

    // grab what we need
    if (format == "Format1") {
      this->readProfileFormat1(tp6File, response);
    } else if (format == "Format2") {
      this->readProfileFormat2(tp6File, response);
    }
  })
}

/**
 * This function reads the 2nd atmospheric profile from a "tape6"
 * file generated by MODTRAN4.
 */
void ModtranTapeBackend::readProfileFormat1(
    std::istream& tapeStream, ModelResponse& response) const {
  // find the next table (has temp, pressure, humidity)
  std::string line;
  while (std::getline(tapeStream, line)) {
    // check if this is the start of a profile table
    if (!line.compare(0, kFormat1ProfileHeader.size(), kFormat1ProfileHeader)) {
      break;
    }
  }

  // read the blank line, the two header lines and the next blank line
  std::getline(tapeStream, line);
  std::getline(tapeStream, line);
  std::getline(tapeStream, line);
  std::getline(tapeStream, line);

  // read the whole table
  int altIndex = 0;
  std::vector<double> lineTokens;
  while (std::getline(tapeStream, line) && !line.empty()) {
    // make sure we don't have too many altitudes
    DIRSIG_THROW_IF(
        altIndex >= 100, "too many altitudes in the atmospheric profile table");

    // split the line into the column tokens and check the count
    lineTokens.clear();
    double val;
    std::istringstream ss(line);
    while (ss >> val) lineTokens.push_back(val);
    DIRSIG_THROW_IF(
        lineTokens.size() != 14, "row " + std::to_string(altIndex + 1) +
                                     ": unexpected number of columns in the "
                                     "atmospheric profile table (" +
                                     std::to_string(lineTokens.size()) +
                                     ", expected 14)");

    // parse the columns and convert
    double altKm = lineTokens[1];
    double tempC = lineTokens[3] - 273.15;
    double pressureMb = lineTokens[2];
    double rh = lineTokens[10];
    response.profileTable[(altIndex * 4) + 0] = altKm;
    response.profileTable[(altIndex * 4) + 1] = tempC;
    response.profileTable[(altIndex * 4) + 2] = pressureMb;
    response.profileTable[(altIndex * 4) + 3] = Tdp(tempC, rh);

    altIndex++;
  }
}

/**
 * This function reads the 2nd atmospheric profile from a "tape6"
 * file generated by MODTRAN5 or MODTRAN6.
 */
void ModtranTapeBackend::readProfileFormat2(
    std::istream& tapeStream, ModelResponse& response) const {
  // find the next table
  std::string line;
  while (std::getline(tapeStream, line)) {
    // check if this is the start of a profile table
    if (!line.compare(0, kFormat2ProfileHeader.size(), kFormat2ProfileHeader)) {
      break;
    }
  }

  // read the blank line and the two header lines
  std::getline(tapeStream, line);
  std::getline(tapeStream, line);
  std::getline(tapeStream, line);

  // read the whole table
  int altIndex = 0;
  std::vector<double> lineTokens;
  while (std::getline(tapeStream, line) && !line.empty()) {
    // make sure we don't have too many altitudes
    DIRSIG_THROW_IF(
        altIndex >= 100, "too many altitudes in the atmospheric profile table");

    // split the line into the column tokens and check the count
    lineTokens.clear();
    double val;
    std::istringstream ss(line);
    while (ss >> val) lineTokens.push_back(val);
    DIRSIG_THROW_IF(
        lineTokens.size() != 14, "row " + std::to_string(altIndex + 1) +
                                     ": unexpected number of columns in the "
                                     "atmospheric profile table (" +
                                     std::to_string(lineTokens.size()) +
                                     ", expected 14)");

    // parse the columns and convert
    double altKm = lineTokens[1];
    double tempC = lineTokens[3] - 273.15;
    double pressureMb = lineTokens[2];
    double rh = lineTokens[10];
    response.profileTable[(altIndex * 4) + 0] = altKm;
    response.profileTable[(altIndex * 4) + 1] = tempC;
    response.profileTable[(altIndex * 4) + 2] = pressureMb;
    response.profileTable[(altIndex * 4) + 3] = Tdp(tempC, rh);

    altIndex++;
  }
}

void ModtranTapeBackend::readExoSourceResult(
    const fs::path& runPath,
    const ModelRequest& request,
    ModelResponse& response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest) return;

  // open the MODTRAN tape7 output file
  auto tp7Path = PathCat(runPath, "tmp.7sc").string();
  if (request.respData != nullptr) {
    tp7Path = PathCat(runPath, "tmp.tp7").string();
  }
  std::ifstream tp7File;
  try {
    tp7File = Ifstream(tp7Path);
  } catch (const std::exception& e) {
    DIRSIG_THROW(
        std::string("ModtranTapeBackend::readExoSourceResult:\n") +
        "Unable to open MODTRAN output file '" + tp7Path + "'\n" +
        e.what());
  }

  // check if the file is empty
  DIRSIG_THROW_IF( StreamSize(tp7File) == 0,
      "MODTRAN output file '" + tp7Path + "' is empty!");

#ifdef _DEBUG
  std::cout << "ReadExoSourceResult:" << std::endl
            << "Filename = " << tp7Path << std::endl;
#endif

  // read the tape7 header
  float h1, h2, angle, range, v1, v2, dv;
  DIRSIG_ERROR_TRACE(
      "reading header of MODTRAN output file \"" + tp7Path + "\"",
      { Tape7Output::readHeader(tp7File, h1, h2, angle, range, v1, v2, dv); })
  size_t vCount = static_cast<size_t>((v2 - v1) / dv);

  if (request.respData == nullptr) {
    // read the result directly into the output
    response.buffer1.resize(request.wavelengthSamples.size());
    response.buffer2.resize(request.wavelengthSamples.size());
    Tape7Output::readIrradianceData(
        tp7File, response.buffer1, response.buffer2);
  } else {
    // read the raw data into a temporary buffer
    std::vector<float> totalRad(vCount);
    std::vector<float> totalTau(vCount);
    Tape7Output::readIrradianceData(tp7File, totalRad, totalTau);

    // integrate the raw data with the channels
    std::vector<ChannelResponse> channels;
    setupChannelResponses(v1, v2, dv, vCount, request.respData, channels);
    response.buffer1.resize(request.respData->size());
    response.buffer2.resize(request.respData->size());
    for (size_t ci = 0; ci < channels.size(); ci++) {
      response.buffer1[ci] = channels[ci].integrate(totalRad, true);
      response.buffer2[ci] = channels[ci].average(totalTau);
    }
  }

  // make sure we didn't have a read fail
  DIRSIG_THROW_IF( !tp7File,
    std::string("ModtranTapeBackend::readExoSourceResult: read failed!\n") +
    "Filename = '" + tp7Path + "'\n" );
    
  // check if we need to "correct" the transmission
  if (request.sourceZenithDeg >= 90.0) {
    for (size_t ii = 0; ii < request.wavelengthSamples.size(); ii++) {
      response.buffer2[ii] = 0.0;
    }
  }
}

void ModtranTapeBackend::readSkyPathResult(
    const fs::path& runPath,
    const ModelRequest& request,
    ModelResponse& response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest) return;

  // open the MODTRAN tape7 output file
  auto tp7Path = PathCat(runPath, "tmp.7sc").string();
  if (request.respData != nullptr) {
    tp7Path = PathCat(runPath, "tmp.tp7").string();
  }
  std::ifstream tp7File;
  try {
    tp7File = Ifstream(tp7Path);
  } catch (const std::exception& e) {
    DIRSIG_THROW(
        std::string("ModtranTapeBackend::readSkyPathResult:\n") +
        "Unable to open MODTRAN output file '" + tp7Path + "'\n" +
        e.what());
  }

  // check if the file is empty
  DIRSIG_THROW_IF( StreamSize(tp7File) == 0,
      "MODTRAN output file '" + tp7Path + "' is empty");

#ifdef _DEBUG
  std::cout << "ReadSkyPathResult:" << std::endl
            << "Filename = " << tp7Path << std::endl;
#endif

  // read the tape7 header
  float h1, h2, angle, range, v1, v2, dv;
  DIRSIG_ERROR_TRACE(
      "reading header of MODTRAN output file \"" + tp7Path + "\"",
      { Tape7Output::readHeader(tp7File, h1, h2, angle, range, v1, v2, dv); })
  size_t vCount = static_cast<size_t>((v2 - v1) / dv);

  // store the path length (convert to meters)
  response.range = range * 1000.0;

  if (request.respData == nullptr) {
    // read the result directly into the output
    response.buffer1.resize(request.wavelengthSamples.size());
    response.buffer2.resize(request.wavelengthSamples.size());
    Tape7Output::readTotalRadianceData(
        tp7File, response.buffer1, response.buffer2);
  } else {
    // read the raw data into a temporary buffer
    std::vector<float> totalRad(vCount);
    std::vector<float> totalTau(vCount);
    Tape7Output::readTotalRadianceData(tp7File, totalRad, totalTau);

    // integrate the raw data with the channels
    std::vector<ChannelResponse> channels;
    setupChannelResponses(v1, v2, dv, vCount, request.respData, channels);
    response.buffer1.resize(request.respData->size());
    response.buffer2.resize(request.respData->size());
    for (size_t ci = 0; ci < channels.size(); ci++) {
      response.buffer1[ci] = channels[ci].integrate(totalRad, true);
      response.buffer2[ci] = channels[ci].average(totalTau);
    }
  }

  // make sure we didn't have a read fail
  DIRSIG_THROW_IF( !tp7File,
    std::string("ModtranTapeBackend::readSkyPathResult: read failed!\n") +
    "Filename = '" + tp7Path + "'\n" );
}

void ModtranTapeBackend::readGeneralPathResult(
    const fs::path& runPath,
    const ModelRequest& request,
    ModelResponse& response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest) return;

  // open the MODTRAN tape7 output file
  auto tp7Path = PathCat(runPath, "tmp.7sc").string();
  if (request.respData != nullptr) {
    tp7Path = PathCat(runPath, "tmp.tp7").string();
  }
  std::ifstream tp7File;
  try {
    tp7File = Ifstream(tp7Path);
  } catch (const std::exception& e) {
    DIRSIG_THROW(
        std::string("ModtranTapeBackend::readGeneralPathResult:\n") +
        "Unable to open MODTRAN output file '" + tp7Path + "'\n" +
        e.what());
  }

  // check if the file is empty
  DIRSIG_THROW_IF( StreamSize(tp7File) == 0,
      "MODTRAN output file '" + tp7Path + "' is empty!");

#ifdef _DEBUG
  std::cout << "ReadGeneralPathResult:" << std::endl
            << "Filename = " << tp7Path << std::endl;
#endif

  // read the tape7 header
  float h1, h2, angle, range, v1, v2, dv;
  DIRSIG_ERROR_TRACE(
      "reading header of MODTRAN output file \"" + tp7Path + "\"",
      { Tape7Output::readHeader(tp7File, h1, h2, angle, range, v1, v2, dv); })
  size_t vCount = static_cast<size_t>((v2 - v1) / dv);

  // if we started above the atmosphere, don't let MODTRAN reduce the range
  auto origRangeKm = request.pathLengthM / 1000.0;
  if ((request.pathAlt1M > 100.0e+03) && (range < origRangeKm)) {
    range = origRangeKm;
  }

  // store the path length (convert to meters)
  response.range = range * 1000.0;

  if (request.respData == nullptr) {
    // read the result directly into the output
    response.buffer1.resize(request.wavelengthSamples.size());
    response.buffer2.resize(request.wavelengthSamples.size());
    Tape7Output::readPathRadianceData(
        tp7File, response.buffer1, response.buffer2);
  } else {
    // read the raw data into a temporary buffer
    std::vector<float> totalRad(vCount);
    std::vector<float> totalTau(vCount);
    Tape7Output::readPathRadianceData(tp7File, totalRad, totalTau);

    // integrate the raw data with the channels
    std::vector<ChannelResponse> channels;
    setupChannelResponses(v1, v2, dv, vCount, request.respData, channels);
    response.buffer1.resize(request.respData->size());
    response.buffer2.resize(request.respData->size());
    for (size_t ci = 0; ci < channels.size(); ci++) {
      response.buffer1[ci] = channels[ci].integrate(totalRad, true);
      response.buffer2[ci] = channels[ci].average(totalTau);
    }
  }

  // make sure we didn't have a read fail
  DIRSIG_THROW_IF( !tp7File,
    std::string("ModtranTapeBackend::ReadGeneralPathResult: read failed!\n") +
    "Filename = '" + tp7Path + "'\n" );

  // normalize and convert the data
  for (size_t ii = 0; ii < response.buffer1.size(); ii++) {
    response.buffer1[ii] = response.buffer1[ii] / response.range;
    if (response.buffer2[ii] > 0.0) {
      response.buffer2[ii] = -log(response.buffer2[ii]) / response.range;
    } else {
      response.buffer2[ii] = infiniteExt;
    }
  }
}

void ModtranTapeBackend::readTotalSkyResult(
    const fs::path& runPath,
    const ModelRequest& request,
    ModelResponse& response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest) return;

  // open the MODTRAN tape7 output file
  auto tp7Path = PathCat(runPath, "tmp.7sc").string();
  if (request.respData != nullptr) {
    tp7Path = PathCat(runPath, "tmp.tp7").string();
  }
  std::ifstream tp7File;
  try {
    tp7File = Ifstream(tp7Path);
  } catch (const std::exception& e) {
    DIRSIG_THROW(
        std::string("ModtranTapeBackend::readTotalSkyResult:\n") +
        "Unable to open MODTRAN output file '" + tp7Path + "'\n" +
        e.what());
  }

  // check if the file is empty
  DIRSIG_THROW_IF( StreamSize(tp7File) == 0,
      "MODTRAN output file '" + tp7Path + "' is empty!");

#ifdef _DEBUG
  std::cout << "ReadTotalSkyResult:" << std::endl
            << "Filename = " << tp7Path << std::endl;
#endif

  // read the tape7 header
  float h1, h2, angle, range, v1, v2, dv;
  DIRSIG_ERROR_TRACE(
      "reading header of MODTRAN output file \"" + tp7Path + "\"",
      { Tape7Output::readHeader(tp7File, h1, h2, angle, range, v1, v2, dv); })
  size_t vCount = static_cast<size_t>((v2 - v1) / dv);

  // store the path length (convert to meters)
  response.range = range * 1000.0;

  if (request.respData == nullptr) {
    // read the result directly into the output
    response.buffer1.resize(request.wavelengthSamples.size());
    response.buffer2.resize(request.wavelengthSamples.size());
    Tape7Output::readTotalRadianceData(
        tp7File, response.buffer1, response.buffer2);
  } else {
    // read the raw data into a temporary buffer
    std::vector<float> totalRad(vCount);
    std::vector<float> totalTau(vCount);
    Tape7Output::readTotalRadianceData(tp7File, totalRad, totalTau);

    // integrate the raw data with the channels
    std::vector<ChannelResponse> channels;
    setupChannelResponses(v1, v2, dv, vCount, request.respData, channels);
    response.buffer1.resize(request.respData->size());
    response.buffer2.resize(request.respData->size());
    for (size_t ci = 0; ci < channels.size(); ci++) {
      response.buffer1[ci] = channels[ci].integrate(totalRad, true);
      response.buffer2[ci] = channels[ci].average(totalTau);
    }
  }

  // make sure we didn't have a read fail
  DIRSIG_THROW_IF( !tp7File,
    std::string("ModtranTapeBackend::readTotalSkyResult: read failed!\n") +
    "Filename = '" + tp7Path + "'\n" );
}

/**
 * This is the reader for a "total" (L-shaped path) run in MODTRAN.
 */
void ModtranTapeBackend::readTotalPathResult(
    const fs::path& runPath,
    const ModelRequest& request,
    ModelResponse& response) const {
  // if this request was disabled, skip it
  if (request.options & ModelRequestOptions::SkipRequest) return;

  // open the MODTRAN tape7 output file
  auto tp7Path = PathCat(runPath, "tmp.7sc").string();
  if (request.respData != nullptr) {
    tp7Path = PathCat(runPath, "tmp.tp7").string();
  }
  std::ifstream tp7File;
  try {
    tp7File = Ifstream(tp7Path);
  } catch (const std::exception& e) {
    DIRSIG_THROW(
        std::string("ModtranTapeBackend::readTotalPathResult:\n") +
        "Unable to open MODTRAN output file '" + tp7Path + "'\n" +
        e.what());
  }

  // check if the file is empty
  DIRSIG_THROW_IF( StreamSize(tp7File) == 0,
      "MODTRAN output file '" + tp7Path + "' is empty!");

#ifdef _DEBUG
  std::cout << "ReadTotalPathResult:" << std::endl
            << "Filename = " << tp7Path << std::endl;
#endif

  // read the tape7 header
  float h1, h2, angle, range, v1, v2, dv;
  DIRSIG_ERROR_TRACE(
      "reading header of MODTRAN output file \"" + tp7Path + "\"",
      { Tape7Output::readHeader(tp7File, h1, h2, angle, range, v1, v2, dv); })
  size_t vCount = static_cast<size_t>((v2 - v1) / dv);

  if (request.respData == nullptr) {
    // read the result directly into the output
    response.buffer1.resize(request.wavelengthSamples.size());
    response.buffer2.resize(request.wavelengthSamples.size());
    Tape7Output::readSensorAndSurfaceData(
        tp7File, response.buffer1, response.buffer2);
  } else {
    // read the raw data into a temporary buffer
    std::vector<float> sensorRad(vCount);
    std::vector<float> surfaceRad(vCount);
    Tape7Output::readSensorAndSurfaceData(tp7File, sensorRad, surfaceRad);

    // integrate the raw data with the channels
    std::vector<ChannelResponse> channels;
    setupChannelResponses(v1, v2, dv, vCount, request.respData, channels);
    response.buffer1.resize(request.respData->size());
    response.buffer2.resize(request.respData->size());
    for (size_t ci = 0; ci < channels.size(); ci++) {
      response.buffer1[ci] = channels[ci].integrate(sensorRad, true);
      response.buffer2[ci] = channels[ci].integrate(surfaceRad, true);
    }
  }

  // make sure we didn't have a read fail
  DIRSIG_THROW_IF( !tp7File,
    std::string("ModtranTapeBackend::readTotalPathResult: read failed!\n") +
    "Filename = '" + tp7Path + "'\n" );
}

void ModtranTapeBackend::setupChannelResponses(
    float v1,
    float v2,
    float dv,
    size_t vCount,
    const SpectralResponseData* respData,
    std::vector<ChannelResponse>& channels) const {
  channels.resize(respData->size());

  for (size_t ci = 0; ci < respData->size(); ci++) {
    auto& rsp = respData->at(ci);
    float firstFreq = 10000.0 / rsp.back().wavelength;
    firstFreq = std::min(std::max(firstFreq, v1), v2);
    float lastFreq = 10000.0 / rsp.front().wavelength;
    lastFreq = std::min(std::max(lastFreq, v1), v2);
    size_t freqCount = static_cast<size_t>((lastFreq - firstFreq) / dv);

    //
    // The interpolator needs the old and new data to be on the same
    // spectral units.  But we are interpolating to wavenumbers, which
    // are also going to be in decreasing order as wavelengths. So we
    // compute the target wavelengths in increasing order, perform the
    // interpolation and then reverse them.
    //
    std::vector<float> oldWavelengths(rsp.size());
    std::vector<float> oldData(rsp.size());
    for (size_t ii = 0; ii < rsp.size(); ii++) {
      oldWavelengths[ii] = rsp[ii].wavelength;
      oldData[ii] = rsp[ii].value;
    }
    std::vector<float> newWavelengths(freqCount);
    std::vector<float> newWidths(freqCount);
    for (size_t ii = 0; ii < freqCount; ii++) {
      float centerFreq = lastFreq - (ii * dv);
      newWavelengths[ii] = 10000.0 / centerFreq;
      float lowerWave = 10000.0 / (centerFreq + dv / 2);
      float upperWave = 10000.0 / (centerFreq - dv / 2);
      newWidths[ii] = upperWave - lowerWave;
    }

    std::vector<float> newData(freqCount);
    d5::interpolate(oldWavelengths, oldData, newWavelengths, newData);

    // create the channel response (reverse the weights)
    ChannelResponse& channel = channels[ci];
    channel.startIndex = static_cast<long>((firstFreq - v1) / dv);
    channel.stopIndex = static_cast<long>((lastFreq - v1) / dv);
    channel.pointCount = channel.stopIndex - channel.startIndex;
    channel.weights.resize(newData.size());
    std::reverse_copy(
        std::begin(newData), std::end(newData), std::begin(channel.weights));
    channel.widths.resize(newWidths.size());
    std::reverse_copy(
        std::begin(newWidths), std::end(newWidths), std::begin(channel.widths));

#ifdef _DEBUG
    std::cout << "firstFreq = " << firstFreq << std::endl;
    std::cout << "lastFreq  = " << lastFreq << std::endl;
    std::cout << "freqCount = " << freqCount << std::endl;
    std::cout << "Channel #" << ci << std::endl
              << "    Start index = " << channel.startIndex << std::endl
              << "    Start freq = " << (v1 + channel.startIndex * dv)
              << std::endl
              << "    Stop index = " << channel.stopIndex << std::endl
              << "    Stop freq = " << (v1 + channel.stopIndex * dv)
              << std::endl
              << "    Freq count = " << freqCount << std::endl;
    for (size_t ii = 0; ii < freqCount; ii++) {
      std::cout << 10000.0 / newWavelengths[ii] << " " << newWavelengths[ii]
                << " " << newData[ii] << std::endl;
    }
#endif
  }
}

}
