#include "TapeHelper.h"

#include <cassert>
#include <cmath>
#include <fstream>
#include <iostream>
#include <limits>
#include <sstream>

#ifdef WIN32
#include <stdio.h>
#endif

#include <filesystem>

#define CARD_LENGTH 256
#define TOKEN_LENGTH 256

char *Tape5::readCard(std::istream &inStream) {
  static char cardStr[CARD_LENGTH];

  // read the next line and then gobble the remaining newline
  inStream.getline(cardStr, CARD_LENGTH, '\n');
  int cardLength = strlen(cardStr);

  for (int ii = 0; ii < cardLength; ii++) {
    if ((cardStr[ii] == '\n') || (cardStr[ii] == '\r')) {
      for (int jj = ii; jj < cardLength; jj++) {
        cardStr[jj] = '\0';
      }
      break;
    }
  }

  return cardStr;
}

void Tape5::fillVariable(char *_var, char _fillValue, int _count) {
  for (int ii = 0; ii < _count; ii++) {
    _var[ii] = _fillValue;
  }
}

/**
 * Extract an integer variable from a card.
 */
void Tape5::extractVariable(const char *_card, size_t _start, size_t _width,
                            int &_var) {
  assert(_width <= TOKEN_LENGTH);

  // set the default value
  _var = 0;

  // check if the variable starts past the end of the card
  if (_start >= strlen(_card)) {
    return;
  }
  // check if the variable ends past the end of the card
  else if ((_start + _width) > strlen(_card)) {
    _width = strlen(_card) - _start;
  }

  char tokenStr[TOKEN_LENGTH];
  fillVariable(tokenStr, '\0', TOKEN_LENGTH);
  strncpy(tokenStr, _card + _start, _width);
  sscanf(tokenStr, "%d", &_var);
}

/**
 * Extract an float variable from a card.
 */
void Tape5::extractVariable(const char *_card, size_t _start, size_t _width,
                            float &_var) {
  assert(_width <= TOKEN_LENGTH);

  // set the default value
  _var = 0.0;

  // check if the variable starts past the end of the card
  if (_start >= strlen(_card)) {
    return;
  }
  // check if the variable ends past the end of the card
  else if ((_start + _width) > strlen(_card)) {
    _width = strlen(_card) - _start;
  }

  char tokenStr[TOKEN_LENGTH];
  fillVariable(tokenStr, '\0', TOKEN_LENGTH);
  strncpy(tokenStr, _card + _start, _width);
  sscanf(tokenStr, "%f", &_var);
}

/*
 * We assume the character array is width+1 long, so we have room for a
 * null at the end.
 */
void Tape5::extractVariable(const char *_card, size_t _start, size_t _width,
                            char *_var) {
  assert(_width <= TOKEN_LENGTH);

  // set the default value
  fillVariable(_var, ' ', _width);
  _var[_width] = '\0';

  // check if the variable starts past the end of the card
  if (_start >= strlen(_card)) {
    return;
  }
  // check if the variable ends past the end of the card
  else if ((_start + _width) > strlen(_card)) {
    if ((strlen(_card) - _start) <= 0) {
      return;
    }
    _width = strlen(_card) - _start;
  }

  char tokenStr[TOKEN_LENGTH];
  fillVariable(tokenStr, '\0', TOKEN_LENGTH);
  strncpy(tokenStr, _card + _start, _width);
  for (size_t ii = 0; ii < _width; ii++) {
    _var[ii] = tokenStr[ii];
    if ((_var[ii] == '\n') || (_var[ii] == '\0')) {
      _var[ii] = ' ';
      break;
    }
  }
}

void Tape5::extractVariable(const char *_card, size_t _start, char &_var) {
  char tokenStr[TOKEN_LENGTH];
  if (_start >= strlen(_card)) {
    _var = ' ';
  } else {
    fillVariable(tokenStr, '\0', TOKEN_LENGTH);
    strncpy(tokenStr, _card + _start, 1);
    sscanf(tokenStr, "%c", &_var);
    if ((_var == '\n') || (_var == '\r') || (_var == '\0')) {
      _var = ' ';
    }
  }
}

/*
 * Read tape5
 */
void Tape5::read(const std::string &_filename, Verbosity _verbosity) {
  std::string errMsg;
  if (_verbosity == VerboseIO) {
    std::cout << "Reading input card tape5 file ... \n" << std::flush;
  }

  // check if file is actually a directory
  if (std::filesystem::is_directory(_filename)) {
    errMsg = "Supplied tape5 filename is an existing directory!\n";
    errMsg += "Filename = '" + _filename + "'";
    throw std::runtime_error(errMsg);
  }

  // open up the file as a stream
  std::ifstream tpStream(_filename.c_str());
  if (!tpStream) {
    errMsg = "Unable to open input tape5 file!\n";
    errMsg += "Filename = '" + _filename + "'";
    throw std::runtime_error(errMsg);
  }

  // read the file
  this->read(tpStream);

  // close the file
  this->filename = _filename;
}

/*
 * Read tape5
 */
void Tape5::read(std::istream &fp) {
  char *cardStr = nullptr;

  /* initialize a couple things */
  card1.modtrn = 'm';
  card1.speed = 's';
  card1a.dis = 'f';
  card1a.disazm = 'f';
  card1a.disalb = ' ';
  card1a.nstr = 0;
  card1a.sfwhm = 0.0;
  card1a.co2mx = 0.0;
  card1a.lsunfl = 'f';
  card1a.lbmnam = 'f';
  card1a.lfltnm = 'f';
  card1a.h2oaer = 'f';
  card1a.cdtdir = 'f';
  card1a.solcon = 0.0;
  for (int ii = 0; ii < 11; ii++) {
    card1a.h2ostr[ii] = '\0';
    card1a.o3str[ii] = '\0';
  }

  /* read out the values for CARD1 */
  cardStr = readCard(fp);
  extractVariable(cardStr, 0, card1.modtrn);
  extractVariable(cardStr, 1, card1.speed);
  extractVariable(cardStr, 2, card1.cbinout);
  extractVariable(cardStr, 3, card1.lymolc);
  extractVariable(cardStr, 4, 1, card1.model);
  extractVariable(cardStr, 5, card1.tbest);
  extractVariable(cardStr, 6, 4, card1.itype);
  extractVariable(cardStr, 10, 5, card1.iemsct);
  extractVariable(cardStr, 15, 5, card1.imult);
  extractVariable(cardStr, 20, 5, card1.m1);
  extractVariable(cardStr, 25, 5, card1.m2);
  extractVariable(cardStr, 30, 5, card1.m3);
  extractVariable(cardStr, 35, 5, card1.m4);
  extractVariable(cardStr, 40, 5, card1.m5);
  extractVariable(cardStr, 45, 5, card1.m6);
  extractVariable(cardStr, 50, 5, card1.mdef);
  extractVariable(cardStr, 55, 5, card1.im);
  extractVariable(cardStr, 61, 4, card1.noprt);
  extractVariable(cardStr, 65, 8, card1.tptemp);
  extractVariable(cardStr, 73, 7, card1.surref);

  /* read the next card */
  cardStr = readCard(fp);

  /* this card could be CARD #2 or CARD #1A (depending on the version) */
  if ((cardStr[0] == 'F') || (cardStr[0] == 'f') || (cardStr[0] == 'T') ||
      (cardStr[0] == 't') || (cardStr[0] == 'S') || (cardStr[0] == 's')) {
    for (int ii = 0; ii < 11; ii++) {
      card1a.h2ostr[ii] = '\0';
      card1a.o3str[ii] = '\0';
    }

    /* read the CARD #1A values */
    extractVariable(cardStr, 0, card1a.dis);
    extractVariable(cardStr, 1, card1a.disazm);
    extractVariable(cardStr, 2, card1a.disalb);
    extractVariable(cardStr, 3, 3, card1a.nstr);
    extractVariable(cardStr, 6, 4, card1a.sfwhm);
    extractVariable(cardStr, 10, 10, card1a.co2mx);
    extractVariable(cardStr, 20, 10, card1a.h2ostr);
    extractVariable(cardStr, 30, 10, card1a.o3str);
    extractVariable(cardStr, 40, 1, card1a.c_prof);
    extractVariable(cardStr, 41, card1a.lsunfl);
    extractVariable(cardStr, 43, card1a.lbmnam);
    extractVariable(cardStr, 45, card1a.lfltnm);
    extractVariable(cardStr, 47, card1a.h2oaer);
    extractVariable(cardStr, 49, card1a.cdtdir);
    extractVariable(cardStr, 50, 10, card1a.solcon);
    extractVariable(cardStr, 60, card1a.cdastm);
    extractVariable(cardStr, 61, 9, card1a.astmc);
    extractVariable(cardStr, 70, 10, card1a.astmx);
    extractVariable(cardStr, 80, 10, card1a.astmo);
    extractVariable(cardStr, 90, 10, card1a.aerrh);
    extractVariable(cardStr, 100, 10, card1a.aerrh);

    /* check for a valid DIS variable */
    if ((card1a.dis != 'T') && (card1a.dis != 't') && (card1a.dis != 'F') &&
        (card1a.dis != 'f') && (card1a.dis != 'S') && (card1a.dis != 's') &&
        (card1a.dis != ' ')) {
      card1a.dis = ' ';
    }

    /* check for a valid DISAZM variable */
    if ((card1a.disazm != 'T') && (card1a.disazm != 't') &&
        (card1a.disazm != 'F') && (card1a.disazm != 'f') &&
        (card1a.disazm != ' ')) {
      card1a.disazm = ' ';
    }

    /* check for a valid LSUNFL variable */
    if ((card1a.lsunfl != 'T') && (card1a.lsunfl != 't') &&
        (card1a.lsunfl != 'F') && (card1a.lsunfl != 'f') &&
        (card1a.lsunfl != ' ')) {
      card1a.lsunfl = ' ';
    }

    /* check for a valid LBMNAM variable */
    if ((card1a.lbmnam != 'T') && (card1a.lbmnam != 't') &&
        (card1a.lbmnam != 'F') && (card1a.lbmnam != 'f') &&
        (card1a.lbmnam != ' ')) {
      card1a.lbmnam = ' ';
    }

    /* check for a valid LFLTNM variable */
    if ((card1a.lfltnm != 'T') && (card1a.lfltnm != 't') &&
        (card1a.lfltnm != 'F') && (card1a.lfltnm != 'f') &&
        (card1a.lfltnm != ' ')) {
      card1a.lfltnm = ' ';
    }

    /* check for a valid H2OAER variable */
    if ((card1a.h2oaer != 'T') && (card1a.h2oaer != 't') &&
        (card1a.h2oaer != 'F') && (card1a.h2oaer != 'f') &&
        (card1a.h2oaer != ' ')) {
      card1a.h2oaer = ' ';
    }

    /* check for a valid CDTDIR variable */
    if ((card1a.cdtdir != 'T') && (card1a.cdtdir != 't') &&
        (card1a.cdtdir != 'F') && (card1a.cdtdir != 'f') &&
        (card1a.cdtdir != ' ')) {
      card1a.cdtdir = ' ';
    }

    /* check to read CARD 1A1 */
    fillVariable(card1a1.usrsun, '\0', 256);
    if ((card1a.lsunfl == 'T') || (card1a.lsunfl == 't')) {
      cardStr = readCard(fp);
      strncpy(card1a1.usrsun, cardStr, CARD_LENGTH);
    }

    /* check to read CARD 1A2 */
    fillVariable(card1a2.bmname, '\0', 256);
    if ((card1a.lbmnam == 'T') || (card1a.lbmnam == 't')) {
      cardStr = readCard(fp);
      strncpy(card1a2.bmname, cardStr, CARD_LENGTH);
    }

    /* check to read CARD 1A3 */
    fillVariable(card1a3.filtnm, '\0', 256);
    if ((card1a.lfltnm == 'T') || (card1a.lfltnm == 't')) {
      cardStr = readCard(fp);
      strncpy(card1a3.filtnm, cardStr, CARD_LENGTH);
    }

    /* check to read CARD 1A4 */
    fillVariable(card1a4.datdir, '\0', 80);
    if ((card1a.cdtdir == 'T') || (card1a.cdtdir == 't')) {
      cardStr = readCard(fp);
      strncpy(card1a4.datdir, cardStr, CARD_LENGTH);
    }

    /* check to read CARD 1A5 */
    if ((card1a.c_prof == 1) || (card1a.c_prof == 3) || (card1a.c_prof == 5) ||
        (card1a.c_prof == 7)) {
      cardStr = readCard(fp);
      for (int ii = 0; ii < 10; ii++) {
        extractVariable(cardStr, ii * 5, 5, card1a5.s_umix[ii]);
      }
    }

    /* check to read CARD 1A6 */
    if ((card1a.c_prof == 2) || (card1a.c_prof == 3) || (card1a.c_prof == 6) ||
        (card1a.c_prof == 7)) {
      cardStr = readCard(fp);
      for (int ii = 0; ii < 13; ii++) {
        extractVariable(cardStr, ii * 5, 5, card1a6.s_xsec[ii]);
      }
    }

    /* check to write CARD 1A7 */
    if ((card1a.c_prof == 4) || (card1a.c_prof == 5) || (card1a.c_prof == 6) ||
        (card1a.c_prof == 7)) {
      cardStr = readCard(fp);
      for (int ii = 0; ii < 16; ii++) {
        extractVariable(cardStr, ii * 5, 5, card1a7.s_trac[ii]);
      }
    }

    /* read the next line */
    cardStr = readCard(fp);
  }

  /* read the CARD #2 values */
  extractVariable(cardStr, 0, 2, card2.aplus);
  extractVariable(cardStr, 2, 3, card2.ihaze);
  extractVariable(cardStr, 5, card2.cnovam);
  extractVariable(cardStr, 6, 4, card2.iseasn);
  extractVariable(cardStr, 10, 3, card2.aruss);
  extractVariable(cardStr, 13, 2, card2.ivulcn);
  extractVariable(cardStr, 15, 5, card2.icstl);
  extractVariable(cardStr, 20, 5, card2.icld);
  extractVariable(cardStr, 25, 5, card2.ivsa);
  extractVariable(cardStr, 30, 10, card2.vis);
  extractVariable(cardStr, 40, 10, card2.wss);
  extractVariable(cardStr, 50, 10, card2.whh);
  extractVariable(cardStr, 60, 10, card2.rainrt);
  extractVariable(cardStr, 70, 10, card2.gndalt);

  /* check for flexible aeorsol marker (A+) */
  if (strncmp(card2.aplus, "A+", 2) == 0) {
    cardStr = readCard(fp);
    extractVariable(cardStr, 1, 9, card2a_plus.zaer11);
    extractVariable(cardStr, 11, 9, card2a_plus.zaer12);
    extractVariable(cardStr, 21, 9, card2a_plus.scale1);
    extractVariable(cardStr, 51, 9, card2a_plus.zaer21);
    extractVariable(cardStr, 61, 9, card2a_plus.zaer22);
    extractVariable(cardStr, 71, 9, card2a_plus.scale2);

    cardStr = readCard(fp);
    extractVariable(cardStr, 1, 9, card2a_plus.zaer31);
    extractVariable(cardStr, 11, 9, card2a_plus.zaer32);
    extractVariable(cardStr, 21, 9, card2a_plus.scale3);
    extractVariable(cardStr, 51, 9, card2a_plus.zaer41);
    extractVariable(cardStr, 61, 9, card2a_plus.zaer42);
    extractVariable(cardStr, 71, 9, card2a_plus.scale4);
  }

  /* check if this file contains CARD 2A (Standard Form) */
  if ((card2.icld == 18) || (card2.icld == 19)) {
    cardStr = readCard(fp);
    extractVariable(cardStr, 0, 8, card2a.cthik);
    extractVariable(cardStr, 8, 8, card2a.calt);
    extractVariable(cardStr, 16, 8, card2a.cext);
  }

  /* check if this file contains CARD 2A (Alternate Form) */
  if ((card2.icld >= 1) && (card2.icld <= 10)) {
    cardStr = readCard(fp);
    extractVariable(cardStr, 0, 8, card2a_alt.cthik);
    extractVariable(cardStr, 8, 8, card2a_alt.calt);
    extractVariable(cardStr, 16, 8, card2a_alt.cext);
    extractVariable(cardStr, 24, 4, card2a_alt.ncralt);
    extractVariable(cardStr, 28, 4, card2a_alt.ncrspc);
    extractVariable(cardStr, 32, 8, card2a_alt.cwavln);
    extractVariable(cardStr, 40, 8, card2a_alt.ccolwd);
    extractVariable(cardStr, 48, 8, card2a_alt.ccolip);
    extractVariable(cardStr, 56, 8, card2a_alt.chumid);
    extractVariable(cardStr, 64, 8, card2a_alt.asymwd);
    extractVariable(cardStr, 72, 8, card2a_alt.asymip);
  }

  /* check if we need to read CARD_2B */
  if (card2.ivsa == 1) {
    cardStr = readCard(fp);
    extractVariable(cardStr, 0, 10, card2b.zcvsa);
    extractVariable(cardStr, 10, 10, card2b.ztvsa);
    extractVariable(cardStr, 20, 10, card2b.zinvsa);
  }

  /* check if this tape5 contains a user atmosphere (radiosonde) */
  if (((card1.model == 0) || (card1.model == 7)) && (card1.im == 1)) {
    /* Cards 2C and 2C1 are always read for MODEL 0 or 7 */
    cardStr = readCard(fp);
    extractVariable(cardStr, 0, 5, card2c.ml);
    extractVariable(cardStr, 5, 5, card2c.ird1);
    extractVariable(cardStr, 10, 5, card2c.ird2);
    extractVariable(cardStr, 15, 20, card2c.hmodel);
    extractVariable(cardStr, 35, 10, card2c.ree);
    extractVariable(cardStr, 45, 5, card2c.nmolyc);
    extractVariable(cardStr, 50, 10, card2c.e_mass);
    extractVariable(cardStr, 60, 10, card2c.airmwt);

    /* read the cards for each layer */
    for (int layer = 0; layer < card2c.ml; layer++) {
      cardStr = readCard(fp);

      extractVariable(cardStr, 0, 10, card2c1[layer].zmdl);
      extractVariable(cardStr, 10, 10, card2c1[layer].p);
      extractVariable(cardStr, 20, 10, card2c1[layer].t);
      extractVariable(cardStr, 30, 10, card2c1[layer].wmol[0]);
      extractVariable(cardStr, 40, 10, card2c1[layer].wmol[1]);
      extractVariable(cardStr, 50, 10, card2c1[layer].wmol[2]);
      extractVariable(cardStr, 60, 14, card2c1[layer].jchar);
      extractVariable(cardStr, 74, card2c1[layer].jcharx);
      extractVariable(cardStr, 75, card2c1[layer].jchary);

      if (card2c.ird1 == 1) {
        // read in Card 2C2 (two lines!)
        cardStr = readCard(fp);
        extractVariable(cardStr, 0, 10, card2c2[layer].wmol[3]);
        extractVariable(cardStr, 10, 10, card2c2[layer].wmol[4]);
        extractVariable(cardStr, 20, 10, card2c2[layer].wmol[5]);
        extractVariable(cardStr, 30, 10, card2c2[layer].wmol[6]);
        extractVariable(cardStr, 40, 10, card2c2[layer].wmol[7]);
        extractVariable(cardStr, 50, 10, card2c2[layer].wmol[8]);
        extractVariable(cardStr, 60, 10, card2c2[layer].wmol[9]);
        extractVariable(cardStr, 70, 10, card2c2[layer].wmol[10]);
        cardStr = readCard(fp);
        extractVariable(cardStr, 0, 10, card2c2[layer].wmol[11]);
      }

      if (card2c.ird2 == 1) {
        // read in Card 2C3, format #1
        cardStr = readCard(fp);
        extractVariable(cardStr, 10, 10, card2c3[layer].ahaze);
        extractVariable(cardStr, 20, 10, card2c3[layer].eqlwcz);
        extractVariable(cardStr, 30, 10, card2c3[layer].rratz);
        extractVariable(cardStr, 40, 5, card2c3[layer].iha1);
        extractVariable(cardStr, 45, 5, card2c3[layer].icld1);
        extractVariable(cardStr, 50, 5, card2c3[layer].ivul1);
        extractVariable(cardStr, 55, 5, card2c3[layer].isea1);
        extractVariable(cardStr, 60, 5, card2c3[layer].ichr1);
      } else if (card2c.ird2 == 2) {
        // read in Card 2C3, format #2
        cardStr = readCard(fp);
        extractVariable(cardStr, 10, 10, card2c3[layer].ahaze);
        extractVariable(cardStr, 30, 10, card2c3[layer].rratz);
        extractVariable(cardStr, 40, 10, card2c3[layer].ahaze1);
        extractVariable(cardStr, 50, 10, card2c3[layer].ahaze2);
        extractVariable(cardStr, 60, 10, card2c3[layer].ahaze3);
        extractVariable(cardStr, 70, 10, card2c3[layer].ahaze4);
      }
    }
  }

  /* Cards 2D, 2D1 and 2D2 are read if IHAZE = 7 or ICLD = 11 */
  if ((card2.ihaze == 7) || (card2.icld == 11) ||
      (strncmp(card2.aruss, "USS", 3) == 0)) {
    cardStr = readCard(fp);
    extractVariable(cardStr, 0, 5, card2d.ireg[0]);
    extractVariable(cardStr, 5, 5, card2d.ireg[1]);
    extractVariable(cardStr, 10, 5, card2d.ireg[2]);
    extractVariable(cardStr, 15, 5, card2d.ireg[3]);

    /* check each if parameters for each region will be read in */
    for (int rr = 0; rr < 4; rr++) {
      if (card2d.ireg[rr] > 0) {
        // check if we have to many coefficients
        if (card2d.ireg[rr] > MAX_COEFF) {
          throw std::runtime_error("unsupported option: CARD2D IREG(" +
                                   std::to_string(rr + 1) +
                                   ") = " + std::to_string(card2d.ireg[rr]) +
                                   " > " + std::to_string(MAX_COEFF));
        }

        // read/parse Card 2D1
        cardStr = readCard(fp);
        extractVariable(cardStr, 0, 10, card2d1[rr].awccon);
        extractVariable(cardStr, 10, 72, card2d1[rr].title);

        /*
         * Each Card2D2 term must be parsed individually
         */
        cardStr = readCard(fp);
        for (int ww = 0; ww < card2d.ireg[rr]; ww++) {
          int wOffset = (ww % 3) * 26;

          extractVariable(cardStr, wOffset + 0, 6, card2d2[rr][ww].vx);
          extractVariable(cardStr, wOffset + 6, 7, card2d2[rr][ww].extc);
          extractVariable(cardStr, wOffset + 13, 7, card2d2[rr][ww].absc);
          extractVariable(cardStr, wOffset + 20, 6, card2d2[rr][ww].asym);

          /* check to absorb the end-of-line character */
          if ((ww + 1) % 3 == 0) {
            cardStr = readCard(fp);
          }
        }
        // cardStr = readCard( fp );
      }
    }
  }

  /* Card 3ALT is only read if IEMSCT == 3 */
  if (card1.iemsct != 3) {
    cardStr = readCard(fp);
    extractVariable(cardStr, 0, 10, card3.h1);
    extractVariable(cardStr, 10, 10, card3.h2);
    extractVariable(cardStr, 20, 10, card3.obszen);
    extractVariable(cardStr, 30, 10, card3.range);
    extractVariable(cardStr, 40, 10, card3.beta);
    extractVariable(cardStr, 50, 10, card3.ro);
    extractVariable(cardStr, 60, 5, card3.lenn);
    extractVariable(cardStr, 70, 10, card3.bckzen);
  } else {
    cardStr = readCard(fp);
    extractVariable(cardStr, 0, 10, card3alt.h1);
    extractVariable(cardStr, 10, 10, card3alt.h2);
    extractVariable(cardStr, 20, 10, card3alt.angle);
    extractVariable(cardStr, 30, 5, card3alt.iday);
    extractVariable(cardStr, 40, 10, card3alt.ro);
    extractVariable(cardStr, 50, 5, card3alt.isourc);
    extractVariable(cardStr, 55, 10, card3alt.anglem);
  }

  /* Card 3A1 and 3A2 are only read if IEMSCT == 2 or 4 */
  if ((card1.iemsct == 2) || (card1.iemsct == 4)) {
    cardStr = readCard(fp);
    extractVariable(cardStr, 0, 5, card3a1.iparm);
    extractVariable(cardStr, 5, 5, card3a1.iph);
    extractVariable(cardStr, 10, 5, card3a1.iday);
    extractVariable(cardStr, 15, 5, card3a1.isourc);

    cardStr = readCard(fp);
    extractVariable(cardStr, 0, 10, card3a2.parm1);
    extractVariable(cardStr, 10, 10, card3a2.parm2);
    extractVariable(cardStr, 20, 10, card3a2.parm3);
    extractVariable(cardStr, 30, 10, card3a2.parm4);
    extractVariable(cardStr, 40, 10, card3a2.time);
    extractVariable(cardStr, 50, 10, card3a2.psipo);
    extractVariable(cardStr, 60, 10, card3a2.anglem);
    extractVariable(cardStr, 70, 10, card3a2.g);
  }

  cardStr = readCard(fp);
  extractVariable(cardStr, 0, 10, card4.v1);
  extractVariable(cardStr, 10, 10, card4.v2);
  extractVariable(cardStr, 20, 10, card4.dv);
  extractVariable(cardStr, 30, 10, card4.fwhm);
  extractVariable(cardStr, 40, card4.yflag);
  extractVariable(cardStr, 41, card4.xflag);
  extractVariable(cardStr, 42, 8, card4.dlimit);
  extractVariable(cardStr, 50, 7, card4.flags);
  extractVariable(cardStr, 57, 3, card4.mlflx);
  extractVariable(cardStr, 60, 10, card4.vrfrac);

  /* check YFLAG and XFLAG */
  if ((card4.yflag == '\n') || (card4.yflag == '\r') || (card4.yflag == '\0')) {
    card4.yflag = ' ';
  }
  if ((card4.xflag == '\n') || (card4.xflag == '\r') || (card4.xflag == '\0')) {
    card4.xflag = ' ';
  }

  /* fill DLIMIT */
  for (int ii = 0; ii < 8; ii++) {
    if ((card4.dlimit[ii] == '\n') || (card4.dlimit[ii] == '\r') ||
        (card4.dlimit[ii] == '\0')) {
      for (int jj = ii; jj < 8; jj++) {
        card4.dlimit[jj] = ' ';
      }
    }
  }
  card4.dlimit[8] = '\0';

  /* fill DFLAGS */
  for (int ii = 0; ii < 7; ii++) {
    if ((card4.flags[ii] == '\n') || (card4.flags[ii] == '\r') ||
        (card4.flags[ii] == '\0')) {
      for (int jj = ii; jj < 7; jj++) {
        card4.flags[jj] = ' ';
      }
    }
  }
  card4.flags[7] = '\0';

  // check if we need to read fancy Card4 stuff
  if (strncmp(card1.surref, "BRDF", 4) == 0) {
    throw std::runtime_error(
        "unsupported option: Not supporting 'BRDF' option (yet)");
  } else if (strncmp(card1.surref, "LAMBER", 6) == 0) {
    throw std::runtime_error(
        "unsupported option: Not supporting 'LAMBER' option (yet)");
  }

  // read in Card5
  cardStr = readCard(fp);
  extractVariable(cardStr, 0, 5, card5.irpt);

  // make sure the stream is in good shape
  if (fp.fail()) {
    throw std::runtime_error(
        "Unexpected end of file reached while reading tape5!");
  }
}

/*
 * Write tape5
 */
void Tape5::write(const std::string &filename, Verbosity verbosity) {
  FILE *fp = nullptr;

#if defined(_MSC_VER) && (_MSC_VER < 1900)
  _set_output_format(_TWO_DIGIT_EXPONENT);
#endif

  /* we need to make sure the lowest CARD 2C1 jives with GNDALT */
  if ((card1.model == 7) && (card1.im == 1)) {
    /* look for the correct lowest altitude */
    int startIndex = 0;
    for (int layerIndex = 0; layerIndex < card2c.ml; layerIndex++) {
      if (card2c1[layerIndex].zmdl < card2.gndalt) {
        startIndex = layerIndex;
      } else {
        break;
      }
    }

    /* build the new atmospheric profiles */
    int shiftIndex = 0;
    for (int layerIndex = startIndex; layerIndex < card2c.ml; layerIndex++) {
      card2c1[shiftIndex] = card2c1[layerIndex];
      card2c2[shiftIndex] = card2c2[layerIndex];
      card2c3[shiftIndex] = card2c3[layerIndex];
      shiftIndex++;
    }
    card2c.ml = shiftIndex;

    /* set the altitude of the lowest layer to GNDALT */
    card2c1[0].zmdl = card2.gndalt;

    /* make sure the first three pressures are different */
    if (fabs(card2c1[0].p - card2c1[1].p) < 0.01) {
      card2c1[1].p += 0.01f;
    }
    if (fabs(card2c1[1].p - card2c1[2].p) < 0.01) {
      card2c1[2].p += 0.01f;
    }
  }

  if (verbosity == VerboseIO) {
    std::cout << std::endl
              << "Writing output card tape5 file ... " << std::flush;
  }

  /* open the output file */
  if ((fp = fopen(filename.c_str(), "w")) == nullptr) {
    throw std::runtime_error(std::string("unable to open file: ") +
                             strerror(errno));
  }

  fprintf(fp, CARD1_OUTPUT, card1.modtrn, card1.speed, card1.cbinout,
          card1.model, card1.itype, card1.iemsct, card1.imult, card1.m1,
          card1.m2, card1.m3, card1.m4, card1.m5, card1.m6, card1.mdef,
          card1.im, card1.noprt, card1.tptemp, card1.surref);

  char card1aMod4Format[] = CARD1A_MOD4_OUTPUT;
  char card1aMod5Format[] = CARD1A_MOD5_OUTPUT;
  char *card1aFormat = card1aMod4Format;
  if (card1a.nstr > 100) {
    card1aFormat = card1aMod5Format;
  }

  fprintf(fp, card1aFormat, card1a.dis, card1a.disazm, card1a.disalb,
          card1a.nstr, card1a.sfwhm, card1a.co2mx, card1a.h2ostr, card1a.o3str,
          card1a.c_prof, card1a.lsunfl, card1a.lbmnam, card1a.lfltnm,
          card1a.h2oaer, card1a.cdtdir, card1a.solcon, card1a.cdastm,
          card1a.astmc, card1a.astmx, card1a.astmo, card1a.aerrh,
          card1a.nssalb);
  fflush(fp);

  /* check to write CARD 1A1 */
  if ((card1a.lsunfl == 'T') || (card1a.lsunfl == 't')) {
    fprintf(fp, "%s\n", card1a1.usrsun);
  }

  /* check to write CARD 1A2 */
  if ((card1a.lbmnam == 'T') || (card1a.lbmnam == 't')) {
    fprintf(fp, "%s\n", card1a2.bmname);
  }

  /* check to write CARD 1A3 */
  if ((card1a.lfltnm == 'T') || (card1a.lfltnm == 't')) {
    fprintf(fp, "%s\n", card1a3.filtnm);
  }

  /* check to write CARD 1A4 */
  if ((card1a.cdtdir == 'T') || (card1a.cdtdir == 't')) {
    fprintf(fp, "%s\n", card1a4.datdir);
  }

  /* check to write CARD 1A5 */
  if ((card1a.c_prof == 1) || (card1a.c_prof == 3) || (card1a.c_prof == 5) ||
      (card1a.c_prof == 7)) {
    for (float ii : card1a5.s_umix)
      fprintf(fp, "%5.3f", ii);
    fprintf(fp, "\n");
  }

  /* check to write CARD 1A6 */
  if ((card1a.c_prof == 2) || (card1a.c_prof == 3) || (card1a.c_prof == 6) ||
      (card1a.c_prof == 7)) {
    for (float ii : card1a6.s_xsec)
      fprintf(fp, "%5.3f", ii);
    fprintf(fp, "\n");
  }

  /* check to write CARD 1A6 */
  if ((card1a.c_prof == 4) || (card1a.c_prof == 5) || (card1a.c_prof == 6) ||
      (card1a.c_prof == 7)) {
    for (float ii : card1a7.s_trac)
      fprintf(fp, "%5.3f", ii);
    fprintf(fp, "\n");
  }

  fflush(fp);
  fprintf(fp, CARD2_OUTPUT, card2.aplus[0], card2.aplus[1], card2.ihaze,
          card2.cnovam, card2.iseasn, card2.aruss[0], card2.aruss[1],
          card2.aruss[2], card2.ivulcn, card2.icstl, card2.icld, card2.ivsa,
          card2.vis, card2.wss, card2.whh, card2.rainrt, card2.gndalt);

  if (strncmp(card2.aplus, "A+", 2) == 0) {
    fprintf(fp, CARD2A_PLUS_OUTPUT, card2a_plus.zaer11, card2a_plus.zaer12,
            card2a_plus.scale1, card2a_plus.zaer21, card2a_plus.zaer22,
            card2a_plus.scale2);
    fprintf(fp, CARD2A_PLUS_OUTPUT, card2a_plus.zaer31, card2a_plus.zaer32,
            card2a_plus.scale3, card2a_plus.zaer41, card2a_plus.zaer42,
            card2a_plus.scale4);
  }

  if ((card2.icld == 18) || (card2.icld == 19)) {
    fprintf(fp, CARD2A_OUTPUT, card2a.cthik, card2a.calt, card2a.cext);
  }

  if ((card2.icld >= 1) && (card2.icld <= 10)) {
    fprintf(fp, CARD2A_ALT_OUTPUT, card2a_alt.cthik, card2a_alt.calt,
            card2a_alt.cext, card2a_alt.ncralt, card2a_alt.ncrspc,
            card2a_alt.cwavln, card2a_alt.ccolwd, card2a_alt.ccolip,
            card2a_alt.chumid, card2a_alt.asymwd, card2a_alt.asymip);
  }

  /* check if we need to read CARD_2B */
  if (card2.ivsa == 1) {
    fprintf(fp, CARD2B_OUTPUT, card2b.zcvsa, card2b.ztvsa, card2b.zinvsa);
  }

  if (((card1.model == 0) || (card1.model == 7)) && (card1.im == 1)) {
    fprintf(fp, CARD2C_OUTPUT, card2c.ml, card2c.ird1, card2c.ird2,
            card2c.hmodel, card2c.ree, card2c.nmolyc, card2c.e_mass,
            card2c.airmwt);
    // check for unsupported modelcular species stuff
    if (card2c.nmolyc > 0) {
      fclose(fp);
      throw std::runtime_error("unsupported feature: NMOLYC in CARD2C is > 0");
    }

    for (int layer = 0; layer < card2c.ml; layer++) {
      fprintf(fp, CARD2C1_OUTPUT, card2c1[layer].zmdl, card2c1[layer].p,
              card2c1[layer].t, card2c1[layer].wmol[0], card2c1[layer].wmol[1],
              card2c1[layer].wmol[2], card2c1[layer].jchar,
              card2c1[layer].jcharx, card2c1[layer].jchary);

      if (card2c.ird1 == 1) {
        fprintf(fp, CARD2C2_OUTPUT, card2c2[layer].wmol[3],
                card2c2[layer].wmol[4], card2c2[layer].wmol[5],
                card2c2[layer].wmol[6], card2c2[layer].wmol[7],
                card2c2[layer].wmol[8], card2c2[layer].wmol[9],
                card2c2[layer].wmol[10], card2c2[layer].wmol[11]);
      }

      if (card2c.ird2 == 1) {
        fprintf(fp, CARD2C3a_OUTPUT, card2c3[layer].ahaze,
                card2c3[layer].eqlwcz, card2c3[layer].rratz,
                card2c3[layer].iha1, card2c3[layer].icld1, card2c3[layer].ivul1,
                card2c3[layer].isea1, card2c3[layer].ichr1);
      }
      if (card2c.ird2 == 2) {
        fprintf(fp, CARD2C3b_OUTPUT, card2c3[layer].ahaze1,
                card2c3[layer].rratz, card2c3[layer].ahaze2,
                card2c3[layer].ahaze3, card2c3[layer].ahaze4);
      }
    }
  }

  /* Cards 2D, 2D1 and 2D2 are read if IHAZE = 7 or ICLD = 11 */
  if ((card2.ihaze == 7) || (card2.icld == 11) ||
      (strncmp(card2.aruss, "USS", 3) == 0)) {
    fprintf(fp, CARD2D_OUTPUT, card2d.ireg[0], card2d.ireg[1], card2d.ireg[2],
            card2d.ireg[3]);

    /* check each if parameters for each region will be written out */
    for (int rr = 0; rr < 4; rr++) {
      if (card2d.ireg[rr] > 0) {
        fprintf(fp, CARD2D1_OUTPUT, card2d1[rr].awccon, card2d1[rr].title);
        for (int ww = 0; ww < card2d.ireg[rr]; ww++) {
          fprintf(fp, CARD2D2_OUTPUT, card2d2[rr][ww].vx, card2d2[rr][ww].extc,
                  card2d2[rr][ww].absc, card2d2[rr][ww].asym);
          if ((ww + 1) % 3 == 0) {
            fprintf(fp, "\n");
          }
        }
        fprintf(fp, "\n");
      }
    }
  }

  /* Card 3ALT is only written if IEMSCT == 3 */
  if (card1.iemsct != 3) {
    fprintf(fp, CARD3_OUTPUT, card3.h1, card3.h2, card3.obszen, card3.range,
            card3.beta, card3.ro, card3.lenn, card3.bckzen);
  } else {
    fprintf(fp, CARD3ALT_OUTPUT, card3alt.h1, card3alt.h2, card3alt.angle,
            card3alt.iday, card3alt.ro, card3alt.isourc, card3alt.anglem);
  }

  /* Card 3A1 and 3A2 are only written if IEMSCT == 2 */
  if ((card1.iemsct == 2) || (card1.iemsct == 4)) {
    fprintf(fp, CARD3A1_OUTPUT, card3a1.iparm, card3a1.iph, card3a1.iday,
            card3a1.isourc);
    fprintf(fp, CARD3A2_OUTPUT, card3a2.parm1, card3a2.parm2, card3a2.parm3,
            card3a2.parm4, card3a2.time, card3a2.psipo, card3a2.anglem,
            card3a2.g);
  }

  // tweak the float formatting based on spectral unit type
  char card4WavenumbersFormat[] = CARD4_WAVENUMBER_OUTPUT;
  char card4MicronFormat[] = CARD4_MICRON_OUTPUT;
  char card4NanometerFormat[] = CARD4_NANOMETER_OUTPUT;

  char *card4FormatStr = card4WavenumbersFormat;

  if (auto c = card4.flags[0]; c == ' ' || c == 'W' || c == 'w') {
    card4FormatStr = card4WavenumbersFormat;
  } else if (c == 'M' || c == 'm') {
    card4FormatStr = card4MicronFormat;
  } else if (c == 'N' || c == 'n') {
    card4FormatStr = card4NanometerFormat;
  } else {
    fclose(fp);
    throw std::runtime_error(std::string("unknown spectral units (") + c +
                             " in CARD4");
  }

  // write card4
  fprintf(fp, card4FormatStr, card4.v1, card4.v2, card4.dv, card4.fwhm,
          card4.yflag, card4.xflag, card4.dlimit[0], card4.dlimit[1],
          card4.dlimit[2], card4.dlimit[3], card4.dlimit[4], card4.dlimit[5],
          card4.dlimit[6], card4.dlimit[7], card4.flags[0], card4.flags[1],
          card4.flags[2], card4.flags[3], card4.flags[4], card4.flags[5],
          card4.flags[6], card4.mlflx, card4.vrfrac);

  if (strncmp(card1.surref, "LAMBER", 6) == 0) {
    fprintf(fp, card4a.fmtStr.c_str(), card4a.nsurf, card4a.aatemp, card4a.dh20,
            card4a.mltrfl);
    fprintf(fp, "%s\n", card4l1.salbfl);
    fprintf(fp, "%s\n", card4l2.csalb);
  }

  fprintf(fp, CARD5_OUTPUT, card5.irpt);

  fclose(fp);
}

/*
 * Adjust radiosonde data.
 */
bool Tape5::adjustRadiosonde(double maxInversionAlt, double airTemp,
                             double airPressure, double dewPoint,
                             bool verboseOutput) {
  int layer;
  int low_layer = 0, layer_index;
  int inversion_layer = 0;
  double maxTemp;

  /* make the boundary temperature (K) equal to the air temperature (C) */
  card1.tptemp = static_cast<float>(airTemp + 273.15);

  /* we need to make sure the lowest CARD 2C1 jives with GNDALT */
  if ((card1.model == 7) && (card1.im == 1)) {
    /* look for the layer closest to GNDALT */
    for (layer = 0; layer < card2c.ml; layer++) {
      if (card2c1[layer].zmdl < card2.gndalt) {
        low_layer = layer + 1;
      }
    }

    // shift the atmospheric layers to match the GNDALT
    if (low_layer > 0) {
      if (verboseOutput) {
        std::cout << "The ground altitude was higher than the lowest layer!\n"
                  << "    Removing lower " << low_layer << "layer(s)\n";
      }

      layer_index = 0;
      for (layer = low_layer; layer < card2c.ml; layer++) {
        card2c1[layer_index] = card2c1[layer];
        layer_index++;
      }
      card2c.ml = layer_index;

      /* set the altitude of the lowest layer to GNDALT */
      card2c1[0].zmdl = card2.gndalt;
    }

    /* make sure the first three pressures are different */
    if (fabs(card2c1[0].p - card2c1[1].p) < 0.01) {
      card2c1[1].p += 0.01f;
    }
    if (fabs(card2c1[1].p - card2c1[2].p) < 0.01) {
      card2c1[2].p += 0.01f;
    }
  }

  if (verboseOutput) {
    std::cout << "Adjusting model atmosphere to match weather data ... ";
  }

  /* check if this is a model atmosphere */
  if (card1.model != 7) {
    if (verboseOutput) {
      std::cout << "done." << std::endl
                << "    Set boundary layer temperature to " << card1.tptemp
                << std::endl;
    }
    return true;
  }

  /* look for the inversion layer by searching for the maximum temp*/
  maxTemp = -std::numeric_limits<double>::max();
  for (layer = 0; layer < card2c.ml; layer++) {
    /* check if we reached the maximum inversion altitude */
    if (card2c1[layer].zmdl > maxInversionAlt) {
      maxTemp = card2c1[layer].t;
      inversion_layer = layer - 1;
      break;
    }
    if (card2c1[layer].t > maxTemp) {
      maxTemp = card2c1[layer].t;
      inversion_layer = layer;
    }
  }

  /* throw out inversion layers that are the lowest or highest */
  if ((inversion_layer == 0) || (inversion_layer == card2c.ml - 1)) {
    if (verboseOutput) {
      std::cout << "done." << std::endl
                << "    No inversion layer identified" << std::endl;
    }
    return true;
  }

  /* get the air temp, pressure and dew point for the current time */
  card2c1[layer].t = static_cast<float>(airTemp);
  card2c1[layer].p = static_cast<float>(airPressure);
  card2c1[layer].wmol[0] = static_cast<float>(dewPoint);

  /* reinterpolate from the ground to the inversion layer */
  for (layer = 1; layer < inversion_layer; layer++) {
    /* air temperature */
    card2c1[layer].t =
        card2c1[0].t + (card2c1[layer].zmdl - card2c1[0].zmdl) *
                           (card2c1[inversion_layer].t - card2c1[0].t) /
                           (card2c1[inversion_layer].zmdl - card2c1[0].zmdl);

    /* air pressure */
    card2c1[layer].p =
        card2c1[0].p + (card2c1[layer].zmdl - card2c1[0].zmdl) *
                           (card2c1[inversion_layer].p - card2c1[0].p) /
                           (card2c1[inversion_layer].zmdl - card2c1[0].zmdl);

    /* dew point */
    card2c1[layer].wmol[0] =
        card2c1[0].wmol[0] +
        (card2c1[layer].zmdl - card2c1[0].zmdl) *
            (card2c1[inversion_layer].wmol[0] - card2c1[0].wmol[0]) /
            (card2c1[inversion_layer].zmdl - card2c1[0].zmdl);
  }

  /* report the guess at the inversion layer */
  if (verboseOutput) {
    std::cout << "done." << std::endl
              << "    Inversion layer altitude = "
              << card2c1[inversion_layer].zmdl << std::endl;
  }

  return true;
}

void Tape5::useLambertianReflectance(const std::string &specAlbFilename,
                                     const std::string &specAlbLabel) {
  if (specAlbFilename.size() > 256) {
    throw std::runtime_error("spectral albedo filename is too long!");
  }

  if (specAlbLabel.size() > 80) {
    throw std::runtime_error("spectral albedo label is too long!");
  }

  strncpy(card1.surref, "LAMBER ", 8);
  card4a.nsurf = 1;
  strncpy(card4l1.salbfl, specAlbFilename.c_str(), specAlbFilename.size());
  strncpy(card4l2.csalb, specAlbLabel.c_str(), specAlbLabel.size());
}

void Tape5::useFilterFunction(const std::string &fltFilename) {
  if (fltFilename.size() > 256) {
    throw std::runtime_error("filter function filename is too long!");
  }

  // enable the filter function
  card1a.lfltnm = 't';
  strncpy(card1a3.filtnm, fltFilename.c_str(), fltFilename.size());
}

namespace {

using namespace d5::atm::modtran;

void SetHaze(const HazeModel &hz, Tape5 &tp5) {
  tp5.card2.ihaze = static_cast<int>(hz.type());
  tp5.card2.vis = static_cast<int>(hz.visibility());
}

void SetHaze(const NavyMaritimeHazeModel &hz, Tape5 &tp5) {
  SetHaze(static_cast<const HazeModel &>(hz), tp5);
  tp5.card2.wss = hz.currentWindSpeedMS;
  tp5.card2.whh = hz.averageWindSpeedMS;
  tp5.card2.icstl = hz.airMassInfluence;
}

void SetHaze(const DesertHazeModel &hz, Tape5 &tp5) {
  SetHaze(static_cast<const HazeModel &>(hz), tp5);
  tp5.card2.wss = hz.currentWindSpeedMS;
}

void SetMultiScat(const MultipleScattering &ms, Tape5 &tp5) {
  tp5.card1.imult = ms.type() != MultipleScattering::Type::None;
  if (ms.type() == MultipleScattering::Type::Isaac) {
    tp5.card1a.dis = 'f';
  }
}

void SetMultiScat(const DisortMultipleScattering &ms, Tape5 &tp5) {
  tp5.card1.imult = 1;
  tp5.card1a.dis = 't';
  tp5.card1a.disazm = 't';
  tp5.card1a.nstr = ms.streams();
}
} // namespace

void Tape5::fromParameters(const d5::atm::modtran::Tape5Parameters &params) {
  using namespace d5::atm::modtran;

  // Handle the atmospheric model
  int atmModel = -1;
  switch (params.atmosphericModel) {
  case ModelType::UserDefinedSingleAltitude:
  case ModelType::UserDefined:
  case ModelType::UserDefinedPressureDependent:
    throw std::runtime_error(
        "user-defined atmospheres are not supported in parameters!");
    break;
  case ModelType::Tropical:
  case ModelType::MidLatitudeSummer:
  case ModelType::MidLatitudeWinter:
  case ModelType::SubArcticSummer:
  case ModelType::SubArcticWinter:
  case ModelType::UsStandard1976:
    atmModel = static_cast<int>(params.atmosphericModel);
    break;
  default:
    throw std::runtime_error("unknown atmospheric model type!");
  }
  card1.model = atmModel;

  // Setup the haze model
  if (params.hazeModel) {
    params.hazeModel->dispatch([this](auto &&hz) { SetHaze(hz, *this); });
  } else {
    card2.ihaze = static_cast<int>(HazeModel::Type::RuralVis23Km);
  }

  // Setup the multiple scattering
  if (params.multipleScattering) {
    params.multipleScattering->dispatch(
        [this](auto &&ms) { SetMultiScat(ms, *this); });
  } else {
    card1.imult = 1;
    card1a.dis = 'f';
  }

  // Setup the CO2 mixing ratio
  card1a.co2mx = params.co2MixingRatioPpm;

  // Setup the (optional) band model name
  if (!params.bandModelName.empty()) {
    card1a.lbmnam = 't';
    strncpy(card1a2.bmname, params.bandModelName.c_str(), 257);
  }

  // Default to using the Mie-generated internal aerosol phase functions
  card3a1.iph = 2;
}

void Tape7Output::readHeader(std::istream &tapeStream, float &h1, float &h2,
                             float &angle, float &range, float &v1, float &v2,
                             float &dv) {
  // read the 11 line header
  std::string line;
  for (size_t lineIndex = 1; lineIndex <= 11; lineIndex++) {
    if (!std::getline(tapeStream, line)) {
      throw std::runtime_error("unexpected end of file while reading header");
    }

    // if this is the 6th line, this is card 3 and we want to extract stuff
    if (lineIndex == 6) {
      std::istringstream cardStream(line);
      cardStream >> h1 >> h2 >> angle >> range;
      continue;
    }

    // if this is the 9th line, this is card 4 and we want to extract stuff
    if (lineIndex == 9) {
      std::istringstream cardStream(line);
      cardStream >> v1 >> v2 >> dv;
      continue;
    }
  }
}

void Tape7Output::readIrradianceData(std::istream &tapeStream,
                                     std::vector<float> &totalIrrad,
                                     std::vector<float> &totalTau) {
#ifdef _DEBUG
  std::cout << "Tape7Output::readIrradianceData" << std::endl;
#endif
  // read the rest of the file, which should be nothing but data
  double lambdaValue, irradValue, tauValue, skipValue;
  std::string line;
  for (size_t ii = 0; ii < totalIrrad.size(); ii++) {
    std::getline(tapeStream, line);
    std::istringstream lineStream(line);
    lineStream >> lambdaValue >> tauValue >> skipValue >> irradValue;

#ifdef _DEBUG
    std::cout << ii << ", " << lambdaValue << ", " << irradValue << ", "
              << tauValue << std::endl;
#endif

    // store the values in the response
    totalIrrad[ii] = irradValue;
    totalTau[ii] = tauValue;
  }
}

/**
 * Summary of columns in a "tape7" file in radiance mode
 * 0  -> Frequency/Wavelength
 * 1  -> Total transmission
 * 2  -> Path thermal
 * 3  -> Thermal scattered (empty with DISORT)
 * 4  -> Surface emission
 * 5  -> Total solar scattered (single + multiple)
 * 6  -> Single solar scattered
 * 7  -> Total ground reflected (includes direct and diffuse)
 * 8  -> Direct reflected (direct only)
 * 9  -> Total radiance
 * For just the path radiance data, we want the "path thermal" and "solar
 * scattered" (combines single and multiple):
 */
void Tape7Output::readPathRadianceData(std::istream &tapeStream,
                                       std::vector<float> &pathRad,
                                       std::vector<float> &pathTau) {
#ifdef _DEBUG
  std::cout << "Tape7Output::readPathRadianceData" << std::endl;
#endif
  // read the rest of the file, which should be nothing but data
  double lambdaValue, trans, thermalPath, thermalScat, surfEmis;
  double solarScat, singleScat, grndRefl, drctRefl, totalRad;
  std::string line;
  for (size_t ii = 0; ii < pathRad.size(); ii++) {
    std::getline(tapeStream, line);
    std::istringstream lineStream(line);
    auto pos = lineStream.tellg();
    lineStream >> lambdaValue >> trans >> thermalPath >> thermalScat;

    // correct if THRML_SCT was a missing (MODTRAN4/MODTRAN5 w/ DISORT)
    if (lineStream.tellg() - pos > 50) {
      surfEmis = thermalScat;
      thermalScat = 0.0;
    } else {
      lineStream >> surfEmis;
    }
    lineStream >> solarScat >> singleScat >> grndRefl >> drctRefl >> totalRad;

#ifdef _DEBUG
    std::cout << ii << std::fixed << ", " << lambdaValue << ", " << std::fixed
              << trans << ", " << std::scientific << thermalPath << ", "
              << std::scientific << thermalScat << ", " << std::scientific
              << solarScat << ", " << std::scientific << singleScat << ", "
              << std::scientific << totalRad << ", " << std::endl;
#endif
    pathRad[ii] = thermalPath + solarScat;
    pathTau[ii] = trans;

    // address SOL_SCAT vs. MULT_SCAT + SING_SCAT changes
    if (solarScat < singleScat) {
      pathRad[ii] += singleScat;
    }
  }
}

void Tape7Output::readTotalRadianceData(std::istream &tapeStream,
                                        std::vector<float> &totalRad,
                                        std::vector<float> &totalTau) {
#ifdef _DEBUG
  std::cout << "Tape7Output::readTotalRadianceData" << std::endl;
#endif
  // read the rest of the file, which should be nothing but data
  double lambdaValue, radValue, tauValue, skipValue;
  std::string line;
  for (size_t ii = 0; ii < totalRad.size(); ii++) {
    std::getline(tapeStream, line);
    std::istringstream lineStream(line);
    auto pos = lineStream.tellg();
    lineStream >> lambdaValue >> tauValue >> skipValue >> skipValue;

    // check if there was a missing column (MODTRAN4/MODTRAN5 w/ DISORT)
    if (lineStream.tellg() - pos < 50) {
      lineStream >> skipValue;
    }
    lineStream >> skipValue >> skipValue >> skipValue >> skipValue;
    lineStream >> radValue;

#ifdef _DEBUG
    std::cout << std::fixed << ii << ", " << lambdaValue << ", "
              << std::scientific << radValue << ", " << tauValue << std::endl;
#endif

    totalRad[ii] = radValue;
    totalTau[ii] = tauValue;
  }
}

void Tape7Output::readSensorAndSurfaceData(std::istream &tapeStream,
                                           std::vector<float> &sensorRad,
                                           std::vector<float> &surfaceRad) {
#ifdef _DEBUG
  std::cout << "Tape7Output::readSensorAndSurfaceData" << std::endl;
#endif
  // read the rest of the file, which should be nothing but data
  float lambdaValue, totalTrans, pathThermal, scatThermal, surfEmis, singScat,
      multScat, grndRefl, drctRefl, totalRad;
  std::string line;
  for (size_t ii = 0; ii < sensorRad.size(); ii++) {
    std::getline(tapeStream, line);
    std::istringstream lineStream(line);
    auto pos = lineStream.tellg();
    lineStream >> lambdaValue >> totalTrans >> pathThermal >> scatThermal;

    // if there was a missing column (MODTRAN4/MODTRAN5 w/ DISORT)
    if (lineStream.tellg() - pos < 50) {
      lineStream >> surfEmis;
    } else {
      // "THRML_SCT" was missing, we read "SURF_EMIS" instead
      surfEmis = scatThermal;
      scatThermal = 0.0;
    }
    lineStream >> singScat >> multScat >> grndRefl >> drctRefl >> totalRad;

    // build up the surface leaving radiance
    double surfRad = grndRefl + surfEmis;

#ifdef _DEBUG
    std::cout << std::fixed << ii << ", " << lambdaValue << ", "
              << std::scientific << totalRad << ", " << surfRad << std::endl;
#endif

    sensorRad[ii] = totalRad;
    surfaceRad[ii] = surfRad;
  }
}

void ChannelOutput::readHeader(std::istream &tapeStream) {
  // read the 5 line header
  std::string line;
  for (size_t lineIndex = 1; lineIndex <= 5; lineIndex++) {
    if (!std::getline(tapeStream, line)) {
      throw std::runtime_error("unexpected end of file while reading header");
    }
  }
}

void ChannelOutput::readIrradianceData(std::istream &tapeStream,
                                       std::vector<float> &totalIrrad,
                                       std::vector<float> &totalTau) {
  // read the rest of the file, which should be nothing but data
  double lambdaValue, irradValue, tauValue, skipValue;
  std::string line;
  for (size_t ii = 0; ii < totalIrrad.size(); ii++) {
    std::getline(tapeStream, line);
    std::istringstream lineStream(line);
    lineStream >> lambdaValue >> tauValue >> skipValue >> irradValue;

#ifdef _DEBUG
    std::cout << ii << ", " << lambdaValue << ", " << irradValue << ", "
              << tauValue << std::endl;
#endif

    // store the values in the response
    totalIrrad[ii] = irradValue;
    totalTau[ii] = tauValue;
  }
}

void ChannelOutput::readRadianceData(std::istream &tapeStream,
                                     std::vector<float> &totalRad,
                                     std::vector<float> &totalTau) {
  // read the rest of the file, which should be nothing but data
  double lambdaValue, widthValue, radValue, tauValue, skipValue;
  std::string line;
  for (size_t ii = 0; ii < totalRad.size(); ii++) {
    std::getline(tapeStream, line);
    std::istringstream lineStream(line);

    lineStream >> lambdaValue >> skipValue >> skipValue >> skipValue >>
        radValue >> skipValue >> skipValue >> skipValue >> widthValue >>
        skipValue >> skipValue >> skipValue >> skipValue >> skipValue >>
        skipValue >> skipValue >> skipValue >> skipValue >> skipValue >>
        skipValue >> skipValue >> tauValue >> skipValue >> skipValue;

#ifdef _DEBUG
    std::cout << std::fixed << ii << ", " << lambdaValue << ", "
              << std::scientific << radValue << ", " << tauValue << std::endl;
#endif

    totalRad[ii] = radValue;
    totalTau[ii] = tauValue;
  }
}
