#pragma once

#include <cassert>
#include <cstring>
#include <istream>
#include <vector>

#include "dirsig5/atmosphere/backends/ModtranCommon.h"

/*
 * General purpose defines.
 */
#define MAX_ANGLES 50
#define MAX_ML 128
#define MAX_REGIONS 4
#define MAX_COEFF 47

/*
 * I/O defines
 * These will be moved in to the card structs themselves as constants.
 */
#define CARD1_OUTPUT "%c%c%c%2d%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d%8.3f%s\n"

/*
 * The CARD 1A input is handled using the extractVariable() functions
 */
#define CARD1A_MOD4_OUTPUT                       \
  "%c%c%c%2d %4.0f%10.3f%10s%10s%1d%c %c %c %c " \
  "%c%10.4f%c%9.4f%10.4f%10.4f%10.4f%10d\n"
#define CARD1A_MOD5_OUTPUT                      \
  "%c%c%c%3d%4.0f%10.3f%10s%10s%1d%c %c %c %c " \
  "%c%10.4f%c%9.4f%10.4f%10.4f%10.4f%10d\n"

/*
 * The CARD 2 input is handled using the extractVariable() functions
 */
#define CARD2_OUTPUT \
  "%c%c%3d%c%4d%c%c%c%2d%5d%5d%5d%10.5f%10.5f%10.5f%10.5f%10.5f\n"

#define CARD2A_PLUS_OUTPUT \
  " %9.0f%9.0f%9.0f                   %9.0f%9.0f%9.0f\n"

#define CARD2A_OUTPUT "%8.3f%8.3f%8.3f\n"
#define CARD2A_ALT_OUTPUT \
  "%8.3f%8.3f%8.3f%4d%4d%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f\n"

#define CARD2B_OUTPUT "%10.3f%10.3f%10.3f\n"

/*
 * The CARD 2C1 input is handled using the extractVariable() functions
 */
#define CARD2C_OUTPUT "%5d%5d%5d%s%10.3f%5d%10.3f%10.3f\n"

/*
 * The CARD 2C1 input is handled using the extractVariable() functions
 */
#define CARD2C1_OUTPUT "%10.6f%10.3e%10.3e%10.3e%10.3e%10.3e%s%c%c\n"
#define CARD2C2_OUTPUT \
  "%10.3e%10.3e%10.3e%10.3e%10.3e%10.3e%10.3e%10.3e\n%10.3e\n"

#define CARD2C3a_OUTPUT "          %10.3e%10.3e%10.3e%5d%5d%5d%5d%5d\n"
#define CARD2C3b_OUTPUT "          %10e          %10e%10e%10e%10e\n"

#define CARD2D_OUTPUT "%5d%5d%5d%5d\n"
#define CARD2D1_OUTPUT "%10.3e%s\n"
#define CARD2D2_OUTPUT "%6.2f %6.4f %6.4f %5.3f"

#define CARD3_OUTPUT "%10.5f%10.5f%10.3f%10.3f%10.3f%10.3f%5d     %10.3f\n"
#define CARD3ALT_OUTPUT "%10.5f%10.5f%10.3f%5d     %10.3f%5d%10.3f\n"

#define CARD3A1_OUTPUT "%5d%5d%5d%5d\n"
#define CARD3A2_OUTPUT "%10.3f%10.3f%10.3f%10.3f%10.3f%10.3f%10.3f%10.3f\n"

#define CARD3B1_OUTPUT "%5d\n"
#define CARD3B2_OUTPUT "%10.3e%10.3e%10.3e%10.3e%10.3e\n"

#define CARD4_WAVENUMBER_OUTPUT \
  "%10.1f%10.1f%10.1f%10.1f%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%3d%10f\n"
#define CARD4_MICRON_OUTPUT \
  "%10.6f%10.6f%10.6f%10.6f%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%3d%10f\n"
#define CARD4_NANOMETER_OUTPUT \
  "%10.0f%10.0f%10.0f%10.0f%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%3d%10f\n"

#define CARD5_OUTPUT "%5d\n"

struct CARD_1 {
  char modtrn = 't';
  char speed = 's';
  char cbinout = 'f';
  char lymolc = ' ';
  int model = 0;
  char tbest = ' ';
  int itype = 0;
  int iemsct = 0;
  int imult = 0;
  int m1 = 0;
  int m2 = 0;
  int m3 = 0;
  int m4 = 0;
  int m5 = 0;
  int m6 = 0;
  int mdef = 0;
  int im = 0;
  int noprt = 0;
  float tptemp = 0;
  char surref[8];

  /// Indicates the type of atmospheric line-of-sight (LOS) path.
  enum PathGeometryType {
    /** Horizontal (constant-pressure) path, i.e. single layer, no refraction
     *  calculation.
     */
    HorizontalPath = 1,

    /// Vertical or slant path between two altitudes
    SlantBetweenTwoAltitudesPath = 2,

    /// Vertical or slant path to space or ground.
    SlantToSpaceOrGroundPath = 3
  };

  /// Determines the mode of execution of MODTRAN.
  enum ExecutionMode {
    /// Spectral transmittance mode
    TransmittanceMode = 0,

    /// Spectral thermal radiance (no sun/moon) mode
    ThermalRadianceMode = 1,

    /** Spectral thermal plus solar/lunar radiance mode (if IMULT=0, only
     *  single scatter solar radiance is included).
     */
    SourceWithThermalRadianceMode = 2,

    /// Directly transmitted solar/lunar irradiance
    SourceIrradianceMode = 3,

    /** Solar/lunar radiance mode with no thermal emission or scatter (if
     *  IMULT=0, only single scatter solar radiance is included).
     */
    SourceWithoutThermalRadianceMode = 4
  };

  CARD_1() { memset(surref, '\0', 8 * sizeof(char)); }

  void setExecutionMode(ExecutionMode executionMode) {
    this->iemsct = executionMode;
  }

  void setPathGeometryType(PathGeometryType pathGeometryType) {
    this->itype = pathGeometryType;
  }
};

struct CARD_1A {
  char dis = 'f';
  char disazm = 'f';
  char disalb = 'f';
  int nstr = 0;
  float sfwhm = 0;
  float co2mx = 0;
  char h2ostr[11];
  char o3str[11];
  int c_prof = 0;
  char lsunfl = 'f';
  char lbmnam = 'f';
  char lfltnm = 'f';
  char h2oaer = ' ';
  char cdtdir = ' ';
  float solcon = 0;
  char cdastm = ' ';
  float astmc = 0;
  float astmx = 0;
  float astmo = 0;
  float aerrh = 0;
  int nssalb = 0;

  CARD_1A() {
    memset(h2ostr, ' ', 11 * sizeof(char));
    h2ostr[10] = '\0';
    memset(o3str, ' ', 11 * sizeof(char));
    o3str[10] = '\0';
  }
};

struct CARD_1A1 {
  char usrsun[257];

  CARD_1A1() {
    memset(usrsun, ' ', 256 * sizeof(char));
    usrsun[256] = '\0';
  }
};

struct CARD_1A2 {
  char bmname[257];

  CARD_1A2() {
    memset(bmname, ' ', 256 * sizeof(char));
    bmname[256] = '\0';
  }
};

struct CARD_1A3 {
  char filtnm[257];

  CARD_1A3() {
    memset(filtnm, ' ', 256 * sizeof(char));
    filtnm[256] = '\0';
  }
};

struct CARD_1A4 {
  char datdir[257];

  CARD_1A4() {
    memset(datdir, ' ', 256 * sizeof(char));
    datdir[256] = '\0';
  }
};

struct CARD_1A5 {
  float s_umix[10];

  CARD_1A5() {
    for (float& ii : s_umix) ii = 1.0;
  }
};

struct CARD_1A6 {
  float s_xsec[13];

  CARD_1A6() {
    for (float& ii : s_xsec) ii = 1.0;
  }
};

struct CARD_1A7 {
  float s_trac[16];

  CARD_1A7() {
    for (float& ii : s_trac) ii = 1.0;
  }
};

struct CARD_1B {
  float awavln;
  float assalb;
};

struct CARD_2 {
  char aplus[2];
  int ihaze = 0;
  char cnovam = ' ';
  int iseasn = 0;
  char aruss[3];
  int ivulcn = 0;
  int icstl = 0;
  int icld = 0;
  int ivsa = 0;
  float vis = 0;
  float wss = 0;
  float whh = 0;
  float rainrt = 0;
  float gndalt = 0;

  CARD_2() {
    memset(aplus, ' ', 2 * sizeof(char));
    memset(aruss, ' ', 3 * sizeof(char));
  }
};

struct CARD_2A_PLUS {
  float zaer11 = 0;
  float zaer12 = 0;
  float scale1 = 0;
  float zaer21 = 0;
  float zaer22 = 0;
  float scale2 = 0;
  float zaer31 = 0;
  float zaer32 = 0;
  float scale3 = 0;
  float zaer41 = 0;
  float zaer42 = 0;
  float scale4 = 0;
};

struct CARD_2A {
  float cthik = 0;
  float calt = 0;
  float cext = 0;
  int iseed = 0;
};

struct CARD_2A_ALT {
  float cthik = 0;
  float calt = 0;
  float cext = 0;
  int ncralt = 0;
  int ncrspc = 0;
  float cwavln = 0;
  float ccolwd = 0;
  float ccolip = 0;
  float chumid = 0;
  float asymwd = 0;
  float asymip = 0;
};

struct CARD_2B {
  float zcvsa = 0;
  float ztvsa = 0;
  float zinvsa = 0;
};

struct CARD_2C {
  int ml = 0;
  int ird1 = 0;
  int ird2 = 0;
  char hmodel[21];
  float ree = 0;
  int nmolyc = 0;
  float e_mass = 0;
  float airmwt = 0;

  CARD_2C() {
    memset(hmodel, ' ', 20 * sizeof(char));
    hmodel[20] = '\0';
  }
};

struct CARD_2C1 {
  float zmdl = 0;
  float p = 0;
  float t = 0;
  float wmol[3];
  char jchar[15];
  char jcharx;
  char jchary;

  CARD_2C1() {
    for (auto& x : wmol) x = 0;
    memset(jchar, ' ', 14 * sizeof(char));
    jchar[14] = '\0';
    jcharx = ' ';
    jchary = ' ';
  }
};

struct CARD_2C2 {
  float wmol[12];

  CARD_2C2() {
    for (auto& x : wmol) x = 0;
  }
};

struct CARD_2C3 {
  float ahaze = 0;
  float eqlwcz = 0;
  float rratz = 0;
  int iha1 = 0;
  int icld1 = 0;
  int ivul1 = 0;
  int isea1 = 0;
  int ichr1 = 0;
  float ahaze1 = 0;
  float ahaze2 = 0;
  float ahaze3 = 0;
  float ahaze4 = 0;
};

struct CARD_2D {
  int ireg[4];

  CARD_2D() {
    for (auto& x : ireg) x = 0;
  }
};

struct CARD_2D1 {
  float awccon = 0;
  char title[73];

  CARD_2D1() {
    memset(title, ' ', 72 * sizeof(char));
    title[72] = '\0';
  }
};

struct CARD_2D2 {
  float vx = 0;
  float extc = 0;
  float absc = 0;
  float asym = 0;
};

struct CARD_3 {
  float h1 = 0;
  float h2 = 0;
  float obszen = 0;
  float range = 0;
  float beta = 0;
  float ro = 0;
  int lenn = 0;
  float bckzen = 0;
};

struct CARD_3ALT {
  float h1 = 0;
  float h2 = 0;
  float angle = 0;
  int iday = 0;
  float ro = 0;
  int isourc = 0;
  float anglem = 0;
};

struct CARD_3A1 {
  int iparm = 0;
  int iph = 0;
  int iday = 0;
  int isourc = 0;
};

struct CARD_3A2 {
  float parm1 = 0;
  float parm2 = 0;
  float parm3 = 0;
  float parm4 = 0;
  float time = 0;
  float psipo = 0;
  float anglem = 0;
  float g = 0;
};

struct CARD_3B1 {
  int nangls = 0;
};

struct CARD_3B2 {
  float angf = 0;
  float f[4];

  CARD_3B2() {
    for (auto& x : f) x = 0;
  }
};

struct CARD_4 {
  float v1 = 0;
  float v2 = 0;
  float dv = 0;
  float fwhm = 0;
  char yflag = ' ';
  char xflag = ' ';
  char dlimit[9];
  char flags[8];
  int mlflx = 0;
  float vrfrac = 0;

  CARD_4() {
    memset(dlimit, ' ', 9 * sizeof(char));
    dlimit[8] = '\0';

    memset(flags, ' ', 8 * sizeof(char));
    flags[7] = '\0';
  }
};

struct CARD_4A {
  int nsurf = 0;
  float aatemp = 0;
  float dh20 = 0;
  char mltrfl = 'F';
  std::string fmtStr = "%1d%9f%9f%c\n";
};

struct CARD_4B1 {
  char cbrdf[81];

  CARD_4B1() { memset(cbrdf, '\0', 81 * sizeof(char)); }
};

struct CARD_4L1 {
  char salbfl[257];

  CARD_4L1() { memset(salbfl, '\0', 257 * sizeof(char)); }
};

struct CARD_4L2 {
  char csalb[81];

  CARD_4L2() { memset(csalb, '\0', 81 * sizeof(char)); }
};

struct CARD_5 {
  int irpt = 0;
};

/**
 * \class Tape5
 *
 * \ingroup modtran
 *
 * MODTRAN tape5 representation.
 */
class Tape5 {
public:
  Tape5() = default;

  /**
   * Type for indicating verbosity of I/O routines.
   */
  enum Verbosity {
    QuietIO,  /**< Quiet I/O */
    VerboseIO /**< Verbose I/O */
  };

public:
  /**
   * Set the name of the file that this class was loaded from.
   */
  void setFilename(const std::string& _filename) { this->filename = _filename; }

  /**
   * Get the name of the file that this class was loaded from.
   */
  const std::string& getFilename() { return this->filename; }

  /**
   * Read the data from a file
   */
  void read(const std::string& _filename, Verbosity _verbosity);

  /**
   * Read the data from a stream
   */
  void read(std::istream& fp);

  /**
   * Write the data to an output file
   */
  void write(const std::string& filename, Verbosity verbosity);

  /**
   * Match weather in tape5 to DIRSIG weather history.
   *
   * \param    maxInversionAlt   Maximum inversion altitude [km]
   * \param    airTemp           Air temperature [C]
   * \param    airPressure       Air pressure [mbar]
   * \param    dewPoint          Dew point [C]
   *
   * \return   tape5 matched to DIRSIG weather.
   */
  bool adjustRadiosonde(
      double maxInversionAlt,
      double airTemp,
      double airPressure,
      double dewPoint,
      bool verboseOutput = true);

  /**
   * Setup the tape5 to use a background reflectance.
   */
  void useLambertianReflectance(
      const std::string& specAlbFilename, const std::string& specAlbLabel);

  /**
   * Setup the tape5 to use a spectral filter function.
   */
  void useFilterFunction(const std::string& fltFilename);

  /**
   * Setup the tape5 using some user-friendly parameters
   */
  void fromParameters(const d5::atm::modtran::Tape5Parameters& params);

private:
  static void fillVariable(char* _var, char _fillValue, int _count);
  static char* readCard(std::istream& inStream);
  static void extractVariable(
      const char* _card, size_t _start, size_t _width, int& _var);
  static void extractVariable(
      const char* _card, size_t _start, size_t _width, float& _var);
  static void extractVariable(
      const char* _card, size_t _start, size_t _width, char* _var);
  static void extractVariable(const char* _card, size_t _start, char& _var);

public:
  std::string filename;
  CARD_1 card1;
  CARD_1A card1a;
  CARD_1A1 card1a1;
  CARD_1A2 card1a2;
  CARD_1A3 card1a3;
  CARD_1A4 card1a4;
  CARD_1A5 card1a5;
  CARD_1A6 card1a6;
  CARD_1A7 card1a7;
  CARD_2 card2;
  CARD_2A_PLUS card2a_plus;
  CARD_2A card2a;
  CARD_2A_ALT card2a_alt;
  CARD_2B card2b;
  CARD_2C card2c;
  CARD_2C1 card2c1[MAX_ML];
  CARD_2C2 card2c2[MAX_ML];
  CARD_2C3 card2c3[MAX_ML];
  CARD_2D card2d;
  CARD_2D1 card2d1[MAX_REGIONS];
  CARD_2D2 card2d2[MAX_REGIONS][MAX_COEFF];
  CARD_3 card3;
  CARD_3ALT card3alt;
  CARD_3A1 card3a1;
  CARD_3A2 card3a2;
  CARD_3B1 card3b1;
  CARD_3B2 card3b2[MAX_ANGLES];
  CARD_4 card4;
  CARD_4A card4a;
  CARD_4L1 card4l1;
  CARD_4L2 card4l2;
  CARD_5 card5;
};

namespace Tape7Output {
/**
 * Read the 11-line header (a simplified "tape5") from the tape7 file.
 */
void readHeader(
    std::istream& tapeStream,
    float& h1,
    float& h2,
    float& angle,
    float& range,
    float& v1,
    float& v2,
    float& dv);

/**
 * Read the remaining data lines in the file from an irradiance run.
 */
void readIrradianceData(
    std::istream& tapeStream,
    std::vector<float>& totalIrrad,
    std::vector<float>& totalTau);

/**
 * Read the remaining data lines in the file from an radiance run
 * and scoop out the path radiance and transmission.
 */
void readPathRadianceData(
    std::istream& tapeStream,
    std::vector<float>& pathRad,
    std::vector<float>& pathTau);

/**
 * Read the remaining data lines in the file from an radiance run
 * and scoop out the total radiance and transmission.
 */
void readTotalRadianceData(
    std::istream& tapeStream,
    std::vector<float>& totalRad,
    std::vector<float>& totalTau);

/**
 * Read the remaining data lines in the file from an radiance run
 * and scoop out the sensor-reaching and surface-leaving radiance.
 */
void readSensorAndSurfaceData(
    std::istream& tapeStream,
    std::vector<float>& totalRad,
    std::vector<float>& surfaceRad);
}

namespace ChannelOutput {
/**
 * Read the 5-line header from the channel file.
 */
void readHeader(std::istream& tapeStream);

/**
 * Read the remaining data lines in the file from an irradiance run.
 */
void readIrradianceData(
    std::istream& tapeStream,
    std::vector<float>& totalIrrad,
    std::vector<float>& totalTau);

/**
 * Read the remaining data lines in the file from an radiance run.
 */
void readRadianceData(
    std::istream& tapeStream,
    std::vector<float>& totalRad,
    std::vector<float>& totalTau);
}
