#include "dirsig5/base/CacheDirectory.h"

#include <fstream>
#include <iostream>
#include <regex>
#include <streambuf>

#include "Config.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/Logger.h"
#include "dirsig5/base/StandardPaths.h"

using namespace std::string_literals;

namespace d5 {

CacheDirectory::CacheDirectory(const std::string &dirPath, bool noPurge)
    : directoryPath_(dirPath), purgeDisabled_(noPurge) {
  if (!directoryPath_.empty())
    DIRSIG_THROW_IF(!stdfs::create_directories(directoryPath_),
                    "Invalid path "s + directoryPath_);

  auto getDirectoryPathFromEnv = [&](const char *varName) {
    const char *var = std::getenv(varName);
    if (!var)
      return;
    if (!stdfs::is_directory(var) and !stdfs::create_directories(var)) {
      DIRSIG_WARN("{}={} is not a usable cache directory!", varName, var);
      return;
    }
    directoryPath_ = var;
  };
  if (directoryPath_.empty())
    getDirectoryPathFromEnv("DIRSIG_CACHE_DIR");
  if (directoryPath_.empty())
    getDirectoryPathFromEnv("SQT_CACHE_DIR");
  if (std::getenv("SQT_CACHE_DIR"))
    DIRSIG_WARN("Setting SQT_CACHE_DIR is deprecated. Use DIRSIG_CACHE_DIR "
                "instead.");

  // Is cache dir still empty?
  stdfs::path docsPath = StandardPaths::Documents();
  if (directoryPath_.empty() and not docsPath.empty()) {
    stdfs::path cachePath = docsPath / "DIRSIG" / "cache";
    if (stdfs::create_directories(cachePath) or
        stdfs::is_directory(cachePath)) {
      directoryPath_ = stdfs::absolute(cachePath).string();
      DIRSIG_INFO("Using default DIRSIG cache location {}. Set "
                  "DIRSIG_CACHE_DIR in your shell environment to "
                  "override.",
                  directoryPath_);
    }
  }

  // If cache dir is still empty, then default to pwd.
  if (directoryPath_.empty()) {
    directoryPath_ = stdfs::current_path().string();
    purgeDisabled_ = true;
    DIRSIG_WARN(
        "Cache defaulting to working directory because {} is not writable. "
        "Purging is disabled. Set DIRSIG_CACHE_DIR in your shell "
        "environment to explicitly choose a cache directory.",
        docsPath.string());
  }
}

bool CacheDirectory::contains(const std::string &filename) const {
  stdfs::path path = directoryPath_;
  path /= filename;
  return stdfs::is_regular_file(path);
}

std::string CacheDirectory::read(const std::string &filename) const {
  DIRSIG_INFO("Cache directory {}: reading {}", directoryPath_, filename);
  stdfs::path path = directoryPath_;
  path /= filename;
  auto ifs = d5::OpenIfstream(path.string(), std::ios_base::binary);
  return std::string((std::istreambuf_iterator<char>(ifs)), //
                     std::istreambuf_iterator<char>());
}

void CacheDirectory::write(const std::string &filename,
                           const std::string &data) const {
  DIRSIG_INFO("Cache directory {}: writing {}", directoryPath_, filename);
  stdfs::path path = directoryPath_;
  path /= filename;
  auto ofs = d5::OpenOfstream(path.string(), std::ios_base::binary);
  ofs.write(data.data(), data.size());
}

static std::regex globToRegex(const std::string &glob) {
  std::string re;
  for (const char *ch = glob.c_str(); ch and *ch != '\0'; ch++) {
    switch (*ch) {
    case '*':
      re += ".*";
      break;
    case '.':
      re += "\\.";
      break;
    case '?':
      re += ".";
      break;
    case '\\':
      re += "\\";
      re += *ch++;
      break;
    default:
      re += *ch;
      break;
    }
  }
  return std::regex(re, std::regex::icase);
}

void CacheDirectory::purge(const std::vector<std::string> &patterns) {
  if (purgeDisabled_) {
    DIRSIG_WARN("Trying to purge {}, but purging is disabled", directoryPath_);
    return;
  }
  std::vector<std::regex> regexes;
  for (const auto &pattern : patterns)
    regexes.push_back(globToRegex(pattern));
  for (const auto &direntry : stdfs::directory_iterator{directoryPath_}) {
    std::string filename = direntry.path().string();
    for (const auto &regex : regexes)
      if (std::regex_match(filename, regex)) {
        DIRSIG_INFO("Cache directory {}: purging {}", directoryPath_, filename);
        stdfs_error_code ec;
        if (!stdfs::remove(filename, ec))
          DIRSIG_WARN("Can't purge {}: {}", filename, ec.message());
      }
  }
}

} // namespace d5
