#include "dirsig5/base/CommandLine.h"

#include <spdlog/fmt/fmt.h>

#include "dirsig5/base/Logger.h"
#include "dirsig5/base/ScopeGuard.h"
#include "dirsig5/sdk_info.h"

#ifdef _WIN32
#define NO_TERMINAL_STYLE 1
#endif // #if _WIN32
#if NO_TERMINAL_STYLE
#define TERMINAL_STYLE(X) ""
#else
#define TERMINAL_STYLE(X) "\033[" X "m"
#endif // #if NO_TERMINAL_STYLE
#define TERMINAL_RESET TERMINAL_STYLE("0")
#define TERMINAL_BOLD TERMINAL_STYLE("1")
#define TERMINAL_DIM TERMINAL_STYLE("2")
#define TERMINAL_ITALIC TERMINAL_STYLE("3")
#define TERMINAL_UNDERLINE TERMINAL_STYLE("4")
#if __linux__
#define TERMINAL_FG_AMBER TERMINAL_STYLE("38;2;255;191;0")
#define TERMINAL_FG_GREEN TERMINAL_STYLE("38;2;135;250;179")
#define TERMINAL_FG_BLUE TERMINAL_STYLE("38;2;135;206;250")
#else
#define TERMINAL_FG_AMBER TERMINAL_STYLE("33")
#define TERMINAL_FG_GREEN TERMINAL_STYLE("32")
#define TERMINAL_FG_BLUE TERMINAL_STYLE("36")
#endif // #if __linux__

namespace d5 {

CommandLine::CommandLine(bool isTopLevel) : mIsTopLevel(isTopLevel) {
  mCommands.emplace_back();
  mCommands.back().argumentGroups.emplace_back();
  if (mIsTopLevel) {
    optionalArgument("-h/--help", [&] {
      mCommands.front().printHelpAndExit(*this);
    }) = "Display this help and exit.";
    optionalArgument("--log_level",
                     [&](std::string str) {
                       setupLogger(LoggerOptions().withMinLevel(
                           spdlog::level::from_str(str)));
                     }) =
        "Sets the minimum logging level (debug, info, warning, error, "
        "critical, off). Defaults to info.";
  } else {
    // Don't complain about other options if not top-level
    mIgnoreUnknown = true;
  }
}

void CommandLine::Command::reset() {
  if (onReset)
    onReset();
  for (Argument &argument : arguments)
    argument.isPresent = false;
  for (Argument &argument : requiredPositionalArguments)
    argument.isPresent = false;
  requiredPositionalArgumentItr = requiredPositionalArguments.begin();
}

void CommandLine::Command::execute() {
  for (Argument &argument : arguments)
    if (argument.isRequired && !argument.isPresent)
      DIRSIG_THROW_FMT("Missing required argument: {}", argument.niceName());
  if (requiredPositionalArgumentItr != requiredPositionalArguments.end() &&
      !dontRequirePositionalArguments)
    DIRSIG_THROW_FMT("Missing required positional arguments!");
  if (onExecute)
    onExecute();
}

static std::string tabOverNewLines(std::string inStr) {
  std::string outStr;
  outStr.reserve(inStr.size());
  for (char ch : inStr) {
    outStr += ch;
    if (ch == '\n')
      outStr += '\t';
  }
  return outStr;
}

void CommandLine::Command::printHelpAndExit(CommandLine &owner) {
  // Usage line:
  {
    std::vector<std::string> usage;
    usage.reserve(8);
    // Usage: my_program
    usage.emplace_back(TERMINAL_DIM "Usage:" TERMINAL_RESET);
    usage.emplace_back(owner.programName());
    // Usage: my_program ... my_command
    if (!name.empty())
      usage.emplace_back(fmt::format(TERMINAL_DIM
                                     "... " TERMINAL_RESET TERMINAL_FG_BLUE
                                     "{}" TERMINAL_RESET,
                                     name));
    usage.emplace_back(TERMINAL_FG_AMBER "[options]" TERMINAL_RESET);
    // Usage: my_program (... my_command) [options] filename
    for (const auto &argument : requiredPositionalArguments)
      usage.emplace_back(fmt::format(TERMINAL_FG_GREEN "{}" TERMINAL_RESET,
                                     argument.names[0]));
    // Usage: my_program (... my_command) [options] filename params+
    if (!defaultPositionalName.empty())
      usage.emplace_back(fmt::format(TERMINAL_FG_GREEN "{}+" TERMINAL_RESET,
                                     defaultPositionalName));
    // Usage: my_program [options] filename params+ <command> ...
    if (name.empty() && owner.commands().size() > 1)
      usage.emplace_back(TERMINAL_FG_BLUE
                         "<command>" TERMINAL_RESET TERMINAL_DIM
                         " ..." TERMINAL_RESET);
    DIRSIG_INFO("{}", fmt::join(usage, " "));
  }
  // Description:
  {
    if (!description.empty())
      DIRSIG_INFO("\n{}\n", description);
  }
  // Options:
  {
    if (!requiredPositionalArguments.empty() ||
        !defaultPositionalName.empty()) {
      DIRSIG_INFO("");
      DIRSIG_INFO(TERMINAL_UNDERLINE "Positional Arguments:" TERMINAL_RESET);
      for (auto &arg : requiredPositionalArguments) {
        DIRSIG_INFO("  " TERMINAL_FG_GREEN "{}" TERMINAL_RESET " " TERMINAL_BOLD
                    "(required!)" TERMINAL_RESET,
                    arg.names[0]);
        if (!arg.helpString.empty())
          DIRSIG_INFO("\t{}", tabOverNewLines(arg.helpString));
      }
      if (!defaultPositionalName.empty()) {
        DIRSIG_INFO("  " TERMINAL_FG_GREEN "{}" TERMINAL_RESET,
                    defaultPositionalName);
        if (!defaultPositionalHelpString.empty())
          DIRSIG_INFO("\t{}", tabOverNewLines(defaultPositionalHelpString));
      }
    }
    if (!arguments.empty()) {
      DIRSIG_INFO("");
      DIRSIG_INFO(TERMINAL_UNDERLINE "Options:" TERMINAL_RESET);
      for (auto &group : argumentGroups) {
        if (!group.name.empty()) {
          DIRSIG_INFO("");
          DIRSIG_INFO("  [" TERMINAL_ITALIC "{}" TERMINAL_RESET "]",
                      group.name);
        }
        for (auto &arg : group.arguments) {
          DIRSIG_INFO(
              "  " TERMINAL_FG_AMBER "{}" TERMINAL_RESET
              " " TERMINAL_DIM TERMINAL_ITALIC "{}" TERMINAL_RESET "{}", //
              arg->niceName(), arg->typeString,
              arg->isRequired ? TERMINAL_BOLD "(required!)" TERMINAL_RESET
                              : "");
          if (!arg->helpString.empty())
            DIRSIG_INFO("\t{}", tabOverNewLines(arg->helpString));
        }
      }
    }
  }
  // Command listing:
  {
    if (name.empty() && owner.commands().size() > 1) {
      DIRSIG_INFO("");
      DIRSIG_INFO(TERMINAL_UNDERLINE "Commands:" TERMINAL_RESET);
      std::vector<std::string_view> commandNames;
      auto itr = owner.commands().begin(); // Skip default
      for (itr++; itr != owner.commands().end(); itr++)
        commandNames.push_back(itr->name);
      std::sort(commandNames.begin(), commandNames.end());
      for (auto &commandName : commandNames)
        DIRSIG_INFO("  " TERMINAL_FG_BLUE "{}" TERMINAL_RESET, commandName);
      DIRSIG_INFO("\nFor command help: {} " TERMINAL_FG_BLUE
                  "<command>" TERMINAL_RESET " " TERMINAL_FG_AMBER
                  "-h" TERMINAL_RESET,
                  owner.programName());
    }
  }
  std::exit(EXIT_SUCCESS);
}

[[nodiscard]] static bool isArgument(std::string_view strv) {
  // ^--?[a-zA-Z](?:-?[a-zA-Z0-9]+)*$
  std::string str = trimmed(strv);
  auto pos = str.begin();
  if (pos == str.end() || *pos++ != '-')
    return false;
  if (*pos == '-')
    ++pos;
  if (!std::isalpha(static_cast<unsigned char>(*pos)))
    return false;
  for (auto tmp = ++pos; tmp != str.end(); pos = ++tmp) {
    if (*tmp == '-')
      ++tmp;
    if (*tmp == '_')
      ++tmp;
    if (!std::isalnum(static_cast<unsigned char>(*tmp)))
      break;
  }
  return pos == str.end();
}

int CommandLine::run(int argc, char **argv) {
  mProgramName = argv[0];
  ++argv;
  --argc;
  if (mIsTopLevel) {
    // Default to --log_level=info.
    setupLogger(LoggerOptions().withMinLevel(spdlog::level::info));
  }
  DIRSIG_TRY {
    auto active = mCommands.begin();
    active->reset();
    while (argc > 0 && std::string_view(*argv) != "--") {
      // Look for '=', if present, truncate *argv.
      char *eq = nullptr;
      auto eqPos = std::string_view(*argv).find('=');
      if (eqPos != std::string_view::npos)
        eq = *argv + eqPos, *eq = '\0';
      auto leaveNoTrace = make_guard([=, backup = *argv] {
        // Be careful here. It is nice to support the optional equals sign
        // like this, but modifying arguments entirely in-place can screw
        // up other parsing logic the user may run separately.
        if (eq)
          *eq = '=';
        *argv = backup;
      });
      if (isArgument(*argv)) {
        bool argOk = false;
        for (Argument &arg : active->arguments) {
          if (arg.names[0] == *argv || arg.names[1] == *argv) {
            if (eq)
              *argv = eq + 1; // Shift.
            else {
              --argc;
              ++argv;
            }
            // This is a backward-compatibility hack to support comma-separated
            // strings of numbers as a single argument, instead of multiple
            // space-separated arguments.
            if (argc > 0 && arg.argc > 1 && arg.isNumericOnly) {
              if (TokenList tokens{*argv, ","};
                  arg.argc == int(tokens.size())) {
                std::vector<std::string> newArgStrs;
                std::vector<char *> newArgv;
                for (auto token : tokens)
                  newArgStrs.push_back(trimmed(token));
                for (auto &argStr : newArgStrs)
                  newArgv.push_back(argStr.data());
                arg.onArgument(newArgv.data());
                arg.isPresent = true;
                argOk = true;
                --argc;
                ++argv;
                break;
              }
            }
            // Not enough args? or expect no args, but ends like --arg=?
            if (arg.argc > argc || (arg.argc == 0 && eq)) {
              if (arg.typeString.empty())
                DIRSIG_THROW_FMT(
                    "{} expects no arguments (should be a pure flag)",
                    arg.niceName());
              else
                DIRSIG_THROW_FMT("{} expects {}", arg.niceName(),
                                 arg.typeString);
            }
            arg.onArgument(argv);
            arg.isPresent = true;
            argOk = true;
            argc -= arg.argc;
            argv += arg.argc;
            break;
          }
        }
        if (!argOk) {
          if (mIgnoreUnknown) {
            DIRSIG_DEBUG("Ignoring argument: {}", *argv);
            argc--;
            argv++;
          } else {
            DIRSIG_THROW_FMT("Unexpected argument: {}", *argv);
          }
        }
      } else {
        if (eq)
          *eq = '='; // Undo truncate.
        bool found = false;
        for (auto itr = mCommands.begin(); itr != mCommands.end(); ++itr) {
          if (itr->name == *argv) {
            active->execute();
            active = itr, active->reset();
            found = true;
            break;
          }
        }
        if (!found) {
          if (active->requiredPositionalArgumentItr !=
              active->requiredPositionalArguments.end()) {
            active->requiredPositionalArgumentItr->onArgument(argv);
            /* argc -= active->requiredPositionalArgumentItr->argc;
               argv += active->requiredPositionalArgumentItr->argc; */
            ++active->requiredPositionalArgumentItr;
          } else {
            if (!active->defaultPositional && !mIgnoreUnknown) {
              DIRSIG_THROW_FMT("Unexpected positional argument: {}", *argv);
            }
            if (active->defaultPositional) {
              active->defaultPositional(std::string_view(*argv));
            }
          }
        }
        --argc;
        ++argv;
      }
    }
    active->execute();
    return EXIT_SUCCESS;
  }
  DIRSIG_CATCH(error, {
    dumpError(error);
    if (mIsTopLevel) {
      DIRSIG_INFO("");
      DIRSIG_INFO("For help: {} " TERMINAL_FG_AMBER "-h" TERMINAL_RESET
                  " (or " TERMINAL_FG_AMBER "--help" TERMINAL_RESET ")",
                  mProgramName);
    }
    return EXIT_FAILURE;
  })
}

} // namespace d5

#undef TERMINAL_RESET
#undef TERMINAL_BOLD
#undef TERMINAL_DIM
#undef TERMINAL_ITALIC
#undef TERMINAL_UNDERLINE
#undef TERMINAL_FG_AMBER
#undef TERMINAL_FG_GREEN
#undef TERMINAL_FG_BLUE
#undef TERMINAL_STYLE
