#include "dirsig5/base/Library.h"

#include "Config.h"
#include "dirsig5/base/Logger.h"
#include "dirsig5/base/StringUtils.h"

#if HAS_UNIX
#include <dlfcn.h>
#else
#include <Windows.h>
#include <strsafe.h>
#endif // #if HAS_UNIX

namespace native {

#if HAS_UNIX

using handle = void *;

static std::string error_string() { return ::dlerror(); }

static handle open(const stdfs::path &filepath) {
  auto flags = 
#if HAS_APPLE
    RTLD_LAZY | RTLD_LOCAL; // DEEPBIND not available on Mac?
#else
    RTLD_LAZY | RTLD_LOCAL | RTLD_DEEPBIND;
#endif
  if (filepath.extension().empty()) {
    auto attempt0 = stdfs::path(filepath).replace_extension("so");
    auto attempt1 = stdfs::path(filepath).replace_extension("dylib");
    if (handle lib = ::dlopen(attempt0.c_str(), flags); lib)
      return lib;
    if (handle lib = ::dlopen(attempt1.c_str(), flags); lib)
      return lib;
  }
  return ::dlopen(filepath.c_str(), flags);
}

static bool close(handle lib) { return ::dlclose(lib) == 0; }

static void *resolve(handle lib, const std::string &name) {
  return ::dlsym(lib, name.c_str());
}

#else // We're on Windows

using handle = HMODULE;

static const handle invalid_handle = nullptr;

static std::string error_string() { return WindowsErrorString(); }

template <typename Char>
static handle WindowsLoadLibrary(const Char *filename) {
  if constexpr (sizeof(Char) == 1)
    return ::LoadLibraryA(filename);
  else
    return ::LoadLibraryW(filename);
}

static handle open(const stdfs::path &filepath) {
  if (filepath.extension().empty()) {
    auto attempt = stdfs::path(filepath).replace_extension("dll");
    attempt.make_preferred(); // Just in case?
    if (handle lib = WindowsLoadLibrary(attempt.c_str()); lib)
      return lib;
  }
  return WindowsLoadLibrary(filepath.c_str());
}

static bool close(handle lib) { return ::FreeLibrary(lib) != 0; }

static void *resolve(handle lib, const std::string &name) {
  return reinterpret_cast<void *>(::GetProcAddress(lib, name.c_str()));
}

#endif // #if HAS_UNIX

} // namespace native

namespace d5 {

std::string Library::error() const { return native::error_string(); }

void Library::open(const std::string &filename) noexcept try {
  filename_ = filename;
  native::handle handle = native::open(stdfs::path(filename).make_preferred());
  if (handle)
    handle_ = handle;
  else
    DIRSIG_DEBUG("Failed to open dynamic library '{}': {}", filename,
                 native::error_string());
} catch (const std::exception &error) {
  DIRSIG_DEBUG("Failed to open dynamic library '{}': {}", filename,
               error.what());
} catch (...) {
  DIRSIG_DEBUG("Failed to open dynamic library '{}': caught unknown exception",
               filename);
}

void Library::close() noexcept try {
  if (handle_.has_value()) {
    if (!native::close(std::any_cast<native::handle>(handle_)))
      DIRSIG_DEBUG("Failed to close dynamic library '{}': {}", filename_,
                   native::error_string());
    handle_.reset();
  }
} catch (const std::exception &error) {
  DIRSIG_DEBUG("Failed to open dynamic library '{}': {}", filename_,
               error.what());
} catch (...) {
  DIRSIG_DEBUG("Failed to close dynamic library '{}': caught unknown exception",
               filename_);
}

void *Library::resolve(const std::string &name) const {
  if (handle_.has_value())
    return native::resolve(std::any_cast<native::handle>(handle_), name);
  else
    return nullptr;
}

} // namespace d5
