#include "dirsig5/base/Logger.h"

#include <map>
#include <sstream>
#include <vector>

namespace d5 {

Logger &Logger::getInstance() {
  static Logger instance;
  return instance;
}

/** Custom formatter to use a different format string per level. Under the hood,
 *  it just dispatches to a number of spdlog::pattern_formatter's.
 *  \ingroup logging
 */
class PerLevelFormatter : public spdlog::formatter {
public:
  /** Constructor
   *
   *  \param basePattern  The default pattern to use if no others are assigned
   *                      to a level.
   */
  explicit PerLevelFormatter(const std::string &basePattern)
      : spdlog::formatter() {
    // Seed the formatters all with the base pattern
    auto res = formatters_.emplace(
        basePattern, std::make_unique<spdlog::pattern_formatter>(basePattern));
    formatterForLevel_.fill(res.first->second.get());
  }

  /** Formats the given message.
   *
   *  \param[in] msg    The message to format
   *  \param[out] dest  The destination memory buffer
   */
  void format(const spdlog::details::log_msg &msg,
              spdlog::memory_buf_t &dest) override {
    // Dispatch to the correct formatter for this level
    formatterForLevel_[msg.level]->format(msg, dest);
  }

  /** Sets the pattern for the given level
   *
   *  \param level    The level for which to set the pattern
   *  \param pattern  The pattern to use at that level
   */
  void setPatternForLevel(spdlog::level::level_enum level,
                          const std::string &pattern) {
    // Make sure this is a valid level
    if (static_cast<size_t>(level) >= formatterForLevel_.size())
      return;

    // If we have already seen this pattern, use the existing formatter, if not
    // make a new one.
    auto it = formatters_.find(pattern);
    if (it != formatters_.end()) {
      formatterForLevel_[level] = it->second.get();
    } else {
      formatters_[pattern] =
          std::make_unique<spdlog::pattern_formatter>(pattern);
      formatterForLevel_[level] = formatters_.at(pattern).get();
    }
  }

  /** Sets the pattern for a range of levels
   *
   *  \param minLevel  The minimum (inclusive) level at which to use the pattern
   *  \param maxLevel  The maximum (inclusive) level at which to use the pattern
   *  \param format    The pattern to use
   */
  void setPatternForLevel(spdlog::level::level_enum minLevel,
                          spdlog::level::level_enum maxLevel,
                          const std::string &format) {
    for (size_t l = (size_t)minLevel; l <= (size_t)maxLevel; l++) {
      setPatternForLevel(static_cast<spdlog::level::level_enum>(l), format);
    }
  }

  /** Clones this formatter */
  [[nodiscard]] std::unique_ptr<spdlog::formatter> clone() const override {
    std::unique_ptr<spdlog::formatter> f(new PerLevelFormatter);

    // Go through each level and set the pattern
    for (size_t i = 0; i < formatterForLevel_.size(); i++) {
      for (const auto &str_and_fmt : formatters_) {
        if (formatterForLevel_[i] == str_and_fmt.second.get()) {
          static_cast<PerLevelFormatter *>(f.get())->setPatternForLevel(
              static_cast<spdlog::level::level_enum>(i), str_and_fmt.first);
        }
      }
    }

    return f;
  }

private:
  // Default constructor (makes an invalid object, but useful for clone())
  PerLevelFormatter() = default;

private:
  // Array of formatters, indexed by level (trace = 0)
  std::array<spdlog::formatter *, spdlog::level::off> formatterForLevel_;

  // Map of pattern to the formatter using that pattern
  std::map<std::string, std::unique_ptr<spdlog::formatter>> formatters_;
};

class PerLevelSink : public spdlog::sinks::sink {
public:
  PerLevelSink(std::unique_ptr<spdlog::sinks::sink> defaultSink) {
    sinks_.push_back(std::move(defaultSink));
    for (auto &i : sinkForLevel_)
      i = 0;
  }

  ~PerLevelSink() override = default;

  void
  SetSinkForLevels(std::unique_ptr<spdlog::sinks::sink> sink,
                   std::initializer_list<spdlog::level::level_enum> levels) {
    auto idx = sinks_.size();
    sinks_.push_back(std::move(sink));
    for (auto l : levels)
      sinkForLevel_[l] = idx;
  }

  void log(const spdlog::details::log_msg &msg) override {
    if (std::find(msg.payload.begin(), msg.payload.end(), '\n') ==
        msg.payload.end())
      sinks_[sinkForLevel_[msg.level]]->log(msg);
    else {
      std::string line = {msg.payload.begin(), msg.payload.end()};
      std::stringstream ss(line);
      while (std::getline(ss, line)) {
        auto tmp = msg;
        tmp.payload = line;
        sinks_[sinkForLevel_[msg.level]]->log(tmp);
      }
    }
  }

  void flush() override {
    for (auto &s : sinks_)
      s->flush();
  }

  void set_pattern(const std::string &pattern) override {
    for (auto &s : sinks_)
      s->set_pattern(pattern);
  }

  void
  set_formatter(std::unique_ptr<spdlog::formatter> sink_formatter) override {
    for (std::size_t i = 0; i < sinks_.size() - 1; i++) {
      sinks_[i]->set_formatter(sink_formatter->clone());
    }
    sinks_.back()->set_formatter(std::move(sink_formatter));
  }

private:
  std::array<std::size_t, spdlog::level::level_enum::n_levels> sinkForLevel_;
  std::vector<std::unique_ptr<spdlog::sinks::sink>> sinks_;
};

void Logger::reinit(const LoggerOptions &options) {
  std::vector<spdlog::sink_ptr> sinks;

  // We want to output source locations on everything >= warnings
  const std::string kLevelFmt = "%^[%l]%$ %v";

  // Make our custom formatter for using the 2 patterns
  std::unique_ptr<spdlog::formatter> format(new PerLevelFormatter(kLevelFmt));
  static_cast<PerLevelFormatter *>(format.get())
      ->setPatternForLevel(spdlog::level::info, "%v");

  // Stderr/stdout sinks
  auto consoleSink = std::make_shared<PerLevelSink>(
      std::make_unique<spdlog::sinks::stderr_color_sink_mt>());
  consoleSink->SetSinkForLevels(
      std::make_unique<spdlog::sinks::stdout_color_sink_mt>(),
      {spdlog::level::info});
  sinks.push_back(std::move(consoleSink));

  // File sink if requested
  if (!options.filename().empty()) {
    sinks.push_back(std::make_shared<spdlog::sinks::basic_file_sink_mt>(
        options.filename()));
  }

  // Make and configure the logger
  logger_ =
      std::make_unique<spdlog::logger>("d5log", sinks.begin(), sinks.end());
  logger_->set_level(options.minLevel());
  logger_->set_formatter(std::move(format));
}

} // namespace d5
