#include "dirsig5/base/StandardPaths.h"

#include <regex>

#include "Config.h"
#include "dirsig5/base/Logger.h"
#include "dirsig5/base/StringUtils.h"

#if HAS_LINUX
#include <unistd.h>
#elif HAS_APPLE
#include <mach-o/dyld.h>
#else // We're on Windows
#include <Windows.h>
#include <strsafe.h>
#endif

namespace d5::StandardPaths {

static std::string GetExePath() {
  std::string str;
  str.resize(8192 + 1);
#if HAS_LINUX
  int ok = int(readlink("/proc/self/exe", str.data(), 8192));
  if (ok < 0)
    return std::string();
  str[ok] = '\0';
#elif HAS_APPLE
  uint32_t size = 8192;
  int ok = _NSGetExecutablePath(str.data(), &size);
  if (ok < 0)
    return std::string();
  str[size] = '\0';
#else // We're on Windows
  DWORD ok = GetModuleFileNameA(NULL, str.data(), 8192);
  if (ok == 0)
    return std::string();
  str[ok] = '\0';
  std::replace(str.data(), str.data() + ok, '\\', '/');
#endif
  return str;
}

static std::string GetApplicationInstall() {
  return stdfs::canonical(stdfs::path(GetExePath()))
      .parent_path()
      .parent_path()
      .string();
}

static std::string ApplicationInstall() {
  static const std::string path = GetApplicationInstall();
  return path;
}

std::string Home() { return stdfs::canonical(std::getenv(HOME_VAR)).string(); }

std::string Documents() { return (stdfs::path(Home()) / "Documents").string(); }

static std::string GetDirsigHome() {
  const char *envVar = std::getenv("DIRSIG_HOME");
  const char *envVarError = nullptr;
  if (envVar && !trimmed(std::string_view(envVar)).empty()) {
    if (!stdfs::exists(envVar))
      envVarError = "does not exist";
    else if (!stdfs::is_directory(envVar))
      envVarError = "is not a directory";
    else if (!stdfs::is_directory(stdfs::path(envVar) / "plugins"))
      envVarError = "does not contain a plugins subdirectory";
    else {
      DIRSIG_DEBUG("Using user-specified 'DIRSIG_HOME={}'", envVar);
      return stdfs::canonical(envVar).string();
    }
  }
  std::string str = ApplicationInstall();
  if (envVarError)
    DIRSIG_WARN("User-specified 'DIRSIG_HOME={}' {}, falling back to "
                "default location '{}'.",
                envVar, envVarError, str);
  else
    DIRSIG_DEBUG("Using 'DIRSIG_HOME={}'", str);
  return str;
}

std::string DirsigHome() {
  static const std::string path = GetDirsigHome();
  return path;
}

std::string DirsigExecutables() {
  auto exePath = stdfs::canonical(stdfs::path(GetExePath())).parent_path();
  return exePath.string();
}

std::string DirsigPlugins(const std::string &plugin) {
  stdfs::path path = DirsigHome();
  path /= "plugins";
  if (!plugin.empty())
    path /= plugin;
  return path.string();
}

static std::string GetUserPlugins() {
  const char *envVar = std::getenv("DIRSIG_PLUGIN_PATH");
  const char *envVarError = nullptr;
  if (envVar && !trimmed(std::string_view(envVar)).empty()) {
    if (!stdfs::exists(envVar))
      envVarError = "does not exist";
    else if (!stdfs::is_directory(envVar))
      envVarError = "is not a directory";
    else {
      DIRSIG_DEBUG("Using user-specified 'DIRSIG_PLUGIN_PATH={}'", envVar);
      return stdfs::canonical(envVar).string();
    }
  }
  stdfs::path path = Documents();
  path /= "DIRSIG";
  path /= "plugins";
  std::string str = path.string();
  if (envVarError)
    DIRSIG_WARN("User-specified 'DIRSIG_PLUGIN_PATH={}' {}, falling back to "
                "default location '{}'.",
                envVar, envVarError, str);
  else
    DIRSIG_DEBUG("Using 'DIRSIG_PLUGIN_PATH={}'", str);
  return str;
}

std::string UserPlugins(const std::string &plugin) {
  static const std::string userPlugins = GetUserPlugins();
  stdfs::path path{userPlugins};
  if (!plugin.empty())
    path /= plugin;
  return path.string();
}

std::string DirsigIniFilename() {
  return (stdfs::path(DirsigHome()) / "dirsig.ini").string();
}

std::string
withReplacements(const std::string &path,
                 const std::unordered_map<std::string, std::string> &what,
                 bool useSysEnv) {
  static const std::regex varRegex(R"(\$\{([^}]+)\}|\$([a-zA-Z_]+))");
  std::ptrdiff_t offset = 0;
  std::string result;
  do {
    std::smatch match;
    if (!std::regex_search(path.begin() + offset, path.end(), match,
                           varRegex)) {
      result += path.substr(offset);
      break;
    }
    result += path.substr(offset, match.position());
    offset += match.position() + match.length();
    std::string varName;
    if (match[1].matched)
      varName = match[1];
    if (match[2].matched)
      varName = match[2];
    if (varName == "DIRSIG_HOME")
      result += DirsigHome();
    else if (auto it = what.find(varName); it != what.end())
      result += it->second;
    else if (auto var = std::getenv(varName.c_str()); var and useSysEnv)
      result += var;
    else
      result += match[0];
  } while (true);
  return result;
}

std::string withNativeSeparators(const std::string &path) {
  return stdfs::path(path).make_preferred().string();
}

} // namespace d5::StandardPaths
