#include "dirsig5/base/Timer.h"

#include <cmath>
#include <iostream>
#include <regex>
#include <sstream>

using namespace std::string_literals;

namespace d5 {

struct TimeUnitInfo {
  char const *suffix;
  double scaleFromS;
};

static constexpr TimeUnitInfo kTimeUnitInfos[] = {
    {"h", 1. / (60 * 60)},
    {"m", 1. / 60},
    {"s", 1},
    {"v", 1000},
};

/** Finds the format specifier for the given time unit in the format string
 *
 *  \param[in] format    The format string
 *  \param[in] timeUnit  The unit to search for
 *  \param[out] match    The match (if found)
 *
 *  \return Whether we found a match
 */
bool FindTimeUnitFormat(const std::string &format, Timer::TimeUnit timeUnit,
                        std::ssub_match &match) {
  // Make a regex that fits the base format plus optional precision
  std::regex pattern(R"(%[0-9.]*)"s + kTimeUnitInfos[timeUnit].suffix);
  std::smatch m;
  if (std::regex_search(format, m, pattern)) {
    match = m[0];
    return true;
  } else
    return false;
}

/** Extracts the given time units out of the given time
 *
 *  \param[in] timeUnit      The time unit to extract
 *  \param[in] smallest       Whether this is the smallest unit that will be
 *                            extracted (leave fractional amount)
 *  \param[in,out] remainder  Remaining time from which to consume [seconds]
 */
double ConsumeTimeUnitAmount(Timer::TimeUnit timeUnit, bool smallest,
                             double &remainder) {
  // Get the conversion to from seconds to this unit of time
  double scale = kTimeUnitInfos[timeUnit].scaleFromS;

  double amount;
  if (smallest) {
    // For the smallest time unit, get the number of whole units and
    // fractional units, clearing the remainder
    amount = remainder * scale;
    remainder = 0;
  } else {
    // For other time units, only get the number of whole units and leave
    // the remainder
    amount = std::floor(remainder * scale);
    remainder -= amount / scale;
  }

  return amount;
}

/** Substitutes in the time for the format specifier
 *
 *  \param[in] timeUnit  The time unit to replace
 *  \param[in] amount    The time amount to substitute
 *  \param[in] match     The match for the format specifier
 *  \param[out] output   The string to format
 */
void ReplaceTimeUnitOutput(Timer::TimeUnit timeUnit, double amount,
                           const std::ssub_match &match, std::string &output) {
  // get the strings for the match and the format
  std::string replacement;
  auto matchStr = match.str();
  auto baseFormat = "%"s + kTimeUnitInfos[timeUnit].suffix;

  if (matchStr.size() == baseFormat.size()) {
    // if this time unit was specified as the base format, just print as an
    // integer
    replacement = std::to_string(static_cast<int>(amount));
  } else {
    // if this time unit had more characters within the base format, pass
    // them along to print as floating point
    auto matchInside = matchStr.substr(1, matchStr.size() - baseFormat.size());
    auto format = "%"s + matchInside + "f"s;
    char buffer[64];
    snprintf(buffer, 64, format.c_str(), amount);
    replacement = buffer;
  }

  // replace the format (find the position manually, as the output has
  // changed)
  auto position = output.find(matchStr);
  output.replace(position, matchStr.size(), replacement);
}

double Timer::duration(TimeUnit timeUnit) const {
  if ((timeUnit >= kTimeUnitCount) || running_) {
    return 0;
  }

  // take the total number of seconds between end and start
  double duration = std::chrono::duration_cast<std::chrono::duration<double>>(
                        endTime_ - beginTime_)
                        .count();

  // Convert to the given time unit
  return duration * kTimeUnitInfos[timeUnit].scaleFromS;
}

std::string Timer::format(const char *format) const {
  if (running_)
    return {};

  std::string search(format);

  // find any matches for the possible time units in the format string,
  // and record the smallest one we found
  std::ssub_match matches[kTimeUnitCount];
  TimeUnit smallestTimeUnit = kTimeUnitCount;
  for (int i = 0; i < kTimeUnitCount; ++i) {
    auto timeUnit = static_cast<TimeUnit>(i);
    if (FindTimeUnitFormat(search, timeUnit, matches[i])) {
      smallestTimeUnit = timeUnit;
    }
  }

  // start with the total number of milliseconds between end and start
  double duration = std::chrono::duration_cast<std::chrono::duration<double>>(
                        endTime_ - beginTime_)
                        .count();
  double remainder = duration;

  std::string output(format);

  // for each time unit found in the format, consume the amount and place
  // it into the output string
  for (int i = 0; i < kTimeUnitCount; ++i) {
    const auto &match = matches[i];
    if (match.matched) {
      auto timeUnit = static_cast<TimeUnit>(i);
      double amount = ConsumeTimeUnitAmount(
          timeUnit, (timeUnit == smallestTimeUnit), remainder);
      ReplaceTimeUnitOutput(timeUnit, amount, match, output);
    }
  }

  return output;
}

} // namespace d5
