#include "dirsig5/base/TruthRequests.h"

#include <algorithm>
#include <iostream>
#include <stdexcept>

#include <spdlog/fmt/fmt.h>

#include "dirsig5/base/StringUtils.h"

bool d5::AddTruthCollector(const std::string &name,
                           const std::vector<d5::Truth::Parameter> &params,
                           std::vector<d5::Truth> &truthRequests) {
  using namespace d5;

  // make an all lower-case, no whitespace version of the name
  auto n = d5::toLower(name);
  n.erase(std::remove(n.begin(), n.end(), ' '), n.end());
  n.erase(std::remove(n.begin(), n.end(), '_'), n.end());
  bool added = false;
  bool isAll = n == "all";
  bool isAtmosphere = n == "atmosphere";
  bool isMaterial = n == "material";
  bool isIntersection = n == "intersection";
  bool isIrradiance = n == "irradiance";
  bool isTextureCoords = n == "texturecoords";
  bool isNormal = n == "normal";
  bool isShadow = n == "shadow";
  bool isGeoLocation = n == "geolocation";
  bool isGeoIds = n == "geoids";
  bool isBarycentricCoords = n == "barycentriccoords";
  bool isPathAngles = n == "pathangles";
  bool isPaths = n == "paths";
  bool isPropagation = n == "propagation";
  bool isMediumIntersection = n == "mediumintersection";
  bool isVolume = (n == "plume" || n == "volume" || n == "cloud");

  auto AddRequest = [&](TruthType t) {
    truthRequests.emplace_back(t);
    added = true;
  };

  auto Check = [&n, isAll, &AddRequest](const auto &expected, TruthType t,
                                        auto... groups) {
    if (n == expected || (groups || ...) || isAll)
      AddRequest(t);
  };
  auto MultiNameCheck =
      [isAll, &AddRequest,
       &n](const std::initializer_list<std::string_view> &expected, TruthType t,
           auto... groups) {
        if ((groups || ...) || isAll) {
          AddRequest(t);
          return;
        }
        for (const auto &p : expected) {
          if (n == p) {
            AddRequest(t);
            return;
          }
        }
      };

  Check("materialindex", TOP_MAT_ID, isMaterial);
  MultiNameCheck({"materialvariantindex", "opticalpropertyindex"}, TOP_OPT_ID,
                 isMaterial);
  Check("scenex", AVG_HIT_X, isIntersection);
  Check("sceney", AVG_HIT_Y, isIntersection);
  Check("scenez", AVG_HIT_Z, isIntersection);
  Check("distance", AVG_DISTANCE, isIntersection);
  Check("viewcosine", AVG_COSINE, isIntersection);
  Check("textureu", AVG_U, isTextureCoords, isIntersection);
  Check("texturev", AVG_V, isTextureCoords, isIntersection);
  Check("normx", AVG_NRM_X, isNormal);
  Check("normy", AVG_NRM_Y, isNormal);
  Check("normz", AVG_NRM_Z, isNormal);
  Check("sunshadow", SUN_SHADOW, isShadow);
  Check("sunexposure", SUN_EXPOSURE, isShadow);
  Check("moonshadow", MOON_SHADOW, isShadow);
  Check("moonexposure", MOON_EXPOSURE, isShadow);
  MultiNameCheck({"shapefactor", "skyfraction"}, SHAPE_FACTOR, isShadow);
  Check("samplecount", SAMPLE_COUNT, isPaths);
  Check("bouncecount", BOUNCE_COUNT, isPaths);
  Check("sourcecount", SOURCE_COUNT, isPaths);
  Check("latitude", AVG_GEO_LAT, isGeoLocation);
  Check("longitude", AVG_GEO_LON, isGeoLocation);
  Check("altitude", AVG_GEO_ALT, isGeoLocation);
  Check("sceneindex", TOP_SCENE_ID, isGeoIds);
  Check("geometryindex", TOP_GEOM_ID, isGeoIds);
  Check("instanceindex", TOP_INST_ID, isGeoIds);
  Check("primitiveindex", TOP_PRIM_ID, isGeoIds);
  Check("temperature", AVG_TEMP);
  Check("baryu", BARY_U, isBarycentricCoords);
  Check("baryv", BARY_V, isBarycentricCoords);
  Check("baryw", BARY_W, isBarycentricCoords);
  Check("pathradiance", PATH_RADIANCE, isAtmosphere);
  Check("pathtransmission", PATH_TRANSMISSION, isAtmosphere);
  Check("pathzenith", PATH_ANGLE_ZENITH, isPathAngles, isAtmosphere);
  Check("pathazimuth", PATH_ANGLE_AZIMUTH, isPathAngles, isAtmosphere);
  Check("pathrightaccension", PATH_RIGHT_ACCENSION, isPathAngles);
  Check("pathdeclination", PATH_DECLINATION, isPathAngles);
  Check("pathlighttime", PATH_LIGHT_TIME, isAtmosphere, isPropagation);
  Check("propstartdirx", PROP_START_DIR_X, isPropagation);
  Check("propstartdiry", PROP_START_DIR_Y, isPropagation);
  Check("propstartdirz", PROP_START_DIR_Z, isPropagation);
  Check("propenddirx", PROP_END_DIR_X, isPropagation);
  Check("propenddiry", PROP_END_DIR_Y, isPropagation);
  Check("propenddirz", PROP_END_DIR_Z, isPropagation);
  Check("solarirradiance", SOLAR_IRRADIANCE, isIrradiance);
  Check("lunarirradiance", LUNAR_IRRADIANCE, isIrradiance);
  Check("mediumx", MED_HIT_X, isMediumIntersection, isVolume);
  Check("mediumy", MED_HIT_Y, isMediumIntersection, isVolume);
  Check("mediumz", MED_HIT_Z, isMediumIntersection, isVolume);
  Check("volpathlength", VOL_PATH_LENGTH, isVolume);
  Check("volconclength", VOL_CONC_LENGTH, isVolume);
  Check("ecefx", AVG_ECEF_X, isGeoLocation);
  Check("ecefy", AVG_ECEF_Y, isGeoLocation);
  Check("ecefz", AVG_ECEF_Z, isGeoLocation);
  Check("convergencestatus", CONVERGENCE_STATUS);
  Check("cloudlayerfraction", CLOUD_LAYER_FRACTION);

  if (n == "abundance") {
    // abundance has parameters so don't put it in "all" and check size
    if (params.empty()) {
      std::string errMsg = "Abundance truth requires a materials list!\n";
      throw std::runtime_error(errMsg);
    }

    for (const auto &param : params) {
      if (param.name == "materials") {
        auto mats = d5::TokenList(param.value, ",");
        for (const auto &mat : mats) {
          auto label = d5::trimmed(mat);
          auto &r = truthRequests.emplace_back(d5::MAT_ABUNDANCE);
          r.parameters.push_back({param.name, label});
        }
      } else if (param.name == "tags") {
        auto tags = d5::TokenList(param.value, ",");
        for (const auto &t : tags) {
          auto tag = d5::trimmed(t);
          auto &r = truthRequests.emplace_back(d5::INST_ABUNDANCE);
          r.parameters.push_back({param.name, tag});
        }
      }
    }

    added = true;
  }

  return added;
}

std::string d5::TruthBandName(const d5::Truth &truth) {
  if (truth.parameters.empty()) {
    return fmt::format(d5::TruthBandNameFmtStr(truth.type));
  } else {
    using ctx = fmt::format_context;
    std::vector<fmt::basic_format_arg<ctx>> fmtArgs;
    for (auto const &a : truth.parameters) {
      fmtArgs.push_back(fmt::detail::make_arg<ctx>(a.value));
    }
    return fmt::vformat(
        d5::TruthBandNameFmtStr(truth.type),
        fmt::format_args(fmtArgs.data(), static_cast<int>(fmtArgs.size())));
  }
}
