#include <gtest/gtest.h>

#include "dirsig5/base/DualQuaternion.h"
#include "dirsig5/base/Geometry.h"

using namespace d5;

/*
    The following tests are ports of DIRSIG4 unit tests
    found in the base_geometry subdirectory.

    They have been ported to use the Eigen and gtest
    framework.  We are number the D4 tests by case numbers.

    Case01 - testTranslationX
    Case02 - testTranslationY
    Case03 - testTranslationZ
    *Case04 - testScaleX
    *Case05 - testScaleY
    *Case06 - testScaleZ
    Case07 - testRotateX
    Case08 - testRotateY
    Case09 - testRotateZ
    x**Case10 - testChain1
    x**Case11 - testChain2

    *NOT IMPLEMENTED
        (Scaling is not part of rigid body rotation/translation)
        (Dynamic scaling will become part of deformable transforms)

    **Modified to remove any scaling steps.
    x Failing tests.
*/

//
// Trivial test defining a rotation and translation Quaternions
//
//    rot( 1, 0, 0, 0) = no rotation
//    translate( 0, 0, 0, 0) = no translation
//
//
TEST(AffineDualQuaternion, AffineDualQuaternionCase00) {
  // Define a test point
  //
  Vector testPoint(0, 0, 1);

  // Define no rotation and no translation
  //
  Quaternion rot(1, 0, 0, 0), translate(0, 0, 0, 0);

  // Define a DualQuaternion data structure
  //
  DualQuaternion qtform(rot, translate);

  // Call the normalize member function.
  //
  qtform.normalize();

  // Apply the translation transformation on the test
  // point
  //

  Vector fPt;
  fPt = qtform.pointForward(testPoint);

  EXPECT_NEAR(testPoint(0), fPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), fPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), fPt(2), 1E-6);

  Vector bPt;
  bPt = qtform.pointInverse(fPt);

  EXPECT_NEAR(testPoint(0), bPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), bPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), bPt(2), 1E-6);
}

TEST(AffineDualQuaternionTest, AffineDualQuaternionTestCase01) {
  // Define a test point
  //
  const Point testPoint(0, 0, 1);

  // Define a rotation transformation variable
  // with no rotation.
  //
  Quaternion rotate(Eigen::AngleAxisd(0, Vector::UnitX()));

  // Define a translation transformation variable
  //
  Vector translate(-3, 0, 0);

  // Define a DualQuaternion to hold the rotation
  // and translation transforms.
  //
  DualQuaternion tform(rotate, translate);
  tform.normalize();

  // Apply the translation transformation on the test
  // point
  //
  Point fPt = tform.pointForward(testPoint);

  // Check the answers
  //
  EXPECT_EQ(-3, fPt(0));
  EXPECT_EQ(0, fPt(1));
  EXPECT_EQ(1, fPt(2));

  // Verify the backwards transform
  //
  Point iPt = tform.pointInverse(fPt);

  // Check the answers
  //
  EXPECT_EQ(testPoint(0), iPt(0));
  EXPECT_EQ(testPoint(1), iPt(1));
  EXPECT_EQ(testPoint(2), iPt(2));
}

TEST(AffineDualQuaternionTest, AffineDualQuaternionTestCase02) {
  // Define a test point
  //
  const Point testPoint(0, 0, 1);

  // Define a rotation transformation variable
  // with no rotation.
  //
  Quaternion rotate(Eigen::AngleAxisd(0, Vector::UnitY()));

  // Define a translation transformation variable
  // for a y-tranlation
  //
  Vector translate(0, 66.4, 0);

  // Define a DualQuaternion to hold the rotation
  // and translation transforms.
  //
  DualQuaternion tform(rotate, translate);
  tform.normalize();

  // Apply the translation transformation on the test
  // point
  //
  Point fPt = tform.pointForward(testPoint);

  // Check the answers
  //
  EXPECT_EQ(0, fPt(0));
  EXPECT_EQ(66.4, fPt(1));
  EXPECT_EQ(1, fPt(2));

  // Verify the backwards transform
  //
  Point iPt = tform.pointInverse(fPt);

  // Check the answers
  //
  EXPECT_EQ(testPoint(0), iPt(0));
  EXPECT_EQ(testPoint(1), iPt(1));
  EXPECT_EQ(testPoint(2), iPt(2));
}

TEST(AffineDualQuaternionTest, AffineDualQuaternionTestCase03) {
  // Define a test point
  //
  const Point testPoint(0, 0, 1);

  // Define a rotation transformation variable
  // with no rotation.
  //
  Quaternion rotate(Eigen::AngleAxisd(0, Vector::UnitZ()));

  // Define a translation transformation variable
  // Specify an z-translation factor
  //
  Vector translate(0, 0, 8.8);

  // Define a DualQuaternion to hold the rotation
  // and translation transforms.
  //
  DualQuaternion tform(rotate, translate);
  tform.normalize();

  // Apply the translation transformation on the test
  // point
  //
  Point fPt = tform.pointForward(testPoint);

  // Check the answers
  //
  EXPECT_EQ(0, fPt(0));
  EXPECT_EQ(0.0, fPt(1));
  EXPECT_EQ(9.8, fPt(2));

  // Verify the backwards transform
  //
  Point iPt = tform.pointInverse(fPt);

  // Check the answers
  //
  EXPECT_EQ(testPoint(0), iPt(0));
  EXPECT_EQ(testPoint(1), iPt(1));
  EXPECT_EQ(testPoint(2), iPt(2));
}

/*

TEST( AffineDualQuaternionTest, AffineDualQuaternionTestCase04){

    // Define a test point
    //
    const Point testPoint( 1,1,1);

    // Define no rotation and no translation
    //
    Quaternion rot( 1,0,0,0), translate(0,0,0,0);


    // Define a DualQuaternion data structure
    //
    DualQuaternion qtform( rot,  translate );

    // Call the normalize member function.
    //
    qtform.normalize();

    // Define a scaling transformation variable
    //
    Eigen::Transform<double,3,Eigen::Affine> tform;


    // Specify an x-scale factor
    //
    tform=Eigen::Scaling(1.2,1.0,1.0);

    // Apply the scaling transformation on the test
    // point
    //
    //Point fPt = qtform.pointForward( testPoint ) * tform;
    Point fPt = tform * qtform.pointForward( testPoint );

    // Check the answers
    //
    EXPECT_EQ( 1.2, fPt(0));
    EXPECT_EQ( 1, fPt(1));
    EXPECT_EQ( 1, fPt(2));

    // Verify the backwards transform
    //
    Eigen::Transform<double,3,Eigen::Affine> tform_inv;
    tform_inv = tform.inverse();

    // Point iPt = tform.pointInverse(fPt) * tform_inv;
    Point iPt = tform_inv * tform.pointInverse(fPt) * tform_inv;

    // Check the answers
    //
    EXPECT_EQ( testPoint(0), iPt(0));
    EXPECT_EQ( testPoint(1), iPt(1));
    EXPECT_EQ( testPoint(2), iPt(2));
}

*/

TEST(AffineDualQuaternionTest, AffineDualQuaternionTestCase07) {
  // Define a test point
  //
  const Point testPoint(0, 0, 1);

  // Define an rotation angle and no translation
  //
  Quaternion rotate, translate(0, 0, 0, 0);
  rotate = Quaternion(Eigen::AngleAxis<double>(PI / 2.0, Vector::UnitX()));

  // Define a DualQuaternion data structure
  //
  DualQuaternion qtform(rotate, translate);

  // Call the normalize member function.
  //
  qtform.normalize();

  // Apply the rotation transformation on the test
  // point
  //
  Point fPt = qtform.pointForward(testPoint);

  // Check the answers
  //
  EXPECT_NEAR(0, fPt(0), 1E-6);
  EXPECT_NEAR(-1, fPt(1), 1E-6);
  EXPECT_NEAR(0, fPt(2), 1E-6);

  // Verify the backwards transform
  //
  Point iPt = qtform.pointInverse(fPt);

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineDualQuaternionTest, AffineDualQuaternionTestCase08) {
  // Define a test point
  //
  const Point testPoint(0, 0, 1);

  // Define a rotation transformation variable and
  // no translation.
  //
  Quaternion rotate, translate(0, 0, 0, 0);
  rotate = Quaternion(Eigen::AngleAxis<double>(-PI / 2.0, Vector::UnitY()));

  // Define a DualQuaternion data structure
  //
  DualQuaternion qtform(rotate, translate);
  qtform.normalize();

  // Apply the scaling transformation on the test
  // point
  //
  Point fPt = qtform.pointForward(testPoint);

  // Check the answers
  //
  EXPECT_NEAR(-1, fPt(0), 1E-6);
  EXPECT_NEAR(0, fPt(1), 1E-6);
  EXPECT_NEAR(0, fPt(2), 1E-6);

  // Verify the backwards transform
  //
  Point iPt = qtform.pointInverse(fPt);

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineDualQuaternionTest, AffineDualQuaternionTestCase09) {
  // Define a test point
  //
  const Point testPoint(0.5, 0, 0);

  // Define a rotation transformation variable
  //
  Quaternion rotate, translate(0, 0, 0, 0);
  rotate = Quaternion(Eigen::AngleAxis<double>(PI / 2.0, Vector::UnitZ()));

  // Define a DualQuaternion data structure
  //
  DualQuaternion qtform(rotate, translate);
  qtform.normalize();

  // Apply the scaling transformation on the test
  // point
  //
  Point fPt = qtform.pointForward(testPoint);

  // Check the answers
  //
  EXPECT_NEAR(0, fPt(0), 1E-6);
  EXPECT_NEAR(0.5, fPt(1), 1E-6);
  EXPECT_NEAR(0, fPt(2), 1E-6);

  // Verify the backwards transform
  //

  Point iPt = qtform.pointInverse(fPt);

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineDualQuaternionTest, AffineDualQuaternionTestCase10a) {
  // The following test is the similar to Case10,
  // but scaling was removed so that a baseline for
  // DualQuaternion tests can be vetted.

  // Define a test point
  //
  const Point testPoint(0, 0, 1);

  // Define No Rotation Quaternion
  //
  Quaternion norotate(1, 0, 0, 0);

  // Define No Translate Quaternion
  //
  Quaternion dqnotranslate(0, 0, 0, 0);

  // Define No Translate Vector
  //
  Vector notranslate(0, 0, 0);

  // Convert into a Quaternion
  //
  Quaternion qxrotate;
  qxrotate = Quaternion(Eigen::AngleAxisd(PI / 2.0, Vector::UnitX()));

  // Convert into a DualQuaternion
  //
  DualQuaternion dqxrotate(qxrotate, notranslate);
  dqxrotate.normalize();

  // Specify a z-rotation factor
  //
  // zrotate=Eigen::AngleAxis<double>(PI,
  //     Vector::UnitZ());

  // Convert into a Quaternion
  //
  Quaternion qzrotate;
  qzrotate = Quaternion(Eigen::AngleAxisd(PI, Vector::UnitZ()));

  // Convert into a DualQuaternion
  //
  DualQuaternion dqzrotate(qzrotate, notranslate);
  dqzrotate.normalize();

  // Specify an x-translation factor
  //
  Vector xtranslate(5, 0, 0);

  // Convert into a Dual Quaternion
  //
  DualQuaternion dqxtranslate(norotate, xtranslate);
  dqxtranslate.normalize();

  // Complete transform
  //
  // tform = xtranslate * zrotate * xrotate;
  DualQuaternion qtform;
  qtform = dqxtranslate * dqzrotate * dqxrotate;
  qtform.normalize();

  // Apply the scaling transformation on the test
  // point.
  //
  Point fPt = qtform.pointForward(testPoint);

  // Check the answers
  //
  EXPECT_NEAR(5, fPt(0), 1E-6);
  EXPECT_NEAR(1, fPt(1), 1E-6);
  EXPECT_NEAR(0, fPt(2), 1E-6);

  // Verify the backwards transform
  //
  Point iPt = qtform.pointInverse(fPt);

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineDualQuaternionTest, AffineDualQuaternionTestCase11a) {
  // Define a test point
  //
  const Point testPoint(1, 1, 0);

  // Define No Rotation Quaternion
  //
  Quaternion norotate(1, 0, 0, 0);

  // Define No Translate Quaternion
  //
  Quaternion dqnotranslate(0, 0, 0, 0);

  // Define No Translate Vector
  //
  Vector notranslate(0, 0, 0);

  // Specify an z-translation factor
  //
  Vector ztranslate(0, 0, -1);

  // Convert into a DualQuaternion
  //
  DualQuaternion dqztranslate(norotate, ztranslate);
  dqztranslate.normalize();

  // Specify an x-rotation as a Quaternion
  //
  Quaternion qxrotate;
  qxrotate = Quaternion(Eigen::AngleAxis<double>(PI, Vector::UnitX()));

  // Specify x-rotation as a DualQuaternion
  //
  DualQuaternion dqxrotate(qxrotate, notranslate);
  dqxrotate.normalize();

  // Specify an x-translation factor
  //
  Vector xtranslate(0.9, 0, 0);

  // Convert into a DualQuaternion
  //
  DualQuaternion dqxtranslate(norotate, xtranslate);
  dqxtranslate.normalize();

  // Specify an z-translation factor
  //
  Vector ztranslate2(0, 0, -1);

  // Convert into a DualQuaternion
  //
  DualQuaternion dqztranslate2(norotate, ztranslate2);
  dqztranslate2.normalize();

  // Specify an y-translation factor
  //
  Vector ytranslate(0, 3, 0);

  // Convert into a DualQuaternion
  //
  DualQuaternion dqytranslate(norotate, ytranslate);
  dqytranslate.normalize();

  // Complete transform
  //
  // tform = ytranslate * yscale * ztranslate2 *
  //      xtranslate * xrotate * xscale *
  //      ztranslate;
  DualQuaternion qtform;
  qtform =
      dqytranslate * dqztranslate2 * dqxtranslate * dqxrotate * dqztranslate;

  // Apply the scaling transformation on the test
  // point.
  //
  Point fPt = qtform.pointForward(testPoint);

  // Check the answers
  //
  EXPECT_NEAR(1.9, fPt(0), 1E-6);
  EXPECT_NEAR(2, fPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), fPt(2), 1E-6);

  // Verify the backwards transform
  //

  Point iPt = qtform.pointInverse(fPt);

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}
