#include <gtest/gtest.h>

#include "dirsig5/base/Constants.h"
#include "dirsig5/base/Geometry.h"

using namespace d5;
/*
        The following tests are ports of DIRSIG4 unit tests
        found in the base_geometry subdirectory.

        They have been ported to use the Eigen and gtest
        framework.  We are number the D4 tests by case numbers.

        Case01 - testTranslationX
        Case02 - testTranslationY
        Case03 - testTranslationZ
        Case04 - testScaleX
        Case05 - testScaleY
        Case06 - testScaleZ
        Case07 - testRotateX
        Case08 - testRotateY
        Case09 - testRotateZ
        Case10 - testChain1
        Case11 - testChain2

        Case10a,Case11a - Same as Case10 and Case11, but scale
                        transformation removed to serve as baseline
                        comparison for DualQuaternion tests.

        *NOT IMPLEMENTED (YET)
        xCurrently Failing Test
*/

TEST(AffineTransformTest, AffineTransformTestCase01) {
  // Define a test point
  const Point testPoint(0, 0, 1);

  // Define a X translation transform
  Eigen::Translation<double, 3> tform(-3.0, 0.0, 0.0);

  // Apply the translation transformation on the test point
  Point fPt = tform * testPoint;

  // Check the answers
  EXPECT_EQ(-3, fPt(0));
  EXPECT_EQ(0, fPt(1));
  EXPECT_EQ(1, fPt(2));

  // Verify the backwards transform
  Eigen::Translation3d tform_inv = tform.inverse();
  Point iPt = tform_inv * tform * testPoint;

  // Check the that the backtransformed point matches the original
  EXPECT_EQ(testPoint(0), iPt(0));
  EXPECT_EQ(testPoint(1), iPt(1));
  EXPECT_EQ(testPoint(2), iPt(2));
}

TEST(AffineTransformTest, AffineTransformTestCase02) {
  // Define a test point
  const Point testPoint(0.0, 0.0, 1.0);

  // Define a Y translation transform
  Eigen::Translation<double, 3> tform(0.0, 66.4, 0.0);

  // Apply the translation transformation on the test point
  Point fPt = tform * testPoint;

  // Check the answers
  EXPECT_EQ(0.0, fPt(0));
  EXPECT_EQ(66.4, fPt(1));
  EXPECT_EQ(1.0, fPt(2));

  // Verify the backwards transform
  Eigen::Translation3d tform_inv = tform.inverse();
  Point iPt = tform_inv * tform * testPoint;

  // Check the that the backtransformed point matches the original
  EXPECT_EQ(testPoint(0), iPt(0));
  EXPECT_EQ(testPoint(1), iPt(1));
  EXPECT_EQ(testPoint(2), iPt(2));
}

TEST(AffineTransformTest, AffineTransformTestCase03) {
  // Define a test point
  const Point testPoint(0.0, 0.0, 1.0);

  // Define a Z translation transform
  Eigen::Translation<double, 3> tform(0.0, 0.0, 8.8);

  // Apply the translation transformation on the test point
  Point fPt = tform * testPoint;

  // Check the answers
  EXPECT_EQ(0.0, fPt(0));
  EXPECT_EQ(0.0, fPt(1));
  EXPECT_EQ(9.8, fPt(2));

  // Verify the backwards transform
  Eigen::Translation3d tform_inv = tform.inverse();
  Point iPt = tform_inv * tform * testPoint;

  // Check the that the backtransformed point matches the original
  EXPECT_EQ(testPoint(0), iPt(0));
  EXPECT_EQ(testPoint(1), iPt(1));
  EXPECT_EQ(testPoint(2), iPt(2));
}

TEST(AffineTransformTest, AffineTransformTestCase04) {
  // Define a test point
  const Point testPoint(1.0, 1.0, 1.0);

  // Define a scaling transformation variable, specify an x-scale factor
  Eigen::Transform<double, 3, Eigen::Affine> tform;
  tform = Eigen::Scaling(1.2, 1.0, 1.0);

  // Apply the scaling transformation on the test point
  Point fPt = tform * testPoint;

  // Check the answers
  EXPECT_EQ(1.2, fPt(0));
  EXPECT_EQ(1, fPt(1));
  EXPECT_EQ(1, fPt(2));

  // Verify the backwards transform
  Eigen::Transform<double, 3, Eigen::Affine> tform_inv;
  tform_inv = tform.inverse();
  Point iPt = tform_inv * tform * testPoint;

  // Check the that the backtransformed point matches the original
  EXPECT_EQ(testPoint(0), iPt(0));
  EXPECT_EQ(testPoint(1), iPt(1));
  EXPECT_EQ(testPoint(2), iPt(2));
}

TEST(AffineTransformTest, AffineTransformTestCase05) {
  // Define a test point
  //
  const Point testPoint(1, 1, 1);

  // Define a scaling transformation variable
  //

  Eigen::Transform<double, 3, Eigen::Affine> tform;

  // Specify an y-scale factor
  //
  tform = Eigen::Scaling(1.0, 0.1, 1.0);

  // Apply the scaling transformation on the test
  // point
  //
  Point fPt = tform * testPoint;

  // Check the answers
  //
  EXPECT_EQ(1, fPt(0));
  EXPECT_EQ(0.1, fPt(1));
  EXPECT_EQ(1, fPt(2));

  // Verify the backwards transform
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform_inv;
  tform_inv = tform.inverse();

  Point iPt = tform_inv * tform * testPoint;

  // Check the answers
  //
  EXPECT_EQ(testPoint(0), iPt(0));
  EXPECT_EQ(testPoint(1), iPt(1));
  EXPECT_EQ(testPoint(2), iPt(2));
}

TEST(AffineTransformTest, AffineTransformTestCase06) {
  // Define a test point
  //
  const Point testPoint(1, 1, 1);

  // Define a scaling transformation variable
  //

  Eigen::Transform<double, 3, Eigen::Affine> tform;

  // Specify an z-scale factor
  //
  tform = Eigen::Scaling(1.0, 1.0, 3.8);

  // Apply the scaling transformation on the test
  // point
  //
  Point fPt = tform * testPoint;

  // Check the answers
  //
  EXPECT_EQ(1, fPt(0));
  EXPECT_EQ(1, fPt(1));
  EXPECT_EQ(3.8, fPt(2));

  // Verify the backwards transform
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform_inv;
  tform_inv = tform.inverse();

  Point iPt = tform_inv * tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineTransformTest, AffineTransformTestCase07) {
  // Define a test point
  //
  const Point testPoint(0, 0, 1);

  // Define a rotation transformation variable
  //

  Eigen::Transform<double, 3, Eigen::Affine> tform;

  // Specify an x-rotation factor
  //
  tform = Eigen::AngleAxis<double>(PI / 2.0, Vector::UnitX());

  // Apply the scaling transformation on the test
  // point
  //
  Point fPt = tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(0, fPt(0), 1E-6);
  EXPECT_NEAR(-1, fPt(1), 1E-6);
  EXPECT_NEAR(0, fPt(2), 1E-6);

  // Verify the backwards transform
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform_inv;
  tform_inv = tform.inverse();

  Point iPt = tform_inv * tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineTransformTest, AffineTransformTestCase08) {
  // Define a test point
  //
  const Point testPoint(0, 0, 1);

  // Define a rotation transformation variable
  //

  Eigen::Transform<double, 3, Eigen::Affine> tform;

  // Specify an y-rotation factor
  //
  tform = Eigen::AngleAxis<double>(-PI / 2.0, Vector::UnitY());

  // Apply the scaling transformation on the test
  // point
  //
  Point fPt = tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(-1, fPt(0), 1E-6);
  EXPECT_NEAR(0, fPt(1), 1E-6);
  EXPECT_NEAR(0, fPt(2), 1E-6);

  // Verify the backwards transform
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform_inv;
  tform_inv = tform.inverse();

  Point iPt = tform_inv * tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineTransformTest, AffineTransformTestCase09) {
  // Define a test point
  //
  const Point testPoint(0.5, 0, 0);

  // Define a rotation transformation variable
  //

  Eigen::Transform<double, 3, Eigen::Affine> tform;

  // Specify an z-rotation factor
  //
  tform = Eigen::AngleAxis<double>(PI / 2.0, Vector::UnitZ());

  // Apply the scaling transformation on the test
  // point
  //
  Point fPt = tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(0, fPt(0), 1E-6);
  EXPECT_NEAR(0.5, fPt(1), 1E-6);
  EXPECT_NEAR(0, fPt(2), 1E-6);

  // Verify the backwards transform
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform_inv;
  tform_inv = tform.inverse();

  Point iPt = tform_inv * tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineTransformTest, AffineTransformTestCase10) {
  // Define a test point
  //
  const Point testPoint(0, 0, 1);

  // Define a rotation transformation variable
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform, xrotate, zrotate;

  // Specify an x-rotation factor
  //
  xrotate = Eigen::AngleAxis<double>(PI / 2.0, Vector::UnitX());

  // Specify a z-rotation factor
  //
  zrotate = Eigen::AngleAxis<double>(PI, Vector::UnitZ());

  // Specify an x-translation factor
  //
  Eigen::Translation<double, 3> xtranslate(0, 0, 0);
  xtranslate.x() = 5.0;

  // Define a y-scaling transformation variable
  //
  Eigen::Transform<double, 3, Eigen::Affine> yscale;

  // Specify an y-scale factor
  //
  yscale = Eigen::Scaling(1.0, 0.25, 1.0);

  // Complete transform
  //
  tform = yscale * xtranslate * zrotate * xrotate;

  // Apply the scaling transformation on the test
  // point.
  //
  Point fPt = tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(5, fPt(0), 1E-6);
  EXPECT_NEAR(0.25, fPt(1), 1E-6);
  EXPECT_NEAR(0, fPt(2), 1E-6);

  // Verify the backwards transform
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform_inv;
  tform_inv = tform.inverse();

  Point iPt = tform_inv * tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineTransformTest, AffineTransformTestCase10a) {
  // The following test is the similar to Case10,
  // but scaling was removed so that a baseline for
  // DualQuaternion tests can be vetted.

  // Define a test point
  //
  const Point testPoint(0, 0, 1);

  // Define a rotation transformation variable
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform, xrotate, zrotate;

  // Specify an x-rotation factor
  //
  xrotate = Eigen::AngleAxis<double>(PI / 2.0, Vector::UnitX());

  // Specify a z-rotation factor
  //
  zrotate = Eigen::AngleAxis<double>(PI, Vector::UnitZ());

  // Specify an x-translation factor
  //
  Eigen::Translation<double, 3> xtranslate(5, 0, 0);

  // Complete transform
  //
  tform = xtranslate * zrotate * xrotate;

  // Apply the transformation on the test
  // point.
  //
  Point fPt = tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(5, fPt(0), 1E-6);
  EXPECT_NEAR(1, fPt(1), 1E-6);
  EXPECT_NEAR(0, fPt(2), 1E-6);

  // Verify the backwards transform
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform_inv;
  tform_inv = tform.inverse();

  Point iPt = tform_inv * tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineTransformTest, AffineTransformTestCase11) {
  // Define a test point
  //
  const Point testPoint(1, 1, 0);

  // Define a transformation variable
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform, xrotate;

  // Specify an z-translation factor
  //
  Eigen::Translation<double, 3> ztranslate;
  ztranslate.x() = 0.0;
  ztranslate.y() = 0.0;
  ztranslate.z() = -1.0;

  // Define a x-scaling transformation variable
  //
  Eigen::Transform<double, 3, Eigen::Affine> xscale;
  // Specify an x-scale factor
  //
  xscale = Eigen::Scaling(0.1, 1.00, 1.00);

  // Specify an x-rotation factor
  //
  xrotate = Eigen::AngleAxis<double>(PI, Vector::UnitX());

  // Specify an x-translation factor
  //
  Eigen::Translation<double, 3> xtranslate;
  xtranslate.x() = 0.9;
  xtranslate.y() = 0.0;
  xtranslate.z() = 0.0;

  // Specify an z-translation factor
  //
  Eigen::Translation<double, 3> ztranslate2;
  ztranslate2.x() = 0.0;
  ztranslate2.y() = 0.0;
  ztranslate2.z() = -1.0;

  // Define a y-scaling transformation variable
  //
  Eigen::Transform<double, 3, Eigen::Affine> yscale;
  // Specify an y-scale factor
  //
  yscale = Eigen::Scaling(1.0, 2.0, 1.0);

  // Specify an y-translation factor
  //
  Eigen::Translation<double, 3> ytranslate;
  ytranslate.x() = 0.0;
  ytranslate.y() = 3.0;
  ytranslate.z() = 0.0;

  // Complete transform
  //
  tform = ytranslate * yscale * ztranslate2 * xtranslate * xrotate * xscale *
          ztranslate;

  // Apply the scaling transformation on the test
  // point.
  //
  Point fPt = tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), fPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), fPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), fPt(2), 1E-6);

  // Verify the backwards transform
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform_inv;
  tform_inv = tform.inverse();

  Point iPt = tform_inv * tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}

TEST(AffineTransformTest, AffineTransformTestCase11a) {
  // Define a test point
  //
  const Point testPoint(1, 1, 0);

  // Define a transformation variable
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform, xrotate;

  // Specify an z-translation factor
  //
  Eigen::Translation<double, 3> ztranslate(0, 0, -1);

  // Specify an x-rotation factor
  //
  xrotate = Eigen::AngleAxis<double>(PI, Vector::UnitX());

  // Specify an x-translation factor
  //
  Eigen::Translation<double, 3> xtranslate(0.9, 0, 0);

  // Specify an z-translation factor
  //
  Eigen::Translation<double, 3> ztranslate2(0, 0, -1);

  // Specify an y-translation factor
  //
  Eigen::Translation<double, 3> ytranslate(0.0, 3.0, 0.0);

  // Complete transform
  //
  tform = ytranslate * ztranslate2 * xtranslate * xrotate * ztranslate;

  // Apply the scaling transformation on the test
  // point.
  //
  Point fPt = tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(1.9, fPt(0), 1E-6);
  EXPECT_NEAR(2, fPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), fPt(2), 1E-6);

  // Verify the backwards transform
  //
  Eigen::Transform<double, 3, Eigen::Affine> tform_inv;
  tform_inv = tform.inverse();

  Point iPt = tform_inv * tform * testPoint;

  // Check the answers
  //
  EXPECT_NEAR(testPoint(0), iPt(0), 1E-6);
  EXPECT_NEAR(testPoint(1), iPt(1), 1E-6);
  EXPECT_NEAR(testPoint(2), iPt(2), 1E-6);
}
