#pragma once

#include <array>
#include <cmath>
#include <utility>
#include <vector>

#include <Eigen/Geometry>
#include <gtest/gtest.h>

namespace d5 {

template <typename T, int N>
::testing::AssertionResult IsNear(
  const Eigen::Matrix<T, N, 1>& test,
  const Eigen::Matrix<T, N, 1>& truth,
  T eps) {
  for (int i = 0; i < N; i++) {
    auto diff = std::abs(test[i] - truth[i]);
    if (diff > eps) {
      return ::testing::AssertionFailure()
             << "Test value at index " << i << " (" << test[i]
             << ") differs from truth value (" << truth[i] << ") by " << diff
             << ", which is greater than the threshold of " << eps;
    }
  }
  return ::testing::AssertionSuccess();
}

/** Custom predicate for testing if two arrays are near to each other */
template <typename T, size_t N>
::testing::AssertionResult IsNear(
  const std::array<T, N>& test, const std::array<T, N>& truth, T eps) {
  for (size_t i = 0; i < N; i++) {
    auto diff = std::abs(test[i] - truth[i]);
    if (diff > eps) {
      return ::testing::AssertionFailure()
             << "Test value at index " << i << " (" << test[i]
             << ") differs from truth value (" << truth[i] << ") by " << diff
             << ", which is greater than the threshold of " << eps;
    }
  }
  return ::testing::AssertionSuccess();
}

/** Custom predicate for testing if two vectors are near to each other */
template <typename T>
::testing::AssertionResult IsNear(
  const std::vector<T>& test, const std::vector<T>& truth, T eps) {
  if (test.size() != truth.size()) {
    return ::testing::AssertionFailure()
           << "Test vector had a different size (" << test.size()
           << ") than truth vector (" << truth.size() << ")";
  }

  for (size_t i = 0; i < test.size(); i++) {
    auto diff = std::abs(test[i] - truth[i]);
    if (diff > eps) {
      return ::testing::AssertionFailure()
             << "Test value at index " << i << " (" << test[i]
             << ") differs from truth value (" << truth[i] << ") by " << diff
             << ", which is greater than the threshold of " << eps;
    }
  }
  return ::testing::AssertionSuccess();
}

template <typename T>
::testing::AssertionResult IsNear(
  const std::pair<T, T>& test, const std::pair<T, T>& truth, T eps) {
  auto diff = std::abs(test.first - truth.first);
  if (diff > eps) {
    return ::testing::AssertionFailure()
           << "First item in pair (" << test.first
           << ") differs from truth value (" << truth.first << ") by " << diff
           << ", which is greater than the threshold of " << eps;
  }

  diff = std::abs(test.second - truth.second);
  if (diff > eps) {
    return ::testing::AssertionFailure()
           << "Second item in pair (" << test.second
           << ") differs from truth value (" << truth.second << ") by " << diff
           << ", which is greater than the threshold of " << eps;
  }

  return ::testing::AssertionSuccess();
}

}  // namespace d5
