#include <fstream>
#include <iostream>
#include <random>
#include <stdexcept>
#include <string_view>

#if USE_BOOST_FILESYSTEM
#include <boost/filesystem.hpp>
namespace stdfs = boost::filesystem;
using stdfs_error_code = boost::system::error_code;
#else
#include <filesystem>
namespace stdfs = std::filesystem;
using stdfs_error_code = std::error_code;
#endif

namespace d5 {

/** Creates a filename for a temporary file with the given extension */
inline std::string TempFilename(std::string_view ext) {
  thread_local static std::mt19937 rng(std::random_device{}());
  thread_local static std::uniform_int_distribution<int> pick(0, 26);

  auto p = stdfs::temp_directory_path();
  std::string name(9, 'a');
  for (auto& c : name) c += pick(rng);
  p /= name + std::string(ext);
  return p.string();
}

/** A simple wrapper around ofstream that removes the file upon exiting scope */
class TempOfstream {
public:
  // We only support ofstream constructors that open the file. They throw if
  // we can't open the file
  explicit TempOfstream(const char* filename)
      : ofs_(filename), filename_(filename) {
    ThrowIfNotOpen();
  }
  explicit TempOfstream(const std::string& filename)
      : ofs_(filename), filename_(filename) {
    ThrowIfNotOpen();
  }

  /** Destructor, deletes the file */
  ~TempOfstream() {
    stdfs_error_code ec;
    if (!stdfs::remove(filename_, ec))
      std::cerr << "Failed to remove file \"" << filename_ << "\"" << std::endl;
  }

  /** Returns the filename */
  [[nodiscard]] const std::string& filename() const { return filename_; }

  /** Closes (but doesn't delete the file) */
  void close() { ofs_.close(); }

  /** Pass-through of the output operator to the ofstream */
  template <typename T>
  friend TempOfstream& operator<<(TempOfstream& os, const T& x) {
    os.ofs_ << x;
    return os;
  }

private:
  void ThrowIfNotOpen() {
    if (!ofs_.is_open()) {
      throw std::runtime_error("Unable to open file: \"" + filename_ + "\"");
    }
  }

private:
  std::ofstream ofs_;
  std::string filename_;
};

/** Helper function for making a temporary file with the given extension */
inline auto MakeTempFile(std::string_view ext) {
  return TempOfstream(TempFilename(ext));
}

}  // namespace d5
