#include <gtest/gtest.h>

#include "dirsig5/base/DateTime.h"

using namespace d5;

/**
 * Test the leap year function.
 */
TEST(CLDateTimeTest, CLDateTimeIsLeapYearTests) {
  EXPECT_EQ(DateTime::isLeapYear(1800), false);
  EXPECT_EQ(DateTime::isLeapYear(1900), false);
  EXPECT_EQ(DateTime::isLeapYear(1984), true);
  EXPECT_EQ(DateTime::isLeapYear(2000), true);
  EXPECT_EQ(DateTime::isLeapYear(2096), true);
  EXPECT_EQ(DateTime::isLeapYear(2100), false);
  EXPECT_EQ(DateTime::isLeapYear(2201), false);
}

/**
 * Test the number of days in a year.
 */
TEST(CLDateTimeTest, CLDateTimeDaysInYearTests) {
  EXPECT_EQ(DateTime::daysInYear(1800), 365);
  EXPECT_EQ(DateTime::daysInYear(1900), 365);
  EXPECT_EQ(DateTime::daysInYear(1988), 366);
  EXPECT_EQ(DateTime::daysInYear(2000), 366);
  EXPECT_EQ(DateTime::daysInYear(2092), 366);
  EXPECT_EQ(DateTime::daysInYear(2100), 365);
  EXPECT_EQ(DateTime::daysInYear(2201), 365);
}

/**
 * Test the date to seconds conversion.
 * Truth values computed using https://www.epochconverter.com.
 */
TEST(CLDateTimeTest, CLDateTimeDateToSecondsTest) {
  EXPECT_EQ(DateTime::dateToSeconds(1, 1, 1970), 0);
  EXPECT_EQ(DateTime::dateToSeconds(1, 1, 1980), 315532800);
  EXPECT_EQ(DateTime::dateToSeconds(1, 7, 2000), 962409600);
  EXPECT_EQ(DateTime::dateToSeconds(14, 1, 2018), 1515888000);
}

/**
 * Test the seconds to date conversion.
 * Truth values computed using https://www.epochconverter.com.
 */
TEST(CLDateTimeTest, CLDateTimeSecondsToDateTests) {
  int dayOfYear, day, month, year;

  DateTime::secondsToDate(16761600, day, month, year, dayOfYear);
  ASSERT_TRUE((day == 14) && (month == 7) && (year == 1970));
  ASSERT_TRUE(dayOfYear == 195);

  DateTime::secondsToDate(662601600, day, month, year, dayOfYear);
  ASSERT_TRUE((day == 31) && (month == 12) && (year == 1990));
  ASSERT_TRUE(dayOfYear == 365);

  DateTime::secondsToDate(962409600, day, month, year, dayOfYear);
  ASSERT_TRUE((day == 1) && (month == 7) && (year == 2000));
  ASSERT_TRUE(dayOfYear == 183);

  DateTime::secondsToDate(1515888000, day, month, year, dayOfYear);
  ASSERT_TRUE((day == 14) && (month == 1) && (year == 2018));
  ASSERT_TRUE(dayOfYear == 14);
}

/**
 * Test the converting dates/times to total seconds.
 * Truth values computed using https://www.epochconverter.com.
 */
TEST(CLDateTimeTest, CLDateTimeTotalSecondsTests) {
  DateTime dateTime;

  // 31-Dec-2000 @ 11:59:59 PM GMT/UTC
  dateTime.set(2000, 12, 31, 23, 59, 59, 0);
  EXPECT_EQ(978307199.0, dateTime.getTotalSeconds());

  // 31-Dec-2000 @ 11:59:59 PM EST (UTC-5)
  dateTime.set(2000, 12, 31, 23, 59, 59, -5);
  EXPECT_EQ(978325199.0, dateTime.getTotalSeconds());

  // 31-Dec-2000 @ 11:59:59 PM EST (UTC+8)
  dateTime.set(2000, 12, 31, 23, 59, 59, +8);
  EXPECT_EQ(978278399.0, dateTime.getTotalSeconds());

  // 14-Jul-2020 @ 04:59:59 AM EST (UTC-4)
  dateTime.set(2020, 7, 14, 4, 59, 59, -4);
  EXPECT_EQ(1594717199.0, dateTime.getTotalSeconds());

  // 20-Jul-1970 @ 08:17:01.2345 PM EST (UTC)
  dateTime.set(1970, 7, 20, 20, 17, 1.2345, 0);
  EXPECT_EQ(17353021.2345, dateTime.getTotalSeconds());
}

/**
 * Test the absolute mutator and ISO8601 conversion.
 */
TEST(CLDateTimeTest, CLDateTimeToIso8601Tests) {
  DateTime dateTime;

  // the start of the epoch
  dateTime.set(1970, 1, 1, 0, 0, 0, 0);
  EXPECT_EQ("1970-01-01T00:00:00.0000+00:00", dateTime.toIso8601());

  dateTime.set(2000, 12, 31, 5, 23, 59.59, -5);
  EXPECT_EQ("2000-12-31T05:23:59.5900-05:00", dateTime.toIso8601());
}

/**
 * Test the absolute mutator and ISO8601 conversion.
 */
TEST(CLDateTimeTest, CLDateTimeToFromIso8601Tests) {
  DateTime dateTime;
  std::string dateTimeStr;

  dateTimeStr = "2000-12-31T05:23:59.5900-05:30";
  dateTime = DateTime::fromIso8601(dateTimeStr);
  EXPECT_EQ(dateTimeStr, dateTime.toIso8601());

  dateTimeStr = "2000-12-31T05:23:59.5900+08:15";
  dateTime = DateTime::fromIso8601(dateTimeStr);
  EXPECT_EQ(dateTimeStr, dateTime.toIso8601());
}

/**
 * Test the absolute mutator to absolute seconds
 */
TEST(CLDateTimeTest, CLDateTimeFromIso8601Tests) {
  DateTime dateTime;

  // start of the epoch (negative 0 offset)
  dateTime = DateTime::fromIso8601("1970-01-01T00:00:00.0000-00:00");
  EXPECT_EQ(0.0, dateTime.getTotalSeconds());

  // start of the epoch (positive 0 offset)
  dateTime = DateTime::fromIso8601("1970-01-01T00:00:00.0000+00:00");
  EXPECT_EQ(0.0, dateTime.getTotalSeconds());

  // start of epoch at 1200
  dateTime = DateTime::fromIso8601("1970-01-01T12:00:00.0000-00:00");
  EXPECT_EQ(43200.0, dateTime.getTotalSeconds());

  // start of epoch at 1200 -05:00
  dateTime = DateTime::fromIso8601("1970-01-01T12:00:00.0000-05:00");
  EXPECT_EQ(61200.0, dateTime.getTotalSeconds());

  // 20-Jul-1970 @ 08:17:01.2345 PM EST (UTC)
  dateTime = DateTime::fromIso8601("1970-07-20T20:17:01.2345+00:00");
  EXPECT_EQ(17353021.2345, dateTime.getTotalSeconds());
}
