#include <random>

#include <gtest/gtest.h>

#include "dirsig5/base/DualQuaternion.h"
#include "dirsig5/base/Pcg.h"

using namespace d5;

TEST(DualQuaternionTest, RoundTripConversion) {
  // Initialize some dual quaternions.
  DualQuaternion dq1 = {Quaternion{+1, +2, +3, -4}, Quaternion{+5, -6, +7, +8}};
  DualQuaternion dq2 = {Quaternion{-1, -1, +7, +2}, Quaternion{+5, -9, +2, -1}};
  dq1.addScaling({1.23, 1.23, 1.23});
  dq2.addScaling({2.41, 2.41, 2.41});

  // Product.
  DualQuaternion dq = dq1 * dq2;

  {
    // Round-trip.
    auto u = dq.pointInverse(dq.pointForward({1, 2, 3}));
    auto v = dq.pointForward(dq.pointInverse({1, 2, 3}));
    EXPECT_NEAR(1, u[0], 1e-5);
    EXPECT_NEAR(2, u[1], 1e-5);
    EXPECT_NEAR(3, u[2], 1e-5);
    EXPECT_NEAR(1, v[0], 1e-5);
    EXPECT_NEAR(2, v[1], 1e-5);
    EXPECT_NEAR(3, v[2], 1e-5);

    // Chaining.
    auto u1 = dq.pointForward({1, 2, 3});
    auto v1 = dq.pointInverse({1, 2, 3});
    auto u2 = dq1.pointForward(dq2.pointForward({1, 2, 3}));
    auto v2 = dq2.pointInverse(dq1.pointInverse({1, 2, 3}));
    EXPECT_NEAR(u1[0], u2[0], 1e-5);
    EXPECT_NEAR(u1[1], u2[1], 1e-5);
    EXPECT_NEAR(u1[2], u2[2], 1e-5);
    EXPECT_NEAR(v1[0], v2[0], 1e-5);
    EXPECT_NEAR(v1[1], v2[1], 1e-5);
    EXPECT_NEAR(v1[2], v2[2], 1e-5);
  }

  {
    // Round-trip.
    auto u = dq.vectorInverse(dq.vectorForward({1, 2, 3}));
    auto v = dq.vectorForward(dq.vectorInverse({1, 2, 3}));
    EXPECT_NEAR(1, u[0], 1e-5);
    EXPECT_NEAR(2, u[1], 1e-5);
    EXPECT_NEAR(3, u[2], 1e-5);
    EXPECT_NEAR(1, v[0], 1e-5);
    EXPECT_NEAR(2, v[1], 1e-5);
    EXPECT_NEAR(3, v[2], 1e-5);

    // Chaining.
    auto u1 = dq.vectorForward({1, 2, 3});
    auto v1 = dq.vectorInverse({1, 2, 3});
    auto u2 = dq1.vectorForward(dq2.vectorForward({1, 2, 3}));
    auto v2 = dq2.vectorInverse(dq1.vectorInverse({1, 2, 3}));
    EXPECT_NEAR(u1[0], u2[0], 1e-5);
    EXPECT_NEAR(u1[1], u2[1], 1e-5);
    EXPECT_NEAR(u1[2], u2[2], 1e-5);
    EXPECT_NEAR(v1[0], v2[0], 1e-5);
    EXPECT_NEAR(v1[1], v2[1], 1e-5);
    EXPECT_NEAR(v1[2], v2[2], 1e-5);
  }
}

TEST(DualQuaternionTest, MicroRotations) {
  {
    int seed = testing::UnitTest::GetInstance()->random_seed();
    RandEngine gen(seed);
    std::uniform_real_distribution<double> uniform;
    auto makeQuaternion = [](double theta, double axisX, double axisY,
                             double axisZ) {
      double cosTheta = std::cos(theta / 2);
      double sinTheta = std::sin(theta / 2);
      double len = std::sqrt(axisX * axisX + axisY * axisY + axisZ * axisZ);
      axisX /= len;
      axisY /= len;
      axisZ /= len;
      return Quaternion{cosTheta, sinTheta * axisX, sinTheta * axisY,
                        sinTheta * axisZ};
    };
    double theta1 = uniform(gen) * 2 * PI;
    double theta2 = theta1 + (2 * uniform(gen) - 1) * 1e-6;
    double axisX = 2 * uniform(gen) - 1;
    double axisY = 2 * uniform(gen) - 1;
    double axisZ = 2 * uniform(gen) - 1;
    Vector location1 = {1000 * (2 * uniform(gen) - 1),
                        1000 * (2 * uniform(gen) - 1),
                        1000 * (2 * uniform(gen) - 1)};
    Vector location2 = {location1[0] + (2 * uniform(gen) - 1) * 1e-6,
                        location1[1] + (2 * uniform(gen) - 1) * 1e-6,
                        location1[2] + (2 * uniform(gen) - 1) * 1e-6};
    if (location1.norm() == 0.0 || location2.norm() == 0.0) { // pedantic check
      location1 = Vector{1, 0, 0};
      location2 = Vector{1 + 1e-6, 0, 0};
    }
    DualQuaternion dq1(makeQuaternion(theta1, axisX, axisY, axisZ), location1);
    DualQuaternion dq2(makeQuaternion(theta2, axisX, axisY, axisZ), location2);
    DualQuaternion dq = ScLERP(dq1, dq2, uniform(gen));

    // All normalized!
    EXPECT_NEAR(dq1.real.norm(), 1.0, 1e-6);
    EXPECT_NEAR(dq2.real.norm(), 1.0, 1e-6);
    EXPECT_NEAR(dq.real.norm(), 1.0, 1e-6);

    // Interpolated dual part should be very close to end-point dual parts.
    const auto &dual1 = dq1.dual.coeffs();
    const auto &dual2 = dq2.dual.coeffs();
    const auto &dual = dq.dual.coeffs();
    EXPECT_NEAR(1 + (dual - dual1).norm() / dual1.norm(), 1.0, 1e-6);
    EXPECT_NEAR(1 + (dual - dual2).norm() / dual2.norm(), 1.0, 1e-6);
  }

  {
    // L3H problem case
    DualQuaternion dq1;
    DualQuaternion dq2;
    dq1.real = {-0.707107, 0.000180, 0.000035, 0.707107};
    dq2.real = {-0.707107, 0.000181, 0.000034, 0.707107};
    dq1.dual = {-176.790878, -53.667645, -57.149372, -176.774368};
    dq2.dual = {-176.790985, -53.692169, -57.173817, -176.774429};
    for (int k = 0; k < 100; k++) {
      double t = 0.2 + k / 100.0 * 0.6;
      DualQuaternion dq = ScLERP(dq1, dq2, t);
      EXPECT_NEAR(dq.dual.w(), dq1.dual.w(), 0.05);
      EXPECT_NEAR(dq.dual.x(), dq1.dual.x(), 0.05);
      EXPECT_NEAR(dq.dual.y(), dq1.dual.y(), 0.05);
      EXPECT_NEAR(dq.dual.z(), dq1.dual.z(), 0.05);
    }
  }
}
