#include <gtest/gtest.h>

#include "dirsig5/base/Json.h"

/**
 * Test the Eigen functionality
 */
TEST(JsonTest, EigenTest) {
  // Test a matrix
  {
    Eigen::Matrix3d m;
    m << 1e0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8;

    d5::Json j = m;
    ASSERT_TRUE(j.is_array());
    ASSERT_EQ(j.size(), 3u);
    for (int r = 0; r < m.rows(); r++) {
      const auto &row = j.at(r);
      ASSERT_TRUE(j.is_array());
      ASSERT_EQ(j.size(), 3u);
      for (int c = 0; c < m.cols(); c++) {
        EXPECT_NEAR(row.at(c), m(r, c), 1e-10);
      }
    }

    Eigen::Matrix3d read;
    ASSERT_NO_THROW(j.get_to(read));
    for (int r = 0; r < read.rows(); r++) {
      for (int c = 0; c < read.cols(); c++) {
        EXPECT_NEAR(read(r, c), m(r, c), 1e-10);
      }
    }
  }

  // Test a vector
  {
    Eigen::Vector3d v;
    v << 1, 2, 3.5;
    d5::Json j = v;
    ASSERT_TRUE(j.is_array());
    for (int i = 0; i < 3; i++) {
      ASSERT_TRUE(j.at(i).is_number());
      EXPECT_NEAR(j.at(i).get<double>(), v(i), 1e-10);
    }

    Eigen::Vector3d read;
    ASSERT_NO_THROW(j.get_to(read));
    for (int r = 0; r < read.rows(); r++) {
      EXPECT_NEAR(read(r), v(r), 1e-10);
    }

    // try the 2D vector read
    d5::Json j2d = "[[1], [2], [3.5]]";
    ASSERT_NO_THROW(j.get_to(read));
    for (int r = 0; r < read.rows(); r++) {
      EXPECT_NEAR(read(r), v(r), 1e-10);
    }
  }

  // Test a type alias for a vector
  {
    using Banana = Eigen::Vector3d;
    Banana v;
    v << 1, 2, 3.5;
    d5::Json j = v;
    ASSERT_TRUE(j.is_array());
    for (int i = 0; i < 3; i++) {
      ASSERT_TRUE(j.at(i).is_number());
      EXPECT_NEAR(j.at(i).get<double>(), v(i), 1e-10);
    }

    Banana read;
    ASSERT_NO_THROW(j.get_to(read));
    for (int r = 0; r < read.rows(); r++) {
      EXPECT_NEAR(read(r), v(r), 1e-10);
    }
  }

  // Test a quaternion
  {
    Eigen::Quaterniond q(5.8, 3.2, 1.6, 9.3);
    d5::Json j = q;
    ASSERT_TRUE(j.is_array());
    ASSERT_EQ(j.size(), 4);
    EXPECT_NEAR(j[0], q.w(), 1e-10);
    EXPECT_NEAR(j[1], q.x(), 1e-10);
    EXPECT_NEAR(j[2], q.y(), 1e-10);
    EXPECT_NEAR(j[3], q.z(), 1e-10);

    Eigen::Quaterniond read;
    ASSERT_NO_THROW(j.get_to(read));
    EXPECT_NEAR(read.w(), q.w(), 1e-10);
    EXPECT_NEAR(read.x(), q.x(), 1e-10);
    EXPECT_NEAR(read.y(), q.y(), 1e-10);
    EXPECT_NEAR(read.z(), q.z(), 1e-10);
  }
}
