#include <gtest/gtest.h>

#include "dirsig5/base/Geometry.h"

using namespace d5;
/*
        The following tests are ports of DIRSIG4 unit tests
        found in the base_geometry subdirectory.

        They have been ported to use the Eigen and gtest
        framework.  We are number the D4 tests by case numbers.

        Case01 - testRotationVecVec
        Case02 - testRotationVecVecClose
        Case03 - testRocationVecVecOpposite
        *Case04 - testRotationVecVecRandom
        Case05 - testRotationAxis1
        Case06 - testRotationAxis2
        *Case07 - testSetPointingTform
        Case08 - testMakeScaling
        Case09 - testMakeTranslation
        Case10 - testMakeXYZRotation
        Case11 - testEulerConversion1
        Case12 - testEulerConversion2
        Case13 - testEulerConversionNorthSingular
        Case14 - testEulerConversionSouthSingular



        *NOT IMPLEMENTED (YET)

*/

/**
 * Test vector-vector make rotation.
 */
TEST(CLMatrixTest, CLMatrixTestCase01) {
  Vector fromVec(0.0, 1.0, 0.0);
  Vector toVec(0.0, 0.0, 1.0);

  Matrix rotationMtx = makeRotation(fromVec, toVec);

  Vector tformVec = rotationMtx * fromVec;

  // Check elements to see if they resolve to the same
  // value
  //

  //
  //	The following ASSERT fails because of precision
  //	issues.  This is resolved in the next statememtns.
  //
  //      ASSERT_EQ( 1, tformVec == toVec );

  ASSERT_DOUBLE_EQ(tformVec(0), toVec(0));
  ASSERT_NEAR(tformVec(1), toVec(1), 1E-6);
  ASSERT_DOUBLE_EQ(tformVec(2), toVec(2));
}

/**
 * Test vector-vector make rotation for close vectors.
 */
TEST(CLMatrixTest, CLMatrixTestCase02) {
  Vector fromVec(0.01, 0.0, 1.0);
  fromVec.normalize();
  Vector toVec(0.0, 0.0, 1.0);

  fromVec.normalize();

  Matrix rotationMtx = makeRotation(fromVec, toVec);

  Vector tformVec = rotationMtx * fromVec;

  // Check elements to see if they resolve to the same
  // value
  //
  EXPECT_NE(fromVec, toVec);

  EXPECT_NEAR(tformVec(0), toVec(0), 1E-6);
  EXPECT_NEAR(tformVec(1), toVec(1), 1E-6);
  EXPECT_NEAR(tformVec(2), toVec(2), 1E-6);
}

/**
 * Test vector-vector make rotation vectors 180 degrees apart.
 */
TEST(CLMatrixTest, CLMatrixTestCase03) {
  Vector fromVec(0.00, 0.0, 1.0);
  Vector toVec(0.00, 0.0, -1.0);

  Matrix rotationMtx = makeRotation(fromVec, toVec);

  Vector tformVec = rotationMtx * fromVec;

  ASSERT_EQ(tformVec, toVec);
}

/**
 * Test make axis rotation.
 */
TEST(CLMatrixTest, CLMatrixTestCase05) {
  Vector rotationAxis(0.0, 1.0, 0.0);
  Matrix axisRotation = makeRotation(rotationAxis, -PI / 2.0);

  Point testPt(1.0, 0.0, 0.0);
  testPt = axisRotation * testPt;

  //
  // The call below fails because of precision issues (I think) RR
  //    EXPECT_EQ( Point( 0.0, 0.0, 1.0 ) ,  testPt );

  EXPECT_NEAR(0.0, testPt(0), 1E-6);
  EXPECT_NEAR(0.0, testPt(1), 1E-6);
  EXPECT_NEAR(1.0, testPt(2), 1E-6);
}

/**
 * Test make axis rotation.
 */
TEST(CLMatrixTest, CLMatrixTestCase06) {
  Vector rotationAxis(1.0, 1.0, 0.0);
  rotationAxis.normalize();
  Matrix axisRotation = makeRotation(rotationAxis, -PI / 2.0);

  Point testPt(0.5, -0.5, 0.0);
  testPt = axisRotation * testPt;

  //    ASSERT_EQ(1,  Point(0.0, 0.0, 0.707107)== testPt);

  EXPECT_NEAR(0.0, testPt(0), 1e-6);
  EXPECT_NEAR(0.0, testPt(1), 1e-6);
  EXPECT_NEAR(0.707107, testPt(2), 1e-6);
}

/**
 * Test pointing transform setup
 */
/*
TEST( CLMatrixTest, CLMatrixTestCase07 ) {

    Vector focalPlaneDefault( 0, 0, -1 );

    Vector globalUp( 0, 1, 0 );
    Vector globalUp2( 1, 0, 0 );
    Vector focalPlaneNew( 0, 0.25, -1 );

    // Should result in identity matrix
    Matrix tformMtx;
    tformMtx.set( focalPlaneNew, globalUp );

    Matrix tformMtx2;
    tformMtx2.set( focalPlaneNew, globalUp2 );

    // Transform focal plane nomical direction by tform
    Vector tformVec = tformMtx * focalPlaneDefault;
    Vector tformVec2 = tformMtx2 * focalPlaneDefault;

    CPPUNIT_ASSERT( tformVec == focalPlaneNew );
    CPPUNIT_ASSERT( tformVec2 == focalPlaneNew );
}
*/

/**
 * Test make scaling matrix.
 */
TEST(CLMatrixTest, CLMAtrixTestCase08) {
  Eigen::Transform<double, 3, Eigen::Affine> sclMtx;
  sclMtx = Eigen::Scaling(2.0, 3.0, 4.0);

  Point testPt(1.0, 1.0, 1.0);
  testPt = sclMtx * testPt;

  ASSERT_EQ(Point(2.0, 3.0, 4.0), testPt);
}

/**
 * Test make translation matrix.
 */
TEST(CLMatrixTest, CLMatrixTestCase09) {
  Eigen::Transform<double, 3, Eigen::Affine> tnsMtx;
  tnsMtx = Eigen::Translation<double, 3>(-1.0, 2.2, -4.0);

  Point testPt(6.0, 6.0, 6.0);
  testPt = tnsMtx * testPt;

  ASSERT_EQ(Point(5.0, 8.2, 2.0), testPt);
}

/**
 * Test make x/y/z rotation matrix.
 */
TEST(CLMatrixTest, CLMatrixTestCase10) {
  Point testPt;

  // test makeXRotation
  Eigen::Transform<double, 3, Eigen::Affine> xRotMtx;
  xRotMtx = Eigen::AngleAxis<double>((PI / 2.0), Vector::UnitX());

  testPt = Point(0.0, 1.0, 0.0);
  testPt = xRotMtx * testPt;

  //   Fails because of precision issues - RR
  //	ASSERT_EQ( Point( 0.0, 0.0, 1.0 ), testPt );

  //	Needed to use the following to allow a tolerance
  //	factor to be applied.
  //
  EXPECT_NEAR(0.0, testPt(0), 1e-16);
  EXPECT_NEAR(0.0, testPt(1), 1e-16);
  EXPECT_NEAR(1.0, testPt(2), 1e-16);

  // test makeYRotation
  Eigen::Transform<double, 3, Eigen::Affine> yRotMtx;
  yRotMtx = Eigen::AngleAxis<double>((PI / 2.0), Vector::UnitY());

  testPt = Point(2.0, 0.0, 0.0);
  testPt = yRotMtx * testPt;

  //   Fails because of precision issues - RR
  //    EXPECT_EQ( 1, Point( 0.0, 0.0, -2.0 ) == testPt );

  //	Needed to use the following to allow a tolerance
  //	factor to be applied (which is different from
  //	the x-rotation case.
  //
  EXPECT_NEAR(0.0, testPt(0), 1e-15);
  EXPECT_NEAR(0.0, testPt(1), 1e-15);
  EXPECT_NEAR(-2.0, testPt(2), 1e-15);

  // test makeZRotation
  Eigen::Transform<double, 3, Eigen::Affine> zRotMtx;
  zRotMtx = Eigen::AngleAxis<double>((-PI / 2.0), Vector::UnitZ());

  testPt = Point(2.0, 0.0, 0.0);
  testPt = zRotMtx * testPt;

  //    ASSERT_EQ(1, Point( 0.0, -2.0, 0.0 ) == testPt );

  EXPECT_NEAR(0.0, testPt(0), 1e-15);
  EXPECT_NEAR(-2.0, testPt(1), 1e-15);
  EXPECT_NEAR(0.0, testPt(2), 1e-15);
}

/**
 * Test euler angle extraction from matrix built using XZY order
 */

TEST(CLMatrixTest, CDMatrixTestCase11) {
  double yTruth = 0.123;
  double zTruth = 0.400;
  double xTruth = 3.034;

  Eigen::Transform<double, 3, Eigen::Affine> xMtx, yMtx, zMtx, rotMtx;

  xMtx = Eigen::AngleAxis<double>(xTruth, Vector::UnitX());

  yMtx = Eigen::AngleAxis<double>(yTruth, Vector::UnitY());

  zMtx = Eigen::AngleAxis<double>(zTruth, Vector::UnitZ());

  rotMtx = xMtx * zMtx * yMtx;

  //
  //
  Vector angles;
  angles = rotMtx.rotation().eulerAngles(0, 2, 1);

  double xRot, yRot, zRot;

  xRot = angles(0);
  yRot = angles(2);
  zRot = angles(1);

  //
  //  Old D4 call.
  //    rotMtx.toYzxEulerAngles( xRot, yRot, zRot );

  EXPECT_NEAR(xTruth, xRot, 1E-6);
  EXPECT_NEAR(yTruth, yRot, 1E-6);
  EXPECT_NEAR(zTruth, zRot, 1E-6);
}

/**
 * Test euler angle extraction from matrix using ZYX
 */
TEST(CLMatrixTest, CLMatrixTestCase12) {
  Eigen::Transform<double, 3, Eigen::Affine> xMatrix, yMatrix, zMatrix,
      zyxMatrix;

  // build the composite matrix using ZYX order
  xMatrix = Eigen::AngleAxis<double>(3.034, Vector::UnitX());
  yMatrix = Eigen::AngleAxis<double>(0.123, Vector::UnitY());
  zMatrix = Eigen::AngleAxis<double>(0.400, Vector::UnitZ());

  zyxMatrix = xMatrix * yMatrix * zMatrix;

  // extract the equivelant angles assuming YZX order
  Vector angles;
  angles = zyxMatrix.rotation().eulerAngles(0, 2, 1);

  double xAngle, yAngle, zAngle;

  xAngle = angles(0);
  yAngle = angles(2);
  zAngle = angles(1);

  //
  // Old D4 Call
  //
  //    zyxMatrix.toYzxEulerAngles( xAngle, yAngle, zAngle );

  // build the equivelant composite matrix using YZX order
  xMatrix = Eigen::AngleAxisd(xAngle, Vector::UnitX());
  yMatrix = Eigen::AngleAxisd(yAngle, Vector::UnitY());
  zMatrix = Eigen::AngleAxisd(zAngle, Vector::UnitZ());
  Matrix yzxMatrix = xMatrix * zMatrix * yMatrix;

  // check that the two matrices match, element by element
  for (int jj = 0; jj < 4; jj++) {
    for (int ii = 0; ii < 4; ii++) {
      double zyxValue = zyxMatrix(ii, jj);
      double yzxValue = yzxMatrix(ii, jj);
      EXPECT_NEAR(zyxValue, yzxValue, 1E-6);
    }
  }
}

/**
 * Test euler angle extraction from matrix for singular north pointing case.
 */
TEST(CLMatrixTest, CLMatrixTestCase13) {
  double yTruth = 0.000;
  double zTruth = PI / 2.0;
  double xTruth = 0.000;

  Eigen::Transform<double, 3, Eigen::Affine> xMtx, yMtx, zMtx, rotMtx;

  xMtx = Eigen::AngleAxis<double>(xTruth, Vector::UnitX());
  yMtx = Eigen::AngleAxis<double>(yTruth, Vector::UnitY());
  zMtx = Eigen::AngleAxis<double>(zTruth, Vector::UnitZ());

  rotMtx = yMtx * zMtx * xMtx;

  Vector angles;
  angles = rotMtx.rotation().eulerAngles(0, 2, 1);

  double xRot, yRot, zRot;

  xRot = angles(0);
  yRot = angles(2);
  zRot = angles(1);

  //
  //  Old D4 Call
  //
  //    rotMtx.toYzxEulerAngles( xRot, yRot, zRot );

  EXPECT_NEAR(xTruth, xRot, 1E-6);
  EXPECT_NEAR(yTruth, yRot, 1E-6);
  EXPECT_NEAR(zTruth, zRot, 1E-6);
}

/**
 * Test euler angle extraction from matrix for singular south pointing case.
 */
TEST(CLMatrixTest, CLMatrixTestCase14) {
  double yTruth = 0.000;
  double zTruth = -PI / 2.0;
  double xTruth = 0.000;

  Eigen::Transform<double, 3, Eigen::Affine> xMtx, yMtx, zMtx, rotMtx;

  xMtx = Eigen::AngleAxis<double>(xTruth, Vector::UnitX());
  yMtx = Eigen::AngleAxis<double>(yTruth, Vector::UnitY());
  zMtx = Eigen::AngleAxis<double>(zTruth, Vector::UnitZ());

  rotMtx = yMtx * zMtx * xMtx;

  Vector angles;
  angles = rotMtx.rotation().eulerAngles(0, 2, 1);

  double xRot, yRot, zRot;

  xRot = angles(0);
  yRot = angles(2);
  zRot = angles(1);

  //
  //  Old D4 Call
  //
  //    rotMtx.toYzxEulerAngles( xRot, yRot, zRot );

  EXPECT_NEAR(xTruth, xRot, 1E-6);
  EXPECT_NEAR(yTruth, yRot, 1E-6);
  EXPECT_NEAR(zTruth, zRot, 1E-6);
}
