/** Unit test for mixins
 *
 *  \file MixinTests.cpp
 */

#include "dirsig5/base/Mixin.h"

#include <gtest/gtest.h>

using namespace d5;

template <typename Base> struct A : Base {
  int a;

  template <int V> void setA() { a = V; }

  void setC(char val) { this->self().c = val; }
};

template <typename Base> struct B : Base {
  double b;
};

template <typename Base> struct C : Base {
  char c;
};

template <typename Base> struct D : Base {
  bool d;
};

template <typename T, typename Base> struct E : Base {
  T e;
};

template <typename Base> using Eu32 = E<std::uint32_t, Base>;

template <typename Base> using Ef64 = E<std::uint32_t, Base>;

struct ABC : mixin<A, B, C> {};
struct ABEC : mixin<A, B, Eu32, C> {
  void setATo9() { setA<9>(); }
  void setCToE() { setC('e'); }
};
struct EAC : mixin<Eu32, A, C> {};

TEST(MixinTests, MixinBaseTest) {
  using mixin_t = mixin<A, B, C>;
  struct F : MixinBase<A, mixin<A, B, C>> {};

  F f;
  const F &cl_f(f);
  static_assert(std::is_same_v<decltype(f.self()), mixin_t &>);
  static_assert(std::is_same_v<decltype(cl_f.self()), const mixin_t &>);
  static_assert(std::is_same_v<decltype(static_cast<const F &&>(f).self()),
                               const mixin_t &&>);
  static_assert(std::is_same_v<decltype(std::move(f).self()), mixin_t &&>);
}

// Test mixin_type_t, which should get the mixin type for each of our test cases
static_assert(std::is_same_v<mixin_type_t<ABC>, mixin<A, B, C>>);
static_assert(std::is_same_v<mixin_type_t<ABEC>, mixin<A, B, Eu32, C>>);
static_assert(std::is_same_v<mixin_type_t<EAC>, mixin<Eu32, A, C>>);

template <template <class...> class... Ts>
struct TestMetafunc : std::true_type {};

static_assert(std::is_same_v<mixin_metafunction_t<ABC, TestMetafunc>,
                             TestMetafunc<A, B, C>>);
static_assert(std::is_same_v<mixin_metafunction_t<ABEC, TestMetafunc>,
                             TestMetafunc<A, B, Eu32, C>>);

TEST(MixinTests, MixinTest) {
  ABEC x;
  x.a = 5;
  x.b = 4.5;
  x.c = 'a';
  x.e = 3;
  EXPECT_EQ(x.a, 5);
  EXPECT_EQ(x.b, 4.5);
  EXPECT_EQ(x.c, 'a');
  EXPECT_EQ(x.e, 3);
  static_assert(has_mixin_v<ABEC, A>);
  static_assert(has_mixin_v<ABEC, B>);
  static_assert(has_mixin_v<ABEC, C>);
  static_assert(has_mixin_v<ABEC, Eu32>);
  static_assert(!has_mixin_v<ABEC, Ef64>);
  static_assert(!has_mixin_v<ABEC, D>);
}

TEST(MixinTests, IsMixinTest) {
  static_assert(is_mixin_v<ABC>);
  static_assert(is_mixin_v<ABEC>);
  static_assert(is_mixin_v<EAC>);
  static_assert(!is_mixin_v<uint32_t>);
  static_assert(!is_mixin_v<A<MixinBase<A, ABC>>>);
}

TEST(MixinTests, mixin_baseTest) {
  // Test ABC
  static_assert(std::is_base_of_v<mixin_base_t<ABC, A>, ABC>);
  static_assert(std::is_base_of_v<mixin_base_t<ABC, B>, ABC>);
  static_assert(std::is_base_of_v<mixin_base_t<ABC, C>, ABC>);
  static_assert(
      std::is_same_v<mixin_base_t<ABC, A>, A<MixinBase<A, mixin<A, B, C>>>>);
  static_assert(
      std::is_same_v<mixin_base_t<ABC, B>, B<MixinBase<B, mixin<A, B, C>>>>);
  static_assert(
      std::is_same_v<mixin_base_t<ABC, C>, C<MixinBase<C, mixin<A, B, C>>>>);

  // Test ABEC
  static_assert(std::is_base_of_v<mixin_base_t<ABEC, A>, ABEC>);
  static_assert(std::is_base_of_v<mixin_base_t<ABEC, B>, ABEC>);
  static_assert(std::is_base_of_v<mixin_base_t<ABEC, Eu32>, ABEC>);
  static_assert(std::is_base_of_v<mixin_base_t<ABEC, C>, ABEC>);
  static_assert(std::is_same_v<mixin_base_t<ABEC, A>,
                               A<MixinBase<A, mixin<A, B, Eu32, C>>>>);
  static_assert(std::is_same_v<mixin_base_t<ABEC, B>,
                               B<MixinBase<B, mixin<A, B, Eu32, C>>>>);
  static_assert(std::is_same_v<mixin_base_t<ABEC, Eu32>,
                               Eu32<MixinBase<Eu32, mixin<A, B, Eu32, C>>>>);
  static_assert(std::is_same_v<mixin_base_t<ABEC, C>,
                               C<MixinBase<C, mixin<A, B, Eu32, C>>>>);

  // Test EAC
  static_assert(std::is_base_of_v<mixin_base_t<EAC, Eu32>, EAC>);
  static_assert(std::is_base_of_v<mixin_base_t<EAC, A>, EAC>);
  static_assert(std::is_base_of_v<mixin_base_t<EAC, C>, EAC>);
  static_assert(
      std::is_same_v<mixin_base_t<EAC, Eu32>,
                     E<std::uint32_t, MixinBase<Eu32, mixin<Eu32, A, C>>>>);
  static_assert(
      std::is_same_v<mixin_base_t<EAC, A>, A<MixinBase<A, mixin<Eu32, A, C>>>>);
  static_assert(
      std::is_same_v<mixin_base_t<EAC, C>, C<MixinBase<C, mixin<Eu32, A, C>>>>);
}

TEST(MixinTests, MixinCastABCTest) {
  ABC x;
  mixin_cast<A>(x).a = 5;
  mixin_cast<B>(x).b = 4.5;
  mixin_cast<C>(x).c = 'a';
  EXPECT_EQ(x.a, 5);
  EXPECT_EQ(x.b, 4.5);
  EXPECT_EQ(x.c, 'a');
  mixin_cast<A>(x).setC('b');
  EXPECT_EQ(x.c, 'b');
}

TEST(MixinTests, MixinCastABECTest) {
  ABEC x;
  mixin_cast<A>(x).a = 5;
  mixin_cast<B>(x).b = 4.5;
  mixin_cast<C>(x).c = 'a';
  mixin_cast<Eu32>(x).e = 3;
  EXPECT_EQ(x.a, 5);
  EXPECT_EQ(x.b, 4.5);
  EXPECT_EQ(x.c, 'a');
  EXPECT_EQ(x.e, 3);
  mixin_cast<A>(x).setC('b');
  EXPECT_EQ(x.c, 'b');
  x.setC('c');
  EXPECT_EQ(x.c, 'c');
  x.setCToE();
  EXPECT_EQ(x.c, 'e');

  x.setA<7>();
  EXPECT_EQ(x.a, 7);
  x.setATo9();
  EXPECT_EQ(x.a, 9);
}
