#include <gtest/gtest.h>

#include "dirsig5/base/Geometry.h"

using namespace d5;
/*
        The following tests are ports of DIRSIG4 unit tests
        found in the base_geometry subdirectory.

        They have been ported to use the Eigen and gtest
        framework.  We are number the D4 tests by case numbers.

        Case01 - testConstruct
        Case02 - testManipulate
        Case03 - testLength
        Case04 - testUnit
        Case05 - testConjugate
        Case06 - testInverse
        Case07 - testToMatrix

        *NOT IMPLEMENTED (YET)
        xCurrently Failing Test
*/

/**
 * Test the Quaternion construct
 */
TEST(QuaternionTest, QuaternionTestCase01) {
  Quaternion q(0, 0, 0, 0);

  EXPECT_NEAR(0.0, q.w(), 1E-6);
  EXPECT_NEAR(0.0, q.x(), 1E-6);
  EXPECT_NEAR(0.0, q.y(), 1E-6);
  EXPECT_NEAR(0.0, q.z(), 1E-6);

  Quaternion r(+5, +2, -1, -3);

  EXPECT_NEAR(+5, r.w(), 1E-6);
  EXPECT_NEAR(+2, r.x(), 1E-6);
  EXPECT_NEAR(-1, r.y(), 1E-6);
  EXPECT_NEAR(-3, r.z(), 1E-6);
}

/**
 * Test the Quaternion manipulation
 */
TEST(QuaternionTest, QuaternionTestCase02) {
  Quaternion q;

  q.w() = +5;
  q.x() = +2;
  q.y() = -1;
  q.z() = -3;

  EXPECT_NEAR(+5, q.w(), 1E-6);
  EXPECT_NEAR(+2, q.x(), 1E-6);
  EXPECT_NEAR(-1, q.y(), 1E-6);
  EXPECT_NEAR(-3, q.z(), 1E-6);
}

/**
 * Test the Quaternion length
 */
TEST(QuaternionTest, QuaternionTestCase03) {
  Quaternion q;

  q.w() = +5;
  q.x() = +2;
  q.y() = -1;
  q.z() = -3;

  EXPECT_NEAR(sqrt(39.0), q.norm(), 1E-6);
}

/**
 * Test the Quaternion unit
 */
TEST(QuaternionTest, QuaternionTestCase04) {
  Quaternion q(+5, +2, -1, -3);

  q.normalize();

  EXPECT_NEAR(+5 / (sqrt(39.0)), q.w(), 1E-6);
  EXPECT_NEAR(+2 / (sqrt(39.0)), q.x(), 1E-6);
  EXPECT_NEAR(-1 / (sqrt(39.0)), q.y(), 1E-6);
  EXPECT_NEAR(-3 / (sqrt(39.0)), q.z(), 1E-6);
}

/**
 * Test the Quaternion conjugate
 */
TEST(QuaternionTest, QuaternionTestCase05) {
  Quaternion q(+5, +2, -1, -3);
  Quaternion conj;
  conj = q.conjugate();

  EXPECT_NEAR(+5, conj.w(), 1E-6);
  EXPECT_NEAR(-2, conj.x(), 1E-6);
  EXPECT_NEAR(+1, conj.y(), 1E-6);
  EXPECT_NEAR(+3, conj.z(), 1E-6);
}

/**
 * Test the Quaternion inverse
 */
TEST(QuaternionTest, QuaternionTestCase06) {
  Quaternion q(+5, +2, -1, -3);
  Quaternion inv;
  inv = q.inverse();

  EXPECT_NEAR(+5.0 / 39.0, inv.w(), 1E-6);
  EXPECT_NEAR(-2.0 / 39.0, inv.x(), 1E-6);
  EXPECT_NEAR(+1.0 / 39.0, inv.y(), 1E-6);
  EXPECT_NEAR(+3.0 / 39.0, inv.z(), 1E-6);
}

/**
 * The truth for this test case was produced by WolframAlpha
 * according to CDQuaternionTest.cpp
 */
TEST(QuaternionTest, QuaternionTestCase07) {
  Quaternion q(+5, +2, -1, -3);

  Eigen::Matrix3d m;
  m = q.normalized().toRotationMatrix();

  EXPECT_NEAR(+19.0 / 39.0, m(0, 0), 1E-6);
  EXPECT_NEAR(+2.0 / 3.0, m(0, 1), 1E-6);
  EXPECT_NEAR(-22.0 / 39.0, m(0, 2), 1E-6);

  EXPECT_NEAR(-34.0 / 39.0, m(1, 0), 1E-6);
  EXPECT_NEAR(+1.0 / 3.0, m(1, 1), 1E-6);
  EXPECT_NEAR(-14.0 / 39.0, m(1, 2), 1E-6);

  EXPECT_NEAR(-2.0 / 39.0, m(2, 0), 1E-6);
  EXPECT_NEAR(+2.0 / 3.0, m(2, 1), 1E-6);
  EXPECT_NEAR(+29.0 / 39.0, m(2, 2), 1E-6);

  Quaternion q2(0, +1, 0, 0);

  Eigen::Matrix3d m2;
  m2 = q2.normalized().toRotationMatrix();

  EXPECT_NEAR(+1.0, m2(0, 0), 1E-6);
  EXPECT_NEAR(0.0, m2(0, 1), 1E-6);
  EXPECT_NEAR(0.0, m2(0, 2), 1E-6);

  EXPECT_NEAR(0.0, m2(1, 0), 1E-6);
  EXPECT_NEAR(-1.0, m2(1, 1), 1E-6);
  EXPECT_NEAR(0.0, m2(1, 2), 1E-6);

  EXPECT_NEAR(0.0, m2(2, 0), 1E-6);
  EXPECT_NEAR(0.0, m2(2, 1), 1E-6);
  EXPECT_NEAR(-1.0, m2(2, 2), 1E-6);

  Quaternion q3(0, 0, -1, 0);

  Eigen::Matrix3d m3;
  m3 = q3.normalized().toRotationMatrix();

  EXPECT_NEAR(-1.0, m3(0, 0), 1E-6);
  EXPECT_NEAR(0.0, m3(0, 1), 1E-6);
  EXPECT_NEAR(0.0, m3(0, 2), 1E-6);

  EXPECT_NEAR(0.0, m3(1, 0), 1E-6);
  EXPECT_NEAR(+1.0, m3(1, 1), 1E-6);
  EXPECT_NEAR(0.0, m3(1, 2), 1E-6);

  EXPECT_NEAR(0.0, m3(2, 0), 1E-6);
  EXPECT_NEAR(0.0, m3(2, 1), 1E-6);
  EXPECT_NEAR(-1.0, m3(2, 2), 1E-6);
}
