#include <gtest/gtest.h>

#include "dirsig5/base/SpectralVector.h"

#define N_POINTS 214

using namespace d5;

/*
The following will test the interface SpectralVector.h.

http://twiki.cis.rit.edu/twiki/bin/view/DIRS/EnviSpectralLibraryBigEquationExtractionTest


The above page contains a spectral library that has been pulled out
for the specific set of LUT parameters of


AEROSOL=20.0000; ELEVATION=0.315000; WATER=1.05000; RAD:EsCosTau1Tau
2/Pi
AEROSOL=20.0000; ELEVATION=0.315000; WATER=1.05000; RAD:LdTau2
AEROSOL=20.0000; ELEVATION=0.315000; WATER=1.05000; RAD:Lu

The above expressions will then combined using the following
expression of

        Lsensor = (EsCosTau1Tau2/Pi)*r*LdTau2+Lu

Where r=1.0 (perfect Lambertian reflector)

It will then be compared to the answer from the spectral math
calculation using ENVI for the same set of MODTRAN parameters.

AEROSOL=20.0000; ELEVATION=0.315000; WATER=1.05000; RAD:Lsensor


The tests will be as follows

        Case00
                Deposit hardwired wavelength values into a
                spectral library and compare to make sure
                that what was deposited into the spectral
                vector is the same value.

        Case01
                Same as Case00, but we will make a copy of
                the wavelength value and make sure that
                what went in is what comes out.


        Case02
                Same as Case01, but we will make a copy of
                the wavelength spectral vector and do a
                SpectralVector subtraction on the source and
                the copy.

        Case03
                Assign a new spectral vector for
                EsCosThetaTau1.  We will test minimum and
                maximum location of the values.


        Case04
                Big Equation test where hardwired values for
                the terms, EsCosTheta1, LdTau2, Lu for a 100%
                reflector.  These terms were extracted from
                an ENVI spectral library and the same equations
                were computed in ENVI using "Spectral Math".
                The results of this calculation in ENVI was
                hardwired into this test and compared against
                the SpectralVector calculation by computing
                an RMS (further testing the "pow" function).

                Several terms in the comments have been tested
                using EXPECT_EQ, but have failed because of
                precision errors.
                A sanity check for the very last vector element
                [213] was also made to make sure the calculations
                made sense.

                We are leaving all these tests  as part
                of the comments to document the expected values
                of the tests.  They can be uncommented to invoke
                a fail, but can be visually inspected in the logs
                that they are indeed the same values.

RR - 11/02/2015

*/

TEST(SpectralVectorBigEquation, Case00) {
  double wavelength[N_POINTS] = {
      3.800000e-01, 3.900000e-01, 4.000000e-01, 4.100000e-01, 4.200000e-01,
      4.300000e-01, 4.400000e-01, 4.500000e-01, 4.600000e-01, 4.700000e-01,
      4.800000e-01, 4.900000e-01, 5.000000e-01, 5.100000e-01, 5.200000e-01,
      5.300000e-01, 5.400000e-01, 5.500000e-01, 5.600000e-01, 5.700000e-01,
      5.800000e-01, 5.900000e-01, 6.000000e-01, 6.100000e-01, 6.200000e-01,
      6.300000e-01, 6.400000e-01, 6.500000e-01, 6.600000e-01, 6.700000e-01,
      6.800000e-01, 6.900000e-01, 7.000000e-01, 7.100000e-01, 7.200000e-01,
      7.300000e-01, 7.400000e-01, 7.500000e-01, 7.600000e-01, 7.700000e-01,
      7.800000e-01, 7.900000e-01, 8.000000e-01, 8.100000e-01, 8.200000e-01,
      8.300000e-01, 8.400000e-01, 8.500000e-01, 8.600000e-01, 8.700000e-01,
      8.800000e-01, 8.900000e-01, 9.000000e-01, 9.100000e-01, 9.200000e-01,
      9.300000e-01, 9.400000e-01, 9.500000e-01, 9.600000e-01, 9.700000e-01,
      9.800000e-01, 9.900000e-01, 1.000000e+00, 1.010000e+00, 1.020000e+00,
      1.030000e+00, 1.040000e+00, 1.050000e+00, 1.060000e+00, 1.070000e+00,
      1.080000e+00, 1.090000e+00, 1.100000e+00, 1.110000e+00, 1.120000e+00,
      1.130000e+00, 1.140000e+00, 1.150000e+00, 1.160000e+00, 1.170000e+00,
      1.180000e+00, 1.190000e+00, 1.200000e+00, 1.210000e+00, 1.220000e+00,
      1.230000e+00, 1.240000e+00, 1.250000e+00, 1.260000e+00, 1.270000e+00,
      1.280000e+00, 1.290000e+00, 1.300000e+00, 1.310000e+00, 1.320000e+00,
      1.330000e+00, 1.340000e+00, 1.350000e+00, 1.360000e+00, 1.370000e+00,
      1.380000e+00, 1.390000e+00, 1.400000e+00, 1.410000e+00, 1.420000e+00,
      1.430000e+00, 1.440000e+00, 1.450000e+00, 1.460000e+00, 1.470000e+00,
      1.480000e+00, 1.490000e+00, 1.500000e+00, 1.510000e+00, 1.520000e+00,
      1.530000e+00, 1.540000e+00, 1.550000e+00, 1.560000e+00, 1.570000e+00,
      1.580000e+00, 1.590000e+00, 1.600000e+00, 1.610000e+00, 1.620000e+00,
      1.630000e+00, 1.640000e+00, 1.650000e+00, 1.660000e+00, 1.670000e+00,
      1.680000e+00, 1.690000e+00, 1.700000e+00, 1.710000e+00, 1.720000e+00,
      1.730000e+00, 1.740000e+00, 1.750000e+00, 1.760000e+00, 1.770000e+00,
      1.780000e+00, 1.790000e+00, 1.800000e+00, 1.810000e+00, 1.820000e+00,
      1.830000e+00, 1.840000e+00, 1.850000e+00, 1.860000e+00, 1.870000e+00,
      1.880000e+00, 1.890000e+00, 1.900000e+00, 1.910000e+00, 1.920000e+00,
      1.930000e+00, 1.940000e+00, 1.950000e+00, 1.960000e+00, 1.970000e+00,
      1.980000e+00, 1.990000e+00, 2.000000e+00, 2.010000e+00, 2.020000e+00,
      2.030000e+00, 2.040000e+00, 2.050000e+00, 2.060000e+00, 2.070000e+00,
      2.080000e+00, 2.090000e+00, 2.100000e+00, 2.110000e+00, 2.120000e+00,
      2.130000e+00, 2.140000e+00, 2.150000e+00, 2.160000e+00, 2.170000e+00,
      2.180000e+00, 2.190000e+00, 2.200000e+00, 2.210000e+00, 2.220000e+00,
      2.230000e+00, 2.240000e+00, 2.250000e+00, 2.260000e+00, 2.270000e+00,
      2.280000e+00, 2.290000e+00, 2.300000e+00, 2.310000e+00, 2.320000e+00,
      2.330000e+00, 2.340000e+00, 2.350000e+00, 2.360000e+00, 2.370000e+00,
      2.380000e+00, 2.390000e+00, 2.400000e+00, 2.410000e+00, 2.420000e+00,
      2.430000e+00, 2.440000e+00, 2.450000e+00, 2.460000e+00, 2.470000e+00,
      2.480000e+00, 2.490000e+00, 2.500000e+00, 2.510000e+00};

  for (size_t i = 0; i < N_POINTS; i++) {
    EXPECT_EQ(wavelength[i], wavelength[i]);
  }

  size_t pointCount = N_POINTS;
  double value = 0.0;

  SpectralVector wavelengthSpectrum(pointCount, value);

  for (size_t i = 0; i < N_POINTS; i++) {
    double data = wavelength[i];
    wavelengthSpectrum.setData(i, data);
  }
}

TEST(SpectralVectorBigEquation, Case01) {
  double wavelength[N_POINTS] = {
      3.800000e-01, 3.900000e-01, 4.000000e-01, 4.100000e-01, 4.200000e-01,
      4.300000e-01, 4.400000e-01, 4.500000e-01, 4.600000e-01, 4.700000e-01,
      4.800000e-01, 4.900000e-01, 5.000000e-01, 5.100000e-01, 5.200000e-01,
      5.300000e-01, 5.400000e-01, 5.500000e-01, 5.600000e-01, 5.700000e-01,
      5.800000e-01, 5.900000e-01, 6.000000e-01, 6.100000e-01, 6.200000e-01,
      6.300000e-01, 6.400000e-01, 6.500000e-01, 6.600000e-01, 6.700000e-01,
      6.800000e-01, 6.900000e-01, 7.000000e-01, 7.100000e-01, 7.200000e-01,
      7.300000e-01, 7.400000e-01, 7.500000e-01, 7.600000e-01, 7.700000e-01,
      7.800000e-01, 7.900000e-01, 8.000000e-01, 8.100000e-01, 8.200000e-01,
      8.300000e-01, 8.400000e-01, 8.500000e-01, 8.600000e-01, 8.700000e-01,
      8.800000e-01, 8.900000e-01, 9.000000e-01, 9.100000e-01, 9.200000e-01,
      9.300000e-01, 9.400000e-01, 9.500000e-01, 9.600000e-01, 9.700000e-01,
      9.800000e-01, 9.900000e-01, 1.000000e+00, 1.010000e+00, 1.020000e+00,
      1.030000e+00, 1.040000e+00, 1.050000e+00, 1.060000e+00, 1.070000e+00,
      1.080000e+00, 1.090000e+00, 1.100000e+00, 1.110000e+00, 1.120000e+00,
      1.130000e+00, 1.140000e+00, 1.150000e+00, 1.160000e+00, 1.170000e+00,
      1.180000e+00, 1.190000e+00, 1.200000e+00, 1.210000e+00, 1.220000e+00,
      1.230000e+00, 1.240000e+00, 1.250000e+00, 1.260000e+00, 1.270000e+00,
      1.280000e+00, 1.290000e+00, 1.300000e+00, 1.310000e+00, 1.320000e+00,
      1.330000e+00, 1.340000e+00, 1.350000e+00, 1.360000e+00, 1.370000e+00,
      1.380000e+00, 1.390000e+00, 1.400000e+00, 1.410000e+00, 1.420000e+00,
      1.430000e+00, 1.440000e+00, 1.450000e+00, 1.460000e+00, 1.470000e+00,
      1.480000e+00, 1.490000e+00, 1.500000e+00, 1.510000e+00, 1.520000e+00,
      1.530000e+00, 1.540000e+00, 1.550000e+00, 1.560000e+00, 1.570000e+00,
      1.580000e+00, 1.590000e+00, 1.600000e+00, 1.610000e+00, 1.620000e+00,
      1.630000e+00, 1.640000e+00, 1.650000e+00, 1.660000e+00, 1.670000e+00,
      1.680000e+00, 1.690000e+00, 1.700000e+00, 1.710000e+00, 1.720000e+00,
      1.730000e+00, 1.740000e+00, 1.750000e+00, 1.760000e+00, 1.770000e+00,
      1.780000e+00, 1.790000e+00, 1.800000e+00, 1.810000e+00, 1.820000e+00,
      1.830000e+00, 1.840000e+00, 1.850000e+00, 1.860000e+00, 1.870000e+00,
      1.880000e+00, 1.890000e+00, 1.900000e+00, 1.910000e+00, 1.920000e+00,
      1.930000e+00, 1.940000e+00, 1.950000e+00, 1.960000e+00, 1.970000e+00,
      1.980000e+00, 1.990000e+00, 2.000000e+00, 2.010000e+00, 2.020000e+00,
      2.030000e+00, 2.040000e+00, 2.050000e+00, 2.060000e+00, 2.070000e+00,
      2.080000e+00, 2.090000e+00, 2.100000e+00, 2.110000e+00, 2.120000e+00,
      2.130000e+00, 2.140000e+00, 2.150000e+00, 2.160000e+00, 2.170000e+00,
      2.180000e+00, 2.190000e+00, 2.200000e+00, 2.210000e+00, 2.220000e+00,
      2.230000e+00, 2.240000e+00, 2.250000e+00, 2.260000e+00, 2.270000e+00,
      2.280000e+00, 2.290000e+00, 2.300000e+00, 2.310000e+00, 2.320000e+00,
      2.330000e+00, 2.340000e+00, 2.350000e+00, 2.360000e+00, 2.370000e+00,
      2.380000e+00, 2.390000e+00, 2.400000e+00, 2.410000e+00, 2.420000e+00,
      2.430000e+00, 2.440000e+00, 2.450000e+00, 2.460000e+00, 2.470000e+00,
      2.480000e+00, 2.490000e+00, 2.500000e+00, 2.510000e+00};

  for (size_t i = 0; i < N_POINTS; i++) {
    EXPECT_EQ(wavelength[i], wavelength[i]);
  }

  size_t pointCount = N_POINTS;
  double value = 0.0;

  SpectralVector wavelengthSpectrum(pointCount, value);

  for (size_t i = 0; i < N_POINTS; i++) {
    double data = wavelength[i];
    wavelengthSpectrum.setData(i, data);
  }

  for (size_t i = 0; i < N_POINTS; i++) {
    EXPECT_EQ(wavelength[i], wavelengthSpectrum.getData(i));
  }
}

TEST(SpectralVectorBigEquation, Case02) {
  double wavelength[N_POINTS] = {
      3.800000e-01, 3.900000e-01, 4.000000e-01, 4.100000e-01, 4.200000e-01,
      4.300000e-01, 4.400000e-01, 4.500000e-01, 4.600000e-01, 4.700000e-01,
      4.800000e-01, 4.900000e-01, 5.000000e-01, 5.100000e-01, 5.200000e-01,
      5.300000e-01, 5.400000e-01, 5.500000e-01, 5.600000e-01, 5.700000e-01,
      5.800000e-01, 5.900000e-01, 6.000000e-01, 6.100000e-01, 6.200000e-01,
      6.300000e-01, 6.400000e-01, 6.500000e-01, 6.600000e-01, 6.700000e-01,
      6.800000e-01, 6.900000e-01, 7.000000e-01, 7.100000e-01, 7.200000e-01,
      7.300000e-01, 7.400000e-01, 7.500000e-01, 7.600000e-01, 7.700000e-01,
      7.800000e-01, 7.900000e-01, 8.000000e-01, 8.100000e-01, 8.200000e-01,
      8.300000e-01, 8.400000e-01, 8.500000e-01, 8.600000e-01, 8.700000e-01,
      8.800000e-01, 8.900000e-01, 9.000000e-01, 9.100000e-01, 9.200000e-01,
      9.300000e-01, 9.400000e-01, 9.500000e-01, 9.600000e-01, 9.700000e-01,
      9.800000e-01, 9.900000e-01, 1.000000e+00, 1.010000e+00, 1.020000e+00,
      1.030000e+00, 1.040000e+00, 1.050000e+00, 1.060000e+00, 1.070000e+00,
      1.080000e+00, 1.090000e+00, 1.100000e+00, 1.110000e+00, 1.120000e+00,
      1.130000e+00, 1.140000e+00, 1.150000e+00, 1.160000e+00, 1.170000e+00,
      1.180000e+00, 1.190000e+00, 1.200000e+00, 1.210000e+00, 1.220000e+00,
      1.230000e+00, 1.240000e+00, 1.250000e+00, 1.260000e+00, 1.270000e+00,
      1.280000e+00, 1.290000e+00, 1.300000e+00, 1.310000e+00, 1.320000e+00,
      1.330000e+00, 1.340000e+00, 1.350000e+00, 1.360000e+00, 1.370000e+00,
      1.380000e+00, 1.390000e+00, 1.400000e+00, 1.410000e+00, 1.420000e+00,
      1.430000e+00, 1.440000e+00, 1.450000e+00, 1.460000e+00, 1.470000e+00,
      1.480000e+00, 1.490000e+00, 1.500000e+00, 1.510000e+00, 1.520000e+00,
      1.530000e+00, 1.540000e+00, 1.550000e+00, 1.560000e+00, 1.570000e+00,
      1.580000e+00, 1.590000e+00, 1.600000e+00, 1.610000e+00, 1.620000e+00,
      1.630000e+00, 1.640000e+00, 1.650000e+00, 1.660000e+00, 1.670000e+00,
      1.680000e+00, 1.690000e+00, 1.700000e+00, 1.710000e+00, 1.720000e+00,
      1.730000e+00, 1.740000e+00, 1.750000e+00, 1.760000e+00, 1.770000e+00,
      1.780000e+00, 1.790000e+00, 1.800000e+00, 1.810000e+00, 1.820000e+00,
      1.830000e+00, 1.840000e+00, 1.850000e+00, 1.860000e+00, 1.870000e+00,
      1.880000e+00, 1.890000e+00, 1.900000e+00, 1.910000e+00, 1.920000e+00,
      1.930000e+00, 1.940000e+00, 1.950000e+00, 1.960000e+00, 1.970000e+00,
      1.980000e+00, 1.990000e+00, 2.000000e+00, 2.010000e+00, 2.020000e+00,
      2.030000e+00, 2.040000e+00, 2.050000e+00, 2.060000e+00, 2.070000e+00,
      2.080000e+00, 2.090000e+00, 2.100000e+00, 2.110000e+00, 2.120000e+00,
      2.130000e+00, 2.140000e+00, 2.150000e+00, 2.160000e+00, 2.170000e+00,
      2.180000e+00, 2.190000e+00, 2.200000e+00, 2.210000e+00, 2.220000e+00,
      2.230000e+00, 2.240000e+00, 2.250000e+00, 2.260000e+00, 2.270000e+00,
      2.280000e+00, 2.290000e+00, 2.300000e+00, 2.310000e+00, 2.320000e+00,
      2.330000e+00, 2.340000e+00, 2.350000e+00, 2.360000e+00, 2.370000e+00,
      2.380000e+00, 2.390000e+00, 2.400000e+00, 2.410000e+00, 2.420000e+00,
      2.430000e+00, 2.440000e+00, 2.450000e+00, 2.460000e+00, 2.470000e+00,
      2.480000e+00, 2.490000e+00, 2.500000e+00, 2.510000e+00};

  for (size_t i = 0; i < N_POINTS; i++) {
    EXPECT_EQ(wavelength[i], wavelength[i]);
  }

  size_t pointCount = N_POINTS;
  double value = 0.0;

  SpectralVector wavelengthSpectrum(pointCount, value);

  for (size_t i = 0; i < N_POINTS; i++) {
    double data = wavelength[i];
    wavelengthSpectrum.setData(i, data);
  }

  for (size_t i = 0; i < N_POINTS; i++) {
    EXPECT_EQ(wavelength[i], wavelengthSpectrum.getData(i));
  }

  SpectralVector copySpectrum(wavelengthSpectrum);

  //
  // 	Test the in-place subtraction.
  //

  copySpectrum -= wavelengthSpectrum;

  for (size_t i = 0; i < N_POINTS; i++) {
    EXPECT_EQ(copySpectrum[i], 0.0);
  }

  //
  //	Test the in-place addition.
  //

  copySpectrum += wavelengthSpectrum;

  for (size_t i = 0; i < N_POINTS; i++) {
    EXPECT_EQ(copySpectrum[i], wavelengthSpectrum[i]);
  }

  //
  //	Test the in-place multiplication.
  //

  copySpectrum *= wavelengthSpectrum;

  for (size_t i = 0; i < N_POINTS; i++) {
    EXPECT_EQ(copySpectrum[i], wavelengthSpectrum[i] * wavelengthSpectrum[i]);
  }

  //
  // 	Test the in-place division
  //
  copySpectrum.eigenData /= copySpectrum.eigenData;

  for (size_t i = 0; i < N_POINTS; i++) {
    EXPECT_EQ(copySpectrum[i], 1.0);
  }
}

TEST(SpectralVectorBigEquation, Case03) {
  double escosthetatau1[N_POINTS] = {
      3.707000e-03, 4.198700e-03, 6.627000e-03, 8.542200e-03, 9.099000e-03,
      8.852500e-03, 1.103600e-02, 1.325700e-02, 1.439500e-02, 1.532300e-02,
      1.594900e-02, 1.592400e-02, 1.627200e-02, 1.662100e-02, 1.635000e-02,
      1.736700e-02, 1.775100e-02, 1.818100e-02, 1.837800e-02, 1.859200e-02,
      1.913100e-02, 1.804800e-02, 1.885300e-02, 1.930200e-02, 1.927900e-02,
      1.920200e-02, 1.951600e-02, 1.868300e-02, 1.924500e-02, 2.007800e-02,
      1.968700e-02, 1.725600e-02, 1.784100e-02, 1.818600e-02, 1.440900e-02,
      1.488400e-02, 1.797300e-02, 1.900400e-02, 1.259200e-02, 1.610900e-02,
      1.850900e-02, 1.742900e-02, 1.698100e-02, 1.554000e-02, 1.287000e-02,
      1.403200e-02, 1.594100e-02, 1.607400e-02, 1.614300e-02, 1.603500e-02,
      1.599800e-02, 1.473600e-02, 1.042300e-02, 9.419700e-03, 9.881000e-03,
      5.271500e-03, 2.511000e-03, 2.932400e-03, 4.703200e-03, 8.651300e-03,
      1.026400e-02, 1.291700e-02, 1.331400e-02, 1.311100e-02, 1.302700e-02,
      1.293600e-02, 1.277200e-02, 1.260300e-02, 1.233200e-02, 1.185400e-02,
      1.119400e-02, 1.016100e-02, 7.990000e-03, 5.005100e-03, 1.130400e-03,
      8.183000e-04, 1.573500e-03, 1.718300e-03, 4.233700e-03, 6.714700e-03,
      7.177700e-03, 7.441200e-03, 7.507700e-03, 7.761500e-03, 8.387700e-03,
      8.968200e-03, 9.331600e-03, 9.217600e-03, 8.462600e-03, 7.723800e-03,
      8.247900e-03, 7.993600e-03, 6.912200e-03, 5.705900e-03, 4.527600e-03,
      2.677100e-03, 1.431200e-03, 1.355200e-04, 6.304900e-09, 3.900900e-10,
      1.026600e-09, 6.989000e-10, 2.080300e-08, 3.663600e-06, 2.598300e-05,
      1.039200e-04, 1.176000e-04, 3.856900e-04, 7.802000e-04, 4.824500e-04,
      7.918900e-04, 1.909500e-03, 3.434800e-03, 4.349400e-03, 5.135600e-03,
      5.655600e-03, 5.878200e-03, 5.945600e-03, 5.888400e-03, 5.358100e-03,
      5.251200e-03, 5.466600e-03, 5.188800e-03, 5.019200e-03, 5.322300e-03,
      5.329300e-03, 4.962300e-03, 4.906500e-03, 4.907800e-03, 4.758300e-03,
      4.635100e-03, 4.549700e-03, 4.387800e-03, 4.171500e-03, 3.975400e-03,
      3.649700e-03, 3.310500e-03, 3.199800e-03, 2.885400e-03, 2.217900e-03,
      1.410200e-03, 8.046800e-04, 1.770200e-04, 1.369700e-05, 2.641100e-07,
      1.351600e-08, 1.148400e-13, 1.149100e-13, 5.224700e-13, 7.510800e-13,
      2.492700e-10, 1.011700e-09, 2.252800e-10, 3.023100e-10, 1.118100e-08,
      4.505200e-07, 9.457300e-06, 4.433600e-05, 1.414000e-04, 6.234700e-04,
      1.208500e-03, 1.414700e-03, 6.280600e-04, 3.566800e-04, 6.818400e-04,
      1.676500e-03, 1.940100e-03, 1.379200e-03, 1.169200e-03, 1.299200e-03,
      1.831500e-03, 2.001400e-03, 2.047200e-03, 2.126000e-03, 2.047500e-03,
      2.118500e-03, 2.154900e-03, 1.973200e-03, 1.865700e-03, 1.836400e-03,
      1.785600e-03, 1.756800e-03, 1.652600e-03, 1.779300e-03, 1.795500e-03,
      1.778100e-03, 1.729300e-03, 1.671600e-03, 1.587700e-03, 1.554900e-03,
      1.521300e-03, 1.467900e-03, 1.366100e-03, 1.397800e-03, 1.163200e-03,
      1.248400e-03, 1.040200e-03, 8.821400e-04, 9.897701e-04, 7.387200e-04,
      6.456600e-04, 6.346300e-04, 6.731300e-04, 4.484400e-04, 3.743200e-04,
      4.957000e-04, 4.089100e-04, 1.699800e-04, 2.972400e-04, 1.364900e-04,
      4.309100e-05, 4.828300e-06, 6.678000e-06, 1.201800e-06};

  SpectralVector EsCosThetaTau1(N_POINTS, 0.0);

  for (size_t i = 0; i < N_POINTS; i++) {
    EsCosThetaTau1.setData(i, escosthetatau1[i]);
  }

  size_t n_points = EsCosThetaTau1.getPointCount();

  for (size_t i = 0; i < N_POINTS; i++) {
    EXPECT_EQ(EsCosThetaTau1[i], escosthetatau1[i]);
  }

  //
  // Test to see the the number of points in the SpectralVector
  // are still the same.
  //

  EXPECT_EQ(n_points, (size_t)N_POINTS);

  EXPECT_EQ(EsCosThetaTau1.min(), 1.148400e-13);
  EXPECT_EQ(EsCosThetaTau1.max(), 2.007800e-02);
}

TEST(SpectralVectorBigEquation, Case04) {
  double escosthetatau1[N_POINTS] = {
      3.707000e-03, 4.198700e-03, 6.627000e-03, 8.542200e-03, 9.099000e-03,
      8.852500e-03, 1.103600e-02, 1.325700e-02, 1.439500e-02, 1.532300e-02,
      1.594900e-02, 1.592400e-02, 1.627200e-02, 1.662100e-02, 1.635000e-02,
      1.736700e-02, 1.775100e-02, 1.818100e-02, 1.837800e-02, 1.859200e-02,
      1.913100e-02, 1.804800e-02, 1.885300e-02, 1.930200e-02, 1.927900e-02,
      1.920200e-02, 1.951600e-02, 1.868300e-02, 1.924500e-02, 2.007800e-02,
      1.968700e-02, 1.725600e-02, 1.784100e-02, 1.818600e-02, 1.440900e-02,
      1.488400e-02, 1.797300e-02, 1.900400e-02, 1.259200e-02, 1.610900e-02,
      1.850900e-02, 1.742900e-02, 1.698100e-02, 1.554000e-02, 1.287000e-02,
      1.403200e-02, 1.594100e-02, 1.607400e-02, 1.614300e-02, 1.603500e-02,
      1.599800e-02, 1.473600e-02, 1.042300e-02, 9.419700e-03, 9.881000e-03,
      5.271500e-03, 2.511000e-03, 2.932400e-03, 4.703200e-03, 8.651300e-03,
      1.026400e-02, 1.291700e-02, 1.331400e-02, 1.311100e-02, 1.302700e-02,
      1.293600e-02, 1.277200e-02, 1.260300e-02, 1.233200e-02, 1.185400e-02,
      1.119400e-02, 1.016100e-02, 7.990000e-03, 5.005100e-03, 1.130400e-03,
      8.183000e-04, 1.573500e-03, 1.718300e-03, 4.233700e-03, 6.714700e-03,
      7.177700e-03, 7.441200e-03, 7.507700e-03, 7.761500e-03, 8.387700e-03,
      8.968200e-03, 9.331600e-03, 9.217600e-03, 8.462600e-03, 7.723800e-03,
      8.247900e-03, 7.993600e-03, 6.912200e-03, 5.705900e-03, 4.527600e-03,
      2.677100e-03, 1.431200e-03, 1.355200e-04, 6.304900e-09, 3.900900e-10,
      1.026600e-09, 6.989000e-10, 2.080300e-08, 3.663600e-06, 2.598300e-05,
      1.039200e-04, 1.176000e-04, 3.856900e-04, 7.802000e-04, 4.824500e-04,
      7.918900e-04, 1.909500e-03, 3.434800e-03, 4.349400e-03, 5.135600e-03,
      5.655600e-03, 5.878200e-03, 5.945600e-03, 5.888400e-03, 5.358100e-03,
      5.251200e-03, 5.466600e-03, 5.188800e-03, 5.019200e-03, 5.322300e-03,
      5.329300e-03, 4.962300e-03, 4.906500e-03, 4.907800e-03, 4.758300e-03,
      4.635100e-03, 4.549700e-03, 4.387800e-03, 4.171500e-03, 3.975400e-03,
      3.649700e-03, 3.310500e-03, 3.199800e-03, 2.885400e-03, 2.217900e-03,
      1.410200e-03, 8.046800e-04, 1.770200e-04, 1.369700e-05, 2.641100e-07,
      1.351600e-08, 1.148400e-13, 1.149100e-13, 5.224700e-13, 7.510800e-13,
      2.492700e-10, 1.011700e-09, 2.252800e-10, 3.023100e-10, 1.118100e-08,
      4.505200e-07, 9.457300e-06, 4.433600e-05, 1.414000e-04, 6.234700e-04,
      1.208500e-03, 1.414700e-03, 6.280600e-04, 3.566800e-04, 6.818400e-04,
      1.676500e-03, 1.940100e-03, 1.379200e-03, 1.169200e-03, 1.299200e-03,
      1.831500e-03, 2.001400e-03, 2.047200e-03, 2.126000e-03, 2.047500e-03,
      2.118500e-03, 2.154900e-03, 1.973200e-03, 1.865700e-03, 1.836400e-03,
      1.785600e-03, 1.756800e-03, 1.652600e-03, 1.779300e-03, 1.795500e-03,
      1.778100e-03, 1.729300e-03, 1.671600e-03, 1.587700e-03, 1.554900e-03,
      1.521300e-03, 1.467900e-03, 1.366100e-03, 1.397800e-03, 1.163200e-03,
      1.248400e-03, 1.040200e-03, 8.821400e-04, 9.897701e-04, 7.387200e-04,
      6.456600e-04, 6.346300e-04, 6.731300e-04, 4.484400e-04, 3.743200e-04,
      4.957000e-04, 4.089100e-04, 1.699800e-04, 2.972400e-04, 1.364900e-04,
      4.309100e-05, 4.828300e-06, 6.678000e-06, 1.201800e-06};

  SpectralVector EsCosThetaTau1(N_POINTS, 0.0);

  for (size_t i = 0; i < N_POINTS; i++) {
    EsCosThetaTau1.setData(i, escosthetatau1[i]);
  }

  //	double EsCosThetaTau1_mean = EsCosThetaTau1.average();

  //
  // 	The following does not work probably because of
  //	precision discrepancies.
  //
  //	EXPECT_EQ( 0.0063908, EsCosTheta1_mean);

  double ldtau2[N_POINTS] = {
      7.855000e-03, // 1
      8.020299e-03, // 2
      1.152600e-02, // 3
      1.373380e-02, // 4
      1.355700e-02, // 5
      1.224350e-02, // 6
      1.422900e-02, // 7
      1.604800e-02, // 8
      1.642500e-02, // 9
      1.650600e-02, // 10
      1.629400e-02, // 11
      1.544400e-02, // 12
      1.498800e-02, // 13
      1.459400e-02, // 14
      1.380100e-02, // 15
      1.404700e-02, // 16
      1.375400e-02, // 17
      1.355200e-02, // 18
      1.316900e-02, // 19
      1.269300e-02, // 20
      1.259200e-02, // 21
      1.102800e-02, // 22
      1.133900e-02, // 23
      1.146100e-02, // 24
      1.106100e-02, // 25
      1.055400e-02, // 26
      1.043600e-02, // 27
      9.457000e-03, // 28
      9.639999e-03, // 29
      9.909000e-03, // 30
      9.392999e-03, // 31
      7.475002e-03, // 32
      7.563001e-03, // 33
      7.619001e-03, // 34
      5.325001e-03, // 35
      5.353001e-03, // 36
      6.932000e-03, // 37
      7.389000e-03, // 38
      4.513000e-03, // 39
      5.686998e-03, // 40
      6.619999e-03, // 41
      5.895000e-03, // 42
      5.549001e-03, // 43
      4.791999e-03, // 44
      3.606000e-03, // 45
      3.982001e-03, // 46
      4.708000e-03, // 47
      4.787000e-03, // 48
      4.737001e-03, // 49
      4.613001e-03, // 50
      4.485000e-03, // 51
      3.901999e-03, // 52
      2.364000e-03, // 53
      2.012300e-03, // 54
      2.077000e-03, // 55
      9.826999e-04, // 56
      3.805999e-04, // 57
      4.395000e-04, // 58
      7.602000e-04, // 59
      1.636700e-03, // 60
      2.005000e-03, // 61
      2.791000e-03, // 62
      2.905999e-03, // 63
      2.812999e-03, // 64
      2.758999e-03, // 65
      2.700999e-03, // 66
      2.633000e-03, // 67
      2.554000e-03, // 68
      2.452000e-03, // 69
      2.286999e-03, // 70
      2.062000e-03, // 71
      1.766000e-03, // 72
      1.222900e-03, // 73
      6.591002e-04, // 74
      1.199000e-04, // 75
      8.102000e-05, // 76
      1.681000e-04, // 77
      1.808000e-04, // 78
      4.880000e-04, // 79
      8.764002e-04, // 80
      9.779003e-04, // 81
      1.003800e-03, // 82
      9.985999e-04, // 83
      1.042000e-03, // 84
      1.145600e-03, // 85
      1.258801e-03, // 86
      1.334400e-03, // 87
      1.295400e-03, // 88
      1.130699e-03, // 89
      9.792000e-04, // 90
      1.079600e-03, // 91
      1.007900e-03, // 92
      7.952000e-04, // 93
      6.034002e-04, // 94
      4.321001e-04, // 95
      2.237000e-04, // 96
      1.091000e-04, // 97
      8.320014e-06, // 98
      1.733000e-10, // 99
      1.766001e-11, // 100
      4.909995e-11, // 101
      2.413003e-11, // 102
      9.820003e-10, // 103
      1.815999e-07, // 104
      1.088001e-06, // 105
      5.709997e-06, // 106
      5.689995e-06, // 107
      2.144001e-05, // 108
      4.626997e-05, // 109
      2.909999e-05, // 110
      4.746998e-05, // 111
      1.260999e-04, // 112
      2.581999e-04, // 113
      3.511002e-04, // 114
      4.400001e-04, // 115
      5.108998e-04, // 116
      5.445001e-04, // 117
      5.524997e-04, // 118
      5.429997e-04, // 119
      4.753000e-04, // 120
      4.604999e-04, // 121
      4.853001e-04, // 122
      4.444001e-04, // 123
      4.241997e-04, // 124
      4.568002e-04, // 125
      4.530000e-04, // 126
      4.131999e-04, // 127
      4.005996e-04, // 128
      3.966000e-04, // 129
      3.791000e-04, // 130
      3.662002e-04, // 131
      3.504003e-04, // 132
      3.303001e-04, // 133
      3.052000e-04, // 134
      2.852003e-04, // 135
      2.532999e-04, // 136
      2.207002e-04, // 137
      2.093001e-04, // 138
      1.802000e-04, // 139
      1.216999e-04, // 140
      6.969995e-05, // 141
      3.767002e-05, // 142
      6.759990e-06, // 143
      4.379999e-07, // 144
      8.579974e-09, // 145
      4.500000e-10, // 146
      6.480005e-15, // 147
      8.650009e-15, // 148
      2.095004e-14, // 149
      2.649996e-14, // 150
      7.130019e-12, // 151
      2.940004e-11, // 152
      6.549997e-12, // 153
      9.289985e-12, // 154
      3.330003e-10, // 155
      1.365001e-08, // 156
      2.871002e-07, // 157
      1.105000e-06, // 158
      4.290006e-06, // 159
      2.275000e-05, // 160
      5.329994e-05, // 161
      6.410002e-05, // 162
      2.271996e-05, // 163
      1.083000e-05, // 164
      2.436002e-05, // 165
      7.920002e-05, // 166
      9.769993e-05, // 167
      5.979999e-05, // 168
      4.699989e-05, // 169
      5.370006e-05, // 170
      9.019999e-05, // 171
      1.015998e-04, // 172
      1.050001e-04, // 173
      1.103000e-04, // 174
      1.045000e-04, // 175
      1.102001e-04, // 176
      1.129000e-04, // 177
      9.890017e-05, // 178
      9.210000e-05, // 179
      9.079999e-05, // 180
      8.620007e-05, // 181
      8.380006e-05, // 182
      7.750001e-05, // 183
      8.579995e-05, // 184
      8.709996e-05, // 185
      8.659996e-05, // 186
      8.389994e-05, // 187
      8.029991e-05, // 188
      7.509999e-05, // 189
      7.290009e-05, // 190
      7.070007e-05, // 191
      6.760005e-05, // 192
      6.130000e-05, // 193
      6.190001e-05, // 194
      4.880002e-05, // 195
      5.300005e-05, // 196
      4.139997e-05, // 197
      3.323000e-05, // 198
      3.892998e-05, // 199
      2.719002e-05, // 200
      2.230000e-05, // 201
      2.170005e-05, // 202
      2.338999e-05, // 203
      1.412001e-05, // 204
      1.186001e-05, // 205
      1.670001e-05, // 206
      1.368998e-05, // 207
      5.119990e-06, // 208
      9.309995e-06, // 209
      3.820009e-06, // 210
      1.004999e-06, // 211
      9.940004e-08, // 212
      1.385001e-07, // 213
      3.349999e-08  // 214
  };

  SpectralVector LdTau2(N_POINTS, 0.0);

  for (size_t i = 0; i < N_POINTS; i++) {
    LdTau2.setData(i, ldtau2[i]);
  }

  //	double LdTau2_mean = LdTau2.average();

  //
  // 	The following does not work probably because of
  //	precision discrepancies.
  //
  //	EXPECT_EQ( 0.00270336 , LdTau2_mean );

  double lu[N_POINTS] = {
      5.303300e-03, // 1
      4.835500e-03, // 2
      6.269100e-03, // 3
      6.849800e-03, // 4
      6.213900e-03, // 5
      5.164800e-03, // 6
      5.558800e-03, // 7
      5.833900e-03, // 8
      5.576800e-03, // 9
      5.255800e-03, // 10
      4.880700e-03, // 11
      4.360900e-03, // 12
      4.026200e-03, // 13
      3.727500e-03, // 14
      3.324100e-03, // 15
      3.225800e-03, // 16
      3.030400e-03, // 17
      2.853700e-03, // 18
      2.648800e-03, // 19
      2.477000e-03, // 20
      2.353200e-03, // 21
      2.107400e-03, // 22
      2.016100e-03, // 23
      1.901400e-03, // 24
      1.774400e-03, // 25
      1.660600e-03, // 26
      1.586000e-03, // 27
      1.448900e-03, // 28
      1.389500e-03, // 29
      1.359700e-03, // 30
      1.264800e-03, // 31
      1.079300e-03, // 32
      1.064900e-03, // 33
      1.017800e-03, // 34
      8.451200e-04, // 35
      8.118200e-04, // 36
      8.509400e-04, // 37
      8.371200e-04, // 38
      5.737100e-04, // 39
      6.580000e-04, // 40
      7.064400e-04, // 41
      6.460300e-04, // 42
      6.048300e-04, // 43
      5.482100e-04, // 44
      4.687800e-04, // 45
      4.693400e-04, // 46
      4.819000e-04, // 47
      4.577800e-04, // 48
      4.403700e-04, // 49
      4.206500e-04, // 50
      4.052900e-04, // 51
      3.709800e-04, // 52
      2.923900e-04, // 53
      2.655100e-04, // 54
      2.597700e-04, // 55
      1.841700e-04, // 56
      1.340500e-04, // 57
      1.367800e-04, // 58
      1.568700e-04, // 59
      1.996800e-04, // 60
      2.111400e-04, // 61
      2.329900e-04, // 62
      2.292500e-04, // 63
      2.191400e-04, // 64
      2.112400e-04, // 65
      2.037000e-04, // 66
      1.953500e-04, // 67
      1.876300e-04, // 68
      1.788600e-04, // 69
      1.680900e-04, // 70
      1.566100e-04, // 71
      1.421300e-04, // 72
      1.194700e-04, // 73
      9.158900e-05, // 74
      4.907500e-05, // 75
      4.300600e-05, // 76
      5.107100e-05, // 77
      5.138000e-05, // 78
      7.210500e-05, // 79
      8.772000e-05, // 80
      8.773600e-05, // 81
      8.692100e-05, // 82
      8.444200e-05, // 83
      8.383700e-05, // 84
      8.595200e-05, // 85
      8.736800e-05, // 86
      8.737800e-05, // 87
      8.426000e-05, // 88
      7.596200e-05, // 89
      6.833100e-05, // 90
      7.159700e-05, // 91
      6.952000e-05, // 92
      6.243500e-05, // 93
      5.442100e-05, // 94
      4.685500e-05, // 95
      3.517300e-05, // 96
      2.632800e-05, // 97
      1.091100e-05, // 98
      4.869800e-06, // 99
      4.772200e-06, // 100
      4.938800e-06, // 101
      5.047700e-06, // 102
      5.029600e-06, // 103
      5.994800e-06, // 104
      7.695300e-06, // 105
      8.795100e-06, // 106
      9.074900e-06, // 107
      1.199900e-05, // 108
      1.529900e-05, // 109
      1.221700e-05, // 110
      1.428400e-05, // 111
      2.010500e-05, // 112
      2.634900e-05, // 113
      2.938300e-05, // 114
      3.191900e-05, // 115
      3.309900e-05, // 116
      3.314400e-05, // 117
      3.274800e-05, // 118
      3.181800e-05, // 119
      2.857900e-05, // 120
      2.749600e-05, // 121
      2.808500e-05, // 122
      2.630300e-05, // 123
      2.502900e-05, // 124
      2.600100e-05, // 125
      2.562500e-05, // 126
      2.357000e-05, // 127
      2.299800e-05, // 128
      2.269500e-05, // 129
      2.168600e-05, // 130
      2.078500e-05, // 131
      2.028100e-05, // 132
      1.942200e-05, // 133
      1.835600e-05, // 134
      1.745500e-05, // 135
      1.609100e-05, // 136
      1.480800e-05, // 137
      1.435100e-05, // 138
      1.328800e-05, // 139
      1.126500e-05, // 140
      8.664500e-06, // 141
      6.535200e-06, // 142
      3.529000e-06, // 143
      2.010900e-06, // 144
      1.289900e-06, // 145
      9.659300e-07, // 146
      7.455000e-07, // 147
      7.502100e-07, // 148
      7.677500e-07, // 149
      6.233400e-07, // 150
      7.696600e-07, // 151
      8.393300e-07, // 152
      6.743200e-07, // 153
      6.904800e-07, // 154
      7.991200e-07, // 155
      9.264100e-07, // 156
      1.156900e-06, // 157
      1.388600e-06, // 158
      1.674600e-06, // 159
      3.337400e-06, // 160
      5.009000e-06, // 161
      5.423300e-06, // 162
      2.739400e-06, // 163
      1.758800e-06, // 164
      2.781400e-06, // 165
      5.325800e-06, // 166
      5.801000e-06, // 167
      4.277800e-06, // 168
      3.705400e-06, // 169
      3.991000e-06, // 170
      5.155900e-06, // 171
      5.450200e-06, // 172
      5.502300e-06, // 173
      5.557600e-06, // 174
      5.338800e-06, // 175
      5.382600e-06, // 176
      5.361300e-06, // 177
      4.973900e-06, // 178
      4.679200e-06, // 179
      4.530500e-06, // 180
      4.417000e-06, // 181
      4.312300e-06, // 182
      4.038500e-06, // 183
      4.197100e-06, // 184
      4.152000e-06, // 185
      4.038200e-06, // 186
      3.886800e-06, // 187
      3.725600e-06, // 188
      3.515200e-06, // 189
      3.408100e-06, // 190
      3.299900e-06, // 191
      3.164100e-06, // 192
      2.956800e-06, // 193
      3.006400e-06, // 194
      2.591000e-06, // 195
      2.711400e-06, // 196
      2.355100e-06, // 197
      2.070200e-06, // 198
      2.220500e-06, // 199
      1.786000e-06, // 200
      1.681200e-06, // 201
      1.650600e-06, // 202
      1.696400e-06, // 203
      1.347400e-06, // 204
      1.199000e-06, // 205
      1.386400e-06, // 206
      1.219700e-06, // 207
      7.663000e-07, // 208
      1.055800e-06, // 209
      7.335300e-07, // 210
      4.550700e-07, // 211
      2.773000e-07, // 212
      2.998100e-07, // 213
      2.014700e-07  // 214
  };

  SpectralVector Lu(N_POINTS, 0.0);

  for (size_t i = 0; i < N_POINTS; i++) {
    Lu.setData(i, lu[i]);
  }

  // double Lu_mean = Lu.average();

  //
  // 	The following does not work probably because of
  //	precision discrepancies.
  //
  //	EXPECT_EQ( 0.00061113 , Lu_mean );

  double lsensor[N_POINTS] = {
      1.686530e-02, // 1
      1.705450e-02, // 2
      2.442210e-02, // 3
      2.912580e-02, // 4
      2.886990e-02, // 5
      2.626080e-02, // 6
      3.082380e-02, // 7
      3.513890e-02, // 8
      3.639680e-02, // 9
      3.708480e-02, // 10
      3.712370e-02, // 11
      3.572890e-02, // 12
      3.528620e-02, // 13
      3.494250e-02, // 14
      3.347510e-02, // 15
      3.463980e-02, // 16
      3.453540e-02, // 17
      3.458670e-02, // 18
      3.419580e-02, // 19
      3.376200e-02, // 20
      3.407620e-02, // 21
      3.118340e-02, // 22
      3.220810e-02, // 23
      3.266440e-02, // 24
      3.211440e-02, // 25
      3.141660e-02, // 26
      3.153800e-02, // 27
      2.958890e-02, // 28
      3.027450e-02, // 29
      3.134670e-02, // 30
      3.034480e-02, // 31
      2.581030e-02, // 32
      2.646890e-02, // 33
      2.682280e-02, // 34
      2.057912e-02, // 35
      2.104882e-02, // 36
      2.575594e-02, // 37
      2.723012e-02, // 38
      1.767871e-02, // 39
      2.245400e-02, // 40
      2.583544e-02, // 41
      2.397003e-02, // 42
      2.313483e-02, // 43
      2.088021e-02, // 44
      1.694478e-02, // 45
      1.848334e-02, // 46
      2.113090e-02, // 47
      2.131878e-02, // 48
      2.132037e-02, // 49
      2.106865e-02, // 50
      2.088829e-02, // 51
      1.900898e-02, // 52
      1.307939e-02, // 53
      1.169751e-02, // 54
      1.221777e-02, // 55
      6.438370e-03, // 56
      3.025650e-03, // 57
      3.508680e-03, // 58
      5.620270e-03, // 59
      1.048768e-02, // 60
      1.248014e-02, // 61
      1.594099e-02, // 62
      1.644925e-02, // 63
      1.614314e-02, // 64
      1.599724e-02, // 65
      1.584070e-02, // 66
      1.560035e-02, // 67
      1.534463e-02, // 68
      1.496286e-02, // 69
      1.430909e-02, // 70
      1.341261e-02, // 71
      1.206913e-02, // 72
      9.332370e-03, // 73
      5.755789e-03, // 74
      1.299375e-03, // 75
      9.423260e-04, // 76
      1.792671e-03, // 77
      1.950480e-03, // 78
      4.793805e-03, // 79
      7.678820e-03, // 80
      8.243336e-03, // 81
      8.531922e-03, // 82
      8.590742e-03, // 83
      8.887337e-03, // 84
      9.619252e-03, // 85
      1.031437e-02, // 86
      1.075338e-02, // 87
      1.059726e-02, // 88
      9.669262e-03, // 89
      8.771331e-03, // 90
      9.399097e-03, // 91
      9.071020e-03, // 92
      7.769835e-03, // 93
      6.363721e-03, // 94
      5.006555e-03, // 95
      2.935973e-03, // 96
      1.566628e-03, // 97
      1.547510e-04, // 98
      4.876278e-06, // 99
      4.772608e-06, // 100
      4.939875e-06, // 101
      5.048423e-06, // 102
      5.051385e-06, // 103
      9.840000e-06, // 104
      3.476630e-05, // 105
      1.184251e-04, // 106
      1.323649e-04, // 107
      4.191290e-04, // 108
      8.417690e-04, // 109
      5.237670e-04, // 110
      8.536440e-04, // 111
      2.055705e-03, // 112
      3.719349e-03, // 113
      4.729883e-03, // 114
      5.607519e-03, // 115
      6.199599e-03, // 116
      6.455844e-03, // 117
      6.530848e-03, // 118
      6.463218e-03, // 119
      5.861979e-03, // 120
      5.739196e-03, // 121
      5.979985e-03, // 122
      5.659503e-03, // 123
      5.468429e-03, // 124
      5.805101e-03, // 125
      5.807925e-03, // 126
      5.399070e-03, // 127
      5.330098e-03, // 128
      5.327095e-03, // 129
      5.159086e-03, // 130
      5.022085e-03, // 131
      4.920381e-03, // 132
      4.737522e-03, // 133
      4.495056e-03, // 134
      4.278055e-03, // 135
      3.919091e-03, // 136
      3.546008e-03, // 137
      3.423451e-03, // 138
      3.078888e-03, // 139
      2.350865e-03, // 140
      1.488564e-03, // 141
      8.488852e-04, // 142
      1.873090e-04, // 143
      1.614590e-05, // 144
      1.562590e-06, // 145
      9.798960e-07, // 146
      7.455001e-07, // 147
      7.502101e-07, // 148
      7.677506e-07, // 149
      6.233408e-07, // 150
      7.699164e-07, // 151
      8.403711e-07, // 152
      6.745518e-07, // 153
      6.907916e-07, // 154
      8.106340e-07, // 155
      1.390580e-06, // 156
      1.090130e-05, // 157
      4.682960e-05, // 158
      1.473646e-04, // 159
      6.495574e-04, // 160
      1.266809e-03, // 161
      1.484223e-03, // 162
      6.535194e-04, // 163
      3.692688e-04, // 164
      7.089814e-04, // 165
      1.761026e-03, // 166
      2.043601e-03, // 167
      1.443278e-03, // 168
      1.219905e-03, // 169
      1.356891e-03, // 170
      1.926856e-03, // 171
      2.108450e-03, // 172
      2.157702e-03, // 173
      2.241858e-03, // 174
      2.157339e-03, // 175
      2.234083e-03, // 176
      2.273161e-03, // 177
      2.077074e-03, // 178
      1.962479e-03, // 179
      1.931731e-03, // 180
      1.876217e-03, // 181
      1.844912e-03, // 182
      1.734138e-03, // 183
      1.869297e-03, // 184
      1.886752e-03, // 185
      1.868738e-03, // 186
      1.817087e-03, // 187
      1.755626e-03, // 188
      1.666315e-03, // 189
      1.631208e-03, // 190
      1.595300e-03, // 191
      1.538664e-03, // 192
      1.430357e-03, // 193
      1.462706e-03, // 194
      1.214591e-03, // 195
      1.304111e-03, // 196
      1.083955e-03, // 197
      9.174402e-04, // 198
      1.030921e-03, // 199
      7.676960e-04, // 200
      6.696412e-04, // 201
      6.579806e-04, // 202
      6.982164e-04, // 203
      4.639074e-04, // 204
      3.873790e-04, // 205
      5.137864e-04, // 206
      4.238197e-04, // 207
      1.758663e-04, // 208
      3.076058e-04, // 209
      1.410435e-04, // 210
      4.455107e-05, // 211
      5.205000e-06, // 212
      7.116310e-06, // 213
      1.436770e-06  // 214
  };

  SpectralVector Lsensor(N_POINTS, 0.0);

  for (size_t i = 0; i < N_POINTS; i++) {
    Lsensor.setData(i, lsensor[i]);
  }

  //	double Lsensor_mean = Lsensor.average();

  //
  // 	The following does not work probably because of
  //	precision discrepancies, but the numbers below
  // 	are reported as the same in the gtest output.
  //
  //	EXPECT_EQ(  0.00970529 , Lsensor_mean );

  //
  // We will now compute Lsensor vector from the following
  // components.
  //
  //	Lsensor = (EsCosThetaTau1)*r+LdTau2*r+Lu
  //

  //
  // The following would contain the reflectance
  // spectrum that would be multiplied by EsCosThetaTau1
  // and LdTau2, respectively.
  //
  SpectralVector rE(N_POINTS, 1.0);
  SpectralVector rL(N_POINTS, 1.0);

  SpectralVector answer;

  EXPECT_EQ(1.0, rE[213]);
  EXPECT_EQ(1.0, rL[213]);
  EXPECT_EQ(1.201800e-06, EsCosThetaTau1[213]);

  rE *= EsCosThetaTau1;

  EXPECT_EQ(1.201800e-06, EsCosThetaTau1[213]);
  EXPECT_EQ(1.201800e-06, rE[213]);

  //	EXPECT_EQ( 1.201800e-06   , r[213] );
  //	EXPECT_EQ(  3.349999e-08, LdTau2[213] );

  rL *= LdTau2;
  EXPECT_EQ(3.349999e-08, LdTau2[213]);
  EXPECT_EQ(3.349999e-08, rL[213]);

  //	EXPECT_EQ(   1.201800e-06 * 3.349999e-08, r[213] );
  //	EXPECT_EQ(   4.02603e-14 , r[213] );

  /*
          answer+=rE;
          answer+=rL;
          answer+=Lu;
  */
  answer = rE;
  EXPECT_EQ(1.201800e-06, answer[213]);

  answer += rL;
  EXPECT_EQ(1.201800e-06 + 3.349999e-08, answer[213]);

  answer += Lu;
  EXPECT_EQ(2.014700e-07, Lu[213]);
  EXPECT_EQ(1.201800e-06 + 3.349999e-08 + 2.014700e-07, answer[213]);
  EXPECT_EQ(EsCosThetaTau1[213] + LdTau2[213] + 2.014700e-07, answer[213]);
  //	EXPECT_EQ( 1.436770e-06, answer[213] );

  //	int n_points = answer.getPointCount();

  //
  // Although the following test block fails under gtest because
  // of precision errors, it does display the fact that the numbers
  // from the ENVI calculated spectral vectors are similar at that
  // computed by the SpectralVector routines.
  // RR - 10/30/2015
  //

  /*
          for (int i=0; i<n_points; i++) {

                  EXPECT_EQ( Lsensor[i], answer[i]);
          }
  */

  //
  // We want to test the SpectralVector pow() by computing the RMS
  // between the ENVI calculated and SpectralVector calculated
  // results.
  //

  SpectralVector error = Lsensor;
  error -= answer;
  SpectralVector squared_error = error.pow(2.0);

  EXPECT_EQ(error[213] * error[213], squared_error[213]);

  //	double error_mean = squared_error.average();

  //	double rms = sqrt( error_mean );

  //
  // 	The following does not work probably because of
  //	precision discrepancies, but the numbers below
  // 	are reported as the same in the gtest output.
  //
  //	EXPECT_EQ( 4.21081e-10, rms );
}
