#include "SpectralFixture.h"
#include "dirsig5/base/SpectralVector.h"

using namespace d5;

/*
The following will test the interface SpectralVector.h.


The tests will be as follows

    Case00
                Check null input wavelengths
                ASSERT_ANY_THROW

    Case01
                Check null output wavelengths
                ASSERT_ANY_THROW

    Case02
                Check single valued inputs
                ASSERT_NO_THROW

    Case03
                Test random vectors as input
                ASSERT_NO_THROW

    Case03
                Test NaN wavelengths on i/o
                ASSERT_NO_THROW

    mixed vector types are tested through


*/

TEST_F(SpectralFixture, InterpCase00) {
  // null input wavelengths
  std::vector<float> inputWavelengths(0);

  std::vector<float> outputData(30);

  ASSERT_ANY_THROW(
      interpolate(inputWavelengths, genRandomVector<std::vector<float>>(10),
                  genWavelengthVector<std::vector<float>>(0.4, 0.7, 30),
                  outputData););

  ASSERT_ANY_THROW(
      interpolate(inputWavelengths, genRandomVector<SpectralVector>(10),
                  genWavelengthVector<std::vector<float>>(0.4, 0.7, 30),
                  outputData););

  ASSERT_ANY_THROW(
      interpolate(inputWavelengths, genRandomVector<SpectralVector>(10),
                  genWavelengthVector<SpectralVector>(0.4, 0.7, 30),
                  outputData););
}

TEST_F(SpectralFixture, InterpCase01) {
  // null input wavelengths
  std::vector<float> outputWavelengths(0);

  std::vector<float> outputData(30);

  ASSERT_ANY_THROW(
      interpolate(genWavelengthVector<std::vector<float>>(0.4, 0.7, 30),
                  genRandomVector<std::vector<float>>(30), outputWavelengths,
                  outputData););

  ASSERT_ANY_THROW(
      interpolate(genWavelengthVector<std::vector<float>>(0.4, 0.7, 30),
                  genRandomVector<SpectralVector>(30), outputWavelengths,
                  outputData););
}

TEST_F(SpectralFixture, InterpCase02) {
  // output is size one
  std::vector<float> outputData(1);
  ASSERT_NO_THROW(
      interpolate(genWavelengthVector<std::vector<float>>(0.4, 0.7, 30),
                  genRandomVector<std::vector<float>>(30),
                  genWavelengthVector<std::vector<float>>(0.5, 0.6, 1),
                  outputData););

  // input is size one
  outputData.resize(10);
  ASSERT_NO_THROW(
      interpolate(genWavelengthVector<std::vector<float>>(0.4, 0.7, 1),
                  genRandomVector<std::vector<float>>(1),
                  genWavelengthVector<std::vector<float>>(0.5, 0.6, 10),
                  outputData););
}

TEST_F(SpectralFixture, InterpCase03) {
  size_t li = randomLength();
  size_t lo = randomLength();
  double min_i, max_i, min_o, max_o;
  randomLimits(min_i, max_i);
  randomLimits(min_o, max_o);

  // randomly generate some data
  std::vector<float> inputData = genRandomVector<std::vector<float>>(li);
  std::vector<float> inputWavelengths =
      genWavelengthVector<std::vector<float>>(min_i, max_i, li);
  std::vector<float> outputData(lo);
  std::vector<float> outputWavelengths =
      genWavelengthVector<std::vector<float>>(min_o, max_o, lo);

  ASSERT_NO_THROW(
      interpolate(genWavelengthVector<std::vector<float>>(min_i, max_i, li),
                  inputData, outputWavelengths, outputData););

  // check the interpolated data
  size_t idx = randomIndex(outputData.size());
  float wlo = outputWavelengths[idx];
  float valo = outputData[idx];

  // check extrapolation
  if (wlo >= max_i) {
    EXPECT_NEAR(valo, inputData.back(), 1e-6);
  } else if (wlo <= min_i) {
    EXPECT_NEAR(valo, inputData.front(), 1e-6);
  } else {
    size_t idxi2 = 1;
    for (; idxi2 < inputWavelengths.size(); idxi2++) {
      if (wlo < inputWavelengths[idxi2]) {
        break;
      }
    }
    size_t idxi1 = idxi2 - 1;
    ASSERT_LE(valo, std::max(inputData[idxi1], inputData[idxi2]));
    ASSERT_GE(valo, std::min(inputData[idxi1], inputData[idxi2]));
  }
}

TEST_F(SpectralFixture, InterpCase04) {
  std::vector<float> outputWavelengths =
      genWavelengthVector<std::vector<float>>(0.4, 0.7, 300);
  std::vector<float> outputData;

#ifndef _MSC_VER
  ASSERT_ANY_THROW(std::vector<float> inNaN =
                       genWavelengthVector<std::vector<float>>(0.4, 0.7, 30);
                   inNaN[12] = 1.0 / 0.0;
                   interpolate(inNaN, genRandomVector<std::vector<float>>(30),
                               outputWavelengths, outputData););

  ASSERT_ANY_THROW(outputWavelengths[194] = 1.0 / 0.0; interpolate(
                       genWavelengthVector<std::vector<float>>(0.4, 0.7, 30),
                       genRandomVector<SpectralVector>(30), outputWavelengths,
                       outputData););
#endif
}
