#include <gtest/gtest.h>

#include "dirsig5/base/SpectralVector.h"

using namespace d5;

/*
The following will test the interface SpectralVector.h.


The tests will be as follows

        Case00
                Testing the default constructor.
                ASSERT_NO_THROW

        Case01
                Testing the constructor with a
                negative point count.
                ASSERT_ANY_THROW

        Case02
                Testing the constructor with a
                zero point count.
                ASSERT_NO_THROW

        Case03
                Testing the constructor with a
                very large 64-bit point count.
                ASSERT_ANY_THROW

        Case04
                Testing the constructor with a
                very large 32-bit point count.
                ASSERT_ANY_THROW

        Case05
                Testing the constructor with a
                very large 31-bit point count.
                2^31.
                ASSERT_ANY_THROW

        Case06
                Testing the constructor with a
                very large 31-bit point count.
                2^31-1
                ASSERT_ANY_THROW

        Case07
                Testing the constructor with a
                very large 16-bit point count.
                ASSERT_NO_THROW

        Case08
                Testing the copy constructor using
                different syntaxes.
                ASSERT_NO_THROW

        Case09
                Testing the setData value for setting
                an overall value.
                ASSERT_NO_THROW

        Case10
                Testing the setData value for a specific
                index location which is negative (-1).
                ASSERT_ANY_THROW


        Case11
                Testing the setData value for a specific
                index location which is outside of the
                array bounds (65537 > 65536).


        Case12
                Deposit a small array of values into
                a spectral vector for testing.


        Case13
                A simplified case of a probem found when
                testing SpectralVectorBigEquation which
                is causing a core dump.
                SpectralVector created by the default
                constructor and then immediately used as
                a destination of an inplace SpectralVector
                addition.

*/

TEST(SpectralVector, Case00) {
  //
  // Call should NOT throw an exception because
  // it is just the default constructor
  //

  ASSERT_NO_THROW(SpectralVector spectrum;);
}

TEST(SpectralVector, Case01) {
  //
  // Call should throw an exception because we
  // are setting the pointCount to a negative value
  //
  size_t pointCount = -1;
  double value = 0.0;

  ASSERT_ANY_THROW(SpectralVector spectrum(pointCount, value););
}

TEST(SpectralVector, Case02) {
  //
  // Call should throw an exception because we
  // are setting the pointcount to a zero pointCount
  //
  size_t pointCount = 0;
  double value = 0.0;

  ASSERT_NO_THROW(SpectralVector spectrum(pointCount, value););
}

TEST(SpectralVector, Case03) {
  //
  // Call should not throw an exception because we
  // are setting the pointcount to within 32 bits
  //
  size_t pointCount = 4294967296;
  double value = 0.0;

  ASSERT_ANY_THROW(SpectralVector spectrum(pointCount, value););
}

TEST(SpectralVector, Case04) {
  //
  // Call should not throw an exception because we
  // are setting the pointcount to within 32 bits
  //
  size_t pointCount = 4294967295;
  double value = 0.0;

  ASSERT_ANY_THROW(SpectralVector spectrum(pointCount, value););
}

TEST(SpectralVector, Case05) {
  //
  // Call should not throw an exception because we
  // are setting the pointcount to within 32 bits
  //
  // 2^31
  size_t pointCount = 2147483648;
  double value = 0.0;

  ASSERT_ANY_THROW(SpectralVector spectrum(pointCount, value););
}

TEST(SpectralVector, Case06) {
  //
  // Call should not throw an exception because we
  // are setting the pointcount to within 32 bits
  //
  // 2^31-1
  size_t pointCount = 2147483647;
  double value = 0.0;

  ASSERT_ANY_THROW(SpectralVector spectrum(pointCount, value););
}

TEST(SpectralVector, Case07) {
  //
  // Call should not throw an exception because we
  // are setting the pointcount to within 16 bits
  //
  size_t pointCount = 65536;
  double value = 0.0;

  ASSERT_NO_THROW(SpectralVector spectrum(pointCount, value););
}

TEST(SpectralVector, Case08) {
  //
  // Call should not throw an exception because we
  // are setting the pointcount to within 16 bits
  //
  size_t pointCount = 65536;
  double value = 0.0;

  // We are testing the copy constructor

  ASSERT_NO_THROW(SpectralVector spectrum(pointCount, value);
                  SpectralVector spectrumCopy(spectrum););

  ASSERT_NO_THROW(SpectralVector spectrum(pointCount, value);
                  SpectralVector spectrumCopy = spectrum;);

  /*
  Why does the following NOT WORK?
  I thought that the scope of a defined variable went outside
  of the ASSERT_* directives :-(

  AAG verified that it DOES NOT (October 23, 2015).

          ASSERT_NO_THROW(
                  SpectralVector spectrum( pointCount , value );
          );

          ASSERT_NO_THROW(
                  SpectralVector spectrumCopy(spectrum);
          );
  */

  SpectralVector spectrum(pointCount, value);

  ASSERT_NO_THROW(SpectralVector spectrum2 = spectrum;
                  SpectralVector spectrum3(spectrum2););
}

TEST(SpectralVector, Case09) {
  //
  // Call should not throw an exception because we
  // are setting the pointcount to within 16 bits
  //
  size_t pointCount = 65536;
  double value = 0.0;
  double newValue = 1e308;
  SpectralVector spectrum(pointCount, value);

  ASSERT_NO_THROW(spectrum.setData(newValue););
}

TEST(SpectralVector, Case10) {
  //
  // Call should throw an exception because we
  // are setting the pointcount to within 16 bits
  //
  size_t pointCount = 65536;
  double value = 0.0;
  double newValue = 1e308;
  SpectralVector spectrum(pointCount, value);

  size_t index = -1;

  ASSERT_ANY_THROW(spectrum.setData(index, newValue););
}

TEST(SpectralVector, Case11) {
  //
  // Call should not throw an exception because we
  // are setting the pointcount to within 16 bits
  //
  size_t pointCount = 65536;
  double value = 0.0;
  double newValue = 1e308;
  SpectralVector spectrum(pointCount, value);

  size_t index = 65537;

  ASSERT_ANY_THROW(spectrum.setData(index, newValue););
}

TEST(SpectralVector, Case12) {
  //
  //  Deposit a small array of doubles into
  //  a newly formed spectral vector
  //
  double testArray[3] = {0.0, -1.0, 2.0};
  size_t pointCount = 3;
  double value = 0.0;
  SpectralVector spectrum(pointCount, value);

  /*
          for (int i=0; i< 2; i++ ){

                  newValue = testArray[i];
                  spectrum.setData( i, newValue );
          }
  */
  spectrum.setData(0, testArray[0]);
  spectrum.setData(1, testArray[1]);
  spectrum.setData(2, testArray[2]);

  EXPECT_EQ(spectrum.min(), -1.0);
  EXPECT_EQ(spectrum.max(), 2.0);
}

TEST(SpectralVector, Case13) {
  //
  //  Deposit a small array of doubles into
  //  a newly formed spectral vector
  //
  double testArray[3] = {0.0, -1.0, 2.0};
  size_t pointCount = 3;
  double newValue = 0.0;
  SpectralVector spectrum(pointCount, newValue);

  spectrum.setData(0, testArray[0]);
  spectrum.setData(1, testArray[1]);
  spectrum.setData(2, testArray[2]);

  SpectralVector answer;
  ASSERT_ANY_THROW(answer += spectrum;);
}

TEST(SpectralVector, Case14) {
  //
  //  Deposit a small array of doubles into
  //  a newly formed spectral vector
  //
  double testArray[3] = {0.0, -1.0, 2.0};
  size_t pointCount = 3;
  double newValue = 0.0;
  SpectralVector spectrum(pointCount, newValue);

  spectrum.setData(0, testArray[0]);
  spectrum.setData(1, testArray[1]);
  spectrum.setData(2, testArray[2]);

  SpectralVector answer;
  ASSERT_ANY_THROW(answer -= spectrum;);
}

TEST(SpectralVector, Case15) {
  //
  //  Deposit a small array of doubles into
  //  a newly formed spectral vector
  //
  double testArray[3] = {0.0, -1.0, 2.0};
  size_t pointCount = 3;
  double newValue = 0.0;
  SpectralVector spectrum(pointCount, newValue);

  spectrum.setData(0, testArray[0]);
  spectrum.setData(1, testArray[1]);
  spectrum.setData(2, testArray[2]);

  SpectralVector answer;
  ASSERT_ANY_THROW(answer *= spectrum;);
}

TEST(SpectralVector, Case16) {
  //
  //  Deposit a small array of doubles into
  //  a newly formed spectral vector
  //
  double testArray[3] = {0.0, -1.0, 2.0};
  size_t pointCount = 3;
  double newValue = 0.0;
  SpectralVector spectrum(pointCount, newValue);

  spectrum.setData(0, testArray[0]);
  spectrum.setData(1, testArray[1]);
  spectrum.setData(2, testArray[2]);

  SpectralVector answer;
  ASSERT_ANY_THROW(answer /= spectrum;);
}

TEST(SpectralVector, Case17) {
  //
  //  Deposit a small array of doubles into
  //  a newly formed spectral vector
  //
  double testArray[3] = {0.0, -1.0, 2.0};
  size_t pointCount = 3;
  double newValue = 0.0;
  SpectralVector spectrum(pointCount, newValue);

  spectrum.setData(0, testArray[0]);
  spectrum.setData(1, testArray[1]);
  spectrum.setData(2, testArray[2]);

  SpectralVector answer;
  ASSERT_NO_THROW(answer.copyData(spectrum););
}
