/** Unit tests for StringUtils.h
 *
 *  \file StringUtilsTest.cpp
 */

#include "dirsig5/base/StringUtils.h"

#include <gtest/gtest.h>

#include "CLStdComparisons.h"

using namespace d5;

struct A {
  int a;
};
std::ostream &operator<<(std::ostream &os, const A &a) {
  os << "a=" << a.a;
  return os;
}

struct B {
  int b;
};

TEST(StringUtilsTest, JoinTest) {
  EXPECT_EQ(Join(std::vector<int>{5, 10, 15}, ", "), "5, 10, 15");
  EXPECT_EQ(Join(std::vector<double>{}, ", "), "");
  EXPECT_EQ(Join(std::vector<A>{{1}, {2}, {3}}, ", "), "a=1, a=2, a=3");

  std::vector<A> as{{1}, {2}, {3}};
  EXPECT_EQ(Join(as.begin(), as.end(), ", "), "a=1, a=2, a=3");

  std::vector<B> bs{{1}, {2}, {3}};
  EXPECT_EQ(Join(bs.begin(), bs.end(), ", ", [](const auto &b) { return b.b; }),
            "1, 2, 3");
}

TEST(StringUtilsTest, ParseNumberTest) {
  using d5::NumericParseStatus;
  using d5::ParseNumber;

  // Easy, high-level cases
  EXPECT_EQ(ParseNumber<std::uint32_t>("   98"), 98);
  EXPECT_EQ(ParseNumber<std::uint32_t>("876"), 876);
  EXPECT_EQ(ParseNumber<std::uint32_t>("\n 12"), 12);
  EXPECT_EQ(ParseNumber<std::uint32_t>("\t 13"), 13);
  EXPECT_EQ(ParseNumber<int>("   98"), 98);
  EXPECT_EQ(ParseNumber<int>("876"), 876);
  EXPECT_EQ(ParseNumber<int>("\n -15"), -15);
  EXPECT_EQ(ParseNumber<int>("\t -1587"), -1587);
  EXPECT_DOUBLE_EQ(ParseNumber<double>("\t -15.87"), -15.87);
  EXPECT_DOUBLE_EQ(ParseNumber<double>("\n +1.687e1"), 16.87);
  EXPECT_DOUBLE_EQ(ParseNumber<double>("-1.687E-2"), -.01687);

  // Test the other overload happy path
  std::uint32_t u32;
  EXPECT_NO_THROW(ParseNumber("\n  53897", u32));
  EXPECT_EQ(u32, 53897);

  //                   0         1           2          3         4
  //                   012345678901 23456 78901 2345678901234567890123456789
  std::string input = "   98  876 \n 12 \t -13\n-15.87 +1.687e1 -1.687E-2  \n";

  // Test the behavior of the lower-level overload
  char *end;
  EXPECT_EQ(ParseNumber(input.data(), &end, u32), NumericParseStatus::Success);
  EXPECT_EQ(end, input.data() + 5);
  EXPECT_EQ(u32, 98);

  EXPECT_EQ(ParseNumber(input.data() + 5, &end, u32),
            NumericParseStatus::Success);
  EXPECT_EQ(end, input.data() + 10);
  EXPECT_EQ(u32, 876);

  double d;
  EXPECT_EQ(ParseNumber(input.data() + 37, &end, d),
            NumericParseStatus::Success);
  EXPECT_EQ(end, input.data() + 47);
  EXPECT_DOUBLE_EQ(d, -0.01687);

  // Test error behavior
  //                     0         1         2         3         4
  //                     0123456789012345678901234567890123456789012
  std::string uintBad = "65536 4294967296 18446744073709551616 hello";

  std::uint16_t u16;
  EXPECT_EQ(ParseNumber(uintBad.data(), &end, u16),
            NumericParseStatus::Overflow);
  EXPECT_EQ(end, uintBad.data() + 5);

  EXPECT_EQ(ParseNumber(uintBad.data() + 5, &end, u32),
            NumericParseStatus::Overflow);
  EXPECT_EQ(end, uintBad.data() + 16);

  std::uint64_t u64;
  EXPECT_EQ(ParseNumber(uintBad.data() + 16, &end, u64),
            NumericParseStatus::Overflow);
  EXPECT_EQ(end, uintBad.data() + 37);

  EXPECT_EQ(ParseNumber(uintBad.data() + 37, &end, u16),
            NumericParseStatus::Invalid);

  //                    0         1         2         3
  //                    012345678901234567890123456789012345
  std::string intBad = "-32769 32768 2147483648 -2147483649 ";

  std::int16_t i16;
  EXPECT_EQ(ParseNumber(intBad.data(), &end, i16),
            NumericParseStatus::Underflow);
  EXPECT_EQ(end, intBad.data() + 6);
  EXPECT_EQ(ParseNumber(intBad.data() + 6, &end, i16),
            NumericParseStatus::Overflow);
  EXPECT_EQ(end, intBad.data() + 12);

  std::int32_t i32;
  EXPECT_EQ(ParseNumber(intBad.data() + 12, &end, i32),
            NumericParseStatus::Overflow);
  EXPECT_EQ(end, intBad.data() + 23);
  EXPECT_EQ(ParseNumber(intBad.data() + 23, &end, i32),
            NumericParseStatus::Underflow);
  EXPECT_EQ(end, intBad.data() + 35);

  //        0         1         2         3         4
  //        01234567890123456789012345678901234567890123
  intBad = "9223372036854775808 -9223372036854775809 foo";
  std::int64_t i64;
  EXPECT_EQ(ParseNumber(intBad.data() + 0, &end, i64),
            NumericParseStatus::Overflow);
  EXPECT_EQ(end, intBad.data() + 19);
  EXPECT_EQ(ParseNumber(intBad.data() + 19, &end, i64),
            NumericParseStatus::Underflow);
  EXPECT_EQ(end, intBad.data() + 40);
  EXPECT_EQ(ParseNumber(uintBad.data() + 40, &end, i64),
            NumericParseStatus::Invalid);

  //                    0         1
  //                    012345678901234567
  std::string dblBad = "1e+309 -1e309 foo ";

  EXPECT_EQ(ParseNumber(dblBad.data() + 0, &end, d),
            NumericParseStatus::Overflow);
  EXPECT_EQ(end, dblBad.data() + 6);
  EXPECT_EQ(ParseNumber(dblBad.data() + 6, &end, d),
            NumericParseStatus::Overflow);
  EXPECT_EQ(end, dblBad.data() + 13);
  EXPECT_EQ(ParseNumber(dblBad.data() + 13, &end, d),
            NumericParseStatus::Invalid);

  EXPECT_THROW(ParseNumber<std::uint8_t>("256"),
               d5::NumericParseError<std::uint8_t>);
  EXPECT_THROW(ParseNumber<std::uint8_t>("foo"),
               d5::NumericParseError<std::uint8_t>);

  EXPECT_THROW(ParseNumber("foo", u16), d5::NumericParseError<std::uint16_t>);
}
