/** Unit test for tag_invoke
 *
 *  \file TagInvokeTest.cpp
 */

#include "dirsig5/base/TagInvoke.h"

#include <gtest/gtest.h>

// Set up a small customization point scenario

namespace lib {

D5_DEFINE_TAG_INVOKE_CPO(cp);

}

namespace client {

struct InClassTest {
  int a;

  friend constexpr auto tag_invoke(lib::cp_t, const InClassTest &self, int b) {
    return self.a + b;
  }
};

struct OutOfClassTest {
  int a;
};

inline auto tag_invoke(lib::cp_t, const OutOfClassTest &self, int b) {
  return self.a * b;
}

struct VoidTest {
  int a;

  friend auto tag_invoke(lib::cp_t, const VoidTest &self, int &b) noexcept {
    b = self.a;
  }
};

} // namespace client

// Test d5::is_tag_invocable_v
static_assert(
    d5::is_tag_invocable_v<lib::cp_t, const client::InClassTest &, int>);
static_assert(
    d5::is_tag_invocable_v<lib::cp_t, const client::OutOfClassTest &, int>);
static_assert(
    !d5::is_tag_invocable_v<lib::cp_t, const client::VoidTest &, int>);
static_assert(
    d5::is_tag_invocable_v<lib::cp_t, const client::VoidTest &, int &>);

// Test d5::is_nothrow_tag_invocable_v
static_assert(
    d5::is_nothrow_tag_invocable_v<lib::cp_t, const client::VoidTest &, int &>);
static_assert(!d5::is_nothrow_tag_invocable_v<
              lib::cp_t, const client::InClassTest &, int>);
static_assert(!d5::is_nothrow_tag_invocable_v<
              lib::cp_t, const client::OutOfClassTest &, int>);

// Test d5::tag_invoke_result_t
static_assert(
    std::is_same_v<
        d5::tag_invoke_result_t<lib::cp_t, const client::InClassTest &, int>,
        int>);
static_assert(
    std::is_same_v<
        d5::tag_invoke_result_t<lib::cp_t, const client::OutOfClassTest &, int>,
        int>);
static_assert(
    std::is_same_v<
        d5::tag_invoke_result_t<lib::cp_t, const client::VoidTest &, int &>,
        void>);

TEST(TagInvokeTests, RuntimeUsage) {
  client::InClassTest a{5};
  client::OutOfClassTest b{6};
  client::VoidTest c{7};

  EXPECT_EQ(d5::tag_invoke(lib::cp, a, 7), 12);
  EXPECT_EQ(d5::tag_invoke(lib::cp, b, 7), 42);

  int d;
  d5::tag_invoke(lib::cp_t{}, c, d);
  EXPECT_EQ(c.a, d);
}

TEST(TagInvokeTests, ConstexprUsage) {
  constexpr int a = d5::tag_invoke(lib::cp, client::InClassTest{5}, 7);
  constexpr int b = d5::tag_invoke(lib::cp_t{}, client::InClassTest{6}, 9);
  EXPECT_EQ(a, 12);
  EXPECT_EQ(b, 15);
}
