#include <array>
#include <vector>

#include <gtest/gtest.h>

#include "dirsig5/base/Temperature.h"

using namespace d5;

TEST(TemperatureTest, ConversionTests) {
  std::vector<std::array<double, 3>> testCases{{{283.15, 10, 50}},
                                               {{268.15, -5, 23}},
                                               {{273.15, 0, 32}},
                                               {{278.15, 5, 41}}};
  for (const auto &kcf : testCases) {
    EXPECT_NEAR(CelsiusToKelvin(kcf[1]), kcf[0], 1e-10)
        << kcf[1] << "C to K failed";
    EXPECT_NEAR(KelvinToCelsius(kcf[0]), kcf[1], 1e-10)
        << kcf[0] << "K to C failed";
    EXPECT_NEAR(FahrenheitToCelsius(kcf[2]), kcf[1], 1e-10)
        << kcf[2] << "F to C failed";
    EXPECT_NEAR(CelsiusToFahrenheit(kcf[1]), kcf[2], 1e-10)
        << kcf[1] << "C to F failed";
    EXPECT_NEAR(FahrenheitToKelvin(kcf[2]), kcf[0], 1e-10)
        << kcf[2] << "F to K failed";
    EXPECT_NEAR(KelvinToFahrenheit(kcf[0]), kcf[2], 1e-10)
        << kcf[0] << "K to F failed";
    EXPECT_EQ(
        Temperature::Convert(kcf[0], Temperature::Kelvin, Temperature::Kelvin),
        kcf[0]);
    EXPECT_NEAR(
        Temperature::Convert(kcf[0], Temperature::Kelvin, Temperature::Celsius),
        kcf[1], 1e-10);
    EXPECT_NEAR(Temperature::Convert(kcf[0], Temperature::Kelvin,
                                     Temperature::Fahrenheit),
                kcf[2], 1e-10);
    EXPECT_EQ(
        Temperature::Convert(kcf[1], Temperature::Celsius, Temperature::Kelvin),
        kcf[0]);
    EXPECT_EQ(Temperature::Convert(kcf[1], Temperature::Celsius,
                                   Temperature::Celsius),
              kcf[1]);
    EXPECT_EQ(Temperature::Convert(kcf[1], Temperature::Celsius,
                                   Temperature::Fahrenheit),
              kcf[2]);
    EXPECT_EQ(Temperature::Convert(kcf[2], Temperature::Fahrenheit,
                                   Temperature::Kelvin),
              kcf[0]);
    EXPECT_EQ(Temperature::Convert(kcf[2], Temperature::Fahrenheit,
                                   Temperature::Celsius),
              kcf[1]);
    EXPECT_EQ(Temperature::Convert(kcf[2], Temperature::Fahrenheit,
                                   Temperature::Fahrenheit),
              kcf[2]);
  }
}

TEST(TemperatureTest, ConstructorTests) {
  std::vector<Temperature> testCases{Temperature::fromKelvin(273.15),
                                     Temperature::fromCelsius(0),
                                     Temperature::fromFahrenheit(32),
                                     Temperature(273.15, Temperature::Kelvin),
                                     Temperature(0, Temperature::Celsius),
                                     Temperature(32, Temperature::Fahrenheit)};
  for (const auto &t : testCases) {
    EXPECT_DOUBLE_EQ(t.get<Temperature::Kelvin>(), 273.15);
    EXPECT_DOUBLE_EQ(t.kelvin(), 273.15);
    EXPECT_DOUBLE_EQ(t.get<Temperature::Celsius>(), 0);
    EXPECT_DOUBLE_EQ(t.celsius(), 0);
    EXPECT_DOUBLE_EQ(t.get<Temperature::Fahrenheit>(), 32);
    EXPECT_DOUBLE_EQ(t.fahrenheit(), 32);
  }
}

TEST(TemperatureTest, CompareTests) {
  EXPECT_EQ(Temperature::fromKelvin(273.15), Temperature::fromCelsius(0));
  EXPECT_LE(Temperature::fromKelvin(273.15), Temperature::fromCelsius(0));
  EXPECT_GE(Temperature::fromKelvin(273.15), Temperature::fromCelsius(0));
  EXPECT_NE(Temperature::fromKelvin(273.15), Temperature::fromCelsius(1));
  EXPECT_LT(Temperature::fromKelvin(273.15), Temperature::fromCelsius(1));
  EXPECT_LE(Temperature::fromKelvin(273.15), Temperature::fromCelsius(1));
  EXPECT_GT(Temperature::fromFahrenheit(65), Temperature::fromCelsius(1));
  EXPECT_GE(Temperature::fromFahrenheit(65), Temperature::fromCelsius(1));
}
