/** Implementation file for BrdfModel
 *
 *  \file BrdfModel.cpp
 */

#include "dirsig5/brdf/BrdfModel.h"

#include <algorithm>

namespace d5::brdf {

BrdfModel::BrdfModel(std::vector<double> wavelengthMicrons)
    : wavelengthMicrons_(std::move(wavelengthMicrons)) {
  std::sort(wavelengthMicrons_.begin(), wavelengthMicrons_.end());
}

void BrdfModel::simplifySpectralBrdf(std::vector<double>& nadirDHR) {
  auto newWavelengths = doSimplifySpectralBrdf(nadirDHR);
  if (!newWavelengths.empty()) {
    wavelengthMicrons_ = std::move(newWavelengths);
  }

  // Re-initialize after simplifying the wavelengths
  init();
}

std::vector<BrdfSample> BrdfModel::sampleAngleCombinations(
    const std::vector<double>& illumZenithRad,
    const std::vector<double>& illumAzimuthRad,
    const std::vector<double>& viewZenithRad,
    const std::vector<double>& viewAzimuthRad) const {
  std::vector<BrdfSample> samples(
      illumZenithRad.size() * illumAzimuthRad.size() * viewZenithRad.size() *
      viewAzimuthRad.size());
  size_t index = 0;

  for (auto illumZenith : illumZenithRad) {
    for (auto illumAzimuth : illumAzimuthRad) {
      for (auto viewZenith : viewZenithRad) {
        for (auto viewAzimuth : viewAzimuthRad) {
          auto& s = samples[index++];
          s.viewZenithRad = viewZenith;
          s.viewAzimuthRad = viewAzimuth;
          s.illuminantZenithRad = illumZenith;
          s.illuminantAzimuthRad = illumAzimuth;
          s.brdf = sample(illumZenith, illumAzimuth, viewZenith, viewAzimuth);
        }
      }
    }
  }

  return samples;
}

void BrdfModel::getHashInputs(const HashCallback& callback) const {
  AddHashData(wavelengthMicrons_, callback);

  // Version number (used as hash salt, update on any implementation changes)
  static constexpr int kVersion = 1;
  AddHashData(kVersion, callback);

  doGetHashInputs(callback);
}
}
