/** Implementation file for Fresnel reflectance
 *
 *  \file FresnelReflectance.cpp
 */

#include "dirsig5/brdf/FresnelReflectance.h"

#include "dirsig5/base/Constants.h"

namespace d5::brdf {

FresnelReflectance::BrdfAngles::BrdfAngles(
    double _thetaI, double _phiI, double _thetaR, double _phiR)
    : thetaI(_thetaI), phiI(_phiI), thetaR(_thetaR), phiR(_phiR) {
  // Calculate the difference in view and illuminant azimuth
  deltaPhi = phiR - phiI;

  // Pre-compute some trig functions
  cosThetaI = cos(thetaI);
  sinThetaI = sin(thetaI);
  cosThetaR = cos(thetaR);
  sinThetaR = sin(thetaR);
  tanThetaI = sinThetaI / cosThetaI;
  tanThetaR = sinThetaR / cosThetaR;

  // Compute beta and cos( beta ) -- Priest and Germer, Equation (12)
  beta = computeBeta(cosThetaI, cosThetaR, sinThetaI, sinThetaR, deltaPhi);
  cosBeta = cos(beta);
  sinBeta = sin(beta);

  // Compute theta -- Priest and Germer (12)
  cosTheta = computeCosTheta(cosThetaI, cosThetaR, cosBeta);

  theta = acos(cosTheta);
  tanTheta = tan(theta);

  // Compute the Etas and some other angles
  computeEtas(*this);
}

void FresnelReflectance::computeEtas(BrdfAngles& angles) {
  // Compute CosEtaI and CosEtaR

  // First check for the 0 / 0 case
  if ((fabs(angles.beta) < 1.0e-10) || (fabs(angles.thetaI) < 1.0e-10) ||
      (fabs(angles.thetaR) < 1.0e-10)) {
    // in-plane scattering, etaI and etaR = 0.0
    angles.cosEtaI = 1.0;
    angles.cosEtaR = 1.0;
    angles.signEta = 0.0;
  } else {
    // compute cosEtaI using Priest and Germer (19)
    angles.cosEtaI = angles.cosTheta;
    angles.cosEtaI -= (angles.cosThetaI * angles.cosBeta);
    angles.cosEtaI /= (angles.sinThetaI * angles.sinBeta);

    // Ensure that: -1 <= cosEtaI <= 1
    if (angles.cosEtaI > 1.0) angles.cosEtaI = 1.0;
    if (angles.cosEtaI < -1.0) angles.cosEtaI = -1.0;

    // compute cosEtaR using Priest and Germer (19)
    angles.cosEtaR = angles.cosTheta;
    angles.cosEtaR -= (angles.cosThetaR * angles.cosBeta);
    angles.cosEtaR /= (angles.sinThetaR * angles.sinBeta);

    // Ensure that: -1 <= cosEtaR <= 1
    if (angles.cosEtaR > 1.0) angles.cosEtaR = 1.0;
    if (angles.cosEtaR < -1.0) angles.cosEtaR = -1.0;

    // Make sure deltaPhi is between -360 and 360
    while (angles.deltaPhi <= -TwoPI) angles.deltaPhi += TwoPI;
    while (angles.deltaPhi >= TwoPI) angles.deltaPhi -= TwoPI;

    // Determine the sign of etaI and etaR
    if ((angles.deltaPhi > 0) && (angles.deltaPhi < PI)) {
      angles.signEta = -1.0;
    } else if ((angles.deltaPhi == 0) || (angles.deltaPhi == PI)) {
      angles.signEta = 0.0;
    } else {
      angles.signEta = 1.0;
    }
  }

  //
  // The case of a straight-up (i.e. a specular return centered about
  // the facet normal) is degenerate and will lead to an incorrect zero
  // result.  The following check prevents this problem.
  //
  if ((angles.cosEtaI == 0) && (angles.cosEtaR == 0) && (angles.signEta == 0)) {
    angles.signEta = 1.0;
  }

  // Precalculate come commonly used sines and cosines
  angles.sinEtaI = angles.signEta * sqrt(1.0 - angles.cosEtaI * angles.cosEtaI);
  angles.sinEtaR = angles.signEta * sqrt(1.0 - angles.cosEtaR * angles.cosEtaR);

  // in loop angle multiplies that we'll need later on
  angles.sinEtaRsinEtaI = angles.sinEtaR * angles.sinEtaI;
  angles.cosEtaRcosEtaI = angles.cosEtaR * angles.cosEtaI;
  angles.sinEtaRcosEtaI = angles.sinEtaR * angles.cosEtaI;
  angles.cosEtaRsinEtaI = angles.cosEtaR * angles.sinEtaI;
}

std::vector<MuellerMatrix> FresnelReflectance::computeReflectance(
    const std::vector<std::complex<double>>& nk, const BrdfAngles& angles) {
  // Create a polaraized vector variable for the answer
  std::vector<MuellerMatrix> ans(nk.size());

  // Create variables for the indices of refraction
  // For now, assume we are strictly working in air (actually a vacuum)
  const std::complex<double> nI(1.0, 0.0);

  // Create variables for the Fresnel amplitudes
  std::complex<double> Ass;
  std::complex<double> App;

  // Create variables for the elements of the Jones Matrix used in
  // eq (14) of Priest and Germer
  std::complex<double> Tss, Tps, Tsp, Tpp;

  // The following are for the complex conjugates of the above
  std::complex<double> TssCC, TpsCC, TspCC, TppCC;

  // The following are for the squared magnitudes of the above
  double magSqTss, magSqTps, magSqTsp, magSqTpp;
  std::complex<double> sqrt1mbb;
  std::complex<float> bbf;

  double rTssTpsCC, iTppTscCC, rTspTppCC, iTpsTssCC, rTssTspCC, rTpsTppCC,
      rTssTppCC, rTpsTspCC, iTpsTspCC, iTssTppCC, iTssTspCC, iTpsTppCC;

  // Calculate the Mueller matrix elements for each spectral location
  for (size_t ii = 0; ii < nk.size(); ii++) {
    // Calculate the Fresnel amplitudes
    const std::complex<double>& nT = nk[ii];

    // reuse part of each reflectance calculation
    bbf = (nI / nT) * angles.sinBeta;
    bbf = bbf * bbf;
    sqrt1mbb = sqrt(1.f - bbf);

    Ass = fresnelReflectionPerpendicular(nI, nT, angles.cosBeta, sqrt1mbb);
    App = fresnelReflectionParallel(nI, nT, angles.cosBeta, sqrt1mbb);

    // Calculate Tss, Tps, Tsp, Tpp
    Tss = App * angles.sinEtaRsinEtaI + Ass * angles.cosEtaRcosEtaI;
    Tps = App * angles.sinEtaRcosEtaI - Ass * angles.cosEtaRsinEtaI;
    Tsp = App * angles.cosEtaRsinEtaI - Ass * angles.sinEtaRcosEtaI;
    Tpp = App * angles.cosEtaRcosEtaI + Ass * angles.sinEtaRsinEtaI;

    // Calculate TssCC, TpsCC, TspCC, TppCC
    TssCC = conj(Tss);
    TpsCC = conj(Tps);
    TspCC = conj(Tsp);
    TppCC = conj(Tpp);

    // Calculate magSqTss, magSqTps, magSqTsp, magSqTpp
    magSqTss = real(Tss * TssCC);
    magSqTps = real(Tps * TpsCC);
    magSqTsp = real(Tsp * TspCC);
    magSqTpp = real(Tpp * TppCC);

    // Now we are ready to calculate the Mueller matrix elements
    // The signs of the 4th row and column follow the convention of:
    //   H. C. Van de Hulst, Light Scattering by Small Particles,
    //   John Wiley & Sons, New York (1957).
    // another sign convention is described in:
    //   Flynn and Alexander, Polarized surface scattering expressed
    //   in terms of a bidirectional reflectance distribution function
    //   matrix
    rTssTpsCC = real(Tss * TpsCC);
    iTppTscCC = imag(Tpp * TspCC);
    rTspTppCC = real(Tsp * TppCC);
    iTpsTssCC = imag(Tps * TssCC);
    rTssTspCC = real(Tss * TspCC);
    rTpsTppCC = real(Tps * TppCC);
    rTssTppCC = real(Tss * TppCC);
    rTpsTspCC = real(Tps * TspCC);
    iTpsTspCC = imag(Tps * TspCC);
    iTssTppCC = imag(Tss * TppCC);
    iTssTspCC = imag(Tss * TspCC);
    iTpsTppCC = imag(Tps * TppCC);

    auto& mm = ans[ii];
    mm(0, 0) = 0.5 * (magSqTss + magSqTsp + magSqTps + magSqTpp);
    mm(0, 1) = 0.5 * (magSqTss + magSqTsp - magSqTps - magSqTpp);
    mm(0, 2) = rTssTpsCC + rTspTppCC;
    mm(0, 3) = iTpsTssCC + iTppTscCC;
    mm(1, 0) = 0.5 * (magSqTss - magSqTsp + magSqTps - magSqTpp);
    mm(1, 1) = 0.5 * (magSqTss - magSqTsp - magSqTps + magSqTpp);
    mm(1, 2) = rTssTpsCC - rTspTppCC;
    mm(1, 3) = iTpsTssCC - iTppTscCC;
    mm(2, 0) = rTssTspCC + rTpsTppCC;
    mm(2, 1) = rTssTspCC - rTpsTppCC;
    mm(2, 2) = rTssTppCC + rTpsTspCC;
    mm(2, 3) = iTpsTspCC - iTssTppCC;
    mm(3, 0) = iTssTspCC + iTpsTppCC;
    mm(3, 1) = iTssTspCC - iTpsTppCC;
    mm(3, 2) = iTssTppCC + iTpsTspCC;
    mm(3, 3) = rTssTppCC - rTpsTspCC;
  }

  return ans;
}

std::vector<MuellerMatrix> FresnelReflectance::computeTransmission(
    const std::vector<std::complex<double>>& nk, const BrdfAngles& angles) {
  // Create a variable for the answer
  std::vector<MuellerMatrix> ans(nk.size());

  // Create variables for the elements of the Jones Matrix used in
  // eq (14) of Priest and Germer
  std::complex<double> Tss, Tpp;

  // The following are for the complex conjugates of the above
  std::complex<double> TssCC, TppCC;

  // The following are for the squared magnitudes of the above
  double magSqTss, magSqTpp;

  // Create variables for the Fresnel amplitudes
  std::complex<double> Ass;
  std::complex<double> App;

  // Create variables for the indices of refraction
  // For now, assume we are strictly working in air (actually a vacuum)
  const std::complex<double> nI(1.0, 0.0);
  std::complex<double> sqrt1mbb;
  std::complex<float> bbf;

  double tempVal;

  // Calculate the Mueller matrix elements for each spectral location
  for (size_t ii = 0; ii < nk.size(); ii++) {
    // Calculate the Fresnel amplitudes
    const std::complex<double>& nT = nk[ii];

    // reuse part of each reflectance calculation
    bbf = (nI / nT) * angles.sinThetaR;
    bbf = bbf * bbf;
    sqrt1mbb = sqrt(1.f - bbf);

    Ass = fresnelReflectionPerpendicular(nI, nT, angles.cosThetaR, sqrt1mbb);
    App = fresnelReflectionParallel(nI, nT, angles.cosThetaR, sqrt1mbb);

    // Calculate Tss, Tps, Tsp, Tpp
    Tss = Ass;
    Tpp = App;

    // Calculate TssCC, TpsCC, TspCC, TppCC
    TssCC = conj(Tss);
    TppCC = conj(Tpp);

    // Calculate magTss, magTpp
    magSqTss = 1.0 - real(Tss * TssCC);
    magSqTpp = 1.0 - real(Tpp * TppCC);

    // Now we are ready to calculate the Mueller matrix elements
    // The signs of the 4th row and column follow the convention of:
    //   H. C. Van de Hulst, Light Scattering by Small Particles,
    //   John Wiley & Sons, New York (1957).
    // another sign convention is described in:
    //   Flynn and Alexander, Polarized surface scattering expressed
    //   in terms of a bidirectional reflectance distribution function
    //   matrix
    auto& mm = ans[ii];
    tempVal = 0.5 * (magSqTss + magSqTpp);
    mm(0, 0) = tempVal;
    mm(1, 1) = tempVal;

    tempVal = 0.5 * (magSqTss - magSqTpp);
    mm(0, 1) = tempVal;
    mm(1, 0) = tempVal;

    mm(0, 2) = 0;
    mm(0, 3) = 0;
    mm(1, 2) = 0;
    mm(1, 3) = 0;
    mm(2, 0) = 0;
    mm(2, 1) = 0;

    tempVal = real(Tss * TppCC);
    mm(2, 2) = tempVal;
    mm(3, 3) = tempVal;

    tempVal = -imag(Tss * TppCC);
    mm(2, 3) = tempVal;

    mm(3, 0) = 0;
    mm(3, 1) = 0;

    tempVal = 1.0 * imag(Tss * TppCC);
    mm(3, 2) = tempVal;
  }

  return ans;
}

void FresnelReflectance::compute(
    const std::vector<std::complex<double>>& nk,
    const BrdfAngles& angles,
    std::vector<MuellerMatrix>& r,
    std::vector<MuellerMatrix>& t) {
  r.resize(nk.size());
  t.resize(nk.size());

  // Create variables for the indices of refraction
  // For now, assume we are strictly working in air (actually a vacuum)
  const std::complex<double> nI(1.0, 0.0);

  // Create variables for the Fresnel amplitudes
  std::complex<double> Ass;
  std::complex<double> App;

  // Create variables for the elements of the Jones Matrix used in
  // eq (14) of Priest and Germer
  std::complex<double> Tss, Tps, Tsp, Tpp;

  // The following are for the complex conjugates of the above
  std::complex<double> TssCC, TpsCC, TspCC, TppCC;

  // The following are for the squared magnitudes of the above
  double magSqTss, magSqTps, magSqTsp, magSqTpp;
  std::complex<double> sqrt1mbb;
  std::complex<float> bbf;

  double rTssTpsCC, iTppTscCC, rTspTppCC, iTpsTssCC, rTssTspCC, rTpsTppCC,
      rTssTppCC, rTpsTspCC, iTpsTspCC, iTssTppCC, iTssTspCC, iTpsTppCC;
  double tempVal;

  // Calculate the Mueller matrix elements for each spectral location
  for (size_t ii = 0; ii < nk.size(); ii++) {
    const std::complex<double>& nT = nk[ii];

    // reuse part of each reflectance calculation
    bbf = (nI / nT) * angles.sinBeta;
    bbf = bbf * bbf;
    sqrt1mbb = sqrt(1.f - bbf);

    Ass = fresnelReflectionPerpendicular(nI, nT, angles.cosBeta, sqrt1mbb);
    App = fresnelReflectionParallel(nI, nT, angles.cosBeta, sqrt1mbb);

    // Calculate Tss, Tps, Tsp, Tpp
    Tss = App * angles.sinEtaRsinEtaI + Ass * angles.cosEtaRcosEtaI;
    Tps = App * angles.sinEtaRcosEtaI - Ass * angles.cosEtaRsinEtaI;
    Tsp = App * angles.cosEtaRsinEtaI - Ass * angles.sinEtaRcosEtaI;
    Tpp = App * angles.cosEtaRcosEtaI + Ass * angles.sinEtaRsinEtaI;

    // Calculate TssCC, TpsCC, TspCC, TppCC
    TssCC = conj(Tss);
    TpsCC = conj(Tps);
    TspCC = conj(Tsp);
    TppCC = conj(Tpp);

    // Calculate magSqTss, magSqTps, magSqTsp, magSqTpp
    magSqTss = real(Tss * TssCC);
    magSqTps = real(Tps * TpsCC);
    magSqTsp = real(Tsp * TspCC);
    magSqTpp = real(Tpp * TppCC);

    // Now we are ready to calculate the Mueller matrix elements
    // The signs of the 4th row and column follow the convention of:
    //   H. C. Van de Hulst, Light Scattering by Small Particles,
    //   John Wiley & Sons, New York (1957).
    // another sign convention is described in:
    //   Flynn and Alexander, Polarized surface scattering expressed
    //   in terms of a bidirectional reflectance distribution function
    //   matrix
    rTssTpsCC = real(Tss * TpsCC);
    iTppTscCC = imag(Tpp * TspCC);
    rTspTppCC = real(Tsp * TppCC);
    iTpsTssCC = imag(Tps * TssCC);
    rTssTspCC = real(Tss * TspCC);
    rTpsTppCC = real(Tps * TppCC);
    rTssTppCC = real(Tss * TppCC);
    rTpsTspCC = real(Tps * TspCC);
    iTpsTspCC = imag(Tps * TspCC);
    iTssTppCC = imag(Tss * TppCC);
    iTssTspCC = imag(Tss * TspCC);
    iTpsTppCC = imag(Tps * TppCC);

    auto& rmm = r[ii];
    rmm(0, 0) = 0.5 * (magSqTss + magSqTsp + magSqTps + magSqTpp);
    rmm(0, 1) = 0.5 * (magSqTss + magSqTsp - magSqTps - magSqTpp);
    rmm(0, 2) = rTssTpsCC + rTspTppCC;
    rmm(0, 3) = iTpsTssCC + iTppTscCC;
    rmm(1, 0) = 0.5 * (magSqTss - magSqTsp + magSqTps - magSqTpp);
    rmm(1, 1) = 0.5 * (magSqTss - magSqTsp - magSqTps + magSqTpp);
    rmm(1, 2) = rTssTpsCC - rTspTppCC;
    rmm(1, 3) = iTpsTssCC - iTppTscCC;
    rmm(2, 0) = rTssTspCC + rTpsTppCC;
    rmm(2, 1) = rTssTspCC - rTpsTppCC;
    rmm(2, 2) = rTssTppCC + rTpsTspCC;
    rmm(2, 3) = iTpsTspCC - iTssTppCC;
    rmm(3, 0) = iTssTspCC + iTpsTppCC;
    rmm(3, 1) = iTssTspCC - iTpsTppCC;
    rmm(3, 2) = iTssTppCC + iTpsTspCC;
    rmm(3, 3) = rTssTppCC - rTpsTspCC;

    Tss = fresnelReflectionPerpendicular(nI, nT, angles.cosThetaR, sqrt1mbb);
    Tpp = fresnelReflectionParallel(nI, nT, angles.cosThetaR, sqrt1mbb);

    // Calculate TssCC, TpsCC, TspCC, TppCC
    TssCC = conj(Tss);
    TppCC = conj(Tpp);

    // Calculate magTss, magTpp
    magSqTss = 1.0 - real(Tss * TssCC);
    magSqTpp = 1.0 - real(Tpp * TppCC);

    // Now we are ready to calculate the Mueller matrix elements
    // The signs of the 4th row and column follow the convention of:
    //   H. C. Van de Hulst, Light Scattering by Small Particles,
    //   John Wiley & Sons, New York (1957).
    // another sign convention is described in:
    //   Flynn and Alexander, Polarized surface scattering expressed
    //   in terms of a bidirectional reflectance distribution function
    //   matrix
    auto& tmm = t[ii];
    tempVal = 0.5 * (magSqTss + magSqTpp);
    tmm(0, 0) = tempVal;
    tmm(1, 1) = tempVal;

    tempVal = 0.5 * (magSqTss - magSqTpp);
    tmm(0, 1) = tempVal;
    tmm(1, 0) = tempVal;

    tmm(0, 2) = 0;
    tmm(0, 3) = 0;
    tmm(1, 2) = 0;
    tmm(1, 3) = 0;
    tmm(2, 0) = 0;
    tmm(2, 1) = 0;

    tempVal = real(Tss * TppCC);
    tmm(2, 2) = tempVal;
    tmm(3, 3) = tempVal;

    tempVal = -imag(Tss * TppCC);
    tmm(2, 3) = tempVal;

    tmm(3, 0) = 0;
    tmm(3, 1) = 0;

    tempVal = 1.0 * imag(Tss * TppCC);
    tmm(3, 2) = tempVal;
  }
}

// Compute parallel component of Fresnel Reflectance
std::complex<double> FresnelReflectance::fresnelReflectionParallel(
    const std::complex<double>& nI,
    const std::complex<double>& nT,
    double cosThetaI,
    const std::complex<double>& sqrt1mbb) {
  std::complex<double> A = nT * cosThetaI;
  std::complex<double> B = nI * sqrt1mbb;

  return (A - B) / (A + B);
}

// Compute perpendicular componenet of Fresnel Reflectance
std::complex<double> FresnelReflectance::fresnelReflectionPerpendicular(
    const std::complex<double>& nI,
    const std::complex<double>& nT,
    double cosThetaI,
    const std::complex<double>& sqrt1mbb) {
  std::complex<double> A = nI * cosThetaI;
  std::complex<double> B = nT * sqrt1mbb;

  return (A - B) / (A + B);
}
}
