/** Implementation file for microfact models
 *
 *  \file MicroFacetModels.cpp
 */

#include "dirsig5/brdf/MicroFacetModels.h"

#include <cmath>

namespace d5::brdf {

// #############################################
// #         OPGaussian Implementation         #
// #############################################

double OPGaussian::operator()(
    const FresnelReflectance::BrdfAngles& angles) const {
  double cosTn = angles.cosTheta;
  double tanTn = angles.tanTheta;

  return B_ / (2.0 * M_PI * var_ * cosTn * cosTn * cosTn) *
         exp(-(tanTn * tanTn) / (2.0 * var_));
}

void OPGaussian::getHashInputs(const HashCallback& callback) const {
  // Version number (used as hash salt, update on any implementation changes)
  static constexpr int kVersion = 1;
  AddHashData(kVersion, callback);
  AddHashData(B_, callback);
  AddHashData(var_, callback);
}

// #############################################
// #          OPCauchy Implementation          #
// #############################################

double OPCauchy::operator()(
    const FresnelReflectance::BrdfAngles& angles) const {
  double cosTn = angles.cosTheta;
  double tanTn = angles.tanTheta;

  return B_ / (cosTn * (var_ + tanTn * tanTn));
}

void OPCauchy::getHashInputs(const HashCallback& callback) const {
  // Version number (used as hash salt, update on any implementation changes)
  static constexpr int kVersion = 1;
  AddHashData(kVersion, callback);
  AddHashData(B_, callback);
  AddHashData(var_, callback);
}

// #############################################
// #       SFMaxwellBeard Implementation       #
// #############################################

double SFMaxwellBeard::operator()(
    const FresnelReflectance::BrdfAngles& angles) const {
  double k = (angles.theta * invOmega_);
  double eVal = k * exp(angles.beta * invTau_);
  return (1.0 + eVal) / (1 + angles.theta * invOmega_);
}

void SFMaxwellBeard::getHashInputs(const HashCallback& callback) const {
  // Version number (used as hash salt, update on any implementation changes)
  static constexpr int kVersion = 1;
  AddHashData(kVersion, callback);
  AddHashData(invOmega_, callback);
  AddHashData(invTau_, callback);
}

// #############################################
// #          SFSancer Implementation          #
// #############################################

double SFSancer::operator()(
    const FresnelReflectance::BrdfAngles& angles) const {
  static const double sqrt2dpi = sqrt(2.0 / M_PI);
  static const double sqrt2 = sqrt(2.0);

  double cotI = 1.0 / angles.tanThetaI;
  double cotR = 1.0 / angles.tanThetaR;

  double shadow_I = 0.5 * (sqrt2dpi * sigma_ / cotI *
                           exp(-cotI * cotI / 2.0 / sigma_ / sigma_));

  shadow_I -= erfc(cotI / sqrt2 / sigma_);
  shadow_I *= 0.5;

  double shadow_R = 0.5 * (sqrt(2.0 / M_PI) * sigma_ / cotR *
                           exp(-cotR * cotR / 2.0 / sigma_ / sigma_));

  shadow_R -= erfc(cotR / sqrt2 / sigma_);
  shadow_R *= 0.5;

  return 1.0 / (1.0 + shadow_I + shadow_R);
}

void SFSancer::getHashInputs(const HashCallback& callback) const {
  // Version number (used as hash salt, update on any implementation changes)
  static constexpr int kVersion = 1;
  AddHashData(kVersion, callback);
  AddHashData(sigma_, callback);
}

// #############################################
// #       VTMaxwellBeard Implementation       #
// #############################################

double VTMaxwellBeard::operator()(
    const FresnelReflectance::BrdfAngles& angles) const {
  return rho_D_ + ((2.0 * rho_V_) / (angles.cosThetaI + angles.cosThetaR));
}

void VTMaxwellBeard::getHashInputs(const HashCallback& callback) const {
  // Version number (used as hash salt, update on any implementation changes)
  static constexpr int kVersion = 1;
  AddHashData(kVersion, callback);
  AddHashData(rho_D_, callback);
  AddHashData(rho_V_, callback);
}
}
