/** Implementation file for Priest-Germer
 *
 *  \file PriestGermer.cpp
 */

#include "dirsig5/brdf/PriestGermer.h"

#include <cmath>

#include "PriestGermerDhr.h"
#include "dirsig5/brdf/MicroFacetModels.h"
#include "dirsig5/base/SpectralVector.h"

namespace d5::brdf {

std::vector<double> PriestGermer::sample(
    double illumZenithRad,
    double illumAzimuthRad,
    double viewZenithRad,
    double viewAzimuthRad) const {
  std::vector<double> result(wavelengthMicrons().size());

  // Localize azimuth.
  viewAzimuthRad -= illumAzimuthRad;
  if (viewAzimuthRad < 0) viewAzimuthRad += 2 * M_PI;
  if (viewAzimuthRad >= 2 * M_PI) viewAzimuthRad -= 2 * M_PI;
  illumAzimuthRad = 0;

  // Is either theta is out of range?
  constexpr double maxTheta = 89.0 * M_PI / 180.0;
  if (!(illumZenithRad >= 0.0 && illumZenithRad <= maxTheta &&  //
        viewZenithRad >= 0.0 && viewZenithRad <= maxTheta))
    return result;  // Just return 0.

  // Use FresnelReflectance::BrdfAngles to compute trigonometric values.
  FresnelReflectance::BrdfAngles angles(
      illumZenithRad, illumAzimuthRad, viewZenithRad, viewAzimuthRad);

  // Initialize result.
  std::vector<MuellerMatrix> fresnelR =
      FresnelReflectance::computeReflectance(ior_, angles);
  double tanTheta2 = angles.tanTheta * angles.tanTheta;
  double cosTheta2 = angles.cosTheta * angles.cosTheta;
  double invCosThetaTerm =
      0.25 * M_1_PI /
      (cosTheta2 * cosTheta2 * angles.cosThetaR * angles.cosThetaI);
  for (size_t pos = 0; pos < wavelengthMicrons().size(); pos++) {
    result[pos] = std::exp(-tanTheta2 * halfInvSigma2_[pos]) *
                  halfInvSigma2_[pos] * invCosThetaTerm * fresnelR[pos](0, 0) *
                  SFSancer(sigma_[pos])(angles);
    // Optionally do diffuse renormalization.
    double dhr = Dhr(sigma_[pos], angles.thetaR);
    if (params_.diffuseRenormalization) dhr *= fresnelR[pos](0, 0);
    result[pos] += (1 - dhr) * M_1_PI;
  }

  return result;
}

void PriestGermer::init() {
  // Unzip parameters.
  std::vector<double> fitLambda;
  std::vector<double> fitSigma;
  std::vector<double> fitN;
  std::vector<double> fitK;
  fitLambda.reserve(params_.spectral.size());
  fitSigma.reserve(params_.spectral.size());
  fitN.reserve(params_.spectral.size());
  fitK.reserve(params_.spectral.size());
  for (const auto& sp : params_.spectral) {
    fitLambda.push_back(sp.wavelengthMicrons);
    fitSigma.push_back(sp.sigma);
    fitN.push_back(sp.ior.real());
    fitK.push_back(sp.ior.imag());
  }

  // Interpolate sigma.
  sigma_.clear();
  interpolate(fitLambda, fitSigma, wavelengthMicrons(), sigma_);
  halfInvSigma2_.clear();
  halfInvSigma2_.reserve(sigma_.size());
  for (double sigma : sigma_) {
    halfInvSigma2_.emplace_back(0.5 / (sigma * sigma));
  }

  // Interpolate IOR.
  ior_.resize(wavelengthMicrons().size());
  std::vector<double> interpN;
  std::vector<double> interpK;
  interpolate(fitLambda, fitN, wavelengthMicrons(), interpN);
  interpolate(fitLambda, fitK, wavelengthMicrons(), interpK);
  auto itrInterpN = interpN.begin();
  auto itrInterpK = interpK.begin();
  for (auto& ior : ior_) ior = {*itrInterpN++, *itrInterpK++};
}

std::vector<double> PriestGermer::doSimplifySpectralBrdf(
    std::vector<double>& nadirDHR) {
  FresnelReflectance::BrdfAngles angles(0, 0, 0, 0);
  std::vector<double> dhrWave;
  std::vector<double> dhr;
  for (const auto& sp : params_.spectral) {
    dhrWave.push_back(sp.wavelengthMicrons);
    if (params_.diffuseRenormalization) {
      dhr.push_back(1.0);
    } else {
      std::vector<MuellerMatrix> fresnelR =
          FresnelReflectance::computeReflectance({sp.ior}, angles);
      dhr.push_back(fresnelR[0](0, 0));
    }
  }
  interpolate(dhrWave, dhr, wavelengthMicrons(), nadirDHR);
  return dhrWave;
}

void PriestGermer::doGetHashInputs(const HashCallback& callback) const {
  auto Add = [&](const auto& data) { AddHashData(data, callback); };

  Add(params_.diffuseRenormalization);
  Add(params_.spectral);

  // Version number (used as hash salt, update on any implementation changes)
  static constexpr int kVersion = 1;
  Add(kVersion);
}
}  // namespace d5::brdf
