/** Implementation file for the Ross-Li BRDF
 *
 *  \file RossLi.cpp
 */

#include "dirsig5/brdf/RossLi.h"

#include <cmath>

#include "dirsig5/base/SpectralVector.h"

namespace d5::brdf {

/// Helper struct for holding angular information and precompute sin's and cos's
struct Angles {
  Angles(double _cos_i, double _cos_v, double _raa)
      : cos_i(_cos_i), cos_v(_cos_v), raa(_raa) {
    this->computeDerivedTerms();
  }

  // input angles (source zen, view zen, relative az)
  double cos_i, cos_v, raa;

  // derived trig angles
  double sza, vza, sin_i, sin_v, cos_phi;

  // derived phase angles
  double Xi, cos_Xi, sin_Xi;

private:
  /// Calculates the trig/phase functions of s/v angles.
  void computeDerivedTerms() {
    sza = acos(cos_i);
    sin_i = sin(sza);
    vza = acos(cos_v);
    sin_v = sin(vza);
    cos_phi = cos(raa);

    // equation 3
    cos_Xi = cos_i * cos_v + sin_i * sin_v * cos_phi;

    // verify the angles are within bounds
    if (cos_Xi < -1) cos_Xi = -1;
    if (cos_Xi > +1) cos_Xi = 1;
    Xi = acos(cos_Xi);
    sin_Xi = sin(Xi);
  }
};

/// Returns the common component (numerator) of Ross thick/thin kernel
inline double RossCommon(const Angles& angles) {
  return (((0.5 * M_PI - angles.Xi) * angles.cos_Xi) + angles.sin_Xi);
}

/** Returns the Ross Thick kernel function for given solar zenith, view zenith
 *  and relative azimuth angles
 */
inline double RossThick(const Angles& angles) {
  // equation 7
  return (RossCommon(angles) / (angles.cos_i + angles.cos_v)) - 0.25 * M_PI;
}

/** Returns the Ross Thin kernel function for given solar zenith, view zenith
 *  and relative azimuth angles
 */
inline double RossThin(const Angles& angles) {
  // equation 13
  return (RossCommon(angles) / (angles.cos_i * angles.cos_v)) - 0.5 * M_PI;
}

/// Computes the squared distance between ellipse centers
inline double getDistance(double tan_ip, double tan_vp, const Angles& angles) {
  // equation 35
  double d_sq = (tan_ip * tan_ip) + (tan_vp * tan_vp) -
                2 * (tan_ip * tan_vp * angles.cos_phi);
  if (d_sq < 0.) d_sq = 0.;
  return d_sq;
}

/** Estimates the overlap betweeen shadow of the illumination and shadow of
 *  viewing. All given angles are corrected for elliptical crown shape.
 */
inline void computeGroundOverlap(
    const Angles& angles,
    double tan_ip,
    double tan_vp,
    double HB,
    double& overlap,
    double& sec_ip_vp) {
  // get the distance between ellipse centers
  double D_sq = getDistance(tan_ip, tan_vp, angles);
  double sin_phi = sin(angles.raa);
  if (angles.cos_i == 0 || angles.cos_v == 0) {
    overlap = 0;
    sec_ip_vp = 0;
    return;
  }
  sec_ip_vp = (1. / angles.cos_i) + (1. / angles.cos_v);

  // equation 34
  double tantansin = tan_ip * tan_vp * sin_phi;
  double cos_t = (HB * sqrt(D_sq + (tantansin * tantansin))) / sec_ip_vp;
  // check bounds
  if (cos_t < -1) cos_t = -1;
  if (cos_t > +1) cos_t = +1;
  double t = acos(cos_t);
  double sin_t = sin(t);

  // equation 33
  overlap = (1. / M_PI) * (t - sin_t * cos_t) * sec_ip_vp;
  if (overlap < 0) overlap = 0;
}

/// Returns the Li Sparse kernel function for given angles and BR, HB ratios
inline double LiSparse(
    Angles angles, double BR, double HB, bool useReciprocalLi) {
  // Li angular terms
  double tan_ip = BR * angles.sin_i / angles.cos_i;
  double tan_vp = BR * angles.sin_v / angles.cos_v;
  if (tan_ip < 0) tan_ip = 0;
  if (tan_vp < 0) tan_vp = 0;

  if (BR != 1.0) {
    double theta_ip = atan(tan_ip);
    double theta_vp = atan(tan_vp);
    angles = Angles(cos(theta_ip), cos(theta_vp), angles.raa);
  }

  double overlap, sec_ip_vp;
  computeGroundOverlap(angles, tan_ip, tan_vp, HB, overlap, sec_ip_vp);

  if (useReciprocalLi) {
    // equation 32 with added cosine term
    return overlap - sec_ip_vp +
           (0.5 * (1. + angles.cos_Xi) / angles.cos_v / angles.cos_i);
  } else {
    // equation 32
    return overlap - sec_ip_vp + (0.5 * (1. + angles.cos_Xi) / angles.cos_v);
  }
}

/// Returns the Li Dense kernel function for given angles and BR, HB ratios
inline double LiDense(const Angles& angles, double BR, double HB) {
  // Li angular terms
  double tan_ip = BR * angles.sin_i / angles.cos_i;
  double tan_vp = BR * angles.sin_v / angles.cos_v;
  if (tan_ip < 0) tan_ip = 0;
  if (tan_vp < 0) tan_vp = 0;
  double theta_ip = atan(tan_ip);
  double theta_vp = atan(tan_vp);

  Angles angles_p(cos(theta_ip), cos(theta_vp), angles.raa);

  double overlap, sec_ip_vp;
  computeGroundOverlap(angles_p, tan_ip, tan_vp, HB, overlap, sec_ip_vp);

  // equation 47
  double ldense =
      ((1. + angles_p.cos_Xi) / angles_p.cos_v) / (sec_ip_vp - overlap) - 2.0;

  return ldense;
}

/** Helper function for getting the BRDF coefficients
 *
 *  \param[in] angles  The angles at which to evaluate the kernels
 *  \param[in] params  The Ross-Li parameters
 *  \param[out] kVol   The volume kernel coefficient
 *  \param[out] kGeo   The geometric kernel coefficient
 */
void getCoeffs(
    const Angles& angles,
    const RossLi::Parameters& params,
    double& kVol,
    double& kGeo) {
  kVol = params.rossKernel == RossLi::RossKernel::Thick ? RossThick(angles)
                                                        : RossThin(angles);
  kGeo = params.liKernel == RossLi::LiKernel::Dense
             ? LiDense(angles, params.br, params.hb)
             : LiSparse(angles, params.br, params.hb, params.reciprocalLi);
}

/** Helper function for computing the BRDF */
double computeBrdf(
    double cIso, double cVol, double kVol, double cGeo, double kGeo) {
  return (cIso + cVol * kVol + cGeo * kGeo) / M_PI;
}

std::vector<double> RossLi::sample(
    double illumZenithRad,
    double illumAzimuthRad,
    double viewZenithRad,
    double viewAzimuthRad) const {
  // Calculate the angles and the kernel values
  Angles angles(
      cos(illumZenithRad), cos(viewZenithRad),
      viewAzimuthRad - illumAzimuthRad);
  double kVol, kGeo;
  getCoeffs(angles, params_, kVol, kGeo);

  // Compute the BRDF for each wavelength
  std::vector<double> brdf(wavelengthMicrons().size());
  for (size_t i = 0; i < brdf.size(); i++) {
    brdf[i] =
        computeBrdf(isoParams_[i], volParams_[i], kVol, geoParams_[i], kGeo);
  }
  return brdf;
}

void RossLi::init() {
  const size_t paramSize = params_.spectral.size();
  std::vector<double> paramWaves(paramSize), iso(paramSize), vol(paramSize),
      geo(paramSize);
  for (size_t i = 0; i < paramSize; i++) {
    const auto& p = params_.spectral[i];
    paramWaves[i] = p.wavelengthMicrons;
    iso[i] = p.isotropic;
    vol[i] = p.volume;
    geo[i] = p.geometric;
  }

  const auto& wavelengths = wavelengthMicrons();
  interpolate(paramWaves, iso, wavelengths, isoParams_);
  interpolate(paramWaves, vol, wavelengths, volParams_);
  interpolate(paramWaves, geo, wavelengths, geoParams_);
}

std::vector<double> RossLi::doSimplifySpectralBrdf(
    std::vector<double>& nadirDHR) {
  // compute total reflectances
  double step = 0.02;
  double kVol, kGeo;

  const auto& wavelengths = wavelengthMicrons();
  nadirDHR.resize(wavelengths.size());
  for (size_t is = 0; is < wavelengths.size(); is++) {
    // find the average value above 20 deg
    double total = 0;
    size_t count = 0;

    for (double muS = 0.35 + 0.5 * step; muS < 1; muS += step) {
      for (double phSf = 0.5 * step; phSf < 1; phSf += step) {
        double phS = phSf * M_PI;
        Angles angles(muS, 1, phS);
        getCoeffs(angles, params_, kVol, kGeo);

        double brdf = computeBrdf(
            isoParams_[is], volParams_[is], kVol, geoParams_[is], kGeo);
        if (brdf < 0) continue;
        total += brdf;
        count++;
      }
    }
    total /= count;
    if (total < 0) total = 0;
    if (total > 1) total = 1;

    nadirDHR[is] = total * M_PI;
  }

  // compute new wavelengths (the parameter wavelengths)
  std::vector<double> newWave(params_.spectral.size());
  for (size_t i = 0; i < newWave.size(); i++) {
    newWave[i] = params_.spectral[i].wavelengthMicrons;
  }
  return newWave;
}

void RossLi::doGetHashInputs(const HashCallback& callback) const {
  auto Add = [&](const auto& data) { AddHashData(data, callback); };

  Add(params_.rossKernel);
  Add(params_.liKernel);
  Add(params_.reciprocalLi);
  Add(params_.br);
  Add(params_.hb);
  Add(params_.spectral);

  // Version number (used as hash salt, update on any implementation changes)
  static constexpr int kVersion = 1;
  Add(kVersion);
}
}
