/** Implementation file for Shell-Background
 *
 *  \file ShellBackground.cpp
 */
#include "dirsig5/brdf/ShellBackground.h"

#include <cmath>

#include "dirsig5/base/SpectralVector.h"

namespace d5::brdf {

std::vector<double> ShellBackground::sample(
    double illumZenithRad,
    double illumAzimuthRad,
    double viewZenithRad,
    double viewAzimuthRad) const {
  std::vector<double> result(wavelengthMicrons().size());

  // Localize azimuth.
  viewAzimuthRad -= illumAzimuthRad;
  if (viewAzimuthRad < 0) viewAzimuthRad += 2 * M_PI;
  if (viewAzimuthRad >= 2 * M_PI) viewAzimuthRad -= 2 * M_PI;
  illumAzimuthRad = 0;

  // Is either theta is out of range?
  constexpr double maxTheta = 89.0 * M_PI / 180.0;
  if (!(illumZenithRad >= 0.0 && illumZenithRad <= maxTheta &&  //
        viewZenithRad >= 0.0 && viewZenithRad <= maxTheta))
    return result;  // Just return 0.

  // Some trigonometric terms.
  double sinThetaO = std::sin(viewZenithRad),
         cosThetaO = std::cos(viewZenithRad);
  double sinThetaI = std::sin(illumZenithRad),
         cosThetaI = std::cos(illumZenithRad);
  double sinPhiO = std::sin(viewAzimuthRad), cosPhiO = std::cos(viewAzimuthRad);
  double tanThetaI = sinThetaI / cosThetaI;
  double tanThetaO = sinThetaO / cosThetaO;

  // Phase angle.
  double xi = 0;
  double cosXi = cosThetaI * cosThetaO + sinThetaI * sinThetaO * cosPhiO;
  double sinXi = std::sqrt(std::max(1 - cosXi * cosXi, 0.0));
  if (std::abs(cosXi) < 1) xi = std::acos(cosXi);

  // Roujean parameters.
  double f1 = 0.0;
  double f2 = 0.0;
  {
    double phi = viewAzimuthRad;
    double sinPhi = sinPhiO;
    double cosPhi = cosPhiO;
    if (phi > M_PI) {
      phi = 2 * M_PI - phi;
      sinPhi = std::sin(phi);
      cosPhi = std::cos(phi);
    }
    double tan2ThetaI = tanThetaI * tanThetaI;
    double tan2ThetaO = tanThetaO * tanThetaO;
    double tmp1 = tan2ThetaI + tan2ThetaO - 2 * tanThetaI * tanThetaO * cosPhi;
    double tmp2 = tanThetaI + tanThetaO + std::sqrt(std::max(tmp1, 0.0));
    double tmp3 = tanThetaI * tanThetaO * ((M_PI - phi) * cosPhi + sinPhi);
    double tmp4 = (M_PI - 2 * xi) * cosXi + 2 * sinXi;
    f1 = M_1_PI * (0.5 * tmp3 - tmp2);
    f2 = M_1_PI * (2.0 / 3.0) * tmp4 / (cosThetaI + cosThetaO) - 1.0 / 3.0;
  }

  for (size_t pos = 0; pos < wavelengthMicrons().size(); pos++) {
    double rho = 1 - emissivity_[pos];
    auto fitItr = std::lower_bound(
        params_.fit.begin(), params_.fit.end(), rho,
        [](const auto& lhs, const auto& rhs) { return lhs.rho < rhs; });
    if (fitItr == params_.fit.end()) --fitItr;
    if (fitItr != params_.fit.begin()) --fitItr;
    const auto& fitA = *(fitItr + 0);
    const auto& fitB = *(fitItr + 1);
    double fac = (rho - fitA.rho) / (fitB.rho - fitA.rho);
    double f00A = fitA.k0 + fitA.k1 * f1 + fitA.k2 * f2;
    double f00B = fitB.k0 + fitB.k1 * f1 + fitB.k2 * f2;
    result[pos] = (1 - fac) * f00A + fac * f00B;
    result[pos] *= 1e-2 * M_1_PI;
    result[pos] += f00Sigma_;
    result[pos] = std::max(result[pos], 0.0);
  }

  return result;
}

void ShellBackground::init() {
  // Interpolate emissivity.
  std::vector<double> emissivityLambda;
  std::vector<double> emissivity;
  for (const auto& sample : params_.emissivity) {
    emissivityLambda.push_back(sample.wavelengthMicrons);
    emissivity.push_back(sample.emissivity);
  }
  emissivity_.clear();
  interpolate(
      emissivityLambda, emissivity, wavelengthMicrons(), emissivity_);
  f00Sigma_ = +params_.a * std::pow(params_.gsdInches, -params_.b);
  f00Sigma_ *= 0.5;
}

std::vector<double> ShellBackground::doSimplifySpectralBrdf(
    std::vector<double>& nadirDHR) {
  // Brute-force DHR.
  nadirDHR.resize(wavelengthMicrons().size());
  std::fill(nadirDHR.begin(), nadirDHR.end(), 0.0);
  constexpr size_t thetaNum = 128;
  constexpr size_t phiNum = 32;  // Should be azimuthally symmetric?
  for (size_t thetaInd = 0; thetaInd < thetaNum; thetaInd++) {
    double thetaI = thetaInd / double(thetaNum) * M_PI_2;
    std::vector<double> brdfSum(wavelengthMicrons().size());
    for (size_t phiInd = 0; phiInd < phiNum; phiInd++) {
      double phiI = phiInd / double(phiNum) * 2 * M_PI;
      auto brdf = sample(thetaI, phiI, 0, 0);
      for (size_t pos = 0; pos < wavelengthMicrons().size(); pos++) {
        brdfSum[pos] += brdf[pos];
      }
    }
    double fac = std::sin(thetaI) * std::cos(thetaI) / phiNum * 2 * M_PI;
    for (size_t pos = 0; pos < wavelengthMicrons().size(); pos++) {
      nadirDHR[pos] += brdfSum[pos] * fac;
    }
  }
  for (size_t pos = 0; pos < wavelengthMicrons().size(); pos++) {
    nadirDHR[pos] *= M_PI_2 / thetaNum;
  }

  // New wavelengths.
  std::vector<double> newWave(params_.emissivity.size());
  for (size_t pos = 0; pos < newWave.size(); pos++) {
    newWave[pos] = params_.emissivity[pos].wavelengthMicrons;
  }

  // Too slow and uses too much memory for many wavelengths. So, reduce
  // new wavelengths by discarding every other one until less than 16 remain.
  auto reduce = [](auto wave) {
    std::vector<double> reduced;
    reduced.reserve(wave.size() / 2);
    bool addedBack = false;
    for (auto waveItr = wave.begin(); waveItr < wave.end();) {
      reduced.push_back(*waveItr);
      if (waveItr + 1 == wave.end()) addedBack = true;
      ++waveItr;
      ++waveItr;
    }
    if (!addedBack) reduced.push_back(wave.back());
    return reduced;
  };
  while (newWave.size() > 16) newWave = reduce(newWave);
  return newWave;
}

void ShellBackground::doGetHashInputs(const HashCallback& callback) const {
  auto Add = [&](const auto& data) { AddHashData(data, callback); };

  Add(params_.a);
  Add(params_.b);
  Add(params_.gsdInches);
  Add(params_.fit);
  Add(params_.emissivity);

  // Version number (used as hash salt, update on any implementation changes)
  static constexpr int kVersion = 1;
  Add(kVersion);
}
}  // namespace d5::brdf
