/** Implementation file for the Shell Target BRDF
 *
 *  \file ShellTarget.cpp
 */

#include "dirsig5/brdf/ShellTarget.h"

#include "dirsig5/base/Constants.h"
#include "dirsig5/base/SpectralVector.h"

namespace d5::brdf {

ShellTarget::ShellTarget(
    std::vector<double> wavelengthMicrons, Parameters params)
    : BrdfModel(std::move(wavelengthMicrons)), params_(std::move(params)) {
  // Sort the spectral samples by wavelength (ascending)
  std::sort(
      params_.spectral.begin(), params_.spectral.end(),
      [](const auto& a, const auto& b) {
        return a.wavelengthMicrons < b.wavelengthMicrons;
      });

  init();
}

std::vector<double> ShellTarget::sample(
    double illumZenithRad,
    double illumAzimuthRad,
    double viewZenithRad,
    double viewAzimuthRad) const {
  const auto& wavelengths = this->wavelengthMicrons();

  constexpr double k2Pi = TwoPI;

  // We only care about the difference in azimuth
  viewAzimuthRad = viewAzimuthRad - illumAzimuthRad;
  if (viewAzimuthRad < 0) viewAzimuthRad += k2Pi;
  if (viewAzimuthRad >= k2Pi) viewAzimuthRad -= k2Pi;
  illumAzimuthRad = 0;

  // Use the Fresnel struct to store all of the trig values
  FresnelReflectance::BrdfAngles angles(
      illumZenithRad, illumAzimuthRad, viewZenithRad, viewAzimuthRad);

  //
  // Perform Lambda Independent Computation
  //

  // Compute Fresnel reflectance/transmission
  // don't need transmission for diffuse/specular interpolation
  std::vector<MuellerMatrix> fresnelRefl, fresnelTran;
  if (params_.interpolation == Interpolation::DiffuseAndSpecular) {
    fresnelRefl = FresnelReflectance::computeReflectance(nk_, angles);
  } else {
    FresnelReflectance::compute(nk_, angles, fresnelRefl, fresnelTran);
  }

  // Compute geometric scaling factor
  double geomScale = 1.0 / (4.0 * angles.cosThetaI * angles.cosThetaR);

  //
  // Perform Lambda Dependent Computations
  //

  std::vector<double> rscale(wavelengths.size());
  std::vector<double> txscale(wavelengths.size());
  std::vector<double> txadd(wavelengths.size());
  std::vector<double> scaledVol(wavelengths.size());
  for (size_t ii = 0; ii < wavelengths.size(); ii++) {
    // compute and interpolate the combined P and S terms
    double l = wavelengths[ii];
    auto [fitA, fitB] = fitIters_[ii];

    //
    // Compute specular contribution
    //

    // compute orientation probability
    double P_a = (*fitA->orientPdf)(angles);
    double P_b = (*fitB->orientPdf)(angles);

    // compute shadowing term
    double S_a = (*fitA->shadowFunc)(angles);
    double S_b = (*fitB->shadowFunc)(angles);

    // compute some interpolation params
    double l_a = fitA->wavelengthMicrons;
    double l_b = fitB->wavelengthMicrons;
    double delta_ab = (l_b - l_a);
    double fraction_a = ((l_b - l) / delta_ab);
    double fraction_b = ((l - l_a) / delta_ab);

    double mu_a = P_a * S_a;
    double mu_b = P_b * S_b;
    double mu = mu_a * fraction_a + mu_b * fraction_b;

    // apply all multiplicative scales (all methods)
    rscale[ii] = geomScale * mu;

    //
    // Handle interpolation methods
    //

    // evaluate volumetric term at both reference wavelengths - Eqn 8.36
    double f_00_vol_a = (*fitA->volumeTerm)(angles);
    double f_00_vol_b = (*fitB->volumeTerm)(angles);
    double f_00_vol_p = f_00_vol_a * fraction_a + f_00_vol_b * fraction_b;

    // early exit for interp method 2 (use provided reflectance)
    if (params_.interpolation == Interpolation::DiffuseAndSpecular) {
      double rho = params_.dhr[ii];
      double fitDHR = fitA->dhr * fraction_a + fitB->dhr * fraction_b;

      // calc wlsf and apply to the result
      rscale[ii] *= rho / fitDHR;
      scaledVol[ii] = f_00_vol_p * rho / fitDHR;
      continue;
    }

    // if interpolation method = 0 set, then ignore the measured
    // reflectance values
    if (params_.interpolation == Interpolation::IgnoreEmissivityCurve) {
      txscale[ii] = f_00_vol_p;
      txadd[ii] = (1 - fitA->tgain) * f_00_vol_p;
      continue;
    }

    //
    // Compute volumetric contribution for interp method 1
    //

    // Jim's interpolation method is MaxwellBeard specific
    auto* volA = static_cast<VTMaxwellBeard*>(fitA->volumeTerm.get());
    auto* volB = static_cast<VTMaxwellBeard*>(fitB->volumeTerm.get());

    // Approximate hemispherical integration of volume terms - Eqn 8.33
    double rho_vol_dhr_a = PI * volA->rho_D_ + 4 * volA->rho_V_;
    double rho_vol_dhr_b = PI * volB->rho_D_ + 4 * volB->rho_V_;

    // Eqn 8.29
    double rho_spec_dhr_a = fitA->dhr - rho_vol_dhr_a;
    double rho_spec_dhr_b = fitB->dhr - rho_vol_dhr_b;

    // Eqn 8.34 (solve for y-isect)
    double delta_rho = (rho_spec_dhr_b - rho_spec_dhr_a);
    double fraction_rho = delta_rho / delta_ab;
    double b = rho_spec_dhr_b - fraction_rho * l_b;
    double rho_spec_dhr = fraction_rho * l + b;

    double rho = params_.dhr[ii];

    // Calculate difference between fit interpolated and measured DHR
    double rho_vol_dhr_p = rho - rho_spec_dhr;
    if (rho_vol_dhr_p < 0.0) rho_vol_dhr_p = 0.0;
    double f_00_vol =
        rho_vol_dhr_p * ((f_00_vol_a / rho_vol_dhr_a) * fraction_a +
                         (f_00_vol_b / rho_vol_dhr_b) * fraction_b);

    txscale[ii] = f_00_vol;
    txadd[ii] = (1 - fitA->tgain) * f_00_vol;
  }

  // Compute the final BRDF
  std::vector<double> brdf(wavelengths.size());
  for (size_t ii = 0; ii < brdf.size(); ii++) {
    brdf[ii] = fresnelRefl[ii](0, 0) * rscale[ii];
    if (params_.interpolation != Interpolation::DiffuseAndSpecular) {
      brdf[ii] +=
          params_.spectral[ii].tgain * fresnelTran[ii](0, 0) * txscale[ii] +
          txadd[ii];
    } else {
      brdf[ii] += scaledVol[ii];
    }
    if (brdf[ii] < 0) brdf[ii] = 0;
  }

  return brdf;
}

void ShellTarget::init() {
  const auto& wavelengths = wavelengthMicrons();

  // Interpolate the DHR to be on the same wavelengths as the BRDF spectral
  // sampling
  std::vector<double> newDhr;
  interpolate(params_.dhrWavelengths, params_.dhr, wavelengths, newDhr);
  params_.dhrWavelengths = wavelengths;
  params_.dhr = newDhr;

  // Prepare the Fresnel inputs
  std::vector<double> fitLambda, fitN, fitK, interpN, interpK;
  fitLambda.reserve(params_.spectral.size());
  fitN.reserve(params_.spectral.size());
  fitK.reserve(params_.spectral.size());
  for (const auto& sp : params_.spectral) {
    fitLambda.push_back(sp.wavelengthMicrons);
    fitN.push_back(sp.ior.real());
    fitK.push_back(sp.ior.imag());
  }

  // Interpolate to the wavelength spectrum
  interpolate(fitLambda, fitN, wavelengths, interpN);
  interpolate(fitLambda, fitK, wavelengths, interpK);
  nk_.clear();
  nk_.reserve(wavelengths.size());
  for (size_t i = 0; i < wavelengths.size(); i++) {
    nk_.emplace_back(interpN[i], interpK[i]);
  }

  // Cache the interpolation fit parameters for each wavelength
  fitIters_.clear();
  fitIters_.reserve(wavelengths.size());
  for (auto wavelength : wavelengths) {
    auto it = std::lower_bound(
        params_.spectral.begin(), params_.spectral.end(), wavelength,
        [](const auto& a, double wave) { return a.wavelengthMicrons < wave; });

    if (it == params_.spectral.end()) {
      fitIters_.emplace_back(it - 2, it - 1);
    } else if (it == params_.spectral.begin()) {
      fitIters_.emplace_back(it, it + 1);
    } else {
      fitIters_.emplace_back(it - 1, it);
    }
  }
}

std::vector<double> ShellTarget::doSimplifySpectralBrdf(
    std::vector<double>& nadirDHR) {
  // The new wavelengths will be the fit sample wavelengths
  std::vector<double> fitWavelengths;
  fitWavelengths.reserve(params_.spectral.size());
  for (const auto& p : params_.spectral) {
    fitWavelengths.push_back(p.wavelengthMicrons);
  }

  // Populate the nadir DHR with out high-res DHR, unless we are explicitly
  // ignoring it, in which case, use the fit parameters.
  if (params_.interpolation == Interpolation::IgnoreEmissivityCurve) {
    std::vector<double> dhr, dhrWave;
    for (const auto& p : params_.spectral) {
      dhrWave.push_back(p.wavelengthMicrons);
      dhr.push_back(p.dhr);
    }
    interpolate(dhrWave, dhr, wavelengthMicrons(), nadirDHR);
  } else {
    nadirDHR = params_.dhr;
  }

  return fitWavelengths;
}

void ShellTarget::doGetHashInputs(const HashCallback& callback) const {
  auto Add = [&](const auto& data) { AddHashData(data, callback); };

  Add(params_.interpolation);
  Add(params_.dhrWavelengths);
  Add(params_.dhr);
  for (const auto& sp : params_.spectral) {
    Add(sp.wavelengthMicrons);
    Add(sp.ior);
    Add(sp.dhr);
    Add(sp.tgain);
    if (sp.orientPdf) sp.orientPdf->getHashInputs(callback);
    if (sp.shadowFunc) sp.shadowFunc->getHashInputs(callback);
    if (sp.volumeTerm) sp.volumeTerm->getHashInputs(callback);
  }

  // Version number (used as hash salt, update on any implementation changes)
  static constexpr int kVersion = 1;
  Add(kVersion);
}
}
