/** Unit tests for ShellTarget.h
 *
 *  \file ShellTargetTest.cpp
 */

#include <memory>

#include <gtest/gtest.h>

#include "dirsig5/brdf/ShellTarget.h"
#include "dirsig5/base/Constants.h"

using namespace d5::brdf;
using d5::PI;

class BrdfShellTargetTest : public ::testing::Test {
protected:
  void SetUp() override {
    std::vector<double> wavelengths{0.4, 3.8, 4.2};

    ShellTarget::Parameters params;
    params.dhrWavelengths = wavelengths;
    params.dhr.resize(3, 1.0 - 0.93005118);
    params.spectral.resize(3);

    params.spectral[0].wavelengthMicrons = 0.3;
    params.spectral[0].ior = std::complex<double>(1.4, 0.4);
    params.spectral[0].dhr = 0.069948816;
    params.spectral[0].tgain = 0;
    params.spectral[0].orientPdf = std::make_shared<OPGaussian>(1.3, 0.125);
    params.spectral[0].shadowFunc = std::make_shared<SFMaxwellBeard>(10, 5);
    params.spectral[0].volumeTerm =
        std::make_shared<VTMaxwellBeard>(1.1e-5, 1e-7);

    params.spectral[1].wavelengthMicrons = 0.4;
    params.spectral[1].ior = std::complex<double>(1.4, 0.4);
    params.spectral[1].dhr = 0.069948816;
    params.spectral[1].tgain = 0;
    params.spectral[1].orientPdf = params.spectral[0].orientPdf;
    params.spectral[1].shadowFunc = params.spectral[0].shadowFunc;
    params.spectral[1].volumeTerm = params.spectral[0].volumeTerm;

    params.spectral[2].wavelengthMicrons = 14.4;
    params.spectral[2].ior = std::complex<double>(1.2, 0.5);
    params.spectral[2].dhr = 0.069948816;
    params.spectral[2].tgain = 0;
    params.spectral[2].orientPdf = params.spectral[0].orientPdf;
    params.spectral[2].shadowFunc = params.spectral[0].shadowFunc;
    params.spectral[2].volumeTerm = params.spectral[0].volumeTerm;

    shellRef_ = std::make_unique<ShellTarget>(
        std::move(wavelengths), std::move(params));
  }

  std::unique_ptr<ShellTarget> shellRef_;
};

/**
 * Test Gartley verification:
 *
 *     Angles IN (theta, phi):  (10, 0)
 *     Angles OUT (theta, phi): (10, 90)
 *
 * This verification data was generated using Mathcad
 * implementation of ShellTarget.
 */
TEST_F(BrdfShellTargetTest, GartleyCase1) {
  double sourceTheta = 10 * PI / 180;
  double sourcePhi = 0;
  double viewTheta = 10 * PI / 180;
  double viewPhi = PI / 2;

  auto brdf = shellRef_->sample(sourceTheta, sourcePhi, viewTheta, viewPhi);
  EXPECT_NEAR(brdf[0], 0.11478523, 1e-4);
  EXPECT_NEAR(brdf[2], 0.11278871, 1e-4);
}

TEST_F(BrdfShellTargetTest, GartleyCase2) {
  double sourceTheta = 20 * PI / 180;
  double sourcePhi = 0;
  double viewTheta = 10 * PI / 180;
  double viewPhi = PI;

  auto brdf = shellRef_->sample(sourceTheta, sourcePhi, viewTheta, viewPhi);
  EXPECT_NEAR(brdf[0], 1.533036E-01, 1e-4);
  EXPECT_NEAR(brdf[2], 1.506752E-01, 1e-4);
}

TEST_F(BrdfShellTargetTest, GartleyCase3) {
  double sourceTheta = 50 * PI / 180;
  double sourcePhi = 0;
  double viewTheta = 50 * PI / 180;
  double viewPhi = 200 * PI / 180;

  auto brdf = shellRef_->sample(sourceTheta, sourcePhi, viewTheta, viewPhi);
  EXPECT_NEAR(brdf[0], 1.622258E-01, 1e-4);
  EXPECT_NEAR(brdf[2], 1.627673E-01, 1e-4);
}

TEST_F(BrdfShellTargetTest, GartleyCase4) {
  double sourceTheta = 50 * PI / 180;
  double sourcePhi = 0;
  double viewTheta = 50 * PI / 180;
  double viewPhi = 275 * PI / 180;

  auto brdf = shellRef_->sample(sourceTheta, sourcePhi, viewTheta, viewPhi);
  EXPECT_NEAR(brdf[0], 1.115687E-005, 1e-4);
  EXPECT_NEAR(brdf[2], 1.115687E-05, 1e-4);
}
