#include "dirsig5/d4legacy/geometry/FacetObject.h"

#include <algorithm>
#include <limits>
#include <stdexcept>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Logger.h"

namespace d5::d4legacy {

// ###########################################
// #         Part API Implementation         #
// ###########################################

uint32_t FacetObject::addPart(const std::string& name) {
  auto it = std::find_if(parts_.begin(), parts_.end(), [&](const auto& p) {
    return p.name == name;
  });
  if (it != parts_.end()) return it - parts_.begin();
  auto& p = parts_.emplace_back();
  p.name = name;
  return parts_.size() - 1;
}

auto FacetObject::getPartByName(const std::string& name) const -> const Part* {
  auto it = std::find_if(parts_.begin(), parts_.end(), [&](const auto& p) {
    return p.name == name;
  });
  if (it != parts_.end()) return &(*it);
  return nullptr;
}

int FacetObject::getPartIndexByName(const std::string& name) const {
  auto it = std::find_if(parts_.begin(), parts_.end(), [&](const auto& p) {
    return p.name == name;
  });
  if (it != parts_.end()) return it - parts_.begin();
  return -1;
}

template <size_t N>
void SwapStride(std::vector<uint32_t>& list, size_t a, size_t b) {
  a *= N;
  b *= N;
  for (size_t i = 0; i < N; i++) {
    std::swap(list[a + i], list[b + i]);
  }
}

template <std::size_t N, typename T, typename GetItem>
std::vector<std::uint32_t> RemoveUnused(
  std::vector<T>& buffer, const std::vector<bool>& used, GetItem getItem) {
  // Make a map of old -> new index and remap values
  std::vector<std::uint32_t> remap(used.size());
  std::size_t newSize = remap.size();
  for (std::size_t i = 0; i < newSize; i++) {
    if (used[i]) {
      remap[i] = i;
    } else {
      newSize--;
      while (newSize > i && !used[newSize]) newSize--;
      if (newSize == i) break;
      remap[i] = 0;
      remap[newSize] = i;
      const auto src = getItem(newSize);
      auto dst = getItem(i);
      for (std::size_t i = 0; i < N; i++) dst[i] = src[i];
    }
  }

  // Clear out the remainder of the buffer
  buffer.resize(newSize * N);

  return remap;
}

void FacetObject::removePart(const std::string& name) {
  // Determine the index of the part
  size_t partIdx;
  {
    auto it = std::find_if(parts_.begin(), parts_.end(), [&](const auto& p) {
      return p.name == name;
    });
    if (it == parts_.end()) return;
    partIdx = it - parts_.begin();
  }

  // Determine which of our lists we need to clean up
  bool cleanupVtF = !vtFacets_.empty(), cleanupVnF = !vnFacets_.empty(),
       cleanupTempK = !facetTempK_.empty(),
       cleanupThickness = !facetThickness_.empty();

  // We will clean up lists by simply swapping facets to the end and truncating,
  // newEnd will be the new number of facets
  size_t newEnd = numFacets();
  while (newEnd != 0 && partForFacet_[newEnd - 1] == partIdx) newEnd--;

  // Go through each facet and swap it to the end if it is in the part being
  // removed
  for (size_t i = 0; i < newEnd; i++) {
    if (partForFacet_[i] != partIdx) continue;

    size_t newEndInc = newEnd - 1;
    SwapStride<kFacetStride>(f_, i, newEndInc);
    std::swap(matForFacet_[i], matForFacet_[newEndInc]);
    std::swap(partForFacet_[i], partForFacet_[newEndInc]);
    if (cleanupThickness)
      std::swap(facetThickness_[i], facetThickness_[newEndInc]);
    if (cleanupTempK) std::swap(facetTempK_[i], facetTempK_[newEndInc]);
    if (cleanupVtF) SwapStride<kFacetStride>(vtFacets_, i, newEndInc);
    if (cleanupVnF) SwapStride<kFacetStride>(vnFacets_, i, newEndInc);
    newEnd--;
    while (newEnd != 0 && partForFacet_[newEnd - 1] == partIdx) newEnd--;
  }

  // Truncate all of the lists
  f_.resize(newEnd * kFacetStride);
  matForFacet_.resize(newEnd);
  partForFacet_.resize(newEnd);
  if (cleanupThickness) facetThickness_.resize(newEnd);
  if (cleanupTempK) facetTempK_.resize(newEnd);
  if (cleanupVtF) vtFacets_.resize(newEnd * kFacetStride);
  if (cleanupVnF) vnFacets_.resize(newEnd * kFacetStride);

  // we now need to update partForFacet_ to reflect that the part has been
  // removed
  for (auto& p : partForFacet_) {
    if (p > partIdx) p--;
  }

  // Now clean up point sources
  newEnd = pointSources_.size();
  while (newEnd != 0 && pointSources_[newEnd - 1].partIndex == partIdx) {
    newEnd--;
  }
  for (size_t i = 0; i < newEnd; i++) {
    if (pointSources_[i].partIndex != partIdx) continue;
    std::swap(pointSources_[i], pointSources_[newEnd - 1]);
    newEnd--;
    while (newEnd != 0 && pointSources_[newEnd - 1].partIndex == partIdx) {
      newEnd--;
    }
  }
  pointSources_.resize(newEnd);
  for (auto& ps : pointSources_) {
    if (ps.partIndex > partIdx) ps.partIndex--;
  }

  // Erase any keypoints for this part, reindex the remaining keypoints
  for (auto it = keypoints_.begin(); it != keypoints_.end();) {
    if (it->second.partIndex == partIdx) {
      it = keypoints_.erase(it);
    } else {
      if (it->second.partIndex > partIdx) it->second.partIndex--;
      ++it;
    }
  }

  // Now, clean up unused vertices (vertices only in the removed part)
  if (!v_.empty()) {
    // Determine the vertices that are used
    std::vector<bool> used(numVertices(), false);
    for (auto vi : f_) used[vi] = true;
    for (const auto& s : pointSources()) used[s.vertexIndex] = true;

    // Cleanup the buffer and get the remapping
    auto remap = RemoveUnused<kVertexStride>(
      v_, used, [&](std::size_t i) { return vertex(i); });

    // Update vertex indices in facets/point sources
    for (auto& vi : f_) vi = remap[vi];
    for (auto& s : pointSources_) s.vertexIndex = remap[s.vertexIndex];
  }

  // Do the same for texture coordinates
  if (!vt_.empty()) {
    // Determine the texture coordinates that are used
    std::vector<bool> used(numTextureCoords(), false);
    for (auto i : vtFacets_) used[i] = true;

    // Make a map of old -> new index and remap values
    auto remap = RemoveUnused<kVtStride>(
      vt_, used, [&](std::size_t i) { return textureCoord(i); });

    // Update the indices
    for (auto& i : vtFacets_) i = remap[i];
  }

  // Do the same for vertex normals
  if (!vn_.empty()) {
    // Determine the texture coordinates that are used
    std::vector<bool> used(numVertexNormals(), false);
    for (auto i : vnFacets_) {
      if (i < used.size()) used[i] = true;
    }

    // Make a map of old -> new index and remap values
    auto remap = RemoveUnused<kVnStride>(
      vn_, used, [&](std::size_t i) { return vertexNormal(i); });

    // Update the indices
    for (auto& i : vnFacets_) {
      if (i < remap.size()) i = remap[i];
    }
  }

  // Finally, erase the part from the parts list
  parts_.erase(parts_.begin() + partIdx);

  // Once the part has been removed, we need to clean up any unused materials.
  // So, we need to go through our facets and determine what materials are
  // now unused
  std::vector<bool> matUsed(numMaterials(), false);
  for (auto m : matForFacet_) matUsed[m] = true;
  if (std::any_of(matUsed.begin(), matUsed.end(), [](bool u) { return !u; })) {
    // Determine a remapping of old material index to new material index
    std::vector<uint32_t> matRemap(numMaterials());
    std::vector<std::string> newMats;
    uint32_t newIndex = 0;
    for (size_t i = 0; i < matRemap.size(); i++) {
      matRemap[i] = newIndex;
      if (matUsed[i]) {
        newMats.push_back(materialLabel(i));
        newIndex++;
      }
    }

    // Perform remapping of facet indices
    for (auto& m : matForFacet_) m = matRemap[m];

    // Make the new list of materials
    materialLabels_.swap(newMats);
  }

  // We then do the same thing for point source materials
  std::vector<bool> psMatUsed(numPointSourceMaterials(), false);
  for (const auto& s : pointSources()) psMatUsed[s.materialIndex] = true;
  if (!std::all_of(
        psMatUsed.begin(), psMatUsed.end(), [](bool u) { return u; })) {
    // Determing a remapping of old index to new index
    std::vector<std::uint32_t> remap(psMatUsed.size());
    std::vector<std::string> newMats;
    std::uint32_t newIndex = 0;
    for (std::size_t i = 0; i < remap.size(); i++) {
      remap[i] = newIndex;
      if (psMatUsed[i]) {
        newMats.push_back(pointSourceMaterialLabel(i));
        newIndex++;
      }
    }

    // Remap and replace the materials
    for (auto& s : pointSources_) s.materialIndex = remap[s.materialIndex];
    pointSourceMaterialLabels_.swap(newMats);
  }
}

// ###############################################
// #         Material API Implementation         #
// ###############################################

uint32_t FacetObject::addMaterialLabel(const std::string& label) {
  auto it = matLabelToLocalId_.find(label);
  if (it != matLabelToLocalId_.end()) return it->second;
  auto id = materialLabels_.size();
  matLabelToLocalId_[label] = id;
  materialLabels_.push_back(label);
  return id;
}

// #########################################################
// #        Point Source Material API Implementation       #
// #########################################################

uint32_t FacetObject::addPointSourceMaterialLabel(const std::string& label) {
  auto it = pointSourceMatLabelToLocalId_.find(label);
  if (it != pointSourceMatLabelToLocalId_.end()) return it->second;
  auto id = pointSourceMaterialLabels_.size();
  pointSourceMatLabelToLocalId_[label] = id;
  pointSourceMaterialLabels_.push_back(label);
  return id;
}

// #############################################
// #        SplitOffParts Implementation       #
// #############################################

template <typename Callable>
std::array<uint32_t, 3> RemapVertices(
  const uint32_t* facet, std::vector<uint32_t>& remap, Callable callback) {
  std::array<uint32_t, 3> remapped;
  for (size_t i = 0; i < 3; i++) {
    auto vIdx = facet[i];
    if (vIdx != FacetObject::kNoVertexNormal) {
      auto& remappedVertexIdx = remap[vIdx];
      if (remappedVertexIdx == FacetObject::kNoVertexNormal) {
        remappedVertexIdx = callback(vIdx);
      }
      remapped[i] = remappedVertexIdx;
    } else {
      remapped[i] = FacetObject::kNoVertexNormal;
    }
  }
  return remapped;
}

std::vector<FacetObject> SplitOffParts(
  FacetObject& fo, const std::vector<std::string>& partNamesToSplit) {
  std::vector<size_t> partsToSplit;
  partsToSplit.reserve(partNamesToSplit.size());
  for (const auto& n : partNamesToSplit) {
    auto it = std::find_if(
      fo.parts().begin(), fo.parts().end(),
      [&](const FacetObject::Part& p) { return p.name == n; });
    DIRSIG_THROW_IF(it == fo.parts().end(), "unknown part name \"" + n + "\"");
    partsToSplit.push_back(it - fo.parts().begin());
  }

  // We'll have a persistent remapping of source vertex idx -> part vertex idx
  // and source material idx -> part material idx
  std::vector<uint32_t> vertexRemap(fo.numVertices()),
    materialRemap(fo.numMaterials()), vtRemap(fo.numTextureCoords()),
    vnRemap(fo.numVertexNormals()),
    psMaterialRemap(fo.numPointSourceMaterials());
  auto ClearRemap = [](std::vector<uint32_t>& r) {
    std::fill(r.begin(), r.end(), FacetObject::kNoVertexNormal);
  };

  const auto nFacets = fo.numFacets();
  const auto nTexFacets = fo.numTextureFacets();
  const auto nVertexNormalFacets = fo.numVertexNormalFacets();

  bool hasThicknesses = fo.hasFacetThicknesses(),
       hasTemps = fo.hasFacetTemperatures();

  // Split each part in turn
  std::vector<FacetObject> partList(partsToSplit.size());
  for (size_t i = 0; i < partsToSplit.size(); i++) {
    auto partIdx = partsToSplit[i];
    auto& dst = partList[i];

    // Clear all of the remappings for each part
    ClearRemap(vertexRemap);
    ClearRemap(materialRemap);
    ClearRemap(vtRemap);
    ClearRemap(vnRemap);
    ClearRemap(psMaterialRemap);

    // The split parts will only have one part each, but keep the name
    dst.addPart(fo.parts()[partIdx].name);

    // Process each facet, moving over vertices and materials as needed
    dst.reserveFacets(fo.parts()[partIdx].numFacets);
    for (size_t facetIndex = 0; facetIndex < nFacets; facetIndex++) {
      if (fo.partForFacet(facetIndex) != partIdx) continue;

      // Remap the three vertices into the part
      auto vIdx =
        RemapVertices(fo.facet(facetIndex), vertexRemap, [&](uint32_t idx) {
          const auto* v = fo.vertex(idx);
          return dst.addVertex(v[0], v[1], v[2]);
        });

      // Remap the facet material into the part
      auto matIndex = fo.facetMaterialIndices()[facetIndex];
      auto& remappedMatIndex = materialRemap[matIndex];
      if (remappedMatIndex == FacetObject::kNoVertexNormal) {
        remappedMatIndex = dst.addMaterialLabel(fo.materialLabel(matIndex));
      }

      // Add the facet to the part
      auto newFacetIdx =
        dst.addFacetToPart(0, vIdx[0], vIdx[1], vIdx[2], remappedMatIndex);

      // Move over facet temperature/thickness if they are there
      if (hasThicknesses) {
        dst.setFacetThicknessM(newFacetIdx, fo.facetThicknessesM()[facetIndex]);
      }
      if (hasTemps) {
        dst.setFacetTemperatureK(
          newFacetIdx, fo.facetTemperaturesK()[facetIndex]);
      }

      // Move over texture coordinates and vertex normals if they are there
      if (facetIndex < nTexFacets) {
        auto vtIdx = RemapVertices(
          fo.textureFacet(facetIndex), vtRemap, [&](uint32_t idx) {
            const auto* vt = fo.textureCoord(idx);
            return dst.addTextureCoord(vt[0], vt[1]);
          });
        dst.setTextureFacet(newFacetIdx, vtIdx[0], vtIdx[1], vtIdx[2]);
      }
      if (facetIndex < nVertexNormalFacets) {
        auto vnIdx = RemapVertices(
          fo.vertexNormalFacet(facetIndex), vnRemap, [&](uint32_t idx) {
            const auto* vn = fo.vertexNormal(idx);
            return dst.addVertexNormal(vn[0], vn[1], vn[2]);
          });
        dst.setVertexNormalFacet(newFacetIdx, vnIdx[0], vnIdx[1], vnIdx[2]);
      }
    }

    // Process the point sources
    for (const auto& ps : fo.pointSources()) {
      // Only worry about ones in this part
      if (ps.partIndex != partIdx) continue;

      // Move over the material if needed
      auto& remappedMatIndex = psMaterialRemap[ps.materialIndex];
      if (remappedMatIndex == FacetObject::kNoVertexNormal) {
        remappedMatIndex = dst.addPointSourceMaterialLabel(
          fo.pointSourceMaterialLabel(ps.materialIndex));
      }

      auto& newPs = dst.addPointSource();
      newPs.vertexIndex = vertexRemap[ps.vertexIndex];
      if (newPs.vertexIndex == FacetObject::kNoVertexNormal) {
        const auto* v = fo.vertex(ps.vertexIndex);
        newPs.vertexIndex = dst.addVertex(v[0], v[1], v[2]);
      }
      newPs.materialIndex = remappedMatIndex;
      newPs.direction = ps.direction;
      newPs.partIndex = 0;
    }

    // Move over any keypoints
    for (const auto& kp : fo.keypoints()) {
      if (kp.second.partIndex == partIdx) {
        dst.addKeypointToPart(0, kp.first, kp.second.points);
      }
    }
  }

  for (auto& part : partNamesToSplit) fo.removePart(part);

  return partList;
}

void FacetObject::removeAllUnusedAttributes() {
  auto AllZero = [](auto& elems) {
    for (auto& elem : elems)
      if (elem != 0) return false;
    return true;
  };
  if (AllZero(facetTempK_)) removeFacetTemperatures();
  if (AllZero(facetThickness_)) removeFacetThicknesses();
  if (AllZero(facetThermalEmissivity_)) removeFacetThermalEmissivities();
}

// ###############################
// #       Transform API         #
// ###############################

void FacetObject::transform(Eigen::Matrix4f m) {
  if (m(3, 0) == 0 && m(3, 1) == 0 && m(3, 2) == 0 && m(3, 3) == 1) {
    transform(Eigen::Matrix3f(m.block<3, 3>(0, 0)));
    translate(Eigen::Vector3f(m.block<3, 1>(0, 3)));
  } else {
    for (uint32_t i = 0; i < numVertices(); i++) {
      Eigen::Vector4f v = {mapv(i)[0], mapv(i)[1], mapv(i)[2], 1.0f};
      v = m * v;
      v /= v[3];
      mapv(i) << v[0], v[1], v[2];
    }
    Eigen::Matrix4f mI = m.inverse().transpose();
    for (uint32_t i = 0; i < numVertexNormals(); i++) {
      Eigen::Vector4f n = {
        mapvn(i)[0],  //
        mapvn(i)[1],  //
        mapvn(i)[2], 0.0f};
      n = mI * n;
      mapvn(i) << n[0], n[1], n[2];
    }
  }
}

void FacetObject::transform(Eigen::Matrix3f m) {
  for (uint32_t i = 0; i < numVertices(); i++) {
    mapv(i) << m * mapv(i);
  }
  Eigen::Matrix3f mI = m.inverse().transpose();
  for (uint32_t i = 0; i < numVertexNormals(); i++) {
    mapvn(i) << mI * mapvn(i);
  }
}

void FacetObject::translate(Vectorf t) {
  for (uint32_t i = 0; i < numVertices(); i++) mapv(i) += t;
}

void FacetObject::scale(Vectorf s) {
  transform(Eigen::Matrix3f{
    {s[0], 0.0f, 0.0f},  //
    {0.0f, s[1], 0.0f},  //
    {0.0f, 0.0f, s[2]}});
}

void FacetObject::scale(float s) {
  transform(Eigen::Matrix3f{
    {s, 0.0f, 0.0f},  //
    {0.0f, s, 0.0f},  //
    {0.0f, 0.0f, s}});
}

void FacetObject::swapAxes(uint32_t ax0, uint32_t ax1) {
  for (uint32_t i = 0; i < numVertices(); i++)
    std::swap(vertex(i)[ax0], vertex(i)[ax1]);
  for (uint32_t i = 0; i < numVertexNormals(); i++)
    std::swap(vertexNormal(i)[ax0], vertexNormal(i)[ax1]);
}

BBoxf FacetObject::computeExtent() const {
  BBoxf extent;
  for (uint32_t i = 0; i < numVertices(); i++) extent |= Vectorf(mapv(i));
  return extent;
}

Vectorf FacetObject::computeExtentCenter() const {
  auto extent = computeExtent();
  return 0.5f * extent.min + 0.5f * extent.max;
}

Vectorf FacetObject::computeCentroid() const {
  Vector c = {0.0, 0.0, 0.0};
  for (uint32_t i = 0; i < numVertices(); i++) c += mapv(i).cast<double>();
  if (numVertices() > 0) c /= double(numVertices());
  return c.cast<float>();
}

// ##########################################
// #       File Write Implementation        #
// ##########################################

void FacetObject::writeObj(std::ostream& os) const {
  os << "# Written by DIRSIG5 d4legacy::FacetObject\n";
  os << "#\n";
  os << "# - Vertex count = " << numVertices() << '\n';
  os << "# - Vertex normal count = " << numVertexNormals() << '\n';
  os << "# - Texture coord count = " << numTextureCoords() << '\n';
  os << "# - Facet count = " << numFacets() << '\n';
  os << "# - Part count = " << parts_.size() << '\n';
  for (const auto& part : parts_) os << "#   * " << part.name << '\n';
  os << "# - Material count = " << materialLabels_.size() << '\n';
  for (const auto& label : materialLabels_) os << "#   * " << label << '\n';
  os << "\n# v\n";
  for (uint32_t i = 0; i < numVertices(); i++)
    os << "v "                 //
       << vertex(i)[0] << ' '  //
       << vertex(i)[1] << ' '  //
       << vertex(i)[2] << '\n';
  os << "# vn\n";
  for (uint32_t i = 0; i < numVertexNormals(); i++)
    os << "vn "                      //
       << vertexNormal(i)[0] << ' '  //
       << vertexNormal(i)[1] << ' '  //
       << vertexNormal(i)[2] << '\n';
  os << "# vt\n";
  for (uint32_t i = 0; i < numTextureCoords(); i++)
    os << "vt "                      //
       << textureCoord(i)[0] << ' '  //
       << textureCoord(i)[1] << '\n';
  os << "# f\n";
  uint32_t partIndex = 0;
  for (const Part& part : parts_) {
    os << "g " << part.name << '\n';
    std::string_view lastMat;
    for (uint32_t i = 0; i < numFacets(); i++)
      if (partForFacet_[i] == partIndex) {
        std::string_view mat = materialLabels_[matForFacet_[i]];
        if (mat != lastMat) os << "usemtl " << mat << '\n';
        lastMat = mat;
        os << 'f';
        for (uint32_t j = 0; j < 3; j++) {
          os << ' ';
          os << facet(i)[j] + 1;
          if (!vtFacets_.empty()) os << '/' << vtFacets_[3 * i + j] + 1;
          if (!vnFacets_.empty()) {
            if (vtFacets_.empty()) os << '/';
            os << '/' << vnFacets_[3 * i + j] + 1;
          }
        }
        os << '\n';
      }
    partIndex++;
  }
}

void FacetObject::writeGdb(std::ostream& os) const {
  os << "OBJECT\n";
  os << "Object\n";  // Name?
  os << "1-0-0\n";
  uint32_t partIndex = 0;
  for (const Part& part : parts_) {
    os << "PART\n";
    os << "1-" << partIndex << "-0\n";
    os << part.name << '\n';
    uint32_t facetIndex = 0;
    for (uint32_t i = 0; i < numFacets(); i++) {
      if (partForFacet_[i] != partIndex) continue;
      os << "FACE\n";
      os << part.name << '_' << facetIndex << '\n';
      os << "1-" << partIndex << '-' << facetIndex << '\n';
      os << "UNKNOWN\n";  // Material name (not used)
      os << materialLabels_[matForFacet_[i]] << '\n';
      os << part.name << '_' << facetIndex << '\n';
      if (hasFacetTemperatures())
        os << facetTempK_[i] << '\n';
      else
        os << "-1\n";
      if (hasFacetThicknesses())
        os << facetThickness_[i] << '\n';
      else
        os << "0\n";
      os << "0\n"  // Self-generated power
            "0\n"  // Shape factor
            "NULL\n"
            "NULL\n"
            "NULL\n"
            "3\n";  // Vertex count
      facetIndex++;
      for (int j = 0; j < 3; j++) {
        const float* v = vertex(facet(i)[j]);
        os << v[0] << ' '  //
           << v[1] << ' '  //
           << v[2] << '\n';
      }
      auto tmpn = facetNormal(i);
      Vector n = {tmpn[0], tmpn[1], tmpn[2]};
      os << n[0] << ' '  //
         << n[1] << ' '  //
         << n[2] << '\n';
      os << getZenith(n) << '\n';
      os << getGeoAzimuth(n) << '\n';
      os << "0.0\n";
    }
    partIndex++;
  }
}

std::string FacetObject::summary() const {
  std::string s = fmt::format(
    "FacetObject \"{}\"\n"
    "  Vertex count = {}\n"
    "  Facet count = {}\n"
    "  Part count = {}\n",
    originalFilename(), numVertices(), numFacets(), parts().size());
  if (parts().size() > 1)
    for (const auto& part : parts())
      s += fmt::format(
        "    - \"{}\" (Facet count = {})\n", part.name, part.numFacets);
  BBoxf extent = computeExtent();
  s += fmt::format(
    "  Extent min = {:+.3f}, {:+.3f}, {:+.3f}\n"
    "  Extent max = {:+.3f}, {:+.3f}, {:+.3f}\n",
    extent.min[0], extent.min[1], extent.min[2],  //
    extent.max[0], extent.max[1], extent.max[2]);
  s += fmt::format(
    "  Has facet thicknesses? = {}\n"
    "  Has facet temperatures? = {}\n"
    "  Has facet thermal emissivities? = {}\n",
    hasFacetThicknesses() ? "Yes" : "No", hasFacetTemperatures() ? "Yes" : "No",
    hasFacetThermalEmissivities() ? "Yes" : "No");
  if (numMaterials() > 0) {
    s += fmt::format("  Material count = {}\n", numMaterials());
    std::vector<long> matUsage(numMaterials());
    for (uint32_t i : facetMaterialIndices()) matUsage[i]++;
    for (uint32_t i = 0; i < numMaterials(); i++)
      s += fmt::format(
        "    - \"{}\" (Facet count = {})\n", materialLabel(i), matUsage[i]);
  }
  return s;
}

void FacetObject::validateIndexes() const {
  for (auto index : f_)
    DIRSIG_THROW_IF_FMT(
      index >= v_.size() / 3, "Out-of-range position index! ({} >= {})", index,
      v_.size() / 3);
  for (auto index : vnFacets_)
    DIRSIG_THROW_IF_FMT(
      index >= vn_.size() / 3 && index != UINT32_MAX,
      "Out-of-range normal index! ({} >= {})", index, vn_.size() / 3);
  for (auto index : vtFacets_)
    DIRSIG_THROW_IF_FMT(
      index >= vt_.size() / 2,
      "Out-of-range texture coordinate index! ({} >= {})", index,
      vt_.size() / 2);
}

}  // namespace d5::d4legacy
