#include "dirsig5/d4legacy/geometry/FacetObjectIO.h"

#include <stdexcept>

#include "dirsig5/base/StringUtils.h"
#include "readers/AssimpReader.h"
#include "readers/GdbReader.h"
#include "readers/ObjReader.h"
#include "readers/TdfReader.h"

namespace d5::d4legacy {

FacetObject ReadFacetObjectFromFile(const FacetObjectReadRequest& request) {
  auto lowerFilename = toLower(request.filename());

  FacetObject obj;
  if (endsWith(lowerFilename, ".obj")) {
    obj = ReadObjFile(request);
  } else if (endsWith(lowerFilename, ".gdb")) {
    obj = ReadGdbFile(request);
#if WITH_MUSES
  } else if (endsWith(lowerFilename, ".tdf")) {
    obj = ReadTdfFile(request);
#endif  // #if WITH_MUSES
  } else {
#if WITH_ASSIMP
    obj = ReadAssimpFile(request);
#else
    DIRSIG_THROW("Unrecognized geometry file extension");
#endif // #if WITH_ASSIMP
  }

  // Attach any keypoints
  for (const auto& kp : request.keypoints()) {
    size_t partIndex = 0;
    if (!kp.part.empty()) {
      if (int idx = obj.getPartIndexByName(kp.part); idx != -1)
        partIndex = static_cast<size_t>(idx);
      else
        DIRSIG_THROW(
          "A keypoint was attached to a part with the name \"" + kp.part +
          "\", but this part was not found in the object");
    }
    obj.addKeypointToPart(partIndex, kp.name, kp.points);
  }

  // Be sure we don't keep thicknesses, emissivities, and temperatures
  // if they aren't actually used.
  obj.removeAllUnusedAttributes();

  if (request.temperature() >= 0) {
    auto tempK = float(request.temperature());
    for (uint32_t index = 0; index < obj.numFacets(); index++)
      obj.setFacetTemperatureK(index, tempK);
  }

  obj.setOriginalFilename(request.filename());
  obj.validateIndexes();
  return obj;
}

}  // namespace d5::d4legacy
