#include "dirsig5/d4legacy/geometry/Geometry.h"

#include "dirsig5/base/Error.h"
#include "dirsig5/d4legacy/geometry/Primitives.h"
#include "spdlog/fmt/fmt.h"

using namespace std::string_view_literals;

namespace d5::d4legacy {

std::shared_ptr<Geometry> Geometry::FromXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  xml::Node subnode = node;
  if (subnode.name() == "basegeometry"sv or subnode.name() == "basesource"sv)
    subnode = node.first_child();
  std::shared_ptr<Geometry> geometry;
  std::string tagName = subnode.name();
  if (tagName == "box")
    geometry = std::make_shared<Box>();
  else if (tagName == "catenary" || tagName == "catenarycurve")
    geometry = std::make_shared<CatenaryCurve>();
  else if (tagName == "curvedfrustum")
    geometry = std::make_shared<CurvedFrustum>();
  else if (tagName == "cylinder")
    geometry = std::make_shared<Cylinder>();
  else if (tagName == "sphere")
    geometry = std::make_shared<Sphere>();
  else if (tagName == "sinusoid")
    geometry = std::make_shared<Sinusoid>();
  else if (tagName == "secchidisk")
    geometry = std::make_shared<SecchiDisk>();
  else if (tagName == "disk")
    geometry = std::make_shared<Disk>();
  else if (tagName == "groundplane" || tagName == "surfaceplane")
    geometry = std::make_shared<SurfacePlane>();
  else if (tagName == "obj" || tagName == "fbx" || tagName == "assimp") // Hack
    geometry = std::make_shared<ObjFacetObject>();
  else if (tagName == "gdb")
    geometry = std::make_shared<GdbFacetObject>();
  else if (tagName == "tdf")
    geometry = std::make_shared<TdfFacetObject>();
  else if (tagName == "hdf")
    geometry = std::make_shared<HdfFacetObject>();
  else if (tagName == "odb")
    geometry = std::make_shared<OdbInclude>();
  else if (tagName == "glist")
    geometry = std::make_shared<GlistInclude>();
  else if (tagName == "regulargrid")
    geometry = std::make_shared<RegularGrid>();
  else if (tagName == "beziercurveset")
    geometry = std::make_shared<BezierCurveSet>();
  else if (tagName == "pointsource")
    geometry = std::make_shared<PointSource>();
  else
    DIRSIG_THROW_FMT(
      "line {}: unrecognized base geometry <{}>", xml::lineno(subnode),
      tagName);
  if (geometry) {
    geometry->lineNumber = xml::lineno(subnode);
    geometry->ReadXML(subnode, facade, fileLocator);
  }
  return geometry;
}

void Geometry::ReadXML(
  const xml::Node&, const geolib::Facade*, const FileLocator&) {
  DIRSIG_THROW_FMT(
    "Geometry type '{}' doesn't implement ReadXML()", to_string(type_));
}

void Geometry::ReadTag(const TagFileSection&) {
  DIRSIG_THROW_FMT(
    "Geometry type '{}' doesn't implement ReadTag()", to_string(type_));
}

void FacetObjectWithProperties::FacetObjectReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  filename = xml::readFilename(
    node, "filename", fileLocator, FileLocator::GeometryList);
  units = xml::readDistanceUnitsAttribute(node);
  swapYZ = node.attribute("swapyz").as_bool(false);
  defaultTemperature = xml::readNumber(node, "temperature", -1);
  for (auto assign : node.children("assign"))
    materialRemap[assign.text().as_string()] =
      assign.attribute("id").as_string();
  for (auto part : node.child("properties").children("part")) {
    bool hasProps = false;
    Part partProps;
    partProps.name = part.attribute("name").as_string();
    if (!part.attribute("enabled").as_bool(true)) {
      partProps.enabled = false;
      hasProps = true;
    }
    if (part.child("matid")) {
      partProps.materialId = part.child("matid").text().as_string();
      hasProps = true;
    }
    if (part.child("origin")) {
      partProps.origin = xml::readCoordinate(part, "origin");
      hasProps = true;
    }
    auto motion = part.child("movement");
    if (!motion) motion = part.child("motion");
    if (motion) {
      std::string_view type = motion.attribute("type").as_string("delta");
      if (type == "delta") {
        partProps.motion = xml::readFilename(
          motion, "movement", fileLocator, FileLocator::GeometryList);
        hasProps = true;
      } else if (type == "spin") {
        partProps.motion.emplace<SpinMotion>().ReadXML(motion, facade);
        hasProps = true;
      } else if (type == "flexible" or type == "flex") {
        partProps.motion.emplace<FlexMotion>().ReadXML(
          motion, facade, fileLocator);
        hasProps = true;
      }
    }
    if (hasProps) partProperties.push_back(std::move(partProps));
  }
  for (auto kpt : node.child("keypointlist").children("keypoint")) {
    auto& keypoint = keypoints.emplace_back();
    keypoint.name = xml::requiredAttribute(kpt, "name").as_string();
    keypoint.part = kpt.attribute("part").as_string();
    for (auto pt : kpt.children("point"))
      keypoint.points.push_back(xml::readCoordinate(pt));
    if (keypoint.points.empty())
      DIRSIG_THROW_FMT("all <keypoint> entries must have at least 1 <point>");
  }
}

void TdfFacetObject::ReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  FacetObjectReadXML(node, facade, fileLocator);
  timeStepIndex = node.child("resulttimestepindex").text().as_int(-1);
  useTdfMaterialNames = true;
  // Remap with <materialidremappings> for backwards compatibility
  if (auto remapping = node.child("materialidremappings")) {
    useTdfMaterialNames = false;
    for (auto remap : remapping.children("remap")) {
      auto key = xml::requiredChild(remap, "tdfid").text().as_string();
      auto val = xml::requiredChild(remap, "dirsigid").text().as_string();
      materialRemap[key] = val;
    }
  }
}

void RegularGrid::ReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  mediumType = node.attribute("type").as_string();
  filename =
    xml::readFilename(node, "gridfilename", fileLocator, FileLocator::Any);
  insertPoint = xml::readCoordinate(node, "insertpoint");
  voxelSize[0] = xml::readNumber(node, "voxeldeltax", 1.0);
  voxelSize[1] = xml::readNumber(node, "voxeldeltay", 1.0);
  voxelSize[2] = xml::readNumber(node, "voxeldeltaz", 1.0);
}

void HeightField::ReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {}

void PointSource::ReadXML(
  const xml::Node& node, const geolib::Facade*, const FileLocator&) {
  matId = xml::requiredAttribute(node, "matid").as_string();
  pointingVec = xml::readCoordinate(node, "pointing", {0, 0, -1});
}

void BezierCurveSet::ReadXML(
  const xml::Node& node, const geolib::Facade*, const FileLocator&) {
  matId = xml::requiredChild(node, "matid").text().as_string();
  vertexData.reserve(32);
  {
    std::istringstream iss(
      xml::requiredChild(node, "vertexdata").text().as_string());
    while (!iss.eof()) {
      float x, y, z, r;
      iss >> x;
      iss >> y;
      iss >> z;
      iss >> r;
      if (!iss) break;
      vertexData.emplace_back(x);
      vertexData.emplace_back(y);
      vertexData.emplace_back(z);
      vertexData.emplace_back(r);
    }
  }
  if (std::string_view s = node.child("normaldata").text().as_string();
      !s.empty()) {
    normalData.reserve(32);
    std::istringstream iss(s.data());
    while (!iss.eof()) {
      float x, y, z;
      iss >> x;
      iss >> y;
      iss >> z;
      if (!iss) break;
      normalData.emplace_back(x);
      normalData.emplace_back(y);
      normalData.emplace_back(z);
    }
    if (!normalData.empty() && normalData.size() / 3 != vertexData.size() / 4)
      DIRSIG_THROW_FMT(
        "line {}: if present, bezier curve normals must be "
        "specified per-vertex (read {} normals, expected {})",
        xml::lineno(node),
        normalData.size() / 3,  //
        vertexData.size() / 4);
  }
  if (std::string_view s = node.child("firstindexes").text().as_string();
      !s.empty()) {
    firstIndexes.reserve(32);
    std::istringstream iss(s.data());
    while (!iss.eof()) {
      uint32_t u = 0;
      iss >> u;
      if (!iss) break;
      firstIndexes.emplace_back(u);
    }
  }
}

std::string to_string(Geometry::Type type) {
  switch (type) {
    case Geometry::Type::GdbFacetObject:
      return "GdbFacetObject";
    case Geometry::Type::ObjFacetObject:
      return "ObjFacetObject";
    case Geometry::Type::TdfFacetObject:
      return "TdfFacetObject";
    case Geometry::Type::HdfFacetObject:
      return "HdfFacetObject";
    case Geometry::Type::OdbInclude:
      return "OdbInclude";
    case Geometry::Type::GlistInclude:
      return "GlistInclude";
    case Geometry::Type::Box:
      return "Box";
    case Geometry::Type::CurvedFrustum:
      return "CurvedFrustum";
    case Geometry::Type::Cylinder:
      return "Cylinder";
    case Geometry::Type::Sphere:
      return "Sphere";
    case Geometry::Type::Sinusoid:
      return "Sinusoid";
    case Geometry::Type::SecchiDisk:
      return "SecchiDisk";
    case Geometry::Type::Disk:
      return "Disk";
    case Geometry::Type::RegularGrid:
      return "RegularGrid";
    case Geometry::Type::SurfacePlane:
      return "SurfacePlane";
    case Geometry::Type::CatenaryCurve:
      return "CatenaryCurve";
    case Geometry::Type::HeightField:
      return "HeightField";
    case Geometry::Type::PointSource:
      return "PointSource";
    case Geometry::Type::Unknown:
    default:
      return "Unknown";
  }
}

}  // namespace d5::d4legacy
