#include "dirsig5/d4legacy/geometry/GeometryList.h"

#include <fstream>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/BinaryStream.h"
#include "dirsig5/base/Logger.h"
#include "readers/OdbReader.h"

using namespace std::string_view_literals;

namespace d5::d4legacy {

void Object::ReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator,
  const FileLocator& fileLocatorLocal) {
  lineNumber = xml::lineno(node);
  useLocalSearchPaths = node.attribute("search_paths").as_string() == "local"sv;
  const FileLocator& preferredLocator =
    useLocalSearchPaths ? fileLocatorLocal : fileLocator;
  localMaterialsFilename = node.child("localmaterials").text().as_string();
  tags = xml::readTags(node);
  auto process_children = [&](xml::Node n) {
    for (auto geom : n.children("basegeometry")) {
      geometries.push_back(Geometry::FromXML(geom, facade, preferredLocator));
      weights.push_back(geom.attribute("weight").as_double(1));
    }
    for (auto geom : n.children("basesource")) {
      geometries.push_back(Geometry::FromXML(geom, facade, preferredLocator));
      weights.push_back(geom.attribute("weight").as_double(1));
    }
    for (auto inst : n.children("staticinstance"))
      instances.push_back(Instance::FromXML(inst, facade, preferredLocator));
    for (auto inst : n.children("dynamicinstance"))
      instances.push_back(Instance::FromXML(inst, facade, preferredLocator));
    for (auto inst : n.children("staticinstancebinaryfile")) {
      auto tags{std::vector<std::string>()};
      auto anchor{std::string()};
      bool anchorRotation{false};
      if (auto attr{inst.attribute("tags")}) {
        tags = xml::readTags(inst);
      }
      if (auto attr{inst.attribute("anchor")}) {
        anchor = attr.as_string();
        anchorRotation = inst.attribute("anchorrotation").as_bool();
      }
      auto fileName = preferredLocator.requiredFile(inst.text().as_string());
      auto fileStrm = OpenIfstream(fileName, std::ios::binary);
      auto binStrm = BinaryStream(&fileStrm);
      uint32_t numInstances{binStrm.read<uint32_t>()};
      for (uint32_t i{}; i < numInstances; i++) {
        auto newInst{std::make_shared<AffineInstance>()};
        float affine[12]{};
        binStrm.read(&affine[0], 12);
        std::copy(&affine[0], &affine[0] + 12, newInst->affine.begin());
        newInst->affine[12] = 0;
        newInst->affine[13] = 0;
        newInst->affine[14] = 0;
        newInst->affine[15] = 1;
        newInst->tags = tags;
        newInst->anchor = anchor;
        newInst->anchorRotation = anchorRotation;
        instances.push_back(newInst);
      }
    }
    if (auto popu = n.child("population"); popu) {
      population = Population();
      population->ReadXML(popu);
    }
    for (auto fill : n.children("randomfill"))
      randomFills.emplace_back().ReadXML(fill);
  };
  process_children(node);
  for (auto xinc : node.children("xi:include")) {
    auto fname = fileLocatorLocal.requiredFile(
      xml::requiredAttribute(xinc, "href").as_string());
    process_children(xml::loadFile(fname));
  }
  DIRSIG_THROW_IF_FMT(
    instances.empty() and randomFills.empty(),
    "line {}: object has no instances or random fills, at least 1 is "
    "required",
    lineNumber);
}

void GeometryList::ReadODBFile(
  const std::string& filename, const geolib::Facade*, const FileLocator&) {
  *this = std::move(*LoadOdb(filename));
}

void GeometryList::ReadXMLFile(
  const std::string& filename,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  DIRSIG_DEBUG("Reading GList '{}'...", filename);
  auto document = xml::loadFile(filename);
  auto fileLocatorLocal = FileLocator::MakeRelativeToFile(filename);
  fileLocatorLocal.setSearchCwd(false);
  ReadXML(document.document_element(), facade, fileLocator, fileLocatorLocal);
}

void GeometryList::ReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator,
  const FileLocator& fileLocatorLocal) {
  tags = xml::readTags(node);
  for (auto object : node.children("object"))
    if (object.attribute("enabled").as_bool(true))
      objects.emplace_back().ReadXML(
        object, facade, fileLocator, fileLocatorLocal);
  for (auto inc : node.children("geometrylistinclude")) {
    if (inc.attribute("enabled").as_bool(true)) {
      GeometryList glist;
      std::string filename =
        fileLocatorLocal.locateFile(inc.text().as_string());
      std::string lowerFilename = toLower(filename);
      if (endsWith(lowerFilename, ".glist") or endsWith(lowerFilename, ".xml"))
        glist.ReadXMLFile(filename, facade, fileLocator);
      else if (endsWith(lowerFilename, ".odb"))
        glist.ReadODBFile(filename, facade, fileLocator);
      else
        DIRSIG_THROW_FMT(
          "line {}: unrecognized file extension", xml::lineno(inc));
      append(std::move(glist));
      includes.push_back(filename);
    }
  }
}

}  // namespace d5::d4legacy
