#include "dirsig5/d4legacy/geometry/Instance.h"

#include "dirsig5/base/Error.h"
#include "spdlog/fmt/fmt.h"

using namespace std::string_view_literals;

namespace d5::d4legacy {

std::shared_ptr<Instance> Instance::FromXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  auto dataNode = node;
  std::string nodeName = node.name();
  std::shared_ptr<Instance> instance;
  if (nodeName == "staticinstance") {
    if (node.child("matrix"))
      instance = std::make_shared<AffineInstance>();
    else
      instance = std::make_shared<TsrInstance>();
  } else if (nodeName == "dynamicinstance") {
    if (node.child("keyframemovement")) {
      dataNode = node.child("keyframemovement");
      instance = std::make_shared<DeltaMotionInstance>();
    } else if (node.child("genericmovement")) {
      dataNode = node.child("genericmovement");
      instance = std::make_shared<GenericMotionInstance>();
    } else if (node.child("motion")) {
      dataNode = node.child("motion");
      std::string type = dataNode.attribute("type").as_string();
      if (type == "delta" or type.empty())
        instance = std::make_shared<DeltaMotionInstance>();
      else if (type == "generic")
        instance = std::make_shared<GenericMotionInstance>();
      else if (type == "flexible")
        instance = std::make_shared<FlexMotionInstance>();
      else 
        DIRSIG_THROW_FMT(
          "line {}: unknown motion type '{}'",
            xml::lineno(node), type);
    } else {
      DIRSIG_THROW_FMT(
        "line {}: missing/unknown motion type", xml::lineno(node));
    }
  } else {
    DIRSIG_THROW_FMT( "line {}: unrecognized instance type <{}>",
        xml::lineno(node), nodeName);
  }

  // we should have an instance at this point
  if (instance == nullptr) {
    DIRSIG_THROW_FMT("Unable to read instance begining on line {}!",
        xml::lineno(node));
  }

  instance->ReadXML(node, dataNode, facade, fileLocator);
  return instance;
}

void Instance::ReadXML(
  const xml::Node& node,
  const xml::Node& dataNode,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  // This is catastrophic for millions of instances!
  /* lineNumber = xml::lineno(node); */
  lineNumber = 0;
  name = node.attribute("name").as_string();
  tags = xml::readTags(node);
  origToNewMaterialId.clear();
  for (auto subnode : node.children("assign")) {
    origToNewMaterialId[subnode.text().as_string()] =
      subnode.attribute("id").as_string();
  }
  timeWindow = node.attribute("timewindow").as_string();
  controlTableFilename = node.attribute("control").as_string();
  if (auto attr = node.attribute("geometryindex"); attr)
    geometryPopulationIndex = attr.as_int();
  else
    geometryPopulationIndex = std::nullopt;
  FinishReadXML(dataNode, facade, fileLocator);
}

void TsrInstance::ReallyFinishReadXML(
  const xml::Node& node, const geolib::Facade* facade) {
  for (auto subnode : node.children()) {
    if (subnode.name() == "translation"sv) {
      auto e = subnode.first_child();
      if (e.name() == "point"sv) {
        xTranslate = xml::readNumber(e, "x");
        yTranslate = xml::readNumber(e, "y");
        zTranslate = xml::readNumber(e, "z");
      } else {
        auto ip = xml::readLocation(e, facade).getLocalCartesian();
        xTranslate = ip[0];
        yTranslate = ip[1];
        zTranslate = ip[2];
      }
    } else if (subnode.name() == "rotation"sv) {
      auto e = subnode.first_child();
      xRotate = xml::readNumber(e, "x");
      yRotate = xml::readNumber(e, "y");
      zRotate = xml::readNumber(e, "z");
      if (subnode.attribute("units").as_string() == "degrees"sv) {
        xRotate *= M_PI / 180.0;
        yRotate *= M_PI / 180.0;
        zRotate *= M_PI / 180.0;
      }
      std::string order = subnode.attribute("rotationorder").as_string();
      if (order.empty()) order = "zyx";
      rotationOrder = ParseRotationOrder(order);
    } else if (subnode.name() == "scale"sv) {
      auto e = subnode.first_child();
      xScale = xml::readNumber(e, "x");
      yScale = xml::readNumber(e, "y");
      zScale = xml::readNumber(e, "z");
    } else if (subnode.name() == "matrix"sv)
      DIRSIG_THROW_FMT("cannot mix matrix and translation/scale/rotation in a "
                       "staticinstance");
    else if (subnode.name() != "assign"sv)
      DIRSIG_THROW_FMT(
        "unrecognized tag '{}' in staticinstance", subnode.name());
  }
}

void AffineInstance::ReallyFinishReadXML(
  const xml::Node& node, const geolib::Facade*) {
  for (auto subnode : node.children()) {
    if (subnode.name() == "matrix"sv) {
      std::stringstream ss(subnode.text().as_string());
      char dummy;
      for (int k = 0; k < 16; k++) {
        ss >> affine[k];
        if (k != 15) {
          ss >> dummy;
          if (dummy != ',')
            DIRSIG_THROW_FMT(
              "invalid matrix string '{}'", subnode.text().as_string());
        }
      }
    } else if (
      subnode.name() == "scale"sv || subnode.name() == "rotation"sv ||
      subnode.name() == "translation"sv)
      DIRSIG_THROW_FMT("cannot mix matrix and translation/scale/rotation in a "
                       "staticinstance");
    else if (subnode.name() != "assign"sv)
      DIRSIG_THROW_FMT(
        "unrecognized tag '{}' in staticinstance", subnode.name());
  }
}

void DeltaMotionInstance::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  filename = xml::readFilename(node, "filename", fileLocator);
}

void GenericMotionInstance::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  filename = xml::readFilename(node, "filename", fileLocator);
  if (auto options = node.child("options"); options) {
    startHidden = options.attribute("starthidden").as_bool();
    endHidden = options.attribute("endhidden").as_bool();
  }
}

void FlexMotionInstance::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  flexMotion.ReadXML(node, facade, fileLocator);
}

void DecalInstance::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  // TODO
}

}  // namespace d5::d4legacy
