#include "dirsig5/d4legacy/geometry/Population.h"

#include <fstream>
#include <ostream>
#include <sstream>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/d4legacy/xml/xml.h"
#include "spdlog/fmt/fmt.h"

using namespace std::string_view_literals;

namespace d5::d4legacy {

auto RandomFill::Anchor::ReadAnchorListFile(const std::string& filename)
  -> std::vector<AnchorEntry> {
  auto ifs = OpenIfstream(filename);

  std::vector<RandomFill::Anchor::AnchorEntry> entries;
  std::string line;
  for (size_t lineNo = 1; std::getline(ifs, line); lineNo++) {
    if (line.empty()) continue;
    auto tokens = TokenList(line, " ");
    if (tokens.size() < 4) {
      DIRSIG_THROW_FMT(
        "line {}: each anchor list entry must have at least 4 values", lineNo);
    }
    auto& entry = entries.emplace_back();
    entry.x = ParseNumber<double>(tokens[0]);
    entry.y = ParseNumber<double>(tokens[1]);
    entry.rx = ParseNumber<double>(tokens[2]);
    entry.ry = ParseNumber<double>(tokens[3]);
  }

  return entries;
}

auto RandomFill::PolyFill::ReadPolyFile(const std::string& filename)
  -> std::vector<Polygon> {
  auto ifs = OpenIfstream(filename);

  Polygon polygon;
  std::vector<Polygon> polygons;
  size_t lineNo = 0;
  auto addPolygon = [&]() {
    if (!polygon.points.empty()) {
      if (polygon.points.size() < 4)
        DIRSIG_THROW_FMT(
          "line {}: each polygon must have at least 4 points, where "
          "the last is a duplicate of the first",
          lineNo);
      // Compute center and bounds.
      polygon.centerPoint = {};
      polygon.boundMin = polygon.points.front();
      polygon.boundMax = polygon.points.front();
      for (size_t pointInd = 1; pointInd < polygon.points.size(); pointInd++) {
        const auto& point = polygon.points[pointInd];
        polygon.centerPoint[0] += point[0];
        polygon.centerPoint[1] += point[1];
        polygon.boundMin[0] = std::min(polygon.boundMin[0], point[0]);
        polygon.boundMin[1] = std::min(polygon.boundMin[1], point[1]);
        polygon.boundMax[0] = std::max(polygon.boundMax[0], point[0]);
        polygon.boundMax[1] = std::max(polygon.boundMax[1], point[1]);
      }
      polygon.centerPoint[0] /= polygon.points.size() - 1;
      polygon.centerPoint[1] /= polygon.points.size() - 1;
      polygons.push_back(polygon);
    }
    // Clear.
    polygon.points.clear();
  };
  std::string line;
  for (lineNo = 1; std::getline(ifs, line); lineNo++) {
    if (line.empty()) {
      addPolygon();
      continue;
    }
    auto tokens = TokenList(line, " ");
    if (tokens.size() != 2)
      DIRSIG_THROW_FMT(
        "line {}: each polygon point must have 2 values", lineNo);
    polygon.points.emplace_back();
    polygon.points.back()[0] = ParseNumber<double>(tokens[0]);
    polygon.points.back()[1] = ParseNumber<double>(tokens[1]);
  }
  addPolygon();

  return polygons;
}

bool RandomFill::PolyFill::Polygon::contains(
  const std::array<double, 2>& point) const {
  // Test triangles.
  for (size_t pointInd = 0; pointInd + 1 < points.size(); pointInd++) {
    const auto& v0 = centerPoint;
    const auto& v1 = points[pointInd];
    const auto& v2 = points[pointInd + 1];
    auto sign = [&](const auto& p1, const auto& p2) {
      const auto& p0 = point;
      return (p0[0] - p2[0]) * (p1[1] - p2[1]) -
             (p1[0] - p2[0]) * (p0[1] - p2[1]);
    };
    double d1 = sign(v0, v1);
    double d2 = sign(v1, v2);
    double d3 = sign(v2, v0);
    bool hasNeg = (d1 < 0) || (d2 < 0) || (d3 < 0);
    bool hasPos = (d1 > 0) || (d2 > 0) || (d3 > 0);
    if (!(hasNeg && hasPos)) {
      return true;  // Triangle contains point, so polygon contains point.
    }
  }
  return false;
}

double RandomFill::PolyFill::Polygon::calculateArea() const {
  // Sum areas of triangles.
  double area = 0;
  for (size_t pointInd = 0; pointInd + 1 < points.size(); pointInd++) {
    const auto& v0 = centerPoint;
    const auto& v1 = points[pointInd];
    const auto& v2 = points[pointInd + 1];
    area += 0.5 * std::abs(
                    v0[0] * (v1[1] - v2[1]) + v1[0] * (v2[1] - v0[1]) +
                    v2[0] * (v0[1] - v1[1]));
  }
  return area;
}

template <typename T>
static std::shared_ptr<T> ReadStaticInstance(const xml::Node& node) {
  auto inst = std::make_shared<T>();
  inst->ReadXML(node);
  return inst;
}

std::vector<std::shared_ptr<Instance>> RandomFill::ReadSavefile(
  const std::string& filename) {
  std::vector<std::shared_ptr<Instance>> instances;
  // We have to read the contents without using the XML
  // utility helper because we need to wrap the raw data
  // with a root element!
  std::string filecontent;
  {
    auto ifs = OpenIfstream(filename);
    filecontent = "<staticinstances>\n";
    filecontent += std::string(
      std::istreambuf_iterator<char>(ifs), std::istreambuf_iterator<char>());
    filecontent += "\n</staticinstances>";
  }
  auto doc = xml::loadString(filecontent);
  auto root = doc.document_element();
  geolib::Facade defaultFacade(0, 0, 0);
  for (auto node : root.children("staticinstance"))
    instances.emplace_back(Instance::FromXML(node, &defaultFacade));
  return instances;
}

void RandomFill::WriteSavefile(
  const std::string& filename,
  const std::vector<std::array<float, 12>>& instances,
  const std::vector<size_t>& geometryIndexes) {
  auto indexItr = geometryIndexes.begin();
  bool indexItrFlag = instances.size() == geometryIndexes.size();
  auto ofs = OpenOfstream(filename);
  for (const auto& instance : instances) {
    ofs << "<staticinstance";
    if (indexItrFlag) ofs << " geometryindex=\"" << *indexItr++ << "\"";
    ofs << "><matrix>";
    for (const auto& value : instance) ofs << value << ',';
    ofs << "0,0,0,1"
           "</matrix>"
           "</staticinstance>\n";
  }
}

void Population::ReadXML(const xml::Node& node) {
  *this = {};
  for (auto matvar : node.children("materialvariant")) {
    auto& var = materialVariants.emplace_back();
    var.id = xml::requiredChild(matvar, "variantid").text().as_string();
    for (auto matid : matvar.child("distribution").children("matid"))
      var.distribution.push_back(
        {matid.attribute("weight").as_double(1.0), matid.text().as_string()});
    if (auto subnode = matvar.child("seed"))
      var.seed = subnode.text().as_llong(0);
  }
}

void RandomFill::Box::ReadXML(const xml::Node& node) {
  *this = {};
  lowerExtent = xml::readCoordinate(node, "lowerextent");
  upperExtent = xml::readCoordinate(node, "upperextent");
  count = xml::requiredChild(node, "count").text().as_llong();
}

void RandomFill::Anchor::ReadXML(const xml::Node& node) {
  *this = {};
  filename = xml::requiredChild(node, "filename").text().as_string();
  fillFraction =
    xml::requiredChild(node, "fillpercent").text().as_double() / 100;
  reverseFraction =
    xml::requiredChild(node, "reversepercent").text().as_double() / 100;
  anchor = xml::requiredChild(node, "anchor").text().as_string();
  posDev = xml::readCoordinate(node, "posdev");
  rotDevRad = xml::readCoordinate(node, "rotdev");
  if (node.child("rotdev").attribute("units").as_string() == "degrees"sv)
    rotDevRad *= PI / 180;
  minScale = node.child("minscale").text().as_double(1.0);
  maxScale = node.child("maxscale").text().as_double(1.0);
  if (auto subnode = node.child("seed")) seed = subnode.text().as_llong(0);
}

void RandomFill::DensityMap::ReadXML(const xml::Node& node) {
  *this = {};
  if (auto subnode = node.child("seed")) seed = subnode.text().as_llong(0);
  anchor = xml::requiredChild(node, "anchor").text().as_string();
  count = xml::requiredChild(node, "count").text().as_llong();
  minDist = xml::requiredChild(node, "mindist").text().as_double();
  maxDist = xml::requiredChild(node, "maxdist").text().as_double();
  minScale = node.child("minscale").text().as_double(1.0);
  maxScale = node.child("maxscale").text().as_double(1.0);
  materialId = xml::requiredChild(node, "matid").text().as_string();
  for (auto subnode : node.child("morematids").children("matid"))
    moreMaterialIds.push_back(subnode.text().as_string());
  randomOrientation = node.child("randomorientation").text().as_bool(false);
}

void RandomFill::PolyFill::ReadXML(const xml::Node& node) {
  *this = {};
  polyFilename = xml::requiredChild(node, "polyfile").text().as_string();
  if (auto subnode = node.child("seed")) seed = subnode.text().as_llong(0);
  minCount = xml::requiredChild(node, "mincount").text().as_llong();
  maxCount = xml::requiredChild(node, "maxcount").text().as_llong();
  minScale = node.child("minscale").text().as_double(1.0);
  maxScale = node.child("maxscale").text().as_double(1.0);
  materialId = xml::requiredChild(node, "matid").text().as_string();
  randomOrientation = node.child("randomorientation").text().as_bool(false);
}

void RandomFill::ReadXML(const xml::Node& node) {
  if (xml::size(node.children()) != 1)
    DIRSIG_THROW_FMT(
      "line {}: random fill node must contain exactly 1 child",
      xml::lineno(node));
  saveFilename = node.attribute("savefile").as_string();
  loadInstances = node.attribute("loadinstances").as_bool();
  if (auto subnode = node.child("box"); subnode)
    fill.emplace<Box>().ReadXML(subnode);
  else if (auto subnode = node.child("anchorlist"); subnode)
    fill.emplace<Anchor>().ReadXML(subnode);
  else if (auto subnode = node.child("densitymap"); subnode)
    fill.emplace<DensityMap>().ReadXML(subnode);
  else if (auto subnode = node.child("polyfill"); subnode)
    fill.emplace<PolyFill>().ReadXML(subnode);
  else
    DIRSIG_THROW_FMT(
      "line {}: unrecognized random fill type <{}>", xml::lineno(node),
      node.first_child().name());
}

}  // namespace d5::d4legacy
