#include "dirsig5/d4legacy/geometry/Primitives.h"

using namespace std::string_view_literals;

namespace d5::d4legacy {

void Box::FinishReadXML(const xml::Node& node) {
  lowerExtent = xml::readCoordinate(node, "lowerextent", {-0.5, -0.5, -0.5});
  upperExtent = xml::readCoordinate(node, "upperextent", {+0.5, +0.5, +0.5});
}

void Box::FinishReadTag(const TagFileSection& section) {
  section.optionalVariable("LOWER_EXTENT", 0, lowerExtent);
  section.optionalVariable("UPPER_EXTENT", 0, upperExtent);
}

void CatenaryCurve::FinishReadXML(const xml::Node& node) {
  thickness = xml::readNumber(node, "thickness", 0.01);
  sag = xml::readNumber(node, "sag", 20);
  pointA = xml::readCoordinate(node, "a", {-15, 0, 30});
  pointB = xml::readCoordinate(node, "b", {+15, 0, 30});
}

void CatenaryCurve::FinishReadTag(const TagFileSection& section) {
  section.optionalVariable("THICKNESS", 0, thickness).Or(0.01);
  section.optionalVariable("SAG", 0, sag).Or(20);
  section.optionalVariable("POINT_A", 0, pointA).Or({-15, 0, 30});
  section.optionalVariable("POINT_B", 0, pointA).Or({+15, 0, 30});
}

void CurvedFrustum::FinishReadXML(const xml::Node& node) {
  // TODO rotate?
  auto subelem_bottom = xml::requiredChild(node, "bottom");
  auto subelem_top = xml::requiredChild(node, "top");
  topRadius = xml::readNumber(subelem_top, "radius", 0.0);
  bottomRadius = xml::readNumber(subelem_bottom, "radius", 1.0);
  bottomXWidth = xml::readNumber(subelem_bottom, "xwidth", 10.0);
  bottomYWidth = xml::readNumber(subelem_bottom, "ywidth", 10.0);
  center = xml::readCoordinate(node, "center", {0.0, 0.0, 0.0});
  height = xml::readNumber(node, "height", 1.0);
  rampLength = xml::readNumber(node, "ramplength", 0.0);
}

void CurvedFrustum::FinishReadTag(const TagFileSection& section) {
  section.optionalVariable("TOP_RADIUS", 0, topRadius).Or(0);
  section.optionalVariable("BOTTOM_RADIUS", 0, bottomRadius).Or(1);
  section.optionalVariable("BOTTOM_WIDTH_X", 0, bottomXWidth).Or(10);
  section.optionalVariable("BOTTOM_WIDTH_Y", 0, bottomYWidth).Or(10);
  section.optionalVariable("ROTATE", 0, zRotation).Or(0);
  section.optionalVariable("HEIGHT", 0, height).Or(1);
  section.optionalVariable("CENTER", 0, center);
  section.optionalVariable("RAMP", 0, rampLength).Or(0);
}

void Cylinder::FinishReadXML(const xml::Node& node) {
  radius = xml::readNumber(node, "radius", 1.0);
  pointA = xml::readCoordinate(node, "point_a", {0.0, 0.0, -0.5});
  pointB = xml::readCoordinate(node, "point_b", {0.0, 0.0, 0.5});
  capA = node.attribute("cap_a").as_string() != "false"sv;
  capB = node.attribute("cap_b").as_string() != "false"sv;
}

void Cylinder::FinishReadTag(const TagFileSection& section) {
  section.optionalVariable("RADIUS", 0, radius).Or(1);
  section.optionalVariable("POINT_A", 0, pointA).Or({0.0, 0.0, -0.5});
  section.optionalVariable("POINT_B", 0, pointB).Or({0.0, 0.0, +0.5});
  section.optionalVariable("CAP_A", 0, capA).Or(true);
  section.optionalVariable("CAP_B", 0, capB).Or(true);
}

void Sphere::FinishReadXML(const xml::Node& node) {
  radius = xml::readNumber(node, "radius", 1.0);
  center = xml::readCoordinate(node, "center", {0.0, 0.0, 0.0});
}

void Sphere::FinishReadTag(const TagFileSection& section) {
  section.optionalVariable("RADIUS", 0, radius).Or(1);
  section.optionalVariable("CENTER", 0, center);
}

void Sinusoid::FinishReadXML(const xml::Node& node) {
  lowerExtent = xml::readCoordinate(node, "lowerextent", {-0.5, -0.5, -0.5});
  upperExtent = xml::readCoordinate(node, "upperextent", {+0.5, +0.5, +0.5});
  magnitude = xml::readNumber(node, "magnitude", 1.0);
  period = xml::readNumber(node, "period", 0.5);
  orientation = xml::readCoordinate(node, "orientation", {1.0, 0.0, 0.0});
  phaseShift = xml::readNumber(node, "phaseshift", 0.0);
}

void Sinusoid::FinishReadTag(const TagFileSection& section) {
  std::array<double, 6> boundingBox = {};
  section.requiredVariable("BOUNDING_BOX", 0, boundingBox);
  lowerExtent = {boundingBox[0], boundingBox[1], boundingBox[2]};
  upperExtent = {boundingBox[3], boundingBox[4], boundingBox[5]};
  section.optionalVariable("MAGNITUDE", 0, magnitude).Or(1);
  section.optionalVariable("PERIOD", 0, period).Or(0.5);
  section.optionalVariable("ORIENTATION", 0, orientation).Or({1, 0, 0});
  section.optionalVariable("PHASE_SHIFT", 0, phaseShift).Or(0);
}

void SecchiDisk::FinishReadXML(const xml::Node& node) {
  radius = xml::readNumber(node, "radius", 1.0);
  center = xml::readCoordinate(node, "center", {0.0, 0.0, 0.0});
  normal = xml::readCoordinate(node, "normal", {0.0, 0.0, 1.0});
}

void SecchiDisk::FinishReadTag(const TagFileSection& section) {
  section.optionalVariable("RADIUS", 0, radius).Or(1);
  section.optionalVariable("CENTER", 0, center).Or({0, 0, 0});
  section.optionalVariable("NORMAL", 0, normal).Or({0, 0, 1});
}

void Disk::FinishReadXML(const xml::Node& node) {
  radius = xml::readNumber(node, "radius", 1.0);
  center = xml::readCoordinate(node, "center", {0.0, 0.0, 0.0});
  normal = xml::readCoordinate(node, "normal", {0.0, 0.0, 1.0});
}

void Disk::FinishReadTag(const TagFileSection& section) {
  section.optionalVariable("RADIUS", 0, radius).Or(1);
  section.optionalVariable("CENTER", 0, center).Or({0, 0, 0});
  section.optionalVariable("NORMAL", 0, normal).Or({0, 0, 1});
}

void SurfacePlane::FinishReadXML(const xml::Node& node) {
  anchor = xml::readCoordinate(node, "anchor");
  if (auto subelem = node.child("checkers"); subelem) {
    checkers = true;
    checkersMaterialId = subelem.attribute("matid").value();
    checkerWidth = subelem.attribute("width").as_double(1);
    for (auto paintE : subelem.children("paint")) {
      auto& o = checkerOverrides.emplace_back();
      o.xIndex = paintE.attribute("xindex").as_int();
      o.yIndex = paintE.attribute("yindex").as_int();
      o.materialId = paintE.attribute("matid").value();
    }
  }
}

void SurfacePlane::FinishReadTag(const TagFileSection& section) {
  section.optionalVariable("ANCHOR", 0, anchor);
  if (const auto* checks = section.subsection("ADD_CHECKS"); checks) {
    checkers = true;
    checks->requiredVariable("MATERIAL_ID", 0, checkersMaterialId);
    checks->optionalVariable("WIDTH", 0, checkerWidth).Or(1);
  }
}

}  // namespace d5::d4legacy
