#include "AssimpReader.h"

#if WITH_ASSIMP
#include "assimp/Importer.hpp"
#include "assimp/postprocess.h"
#include "assimp/scene.h"
#endif

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Logger.h"
#include "dirsig5/base/ScopeGuard.h"

namespace d5::d4legacy {

#if WITH_ASSIMP
static void LoadAssimpMesh(
  const aiScene& scene, const aiMesh& mesh, aiMatrix4x4 xf, FacetObject& fo) {
  auto matName = scene.mMaterials[mesh.mMaterialIndex]->GetName();
  const int part = fo.addPart(std::to_string(fo.parts().size()));
  const int firstVert = fo.numVertices();
  const int firstTexCoord = fo.numTextureCoords();
  const int firstNormal = fo.numVertexNormals();
  const int mat = fo.addMaterialLabel(matName.C_Str());
  for (unsigned int j = 0; j < mesh.mNumVertices; j++) {
    aiVector3D v = mesh.mVertices[j];
    fo.addVertex(
      xf[0][0] * v.x + xf[0][1] * v.y + xf[0][2] * v.z + xf[0][3],  //
      xf[1][0] * v.x + xf[1][1] * v.y + xf[1][2] * v.z + xf[1][3],  //
      xf[2][0] * v.x + xf[2][1] * v.y + xf[2][2] * v.z + xf[2][3]);
    if (mesh.mTextureCoords[0])
      fo.addTextureCoord(
        mesh.mTextureCoords[0][j].x, mesh.mTextureCoords[0][j].y);
    if (mesh.mNormals) {
      v = mesh.mNormals[j];
      fo.addVertexNormal(
        xf[0][0] * v.x + xf[0][1] * v.y + xf[0][2] * v.z,  //
        xf[1][0] * v.x + xf[1][1] * v.y + xf[1][2] * v.z,  //
        xf[2][0] * v.x + xf[2][1] * v.y + xf[2][2] * v.z);
    }
  }
  for (unsigned int j = 0; j < mesh.mNumFaces; j++) {
    auto face = fo.addFacetToPart(
      part,                                    //
      firstVert + mesh.mFaces[j].mIndices[0],  //
      firstVert + mesh.mFaces[j].mIndices[1],  //
      firstVert + mesh.mFaces[j].mIndices[2], mat);
    if (mesh.mTextureCoords[0])
      fo.setTextureFacet(
        face, firstTexCoord + mesh.mFaces[j].mIndices[0],
        firstTexCoord + mesh.mFaces[j].mIndices[1],
        firstTexCoord + mesh.mFaces[j].mIndices[2]);
    if (mesh.mNormals)
      fo.setVertexNormalFacet(
        face, firstNormal + mesh.mFaces[j].mIndices[0],
        firstNormal + mesh.mFaces[j].mIndices[1],
        firstNormal + mesh.mFaces[j].mIndices[2]);
  }
}

static void LoadAssimpNode(
  const aiScene& scene, const aiNode& node, aiMatrix4x4 xf, FacetObject& fo) {
  xf = xf * node.mTransformation;
  for (unsigned int i = 0; i < node.mNumMeshes; i++)
    LoadAssimpMesh(scene, *scene.mMeshes[node.mMeshes[i]], xf, fo);
  for (unsigned int i = 0; i < node.mNumChildren; i++)
    LoadAssimpNode(scene, *node.mChildren[i], xf, fo);
}
#endif  // #if WITH_ASSIMP

FacetObject ReadAssimpFile(const FacetObjectReadRequest& request) {
#if WITH_ASSIMP
  FacetObject fo{};
  Assimp::Importer importer{};
  auto scene = importer.ReadFile(
    request.filename(), aiProcessPreset_TargetRealtime_MaxQuality &
                          ~aiProcess_RemoveRedundantMaterials);
  DIRSIG_THROW_IF_FMT(!scene, "ASSIMP failed to read: {}", request.filename());
  aiMatrix4x4 xf = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};
  LoadAssimpNode(*scene, *scene->mRootNode, xf, fo);
  return fo;
#else
  return {};
#endif
}

}  // namespace d5::d4legacy
