#include "GdbReader.h"

#include <array>
#include <cstring>
#include <fstream>
#include <unordered_set>

#include "spdlog/fmt/fmt.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"

namespace d5::d4legacy {

namespace {

bool readLine(std::istream& in, std::string& str, long& lineNumber) {
  // read the line, trim off ^M, increment line number
  if (!std::getline(in, str)) return false;
  if (!str.empty() && str.back() == '\r') str.resize(str.size() - 1);
  lineNumber++;
  return true;
}

void skipLine(std::istream& in, std::string& str, long& lineNumber) {
  std::getline(in, str);
  lineNumber++;
}

void skipNLines(
    std::istream& in, size_t n, std::string& str, long& lineNumber) {
  for (size_t i = 0; i < n; i++) std::getline(in, str);
  lineNumber += n;
}
}

bool ParseTriplet(std::string& line, double scale, float* t) {
  char* s = &line[0];
  char* end = s + line.size();
  char* p;
  t[0] = static_cast<float>(scale * strtod(s, &p));
  if (p == s || p == end) return false;
  s = p;
  t[1] = static_cast<float>(scale * strtod(s, &p));
  if (p == s || p == end) return false;
  s = p;
  t[2] = static_cast<float>(scale * strtod(s, &p));
  return p != s;
}

bool ParseTriplet(std::string& line, float* t) {
  char* s = &line[0];
  char* end = s + line.size();
  char* p;
  t[0] = static_cast<float>(strtod(s, &p));
  if (p == s || p == end) return false;
  s = p;
  t[1] = static_cast<float>(strtod(s, &p));
  if (p == s || p == end) return false;
  s = p;
  t[2] = static_cast<float>(strtod(s, &p));
  return p != s;
}

void extractPart(
    uint32_t partIdx,
    FacetObject& obj,
    const FacetObjectReadRequest& request,
    std::ifstream& ifs,
    long& lineNumber,
    std::string& lineStr) {
  auto err = [&](const auto& what) {
    return DIRSIG_MAKE_ERROR(fmt::format("line {}: {}", lineNumber, what));
  };
  auto nextLine = [&]() { return readLine(ifs, lineStr, lineNumber); };

  auto unitScale = toMeters(request.units());

  static constexpr int64_t cacheSize = 1000;

  // vertex/normal vars
  std::array<float, 3> vertex;
  std::array<float, 3> tmpNormal;

  skipLine(ifs, lineStr, lineNumber);  // part ID
  nextLine();                          // should be the start of a facet

  while (lineStr == "FACE") {
    // facet name, facet ID string, material name
    skipNLines(ifs, 3, lineStr, lineNumber);
    if (!nextLine()) throw err("invalid facet material ID");
    std::string material;
    std::swap(lineStr, material);

    skipLine(ifs, lineStr, lineNumber);  // facet name
    if (!nextLine()) throw err("invalid facet temperature");
    auto temperature = static_cast<float>(std::atof(lineStr.c_str()));

    //
    // DIRSIG3 allowed either 0 or -1 Celcius as a flag to enable
    // THERM calculations for the facet.  In DIRSIG4, we only use
    // -1 Kelvin for this purpose.  So here we set values of 0 to
    // -1 to eliminate the old 0.0 case, and add the Kelvin offset
    // to all other (override) temperatures.
    //
    if (temperature == 0.0) temperature = -1.0;
    if (temperature != -1.0) temperature += 273.15;

    if (!nextLine()) throw err("invalid facet thickness");
    auto thickness = static_cast<float>(atof(lineStr.c_str()));

    // self-generated power, shape factor, 3 NULLs
    skipNLines(ifs, 5, lineStr, lineNumber);

    // get the number of vertices
    if (!nextLine()) throw err("invalid number of vertices");
    int vertexCount = atoi(lineStr.c_str());
    if (vertexCount != 1 && vertexCount != 3 && vertexCount != 4) {
      throw err("vertex count must be either 1, 3 or 4");
    }

    // load all of the vertices into the object
    std::array<size_t, 4> facetVertexIndices;
    for (int vi = 0; vi < vertexCount; vi++) {
      nextLine();

      // parse current vertex
      if (!ParseTriplet(lineStr, unitScale, vertex.data())) {
        throw err("invalid facet vertex");
      }

      // We are going to look through all of the recent vertices to try to
      // find this one
      auto numVertices = static_cast<int64_t>(obj.numVertices());
      bool found = false;
      for (int64_t i = 1; i <= cacheSize; i++) {
        auto lookup = numVertices - i;
        if (lookup < 0) break;

        const auto* v = obj.vertex(lookup);
        if (fabs(vertex[0] - v[0]) <= 1e-8 && fabs(vertex[1] - v[1]) <= 1e-8 &&
            fabs(vertex[2] - v[2]) <= 1e-8) {
          facetVertexIndices[vi] = lookup;
          found = true;
          break;
        }
      }

      if (!found) {
        facetVertexIndices[vi] = obj.addVertex(vertex[0], vertex[1], vertex[2]);
      }
    }

    nextLine();  // normal vector

    // parse current normal
    if (!ParseTriplet(lineStr, tmpNormal.data())) {
      throw err("invalid facet normal");
    }

    if (tmpNormal[0] == 0 && tmpNormal[1] == 0 && tmpNormal[2] == 0) {
      vertexCount = 0;
    }

    // slope, azimuth, dummy
    skipNLines(ifs, 3, lineStr, lineNumber);

    bool isPointSource = vertexCount == 1;
    auto matId = isPointSource ? obj.addPointSourceMaterialLabel(material)
                               : obj.addMaterialLabel(material);

    // add a triangle if the vertex count is 3 or 4
    if (vertexCount > 1) {
      auto facetIdx = obj.addFacetToPart(
          partIdx, facetVertexIndices[0], facetVertexIndices[1],
          facetVertexIndices[2], matId);
      obj.setFacetThicknessM(facetIdx, thickness / 100.0f);
      obj.setFacetTemperatureK(facetIdx, temperature);
    }

    // add another if it is 4
    if (vertexCount == 4) {
      auto facetIdx = obj.addFacetToPart(
          partIdx, facetVertexIndices[2], facetVertexIndices[3],
          facetVertexIndices[0], matId);
      obj.setFacetThicknessM(facetIdx, thickness / 100.0f);
      obj.setFacetTemperatureK(facetIdx, temperature);
    }

    // secondary point source
    if (vertexCount == 1) {
      auto& ptSrc = obj.addPointSource();
      ptSrc.vertexIndex = facetVertexIndices[0];
      ptSrc.materialIndex = matId;
      ptSrc.direction = tmpNormal;
      ptSrc.partIndex = partIdx;
    }

    nextLine();
  }
}

FacetObject ReadGdbFile(const FacetObjectReadRequest& request) {
  FacetObject obj;

  auto ifs = d5::OpenIfstream(request.filename());

  // create a string for the current line, line number counter, etc.
  std::string lineStr;
  long lineNumber = 0;

  auto err = [&](const auto& what) {
    return DIRSIG_MAKE_ERROR(fmt::format("line {}: {}", lineNumber, what));
  };
  auto nextLine = [&]() { readLine(ifs, lineStr, lineNumber); };

  // read the first line
  nextLine();

  // loop until we hit the end of the file
  while (lineStr != "END") {
    // this line should be OBJECT
    if (lineStr != "OBJECT") {
      throw err(fmt::format(
          R"(expected to see "OBJECT" line, found ("{}"))", lineStr));
    }

    // object name/ID
    skipNLines(ifs, 2, lineStr, lineNumber);
    nextLine();  // PART or END
    while (lineStr == "PART") {
      nextLine();  // part name
      auto partName = std::move(lineStr);
      extractPart(
          obj.addPart(partName), obj, request, ifs, lineNumber, lineStr);
    }
  }

  return obj;
}
}
