#include "ObjReader.h"

#include <array>
#include <cstdint>
#include <fstream>
#include <limits>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/Logger.h"

#if defined(_WIN32)
#define strtok_r strtok_s
#endif

constexpr auto u32Max = std::numeric_limits<uint32_t>::max();

namespace d5::d4legacy {

/// Helper function for reading a line
static bool readLine(std::istream& in, std::string& str, long& lineNumber) {
  // try to read line, string carraige returns, increment line number
  if (!std::getline(in, str)) return false;
  if (!str.empty() && str.back() == '\r') str.resize(str.size() - 1);
  lineNumber++;
  return true;
}

template <typename T, size_t N>
bool parseArray(std::array<T, N>& a, char** rest) {
  char* pEnd;
  for (auto& c : a) {
    auto* vTokTmp = strtok_r(nullptr, " ", rest);
    if (vTokTmp == nullptr) return false;
    if constexpr (std::is_same_v<T, float>) {
      c = std::strtof(vTokTmp, &pEnd);
    } else {
      c = std::strtod(vTokTmp, &pEnd);
    }
  }

  return true;
}

FacetObject ReadObjFile(const FacetObjectReadRequest& request) {
  FacetObject obj;

  // temporary buffers
  std::array<int64_t, 100> vIndex, vtIndex, vnIndex;

  // try to directly open the file for reading
  auto ifs = d5::OpenIfstream(request.filename());

  auto unitScale = static_cast<float>(toMeters(request.units()));

  size_t vCount = 0;
  size_t vtCount = 0;
  size_t vnCount = 0;

  // loop vars
  std::string tmpStr, tmpElem, line;
  long lineNumber = 0;

  auto nextLine = [&](std::string& s) { return readLine(ifs, s, lineNumber); };

  // pre-read the file for sizing arrays (faster than dynamically sizing)
  while (nextLine(tmpStr)) {
    if (tmpStr.size() < 2) continue;
    if (tmpStr[0] == 'v') {
      char c2 = tmpStr[1];
      if (c2 == ' ') {
        vCount++;
      } else if (c2 == 't') {
        vtCount++;
      } else if (c2 == 'n') {
        vnCount++;
      }
    }
  }

  // pre-allocate arrays
  obj.reserveVertices(vCount);
  obj.reserveTextureCoords(vtCount);
  obj.reserveVertexNormals(vnCount);

  // clear eof and rewind to the start of the stream
  ifs.clear();
  ifs.seekg(0);
  lineNumber = 0;

  // additional tags
  auto matId = u32Max;
  auto curPart = obj.addPart(std::string());

  // re-usable string stream
  std::stringstream matIdStream;

  // re-usable in-loop string vars
  const char* c_keyword;
  const char* groupName;
  std::string usemtlStream;
  std::string matIdStr;

  // re-usable arrays of facet data
  std::array<int64_t, 4> indices;
  std::array<char*, 4> vertStrings;

  // vectors that maintain the order of vertex normals
  std::vector<uint32_t> orderedNormals;

  // vars for tokenization
  char* tok;
  constexpr size_t maxLine = 400;
  std::array<char, maxLine + 1> c_line;

  auto err = [&](const auto& what) {
    return DIRSIG_MAKE_ERROR(fmt::format("line {}: {}", lineNumber, what));
  };

  // read the whole file again
  while (nextLine(tmpStr)) {
    if (tmpStr.empty()) continue;

    // handle line continuation
    if (tmpStr.back() == '\\') {
      while (tmpStr.back() == '\\') {
        line += tmpStr.substr(0, tmpStr.size() - 2);
        nextLine(tmpStr);
      }
      line += tmpStr.substr(0, tmpStr.size() - 2);
    } else {
      line = std::move(tmpStr);
    }

    strncpy(c_line.data(), line.c_str(), maxLine);
    auto* rest = c_line.data();

    // extract out the keyword (first string on the line)
    c_keyword = strtok_r(c_line.data(), " ", &rest);

    // check for blank lines
    if (c_keyword == nullptr) continue;

    if (c_keyword[0] == 'g' || c_keyword[0] == 'o') {  // part declaration;
      groupName = strtok_r(nullptr, " ", &rest);
      if (groupName == nullptr) {
        curPart = obj.addPart(std::string());
      } else {
        curPart = obj.addPart(groupName);
      }
    } else if (strncmp(c_keyword, "usemtl", 6) == 0) {
      // parse out the material id
      matIdStream.clear();
      matIdStream.str(line);
      matIdStream >> usemtlStream >> matIdStr;
      matId = obj.addMaterialLabel(matIdStr);
    } else if (strncmp(c_keyword, "vt", 2) == 0) {  // texture vertex
      std::array<float, 2> vt;
      if (!parseArray(vt, &rest)) {
        throw err("missing or invalid 'vt' (texture vertex) entry");
      }
      obj.addTextureCoord(vt[0], vt[1]);
    } else if (strncmp(c_keyword, "vn", 2) == 0) {  // vertex normal
      std::array<double, 3> n;
      if (!parseArray(n, &rest)) {
        throw err("missing or invalid 'vn' (vertex normal) entry");
      }
      if (request.swapYZ()) std::swap(n[1], n[2]);

      // check for unit length
      double norm2 = n[0] * n[0] + n[1] * n[1] + n[2] * n[2];
      if (norm2 > 0.0) {
        norm2 = std::sqrt(norm2);
        n[0] /= norm2;
        n[1] /= norm2;
        n[2] /= norm2;
        orderedNormals.push_back(obj.addVertexNormal(n[0], n[1], n[2]));
      } else {
        orderedNormals.push_back(u32Max);
      }
    } else if (c_keyword[0] == 'v') {  // vertex
      // parse out the vertex coordinates
      std::array<float, 3> v;
      if (!parseArray(v, &rest)) {
        throw err("missing or invalid 'v' (vertex) entry");
      }
      if (request.swapYZ()) std::swap(v[1], v[2]);

      // add the vertex to the list
      obj.addVertex(v[0] * unitScale, v[1] * unitScale, v[2] * unitScale);
    } else if (c_keyword[0] == 'f') {  // facet
      // make sure we have a valid matid
      if (matId == u32Max) {
        matId = obj.addMaterialLabel("default");
      }

      // tokenize the line and get the number of vertices
      int nVerts = 0;
      tok = strtok_r(nullptr, " ", &rest);
      while (tok != nullptr) {
        if (nVerts >= 4) {
          throw err("missing or invalid 'f' (facet) entry, facets must "
                    "have 3 or 4 vertices");
        }
        vertStrings[nVerts] = tok;
        nVerts++;
        tok = strtok_r(nullptr, " ", &rest);
      }

      // we only handle a maximum of 4 vertexes
      if (nVerts < 3 || nVerts > 4) {
        throw err("missing or invalid 'f' (facet) entry, facets must "
                  "have 3 or 4 vertices");
      }

      for (int ii = 0; ii < nVerts; ii++) {
        int jj = ii;
        if (request.swapYZ()) jj = nVerts - ii - 1;

        vIndex[jj] = 0;
        vtIndex[jj] = 0;
        vnIndex[jj] = 0;

        // check for "//"
        bool vnFlag = false;
        if (strstr(vertStrings[ii], "//") != nullptr) {
          vnFlag = true;
        }
        tok = strtok_r(vertStrings[ii], "/", &rest);
        vIndex[jj] = atoi(tok);
        tok = strtok_r(nullptr, "/", &rest);

        // handle negative (relative indexes)
        if (vIndex[jj] < 0) {
          vIndex[jj] = obj.numVertices() + vIndex[jj] + 1;
        }

        // only vertex info
        if (tok == nullptr) continue;

        int tempElem = atoi(tok);
        tok = strtok_r(nullptr, "/", &rest);
        if (tok == nullptr) {
          // v//n or v/t
          if (vnFlag) {
            vnIndex[jj] = tempElem;

            // handle negative (relative indexes)
            if (vnIndex[jj] < 0) {
              vnIndex[jj] = obj.numVertexNormals() + vnIndex[jj] + 1;
            }
          } else {
            vtIndex[jj] = tempElem;

            // handle negative (relative indexes)
            if (vtIndex[jj] < 0) {
              vtIndex[jj] = obj.numTextureCoords() + vtIndex[jj] + 1;
            }
          }

          continue;
        }

        // v/t/n
        vtIndex[jj] = tempElem;
        vnIndex[jj] = atoi(tok);

        // handle negative (relative indexes)
        if (vtIndex[jj] < 0) {
          vtIndex[jj] = obj.numTextureCoords() + vtIndex[jj] + 1;
        }
        if (vnIndex[jj] < 0) {
          vnIndex[jj] = obj.numVertexNormals() + vIndex[jj] + 1;
        }
      }

      auto tCount = static_cast<int64_t>(obj.numTextureCoords());
      int texCoordCount = 0;
      std::array<uint32_t, 4> texCoords;
      std::array<uint32_t, 4> vertNormals;

      // populate lists of texture coordinates and vertex normals
      for (int ii = 0; ii < nVerts; ii++) {
        // the DIRSIG vertex list starts at 0 not 1 (shift them all)
        indices[ii] = vIndex[ii] - 1;

        // create the list of texture vertexes
        if (vtIndex[ii] != 0) {
          long textureIdx = vtIndex[ii] - 1;

          if (textureIdx >= tCount) {
            throw err(fmt::format(
              "invalid texture vertex ({}) in face, must be < {}",
              textureIdx + 1, obj.numTextureCoords()));
          }

          texCoords[ii] = textureIdx;
          texCoordCount++;
        }

        // create the list of normal vertexes
        if (vnIndex[ii] > 0) {
          long normalIdx = vnIndex[ii] - 1;
          if (normalIdx >= static_cast<long>(orderedNormals.size())) {
            vertNormals[ii] = u32Max;
          } else {
            auto index = orderedNormals[normalIdx];
            if (index != u32Max) {
              vertNormals[ii] = index;
            } else {
              vertNormals[ii] = u32Max;
            }
          }
        } else {
          vertNormals[ii] = u32Max;
        }
      }

      // add this facet to the part list if it's valid
      auto facetIdx =
        obj.addFacetToPart(curPart, indices[0], indices[1], indices[2], matId);
      obj.setFacetTemperatureK(facetIdx, -1.0);

      // if we have valid texture vertexes, add them
      if (texCoordCount != 0) {
        obj.setTextureFacet(facetIdx, texCoords[0], texCoords[1], texCoords[2]);
      }

      // if we have valid vertex normals, add them
      if (!orderedNormals.empty()) {
        obj.setVertexNormalFacet(
          facetIdx, vertNormals[0], vertNormals[1], vertNormals[2]);
      }

      // if we had a 4-sided, split it into a second 3-sided
      if (nVerts == 4) {
        facetIdx = obj.addFacetToPart(
          curPart, indices[2], indices[3], indices[0], matId);
        obj.setFacetTemperatureK(facetIdx, -1.0);

        // if we have valid texture vertexes, add them
        if (texCoordCount != 0) {
          obj.setTextureFacet(
            facetIdx, texCoords[2], texCoords[3], texCoords[0]);
        }

        // if we have valid vertexs normals, add them
        if (!orderedNormals.empty()) {
          obj.setVertexNormalFacet(
            facetIdx, vertNormals[2], vertNormals[3], vertNormals[0]);
        }
      }
    }
  }

  obj.removeEmptyParts();
  return obj;
}
}  // namespace d5::d4legacy
