#include "OdbReader.h"

#include <fstream>
#include <unordered_map>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/Filesystem.h"
#include "dirsig5/d4legacy/geometry/Primitives.h"
#include "dirsig5/d4legacy/misc/FileLocator.h"
#include "dirsig5/d4legacy/misc/Units.h"
#include "dirsig5/d4legacy/tag/TagFile.h"
#include "dirsig5/base/Logger.h"

static const size_t kMaxIncludeDepth = 100;
static const std::string kCenter = "CENTER";
static const std::string kConcentrations = "CONCENTRATIONS";
static const std::string kInfo = "INFO";
static const std::string kInstances = "INSTANCES";
static const std::string kMaterialId = "MATERIAL_ID";
static const std::string kMaterialIds = "MATERIAL_IDS";
static const std::string kNormal = "NORMAL";
static const std::string kRadius = "RADIUS";
static const std::string kTemperature = "TEMPERATURE";
static const std::string kUnits = "UNITS";
static const std::string kWeightings = "WEIGHTINGS";

namespace d5::d4legacy {

template <typename T>
struct tag_file_type_name {};
template <>
struct tag_file_type_name<TagFileSection> {
  static constexpr auto name = "section";
};
template <>
struct tag_file_type_name<TagFileVariable> {
  static constexpr auto name = "variable";
};

template <typename T>
const T& ensure(const TagFileElement& e) {
  DIRSIG_THROW_IF_FMT(
      !e.is<T>(), "line {}: expected element \"{}\" to be a {}", e.lineNumber(),
      e.tagName(), tag_file_type_name<T>::name);
  return e.as<T>();
}

static std::shared_ptr<Geometry> LoadRegularGrid(
    const TagFileSection& section) {
  auto g = std::make_shared<RegularGrid>();
  g->lineNumber = section.lineNumber();
  section.requiredVariable("INSERT_POINT", 0, g->insertPoint);
  section.requiredVariable("DELTA_X", 0, g->voxelSize[0]);
  section.requiredVariable("DELTA_Y", 0, g->voxelSize[1]);
  section.requiredVariable("DELTA_Z", 0, g->voxelSize[2]);
  section.requiredVariable("GRID_FILENAME", 0, g->filename);
  DIRSIG_THROW_IF(
      g->voxelSize[0] <= 0 || g->voxelSize[1] <= 0 || g->voxelSize[2] <= 0,
      "DELTA_[XYZ] must be > 0 for regular grids");
  return g;
}

static std::shared_ptr<Geometry> LoadGeometricEntity(
    const TagFileElement& elem, DistanceUnits units) {
  DIRSIG_ERROR_TRACE(
      fmt::format("parsing {} on line {}", elem.tagName(), elem.lineNumber()), {
        const auto& elemName = elem.tagName();
        std::shared_ptr<Geometry> primitive;
        const auto& tagName = elem.tagName();
        if (tagName == "BOX")
          primitive = std::make_shared<Box>();
        else if (tagName == "CATENARY_CURVE")
          primitive = std::make_shared<CatenaryCurve>();
        else if (tagName == "CURVED_FRUSTUM")
          primitive = std::make_shared<CurvedFrustum>();
        else if (tagName == "CYLINDER")
          primitive = std::make_shared<Cylinder>();
        else if (tagName == "SPHERE")
          primitive = std::make_shared<Sphere>();
        else if (tagName == "SINUSOID")
          primitive = std::make_shared<Sinusoid>();
        else if (tagName == "SECCHI_DISK")
          primitive = std::make_shared<SecchiDisk>();
        else if (tagName == "DISK")
          primitive = std::make_shared<Disk>();
        else if (tagName == "GROUND_PLANE" || tagName == "SURFACE_PLANE")
          primitive = std::make_shared<SurfacePlane>();
        if (primitive) {
          primitive->ReadTag(ensure<TagFileSection>(elem));
          return primitive;
        }
        if (elemName == "GDB_FILENAME") {
          auto g = std::make_shared<GdbFacetObject>();
          g->lineNumber = elem.lineNumber();
          g->units = units;
          ensure<TagFileVariable>(elem).get_to(g->filename);
          return g;
        }
        if (elemName == "OBJ_FILENAME") {
          auto g = std::make_shared<ObjFacetObject>();
          g->lineNumber = elem.lineNumber();
          g->units = units;
          ensure<TagFileVariable>(elem).get_to(g->filename);
          return g;
        }
        if (elemName == "ODB_FILENAME") {
          auto g = std::make_shared<OdbInclude>();
          g->lineNumber = elem.lineNumber();
          ensure<TagFileVariable>(elem).get_to(g->filename);
          return g;
        }
        if (elemName == "REGULAR_GRID")
          return LoadRegularGrid(ensure<TagFileSection>(elem));
        DIRSIG_THROW("unknown geometric element name");
      })
}

static void LoadObject(const TagFileSection& section, Object& obj) {
  DIRSIG_ERROR_TRACE(
      fmt::format("loading an object on line {}", section.lineNumber()), {
        DistanceUnits units;
        DIRSIG_THROW_IF(
            !stringToUnits(
                section.requiredVariable<std::string>(kUnits), units),
            "invalid UNITS value");

        // find element for the geometric entity
        const TagFileElement* geomElem = nullptr;
        for (const auto& e : section) {
          if (e.tagName() != kUnits && e.tagName() != kInstances) {
            geomElem = &e;
          }
        }
        DIRSIG_THROW_IF(!geomElem, "no geometry element found");
        obj.addGeometry(LoadGeometricEntity(*geomElem, units));

        // hack to sneak zRotation into instances
        const auto& geom = *obj.geometries[0];
        double zRotation = 0;
        if (geom.type() == Geometry::Type::CurvedFrustum) {
          zRotation = geom.as<CurvedFrustum>().zRotation;
          zRotation *= M_PI / 180;
        }

        for (const auto& e : section.requiredSubsection(kInstances)) {
          if (e.tagName() != kInfo) continue;
          auto tokens =
              ensure<TagFileVariable>(e).get<std::vector<std::string>>();
          if (tokens.size() == 9) {
            auto inst = std::make_shared<TsrInstance>();
            inst->lineNumber = e.lineNumber();
            inst->xTranslate = std::atof(tokens[0].c_str());
            inst->yTranslate = std::atof(tokens[1].c_str());
            inst->zTranslate = std::atof(tokens[2].c_str());
            inst->xScale = std::atof(tokens[3].c_str());
            inst->yScale = std::atof(tokens[4].c_str());
            inst->zScale = std::atof(tokens[5].c_str());
            inst->xRotate = std::atof(tokens[6].c_str()) * M_PI / 180;
            inst->yRotate = std::atof(tokens[7].c_str()) * M_PI / 180;
            inst->zRotate = std::atof(tokens[8].c_str()) * M_PI / 180;
            inst->zRotate += zRotation;
            obj.instances.emplace_back(std::move(inst));
          } else if (tokens.size() == 16) {
            auto inst = std::make_shared<AffineInstance>();
            inst->lineNumber = e.lineNumber();
            for (size_t i = 0; i < inst->affine.size(); i++) {
              inst->affine[i] = std::atof(tokens[i].c_str());
            }
            // TODO Account for center in affine matrix?
            obj.instances.emplace_back(std::move(inst));
          } else if (tokens.size() == 1) {
            auto inst = std::make_shared<DeltaMotionInstance>();
            inst->lineNumber = e.lineNumber();
            inst->filename = std::move(tokens[0]);
            obj.instances.emplace_back(std::move(inst));
          } else {
            DIRSIG_THROW_FMT("line {}: invalid INFO entry", e.lineNumber());
          }
        }
        DIRSIG_THROW_IF_FMT(
            obj.instances.size() == 0,
            "line {}: empty INSTANCES section, must have at least 1 instance",
            section.requiredSubsection(kInstances).lineNumber());
      })
}

static std::unique_ptr<GeometryList> LoadOdbImpl(
    const std::string& filename, size_t depth) {
  DIRSIG_THROW_IF(depth >= kMaxIncludeDepth, "maximum include depth exceeded");

  // Open the given file
  auto ifs = d5::OpenIfstream(filename);

  // Load the file as a tag file
  auto file = TagFile::LoadFromIstream(ifs);
  ifs.close();

  const auto& top = file->topSection();

  // Make sure we have a valid 1.0 ODB
  if (auto v = top.requiredVariable<std::string>("DIRSIG_ODB"); v != "1.0") {
    DIRSIG_WARN("Unsupported ODB version \"{}\" only \"1.0\" is supported.", v);
  }

  auto glist = std::make_unique<GeometryList>();
  auto& objects = glist->objects;
  for (const auto& element : top) {
    const auto& name = element.tagName();
    if (name == "DIRSIG_ODB") continue;

    if (name == "OBJECT") {
      DIRSIG_THROW_IF_FMT(
          element.type() != TagFileElement::SectionBegin,
          "line {}: OBJECT must be a section", element.lineNumber());
      LoadObject(element.as<TagFileSection>(), objects.emplace_back());
    } else if (name == "INCLUDE_FILE") {
      FileLocator locator;
      locator.addPath(
          FileLocator::Any, {stdfs::absolute(filename).parent_path().string()});
      auto& f = glist->includes.emplace_back(locator.requiredFile(
          ensure<TagFileVariable>(element).get<std::string>()));

      DIRSIG_ERROR_TRACE(
          fmt::format(
              "loading file included from line {} of \"{}\"",
              element.lineNumber(), filename),
          { glist->append(std::move(*LoadOdbImpl(f.c_str(), depth + 1))); });
    } else {
      auto& obj = objects.emplace_back();
      obj.addGeometry(LoadGeometricEntity(element, DistanceUnits::Meters));
      auto inst = std::make_shared<AffineInstance>();
      inst->affine = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};
      const auto& geom = *obj.geometries[0];
      if (geom.type() == Geometry::Type::CurvedFrustum) {
        auto& frustum = const_cast<CurvedFrustum&>(geom.as<CurvedFrustum>());
        double zRotation = 0;
        zRotation = frustum.zRotation;
        zRotation *= M_PI / 180;
        // clang-format off
        inst->affine = {
            std::cos(zRotation), -std::sin(zRotation), 0, frustum.center[0],
            std::sin(zRotation),  std::cos(zRotation), 0, frustum.center[1],
            0, 0, 1, frustum.center[2],
            0, 0, 0, 1};
        // clang-format on
        frustum.center = {0, 0, 0};
      }
      obj.instances.emplace_back(std::move(inst));
    }
  }

  return glist;
}

std::unique_ptr<GeometryList> LoadOdb(const std::string& filename) {
  return LoadOdbImpl(filename, 0);
}
}
