#include "TdfReader.h"

#include <iostream>
#include <map>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Logger.h"

#if WITH_MUSES
#include "tdfbase.h"
#include "tdfelement.h"
#include "tdfenvironment.h"
#include "tdfio.h"
#include "tdfmaterial.h"
#include "tdfpaint.h"
#include "tdfpart.h"
#include "tdfproperty.h"
#include "tdfsolnparam.h"
#include "tdfterrain.h"
#include "tdftime.h"
#endif  // #if WITH_MUSES

namespace d5::d4legacy {

FacetObject ReadTdfFile(const FacetObjectReadRequest& request) {
#if WITH_MUSES
  static std::mutex tdfMutex;
  std::lock_guard tdfMutexLock{tdfMutex};

  TdfIO tdfio;
  TdfProperty writeFrequency;

  // MuSES plug-in needs to save off a new TDF file for MuSES to run against.
  FacetObject obj;
  std::string filename = request.filename();
  TdfIO::setVerboseMode(TDF::NoMessage);
  // check the file version
  int versionMajor = 0;
  int versionMinor = 0;
  char versionMod[32] = {};
  if (
    TdfIO::modelVersion(
      filename.c_str(), versionMajor, versionMinor, versionMod, 32) ==
    TDF::Success) {
  }
  DIRSIG_INFO(
    "Reading in MuSES TDF file {} (version {}.{}.{})", filename, versionMajor,
    versionMinor, versionMod);

  DIRSIG_THROW_IF_FMT(tdfio.readFile(filename.c_str()) != TDF::Success,
    "Unable to read MuSES TDF file '{}'", filename.c_str());

  DIRSIG_THROW_IF_FMT(tdfio.readResults() != TDF::Success,
    "Unable to read results from MuSES TDF file '{}'", filename.c_str());

  int timeStepIndex = request.timeStepIndex();
  int timeStepCount = tdfio.timeStepCount();
  if (timeStepIndex == -1) timeStepIndex = timeStepCount - 1;

  DIRSIG_DEBUG( "  Number of time results in TDF = {}", tdfio.timeStepCount());
  DIRSIG_DEBUG( "  Requested time result in TDF  = {}", timeStepIndex);

  const int vertCount = tdfio.numberOfVertices();
  const int elemCount = tdfio.numberOfElements();
  const int partCount = tdfio.numberOfParts();
  std::map<std::string, int> matLookup;
  std::map<int, std::string> partMats;
  std::map<int, int> partMap;
  std::map<int, int> vertMap;
  // Read part names and materials
  for (int part = 0; part < partCount; part++) {
    int partId = tdfio.partID(part);
    TdfPart curPart;
    tdfio.part(partId, curPart);
    std::string matId;
    int matNum;
    TDF::TdfDataLocation matDataLocation;
    if (curPart.material(matNum, matDataLocation) != TDF::Success)
      DIRSIG_THROW_FMT(
        "unable to parse material for TDF part #{} in {}", matNum, filename);
    matId = std::to_string(matNum);
    if (request.useTdfMaterialNames()) {
      TdfMaterial material;
      if (tdfio.material(matNum, material) != TDF::Success)
        DIRSIG_THROW_FMT(
          "unable to get material for TDF part #{} in {}", matNum, filename);
      matId = material.name();
    }
    auto iter = matLookup.find(matId);
    if (iter == matLookup.end()) {
      matLookup[matId] = obj.addMaterialLabel(matId);
    }
    partMap[partId] = obj.addPart(curPart.name());
    partMats[partId] = matId;
  }

  // Read verts
  {
    std::array<float, 3> vertBoxMin = {+INFINITY, +INFINITY, +INFINITY};
    std::array<float, 3> vertBoxMax = {-INFINITY, -INFINITY, -INFINITY};
    std::vector<std::array<float, 3>> verts;
    verts.reserve(vertCount);
    for (int vert = 0; vert < vertCount; vert++) {
      int vertId = tdfio.vertexID(vert);
      if (vertId == TDF::InvalidID)
        DIRSIG_THROW_FMT("encountered an invalid vertex ID, vertex #{}", vert);
      float x, y, z;
      tdfio.vertex(vertId, x, y, z);
      verts.push_back(std::array<float, 3>{x, y, z});
      for (int k = 0; k < 3; k++) {
        vertBoxMin[k] = std::fmin(vertBoxMin[k], verts.back()[k]);
        vertBoxMax[k] = std::fmax(vertBoxMax[k], verts.back()[k]);
      }
    }
    std::array<float, 3> offset = {
      0.5f * (vertBoxMin[0] + vertBoxMax[0]),
      0.5f * (vertBoxMin[1] + vertBoxMax[1]), vertBoxMin[2]};
    for (int vert = 0; vert < vertCount; vert++) {
      int vertId = tdfio.vertexID(vert);
      if (vertId == TDF::InvalidID)
        DIRSIG_THROW_FMT("invalid vertex ID, vertex #{}", vert);
      float x = verts[vert][0] - offset[0];
      float y = verts[vert][1] - offset[1];
      float z = verts[vert][2] - offset[2];
      vertMap[vertId] = obj.addVertex(-y, x, z);
    }
  }

  // Read facets
  {
    for (int elem = 0; elem < elemCount; elem++) {
      int elemId = tdfio.elementID(elem);
      if (elemId == 0)
        DIRSIG_THROW_FMT("invalid element ID, element #{}", elem);
      TdfElement curElem;
      if (tdfio.element(elemId, curElem) != TDF::Success)
        DIRSIG_THROW_FMT("can't retrieve element ID={}", elemId);
      // There are two types of elements, Shell and Volume
      // We cannot facetize a volume so check that it's a Shell
      if (curElem.type() == TdfElement::TdfElementType::Shell) {
        float elemTemp = tdfio.temperature(elemId, timeStepIndex, TDF::Front);
        float elemThermalEmissivity =
          tdfio.thermalEmissivity(elemId, timeStepIndex, TDF::Front);

        if (elemThermalEmissivity == TDF::InvalidEmissivity) {
          TdfPart myPart;
          TdfPaint myPaint;
          int paintID;
          int elemPartID;
          float elemThermalEmissivity;
          elemPartID = tdfio.elementPartID(elemId);
          tdfio.part(elemPartID, myPart);  // put part info into myPart
          if (myPart.paintUsed(TDF::Front)) {
            myPart.paint(paintID);          // grab the paintID
            tdfio.paint(paintID, myPaint);  // grab the paint info for myPaint
            elemThermalEmissivity = myPaint.thermalEmissivity();
          } else {
            DIRSIG_THROW_FMT(
              "can't retrieve thermal emissivity for element ID={}", elemId);
          }
        }
        if (elemTemp == TDF::InvalidTemperature) {
          DIRSIG_THROW_FMT(
            "can't retrieve temperature for element ID={}", elemId);
        }
        int elemVertCount = curElem.numberOfVertices();

        int vertIds[4];
        int vertIndexes[4];
        curElem.vertices(vertIds[0], vertIds[1], vertIds[2], vertIds[3]);
        for (int vert = 0; vert < elemVertCount; vert++) {
          auto iter = vertMap.find(vertIds[vert]);
          if (iter == vertMap.end())
            DIRSIG_THROW_FMT(
              "can't find index for vertex ID={}", vertIds[vert]);
          vertIndexes[vert] = iter->second;
        }

        int partId = curElem.part();
        if (partMap.count(partId) != 1)
          DIRSIG_THROW_FMT(
            "can't find part ID={} for element ID={}", partId, elemId);
        int partIndex = partMap.at(partId);
        int partMatIndex = matLookup.at(partMats.at(partId));
        for (int vert = 1; vert + 1 < std::min(elemVertCount, 4); vert++) {
          auto facetIndex = obj.addFacetToPart(
            partIndex,              //
            vertIndexes[0],         //
            vertIndexes[vert],      //
            vertIndexes[vert + 1],  //
            partMatIndex);
          obj.setFacetTemperatureK(facetIndex, elemTemp);
          obj.setFacetThermalEmissivity(facetIndex, elemThermalEmissivity);
        }
      }
    }
  }

  // Sets all of the environment and solution parameters
  // Then writes out a new tdf file
  if (request.MuSES_Plugin()) {
    TdfSolutionParameters soln;
    DIRSIG_THROW_IF(
      tdfio.solutionParameters(soln) != TDF::Success,
      "can't get solution parameters");
    soln.setStepSize(request.MuSES_StepSize());
    soln.setStartTime(request.MuSES_solutionStartTime());
    soln.setEndTime(request.MuSES_solutionEndTime());
    soln.setResultOutputMode(TdfSolutionParameters::EntireModel);
    soln.setSaveTransientRecord(true);
    writeFrequency.setValue(1.0);
    soln.setWriteFrequency(writeFrequency);
    TdfEnvironment myEnv;
    tdfio.environment(myEnv);
    myEnv.setEnvironmentType(TdfEnvironment::Natural);
    d5::Point origin = request.MuSES_Origin();
    myEnv.setLatitude(origin[0]);
    myEnv.setLongitude(origin[1] * -1);
    myEnv.setElevation(origin[2]);
    myEnv.setTimeZone(0); /* Ignoring request.MuSES_timeZone() for now ... */
    myEnv.setSolarModeled(0);
    myEnv.setWeatherFile(
      request.MuSES_WeatherFilename().c_str(), request.MuSES_WeatherStart(),
      request.MuSES_WeatherDuration());
    tdfio.setEnvironment(myEnv);
    tdfio.setSolutionParameters(soln);
    tdfio.writeFile(request.MuSES_Filename().c_str());
  }

  obj.removeEmptyParts();

  return obj;
#else
  return {};
#endif  // #if WITH_MUSES
}

}  // namespace d5::d4legacy
