/** Implementation file for bulk property I/O
 *
 *  \file BulkPropertiesIO.cpp
 */

#include "BulkPropertiesIO.h"

#include "dirsig5/base/Error.h"
#include "spdlog/fmt/fmt.h"

namespace d5::d4legacy {

namespace {

// ########################################
// #    Tag file I/O helper functions     #
// ########################################

template <typename T>
void LoadSimpleFilename(const TagFileSection& section, T& p) {
  if (
    section.optionalVariable("FILENAME", 0, p.filename).found ||
    section.optionalVariable("XML_FILENAME", 0, p.filename).found) {
    p.format = SimpleSpectralData::Format::XML;
  } else if (section.optionalVariable("TXT_FILENAME", 0, p.filename).found) {
    p.format = SimpleSpectralData::Format::Text;
  } else {
    DIRSIG_THROW("missing simple property filename");
  }
}

template <typename T>
std::unique_ptr<T> LoadPropFilenameOnly(const TagFileSection& section) {
  auto p = std::make_unique<T>();
  section.requiredVariable("FILENAME", 0, p->filename);
  return p;
}

// ####################################################
// #    AbsorptionProperty tag file I/O functions     #
// ####################################################

std::unique_ptr<SimpleAbsorption> LoadSimpleAbsorption(
  const TagFileSection& section) {
  auto p = std::make_unique<SimpleAbsorption>();
  LoadSimpleFilename(section, *p);

  std::string boolVar;
  if (section.optionalVariable("USE_CONCENTRATION", 0, boolVar).found) {
    if (boolVar == "TRUE") {
      p->useConcentrationScaling = true;
    } else if (boolVar == "FALSE") {
      p->useConcentrationScaling = false;
    } else {
      DIRSIG_THROW(R"(USE_CONCENTRATION must be either "TRUE" or "FALSE")");
    }
  }

  double abs;
  if (section.optionalVariable("BROAD_BAND_SOLAR_ABSORPTION", 0, abs).found) {
    p->solarAbsorption = abs;
  }

  return p;
}

std::unique_ptr<AbsorptionProperty> LoadAbsorptionProperty(
  const TagFileSection& section) {
  std::string name;
  if (!section.optionalVariable("ABSORPTION_PROP_NAME", 0, name).found) {
    return nullptr;
  }

  const auto& propSec = section.requiredSubsection("ABSORPTION_PROP");
  if (name == "ClassicAbsorption") {
    return LoadPropFilenameOnly<ClassicAbsorption>(propSec);
  } else if (name == "SimpleAbsorption") {
    return LoadSimpleAbsorption(section.requiredSubsection("ABSORPTION_PROP"));
  } else if (name == "VoxelizedAbsorption") {
    return LoadPropFilenameOnly<VoxelizedAbsorption>(propSec);
  }

  DIRSIG_THROW_FMT("unknown absorption property \"{}\"", name);
}

// ####################################################
// #    ScatteringProperty tag file I/O functions     #
// ####################################################

std::unique_ptr<SimpleScattering> LoadSimpleScattering(
  const TagFileSection& section) {
  auto p = std::make_unique<SimpleScattering>();
  LoadSimpleFilename(section, *p);

  std::string boolVar;
  if (section.optionalVariable("USE_CONCENTRATION", 0, boolVar).found) {
    if (boolVar == "TRUE") {
      p->useConcentrationScaling = true;
    } else if (boolVar == "FALSE") {
      p->useConcentrationScaling = false;
    } else {
      DIRSIG_THROW(R"(USE_CONCENTRATION must be either "TRUE" or "FALSE")");
    }
  }

  return p;
}

std::unique_ptr<ScatteringProperty> LoadScatteringProperty(
  const TagFileSection& section) {
  std::string name;
  if (!section.optionalVariable("SCATTERING_PROP_NAME", 0, name).found) {
    return nullptr;
  }

  const auto& propSec = section.requiredSubsection("SCATTERING_PROP");
  if (name == "SimpleScattering") {
    return LoadSimpleScattering(propSec);
  }

  DIRSIG_THROW_FMT("unknown scattering propery \"{}\"", name);
}

// #######################################################
// #    PhaseFunctionProperty tag file I/O functions     #
// #######################################################

std::unique_ptr<HenyeyGreensteinPhaseFunction> LoadHenyeyGreenstein(
  const TagFileSection& section) {
  auto p = std::make_unique<HenyeyGreensteinPhaseFunction>();
  LoadSimpleFilename(section, *p);
  return p;
}

std::unique_ptr<PhaseFunctionProperty> LoadPhaseFunctionProperty(
  const TagFileSection& section) {
  std::string name;
  if (!section.optionalVariable("PHASE_FUNCTION_PROP_NAME", 0, name).found) {
    return nullptr;
  }

  const auto& propSec = section.requiredSubsection("PHASE_FUNCTION_PROP");
  if (name == "HenyeyGreenstein") {
    return LoadHenyeyGreenstein(propSec);
  } else if (name == "TabulatedPhaseFunction") {
    return LoadPropFilenameOnly<TabulatedPhaseFunction>(propSec);
  }

  DIRSIG_THROW_FMT("unknown phase function propery \"{}\"", name);
}
}  // namespace

BulkProperties LoadBulkProperties(const TagFileSection& section) {
  return BulkProperties{
    LoadAbsorptionProperty(section), LoadScatteringProperty(section),
    LoadPhaseFunctionProperty(section)};
}
}  // namespace d5::d4legacy
