#include "dirsig5/d4legacy/mat/EmissivityData.h"

#include <fstream>

#include "../misc/LineReader.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/StringUtils.h"
#include "dirsig5/base/Logger.h"

namespace d5::d4legacy {

std::vector<double> EmissivityData::averageReflectance(
  const std::vector<double>& wavelengthMicrons) const {
  std::vector<double> average(wavelengthMicrons.size(), 0),
    scratch(wavelengthMicrons.size());
  if (this->curves.empty()) return average;

  for (const auto& curve : this->curves) {
    interpolate(
      curve.wavelengthMicrons, curve.reflectance, wavelengthMicrons, scratch);
    for (size_t i = 0; i < wavelengthMicrons.size(); i++) {
      average[i] += scratch[i];
    }
  }

  auto n = static_cast<double>(this->curves.size());
  for (auto& x : average) x /= n;
  return average;
}

void LoadEmsFile(
  const char* filename,
  span<const float> wavelengthsScene,
  EmissivityData& ems) {
  ems.curves.clear();
  ems.angleWeighting.clear();

  auto ifs = OpenIfstream(filename);

  LineReader lineReader(ifs);
  lineReader.nextLineRequired();
  size_t curveCount = 0;
  try {
    curveCount = std::stoul(lineReader.line);
  } catch (const std::exception&) {
    lineReader.fail("cannot parse curve count");
  }
  if (curveCount > 100000) {
    lineReader.fail("too many curves");
  }

  ems.curves.resize(curveCount);
  ems.angleWeighting.resize(91);

  for (auto& angleWeighting : ems.angleWeighting) {
    lineReader.nextLineRequired();
    try {
      angleWeighting = std::stod(lineReader.line);
    } catch (const std::exception&) {
      lineReader.fail("cannot parse angle weighting");
    }
  }
  if (curveCount == 0) return;

  // Infer the file type
  size_t curveIndex = 0;
  bool newFormat = false;
  lineReader.nextLineRequired();
  try {
    if (trimmed(lineReader.line) == "CURVE_BEGIN") {
      newFormat = true;
    } else {
      curveIndex = std::stoull(lineReader.line);
    }
  } catch (const std::exception&) {
    lineReader.fail("cannot infer EMS version from curve header");
  }

  double wave, emis;
  bool warnGiven = false;
  bool warnGiven2 = false;
  auto sanitizeCurve{[&](auto& curve) {
    // Sanitize
    if (!wavelengthsScene.empty() && !warnGiven2) {
      double expectedMinWaveLen = wavelengthsScene.front();
      double expectedMaxWaveLen = wavelengthsScene.back();
      double minWaveLen = curve.wavelengthMicrons.front();
      double maxWaveLen = curve.wavelengthMicrons.back();
      if (
        expectedMinWaveLen < minWaveLen - 0.001 ||
        expectedMaxWaveLen > maxWaveLen + 0.001) {
        warnGiven2 = true;
        DIRSIG_WARN(
          "{}: has wavelengths {:1.3f}-{:1.3f}μm, but this is "
          "insufficient for the scene wavelengths {:1.3f}-{:1.3f}μm",
          ShowFilename(filename), minWaveLen, maxWaveLen, expectedMinWaveLen,
          expectedMaxWaveLen);
      }
    }
  }};
  if (newFormat) {
    for (curveIndex = 0; curveIndex < curveCount; curveIndex++) {
      auto& curve = ems.curves[curveIndex];
      while (lineReader.nextLineOptional()) {
        auto lineSize = lineReader.line.size();
        if (lineSize == 0) continue;

        if (
          lineSize >= 11 &&
          std::string_view(lineReader.line.data(), 11) == "CURVE_BEGIN") {
          break;
        }

        char* s = lineReader.line.data();
        char* end = s + lineSize;
        char* p;
        wave = strtod(s, &p);
        if (p == s || p == end) continue;
        s = p;
        emis = strtod(s, &p);
        if (p == s) continue;
        if (!(0 <= emis && emis <= 1)) {
          emis = std::fmax(emis, 0.0);
          emis = std::fmin(emis, 1.0);
          if (warnGiven == false) {
            warnGiven = true;
            DIRSIG_WARN(
              "{}: has emissivity values outside the expected range [0:1]",
              ShowFilename(filename));
          }
        }
        curve.wavelengthMicrons.push_back(wave);
        curve.reflectance.push_back(1 - emis);
      }
      sanitizeCurve(curve);
    }
  } else {
    DIRSIG_WARN("{}: has deprecated wavenumber format", ShowFilename(filename));
    size_t parsedCurveCount = 1;
    while (true) {
      auto& curve = ems.curves[curveIndex];
      for (size_t i = 0; i < 396; i++) {
        lineReader.nextLineRequired();
        std::istringstream line(lineReader.line);
        double wave = 0;
        double emis = 0;
        if (!(line >> wave >> emis))
          lineReader.fail("curve format should be: \"wavenumber emissivity\"");
        if (!(0 <= emis && emis <= 1)) {
          emis = std::fmax(emis, 0.0);
          emis = std::fmin(emis, 1.0);
          if (warnGiven == false) {
            warnGiven = true;
            DIRSIG_WARN(
              "{}: has emissivity values outside the expected range [0:1]",
              ShowFilename(filename));
          }
        }
        curve.wavelengthMicrons.push_back(1e+4 / wave);
        curve.reflectance.push_back(1 - emis);
      }
      sanitizeCurve(curve);
      if (parsedCurveCount == curveCount) break;
      // Get the next curve index
      lineReader.nextLineRequired();
      try {
        curveIndex = std::stoull(lineReader.line);
      } catch (const std::exception&) {
        lineReader.fail("cannot parse curve index");
      }
      parsedCurveCount++;
    }
  }
}

}  // namespace d5::d4legacy
