/** Implementation file for ExtinctionProperty.h
 *
 *  \file ExtinctionProperty.cpp
 */

#include "dirsig5/d4legacy/mat/ExtinctionProperty.h"

#include <fstream>

#include "../misc/LineReader.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/StringUtils.h"
#include "dirsig5/base/Logger.h"
#include "dirsig5/d4legacy/tag/TagFile.h"
#include "dirsig5/d4legacy/tag/TagFileElement.h"

namespace d5::d4legacy {

ClassicExtinction::ExtinctionData ReadExtinctionFile(
  const char* filename, span<const float> wavelengthsScene) {
  ClassicExtinction::ExtinctionData ed;

  auto ifs = OpenIfstream(filename);

  LineReader lineReader(ifs);
  lineReader.nextLineRequired();
  if (trimmed(lineReader.line) != "1") {
    lineReader.fail("first line must be \"1\"");
  }

  bool newFormat = false;
  lineReader.nextLineRequired();
  if (lineReader.line == "CURVE_BEGIN") {
    newFormat = true;
  } else if (lineReader.line != "0") {
    lineReader.fail(R"(second line must be either "CURVE_BEGIN" or "0")");
  }
  if (!newFormat)
    DIRSIG_WARN("{}: has deprecated wavenumber format", ShowFilename(filename));

  while (lineReader.nextLineOptional()) {
    std::istringstream line(lineReader.line);
    float wavelength = 0, extinction = 0;
    if (!(line >> wavelength >> extinction)) {
      lineReader.fail("line format should be: \"wavelength extinction\"");
    }
    ed.wavelengthsMicron.push_back(newFormat ? wavelength : 1e4 / wavelength);
    ed.extinctionsInvKm.push_back(extinction);
  }

  // Sanitize!
  if (!wavelengthsScene.empty()) {
    double expectedMinWaveLen = wavelengthsScene.front();
    double expectedMaxWaveLen = wavelengthsScene.back();
    double minWaveLen = ed.wavelengthsMicron.front();
    double maxWaveLen = ed.wavelengthsMicron.back();
    if (
      expectedMinWaveLen < minWaveLen - 0.001 ||
      expectedMaxWaveLen > maxWaveLen + 0.001) {
      DIRSIG_WARN(
        "{}: has wavelengths {:1.3f}-{:1.3f}μm, but this is "
        "insufficient for the scene wavelengths {:1.3f}-{:1.3f}μm",
        ShowFilename(filename), minWaveLen, maxWaveLen, expectedMinWaveLen,
        expectedMaxWaveLen);
    }
  }

  return ed;
}

}  // namespace d5::d4legacy
