#include "dirsig5/d4legacy/mat/MatFile.h"

#include <fstream>
#include <unordered_set>

#include "../tag/CommonParsers.h"
#include "BulkPropertiesIO.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/StringUtils.h"
#include "dirsig5/d4legacy/tag/TagFile.h"
#include "dirsig5/d4legacy/tag/TagFileElement.h"
#include "dirsig5/base/Logger.h"

namespace d5::d4legacy {

static std::unique_ptr<ReflectanceProperty> LoadReflectancePropertyByName(
  const std::string& reflPropName, TagFileSection& reflPropSec);

// Load ClassicEmissivity from tag file section.
static std::unique_ptr<EmissivityProperty> LoadClassicEmissivity(
  TagFileSection& section) {
  auto e = std::make_unique<ClassicEmissivity>();

  // Populate the optional texture map
  if (auto s = section.subsection("TEXTURE_MAP"); s != nullptr) {
    auto& tm = e->textureMap.emplace();

    s->optionalVariable("NAME", 0, tm.name);
    s->optionalVariable("ENABLED", 0, tm.enabled);
    tm.projector.ReadTag(*s);

    const auto& imgList = s->requiredSubsection("IMAGE_LIST");
    for (const auto& entry : imgList) {
      if (
        entry.tagName() != "IMAGE" ||
        entry.type() != TagFileElement::Type::SectionBegin) {
        continue;
      }

      const auto& imageSec = entry.as<TagFileSection>();
      auto& band = tm.bands.emplace_back();
      band.source = ImageSource(
        ImageSource::Type::Filename,
        imageSec.requiredVariable<std::string>("FILENAME"));
      imageSec.requiredVariable("MIN_WAVELENGTH", 0, band.bandpassMin);
      imageSec.requiredVariable("MAX_WAVELENGTH", 0, band.bandpassMax);

      // read the (optional) ignore DC list
      auto ignoreDcStr = imageSec.optionalVariable<std::string>("IGNORE_DCS");
      if (!ignoreDcStr.empty()) {
        auto ignoreDcTokens = TokenList(ignoreDcStr, ",");
        for (const auto& token : ignoreDcTokens) {
          int dcValue = ParseNumber<int>(token);
          if (dcValue < 0 || dcValue > 255) {
            DIRSIG_THROW_FMT("invalid DC ({}) in IGNORE_DCS list", dcValue);
          }
          band.ignoreDCs.push_back(static_cast<uint8_t>(dcValue));
        }
      }

      // read the (optional) mean and variance
      auto imStatsStr = imageSec.optionalVariable<std::string>("STATS");
      if (!imStatsStr.empty()) {
        auto tokens = TokenList(imStatsStr, ",");
        if (tokens.size() != 2) DIRSIG_THROW_FMT("invalid STATS pair");
        band.dcMean = ParseNumber<double>(tokens[0]);
        band.dcStdDev = ParseNumber<double>(tokens[1]);
      }
    }

    if (tm.bands.empty()) {
      DIRSIG_THROW_FMT("at least one IMAGE band needs to be defined");
    }
  }

  e->filename = section.requiredVariable<std::string>("FILENAME");
  return e;
}

// WardDataPair for use with TagFileSection::requiredVariable().
using WardDataPair = std::array<float, 2>;

// Load WardBrdf from tag file section.
static std::unique_ptr<ReflectanceProperty> LoadWardBrdf(
  TagFileSection& section) {
  auto r = std::make_unique<WardBrdf>();
  try {
    // Try to parse diffuse/specular weights
    auto dsWeights = section.requiredVariable<WardDataPair>("DS_WEIGHTS");
    r->diffuseWeight = dsWeights[0];
    r->specularWeight = dsWeights[1];
  } catch (const std::exception&) {
    DIRSIG_THROW_FMT("cannot parse DS_WEIGHTS");
  }
  try {
    // Try to parse XY specular sigmas
    auto xySigmas = section.requiredVariable<WardDataPair>("XY_SIGMAS");
    r->specularSigmaParallel = xySigmas[0];
    r->specularSigmaOrthogonal = xySigmas[1];
  } catch (const std::exception&) {
    DIRSIG_THROW_FMT("cannot parse XY_SIGMAS");
  }
  return r;
}

// Load WardBtdf from tag file section.
static std::unique_ptr<ExtinctionProperty> LoadWardBtdf(
  TagFileSection& section) {
  auto t = std::make_unique<WardBtdf>();
  try {
    // Try to parse diffuse/specular weights
    auto dsWeights = section.requiredVariable<WardDataPair>("DS_WEIGHTS");
    t->diffuseWeight = dsWeights[0];
    t->specularWeight = dsWeights[1];
  } catch (const std::exception&) {
    DIRSIG_THROW_FMT("cannot parse DS_WEIGHTS");
  }
  try {
    // Try to parse XY specular sigmas
    auto xySigmas = section.requiredVariable<WardDataPair>("XY_SIGMAS");
    t->specularSigmaParallel = xySigmas[0];
    t->specularSigmaOrthogonal = xySigmas[1];
  } catch (const std::exception&) {
    DIRSIG_THROW_FMT("cannot parse XY_SIGMAS");
  }
  return t;
}

template <typename Prop>
static void LoadSimpleSpectralDataMembers(
  TagFileSection& section, std::unique_ptr<Prop>& r) {
  r->format = SimpleSpectralData::Format::XML;
  section.optionalVariable("FILENAME", 0, r->filename);
  section.optionalVariable("XML_FILENAME", 0, r->filename);
  if (r->filename.empty()) {
    section.optionalVariable("TXT_FILENAME", 0, r->filename);
    r->format = SimpleSpectralData::Format::Text;
  }
  if (r->filename.empty()) {
    r->filename = section.requiredVariable<std::string>("SLI_FILENAME");
    if (section.element("SLI_INDEX"))
      r->sliIndex = section.requiredVariable<int>("SLI_INDEX");
    else
      r->sliIndex = section.requiredVariable<std::string>("SLI_SPECTRUM");
    r->format = SimpleSpectralData::Format::SLI;
  }
}

// Load SimpleReflectance from tag file section.
static std::unique_ptr<ReflectanceProperty> LoadSimpleReflectance(
  TagFileSection& section) {
  auto r = std::make_unique<SimpleReflectance>();
  LoadSimpleSpectralDataMembers(section, r);
  return r;
}

// Load DeltaReflectance from tag file section.
static std::unique_ptr<ReflectanceProperty> LoadDeltaReflectance(
  TagFileSection& section) {
  auto r = std::make_unique<DeltaReflectance>();
  LoadSimpleSpectralDataMembers(section, r);
  return r;
}

// Load PriestGermer from tag file section.
static std::unique_ptr<ReflectanceProperty> LoadPriestGermer(
  TagFileSection& section) {
  auto r = std::make_unique<PriestGermer>();
  r->polarizationEnabled =
    section.requiredVariable<bool>("ENABLE_POLARIZATION");
  r->diffuseContributionEnabled =
    section.requiredVariable<bool>("ENABLE_DIFFUSE_CONTRIBUTION");
  bool enableDiffuseRenormalization = false;
  section.optionalVariable<bool>(
    "ENABLE_DIFFUSE_RENORMALIZATION", 0, enableDiffuseRenormalization);
  r->diffuseRenormalizationEnabled = enableDiffuseRenormalization;
  r->shadowingFunctionEnabled =
    section.requiredVariable<bool>("ENABLE_SHADOWING_FUNCTION");
  r->paramFilename = section.requiredVariable<std::string>("PARAMETER_FILE");
  return r;
}

// Load ShellTarget from tag file section.
static std::unique_ptr<ReflectanceProperty> LoadShellTarget(
  TagFileSection& section) {
  auto r = std::make_unique<ShellTarget>();
  section.requiredVariable("BRDF_FIT_FILE", 0, r->fitFilename);
  section.optionalVariable("INTERPOLATION_METHOD", 0, r->interpolation);
  if (r->interpolation == 0) {
    section.requiredVariable<std::string>("EMISSIVITY_FILE", 0, r->emsFilename);
  } else {
    section.optionalVariable("EMISSIVITY_FILE", 0, r->emsFilename);
  }
  return r;
}

// Load ShellBackground from tag file section.
static std::unique_ptr<ReflectanceProperty> LoadShellBackground(
  TagFileSection& section) {
  auto r = std::make_unique<ShellBackground>();
  section.requiredVariable("BRDF_FIT_FILE", 0, r->fitFilename);
  section.optionalVariable("EMISSIVITY_FILE", 0, r->emsFilename);
  return r;
}

// Load ReflectanceMap from tag file section
static std::unique_ptr<ReflectanceProperty> LoadReflectanceMap(
  TagFileSection& section) {
  auto r = std::make_unique<ReflectanceMap>();
  auto& map = r->map;
  map.source.setFilename(
    section.requiredVariable<std::string>("IMAGE_FILENAME"));
  map.projector.ReadTag(section);
  if (auto c = section.optionalVariable<std::string>("BANDS_TO_WAVELENGTHS");
      !c.empty()) {
    map.bandCenters = tag::ParseVector<double>(c);
    for (auto w : map.bandCenters) {
      if (w <= 0) DIRSIG_THROW_FMT("invalid value in BANDS_TO_WAVELENGTHS");
    }
  }

  // Options are either in the OPTIONS section or just in this section
  TagFileSection* options = &section;
  if (auto* s = section.subsection("OPTIONS"); s != nullptr) options = s;

  // Load in options
  options->optionalVariable("CLIP_VALUES", 0, map.clipValues);
  options->optionalVariable("CONVERT_TO_MIXTURE", 0, map.convertToMixture);
  options->optionalVariable("GAIN", 0, map.gain);
  options->optionalVariable("BIAS", 0, map.bias);
  options->optionalVariable("SQT_FILENAME", 0, map.sqtFilename);

  return r;
}

// Load RgbImage from tag file section
static std::unique_ptr<ReflectanceProperty> LoadRgbImage(
  TagFileSection& section) {
  auto r = std::make_unique<RgbImage>();
  section.requiredVariable("IMAGE_FILENAME", 0, r->filename);
  section.requiredVariable("SCALE", 0, r->scale);
  r->projector.ReadTag(section);
  return r;
}

static std::unique_ptr<ReflectanceProperty> LoadCurveMap(
  TagFileSection& section) {
  auto r = std::make_unique<CurveMap>();
  section.requiredVariable("SLI_FILENAME", 0, r->sliFilename);
  section.optionalVariable("SLI_INDEX_OFFSET", 0, r->sliIndexOffset);
  section.requiredVariable("IMAGE_FILENAME", 0, r->imageFilename);
  r->projector.ReadTag(section);
  return r;
}

// Load RossLi from tag file section
static std::unique_ptr<ReflectanceProperty> LoadRossLi(
  const TagFileSection& section) {
  auto r = std::make_unique<RossLi>();

  // Parse the Ross kernel
  std::string tmp;
  if (section.optionalVariable("ROSS", 0, tmp); !tmp.empty()) {
    if (tmp == "THIN") {
      r->rossKernel = RossLi::Thin;
    } else if (tmp == "THICK") {
      r->rossKernel = RossLi::Thick;
    } else {
      DIRSIG_THROW_FMT("unknown Ross kernel type: \"{}\"", tmp);
    }
  }

  // Parse the Li kernel
  tmp.clear();
  if (section.optionalVariable("LI", 0, tmp); !tmp.empty()) {
    if (tmp == "DENSE") {
      r->liKernel = RossLi::Dense;
    } else if (tmp == "SPARSE") {
      r->liKernel = RossLi::Sparse;
    } else {
      DIRSIG_THROW_FMT("unknown Li kernel type: \"{}\"", tmp);
    }
  }

  // Read the wavelength-invariant parameters
  section.optionalVariable("BR", 0, r->br);
  section.optionalVariable("HB", 0, r->hb);
  section.optionalVariable("VERTICAL_FLAG", 0, r->verticalFlag);
  if (r->liKernel == RossLi::Sparse) {
    section.optionalVariable("USE_RECIPROCAL_SPARSE_FORM", 0, r->reciprocalLi);
  }

  // Read the emissivity/texture parameters
  if (section.optionalVariable("EMISSIVITY_FILE", 0, r->emissivityFilename);
      !r->emissivityFilename.empty()) {
    if (section.optionalVariable("TEXTURE_FILE", 0, r->textureFilename);
        !r->textureFilename.empty()) {
      tag::ParseArray(
        section.optionalVariable<std::string>("TEXTURE_BANDPASS"),
        r->textureBandpass);
      tag::ParseArray(
        section.optionalVariable<std::string>("TEXTURE_STATS"),
        r->textureStats);
    }
  } else if (section.elementCount("TEXTURE_FILE") > 0) {
    DIRSIG_THROW_FMT(
      "EMISSIVITY_FILE and TEXTURE_FILE must be supplied together");
  }

  // If they used a fit file, that takes precendence, so parse and return
  if (auto fCount = section.elementCount("BRDF_FIT_FILE"); fCount > 0) {
    if (fCount > 1) DIRSIG_THROW_FMT("Only one BRDF_FIT_FILE allowed");
    section.optionalVariable("BRDF_FIT_FILE", 0, r->fitFilename);
    return r;
  }

  // Go through brdf fit sections
  auto fitCount = section.elementCount("BRDF_FIT");
  for (size_t i = 0; i < fitCount; i++) {
    auto& s = section.requiredSubsection("BRDF_FIT", i);
    auto& params = r->fitParameters.emplace_back();
    s.requiredVariable("LAMBDA", 0, params.wavelengthMicron);
    s.requiredVariable("FISO", 0, params.fIso);
    s.requiredVariable("FVOL", 0, params.fVol);
    s.requiredVariable("FGEO", 0, params.fGeo);
  }

  // look for the (very optional) range based coeffs
  size_t rangeCount = section.elementCount("RANGE_PARAMS");
  if (rangeCount > 0u && fitCount == 0u) {
    DIRSIG_THROW_FMT(
      "RANGE_PARAMS can only be used to override angles in an existing "
      "setup, make sure there are BRDF fit parameters defined.);");
  }
  for (size_t i = 0; i < rangeCount; i++) {
    std::array<std::string, 5> tokens;
    tag::ParseArray(
      section.requiredVariable<std::string>("RANGE_PARAMS", i), tokens);
    r->rangeParameters.push_back(
      {atoi(tokens[0].c_str()), atoi(tokens[1].c_str()),
       atoi(tokens[2].c_str()), atoi(tokens[3].c_str()), tokens[4]});
  }

  return r;
}

// Load SphericalDataBrdf from tag file section
static std::unique_ptr<ReflectanceProperty> LoadSphericalDataBrdf(
  const TagFileSection& section) {
  auto r = std::make_unique<SphericalDataBrdf>();
  section.requiredVariable("DATA_FILENAME", 0, r->sqtFilename);
  section.optionalVariable("DHR_FILENAME", 0, r->dhrFilename);
  section.optionalVariable("NORMALIZE_DHR", 0, r->normalizeDhr);
  return r;
}

static std::unique_ptr<ReflectanceProperty> LoadReflectancePropertyByName(
  const std::string& reflPropName, TagFileSection& reflPropSec) {
  if (reflPropName == "WardBRDF") {
    return LoadWardBrdf(reflPropSec);
  } else if (
    reflPropName == "SimpleReflectance" || reflPropName == "SimpleReflection") {
    return LoadSimpleReflectance(reflPropSec);
  } else if (
    reflPropName == "DeltaReflectance" || reflPropName == "DeltaReflection") {
    return LoadDeltaReflectance(reflPropSec);
  } else if (reflPropName == "PriestGermer") {
    return LoadPriestGermer(reflPropSec);
  } else if (reflPropName == "ShellTarget") {
    return LoadShellTarget(reflPropSec);
  } else if (reflPropName == "ShellBackground") {
    return LoadShellBackground(reflPropSec);
  } else if (reflPropName == "ImageMapped") {
    return LoadReflectanceMap(reflPropSec);
  } else if (reflPropName == "RossLi") {
    return LoadRossLi(reflPropSec);
  } else if (reflPropName == "RgbImage") {
    return LoadRgbImage(reflPropSec);
  } else if (reflPropName == "CurveMap") {
    return LoadCurveMap(reflPropSec);
  } else if (reflPropName == "SphericalDataBRDF") {
    return LoadSphericalDataBrdf(reflPropSec);
  } else if (reflPropName == "PMFresnelBRDF") {
    return std::make_unique<PMFresnelBrdf>();
  } else {
    DIRSIG_THROW_FMT("unsupported reflectance type: {}", reflPropName);
  }
}

// Load ClassicExtinction from tag file section.
static std::unique_ptr<ExtinctionProperty> LoadClassicExtinction(
  TagFileSection& section) {
  auto e = std::make_unique<ClassicExtinction>();
  section.requiredVariable<std::string>("FILENAME", 0, e->filename);
  return e;
}

// Load SimpleTransmission from tag file section.
static std::unique_ptr<ExtinctionProperty> LoadSimpleTransmission(
  TagFileSection& section) {
  auto t = std::make_unique<SimpleTransmission>();
  t->format = SimpleSpectralData::Format::XML;
  LoadSimpleSpectralDataMembers(section, t);
  return t;
}

// Load SimpleTransmission from tag file section.
static std::unique_ptr<ExtinctionProperty> LoadDeltaTransmission(
  TagFileSection& section) {
  auto t = std::make_unique<DeltaTransmission>();
  LoadSimpleSpectralDataMembers(section, t);
  return t;
}

// Load SphericalDataBtdf from tag file section
static std::unique_ptr<ExtinctionProperty> LoadSphericalDataBtdf(
  TagFileSection& section) {
  auto e = std::make_unique<SphericalDataBtdf>();
  section.requiredVariable("DATA_FILENAME", 0, e->sqtFilename);
  section.optionalVariable("DHR_FILENAME", 0, e->dhrFilename);
  section.optionalVariable("NORMALIZE_DHR", 0, e->normalizeDhr);
  return e;
}

static auto LoadThermMaterialProps(
  const std::string& filename, TagFileSection& section) {
  auto loadRequiredPositive{[&](const char* name, auto& v) {
    v = section.requiredVariable<double>(name);
    if (!(v > 0))
      DIRSIG_WARN(
        "{}: line {}: has {} = {}, but this should be strictly positive",
        ShowFilename(filename), section.lineNumber(), name, v);
  }};
  auto loadOptionalPositive{[&](const char* name, auto& v) {
    if (section.optionalVariable(name, 0, v).found && !(v > 0))
      DIRSIG_WARN(
        "{}: line {}: has {} = {}, but this should be strictly positive",
        ShowFilename(filename), section.lineNumber(), name, v);
  }};
  auto loadOptionalPositive2{[&](const char* name, auto& v) {
    if (section.optionalVariable(name, 0, v).found && !((v > 0) || (v == -1)))
      DIRSIG_WARN(
        "{}: line {}: has {} = {}, but this should be strictly positive",
        ShowFilename(filename), section.lineNumber(), name, v);
  }};
  ThermMaterialProps p;
  loadRequiredPositive("SPECIFIC_HEAT", p.specificHeat);
  loadRequiredPositive("MASS_DENSITY", p.massDensity);
  loadRequiredPositive("THERMAL_CONDUCTIVITY", p.thermalConductivity);
  p.exposedArea =
    section.requiredVariable<double>("EXPOSED_AREA");  // This can be negative
  loadOptionalPositive2("VISIBLE_EMISSIVITY", p.solarAbsorption);
  loadOptionalPositive2("SOLAR_ABSORPTION", p.solarAbsorption);
  loadOptionalPositive2("THERMAL_EMISSIVITY", p.thermalEmissivity);
  loadOptionalPositive("THICKNESS", p.thickness);  // [cm]
  loadOptionalPositive("SELF_GENERATED_POWER", p.selfGeneratedPower);
  return p;
}

// Load ThermTemperatureSolver from tag file section.
static auto LoadThermTemperatureSolver(
  const std::string& filename, TagFileSection& tempSolvSec) {
  auto tempSolv = std::make_unique<ThermTemperatureSolver>();
  tempSolv->material = LoadThermMaterialProps(filename, tempSolvSec);
  tempSolvSec.optionalVariable(
    "ENABLE_SKY_SHADOWS", 0, tempSolv->enableSkyShadows);
  tempSolvSec.optionalVariable(
    "ENABLE_TIMESTEP_OPTIMIZATION", 0, tempSolv->enableTimestepOptimization);
  tempSolvSec.optionalVariable(
    "MINIMUM_OPTIMIZED_TIMESTEPS", 0, tempSolv->minimumOptimizedTimesteps);
  return tempSolv;
}

// Load BalfourTemperatureSolver from tag file section.
static std::unique_ptr<TemperatureSolver> LoadBalfourTemperatureSolver(
  TagFileSection& section) {
  auto t = std::make_unique<BalfourTemperatureSolver>();
  t->coefficients[0] = section.requiredVariable<double>("A0");
  t->coefficients[1] = section.requiredVariable<double>("A1");
  t->coefficients[2] = section.requiredVariable<double>("A2");
  t->coefficients[3] = section.requiredVariable<double>("A3");
  t->coefficients[4] = section.requiredVariable<double>("A4");
  section.optionalVariable("A5", 0, t->coefficients[5]).Or(0);
  return t;
}

// Load MappedBalfourTemperatureSolver from tag file section.
static std::unique_ptr<TemperatureSolver> LoadMappedBalfourTemperatureSolver(
  TagFileSection& section) {
  auto t = std::make_unique<MappedBalfourSolver>();
  auto& map = t->map;
  section.optionalVariable("NAME", 0, map.name);
  map.source = ImageSource(
    ImageSource::Type::Filename,
    section.requiredVariable<std::string>("IMAGE_FILENAME"));
  map.projector.ReadTag(section);
  auto enabledStr = section.optionalVariable<std::string>("ENABLED");
  map.enabled = enabledStr.empty() || enabledStr == "TRUE";
  auto type = section.optionalVariable<std::string>("TYPE");
  if (type == "Iterative") map.mapType = BalfourMapEntry::Type::Iterative;
  return t;
}

// Load MappedThermTemperatureSolver from tag file section.
static std::unique_ptr<TemperatureSolver> LoadMappedThermTemperatureSolver(
  const std::string& filename, TagFileSection& section) {
  auto t = std::make_unique<MappedThermSolver>();
  auto& map = t->map;
  section.optionalVariable("NAME", 0, map.name);
  map.source = ImageSource(
    ImageSource::Type::Filename,
    section.requiredVariable<std::string>("IMAGE_FILENAME"));
  map.projector.ReadTag(section);
  auto enabledStr = section.optionalVariable<std::string>("ENABLED");
  map.enabled = enabledStr.empty() || enabledStr == "TRUE";
  map.thermDefaults = LoadThermMaterialProps(
    filename, section.requiredSubsection("THERM_DEFAULTS"));
  return t;
}

// Load DataDrivenTemperatureSolver from tag file section.
static std::unique_ptr<TemperatureSolver> LoadDataDrivenTemperatureSolver(
  TagFileSection& tempSolvSec) {
  auto t = std::make_unique<DataDrivenTemperatureSolver>();
  if (!tempSolvSec.optionalVariable("TEMPERATURE", 0, t->tempK).found) {
    t->filename = tempSolvSec.requiredVariable<std::string>("FILENAME");
  }
  return t;
}

// Load TemperatureMapSolver from tag file section
static std::unique_ptr<TemperatureSolver> LoadTemperatureMapSolver(
  TagFileSection& section) {
  auto tempSolv = std::make_unique<TemperatureMapSolver>();
  auto& map = tempSolv->map;
  section.optionalVariable("NAME", 0, map.name);
  tempSolv->map.source = ImageSource(
    ImageSource::Type::Filename,
    section.requiredVariable<std::string>("IMAGE_FILENAME"));
  map.projector.ReadTag(section);
  auto enabledStr = section.optionalVariable<std::string>("ENABLED");
  map.enabled = enabledStr.empty() || enabledStr == "TRUE";

  if (auto opSec = section.subsection("OPTIONS"); opSec != nullptr) {
    opSec->optionalVariable("GAIN", 0, map.gain);
    opSec->optionalVariable("BIAS", 0, map.bias);
    std::string units;
    opSec->optionalVariable("UNITS", 0, units);
    if (units == "Celsius") {
      // Do nothing
    } else if (units == "Kelvin") {
      map.bias -= 273.15;
    } else if (units == "Fahrenheit") {
      map.bias -= 32;
      map.gain *= 5.0 / 9.0;
      map.bias *= 5.0 / 9.0;
    } else if (!units.empty()) {
      DIRSIG_THROW_FMT(
        "Unrecognized units \"{}\" for TemperatureMap! Expected "
        "\"Celsius\", \"Kelvin\", or \"Fahrenheit\"",
        units);
    }
  }
  return tempSolv;
}

static void LoadOldMatProperties(TagFileSection& section, MatFileEntry& entry) {
  auto optDescription =
    section.optionalVariable<std::string>("OPTICAL_DESCRIPTION");
  if (optDescription == "SOURCE") {
    auto& src = entry.source.emplace();

    section.optionalVariable("NAME", 0, src.name);
    section.optionalVariable("ID", 0, src.id);

    // Determine the intensity filename
    section.optionalVariable("INTENSITY_FILENAME", 0, src.intensityFilename);
    if (src.intensityFilename.empty()) {
      section.optionalVariable("RADIANCE_FILENAME", 0, src.intensityFilename);
      if (src.intensityFilename.empty())
        DIRSIG_THROW_FMT("source descriptions must have an INTENSITY_FILENAME");
    }

    // Read in the shaping parameters
    try {
      section.optionalVariable("SOURCE_SHAPE", 0, src.shapeFactor);
    } catch (const NumericParseError<double>&) {
      section.requiredVariable("SOURCE_SHAPE", 0, src.shapeFactorIESFilename);
    }
    if (auto* s = section.subsection("SHAPING_LIST"); s != nullptr) {
      auto entryCount = s->elementCount("SHAPE_MAG_PAIR");
      for (size_t i = 0; i < entryCount; i++) {
        auto sm = tag::ParseArray<double, 2>(
          s->optionalVariable<std::string>("SHAPE_MAG_PAIR", i));
        src.shapeFactorMagnitudes.emplace_back(sm[0], sm[1]);
      }
    }

    section.optionalVariable("NORMALIZE_SHAPE", 0, src.normalizeShape);
    section.optionalVariable("PSD_FILENAME", 0, src.psdFilename);

    if (auto f = section.optionalVariable<double>("BLINK_FREQUENCY", 0, 0);
        f != 0) {
      src.blinkFrequency = f;
      section.optionalVariable("BLINK_OFFSET", 0, src.blinkOffset);
    }

    section.optionalVariable("PHOTON_MAPPED", 0, src.photonMapped);
  } else {
    // currently, all materials default to the classic emissivity model
    auto emsFile = section.optionalVariable<std::string>("EMISSIVITY_FILE");
    if (emsFile.empty()) return;

    auto e = std::make_unique<ClassicEmissivity>();
    e->filename = std::move(emsFile);
    entry.surfaceProperties().setEmissivityProperty(std::move(e));

    if (
      optDescription == "UNIFORM_TRANSMISSION" ||
      optDescription == "NON_UNIFORM_TRANSMISSION") {
      auto ext = std::make_unique<ClassicExtinction>();
      section.requiredVariable("EXTINCTION_FILE", 0, ext->filename);
      entry.surfaceProperties().setExtinctionProperty(std::move(ext));
    }
  }
}

static void LoadClassicTherm(
  const std::string& filename, TagFileSection& entrySec, MatFileEntry& entry) {
  if (entrySec.elementCount("SPECIFIC_HEAT") > 0) {
    entry.setTempSolver(LoadThermTemperatureSolver(filename, entrySec));
  }
}

static std::unique_ptr<MaterialMapEntry> LoadMaterialMap(
  const TagFileSection& section) {
  auto m = std::make_unique<MaterialMapEntry>();
  section.optionalVariable("NAME", 0, m->name);
  section.optionalVariable("ENABLED", 0, m->enabled);
  m->projector.ReadTag(section);
  m->source = ImageSource(
    ImageSource::Type::Filename,
    section.requiredVariable<std::string>("IMAGE_FILENAME"));

  auto& lut = m->materialLut;
  const auto& lutSec = section.requiredSubsection("LUT");
  for (const auto& entry : lutSec) {
    if (entry.type() != TagFileElement::Type::UnformattedData) {
      DIRSIG_THROW_FMT("material map LUTs can only contain unformatted data");
    }
    auto lutTokens = TokenList(entry.as<TagFileData>().data(), ":");
    if (lutTokens.size() != 2) {
      DIRSIG_THROW_FMT("invalid LUT entry, must be \"DC:label\"");
    }

    auto dc = ParseNumber<int>(lutTokens[0]);
    if (dc < 0 || dc > 255) {
      DIRSIG_THROW_FMT(
        "out-of-range digital count in LUT, must be in [0, 255]");
    }

    lut[static_cast<uint8_t>(dc)] = trimmed(lutTokens[1]);
  }

  const auto& optionsSec = section.requiredSubsection("OPTIONS");
  bool isMixing = true;
  optionsSec.optionalVariable("ENABLE_MIXING", 0, isMixing);
  m->isMixing = isMixing;

  return m;
}

static std::unique_ptr<MixtureMapEntry> LoadMixtureMap(
  const TagFileSection& section) {
  auto m = std::make_unique<MixtureMapEntry>();
  section.optionalVariable("NAME", 0, m->name);
  section.optionalVariable("ENABLED", 0, m->enabled);
  m->projector.ReadTag(section);
  m->source = ImageSource(
    ImageSource::Type::Filename,
    section.requiredVariable<std::string>("IMAGE_FILENAME"));

  auto labelsVar =
    section.requiredVariable<std::string>("BAND_MATERIAL_LABELS");
  for (auto l : TokenList(labelsVar, ",")) {
    m->bandMaterialIds.push_back(trimmed(l));
  }
  return m;
}

static std::unique_ptr<BumpMapEntry> LoadBumpMap(
  const TagFileSection& section) {
  auto m = std::make_unique<BumpMapEntry>();
  section.optionalVariable("NAME", 0, m->name);
  section.optionalVariable("ENABLED", 0, m->enabled);
  m->projector.ReadTag(section);
  m->source = ImageSource(
    ImageSource::Type::Filename,
    section.requiredVariable<std::string>("IMAGE_FILENAME"));

  const auto& optionsSec = section.requiredSubsection("OPTIONS");
  optionsSec.requiredVariable<double>("SCALE", 0, m->scale);
  return m;
}

static std::unique_ptr<NormalMapEntry> LoadNormalMap(
  const TagFileSection& section) {
  auto m = std::make_unique<NormalMapEntry>();
  section.optionalVariable("NAME", 0, m->name);
  section.optionalVariable("ENABLED", 0, m->enabled);
  m->projector.ReadTag(section);
  section.requiredVariable("IMAGE_FILENAME", 0, m->imageFilename);

  auto* optionsSec = section.subsection("OPTIONS");
  if (optionsSec) {
    optionsSec->optionalVariable<float>("SCALE", 0, m->scale);
  }
  return m;
}

static const std::unordered_set<std::string> kUnsupportedRadSolvers{
  "Cloud",
  "Gas",
  "GeoDensity",
  "Grid",
  "PolarimetricCloud",
  "Plume",
  "Portal",
  "SurfacePM",
  "OpenWaterSurface",
  "Medium",
  "MediumPM",
  "FixedMedium",
  "OpenWaterVolume",
  "FixedRadiance",
  "RadianceMap"};

// Load a MatFileEntry
static MatFileEntry LoadMatFileEntry(
  const std::string& filename, TagFileSection& entrySec) {
  MatFileEntry entry;
  entry.setId(entrySec.requiredVariable<std::string>("ID"));
  entry.setName(entrySec.optionalVariable<std::string>("NAME"));
  entry.author = entrySec.optionalVariable<std::string>("AUTHOR");
  entry.organization = entrySec.optionalVariable<std::string>("ORGANIZATION");
  entry.description = entrySec.optionalVariable<std::string>("DESCRIPTION");

  // Bundles have first priority
  if (auto s = entrySec.optionalVariable<std::string>("BUNDLE_FILENAME");
      !s.empty()) {
    entry.bundleFilename = std::move(s);
    return entry;
  }

  // Check for various map types
  if (auto* s = entrySec.subsection("MATERIAL_MAP")) {
    entry.setMaterialMap(LoadMaterialMap(*s));
  }
  if (auto* s = entrySec.subsection("MIXTURE_MAP")) {
    entry.setMixtureMap(LoadMixtureMap(*s));
  }
  if (auto* s = entrySec.subsection("BUMP_MAP")) {
    entry.setBumpMap(LoadBumpMap(*s));
  }
  if (auto* s = entrySec.subsection("NORMAL_MAP")) {
    entry.setNormalMap(LoadNormalMap(*s));
  }

  // Handle the rad solver
  auto radSolverName =
    entrySec.optionalVariable<std::string>("RAD_SOLVER_NAME");
  if (!radSolverName.empty()) {
    if (radSolverName == "Null") {
      entry.setRadSolverType(RadSolverType::Null);
    } else if (radSolverName == "Classic") {
      entry.setRadSolverType(RadSolverType::Classic);
    } else if (radSolverName == "Simple") {
      entry.setRadSolverType(RadSolverType::Simple);
    } else if (radSolverName == "Mirror") {
      entry.setRadSolverType(RadSolverType::Mirror);
    } else if (radSolverName == "Generic") {
      entry.setRadSolverType(RadSolverType::Generic);
    } else if (radSolverName == "Fresnel") {
      entry.setRadSolverType(RadSolverType::Fresnel);
    } else if (kUnsupportedRadSolvers.count(radSolverName) > 0) {
      DIRSIG_WARN(
        "\"{}\" rad solvers are not supported, treating as unknown",
        radSolverName);
      entry.setRadSolverType(RadSolverType::Unknown);
    } else {
      DIRSIG_THROW_FMT("unknown RAD_SOLVER_NAME \"{}\"", radSolverName);
    }
  } else {
    // if we are in the new format, the user must give a rad solver if there
    // are surface properties
    if (auto* s = entrySec.subsection("SURFACE_PROPERTIES"); s != nullptr) {
      DIRSIG_THROW_FMT("a RAD_SOLVER_NAME must be given if there are "
                       "SURFACE_PROPERTIES");
    }

    // Allow for empty entries (dummy/mapped)
    auto* bp = entrySec.subsection("BULK_PROPERTIES");
    auto od = entrySec.optionalVariable<std::string>("OPTICAL_DESCRIPTION");
    auto ef = entrySec.optionalVariable<std::string>("EMISSIVITY_FILE");
    if (!bp && od.empty() && ef.empty()) return entry;

    entry.setRadSolverType(RadSolverType::Classic);
  }

  // Handle bulk properties
  if (auto* s = entrySec.subsection("BULK_PROPERTIES"); s != nullptr) {
    entry.bulkProperties = LoadBulkProperties(*s);
  }

  // Parse the temperature solver if one is there
  if (auto ts = entrySec.optionalVariable<std::string>("TEMP_SOLVER_NAME");
      !ts.empty()) {
    auto& tsSec = entrySec.requiredSubsection("TEMP_SOLVER");
    if (ts == "Therm" || ts == "RitTherm") {
      entry.setTempSolver(LoadThermTemperatureSolver(filename, tsSec));
    } else if (ts == "Balfour") {
      entry.setTempSolver(LoadBalfourTemperatureSolver(tsSec));
    } else if (ts == "MappedBalfour") {
      entry.setTempSolver(LoadMappedBalfourTemperatureSolver(tsSec));
    } else if (ts == "MappedTherm") {
      entry.setTempSolver(LoadMappedThermTemperatureSolver(filename, tsSec));
    } else if (ts == "Fixed" || ts == "DataDriven") {
      entry.setTempSolver(LoadDataDrivenTemperatureSolver(tsSec));
    } else if (ts == "Map") {
      entry.setTempSolver(LoadTemperatureMapSolver(tsSec));
    } else if (ts == "Null") {
    } else {
      DIRSIG_THROW_FMT("unsupported temperature solver type: {}", ts);
    }
  } else {
    LoadClassicTherm(filename, entrySec, entry);
  }

  float thicknessCm = -1;
  entrySec.optionalVariable("THICKNESS", 0, thicknessCm);
  if (thicknessCm != -1) {
    entry.setThicknessCm(thicknessCm);
  }

  if (entry.radSolverType() == RadSolverType::Mirror) return entry;

  // Get the surface properties section(s)
  bool anySurfProp = false;
  for (size_t surfPropIndex = 0; true; surfPropIndex++)
    if (auto* surfPropSec =
          entrySec.subsection("SURFACE_PROPERTIES", surfPropIndex);
        surfPropSec) {
      anySurfProp = true;
      auto& surfProps = entry.surfaceProperties(surfPropIndex);
      surfPropSec->optionalVariable(
        "REFLECTANCE_PROP_WEIGHT", 0, surfProps.reflPropWeight);
      surfPropSec->optionalVariable(
        "EMISSIVITY_PROP_WEIGHT", 0, surfProps.reflPropWeight);
      surfPropSec->optionalVariable(
        "TRANSMITTANCE_PROP_WEIGHT", 0, surfProps.tranPropWeight);

      // Check for reflectance properties.
      std::string reflPropName;
      if (surfPropSec
            ->optionalVariable("REFLECTANCE_PROP_NAME", 0, reflPropName)
            .found)
        surfProps.setReflectanceProperty(LoadReflectancePropertyByName(
          reflPropName, surfPropSec->requiredSubsection("REFLECTANCE_PROP")));

      // Check for emissivity properties
      std::string emsPropName;
      if (surfPropSec->optionalVariable("EMISSIVITY_PROP_NAME", 0, emsPropName)
            .found) {
        auto& emsPropSec = surfPropSec->requiredSubsection("EMISSIVITY_PROP");
        if (emsPropName == "ClassicEmissivity")
          surfProps.setEmissivityProperty(LoadClassicEmissivity(emsPropSec));
        else
          DIRSIG_THROW_FMT("unsupported emissivity type: {}", emsPropName);
      }

      // Check for extinction properties
      std::string extPropName;
      if (surfPropSec->optionalVariable("EXTINCTION_PROP_NAME", 0, extPropName)
            .found) {
        auto& extPropSec = surfPropSec->requiredSubsection("EXTINCTION_PROP");
        if (extPropName == "ClassicExtinction")
          surfProps.setExtinctionProperty(LoadClassicExtinction(extPropSec));
        else
          DIRSIG_THROW_FMT("unsupported extinction type: {}", extPropName);
      }

      // Check for transmission properties
      std::string tranPropName;
      if (surfPropSec
            ->optionalVariable("TRANSMITTANCE_PROP_NAME", 0, tranPropName)
            .found) {
        auto& tranPropSec =
          surfPropSec->requiredSubsection("TRANSMITTANCE_PROP");
        if (
          tranPropName == "SimpleTransmittance" ||
          tranPropName == "SimpleTransmission")
          surfProps.setExtinctionProperty(LoadSimpleTransmission(tranPropSec));
        else if (
          tranPropName == "DeltaTransmittance" ||
          tranPropName == "DeltaTransmission")
          surfProps.setExtinctionProperty(LoadDeltaTransmission(tranPropSec));
        else if (tranPropName == "WardBTDF")
          surfProps.setExtinctionProperty(std::make_unique<WardBtdf>());
        else if (tranPropName == "PMFresnelBTDF")
          surfProps.setExtinctionProperty(std::make_unique<PMFresnelBtdf>());
        else if (tranPropName == "SphericalDataBTDF")
          surfProps.setExtinctionProperty(LoadSphericalDataBtdf(tranPropSec));
        else
          DIRSIG_THROW_FMT("unsupported transmission type: {}", tranPropName);
      }
    } else
      break;

  if (not anySurfProp) LoadOldMatProperties(entrySec, entry);

  return entry;
}

std::unique_ptr<MatFile> DoLoadFromFile(
  const std::string& filename,
  const std::string& specificId,
  MatFile::ErrorMode errorMode,
  size_t depth) {
  if (depth == 10) {
    DIRSIG_THROW_FMT("maximum material file include depth exceeded");
  }

  auto RaiseError = [&](const std::string& e) {
    if (errorMode == MatFile::ErrorMode::Permissive) {
      std::cerr << "Warning: " << e << '\n';
    } else {
      DIRSIG_THROW_FMT(e);
    }
  };

  std::ifstream ifs;
  try {
    ifs = OpenIfstream(filename);
  } catch (...) {
    if (errorMode == MatFile::ErrorMode::Permissive) {
      DIRSIG_WARN("Can't open \"{}\", skipping", filename);
      return nullptr;
    } else
      throw;
  }

  // Make a file locator that can look relative to this file
  auto locator = FileLocator::MakeRelativeToFile(filename);

  auto matFile = std::make_unique<MatFile>();
  auto tagFile = TagFile::LoadFromIstream(ifs);
  auto& tagFileRoot = tagFile->topSection();

  // Loop over all entries in the file
  for (size_t entryIndex = 0;; entryIndex++) {
    auto* entrySec = tagFileRoot.subsection("MATERIAL_ENTRY", entryIndex);
    if (!entrySec) break;

    // If looking for specific ID, skip any non-matching IDs.
    if (
      !specificId.empty() &&
      specificId != entrySec->requiredVariable<std::string>("ID"))
      continue;

    try {
      // Try to load the material.
      matFile->entries().emplace_back(LoadMatFileEntry(filename, *entrySec));
    } catch (const std::runtime_error& err) {
      RaiseError(fmt::format(
        "cannot load material ID '{}': {}",
        entrySec->requiredVariable<std::string>("ID"), err.what()));
    }

    // If this was the only ID we cared about, bail out
    if (!specificId.empty()) break;
  }

  // If looking for specific ID and we found it, we are done, no need to look
  // at includes
  if (!specificId.empty() && !matFile->entries().empty()) {
    return matFile;
  }

  // Process all includes in the file
  for (size_t incIndex = 0;; incIndex++) {
    auto incElem = tagFileRoot.element("INCLUDE_FILE", incIndex);
    if (incElem == nullptr) break;
    if (incElem->type() != TagFileElement::Variable) {
      RaiseError(fmt::format(
        "line {}: INCLUDE_FILE entries must be variables",
        incElem->lineNumber()));
      continue;
    }

    auto incFilename = incElem->as<TagFileVariable>().get<std::string>();

    auto f = locator.locateFile(incFilename);
    if (f.empty()) {
      RaiseError(fmt::format(
        "line {}: unable to locate include file \"{}\"", incElem->lineNumber(),
        incFilename));
      continue;
    }

    std::unique_ptr<MatFile> incFile;
    try {
      incFile = DoLoadFromFile(
        f, std::string(), MatFile::ErrorMode::Default, depth + 1);
    } catch (const std::exception& e) {
      RaiseError(fmt::format(
        "line {}: while processing included file \"{}\":\n{}",
        incElem->lineNumber(), f, e.what()));
      continue;
    }

    matFile->includes().push_back(std::move(f));

    // If we were looking for a specific ID and we found it, grab it an
    // we're done. Otherwise, add all of the entries
    if (!specificId.empty()) {
      for (auto& e : incFile->entries())
        if (e.id() == specificId) {
          matFile->entries().push_back(std::move(e));
          return matFile;
        }
    } else {
      for (auto& e : incFile->entries())
        matFile->entries().push_back(std::move(e));
      for (auto& i : incFile->includes())
        matFile->includes().push_back(std::move(i));
    }
  }

  // If looking for specific ID, but didn't find it, report an error
  if (!specificId.empty() && matFile->entries().empty()) {
    RaiseError(fmt::format("cannot find material ID '{}'", specificId));
  }

  return matFile;
}

std::unique_ptr<MatFile> MatFile::LoadFromFile(
  const std::string& filename,
  const std::string& specificId,
  ErrorMode errorMode) {
  return DoLoadFromFile(filename, specificId, errorMode, 0);
}
}  // namespace d5::d4legacy
