/** Implementation file for ReflectanceProperty.h
 *
 *  \file ReflectanceProperty.cpp
 */

#include "dirsig5/d4legacy/mat/ReflectanceProperty.h"

#include <fstream>

#include "../misc/LineReader.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/d4legacy/tag/TagFile.h"

namespace d5::d4legacy {

std::vector<PriestGermer::SpectralParameters> ReadPriestGermerParametersFile(
  const char* filename) {
  auto ifs = OpenIfstream(filename);
  LineReader lineReader(ifs);

  std::vector<PriestGermer::SpectralParameters> params;
  while (lineReader.nextLineOptional()) {
    if (lineReader.line.empty() || lineReader.line[0] == '#') {
      continue;  // Skip blank lines and comments
    }
    std::istringstream line(lineReader.line);
    double wavelength;
    double paramS;
    double paramN;
    double paramK;
    if (!(line >> wavelength >> paramS >> paramN >> paramK)) {
      lineReader.fail(
        "PriestGermer format should be: \"wavelength paramS paramN paramK\"");
    }
    params.push_back({wavelength, paramS, paramN, paramK});
  }
  // Sort by wavelength.
  std::sort(params.begin(), params.end(), [](const auto& lhs, const auto& rhs) {
    return lhs.wavelengthMicron < rhs.wavelengthMicron;
  });
  return params;
}

std::vector<ShellTarget::FitParams> ReadShellTargetFitFile(
  const char* filename) {
  auto ifs = OpenIfstream(filename);
  auto tagFile = TagFile::LoadFromIstream(ifs);
  auto& tagFileRoot = tagFile->topSection();
  if (tagFileRoot.requiredVariable<std::string>("SHELL_TARGET") != "1.0") {
    DIRSIG_THROW_FMT("invalid file version");
  }

  std::vector<ShellTarget::FitParams> fitParams(
    tagFileRoot.elementCount("FIT_PARAMS"));
  std::size_t fitSecIndex = 0;
  for (auto& fit : fitParams) {
    auto& fitSec = tagFileRoot.requiredSubsection("FIT_PARAMS", fitSecIndex++);
    fit.wavelengthMicrons = fitSec.requiredVariable<double>("LAMBDA");
    fit.n = fitSec.requiredVariable<double>("N");
    fit.k = fitSec.requiredVariable<double>("K");
    fit.dhr = fitSec.requiredVariable<double>("DHR");
    fitSec.optionalVariable("TGAIN", 0, fit.tgain);
    auto orientName = fitSec.requiredVariable<std::string>("ORIENT_PROB_NAME");
    if (orientName == "Cauchy") {
      fit.orient = ShellTarget::Orient::Cauchy;
    } else if (orientName == "Gaussian") {
      fit.orient = ShellTarget::Orient::Gaussian;
    } else {
      DIRSIG_THROW_FMT("unknown ORIENT_PROB_NAME");
    }

    auto& orientSec = fitSec.requiredSubsection("ORIENT_PROB", 0);
    fit.orientMu = orientSec.requiredVariable<double>("BIAS");
    fit.orientSigma = orientSec.requiredVariable<double>("SIGMA");
    if (
      fitSec.requiredVariable<std::string>("SHADOW_FUNCT_NAME") !=
      "Maxwell-Beard") {
      DIRSIG_THROW_FMT("unknown SHADOW_FUNCT_NAME");
    }

    auto& shadowSec = fitSec.requiredSubsection("SHADOW_FUNCT", 0);
    fit.shadowFuncTau = shadowSec.requiredVariable<double>("TAU");
    fit.shadowFuncOmega = shadowSec.requiredVariable<double>("OMEGA");
    if (
      fitSec.requiredVariable<std::string>("VOLUME_TERM_NAME") !=
      "Maxwell-Beard") {
      DIRSIG_THROW_FMT("unknown VOLUME_TERM_NAME");
    }

    auto& volumeSec = fitSec.requiredSubsection("VOLUME_TERM", 0);
    fit.volumeFuncRhoD = volumeSec.requiredVariable<double>("RHO_D");
    fit.volumeFuncRhoV = volumeSec.requiredVariable<double>("RHO_V");
  }

  return fitParams;
}

ShellBackground::FitParameters ReadShellBackgroundFitFile(
  const char* filename) {
  auto ifs = OpenIfstream(filename);
  auto tagFile = TagFile::LoadFromIstream(ifs);
  auto& tagFileRoot = tagFile->topSection();
  if (tagFileRoot.requiredVariable<std::string>("SHELL_BACKGROUND") != "1.0") {
    DIRSIG_THROW_FMT("invalid file version");
  }

  ShellBackground::FitParameters params;

  // Many of these parameters concern polarization, which is not
  // modeled in DIRSIG5. We only need A, B coefficients to resolve the
  // unpolarized term. However, we also need GSD_INCHES, a new parameter
  // not originally in DIRSIG4, where it was available implicitly. To be
  // backward compatible, GSD_INCHES is optional. If unspecified, it
  // defaults as assigned in d4legacy/mat/ReflectanceProperty.h.
  auto& sharedFitSec = tagFileRoot.requiredSubsection("SHARED_FIT_PARAMS");
  sharedFitSec.optionalVariable("RHO_0", 0, params.shared.rho0);
  sharedFitSec.optionalVariable("P1", 0, params.shared.p1);
  sharedFitSec.optionalVariable("P2", 0, params.shared.p2);
  sharedFitSec.optionalVariable("P3", 0, params.shared.p3);
  sharedFitSec.optionalVariable("P4", 0, params.shared.p4);
  sharedFitSec.optionalVariable("PF1", 0, params.shared.pf1);
  sharedFitSec.optionalVariable("PF2", 0, params.shared.pf2);
  sharedFitSec.optionalVariable("PF3", 0, params.shared.pf3);
  sharedFitSec.optionalVariable("PF4", 0, params.shared.pf4);
  sharedFitSec.requiredVariable("A", 0, params.shared.a);
  sharedFitSec.requiredVariable("B", 0, params.shared.b);
  sharedFitSec.optionalVariable("C", 0, params.shared.c);
  sharedFitSec.optionalVariable("D", 0, params.shared.d);
  sharedFitSec.optionalVariable("E", 0, params.shared.e);
  sharedFitSec.optionalVariable("F", 0, params.shared.f);
  sharedFitSec.optionalVariable("GSD_INCHES", 0, params.shared.gsdInches);

  params.spectral.resize(tagFileRoot.elementCount("FIT_PARAMS"));
  std::size_t fitSecIndex = 0;
  for (auto& fit : params.spectral) {
    auto& fitSec = tagFileRoot.requiredSubsection("FIT_PARAMS", fitSecIndex++);
    fitSec.requiredVariable("RHO", 0, fit.rho);
    fitSec.requiredVariable("K0", 0, fit.k0);
    fitSec.requiredVariable("K1", 0, fit.k1);
    fitSec.requiredVariable("K2", 0, fit.k2);

    // Fit parameters are sorted and interpolated by reflectance, so
    // the wavelength of each measurement is just metadata.
    fitSec.optionalVariable("LAMBDA", 0, fit.wavelengthMicrons);
  }

  return params;
}

std::vector<RossLi::FitParameters> ReadRossLiFitFile(const char* filename) {
  auto ifs = OpenIfstream(filename);
  LineReader lineReader(ifs);

  std::vector<RossLi::FitParameters> params;
  while (lineReader.nextLineOptional()) {
    if (lineReader.line.empty() || lineReader.line[0] == '#') {
      continue;  // Skip blank lines and comments
    }

    std::istringstream line(lineReader.line);
    double wavelength, fIso, fVol, fGeo;
    if (!(line >> wavelength >> fIso >> fVol >> fGeo)) {
      lineReader.fail("RossLi format should be: \"wavelength fIso fVol fGeo\"");
    }
    params.push_back({wavelength, fIso, fVol, fGeo});
  }
  // Sort by wavelength.
  std::sort(params.begin(), params.end(), [](const auto& lhs, const auto& rhs) {
    return lhs.wavelengthMicron < rhs.wavelengthMicron;
  });
  return params;
}
}  // namespace d5::d4legacy
