/** Implementation file for SimpleSpectralData.h
 *
 *  \file SimpleSpectralData.cpp
 */

#include "dirsig5/d4legacy/mat/SimpleSpectralData.h"

#include <fstream>
#include <string>

#include "../misc/LineReader.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/Filesystem.h"
#include "dirsig5/base/StringUtils.h"
#include "dirsig5/d4legacy/misc/Units.h"
#include "dirsig5/d4legacy/xml/xml.h"
#include "dirsig5/envi/SpectralLibrary.h"
#include "dirsig5/base/Logger.h"

namespace d5::d4legacy {

enum SpectralUnits { Microns, Nanometers, Wavenumbers };

std::vector<double> ReadSpectralSamples(const char* filename) {
  auto ifs = OpenIfstream(filename);

  // figure out the spectral units
  SpectralUnits units;
  std::string line;
  std::getline(ifs, line);
  if (line == "microns")
    units = Microns;
  else if (line == "nanometers")
    units = Nanometers;
  else if (line == "wavenumbers")
    units = Wavenumbers;
  else
    DIRSIG_THROW_FMT(R"(unknown spectral units in "{}": "{}")", filename, line);

  // read the rest of the file, which is the points to use
  std::vector<double> samples;
  while (std::getline(ifs, line)) {
    if (line.empty()) continue;
    double value = std::stof(line);
    if (units == Microns)
      samples.push_back(value);
    else if (units == Nanometers)
      samples.push_back(value / 1000.0);
    else if (units == Wavenumbers)
      samples.push_back(10000.0 / value);
  }

  // make sure the points are sorted and unique
  std::sort(samples.begin(), samples.end());
  auto last = std::unique(
    samples.begin(), samples.end(),
    [](const double& a, const double& b) { return std::abs(a - b) < 1e-5; });
  samples.erase(last, samples.end());
  return samples;
}

static SimpleSpectralData ReadText(const char* filename) {
  auto ifs = OpenIfstream(filename);
  SimpleSpectralData ssd;
  LineReader lineReader(ifs);
  while (lineReader.nextLineOptional()) {
    if (lineReader.line.empty() || lineReader.line[0] == '#') continue;

    std::istringstream line(lineReader.line);
    double wavelength = 0;
    double spectralValue = 0;
    if (!(line >> wavelength >> spectralValue)) {
      lineReader.fail("format should be: \"wavelength value\"");
    }

    ssd.wavelengthsMicron.push_back(wavelength);
    ssd.spectrum.push_back(spectralValue);
  }

  return ssd;
}

static SimpleSpectralData ReadXml(const char* filename) {
  // load the file as XML
  auto doc = xml::loadFile(filename);
  auto root = doc.document_element();
  if (std::string_view(root.name()) != "spectraldata")
    DIRSIG_THROW_FMT("root element should be <spectraldata>");

  std::vector<std::pair<float, float>> spectralPairs;
  for (auto node : root.children("point"))
    spectralPairs.emplace_back(
      xml::requiredChild(node, "spectralpoint").text().as_float(),
      xml::requiredChild(node, "value").text().as_float());

  std::string unitsStr = xml::requiredAttribute(root, "spectralunits").value();
  WavelengthUnits units;
  if (!stringToUnits(unitsStr, units))
    DIRSIG_THROW_FMT("unknown spectral units '{}'", unitsStr);
  if (
    units != WavelengthUnits::Microns && units != WavelengthUnits::Nanometers) {
    DIRSIG_WARN(
      "{}: has deprecated '{}' units! Prefer to use microns or nanometers",
      ShowFilename(filename), unitsStr);
  }
  for (auto& [spectralPoint, value] : spectralPairs)
    spectralPoint = toMicrons(spectralPoint, units);

  // Sort the points
  std::sort(spectralPairs.begin(), spectralPairs.end());

  // Split up the pairs and return
  SimpleSpectralData ssd;
  ssd.wavelengthsMicron.reserve(spectralPairs.size());
  ssd.spectrum.reserve(spectralPairs.size());
  for (const auto& p : spectralPairs) {
    ssd.wavelengthsMicron.push_back(p.first);
    ssd.spectrum.push_back(p.second);
  }

  return ssd;
}

static SimpleSpectralData ReadSli(
  const char* filename, const std::variant<int, std::string>& sliIndex) {
  envi::SpectralLibrary sli;
  sli.read(filename, sliIndex);
  SimpleSpectralData ssd;
  ssd.spectrum.resize(sli.spectra[0].size());
  std::copy(sli.spectra[0].begin(), sli.spectra[0].end(), ssd.spectrum.begin());
  ssd.wavelengthsMicron = std::move(sli.wavelength);
  for (double& value : ssd.spectrum) value *= sli.reflectanceScale;
  return ssd;
}

SimpleSpectralData ReadSimpleSpectralDataFromFile(
  const char* filename,
  SimpleSpectralData::Format format,
  const std::variant<int, std::string>& sliIndex,
  span<const float> wavelengthsScene) {
  SimpleSpectralData ssd{};
  switch (format) {
    case SimpleSpectralData::Format::Text:
      ssd = ReadText(filename);
      break;
    case SimpleSpectralData::Format::XML:
      ssd = ReadXml(filename);
      break;
    case SimpleSpectralData::Format::SLI:
      ssd = ReadSli(filename, sliIndex);
      break;
    case SimpleSpectralData::Format::Unknown:
    default:
      throw std::invalid_argument("unknown simple spectral data format");
  }

  // Sanitize!
  if (!wavelengthsScene.empty()) {
    double expectedMinWaveLen = wavelengthsScene.front();
    double expectedMaxWaveLen = wavelengthsScene.back();
    double minWaveLen = ssd.wavelengthsMicron.front();
    double maxWaveLen = ssd.wavelengthsMicron.back();
    if (
      expectedMinWaveLen < minWaveLen - 0.001 ||
      expectedMaxWaveLen > maxWaveLen + 0.001) {
      DIRSIG_WARN(
        "{}: has wavelengths {:1.3f}-{:1.3f}μm, but this is "
        "insufficient for the scene wavelengths {:1.3f}-{:1.3f}μm",
        ShowFilename(filename), minWaveLen, maxWaveLen, expectedMinWaveLen,
        expectedMaxWaveLen);
    }
    int maxOversampling = 0;
    std::vector<int> numSamplesPerBin{};
    numSamplesPerBin.resize(ssd.wavelengthsMicron.size() - 1);
    for (double waveLen : wavelengthsScene) {
      for (size_t i = 0; i < numSamplesPerBin.size(); i++) {
        if (
          ssd.wavelengthsMicron[i] <= waveLen &&
          waveLen < ssd.wavelengthsMicron[i + 1]) {
          numSamplesPerBin[i]++;
          maxOversampling = std::max(maxOversampling, numSamplesPerBin[i]);
          break;
        }
      }
    }
    if (maxOversampling > 10) {
      DIRSIG_WARN(
        "{}: is ~{}x oversampled by the scene wavelengths",
        ShowFilename(filename), maxOversampling);
    }
  }

  return ssd;
}

}  // namespace d5::d4legacy
