#include "dirsig5/d4legacy/misc/FileLocator.h"

#include <algorithm>
#include <sstream>

#include "dirsig5/base/Filesystem.h"

namespace d5::d4legacy {

static std::string ApplyFilters(
  const std::vector<std::function<std::string(const std::string&)>>& filters,
  std::string filename) {
  for (const auto& filter : filters) filename = filter(filename);
  return filename;
}

static std::string LocateFileIn(
  const std::vector<std::string>& dirs, const std::string& filename) {
  for (const auto& dir : dirs) {
    stdfs::path dirPath = dir;
    stdfs::path completePath = dirPath / filename;
    if (stdfs::exists(completePath))
      return stdfs::canonical(completePath).string();
  }
  return std::string();
}

std::string FileLocator::locateFile(const std::string& filename) const {
  stdfs::path filename_path = ApplyFilters(filters_, filename);

  // If we are searching the CWD, then see if the file exists, otherwise, only
  // test only if it is an absolute path
  if (
    (searchCwd_ || filename_path.is_absolute()) && stdfs::exists(filename_path))
    return stdfs::canonical(filename_path).string();

  for (const auto& nameAndPaths : pathLookups_) {
    if (auto s = LocateFileIn(nameAndPaths.second, filename); !s.empty())
      return s;
  }

  return filename;
  /* return std::string(); */
}

std::string FileLocator::locateFile(
  const std::string& filename, FileType type) const {
  if (type == FileType::Any) return locateFile(filename);

  stdfs::path filename_path = ApplyFilters(filters_, filename);

  // If we are searching the CWD, then see if the file exists, otherwise, only
  // test only if it is an absolute path
  if (
    (searchCwd_ || filename_path.is_absolute()) && stdfs::exists(filename_path))
    return stdfs::canonical(filename_path).string();

  if (auto it = find(type); it != end())
    if (auto s = LocateFileIn(it->second, filename); !s.empty()) return s;
  if (auto it = find(Any); it != end())
    if (auto s = LocateFileIn(it->second, filename); !s.empty()) return s;

  return filename;
  /* return std::string(); */
}

std::string FileLocator::requiredFile(const std::string& filename) const {
  if (auto f = locateFile(filename); !f.empty()) return f;
  std::stringstream e;
  e << "unable to locate file: \"" << filename
    << "\"\nlooked in the following directories:\n";
  if (searchCwd_) e << "  " << stdfs::current_path().string() << "\n";
  for (const auto& nameAndPaths : pathLookups_)
    for (const auto& d : nameAndPaths.second) e << "  " << d << "\n";
  throw std::runtime_error(e.str());
}

std::string FileLocator::requiredFile(
  const std::string& filename, FileType type) const {
  if (auto f = locateFile(filename, type); !f.empty()) return f;
  std::stringstream e;
  e << "unable to locate file: \"" << filename
    << "\"\nlooked in the following directories:\n";
  if (searchCwd_) e << "  " << stdfs::current_path().string() << "\n";
  if (auto it = find(type); it != end())
    for (const auto& d : it->second) e << "  " << d << "\n";
  if (type != Any)
    if (auto it = find(Any); it != end())
      for (const auto& d : it->second) e << "  " << d << "\n";
  throw std::runtime_error(e.str());
}

FileLocator FileLocator::MakeRelativeToFile(const std::string& filename) {
  FileLocator loc;
  loc.addPath(
    FileType::Any, {stdfs::absolute(filename).parent_path().string()});
  return loc;
}

}  // namespace d5::d4legacy
