/** Implementation file for dealing with units
 *
 *  \file Units.cpp
 */

#include "dirsig5/d4legacy/misc/Units.h"

#include <algorithm>
#include <cctype>
#include <stdexcept>

namespace d5::d4legacy {
bool stringToUnits(std::string unitString, DistanceUnits& units) {
  std::transform(
    unitString.begin(), unitString.end(), unitString.begin(),
    [](unsigned char c) { return std::toupper(c); });

  if (unitString == "INCHES") {
    units = DistanceUnits::Inches;
  } else if (unitString == "FEET") {
    units = DistanceUnits::Feet;
  } else if (unitString == "CENTIMETERS") {
    units = DistanceUnits::Centimeters;
  } else if (unitString == "METERS") {
    units = DistanceUnits::Meters;
  } else {
    return false;
  }

  return true;
}

std::string to_string(DistanceUnits units) {
  switch (units) {
    case DistanceUnits::Inches:
      return "inches";
    case DistanceUnits::Feet:
      return "feet";
    case DistanceUnits::Centimeters:
      return "centimeters";
    case DistanceUnits::Meters:
      return "meters";
  }
  return "unknown";
}

double toMeters(const DistanceUnits& units) {
  switch (units) {
    case DistanceUnits::Inches:
      return 2.540000e-02;
    case DistanceUnits::Feet:
      return 3.048000e-01;
    case DistanceUnits::Centimeters:
      return 1.000000e-02;
    case DistanceUnits::Meters:
    default:
      return 1.000000e+00;
  }
}

double fromMeters(const DistanceUnits& units) {
  switch (units) {
    case DistanceUnits::Inches:
      return 39.3700787;
    case DistanceUnits::Feet:
      return 3.2808399;
    case DistanceUnits::Centimeters:
      return 100.0000000;
    case DistanceUnits::Meters:
    default:
      return 1.0000000;
  }
}

bool stringToUnits(const std::string& unitString, WavelengthUnits& units) {
  std::string ucUnits = unitString;
  std::transform(
    ucUnits.begin(), ucUnits.end(), ucUnits.begin(),
    [](unsigned char c) { return std::toupper(c); });

  if (ucUnits == "MICRONS" || ucUnits == "MICROMETERS" || ucUnits == "UM") {
    units = WavelengthUnits::Microns;
  } else if (ucUnits == "NANOMETERS" || ucUnits == "NM") {
    units = WavelengthUnits::Nanometers;
  } else if (ucUnits == "WAVENUMBERS") {
    units = WavelengthUnits::Wavenumbers;
  } else if (ucUnits == "GHZ") {
    units = WavelengthUnits::GHz;
  } else {
    return false;
  }

  return true;
}

std::string to_string(WavelengthUnits units) {
  switch (units) {
    case WavelengthUnits::Microns:
      return "microns";
    case WavelengthUnits::Nanometers:
      return "nanometers";
    case WavelengthUnits::Millimeters:
      return "millimeters";
    case WavelengthUnits::Centimeters:
      return "centimeters";
    case WavelengthUnits::Meters:
      return "meters";
    case WavelengthUnits::Angstroms:
      return "angstroms";
    case WavelengthUnits::Wavenumbers:
      return "wavenumbers";
    case WavelengthUnits::GHz:
      return "GHz";
    case WavelengthUnits::MHz:
      return "MHz";
  }
  return "unknown";
}

double toMicrons(double val, const WavelengthUnits& units) {
  switch (units) {
    case WavelengthUnits::Microns:
      break;
    case WavelengthUnits::Nanometers:
      return val * 1e-3;
    case WavelengthUnits::Millimeters:
      return val * 1e3;
    case WavelengthUnits::Centimeters:
      return val * 1e4;
    case WavelengthUnits::Meters:
      return val * 1e6;
    case WavelengthUnits::Angstroms:
      return val * 1e-4;
    case WavelengthUnits::Wavenumbers:
      return 1e+5 / val;
    case WavelengthUnits::GHz:
      return 299792.456 / val;
    case WavelengthUnits::MHz:
      return 299792456 / val;
  }

  return val;
}
}  // namespace d5::d4legacy
