#include "dirsig5/d4legacy/motion/DeltaMotion.h"

#include <fstream>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/StringUtils.h"
#include "dirsig5/d4legacy/tag/TagFile.h"
#include "dirsig5/base/Logger.h"
#include "spdlog/fmt/fmt.h"

namespace d5::d4legacy {

static void ReadInitPositionSection(
  const TagFileSection& section, DeltaMotion& dm) {
  // Some temp vars for extraction
  char comma;
  int day = 0, month = 0, year = 0;
  int hours = 0, minutes = 0;
  double seconds = 0;
  double gmtOffset = 0;

  // extract out the date stuff
  bool haveDate = false;
  if (section.elementCount("DATE") > 0) {
    std::istringstream dateStream(
      section.optionalVariable<std::string>("DATE"));
    dateStream >> month >> day >> year;
    haveDate = true;
  }

  // extract out the time stuff
  bool haveTime = false;
  if (
    section.elementCount("LOCAL_TIME") > 0 ||
    section.elementCount("GMT_TIME") > 0) {
    // make sure we have a GMT_OFFSET
    gmtOffset = -section.requiredVariable<double>("GMT_OFFSET");

    // guess which one is correct
    if (section.elementCount("LOCAL_TIME") > 0) {
      auto localTime = section.requiredVariable<double>("LOCAL_TIME");
      hours = static_cast<int>(localTime);
      minutes = static_cast<int>(60 * (localTime - hours));
      seconds = 3600 * (localTime - hours) - 60 * minutes;
    } else if (section.elementCount("GMT_TIME") > 0) {
      auto gmtTime = section.requiredVariable<double>("GMT_TIME");
      hours = static_cast<int>(gmtTime + gmtOffset);
      minutes = static_cast<int>(60 * (gmtTime - hours + gmtOffset));
      seconds = 3600 * (gmtTime - hours - gmtOffset) - 60 * minutes;
    }
    haveTime = true;
  }

  // check for an ISO 8601 date/time string
  std::string dateTimeString;
  if (section.elementCount("DATE_TIME") > 0) {
    dateTimeString = section.optionalVariable<std::string>("DATE_TIME");
    haveDate = true;
    haveTime = true;
  }

  // make sure the user provided both date and time
  if ((haveDate || haveTime) && (haveDate != haveTime)) {
    DIRSIG_THROW_FMT("Must include both the DATE and TIME, or neither");
  }

  auto& initPos = dm.initialPosition;

  // if we were supplied a reference date and time, store it
  if (haveDate && haveTime) {
    // stuff the reference data/time
    DateTime refDateTime;
    if (dateTimeString.empty() == false) {
      refDateTime.setIso8601(dateTimeString);
    } else {
      refDateTime.set(year, month, day, hours, minutes, seconds, gmtOffset);
    }
    initPos.refDateTime = refDateTime;
  }

  // read starting transform
  std::istringstream trnStream(
    section.requiredVariable<std::string>("TRANSLATION"));
  trnStream >> initPos.translation[0] >> comma >> initPos.translation[1] >>
    comma >> initPos.translation[2];
  std::istringstream sclStream(section.requiredVariable<std::string>("SCALE"));
  sclStream >> initPos.scale[0] >> comma >> initPos.scale[1] >> comma >>
    initPos.scale[2];
  std::istringstream rotStream(
    section.requiredVariable<std::string>("ROTATION"));
  rotStream >> initPos.rotation[0] >> comma >> initPos.rotation[1] >> comma >>
    initPos.rotation[2];
}

static void ReadMovesSection(
  const TagFileSection& section,
  DeltaMotion& dm,
  const char* filename /* Only needed for reference in a warning */) {
  // temporary variables for parsing the line
  char movStr[512];

  // an entry counter used to tell people which entry is bad (if need be)
  int entryCount = 1;

  // loop through all the instances in the section
  for (const auto& element : section) {
    auto& mv = dm.moves.emplace_back();

    // make sure this is a "MOVE = " line
    if (element.tagName() != "MOVE") {
      DIRSIG_THROW_FMT(
        "unexpected tag name in entry #{} in the MOVES section (was "
        "{}, "
        "should have been MOVE)",
        entryCount, element.tagName());
    }

    // grab the value string and trim it
    if (element.type() != TagFileElement::Variable) {
      DIRSIG_THROW_FMT(
        "entry {}: each MOVE in the MOVES section must be a variable",
        entryCount);
    }
    std::string str;
    element.as<TagFileVariable>().get_to(str);
    str.erase(str.find_last_not_of(" \t") + 1);

    // make copy of the value string to parse out the move information
    if (str.size() > 400) {
      DIRSIG_THROW_FMT("MOVE entry #{} is unexpectedly large", entryCount);
    }
    strcpy(movStr, str.c_str());

    // read the delta time from the line
    char* tokenPtr = strtok(movStr, ",");
    if (tokenPtr == nullptr) {
      DIRSIG_THROW_FMT(
        "failed to parse entry #{} in MOVES section", entryCount);
    }
    double deltaTime = std::strtod(tokenPtr, nullptr);

    // check if the deltaTime is = 0 (which is bad!)
    if (deltaTime < 0.0) {
      DIRSIG_THROW_FMT(
        "Found a MOVE entry (#{}) with a delta time < 0", entryCount);
    } else if (deltaTime == 0.0) {
      DIRSIG_WARN(
        "Found a MOVE entry with a delta time = 0!\n"
        "    The delta time was changed to 1 us to avoid a state "
        "inconsistency\n"
        "    Line number = {}\n"
        "    Filename = {}\n",
        element.lineNumber(), filename);
      deltaTime = 1.0e-06;
    }
    mv.dSec = deltaTime;

    // extract the delta translation
    for (int ii = 0; ii < 3; ii++) {
      tokenPtr = strtok(nullptr, ",");
      if (tokenPtr == nullptr) {
        DIRSIG_THROW_FMT(
          "failed to parse entry #{} in MOVES section", entryCount);
      }
      mv.dTranslation[ii] = std::strtod(tokenPtr, nullptr);
    }

    // extract and add the scale to the accumulated scale
    for (int ii = 0; ii < 3; ii++) {
      tokenPtr = strtok(nullptr, ",");
      if (tokenPtr == nullptr) {
        DIRSIG_THROW_FMT(
          "failed to parse entry #{} in MOVES section", entryCount);
      }
      mv.dScale[ii] = std::strtod(tokenPtr, nullptr);
    }

    // extract and add the rotation to the accumulated rotation
    for (int ii = 0; ii < 3; ii++) {
      tokenPtr = strtok(nullptr, ",");
      if (tokenPtr == nullptr) {
        DIRSIG_THROW_FMT(
          "failed to parse entry #{} in MOVES section", entryCount);
      }
      mv.dRotation[ii] += std::strtod(tokenPtr, nullptr);
    }

    entryCount++;
  }
}

void DeltaMotion::ReadMOVFile(const std::string& filename) {
  auto ifs = OpenIfstream(filename);
  auto tagFile = TagFile::LoadFromIstream(ifs);
  auto& tagFileRoot = tagFile->topSection();

  // Read the options
  if (const auto* section = tagFileRoot.subsection("OPTIONS");
      section != nullptr) {
    startHidden =
      toLower(section->optionalVariable<std::string>("START_HIDDEN")) == "true";
    endHidden =
      toLower(section->optionalVariable<std::string>("END_HIDDEN")) == "true";
  }
  ReadInitPositionSection(
    tagFileRoot.requiredSubsection("INIT_POSITION"), *this);
  ReadMovesSection(
    tagFileRoot.requiredSubsection("MOVES"), *this, filename.c_str());
}

std::vector<std::pair<double, DualQuaternion>> DeltaMotionToDualQuaternions(
  const DeltaMotion& motion, const std::optional<DateTime>& refDateTime) {
  std::vector<std::pair<double, DualQuaternion>> dqs;
  dqs.reserve(1 + motion.moves.size());
  const auto& initial = motion.initialPosition;

  double sec = 0;
  if (initial.refDateTime) {
    if (refDateTime) {
      sec = initial.refDateTime->getTotalSeconds() -
            refDateTime->getTotalSeconds();
    } else {
      DIRSIG_THROW(
        "delta motion initial position had a reference date time, so "
        "a reference date/time is also required on the "
        "scene2hdf command line.");
    }
  }

  std::array<double, 3> translation = initial.translation;
  std::array<double, 3> rotation = initial.rotation;
  std::array<double, 3> scale = initial.scale;
  auto makeDqRotation = [&](int axis) {
    double cosTheta = std::cos(rotation[axis] * 0.5);
    double sinTheta = std::sin(rotation[axis] * 0.5);
    double vec[3] = {};
    vec[axis] = sinTheta;
    return DualQuaternion{
      Quaternion{cosTheta, vec[0], vec[1], vec[2]}, Quaternion{0, 0, 0, 0}};
  };
  auto makeDq = [&]() {
    DualQuaternion dq = {
      Quaternion{1, 0, 0, 0},
      Quaternion{
        0, translation[0] * 0.5, translation[1] * 0.5, translation[2] * 0.5}};
    dq = dq * makeDqRotation(0);
    dq = dq * makeDqRotation(1);
    dq = dq * makeDqRotation(2);
    double len = std::sqrt(dq.real.squaredNorm());
    dq.real.x() /= len;
    dq.real.y() /= len;
    dq.real.z() /= len;
    dq.real.w() /= len;
    dq.dual.x() /= len;
    dq.dual.y() /= len;
    dq.dual.z() /= len;
    dq.dual.w() /= len;
    dq.addScaling({scale[0], scale[1], scale[2]});
    dqs.emplace_back(sec, dq);
  };
  makeDq();
  for (const auto& move : motion.moves) {
    // Bump deltas.
    sec += move.dSec;
    translation[0] += move.dTranslation[0];
    translation[1] += move.dTranslation[1];
    translation[2] += move.dTranslation[2];
    rotation[0] += move.dRotation[0];
    rotation[1] += move.dRotation[1];
    rotation[2] += move.dRotation[2];
    scale[0] += move.dScale[0];
    scale[1] += move.dScale[1];
    scale[2] += move.dScale[2];
    if (!(scale[0] > 0 && scale[1] > 0 && scale[2] > 0)) {
      DIRSIG_THROW("Delta motion integrated scale must always be positive");
    }
    makeDq();
  }
  return dqs;
}

}  // namespace d5::d4legacy
