#include "dirsig5/d4legacy/motion/FlexMotion.h"

#include <fstream>
#include <iostream>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/flex_motion/Motions.h"
#include "spdlog/fmt/fmt.h"

namespace d5::d4legacy {

std::unique_ptr<JitterModel> JitterModel::FromXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  std::unique_ptr<JitterModel> jm;
  std::string_view type = xml::requiredAttribute(node, "type").as_string();
  if (type == "correlated")
    jm = std::make_unique<JitterModels::Correlated>();
  else if (type == "normal")
    jm = std::make_unique<JitterModels::Uncorrelated>();
  else
    DIRSIG_THROW_FMT("unknown jitter model type: {}", type);
  jm->ReadXML(node, facade, fileLocator);
  return jm;
}

std::unique_ptr<LocationEngine> LocationEngine::FromXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  std::unique_ptr<LocationEngine> le;
  std::string_view type = xml::requiredAttribute(node, "type").as_string();
  if (type == "fixed")
    le = std::make_unique<LocationEngines::Fixed>();
  else if (type == "straight")
    le = std::make_unique<LocationEngines::Straight>();
  else if (type == "circling")
    le = std::make_unique<LocationEngines::Circling>();
  else if (type == "sgp4")
    le = std::make_unique<LocationEngines::Sgp4>();
  else if (type == "waypoints")
    le = std::make_unique<LocationEngines::Waypoints>();
  else
    DIRSIG_THROW_FMT("unknown location engine type: {}", type);
  le->ReadXML(node, facade, fileLocator);
  return le;
}

std::unique_ptr<RotationEngine> RotationEngine::FromXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  std::unique_ptr<RotationEngine> re;
  std::string_view type = xml::requiredAttribute(node, "type").as_string();
  if (type == "euler")
    re = std::make_unique<RotationEngines::EulerAngles>();
  else if (type == "spin")
    re = std::make_unique<RotationEngines::Spin>();
  else if (type == "quaternions")
    re = std::make_unique<RotationEngines::Quaternions>();
  else if (type == "velocity")
    re = std::make_unique<RotationEngines::Velocity>();
  else if (type == "lookat")
    re = std::make_unique<RotationEngines::LookAt>();
  else
    DIRSIG_THROW_FMT("unknown orientation engine type: {}", type);
  re->ReadXML(node, facade, fileLocator);
  return re;
}

void FlexMotion::ReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  le_.reset();
  re_.reset();
  if (node.child("locationengine")) {
    le_ = LocationEngine::FromXML(
      node.child("locationengine"), facade, fileLocator);
    if (auto jitter = node.child("locationjitter")) {
      if (auto xaxis = jitter.child("xaxis"))
        le_->setJitter(0, JitterModel::FromXML(xaxis, facade, fileLocator));
      if (auto yaxis = jitter.child("yaxis"))
        le_->setJitter(1, JitterModel::FromXML(yaxis, facade, fileLocator));
      if (auto zaxis = jitter.child("zaxis"))
        le_->setJitter(2, JitterModel::FromXML(zaxis, facade, fileLocator));
    }
  }
  if (node.child("orientationengine")) {
    re_ = RotationEngine::FromXML(
      node.child("orientationengine"), facade, fileLocator);
    if (auto jitter = node.child("orientationjitter")) {
      if (auto xaxis = jitter.child("xaxis"))
        re_->setJitter(0, JitterModel::FromXML(xaxis, facade, fileLocator));
      if (auto yaxis = jitter.child("yaxis"))
        re_->setJitter(1, JitterModel::FromXML(yaxis, facade, fileLocator));
      if (auto zaxis = jitter.child("zaxis"))
        re_->setJitter(2, JitterModel::FromXML(zaxis, facade, fileLocator));
    }
  }
}

namespace LocationEngines {

void Fixed::ReadXML(
  const xml::Node& node, const geolib::Facade* facade, const FileLocator&) {
  location = xml::readLocation(xml::requiredChild(node, "location"), facade);
}

void Straight::ReadXML(
  const xml::Node& node, const geolib::Facade* facade, const FileLocator&) {
  headingDegEofN = xml::requiredChild(node, "heading").text().as_double();
  velocityMS = xml::requiredChild(node, "velocity").text().as_double();
  startLocation = geolib::Location(
    xml::readCoordinate(xml::requiredChild(node, "start")), facade);
  startTime = DateTime::fromRelativeSeconds(0);
}

void Circling::ReadXML(
  const xml::Node& node, const geolib::Facade* facade, const FileLocator&) {
  radiusM = xml::requiredChild(node, "radius").text().as_double();
  velocityMS = xml::requiredChild(node, "velocity").text().as_double();
  centerLocation = geolib::Location(
    xml::readCoordinate(xml::requiredChild(node, "center")), facade);
  startTime = DateTime::fromRelativeSeconds(0);
}

void Sgp4::ReadXML(
  const xml::Node& node, const geolib::Facade*, const FileLocator&) {
  auto data = xml::requiredChild(node, "data");
  tleLine1 = xml::requiredChild(data, "tle1").text().as_string();
  tleLine2 = xml::requiredChild(data, "tle2").text().as_string();
  if (tleLine1.size() != 69)
    DIRSIG_THROW_FMT("TLE line #1 was not 69 characters!");
  if (tleLine2.size() != 69)
    DIRSIG_THROW_FMT("TLE line #2 was not 69 characters!");
}

void Waypoints::ReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  auto data = xml::requiredChild(node, "data");
  std::string_view sourceType =
    xml::requiredAttribute(data, "source").as_string();
  if (sourceType == "internal" or sourceType == "external") {
    frameType = xml::requiredAttribute(data, "frame").as_string();
    timeType = xml::requiredAttribute(data, "datetime").as_string();
    delimiter = xml::requiredAttribute(data, "delimiter").as_string();
    loop = data.attribute("loop").as_bool(false);
    if (sourceType == "internal")
      waypoints = LoadWaypointsFromString(data.child_value(), facade);
    else
      externalFilename = xml::readFilename(
        data, "filename", fileLocator, FileLocator::GeometryList);
  } else if (sourceType == "stk_report") {
    stkFilename = xml::readFilename(
      data, "filename", fileLocator, FileLocator::GeometryList);
  } else
    DIRSIG_THROW_FMT("unknown waypoints data source: {}", sourceType);
}

}  // namespace LocationEngines

namespace RotationEngines {

void ConstraintVector::ReadXML(const xml::Node& node) {
  if (!node) return;
  std::string_view typeStr = node.attribute("type").as_string();
  if (typeStr.empty() || typeStr == "constant") {
    std::string_view frameStr =
      xml::requiredAttribute(node, "frame").as_string();
    if (frameStr == "scene")
      option = UseConstantEnu;
    else if (frameStr == "ecef")
      option = UseConstantEcef;
    else
      DIRSIG_THROW_FMT(
        "unknown 'frame' attribute for <constraint>: {}", frameStr);
    auto tokens =
      TokenList(xml::requiredAttribute(node, "vector").as_string(), " ,");
    if (tokens.size() != 3)
      DIRSIG_THROW_FMT("invalid 'vector' attribute for <constraint>");
    up = Vector(
      ParseNumber<double>(tokens[0]), ParseNumber<double>(tokens[1]),
      ParseNumber<double>(tokens[2]));
  } else if (typeStr == "velocity") {
    option = UseVelocity;
  } else {
    DIRSIG_THROW_FMT("unknown 'type' attribute for <constraint>: {}", typeStr);
  }
}

RotationEngine::Frame ParseFrame(const xml::Node& node) {
  std::string_view frame = xml::requiredAttribute(node, "frame").as_string();
  if (frame == "scene") return RotationEngine::Frame::Enu;
  if (frame == "ecef") return RotationEngine::Frame::Ecef;
  DIRSIG_THROW_FMT("unknown Euler angle orientation engine frame: {}", frame);
}

void EulerAngles::ReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  auto data = xml::requiredChild(node, "data");
  std::string_view sourceType = data.attribute("source").as_string();
  frame = ParseFrame(data);
  timeType = xml::requiredAttribute(data, "datetime").as_string();
  delimiter = xml::requiredAttribute(data, "delimiter").as_string();
  loop = data.attribute("loop").as_bool(false);
  rotationOrder =
    ParseRotationOrder(xml::requiredAttribute(data, "order").as_string());
  if (sourceType == "internal")
    eulerAngles = LoadAnglesFromString(data.child_value());
  else if (sourceType == "external")
    externalFilename = xml::readFilename(
      data, "filename", fileLocator, FileLocator::GeometryList);
  else
    DIRSIG_THROW_FMT(
      "unknown Euler angle orientation engine source: {}", sourceType);
}

void Spin::ReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  auto data = xml::requiredChild(node, "data");
  std::string_view sourceType = data.attribute("source").as_string();
  frame = ParseFrame(data);
  timeType = xml::requiredAttribute(data, "datetime").as_string();
  delimiter = xml::requiredAttribute(data, "delimiter").as_string();
  if (sourceType == "internal")
    rates = LoadRatesFromString(data.child_value());
  else
    externalFilename = xml::readFilename(
      data, "filename", fileLocator, FileLocator::GeometryList);
  // see if there was a rotation axis
  std::string_view axisStr = data.attribute("axis").as_string();
  if (!axisStr.empty()) {
    auto tokens = TokenList(axisStr, " ,");
    if (tokens.size() != 3) DIRSIG_THROW_FMT("invalid spin axis");
    axis = Vector(
      ParseNumber<double>(tokens[0]), ParseNumber<double>(tokens[1]),
      ParseNumber<double>(tokens[2]));
  }
}

void Quaternions::ReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  auto data = xml::requiredChild(node, "data");
  std::string_view sourceType = data.attribute("source").as_string();
  if (sourceType == "internal" || sourceType == "external") {
    frame = ParseFrame(data);
    timeType = xml::requiredAttribute(data, "datetime").as_string();
    delimiter = xml::requiredAttribute(data, "delimiter").as_string();
    loop = data.attribute("loop").as_bool(false);
    if (sourceType == "internal")
      quaternions = LoadQuaternionsFromString(data.child_value());
    else
      externalFilename = xml::readFilename(
        data, "filename", fileLocator, FileLocator::GeometryList);
  } else if (sourceType == "stk_report") {
    stkFilename = xml::readFilename(
      data, "filename", fileLocator, FileLocator::GeometryList);
  } else
    DIRSIG_THROW_FMT(
      "unknown quaternion orientation engine source: {}", sourceType);
}

void Velocity::ReadXML(
  const xml::Node& node, const geolib::Facade*, const FileLocator&) {
  auto subNode = node.child("constraint");
  if (!subNode) subNode = node.child("up");
  constraint.ReadXML(subNode);
}

void LookAt::ReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  auto subNode = node.child("constraint");
  if (!subNode) subNode = node.child("up");
  constraint.ReadXML(subNode);
  locEngine = LocationEngine::FromXML(
    xml::requiredChild(node, "locationengine"), facade, fileLocator);
}

}  // namespace RotationEngines

/// Routing helper function for loading from a string
template <typename Func, typename... Args>
decltype(auto) LoadFromString(
  Func&& f, const std::string& data, Args&&... args) {
  return f(data, std::forward<Args>(args)...);
}

/// Routing helper function for loading from a file
template <typename Func, typename... Args>
decltype(auto) LoadFromFile(Func&& f, const char* filename, Args&&... args) {
  auto ifs = d5::OpenIfstream(filename);
  std::string data = std::string(
    std::istreambuf_iterator<char>(ifs), std::istreambuf_iterator<char>());
  return f(data, std::forward<Args>(args)...);
}

std::string to_string(RotationEngine::Frame f) {
  switch (f) {
    case RotationEngine::Frame::Enu:
      return "ENU";
    case RotationEngine::Frame::Ecef:
      return "ECEF";
    default:
      return "unknown";
  }
}

namespace LocationEngines {

// ##############################################
// #   WaypointsLocationEngine Implementation   #
// ##############################################

namespace {

std::vector<Waypoints::LocationEntry> LoadWaypoints(
  const std::string& data, const geolib::Facade* facade, const Waypoints& le) {
  std::vector<Waypoints::LocationEntry> entries;
  if (le.timeType != "relative")
    DIRSIG_THROW_FMT("Only 'relative' date/time is supported (for now)!");
  for (auto line : TokenList(data, "\r\n")) {
    std::string delim = " \t";
    delim += le.delimiter;
    auto tokens = TokenList(line, delim);
    if (tokens.empty()) continue;
    if (
      ((le.frameType == "scene") && (tokens.size() != 4)) ||
      ((le.frameType == "geodetic") && (tokens.size() != 4)) ||
      ((le.frameType == "utm ") && (tokens.size() != 6)) ||
      ((le.frameType == "ecef ") && (tokens.size() != 4)))
      DIRSIG_THROW_FMT("Incorrect number of fields in line!");

    // extract and add the various coordinate values
    auto& entry = entries.emplace_back();
    entry.dateTime =
      DateTime::fromRelativeSeconds(ParseNumber<double>(tokens[0]));
    if (le.frameType == "scene") {
      Point sceneCoord(
        ParseNumber<double>(tokens[1]), ParseNumber<double>(tokens[2]),
        ParseNumber<double>(tokens[3]));
      entry.location = geolib::Location(sceneCoord, facade);
    } else if (le.frameType == "geodetic") {
      GeodeticCoord geodeticCoord(
        ParseNumber<double>(tokens[1]), ParseNumber<double>(tokens[2]),
        ParseNumber<double>(tokens[3]));
      entry.location = geolib::Location(geodeticCoord, facade);
    } else if (le.frameType == "utm") {
      UtmCoord utmCoord(
        ParseNumber<long>(tokens[1]), tokens[2][0],
        ParseNumber<double>(tokens[3]), ParseNumber<double>(tokens[4]),
        ParseNumber<double>(tokens[5]));
      entry.location = geolib::Location(utmCoord, facade);
    } else if (le.frameType == "ecef") {
      EcefCoord ecefCoord(
        ParseNumber<double>(tokens[1]), ParseNumber<double>(tokens[2]),
        ParseNumber<double>(tokens[3]));
      entry.location = geolib::Location(ecefCoord, facade);
    } else
      DIRSIG_THROW_FMT("unknown frame type: {}", le.frameType);
  }

  return entries;
}

}  // namespace

auto Waypoints::LoadWaypointsFromString(
  const std::string& data,
  const geolib::Facade* facade) const -> std::vector<LocationEntry> {
  return LoadFromString(LoadWaypoints, data, facade, *this);
}

auto Waypoints::LoadWaypointsFromExternalFile(
  const char* filename,
  const geolib::Facade* facade) const -> std::vector<LocationEntry> {
  return LoadFromFile(LoadWaypoints, filename, facade, *this);
}

}  // namespace LocationEngines

namespace RotationEngines {

// ####################################################
// #            EulerAngles Implementation            #
// ####################################################

namespace {

std::vector<EulerAngles::EulerAngleEntry> LoadEulerAngles(
  const std::string& data, const EulerAngles& re) {
  std::vector<EulerAngles::EulerAngleEntry> entries;
  if (re.timeType != "relative")
    DIRSIG_THROW_FMT("Only 'relative' date/time is supported (for now)!");
  for (auto line : TokenList(data, "\r\n")) {
    std::string delim = " \t";
    delim += re.delimiter;
    auto tokens = TokenList(line, delim);
    if (tokens.empty()) continue;
    if (tokens.size() != 4)
      DIRSIG_THROW_FMT("incorrect number of fields in line");
    auto& entry = entries.emplace_back();
    entry.dateTime =
      DateTime::fromRelativeSeconds(ParseNumber<double>(tokens.at(0)));
    entry.angles[0] = ParseNumber<double>(tokens.at(1));
    entry.angles[1] = ParseNumber<double>(tokens.at(2));
    entry.angles[2] = ParseNumber<double>(tokens.at(3));
  }

  return entries;
}

}  // namespace

auto EulerAngles::LoadAnglesFromString(const std::string& data) const
  -> std::vector<EulerAngleEntry> {
  return LoadFromString(LoadEulerAngles, data, *this);
}

auto EulerAngles::LoadAnglesFromFile(const char* filename) const
  -> std::vector<EulerAngleEntry> {
  return LoadFromFile(LoadEulerAngles, filename, *this);
}

// #############################################
// #            Spin Implementation            #
// #############################################

namespace {

std::vector<Spin::RateEntry> LoadSpinRates(
  const std::string& data, const Spin& re) {
  std::vector<Spin::RateEntry> entries;
  if (re.timeType != "relative")
    DIRSIG_THROW_FMT("Only 'relative' date/time is supported (for now)!");
  for (auto line : TokenList(data, "\r\n")) {
    std::string delim = " \t";
    delim += re.delimiter;
    auto tokens = TokenList(line, delim);
    if (tokens.empty()) continue;
    if (tokens.size() != 2)
      DIRSIG_THROW_FMT("incorrect number of fields in line");
    auto& entry = entries.emplace_back();
    entry.dateTime =
      DateTime::fromRelativeSeconds(ParseNumber<double>(tokens.at(0)));
    entry.rateRadS = ParseNumber<double>(tokens.at(1));
  }
  return entries;
}

}  // namespace

auto Spin::LoadRatesFromString(const std::string& data) const
  -> std::vector<RateEntry> {
  return LoadFromString(LoadSpinRates, data, *this);
}
auto Spin::LoadRatesFromFile(const char* filename) const
  -> std::vector<RateEntry> {
  return LoadFromFile(LoadSpinRates, filename, *this);
}

// ###################################################
// #            Quaternion Implementation            #
// ###################################################

namespace {

std::vector<Quaternions::QuaternionEntry> LoadQuaternions(
  const std::string& data, const Quaternions& re) {
  std::vector<Quaternions::QuaternionEntry> entries;
  if (re.timeType != "relative")
    DIRSIG_THROW_FMT("Only 'relative' date/time is supported (for now)!");
  for (auto line : TokenList(data, "\r\n")) {
    std::string delim = " \t";
    delim += re.delimiter;
    auto tokens = TokenList(line, delim);
    if (tokens.empty()) continue;
    if (tokens.size() != 5)
      DIRSIG_THROW_FMT("incorrect number of fields in line");
    auto& entry = entries.emplace_back();
    entry.dateTime =
      DateTime::fromRelativeSeconds(ParseNumber<double>(tokens.at(0)));
    entry.quaternion.w() = ParseNumber<double>(tokens.at(1));
    entry.quaternion.x() = ParseNumber<double>(tokens.at(2));
    entry.quaternion.y() = ParseNumber<double>(tokens.at(3));
    entry.quaternion.z() = ParseNumber<double>(tokens.at(4));
  }
  return entries;
}

}  // namespace

auto Quaternions::LoadQuaternionsFromString(const std::string& data) const
  -> std::vector<QuaternionEntry> {
  return LoadFromString(LoadQuaternions, data, *this);
}

auto Quaternions::LoadQuaternionsFromFile(const char* filename) const
  -> std::vector<QuaternionEntry> {
  return LoadFromFile(LoadQuaternions, filename, *this);
}

}  // namespace RotationEngines

namespace JitterModels {

void Correlated::ReadXML(
  const xml::Node& node, const geolib::Facade*, const FileLocator&) {
  auto psd = xml::requiredChild(node, "psd");
  for (auto entry : psd.children("entry")) {
    double frequency = xml::readNumber(entry, "frequency", 0.0);
    double magnitude = xml::readNumber(entry, "magnitude", 1.0);
    double phase = xml::readNumber(entry, "phase", 0.0);
    components.push_back({frequency, magnitude, phase});
  }
}

void Uncorrelated::ReadXML(
  const xml::Node& node, const geolib::Facade*, const FileLocator&) {
  auto values = xml::requiredChild(node, "values");
  seed = xml::readNumber(values, "seed", 0.0);
  mean = xml::readNumber(values, "mean", 0.0);
  stdev = xml::readNumber(values, "standarddeviation", 1.0);
}

}  // namespace JitterModels

FlexMotion SpinMotionToFlexMotion(const SpinMotion& spin) {
  FlexMotion fm;
  auto le = std::make_unique<LocationEngines::Fixed>();
  le->location = geolib::Location(spin.origin);
  auto oe = std::make_unique<RotationEngines::Spin>();
  oe->axis = spin.axis;
  oe->rates.push_back({DateTime::fromRelativeSeconds(0), spin.rate});
  fm.setLocationEngine(std::move(le));
  fm.setRotationEngine(std::move(oe));
  return fm;
}

FlexMotion GenericMotionToFlexMotion(const GenericMotion& motion) {
  FlexMotion fm;
  fm.setLocationEngine(std::make_unique<LocationEngines::Waypoints>());
  fm.setRotationEngine(std::make_unique<RotationEngines::Quaternions>());
  auto& locationEngine = fm.locationEngine<LocationEngines::Waypoints>();
  auto& rotationEngine = fm.rotationEngine<RotationEngines::Quaternions>();
  locationEngine.waypoints.reserve(motion.entries.size());
  rotationEngine.quaternions.reserve(motion.entries.size());
  for (const auto& entry : motion.entries) {
    // Add waypoint
    locationEngine.waypoints.emplace_back();
    locationEngine.waypoints.back().dateTime = entry.dateTime;
    locationEngine.waypoints.back().location = entry.location;
    // Add quaternion
    Quaternion quaternion(MakeRotationMatrix(
      entry.rotation[0], entry.rotation[1], entry.rotation[2],
      motion.rotationOrder));
    rotationEngine.quaternions.emplace_back();
    rotationEngine.quaternions.back().dateTime = entry.dateTime;
    rotationEngine.quaternions.back().quaternion = quaternion;
  }
  switch (motion.rotationFrame) {
    case GenericMotion::RotationFrame::SceneEnu:
      rotationEngine.frame = RotationEngine::Frame::Enu;
      break;
    case GenericMotion::RotationFrame::PlatformEnu:
      DIRSIG_THROW(
        "unsupported rotation frame 'PlatformEnu' for GenericMotion");
      break;
    case GenericMotion::RotationFrame::Ecef:
      rotationEngine.frame = RotationEngine::Frame::Ecef;
      break;
  }
  return fm;
}

template <typename T>
static std::shared_ptr<flex_motion::LocationEngine> LoadRuntimeLocationEngine(
  const T& le) {
  if constexpr (std::is_same_v<T, LocationEngine>) {
    switch (le.type()) {
      case LocationEngine::Type::Circling:
        return LoadRuntimeLocationEngine(
          le.template as<LocationEngines::Circling>());
      case LocationEngine::Type::Sgp4:
        return LoadRuntimeLocationEngine(
          le.template as<LocationEngines::Sgp4>());
      case LocationEngine::Type::Straight:
        return LoadRuntimeLocationEngine(
          le.template as<LocationEngines::Straight>());
      case LocationEngine::Type::Waypoints:
        return LoadRuntimeLocationEngine(
          le.template as<LocationEngines::Waypoints>());
      default:
        DIRSIG_THROW("Unknown location engine!");
        break;
    }
  } else if constexpr (std::is_same_v<T, LocationEngines::Circling>) {
    auto* engine = new flex_motion::LocationEngines::Circling();
    engine->center = le.centerLocation;
    engine->radius = le.radiusM;
    engine->speed = le.velocityMS;
    engine->startDateTime = le.startTime;
    return std::shared_ptr<flex_motion::LocationEngine>(engine);
  } else if constexpr (std::is_same_v<T, LocationEngines::Sgp4>) {
    auto* engine = new flex_motion::LocationEngines::Sgp4();
    const char* tle1 = le.tleLine1.data();  // TODO Safe?
    const char* tle2 = le.tleLine2.data();
    std::copy(tle1, tle1 + 69, &engine->tleLine1[0]);
    std::copy(tle2, tle2 + 69, &engine->tleLine2[0]);
    return std::shared_ptr<flex_motion::LocationEngine>(engine);
  } else if constexpr (std::is_same_v<T, LocationEngines::Straight>) {
    auto* engine = new flex_motion::LocationEngines::StraightPath();
    engine->heading = le.headingDegEofN;
    engine->speed = le.velocityMS;
    engine->startLocation = le.startLocation;
    engine->startDateTime = le.startTime;
    return std::shared_ptr<flex_motion::LocationEngine>(engine);
  } else if constexpr (std::is_same_v<T, LocationEngines::Waypoints>) {
    auto* engine = new flex_motion::LocationEngines::Waypoints();
    for (auto& [timeSec, waypoint] : le.waypoints)
      engine->keyframes.push_back({timeSec.getTotalSeconds(), waypoint});
    return std::shared_ptr<flex_motion::LocationEngine>(engine);
  }
  return {};
}

template <typename T>
static std::shared_ptr<flex_motion::RotationEngine> LoadRuntimeRotationEngine(
  const T& re, const std::shared_ptr<flex_motion::LocationEngine>& le) {
  if constexpr (std::is_same_v<T, RotationEngine>) {
    switch (re.type()) {
      case RotationEngine::Type::EulerAngles:
        return LoadRuntimeRotationEngine(
          re.template as<RotationEngines::EulerAngles>(), le);
      case RotationEngine::Type::Spin:
        return LoadRuntimeRotationEngine(
          re.template as<RotationEngines::Spin>(), le);
      case RotationEngine::Type::Quaternions:
        return LoadRuntimeRotationEngine(
          re.template as<RotationEngines::Quaternions>(), le);
        break;
      case RotationEngine::Type::Velocity:
        return LoadRuntimeRotationEngine(
          re.template as<RotationEngines::Velocity>(), le);
        break;
      case RotationEngine::Type::LookAt:
        return LoadRuntimeRotationEngine(
          re.template as<RotationEngines::LookAt>(), le);
        break;
      default:
        DIRSIG_THROW("Unknown rotation engine!");
        break;
    }
  } else if constexpr (std::is_same_v<T, RotationEngines::EulerAngles>) {
    DIRSIG_THROW(
      "Convering EulerAngles directly from d4legacy is not supported yet!");
  } else if constexpr (std::is_same_v<T, RotationEngines::Quaternions>) {
    auto* engine = new flex_motion::RotationEngines::Quaternions();
    for (auto& [timeSec, quaternion] : re.quaternions)
      engine->keyframes.push_back({timeSec.getTotalSeconds(), quaternion});
    engine->loop = re.loop;
    return std::shared_ptr<flex_motion::RotationEngine>(engine);
  } else if constexpr (std::is_same_v<T, RotationEngines::Spin>) {
    auto* engine = new flex_motion::RotationEngines::Spin();
    for (auto& [timeSec, spinRate] : re.rates)  // TODO Sufficient?
      engine->keyframes.push_back({timeSec.getTotalSeconds(), spinRate, {}});
    engine->spinAxis = re.axis;
    return std::shared_ptr<flex_motion::RotationEngine>(engine);
  } else if constexpr (std::is_same_v<T, RotationEngines::Velocity>) {
    auto* engine = new flex_motion::RotationEngines::Velocity();
    engine->locationEngine = le;
    engine->constraintVector = re.constraint.up;  // TODO Sufficient?
    return std::shared_ptr<flex_motion::RotationEngine>(engine);
  } else if constexpr (std::is_same_v<T, RotationEngines::LookAt>) {
    auto* engine = new flex_motion::RotationEngines::LookAt();
    engine->locationEngine = le;
    engine->trackingEngine = LoadRuntimeLocationEngine(*re.locEngine);
    return std::shared_ptr<flex_motion::RotationEngine>(engine);
  }
  return {};
}

std::shared_ptr<flex_motion::Motion> LoadRuntimeFlexMotion(
  const FlexMotion& fm) {
  auto result = std::make_shared<flex_motion::Motions::Flex>();
  result->locationEngine = LoadRuntimeLocationEngine(fm.locationEngine());
  result->rotationEngine =
    LoadRuntimeRotationEngine(fm.rotationEngine(), result->locationEngine);
  return result;
}

}  // namespace d5::d4legacy
