#include "dirsig5/d4legacy/motion/GenericMotion.h"

#include "dirsig5/base/Error.h"
#include "spdlog/fmt/fmt.h"

using namespace std::string_view_literals;

namespace d5::d4legacy {

void GenericMotion::ReadXML(
  const xml::Node& node, const geolib::Facade* facade) {
  auto data = xml::requiredChild(node, "data");

  // Rotation frame.
  std::string_view rotationFrameStr =
    data.attribute("rotationframe").as_string();
  if (rotationFrameStr == "sceneenu" || rotationFrameStr.empty())
    rotationFrame = GenericMotion::RotationFrame::SceneEnu;
  else if (rotationFrameStr == "platformenu")
    rotationFrame = GenericMotion::RotationFrame::PlatformEnu;
  else if (rotationFrameStr == "ecef")
    rotationFrame = GenericMotion::RotationFrame::Ecef;
  else
    DIRSIG_THROW_FMT("unsupported rotationframe '{}'", rotationFrameStr);

  // Rotation order.
  if (data.attribute("rotationorder"))
    rotationOrder =
      ParseRotationOrder(data.attribute("rotationorder").as_string());

  // Rotation units conversion factor.
  double rotationUnitsFac = 1.0;
  if (data.attribute("angularunits")) {
    if (data.attribute("angularunits").as_string() == "degrees"sv)
      rotationUnitsFac = PI / 180.0;
    else if (data.attribute("angularunits").as_string() != "radians"sv)
      DIRSIG_THROW_FMT(
        "unsupported angularunits '{}'",
        data.attribute("angularunits").as_string());
  }

  // Angle type. Schema requires this is absolute.
  if (
    data.attribute("angletype") and
    data.attribute("angletype").as_string() != "absolute"sv)
    DIRSIG_THROW_FMT(
      "unsupported angletype '{}'", data.attribute("angletype").as_string());

  // Distance units. Schema requires this is meters.
  if (
    data.attribute("spatialunits") and
    data.attribute("spatialunits").as_string() != "meters"sv)
    DIRSIG_THROW_FMT(
      "unsupported spatialunits '{}'",
      data.attribute("spatialunits").as_string());

  for (auto entry : data.children("entry")) {
    auto& nextEntry = entries.emplace_back();

    // Read <datetime>...</datetime>
    auto dateTime = xml::requiredChild(entry, "datetime");
    bool dateTimeIsRelative = true;
    if (dateTime.attribute("type")) {
      if (dateTime.attribute("type").as_string() == "absolute"sv)
        dateTimeIsRelative = false;
      else if (dateTime.attribute("type").as_string() != "relative"sv)
        DIRSIG_THROW_FMT(
          "unsupported datetime type '{}'",
          dateTime.attribute("type").as_string());
    }
    if (dateTimeIsRelative)
      nextEntry.dateTime =
        DateTime::fromRelativeSeconds(dateTime.text().as_double());
    else
      nextEntry.dateTime = DateTime::fromIso8601(dateTime.text().as_string());
    if (entry.child("position")) {
      // Read <position>...</position>
      nextEntry.location =
        xml::readLocation(entry.child("position").first_child(), facade);
      // Read <orientation>...</orientation>
      auto rotation = xml::readCoordinate(entry.child("orientation"), "eulerangles");
      nextEntry.rotation[0] = rotation[0] * rotationUnitsFac;
      nextEntry.rotation[1] = rotation[1] * rotationUnitsFac;
      nextEntry.rotation[2] = rotation[2] * rotationUnitsFac;
    } else {
      nextEntry.location = geolib::Location(facade);
      nextEntry.location.setLocation(Point{
        entry.child("xlocation").text().as_double(),
        entry.child("ylocation").text().as_double(),
        entry.child("zlocation").text().as_double()});
      nextEntry.rotation = {
        entry.child("xrotation").text().as_double(),
        entry.child("yrotation").text().as_double(),
        entry.child("zrotation").text().as_double()};
      nextEntry.rotation[0] *= rotationUnitsFac;
      nextEntry.rotation[1] *= rotationUnitsFac;
      nextEntry.rotation[2] *= rotationUnitsFac;
    }
  }
}

void GenericMotion::ReadPPDFile(
  const std::string& filename, const geolib::Facade* facade) {
  auto doc = xml::loadFile(filename);
  ReadXML(doc.document_element(), facade);
}

}  // namespace d5::d4legacy
