#include "dirsig5/d4legacy/motion/StkReportFile.h"

#include <fstream>
#include <stdexcept>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/ErrorIO.h"
#include "spdlog/fmt/fmt.h"

namespace d5::d4legacy::stk {

enum CentralBody { UnknownBody, EarthBody };

enum CoordinateSystem { UnknownCoordinates, FixedCoordinates };

std::vector<LocationEngines::Waypoints::LocationEntry> readEphemeris(
  const char* filename, geolib::Facade& geolib) {
  auto reportFile = OpenIfstream(filename);
  int pointCount = 0;
  CentralBody centralBody = UnknownBody;
  CoordinateSystem coordinateSystem = UnknownCoordinates;

  // start reading in lines, looking for stuff we want
  size_t lineNum = 1;
  for (;; lineNum++) {
    std::string lineString;
    if (!std::getline(reportFile, lineString)) break;
    auto lineTokens = TokenList(lineString, " \r\t");
    if (lineTokens.empty()) continue;
    auto firstToken = lineTokens[0];
    if (firstToken == "BEGIN") {
      // check if this is an "Ephemeris" report file
      if (lineTokens.size() < 2 || lineTokens[1] != "Ephemeris")
        DIRSIG_THROW_FMT("this is not an STK ephemeris file");
    } else if (firstToken == "NumberOfEphemerisPoints") {
      // try to grab the number of points
      if (lineTokens.size() < 2)
        DIRSIG_THROW_FMT(
          "NumberOfEphemerisPoints line ({}) did not have a value", lineNum);
      pointCount = ParseNumber<int>(lineTokens[1]);
      if (pointCount <= 0)
        DIRSIG_THROW_FMT(
          "invalid number of ephemeris points on line {}", lineNum);
    } else if (firstToken == "CentralBody") {
      if (lineTokens.size() >= 2 && lineTokens[1] == "Earth")
        centralBody = EarthBody;
      else
        DIRSIG_THROW_FMT("unsupported CentralBody type on line {}", lineNum);
    } else if (firstToken == "CoordinateSystem") {
      if (lineTokens.size() >= 2 && lineTokens[1] == "Fixed")
        coordinateSystem = FixedCoordinates;
      else
        DIRSIG_THROW_FMT(
          "unsupported CoordinateSystem type on line {}", lineNum);
    } else if (firstToken.substr(0, 9) == "Ephemeris") {
      if (
        firstToken != "EphemerisTimePos" && firstToken != "EphemerisTimePosVel")
        DIRSIG_THROW_FMT(
          "unsupported ephemeris data format on line {}", lineNum);
      break;
    }
  }

  // check on our expected variables
  if (centralBody != EarthBody) DIRSIG_THROW_FMT("CentralBody nut supplied");
  if (coordinateSystem != FixedCoordinates)
    DIRSIG_THROW_FMT("CoordinateSystem not supplied");

  // start the read loop
  std::vector<LocationEngines::Waypoints::LocationEntry> ephemeris;
  for (int pointIndex = 0; pointIndex < pointCount; pointIndex++, lineNum++) {
    std::string lineString;
    std::getline(reportFile, lineString);
    auto lineTokens = TokenList(lineString, " \r\t");
    if (lineTokens.empty()) {
      pointIndex--;
      continue;
    }
    if (lineTokens.size() < 4)
      DIRSIG_THROW_FMT("not enough tokens on line {}", lineNum);

    // extract the ECEF position
    auto& entry = ephemeris.emplace_back();
    entry.dateTime =
      DateTime::fromRelativeSeconds(ParseNumber<double>(lineTokens[0]));
    entry.location = geolib::Location(
      EcefCoord(
        ParseNumber<double>(lineTokens[1]), ParseNumber<double>(lineTokens[2]),
        ParseNumber<double>(lineTokens[3])),
      &geolib);
  }

  return ephemeris;
}

AttitudeResult readAttitude(const char* filename) {
  auto reportFile = OpenIfstream(filename);
  int pointCount = 0;
  CentralBody centralBody = UnknownBody;
  CoordinateSystem coordinateSystem = UnknownCoordinates;

  // start reading in lines, looking for stuff we want
  size_t lineNum = 1;
  for (;; lineNum++) {
    std::string lineString;
    if (!std::getline(reportFile, lineString)) break;
    auto lineTokens = TokenList(lineString, " \r\t");
    if (lineTokens.empty()) continue;
    auto firstToken = lineTokens[0];
    if (firstToken == "BEGIN") {
      if (lineTokens.size() < 2 || lineTokens[1] != "Attitude")
        DIRSIG_THROW_FMT("this is not an STK ephemeris file");
    } else if (firstToken == "NumberOfAttitudePoints") {
      // try to grab the number of points
      if (lineTokens.size() < 2)
        DIRSIG_THROW_FMT(
          "NumberOfAttitudePoints line ({}) did not have a value", lineNum);
      pointCount = ParseNumber<int>(lineTokens[1]);
      if (pointCount <= 0)
        DIRSIG_THROW_FMT(
          "invalid number of attitude points on line {}", lineNum);
    } else if (firstToken == "CentralBody") {
      if (lineTokens.size() >= 2 && lineTokens[1] == "Earth")
        centralBody = EarthBody;
      else
        DIRSIG_THROW_FMT("unsupported CentralBody type on line {}", lineNum);
    } else if (firstToken == "CoordinateAxes") {
      if (lineTokens.size() >= 2 && lineTokens[1] == "Fixed")
        coordinateSystem = FixedCoordinates;
      else
        DIRSIG_THROW_FMT(
          "unsupported CoordinateSystem type on line {}", lineNum);
    } else if (firstToken.substr(0, 8) == "Attitude") {
      if (firstToken != "AttitudeTimeQuaternions") {
        DIRSIG_THROW_FMT(
          "unsupported attitude data format on line {}", lineNum);
      }
      break;
    }
  }

  // check on our expected variables
  if (centralBody != EarthBody) DIRSIG_THROW_FMT("CentralBody nut supplied");
  if (coordinateSystem != FixedCoordinates)
    DIRSIG_THROW_FMT("CoordinateSystem not supplied");

  AttitudeResult res;

  //
  // We need to setup the body transform for the STK file.
  // STK assumes the platform frame is:
  //     - +X is velocity aligned, and
  //     - +Z points down to toward the Earth.
  // DIRSIG assumes the platform frame is:
  //     - +Y is velocity aligned, and
  //     - +Z points away from the Earth.
  // To rotate the DIRSIG body frame into the STK body frame, we need to
  //     - rotate about +Z by +90 degrees to make Y the "velocity", then
  //     - rotate about +Y by -180 degrees to make +Z the "down".
  //
  auto zRotation = Eigen::AngleAxisd(M_PI / 2, Eigen::Vector3d::UnitZ());
  auto yRotation = Eigen::AngleAxisd(-M_PI, Eigen::Vector3d::UnitY());
  res.bodyTransform = yRotation * zRotation;

  // start the read loop
  for (int pointIndex = 0; pointIndex < pointCount; pointIndex++, lineNum++) {
    std::string lineString;
    std::getline(reportFile, lineString);
    auto lineTokens = TokenList(lineString, " \r\t");
    if (lineTokens.empty()) {
      pointIndex--;
      continue;
    }
    if (lineTokens.size() < 5)
      DIRSIG_THROW_FMT("not enough tokens on line {}", lineNum);
    auto& entry = res.quaternions.emplace_back();
    entry.dateTime =
      DateTime::fromRelativeSeconds(ParseNumber<double>(lineTokens[0]));

    //
    // Extract the ECEF attitude quaternion
    // NOTE: The STK output report puts the "scalar" component *after*
    // the "vector" component, so the order in the file is: X, Y, Z, W.
    //
    entry.quaternion.x() = ParseNumber<double>(lineTokens[1]);
    entry.quaternion.y() = ParseNumber<double>(lineTokens[2]);
    entry.quaternion.z() = ParseNumber<double>(lineTokens[3]);
    entry.quaternion.w() = ParseNumber<double>(lineTokens[4]);
  }

  return res;
}

}  // namespace d5::d4legacy::stk
