#include "dirsig5/d4legacy/scene/Maps.h"

#include "dirsig5/base/Error.h"
#include "dirsig5/base/StringUtils.h"
#include "dirsig5/d4legacy/misc/Units.h"
#include "dirsig5/base/Logger.h"

using namespace std::string_view_literals;

namespace d5::d4legacy {

static Projector::Extend ParseExtend(std::string str) {
  str = trimmed(str);
  str = toLower(str);
  if (str == "repeat") return Projector::Extend::Repeat;
  if (str == "mirror") return Projector::Extend::Mirror;
  if (str == "clamp") return Projector::Extend::ClampToEdge;
  if (!str.empty())
    DIRSIG_WARN(
      "Unrecognized extend mode in projector: {} (using mirror "
      "by default)\n"
      "Valid modes:\n"
      "  mirror\n"
      "  repeat\n"
      "  clamp",
      str);
  return Projector::Extend::Mirror;
}

static Projector::Sampling ParseSampling(std::string str) {
  str = trimmed(str);
  str = toLower(str);
  if (str == "nearest") return Projector::Sampling::Nearest;
  if (str == "linear") return Projector::Sampling::Linear;
  if (!str.empty())
    DIRSIG_WARN(
      "Unrecognized sampling mode in projector: {} (using nearest by "
      "default)\n"
      "Valid modes:\n"
      "  nearest\n"
      "  linear",
      str);
  return Projector::Sampling::Nearest;
}

void Projector::ReadXML(const xml::Node& node, const geolib::Facade* facade) {
  *this = {};
  auto typeNode = node.first_child();
  if (typeNode.name() == "tileprojector"sv) {
    type = Type::Drape;
    auto insertPoint = xml::requiredChild(typeNode, "insertpoint");
    if (
      insertPoint.child("point") ||  //
      insertPoint.child("cartesianlocation"))
      drapeOrigin = xml::readCoordinate(insertPoint.first_child());
    else
      drapeOrigin = xml::readLocation(insertPoint.first_child(), facade)
                      .getLocalCartesian();
    drapeGsd = xml::requiredChild(typeNode, "gsd").text().as_double();
    if (drapeGsd <= 0.0)
      DIRSIG_THROW_FMT(
        "line {}: maps must have a gsd greater than 0.0",
        xml::lineno(typeNode));
  } else if (typeNode.name() != "uvprojector"sv)
    DIRSIG_THROW_FMT(
      "line {}: unknown map projector <{}>", xml::lineno(typeNode),
      typeNode.name());

  std::string_view originStr = node.attribute("origin").as_string();
  useCartesianOrigin = originStr == "cartesian";
  if (originStr != "cartesian" && originStr != "image" && !originStr.empty())
    DIRSIG_WARN(
      "Unrecognized map origin: {} (using image coordinates by default).",
      originStr);
  flipX = node.attribute("flipx").as_bool();
  flipY = node.attribute("flipy").as_bool();
  extendX = ParseExtend(node.attribute("extendx").as_string());
  extendY = ParseExtend(node.attribute("extendy").as_string());
  minU = node.attribute("minu").as_float(0);
  minV = node.attribute("minv").as_float(0);
  maxU = node.attribute("maxu").as_float(1);
  maxV = node.attribute("maxv").as_float(1);
  sampling = ParseSampling(node.attribute("sampling").as_string());
}

void Projector::ReadTag(const TagFileSection& section) {
  *this = {};
  const auto* drapeSection = section.subsection("DRAPE_PROJECTOR");
  const auto* uvSection = section.subsection("UV_PROJECTOR");

  auto ReadOptions = [&](const TagFileSection& s) {
    useCartesianOrigin =
      s.optionalVariable<std::string>("ORIGIN") == "CARTESIAN";
    auto flipXStr = s.optionalVariable<std::string>("FLIPX");
    auto flipYStr = s.optionalVariable<std::string>("FLIPY");
    flipX = flipXStr.empty() || flipXStr == "TRUE";
    flipY = flipYStr.empty() || flipYStr == "TRUE";
    extendX = ParseExtend(s.optionalVariable<std::string>("EXTENDX"));
    extendY = ParseExtend(s.optionalVariable<std::string>("EXTENDY"));
    minU = s.optionalVariable<float>("MINU", 0, 0.0f);
    minV = s.optionalVariable<float>("MINV", 0, 0.0f);
    maxU = s.optionalVariable<float>("MAXU", 0, 1.0f);
    maxV = s.optionalVariable<float>("MAXV", 0, 1.0f);
    sampling = ParseSampling(s.optionalVariable<std::string>("SAMPLING"));
  };
  if (drapeSection) {
    type = Type::Drape;
    drapeSection->requiredVariable("GSD", 0, drapeGsd);
    drapeOrigin = tag::ParsePoint(
      drapeSection->requiredVariable<std::string>("INSERT_POINT"));
    ReadOptions(*drapeSection);
  } else if (uvSection) {
    type = Type::UV;
    ReadOptions(*uvSection);
  } else if (section.elementCount("INSERT_POINT") > 0) {
    type = Type::Drape;
    section.requiredVariable("GSD", 0, drapeGsd);
    drapeOrigin =
      tag::ParsePoint(section.requiredVariable<std::string>("INSERT_POINT"));
  } else
    DIRSIG_THROW_FMT(
      "can't find either UV_PROJECTOR or DRAPE_PROJECTOR section");
}

void WangTileSource::ReadXML(
  const xml::Node& node, const FileLocator& fileLocator) {
  *this = {};
  xCount = xml::requiredAttribute(node, "xcount").as_int();
  yCount = xml::requiredAttribute(node, "ycount").as_int();
  for (auto tile : node.children("tile")) {
    auto& nextTile = tiles.emplace_back();
    auto neswParts =
      TokenList(xml::requiredAttribute(tile, "nesw").as_string(), ",");
    if (neswParts.size() != 4)
      DIRSIG_THROW_FMT(
        "line {}: invalid wang tile nesw string", xml::lineno(node));
    nextTile.filename = fileLocator.locateFile(tile.text().as_string());
    nextTile.nesw[0] = ParseNumber<int>(neswParts[0]);
    nextTile.nesw[1] = ParseNumber<int>(neswParts[1]);
    nextTile.nesw[2] = ParseNumber<int>(neswParts[2]);
    nextTile.nesw[3] = ParseNumber<int>(neswParts[3]);
  }
}

void ImageSource::ReadXML(
  const xml::Node& node, const FileLocator& fileLocator) {
  if (node.child("wangtiles")) {
    type_ = Type::WangTiles;
    WangTileSource wang;
    wang.ReadXML(node.child("wangtiles"));
    filename_ = Json(wang).dump();
  } else {
    type_ = Type::Filename;
    filename_ = fileLocator.locateFile(
      xml::requiredChild(node, "filename").text().as_string());
  }
}

void ImageSource::ReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  filename_ = fileLocator.locateFile(
    section.requiredVariable<std::string>("IMAGE_FILENAME"));
}

std::unique_ptr<MapEntry> MapEntry::FromXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  std::unique_ptr<MapEntry> mapEntry;
  std::string type = trimmed(std::string(node.name()));
  if (type == "bumpmap")
    mapEntry = std::make_unique<BumpMapEntry>();
  else if (type == "decalmap")
    mapEntry = std::make_unique<DecalMapEntry>();
  else if (type == "densitymap")
    mapEntry = std::make_unique<DensityMapEntry>();
  else if (type == "materialmap")
    mapEntry = std::make_unique<MaterialMapEntry>();
  else if (type == "mixturemap")
    mapEntry = std::make_unique<MixtureMapEntry>();
  else if (type == "noisemap")
    mapEntry = std::make_unique<NoiseMapEntry>();
  else if (type == "texturemap")
    mapEntry = std::make_unique<TextureMapEntry>();
  else if (type == "reflectancemap")
    mapEntry = std::make_unique<ReflectanceMapEntry>();
  else if (type == "temperaturemap")
    mapEntry = std::make_unique<TemperatureMapEntry>();
  else
    DIRSIG_THROW_FMT(
      "line {}: unrecognized map type <{}>", xml::lineno(node), type);
  mapEntry->ReadXML(node, facade, fileLocator);
  return mapEntry;
}

static std::vector<std::string> ReadMaterialIDs(const xml::Node& node) {
  std::vector<std::string> matIds;
  for (auto subnode : node.children("matid"))
    matIds.push_back(trimmed(std::string(subnode.text().as_string())));
  return matIds;
}

void MapEntry::ReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  name = node.attribute("name").as_string();
  enabled = node.attribute("enabled").as_bool(true);
  projector = {};
  if (node.child("projector"))
    projector.ReadXML(node.child("projector"), facade);
  if (node.child("matidlist"))
    materialIds = ReadMaterialIDs(node.child("matidlist"));
  FinishReadXML(node, facade, fileLocator);
}

void MapEntry::ReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  section.optionalVariable("NAME", 0, name);
  section.optionalVariable("ENABLED", 0, enabled);
  projector.ReadTag(section);
  FinishReadTag(section, fileLocator);
}

void BumpMapEntry::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  source.ReadXML(xml::requiredChild(node, "image"), fileLocator);
  scale = xml::requiredChild(node, "scale").text().as_double(1.0);
}

void BumpMapEntry::FinishReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  scale =
    section.requiredSubsection("OPTIONS").requiredVariable<double>("SCALE");
  source.ReadTag(section, fileLocator);
}

void DecalMapEntry::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  glist = fileLocator.locateFile(
    xml::requiredChild(xml::requiredChild(node, "glist"), "filename")
      .text()
      .as_string());
}

void DecalMapEntry::FinishReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  // TODO ?
}

void MaterialMapEntry::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  source.ReadXML(xml::requiredChild(node, "image"), fileLocator);
  std::string_view mixingStr = node.attribute("mixing").as_string();
  isMixing = mixingStr == "true" || mixingStr == "1";
  for (auto entry : node.child("dcmatlut").children("entry")) {
    int dc = xml::requiredChild(entry, "dc").text().as_int();
    if (dc < 0)
      DIRSIG_THROW_FMT(
        "line {}: material map dc must be positive", xml::lineno(entry));
    if (!materialLut[dc].empty())
      DIRSIG_THROW_FMT(
        "line {}: material map has duplicate dc", xml::lineno(entry));
    auto materialId = trimmed(
      std::string(xml::requiredChild(entry, "matid").text().as_string()));
    if (
      std::find(materialIds.begin(), materialIds.end(), materialId) !=
      materialIds.end())
      DIRSIG_THROW_FMT(
        "line {}: map '{}' mapped to itself", xml::lineno(entry), materialId);
    materialLut[dc] = materialId;
  }
}

void MaterialMapEntry::FinishReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  source.ReadTag(section, fileLocator);
  section.requiredSubsection("OPTIONS").optionalVariable(
    "ENABLE_MIXING", 0, isMixing);
  materialLut.clear();
  for (const auto& entry : section.requiredSubsection("LUT")) {
    if (entry.type() != TagFileElement::Type::UnformattedData)
      DIRSIG_THROW_FMT("material map LUT must only contain unformatted data");
    auto data = entry.as<TagFileData>().data();
    auto dataTokens = TokenList(data, ":");
    if (dataTokens.size() != 2)
      DIRSIG_THROW_FMT("invalid LUT entry, must be \"DC:label\"");
    int dc = ParseNumber<int>(dataTokens[0]);
    if (dc < 0 || dc > 255)
      DIRSIG_THROW_FMT("out-of-range DC in LUT, must be in [0, 255]");
    materialLut[dc] = trimmed(dataTokens[1]);
  }
}

void MixtureMapEntry::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  source.ReadXML(xml::requiredChild(node, "image"), fileLocator);
  bandMaterialIds = ReadMaterialIDs(xml::requiredChild(
    xml::requiredChild(node, "bandassignments"), "matidlist"));
}

void MixtureMapEntry::FinishReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  source.ReadTag(section, fileLocator);
  auto labels = section.requiredVariable<std::string>("BAND_MATERIAL_LABELS");
  for (std::string_view label : TokenList(labels, ", "))
    bandMaterialIds.emplace_back(std::string(label));
}

void NoiseMapEntry::FinishReadXML(
  const xml::Node& node, const geolib::Facade*, const FileLocator&) {
  scale = node.child("scale").text().as_double();
  order = node.child("order").text().as_int();
  magnitude = node.child("magnitude").text().as_double();
  beta = node.child("beta").text().as_double();
}

void NoiseMapEntry::FinishReadTag(
  const TagFileSection& section, const FileLocator&) {
  const auto& options = section.requiredSubsection("OPTIONS");
  options.optionalVariable("SCALE", 0, scale);
  options.optionalVariable("ORDER", 0, order);
  options.optionalVariable("MAGNITUDE", 0, magnitude);
  options.optionalVariable("BETA", 0, beta);
}

void TemperatureMapEntry::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  source.ReadXML(xml::requiredChild(node, "image"), fileLocator);
  gain = node.child("gain").text().as_double(1.0);
  bias = node.child("bias").text().as_double(0.0);
  AccountForUnits(node.attribute("units").as_string());
}

void TemperatureMapEntry::FinishReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  source.ReadTag(section, fileLocator);
  const auto& options = section.requiredSubsection("OPTIONS");
  gain = 1;
  bias = 0;
  options.optionalVariable("GAIN", 0, gain);
  options.optionalVariable("BIAS", 0, bias);
  std::string units;
  options.optionalVariable("UNITS", 0, units);
  AccountForUnits(units);
}

void TemperatureMapEntry::AccountForUnits(std::string_view units) {
  if (units == "Celsius") {
    // Do nothing
  } else if (units == "Kelvin") {
    bias -= 273.15;
  } else if (units == "Fahrenheit") {
    bias -= 32;
    gain *= 5.0 / 9.0;
    bias *= 5.0 / 9.0;
  } else if (!units.empty()) {
    DIRSIG_THROW_FMT(
      "Unrecognized units \"{}\" for TemperatureMap! Expected "
      "\"Celsius\", \"Kelvin\", or \"Fahrenheit\"",
      units);
  }
}

void ReflectanceMapEntry::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  source.ReadXML(xml::requiredChild(node, "image"), fileLocator);
  gain = node.child("gain").text().as_double(1.0);
  bias = node.child("bias").text().as_double(0.0);
  clipValues = node.child("clipvalues").text().as_bool(false);
  convertToMixture = node.child("converttomixture").text().as_bool(false);
  sqtFilename = node.child("sqtfilename").text().as_string();
  if (auto list = node.child("bandcenterlist"); list) {
    auto units = WavelengthUnits::Microns;
    if (auto u = list.attribute("units"); u)
      if (!stringToUnits(u.as_string(), units))
        DIRSIG_THROW_FMT(
          "line {}: unknown spectral units \"{}\"", xml::lineno(list),
          u.as_string());
    for (auto band : list.children("bandcenter"))
      bandCenters.push_back(toMicrons(band.text().as_double(), units));
  }
}

void ReflectanceMapEntry::FinishReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  source.ReadTag(section, fileLocator);
  const auto& options = section.requiredSubsection("OPTIONS");
  gain = 1;
  bias = 0;
  options.optionalVariable("GAIN", 0, gain);
  options.optionalVariable("BIAS", 0, bias);
  options.optionalVariable("CLIP_VALUES", 0, clipValues);
  options.optionalVariable("CONVERT_TO_MIXTURE", 0, convertToMixture);
  options.optionalVariable("SQT_FILENAME", 0, sqtFilename);
  // TODO
}

void TextureMapEntry::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade*,
  const FileLocator& fileLocator) {
  bands.clear();
  auto blist = xml::requiredChild(node, "bandlist");
  for (auto band : blist.children("band")) {
    auto& b = bands.emplace_back();
    auto bandpass = xml::requiredChild(band, "bandpass");
    b.bandpassMin = xml::requiredChild(bandpass, "min").text().as_double();
    b.bandpassMax = xml::requiredChild(bandpass, "max").text().as_double();
    b.source.ReadXML(xml::requiredChild(band, "image"), fileLocator);
    if (auto e = band.child("ignoredclist"); e) {
      for (auto dcEntry : e.children("dc")) {
        int dc = dcEntry.text().as_int();
        if (dc < 0 || dc > 255)
          DIRSIG_THROW_FMT(
            "line {}: out-of-bounds DC value {}", xml::lineno(dcEntry), dc);
        b.ignoreDCs.push_back(static_cast<uint8_t>(dc));
      }
    }
    if (auto e = band.child("userstats"); e) {
      b.dcMean = xml::requiredChild(e, "dcmean").text().as_double();
      b.dcStdDev = xml::requiredChild(e, "dcstddev").text().as_double();
    }
  }
}

void TextureMapEntry::FinishReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  //
}

void BalfourMapEntry::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  //
}

void BalfourMapEntry::FinishReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  //
}

void ThermMapEntry::FinishReadXML(
  const xml::Node& node,
  const geolib::Facade* facade,
  const FileLocator& fileLocator) {
  //
}

void ThermMapEntry::FinishReadTag(
  const TagFileSection& section, const FileLocator& fileLocator) {
  //
}

std::string to_string(MapType type) {
  switch (type) {
    case MapType::Bump:
      return "Bump";
    case MapType::Decal:
      return "Decal";
    case MapType::Density:
      return "Density";
    case MapType::Material:
      return "Material";
    case MapType::Mixture:
      return "Mixture";
    case MapType::Noise:
      return "Noise";
    case MapType::Texture:
      return "Texture";
    case MapType::Reflectance:
      return "Reflectance";
    case MapType::Temperature:
      return "Temperature";
    case MapType::Unknown:
    default:
      return "Unknown";
  };
}

}  // namespace d5::d4legacy
