#include "dirsig5/d4legacy/scene/SceneFile.h"

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Filesystem.h"
#include "dirsig5/d4legacy/xml/xml.h"
#include "dirsig5/base/StandardPaths.h"
#include "spdlog/fmt/fmt.h"

using namespace std::string_view_literals;

namespace d5::d4legacy {

static std::string replaceSceneDir(
  const std::string& path, const std::string& sceneDir) {
  const std::string_view sceneDirTag = "$SCENE_DIR"sv;
  if (path.size() < sceneDirTag.size()) return path;

  // check opening text
  if (std::string_view(path).substr(0, sceneDirTag.size()) == sceneDirTag)
    return sceneDir + path.substr(sceneDirTag.size(), path.size());
  return path;
}

std::unique_ptr<SceneFile> SceneFile::LoadFromXmlFile(
  const std::string& filename) {
  auto scene = std::make_unique<SceneFile>();
  auto doc = xml::loadFile(filename);
  auto root = doc.document_element();

  // Read metadata.
  if (auto metadata = root.child("metadata"); metadata)
    for (auto entry : metadata.children("entry"))
      scene->metadata[xml::requiredChild(entry, "name").text().as_string()] =
        xml::requiredChild(entry, "value").text().as_string();
  if (auto it = scene->metadata.find("Name"); it != scene->metadata.end())
    scene->sceneName = it->second;
  else
    scene->sceneName = root.attribute("name").as_string();

  if (auto basedir = root.child("scenebasedirectory"); basedir)
    scene->sceneDir = basedir.text().as_string();
  else
    scene->sceneDir = stdfs::absolute(filename).parent_path().string();
  scene->sceneDir = StandardPaths::withReplacements(scene->sceneDir);

  if (!stdfs::exists(scene->sceneDir))
    throw std::runtime_error(
      "Unable to load scene file, base directory \"" + scene->sceneDir +
      "\" does not exist!");

  // Load all of the search paths
  auto resolvePath = [&](auto dir) {
    return replaceSceneDir(
      StandardPaths::withReplacements(dir), scene->sceneDir);
  };
  auto requiredSearchPath = [&](auto type, const char* tag) {
    scene->fileLocator.addPath(
      type, {resolvePath(xml::requiredChild(root, tag).text().as_string())});
  };
  auto optionalSearchPath = [&](auto type, const char* tag) {
    scene->fileLocator.addPath(
      type, {resolvePath(root.child(tag).text().as_string())});
  };
  requiredSearchPath(FileLocator::GeometryList, "odbdirectory");
  requiredSearchPath(FileLocator::Geometry, "gdbdirectory");
  requiredSearchPath(FileLocator::Emissivity, "emsdirectory");
  requiredSearchPath(FileLocator::Extinction, "extdirectory");
  requiredSearchPath(FileLocator::Absorption, "absdirectory");
  requiredSearchPath(FileLocator::Map, "mapdirectory");
  optionalSearchPath(FileLocator::Source, "srcdirectory");

  // If the user has set the environment variable, add it as a search path
  if (const char* bundle = std::getenv("DIRSIG_BUNDLE_PATH");
      bundle != nullptr) {
    const char* separator = ":";
#if defined(WIN32) || defined(_WIN32) || \
  defined(__WIN32) && !defined(__CYGWIN__)
    separator = ";";
#endif
    for (auto path : TokenList(bundle, separator)) {
      scene->fileLocator.addPath(
        FileLocator::BundledMaterial, {std::string(path)});
      scene->fileLocator.addPath(
        FileLocator::GeometryList, {std::string(path)});
    }
  }

  // Load the scene origin
  {
    static geolib::Facade facade(0, 0, 0);
    auto latLonAlt =
      xml::readLocation(
        xml::requiredChild(root, "sceneorigin").first_child(), &facade)
        .getGeodeticLocation();
    scene->latLonAlt[0] = latLonAlt.latitude;
    scene->latLonAlt[1] = latLonAlt.longitude;
    scene->latLonAlt[2] = latLonAlt.height;
  }
  if (auto alt = root.child("groundaltitude"); alt)
    scene->groundAltitude = alt.text().as_double();

  // Load the material filename
  std::string matFilename = trimmed(std::string_view(
    xml::requiredChild(root, "matfilename").text().as_string()));
  if (!matFilename.empty()) matFilename = resolvePath(matFilename);
  scene->materialFilename = matFilename;

  for (auto glist : xml::requiredChild(root, "geometrylist").children()) {
    if (glist.name() != "geometrylistinclude"sv)
      DIRSIG_THROW_FMT(
        "The <geometrylist> in the top-level scene file must only "
        "contain <geometrylistinclude> tags (it is not allowed "
        "to specify geometry inline)");
    auto& inc = scene->glistIncludes.emplace_back();
    inc.name = glist.attribute("name").as_string();
    inc.enabled = glist.attribute("enabled").as_string() != "false"sv;
    inc.filename = resolvePath(trimmed(std::string(glist.text().as_string())));
  }

  for (auto thumb : root.child("thumbnaillist").children()) {
    auto& info = scene->thumbnailInfo.emplace_back();
    info.name = thumb.child("name").text().as_string();
    info.description = thumb.child("description").text().as_string();
    info.filename = resolvePath(thumb.child("filename").text().as_string());
  }

  if (auto props = root.child("properties"); props) {
    for (auto feature : TokenList(props.attribute("features").as_string(), ","))
      scene->attributes.emplace_back(std::string(feature));
    scene->minGsd = props.attribute("mingsd").as_double();
    scene->maxGsd = props.attribute("maxgsd").as_double();
    scene->spectralDelta = props.attribute("spectral_delta").as_double();
  }

  FileLocator fileLocator;
  fileLocator.addFilter(resolvePath);
  geolib::Facade facade(
    scene->latLonAlt[0], scene->latLonAlt[1], scene->latLonAlt[2]);
  for (auto map : root.child("maplist").children())
    scene->mapEntries.push_back(MapEntry::FromXML(map, &facade, fileLocator));
  for (auto plume : root.child("plumelist").children())
    scene->blackadarPlumes.emplace_back().ReadXML(plume);

  return scene;
}

}  // namespace d5::d4legacy
