/** Implementation file for TagFile
 *
 *  \file TagFile.cpp
 */

#include "dirsig5/d4legacy/tag/TagFile.h"

#include <cctype>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <iterator>

namespace d5::d4legacy {

/** Adds a subsection to the given section
 *
 *  \param[in,out] section  The section to which to add a subsection
 *  \param[in] tag          The tag name of the new subsection
 *  \param[in] lineNumber   The line number at which the subsection was found
 *
 *  \return A pointer to the new subsection
 */
TagFileSection& addSection(
  TagFileSection& section, std::string tag, int64_t lineNumber) {
  auto subSection = std::make_unique<TagFileSection>();
  subSection->setName(std::move(tag));
  subSection->setLineNumber(lineNumber);
  subSection->setSectionDepth(section.sectionDepth() + 1);
  section.push_back(std::move(subSection));
  return section[section.size() - 1].as<TagFileSection>();
}

/** Adds a variable to the given section
 *
 *  \param[in,out] section  The section to which to add a variable
 *  \param[in] tag          The tag name of the variable
 *  \param[in] value        The value of the variable
 *  \param[in] lineNumber   The line number at which the variable appeared
 */
void addVariable(
  TagFileSection& section,
  std::string tag,
  std::string value,
  int64_t lineNumber) {
  auto newVariable = std::make_unique<TagFileVariable>();
  newVariable->setTagName(std::move(tag));
  newVariable->setValue(std::move(value));
  newVariable->setLineNumber(lineNumber);
  section.push_back(std::move(newVariable));
}

/** Add raw data to the given section
 *
 *  \param[in,out] section  The section to which to add the data
 *  \param[in] data         The data to add
 *  \param[in] lineNumber   The line number at which the data was found
 */
void addData(TagFileSection& section, std::string data, int64_t lineNumber) {
  auto newData = std::make_unique<TagFileData>();
  newData->setData(std::move(data));
  newData->setLineNumber(lineNumber);
  section.push_back(std::move(newData));
}

/// A helper struct to keep the context of the file parse
struct ParsingContext {
  int64_t lineNumber = 0;
  int32_t sectionDepth = -1;
  std::string lineString;
  std::string tagString;
  std::string valueString;
};

/** Reads the next line in the tag file
 *
 *  \param[in,out] is       The stream from which to read
 *  \param[in,out] context  The context object to populate with the parsed
 *                          values
 *
 *  \return The type of the element found on the next line
 */
TagFileElement::Type readLine(std::istream& is, ParsingContext& context) {
  unsigned int i, begin_i, end_i;
  auto elementType = TagFileElement::UnknownElement;
  auto& line = context.lineString;

  // clear the strings
  context.lineString.clear();
  context.tagString.clear();
  context.valueString.clear();

  // check some things
  if (!is) return TagFileElement::UnknownElement;

  // read the next non empty, non comment line
  do {
    // try and read a line from the file
    std::getline(is, line);
    if (!is) return TagFileElement::UnknownElement;

    // increment the line number
    context.lineNumber++;

    // get rid of trailing whitespace on every line
    line.erase(line.find_last_not_of(" \t\r") + 1);

    // look for the first non whitespace character
    i = 0;
    while (i < line.size() && std::isspace(line[i])) i++;
  } while (line.empty() || i == line.size() || line[i] == '#');

  // search for the bounds of the tag name (data, variable or section)
  begin_i = i;
  for (i = begin_i; i < line.size(); i++) {
    auto c = line[i];
    if (std::isspace(c)) continue;
    if (c == '=' || c == '\0' || c == '\n' || c == '}' || c == '{') break;
  }
  end_i = i;

  // copy out the tag string
  auto& tag = context.tagString;
  tag = line.substr(begin_i, end_i - begin_i);

  // trim any trailing whitespace from the tag string (e.g. space before "=")
  tag.erase(tag.find_last_not_of(" \t") + 1);

  // figure out what type of element this is
  if (line[i] == '{') {
    elementType = TagFileElement::SectionBegin;
  } else if (line[i] == '}') {
    elementType = TagFileElement::SectionEnd;
  } else if (line[i] == '=') {
    elementType = TagFileElement::Variable;
  } else {
    // otherwise, this must be UnformattedData
    elementType = TagFileElement::UnformattedData;
  }

  // unless this is a variable assignment we are done
  if (elementType != TagFileElement::Variable) return elementType;

  // look for the next non whitespace character
  i = end_i + 1;
  while (i < line.size() && std::isspace(line[i])) i++;

  // extract out the value string
  begin_i = i;
  while (i < line.size() && line[i] != '=' && line[i] != '\n') i++;
  end_i = i;
  context.valueString = line.substr(begin_i, end_i - begin_i);

  return elementType;
}

/** Loads a section of a tag file
 *
 *  \param[in,out] is       The input stream from which to read
 *  \param[in,out] context  The context object to use to parse
 *  \param[out] section     The section into which to read
 */
void loadSection(
  std::istream& is, ParsingContext& context, TagFileSection& section) {
  // increment the overall section depth counter
  context.sectionDepth++;

  // loop until we run out of file to read
  bool sectionComplete = false;
  while (!sectionComplete) {
    // get the next element type
    auto elementType = readLine(is, context);

    // check if we have an incomplete section
    if (
      (elementType == TagFileElement::UnknownElement) &&
      (section.sectionDepth() != 0)) {
      std::ostringstream err;
      err << "TagFile::readSection: Read Error!"
          << "\n    Unexpected end-of-file reached at line #"
          << context.lineNumber
          << "\n    This section or and earlier section may not be closed"
          << "\n    Current section depth = " << section.sectionDepth() << "\n";
      throw std::runtime_error(err.str());
    }

    switch (elementType) {
      case TagFileElement::SectionBegin: {
        auto& subSection =
          addSection(section, std::move(context.tagString), context.lineNumber);
        loadSection(is, context, subSection);
        break;
      }
      case TagFileElement::Variable:
        addVariable(
          section, std::move(context.tagString), std::move(context.valueString),
          context.lineNumber);
        break;
      case TagFileElement::UnformattedData:
        addData(section, std::move(context.lineString), context.lineNumber);
        break;
      case TagFileElement::SectionEnd:
        context.sectionDepth--;
        [[fallthrough]];
      default:
        sectionComplete = true;
    }
  }
}

std::unique_ptr<TagFile> TagFile::LoadFromIstream(std::istream& is) {
  auto tagFile = std::make_unique<TagFile>();

  // store the info of the section
  tagFile->topSection_.setName("Top Level");
  tagFile->topSection_.setLineNumber(0);
  tagFile->topSection_.setSectionDepth(0);

  // start the parse tree by reading the top most elements
  ParsingContext context;
  loadSection(is, context, tagFile->topSection_);
  return tagFile;
}
}  // namespace d5::d4legacy
