/** Implementation file for TagFileElement
 *
 *  \file TagFileElement.cpp
 */

#include "dirsig5/d4legacy/tag/TagFileElement.h"

#include <algorithm>
#include <sstream>
#include <stdexcept>

namespace d5::d4legacy {

size_t TagFileSection::elementCount(std::string_view name) const {
  return std::count_if(
    begin(), end(), [&](const auto& e) { return e.tagName() == name; });
}

const TagFileElement* TagFileSection::element(
  std::string_view name, int instance) const {
  // Search for the requested instance of the tag name
  int instanceCount = 0;
  auto it = std::find_if(begin(), end(), [&](const auto& e) {
    return e.tagName() == name && instanceCount++ == instance;
  });
  return it != end() ? &(*it) : nullptr;
}

const TagFileElement& TagFileSection::requiredElement(
  std::string_view name, int instance) const {
  if (auto e = element(name, instance)) return *e;
  std::ostringstream errorMsg;
  errorMsg << "TagFileSection::requiredElement:\n"
           << "    Failed to find required element!\n"
           << "    Element name = " << name << "\n"
           << "    Section starts at line number = " << lineNumber() << "\n";
  throw std::domain_error(errorMsg.str());
}

const TagFileSection* TagFileSection::subsection(
  std::string_view name, int instance) const {
  auto e = element(name, instance);
  return (e && e->type() == TagFileElement::SectionBegin)
           ? static_cast<const TagFileSection*>(e)
           : nullptr;
}

const TagFileSection& TagFileSection::requiredSubsection(
  std::string_view name, int instance) const {
  if (auto e = element(name, instance)) return e->as<TagFileSection>();
  std::ostringstream errorMsg;
  errorMsg << "TagFileSection::requiredSubsection:\n"
           << "    Failed to find required element!\n"
           << "    Subsection = " << name << "\n"
           << "    Section starts at line number = " << lineNumber() << "\n";
  throw std::domain_error(errorMsg.str());
}

void TagFileSection::display(std::ostream& os) const {
  int depth = sectionDepth();
  depth = (depth > 0) ? depth : 0;

  // Calculate the whitespacing for the section and its elements
  std::string sectionWhitespace(depth * 4, ' ');
  std::string elementWhitespace = sectionWhitespace + "    ";

  os << name() << " {" << std::endl;
  for (const auto& e : *this) {
    os << elementWhitespace;
    e.display(os);
  }
  os << sectionWhitespace << "}" << std::endl;
}
}  // namespace d5::d4legacy
