/** Unit tests for FacetObject.h
 *
 *  \file FacetObjectTest.cpp
 */

#include "dirsig5/d4legacy/geometry/FacetObject.h"

#include <array>
#include <optional>

#include <gtest/gtest.h>

#include "CLStdComparisons.h"

const static std::vector<std::array<std::uint32_t, 3>> kCubeFaces = {
  {0, 2, 1}, {0, 3, 2}, {0, 7, 3}, {0, 4, 7}, {0, 5, 4}, {0, 1, 5},
  {6, 1, 2}, {6, 5, 1}, {6, 3, 7}, {6, 2, 3}, {6, 4, 5}, {6, 7, 4}};

::testing::AssertionResult VerifyInvariants(
  const d5::d4legacy::FacetObject& obj) {
  auto nF = obj.numFacets();

  if (obj.hasFacetThicknesses() && obj.facetThicknessesM().size() != nF) {
    return ::testing::AssertionFailure()
           << "Object had " << nF << " facets, but "
           << obj.facetThicknessesM().size() << " facet thicknesses";
  }

  if (obj.hasFacetTemperatures() && obj.facetTemperaturesK().size() != nF) {
    return ::testing::AssertionFailure()
           << "Object had " << nF << " facets, but "
           << obj.facetThicknessesM().size() << " facet temperatures";
  }

  if (obj.facetMaterialIndices().size() != nF) {
    return ::testing::AssertionFailure()
           << "Object had " << nF << " facets, but "
           << obj.facetMaterialIndices().size() << " facet material indices"
           << std::endl;
  }

  if (auto ntf = obj.numTextureFacets(); ntf != 0 && ntf != nF) {
    return ::testing::AssertionFailure()
           << "Object had " << nF << " facets, but  " << ntf
           << " texture facets";
  }

  if (auto nnf = obj.numVertexNormalFacets(); nnf != 0 && nnf != nF) {
    return ::testing::AssertionFailure()
           << "Object had " << nF << " facets, but  " << nnf
           << " vertex normal facets";
  }

  return ::testing::AssertionSuccess();
}

TEST(D4LegacyFacetObjectTest, DefaultStateTest) {
  d5::d4legacy::FacetObject obj;
  EXPECT_TRUE(obj.parts().empty());
  EXPECT_EQ(obj.numMaterials(), 0u);
  EXPECT_TRUE(obj.vertexData().empty());
  EXPECT_EQ(obj.numFacets(), 0u);
  EXPECT_FALSE(obj.hasFacetThicknesses());
  EXPECT_FALSE(obj.hasFacetTemperatures());
  EXPECT_EQ(obj.numTextureCoords(), 0u);
  EXPECT_EQ(obj.numVertexNormals(), 0u);
  EXPECT_TRUE(obj.pointSources().empty());
  EXPECT_EQ(obj.numPointSourceMaterials(), 0u);
  EXPECT_TRUE(obj.keypoints().empty());
  EXPECT_TRUE(VerifyInvariants(obj));
}

/** Add a cube of the form to the facet object
 *
 *     7     6
 *     /----/|
 *    / |  / |
 * 4 |----|5 | 2
 *   | /3 | /
 *   |____|/
 *  0      1
 */
template <typename OutputIt>
void AddCube(
  d5::d4legacy::FacetObject& obj,
  std::uint32_t part,
  std::array<std::uint32_t, 8> v,
  std::uint32_t matId,
  std::optional<float> thickness,
  std::optional<float> temperature,
  OutputIt out) {
  for (const auto& f : kCubeFaces) {
    auto fi = obj.addFacetToPart(part, v[f[0]], v[f[1]], v[f[2]], matId);
    if (thickness) obj.setFacetThicknessM(fi, *thickness);
    if (temperature) obj.setFacetTemperatureK(fi, *temperature);
    *out++ = fi;
  }
}

d5::d4legacy::FacetObject Make3Cube() {
  d5::d4legacy::FacetObject obj;
  auto mainPart = obj.addPart("main");
  auto p1Part = obj.addPart("p1");
  auto p2Part = obj.addPart("p2");

  auto black = obj.addMaterialLabel("black"), red = obj.addMaterialLabel("red"),
       green = obj.addMaterialLabel("green");

  auto v0 = obj.addVertex(0, 0, 0);
  auto v1 = obj.addVertex(1, 0, 0);
  auto v2 = obj.addVertex(1, 1, 0);
  auto v3 = obj.addVertex(0, 1, 0);
  auto v4 = obj.addVertex(0, 0, 1);
  auto v5 = obj.addVertex(1, 0, 1);
  auto v6 = obj.addVertex(1, 1, 1);
  auto v7 = obj.addVertex(0, 1, 1);
  auto v8 = obj.addVertex(2, 1, 1);
  auto v9 = obj.addVertex(2, 2, 1);
  auto v10 = obj.addVertex(1, 2, 1);
  auto v11 = obj.addVertex(1, 1, 2);
  auto v12 = obj.addVertex(2, 1, 2);
  auto v13 = obj.addVertex(2, 2, 2);
  auto v14 = obj.addVertex(1, 2, 2);
  auto v15 = obj.addVertex(-1, 0, 1);
  auto v16 = obj.addVertex(-1, 1, 1);
  auto v17 = obj.addVertex(-1, 0, 2);
  auto v18 = obj.addVertex(0, 0, 2);
  auto v19 = obj.addVertex(0, 1, 2);
  auto v20 = obj.addVertex(-1, 1, 2);
  auto v21 = obj.addVertex(0, 0, -1);
  auto v22 = obj.addVertex(1, 0, -1);
  auto v23 = obj.addVertex(3, 0, 1);
  auto v24 = obj.addVertex(-2, 0, 2);

  std::vector<std::uint32_t> faces;
  AddCube(
    obj, mainPart, {v0, v1, v2, v3, v4, v5, v6, v7}, black, 0.01f, 300.f,
    std::back_inserter(faces));
  AddCube(
    obj, p1Part, {v6, v8, v9, v10, v11, v12, v13, v14}, red, 0.02f, 310.f,
    std::back_inserter(faces));
  AddCube(
    obj, p2Part, {v15, v4, v7, v16, v17, v18, v19, v20}, green, 0.03f, 320.f,
    std::back_inserter(faces));

  auto srcM0 = obj.addPointSourceMaterialLabel("src0"),
       srcM1 = obj.addPointSourceMaterialLabel("src1"),
       srcM2 = obj.addPointSourceMaterialLabel("src2");

  auto& src0 = obj.addPointSource();
  src0.vertexIndex = v21;
  src0.materialIndex = srcM0;
  src0.direction = {1, 0, 0};
  src0.partIndex = 0;

  auto& src1 = obj.addPointSource();
  src1.vertexIndex = v22;
  src1.materialIndex = srcM0;
  src1.direction = {1, 0, 0};
  src1.partIndex = 0;

  auto& src2 = obj.addPointSource();
  src2.vertexIndex = v23;
  src2.materialIndex = srcM1;
  src2.direction = {0, 1, 0};
  src2.partIndex = 1;

  auto& src3 = obj.addPointSource();
  src3.vertexIndex = v24;
  src3.materialIndex = srcM2;
  src3.direction = {0, 0, 1};
  src3.partIndex = 2;

  return obj;
}

struct TestOrderIndependentResult {
  enum Status { Success, Duplicate, Missing };

  Status status;
  std::size_t index;
};

template <typename Callable>
TestOrderIndependentResult TestOrderIndependent(
  std::size_t n, std::vector<std::uint32_t>& indices, Callable equals) {
  std::vector<bool> has(n, false);
  indices.resize(n, 0);

  for (std::size_t i = 0; i < n; i++) {
    for (std::size_t j = 0; j < n; j++) {
      if (equals(i, j)) {
        if (has[j]) return {TestOrderIndependentResult::Duplicate, j};
        indices[j] = i;
        has[j] = true;
      }
    }
  }

  for (std::size_t i = 0; i < has.size(); i++) {
    if (!has[i]) return {TestOrderIndependentResult::Missing, i};
  }

  return {TestOrderIndependentResult::Success, 0};
}

template <typename T, typename Callable>
::testing::AssertionResult TestOrderIndependentWrapper(
  const std::vector<T>& truthVals,
  std::size_t nTest,
  const std::string& singular,
  const std::string& plural,
  std::vector<std::uint32_t>& indices,
  Callable&& equals) {
  if (nTest != truthVals.size()) {
    return ::testing::AssertionFailure()
           << "expected object to have " << truthVals.size() << " " << plural
           << ", but it had " << nTest;
  }

  auto r = TestOrderIndependent(nTest, indices, std::forward<Callable>(equals));

  auto ToString = [](const auto& t) {
    std::stringstream ss;
    ss << "[" << t[0];
    for (std::size_t i = 1; i < t.size(); i++) ss << ", " << t[i];
    ss << "]";
    return ss.str();
  };

  switch (r.status) {
    case TestOrderIndependentResult::Success:
      break;
    case TestOrderIndependentResult::Duplicate:
      return ::testing::AssertionFailure()
             << "object contained duplicate " << singular << " "
             << ToString(truthVals[r.index]);
    case TestOrderIndependentResult::Missing:
      return ::testing::AssertionFailure()
             << "object is missing " << singular << " "
             << ToString(truthVals[r.index]);
  }
  return ::testing::AssertionSuccess();
}

::testing::AssertionResult TestVertices(
  const d5::d4legacy::FacetObject& obj,
  const std::vector<std::array<float, 3>>& vs,
  std::vector<std::uint32_t>& indices,
  float eps = 1e-5) {
  return TestOrderIndependentWrapper(
    vs, obj.numVertices(), "vertex", "vertices", indices, [&](auto i, auto j) {
      const auto* v = obj.vertex(i);
      const auto& t = vs[j];
      return std::abs(v[0] - t[0]) < eps && std::abs(v[1] - t[1]) < eps &&
             std::abs(v[2] - t[2]) < eps;
    });
}

::testing::AssertionResult TestTextureCoords(
  const d5::d4legacy::FacetObject& obj,
  const std::vector<std::array<float, 2>>& vts,
  std::vector<std::uint32_t>& indices,
  float eps = 1e-5) {
  return TestOrderIndependentWrapper(
    vts, obj.numTextureCoords(), "texture vertex", "texture vertices", indices,
    [&](auto i, auto j) {
      const auto* v = obj.textureCoord(i);
      const auto& t = vts[j];
      return std::abs(v[0] - t[0]) < eps && std::abs(v[1] - t[1]) < eps;
    });
}

::testing::AssertionResult TestFacets(
  const d5::d4legacy::FacetObject& obj,
  const std::vector<std::array<std::uint32_t, 3>>& fs,
  std::vector<std::uint32_t>& indices) {
  return TestOrderIndependentWrapper(
    fs, obj.numFacets(), "facet", "facets", indices, [&](auto i, auto j) {
      const auto* f = obj.facet(i);
      const auto& t = fs[j];
      return f[0] == t[0] && f[1] == t[1] && f[2] == t[2];
    });
}

::testing::AssertionResult TestTextureFacets(
  const d5::d4legacy::FacetObject& obj,
  const std::vector<std::array<std::uint32_t, 3>>& fts,
  std::vector<std::uint32_t>& indices) {
  return TestOrderIndependentWrapper(
    fts, obj.numTextureFacets(), "texture facet", "texture facets", indices,
    [&](auto i, auto j) {
      const auto* f = obj.textureFacet(i);
      const auto& t = fts[j];
      return f[0] == t[0] && f[1] == t[1] && f[2] == t[2];
    });
}

::testing::AssertionResult TestVertexNormals(
  const d5::d4legacy::FacetObject& obj,
  const std::vector<std::array<float, 3>>& vns,
  std::vector<std::uint32_t>& indices,
  float eps = 1e-5) {
  return TestOrderIndependentWrapper(
    vns, obj.numVertexNormals(), "vertex normal", "vertex normals", indices,
    [&](auto i, auto j) {
      const auto* v = obj.vertexNormal(i);
      const auto& t = vns[j];
      return std::abs(v[0] - t[0]) < eps && std::abs(v[1] - t[1]) < eps &&
             std::abs(v[2] - t[2]) < eps;
    });
}

::testing::AssertionResult TestVertexNormalFacets(
  const d5::d4legacy::FacetObject& obj,
  const std::vector<std::array<std::uint32_t, 3>>& fns,
  std::vector<std::uint32_t>& indices) {
  return TestOrderIndependentWrapper(
    fns, obj.numVertexNormalFacets(), "vertex normal facet",
    "vertex normal facets", indices, [&](auto i, auto j) {
      const auto* f = obj.vertexNormalFacet(i);
      const auto& t = fns[j];
      return f[0] == t[0] && f[1] == t[1] && f[2] == t[2];
    });
}

::testing::AssertionResult Equals(
  const d5::d4legacy::FacetObject::Part& test,
  const d5::d4legacy::FacetObject::Part& truth) {
  if (test.name == truth.name && test.numFacets == truth.numFacets) {
    return ::testing::AssertionSuccess();
  }
  return ::testing::AssertionFailure()
         << "Expected parts to be equal:\n  Test: {name: " << test.name
         << ", numFacets: " << test.numFacets
         << "}\n  Truth: {name: " << truth.name
         << ", numFacets: " << truth.numFacets << "}";
}

::testing::AssertionResult Equals(
  const d5::d4legacy::FacetObject::PointSource& test,
  const d5::d4legacy::FacetObject::PointSource& truth) {
  if (
    test.vertexIndex == truth.vertexIndex &&
    test.materialIndex == truth.materialIndex &&
    std::abs(test.direction[0] - truth.direction[0]) < 1e-5 &&
    std::abs(test.direction[0] - truth.direction[0]) < 1e-5 &&
    std::abs(test.direction[0] - truth.direction[0]) < 1e-5 &&
    test.partIndex == truth.partIndex) {
    return ::testing::AssertionSuccess();
  }

  return ::testing::AssertionFailure()
         << "The given point sources were not equal:\n  Test: {vertexIndex: "
         << test.vertexIndex << ", materialIndex: " << test.materialIndex
         << ", direction: {" << test.direction[0] << ", " << test.direction[1]
         << ", " << test.direction[2] << "}, partIndex: " << test.partIndex
         << "}\n  Truth {vertexIndex: " << truth.vertexIndex
         << ", materialIndex: " << truth.materialIndex << ", direction: {"
         << truth.direction[0] << ", " << truth.direction[1] << ", "
         << truth.direction[2] << "}, partIndex: " << truth.partIndex << "}";
}

::testing::AssertionResult Equals(
  const d5::d4legacy::FacetObject::KeypointValue& test,
  const d5::d4legacy::FacetObject::KeypointValue& truth) {
  if (test.partIndex != truth.partIndex) {
    return ::testing::AssertionFailure()
           << "Part indices did not match (" << test.partIndex << " vs. "
           << truth.partIndex << ")";
  }
  if (test.points.size() != truth.points.size()) {
    return ::testing::AssertionFailure()
           << "Different number of points in the keypoint ("
           << test.points.size() << " vs. " << truth.points.size() << ")";
  }
  for (std::size_t i = 0; i < truth.points.size(); i++) {
    if (auto r = d5::IsNear(test.points[i], truth.points[i], 1e-5); !r) {
      return ::testing::AssertionFailure() << "Mismatch on point " << i << "\n"
                                           << r.message();
    }
  }
  return ::testing::AssertionSuccess();
}

TEST(D4LegacyFacetObjectTest, SplitOffPartsTest1) {
  auto obj = Make3Cube();
  auto base = obj;

  std::vector<d5::d4legacy::FacetObject> parts;
  ASSERT_NO_THROW(parts = d5::d4legacy::SplitOffParts(obj, {"p1"}));

  ASSERT_EQ(parts.size(), 1u);
  const auto& p1 = parts[0];

  ASSERT_TRUE(VerifyInvariants(obj));
  ASSERT_TRUE(VerifyInvariants(p1));

  // Make sure the base object has 2 parts ("main", "p1") left and the split
  // part is just "p1"
  ASSERT_EQ(obj.parts().size(), 2u);
  ASSERT_EQ(p1.parts().size(), 1u);
  ASSERT_EQ(obj.getPartIndexByName("main"), 0);
  ASSERT_EQ(obj.getPartIndexByName("p2"), 1);
  ASSERT_EQ(p1.getPartIndexByName("p1"), 0);
  ASSERT_EQ(obj.parts()[0].numFacets, 12u);
  ASSERT_EQ(obj.parts()[1].numFacets, 12u);
  ASSERT_EQ(p1.parts()[0].numFacets, 12u);

  // The base object should have "black" and "green", the split should have
  // "red"
  ASSERT_EQ(obj.numMaterials(), 2u);
  ASSERT_EQ(p1.numMaterials(), 1u);
  EXPECT_EQ(obj.materialLabel(0), "black");
  EXPECT_EQ(obj.materialLabel(1), "green");
  EXPECT_EQ(p1.materialLabel(0), "red");

  // The base object should now have 14 (8 + 8 - 2 shared) vertices left, the
  // split should have 8
  std::vector<std::uint32_t> objVIndices, p1VIndices;
  ASSERT_TRUE(TestVertices(
    obj,
    {{0, 0, 0},
     {1, 0, 0},
     {1, 1, 0},
     {0, 1, 0},
     {0, 0, 1},
     {1, 0, 1},
     {1, 1, 1},
     {0, 1, 1},
     {-1, 0, 1},
     {-1, 1, 1},
     {-1, 0, 2},
     {0, 0, 2},
     {0, 1, 2},
     {-1, 1, 2},
     {0, 0, -1},
     {1, 0, -1},
     {-2, 0, 2}},
    objVIndices));
  ASSERT_TRUE(TestVertices(
    p1,
    {{1, 1, 1},
     {2, 1, 1},
     {2, 2, 1},
     {1, 2, 1},
     {1, 1, 2},
     {2, 1, 2},
     {2, 2, 2},
     {1, 2, 2},
     {3, 0, 1}},
    p1VIndices));

  std::vector<std::uint32_t> objFIndices, p1FIndices;
  std::vector<std::array<std::uint32_t, 3>> facetTruth = kCubeFaces;
  std::array<std::uint32_t, 8> p2Remap{8, 4, 7, 9, 10, 11, 12, 13};
  for (const auto& f : kCubeFaces) {
    facetTruth.push_back(
      {objVIndices[p2Remap[f[0]]], objVIndices[p2Remap[f[1]]],
       objVIndices[p2Remap[f[2]]]});
  }

  ASSERT_TRUE(TestFacets(obj, facetTruth, objFIndices));
  ASSERT_TRUE(obj.hasFacetThicknesses());
  ASSERT_TRUE(obj.hasFacetTemperatures());
  for (std::size_t i = 0; i < facetTruth.size(); i++) {
    EXPECT_EQ(obj.partForFacet(objFIndices[i]), i < 12 ? 0 : 1);
    EXPECT_EQ(obj.facetMaterialIndices()[i], i < 12 ? 0 : 1);
    EXPECT_FLOAT_EQ(obj.facetThicknessesM()[i], i < 12 ? 0.01 : 0.03);
    EXPECT_FLOAT_EQ(obj.facetTemperaturesK()[i], i < 12 ? 300 : 320);
  }

  facetTruth.clear();
  for (const auto& f : kCubeFaces) {
    facetTruth.push_back(
      {p1VIndices[f[0]], p1VIndices[f[1]], p1VIndices[f[2]]});
  }

  ASSERT_TRUE(TestFacets(p1, facetTruth, p1FIndices));
  ASSERT_TRUE(p1.hasFacetThicknesses());
  ASSERT_TRUE(p1.hasFacetTemperatures());
  for (std::size_t i = 0; i < facetTruth.size(); i++) {
    EXPECT_EQ(p1.partForFacet(objFIndices[i]), 0);
    EXPECT_EQ(p1.facetMaterialIndices()[i], 0);
    EXPECT_FLOAT_EQ(p1.facetThicknessesM()[i], 0.02);
    EXPECT_FLOAT_EQ(p1.facetTemperaturesK()[i], 310);
  }

  ASSERT_EQ(obj.pointSources().size(), 3u);
  ASSERT_EQ(p1.pointSources().size(), 1u);
  ASSERT_EQ(obj.numPointSourceMaterials(), 2u);
  ASSERT_EQ(p1.numPointSourceMaterials(), 1u);
  ASSERT_EQ(obj.pointSourceMaterialLabel(0), "src0");
  ASSERT_EQ(obj.pointSourceMaterialLabel(1), "src2");
  ASSERT_EQ(p1.pointSourceMaterialLabel(0), "src1");
  ASSERT_TRUE(
    Equals(obj.pointSources()[0], {objVIndices[14], 0, {1, 0, 0}, 0}));
  ASSERT_TRUE(
    Equals(obj.pointSources()[1], {objVIndices[15], 0, {1, 0, 0}, 0}));
  ASSERT_TRUE(
    Equals(obj.pointSources()[2], {objVIndices[16], 1, {0, 0, 1}, 1}));
  ASSERT_TRUE(Equals(p1.pointSources()[0], {p1VIndices[8], 0, {0, 1, 0}, 0}));
}

TEST(D4LegacyFacetObjectTest, SplitOffPartsTest2) {
  d5::d4legacy::FacetObject obj;
  auto part1 = obj.addPart("p1"), part2 = obj.addPart("p2"),
       part3 = obj.addPart("p3");
  auto red = obj.addMaterialLabel("red"), blue = obj.addMaterialLabel("blue");
  std::array<std::uint32_t, 5> baseVs{
    obj.addVertex(0, 0, 0), obj.addVertex(1, 0, 0), obj.addVertex(0, 1, 0),
    obj.addVertex(-1, 1, 0), obj.addVertex(-1, 0, 0)};
  std::array<std::uint32_t, 3> baseFs{
    obj.addFacetToPart(part1, baseVs[0], baseVs[1], baseVs[2], red),
    obj.addFacetToPart(part2, baseVs[0], baseVs[2], baseVs[3], blue),
    obj.addFacetToPart(part3, baseVs[0], baseVs[3], baseVs[4], red)};
  std::array<std::uint32_t, 5> baseVts{
    obj.addTextureCoord(0.5, 0), obj.addTextureCoord(1, 0),
    obj.addTextureCoord(1, 1), obj.addTextureCoord(0, 1),
    obj.addTextureCoord(0, 0)};
  obj.setTextureFacet(baseFs[0], baseVts[0], baseVts[1], baseVts[2]);
  obj.setTextureFacet(baseFs[1], baseVts[0], baseVts[2], baseVts[3]);
  obj.setTextureFacet(baseFs[2], baseVts[0], baseVts[3], baseVts[4]);
  std::array<std::uint32_t, 5> baseVns{
    obj.addVertexNormal(-1, 0, 1), obj.addVertexNormal(1, 0, 1),
    obj.addVertexNormal(0, 0, 1), obj.addVertexNormal(-1, 1, 1),
    obj.addVertexNormal(0, -1, 1)};
  obj.setVertexNormalFacet(baseFs[0], baseVns[0], baseVns[1], baseVns[2]);
  obj.setVertexNormalFacet(baseFs[1], baseVns[0], baseVns[2], baseVns[3]);
  obj.setVertexNormalFacet(baseFs[2], baseVns[0], baseVns[3], baseVns[4]);
  obj.addKeypointToPart(
    part3, "corners", {{1, 0, 0}, {0, 1, 0}, {-1, 1, 0}, {-1, 0, 0}});
  obj.addKeypointToPart(part2, "center", {{0, 0.5, 0}});

  // Split p1 and p2 off of the base
  auto base = obj;
  std::vector<d5::d4legacy::FacetObject> parts;
  ASSERT_NO_THROW(parts = d5::d4legacy::SplitOffParts(base, {"p1", "p2"}));
  ASSERT_EQ(parts.size(), 2);

  const auto& p1 = parts[0];
  const auto& p2 = parts[1];

  ASSERT_TRUE(VerifyInvariants(base));
  ASSERT_TRUE(VerifyInvariants(p1));
  ASSERT_TRUE(VerifyInvariants(p2));

  // Make sure all three are just the 1 correct part
  ASSERT_EQ(base.parts().size(), 1u);
  ASSERT_EQ(p1.parts().size(), 1u);
  ASSERT_EQ(p2.parts().size(), 1u);
  ASSERT_TRUE(Equals(base.parts()[0], {"p3", 1}));
  ASSERT_TRUE(Equals(p1.parts()[0], {"p1", 1}));
  ASSERT_TRUE(Equals(p2.parts()[0], {"p2", 1}));

  // Make sure the materials are correct
  ASSERT_EQ(base.numMaterials(), 1u);
  ASSERT_EQ(p1.numMaterials(), 1u);
  ASSERT_EQ(base.materialLabel(0), "red");
  ASSERT_EQ(p1.materialLabel(0), "red");
  ASSERT_EQ(p2.materialLabel(0), "blue");

  // Verify the vertices in each object
  std::vector<std::uint32_t> baseVIndices, p1VIndices, p2VIndices;
  ASSERT_TRUE(
    TestVertices(base, {{0, 0, 0}, {-1, 1, 0}, {-1, 0, 0}}, baseVIndices));
  ASSERT_TRUE(TestVertices(p1, {{0, 0, 0}, {1, 0, 0}, {0, 1, 0}}, p1VIndices));
  ASSERT_TRUE(TestVertices(p2, {{0, 0, 0}, {0, 1, 0}, {-1, 1, 0}}, p2VIndices));

  // Verify the facets in each object
  std::vector<std::uint32_t> baseFIndices, p1FIndices, p2FIndices;
  ASSERT_TRUE(TestFacets(
    base, {{baseVIndices[0], baseVIndices[1], baseVIndices[2]}}, baseFIndices));
  ASSERT_TRUE(TestFacets(
    p1, {{p1VIndices[0], p1VIndices[1], p1VIndices[2]}}, p1FIndices));
  ASSERT_TRUE(TestFacets(
    p2, {{p2VIndices[0], p2VIndices[1], p2VIndices[2]}}, p2FIndices));

  // Verify the texture coordinates of each object
  std::vector<std::uint32_t> baseVtIdx, p1VtIdx, p2VtIdx;
  ASSERT_TRUE(TestTextureCoords(base, {{0.5, 0}, {0, 1}, {0, 0}}, baseVtIdx));
  ASSERT_TRUE(TestTextureCoords(p1, {{0.5, 0}, {1, 0}, {1, 1}}, p1VtIdx));
  ASSERT_TRUE(TestTextureCoords(p2, {{0.5, 0}, {1, 1}, {0, 1}}, p2VtIdx));

  // Verify the texture coordinate facets
  std::vector<std::uint32_t> baseVtf, p1Vtf, p2Vtf;
  ASSERT_TRUE(TestTextureFacets(
    base, {{baseVtIdx[0], baseVtIdx[1], baseVtIdx[2]}}, baseVtf));
  ASSERT_TRUE(
    TestTextureFacets(p1, {{p1VtIdx[0], p1VtIdx[1], p1VtIdx[2]}}, p1Vtf));
  ASSERT_TRUE(
    TestTextureFacets(p2, {{p2VtIdx[0], p2VtIdx[1], p2VtIdx[2]}}, p2Vtf));

  // Verify the vertex normals of each object
  std::vector<std::uint32_t> baseVnIdx, p1VnIdx, p2VnIdx;
  ASSERT_TRUE(
    TestVertexNormals(base, {{-1, 0, 1}, {-1, 1, 1}, {0, -1, 1}}, baseVnIdx));
  ASSERT_TRUE(
    TestVertexNormals(p1, {{-1, 0, 1}, {1, 0, 1}, {0, 0, 1}}, p1VnIdx));
  ASSERT_TRUE(
    TestVertexNormals(p2, {{-1, 0, 1}, {0, 0, 1}, {-1, 1, 1}}, p2VnIdx));

  // Verify the texture coordinate facets
  std::vector<std::uint32_t> baseVnf, p1Vnf, p2Vnf;
  ASSERT_TRUE(TestVertexNormalFacets(
    base, {{baseVnIdx[0], baseVnIdx[1], baseVnIdx[2]}}, baseVnf));
  ASSERT_TRUE(
    TestVertexNormalFacets(p1, {{p1VnIdx[0], p1VnIdx[1], p1VnIdx[2]}}, p1Vnf));
  ASSERT_TRUE(
    TestVertexNormalFacets(p2, {{p2VnIdx[0], p2VnIdx[1], p2VnIdx[2]}}, p2Vnf));

  // Verify the keypoints
  ASSERT_EQ(base.keypoints().size(), 1u);
  ASSERT_EQ(p1.keypoints().size(), 0u);
  ASSERT_EQ(p2.keypoints().size(), 1u);
  ASSERT_GT(base.keypoints().count("corners"), 0u);
  ASSERT_GT(p2.keypoints().count("center"), 0u);
  ASSERT_TRUE(Equals(
    base.keypoints().at("corners"),
    {0, {{1, 0, 0}, {0, 1, 0}, {-1, 1, 0}, {-1, 0, 0}}}));
  ASSERT_TRUE(Equals(p2.keypoints().at("center"), {0, {{0, 0.5, 0}}}));

  // Verify a full partition leaves an empty object
  ASSERT_NO_THROW(parts = d5::d4legacy::SplitOffParts(obj, {"p1", "p2", "p3"}));
  ASSERT_TRUE(VerifyInvariants(obj));
  ASSERT_TRUE(obj.parts().empty());
  ASSERT_EQ(obj.numMaterials(), 0u);
  ASSERT_EQ(obj.numVertices(), 0u);
  ASSERT_EQ(obj.numFacets(), 0u);
  ASSERT_EQ(obj.numTextureCoords(), 0u);
  ASSERT_EQ(obj.numVertexNormals(), 0u);
  ASSERT_TRUE(obj.pointSources().empty());
  ASSERT_EQ(obj.numPointSourceMaterials(), 0u);
  ASSERT_TRUE(obj.keypoints().empty());
}
