#include "dirsig5/d4legacy/tag/TagFile.h"

#include <sstream>

#include <gtest/gtest.h>

#include "CLTemporaryFile.h"

namespace test {

struct Filename {
  std::string filename;
  bool operator==(const std::string& str) const { return filename == str; }
};

}  // namespace test

template <>
struct d5::d4legacy::from_tag_string<test::Filename> {
  void operator()(const std::string& str, test::Filename& f) {
    f.filename = str;
  }
};

TEST(D4LegacyTagFileTest, LoadFromStringTest) {
  std::stringstream tagFile;
  tagFile << "MATERIAL_ENTRY {\n"
             "    ID           = 13  \r\n"
             "    NAME      \t   = dirt\t \n"
             "    NAME      \t   = dirt2\t foo\n"
             "    EDITOR_COLOR = 0.500008, 0.2, 0.2\n"
             "    DOUBLE_SIDED = FALSE\n"
             "   \n"
             "    RAD_SOLVER_NAME = Classic\n"
             "    RAD_SOLVER {\n"
             "        ENABLE_SAMPLED_DIFFUSE = FALSE\n"
             "    }\n"
             "\r\n"
             "  \t  TEMP_SOLVER_NAME = Therm\n"
             "    TEMP_SOLVER {\n"
             "        SPECIFIC_HEAT        = 0.2\n"
             "        MASS_DENSITY         = 1.35\n"
             "        EXPOSED_AREA         = -0.5\n"
             "    }\n"
             "\n"
             "    SURFACE_PROPERTIES {\n"
             "\n"
             "        EMISSIVITY_PROP_NAME = ClassicEmissivity\n"
             "        EMISSIVITY_PROP {\n"
             "            FILENAME          = dirt_dark_wet.ems\n"
             "            SPECULAR_FRACTION = 0\n"
             "        }\n"
             "    }\n"
             "}\n"
             "\n"
             "MATERIAL_ENTRY {\n"
             "    ID           = 1000\n"
             "    NAME         = Tire, Rubber, Black, Fair\n"
             "    EDITOR_COLOR = 0, 0, 0\n"
             "    DOUBLE_SIDED = TRUE\n"
             "\n"
             "    RAD_SOLVER_NAME = Classic\n"
             "    RAD_SOLVER {\n"
             "        ENABLE_SAMPLED_DIFFUSE = FALSE\n"
             "    }\n"
             "\n"
             "    TEMP_SOLVER_NAME = Therm\n"
             "    TEMP_SOLVER {\n"
             "        SPECIFIC_HEAT        = 0.2986\n"
             "        MASS_DENSITY         = 1.198\n"
             "        EXPOSED_AREA         = -0.65\n"
             "    }\n"
             "\n"
             "    SURFACE_PROPERTIES {\n"
             "\n"
             "        EMISSIVITY_PROP_NAME = ClassicEmissivity\n"
             "        EMISSIVITY_PROP {\n"
             "            FILENAME          = tire_black.ems   \n"
             "            SPECULAR_FRACTION = 0\n"
             "        }\n"
             "    }\n"
             "}\n";

  std::unique_ptr<d5::d4legacy::TagFile> f;
  ASSERT_NO_THROW(f = d5::d4legacy::TagFile::LoadFromIstream(tagFile));
  ASSERT_NE(f, nullptr);

  const auto& top = f->topSection();
  ASSERT_EQ(top.sectionDepth(), 0);
  ASSERT_EQ(top.size(), 2u);
  ASSERT_EQ(top.elementCount("MATERIAL_ENTRY"), 2u);

  const auto* mat1 = top.subsection("MATERIAL_ENTRY", 0);
  ASSERT_NE(mat1, nullptr);
  EXPECT_EQ(mat1->name(), "MATERIAL_ENTRY");
  EXPECT_EQ(mat1->optionalVariable<int>("ID", 0, 0), 13);
  EXPECT_EQ(mat1->optionalVariable<std::string>("NAME"), "dirt");
  EXPECT_EQ(mat1->requiredVariable<std::string>("NAME", 1), "dirt2\t foo");

  auto* radSolver1 =
    f->topSection().subsection("MATERIAL_ENTRY")->subsection("RAD_SOLVER");
  ASSERT_NE(radSolver1, nullptr);
  EXPECT_EQ(radSolver1->sectionDepth(), 2);
  EXPECT_EQ((*radSolver1)[0].tagName(), "ENABLE_SAMPLED_DIFFUSE");
  EXPECT_FALSE(radSolver1->requiredVariable<bool>("ENABLE_SAMPLED_DIFFUSE"));

  const auto* mat2 = top.subsection("MATERIAL_ENTRY", 1);
  ASSERT_NE(mat2, nullptr);
  EXPECT_EQ(mat2->name(), "MATERIAL_ENTRY");
  EXPECT_TRUE(mat2->requiredVariable<bool>("DOUBLE_SIDED"));
  EXPECT_EQ(
    mat2->optionalVariable<std::string>("NOT_THERE", 0, "default"), "default");
  EXPECT_EQ(mat2->element("ID"), &(*mat2)[0]);
  EXPECT_EQ(mat2->element("foo"), nullptr);
  EXPECT_EQ(&mat2->requiredElement("ID"), &(*mat2)[0]);
  EXPECT_THROW(auto e = mat2->requiredElement("foo"), std::exception);
  EXPECT_EQ(mat2->subsection("ID"), nullptr);
  EXPECT_EQ(mat2->size(), 9u);

  size_t numIter = 0;
  for (const auto& e : *mat2) {
    EXPECT_EQ(&e, &(*mat2)[numIter++]);
  }
  EXPECT_EQ(numIter, mat2->size());

  EXPECT_EQ(
    mat2->requiredSubsection("SURFACE_PROPERTIES")
      .requiredSubsection("EMISSIVITY_PROP")
      .requiredVariable<test::Filename>("FILENAME"),
    "tire_black.ems");
}
