#include "dirsig5/d4legacy/xml/xml.h"

#include "dirsig5/base/Error.h"
#include "spdlog/fmt/fmt.h"

using namespace std::string_view_literals;

namespace d5::d4legacy::xml {

Document loadFile(const char* filename) {
  Document doc;
  if (auto result = doc.load_file(filename); !result)
    DIRSIG_THROW_FMT(
      "can't load XML file \"{}\": {}", filename, result.description());
  return doc;
}

Document loadString(const char* file_content) {
  Document doc;
  if (auto result = doc.load_string(file_content); !result)
    DIRSIG_THROW_FMT("can't load XML from string: {}", result.description());
  return doc;
}

Node requiredChild(const Node& node, const char* name) {
  auto subnode = node.child(name);
  if (!subnode)
    DIRSIG_THROW_FMT(
      "line {}: in <{}>: can't find required XML element <{}>", lineno(node),
      node.name(), name);
  if (size(node.children(name)) > 1)
    DIRSIG_THROW_FMT(
      "line {}: in <{}>: the XML element <{}> must be unique", lineno(node),
      node.name(), name);
  return subnode;
}

Attr requiredAttribute(const Node& node, const char* name) {
  auto attr = node.attribute(name);
  if (!attr)
    DIRSIG_THROW_FMT(
      "line {}: in <{}>: can't find required XML attribute '{}'", lineno(node),
      node.name(), name);
  size_t n = 0;
  for (auto attr : node.attributes())
    if (attr.name() == std::string_view(name)) n++;
  if (n > 1)
    DIRSIG_THROW_FMT(
      "line {}: in <{}>: the XML attribute '{}' must be unique", lineno(node),
      node.name(), name);
  return attr;
}

long lineno(const Node& node) {
  // The offset_debug() observer returns the offset of
  // the relevant pointer in the original buffer if it is possible
  // to do so (i.e., only 1 buffer has ever been loaded). The relevant
  // pointer is the name or the value, if the name is null. To find the
  // line number, we can count the number of newlines from the beginning.
  long offset = node.offset_debug();
  if (offset < 0) return -1;
  const char* ptr = node.name();
  if (!ptr) ptr = node.value();
  if (!ptr) return -1;
  return std::count(ptr - offset, ptr, '\n') + 1;
}

DateTime readDatetime(const Node& node, const char* name) {
  if (auto child = node.child(name)) {
    std::string_view type = requiredAttribute(node, "type").as_string();
    if (type == "relative") {
      return DateTime::fromRelativeSeconds(node.text().as_double());
    } else if (type == "absolute") {
      return DateTime::fromIso8601(node.text().as_string());
    } else {
      DIRSIG_THROW_FMT(
        "line {}: in <{}>: the XML attribute 'type' must be 'relative' or "
        "'absolute'",
        lineno(node), node.name(), name);
    }
  }
  return DateTime::fromRelativeSeconds(0.0);
}

std::string readFilename(
  const Node& node,
  const char* name,
  const FileLocator& fileLocator,
  const FileLocator::FileType type) {
  std::string filename = requiredChild(node, name).text().as_string();
  try {
    return fileLocator.locateFile(filename, type);
  } catch (...) {
    DIRSIG_THROW_FMT("line {}: can't locate file '{}'", lineno(node), filename);
  }
  return {};
}

DistanceUnits readDistanceUnitsAttribute(const Node& node, const char* name) {
  auto attr = node.attribute(name);
  if (!attr) return DistanceUnits::Meters;
  DistanceUnits units;
  if (!stringToUnits(attr.as_string(), units))
    DIRSIG_THROW_FMT(
      "line {}: can't parse distance units '{}'", lineno(node),
      attr.as_string());
  return units;
}

WavelengthUnits readWavelengthUnitsAttribute(
  const Node& node, const char* name) {
  auto attr = node.attribute(name);
  if (!attr) return WavelengthUnits::Microns;
  WavelengthUnits units;
  if (!stringToUnits(attr.as_string(), units))
    DIRSIG_THROW_FMT(
      "line {}: can't parse wavelength units '{}'", lineno(node),
      attr.as_string());
  return units;
}

double readAngularUnitsAttributeAsMultiplier(
  const Node& node, const char* name) {
  auto attr = node.attribute(name);
  if (!attr) return 1.0;
  if (attr.as_string() == "radians"sv) return 1.0;
  if (attr.as_string() == "degrees"sv) return M_PI / 180.0;
  DIRSIG_THROW_FMT(
    "line {}: can't parse angular units '{}'", lineno(node), attr.as_string());
  return 0;
}

RotationOrder readRotationOrderAttribute(const Node& node, const char* name) {
  auto attr = node.attribute(name);
  if (!attr) return RotationOrder::Xyz;
  try {
    return ParseRotationOrder(attr.as_string());
  } catch (...) {
    DIRSIG_THROW_FMT(
      "line {}: can't parse rotation order '{}'", lineno(node),
      attr.as_string());
  }
  return {};
}

static Point read_enu(const Node& node) {
  auto subnode = node.child("point");
  if (!subnode)
    return readCoordinate(node);
  else
    return readCoordinate(subnode);
}

static UtmCoord read_utm(const Node& node) {
  // extract hemisphere
  auto hemisphereText = requiredChild(node, "hemisphere").value();
  char hemisphere = '\0';
  if (hemisphereText == "N"sv)
    hemisphere = 'N';
  else if (hemisphereText == "S"sv)
    hemisphere = 'S';
  else
    throw std::runtime_error("hemisphere value must be N or S");
  return UtmCoord(
    requiredChild(node, "zone").text().as_int(), hemisphere,
    requiredChild(node, "easting").text().as_double(),
    requiredChild(node, "northing").text().as_double(),
    requiredChild(node, "altitude").text().as_double());
}

static GeodeticCoord read_geodetic(const Node& node) {
  return GeodeticCoord(
    requiredChild(node, "latitude").text().as_double(),
    requiredChild(node, "longitude").text().as_double(),
    requiredChild(node, "altitude").text().as_double());
}

static EcefCoord read_ecef(const Node& node) { return readCoordinate(node); }

geolib::Location readLocation(const Node& node, const geolib::Facade* facade) {
  if (node.name() == "location"sv) {
    auto frame = node.attribute("frame");
    if (!frame) frame = node.attribute("type");
    if (frame.value() == "scene"sv)
      return geolib::Location(read_enu(node), facade);
    if (frame.value() == "geodetic"sv)
      return geolib::Location(read_geodetic(node), facade);
    if (frame.value() == "utm"sv)
      return geolib::Location(read_utm(node), facade);
    if (frame.value() == "ecef"sv)
      return geolib::Location(read_ecef(node), facade);
    DIRSIG_THROW_FMT(
      "line {}: unknown location frame type \"{}\"", lineno(node),
      frame.value());
  } else {
    if (node.name() == "scenelocation"sv)
      return geolib::Location(read_enu(node), facade);
    if (node.name() == "geodeticlocation"sv)
      return geolib::Location(read_geodetic(node), facade);
    if (node.name() == "utmlocation"sv)
      return geolib::Location(read_utm(node), facade);
    if (node.name() == "eceflocation"sv)
      return geolib::Location(read_ecef(node), facade);
    DIRSIG_THROW_FMT(
      "line {}: unknown location type \"{}\"", lineno(node), node.name());
  }
}

}  // namespace d5::d4legacy::xml
