/** Interface file for ENVI header I/O
 *
 *  \file HeaderIO.h
 */

#pragma once

#include <algorithm>
#include <functional>
#include <istream>
#include <sstream>
#include <string>
#include <vector>

#include <spdlog/fmt/fmt.h>

#include "dirsig5/base/Error.h"
#include "dirsig5/envi/Header.h"

namespace d5::envi {

/// Base template for I/O of a field from and ENVI header value
template <typename T, typename SFINAE = void>
struct header_field_io {
  template <typename U>
  static constexpr bool always_false = false;

  static void serialize(std::string& s, const T& val) {}

  static void deserialize(const std::string& s, T& val) {}

  static_assert(
      always_false<T>, "No header_field_io specialization for this type");
};

template <typename T>
void serialize_header_field(std::string& s, const T& val) {
  header_field_io<T>::serialize(s, val);
}

template <typename T>
void deserialize_header_field(const std::string& s, T& val) {
  header_field_io<T>::deserialize(s, val);
}

/// Specialization for signed integer types
template <typename T>
struct header_field_io<T, std::enable_if_t<std::is_signed_v<T>>> {
  static void serialize(std::string& s, const T& val) {
    s = std::to_string(val);
  }

  static void deserialize(const std::string& s, T& val) try {
    auto parsed = std::stoll(s);
    DIRSIG_THROW_IF_FMT(
        parsed >= std::numeric_limits<T>::max(),
        "Can't parse signed integer from \"{}\", value out of bounds", s);
    val = static_cast<T>(parsed);
  } catch (Error&) {
    throw;
  } catch (const std::exception&) {
    DIRSIG_THROW_FMT("Can't parse signed integer from \"{}\"", s);
  }
};

/// Specialization for unsigned integer types
template <typename T>
struct header_field_io<T, std::enable_if_t<std::is_unsigned_v<T>>> {
  static void serialize(std::string& s, const T& val) {
    s = std::to_string(val);
  }

  static void deserialize(const std::string& s, T& val) try {
    auto parsed = std::stoull(s);
    DIRSIG_THROW_IF_FMT(
        parsed >= std::numeric_limits<T>::max(),
        "Can't parse unsigned integer from \"{}\", value out of bounds", s);
    val = static_cast<T>(parsed);
  } catch (Error&) {
    throw;
  } catch (const std::exception&) {
    DIRSIG_THROW_FMT("Can't parse unsigned integer from \"{}\"", s);
  }
};

/// Specialization for floating point numbers
template <>
struct header_field_io<double, void> {
  static void serialize(std::string& s, double val) { s = std::to_string(val); }

  static void deserialize(const std::string& s, double& val) try {
    val = std::stod(s);
  } catch (const std::exception&) {
    DIRSIG_THROW_FMT("Can't parse floating point number from \"{}\"", s);
  }
};

/// Specialization for strings
template <>
struct header_field_io<std::string, void> {
  static void serialize(std::string& s, std::string val) { s = std::move(val); }

  static void deserialize(const std::string& s, std::string& val) {
    val = s;

    // Trim off whitespace from the end
    val.erase(
        std::find_if(
            val.rbegin(), val.rend(),
            [](unsigned char ch) { return !std::isspace(ch); })
            .base(),
        val.end());
    auto p = val.find_first_not_of(' ');
    if (p == std::string::npos)
      val.clear();
    else if (p != 0)
      val = val.substr(p);
  }
};

namespace detail {
template <typename RandomAccessIterator>
std::string ImplodeComma(RandomAccessIterator beg, RandomAccessIterator end) {
  std::string s;
  size_t total = 0;
  size_t size = end - beg;
  std::vector<std::string> tokens;
  tokens.reserve(size);
  for (auto it = beg; it != end; ++it) {
    serialize_header_field(tokens.emplace_back(), *it);
    total += tokens.back().size();
  }
  s.reserve(total + tokens.size() + 1);
  s += '{';
  for (size_t i = 0; i < tokens.size(); i++) {
    s += tokens[i];
    if (i + 1 < tokens.size()) s += ',';
  }
  s += '}';
  return s;
}
}

/// Specialization for std::vector's
template <typename T>
struct header_field_io<std::vector<T>, void> {
  static void serialize(std::string& s, const std::vector<T>& val) {
    s = detail::ImplodeComma(val.begin(), val.end());
  }

  static void deserialize(const std::string& s, std::vector<T>& val) {
    // tokenize the string containing the full list
    std::vector<std::string> tokens;

    auto isDelimiter = [](char c) { return c == ','; };

    // walk through the source string and make tokens
    auto stringEnd = s.end();
    for (auto it = s.begin(); it != stringEnd; ++it) {
      std::string token;
      std::insert_iterator<std::string> ii(token, token.begin());
      while (it != stringEnd && isDelimiter(*it)) it++;
      while (it != stringEnd && !isDelimiter(*it)) *ii++ = *it++;
      tokens.push_back(token);
      if (it == stringEnd) break;
    }

    val.resize(tokens.size());
    for (size_t i = 0; i < tokens.size(); i++) {
      DIRSIG_TRY { deserialize_header_field(tokens[i], val[i]); }
      DIRSIG_CATCH(error, {
        throw error.withTraceMessage(
            fmt::format("parsing element {} in a list (\"{}\")", i, tokens[i]));
      })
    }
  }
};

/// Specialization for std::array's
template <typename T, std::size_t N>
struct header_field_io<std::array<T, N>, void> {
  static void serialize(std::string& s, const std::array<T, N>& val) {
    s = detail::ImplodeComma(val.begin(), val.end());
  }

  static void deserialize(const std::string& s, std::array<T, N>& val) {
    std::vector<T> vals;
    deserialize_header_field(s, vals);
    DIRSIG_THROW_IF_FMT(
        vals.size() != N, "expected {} values, but found {}", N, vals.size());
    for (size_t i = 0; i < N; i++) val[i] = std::move(vals[i]);
  }
};

/// Specialization for interleave types
template <>
struct header_field_io<Interleave, void> {
  static void serialize(std::string& s, Interleave val) { s = to_string(val); }

  static void deserialize(std::string s, Interleave& val) {
    std::transform(s.begin(), s.end(), s.begin(), ::tolower);
    if (s == "bsq") {
      val = Interleave::BSQ;
    } else if (s == "bil") {
      val = Interleave::BIL;
    } else if (s == "bip") {
      val = Interleave::BIP;
    } else {
      DIRSIG_THROW_FMT("invalid interleave type \"{}\"", s);
    }
  }
};

/// Specialization for complex functions
template <>
struct header_field_io<Header::ComplexFunction, void> {
  static void serialize(std::string& s, Header::ComplexFunction val) {
    s = to_string(val);
  }

  static void deserialize(const std::string& s, Header::ComplexFunction& val) {
    if (s == "Real") {
      val = Header::ComplexFunction::Real;
    } else if (s == "Imaginary") {
      val = Header::ComplexFunction::Imaginary;
    } else if (s == "Power") {
      val = Header::ComplexFunction::Power;
    } else if (s == "Magnitude") {
      val = Header::ComplexFunction::Magnitude;
    } else if (s == "Phase") {
      val = Header::ComplexFunction::Phase;
    } else {
      DIRSIG_THROW_FMT("invalid complex function type \"{}\"", s);
    }
  }
};

template <typename T>
struct header_field_io<std::optional<T>, void> {
  static void serialize(std::string& s, const std::optional<T>& val) {
    serialize_header_field(s, *val);
  }

  static void deserialize(const std::string& s, std::optional<T>& val) {
    val.reset();
    deserialize_header_field(s, val.emplace());
  }
};
}
