/** Implementation file for ENVI image reading
 *
 *  \file HeaderIO.cpp
 */

#include "dirsig5/envi/HeaderIO.h"

#include <cstdint>
#include <fstream>
#include <iostream>
#include <string_view>
#include <unordered_map>
#include <unordered_set>
#include <utility>

#include <tbb/parallel_for.h>

#include "DataTypeIO.h"
#include "EndianIO.h"
#include "HeaderFieldIO.h"
#include "WavelengthUnitsIO.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/StringUtils.h"

using d5::trimmed;
using d5::trimmedView;

namespace {

/// A simple helper struct for reading files line-by-line
struct LineReader {
  LineReader(std::istream &is) : stream(is) {}

  std::istream &stream;
  std::string line;
  std::size_t lineno = 0;

  bool nextLineOptional() {
    if (std::getline(stream, line)) {
      lineno++;
      return true;
    }
    return false;
  }

  void nextLineRequired() {
    if (!nextLineOptional())
      fail("unexpected EOF");
  }

  void fail(const char *what) { DIRSIG_THROW_FMT("line {}: {}", lineno, what); }

  void fail(const std::string &what) {
    DIRSIG_THROW_FMT("line {}: {}", lineno, what);
  }
};

std::vector<std::string> tokenize(const std::string &source,
                                  const std::string &delims) {
  auto isDelimiter = [](const std::string &delims, char c) {
    return delims.find(c) != std::string::npos;
  };

  // walk through the source string and make tokens
  std::vector<std::string> tokenVector;
  auto stringEnd = source.end();
  for (auto it = source.begin(); it != stringEnd; ++it) {
    std::string token;
    std::insert_iterator<std::string> ii(token, token.begin());
    while (it != stringEnd && isDelimiter(delims, *it))
      it++;
    while (it != stringEnd && !isDelimiter(delims, *it))
      *ii++ = *it++;
    tokenVector.push_back(token);
    if (it == stringEnd)
      break;
  }
  return tokenVector;
}

std::pair<std::string, std::string> ParseDeclaration(LineReader &reader) {
  // Trim off whitespace and skip empty/comment lines
  auto trimmedLine = trimmedView(reader.line);
  auto equalsPos = trimmedLine.find_first_of('=');
  if (equalsPos == 0 || equalsPos == trimmedLine.size() - 1 ||
      equalsPos == std::string_view::npos) {
    reader.fail(
        fmt::format("invalid variable declaration line, expected \"field_name "
                    "= field_value\":\n  \"{}\"",
                    trimmedLine));
  }
  std::string field = trimmed(trimmedLine.substr(0, equalsPos));
  std::string value;
  auto valueLine = trimmedView(trimmedLine.substr(equalsPos + 1));
  if (valueLine[0] == '{') {
    valueLine.remove_prefix(1);
    do {
      auto v = trimmedView(valueLine);
      if (auto p = v.find_first_of('}'); p == std::string_view::npos) {
        value += v;
        reader.nextLineRequired();
        valueLine = reader.line;
      } else {
        if (p == v.size() - 1) {
          value += trimmedView(v.substr(0, v.size() - 1));
          break;
        } else {
          reader.fail("'}' must be the end of the line for an array-value");
        }
      }
    } while (true);
  } else {
    value = valueLine;
  }

  return std::pair<std::string, std::string>{std::move(field),
                                             std::move(value)};
}
#if 0
template <int N, typename T>
bool Parse(
    const std::string& field,
    const std::string& value,
    const char (&fieldName)[N],
    T& dst) {
  DIRSIG_ERROR_TRACE("parsing field \"" + field + "\"",  {
    if (field == fieldName) {
      d5::envi::deserialize_header_field(value, dst);
      return true;
    }
    return false;
  });
}
#endif

} // namespace

d5::envi::Header d5::envi::ReadHeader(std::istream &is) {
  d5::envi::Header hdr;
  LineReader reader(is);

  reader.nextLineRequired();
  DIRSIG_THROW_IF(trimmedView(reader.line) != "ENVI",
                  "ENVI header did not start with \"ENVI\"");

#define HDR_HANDLER(name, dst)                                                 \
  {                                                                            \
    name, [&](int lineno, const std::string &value) {                          \
      DIRSIG_ERROR_TRACE(                                                      \
          fmt::format("parsing field \"{}\" on line {}", name, lineno),        \
          { d5::envi::deserialize_header_field(value, hdr.dst); });            \
    }                                                                          \
  }

  using HandlerMap =
      std::unordered_map<std::string,
                         std::function<void(int, const std::string &)>>;
  HandlerMap handlers{
      HDR_HANDLER("acquisition time", acquisitionTime),
      HDR_HANDLER("band names", bandNames),
      HDR_HANDLER("bands", bands),
      HDR_HANDLER("bbl", badBandMultipliers),
      HDR_HANDLER("byte order", byteOrder),
      HDR_HANDLER("cloud cover", cloudCover),
      HDR_HANDLER("complex function", complexFunction),
      HDR_HANDLER("coordinate system string", coordSystemString),
      HDR_HANDLER("data gain values", dataGainValues),
      HDR_HANDLER("data offset values", dataOffsetValues),
      HDR_HANDLER("data reflectance gain values", dataReflectanceGainValues),
      HDR_HANDLER("data reflectance offset values",
                  dataReflectanceOffsetValues),
      HDR_HANDLER("data type", dataType),
      HDR_HANDLER("data units", dataUnits),
      HDR_HANDLER("default bands", defaultBands),
      HDR_HANDLER("default stretch", defaultStretch),
      HDR_HANDLER("dem band", demBand),
      {"file_type",
       [&](int, const std::string &value) {
         DIRSIG_THROW_IF_FMT(
             value != "ENVI Standard",
             R"(only "ENVI Standard" files are supported, this is a "{}" file)",
             value);
       }},
      HDR_HANDLER("description", description),
      HDR_HANDLER("fwhm", fwhm),
      HDR_HANDLER("geo points", geoPoints),
      HDR_HANDLER("header offset", headerOffset),
      HDR_HANDLER("interleave", interleave),
      HDR_HANDLER("lines", lines),
      HDR_HANDLER("map info", mapInfo),
      HDR_HANDLER("pixel size", pixelSize),
      HDR_HANDLER("reflectance scale factor", reflectanceScaleFactor),
      HDR_HANDLER("rpc info", rpcInfo),
      HDR_HANDLER("samples", samples),
      HDR_HANDLER("security tag", securityTag),
      HDR_HANDLER("sensor type", sensorType),
      HDR_HANDLER("solar irradiance", solarIrradiance),
      HDR_HANDLER("sun azimuth", sunAzimuthDegCwFromN),
      HDR_HANDLER("sun elevation", sunElevationDeg),
      HDR_HANDLER("wavelength", wavelength),
      HDR_HANDLER("wavelength units", wavelengthUnits),
      HDR_HANDLER("x start", xStart),
      HDR_HANDLER("y start", yStart),
      HDR_HANDLER("spectra names", spectraNames)};
#undef HDR_HANDLER

  std::unordered_set<std::string> seenFields;
  std::string requiredFields[] = {"byte order", "data type", "header offset",
                                  "interleave", "lines",     "samples"};

  while (reader.nextLineOptional()) {
    // Trim off whitespace and skip empty/comment lines
    auto trimmedLine = trimmedView(reader.line);
    if (trimmedLine.empty() || trimmedLine[0] == ';')
      continue;

    auto [field, value] = ParseDeclaration(reader);
    seenFields.insert(field);

    if (auto it = handlers.find(field); it != handlers.end()) {
      it->second(reader.lineno, value);
    } else {
      hdr.nonstandardFields[field] = value;
    }
  }

  for (const auto &f : requiredFields) {
    DIRSIG_THROW_IF_FMT(seenFields.count(f) == 0u,
                        "ENVI header missing required field \"{}\"", f);
  }

  return hdr;
}
