#include "dirsig5/envi/SpectralLibrary.h"

#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/envi/ImageIO.h"
#include "dirsig5/raster/Raster.h"

namespace d5::envi {

class SliImageSpecific final : public BsqImageDestination {
public:
  void prepare(const Header &header) override {
    if (indexOrName.index() == 0) {
      index = std::get<int>(indexOrName);
    } else {
      bool found = false;
      for (size_t i = 0; i < header.spectraNames.size(); i++)
        if (std::get<std::string>(indexOrName) == header.spectraNames[i]) {
          index = i;
          found = true;
          break;
        }
      DIRSIG_THROW_IF(not found,
                      "The requested ENVI SLI spectrum name wasn't found");
    }
    if (index < header.spectraNames.size()) {
      library->spectraNames.resize(1);
      library->spectraNames[0] = header.spectraNames[index];
    }
  }
  void allocate(size_t bands, size_t rows, size_t columns,
                DType dataType) override {
    DIRSIG_THROW_IF(bands != 1, "Expected only 1 band in ENVI SLI");
    DIRSIG_THROW_IF(index >= rows,
                    "The requested ENVI SLI index is out of range");
    library->spectra.resize(1);
    library->spectra[0].resize(columns);
    rasterData.create(1, columns, 1, DType(dataType));
  }

  char *row(size_t, size_t r) override {
    return r == index ? reinterpret_cast<char *>(rasterData.rawRow(0))
                      : nullptr;
  }

  void finalize() {
    rasterData.dispatch([&](auto im) {
      for (size_t c = 0; c < im.cols(); c++)
        library->spectra.front()[c] = im.pixel(0, c)[0];
    });
  }

  size_t index = size_t(-1);
  std::variant<int, std::string> indexOrName;
  SpectralLibrary *library = nullptr;
  raster::Raster rasterData;
};

void SpectralLibrary::read(const std::string &filename,
                           const std::variant<int, std::string> &indexOrName) {
  DIRSIG_ERROR_TRACE("reading ENVI SLI", {
    clear();
    SliImageSpecific sliImage;
    sliImage.library = this;
    sliImage.indexOrName = indexOrName;
    auto header =
        LoadImage(ImageReadRequest().withImageFilename(filename), sliImage);
    sliImage.finalize();
    wavelength = header.wavelength;
    if (header.reflectanceScaleFactor)
      reflectanceScale = *header.reflectanceScaleFactor;
    if (auto itr = header.nonstandardFields.find("description");
        itr != header.nonstandardFields.end())
      description = itr->second;
  });
}

class SliImage final : public BsqImageDestination {
public:
  void allocate(size_t bands, size_t rows, size_t columns,
                DType dataType) override {
    DIRSIG_THROW_IF(bands != 1, "Expected only 1 band in ENVI SLI");
    library->spectra.resize(rows);
    for (auto &spectrum : library->spectra)
      spectrum.resize(columns);
    rasterData.create(rows, columns, 1, DType(dataType));
  }

  char *row(size_t, size_t r) override {
    return reinterpret_cast<char *>(rasterData.rawRow(r));
  }

  void finalize() {
    rasterData.dispatch([&](auto im) {
      for (size_t r = 0; r < im.rows(); r++)
        for (size_t c = 0; c < im.cols(); c++)
          library->spectra[r][c] = im.pixel(r, c)[0];
    });
  }

  SpectralLibrary *library = nullptr;
  raster::Raster rasterData;
};

void SpectralLibrary::read(const std::string &filename) {
  DIRSIG_ERROR_TRACE("reading ENVI SLI", {
    clear();
    SliImage sliImage;
    sliImage.library = this;
    auto header =
        LoadImage(ImageReadRequest().withImageFilename(filename), sliImage);
    sliImage.finalize();
    wavelength = header.wavelength;
    if (header.reflectanceScaleFactor)
      reflectanceScale = *header.reflectanceScaleFactor;
    if (auto itr = header.nonstandardFields.find("description");
        itr != header.nonstandardFields.end())
      description = itr->second;
    spectraNames = header.spectraNames;
  });
}

void SpectralLibrary::write(const std::string &filename) const {
  DIRSIG_ERROR_TRACE("writing ENVI SLI", {
    std::ofstream header(filename + ".hdr");
    DIRSIG_THROW_IF(!header.is_open(), "Can't open \"" + filename + ".hdr\"");
    header << "ENVI\n";
    header << "description = {" << description << "}\n";
    header << "samples = " << wavelength.size() << '\n';
    header << "lines = " << spectra.size() << '\n';
    header << "bands = 1\n"
              "header offset = 0\n"
              "file type = ENVI Spectral Library\n"
              "data type = 5\n"
              "interleave = bsq\n"
              "byte order = 0\n"
              "wavelength units = Micrometers\n";
    header << "reflectance scale factor = " << reflectanceScale << '\n';
    header << "spectra names = {";
    for (size_t i = 0; i < spectraNames.size(); i++) {
      header << spectraNames[i];
      if (i + 1 != spectraNames.size())
        header << ", ";
    }
    header << "}\n";
    header << "wavelength = {";
    for (size_t i = 0; i < wavelength.size(); i++) {
      header << wavelength[i];
      if (i + 1 != wavelength.size())
        header << ", ";
    }
    header << "}\n";
    auto binary = OpenOfstream(filename, std::ios::binary);
    for (auto &spectrum : spectra)
      binary.write(reinterpret_cast<const char *>(spectrum.data()),
                   std::streamsize(sizeof(double) * spectrum.size()));
  });
}

SpectralLibrary::Statistics SpectralLibrary::computeStatistics() const {
  Statistics statistics;
  auto numWavelengths = Eigen::Index(wavelength.size());
  auto numSpectra = Eigen::Index(spectra.size());
  auto &mean = statistics.mean;
  auto &var = statistics.var;
  auto &cov = statistics.cov;
  mean.resize(numWavelengths);
  var.resize(numWavelengths);
  cov.resize(numWavelengths, numWavelengths);
  if (numSpectra > 0) {
    for (const SpectralVector &spectrum : spectra)
      mean.array() += spectrum.array();
    mean *= 1.0 / double(numSpectra);
    if (numSpectra > 1) {
      for (const SpectralVector &spectrum : spectra) {
        Eigen::VectorXd offset = spectrum.array() - mean.array();
        cov += offset * offset.transpose();
      }
      cov *= 1.0 / double(numSpectra - 1);
    }
    var = cov.diagonal();
  }
  return statistics;
}

} // namespace d5::envi
