#include <fstream>
#include <iostream>

#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/Logger.h"
#include "dirsig5/envi/StatisticsFile.h"

bool d5::envi::StatisticsFile::read(const std::string &filename) {
  // open the binary STA file
  std::ifstream staFile(filename);
  DIRSIG_THROW_IF(staFile.fail(), "Failed to open ENVI statistics file!");

  // attempt to read the header
  staFile.read((char *)&header, sizeof(header));
  DIRSIG_THROW_IF(staFile.fail(),
                  "Failed to read ENVI statistics file header!");

  // check if we need to byte swap
  swapBytes = false;
  if (header.id == 1313426753) {
    swapBytes = true;
    header.sampleCount = byteSwap(header.sampleCount);
    header.lineCount = byteSwap(header.lineCount);
    header.bandCount = byteSwap(header.bandCount);
    header.dataType = byteSwap(header.dataType);
    header.roiIndex = byteSwap(header.roiIndex);
    header.startSample = byteSwap(header.startSample);
    header.endSample = byteSwap(header.endSample);
    header.startLine = byteSwap(header.startLine);
    header.endLine = byteSwap(header.endLine);
    DIRSIG_DEBUG("  ID = {} (new format, non-native endian)", header.id);
  } else if (header.id == 1095584078) {
    swapBytes = false;
    DIRSIG_DEBUG("  ID = {} (new format, native endian)", header.id);
  } else {
    DIRSIG_THROW("Attempted to read an unsupported or older format STA file!");
  }

  DIRSIG_DEBUG("  sampleCount = {}\n"
               "  lineCount   = {}\n"
               "  bandCount   = {}\n"
               "  dataType    = {}\n"
               "  ROI index   = {}\n"
               "  ROI Box     = [{},{}] -> [{},{}]\n",
               header.sampleCount, header.lineCount, header.bandCount,
               header.dataType, header.roiIndex, header.startSample,
               header.startLine, header.endSample, header.endLine);

  // read the histogram offsets
#ifdef _DEBUG
  std::cout << "histogram offsets:" << std::endl;
#endif
  offsets.histogramOffsets.resize(header.bandCount);
  for (int32_t bi = 0; bi < header.bandCount + 1; bi++) {
    int32_t offset;
    staFile.read((char *)&offset, sizeof(offset));
    if (swapBytes)
      offset = byteSwap(offset);
    if (bi < header.bandCount) {
      offsets.histogramOffsets[bi] = offset;
#ifdef _DEBUG
      if (offset > 0)
        std::cout << offset << std::endl;
#endif
    } else {
      offsets.covarianceOffset = offset;
    }
  }
#ifdef _DEBUG
  std::cout << "Covariance offset = " << offsets.covarianceOffset << std::endl;
#endif

  // read the ROI name
  staFile.read((char *)&offsets.roiNameLength, sizeof(offsets.roiNameLength));
  if (swapBytes)
    offsets.roiNameLength = byteSwap(offsets.roiNameLength);
  offsets.roiName.resize(offsets.roiNameLength);
  staFile.read((char *)offsets.roiName.data(),
               sizeof(char) * offsets.roiNameLength);
  DIRSIG_DEBUG("  ROI name    = '{}'", offsets.roiName);

  // read the band wavelengths
#ifdef _DEBUG
  std::cout << "wavelengths:" << std::endl;
#endif
  offsets.wavelengths.resize(header.bandCount);
  for (int32_t bi = 0; bi < header.bandCount; bi++) {
    float wvl;
    staFile.read((char *)&wvl, sizeof(wvl));
    if (swapBytes)
      wvl = byteSwap(wvl);
    offsets.wavelengths[bi] = wvl;
#ifdef _DEBUG
    std::cout << wvl << std::endl;
#endif
  }

  DIRSIG_THROW_IF(staFile.fail(),
                  "Failed to read offsets in ENVI statistics file!");

  // read the band statistics flags
  bandFlags.hasStats = new char[header.bandCount];
  staFile.read((char *)bandFlags.hasStats, sizeof(char) * header.bandCount);
#ifdef _DEBUG
  for (int32_t bi = 0; bi < header.bandCount; bi++) {
    std::cout << bi << ": " << (int)bandFlags.hasStats[bi] << std::endl;
  }
#endif

  // read the band statistics
  bandStats.bandMin.resize(header.bandCount);
  bandStats.bandMax.resize(header.bandCount);
  bandStats.bandMean.resize(header.bandCount);
  bandStats.bandStddev.resize(header.bandCount);
  auto readStats = [this](std::ifstream &stream, std::vector<double> &stats) {
    for (int32_t bi = 0; bi < header.bandCount; bi++) {
      double val;
      stream.read((char *)&val, sizeof(val));
      if (swapBytes)
        val = byteSwap(val);
      stats[bi] = val;
    }
  };
  readStats(staFile, bandStats.bandMin);
  readStats(staFile, bandStats.bandMax);
  readStats(staFile, bandStats.bandMean);
  readStats(staFile, bandStats.bandStddev);

  DIRSIG_THROW_IF(staFile.fail(),
                  "Failed to read band statistics in ENVI statistics file!");

#ifdef _DEBUG
  for (int32_t bi = 0; bi < header.bandCount; bi++) {
    std::cout << bi << ": " << bandStats.bandMin[bi] << " "
              << bandStats.bandMax[bi] << " " << bandStats.bandMean[bi] << " "
              << bandStats.bandStddev[bi] << std::endl;
  }
#endif

  // seek to the start of the covariance data
  staFile.seekg(offsets.covarianceOffset);

  // read the covariance header
  staFile.read((char *)&covarianceHeader.bandCount,
               sizeof(covarianceHeader.bandCount));
  if (swapBytes)
    covarianceHeader.bandCount = byteSwap(covarianceHeader.bandCount);
  covarianceHeader.bandIndexes.resize(covarianceHeader.bandCount);
  for (int32_t bi = 0; bi < covarianceHeader.bandCount; bi++) {
    int32_t idx;
    staFile.read((char *)&idx, sizeof(idx));
    if (swapBytes)
      idx = byteSwap(idx);
    covarianceHeader.bandIndexes[bi] = idx;
  }

#ifdef _DEBUG
  std::cout << "Covariance matrix is [" << covarianceHeader.bandCount << " x "
            << covarianceHeader.bandCount << "]" << std::endl;
  for (int32_t bi = 0; bi < covarianceHeader.bandCount; bi++) {
    std::cout << bi << ": " << covarianceHeader.bandIndexes[bi] << std::endl;
  }
#endif

  covarianceData.covarianceMatrix.resize(covarianceHeader.bandCount *
                                         covarianceHeader.bandCount);
  for (int32_t ii = 0; ii < covarianceHeader.bandCount; ii++) {
    for (int32_t jj = 0; jj < covarianceHeader.bandCount; jj++) {
      double val;
      staFile.read((char *)&val, sizeof(val));
      if (swapBytes)
        val = byteSwap(val);
      covarianceData.covarianceMatrix[(ii * covarianceHeader.bandCount) + jj] =
          val;
    }
  }
  covarianceData.eigenMatrix.resize(covarianceHeader.bandCount *
                                    covarianceHeader.bandCount);
  for (int32_t ii = 0; ii < covarianceHeader.bandCount; ii++) {
    for (int32_t jj = 0; jj < covarianceHeader.bandCount; jj++) {
      double val;
      staFile.read((char *)&val, sizeof(val));
      if (swapBytes)
        val = byteSwap(val);
      covarianceData.eigenMatrix[(ii * covarianceHeader.bandCount) + jj] = val;
    }
  }
#ifdef _DEBUG
  std::cout << "Eigen values:" << std::endl;
#endif
  covarianceData.eigenValues.resize(covarianceHeader.bandCount);
  for (int32_t bi = 0; bi < covarianceHeader.bandCount; bi++) {
    double val;
    staFile.read((char *)&val, sizeof(val));
    if (swapBytes)
      val = byteSwap(val);
    covarianceData.eigenValues[bi] = val;
#ifdef _DEBUG
    std::cout << bi << ": " << covarianceData.eigenValues[bi] << std::endl;
#endif
  }

  DIRSIG_THROW_IF(staFile.fail(),
                  "Failed to read covariance data in ENVI statistics file!");

  staFile.close();

  return true;
}
