/** Unit tests for the ENVI image reading
 *
 *  \file ImageReadTests.cpp
 */

#include <vector>

#include <gtest/gtest.h>

#include "CLFilesystemTest.h"
#include "CLTemporaryFile.h"
#include "dirsig5/envi/ImageIO.h"

#define FAILURE ::testing::AssertionFailure()
#define SUCCESS ::testing::AssertionSuccess()

template <typename T> struct TestImage {
public:
  TestImage() = default;

  TestImage(size_t bands, size_t rows, size_t cols,
            d5::envi::Interleave interleave) {
    create(bands, rows, cols, interleave);
  }

  [[nodiscard]] size_t bands() const { return bands_; }
  [[nodiscard]] size_t rows() const { return rows_; }
  [[nodiscard]] size_t cols() const { return cols_; }
  [[nodiscard]] d5::envi::Interleave interleave() const { return interleave_; }

  void create(size_t bands, size_t rows, size_t cols,
              d5::envi::Interleave interleave) {
    bands_ = bands;
    rows_ = rows;
    cols_ = cols;
    interleave_ = interleave;
    data_.resize(bands * rows * cols);
  }

  T &operator()(size_t b, size_t r, size_t c) {
    switch (interleave_) {
    case d5::envi::Interleave::BSQ:
      return data_[c + cols_ * (r + rows_ * b)];
    case d5::envi::Interleave::BIL:
      return data_[c + cols_ * (b + bands_ * r)];
    case d5::envi::Interleave::BIP:
    default:
      return data_[b + bands_ * (c + cols_ * r)];
    }
  }

  T operator()(size_t b, size_t r, size_t c) const {
    switch (interleave_) {
    case d5::envi::Interleave::BSQ:
      return data_[c + cols_ * (r + rows_ * b)];
    case d5::envi::Interleave::BIL:
      return data_[c + cols_ * (b + bands_ * r)];
    case d5::envi::Interleave::BIP:
    default:
      return data_[b + bands_ * (c + cols_ * r)];
    }
  }

private:
  size_t bands_ = 0, rows_ = 0, cols_ = 0;
  std::vector<T> data_;
  d5::envi::Interleave interleave_ = d5::envi::Interleave::BSQ;
};

template <typename T> struct TestAdapter : d5::envi::ImageDestination {
public:
  TestAdapter(TestImage<T> &im, d5::envi::Interleave interleave)
      : im_(im), interleave_(interleave) {}

  [[nodiscard]] d5::envi::Interleave interleave() const override {
    return interleave_;
  }

  void allocate(size_t b, size_t r, size_t c, d5::DType dt) override {
    if (dt != d5::dtype_v<T>)
      throw std::runtime_error("wrong data type");
    im_.create(b, r, c, interleave_);
  }

  char *row(size_t b, size_t r) override {
    if (interleave_ == d5::envi::Interleave::BIP) {
      return reinterpret_cast<char *>(&im_(0, r, 0));
    } else {
      return reinterpret_cast<char *>(&im_(b, r, 0));
    }
  }

private:
  TestImage<T> &im_;
  d5::envi::Interleave interleave_;
};

template <class T>
TestAdapter(TestImage<T> &, d5::envi::Interleave) -> TestAdapter<T>;

template <class T> struct is_complex : std::false_type {};
template <class T> struct is_complex<std::complex<T>> : std::true_type {};

template <class T>
TestImage<T> MakeTestImage(size_t bands, size_t rows, size_t cols,
                           d5::envi::Interleave interleave) {
  TestImage<T> im(bands, rows, cols, interleave);
  T val = 0;
  for (size_t b = 0; b < bands; b++) {
    for (size_t r = 0; r < rows; r++) {
      for (size_t c = 0; c < cols; c++) {
        im(b, r, c) = val;
        if constexpr (is_complex<T>::value) {
          val += T{1, 2};
        } else {
          val++;
        }
      }
    }
  }

  return im;
}

struct ImageCleanup {
  ~ImageCleanup() {
    d5::RemoveFile(img);
    d5::RemoveFile(hdr);
  }

  std::string img;
  std::string hdr;
};

template <typename T>
[[nodiscard]] ImageCleanup
WriteImage(TestImage<T> &im, const std::string &filename,
           d5::envi::Endian writeEndian = d5::envi::Endian::Little) {
  std::array<char, 17> padding;
  std::ofstream ofs(filename, std::ios_base::out | std::ios_base::binary);
  if (!ofs)
    throw std::runtime_error("failed to open image");
  ofs.write(padding.data(), padding.size());

  auto native = d5::envi::nativeByteOrder();
  if (native == d5::envi::Endian::Unknown)
    throw std::runtime_error("unknown endian");
  if (writeEndian != native) {
    auto cpy = im;
    for (size_t b = 0; b < im.bands(); b++) {
      for (size_t r = 0; r < im.rows(); r++) {
        for (size_t c = 0; c < im.cols(); c++) {
          if constexpr (is_complex<T>::value) {
            d5::envi::swapEndian(reinterpret_cast<typename T::value_type(&)[2]>(
                cpy(b, r, c))[0]);
            d5::envi::swapEndian(reinterpret_cast<typename T::value_type(&)[2]>(
                cpy(b, r, c))[1]);
          } else {
            d5::envi::swapEndian(cpy(b, r, c));
          }
        }
      }
    }
    ofs.write(reinterpret_cast<char *>(&cpy(0, 0, 0)),
              im.bands() * im.rows() * im.cols() * sizeof(T));
  } else {
    ofs.write(reinterpret_cast<char *>(&im(0, 0, 0)),
              im.bands() * im.rows() * im.cols() * sizeof(T));
  }

  std::ofstream hdr(filename + ".hdr");
  if (!hdr)
    throw std::runtime_error("failed to open header");
  hdr << "ENVI\r\n"
      << "bands = " << im.bands() << "\nbyte order = ";
  hdr << (writeEndian == d5::envi::Endian::Little ? 0 : 1)
      << "\ndata type = " << static_cast<int>(d5::dtype_v<T>)
      << "\r\nfile type = ENVI Standard"
      << "\nheader offset = " << padding.size()
      << "\ninterleave = " << to_string(im.interleave())
      << "\nlines = " << im.rows() << "\nsamples = " << im.cols()
      << "\nband names = {";
  for (std::size_t i = 0; i < im.bands(); i++) {
    hdr << "\n  Band number " << i;
    if (i < im.bands() - 1) {
      hdr << ",";
    } else {
      hdr << "}";
    }
  }

  return ImageCleanup{filename, filename + ".hdr"};
}

template <typename T>
::testing::AssertionResult
ImageEquals(const TestImage<T> &test, const TestImage<T> &truth,
            const std::array<size_t, 4> &roi, size_t stride,
            const std::vector<size_t> &bandRemap) {
  if (bandRemap.empty() && test.bands() != truth.bands()) {
    return FAILURE << "wrong number of bands (" << test.bands() << " vs. "
                   << truth.bands() << ")";
  } else if (!bandRemap.empty() && test.bands() != bandRemap.size()) {
    return FAILURE << "wrong number of bands (" << test.bands() << " vs. "
                   << bandRemap.size() << ")";
  }
  size_t expectedRows = (roi[3] - 1) / stride + 1;
  if (test.rows() != expectedRows) {
    return FAILURE << "wrong number of rows (" << test.rows() << " vs. "
                   << expectedRows << ")";
  }
  size_t expectedCols = (roi[2] - 1) / stride + 1;
  if (test.cols() != expectedCols) {
    return FAILURE << "wrong number of cols (" << test.cols() << " vs. "
                   << expectedCols << ")";
  }

  for (size_t b = 0; b < test.bands(); b++) {
    auto truthBand = bandRemap.empty() ? b : bandRemap[b];

    for (size_t r = 0; r < test.rows(); r++) {
      auto truthRow = roi[1] + r * stride;

      for (size_t c = 0; c < test.cols(); c++) {
        auto truthCol = roi[0] + c * stride;

        if (test(b, r, c) != truth(truthBand, truthRow, truthCol)) {
          return FAILURE << "Test image pixel (b=" << b << ", r=" << r
                         << ", c=" << c
                         << ") is incorrect:\n  value = " << test(b, r, c)
                         << "\n  expected = "
                         << truth(truthBand, truthRow, truthCol)
                         << " (pixel b=" << truthBand << ", r=" << truthRow
                         << ", c=" << truthCol << ")";
        }
      }
    }
  }

  return ::testing::AssertionSuccess();
}

template <typename T>
std::array<size_t, 4> WholeImageRoi(const TestImage<T> &im) {
  return {0, 0, im.cols(), im.rows()};
}

template <typename T>
::testing::AssertionResult HeaderTest(const d5::envi::Header &hdr,
                                      const TestImage<T> &im) {
  if (hdr.bands != im.bands()) {
    return FAILURE << "Header and image bands mismatch (" << hdr.bands
                   << " vs. " << im.bands() << ")";
  }
  if (hdr.lines != im.rows()) {
    return FAILURE << "Header and image rows mismatch (" << hdr.lines << " vs. "
                   << im.rows() << ")";
  }
  if (hdr.samples != im.cols()) {
    return FAILURE << "Header and image columns mismatch (" << hdr.samples
                   << " vs. " << im.cols() << ")";
  }
  if (hdr.interleave != im.interleave()) {
    return FAILURE << "Header and image interleave mismatch ("
                   << to_string(hdr.interleave) << " vs. "
                   << to_string(im.interleave()) << ")";
  }
  if (hdr.bandNames.size() != im.bands()) {
    return FAILURE << "Header did not have enough band names ("
                   << hdr.bandNames.size() << " vs. " << im.bands() << ")";
  }
  for (std::size_t i = 0; i < im.bands(); i++) {
    auto exp = "Band number " + std::to_string(i);
    if (hdr.bandNames[i] != exp) {
      return FAILURE << "Incorrect band name (\"" << hdr.bandNames[i]
                     << "\" vs. \"" << exp << "\")";
    }
  }

  return SUCCESS;
}

template <typename T>
::testing::AssertionResult
ImageTest(size_t bands, size_t rows, size_t cols,
          d5::envi::Interleave interleave,
          const std::optional<std::array<size_t, 4>> &roi = std::nullopt,
          const std::optional<size_t> &stride = std::nullopt,
          const std::vector<size_t> &bandRemap = {},
          d5::envi::Endian writeEndian = d5::envi::Endian::Little) {
  auto truth = MakeTestImage<T>(bands, rows, cols, interleave);
  auto filename = d5::TempFilename(".img");
  auto cleanup = WriteImage(truth, filename, writeEndian);

  auto req = d5::envi::ImageReadRequest().withImageFilename(filename);
  if (roi)
    req.withRoiFromMinSize((*roi)[0], (*roi)[1], (*roi)[2], (*roi)[3]);
  if (stride)
    req.withPixelStride(*stride);
  if (!bandRemap.empty())
    req.withBandsToLoad(bandRemap);

  TestImage<T> readBsq, readBil, readBip;
  d5::envi::Header hdrBsq, hdrBil, hdrBip;
  try {
    hdrBsq = d5::envi::LoadImage(
        req, TestAdapter(readBsq, d5::envi::Interleave::BSQ));
  } catch (const std::exception &e) {
    return FAILURE << "read into BSQ buffer failed with exception:\n"
                   << e.what();
  }
  try {
    hdrBil = d5::envi::LoadImage(
        req, TestAdapter(readBil, d5::envi::Interleave::BIL));
  } catch (const std::exception &e) {
    return FAILURE << "read into BIL buffer failed with exception:\n"
                   << e.what();
  }
  try {
    hdrBip = d5::envi::LoadImage(
        req, TestAdapter(readBip, d5::envi::Interleave::BIP));
  } catch (const std::exception &e) {
    return FAILURE << "read into BIP buffer failed with exception:\n"
                   << e.what();
  }

  auto eq = [&](const TestImage<T> &test) {
    return ImageEquals(test, truth, (roi ? *roi : WholeImageRoi(truth)),
                       (stride ? *stride : 1), bandRemap);
  };

  if (auto r = eq(readBsq); !r) {
    return FAILURE << "image read into BSQ buffer failed equality testing:\n"
                   << r.message();
  }
  if (auto r = eq(readBil); !r) {
    return FAILURE << "image read into BIL buffer failed equality testing:\n"
                   << r.message();
  }
  if (auto r = eq(readBip); !r) {
    return FAILURE << "image read into BIP buffer failed equality testing:\n"
                   << r.message();
  }
  if (auto r = HeaderTest(hdrBsq, truth); !r) {
    return FAILURE << "BSQ header failed testing:\n" << r.message();
  }
  if (auto r = HeaderTest(hdrBil, truth); !r) {
    return FAILURE << "BIL header failed testing:\n" << r.message();
  }
  if (auto r = HeaderTest(hdrBip, truth); !r) {
    return FAILURE << "BIP header failed testing:\n" << r.message();
  }
  return SUCCESS;
}

TEST(ImageReadTests, FullImageReadGrayscaleFloatBip) {
  EXPECT_TRUE(ImageTest<float>(1, 7, 13, d5::envi::Interleave::BIP));
}

TEST(ImageReadTests, FullImageReadDoubleBip) {
  EXPECT_TRUE(ImageTest<double>(3, 7, 13, d5::envi::Interleave::BIP));
}

TEST(ImageReadTests, FullImageReadU32Bsq) {
  EXPECT_TRUE(ImageTest<std::uint32_t>(7, 5, 19, d5::envi::Interleave::BSQ));
}

TEST(ImageReadTests, FullImageReadComplexDoubleBil) {
  EXPECT_TRUE(
      ImageTest<std::complex<double>>(19, 5, 3, d5::envi::Interleave::BIL));
}

TEST(ImageReadTests, StridedReadUint16Bip) {
  EXPECT_TRUE(
      ImageTest<std::uint16_t>(4, 19, 6, d5::envi::Interleave::BIP, {}, 3));
}

TEST(ImageReadTests, StridedReadUint8Bsq) {
  EXPECT_TRUE(
      ImageTest<std::uint8_t>(2, 1, 6, d5::envi::Interleave::BSQ, {}, 3));
}

TEST(ImageReadTests, StridedReadComplexFloatBil) {
  EXPECT_TRUE(ImageTest<std::complex<float>>(2, 17, 21,
                                             d5::envi::Interleave::BIL, {}, 2));
}

template <typename T>
::testing::AssertionResult RoiImageTest(size_t bands, size_t rows, size_t cols,
                                        d5::envi::Interleave interleave,
                                        const std::array<size_t, 4> &roi) {
  return ImageTest<T>(bands, rows, cols, interleave, roi);
}

TEST(ImageReadTests, RoiReadInt16BipTest) {
  EXPECT_TRUE(RoiImageTest<std::int16_t>(4, 19, 5, d5::envi::Interleave::BIP,
                                         {3, 4, 1, 5}));
}

TEST(ImageReadTests, RoiReadFlaot32BsqTest) {
  EXPECT_TRUE(
      RoiImageTest<float>(2, 17, 4, d5::envi::Interleave::BSQ, {1, 5, 3, 7}));
}

TEST(ImageReadTests, RoiReadInt32BilTest) {
  EXPECT_TRUE(RoiImageTest<std::int16_t>(2, 14, 13, d5::envi::Interleave::BIL,
                                         {2, 5, 8, 3}));
}

TEST(ImageReadTests, InvalidRoiTest) {
  auto truth = MakeTestImage<std::int16_t>(1, 5, 7, d5::envi::Interleave::BIL);
  auto filename = d5::TempFilename(".img");
  auto cleanup = WriteImage(truth, filename);

  auto req = d5::envi::ImageReadRequest().withImageFilename(filename);

  TestImage<std::int16_t> read;
  d5::envi::Header hdr;

  EXPECT_ANY_THROW(
      hdr = d5::envi::LoadImage(req.withRoiFromMinSize(0, 0, 6, 7),
                                TestAdapter(read, d5::envi::Interleave::BIL)));
  EXPECT_ANY_THROW(
      hdr = d5::envi::LoadImage(req.withRoiFromMinMaxInclusive(0, 0, 7, 4),
                                TestAdapter(read, d5::envi::Interleave::BIL)));
  EXPECT_ANY_THROW(
      hdr = d5::envi::LoadImage(req.withRoiFromMinSize(6, 5, 1, 1),
                                TestAdapter(read, d5::envi::Interleave::BIL)));
  EXPECT_ANY_THROW(
      hdr = d5::envi::LoadImage(req.withRoiFromMinSize(7, 4, 1, 1),
                                TestAdapter(read, d5::envi::Interleave::BIL)));
}

TEST(ImageReadTests, BandsUint64BipTest) {
  EXPECT_TRUE(ImageTest<std::uint64_t>(5, 3, 4, d5::envi::Interleave::BIP, {},
                                       {}, {4, 3, 0}));
  EXPECT_TRUE(ImageTest<std::uint64_t>(5, 3, 4, d5::envi::Interleave::BIP, {},
                                       {}, {0, 1, 2, 3, 4}));
}

TEST(ImageReadTests, BandsInt64BsqTest) {
  EXPECT_TRUE(ImageTest<std::int64_t>(5, 3, 4, d5::envi::Interleave::BSQ, {},
                                      {}, {0, 1, 0}));
  EXPECT_TRUE(ImageTest<std::int64_t>(5, 3, 4, d5::envi::Interleave::BSQ, {},
                                      {}, {0, 1, 2, 3, 4}));
}

TEST(ImageReadTests, BandsFloat64BilTest) {
  EXPECT_TRUE(
      ImageTest<double>(5, 3, 4, d5::envi::Interleave::BIL, {}, {}, {2, 1, 3}));
  EXPECT_TRUE(ImageTest<double>(5, 3, 4, d5::envi::Interleave::BIL, {}, {},
                                {0, 1, 2, 3, 4}));
}

TEST(ImageReadTests, BandsInvalidTest) {
  auto truth = MakeTestImage<std::int16_t>(4, 5, 7, d5::envi::Interleave::BIL);
  auto filename = d5::TempFilename(".img");
  auto cleanup = WriteImage(truth, filename);

  TestImage<std::int16_t> read;
  d5::envi::Header hdr;

  auto req = d5::envi::ImageReadRequest().withImageFilename(filename);
  EXPECT_ANY_THROW(
      hdr = d5::envi::LoadImage(req.withBandsToLoad({0, 4, 1, 2}),
                                TestAdapter(read, d5::envi::Interleave::BIL)));
}

d5::envi::Endian nonNativeEndian() {
  return d5::envi::nativeByteOrder() == d5::envi::Endian::Little
             ? d5::envi::Endian::Big
             : d5::envi::Endian::Little;
}

TEST(ImageReadTests, NonNativeEndianBIPTest) {
  EXPECT_TRUE(ImageTest<std::uint8_t>(5, 3, 4, d5::envi::Interleave::BIP, {},
                                      {}, {}, nonNativeEndian()));
  EXPECT_TRUE(ImageTest<std::uint16_t>(5, 3, 4, d5::envi::Interleave::BIP, {},
                                       {}, {}, nonNativeEndian()));
  EXPECT_TRUE(ImageTest<std::uint32_t>(5, 3, 4, d5::envi::Interleave::BIP, {},
                                       {}, {}, nonNativeEndian()));
  EXPECT_TRUE(ImageTest<std::uint64_t>(5, 3, 4, d5::envi::Interleave::BIP, {},
                                       {}, {}, nonNativeEndian()));
  EXPECT_TRUE(ImageTest<std::complex<float>>(5, 3, 4, d5::envi::Interleave::BIP,
                                             {}, {}, {}, nonNativeEndian()));
  EXPECT_TRUE(ImageTest<std::complex<double>>(
      5, 3, 4, d5::envi::Interleave::BIP, {}, {}, {}, nonNativeEndian()));
}

TEST(ImageReadTests, NonNativeEndianBILTest) {
  EXPECT_TRUE(ImageTest<std::uint8_t>(5, 3, 4, d5::envi::Interleave::BIL, {},
                                      {}, {}, nonNativeEndian()));
  EXPECT_TRUE(ImageTest<std::uint16_t>(5, 3, 4, d5::envi::Interleave::BIL, {},
                                       {}, {}, nonNativeEndian()));
  EXPECT_TRUE(ImageTest<std::uint32_t>(5, 3, 4, d5::envi::Interleave::BIL, {},
                                       {}, {}, nonNativeEndian()));
  EXPECT_TRUE(ImageTest<std::uint64_t>(5, 3, 4, d5::envi::Interleave::BIL, {},
                                       {}, {}, nonNativeEndian()));
  EXPECT_TRUE(ImageTest<std::complex<float>>(5, 3, 4, d5::envi::Interleave::BIL,
                                             {}, {}, {}, nonNativeEndian()));
  EXPECT_TRUE(ImageTest<std::complex<double>>(
      5, 3, 4, d5::envi::Interleave::BIL, {}, {}, {}, nonNativeEndian()));
}

TEST(ImageReadTests, ComboTest) {
  EXPECT_TRUE(ImageTest<double>(5, 13, 7, d5::envi::Interleave::BIL,
                                std::array<size_t, 4>{1, 2, 5, 10}, 2,
                                {2, 1, 3}, nonNativeEndian()));
}
