#include "Sgp4Facade.h"

#include "dirsig5/base/Constants.h"
#include "sgp4ext.h"
#include "sgp4io.h"
#include "sgp4unit.h"

// some typedefs to make the SGP4 structs less terse
typedef gravconsttype GravConstant;
typedef elsetrec SatRecord;

namespace d5::flex_motion {

void Sgp4Facade::run(const DateTime &desiredTime, char *tleLine1,
                     char *tleLine2, EcefCoord &ecefPosition,
                     EcefCoord &ecefVelocity) {
  static const double minutesPerDay = 60.0 * 24.0;

  // input variables for the stock TLE parser
  char runType = 'c';
  char inputType = 'm';
  char opsMode = 'a';
  GravConstant gravConst = wgs72;

  // output variables for the stock TLE parser
  double dummyTime1, dummyTime2, dummyTime3;
  SatRecord satRecord;

  // run the stock two-line element (TLE) parser included in SGP4
  twoline2rv(tleLine1, tleLine2, runType, inputType, opsMode, gravConst,
             dummyTime1, dummyTime2, dummyTime3, satRecord);

  satRecord.t = 0;

  // compute the days from epoch (DFE) and minutes from epoch (MFE)
  double startDfe = desiredTime.toJulianDate() - satRecord.jdsatepoch;
  double startMfe = startDfe * minutesPerDay;

  // run the calculation (output in TEME)
  double temePosition[3];
  double temeVelocity[3];
  sgp4(gravConst, satRecord, startMfe, temePosition, temeVelocity);

  // convert the position and velocity from TEME to ECEF
  temeToEcef(gstime(desiredTime.toJulianDate()), temePosition, temeVelocity,
             ecefPosition, ecefVelocity);

  return;
}

void Sgp4Facade::temeToEcef(double siderealTime, double temePosition[3],
                            double temeVelocity[3], EcefCoord &ecefPosition,
                            EcefCoord &ecefVelocity) {
  // the earth rotation rate in radians/second
  static const double earthRotRate = 7.292115146706979e-5;

  double siderealAngle = d5::TwoPI - siderealTime;
  Matrix transform1(Eigen::AngleAxisd(siderealAngle, Vector::UnitZ()));

  // this transform accounts for velocity due to the Earth's rotation
  Matrix transform2(Matrix::Identity());
  double sinAngle = std::sin(siderealTime);
  double cosAngle = std::cos(siderealTime);
  transform2(0, 0) = -1 * earthRotRate * sinAngle;
  transform2(0, 1) = +1 * earthRotRate * cosAngle;
  transform2(1, 0) = -1 * earthRotRate * cosAngle;
  transform2(1, 1) = -1 * earthRotRate * sinAngle;
  transform2(2, 2) = 0.0;
  transform2(3, 3) = 0.0;

  Vector tmpPos, tmpVel;
  for (int ii = 0; ii < 3; ii++) {
    tmpPos[ii] = temePosition[ii];
    tmpVel[ii] = temeVelocity[ii];
  }

  // convert the TEME position and velocity to ECEF
  tmpVel = (transform1 * tmpVel) + (transform2 * tmpPos);
  tmpPos = transform1 * tmpPos;

  // convert the ECEF coords from km to m
  ecefPosition.x = tmpPos.x() * 1000.0;
  ecefPosition.y = tmpPos.y() * 1000.0;
  ecefPosition.z = tmpPos.z() * 1000.0;
  ecefVelocity.x = tmpVel.x() * 1000.0;
  ecefVelocity.y = tmpVel.y() * 1000.0;
  ecefVelocity.z = tmpVel.z() * 1000.0;
}

} // namespace d5::flex_motion
