#include "dirsig5/flex_motion/RotationEngines/LookAt.h"
#include <fstream>
#include <iostream>

namespace d5::flex_motion::RotationEngines {

Quaternion LookAt::doGet(const DateTime &absoluteTime, double) const {

  // compute the (reversed) line-of-sight (LOS) vector
  auto locationState = locationEngine->get(absoluteTime);
  auto trackingState = trackingEngine->get(absoluteTime);
  Vector lineOfSight = (locationState.location - trackingState.location);
  double dist = lineOfSight.norm();

  // see whether our target is moving
  if (trackingState.velocity[0] != 0. || trackingState.velocity[1] != 0. ||
      trackingState.velocity[2] != 0.) {

    // estimate the amount of time to get there (but only iterate a few times)
    double lightTimeToTarget, lastDist;
    for (size_t ii = 0; ii < 10; ii++) {
      lightTimeToTarget = dist / speedOfLight;
      lastDist = dist;
      trackingState = trackingEngine->get(absoluteTime - lightTimeToTarget);

      lineOfSight = (locationState.location - trackingState.location);
      dist = lineOfSight.norm();
      // assume we're close enough if our distance isn't changing
      if (fabs(dist - lastDist) < 1e-3)
        break;
    }
  }

  // normalized the line of sight vector
  lineOfSight /= dist;

  // is the location dynamic?
  bool isDynamic = locationState.velocity[0] != 0. ||
                   locationState.velocity[1] != 0 ||
                   locationState.velocity[2] != 0;

  auto testFlag = [](unsigned int var, unsigned int flag) {
    return ((var & flag) == flag);
  };

  auto cVector = constraintVector;
  if (isDynamic && testFlag(constraintFlags, UseVelocity)) {
    cVector = Vector(locationState.velocity).normalized();
  } else if (testFlag(constraintFlags, UseConstant)) {
    // check if we need to convert the frame for the constraint vector
    if ((testFlag(constraintFlags, VectorFrameEcef)) &&
        (frame == d5::geolib::Location::EnuType)) {
      cVector = cache.ecefToEnuTransform * constraintVector;
    } else if ((testFlag(constraintFlags, VectorFrameEnu)) &&
               (frame == d5::geolib::Location::EcefType)) {
      cVector = cache.ecefToEnuTransform.inverse() * constraintVector;
    }
  }

  Vector orthoVector(0, 1, 0);
  if (std::abs(lineOfSight[1]) == 1) {
    orthoVector[1] = -1;
  }
  if (std::abs(lineOfSight.dot(cVector)) != 1) {
    orthoVector = cVector;
  }
  Vector zVector = lineOfSight;
  Vector xVector = orthoVector.cross(zVector).normalized();
  Vector yVector = zVector.cross(xVector).normalized();
  Matrix matrix = Matrix::Identity();
  for (int i = 0; i < 3; i++) {
    matrix(i, 0) = xVector(i);
    matrix(i, 1) = yVector(i);
    matrix(i, 2) = zVector(i);
  }
  return Quaternion(matrix.rotation());
}

} // namespace d5::flex_motion::RotationEngines
