/** Implementation file for OutputFile
 *
 *  \file OutputFile.cpp
 */

#include "dirsig5/io/OutputFile.h"

#include <unordered_map>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Library.h"
#include "dirsig5/base/LibraryManager.h"
#include "dirsig5/base/Logger.h"
#include "dirsig5/base/PluginLibraryManager.h"
#include "dirsig5/io/OutputFileDriver.h"
#include "drivers/local/LocalOutputFileDriver.h"

namespace {

const auto &GetManager() {
  static auto manager = []() {
    d5::PluginLibraryManager<d5::io::OutputFileDriver> m;
    m.loadBuiltin("local_output_file", d5::io::MakeLocalOutputFileDriver);
    m.loadDefaultLocations("file_output");
    return m;
  }();
  return manager;
}
} // namespace

namespace d5::io {

static const std::string kDefaultDriver = "local_output_file";

std::string OutputFile::s_preferredDriver = kDefaultDriver;

OutputFile::OutputFile(std::unique_ptr<OutputFileDriver> plugin)
    : plugin_(std::move(plugin)) {
  DIRSIG_THROW_IF(!plugin_, "null plugin provided");

  DIRSIG_ERROR_TRACE("initializing driver plugin",
                     { callbacks_ = plugin_->getCallbacks(); });

  DIRSIG_THROW_IF(!callbacks_,
                  "the output file driver plugin did not implement "
                  "the full file I/O interface");
}

void OutputFile::SetPreferredDriver(const std::string &name) {
  if (GetManager().hasPlugin(name)) {
    s_preferredDriver = name;
  } else {
    DIRSIG_WARN(
        "OutputFileDriver::SetPreferredPlugin passed non-existant plugin name "
        "\"{}\", defaulting to \"{}\"",
        name, kDefaultDriver);
    s_preferredDriver = kDefaultDriver;
  }
}

OutputFile OutputFile::MakeLocal() { return MakeWithDriver(kDefaultDriver); }

OutputFile OutputFile::MakeWithDriver(const std::string &name) {
  DIRSIG_ERROR_TRACE(
      "trying to make an output file with the \"" + name + "\" driver", {
        if (GetManager().hasPlugin(name)) {
          return OutputFile(GetManager().getInstance(name));
        } else {
          DIRSIG_THROW("no plugin with the given name exists");
        }
      });
}

OutputFile OutputFile::MakeWithPreferredDriver() {
  return MakeWithDriver(s_preferredDriver);
}

} // namespace d5::io
