#include "dirsig5/io/bin/BinOutputFile.h"

#include "dirsig5/base/Constants.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/MpiState.h"
#include "dirsig5/sdk_info.h"

// we are currently writing version "2" of the BIN file
static const int BIN_FILE_VERSION = 2;

namespace {

template <typename T>
bool Write(d5::io::OutputFile &file, const T *buff, std::size_t count) {
  return file.write(reinterpret_cast<const char *>(buff), sizeof(T) * count,
                    file.tellp());
}

/**
 * A helper function to pack a date/time into a 16 character string.
 */
void writeTimestamp(const d5::DateTime &dateTime, char *timeArr) {
  int day, month, year, dayOfYear;
  double daySeconds = dateTime.getCalenderDate(day, month, year, dayOfYear);

  int hour = static_cast<int>(daySeconds / 3600.0);
  daySeconds -= (hour * 3600);

  int minute = static_cast<int>(daySeconds / 60.0);
  daySeconds -= (minute * 60);

  int intSeconds = static_cast<int>(daySeconds);

  auto print1 = [&](int d) { // Print 1-digit decimal number
    if (d < 0)
      d = 0;
    if (d > 9)
      d = 9;
    *timeArr++ = "0123456789"[d];
  };
  auto print2 = [&](int d) { // Print 2-digit decimal number
    print1(d / 10);          // e.g., 23 => 2
    print1(d % 10);          // e.g., 23 => 3
  };
  auto print4 = [&](int d) { // Print 4-digit decimal number
    print2(d / 100);         // e.g., 1995 => 19
    print2(d % 100);         // e.g., 1995 => 95
  };

  print4(year);       // 0, 1, 2, 3
  print2(month);      // 4, 5
  print2(day);        // 6, 7
  print2(hour);       // 8, 9
  print2(minute);     // 10, 11
  *timeArr++ = '.';   // 12,
  print2(intSeconds); // 13, 14
  *timeArr = '\0';    // 15
}

void toYzxAngles(const d5::Quaternion &q, double &xAngle, double &yAngle,
                 double &zAngle) {
  auto rotMatrix = q.toRotationMatrix();
  xAngle = std::atan2(rotMatrix(2, 1), rotMatrix(1, 1));
  yAngle = std::atan2(rotMatrix(0, 2), rotMatrix(0, 0));
  zAngle = -std::asin(rotMatrix(0, 1));
}
} // namespace

namespace d5::io {

static bool isBigEndian() noexcept {
  uint32_t uLong = 1;
  uint8_t uChar = 0;
  std::memcpy(&uChar, &uLong, 1);
  return uChar == 0;
}

BinOutputFile::FileHeader::FileHeader() {
  // populate file identifier
  std::memcpy(this->fileIdentifier, "DIRSIGPROTO\0", 12);

  // we can only write in the current version
  this->fileFormatRev = BIN_FILE_VERSION;

  // host endianness is default
  this->byteOrder = isBigEndian() ? 0 : 1;

  writeTimestamp(d5::DateTime::now(), creationDateTime);

  // populate DIRSIG version
  std::memset(this->dirsigVersion, 0, 33);
  std::strncpy(this->dirsigVersion, kSdkVersion.data(), 31);

  // populate simulation description
  std::memset(this->simulationDesc, 0, 257);

  // populate laser and detector mount types
  std::memcpy(this->laserMountType, "Unknown Mount\0\0\0\0", 17);
  std::memcpy(this->detectorMountType, "Unknown Mount\0\0\0\0", 17);
}

void BinOutputFile::FileHeader::setup(
    const d5::GeodeticCoord &_sceneOrigin, std::uint32_t _xPixelCount,
    std::uint32_t _yPixelCount, double _xPixelPitch, double _yPixelPitch,
    double _xArrayOffset, double _yArrayOffset, std::uint32_t _taskCount) {
  // stash scene origin
  this->sceneOriginLatitude = _sceneOrigin.latitude;
  this->sceneOriginLongitude = _sceneOrigin.longitude;
  this->sceneOriginHeight = _sceneOrigin.height;

  // stash detector layout
  this->xDetectorCount = _xPixelCount;
  this->yDetectorCount = _yPixelCount;
  this->xDetectorPitch = _xPixelPitch * 1.0e+06;
  this->yDetectorPitch = _yPixelPitch * 1.0e+06;
  this->xArrayOffset = _xArrayOffset * 1.0e+06;
  this->yArrayOffset = _yArrayOffset * 1.0e+06;

  // hard code the lens distortion coefficients
  this->lensK1 = 0.0;
  this->lensK2 = 0.0;

  // stash number of tasks contained in file
  this->taskCount = _taskCount;
}

void BinOutputFile::FileHeader::print(std::ostream &stream) const {
  stream << "File Header:" << std::endl
         << "    Created: " << this->creationDateTime << std::endl
         << "    DIRSIG Version: " << this->dirsigVersion << std::endl
         << "    File Version: " << static_cast<int>(this->fileFormatRev)
         << std::endl
         << "    Scene Origin: " << this->sceneOriginLatitude << " [+N], "
         << this->sceneOriginLongitude << " [+W]" << std::endl
         << "    Scene Altitude: " << this->sceneOriginHeight << " [m]"
         << std::endl
         << "    Source Mount: " << this->laserMountType << std::endl
         << "    Receiver Mount: " << this->detectorMountType << std::endl
         << "    Focal Plane Array ID: " << this->fpaId << std::endl
         << "    Array Size: " << this->xDetectorCount << " x "
         << this->yDetectorCount << " [pixels]" << std::endl
         << "    Pixel Dimensions: " << this->xDetectorPitch << " x "
         << this->yDetectorPitch << " [microns]" << std::endl
         << "    Array offsets: " << this->xArrayOffset << " x "
         << this->yArrayOffset << " [microns]" << std::endl
         << "    Lens Distortion Coeffs: " << this->lensK1 << ", "
         << this->lensK2 << std::endl
         << "    Task Count: " << this->taskCount << std::endl
         << std::endl;
}

BinOutputFile::TaskHeader::TaskHeader() {
  std::memset(this->description, 0, 65);
  std::memset(this->startDateTime, 0, 16);
  std::memset(this->stopDateTime, 0, 16);
}

bool BinOutputFile::open() {
  // If we are in MPI mode, append our worker index onto the filename
  if (MPI_STATE.worldSize > 1) {
    if (!MPI_STATE.frameMode) {
      std::cerr << "\nBinOutputFile::open: MPI ERROR!\n"
                   "    Lidar output can only be produced using the "
                   "unique frame "
                   "per process (UFP) schedule!\n"
                   "    Use the '--mpi_one_event_per_node' option to "
                   "enable UFP\n";
      return false;
    }
    this->file.open(this->getFilename() + std::to_string(MPI_STATE.worldRank));
  } else {
    this->file.open(this->getFilename());
  }

  // open the file
  return this->file.is_open();
}

bool BinOutputFile::close() {
  // if the file isn't open, we are done
  if (!this->file.is_open())
    return true;

  // close the file
  this->file.close();

  return true;
}

bool BinOutputFile::writeFileHeader() {
  // write file header data
  bool s = true;
  s &= Write(file, fileHeader.fileIdentifier, 11);
  s &= Write(file, &fileHeader.fileFormatRev, 1);
  s &= Write(file, &fileHeader.byteOrder, 1);
  s &= Write(file, fileHeader.creationDateTime, 15);
  s &= Write(file, fileHeader.dirsigVersion, 32);
  s &= Write(file, fileHeader.simulationDesc, 256);
  s &= Write(file, &fileHeader.sceneOriginLatitude, 1);
  s &= Write(file, &fileHeader.sceneOriginLongitude, 1);
  s &= Write(file, &fileHeader.sceneOriginHeight, 1);
  s &= Write(file, fileHeader.laserMountType, 16);
  s &= Write(file, fileHeader.detectorMountType, 16);
  s &= Write(file, &fileHeader.xDetectorCount, 1);
  s &= Write(file, &fileHeader.yDetectorCount, 1);
  s &= Write(file, &fileHeader.xDetectorPitch, 1);
  s &= Write(file, &fileHeader.yDetectorPitch, 1);
  s &= Write(file, &fileHeader.xArrayOffset, 1);
  s &= Write(file, &fileHeader.yArrayOffset, 1);
  s &= Write(file, &fileHeader.lensK1, 1);
  s &= Write(file, &fileHeader.lensK2, 1);
  s &= Write(file, &fileHeader.taskCount, 1);
  s &= Write(file, &fileHeader.fpaId, 1);

  // flush the file buffer
  file.flush();

  return s;
}

bool BinOutputFile::writeTaskHeader() {
  // read task header data
  bool s = true;
  s &= Write(file, taskHeader.description, 64);
  s &= Write(file, taskHeader.startDateTime, 15);
  s &= Write(file, taskHeader.stopDateTime, 15);
  s &= Write(file, &taskHeader.focalLength, 1);
  s &= Write(file, &taskHeader.pulseRepetitionFreq, 1);
  s &= Write(file, &taskHeader.pulseDuration, 1);
  s &= Write(file, &taskHeader.pulseEnergy, 1);
  s &= Write(file, &taskHeader.laserSpectralCenter, 1);
  s &= Write(file, &taskHeader.laserSpectralWidth, 1);

  this->pulseCountPosition = file.tellp();
  s &= Write(file, &taskHeader.pulseCount, 1);

  // flush the file buffer
  file.flush();

  return s;
}

bool BinOutputFile::writePulseHeader() {
  // read pulse header data
  bool s = true;
  s &= Write(file, &pulseHeader.pulseTime, 1);
  s &= Write(file, &pulseHeader.timeGateStart, 1);
  s &= Write(file, &pulseHeader.timeGateStop, 1);
  s &= Write(file, &pulseHeader.timeGateBinCount, 1);
  s &= Write(file, &pulseHeader.samplesPerTimeBin, 1);
  s &= Write(file, pulseHeader.platformLocation, 3);
  s &= Write(file, pulseHeader.platformRotation, 3);
  s &= Write(file, pulseHeader.transmitterToMount, 16);
  s &= Write(file, pulseHeader.transmitterMountPointing, 3);
  s &= Write(file, pulseHeader.transmitterMountToPlatform, 16);
  s &= Write(file, pulseHeader.receiverToMount, 16);
  s &= Write(file, pulseHeader.receiverMountPointing, 3);
  s &= Write(file, pulseHeader.receiverMountToPlatform, 16);
  s &= Write(file, &pulseHeader.pulseDataType, 1);
  s &= Write(file, &pulseHeader.dataCompression, 1);
  s &= Write(file, &pulseHeader.pulseIndex, 1);
  s &= Write(file, &pulseHeader.pulseDataBytes, 1);
  s &= Write(file, pulseHeader.systemPolTransmit, 16);
  s &= Write(file, pulseHeader.systemPolReceive, 16);

  // flush the file buffer
  file.flush();

  return s;
}

/**
 * The (uncompressed) pulse data needs to be large enough to hold the
 * passive (temporally constant) flux and the waveform
 */
void BinOutputFile::resizePulseBuffer(size_t timeBinCount) {
  size_t bufferSize = this->fileHeader.xDetectorCount *
                      this->fileHeader.yDetectorCount * (timeBinCount + 1);
  assert(bufferSize != 0);

  // check if we need to resize
  if (this->uncompressedPulseBuffer.size() < bufferSize) {
    this->uncompressedPulseBuffer.resize(bufferSize);
  }
}

bool BinOutputFile::writePulseBuffer() {
  bool writeSuccess = true;

  // record the current position so we can go update the pulse count first
  size_t currentPosition = this->file.tellp();

  // update the number of pulses in the task header and write it back out
  this->taskHeader.pulseCount += 1;
  writeSuccess &= this->file.write(&this->taskHeader.pulseCount, 1,
                                   this->pulseCountPosition);

  // now write the pulse data where we were
  writeSuccess &=
      this->file.write(this->uncompressedPulseBuffer.data(),
                       this->uncompressedPulseBuffer.size(), currentPosition);

  return writeSuccess;
}

bool BinOutputFile::writePulseBuffer(const void *buffer, size_t byteCount) {
  bool writeSuccess = true;

  // record the current position so we can go update the pulse count first
  auto currentPosition = this->file.tellp();

  // update the number of pulses in the task header and write it back out
  this->taskHeader.pulseCount += 1;
  writeSuccess &= this->file.write(&this->taskHeader.pulseCount, 1,
                                   this->pulseCountPosition);

  writeSuccess &= this->file.write(reinterpret_cast<const char *>(buffer),
                                   byteCount, currentPosition);

  return writeSuccess;
}

void BinOutputFile::TaskHeader::setupRx(std::uint32_t _taskIndex,
                                        std::uint32_t _pulseCount,
                                        double _focalLength,
                                        const d5::DateTime &_taskStart,
                                        const d5::DateTime &_taskEnd) {
  // store the current task index
  this->taskIndex = _taskIndex;

  // populate task start and end timestamps
  writeTimestamp(_taskStart, this->startDateTime);
  writeTimestamp(_taskEnd, this->stopDateTime);

  // populate instrument data
  this->focalLength = _focalLength;

  // store the number of pulses this task
  this->pulseCount = _pulseCount;
}

void BinOutputFile::TaskHeader::setupTx(double _pulseRate, double _pulseEnergy,
                                        double _pulseDuration,
                                        double _spectralCenter,
                                        double _spectralWidth) {
  // populate laser data
  this->pulseRepetitionFreq = _pulseRate;
  this->pulseDuration = _pulseDuration;
  this->pulseEnergy = _pulseEnergy;
  this->laserSpectralCenter = _spectralCenter;
  this->laserSpectralWidth = _spectralWidth;
}

void BinOutputFile::TaskHeader::print(std::ostream &stream) const {
  stream << "Task Header: " << std::endl
         << "    Index: " << this->taskIndex << std::endl
         << "    Start Date/Time: " << this->startDateTime << std::endl
         << "    Stop Date/Time: " << this->stopDateTime << std::endl
         << "    Focal Length: " << this->focalLength << " [mm]" << std::endl
         << "    Pulse Rate: " << this->pulseRepetitionFreq << " [Hz]"
         << std::endl
         << "    Pulse Duration: " << this->pulseDuration << " [s]" << std::endl
         << "    Pulse Energy: " << this->pulseEnergy << " [J]" << std::endl
         << "    Laser Line: " << this->laserSpectralCenter << " [um]"
         << std::endl
         << "    Laser Width: " << this->laserSpectralWidth << " [um]"
         << std::endl
         << "    Pulse Count: " << this->pulseCount << std::endl
         << std::endl;
  return;
}

BinOutputFile::PulseHeader::PulseHeader() {
  // the platform location and rotation are set to 0s
  memset(this->platformLocation, 0.0, 3 * sizeof(double));
  memset(this->platformRotation, 0.0, 3 * sizeof(double));

  // the transmitter inst attachment is set to an identity matrix
  memset(this->transmitterToMount, 0.0, 16 * sizeof(double));
  this->transmitterToMount[0] = 1.0;
  this->transmitterToMount[5] = 1.0;
  this->transmitterToMount[10] = 1.0;
  this->transmitterToMount[15] = 1.0;

  // the transmitter pointing is set to 0s
  memset(this->transmitterMountPointing, 0.0, 3 * sizeof(double));

  // the transmitter mount attachment is set to an identity matrix
  memset(this->transmitterMountToPlatform, 0.0, 16 * sizeof(double));
  this->transmitterMountToPlatform[0] = 1.0;
  this->transmitterMountToPlatform[5] = 1.0;
  this->transmitterMountToPlatform[10] = 1.0;
  this->transmitterMountToPlatform[15] = 1.0;

  // the receiver inst attachment is set to an identity matrix
  memset(this->receiverToMount, 0.0, 16 * sizeof(double));
  this->receiverToMount[0] = 1.0;
  this->receiverToMount[5] = 1.0;
  this->receiverToMount[10] = 1.0;
  this->receiverToMount[15] = 1.0;

  // the receiver pointing is set to 0s
  memset(this->receiverMountPointing, 0.0, 3 * sizeof(double));

  // the receiver mount attachment is set to an identity matrix
  memset(this->receiverMountToPlatform, 0.0, 16 * sizeof(double));
  this->receiverMountToPlatform[0] = 1.0;
  this->receiverMountToPlatform[5] = 1.0;
  this->receiverMountToPlatform[10] = 1.0;
  this->receiverMountToPlatform[15] = 1.0;

  // initialize polariation states to unpolarized
  memset(systemPolTransmit, 0.0, 16 * sizeof(double));
  this->systemPolTransmit[0] = 1.0;
  memset(systemPolReceive, 0.0, 16 * sizeof(double));
  this->systemPolReceive[0] = 1.0;
}

void BinOutputFile::PulseHeader::setup(
    std::uint32_t _pulseIndex, double _pulseTime, double _gateStartTime,
    double _gateStopTime, std::uint32_t _gateTimeCount,
    const d5::DualQuaternion &_platformState,
    const d5::DualQuaternion & /*_platformAttach*/,
    const d5::DualQuaternion &_mountState,
    const d5::DualQuaternion & /*_mountAttach*/,
    unsigned long _pulseDataBytes) {
  // store the timing information
  this->pulseIndex = _pulseIndex;
  this->pulseTime = _pulseTime;
  this->timeGateStart = _gateStartTime;
  this->timeGateStop = _gateStopTime;
  this->timeGateBinCount = _gateTimeCount;
  this->samplesPerTimeBin = 1;

  // set the platform location
  auto location = _platformState.pointForward(d5::Point(0, 0, 0));
  this->platformLocation[0] = location[0];
  this->platformLocation[1] = location[1];
  this->platformLocation[2] = location[2];

  // set the platform rotation
  toYzxAngles(_platformState.real, this->platformRotation[0],
              this->platformRotation[1], this->platformRotation[2]);

  // set the transmitter to mount transformation

  // set the transmitter pointing Euler angles
  toYzxAngles(_mountState.real, this->transmitterMountPointing[0],
              this->transmitterMountPointing[1],
              this->transmitterMountPointing[2]);

  // set the transmitter mount to platform transformation

  // set the receiver to mount transformation

  // set the receiver pointing Euler angles
  this->receiverMountPointing[0] = this->transmitterMountPointing[0];
  this->receiverMountPointing[1] = this->transmitterMountPointing[1];
  this->receiverMountPointing[2] = this->transmitterMountPointing[2];

  // set the receiver mount to platform transformation

  // we are always using double-precision data for now (type = 5)
  this->pulseDataType = 5;

  // no compression right now
  this->dataCompression = 0;

  // compressed pulse size
  this->pulseDataBytes = _pulseDataBytes;
}

void BinOutputFile::PulseHeader::print(std::ostream &stream) const {
  stream << "Pulse Header:" << std::endl
         << "    Pulse time: " << this->pulseTime << std::endl
         << "    Platform location: " << this->platformLocation[0] << ", "
         << this->platformLocation[1] << ", " << this->platformLocation[2]
         << std::endl
         << "    Platform rotation: " << this->platformRotation[0] << ", "
         << this->platformRotation[1] << ", " << this->platformRotation[2]
         << std::endl
         << "    Mount rotation: " << this->receiverMountPointing[0] << ", "
         << this->receiverMountPointing[1] << ", "
         << this->receiverMountPointing[2] << std::endl
         << "    Time Gate Open: " << this->timeGateStart << " [s], "
         << this->timeGateStart * d5::speedOfLight / 2.0 << " [m]" << std::endl
         << "    Time Gate Close: " << this->timeGateStop << " [s], "
         << this->timeGateStop * d5::speedOfLight / 2.0 << " [m]" << std::endl
         << "    Time Bin Count: " << this->timeGateBinCount << std::endl
         << "    Samples Per Time Bin: " << this->samplesPerTimeBin
         << std::endl;

  if (this->dataCompression == 1) {
    stream << "    Contains compressed pulse data";
  } else {
    stream << "    Contains uncompressed pulse data";
  }
  stream << " (data size = " << this->pulseDataBytes << " bytes)" << std::endl;

  return;
}
} // namespace d5::io
