/** Implementation file for the LocalOutputFileDriver class
 *
 *  \file LocalOutputFileDriver.cpp
 */

#include "LocalOutputFileDriver.h"

#include <cstring>

namespace d5::io {

auto LocalOutputFileDriver::getCallbacks() -> Callbacks {
  Callbacks c;

  c.open = [this](const std::string &filename, std::ios_base::openmode mode) {
    ofs_.open(filename, mode);
    if (!ofs_.is_open())
      lastError_ = strerror(errno);
  };
  c.is_open = [this]() { return ofs_.is_open(); };
  c.close = [this]() { ofs_.close(); };
  c.write2 = [this](const char *data, std::size_t size) {
    ofs_.write(data, static_cast<std::streamsize>(size));
    return ofs_.good();
  };
  c.write3 = [this](const char *data, std::size_t size, std::size_t offset) {
    if (seekp(offset)) {
      ofs_.write(data, static_cast<std::streamsize>(size));
      return ofs_.good();
    } else {
      lastError_ = std::string("write to file failed: ") + strerror(errno);
      return false;
    }
  };
  c.tellp = [this]() { return static_cast<std::size_t>(ofs_.tellp()); };
  c.seekp1 = [this](std::size_t pos) { return seekp(pos); };
  c.seekp2 = [this](std::int64_t off, std::ios_base::seekdir dir) {
    ofs_.seekp(off, dir);
    return ofs_.good();
  };
  c.flush = [this]() {
    ofs_.flush();
    return ofs_.good();
  };
  c.lastError = [this]() { return lastError_; };
  c.writeFile = [this](const std::string &filename, const char *buffer,
                       std::size_t size) {
    std::ofstream ofs(filename);
    if (!ofs) {
      lastError_ = std::string("unable to open file: ") + strerror(errno);
      return false;
    }
    ofs.write(buffer, static_cast<std::streamsize>(size));
    if (!ofs) {
      lastError_ = std::string("write to file failed: ") + strerror(errno);
      return false;
    }
    ofs.close();
    if (!ofs) {
      lastError_ =
          std::string("error detected on file close: ") + strerror(errno);
      return false;
    }
    return true;
  };

  return c;
}

extern "C" OutputFileDriver *MakeLocalOutputFileDriver() {
  return new LocalOutputFileDriver;
}
} // namespace d5::io
