/** Main file for the MpiOutputFileDriver class
 *
 *  \file MpiOutputFileDriver.cpp
 */

#include <fstream>
#include <ios>
#include <limits>

#include "Export.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/MpiState.h"
#include "dirsig5/io/OutputFileDriver.h"
#include "dirsig5/mpi/Mpi.h"

class MpiOutputFileDriver : public d5::io::OutputFileDriver {
public:
  MpiOutputFileDriver() : MPI_(d5::MPI()), f_(MPI_.FILE_NULL()) {}

  ~MpiOutputFileDriver() override = default;

  void open(const std::string &filename, std::ios_base::openmode mode) {
    close();

    int mpiMode = d5::mpi::MODE_CREATE;
    if (mode & std::ios_base::out)
      mpiMode |= d5::mpi::MODE_WRONLY;
    if (mode & std::ios_base::app)
      mpiMode |= d5::mpi::MODE_APPEND;
    DIRSIG_THROW_IF(mode & std::ios_base::in,
                    "MPI output does not support input mode");
    DIRSIG_THROW_IF(mode & std::ios_base::ate,
                    "MPI output does not support \"at end\" mode");

    if (!Try([&]() {
          f_ = MPI_.File_open(MPI_STATE.frameMode ? MPI_.COMM_SELF()
                                                  : MPI_.COMM_WORLD(),
                              filename.c_str(), mpiMode, MPI_.INFO_NULL());
        })) {
      f_ = MPI_.FILE_NULL();
    }
  }

  [[nodiscard]] bool is_open() const { return f_ != MPI_.FILE_NULL(); };

  void close() {
    if (!CheckIsOpen())
      return;

    MPI_.File_close(&f_);
    f_ = MPI_.FILE_NULL();
  }

  bool write(const char *buffer, std::size_t size) {
    if (!CheckIsOpen())
      return false;

    DIRSIG_THROW_IF(size > std::numeric_limits<int>::max(),
                    "Input buffer was too large");

    return Try([&]() {
      MPI_.File_write(f_, buffer, static_cast<int>(size), d5::mpi::CHAR);
    });
  }

  bool write(const char *buffer, std::size_t size, std::size_t offset) {
    if (!CheckIsOpen())
      return false;

    DIRSIG_THROW_IF(offset > std::numeric_limits<int>::max(),
                    "Input offset was too large");
    DIRSIG_THROW_IF(size > std::numeric_limits<int>::max(),
                    "Input buffer was too large");

    DIRSIG_THROW_IF(size + offset > std::numeric_limits<int>::max(),
                    "Write operation would exceed the max MPI file size");

    return Try([&]() {
      MPI_.File_write_at(f_, static_cast<int>(offset), buffer,
                         static_cast<int>(size), d5::mpi::CHAR);
    });
  }

  [[nodiscard]] std::size_t tellp() const {
    if (!is_open())
      return 0;

    return static_cast<std::size_t>(MPI_.File_get_position(f_));
  }

  bool seekp(std::size_t pos) {
    if (!CheckIsOpen())
      return false;

    return Try([&]() {
      MPI_.File_seek(f_, static_cast<d5::mpi::Offset>(pos), SEEK_SET);
    });
  }

  bool seekp(std::int64_t off, std::ios_base::seekdir dir) {
    if (!CheckIsOpen())
      return false;

    int whence = SEEK_SET;
    switch (dir) {
    case std::ios_base::beg:
      whence = SEEK_SET;
      break;
    case std::ios_base::end:
      whence = SEEK_END;
      break;
    case std::ios_base::cur:
      whence = SEEK_CUR;
      break;
    default:
      break;
    }

    return Try([&]() {
      MPI_.File_seek(f_, static_cast<d5::mpi::Offset>(off), whence);
    });
  }

  bool flush() {
    if (!CheckIsOpen())
      return false;
    return Try([&]() { MPI_.File_sync(f_); });
  }

  /**
   * Write the buffer to the file as a single open, write and close
   * operation.  Since the MPI_File_open() function doesn't support
   * the classic O_TRUC option, we have to do a silly dance to make
   * sure we are overwriting the file.
   */
  bool writeFile(const std::string &_filename, const char *_buffer,
                 size_t _bufferSize) try {
    // Verify non-overflow
    DIRSIG_THROW_IF(_bufferSize > std::numeric_limits<int>::max(),
                    "Buffer was too large");

    // we want to make sure to remove an existing file
    auto optFile = MPI_.File_open_try(
        MPI_.COMM_SELF(), _filename.c_str(),
        d5::mpi::MODE_CREATE | d5::mpi::MODE_EXCL | d5::mpi::MODE_WRONLY,
        MPI_.INFO_NULL());

    // if the open failed, the file must have existed already
    if (!optFile) {
      // delete the file
      Try([&]() { MPI_.File_delete(_filename.c_str(), MPI_.INFO_NULL()); });

      // try to open the file again
      optFile = MPI_.File_open_try(MPI_.COMM_SELF(), _filename.c_str(),
                                   d5::mpi::MODE_CREATE | d5::mpi::MODE_WRONLY,
                                   MPI_.INFO_NULL());
      if (!optFile) {
        return false;
      }
    }

    // write the entire buffer
    try {
      auto status = MPI_.File_write(
          *optFile, _buffer, static_cast<int>(_bufferSize), d5::mpi::CHAR);

      // Verify write
      auto countWritten = MPI_.Get_count(&status, d5::mpi::CHAR);
      DIRSIG_THROW_IF(countWritten < 0 ||
                          static_cast<std::size_t>(countWritten) != _bufferSize,
                      "The write operation was not successful");

      MPI_.File_close(&(*optFile));
      return true;
    } catch (const std::exception &e) {
      lastError_ = e.what();
      MPI_.File_close(&(*optFile));
      return false;
    }
  } catch (const std::exception &e) {
    lastError_ = e.what();
    return false;
  }

  Callbacks getCallbacks() override {
    Callbacks c;

    c.open = [this](const std::string &filename, std::ios_base::openmode mode) {
      open(filename, mode);
    };
    c.is_open = [this]() { return is_open(); };
    c.close = [this]() { close(); };
    c.write2 = [this](const char *data, std::size_t size) {
      return write(data, size);
    };
    c.write3 = [this](const char *data, std::size_t size, std::size_t offset) {
      return write(data, size, offset);
    };
    c.tellp = [this]() { return tellp(); };
    c.seekp1 = [this](std::size_t pos) { return seekp(pos); };
    c.seekp2 = [this](std::int64_t off, std::ios_base::seekdir dir) {
      return seekp(off, dir);
    };
    c.flush = [this]() { return flush(); };
    c.lastError = [this]() { return lastError_; };
    c.writeFile = [this](const std::string &filename, const char *buffer,
                         std::size_t bufferSize) {
      return writeFile(filename, buffer, bufferSize);
    };

    return c;
  }

private:
  template <typename Callable> bool Try(Callable &&f) try {
    f();
    return true;
  } catch (const std::exception &e) {
    lastError_ = e.what();
    return false;
  }

  bool CheckIsOpen() {
    if (is_open())
      return true;
    lastError_ = "there was no open file, call open() first";
    return false;
  }

private:
  d5::mpi::Interface &MPI_;
  d5::mpi::File f_;
  std::string lastError_;
};

extern "C" D5_PLUGIN_EXPORT d5::io::OutputFileDriver *entrypoint() {
  return new MpiOutputFileDriver;
}
