#include "dirsig5/io/image/EnviOutputFile.h"

#include <algorithm>
#include <sstream>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/MpiState.h"
#include "dirsig5/base/TypeTraits.h"

namespace d5::io {

bool EnviOutputFile::open() {
  // check for our currently supported data types
  this->dataType = getValue<data_type>();
  if ((this->dataType != ByteType) && (this->dataType != UnsignedIntegerType) &&
      (this->dataType != UnsignedLongType) &&
      (this->dataType != SinglePrecisionType) &&
      (this->dataType != DoublePrecisionType)) {
    std::string errMsg;
    errMsg = "\n";
    errMsg += "EnviOutputFile::open: Unsupported data type!\n";
    errMsg += "    The requested output data type is not supported\n";
    errMsg += "    datatype = " + std::to_string(this->dataType) + "\n";
    DIRSIG_THROW(errMsg);
  }

  // figure out the output data size
  this->dataSize = this->getDataSize();
  assert(this->dataSize != 0);

  // check if we need a scratch buffer to convert the output data */
  if (this->dataType != DoublePrecisionType) {
    // compute the buffer size
    size_t newBufferSize = this->pixelBufferSize * this->dataSize;
    DIRSIG_THROW_IF(newBufferSize == 0,
                    "EnviOutputFile::open: Zero length buffer!\n");

    // see if we need to update the buffer
    if (this->scratchBufferSize != newBufferSize) {
      this->scratchBufferSize = newBufferSize;

      // if the buffer has already been allocated, free it

      delete[] this->scratchBuffer;

      this->scratchBuffer = new char[this->scratchBufferSize];
    }
  }

  // write out the header file
  if (!this->writeHeaderFile()) {
    return false;
  }

  // open the binary file
  try {
    this->imgFile.open(this->getFilename());
  } catch (const std::exception &e) {
    std::cerr << e.what() << std::endl;
    return false;
  }

  offsetCount_ = 0;
  return this->imgFile.is_open();
}

bool EnviOutputFile::writeBuffer(size_t _pixelOffset) {
  // compute the write position
  size_t dataOffset = this->dataSize * _pixelOffset;

  // Test for overflows
#ifdef _DEBUG
  assert((_pixelOffset == 0) ||
         (dataOffset >= this->dataSize && dataOffset >= _pixelOffset));
  size_t test_size = this->pixelBufferSize * this->dataSize;
  assert((this->pixelBufferSize == 0) ||
         (test_size >= this->pixelBufferSize && test_size >= this->dataSize));

  // check if the file is not open
  if (this->imgFile.is_open() == false) {
    std::cerr << "EnviOutputFile::writeBuffer" << std::endl
              << "  File is NOT open!" << std::endl;
    return false;
  }
  if (this->imgFile.lastError().empty() == false) {
    std::cerr << "EnviOutputFile::writeBuffer" << std::endl
              << "  Error = '" << this->imgFile.lastError() << "'" << std::endl;
    return false;
  }
#endif

  // write the data here
  if (this->dataType == DoublePrecisionType) {
    auto dataLength = this->pixelBufferSize * this->dataSize;
    if (!this->imgFile.write(reinterpret_cast<char *>(this->pixelBuffer.get()),
                             dataLength, dataOffset)) {
      this->errorString = "ENVI data file: " + this->imgFile.lastError() +
                          ", dataOffset = " + std::to_string(dataOffset) +
                          ", dataLength = " + std::to_string(dataLength);
      return false;
    }
  } else {
    // need to convert the data type, put converted into the scratch
    if (this->dataType == ByteType) {
      for (size_t ii = 0; ii < this->pixelBufferSize; ii++) {
        (reinterpret_cast<unsigned char *>(this->scratchBuffer))[ii] =
            static_cast<unsigned char>(this->pixelBuffer[ii]);
      }
    } else if (this->dataType == UnsignedIntegerType) {
      for (size_t ii = 0; ii < this->pixelBufferSize; ii++) {
        (reinterpret_cast<std::uint16_t *>(this->scratchBuffer))[ii] =
            static_cast<std::uint16_t>(this->pixelBuffer[ii]);
      }
    } else if (this->dataType == UnsignedLongType) {
      for (size_t ii = 0; ii < this->pixelBufferSize; ii++) {
        (reinterpret_cast<std::uint32_t *>(this->scratchBuffer))[ii] =
            static_cast<std::uint32_t>(this->pixelBuffer[ii]);
      }
    } else if (this->dataType == SinglePrecisionType) {
      for (size_t ii = 0; ii < this->pixelBufferSize; ii++) {
        (reinterpret_cast<float *>(this->scratchBuffer))[ii] =
            static_cast<float>(this->pixelBuffer[ii]);
      }
    }

    // now write the scratch buffer
    auto dataLength = this->pixelBufferSize * this->dataSize;
    if (!this->imgFile.write(this->scratchBuffer, dataLength, dataOffset)) {
      this->errorString = "ENVI data file: " + this->imgFile.lastError() +
                          ", dataOffset = " + std::to_string(dataOffset) +
                          ", dataLength = " + std::to_string(dataLength);
      return false;
    }
  }

  // flush the binary image to keep it in sync with the header
  this->imgFile.flush();

  if (MPI_STATE.worldSize <= 1) {
    if (dataOffset - offsetCount_ > 100000) {
      if (!this->writeHeaderFile()) {
        this->errorString = "ENVI header file: " + this->imgFile.lastError();
        return false;
      }
      offsetCount_ = dataOffset;
    }
  }

  return true;
}

bool EnviOutputFile::close() {
  // if the file isn't open, we are done
  if (!this->imgFile.is_open()) {
    return true;
  }

  // write out the header file
  if (!this->writeHeaderFile())
    return false;

  // close the binary file
  DIRSIG_ERROR_TRACE("closing ENVI output file", { this->imgFile.close(); });

  return true;
}

namespace {

template <typename T>
std::ostream &operator<<(std::ostream &os, const std::vector<T> &v) {
  for (std::size_t i = 0; i < v.size(); i++) {
    os << v[i];
    if (i < v.size() - 1)
      os << ',';
  }

  return os;
}

template <typename T, std::size_t N>
std::ostream &operator<<(std::ostream &os, const std::array<T, N> &v) {
  for (std::size_t i = 0; i < N; i++) {
    os << v[i];
    if (i < N - 1)
      os << ',';
  }

  return os;
}

std::ostream &operator<<(std::ostream &os,
                         ImageOutputFile::InterleaveMethod i) {
  switch (i) {
  case ImageOutputFile::InterleaveMethod::BSQ:
    os << "BSQ";
    break;
  case ImageOutputFile::InterleaveMethod::BIL:
    os << "BIL";
    break;
  case ImageOutputFile::InterleaveMethod::BIP:
    os << "BIP";
    break;
  case ImageOutputFile::InterleaveMethod::Unknown:
  default:
    DIRSIG_THROW("unknown interleave type specified");
  }

  return os;
}

template <typename T> void writeMultiLineList(std::ostream &os, const T &val) {
  os << "{\n  " << val << "}";
}

template <typename T> void writeInlineList(std::ostream &os, const T &val) {
  os << "{" << val << "}";
}

template <typename T> struct EnviHeaderPrinter {
  static void print(std::ostream &os, ImageOutputFile::Tags tag, const T &val) {
    if (tag == ImageOutputFile::Tags::description) {
      writeMultiLineList(os, val);
    } else {
      os << val;
    }
  }
};

template <typename T, std::size_t N>
struct EnviHeaderPrinter<std::array<T, N>> {
  static void print(std::ostream &os, ImageOutputFile::Tags /*tag*/,
                    const std::array<T, N> &val) {
    writeInlineList(os, val);
  }
};

template <typename T> struct EnviHeaderPrinter<std::vector<T>> {
  static void print(std::ostream &os, ImageOutputFile::Tags /*tag*/,
                    const std::vector<T> &val) {
    writeMultiLineList(os, val);
  }
};

template <typename T>
void WriteEnviHeaderField(std::ostream &os, ImageOutputFile::Tags tag,
                          const T &val) {
  EnviHeaderPrinter<T>::print(os, tag, val);
}
} // namespace

bool EnviOutputFile::writeHeaderFile() {
  DIRSIG_ERROR_TRACE("writing an ENVI header file", {
    // skip if not rank 1 while not in frameMode
    if (MPI_STATE.frameMode == false && MPI_STATE.worldRank != 0) {
      return true;
    }

    // make sure that the minimal tags are set
    std::string errMsg;
    if (!metadata_.contains<samples>())
      errMsg = "Tag \"samples\" not set!";
    if (!metadata_.contains<lines>())
      errMsg = "Tag \"lines\" not set!";
    if (!metadata_.contains<bands>())
      errMsg = "Tag \"bands\" not set!";
    // if we have an error, throw an exception
    if (errMsg.empty() == false)
      DIRSIG_THROW("EnviOutputFile::writeHeaderFile: " + errMsg);

    // create a string stream to stuff the ENVI header contents into
    std::ostringstream hdrStream;

    hdrStream << "ENVI" << std::endl;

    visitMetadata([&](Tags tag, const auto &val) {
      auto tagName = to_string(tag);
      std::replace(tagName.begin(), tagName.end(), '_', ' ');
      hdrStream << tagName << " = ";
      WriteEnviHeaderField(hdrStream, tag, val);
      hdrStream << std::endl;
    });

    // write the header to the output as a single operation
    std::string hdrContents = hdrStream.str();
    return this->imgFile.writeFile(hdrFilename, hdrContents.c_str(),
                                   hdrContents.size());
  });
}
} // namespace d5::io
