#include "dirsig5/mpi/Mpi.h"

#include <iostream>
#include <memory>
#include <sstream>
#include <stdexcept>
#include <string>

#include <spdlog/fmt/fmt.h>

#include "Driver.h"
#include "dirsig5/base/PluginLibraryManager.h"
#include "dirsig5/base/StandardPaths.h"

namespace {

/// Helper function to throw formatted errors
template <typename... Args> std::runtime_error Error(Args &&...args) {
  return std::runtime_error(fmt::format(std::forward<Args>(args)...));
}

class PluginLoader final : public d5::PluginLibraryManager<d5::mpi::Driver> {
public:
  PluginLoader() {
    loadDefaultLocations("mpi_adapters");
    if (libraries.empty())
      throw Error("Can't find suitable MPI adapter library in {}",
                  d5::StandardPaths::DirsigPlugins("mpi_adapters"));
  }

  auto *getDriver() { return getInstance(libraries.begin()->first).release(); }
};

PluginLoader &GetLoader() try {
  static PluginLoader s_loader;
  return s_loader;
} catch (const std::exception &e) {
  throw Error("while attempting to load the MPI driver plugin:\n{}", e.what());
}

} // namespace

namespace d5::mpi {

Interface &Interface::getInstance() {
  static Interface mpi;
  return mpi;
}

Interface::Interface() : driver_(GetLoader().getDriver()) {}

Interface::~Interface() { delete driver_; }

void Interface::Barrier(Comm comm) try {
  CheckSuccess(driver_->Barrier(comm));
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Barrier:\n{}", e.what());
}

auto Interface::COMM_SELF() -> Comm { return driver_->COMM_SELF(); }

auto Interface::COMM_WORLD() -> Comm { return driver_->COMM_WORLD(); }

int Interface::Comm_rank(Comm comm) try {
  int rank = -1;
  CheckSuccess(driver_->Comm_rank(comm, &rank));
  if (rank < 0) {
    throw Error("MPI rank was {}, all ranks must be >= 0", rank);
  }
  return rank;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Comm_rank:\n{}", e.what());
}

int Interface::Comm_size(Comm comm) try {
  int size = 0;
  CheckSuccess(driver_->Comm_size(comm, &size));
  if (size <= 0) {
    throw Error("MPI size was {}, world size must be > 0", size);
  }
  return size;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Comm_size:\n{}", e.what());
}

std::string Interface::Error_string(int errorcode) {
  auto buff = std::make_unique<char[]>(driver_->MAX_ERROR_STRING());
  int len = 0;
  int res = driver_->Error_string(errorcode, buff.get(), &len);
  if (res != 0) {
    return "unknown error (MPI_Error_string returned " + std::to_string(res) +
           ")";
  } else {
    return std::string(buff.get(), len);
  }
}

int Interface::Get_count(const Status *status, DataType datatype) try {
  int count;
  CheckSuccess(driver_->Get_count(status, datatype, &count));
  return count;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Get_count:\n{}", e.what());
}

std::string Interface::Get_processor_name() try {
  auto buff = std::make_unique<char[]>(driver_->MAX_PROCESSOR_NAME());
  int len = 0;
  CheckSuccess(driver_->Get_processor_name(buff.get(), &len));
  return std::string(buff.get(), len);
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Get_processor_name:\n{}", e.what());
}

auto Interface::FILE_NULL() -> File { return driver_->FILE_NULL(); }

void Interface::File_close(File *fh) try {
  CheckSuccess(driver_->File_close(fh));
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::File_close:\n{}", e.what());
}

void Interface::File_delete(const char *filename, Info info) try {
  CheckSuccess(driver_->File_delete(filename, info));
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::File_delete:\n{}", e.what());
}

mpi::Offset Interface::File_get_position(File fh) try {
  mpi::Offset p;
  CheckSuccess(driver_->File_get_position(fh, &p));
  return p;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::File_get_position:\n{}", e.what());
}

auto Interface::File_open(Comm comm, const char *filename, int amode,
                          Info info) -> File try {
  File f;
  CheckSuccess(driver_->File_open(comm, filename, amode, info, &f));
  return f;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::File_open:\n{}", e.what());
}

auto Interface::File_open_try(Comm comm, const char *filename, int amode,
                              Info info) -> std::optional<File> {
  File f;
  if (driver_->File_open(comm, filename, amode, info, &f) == 0) {
    return f;
  } else {
    return std::nullopt;
  }
}

void Interface::File_seek(File fh, mpi::Offset offset, int whence) try {
  CheckSuccess(driver_->File_seek(fh, offset, whence));
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::File_seek:\n{}", e.what());
}

void Interface::File_sync(File fh) try {
  CheckSuccess(driver_->File_sync(fh));
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::File_sync:\n{}", e.what());
}

auto Interface::File_write(File fh, const void *buf, int count,
                           DataType datatype) -> Status try {
  Status s;
  CheckSuccess(driver_->File_write(fh, buf, count, datatype, &s));
  return s;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::File_write:\n{}", e.what());
}

auto Interface::File_write_at(File fh, mpi::Offset offset, const void *buf,
                              int count, DataType datatype) -> Status try {
  Status s;
  CheckSuccess(driver_->File_write_at(fh, offset, buf, count, datatype, &s));
  return s;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::File_write_at:\n{}", e.what());
}

void Interface::Finalize() try {
  CheckSuccess(driver_->Finalize());
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Finalize:\n{}", e.what());
}

auto Interface::INFO_NULL() -> Info { return driver_->INFO_NULL(); }

bool Interface::Initialized() try {
  int i = 0;
  CheckSuccess(driver_->Initialized(&i));
  return i != 0;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Initialized:\n{}", e.what());
}

auto Interface::Init_thread(int *argc, char ***argv, Thread required) -> Thread
    try {
  Thread provided;
  CheckSuccess(driver_->Init_thread(argc, argv, required, &provided));
  if (provided < required) {
    throw std::runtime_error(
        "Insufficient threading support provided by MPI, needed " +
        std::to_string(static_cast<int>(required)) + ", but only got " +
        std::to_string(static_cast<int>(provided)));
  }
  return provided;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Init_thread:\n{}", e.what());
}

auto Interface::Probe(int source, int tag, Comm comm) -> Status try {
  Status s;
  CheckSuccess(driver_->Probe(source, tag, comm, &s));
  return s;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Probe:\n{}", e.what());
}

auto Interface::Recv(void *buf, int count, DataType datatype, int source,
                     int tag, Comm comm) -> Status try {
  Status s;
  CheckSuccess(driver_->Recv(buf, count, datatype, source, tag, comm, &s));
  return s;
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Recv:\n{}", e.what());
}

void Interface::Send(const void *buf, int count, DataType datatype, int dest,
                     int tag, Comm comm) try {
  CheckSuccess(driver_->Send(buf, count, datatype, dest, tag, comm));
} catch (const std::exception &e) {
  throw Error("MPI error in Interface::Send:\n{}", e.what());
}

void Interface::CheckSuccess(int errorcode) {
  if (errorcode == 0)
    return;
  throw std::runtime_error("MPI error " + std::to_string(errorcode) + ": " +
                           Error_string(errorcode));
}

} // namespace d5::mpi
