#include "dirsig5/raster/ReadRaster.h"

#include <algorithm>
#include <iostream>
#include <map>
#include <sstream>

#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/Logger.h"
#include "dirsig5/base/PluginLibraryManager.h"
#include "dirsig5/base/StringUtils.h"
#include "dirsig5/raster/RasterMap.h"

using namespace std::string_literals;

namespace d5::raster {

class PluginManager final
    : public PluginLibraryManager<apis::RasterInputInterface> {
public:
  using Super = PluginLibraryManager<apis::RasterInputInterface>;
  PluginManager() : Super("registerPlugin") {}

  void init() {
    loadDefaultLocations("raster_input");
    for (const auto &each : libraries) {
      const auto &filename = each.first;
      const auto &library = each.second;
      DIRSIG_ERROR_TRACE("loading raster input plugin \"" + filename + "\"", {
        std::shared_ptr<apis::RasterInputInterface> plugin;
        DIRSIG_ERROR_TRACE("invoking registerPlugin()", {
          plugin = std::shared_ptr<apis::RasterInputInterface>(
              reinterpret_cast<EntryPointFunc>(library.second)());
        });
        DIRSIG_THROW_IF(plugin == nullptr, "registerPlugin() returned null!");
        apis::RasterInputInit::In in;
        apis::RasterInputInit::Out out;
        try {
          plugin->init(in, out);
        } catch (const std::exception &error) {
          const char *message = error.what();
          if (not out.error.empty())
            message = out.error.c_str();
          DIRSIG_WARN("initialization for raster input plugin \"{}\" "
                      "failed: {}",
                      filename, message);
          return;
        }
        for (const auto &extension : out.extensions)
          pluginsForExtension[toLower(extension)].push_back(plugins.size());
        if (out.tryWithUnrecognizedExtension)
          pluginsForUnrecognizedExtension.push_back(plugins.size());
        if (out.extensions.empty() and not out.tryWithUnrecognizedExtension)
          DIRSIG_WARN("raster input plugin '{}' supports no extensions, "
                      "and did not set tryWithUnrecognizedExtension, so "
                      "it will never be used",
                      filename);
        plugins.push_back(std::move(plugin));
      });
    }
  }

  std::vector<std::shared_ptr<apis::RasterInputInterface>> plugins;
  std::map<std::string, std::vector<std::size_t>> pluginsForExtension;
  std::vector<std::size_t> pluginsForUnrecognizedExtension;
};

void PopulateMetadata(Raster &im, const apis::RasterMetadata &metadata) {
  for (std::size_t i = 0; i < metadata.bands.size(); i++) {
    const auto &b = metadata.bands[i];
    if (!b)
      continue;
    auto &c = im.channelMetadata(i);
    c.name = b->name;
    c.centerWavelengthMicron = b->centerWavelengthMicron;
    c.widthMicron = b->widthMicron;
  }
}

const auto &GetManager() {
  static auto manager = []() {
    PluginManager m;
    m.init();
    return m;
  }();
  return manager;
}

const std::vector<std::size_t> &GetPluginIndices(const std::string &path) {
  const auto &manager = GetManager();
  auto lastDot = path.find_last_of('.');
  if (lastDot != std::string::npos && lastDot + 1 != path.size()) {
    auto extension = toLower(path.substr(lastDot + 1));
    if (auto it = manager.pluginsForExtension.find(extension);
        it != manager.pluginsForExtension.end()) {
      return it->second;
    }
  }
  return manager.pluginsForUnrecognizedExtension;
}

std::string MakeErrorMessage(const std::vector<std::size_t> &pluginIndices,
                             const std::vector<std::string> &errors) {
  const auto &manager = GetManager();
  std::stringstream e;
  e << "no raster plugins could read the image.\n"
       "Tried the following plugins:";
  for (size_t i = 0; i < errors.size(); i++) {
    e << "\n  \"" << manager.plugins[pluginIndices[i]]->name() << "\": ";
    if (errors[i].empty())
      e << "no error provided";
    else
      e << errors[i];
  }
  return e.str();
}

struct RasterDataAdapter : apis::BipRasterDestination {
public:
  RasterDataAdapter(Raster &im) : im_(im) {}

  void allocate(size_t b, size_t r, size_t c,
                apis::RasterDataType dt) override {
    im_.create(r, c, b, static_cast<Raster::DataType>(dt));
  }

  char *row(size_t /*b*/, size_t r) override {
    return reinterpret_cast<char *>(im_.rawRow(r));
  }

private:
  Raster &im_;
};

Raster ReadRaster(const apis::RasterInputRead::In &request) {
  VerifyPathIsFile(request.path);
  const auto &pluginIndices = GetPluginIndices(request.path);
  DIRSIG_THROW_IF_FMT(pluginIndices.empty(),
                      "Can't read raster file '{}'!\n"
                      "Reason: no suitable plugin supports extension",
                      request.path);
  const auto &manager = GetManager();
  Raster dst;
  std::vector<std::string> errors;
  RasterDataAdapter adapter(dst);
  for (auto index : pluginIndices) {
    apis::RasterInputRead::Out out;
    if (manager.plugins[index]->read(request, adapter, out)) {
      PopulateMetadata(dst, out.metadata);
      return dst;
    } else
      errors.push_back(std::move(out.error));
  }
  DIRSIG_THROW_FMT("Can't read raster file '{}'!\nReason: {}", request.path,
                   MakeErrorMessage(pluginIndices, errors));
  return {};
}

bool TryToReadRasterMetadata(const apis::RasterMetadataRead::In &request,
                             apis::RasterMetadataRead::Out &result) {
  VerifyPathIsFile(request.path);
  const auto &pluginIndices = GetPluginIndices(request.path);
  if (pluginIndices.empty()) {
    result.error = "no suitable plugin supports extension";
    return false;
  }

  const auto &manager = GetManager();
  std::vector<std::string> errors;
  for (auto index : pluginIndices) {
    if (manager.plugins[index]->metadata(request, result))
      return true;
    errors.push_back(result.error);
  }
  result.error = MakeErrorMessage(pluginIndices, errors);
  return false;
}

} // namespace d5::raster
