#pragma once

#include "dirsig5/base/Geometry.h"

namespace d5::sqt {

/// Cosinify a direction.
///
/// This function applies a cosine-weighting transformation to the
/// direction `vec`. This function is defined such that, if `vec`
/// is a uniform sample over the hemisphere, then `cosinify(vec)`
/// is a cosine-weighted sample over the hemisphere.
///
/// \note
/// This preserves the hemisphere that `vec` is in, i.e., the
/// sign of `vec[2]`.
///
inline Vectorf cosinify(Vectorf vec) {
  float sinTheta = std::hypot(vec[0], vec[1]);
  if (sinTheta > 1e-4f) {
    float cosTheta = vec[2];
    float newCosTheta = std::sqrt(std::abs(cosTheta));
    float newSinTheta = std::sqrt(std::fmax(1 - std::abs(cosTheta), 0.0f));
    float sinSwap = newSinTheta / sinTheta;
    vec[0] *= sinSwap;
    vec[1] *= sinSwap;
    vec[2] = std::copysign(newCosTheta, cosTheta);
    vec.normalize();
  }
  return vec;
}

/// Uncosinify a direction.
///
/// This function applies the inverse transformation to `cosinify()`,
/// such that `uncosinify(cosinify(vec))` should be the same as `vec`
/// plus or minus some rounding error.
///
/// \note
/// This preserves the hemisphere that `vec` is in, i.e., the
/// sign of `vec[2]`.
///
inline Vectorf uncosinify(Vectorf vec) {
  float sinTheta = std::hypot(vec[0], vec[1]);
  if (sinTheta > 1e-4f) {
    float cosTheta = vec[2];
    float newCosTheta = std::abs(cosTheta) * cosTheta;
    float newSinTheta =
      std::sqrt(std::fmax(1 - newCosTheta * newCosTheta, 0.0f));
    float sinSwap = newSinTheta / sinTheta;
    vec[0] *= sinSwap;
    vec[1] *= sinSwap;
    vec[2] = newCosTheta;
    vec.normalize();
  }
  return vec;
}

}  // namespace d5::sqt
