#include "dirsig5/sqt/Raw2Sqt.h"

#include <array>
#include <cstdint>
#include <fstream>
#include <numeric>
#include <sstream>

#include "dirsig5/sqt/ScatteringFunction.h"
#include "dirsig5/sqt/SpectralSphereQuadTree.h"
#include "dirsig5/sqt/SphereHEI.h"
#include "dirsig5/sqt/SphereQuadTree.h"

Raw2SqtResult raw2sqt(const Raw2SqtRequest& request) {
  using namespace d5;
  using namespace d5::sqt;

  // read the header
  auto& in = *request.rawIn;
  auto& out = *request.sqtOut;
  std::string str;
  std::getline(in, str);

  DIRSIG_THROW_IF(str[5] != '1' or str[6] != '0', "Unsupported version!");
  DIRSIG_THROW_IF(str[7] != 'A', "Unsupported format");
  bool isBidir = str[3] == 'B' or str[3] == 'A';
  bool isAniso = str[3] == 'A';
  bool isUpper = str[4] == 'H' or str[4] == 'R' or str[4] == 'S';
  bool isLower = str[4] == 'T' or str[4] == 'S';
  Flags flags = 0;
  if (isUpper) flags |= UPPER_HEMISPHERE_BIT;
  if (isLower) flags |= LOWER_HEMISPHERE_BIT;
  if (isBidir) flags |= COSINE_WEIGHT_BIT;

  // Read wavelengths.
  size_t waveCount = 0;
  in >> waveCount;
  std::vector<float> waves(waveCount);
  for (size_t waveIndex = 0; waveIndex < waveCount; waveIndex++)
    in >> waves[waveIndex];

  // Read thetas.
  size_t thetaCount = 1;
  std::vector<float> thetas(1);
  if (isBidir) {
    in >> thetaCount;
    thetas.resize(thetaCount);
    for (size_t thetaIndex = 0; thetaIndex < thetaCount; thetaIndex++) {
      in >> thetas[thetaIndex];
      DIRSIG_THROW_IF(
        thetaIndex > 0 and not(thetas[thetaIndex] > thetas[thetaIndex - 1]),
        "Viewing zenith angles must be strictly increasing!");
    }
  }

  // Read phis.
  size_t phiCount = 1;
  std::vector<float> phis(1);
  if (isAniso) {
    in >> phiCount;
    phis.resize(phiCount);
    for (size_t phiIndex = 0; phiIndex < phiCount; phiIndex++) {
      in >> phis[phiIndex];
      DIRSIG_THROW_IF(
        phiIndex > 0 and not(phis[phiIndex] > phis[phiIndex - 1]),
        "Viewing azimuth angles must be strictly increasing!");
      DIRSIG_THROW_IF(
        not(phis[phiIndex] >= 0 and phis[phiIndex] <= M_PI),
        "Viewing azimuth angles must be between 0 and pi!");
    }
    DIRSIG_THROW_IF(phis[0] != 0, "The first viewing azimuth must be 0!");
  }

  struct Sample {
    Vectorf vec;
    std::vector<float> values;
  };
  std::vector<std::vector<Sample>> samplesTable;
  samplesTable.resize(thetaCount * phiCount);
  auto lookupSamples =
    [&](size_t thetaIndex, size_t phiIndex) -> std::vector<Sample>& {
    return samplesTable[phiIndex * thetaCount + thetaIndex];
  };

  while (!in.eof()) {
    std::getline(in, str);
    if (str.empty()) continue;
    std::stringstream sstr(str);
    size_t thetaIndex = 0;
    size_t phiIndex = 0;
    if (isBidir) sstr >> thetaIndex;
    if (isAniso) sstr >> phiIndex;
    Sample& sample = lookupSamples(thetaIndex, phiIndex).emplace_back();
    sstr >> sample.vec[0];
    sstr >> sample.vec[1];
    sstr >> sample.vec[2];
    sample.values.reserve(waveCount);
    float value = 0;
    /**
     * The original while loop below creates an evil error when trying to
     * read very small values that are below the minimum value of a
     * float -- the while ends and the rest of the line is skipped
     * while (sstr >> value) sample.values.push_back(value);
     * Instead we'll just try to read in what we expect
     */
    for (size_t ii = 0; ii < waveCount; ++ii) {
      sstr >> value;
      sample.values.push_back(value);
    }
  }

  ScatteringFunction* scattering = new ScatteringFunction();
  scattering->flags = flags;
  scattering->slices.resize(phiCount);
  SpectralSphereQuadTree::RawData rawData;
  rawData.maxDepth = request.maxDepth;
  rawData.compressThreshold = request.compressThreshold;
  rawData.interpolation = request.interpolation;
  rawData.verboseOutput = request.verboseOutput;
  rawData.flags = flags;
  rawData.waves = waves;
  for (uint32_t phiIndex = 0; phiIndex < phiCount; phiIndex++) {
    ScatteringFunctionSlice* slice = new ScatteringFunctionSlice();
    scattering->slices[phiIndex].phiO = phis[phiIndex];
    scattering->slices[phiIndex].slice = slice;
    slice->spheres.resize(thetaCount);
    if (waves.size() < 5) rawData.waveOptimizeThreshold = 0;
    for (size_t thetaIndex = 0; thetaIndex < thetaCount; thetaIndex++) {
      auto& samples = lookupSamples(thetaIndex, phiIndex);
      rawData.resize(samples.size(), waveCount);
      for (size_t sampleIndex = 0; sampleIndex < samples.size();
           sampleIndex++) {
        rawData.sampleVectors(sampleIndex, 0) = samples[sampleIndex].vec[0];
        rawData.sampleVectors(sampleIndex, 1) = samples[sampleIndex].vec[1];
        rawData.sampleVectors(sampleIndex, 2) = samples[sampleIndex].vec[2];
        for (size_t waveIndex = 0; waveIndex < waveCount; waveIndex++) {
          assert(waveIndex < samples[sampleIndex].values.size());
          rawData.sampleSpectra(sampleIndex, waveIndex) =
            samples[sampleIndex].values[waveIndex];
        }
      }
      slice->spheres[thetaIndex].thetaO = thetas[thetaIndex];
      slice->spheres[thetaIndex].quadTree = new SpectralSphereQuadTree();
      slice->spheres[thetaIndex].quadTree->buildFromRawData(rawData);
    }
  }
  if (request.nadirMagnitudeIn)
    scattering->overrideNadirMagnitude(*request.nadirMagnitudeIn);
  scattering->optimizeMagnitude(0.005f);
  scattering->write(out);
  delete scattering;

  return Raw2SqtResult();
}
