#include "dirsig5/sqt/ScatteringFunction.h"

#include <sstream>

#include "dirsig5/base/BinaryStream.h"
#include "dirsig5/base/ErrorIO.h"
#include "dirsig5/base/LerpUtils.h"
#include "dirsig5/base/Logger.h"

namespace d5::sqt {

ScatteringFunction::ScatteringFunction(void* buf, size_t bufLen) {
  DIRSIG_THROW_IF(buf == nullptr or bufLen == 0, "Invalid SQT buffer!");
  auto rawStreambuf = RawIStreambuf(buf, bufLen);
  auto rawStream = std::istream(&rawStreambuf);
  read(rawStream);
}

void ScatteringFunction::clear() {
  for (Slice& slice : slices) {
    delete slice.slice;
    slice.slice = nullptr;
  }
  slices.clear();
}

void ScatteringFunction::read(const std::string& filename) {
  auto ifs = OpenIfstream(filename, std::ios_base::binary);
  read(ifs);
}

void ScatteringFunction::read(std::istream& in) {
  clear();
  BinaryStream binaryIn(&in);
  char header[8] = {};
  binaryIn.read(&header[0], 8);
  DIRSIG_THROW_IF_FMT(
    header[5] != '2' or header[6] != '0',
    "Out-of-date SQT version {}.{} (Expected 2.0)", header[5], header[6]);
  binaryIn.read(&flags, 1);
  slices.resize(binaryIn.read<uint32_t>());
  for (Slice& slice : slices) {
    slice.phiO = binaryIn.read<float>();
    slice.slice = new ScatteringFunctionSlice(in);
  }
}

void ScatteringFunction::write(const std::string& filename) const {
  auto ofs = OpenOfstream(filename, std::ios_base::binary);
  write(ofs);
}

void ScatteringFunction::write(std::ostream& out) const {
  BinaryStream binaryOut(&out);
  const char* header = "SQTAS20";
  binaryOut.write(&header[0], 8);
  binaryOut.write(&flags, 1);
  binaryOut.write(uint32_t(slices.size()));
  for (const Slice& slice : slices) {
    binaryOut.write(slice.phiO);
    slice.slice->write(out);
  }
}

std::string ScatteringFunction::writeToString() const {
  std::ostringstream out;
  write(out);
  return out.str();
}

float ScatteringFunction::computeBroadbandSolarAverage() const {
  const SpectralCurve& weight = slices[0].slice->spheres[0].quadTree->weight;
  return weight.getIntegral(0.25f, 2.55f) / (2.55f - 0.25f);
}

float ScatteringFunction::computeBroadbandThermalAverage() const {
  const SpectralCurve& weight = slices[0].slice->spheres[0].quadTree->weight;
  return weight.getIntegral(3.00f, 20.00f) / (20.00f - 3.00f);
}

void ScatteringFunction::forceNormalize() {
  const auto& weight = slices[0].slice->spheres[0].quadTree->weight;
  float minWave = weight.getMinWave();
  float maxWave = weight.getMaxWave();
  float waves[2] = {minWave, maxWave};
  float refls[2] = {1.0f, 1.0f};
  SpectralCurve white = {
    span<const float>(&waves[0], 2),  //
    span<const float>(&refls[0], 2)};
  return overrideNadirMagnitude(white);
}

void ScatteringFunction::overrideNadirMagnitude(
  const SpectralCurve& magnitude) {
  if (magnitude.samples.empty()) return;
  SpectralCurve target = magnitude;
  SpectralCurve source = slices[0].slice->spheres[0].quadTree->weight;
  target.resampleToUnion(source.unzipWaves());
  target /= source;
  for (Slice& slice : slices) {
    for (auto& sphere : slice.slice->spheres) {
      sphere.quadTree->weight.resampleToUnion(target.unzipWaves());
      sphere.quadTree->weight *= target;
      sphere.quadTree->weight.clampToZeroOne();
    }
  }
}

void ScatteringFunction::optimizeMagnitude(float tolerance) {
  for (Slice& slice : slices)
    for (auto& sphere : slice.slice->spheres)
      sphere.quadTree->weight.optimizeLinear(tolerance);
}

float ScatteringFunction::getMagnitude(
  Vectorf omegaO, float wave, bool includeUpper, bool includeLower) const {
  OmegaOFrame frame(omegaO);
  auto [fac, slice0, slice1] = findSlices(frame.phiO);
  float magnitude = 0;
  if (fac < 0.0001f)
    magnitude = slice0->slice->getMagnitude(
      frame.thetaO, wave, includeUpper, includeLower);
  else if (fac > 0.9999f)
    magnitude = slice1->slice->getMagnitude(
      frame.thetaO, wave, includeUpper, includeLower);
  else
    magnitude = lerp(
      fac,  //
      slice0->slice->getMagnitude(
        frame.thetaO, wave, includeUpper, includeLower),
      slice1->slice->getMagnitude(
        frame.thetaO, wave, includeUpper, includeLower));
  return magnitude;
}

void ScatteringFunction::getMagnitudeAtStateWavelengths(
  Vectorf omegaO,
  bool includeUpper,
  bool includeLower,
  SpectralVector& magnitude) const {
  OmegaOFrame frame(omegaO);
  auto [fac, slice0, slice1] = findSlices(frame.phiO);
  if (fac < 0.0001f)
    slice0->slice->getMagnitudeAtStateWavelengths(
      frame.thetaO, includeUpper, includeLower, magnitude);
  else if (fac > 0.9999f)
    slice1->slice->getMagnitudeAtStateWavelengths(
      frame.thetaO, includeUpper, includeLower, magnitude);
  else {
    SpectralVector other;
    slice0->slice->getMagnitudeAtStateWavelengths(
      frame.thetaO, includeUpper, includeLower, magnitude);
    slice1->slice->getMagnitudeAtStateWavelengths(
      frame.thetaO, includeUpper, includeLower, other);
    other *= double(fac);
    magnitude *= double(1 - fac);
    magnitude += other;
  }
}

float ScatteringFunction::getValue(
  Vectorf omegaO, Vectorf omegaI, float wave) const {
  OmegaOFrame frame(omegaO);
  omegaI = frame.localOmegaI(omegaI);
  auto [fac, slice0, slice1] = findSlices(frame.phiO);
  float value = 0;
  if (fac < 0.0001f)
    value = slice0->slice->getValue(frame.thetaO, omegaI, wave);
  else if (fac > 0.9999f)
    value = slice1->slice->getValue(frame.thetaO, omegaI, wave);
  else
    value = lerp(
      fac,                                                  //
      slice0->slice->getValue(frame.thetaO, omegaI, wave),  //
      slice1->slice->getValue(frame.thetaO, omegaI, wave));
  return value;
}

void ScatteringFunction::getValueAtStateWavelengths(
  Vectorf omegaO, Vectorf omegaI, SpectralVector& value) const {
  OmegaOFrame frame(omegaO);
  omegaI = frame.localOmegaI(omegaI);
  auto [fac, slice0, slice1] = findSlices(frame.phiO);
  if (fac < 0.0001f)
    slice0->slice->getValueAtStateWavelengths(frame.thetaO, omegaI, value);
  else if (fac > 0.9999f)
    slice1->slice->getValueAtStateWavelengths(frame.thetaO, omegaI, value);
  else {
    SpectralVector other;
    slice0->slice->getValueAtStateWavelengths(frame.thetaO, omegaI, value);
    slice1->slice->getValueAtStateWavelengths(frame.thetaO, omegaI, other);
    other *= double(fac);
    value *= double(1 - fac);
    value += other;
  }
}

float ScatteringFunction::getPdf(
  Vectorf omegaO, Vectorf omegaI, bool allowUpper, bool allowLower) const {
  OmegaOFrame frame(omegaO);
  omegaI = frame.localOmegaI(omegaI);
  auto [fac, slice0, slice1] = findSlices(frame.phiO);
  if (fac < 0.0001f)
    return slice0->slice->getPdf(frame.thetaO, omegaI, allowUpper, allowLower);
  if (fac > 0.9999f)
    return slice1->slice->getPdf(frame.thetaO, omegaI, allowUpper, allowLower);
  return lerp(
    fac,  //
    slice0->slice->getPdf(frame.thetaO, omegaI, allowUpper, allowLower),
    slice1->slice->getPdf(frame.thetaO, omegaI, allowUpper, allowLower));
}

Vectorf ScatteringFunction::samplePdf(
  RandEngine& prng,
  Vectorf omegaO,
  bool allowUpper,
  bool allowLower,
  SpectralVector* weights) const {
  OmegaOFrame frame(omegaO);
  auto [fac, slice0, slice1] = findSlices(frame.phiO);
  Vectorf omegaI =
    float(canonical(prng)) < 1 - fac
      ? slice0->slice->samplePdf(prng, frame.thetaO, allowUpper, allowLower)
      : slice1->slice->samplePdf(prng, frame.thetaO, allowUpper, allowLower);
  // Weights requested?
  if (weights) {
    // This may be incorrect at wavelengths outside the reference, but
    // it tends to be relatively accurate and saves a ton of computation
    // time.
    if (fac < 0.0001f)
      slice0->slice->getMagnitudeAtStateWavelengths(
        frame.thetaO, allowUpper, allowLower, *weights);
    else if (fac > 0.9999f)
      slice1->slice->getMagnitudeAtStateWavelengths(
        frame.thetaO, allowUpper, allowLower, *weights);
    else {
      SpectralVector other;
      slice0->slice->getMagnitudeAtStateWavelengths(
        frame.thetaO, allowUpper, allowLower, *weights);
      slice1->slice->getMagnitudeAtStateWavelengths(
        frame.thetaO, allowUpper, allowLower, other);
      other *= double(fac);
      *weights *= double(1 - fac);
      *weights += other;
    }
  }
  return frame.worldOmegaI(omegaI);
}

ScatteringFunction::FindSlicesResult ScatteringFunction::findSlices(
  float phiO) const {
  DIRSIG_THROW_IF(slices.empty(), "Can't find slices in SQT. No samples!");
  if (slices.size() == 1) return {0, slices.data(), slices.data()};
  if (phiO < 0) phiO += M_PI;
  phiO = std::fmax(phiO, 0.0f);
  phiO = std::fmin(phiO, float(M_PI));
  auto [slice0, slice1] = findLerpIterators(
    slices.begin(), slices.end(), phiO,
    [](const Slice& slice, float phiO) { return slice.phiO < phiO; });
  return {
    findLerpFactorClamped(phiO, slice0->phiO, slice1->phiO),  //
    &*slice0, &*slice1};
}

}  // namespace d5::sqt
