#include "dirsig5/sqt/ScatteringFunctionSlice.h"

#include "dirsig5/base/BinaryStream.h"
#include "dirsig5/base/LerpUtils.h"

namespace d5::sqt {

void ScatteringFunctionSlice::clear() {
  for (Sphere& sphere : spheres) {
    delete sphere.quadTree;
    sphere.quadTree = nullptr;
  }
  spheres.clear();
}

void ScatteringFunctionSlice::read(std::istream& in) {
  clear();
  BinaryStream binaryIn(&in);
  spheres.resize(binaryIn.read<uint32_t>());
  for (Sphere& sphere : spheres) {
    sphere.thetaO = binaryIn.read<float>();
    sphere.quadTree = new SpectralSphereQuadTree(in);
  }
}

void ScatteringFunctionSlice::write(std::ostream& out) const {
  BinaryStream binaryOut(&out);
  binaryOut.write(uint32_t(spheres.size()));
  for (const Sphere& sphere : spheres) {
    binaryOut.write(sphere.thetaO);
    sphere.quadTree->write(out);
  }
}

float ScatteringFunctionSlice::getMagnitude(
  float thetaO, float wave, bool includeUpper, bool includeLower) const {
  auto [fac, sphere0, sphere1] = findSpheres(thetaO);
  if (fac < 0.0001f)
    return sphere0->quadTree->getMagnitude(wave, includeUpper, includeLower);
  if (fac > 0.9999f)
    return sphere1->quadTree->getMagnitude(wave, includeUpper, includeLower);
  return lerp(
    fac,  //
    sphere0->quadTree->getMagnitude(wave, includeUpper, includeLower),
    sphere0->quadTree->getMagnitude(wave, includeUpper, includeLower));
}

void ScatteringFunctionSlice::getMagnitudeAtStateWavelengths(
  float thetaO,
  bool includeUpper,
  bool includeLower,
  SpectralVector& magnitude) const {
  auto [fac, sphere0, sphere1] = findSpheres(thetaO);
  if (fac < 0.0001f)
    sphere0->quadTree->getMagnitudeAtStateWavelengths(
      includeUpper, includeLower, magnitude);
  else if (fac > 0.9999f)
    sphere1->quadTree->getMagnitudeAtStateWavelengths(
      includeUpper, includeLower, magnitude);
  else {
    SpectralVector other;
    sphere0->quadTree->getMagnitudeAtStateWavelengths(
      includeUpper, includeLower, magnitude);
    sphere1->quadTree->getMagnitudeAtStateWavelengths(
      includeUpper, includeLower, other);
    other *= double(fac);
    magnitude *= double(1 - fac);
    magnitude += other;
  }
}

float ScatteringFunctionSlice::getValue(
  float thetaO, Vectorf omegaI, float wave) const {
  auto [fac, sphere0, sphere1] = findSpheres(thetaO);
  if (fac < 0.0001f) return sphere0->quadTree->getValue(omegaI, wave);
  if (fac > 0.9999f) return sphere1->quadTree->getValue(omegaI, wave);
  return lerp(
    fac,                                        //
    sphere0->quadTree->getValue(omegaI, wave),  //
    sphere1->quadTree->getValue(omegaI, wave));
}

void ScatteringFunctionSlice::getValueAtStateWavelengths(
  float thetaO, Vectorf omegaI, SpectralVector& value) const {
  auto [fac, sphere0, sphere1] = findSpheres(thetaO);
  if (fac < 0.0001f)
    sphere0->quadTree->getValueAtStateWavelengths(omegaI, value);
  else if (fac > 0.9999f)
    sphere1->quadTree->getValueAtStateWavelengths(omegaI, value);
  else {
    SpectralVector other;
    sphere0->quadTree->getValueAtStateWavelengths(omegaI, value);
    sphere1->quadTree->getValueAtStateWavelengths(omegaI, other);
    other *= double(fac);
    value *= double(1 - fac);
    value += other;
  }
}

float ScatteringFunctionSlice::getPdf(
  float thetaO, Vectorf omegaI, bool allowUpper, bool allowLower) const {
  auto [fac, sphere0, sphere1] = findSpheres(thetaO);
  if (fac < 0.0001f)
    return sphere0->quadTree->getPdf(omegaI, allowUpper, allowLower);
  if (fac > 0.9999f)
    return sphere1->quadTree->getPdf(omegaI, allowUpper, allowLower);
  return lerp(
    fac,  //
    sphere0->quadTree->getPdf(omegaI, allowUpper, allowLower),
    sphere1->quadTree->getPdf(omegaI, allowUpper, allowLower));
}

Vectorf ScatteringFunctionSlice::samplePdf(
  RandEngine& prng, float thetaO, bool allowUpper, bool allowLower) const {
  auto [fac, sphere0, sphere1] = findSpheres(thetaO);
  return float(canonical(prng)) < 1 - fac
           ? sphere0->quadTree->samplePdf(prng, allowUpper, allowLower)
           : sphere1->quadTree->samplePdf(prng, allowUpper, allowLower);
}

ScatteringFunctionSlice::FindSpheresResult ScatteringFunctionSlice::findSpheres(
  float thetaO) const {
  DIRSIG_THROW_IF(spheres.empty(), "No samples!");
  auto [sphere0, sphere1] = findLerpIterators(
    spheres.begin(), spheres.end(), thetaO,
    [](const Sphere& sphere, float thetaO) { return sphere.thetaO < thetaO; });
  return {
    findLerpFactorClamped(thetaO, sphere0->thetaO, sphere1->thetaO), &*sphere0,
    &*sphere1};
}

}  // namespace d5::sqt
