#include "dirsig5/sqt/SpectralCurve.h"

#include <cstdint>
#include <istream>
#include <ostream>

#include "dirsig5/base/BinaryStream.h"
#include "dirsig5/base/LerpUtils.h"
#include "dirsig5/base/Pcg.h"

namespace d5::sqt {

void SpectralCurve::read(std::istream& in) {
  BinaryStream binaryIn(&in);
  samples.resize(binaryIn.read<uint32_t>());
  binaryIn.read(samples.data(), samples.size());
}

void SpectralCurve::write(std::ostream& out) const {
  BinaryStream binaryOut(&out);
  binaryOut.write(uint32_t(samples.size()));
  binaryOut.write(samples.data(), samples.size());
}

float SpectralCurve::getValue(float wave) const {
  if (samples.empty()) return 0;
  auto [sample0, sample1] = findLerpIterators(
    samples.begin(), samples.end(), wave,
    [](const Sample& sample, float wave) { return sample.wave < wave; });
  return lerp(
    findLerpFactorClamped(wave, sample0->wave, sample1->wave), sample0->value,
    sample1->value);
}

float SpectralCurve::getIntegral(float wave0, float wave1) const {
  if (wave0 >= wave1) std::swap(wave0, wave1);
  if (wave0 == wave1) return 0;
  if (samples.size() == 0) return 0;
  if (samples.size() == 1) return samples[0].value * (wave1 - wave0);
  auto [sample0, sample1] = findLerpIterators(
    samples.begin(), samples.end(), wave0,
    [](const Sample& sample, float wave) { return sample.wave < wave; });
  auto [sample2, sample3] = findLerpIterators(
    samples.begin(), samples.end(), wave1,
    [](const Sample& sample, float wave) { return sample.wave < wave; });
  float result = 0;
  if (wave0 < samples.front().wave) {
    result += samples.front().value * (samples.front().wave - wave0);
    wave0 = samples.front().wave;
  }
  if (wave1 > samples.back().wave) {
    result += samples.back().value * (wave1 - samples.back().wave);
    wave1 = samples.back().wave;
  }
  float value0 = lerp(
    findLerpFactor(wave0, sample0->wave, sample1->wave), sample0->value,
    sample1->value);
  float value1 = lerp(
    findLerpFactor(wave1, sample2->wave, sample3->wave), sample2->value,
    sample3->value);
  result += (sample1->wave - wave0) * (value0 + sample1->value) * 0.5f;
  result += (wave1 - sample2->wave) * (sample2->value + value1) * 0.5f;
  for (auto sample = sample1; sample < sample2; sample++) {
    float waveA = (sample + 0)->wave, valueA = (sample + 0)->value;
    float waveB = (sample + 1)->wave, valueB = (sample + 1)->value;
    result += (waveB - waveA) * (valueB + valueA) * 0.5f;
  }
  return result;
}

template <typename Float>
void SpectralCurve::resampleTo(span<const Float> newWaves) {
  static_assert(std::is_floating_point_v<Float>);
  std::vector<Sample> newSamples;
  newSamples.resize(newWaves.size());
  for (size_t index = 0; index < newSamples.size(); index++) {
    newSamples[index].wave = newWaves[index];
    newSamples[index].value = getValue(newWaves[index]);
  }
  samples = std::move(newSamples);
}

template <typename Float>
void SpectralCurve::resampleToUnion(span<const Float> otherWaves) {
  std::vector<float> newWaves;
  newWaves = unzipWaves();
  newWaves.insert(newWaves.end(), otherWaves.begin(), otherWaves.end());
  std::sort(newWaves.begin(), newWaves.end());
  auto newEnd = std::unique(
    newWaves.begin(), newWaves.end(),
    [](float lhs, float rhs) { return std::fabs(lhs - rhs) <= 1e-4f; });
  newWaves.erase(newEnd, newWaves.end());
  resampleTo(newWaves);
}

template void SpectralCurve::resampleTo<float>(span<const float>);

template void SpectralCurve::resampleTo<double>(span<const double>);

template void SpectralCurve::resampleToUnion<float>(span<const float>);

template void SpectralCurve::resampleToUnion<double>(span<const double>);

void SpectralCurve::optimizeLinear(float tolerance) {
  Pcg32 gen;
  int fail = 0;
  while (fail++ < 32 and samples.size() > 2) {
    int index = gen(samples.size() - 2) + 1;
    auto [wave, value] = samples[index];
    auto [prevWave, prevValue] = samples[index - 1];
    auto [nextWave, nextValue] = samples[index + 1];
    float fac = findLerpFactor(wave, prevWave, nextWave);
    float err = std::fabs(lerp(fac, prevValue, nextValue) - value);
    if (err <= tolerance) {
      samples.erase(samples.begin() + index);
      fail = 0;
    }
  }
}

}  // namespace d5::sqt
