#include "dirsig5/sqt/SpectralSphereQuadTree.h"

#include "SphereInterpolator.h"
#include "dirsig5/base/BinaryStream.h"
#include "dirsig5/base/LerpUtils.h"
#include "dirsig5/base/Logger.h"

namespace d5::sqt {

void SpectralSphereQuadTree::buildFromRawData(const RawData& rawData) {
  clear();
  Flags flags = rawData.flags;
  unsigned maxDepth = rawData.maxDepth;
  const auto& waves = rawData.waves;
  const auto& sampleVectors = rawData.sampleVectors;
  const auto& sampleSpectra = rawData.sampleSpectra;
  int waveCount = waves.size();
  DIRSIG_THROW_IF((flags & FULL_SPHERE) == 0, "Invalid flags!");
  DIRSIG_THROW_IF(maxDepth == 0 or maxDepth > 8, "Invalid depth!");
  DIRSIG_THROW_IF(
    waveCount == 0 or               //
      sampleVectors.rows() == 0 or  //
      sampleSpectra.rows() != sampleVectors.rows() or
      sampleSpectra.cols() != waveCount,
    "Invalid sample data!");
  // Allocate.
  spheres.resize(waves.size());
  for (int waveIndex = 0; waveIndex < waveCount; waveIndex++) {
    spheres[waveIndex].wave = waves[waveIndex];
    spheres[waveIndex].quadTree = new SphereQuadTree(maxDepth, flags);
  }
  if (rawData.interpolation) {
    RTCDevice device = rtcNewDevice(nullptr);
    SphereInterpolator* interpolator =
      new SphereInterpolator(device, waveCount, flags);
    for (int sampleIndex = 0; sampleIndex < int(sampleVectors.rows());
         sampleIndex++) {
      Vectorf sampleVector = {
        sampleVectors(sampleIndex, 0),  //
        sampleVectors(sampleIndex, 1),  //
        sampleVectors(sampleIndex, 2)};
      interpolator->addPoint(
        sampleVector, {&sampleSpectra(sampleIndex, 0), waves.size()});
    }
    interpolator->finalize();
    SphereHEI cells(maxDepth);
    int cellCount = cells.getCellCount();
    for (int cellIndex = 0; cellIndex < cellCount; cellIndex++) {
      Vectorf vec = cells.indexToCenterVector(cellIndex);
      SpectralVector values = interpolator->interpolate(vec);
      for (int waveIndex = 0; waveIndex < waveCount; waveIndex++)
        spheres[waveIndex].quadTree->addVectorData(vec, values[waveIndex]);
    }
    delete interpolator;
    rtcReleaseDevice(device);
  } else {
    for (int sampleIndex = 0; sampleIndex < int(sampleVectors.rows());
         sampleIndex++) {
      Vectorf sampleVector = {
        sampleVectors(sampleIndex, 0),  //
        sampleVectors(sampleIndex, 1),  //
        sampleVectors(sampleIndex, 2)};
      sampleVector.normalize();
      for (int waveIndex = 0; waveIndex < waveCount; waveIndex++)
        spheres[waveIndex].quadTree->addVectorData(
          sampleVector, sampleSpectra(sampleIndex, waveIndex));
    }
  }

  // Finalize and compress.
  for (auto& [wave, quadTree] : spheres) {
    quadTree->finalize();
    if (rawData.clampMagnitude) quadTree->clampMagnitude();
    if (rawData.compressThreshold > 0)
      quadTree->compress(rawData.compressThreshold);
  }

  // Bake magnitude into weight curve.
  weight.samples.clear();
  weight.samples.reserve(spheres.size());
  for (Sphere& sphereSample : spheres) {
    auto& weightSample = weight.samples.emplace_back();
    weightSample.wave = sphereSample.wave;
    weightSample.value = sphereSample.quadTree->getMagnitude();
    sphereSample.quadTree->setMagnitude(1);
  }

  // Optimize wavelengths?
  if (rawData.waveOptimizeThreshold > 0) {
    if (spheres.size() < 2) return;
    Pcg32 prng;
    SphereHEI cells(std::max<unsigned>(maxDepth - 1, 3));
    int cellCount = cells.getCellCount();
    int failCount = 0;
    while (failCount++ < 16 and spheres.size() > 2) {
      int index = prng(spheres.size() - 2) + 1;
      float wave = spheres[index].wave;
      float prevWave = spheres[index - 1].wave;
      float nextWave = spheres[index + 1].wave;
      const SphereQuadTree& quadTree = *spheres[index].quadTree;
      const SphereQuadTree& prevQuadTree = *spheres[index - 1].quadTree;
      const SphereQuadTree& nextQuadTree = *spheres[index + 1].quadTree;
      double fac = findLerpFactor(wave, prevWave, nextWave);
      double mse = 0;
      int mseDenom = 0;
      for (int cellIndex = 0; cellIndex < cellCount; cellIndex++) {
        Vectorf vec = cells.indexToCenterVector(cellIndex);
        if (
          (vec[2] > 0 and (flags & UPPER_HEMISPHERE_BIT)) or
          (vec[2] < 0 and (flags & LOWER_HEMISPHERE_BIT))) {
          double value = quadTree.getPdf(vec);
          double prevValue = prevQuadTree.getPdf(vec);
          double nextValue = nextQuadTree.getPdf(vec);
          double diff = lerp(fac, prevValue, nextValue) - value;
          mse += diff * diff;
          mseDenom++;
        }
      }
      mse /= mseDenom;
      if (std::sqrt(mse) < rawData.waveOptimizeThreshold) {
        delete spheres[index].quadTree;
        spheres[index].quadTree = nullptr;
        spheres.erase(spheres.begin() + index);
        failCount = 0;
      }
    }
    if (rawData.verboseOutput)
      DIRSIG_INFO(
        "{} threshold optimized {} => {} wavelengths",
        rawData.waveOptimizeThreshold, waves.size(), spheres.size());
  }
}

void SpectralSphereQuadTree::clear() {
  for (Sphere& sphere : spheres) {
    delete sphere.quadTree;
    sphere.quadTree = nullptr;
  }
  spheres.clear();
  weight.clear();
  cache = {};
}

void SpectralSphereQuadTree::read(std::istream& in) {
  clear();
  BinaryStream binaryIn(&in);
  spheres.resize(binaryIn.read<uint32_t>());
  for (Sphere& sphere : spheres) {
    sphere.wave = binaryIn.read<float>();
    sphere.quadTree = new SphereQuadTree(in);
  }
  weight.read(in);
}

void SpectralSphereQuadTree::write(std::ostream& out) const {
  BinaryStream binaryOut(&out);
  binaryOut.write(uint32_t(spheres.size()));
  for (const Sphere& sphere : spheres) {
    binaryOut.write(sphere.wave);
    sphere.quadTree->write(out);
  }
  weight.write(out);
}

void SpectralSphereQuadTree::updateState(const State* state) {
  const auto& spectralState = state->currentSpectralState();
  if (
    cache.spectralStateIndex != state->spectralStateIndex or
    cache.spectralState != &spectralState) {
    const auto& waves = spectralState.wavelengthSamples;
    cache.weight = weight.getValue(waves);
    cache.upperMagnitude.resize(waves.size());
    cache.lowerMagnitude.resize(waves.size());
    for (size_t index = 0; index < waves.size(); index++) {
      cache.upperMagnitude[index] = getMagnitude(waves[index], true, false);
      cache.lowerMagnitude[index] = getMagnitude(waves[index], false, true);
    }
    cache.spectralStateIndex = state->spectralStateIndex;
    cache.spectralState = &spectralState;
    auto [fac, pdf0, pdf1] =
      findSpheres(cache.spectralState->referenceWavelength());
    cache.pdfFac = fac;
    cache.pdfQuadTree0 = pdf0->quadTree;
    cache.pdfQuadTree1 = pdf1->quadTree;
    if (cache.pdfFac < 0.0001f) cache.pdfQuadTree1 = cache.pdfQuadTree0;
    if (cache.pdfFac > 0.9999f) cache.pdfQuadTree0 = cache.pdfQuadTree1;
  }
}

float SpectralSphereQuadTree::getMagnitude(
  float wave, bool includeUpper, bool includeLower) const {
  float magnitude = 0;
  if (spheres.size() >= 1) {
    auto [fac, sphere0, sphere1] = findSpheres(wave);
    if (includeUpper)
      magnitude += lerp(
        fac,                                     //
        sphere0->quadTree->getUpperMagnitude(),  //
        sphere1->quadTree->getUpperMagnitude());
    if (includeLower)
      magnitude += lerp(
        fac,                                     //
        sphere0->quadTree->getLowerMagnitude(),  //
        sphere1->quadTree->getLowerMagnitude());
  }
  return weight.getValue(wave) * magnitude;
}

float SpectralSphereQuadTree::getPdf(
  Vectorf vec, bool allowUpper, bool allowLower) const {
  if (cache.pdfQuadTree0 == cache.pdfQuadTree1)
    return cache.pdfQuadTree0->getPdf(vec, allowUpper, allowLower);
  else
    return lerp(
      cache.pdfFac,                                             //
      cache.pdfQuadTree0->getPdf(vec, allowUpper, allowLower),  //
      cache.pdfQuadTree1->getPdf(vec, allowUpper, allowLower));
}

Vectorf SpectralSphereQuadTree::samplePdf(
  RandEngine& prng, bool allowUpper, bool allowLower) const {
  if (cache.pdfQuadTree0 == cache.pdfQuadTree1)
    return cache.pdfQuadTree0->samplePdf(prng, allowUpper, allowLower);
  else
    return float(canonical(prng)) < 1 - cache.pdfFac
             ? cache.pdfQuadTree0->samplePdf(prng, allowUpper, allowLower)
             : cache.pdfQuadTree1->samplePdf(prng, allowUpper, allowLower);
}

float SpectralSphereQuadTree::interpolateSphere(Vectorf vec, float wave) const {
  if (spheres.size() == 0) return 0.0f;
  if (spheres.size() == 1) return spheres[0].quadTree->getValue(vec);
  auto [fac, sphere0, sphere1] = findSpheres(wave);
  if (fac < 0.0001f) return sphere0->quadTree->getValue(vec);
  if (fac > 0.9999f) return sphere1->quadTree->getValue(vec);
  return lerp(
    fac,                               //
    sphere0->quadTree->getValue(vec),  //
    sphere1->quadTree->getValue(vec));
}

void SpectralSphereQuadTree::interpolateSphere(
  Vectorf vec, span<const double> waves, SpectralVector& values) const {
  values.resize(waves.size());
  values.setData(0);
  if (spheres.size() == 1) {
    values.setData(spheres[0].quadTree->getPdf(vec));
  } else if (spheres.size() >= 2) {
    const Sphere* prevSphere0 = nullptr;
    const Sphere* prevSphere1 = nullptr;
    float prevWave0 = 0;
    float prevWave1 = 0;
    float prevValue0 = 0;
    float prevValue1 = 0;
    for (size_t index = 0; index < size_t(values.size()); index++) {
      float wave = waves[index];
      if (index > 0) {
        if (
          (prevWave0 <= wave and prevWave1 >= wave) or
          (prevWave0 > wave and prevSphere0 == &spheres.front()) or
          (prevWave1 < wave and prevSphere1 == &spheres.back())) {
          values[index] = lerp(
            findLerpFactorClamped(wave, prevWave0, prevWave1), prevValue0,
            prevValue1);
          continue;
        }
      }
      auto [fac, sphere0, sphere1] = findSpheres(wave);
      float value0 = prevValue1;
      float value1 = prevValue0;
      if (sphere0 != prevSphere1) value0 = sphere0->quadTree->getPdf(vec);
      if (sphere1 != prevSphere0) value1 = sphere1->quadTree->getPdf(vec);
      values[index] = lerp(fac, value0, value1);
      prevSphere0 = sphere0, prevWave0 = prevSphere0->wave, prevValue0 = value0;
      prevSphere1 = sphere1, prevWave1 = prevSphere1->wave, prevValue1 = value1;
    }
  }
}

SpectralSphereQuadTree::FindSpheresResult SpectralSphereQuadTree::findSpheres(
  float wave) const {
  DIRSIG_THROW_IF(spheres.empty(), "No samples!");
  if (spheres.size() == 1) return {0.0f, &spheres[0], &spheres[0]};
  auto [sphere0, sphere1] = findLerpIterators(
    spheres.begin(), spheres.end(), wave,
    [](const Sphere& sphere, float wave) { return sphere.wave < wave; });
  return {
    findLerpFactorClamped(wave, sphere0->wave, sphere1->wave), &*sphere0,
    &*sphere1};
}

}  // namespace d5::sqt
