#include "dirsig5/sqt/SphereHEI.h"

#include <set>

#include "dirsig5/base/Constants.h"

namespace d5::sqt {

void SphereHEI::indexToCenterVector(size_t index, Vectorf& vec) const {
  DIRSIG_THROW_IF(index >= cellCount, "Out of bounds index!");

  // figure out which region we're in
  unsigned region = index / (1 << 2 * order);
  int f = index - (region * (1 << 2 * order));

  // fractions encoded in binary fraction, Equation 15
  float curSplit = 0.5;
  float f1 = 0;
  float f2 = 0;
  unsigned bits = 2 * order;
  unsigned mask = 2 << (bits - 2);
  if (order > 0) {
    do {
      if (f & mask) f2 += curSplit;
      if (f & (mask >> 1)) f1 += curSplit;
      mask = mask >> 2;
      curSplit /= 2.0;
    } while (mask != 0);
  }

  // get to the centers
  f1 += curSplit;
  f2 += curSplit;
  fractionsToVector(region, f1, f2, vec);
}

void SphereHEI::indexToCellList(size_t index, std::ostream& out) const {
  std::vector<Vectorf> vecs;
  vecs.reserve(4);
  indexToCell(index, vecs);

  out << vecs[0].transpose() << '\n';
  out << vecs[1].transpose() << '\n';
  out << vecs[2].transpose() << '\n';
  out << vecs[3].transpose() << '\n';
  out << vecs[0].transpose() << '\n';
  out << "\n\n";
}

void SphereHEI::indexToCell(size_t index, std::vector<Vectorf>& vecs) const {
  DIRSIG_THROW_IF(index >= cellCount, "Out of bounds index!");

  // figure out which region we're in
  unsigned region = index / (1 << 2 * order);
  int f = index - (region * (1 << 2 * order));
  float f1, f2, curSplit;
  decodeFractions(f, f1, f2, curSplit);

  float hw = curSplit;

  Vectorf vec;
  fractionsToVector(region, f1 - hw, f2 - hw, vec);
  vecs.emplace_back(vec);
  fractionsToVector(region, f1 + hw, f2 - hw, vec);
  vecs.emplace_back(vec);
  fractionsToVector(region, f1 + hw, f2 + hw, vec);
  vecs.emplace_back(vec);
  fractionsToVector(region, f1 - hw, f2 + hw, vec);
  vecs.emplace_back(vec);
}

void SphereHEI::indexToSampleVector(
  size_t index, Vectorf& vec, RandEngine& randEngine) const {
  // make sure we have a valid index
  DIRSIG_THROW_IF(index >= cellCount, "Out of bounds index!");

  // figure out which region we're in
  unsigned regCount = (1 << 2 * order);
  unsigned region = index / regCount;
  int f = index - (region * regCount);
  float f1, f2, curSplit;
  decodeFractions(f, f1, f2, curSplit);

  float hw = curSplit;
  float xi1 = canonical(randEngine);
  float xi2 = canonical(randEngine);

  // compute blinear interp coords
  float A1 = (1.f - xi1) * (1.f - xi2);
  float A2 = xi1 * (1.f - xi2);
  float A3 = xi1 * xi2;
  float A4 = (1.f - xi1) * xi2;

  /**
   * \todo pursue increased interp of angles rather than expensive vectors
   */
  if (region < 4 || region > 7)  // exclude pure equatorial regions
  {
    Vectorf v1, v2, v3, v4;
    fractionsToVector(region, f1 - hw, f2 - hw, v1);
    fractionsToVector(region, f1 + hw, f2 - hw, v2);
    fractionsToVector(region, f1 + hw, f2 + hw, v3);
    fractionsToVector(region, f1 - hw, f2 + hw, v4);

    vec = A1 * v1 + A2 * v2 + A3 * v3 + A4 * v4;
  } else {
    float th1, th2, th3, th4, ph1, ph2, ph3, ph4;
    fractionsToAngles(region, f1 - hw, f2 - hw, th1, ph1);
    fractionsToAngles(region, f1 + hw, f2 - hw, th2, ph2);
    fractionsToAngles(region, f1 + hw, f2 + hw, th3, ph3);
    fractionsToAngles(region, f1 - hw, f2 + hw, th4, ph4);

    anglesToVector(
      A1 * th1 + A2 * th2 + A3 * th3 + A4 * th4,
      A1 * ph1 + A2 * ph2 + A3 * ph3 + A4 * ph4, vec);
  }

  // make sure we're normalized
  vec.normalize();
}

int SphereHEI::getRegionNeighbor(unsigned ri, int d1, int d2) const {
  // create a list of all possible mappings
  static thread_local std::vector<int> rnis(0);
  if (rnis.size() == 0) {
    rnis.resize(96);
    for (int rip = 0; rip < 12; rip++) {
      for (int d1p = -1; d1p <= 1; d1p++)
        for (int d2p = -1; d2p <= 1; d2p++) {
          int dip = 3 * (1 + d1p) + (1 + d2p);
          int iip = 8 * rip + dip;
          int index = -1;
          // top ring
          if (rip <= 3) {
            if ((d1p == 1) && (d2p == 0))
              index = (rip + 1) % 4;
            else if ((d1p == 0) && (d2p == 1))
              index = (rip + 3) % 4;
            else if ((d1p == 1) && (d2p == 1))
              index = (rip + 2) % 4;
            else if ((d1p == -1) && (d2p == 0))
              index = (rip + 4);
            else if ((d1p == 0) && (d2p == -1))
              index = 4 + ((rip + 1) % 4);
            else if ((d1p == -1) && (d2p == -1))
              index = rip + 8;
          } else if (rip >= 8) {
            if ((d1p == 1) && (d2p == 0))
              index = 4 + ((rip + 1) % 4);
            else if ((d1p == 0) && (d2p == 1))
              index = rip - 4;
            else if ((d1p == -1) && (d2p == 0))
              index = 8 + ((rip + 3) % 4);
            else if ((d1p == 0) && (d2p == -1))
              index = 8 + ((rip + 1) % 4);
            else if ((d1p == -1) && (d2p == -1))
              index = 8 + ((rip + 2) % 4);
            else if ((d1p == 1) && (d2p == 1))
              index = rip - 8;
          } else {
            if ((d1p == 1) && (d2p == 0))
              index = rip - 4;
            else if ((d1p == 0) && (d2p == 1))
              index = (rip + 3) % 4;
            else if ((d1p == -1) && (d2p == 0))
              index = 8 + ((rip + 3) % 4);
            else if ((d1p == 0) && (d2p == -1))
              index = rip + 4;
            else if ((d1p == 1) && (d2p == -1))
              index = 4 + ((rip + 1) % 4);
            else if ((d1p == -1) && (d2p == 1))
              index = 4 + ((rip - 1) % 4);
          }

          if ((d1p == 0 || d2p == 0) && d1p != d2p) rnis[iip] = index;
        }
    }
  }

  int di = 3 * (1 + d1) + (1 + d2);
  int ii = 8 * ri + di;
  return rnis[ii];
}

void SphereHEI::getCorners(const Vectorf& v, std::array<RFF, 9>& ns) const {
  unsigned ri;
  float f1, f2;

  float curSplit = getSplit();

  vectorToFractions(v, ri, f1, f2);

  float f1o = (floor(f1 / curSplit) + 0.5) * curSplit;
  float f2o = (floor(f2 / curSplit) + 0.5) * curSplit;

  ns[0] = {ri, f1o, f2o};

  float d1[8] = {1, 1, 0, -1, -1, -1, 0, 1};
  float d2[8] = {0, 1, 1, 1, 0, -1, -1, -1};

  for (size_t id = 0; id < 8; id++) {
    float f1m = f1 + d1[id] * curSplit;
    float f2m = f2 + d2[id] * curSplit;
    size_t in = id + 1;
    if (d1[id] == 0 || d2[id] == 0) {
      if (d1[id] != 0) {
        if (f1m > 1.f) {
          unsigned ri2 = getRegionNeighbor(ri, 1, 0);
          f1m -= 1.f;
          if (ri <= 3)
            ns[in] = {ri2, f2, 1.f - f1m};
          else
            ns[in] = {ri2, f1m, f2};
        } else if (f1m < 0.f) {
          unsigned ri2 = getRegionNeighbor(ri, -1, 0);
          f1m += 1.f;
          if (ri >= 8)
            ns[in] = {ri2, f2, 1.f - f1m};
          else
            ns[in] = {ri2, f1m, f2};
        } else {
          ns[in] = {ri, f1m, f2};
        }
      }

      if (d2[id] != 0) {
        if (f2m > 1.f) {
          unsigned ri2 = getRegionNeighbor(ri, 0, 1);
          f2m -= 1.f;
          if (ri <= 3)
            ns[in] = {ri2, 1.f - f2m, f1};
          else
            ns[in] = {ri2, f1, f2m};
        } else if (f2m < 0.f) {
          unsigned ri2 = getRegionNeighbor(ri, 0, -1);
          f2m += 1.f;
          if (ri >= 8)
            ns[in] = {ri2, 1.f - f2m, f1};
          else
            ns[in] = {ri2, f1, f2m};
        } else {
          ns[in] = {ri, f1, f2m};
        }
      }
    } else {
      ns[in] = {unsigned(-1), 0, 0};
    }
  }
}

void SphereHEI::writeOBJFacet(
  unsigned region,
  size_t index,
  size_t& vIndex,
  std::ostream& out,
  float mag) const {
  float f1, f2, curSplit;
  Vectorf v1, v2, v3, v4;

  // subdivide
  decodeFractions(index, f1, f2, curSplit);
  float hw = curSplit;

  fractionsToVector(region, f1 - hw, f2 - hw, v1);
  fractionsToVector(region, f1 + hw, f2 - hw, v2);
  fractionsToVector(region, f1 + hw, f2 + hw, v3);
  fractionsToVector(region, f1 - hw, f2 + hw, v4);

  out << "v " << v1.transpose() << ' ' << mag << std::endl;
  out << "v " << v2.transpose() << ' ' << mag << std::endl;
  out << "v " << v3.transpose() << ' ' << mag << std::endl;
  out << "v " << v4.transpose() << ' ' << mag << std::endl;
  out << "f " << vIndex << ' ' << vIndex + 1 << ' ' << vIndex + 2 << ' '
      << vIndex + 3 << std::endl
      << std::endl;

  vIndex += 4;
}

void SphereHEI::writeTXTFacet(
  unsigned region, size_t index, std::ostream& out, float stepCount) const {
  float f1, f2, curSplit;
  Vectorf v;

  // subdivide
  decodeFractions(index, f1, f2, curSplit);
  float hw = 0.99 * curSplit;
  float step = 2 * hw / (stepCount - 1);

  for (size_t ii = 0; ii < stepCount; ii++) {
    fractionsToVector(region, f1 - hw + ii * step, f2 - hw, v);
    out << v.transpose() << std::endl;
  }
  for (size_t ii = 0; ii < stepCount; ii++) {
    fractionsToVector(region, f1 + hw, f2 - hw + ii * step, v);
    out << v.transpose() << std::endl;
  }
  for (size_t ii = 0; ii < stepCount; ii++) {
    fractionsToVector(region, f1 + hw - ii * step, f2 + hw, v);
    out << v.transpose() << std::endl;
  }
  for (size_t ii = 0; ii < stepCount; ii++) {
    fractionsToVector(region, f1 - hw, f2 + hw - ii * step, v);
    out << v.transpose() << std::endl;
  }

  fractionsToVector(region, f1 - hw, f2 - hw, v);
  out << v.transpose() << std::endl;
  out << std::endl << std::endl;
}

void SphereHEI::writeSVGFacet(
  unsigned region,
  size_t index,
  unsigned width,
  std::ostream& out,
  float val,
  bool bw) const {
  const float stroke = 0.5f;

  const std::array<float, 11> R = {
    {0.192157, 0.270588, 0.454902, 0.670588, 0.878431, 1, 0.996078, 0.992157,
     0.956863, 0.843137, 0.647059}};
  const std::array<float, 11> G = {
    {0.211765, 0.458824, 0.678431, 0.85098, 0.952941, 1, 0.878431, 0.682353,
     0.427451, 0.188235, 0}};
  const std::array<float, 11> B = {
    {0.584314, 0.705882, 0.819608, 0.913725, 0.972549, 0.74902, 0.564706,
     0.380392, 0.262745, 0.152941, 0.14902}};

  float r, g, b;
  if (val < 0) {
    r = 0;
    g = 0;
    b = 0;
  } else {
    if (bw) {
      // grayscale
      float f = 1 - val;
      if (f < 0) f = 0;

      r = f * 255;
      g = r;
      b = r;

    } else {
      // map the color, interp between entries
      unsigned idx1 = std::floor(val * 10);
      unsigned idx2 = idx1 + 1;

      if (idx1 > 10) {
        r = R[10];
        g = G[10];
        b = B[10];
      } else {
        float fraction = val * 10 - idx1;
        r = (1.f - fraction) * R[idx1] + (fraction)*R[idx2];
        g = (1.f - fraction) * G[idx1] + (fraction)*G[idx2];
        b = (1.f - fraction) * B[idx1] + (fraction)*B[idx2];
      }
    }
  }

  float f1, f2, curSplit;
  float th1, th2, th3, th4;
  float ph1, ph2, ph3, ph4;

  // subdivide
  decodeFractions(index, f1, f2, curSplit);
  float hw = curSplit;

  fractionsToAngles(region, f1 - hw, f2 - hw, th1, ph1);
  fractionsToAngles(region, f1 + hw, f2 - hw, th2, ph2);
  fractionsToAngles(region, f1 + hw, f2 + hw, th3, ph3);
  fractionsToAngles(region, f1 - hw, f2 + hw, th4, ph4);

  float offset = 0;
  if (0.25 * (th1 + th2 + th3 + th4) > PI_2) {
    offset = width;
    th1 = PI - th1;
    th2 = PI - th2;
    th3 = PI - th3;
    th4 = PI - th4;
  }

  float r1 = width * th1 / PI, d12 = r1 * r1;
  float r2 = width * th2 / PI, d22 = r2 * r2;
  float r3 = width * th3 / PI, d32 = r3 * r3;
  float r4 = width * th4 / PI, d42 = r4 * r4;
  float x1 = r1 * std::cos(ph1), y1 = r1 * std::sin(ph1);
  float x2 = r2 * std::cos(ph2), y2 = r2 * std::sin(ph2);
  float x3 = r3 * std::cos(ph3), y3 = r3 * std::sin(ph3);
  float x4 = r4 * std::cos(ph4), y4 = r4 * std::sin(ph4);
  float d2 = 0.25 * width * width;
  bool onEdge = false;
  if (d12 > d2) {
    float d = std::sqrt(d12);
    x1 *= 0.5 * width / d;
    y1 *= 0.5 * width / d;
    onEdge = true;
  }
  if (d22 > d2) {
    float d = std::sqrt(d22);
    x2 *= 0.5 * width / d;
    y2 *= 0.5 * width / d;
    onEdge = true;
  }
  if (d32 > d2) {
    float d = std::sqrt(d32);
    x3 *= 0.5 * width / d;
    y3 *= 0.5 * width / d;
    onEdge = true;
  }
  if (d42 > d2) {
    float d = std::sqrt(d42);
    x4 *= 0.5 * width / d;
    y4 *= 0.5 * width / d;
    onEdge = true;
  }
  x1 += 0.5f * width + offset, y1 += 0.5f * width;
  x2 += 0.5f * width + offset, y2 += 0.5f * width;
  x3 += 0.5f * width + offset, y3 += 0.5f * width;
  x4 += 0.5f * width + offset, y4 += 0.5f * width;
  out << "    <polygon points=\""  //
      << x1 << ',' << y1 << ' '    //
      << x2 << ',' << y2 << ' '    //
      << x3 << ',' << y3 << ' '    //
      << x4 << ',' << y4 << '\"'   //
      << " fill=\"rgb("            //
      << int(255 * r) << ','       //
      << int(255 * g) << ','       //
      << int(255 * b) << ")\""
      << " stroke-width=\"" << stroke << "\" stroke=\"rgb(0,0,0)\" />"
      << std::endl;

  /**
   * \todo this is horribly redundant code
   */
  if (onEdge) {
    offset = 0;
    th1 = PI - th1;
    th2 = PI - th2;
    th3 = PI - th3;
    th4 = PI - th4;
    r1 = width * th1 / PI, d12 = r1 * r1;
    r2 = width * th2 / PI, d22 = r2 * r2;
    r3 = width * th3 / PI, d32 = r3 * r3;
    r4 = width * th4 / PI, d42 = r4 * r4;
    x1 = r1 * std::cos(ph1), y1 = r1 * std::sin(ph1);
    x2 = r2 * std::cos(ph2), y2 = r2 * std::sin(ph2);
    x3 = r3 * std::cos(ph3), y3 = r3 * std::sin(ph3);
    x4 = r4 * std::cos(ph4), y4 = r4 * std::sin(ph4);
    if (d12 > d2) {
      float d = std::sqrt(d12);
      x1 *= 0.5 * width / d;
      y1 *= 0.5 * width / d;
    }
    if (d22 > d2) {
      float d = std::sqrt(d22);
      x2 *= 0.5 * width / d;
      y2 *= 0.5 * width / d;
    }
    if (d32 > d2) {
      float d = std::sqrt(d32);
      x3 *= 0.5 * width / d;
      y3 *= 0.5 * width / d;
    }
    if (d42 > d2) {
      float d = std::sqrt(d42);
      x4 *= 0.5 * width / d;
      y4 *= 0.5 * width / d;
    }
    x1 += 0.5f * width + offset, y1 += 0.5f * width;
    x2 += 0.5f * width + offset, y2 += 0.5f * width;
    x3 += 0.5f * width + offset, y3 += 0.5f * width;
    x4 += 0.5f * width + offset, y4 += 0.5f * width;
    out << "    <polygon points=\""  //
        << x1 << ',' << y1 << ' '    //
        << x2 << ',' << y2 << ' '    //
        << x3 << ',' << y3 << ' '    //
        << x4 << ',' << y4 << '\"'   //
        << " fill=\"rgb("            //
        << int(255 * r) << ','       //
        << int(255 * g) << ','       //
        << int(255 * b) << ")\""
        << " stroke-width=\"" << stroke << "\" stroke=\"rgb(0,0,0)\" />"
        << std::endl;
  }
}

void SphereHEI::writeGridTexture(std::vector<std::vector<float>>& img) const {
  float th, ph;
  size_t w = img.size();
  size_t h = img[0].size();

  float maxV = 0;

  // set to zeros
  for (size_t ii = 0; ii < img.size(); ii++)
    std::fill(img[ii].begin(), img[ii].end(), 0.f);

  float steps = 2000;

  // walk each cell and write the grid
  for (size_t ii = 0; ii < cellCount; ii++) {
    // figure out which region we're in
    unsigned region = ii / (1 << 2 * order);
    int index = ii - (region * (1 << 2 * order));

    float f1, f2, curSplit;

    // subdivide
    decodeFractions(index, f1, f2, curSplit);
    float hw = curSplit;

    std::vector<std::pair<float, float>> corners(4);
    corners[0] = {f1 - hw, f2 - hw};
    corners[1] = {f1 + hw, f2 - hw};
    corners[2] = {f1 + hw, f2 + hw};
    corners[3] = {f1 - hw, f2 + hw};

    for (size_t jj = 0; jj < 4; jj++) {
      size_t kk = jj + 1;
      if (kk == 4) kk = 0;
      for (float t = 0; t <= 1.f; t += 1.f / steps) {
        float fa = (1.f - t) * corners[jj].first + t * corners[kk].first;
        float fb = (1.f - t) * corners[jj].second + t * corners[kk].second;

        fractionsToAngles(region, fa, fb, th, ph);

        if (ph > TwoPI) ph -= TwoPI;
        if (ph < 0) ph += TwoPI;

        // convert to indexes
        size_t ui = std::floor((ph / TwoPI) * w);
        size_t vi = std::floor(((th) / PI) * h);
        if (ui >= w) ui = w - 1;
        if (vi >= h) vi = h - 1;

        img[ui][vi]++;

        if (img[ui][vi] > maxV) maxV = img[ui][vi];
      }
    }
  }

  for (size_t jj = 0; jj < h; jj++) {
    for (size_t ii = 0; ii < w; ii++) {
      img[ii][jj] /= maxV;
      if (img[ii][jj] > 1e-6)
        img[ii][jj] = 1;
      else
        img[ii][jj] *= 1e6;
    }
  }
}

void SphereHEI::fractionsToAngles(
  unsigned region, float f1, float f2, float& theta, float& phi) const {
  // Equation 16
  float x = f1 - f2;
  float y = f1 + f2;

  float sinXi = 0;

  // equatorial regions are numbered [4:7]
  if (region >= 4 && region <= 7) {
    // Equation 19
    float phiOffset = (region - 4) * PI_2;
    phi = phiOffset + x * PI / 4;
    sinXi = (2 / 3.) * (y - 1.0);
  } else {
    float phiOffset = 0;

    // we can still be in the equatorial region
    if (region < 4) {
      phiOffset = region * PI_2;

      if (y <= 1) {
        // Equation 21
        phi = phiOffset + (1 + x) * PI / 4.0;
        sinXi = 2. * y / 3.;
      } else {
        // Equations 28 and 34
        phi = (y != 2) ? phiOffset + (2 - y + x) * PI / (4. * (2 - y)) : 0;
        sinXi = 1.0 - (1 / 3.) * (2. - y) * (2. - y);
      }
    } else {
      phiOffset = (region - 8) * PI_2;

      if (y >= 1) {
        // Equation 21 (modified)
        phi = phiOffset + (1 + x) * PI / 4.0;
        sinXi = 2. * (y - 2) / 3.;
      }

      else {
        // Equation 34 (modified)
        phi = (y != 0) ? phiOffset + (y + x) * PI / (4. * (y)) : 0;
        sinXi = -1.0 + (1 / 3.) * (y) * (y);
      }
    }
  }

  if (phi < 0) phi += TwoPI;
  theta = PI_2 - asinf(sinXi);
}

void SphereHEI::vectorToFractions(
  const Vectorf& vec, unsigned& region, float& f1, float& f2) const {
  double theta = std::acos(vec(2));
  double phi = std::atan2(vec(1), vec(0));
  while (phi < 0) phi += TwoPI;
  if (phi == 0) phi = 1e-9;

  double sinXi = std::sin(PI_2 - theta);
  double x = 0;
  double y = 0;

  if (sinXi > 2.0 / 3.0) {
    // in the top polar region
    region = std::floor(phi / PI_2);
    phi -= region * PI_2;
    if (phi == 0) {
      x = 0, y = 2;
    } else {
      y = 2 - std::sqrt(3 * (1 - sinXi));
      x = (2 - y) * (phi / PI_4 - 1);
    }
    f2 = 0.5 * (y - x);
    f1 = x + f2;
  } else if (sinXi < -2.0 / 3.0) {
    // in the bottom polar region
    region = std::floor(phi / PI_2);
    phi -= region * PI_2;
    region += 8;
    if (phi == 0) {
      x = 0;
      y = 0;
    } else {
      y = std::sqrt(3 * (sinXi + 1));
      x = y * (phi / PI_4 - 1);
    }

    f2 = 0.5 * (y - x);
    f1 = x + f2;
  } else {
    // in the equatorial region, but two possibilities
    double phi0 = phi;
    region = std::floor((phi + PI_4) / PI_2);  // intermediate region
    phi -= region * PI_2;
    y = 1.5 * sinXi + 1;
    x = phi / PI_4;
    f2 = 0.5 * (y - x);
    f1 = x + f2;

    // if its valid, we guessed right
    if (f1 >= 0 && f1 <= 1 && f2 >= 0 && f2 <= 1) {
      if (region < 4) region += 4;
    } else {
      region = std::floor(phi0 / PI_2);
      phi = phi0 - region * PI_2;
      // must be in one of the partial equatorial regions
      if (y < 1) {
        x = phi / PI_4 - 1;
        y = 1.5 * sinXi + 2;
        region += 8;
      } else {
        x = phi / PI_4 - 1;
        y = 1.5 * sinXi;
      }
      f2 = 0.5 * (y - x);
      f1 = x + f2;
    }
  }

  // check for possible precision issues
  if (region > 11) region = 11;
}

}  // namespace d5::sqt
