#pragma once

#include <array>
#include <iostream>
#include <random>

#include <embree3/rtcore.h>
#include <embree3/rtcore_ray.h>

#include "delaunator.h"
#include "dirsig5/base/Error.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/span.h"
#include "dirsig5/sqt/SphereQuadTree.h"

namespace d5::sqt {

class HemisphereInterpolator {
public:
  HemisphereInterpolator(RTCDevice device_, size_t waveCount_)
      : device(device_), scene(rtcNewScene(device_)), waveCount(waveCount_) {}

  HemisphereInterpolator(const HemisphereInterpolator&) = delete;

  ~HemisphereInterpolator() { rtcReleaseScene(scene); }

  /// Add data to the search structure. The vector type must provide
  /// operator[] for indexes [0,1,2]
  void addPoint(Vectorf vec, span<const float> value) {
    DIRSIG_THROW_IF(finalized, "Attempting to add point after finalization!");
    spherePoints.push_back(vec);
    sphereValues.insert(sphereValues.end(), value.begin(), value.end());
    projectedPoints.push_back(double(vec[0]) / (1 + double(vec[2])));
    projectedPoints.push_back(double(vec[1]) / (1 + double(vec[2])));
  }

  /// Finalize the data (no more points can be added and the
  /// search structure is sorted).
  void finalize() try {
    delaunator::Delaunator dU(projectedPoints);
    auto isCompletelyUnder = [&](size_t i) {
      return spherePoints[dU.triangles[i + 0]][2] < -1e-6f and
             spherePoints[dU.triangles[i + 1]][2] < -1e-6f and
             spherePoints[dU.triangles[i + 2]][2] < -1e-6f;
    };
    std::vector<bool> skipPoint(spherePoints.size(), true);
    for (size_t i = 0; i < dU.triangles.size(); i += 3)
      if (not isCompletelyUnder(i)) {
        skipPoint[dU.triangles[i + 0]] = false;
        skipPoint[dU.triangles[i + 1]] = false;
        skipPoint[dU.triangles[i + 2]] = false;
      }

    size_t curIndex = 0;
    std::vector<size_t> remapForward(spherePoints.size(), size_t(-1));
    std::vector<float> vertexValues;
    vertexValues.reserve(sphereValues.size());
    vertexData.reserve(3 * spherePoints.size());
    for (size_t i = 0; i < spherePoints.size(); i++)
      if (not skipPoint[i]) {
        remapForward[i] = curIndex++;
        vertexData.emplace_back(spherePoints[i][0]);
        vertexData.emplace_back(spherePoints[i][1]);
        vertexData.emplace_back(spherePoints[i][2]);
        vertexValues.insert(
            vertexValues.end(),                          //
            sphereValues.begin() + waveCount * (i + 0),  //
            sphereValues.begin() + waveCount * (i + 1));
      }
    triangleData.reserve(dU.triangles.size() / 2);
    for (size_t i = 0; i < dU.triangles.size(); i += 3)
      if (not isCompletelyUnder(i)) {
        triangleData.push_back(remapForward[dU.triangles[i + 0]]);
        triangleData.push_back(remapForward[dU.triangles[i + 1]]);
        triangleData.push_back(remapForward[dU.triangles[i + 2]]);
      }
    RTCGeometry geometry = rtcNewGeometry(device, RTC_GEOMETRY_TYPE_TRIANGLE);
    rtcSetSharedGeometryBuffer(
        geometry,                                 //
        RTC_BUFFER_TYPE_VERTEX, 0,                //
        RTC_FORMAT_FLOAT3, vertexData.data(), 0,  //
        3 * sizeof(float), vertexData.size() / 3);
    rtcSetSharedGeometryBuffer(
        geometry,                                  //
        RTC_BUFFER_TYPE_INDEX, 0,                  //
        RTC_FORMAT_UINT3, triangleData.data(), 0,  //
        3 * sizeof(unsigned), triangleData.size() / 3);
    rtcSetGeometryBuildQuality(geometry, RTC_BUILD_QUALITY_MEDIUM);
    rtcSetGeometryTimeStepCount(geometry, 1);
    rtcCommitGeometry(geometry);
    rtcAttachGeometry(scene, geometry);
    rtcReleaseGeometry(geometry);
    rtcCommitScene(scene);

    sphereValues.swap(vertexValues);
    spherePoints = std::vector<Vectorf>();
    projectedPoints = std::vector<double>();
    finalized = true;
  } catch (const std::runtime_error&) {
    throw std::runtime_error(
        "SQT sphere interpolator Delaunay triangulation step failed!");
  }

  SpectralVector interpolate(Vectorf vec) {
    RTCRayHit rayhit = {};
    auto& ray = rayhit.ray;
    auto& hit = rayhit.hit;
    ray.org_x = ray.org_y = ray.org_z = 0;
    ray.dir_x = vec[0];
    ray.dir_y = vec[1];
    ray.dir_z = vec[2];
    ray.tnear = 0.0f;
    ray.tfar = 1.1f;
    hit.geomID = RTC_INVALID_GEOMETRY_ID;
    RTCIntersectContext context;
    rtcInitIntersectContext(&context);
    rtcIntersect1(scene, &context, &rayhit);
    SpectralVector result;
    result.resize(waveCount);
    result.setData(0);
    if (hit.geomID != RTC_INVALID_GEOMETRY_ID) {
      unsigned index0 = triangleData[hit.primID * 3 + 0];
      unsigned index1 = triangleData[hit.primID * 3 + 0];
      unsigned index2 = triangleData[hit.primID * 3 + 0];
      const float* value0 = &sphereValues[index0 * waveCount];
      const float* value1 = &sphereValues[index1 * waveCount];
      const float* value2 = &sphereValues[index2 * waveCount];
      for (size_t j = 0; j < waveCount; j++)
        result[j] = hit.u * value0[j] + hit.v * value1[j] +
                    (1 - hit.u - hit.v) * value2[j];
    }
    return result;
  }

private:
  /// If true, no more data can be added and the data can be searched.
  bool finalized = false;

  /// The embree device.
  RTCDevice device = {};

  /// The embree scene.
  RTCScene scene = {};

  /// The number of wavelengths.
  size_t waveCount = 0;

  /// The sphere points.
  std::vector<Vectorf> spherePoints;

  /// The sphere values.
  std::vector<float> sphereValues;

  /// The projected points to feed to delaunator.
  std::vector<double> projectedPoints;

  /// Geometry vertex data.
  std::vector<float> vertexData;

  /// Geometry triangle data.
  std::vector<unsigned> triangleData;
};

class SphereInterpolator {
public:
  SphereInterpolator(RTCDevice device_, size_t waveCount_, Flags flags_)
      : device(device_), waveCount(waveCount_), flags(flags_) {}

  SphereInterpolator(const SphereInterpolator&) = delete;

  ~SphereInterpolator() {
    delete upperHemisphere;
    delete lowerHemisphere;
    upperHemisphere = nullptr;
    lowerHemisphere = nullptr;
  }

  void addPoint(Vectorf vec, span<const float> value) {
    if (vec[2] > 0 and (flags & UPPER_HEMISPHERE_BIT) != 0) {
      if (upperHemisphere == nullptr)
        upperHemisphere = new HemisphereInterpolator(device, waveCount);
      upperHemisphere->addPoint(vec, value);
      if (vec[2] < 0.5f)  // Cover the equator!
        upperHemisphere->addPoint({vec[0], vec[1], -vec[2]}, value);
    }
    if (vec[2] < 0 and (flags & LOWER_HEMISPHERE_BIT) != 0) {
      vec[2] *= -1.0f;  // Reflect
      if (lowerHemisphere == nullptr)
        lowerHemisphere = new HemisphereInterpolator(device, waveCount);
      lowerHemisphere->addPoint(vec, value);
      if (vec[2] < 0.5f)  // Cover the equator!
        lowerHemisphere->addPoint({vec[0], vec[1], -vec[2]}, value);
    }
  }

  void finalize() {
    if (upperHemisphere) upperHemisphere->finalize();
    if (lowerHemisphere) lowerHemisphere->finalize();
  }

  SpectralVector interpolate(Vectorf vec) {
    if (vec[2] > 0 and upperHemisphere)
      return upperHemisphere->interpolate(vec);
    if (vec[2] < 0 and lowerHemisphere)
      return lowerHemisphere->interpolate({vec[0], vec[1], -vec[2]});
    SpectralVector result;
    result.resize(waveCount);
    result.setData(0);
    return result;
  }

private:
  RTCDevice device = {};
  size_t waveCount = 0;
  Flags flags = 0;
  HemisphereInterpolator* upperHemisphere = nullptr;
  HemisphereInterpolator* lowerHemisphere = nullptr;
};

}  // namespace d5::sqt
