#include "dirsig5/sqt/SphereNode.h"

#include "dirsig5/sqt/SphereQuadTree.h"

namespace d5::sqt {

SphereNode::~SphereNode() { deleteChildren(); }

void SphereNode::expand(unsigned depth) {
  if (depth > 0) {
    deleteChildren();
    children = new SphereNode[4];
    for (SphereNode* child = children; child < children + 4; child++)
      child->expand(depth - 1);
  }
}

void SphereNode::accumulate(unsigned curDepth) {
  // Leaf?
  if (children == nullptr) {
    if (curDepth == 0) value = 0;
    return;  // Do nothing
  }
  // Accumulate.
  value = 0;
  size_t used = 0;
  for (SphereNode* child = children; child < children + 4; child++) {
    child->accumulate(curDepth + 1);
    if (child->value >= 0) {
      value += child->value;
      used++;
    }
  }
  if (used == 0) {  // Didn't use any children?
    deleteChildren();
    value = curDepth == 0 ? 0 : -1;
    return;
  }
  if (used < 4) {  // Didn't use all children?
    value *= 4.0f / used;
    if (curDepth > 1) {
      deleteChildren();
    } else {
      for (SphereNode* child = children; child < children + 4; child++)
        child->value = value * 0.25f;
    }
  }
}

void SphereNode::averageSamples() {
  if (value > 0 and count > 1) {
    value /= count;
    count = 1;
  }
  if (children)
    for (SphereNode* child = children; child < children + 4; child++)
      child->averageSamples();
}

void SphereNode::scale(float k) {
  if (value >= 0) value *= k;
  if (children)
    for (SphereNode* child = children; child < children + 4; child++)
      child->scale(k);
}

void SphereNode::sample(float xi, unsigned& order, size_t& index) const {
  if (children == nullptr) return;

  index *= 4;
  for (const SphereNode* child = children; child < children + 4; child++) {
    if (xi < child->value) {
      order++;
      child->sample(xi, order, index);
      break;
    }
    if (child->value > 0) xi -= child->value;
    index++;
  }
}

void SphereNode::compress(float threshold, size_t curDepth, size_t index) {
  if (children == nullptr) return;

  // compress each child node then check the results
  index *= 4;
  for (SphereNode* child = children; child < children + 4; child++)
    child->compress(threshold, curDepth + 1, index++);

  // set a threshold for all being the same
  bool shouldCompress = true;
  if (value > 0) {
    // check each child
    for (SphereNode* child = children; child < children + 4; child++) {
      // make sure the child is defined
      // NOTE This is logically inverted in order to detect NaNs.
      DIRSIG_THROW_IF(
        not(child->value >= 0), "A sphere quad-tree with undefined "
                                "values can't be compressed!");

      if (std::abs(child->value - value * 0.25f) > threshold * (value * 0.25f))
        shouldCompress = false;
    }
    if (curDepth < 3) shouldCompress = false;
  }
  if (shouldCompress) deleteChildren();
}

void SphereNode::countLeaves(size_t& counter) const {
  if (children == nullptr)
    counter++;
  else {
    children[0].countLeaves(counter);
    children[1].countLeaves(counter);
    children[2].countLeaves(counter);
    children[3].countLeaves(counter);
  }
}

float SphereNode::walkIndex(
  size_t index,
  unsigned& curDepth,
  uint32_t maxDepth,
  const std::vector<SphereHEI>& partitions,
  uint32_t stopDepth) const {
  // If we've reached a leaf, return our value
  if (children == nullptr || curDepth >= stopDepth) {
    float val = value;
    for (unsigned depth = curDepth; depth < maxDepth; depth++) val *= 0.25f;
    return val;
  }
  size_t childSize = partitions[maxDepth - curDepth - 1].getCellCount() / 12;
  size_t childIndex = index / childSize;
  return children[childIndex].walkIndex(
    index - childIndex * childSize, ++curDepth, maxDepth, partitions,
    stopDepth);
}

void SphereNode::write(unsigned curDepth, std::ostream& out) const {
  // if we've reached a leaf, write out our contents
  if (children == nullptr) {
    uint32_t composed = 0;
    // we'll use 4-bits for our current depth
    composed = curDepth;
    composed = composed << 28;
    // we'll represent the value with 0.12345678 digits (27-bits)
    uint32_t fraction = round(value * 1e8);
    composed |= fraction;
    out.write((const char*)&composed, sizeof(uint32_t));

  } else {
    curDepth++;
    for (const SphereNode* child = children; child < children + 4; child++)
      child->write(curDepth, out);
  }
}

void SphereNode::writeOBJ(
  unsigned region,
  unsigned curDepth,
  uint32_t maxDepth,
  size_t index,
  const std::vector<SphereHEI>& partitions,
  size_t& vIndex,
  std::ostream& out,
  bool useMag) const {
  if (children == nullptr)
    partitions[curDepth].writeOBJFacet(
      region, index, vIndex, out, useMag ? value : 1.f);
  else {
    curDepth++;
    index *= 4;
    for (const SphereNode* child = children; child < children + 4; child++)
      child->writeOBJ(
        region, curDepth, maxDepth, index++, partitions, vIndex, out, useMag);
  }
}

void SphereNode::writeSVG(
  unsigned region,
  unsigned curDepth,
  uint32_t maxDepth,
  float width,
  size_t index,
  float solidAngle,
  const std::vector<SphereHEI>& partitions,
  std::ostream& out,
  bool bw) const {
  if (children == nullptr) {
    float val = value;
    for (unsigned depth = curDepth; depth < maxDepth; depth++) val *= 0.25f;
    val /= solidAngle;
    if (val > 1) val = 1;
    partitions[curDepth].writeSVGFacet(region, index, width, out, val, bw);
  } else {
    curDepth++;
    index *= 4;
    for (const SphereNode* child = children; child < children + 4; child++)
      child->writeSVG(
        region, curDepth, maxDepth, width, index++, solidAngle, partitions, out,
        bw);
  }
}

void SphereNode::writeTXT(
  unsigned region,
  unsigned curDepth,
  uint32_t maxDepth,
  size_t index,
  const std::vector<SphereHEI>& partitions,
  std::ostream& out) const {
  if (children == nullptr)
    partitions[curDepth].writeTXTFacet(region, index, out, 20);
  else {
    curDepth++;
    index *= 4;
    for (const SphereNode* child = children; child < children + 4; child++)
      child->writeTXT(region, curDepth, maxDepth, index++, partitions, out);
  }
}

void SphereNode::read(
  size_t cellCount,
  unsigned& dataDepth,
  float& dataValue,
  unsigned curDepth,
  std::istream& in,
  uint32_t& readCount) {
  if (readCount >= cellCount) return;

  if (curDepth == dataDepth) {
    // set the value
    value = dataValue;

    // read in the data depth and data value
    uint32_t composed;
    in.read((char*)&composed, sizeof(uint32_t));
    readCount++;

    // get the next data point
    uint32_t depth;
    depth = composed >> 28;
    dataValue = (composed ^ (depth << 28)) * 1e-8;
    dataDepth = unsigned(depth);
  } else {
    // need to create the next depth
    deleteChildren();
    children = new SphereNode[4];
    for (SphereNode* child = children; child < children + 4; child++)
      child->read(cellCount, dataDepth, dataValue, curDepth + 1, in, count);
  }
}

}  // namespace d5::sqt
