#include "dirsig5/sqt/SphereQuadTree.h"

#include <cmath>
#include <fstream>
#include <iomanip>
#include <sstream>

#include "Cosinify.h"
#include "dirsig5/base/BinaryStream.h"
#include "dirsig5/base/LerpUtils.h"
#include "elzma.h"

template <typename... Floats>
static constexpr float meanOf(float arg0, Floats... args) noexcept {
  return (arg0 + ... + args) / float(sizeof...(Floats) + 1);
}

namespace d5::sqt {

void SphereQuadTree::addVectorData(Vectorf vec, float value) {
  DIRSIG_THROW_IF(finalized, "Attempted to add data after finalization!");
  if (not std::isfinite(value)) return;

  if (isCosineWeighted()) {
    // If cosine weighting, uncosinify the vector. This causes the
    // integrated magnitude to double, so halve the value.
    vec = uncosinify(vec);
    value *= 0.5f;
  }

  SphereNode* node = indexToNode(partitions.back().vectorToIndex(vec));
  if (node->value < 0) {
    node->value = 0;
    node->count = 0;
  }
  node->value += std::fmax(value, 1e-9f);
  node->count++;
}

void SphereQuadTree::finalize() {
  if (finalized) return;

  size_t cellCount = getCellCount();
  std::vector<Vectorf> vecs;
  bool anyBelow = false;
  bool anyAbove = false;
  auto updateCornerVecs = [&](size_t cell) {
    constexpr float eps = 1e-4f;
    vecs.clear();
    partitions.back().indexToCell(cell, vecs);
    anyBelow = vecs[0][2] < -eps or vecs[1][2] < -eps or  //
               vecs[2][2] < -eps or vecs[3][2] < -eps;
    anyAbove = vecs[0][2] > +eps or vecs[1][2] > +eps or  //
               vecs[2][2] > +eps or vecs[3][2] > +eps;
  };
  for (size_t cell = 0; cell < cellCount; cell++) {
    updateCornerVecs(cell);
    if (
      (not anyAbove and not hasLowerHemisphere()) or  //
      (not anyBelow and not hasUpperHemisphere())) {
      SphereNode* node = indexToNode(cell);
      node->value = 0;
      node->count = 1;
      continue;
    }
  }

  magnitude = 0;
  for (SphereNode& region : regions) {
    region.averageSamples();
    region.accumulate();
    magnitude += region.value;
  }
  if (magnitude > 0) {
    float normalization = 1 / magnitude;
    for (SphereNode& region : regions) region.scale(normalization);
  }

  double upperSum = 0;
  double lowerSum = 0;
  double equatorSum = 0;
  for (size_t cell = 0; cell < cellCount; cell++) {
    updateCornerVecs(cell);
    double value = indexToValue(cell);
    if (anyAbove and not anyBelow) upperSum += value;
    if (anyBelow and not anyAbove) lowerSum += value;
    if (anyAbove and anyBelow) equatorSum += value;
  }
  upperSum += equatorSum * 0.5;
  lowerSum += equatorSum * 0.5;
  if (upperSum == 0 and lowerSum == 0) {  // Shouldn't happen?
    upperFraction = 0.5f;
    lowerFraction = 0.5f;
  } else {
    upperFraction = upperSum / (upperSum + lowerSum);
    lowerFraction = lowerSum / (upperSum + lowerSum);
  }
  if (hasUpperHemisphere() != hasLowerHemisphere()) {
    upperFraction = hasUpperHemisphere() ? 1 : 0;
    lowerFraction = hasUpperHemisphere() ? 0 : 1;
    magnitude -= equatorSum * 0.5;
  }
  magnitude *= solidAngle;
  finalized = true;
}

void SphereQuadTree::compress(float threshold) {
  for (SphereNode& region : regions) region.compress(threshold);

  // Renormalize.
  float normalization = 0;
  for (SphereNode& region : regions) {
    region.accumulate();
    normalization += region.value;
  }
  if (normalization > 0 and normalization != 1) {
    normalization = 1 / normalization;
    for (SphereNode& region : regions) region.scale(normalization);
  }
}

void SphereQuadTree::read(std::istream& in) {
  BinaryStream binaryIn(&in);

  uint64_t cellCount = 0;
  uint64_t compressedSize = 0;
  binaryIn.read(&maxDepth, 1);
  partitions.clear();
  for (uint32_t depth = 0; depth <= maxDepth; depth++)
    partitions.emplace_back(depth);
  binaryIn.read(&flags, 1);
  binaryIn.read(&solidAngle, 1);
  binaryIn.read(&magnitude, 1);
  binaryIn.read(&upperFraction, 1);
  binaryIn.read(&lowerFraction, 1);
  binaryIn.read(&cellCount, 1);
  binaryIn.read(&compressedSize, 1);

  std::vector<unsigned char> compressedBuff(compressedSize);
  binaryIn.read(compressedBuff.data(), compressedBuff.size());
  unsigned char* buff = nullptr;
  size_t buffSize = 0;
  elzma_decompress_handle handle = {};
  int elzmaResult = simpleDecompress(
    ELZMA_lzma, compressedBuff.data(), compressedBuff.size(),  //
    &buff, &buffSize, handle);
  DIRSIG_THROW_IF(
    elzmaResult != ELZMA_E_OK,
    "Unable to decompress memory from input buffer!");
  auto rawStreambuf = RawIStreambuf(buff, buffSize);
  auto rawStream = std::istream(&rawStreambuf);
  uint32_t data = BinaryStream(&rawStream).read<uint32_t>();
  uint32_t dataDepth = data >> 28;
  float dataValue = (data ^ (dataDepth << 28)) * 1e-8;
  uint32_t count = 0;
  for (SphereNode& region : regions) {
    region.read(cellCount, dataDepth, dataValue, 0, rawStream, count);
    region.accumulate();
  }

  free(buff);
  elzma_decompress_free(&handle);
  finalized = true;
}

void SphereQuadTree::write(std::ostream& out) const {
  DIRSIG_THROW_IF(
    not finalized, "Attempted to write an un-finalized SphereQuadTree!");

  uint64_t cellCount = getCellCount();
  std::vector<unsigned char> buff(cellCount * sizeof(uint32_t));
  {
    // Fill buffer.
    auto rawStreambuf = RawOStreambuf(buff);
    auto rawStream = std::ostream(&rawStreambuf);
    for (const SphereNode& region : regions) region.write(0, rawStream);
  }
  // Compress.
  unsigned char* compressedBuff = nullptr;
  size_t compressedSize = 0;
  elzma_compress_handle handle = {};
  int elzmaResult = simpleCompress(
    ELZMA_lzma, buff.data(), buff.size(), &compressedBuff, &compressedSize,
    handle);
  DIRSIG_THROW_IF(
    elzmaResult != ELZMA_E_OK, "Unable to compress memory for output buffer!");

  // Write.
  BinaryStream binaryOut(&out);
  binaryOut.write(maxDepth);
  binaryOut.write(flags);
  binaryOut.write(solidAngle);
  binaryOut.write(magnitude);
  binaryOut.write(upperFraction);
  binaryOut.write(lowerFraction);
  binaryOut.write(cellCount);
  binaryOut.write(uint64_t(compressedSize));
  binaryOut.write(compressedBuff, compressedSize);

  // Clean up.
  free(compressedBuff);
  elzma_compress_free(&handle);
}

void SphereQuadTree::writeSVG(std::ostream& out, bool bw) const {
  // Width doesn't matter, but we need something
  unsigned width = 768;
  out << "<?xml version=\"1.0\" standalone=\"no\" ?>\n"
         "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" "
         "\"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n";
  out << "<svg width=\"" << 2 * width << "px\" height=\"" << width
      << "px\" xmlns=\"http://www.w3.org/2000/svg\" version=\"1.1\" >\n";
  size_t regionIndex = 0;
  for (const SphereNode& region : regions)
    region.writeSVG(
      regionIndex++, 0, maxDepth, width, 0, solidAngle, partitions, out, bw);
  out << "</svg>\n";
  out.flush();
}

float SphereQuadTree::getPdf(
  Vectorf vec, bool allowUpper, bool allowLower) const {
  DIRSIG_THROW_IF(
    not finalized, "Attempted to query an un-finalized SphereQuadTree!");

  float cosTheta = vec[2];
  allowUpper &= hasUpperHemisphere();
  allowLower &= hasLowerHemisphere();
  if (
    (not allowUpper and cosTheta >= 0) or  //
    (not allowLower and cosTheta <= 0))
    return 0;  // Impossible!

  // If cosine weighting, uncosinify the vector.
  if (isCosineWeighted()) vec = uncosinify(vec);

  // Prevent edge-case behavior.
  float eps = 1e-5f;
  if (std::abs(vec[0]) < eps) vec[0] = std::copysign(eps, vec[0]);
  if (std::abs(vec[1]) < eps) vec[1] = std::copysign(eps, vec[1]);

  const auto& deepest = partitions.back();
  int index = deepest.vectorToIndex(vec);
  int quadSize = deepest.getCellCount() / 12;
  unsigned regionIndex = index / quadSize;
  index -= regionIndex * quadSize;

  std::array<SphereHEI::RFF, 9> rffs;
  std::array<float, 9> values;
  std::fill(values.begin(), values.end(), -1);
  unsigned curDepth = 0;
  unsigned solDepth = 0;
  regions[regionIndex].walkIndex(index, curDepth, maxDepth, partitions);
  solDepth = curDepth;
  partitions[solDepth].getCorners(vec, rffs);
  constexpr unsigned Invalid = -1;
  for (int i = 0; i < 9; i++)
    if (rffs[i].R != Invalid) {
      partitions.back().encodeFractions(rffs[i].f1, rffs[i].f2, index);
      curDepth = 0;
      values[i] = regions[rffs[i].R].walkIndex(
                    index, curDepth, maxDepth, partitions, solDepth) /
                  solidAngle;
    }

  // c00 => (0, 0) ( 1, 0) ( 0, 1)
  // c01 => (0, 0) ( 0, 1) (-1, 0)
  // c11 => (0, 0) (-1, 0) ( 0,-1)
  // c10 => (0, 0) ( 0,-1) ( 1, 0)
  float c00 = rffs[2].R != Invalid
                ? meanOf(values[0], values[1], values[2], values[3])
                : meanOf(values[0], values[1], values[3]);
  float c01 = rffs[4].R != Invalid
                ? meanOf(values[0], values[3], values[4], values[5])
                : meanOf(values[0], values[3], values[5]);
  float c11 = rffs[6].R != Invalid
                ? meanOf(values[0], values[5], values[6], values[7])
                : meanOf(values[0], values[5], values[7]);
  float c10 = rffs[8].R != Invalid
                ? meanOf(values[0], values[7], values[8], values[1])
                : meanOf(values[0], values[7], values[1]);

  // Get the fractions for the current point
  float f1 = 0.0f;
  float f2 = 0.0f;
  partitions[solDepth].vectorToFractions(vec, rffs[0].R, f1, f2);
  float fs = partitions[solDepth].getSplit();
  float pdf = bilerp(
    0.5f - (f2 - rffs[0].f2) / fs,  //
    0.5f - (f1 - rffs[0].f1) / fs, c00, c01, c10, c11);

  // If cosine weighting, apply Jacobian of cosinify transform.
  if (isCosineWeighted()) pdf *= 2 * std::abs(cosTheta);

  // Make sure this is normalized.
  if (allowUpper != allowLower)
    pdf /= cosTheta > 0 ? upperFraction : lowerFraction;

  return pdf;
}

Vectorf SphereQuadTree::samplePdf(
  RandEngine& randEngine, bool allowUpper, bool allowLower) const {
  DIRSIG_THROW_IF(
    not finalized, "Attempted to sample an un-finalized SphereQuadTree!");
  allowUpper &= hasUpperHemisphere();
  allowLower &= hasLowerHemisphere();
  if ((not allowUpper and not allowLower) or magnitude <= 1e-20f) {
    return {0, 0, 1};
  }

  Vectorf vec;
  size_t deepestCellCount = partitions.back().getCellCount();
  size_t bailCount = 0;
  while (1) {
    float xi = std::generate_canonical<float, 32>(randEngine);
    size_t regionIndex = 0;
    while (regionIndex < 12) {
      if (xi < regions[regionIndex].value) break;
      xi -= regions[regionIndex].value;
      regionIndex++;
    }
    if (regionIndex > 11) regionIndex = 11;

    unsigned order = 0;
    size_t quadTreeIndex = 0;
    regions[regionIndex].sample(xi, order, quadTreeIndex);

    // have to sample at the same depth because of shape differences
    for (; order + 1 < partitions.size(); order++) {
      // walk down to the deepest cells uniformly
      quadTreeIndex *= 4;
      quadTreeIndex += randEngine(4);
    }

    size_t index = regionIndex * deepestCellCount / 12 + quadTreeIndex;
    if (index >= deepestCellCount) index = deepestCellCount - 1;
    partitions[order].indexToSampleVector(index, vec, randEngine);
    if (
      (vec[2] > 0 and allowUpper) or  //
      (vec[2] < 0 and allowLower))
      break;
    if (++bailCount > 10) {
      return {0, 0, 1};
    }
  }

  // If cosine weighting, cosinify the vector.
  if (isCosineWeighted()) vec = cosinify(vec);

  return vec;
}

SphereNode* SphereQuadTree::indexToNode(size_t nodeIndex) {
  DIRSIG_THROW_IF(finalized, "Attempted to look up node after finalization!");
  int level = partitions.size() - 1;
  size_t cellCount = partitions[level].getCellCount();
  size_t regionCount = cellCount / 12;
  size_t region = nodeIndex / regionCount;
  DIRSIG_THROW_IF(nodeIndex >= cellCount, "Out of bounds index!");
  nodeIndex -= region * regionCount;
  SphereNode* node = &regions[region];
  while (level-- > 0) {
    cellCount = partitions[level].getCellCount() / 12;
    size_t childIndex = nodeIndex / cellCount;
    assert(childIndex < 4);
    node = node->children + childIndex;
    nodeIndex -= cellCount * childIndex;
  }
  return node;
}

float SphereQuadTree::indexToValue(size_t nodeIndex) {
  DIRSIG_THROW_IF(finalized, "Attempted to look up node after finalization!");
  int level = partitions.size() - 1;
  size_t cellCount = partitions[level].getCellCount();
  size_t regionCount = cellCount / 12;
  size_t region = nodeIndex / regionCount;
  DIRSIG_THROW_IF(nodeIndex >= cellCount, "Out of bounds index!");
  nodeIndex -= region * regionCount;
  SphereNode* node = &regions[region];
  while (level-- > 0) {
    if (not node->children) break;
    cellCount = partitions[level].getCellCount() / 12;
    size_t childIndex = nodeIndex / cellCount;
    assert(childIndex < 4);
    node = node->children + childIndex;
    nodeIndex -= cellCount * childIndex;
  }
  float value = node->value;
  while (level-- >= 0) value *= 0.25f;
  return value;
}

}  // namespace d5::sqt
