#include "dirsig5/thermal/RitTherm.h"

#include <cassert>
#include <cmath>

namespace {

/**
 * Compute equilibrum temperature assuming no mass.
 *
 * This is the temperature to which a facet is being driven to at each point in
 * time.  For a facet with actual mass, there is thermal inertia coming in to
 * play.
 *
 * \param    thermalProps
 * \param    surfaceNormal
 * \param    stepData
 * \param    peakAirTemp
 * \param    sunriseAirTemp
 * \param    radiativeRateResult
 * \param    conductiveRateResult
 * \param    convectiveRateResult
 *
 * \return   temperature assuming zero mass.
 */
d5::Temperature txComputeZeroMass(
  const d5::thermal::ThermRecord& thermalProps,
  const d5::Vector& surfaceNormal,
  const d5::thermal::RitThermInvariants::ThermTimeStepInvariants& stepData,
  d5::Temperature peakAirTemp,
  d5::Temperature sunriseAirTemp,
  double directInsolationWeighting,
  double diffuseInsolationWeighting,
  double& radiativeRateResult,
  double& conductiveRateResult,
  double& convectiveRateResult) {
  /*
   * Constants and reused values
   */
  // Celsius to Kelvin conversion
  constexpr double cToKelvin = 273.15;

  // Stefan-Boltzmann constant expressed in [L/hr/K^4]
  constexpr double sb = 4.87185e-9;

  // Cosine of surface normal zenith
  double cosSurfaceZenith = surfaceNormal[2];

  // Sine of surface normal zenith
  double sinSurfaceZenith = sqrt(1 - cosSurfaceZenith * cosSurfaceZenith);

  // Air temperature in [K]
  auto airTempKelvin = stepData.weather.airTemperature.kelvin();

  // Absolute val of exposed area
  double absExposedArea = std::abs(thermalProps.exposedArea);

  // Compute convective heat transfer rate [L/hr/C], DCS eqn 6.2, pg 18
  double convectiveRate;
  if (thermalProps.exposedArea >= 0.0f) {
    convectiveRate =
      (absExposedArea * stepData.weather.windSpeedMS) / 3.0 + 0.5;
  } else {
    // exposed both sides...
    convectiveRate =
      (absExposedArea * stepData.weather.windSpeedMS) / 3.0 + 0.25;
  }

  // Radiative heat transfer rate, initial estimate, DCS eqn 8.3, pg 22
  // midpoint between day's high/low temperatures [K]
  double midpointTempK = 0.5 * (peakAirTemp.kelvin() + sunriseAirTemp.kelvin());
  double radiativeRateGuess = 4 * thermalProps.thermalEmissivity * sb *
                              midpointTempK * midpointTempK * midpointTempK;

  /*
   * Compute conductive heat transfer rate [L/hr/C], DCS eqn 7.2, pg 19
   */
  double conductiveRate = 0.0;
  double conductiveInput = 0.0;

#if 0
  // we only have conduction if there is rain
  if (stepData.precipPresent) {
    conductiveRate =
        absExposedArea * std::max(0.0f, cosSurfaceZenith) * stepData.precipRate;

    // Contribution from rain [L/hr]
    conductiveInput = conductiveRate * stepData.precipTemp;
  }
#endif

  /*
   * Compute total heat transfer rate, DCS eqn 8.3, pg 22
   */
  double totalHeatRateGuess =
    conductiveRate + convectiveRate + radiativeRateGuess;

  /*
   * Compute Biot number, DCS eqn. 2.1, pg 4
   */
  double biot;
  if (thermalProps.thermalConductivity > 0.0) {
    biot = totalHeatRateGuess * std::abs(thermalProps.thickness) /
           thermalProps.thermalConductivity;
  } else {
    biot = std::abs(thermalProps.thermalConductivity);
  }

  /*
   * Compute non-zero Biot adjustment, DCS eqn. 2.7, pg 5
   */
  assert(thermalProps.thickness >= 0.0);

  // T_ext - T_int = bfac * (T_ext,equilib - T_int,equilib)
  double bfac = biot / (biot + 1);

  // FIXME: Reference for bfac application?
  if (thermalProps.exposedArea >= 0.0f) {
    convectiveRate += bfac * (convectiveRate - 0.5f);
  } else {
    convectiveRate += bfac * convectiveRate;
  }
  conductiveRate += bfac * conductiveRate;
  conductiveInput += bfac * conductiveInput;
  absExposedArea += bfac * absExposedArea;

  /*
   * Output parameter store: convective, conductive, radiative rate needed for
   *                         non-zero mass calculation.
   */
  assert(!std::isnan(conductiveRate));
  assert(!std::isnan(convectiveRate));
  assert(!std::isnan(radiativeRateGuess));

  conductiveRateResult = conductiveRate;
  convectiveRateResult = convectiveRate;
  radiativeRateResult = radiativeRateGuess;

  //
  // Compute effective sky temperature [C], DCS eqn 5.2, pg 16
  //
  // Effective diffused insolation transmission relative to exposed
  // horizontal surface.
  //   DCS: TR_dif term, eqn 5.2, pg 16
  double effectiveDiffuse = cosSurfaceZenith * 0.5 + 0.5;

  // if the diffuse weighting is smaller (sky shadowing) us that
  if (effectiveDiffuse > diffuseInsolationWeighting) {
    effectiveDiffuse = diffuseInsolationWeighting;
  }

  // FIXME: THERM does this but it deviates from DCS eqn!
  //        We do this to recreate the floating point numerical noise of the
  //        original THERM.  This should eventually be changed to std::pow.
  double skyEmissSqr = stepData.effSkyEmissivity;
  skyEmissSqr *= skyEmissSqr;

  // Double precision is needed here to cut down floating point error
  double powInner =
    effectiveDiffuse * (skyEmissSqr * skyEmissSqr) + (1 - effectiveDiffuse);

  double effectiveSkyTemp =
    std::pow(powInner, 0.25) * airTempKelvin - cToKelvin;

  /*
   * Compute effects of external temperatures, DCS eqn 3.3, pg 11
   */
  // Sky, air, and precipitation heat input [L/hr]
  double externalTemperatureContrib =
    thermalProps.thermalEmissivity * sb *
      (absExposedArea * std::pow(effectiveSkyTemp + cToKelvin, 4) +
       (1.0f - absExposedArea) * std::pow(airTempKelvin, 4)) +
    convectiveRate * stepData.weather.airTemperature.celsius();
  //+ conductiveRate * stepData.weather.precipTemperature.celsius();

  /*
   * Angle between surface normal and sun position, DCS eqn 9.2, pg 22
   */
  // FIXME: We can use radians instead of degrees
  double deltaPhi = (d5::getGeoAzimuth(surfaceNormal) * (180.0 / d5::PI)) -
                    stepData.sunAzimuthDegrees;

  // Restrict phi difference to [0,180]
  deltaPhi = std::fmod(deltaPhi, 360.0);
  if (std::fabs(deltaPhi) > 180.0) {
    deltaPhi -= (deltaPhi > 0.0) ? 360.0 : -360.0;
  }

  // FIXME: Can we just use vector-vector angle here?
  double cosSurfaceToSunAngle = std::max(
    (cosSurfaceZenith * stepData.cosSunZenith) +
      (sinSurfaceZenith * stepData.sinSunZenith *
       std::cos(deltaPhi * (d5::PI / 180))),
    0.);

  /*
   * Compute effective visible insolation, DCS eqn 4.1, pg 12
   */
  // Scale direct insolation based on sun-shadowing
  double weightedDirectInsolation =
    (stepData.weather.directInsolationWM2 * 0.0858) * directInsolationWeighting;

  double effDirectInsolation = thermalProps.solarAbsorption * absExposedArea *
                               weightedDirectInsolation * cosSurfaceToSunAngle;

  double effIndirectInsolation = thermalProps.solarAbsorption * absExposedArea *
                                 effectiveDiffuse *
                                 stepData.weather.diffuseInsolationWM2 * 0.0858;

  double effectiveVisInsolation = effDirectInsolation + effIndirectInsolation;

  /*
   * Net rate at which heat is entering surface, DCS eqn 3.2, pg 11
   */
  double netInputRate = externalTemperatureContrib + effectiveVisInsolation +
                        thermalProps.selfGeneratedPower;

  /*
   * ??? Attempt to solve equilibrium temperature, DCS eqn 3.2, pg 11
   */
  // ??? Gradient minimization: compute the temp that has flux in equilibriu  m.
  double equilibriumTemp = stepData.weather.airTemperature.celsius();
  double a = 0.0;
  double totalHeatGuess = radiativeRateGuess + convectiveRate + conductiveRate;

  // Iterate on the temperature until we hit equilibrium
  while (1) {
    // Heat inputs minus heat losses (L/hr)
    a = netInputRate -
        thermalProps.thermalEmissivity * sb *
          std::pow(static_cast<double>(equilibriumTemp + cToKelvin), 4) -
        (convectiveRate + conductiveRate) * equilibriumTemp;

    // If we have converged, bail out
    if (std::abs(a) <= 0.01) {
      break;
    }

    // Prepare for another iteration with a new guess at the temperature
    equilibriumTemp += a / totalHeatGuess;
  }

  // temperature equilibrium for zero mass gray body [C]
  assert(!std::isnan(equilibriumTemp));
  assert(std::isfinite(equilibriumTemp));
  return d5::Temperature::fromCelsius(equilibriumTemp);
}

void txBiotExpansion(
  double biotNumber,
  double thermalMass,
  std::vector<double>& biotExpansionWeights,
  std::vector<double>& biotExpansionThermalMasses) {
  // Initialize output parameters
  biotExpansionWeights.clear();
  biotExpansionThermalMasses.clear();
  biotExpansionWeights.resize(5, 0.);
  biotExpansionThermalMasses.resize(5, 0.);

  // No need for expansion if Biot number is small, DCS eqn. 1.10, pg 2
  // FIXME: Can we make this check for Biot number < 0.01?
  if (biotNumber == 0 || thermalMass == 0) {
    biotExpansionThermalMasses[0] = thermalMass;
    biotExpansionWeights[0] = 1;
    return;
  }

  // Taylor series expansion.
  // Sum of expansion weights (series from 1..inf would make weight 1)
  double weightSum = 0;
  double xOldInit = std::atan(std::sqrt(biotNumber));
  for (std::size_t ii = 0; ii < biotExpansionWeights.size(); ii++) {
    // Compute X_i, DCS eqn 2.24, pg 8
    double diff = 1.0;
    double xold = xOldInit;
    double xnew = 0.0;
    auto expansionIdxf = static_cast<double>(ii);
    do {
      // The inclusion of index here gives us a different X_i at each
      // series step
      xnew = std::atan(biotNumber / xold) + expansionIdxf * d5::PI;
      diff = std::fabs(xnew - xold);
      xold = xnew;
    } while (diff > 0.001);

    // Compute apparent thermal mass for series step, DCS eqn 2.32, pg 9
    biotExpansionThermalMasses[ii] = thermalMass * (biotNumber / (xnew * xnew));

    // When we run out of terms or our thermal mass becomes too small, we
    // assign the last term all the remaining weight.
    if (
      biotExpansionThermalMasses[ii] < 0.01 ||
      ii == (biotExpansionWeights.size() - 1)) {
      biotExpansionWeights[ii] = 1 - weightSum;
      break;
    }

    /*
     * Compute C_i term, DCS eqn 2.29, pg 9
     *
     * The cos(X_i) term in the denominator dissapeared.  I think this might
     * be because of eqn 2.24 above.  If X_i were even multiple of Pi, that
     * would always evaluate to 1.
     */
    biotExpansionWeights[ii] =
      2. / (biotNumber + 1 + ((xnew * xnew) / biotNumber));

    // Keep running sum of weights generated in series expansion so far...
    weightSum += biotExpansionWeights[ii];

    // Sanity-Check NaNs
    assert(!std::isnan(biotExpansionThermalMasses[ii]));
    assert(!std::isnan(biotExpansionWeights[ii]));
    assert(std::isfinite(biotExpansionThermalMasses[ii]));
    assert(std::isfinite(biotExpansionWeights[ii]));
  }
}

/**
 * Compute equlibrium temperature for actual mass.
 *
 * \param    timestepSize
 * \param    preZeroMassTemp
 * \param    curZeroMassTemp
 * \param    preRealMassTemp
 * \param    zmAvgConductiveRate
 * \param    zmAvgConvectiveRate
 * \param    zmRadiativeRateGuess
 * \param    thermalProps
 *
 * \return   temperature assuming actual mass.
 */
d5::Temperature txComputeRealMass(
  double timestepSize,
  d5::Temperature preZeroMassTemp,
  d5::Temperature curZeroMassTemp,
  d5::Temperature preRealMassTemp,
  double zmAvgConductiveRate,
  double zmAvgConvectiveRate,
  double zmRadiativeRateGuess,
  const d5::thermal::ThermRecord& thermalProps) {
  /*
   * Constants and reused values
   */
  // Stefan-Boltzmann constant expressed in [L/hr/K^4]
  constexpr double sb = 4.87185e-9;

  // Approx total heat transfer rate [L/hr/C], computed in zero-mass case
  double zmTotalHeatRate =
    zmRadiativeRateGuess + zmAvgConductiveRate + zmAvgConvectiveRate;

  // Compute Biot number, DCS eqn. 2.1, pg 4
  double biot;
  if (thermalProps.thermalConductivity > 0.0) {
    biot = zmTotalHeatRate * std::abs(thermalProps.thickness) /
           thermalProps.thermalConductivity;
  } else {
    biot = std::abs(thermalProps.thermalConductivity);
  }

  // Compute Biot Taylor expansion, DCS eqn 2.32 and 2.33, pg 9
  static thread_local std::vector<double> biotExpansionWeights;
  static thread_local std::vector<double> biotExpansionThermalMasses;
  biotExpansionWeights.clear();
  biotExpansionThermalMasses.clear();
  double thermalMass = thermalProps.specificHeat * thermalProps.massDensity *
                       thermalProps.thickness;
  txBiotExpansion(
    biot, thermalMass, biotExpansionWeights, biotExpansionThermalMasses);

  /*
   * Compute refined total heat transfer rate [L/hr/C], DCS eqn 8.4, pg 21
   *
   * We now have guesses at actual temperatures, which we didn't for zm
   * version of thie calculation.
   */
  d5::Temperature curRealMassTemp = d5::Temperature::fromCelsius(0);

  // Left in powering of curRealMassTemp for clarity, always evals to zero
  double refinedTotalHeatRateGuess = thermalProps.thermalEmissivity * sb *
                                       (std::pow(preZeroMassTemp.kelvin(), 3) +
                                        std::pow(preRealMassTemp.kelvin(), 3) +
                                        std::pow(curZeroMassTemp.kelvin(), 3) +
                                        std::pow(curRealMassTemp.kelvin(), 3)) +
                                     zmAvgConductiveRate + zmAvgConvectiveRate;

  /*
   * Refine current temperature based on Biot taylor expansion
   */
  for (int idx = 0; idx < 5; ++idx) {
    assert(!std::isnan(biotExpansionThermalMasses[idx]));
    // Initialize assuming no Biot thermal mass
    double biotTempRefinement = curZeroMassTemp.celsius();

    // Calculate actual biot adjust for nonzero Biot thermal mass
    if (biotExpansionThermalMasses[idx] > 0) {
      // Approximate exponential heat transfer decay rate [1/hr]
      // DCS eqn 1.15, pg 3
      double rd =
        (refinedTotalHeatRateGuess / biotExpansionThermalMasses[idx]) *
        timestepSize;
      double expNegRd = std::exp(-rd);
      assert(!std::isnan(expNegRd));

      // Refined approximate temperature [C]
      biotTempRefinement =
        preRealMassTemp.celsius() * expNegRd +
        preZeroMassTemp.celsius() * (1 - expNegRd) +
        ((curZeroMassTemp.celsius() - preZeroMassTemp.celsius()) *
         (1 - (1 - expNegRd) / rd));
    }

    curRealMassTemp = d5::Temperature::fromCelsius(
      biotExpansionWeights[idx] * biotTempRefinement +
      curRealMassTemp.celsius());
    assert(!std::isnan(curRealMassTemp.kelvin()));
  }

  return curRealMassTemp;
}
}  // namespace

namespace d5::thermal {

Temperature therm(
  const Vector& surfaceNormal,
  const thermal::ThermRecord& thermalProps,
  const std::vector<float>& sunShadowHistory,
  float diffuseInsolationWeight,
  const RitThermInvariants& thermInvariants,
  std::size_t startIndex) {
  if (startIndex >= thermInvariants.size()) {
    throw std::domain_error(
      "out-of-bounds start THERM index (" + std::to_string(startIndex) +
      ", but must be less than " + std::to_string(thermInvariants.size()) +
      ")");
  }

  // Input sanity checks.
  assert(std::isfinite(thermalProps.solarAbsorption));
  assert(thermalProps.solarAbsorption >= 0.0);
  assert(thermalProps.solarAbsorption <= 1.0);
  assert(std::isfinite(thermalProps.thermalEmissivity));
  assert(thermalProps.thermalEmissivity >= 0.0);
  assert(thermalProps.thermalEmissivity <= 1.0);
  assert(thermalProps.exposedArea >= -1.0);
  assert(thermalProps.exposedArea <= 1.0);
  assert(thermalProps.thickness >= 0.0);

  // Zero mass result storage: temperature [C], radiative, conductive and
  // convective heat transfer rate guess [L/hr/C]
  static thread_local std::vector<Temperature> zmTemperature;
  static thread_local std::vector<double> zmRadiativeRateGuess;
  static thread_local std::vector<double> zmConductiveRate;
  static thread_local std::vector<double> zmConvectiveRate;
  zmTemperature.resize(thermInvariants.size());
  zmRadiativeRateGuess.resize(thermInvariants.size(), 0);
  zmConductiveRate.resize(thermInvariants.size(), 0);
  zmConvectiveRate.resize(thermInvariants.size(), 0);

  // Predict temperature equilibrium assuming surface has zero mass.
  for (std::size_t ts = startIndex; ts < thermInvariants.size(); ts++) {
    zmTemperature[ts] = txComputeZeroMass(
      thermalProps, surfaceNormal, thermInvariants[ts],
      thermInvariants.peakAirTemp(), thermInvariants.sunriseAirTemp(),
      sunShadowHistory[ts], diffuseInsolationWeight, zmRadiativeRateGuess[ts],
      zmConductiveRate[ts], zmConvectiveRate[ts]);
  }

  // Storage for real mass temperature predictions.
  static thread_local std::vector<Temperature> rmTemperature;
  rmTemperature.resize(thermInvariants.size());

#ifdef _DEBUG_OUTPUT
  std::ofstream debugFile("d5_therm_steps.txt");
  debugFile << "# step count = " << thermInvariants.size() << std::endl
            << "# index timeHr rmTemperature zmTemperature zmRadRateGuess "
               "zmAvgCondRate zmAvgConvRate airTemp directInsol diffuseInsol"
            << std::endl;
#endif

  // Assume that we are in thermal equilibrium for our first step.
  rmTemperature[startIndex] = zmTemperature[startIndex];

  // Predict temperature equilibrium accounting for real mass
  for (std::size_t ts = startIndex + 1; ts < thermInvariants.size(); ts++) {
    // Average zero-mass conductive/convective rate over step period
    auto zmAvgConductiveRate =
      0.5 * (zmConductiveRate[ts] + zmConductiveRate[ts - 1]);

    auto zmAvgConvectiveRate =
      0.5 * (zmConvectiveRate[ts] + zmConvectiveRate[ts - 1]);

    rmTemperature[ts] = txComputeRealMass(
      thermInvariants.timeStepDelta(ts), zmTemperature[ts - 1],
      zmTemperature[ts], rmTemperature[ts - 1], zmAvgConductiveRate,
      zmAvgConvectiveRate, zmRadiativeRateGuess[ts], thermalProps);
    assert(std::isfinite(rmTemperature[ts].kelvin()));
#ifdef _DEBUG_OUTPUT
    debugFile << ts << " " << std::fixed << std::setprecision(2)
              << thermInvariants.at(ts).timeHr << " "
              << rmTemperature[ts].celsius() << " "
              << zmTemperature[ts].celsius() << " " << zmRadiativeRateGuess[ts]
              << " " << zmAvgConductiveRate << " " << zmAvgConvectiveRate << " "
              << thermInvariants.at(ts).weather.airTemperature.celsius() << " "
              << thermInvariants.at(ts).weather.directInsolationWM2 << " "
              << thermInvariants.at(ts).weather.diffuseInsolationWM2
              << std::endl;
#endif
  }

  // Final temperature, at prediction time [C]
  const auto& finalTemp = rmTemperature.back();

#ifdef _DEBUG_OUTPUT
  debugFile << "# Final temperature = " << finalTemp.celsius() << std::endl;
  debugFile.close();
#endif

  // Final temperature sanity checks
  assert(!std::isnan(finalTemp.kelvin()));
  assert(finalTemp.kelvin() > 0);

  // Last timestep for real mass prediction is our target prediction date/ti  me
  return finalTemp;
}
}  // namespace d5::thermal
