#include "dirsig5/thermal/RitThermInvariants.h"

#include <cmath>
#include <ctime>

#include "dirsig5/base/Error.h"
#include "dirsig5/base/Logger.h"

namespace d5::thermal {

const std::vector<float>& ThermTimeSteps() {
  static const auto kTimeVector = []() {
    std::vector<float> timeVector(105);

    // start 24 hours back
    constexpr double dayInSec = 24 * 60 * 60;
    constexpr double fifteenMinInSec = 15 * 60;
    constexpr double fiveMinInSec = 5 * 60;
    auto stepTime = -dayInSec;

    // first 23 hours at 15 minute timesteps
    for (std::size_t i = 0; i < 92; i++) {
      timeVector[i] = stepTime;
      stepTime += fifteenMinInSec;
    }

    // last hour at 5 minute timesteps
    for (std::size_t i = 92; i < 105; i++) {
      timeVector[i] = stepTime;
      stepTime += fiveMinInSec;
    }

    return timeVector;
  }();

  return kTimeVector;
}

size_t HistoryLengthToTimeStepIndex(double historyLength) {
  const auto& timeSteps = ThermTimeSteps();
  size_t index = std::distance(
    timeSteps.begin(),
    std::lower_bound(timeSteps.begin(), timeSteps.end(), -historyLength));
  return index == 0 ? 0 : index - 1;
}

void RitThermInvariants::setData(std::vector<value_type> data) {
  data_ = std::move(data);
  if (data_.empty()) return;

  // Compute sunrise time and temperature (look at sun vectors)
  double sunriseTime = -1.0;
  Temperature minAirTemp = Temperature::fromKelvin(1000.0);
  for (std::size_t i = 1; i < data_.size(); i++) {
    // Handles to current and previous timesteps
    const auto& curStep = data_[i];
    const auto& preStep = data_[i - 1];

    // Track the minimum air temp as a backup to the sunrise temp
    if (minAirTemp > curStep.weather.airTemperature) {
      minAirTemp = curStep.weather.airTemperature;
    }

    // Check if we transitioned from below to above horizon
    if (preStep.sunVector(2) <= 0.0 && curStep.sunVector(2) > 0.0) {
      // linear interpolate for time/temp
      auto zenithPre = getZenith(preStep.sunVector);
      auto zenithCur = getZenith(curStep.sunVector);

      auto weightPre = (M_PI_2 - zenithCur) / (zenithPre - zenithCur);
      auto weightCur = (zenithPre - M_PI_2) / (zenithPre - zenithCur);

      sunriseTime = (weightPre * preStep.timeHr) + (weightCur * curStep.timeHr);

      sunriseAirTemp_ = Temperature::fromKelvin(
        (weightPre * preStep.weather.airTemperature.kelvin()) +
        (weightCur * curStep.weather.airTemperature.kelvin()));
      break;
    }
  }
  if (sunriseTime <= 0.0) {
    DIRSIG_WARN("Therm::setup:");
    DIRSIG_WARN("   Unable to determine sunrise air temperature!");
    DIRSIG_WARN(
      "   Using minimum air temperature of {} C instead", minAirTemp.celsius());
    sunriseAirTemp_ = minAirTemp;
  }

  // Determine the peak temperature and its time
  peakAirTemp_ = Temperature::fromKelvin(0);
  peakTempTime_ = 0;
  for (const auto& e : data_) {
    if (e.weather.airTemperature > peakAirTemp_) {
      peakAirTemp_ = e.weather.airTemperature;
      peakTempTime_ = e.timeHr;
    }
  }
}

// Compute effective sky emissivity.
void RitThermInvariants::ComputeEffectiveSkyEmissivity(value_type& stepData) {
  // vapor pressure of air
  double vpa = -666.0;

  const double ac = 17.269;
  const double bc = 35.86;

  auto tak = stepData.weather.airTemperature.kelvin();
  tak = (tak <= bc) ? bc + 1E-36 : tak;

  // attempt to compute vapor press from rel humidity
  if (stepData.weather.relativeHumidity >= 0.0) {
    vpa = stepData.weather.relativeHumidity * 6.108 *
          std::exp(ac * stepData.weather.airTemperature.celsius() / (tak - bc));

    // use rel. humidity to fill in dew point temperature
    stepData.weather.setDewPoint(
      Temperature::fromKelvin(5352.2 / (21.4 - log(vpa))));
  } else {
    // compute vapor press from dew point
    vpa = std::exp(21.4 - 5352.2 / stepData.weather.dewPoint().kelvin());

    // use dew point to fill in relative humidity info
    stepData.weather.relativeHumidity =
      vpa /
      (std::exp(ac * stepData.weather.airTemperature.celsius() / (tak - bc)) *
       6.108);
    if (stepData.weather.relativeHumidity > 1.0f) {
      stepData.weather.relativeHumidity = 1.0f;
    }
  }

  // Emissivity of air (Brunt equation):
  auto ea = 0.05 * sqrt(vpa) + .61;

#if 0
  // Any clouds or overcast skies?
  if ((stepData.cloudType != 0) && (stepData.skyExposure != 1.0f)) {
    // Effect thermal emiss of sky coeff: C(cloudtype)   (DCN pg. 17)
    const float clr[8] = {0.04, 0.08, 0.17, 0.20, 0.22, 0.24, 0.24, 0.25};

    // Sky emissivity adjustment (Pochop, Shanklin, Horner 1968):
    float cc2 = std::pow(1.0f - stepData.skyExposure, 2);
    ea *= clr[stepData.cloudType - 1] * cc2 + 1.0f;
  }
#endif

  // Cold sky coefficient:
  ea = std::pow(static_cast<double>(ea), 0.25);
  stepData.effSkyEmissivity = ea;
}

}  // namespace d5::thermal
