#include "dirsig5/thermal/Balfour.h"

#include <random>

#include <gtest/gtest.h>

using namespace d5::thermal;

TEST(BalfourTests, ClassicBalfourTest) {
  std::vector<float> params{10, 2, 3, 4, 5};

  d5::Weather wth;
  wth.airTemperature = d5::Temperature::fromKelvin(280);
  wth.windSpeedMS = 10;
  wth.directInsolationWM2 = 5;
  wth.diffuseInsolationWM2 = 2;

  EXPECT_NEAR(
      ClassicBalfour(d5::span<const float, 5>(params.data(), 5), wth, 0.5)
          .celsius(),
      15141, 1e-6);
}

TEST(CLBalfourTests, IterativeBalfourTest) {
  std::vector<float> params{10, 2, 3, 4, 5, 0};

  d5::Weather wth, wth2;
  wth.airTemperature = d5::Temperature::fromKelvin(280);
  wth.windSpeedMS = 10;
  wth.directInsolationWM2 = 5;
  wth.diffuseInsolationWM2 = 2;

  wth2.airTemperature = d5::Temperature::fromKelvin(300);
  wth2.windSpeedMS = 2;
  wth2.directInsolationWM2 = 3;
  wth2.diffuseInsolationWM2 = 4;

  auto getWeather = [&](float t) { return t >= -7200 ? wth : wth2; };

  auto getShadowHistory = [&](d5::span<const float> timeS) {
    std::vector<d5::thermal::ShadowHistorySample> h(timeS.size());
    for (std::size_t i = 0; i < h.size(); i++) {
      if (timeS[i] >= -7200) {
        h[i].shadowFraction = 0;
        h[i].solarVector = d5::Vectorf(sqrt(3) / 2, 0, 0.5);
      } else {
        h[i].shadowFraction = 1;
        h[i].solarVector = d5::Vectorf(0, 0, -1);
      }
    }
    return h;
  };

  EXPECT_NEAR(
      IterativeBalfour(
          d5::span<const float, 6>(params.data(), 6), 0, 16,
          d5::Vectorf(0, 0, 1), getShadowHistory, getWeather)
          .celsius(),
      15141, 1e-6);

  // With wth and sun, the answer is 15141 and with wth2 (no sun), the answer is
  // 4222. The integral of the exponential decay from 0-2 hours is a=0.75,
  // so the answer should be a * 15141 + (1 - a) * 4222 = 12411.25
  params[5] = 0.5;
  EXPECT_NEAR(
      IterativeBalfour(
          d5::span<const float, 6>(params.data(), 6), 0, 1000,
          d5::Vectorf(0, 0, 1), getShadowHistory, getWeather)
          .celsius(),
      12411.25, 1e-4);
}
