/** Interface file for a medium boundary plugin
 *
 *  \file MediumBoundaryInterface.h
 */

#pragma once

#include <optional>

#include "dirsig5/apis/Plugin.h"
#include "dirsig5/base/DateTime.h"
#include "dirsig5/base/Geometry.h"
#include "dirsig5/base/Pcg.h"
#include "dirsig5/base/SpectralVector.h"
#include "dirsig5/base/State.h"

namespace d5::apis::plugins {

/** \defgroup plugin_apis_medbound Medium Boundary Plugin
 *  \ingroup plugin_apis
 *  Holds the API definitions for a medium boundary plugin
 */

/// \addtogroup plugin_apis_medbound
/// \{

/** Callback for a material scattering associated with the top surface
 *  of the boundary interface in local coordinates. The vectors
 *  are assumed to be relative to an ideal normal vector (0,0,1).
 */
using BsdfCallback =
    std::function<void(const Vectorf & /*incident*/,
                       const Vectorf & /*exitant*/, SpectralVector &)>;

/** Callback for a material importance sampling (probability of sample
 *  follows the underlying probability density function). The exitant vector
 *  is assumed to be relative to an ideal normal vector (0,0,1). A sampler
 *  only covers a single hemisphere.
 */
using sampleCallback =
    std::function<Vectorf(RandEngine &randEngine, const Vectorf &)>;

/** Callback (spectral) for the total scattering into the given vector from
 *  a single side
 */
using spectralTotalCallback =
    std::function<void(const Vectorf &, SpectralVector &)>;

/** Callback (scalar) for the total scattering into the given vector from
 *  a single side
 */
using scalarTotalCallback = std::function<double(const Vectorf &)>;

/** Callback for the bounding box (if not provided, assumes it will not change)
 */
using BoundingBoxCallback = std::function<void(Pointf &bbMin, Pointf &bbMax)>;

/**
 * Struct for reflective components
 */
struct ScatteringCallbacks {
  ScatteringCallbacks(BsdfCallback bsdf, sampleCallback sample,
                      spectralTotalCallback spectral,
                      scalarTotalCallback scalar)
      : bsdf(bsdf), sample(sample), spectral(spectral), scalar(scalar) {}
  BsdfCallback bsdf;
  sampleCallback sample;
  spectralTotalCallback spectral;
  scalarTotalCallback scalar;
};

/** Callback for an optional solar transmission function meant to exploit
 *  lower level knowledge of the medium
 */
using SolarTransmissionCallback =
    std::function<void(const Pointf &, double time, SpectralVector &)>;

/// Definition of input/output variables for the Init interface
namespace MediumBoundaryInit {

struct In {
  // flag indicating if the default material was requested (i.e. plugin material
  // ignored)
  bool forceDefaultMaterial = false;
};

/** If the boundary is defined with a static facetized object,
 *  the surface can be returned here (for internally optimized
 *  intersections). If used this way, the vectors pointed to
 *  by the supplied output will be moved into the internal simulation
 *  and cannot be accessed further.
 *
 *  If intersections will be handled by the plugin, these
 *  output variables should remain as the defaults and
 *  an intersection callback should be supplied instead.
 *
 *  By default, the interface will be given Fresnel properties
 *  (with refractive index driven by the Medium IOP plugin).
 *  This can be overridden by supplying an index corresponding
 *  to one of a set of user materials provided in the medium
 *  definition. It is not necessary to supply these materials
 *  in order to provide custom intersections -- a returned
 *  material index of -1 will use the default properties
 *  (Fresnel or user-driven).
 *
 *  Note that the bounding box provided here can be overridden
 *  by the bounding box supplied to the medium definition.
 *  If smaller, that bounding box will shrink the box
 *  shown here. If larger, the medium will attempt to tile
 *  the provided interface. The exeption is if the geometry
 *  is provide here (returned geometry will not be tiled).
 *
 *  The boundary plugin can also provide a callback function to be
 *  used in place of the default (Fresnel) material's BRDF.
 */
struct Out {
  /// Minimum bounding corner point of the boundary object
  Pointf boxMin = {0, 0, 0};

  /// Maximum bounding point of the boundary object
  Pointf boxMax = {0, 0, 0};

  /** Spatial resolution of the volume. This should be the suggested length over
   * which properties should be sampled internally and usually corresponds to
   * the underlying representation of the volume (e.g. a voxel size). The value
   * is in meters.
   */
  float resolution = 1.0;

  /// Buffer of vertexes
  std::vector<float> *vertexes = nullptr;

  /// Buffer of vertex index triangles
  std::vector<std::uint32_t> *facets = nullptr;

  /// Buffer of vertex normals
  std::vector<float> *vnData = nullptr;

  /// Buffer of vertex normal indexes (same length as facets)
  std::vector<std::uint32_t> *vnIndexes = nullptr;

  /// Intersection callback.
  IntersectCallback intersectCallback;

  /** An optional transmission callback that will return the full
   *  transmission to a point from a particular direction through
   *  the medium (note that this is usually used with plugins that
   *  define bothe the boundary and the IOPs
   */
  SolarTransmissionCallback transmissionCallback;

  /** An optional bounding box callback for dynamic volumes. If set
   *  it will be used after each updateState call to check for
   *  new bounds. If not set, the bounding box will be assumed
   *  static after init.
   */
  BoundingBoxCallback boundingBoxCallback;

  /// Callbacks to be used in place of the default BRDF
  std::optional<ScatteringCallbacks> userBrdf;

  /// Callbacks to be used in place of the default BTDF
  std::optional<ScatteringCallbacks> userBtdf;

  /** Additional, Lambertian materials that can be assigned to a
   *  hit. The material is referenced by index (in the intersect
   *  interface). The additional materials are defined by a spectrally
   *  constant total reflectance, R, in the vector below.
   */
  std::vector<float> additionalMaterials;
};

} // namespace MediumBoundaryInit

/// Interface for boundary geometry between media.
class MediumBoundaryInterface : virtual public Plugin {
public:
  static void getPluginTypeName(std::vector<std::string> &typeNames) {
    typeNames.push_back("mediumboundary");
  }

  /** Initialization
   *
   *  Requests details about the boundary interface being modeled
   *  including whether intersections should be handled internally
   *  (as static geometry) or in the plugin. For the latter case
   *  and intersection routine can be provided as a callback function
   *
   *  Callbacks are provided as std::function objects
   *  based on the prototypes given for IntersectCallback
   */
  virtual Status init(const MediumBoundaryInit::In &in,
                      MediumBoundaryInit::Out &out) = 0;

  /** Query the number of vertices in the underlying boundary model.
   *  This function is not used if an OBJ is provided (in which case
   *  the virtual function does not need to be implemented).
   */
  [[nodiscard]] virtual size_t getVertexCount() const {
    throw std::runtime_error(
        "Using unimplemented medium boundary virtual function!\n");
  }

  /** Query the number of cells in the underlying boundary model. Cells
   *  are always triangles and have three corresponding vertexes.
   *  This function is not used if an OBJ is provided (in which case
   *  the virtual function does not need to be implemented).
   */
  [[nodiscard]] virtual size_t getCellCount() const {
    throw std::runtime_error(
        "Using unimplemented medium boundary virtual function!\n");
  }

  /** Query the center point corresponding to a cell index.
   *  This function is not used if an OBJ is provided (in which case
   *  the virtual function does not need to be implemented).
   */
  virtual void getCellCenter(size_t /*index*/, Pointf & /*center*/) const {
    throw std::runtime_error(
        "Using unimplemented medium boundary virtual function!\n");
  }

  /** Query for the vertex point corresponding to an index along
   *  with the corresponding normal vector at that point.
   *  This function is not used if an OBJ is provided (in which case
   *  the virtual function does not need to be implemented).
   */
  virtual void getVertexInfo(size_t /*vertexIndex*/, Pointf & /*vertex*/,
                             Vectorf & /*normal*/) const {
    throw std::runtime_error(
        "Using unimplemented medium boundary virtual function!\n");
  }

  /** Query for the vertex indices corresponding to a cell. Cells are
   *  always triangles and three vertex indices are returned.
   *  This function is not used if an OBJ is provided (in which case
   *  the virtual function does not need to be implemented).
   */
  virtual void getCellVertexIndexes(size_t /*facetIndex*/, size_t & /*v0i*/,
                                    size_t & /*v1i*/, size_t & /*v2i*/) const {
    throw std::runtime_error(
        "Using unimplemented medium boundary virtual function!\n");
  }
};

/// \}
} // namespace d5::apis::plugins
